# -*- perl -*-
# $Id: $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Utility for testing iSCSI functionality
## @author mikenel@netapp.com, dl-ssan-automation
## @status Shared
## @pod here

package San::Task::IscsiHelperCM;

# compiler directives
use strict;
use Params::Validate qw( :all );
use Data::Dumper;
use San::SanUtil;
use San::SanMgr;
use IOMan;
use EventMan;

# inherit base class
our @ISA = qw( SanUtil );

# constant variable
use constant NOERRORFOUND => 'NoErrorFound';

################################################################################
#                               Public Methods
################################################################################

################################################################################
# Method:    new
# Objective: Class Constructor.
# Details:   See POD documentation at the end of this file.
################################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {};

    my $debug_level = 1;

    # bless instance into class
    bless $self, $class;

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /debug_level/) {
            if ($value < 0) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $self;
            }
            $debug_level = $value;
        }
    }

    $self->sanlog(-msg => "IscsiHelperCM::new() - Create IscsiHelperCM Object");

    # Set debug level
    $self->set_debug_level($debug_level);

    # Initialize the vserver_info - array of hashes containing an array
    # of hash for vserver name, objects, client names and client objects
    $self->{vserver_info} = ();

    return $self;
} ## end sub new

################################################################################
# initialize_objects
# Objective: It will initialize other library objects which are useful for
#            starting other activities. Activities include running IO or inject
#            events.
# Details:
#     Inputs - none
#     Returns On Success, $SanUtil::TRUE  - all objects created successfully
#             On error, $SanUtil::FALSE - object failed and set_err
################################################################################
sub initialize_objects {

    my ($self) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    $self->_initialize_ioman();
    if ($self->error()) {
        $self->_set_err("Failed to initialize IOMan object", $self->errmsg());
        return $SanUtil::FALSE;
    }
    $self->_initialize_eventman();
    if ($self->error()) {
        $self->_set_err("Failed to initialize EventMan object",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub initialize_objects

################################################################################
# Method:    get_ioman_obj
# Objective: Get the IOMan object that was initialized by IscsiHelperCM object
# Details:
#     Inputs - none
#     Returns On Success, IOMan_Obj
#             On error, undef and set_err
################################################################################
sub get_ioman_obj {
    my ($self) = @_;

    $self->_clear_err();

    if (ref($self->{_ioman_obj}) =~ /IOMan/) {
        return $self->{_ioman_obj};
    } else {
        $self->_set_err("IOMan object not instantiated");
        return undef;
    }
} ## end sub get_ioman_obj

################################################################################
# Method:    get_eventman_obj
# Objective: Get the EventMan object that was initialized by IscsiHelperCM obj
# Details:
#     Inputs - none
#     Returns On Success, EventMan_Obj
#             On error, undef and set_err
################################################################################
sub get_eventman_obj {
    my ($self) = @_;

    $self->_clear_err();

    if (ref($self->{_eventman_obj}) =~ /EventMan/) {
        return $self->{_eventman_obj};
    } else {
        $self->_set_err("EventMan object not instantiated");
        return undef;
    }
} ## end sub get_eventman_obj

################################################################################
# Method:    verify_iscsi_clients
# Objective: Verifies the iSCSI client information agrees with Data ONTAP.
#            Checks both cluster and vserver level verification.
# Details:
#     Inputs - clients (optional) - Array of names of clients you would like to
#              verify. If not specified, all clients in the sanmgr object are
#              verified.
#            - sanmgr_obj (required) - sanmgr object
#     Returns $SanUtil::TRUE  - all clients verified
#             $SanUtil::FALSE - all clients NOT verified and set_err
################################################################################
sub verify_iscsi_clients {
    my ($self, %args) = @_;

    my @client_objs;     # clients objects to be verified
    my @client_names;    # Names of clients user wants to verify
    my $sanmgr_obj;
    my $client_names;    # String of client hostnames to verify

    $self->sanlog(-msg       => "START",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /clients/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            @client_names = @{$value};
        } elsif ($key =~ /sanmgr_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $sanmgr_obj = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(
                "-sanmgr_obj is a required parameter, must be a SanMgr object");
        return $SanUtil::FALSE;
    }

    # If the user passed in some client names to verify, then we need to grab
    # the client objects.
    foreach my $name (@client_names) {

        my @clients =
          $sanmgr_obj->get_objects(type  => "client",
                                   names => [$name]);
        if ($sanmgr_obj->error()) {
            $self->_set_err("get_objects(client,name=$name) failed",
                            $sanmgr_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Verify we only got back 1 object.
        if (scalar(@clients) != 1) {
            $self->_set_err(
                "Expected to find 1 match for get_objects(client,name=$name), but found "
                  . scalar(@clients));
            return $SanUtil::FALSE;
        }

        # Save the client object
        push(@client_objs, $clients[0]);

    } ## end foreach my $name (@client_names)

    # If the user did not specify any clients, all clients will be verified.
    if (scalar(@client_objs) == 0) {
        my @all_client_objs = $sanmgr_obj->get_objects(type => "client");
        if ($sanmgr_obj->error()) {
            $self->_set_err("get_objects(client) failed",
                            $sanmgr_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Scan each client.
        foreach my $client (@all_client_objs) {

            # If it's iscsi, save it.
            if ($client->get_protocol() =~ /^iscsi$/i) {
                push(@client_objs, $client);
            }
        }
    } ## end if (scalar(@client_objs...))

    # Build list of client names for debug msg
    foreach my $client (@client_objs) {
        if (!length($client_names)) {
            $client_names = $client->hostname();
        } else {

            # Else, add a comma and the next hostname.
            $client_names = "$client_names, " . $client->hostname();
        }
    }

    # debug info
    $self->sanlog(-msg       => "Clients to be verified: " . $client_names,
                  -msg_level => 1);

    # Get the cluster object
    my @cluster_objs = $sanmgr_obj->get_objects(type => "cluster");
    if ($sanmgr_obj->error()) {
        $self->_set_err("Failed to get cluster objects", $sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Verify we got back some cluster objects.
    if (!scalar(@cluster_objs)) {
        $self->_set_err("get_objects(cluster) returned 0 objects");
        return $SanUtil::FALSE;
    }

    # Get the cserver object
    my $cserver_obj = $cluster_objs[0]->get_cserver_obj();
    if ($cluster_objs[0]->error()) {
        $self->_set_err("Failed to get cserver object",
                        $cluster_objs[0]->errmsg());
        return $SanUtil::FALSE;
    }

    # Verify each client
    foreach my $client (@client_objs) {

        my $client_name = $client->hostname();
        $self->sanlog(-msg       => "Processing client: $client_name",
                      -msg_level => 1);

        # debug info.  Set Purity=1 for complete dumper output.
        $Data::Dumper::Purity = 1;
        $self->sanlog(-msg       => "Client Object:\n" . Dumper($client),
                      -msg_level => 5);
        $Data::Dumper::Purity = 0;

        # Verify the client obj has > 0 initiators and that the nodename
        # exists.  We need to grab the initiator's nodename.
        if (!( scalar(@{$client->{initiators}})
               && length($client->{initiators}[0]->{nodename}))
          ) {
            $self->_set_err(
                 "Unable to fetch initiator's nodename from the client object");
            return $SanUtil::FALSE;
        }

        # Grab the initiator's nodename from the client object
        my $initiator_iqn = $client->{initiators}[0]->{nodename};

        $self->sanlog(-msg => "$client_name iscsi nodename: $initiator_iqn",
                      -msg_level => 1);

        # Get the client's view of iscsi information
        my $iscsi_info = $client->get_iscsi_info();
        if ($client->error()) {
            $self->_set_err("$client_name get_iscsi_info failed",
                            $client->errmsg());
            return $SanUtil::FALSE;
        }

        # debug info.  Set Purity=1 for complete dumper output.
        $Data::Dumper::Purity = 1;
        $self->sanlog(-msg => "$client_name get_iscsi_info returns:\n"
                        . Dumper($iscsi_info),
                      -msg_level => 1);
        $Data::Dumper::Purity = 0;

        # Verify the cluster data
        $self->_verify_iscsi_client_across_cluster(
                                                -cserver_obj   => $cserver_obj,
                                                -client_hash   => $iscsi_info,
                                                -initiator_iqn => $initiator_iqn
        );
        if ($self->error()) {
            $self->_set_err(
                "_verify_iscsi_client_across_cluster failed for client $client_name",
                $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Now verify each vserver the client is logged in to
        foreach my $nodename (keys %{$iscsi_info}) {

            # Use the target nodename to find the vserver name
            my @iscsi_show =
              $cserver_obj->iscsi_list_info('target-name' => $nodename);
            if ($cserver_obj->error()) {
                $self->_set_err(
                             "iscsi_list_info failed for target-name $nodename",
                             $cserver_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # Verify we only got back 1 object.
            if (scalar(@iscsi_show) != 1) {
                $self->_set_err(
                           "Expected to find 1 match for iscsi show, but found "
                             . scalar(@iscsi_show)
                             . " matches.\n"
                             . "Filter used: target-name => $nodename\n");
                return $SanUtil::FALSE;
            }

            # Now use the vserver name to get the vserver obj
            my @vserver_obj =
              $sanmgr_obj->get_objects(-type  => "vserver",
                                       -names => [$iscsi_show[0]->{vserver}]);
            if ($sanmgr_obj->error()) {
                $self->_set_err("get_objects failed for vserver "
                                  . $iscsi_show[0]->{vserver},
                                $sanmgr_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # Verify we got a single vserver
            if (scalar(@vserver_obj) != 1) {
                $self->_set_err(  "get_objects returned "
                                . scalar(@vserver_obj)
                                . " vservers with name "
                                . $iscsi_show[0]->{vserver});
                return $SanUtil::FALSE;
            }

            # Now use the vserver obj to get the iscsi object
            my $iscsi_obj = $vserver_obj[0]->get_iscsi_obj();
            if ($vserver_obj[0]->error()) {
                $self->_set_err("get_iscsi_obj failed for vserver "
                                  . $iscsi_show[0]->{vserver},
                                $vserver_obj[0]->errmsg());
                return $SanUtil::FALSE;
            }

            # Now call verify iscsi for the client data
            $self->_verify_iscsi_client(
                                       -iscsi_obj   => $iscsi_obj,
                                       -client_hash => $iscsi_info->{$nodename},
                                       -hosttype    => $client->hosttype());
            if ($self->error()) {
                $self->_set_err(
                    "_verify_iscsi_client failed for client $client_name and vserver "
                      . $iscsi_show[0]->{vserver},
                    $self->errmsg());
                return $SanUtil::FALSE;
            }
        } ## end foreach my $nodename (keys ...)

    } ## end foreach my $client (@client_objs)

    # If code reaches here, everything passed.
    $self->sanlog(-msg       => "SUCCESS!",
                  -msg_level => 1);
    return $SanUtil::TRUE;

} ## end sub verify_iscsi_clients

################################################################################
# Method:    create_vserver_info
# Objective: Creates a hash that contains data specific to vservers
#            It creates an array of hashes. Each hash contains the following:
#            name        - contains scalar value, the name of the vserver
#            object      - contains vserver_obj
#            clients     - contains an array of scalars, the names of clients
#                          associated with the vserver
#            client_objs - contains an array of client_objs
# Details:
#     Inputs - sanmgr_obj (required) - SanMgr_Obj that creates the SAN resources
#     Returns On Success, an array (or reference to array) of vserver_info hash
#             On error, an empty array (or reference to array) and set_err
################################################################################
sub create_vserver_info {
    my ($self, %args) = @_;
    my ($sanmgr_obj);

    $self->_clear_err();

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    # Parse arguments
    foreach (keys %args) {
        if (/-?sanmgr_obj/i) {
            $sanmgr_obj = $args{$_};
            if (not length($sanmgr_obj)) {
                $self->_set_err("Invalid value for argument '-sanmgr_obj'");
                return (wantarray()) ? () : [];
            }
        }
    }    # end foreach arg

    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(  "Missing required argument '-sanmgr_obj'."
                        . "Must be a SanMgr object. Object: "
                        . ref($sanmgr_obj));
        return (wantarray()) ? () : [];
    }

    # Initialize vserver_info
    $self->{vserver_info} = ();

    # Get all of the vserver objects.
    my @vserver_objs = $sanmgr_obj->get_objects(type => "vserver");
    if ($sanmgr_obj->error()) {
        $self->_set_err("get_objects(vservers) failed", $sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # If no vservers were returned, no more work to do.
    if (!scalar(@vserver_objs)) {
        return (wantarray()) ? () : [];
    }

    # Get the cluster object so we can get the name below
    my @cluster_obj = $sanmgr_obj->get_nacl_objects(type => "cluster");
    if ($sanmgr_obj->error()) {
        $self->_set_err("Failed to get cluster object", $sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # verify we got an obj back
    if (!scalar (@cluster_obj)) {
        $self->_set_err("get_nacl_objects did not return cluster obj",
                        $sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # Vserver loop
    foreach my $vserver_obj (@vserver_objs) {

        # Get vserver name
        my $vserver_name = $vserver_obj->get_name();
        my $log_label    = "Vserver: $vserver_name";

        # Get all clients associated with the vserver
        my @clients =
          $sanmgr_obj->get_clients_associated_to_vserver(
                                      -cluster_name => $cluster_obj[0]->name(),
                                      -vserver_names => [$vserver_name],
                                      -protocols     => ["iscsi"]);
        if ($sanmgr_obj->error()) {
            $self->_set_err("Cannot get clients associated to vserver",
                            $log_label, $sanmgr_obj->errmsg());
            return (wantarray()) ? () : [];
        }
        $self->sanlog(-msg => "Clients associated to vserver "
                        . $vserver_name
                        . " are: @clients",
                      -msg_level => 1);

        # Get all client objects associated with the sanmgr object
        my @client_objs = ();

        if (scalar(@clients)) {
            my @all_client_objs = $sanmgr_obj->get_objects(type => "client");
            if ($sanmgr_obj->error()) {
                $self->_set_err("get_objects(client) failed",
                                $log_label, $sanmgr_obj->errmsg());
                return (wantarray()) ? () : [];
            }

            # Store the client objects specific to the vserver
            foreach my $client (@clients) {
                foreach my $client_obj (@all_client_objs) {
                    if ($client eq $client_obj->hostname()) {
                        push(@client_objs, $client_obj);
                        last;
                    }
                }
            }
        } ## end if (scalar(@clients))

        # Create a temp hash to store the vserver info
        my %vserver_info = {};
        $vserver_info{name}        = $vserver_name;
        $vserver_info{object}      = $vserver_obj;
        $vserver_info{clients}     = \@clients;
        $vserver_info{client_objs} = \@client_objs;

        $self->sanlog(-msg => "Created vserver_info for vserver $vserver_name",
                      -msg_level => 1);
        $self->sanlog(-msg => "vserver_info dump\n" . Dumper(\%vserver_info),
                      -msg_level => 3);

        # Add it to the vserver_info. It will be saved in the object
        push(@{$self->{vserver_info}}, \%vserver_info);
    } ## end foreach my $vserver_obj (@vserver_objs)

    $self->sanlog(-msg       => "Success: vserver_info created",
                  -msg_level => 1);

    return (wantarray()) ? @{$self->{vserver_info}} : \@{$self->{vserver_info}};
} ## end sub create_vserver_info

################################################################################
# Method:    get_vserver_info
# Objective: Get the hash that contains data specific to vservers
#            It is an array of hashes. Each hash contains the following:
#            name        - contains scalar value, the name of the vserver
#            object      - contains vserver_obj
#            clients     - contains an array of scalars, the names of clients
#                          associated with the vserver
#            client_objs - contains an array of client_objs
# Details:
#     Inputs - none
#     Returns On Success, an array (or reference to array) of vserver_info hash
#             On error, an empty array (or reference to array)
################################################################################
sub get_vserver_info {
    my ($self) = @_;

    $self->_clear_err();

    return (wantarray()) ? @{$self->{vserver_info}} : $self->{vserver_info};
}

################################################################################
# login_all_vserver_info_clients
# Objective: vserver_info contains info of all clients associated to each
#            vserver. Login all clients associated to each vserver from the
#            {vserver_info} data structure. Verify the login succeeded.
# Details:
#     Inputs - -sanmgr_obj (required) - SanMgr_Obj to verify logins for
#
#              -sendtargets (optional) - Do a refresh of sendtargets or not
#              By default, sendtargets info will be refreshed.
#
#              -fail_on_login (optional) - When set, any attempt by the clients
#              to login to target should fail - fail_on_login=>1
#              By default, login clients should not fail - fail_on_login=>0
#              Note: A value of 1 means you expect ALL logins from ALL clients
#              to ALL vservers to FAIL.
#
#              -fail_on_login_vserver_hash (optional) - A hash of vserver names
#              which are expected to fail on login.  All other vservers should
#              succeed on login.  If fail_on_login = 1, then there is no need
#              to specify this hash.
#
#              -isns_address (optional) - Use isns discovery to this isns
#              server address instead of sendtargets.
#
#              -num_expected_luns (optional) - An integer of number of
#              luns that each client is expected to find.  If you provide this,
#              login speed is increased.
#
#     Returns On Success, $SanUtil::TRUE  - all clients login successfully
#             On error, $SanUtil::FALSE - login failed and set_err
################################################################################
sub login_all_vserver_info_clients {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    ###########################################################################
    # All args are passed down to a lower level function call for validation.
    # -sanmgr_obj will be validated in _login_logout_verify_all_vserver_info()
    # -sendtargets is optional param for _client_config()
    # -fail_on_login is optional param for _client_setup() and _client_config()
    # -fail_on_login_vserver_hash is optional param for
    #  _login_logut_verify_all_vserver_info_clients()
    # -isns_address is optional param for _client_config
    # -num_expected_luns is an optional param for _client_discovery()
    ###########################################################################

    $args{'-action'} = "login";

    $self->_login_logout_verify_all_vserver_info_clients(%args);
    if ($self->error()) {
        $self->_set_err("Failed to login and verify all vserver/clients info",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub login_all_vserver_info_clients

################################################################################
# logout_all_vserver_info_clients
# Objective: vserver_info contains info of all clients associated to each
#            vserver. Logout all clients associated to each vserver from the
#            {vserver_info} data structure. Verify the logout succeeded.
# Details:
#     Inputs - -sanmgr_obj (required) - SanMgr_Obj to verify logouts for.
#
#              -keep_portals (optional) - Leave the target portals on the client
#              By default, Portals will be cleaned up
#
#     Returns On Success, $SanUtil::TRUE  - all clients logout successfully
#             On error, $SanUtil::FALSE - logout failed and set_err
################################################################################
sub logout_all_vserver_info_clients {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    ###########################################################################
    # All args are passed down to a lower level function call for validation.
    # -sanmgr_obj will be validated in _login_logout_verify_all_vserver_info()
    # -keep_portals is optional param for _client_cleanup()
    ###########################################################################

    $args{'-action'} = "logout";

    $self->_login_logout_verify_all_vserver_info_clients(%args);
    if ($self->error()) {
        $self->_set_err("Failed to logout and verify all vserver/clients info",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg       => "Success",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub logout_all_vserver_info_clients

################################################################################
# do_iscsi_setup
# Objective: This is a wrapper function to perform an iscsi service action on
#            all vservers found in {vserver_info}. iscsi can be created,
#            deleted, started or stopped using this method.
# Details:
#     Inputs - action (required) - Provide the iscsi action using the iscsi_obj
#              for all vservers in the {vserver_info} function
#            - skip_license(optional) - Skip adding license while creating isci_obj
#     Returns On Success, $SanUtil::TRUE  - iscsi setup successful
#             On error, $SanUtil::FALSE - iscsi setup failed and set_err
################################################################################
sub do_iscsi_setup {
    my ($self, %args) = @_;

    my $action;
    my $skip_license;    # Variable to skip license add during _iscsi_create()

    $self->_clear_err();

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    # Parse arguments
    foreach (keys %args) {
        if (/-?action/i) {
            $action = lc($args{$_});
            if (not length($action)
                or $action !~ /^(create|start|delete|stop)$/i) {
                $self->_set_err("Invalid value for argument '-action'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?skip_license/i) {
            $skip_license = $args{$_};
            if (not length($skip_license)
                or ($skip_license !~ /^(1|0)$/)) {
                $self->_set_err("Invalid value for argument '-skip_license'");
                return $SanUtil::FALSE;
            }
        }
    }    # end foreach arg

    unless ($action =~ /^(create|delete|start|stop)$/i) {
        $self->_set_err(  "Missing required argument '-action'."
                        . "Must be create, delete, start or stop: $action");
        return $SanUtil::FALSE;
    }

    unless (ref($self->{vserver_info}) =~ /ARRAY/) {
        $self->_set_err( "Vserver_Info must be created before using it."
                       . "Call create_vserver_info() to create the vserver_info"
                       . ref($self->{vserver_info}));
        return $SanUtil::FALSE;
    }

    # Vserver loop
    foreach my $vserver_info (@{$self->{vserver_info}}) {

        # Get the vserver object and name
        my $vserver_obj  = $vserver_info->{object};
        my $vserver_name = $vserver_info->{name};

        if ($action eq "create") {

            $self->sanlog(
                    -msg => "Create iscsi service for vserver " . $vserver_name,
                    -msg_level => 1);

            # Create iscsi object
            if ($skip_license) {
                $self->sanlog(
                            -msg => "Skip license add while creating iscsi obj",
                            -msg_level => 1);
                my $iscsi_obj =
                  $vserver_obj->create_iscsi_obj('-no-license' => 'true');

            } else {
                $self->sanlog(-msg => "License is added during iscsi create",
                              -msg_level => 1);
                my $iscsi_obj = $vserver_obj->create_iscsi_obj();
            }

            if ($vserver_obj->error()) {

                # If iscsi object already existed, it is not valid and
                # should be handled by this error call
                $self->_set_err("Cannot create iscsi service for vserver "
                                  . $vserver_name,
                                $vserver_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {    # For all other actions

            my $iscsi_obj;

            # Check if iscsi service was already created
            if ($vserver_obj->iscsi_exist()) {

                # Get iscsi object.
                $iscsi_obj = $vserver_obj->get_iscsi_obj();
                if ($vserver_obj->error()) {
                    $self->_set_err("Cannot get iscsi object for vserver "
                                      . $vserver_name,
                                    $vserver_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            } else {    # fail if iscsi service was not created
                $self->_set_err(
                            "iscsi object not found for vserver $vserver_name");
                return $SanUtil::FALSE;
            }

            if ($action eq "delete") {

                $self->sanlog(
                      -msg => "Stopping iscsi service on vserver $vserver_name",
                      -msg_level => 1);

                # Stop iscsi
                $iscsi_obj->stop();
                if ($iscsi_obj->error()) {
                    $self->_set_err(
                                  "iSCSI Stop failed for vserver $vserver_name",
                                  $iscsi_obj->errmsg());
                    return $SanUtil::FALSE;
                }

                $self->sanlog(
                      -msg => "Deleting iscsi service on vserver $vserver_name",
                      -msg_level => 1);

                # Delete iscsi
                $iscsi_obj->delete();
                if ($iscsi_obj->error()) {
                    $self->_set_err(
                                "iSCSI delete failed for vserver $vserver_name",
                                $iscsi_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            } elsif ($action eq "start") {

                $self->sanlog(
                      -msg => "Starting iscsi service on vserver $vserver_name",
                      -msg_level => 1);

                # Start iscsi
                $iscsi_obj->start();
                if ($iscsi_obj->error()) {
                    $self->_set_err(
                                 "iSCSI Start failed for vserver $vserver_name",
                                 $iscsi_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            } elsif ($action eq "stop") {

                $self->sanlog(
                      -msg => "Stopping iscsi service on vserver $vserver_name",
                      -msg_level => 1);

                # Stop iscsi
                $iscsi_obj->stop();
                if ($iscsi_obj->error()) {
                    $self->_set_err(
                                  "iSCSI Stop failed for vserver $vserver_name",
                                  $iscsi_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            } ## end elsif ($action eq "stop")
        } ## end else [ if ($action eq "create")]
    } ## end foreach my $vserver_info (@...)

    $self->sanlog(-msg       => "$action Successful",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub do_iscsi_setup

################################################################################
# vserver_info_lifs_modify
# Objective: This is a wrapper function to modify lifs on all vservers found in
#            {vserver_info}.
# Details:
#     Inputs - %lif_modify_params (required) - At least one modify action.
#               It supports the exact same params that are used by
#                      ControllerCluster::Filer10M::Vserver::Lif::modify()
#              -vservers (optional) - Reference to an array of vserver names to
#               apply as a filter to the {vserver_info} data structure
#              -ips (optional) - Since each Lif requires a unique IP address,
#               instead of using -address, provide a reference to an array of
#               new IP addresses to modify existing lif IP addresses.
#              -expected_hash(optional) If passed in, method will attempt to
#               update the expected hash.  Only applies to address, home-node,
#               and home port because these are the only items the lif thpl
#               will attempt to modify with this method.
#    Returns On Success, $SanUtil::TRUE  - {vserver_info} lifs modify successful
#             On error, $SanUtil::FALSE - lifs modify failed and set_err
################################################################################
sub vserver_info_lifs_modify {

    my ($self, %args) = @_;

    my ($vservers_ref);
    my (@vservers);
    my $expected_hash;

    $self->_clear_err();

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    # Parse arguments
    foreach (keys %args) {
        if (/^-?vservers$/i) {
            $vservers_ref = $args{$_};
            if (ref($vservers_ref) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-vservers', value must be a reference to an 'ARRAY'"
                );
                return $SanUtil::FALSE;
            }
            @vservers = @{$vservers_ref};
            unless (scalar(@vservers)) {
                $self->_set_err("No entries found in '-vservers' array");
                return $SanUtil::FALSE;
            }
            delete $args{$_};
        } ## end if (/^-?vservers$/i)
    }    # end foreach arg

    unless (ref($self->{vserver_info}) =~ /ARRAY/) {
        $self->_set_err( "Vserver_Info must be created before using it."
                       . "Call create_vserver_info() to create the vserver_info"
                       . ref($self->{vserver_info}));
        return $SanUtil::FALSE;
    }

    ###########################################################################
    # All lif modify args are passed down to a lower level function call for
    # validation.
    # -ips will be validated in $self->lifs_modify()
    #-expected_hash will be validated in $self->lifs_modify()
    ###########################################################################

    my @lif_objs;

    # Vserver loop
    foreach my $vserver_info (@{$self->{vserver_info}}) {

        # Get the vserver object and name
        my $vserver_obj  = $vserver_info->{object};
        my $vserver_name = $vserver_info->{name};

        # If the -vserver filter is specified, only accept those vservers
        if (scalar @vservers and not grep(/^$vserver_name$/, @vservers)) {
            $self->sanlog(
                 -msg =>
                   "Skipping vserver '$vserver_name' as it was not part of the "
                   . "'-vservers' requested for: "
                   . join(', ', @vservers),
                 -msg_level => 1);
            next;
        }

        # Get all lifs objects for the vserver
        $self->sanlog(-msg       => "Get all LIF objects for '$vserver_name'",
                      -msg_level => 1);
        my @vserver_lif_objs = $vserver_obj->get_lif_objs();
        if ($vserver_obj->error()) {
            $self->_set_err(
                          "Cannot get lif objects for vserver " . $vserver_name,
                          $vserver_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Throw an error if no lif objects were found for the vserver
        if (not scalar(@vserver_lif_objs)) {
            $self->_set_err("No lif objects found for vserver $vserver_name");
            return $SanUtil::FALSE;
        }

        # Save them in a global array of lif_objs
        push(@lif_objs, @vserver_lif_objs);
    } ## end foreach my $vserver_info (@...)

    $self->sanlog(-msg       => "Modify LIFs for all vservers",
                  -msg_level => 1);
    $args{'-lif_objs'} = \@lif_objs;

    # Modify lifs
    $self->lifs_modify(%args);
    if ($self->error()) {
        $self->_set_err("Cannot modify lifs for vserver_info vservers",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg       => "Successful",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub vserver_info_lifs_modify

################################################################################
# lifs_modify
# Objective: This is a wrapper function to modify lifs.
# Details:
#     Inputs - -lif_objs (required) - Provide a reference to an array of
#               lif objects to do a modify action
#              %lif_modify_params (required) - At least one modify action.
#               It supports the exact same params that are used by
#                      ControllerCluster::Filer10M::Vserver::Lif::modify()
#              -ips (optional) - Since each Lif requires a unique IP address,
#               instead of using -address, provide a reference to an array of
#               new IP addresses to modify existing lif IP addresses.
#              -expected_hash(optional) If passed in, method will attempt to
#               update the expected hash.  Only applies to address, home-node,
#               and home port because these are the only items the lif thpl
#               will attempt to modify with this method.
#     Returns On Success, $SanUtil::TRUE  - lifs modify successful
#             On error, $SanUtil::FALSE - lifs modify failed and set_err
################################################################################
sub lifs_modify {

    my ($self, %args) = @_;

    my ($value);
    my (@lif_objs, @ips, $expected_hash);
    my $modify_ips = 0;

    $self->_clear_err();

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    # Parse arguments
    foreach (keys %args) {
        if (/^-?lif_objs$/i) {

            # Remove the lif_objs so %args only has lif-modify keys
            $value = delete $args{$_};
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-lif_objs', value must be a reference to an 'ARRAY'"
                );
                return $SanUtil::FALSE;
            }
            @lif_objs = @{$value};
        } elsif (/^-?ips$/i) {

            # Remove the ips so %args only has lif-modify keys
            $value = delete $args{$_};
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-ips', value must be a reference to an 'ARRAY'"
                );
                return $SanUtil::FALSE;
            }
            @ips = @{$value};
        } elsif (/^-?expected_hash$/) {
            $value = delete $args{$_};
            unless ($value) {
                $self->_set_err(
                             "Invalid value for \"-expected_hash\": " . $value);
                return $SanUtil::FALSE;
            }
            $expected_hash = $value;
        }
    }    # end foreach arg

    # If -ips is missing or other modify() args are missing, return error
    unless (scalar keys %args or scalar(@ips)) {
        $self->_set_err("No param passed for lif modify");
        return $SanUtil::FALSE;
    }

    unless (scalar(@lif_objs)) {
        $self->_set_err("No lifs to modify");
        return $SanUtil::FALSE;
    }

    # If -ips were passed, make sure there are enough IPs for the Lifs.
    # Also turn on modify_ips flag which is used below.
    if (scalar(@ips)) {

        $modify_ips = 1;
        if (scalar(@lif_objs) > scalar(@ips)) {
            $self->_set_err("Not enough IP addresses were passed to assign to "
                            . "LIFs: IPs '"
                            . scalar(@ips)
                            . "' < LIFs '"
                            . scalar(@lif_objs)
                            . "'");
            return $SanUtil::FALSE;
        }
    } ## end if (scalar(@ips))

    # lif_obj loop
    foreach my $lif_obj (@lif_objs) {
        my $ip;    # Store the IP from the -ips array

        # if modify param is home-node or home-port, bring the lif down
        if (scalar grep { $_ =~ /(home-node|home-port)/ } keys %args) {

            $self->sanlog(
                   -msg => "Modify LIF "
                     . $lif_obj->get_name()
                     . " status-admin to 'down' to change LIF's attribute that "
                     . "would not be possible with LIF state enabled",
                   -msg_level => 1);
            $lif_obj->modify('-status-admin' => "down");
            if ($lif_obj->error()) {
                $self->_set_err("Cannot modify lif status-admin to down",
                                $lif_obj->errmsg());
                return $SanUtil::FALSE;
            }

        } ## end if (scalar grep { $_ =~...})

        # If IP address is to be modified, get an IP from the array and store it
        # in args. A unique IP is used for every lif_obj.
        if (scalar(@ips)) {
            $ip = shift(@ips);
            $self->sanlog(-msg => "Modify LIF "
                            . $lif_obj->get_name()
                            . " IP address from "
                            . $lif_obj->get_address() . " to "
                            . $ip,
                          -msg_level => 1);
            $args{'-address'} = $ip;
        } ## end if (scalar(@ips))

        # Print all the modify params for the lif
        my $msg = "Modifying LIF " . $lif_obj->get_name() . " with: ";
        foreach (keys %args) {
            $msg .= "$_=>$args{$_} ";
        }
        $self->sanlog(-msg => $msg, -msg_level => 1);

        # Modify lifs
        $lif_obj->modify(%args);
        if ($lif_obj->error()) {
            $self->_set_err("Cannot modify lif status", $lif_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # if modify param is home-node or home-port bring back the lif up
        if (scalar grep { $_ =~ /(home-node|home-port)/ } keys %args) {
            $self->sanlog(-msg => "Reset back LIF "
                            . $lif_obj->get_name()
                            . " status-admin to 'up'",
                          -msg_level => 1);
            $lif_obj->modify('-status-admin' => "up");
            if ($lif_obj->error()) {
                $self->_set_err("Cannot modify lif status-admin to up",
                                $lif_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } ## end if (scalar grep { $_ =~...})

        # If expected hash was passed in, let's attempt to update it.
        if ($expected_hash) {

            my $found = 0;    # flag

            my $vserver = $lif_obj->get_vserver();
            my $name    = $lif_obj->get_name();

            # Need to scan all lifs in the vserver to find the name we are
            # modifying.
            foreach my $lif (@{$expected_hash->{vservers}->{$vserver}->{lif}}) {

                # If the name matches what we are looking for, do some work
                if ($lif->{'lif'} eq $name) {

                    $found = 1;

                    # If scalar ips were passed in, then we updated the ip.
                    if ($modify_ips) {
                        $lif->{'address'} = $ip;
                    }

                    # If home-node was updated, update exp hash.
                    if (scalar grep { $_ =~ /home-node/ } keys %args) {
                        $lif->{'home-node'} = $args{'-home-node'};
                        $lif->{'curr-node'} = $args{'-home-node'};
                    }

                    # If home-port was updated, update exp hash.
                    if (scalar grep { $_ =~ /home-port/ } keys %args) {
                        $lif->{'home-port'} = $args{'-home-port'};
                        $lif->{'curr-port'} = $args{'-home-port'};
                    }
                } ## end if ($lif->{'lif'} eq $name)
            } ## end foreach my $lif (@{$expected_hash...

            # If we didn't find the lif in expected hash, let's return a fail.
            if (!$found) {
                $self->_set_err("Could not find lif in expected hash");
                return $SanUtil::FALSE;
            }
        }
    } ## end foreach my $lif_obj (@lif_objs)

    $self->sanlog(-msg       => "Successful",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub lifs_modify

################################################################################
# cluster_ports_modify
# Objective: This is a wrapper function to modify port up-admin state.
#            Port up-admin can be modified to 'true' or 'false'.
# Details:
#     Inputs - -sanmgr_obj (required) - SanMgr object
#              -ports (required) - Provide the array of port names to be modified
#              -nodes (optional) - Provide the array of node names to be modified
#              <network_adapter->modify() args> - Provide at least 1 arg that
#               is used by 'network interface modify' cmd
#              Examples:
#              '-up-admin' - Provide the correct up-admin value
#               for the data port. Valid values are true or false
#              -mtu (optional) - Provide the correct mtu value
#     Returns On Success, $SanUtil::TRUE  - port modify successful
#             On error, $SanUtil::FALSE - port modify failed and set_err
################################################################################
sub cluster_ports_modify {

    my ($self, %args) = @_;

    my ($sanmgr_obj, $value);
    my (@ports,      @nodes);
    my $port_down_time;
    my $port_up_time;
    my %func_args;    # Store args for _get_port_objs() function

    $self->_clear_err();

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    # Parse arguments
    foreach (keys %args) {
        if (/-?sanmgr_obj/i) {
            $sanmgr_obj = $args{$_};
            if (not length($sanmgr_obj)) {
                $self->_set_err("Invalid value for argument '-sanmgr_obj'");
                return (wantarray()) ? () : [];
            }
            $func_args{$_} = delete $args{$_};
        } elsif (/-?ports/i) {
            $value = $args{$_};
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-ports', value must be a reference to an 'ARRAY'"
                );
                return $SanUtil::FALSE;
            }
            @ports = @{$value};
            $func_args{$_} = delete $args{$_};
        } elsif (/^-?nodes$/i) {
            $value = $args{$_};
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-nodes', value must be a reference to an 'ARRAY'"
                );
                return $SanUtil::FALSE;
            }
            @nodes = @{$value};
            unless (scalar(@nodes)) {
                $self->_set_err("No entries found in '-nodes' array");
                return $SanUtil::FALSE;
            }
            $func_args{$_} = delete $args{$_};
        } ## end elsif (/^-?nodes$/i)
    }    # end foreach arg

    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(  "Missing required argument '-sanmgr_obj'."
                        . "Must be sanmgr_obj: ref($sanmgr_obj)");
        return $SanUtil::FALSE;
    }

    unless (scalar keys %args) {
        $self->_set_err("No param passed for port modify");
        return $SanUtil::FALSE;
    }

    unless (scalar(@ports)) {
        $self->_set_err("Missing required argument '-ports'.");
        return $SanUtil::FALSE;
    }

    # Get Port Objects to modify
    my @netadap_objs = $self->_get_port_objs(%func_args);
    if ($self->error()) {
        $self->_set_err("Failed to get Network Adapter objects for ports "
                          . join(", ", @ports),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # network adapter loop
    foreach my $netadap_obj (@netadap_objs) {

        # Get port name
        my $portname       = $netadap_obj->get_port();
        my $ports_nodename = $netadap_obj->get_node();

        $self->sanlog(
            -msg => "Modify Port '$portname' for Node '$ports_nodename' with:\n"
              . Dumper(%args),
            -msg_level => 1);

        # Modify port status
        $netadap_obj->modify(%args);
        if ($netadap_obj->error()) {
            $self->_set_err(
                "Failed to modify network adapter '$portname' on node '$ports_nodename'",
                $netadap_obj->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end foreach my $netadap_obj (@netadap_objs)

    # Have to sleep here to make sure all lifs are up/down.  There is a delay
    # from when you modify the port and when the lifs reflect the new status.
    # see burt 659255 for details.
    &SanUtil::nate_sleep(30);

    $self->sanlog(-msg => "Success!", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub cluster_ports_modify

################################################################################
# clients_cleanup
# Objective: Cleanup all clients associated to sanmgr_obj
# Details:
#     Inputs - -sanmgr_obj (required) - SanMgr_Obj clients to cleanup
#
#     Returns On Success, ClientUtil::TRUE
#             On error, ClientUtil::FALSE and set_err
################################################################################
sub clients_cleanup {
    my ($self, %args) = @_;

    my ($sanmgr_obj);

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # Parse arguments
    foreach (keys %args) {
        if (/-?sanmgr_obj/i) {
            $sanmgr_obj = $args{$_};
            if (not length($sanmgr_obj)) {
                $self->_set_err("Invalid value for argument '-sanmgr_obj'");
                return $SanUtil::FALSE;
            }
        }
    }    # end foreach arg

    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(  "Missing required argument '-sanmgr_obj'."
                        . "Must be a SanMgr object. Object: "
                        . ref($sanmgr_obj));
        return $SanUtil::FALSE;
    }

    my @client_objs = $sanmgr_obj->get_objects(type => "client");
    if ($sanmgr_obj->error()) {
        $self->_set_err("get_objects(client) failed", $sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    foreach my $client_obj (@client_objs) {
        $self->_client_cleanup(-client_obj  => $client_obj,
                               -clean_level => 2);
        if ($self->error()) {
            $self->_set_err(
                          "Failed to cleanup client " . $client_obj->hostname(),
                          $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end foreach my $client_obj (@client_objs)

    # Insert slight delay to allow time for sessions to be cleaned up.
    &SanUtil::nate_sleep(3);

    # Do a verify for all clients
    $self->verify_iscsi_clients(-sanmgr_obj => $sanmgr_obj);
    if ($self->error()) {
        $self->_set_err("Verify for all clients failed", $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Success!", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub clients_cleanup

################################################################################
# Method:    create_client_info
# Objective: Builds a data structure containing information on each client
#            defined by SanMgr.
#            It creates an array of hashes. Each hash contains the following:
#            name         - contains scalar value, the name of the client
#            object       - contains client_obj
#            vservers     - contains an array of scalars, the names of vservers
#                           associated with the client
#            vserver_objs - contains an array of vserver_objs
# Details:
#     Inputs - sanmgr_obj (required) - SanMgr_Obj that creates the SAN resources
#     Returns On Success, an array (or reference to array) of client_info hash
#             On error, an empty array (or reference to array) and set_err
################################################################################
sub create_client_info {
    my ($self, %args) = @_;
    my ($sanmgr_obj);

    $self->_clear_err();

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    # Parse arguments
    foreach (keys %args) {
        if (/-?sanmgr_obj/i) {
            $sanmgr_obj = $args{$_};
            if (not length($sanmgr_obj)) {
                $self->_set_err("Invalid value for argument '-sanmgr_obj'");
                return (wantarray()) ? () : [];
            }
        }
    }    # end foreach arg

    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(  "Missing required argument '-sanmgr_obj'."
                        . "Must be a SanMgr object. Object: "
                        . ref($sanmgr_obj));
        return (wantarray()) ? () : [];
    }

    # Initialize client_info
    $self->{client_info} = ();

    # Get all of the client objects.
    my @client_objs = $sanmgr_obj->get_objects(type => "client");
    if ($sanmgr_obj->error()) {
        $self->_set_err("get_objects(client) failed", $sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # If no clients were returned, no more work to do.
    if (!scalar(@client_objs)) {
        return (wantarray()) ? () : [];
    }

    # Get the cluster object
    my @cluster_objs = $sanmgr_obj->get_objects(type => "cluster");
    if ($sanmgr_obj->error()) {
        $self->_set_err("Failed to get cluster objects", $sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get the first element in the array.
    my $cluster_obj = $cluster_objs[0];

    # Client loop
    foreach my $client_obj (@client_objs) {

        # Get client name
        my $client_name = $client_obj->hostname();
        my $log_label   = "Client: $client_name";

        # Get all vservers associated with the client
        my @vservers = $sanmgr_obj->get_vservers_associated_to_client(
                                        -client_name => $client_name);
        if ($sanmgr_obj->error()) {
            $self->_set_err("Cannot get vservers associated to $log_label",
                            $sanmgr_obj->errmsg());
            return (wantarray()) ? () : [];
        }
        $self->sanlog(-msg => "Vservers associated to client "
                        . $client_name
                        . " are: @vservers",
                      -msg_level => 1);

        my @vserver_objs = ();
        if (scalar(@vservers)) {

            # Get all vserver objects associated with the sanmgr object
            my @all_vserver_objs = $sanmgr_obj->get_objects(type => "vserver");
            if ($sanmgr_obj->error()) {
                $self->_set_err("get_objects(vserver) failed: $log_label",
                                $sanmgr_obj->errmsg());
                return (wantarray()) ? () : [];
            }

            # Store the vserver objects specific to the vserver
            foreach my $vserver (@vservers) {
                foreach my $vserver_obj (@all_vserver_objs) {
                    if ($vserver eq $vserver_obj->get_name()) {
                        push(@vserver_objs, $vserver_obj);
                        last;
                    }
                }
            }
        } ## end if (scalar(@vservers))

        # Create a temp hash to store the client info
        my %client_info = {};
        $client_info{name}         = $client_name;
        $client_info{object}       = $client_obj;
        $client_info{vservers}     = \@vservers;
        $client_info{vserver_objs} = \@vserver_objs;

        # Verify the client obj has > 0 initiators and that the nodename
        # exists.  We need to grab the initiator's nodename.

        # Grab the initiators
        my @inits = $client_info{object}->get_initiators();
        if ($client_info{object}->error()) {
            $self->_set_err("get_initiators failed for $log_label",
                            $client_info{object}->errmsg());
            return (wantarray()) ? () : [];
        }

        # Verify we got at least 1 initiator back
        if (!(scalar(@inits))) {
            $self->_set_err("No initiators found for $log_label");
            return (wantarray()) ? () : [];
        }

        # Grab the initiator's nodename from the initiator object
        $client_info{nodename} = $inits[0]->get_nodename();
        if ($inits[0]->error()) {
            $self->_set_err("get_nodename failed for $log_label",
                            $inits[0]->errmsg());
            return (wantarray()) ? () : [];
        }

        # Verify we got back a nodename
        if (!(length($client_info{nodename}))) {
            $self->_set_err("Missing nodename for $log_label");
            return (wantarray()) ? () : [];
        }

        $self->sanlog(-msg => "Created client_info for client $client_name",
                      -msg_level => 1);

        # Limit the dumper depth to 3 so the logs don't explode.
        $Data::Dumper::Maxdepth = 3;
        $self->sanlog(-msg => "client_info dump\n" . Dumper(\%client_info),
                      -msg_level => 3);
        $Data::Dumper::Maxdepth = 0;

        # Add it to the client_info. It will be saved in the object
        push(@{$self->{client_info}}, \%client_info);
    } ## end foreach my $client_obj (@client_objs)

    $self->sanlog(-msg       => "Success: client_info created",
                  -msg_level => 1);

    return (wantarray())
      ? @{$self->{client_info}}
      : \@{$self->{client_info}};
} ## end sub create_client_info

################################################################################
# Method:    get_client_info
# Objective: Get the array that contains data specific to clients.
#            It is an array of hashes. Each hash contains the following:
#            name         - contains scalar value, the name of the client
#            object       - contains client_obj
#            vservers     - contains an array of scalars, the names of vservers
#                           associated with the client
#            vserver_objs - contains an array of vserver_objs
# Details:
#     Inputs - none
#     Returns On Success, an array (or reference to array) of client_info hash
#             On error, an empty array (or reference to array)
################################################################################
sub get_client_info {
    my ($self) = @_;

    $self->_clear_err();

    return (wantarray()) ? @{$self->{client_info}} : $self->{client_info};
}

################################################################################
# login_all_client_info_clients
# Objective: client_info contains info of all vservers associated to each
#            client. Login all vservers associated to each client from the
#            {client_info} data structure. Verify the login succeeded.
# Details:
#     Inputs - -sanmgr_obj (required) - SanMgr_Obj to verify logins for
#
#              -sendtargets (optional) - Do a refresh of sendtargets or not
#              By default, sendtargets info will be refreshed.
#
#              -fail_on_login (optional) - When set, any attempt by the clients
#              to login to target should fail - fail_on_login=>1
#              By default, login clients should not fail - fail_on_login=>0
#              Note: A value of 1 means you expect ALL logins from ALL clients
#              to ALL vservers to FAIL.
#
#              -fail_on_login_vserver_hash (optional) - A hash of vserver names
#              which are expected to fail on login.  All other vservers should
#              succeed on login.  If fail_on_login = 1, then there is no need
#              to specify this hash.
#
#              -isns_address (optional) - Use isns discovery to this isns
#              server address instead of sendtargets.
#
#              -clients (optional) -An array of client names which need to login
#              to the iscsi tagets.
#
#              -iscsi_targets (optional) -An array of iscsi target objects.
#
#              -num_expected_luns (optional) - An integer of number of
#              luns that each client is expected to find.  If you provide this,
#              login speed is increased.
#
#     Returns On Success, $SanUtil::TRUE  - all clients login successfully
#             On error, $SanUtil::FALSE - login failed and set_err
################################################################################
sub login_all_client_info_clients {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    ###########################################################################
    # All args are passed down to a lower level function call for validation.
    # -sanmgr_obj will be validated in _login_logout_verify_all_client_info()
    # -sendtargets is optional param for _client_config()
    # -fail_on_login is optional param for _client_setup() and _client_config()
    # -fail_on_login_vserver_hash is optional param for
    #  _login_logut_one_client_info_client()
    # -isns_address is optional param for _client_config
    # -clients is an optional param that will be validated and used in
    # _login_logout_verify_all_client_info() method
    # -iscsi_targets is optional param for _login_logout_verify_one_client_info()
    # -num_expected_luns is an optional param for _client_discovery()
    ###########################################################################

    $args{'-action'} = "login";

    $self->_login_logout_verify_all_client_info_clients(%args);
    if ($self->error()) {
        $self->_set_err("Failed to login and verify all vserver/clients info",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub login_all_client_info_clients

################################################################################
# logout_all_client_info_clients
# Objective: client_info contains info of all vservers associated to each
#            client. Logout all vservers associated to each client from the
#            {client_info} data structure. Verify the logout succeeded.
# Details:
#     Inputs - -sanmgr_obj (required) - SanMgr_Obj to verify logouts for.
#
#              -keep_portals (optional) - Leave the target portals on the client
#              By default, Portals will be cleaned up
#
#     Returns On Success, $SanUtil::TRUE  - all clients logout successfully
#             On error, $SanUtil::FALSE - login failed and set_err
################################################################################
sub logout_all_client_info_clients {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    ###########################################################################
    # All args are passed down to a lower level function call for validation.
    # -sanmgr_obj will be validated in _login_logout_verify_all_client_info()
    # -keep_portals is optional param for _client_cleanup()
    ###########################################################################

    $args{'-action'} = "logout";

    $self->_login_logout_verify_all_client_info_clients(%args);
    if ($self->error()) {
        $self->_set_err("Failed to logout and verify all vserver/clients info",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg       => "Success",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub logout_all_client_info_clients

################################################################################
# Method:    verify_iscsi_expected
# Objective: Verifies the expected iSCSI information agrees with Data ONTAP.
#            Data comes from test case.
# Details:
#     Inputs - sanmgr_obj (required)    - sanmgr object
#            - expected_hash (required) - hash of expected data
#              Details for expected hash can be found here:
#              http://wikid/w/Scalable_SAN_Test_Engineering/CheatSheet/verify_iscsi_expected
#     Returns $SanUtil::TRUE  - all data verified
#             $SanUtil::FALSE - all data NOT verified and set_err
################################################################################
sub verify_iscsi_expected {
    my ($self, %args) = @_;

    my $sanmgr_obj;       # SanMgr object
    my $expected_hash;    # expected hash to verify

    $self->sanlog(-msg       => "START",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /sanmgr_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $sanmgr_obj = $value;
        } elsif ($key =~ /expected_hash/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $expected_hash = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(
                "-sanmgr_obj is a required parameter, must be a SanMgr object");
        return $SanUtil::FALSE;
    }

    # debug info
    $self->sanlog(-msg       => "expected_hash:\n" . Dumper($expected_hash),
                  -msg_level => 1);

    # Make sure vservers or clients have a key to verify.  Check exists first
    # because 'keys' will create the hash if it does not exist.
    if (not((exists($expected_hash->{vservers})
             and keys(%{$expected_hash->{vservers}}))
            or (exists($expected_hash->{clients})
                and keys(%{$expected_hash->{clients}})))
      ) {
        $self->_set_err(
            "No keys defined for vserver or clients. There's nothing to verify."
        );
        return $SanUtil::FALSE;
    } ## end if (not((exists($expected_hash...))))

    # Get the cluster object
    my @cluster_objs = $sanmgr_obj->get_objects(type => "cluster");
    if ($sanmgr_obj->error()) {
        $self->_set_err("Failed to get cluster objects", $sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Verify we got back some cluster objects.
    if (!scalar(@cluster_objs)) {
        $self->_set_err("get_objects(cluster) returned 0 objects");
        return $SanUtil::FALSE;
    }

    # Get the cserver object
    my $cserver_obj = $cluster_objs[0]->get_cserver_obj();
    if ($cluster_objs[0]->error()) {
        $self->_set_err("Failed to get cserver object",
                        $cluster_objs[0]->errmsg());
        return $SanUtil::FALSE;
    }

    # Now verify each vserver in expected_hash
    if (exists($expected_hash->{vservers})) {
        foreach my $vserver (keys %{$expected_hash->{vservers}}) {

            $self->sanlog(-msg       => "Processing vserver $vserver",
                          -msg_level => 1);

            # use the vserver name to get the vserver obj
            my @vserver_obj =
              $sanmgr_obj->get_objects(-type  => "vserver",
                                       -names => [$vserver]);
            if ($sanmgr_obj->error()) {
                $self->_set_err("get_objects failed for vserver $vserver",
                                $sanmgr_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # Verify we got a single vserver
            if (scalar(@vserver_obj) != 1) {
                $self->_set_err(  "get_objects returned "
                                . scalar(@vserver_obj)
                                . " vservers with name $vserver");
                return $SanUtil::FALSE;
            }

            # If a lif key exists, then let's verify the lifs.
            if (exists($expected_hash->{vservers}->{$vserver}->{lif})) {

                $self->_verify_iscsi_array(
                         -object => $vserver_obj[0],
                         -type   => "lif",
                         -value => $expected_hash->{vservers}->{$vserver}->{lif}
                );
                if ($self->error()) {
                    $self->_set_err("Vserver $vserver: failed to verify lifs!",
                                    $self->errmsg());
                    return $SanUtil::FALSE;
                }

            } ## end if (exists($expected_hash...))

            # If an iscsi key exists, then let's verify the iscsi.
            if (exists($expected_hash->{vservers}->{$vserver}->{iscsi})) {

                # Object that will be passed on to other functions.  Will either
                # be a cserver object or an iscsi object.
                my $object;

                # iscsi nodename for logging purposes.
                my $nodename;

                # Check if iscsi exists on the vserver
                my $iscsi_exist = $vserver_obj[0]->iscsi_exist();
                if ($vserver_obj[0]->error()) {
                    $self->_set_err("iscsi_exist failed for vserver $vserver",
                                    $vserver_obj[0]->errmsg());
                    return $SanUtil::FALSE;
                }

                # If iscsi exists, grab the object
                if ($iscsi_exist) {

                    # Use the vserver obj to get the iscsi object
                    $object = $vserver_obj[0]->get_iscsi_obj();
                    if ($vserver_obj[0]->error()) {
                        $self->_set_err(
                                    "get_iscsi_obj failed for vserver $vserver",
                                    $vserver_obj[0]->errmsg());
                        return $SanUtil::FALSE;
                    }

                    # Make sure we got an iscsi object
                    if (ref($object) !~ /iscsi/i) {
                        $self->_set_err("Failed to get iscsi object");
                        return $SanUtil::FALSE;
                    }

                    # Save the nodename for logging
                    $nodename = $object->get_nodename();
                    if ($object->error()) {
                        $self->_set_err("Failed to get iscsi nodename",
                                        $object->errmsg());
                        return $SanUtil::FALSE;
                    }

                } else {

                    # Else iscsi does not exist, we will use the cserver object.
                    $object   = $cserver_obj;
                    $nodename = "iscsi not created";
                }

                $self->sanlog(-msg => "Processing vserver $vserver: $nodename",
                              -msg_level => 1);

                # Verify the iscsi hash
                $self->_verify_expected_iscsi_hash(
                               -object       => $object,
                               -vserver_name => $vserver,
                               -expected_hash =>
                                 $expected_hash->{vservers}->{$vserver}->{iscsi}
                );
                if ($self->error()) {
                    $self->_set_err("_verify_expected_iscsi_hash failed for "
                                      . "vserver $vserver: $nodename",
                                    $self->errmsg());
                    return $SanUtil::FALSE;
                }

                $self->sanlog(
                    -msg => "Successfully verified vserver $vserver: $nodename",
                    -msg_level => 1);
            } ## end if (exists($expected_hash...))

        } ## end foreach my $vserver (keys %...)
    } ## end if (exists($expected_hash...))

    # If a clients key is defined in the expected_hash, perform a cluster level
    # verify for each client.  This checks totals for sessions, connections,
    # and initiators across all vservers in the cluster.  Note that we pass
    # the individual hash for one client, as opposed to all clients at one time.
    # This is because verify_iscsi_client_across_cluster is used by other methods.
    if (exists($expected_hash->{clients})) {
        foreach my $initiator_iqn (keys(%{$expected_hash->{clients}})) {

            # Verify the client across the cluster.
            $self->_verify_iscsi_client_across_cluster(
                    -cserver_obj => $cserver_obj,
                    -client_hash => $expected_hash->{clients}->{$initiator_iqn},
                    -initiator_iqn => $initiator_iqn);
            if ($self->error()) {
                $self->_set_err(
                    "_verify_iscsi_client_across_cluster failed for client $initiator_iqn",
                    $self->errmsg());
                return $SanUtil::FALSE;
            }
        } ## end foreach my $initiator_iqn (...)
    } ## end if (exists($expected_hash...))

    # Call verify_clients_hash for the clients part of the data structure.
    # Verifies: sessions, connections, and initiators totals per vserver and
    # specific elements.  Note that we pass in all clients at one time.  This
    # is because this method is only used by verify_iscsi_expected.
    if (exists($expected_hash->{clients})
        and keys(%{$expected_hash->{clients}})) {
        $self->_verify_expected_clients_hash(
                                      -cserver_obj  => $cserver_obj,
                                      -sanmgr_obj   => $sanmgr_obj,
                                      -clients_hash => $expected_hash->{clients}
        );
        if ($self->error()) {
            $self->_set_err("_verify_clients_hash failed", $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    # If code reaches here, everything passed.
    $self->sanlog(-msg       => "SUCCESS!",
                  -msg_level => 1);
    return $SanUtil::TRUE;

} ## end sub verify_iscsi_expected

################################################################################
# client_info_run_io
# Objective: client_info contains info of all vservers associated to each
#            client. Run IO to devices associated to each client from the
#            {client_info} data structure. It accepts all the params that
#            IOMan::create_io_records() accepts.
# Details:
#     Inputs - -enable_defaults (optional) - Enable default values for IOMan
#               params.
#               By default, enable_defaults is set.
#               If enable_defaults is set to 1, <ioman_create_io_records_params>
#               are set to default recommended values. These params can be
#               overridden by user passed values.
#               For example, blocksize is randomly selected by default but if
#               the user passes -bs => 1k, 1k value will be used along with the
#               other defaults.
#               If enable_defaults is set to 0, the user wanted an entirely
#               different type of IO workload and it is expected that user
#               passes all the correct IOMan params.
#
#              -io_tool (optional) - load generation tool type.
#               Default: dt, type:string, Allowed values: dt|sio
#
#              -bs (optional) - Block size in bytes. By default, if no block
#               size is passed, it will select a random blocksize from the known
#               list of block sizes that's important to test with.
#               An additional method for use with '-io_tool'=>'dt' is to set
#               block size to "variable_record_<min>_<max>"
#               <min> and <max> values are configurable (i.e. if want to write
#               blocks size of 4k to 6k use "variable_record_4k_6k").
#               Default: random, type:string
#
#              <%ioman_create_io_records_params> (optional) - Pass any IOMan
#              specific params. See http://wikid.netapp.com/w/Scalable_SAN_Test_Engineering/Automation/IOMan/user_guide
#
#              -clients (optional) - Array reference to a list of client names.
#               This list will be used to filter running IO to devices of only
#               those clients that are part of the -clients list.
#
#              -wait_finish (optional) - Wait on finishing IO.
#               If set to 1, it will wait on finishing IO thereby making a call
#               to this function blocking. User will not be able to run anything
#               in parallel while running IO (unless it is subtest'd).
#               If set to 0, it will return immediately after starting IO. User
#               will need to make sure it has the IOMan_Obj via get_ioman_obj()
#               and then run ioman_obj->display_results() or
#               ioman_obj->wait_finish()
#
#     Returns On Success, $SanUtil::TRUE  - IO ran successfully
#             On error, $SanUtil::FALSE - IO failed and set_err
################################################################################
sub client_info_run_io {

    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    my ($bs, $enable_defaults, $io_tool, $wait_finish);
    my @io_records;
    my @client_names;    # Names of clients user wants to verify
    my $value;           # String of client hostnames to verify

    # Default value for enable_defaults
    $enable_defaults = 1;
    $wait_finish     = 1;

    # Parse arguments
    foreach (keys %args) {
        if (/^-io_tool$/) {
            $io_tool = lc($args{$_});

            # No need for error check as we will assign 'dt' by default
        } elsif (/^-bs$/) {
            $bs = $args{$_};

            # No need for error check as we will randomly assign a blocksize
            # by default
        } elsif (/^-wait_finish$/) {
            $wait_finish = $args{$_};
            unless ($wait_finish =~ /^(1|0)$/) {
                $self->_set_err(
                    "Error: invalid value for \"-wait_finish\", expected 0 or 1: $args{$_}"
                );
                return (wantarray()) ? () : [];
            }

            # Remove it from the args hash since it is not read by IOMan object
            delete $args{$_};
        } elsif (/^-enable_defaults$/) {
            $enable_defaults = $args{$_};
            unless ($enable_defaults =~ /^(1|0)$/) {
                $self->_set_err(
                    "Error: invalid value for \"-enable_defaults\", expected 0 or 1: $args{$_}"
                );
                return (wantarray()) ? () : [];
            }

            # Remove it from the args hash since it is not read by IOMan object
            delete $args{$_};
        } elsif (/-?clients/) {
            $value = $args{$_};
            if (not length($value)) {
                $self->_set_err("Invalid value for argument '-clients'");
                return (wantarray()) ? () : [];
            }

            # Store the client filter into an array
            @client_names = @{$value};

            # Remove it from the args hash since it is not ready by IOMan object
            delete $args{$_};
        } ## end elsif (/-?clients/)
    }    # end foreach arg

    # Default value (as per IOMan)
    if (not length($io_tool)) {
        $args{'-io_tool'} = $io_tool = "dt";
    }

    unless (ref($self->{client_info}) =~ /ARRAY/) {
        $self->_set_err(  "client_info must be created before using it.\n"
                        . "Call create_client_info() to create the client_info"
                        . ref($self->{client_info}));
        return (wantarray()) ? () : [];
    }

    my $ioman_obj = $self->{_ioman_obj};
    unless (ref($ioman_obj) =~ /^IOMan$/i) {
        $self->_set_err("Failed to find IOMan object. Make sure IOMan object is"
              . " instantiated by calling initialize_objects() before trying to"
              . " run IO. Ref: "
              . ref($ioman_obj));
        return (wantarray()) ? () : [];
    }

    # Cleanup the io_objs to make way for the next set
    $self->sanlog(-msg       => "Cleaning the io_objs for next iteration",
                  -msg_level => 1);
    $ioman_obj->cleanup();
    if ($ioman_obj->error()) {
        $self->_set_err("Failed to cleanup IO Processes", $ioman_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # Sleep a few seconds for IOMan to cleanup the internal processes
    &SanUtil::nate_sleep(10);

    # Make sure the client names in the client filter are valid
    foreach my $client_name (@client_names) {
        unless (scalar grep { $_->{'name'} eq $client_name }
                @{$self->{client_info}}) {
            $self->_set_err(  "-clients filter contains '$client_name' which "
                            . "was not found in {client_info} data structure."
                            . " Make sure -clients filter contains valid client"
                            . " names that were used by SanMgr");
            return (wantarray()) ? () : [];
        }
    } ## end foreach my $client_name (@client_names)

    my @random_block_sizes =
      ("512", "1k", "2k", "4k", "8k", "16k", "32k", "64k", "128k", "256k");
    unless ($bs) {
        $bs = $random_block_sizes[int(rand(scalar(@random_block_sizes)))];
        $args{'-bs'} = $bs;
    }
    $self->sanlog(-msg       => "Selecing a random block size of $bs",
                  -msg_level => 1);

    # An additional method for use with 'dt' is to set block size to
    # "variable_record_1k_8k"
    # These values are configurable (i.e. if want to write blocks size
    # of 4k to 6k use "variable_record_4k_6k").
    if (   $bs =~ /variable_record_(.+)_(.+)/i
        && $io_tool eq "dt") {
        $args{'-incr'} = "var";
        $args{'-min'}  = $1;
        $args{'-max'}  = $2;

        # Remove the variable_record bs from the args hash
        delete $args{'-bs'};
    }

    if ($enable_defaults) {

        # Add default ioman hash for 'dt'
        if ($io_tool eq "dt") {
            unless (exists $args{'-tool_version'}) {
                $args{'-tool_version'} = "current";
            }
            unless (exists $args{'-read'}) {
                $args{'-read'} = 0;
            }
            unless (exists $args{'-oncerr'}) {
                $args{'-oncerr'} = "abort";
            }
            unless (exists $args{'-disable'}) {
                $args{'-disable'} = "pstats,fsync";
            }
            unless (exists $args{'-enable'}) {
                $args{'-enable'} = "noprog,syslog";
            }
            unless (exists $args{'-oflags'}) {
                $args{'-oflags'} = "trunc";
            }
            unless (exists $args{'-errors'}) {
                $args{'-errors'} = 1;
            }
            unless (exists $args{'-pattern'}) {
                $args{'-pattern'} = "iot";
            }
            unless (exists $args{'-noprogt'}) {
                $args{'-noprogt'} = "15s";
            }
            unless (exists $args{'-noprogtt'}) {
                $args{'-noprogtt'} = "300s";
            }
            unless (exists $args{'-alarm'}) {
                $args{'-alarm'} = "7s";
            }
            unless (exists $args{'-pass_thru'}) {
                $args{'-pass_thru'} = "dtype=disk";
            }
        } elsif ($io_tool eq "sio") {

            # Add default ioman hash for 'sio'
            unless (exists $args{'-random_perc'}) {
                $args{'-random_perc'} = "50";
            }
            unless (exists $args{'-read_perc'}) {
                $args{'-read_perc'} = "50";
            }
            unless (exists $args{'-pass_thru'}) {
                $args{'-pass_thru'} = "-direct";
            }
            unless (exists $args{'-threads'}) {
                $args{'-threads'} = 8;
            }
            unless (exists $args{'-verify'}) {
                $args{'-verify'} = 1;
            }
        } ## end elsif ($io_tool eq "sio")

        # runtime default to 10 minutes
        unless (exists $args{'-runtime'}) {
            $args{'-runtime'} = 600;
        }
    } ## end if ($enable_defaults)

    # Do action per client
    foreach my $client_info (@{$self->{client_info}}) {

        # Get Client Name and object
        my $client     = $client_info->{name};
        my $client_obj = $client_info->{object};

        # Apply client filter
        if (scalar(@client_names) and not(grep /^$client$/, @client_names)) {
            $self->sanlog(
                       -msg => "client $client is skipped as a -client filter '"
                         . join(', ', @client_names)
                         . "' is applied",
                       -msg_level => 1);
            next;
        }

        $self->sanlog(-msg       => "Creating IO object for $client",
                      -msg_level => 1);

        # Create argument hash for IOMan object
        $args{'-client'} = $client;

        # Check if devices have been discovered for client_info hash
        unless (ref($client_info->{devices}) =~ /ARRAY/) {
            $self->_set_err(  "Devices must be discovered before running IO. "
                            . "Call discover_devices_client_info_clients() to "
                            . "populate the device list in client_info");
            return (wantarray()) ? () : [];
        }
        unless (scalar(@{$client_info->{devices}})) {
            $self->_set_err(  "No devices found. There could be multiple "
                            . "reasons.\nSome of the reasons could be:"
                            . "\niSCSI service could be down\nLIFs are down"
                            . "\nLUNs are not mapped correctly.");
            return (wantarray()) ? () : [];
        }
        $args{'-devices'} = $client_info->{devices};

        # DT Trigger based on Host OS
        if ($io_tool eq "dt") {
            unless (exists $args{'-trigger'}) {
                if ($client_obj->hosttype() =~ /windows/i) {
                    $args{'-trigger'} =
                      "\"cmd:X:\\SAN\\bin\\common\\dt_noprog_script.bat\"";
                } else {
                    $args{'-trigger'} =
                      "\"cmd:\/x\/eng\/localtest\/noarch\/bin\/dt_noprog_script.ksh\"";
                }
            }
        } ## end if ($io_tool eq "dt")

        # Create IOMan IO records
        my @client_io_records = $ioman_obj->create_io_records(%args);
        if ($ioman_obj->error()) {
            $self->_set_err("_run_io() Failed to create IO object",
                            $ioman_obj->errmsg());
            return (wantarray()) ? () : [];
        }
        push(@io_records, @client_io_records);
    } ## end foreach my $client_info (@{...})

    # Start IO
    $self->sanlog(-msg => "Start IO", -msg_level => 1);
    $ioman_obj->start();
    if ($ioman_obj->error()) {
        $self->_set_err("Failed to start IO", $ioman_obj->errmsg());
        return (wantarray()) ? @io_records : \@io_records;
    }

    if ($wait_finish) {

        # Wait for io_objs run to finish
        $self->sanlog(-msg       => "Wait for io_objs run to finish",
                      -msg_level => 1);
        $ioman_obj->display_results();
        if ($ioman_obj->error()) {
            $self->_set_err("Failed to display results", $ioman_obj->errmsg());
            return (wantarray()) ? @io_records : \@io_records;
        }
    } ## end if ($wait_finish)

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return (wantarray()) ? @io_records : \@io_records;
} ## end sub client_info_run_io

################################################################################
# discover_devices_client_info_clients
# Objective: client_info contains info of all vservers associated to each
#            client. Discover devices associated to each client from the
#            {client_info} data structure and store them in their corresponding
#            {client_info} entry.
# Details:
#     Inputs - -sanmgr_obj (optional) - SanMgr_Obj that configured and setup
#               the clients from the {client_info} data structure.
#
#              -clients (optional) - Array reference to list of client names.
#               This list will be used to filter discovering of devices only
#               for clients as part of the -clients list.
#
#     Returns On Success, $SanUtil::TRUE  - all devices discovered successfully
#             On error, $SanUtil::FALSE - get device failed and set_err
################################################################################
sub discover_devices_client_info_clients {

    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    my ($sanmgr_obj);
    my @client_names;    # Names of clients user wants to verify
    my $value;           # String of client hostnames to verify

    # Parse arguments
    foreach (keys %args) {
        if (/-?sanmgr_obj/i) {
            $sanmgr_obj = $args{$_};
            if (ref($sanmgr_obj) !~ /^San::SanMgr$/i) {
                $self->_set_err("Invalid value for argument '-sanmgr_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?clients/) {
            $value = $args{$_};
            if (not length($value)) {
                $self->_set_err("Invalid value for argument '-clients'");
                return $SanUtil::FALSE;
            }
            @client_names = @{$value};
        }
    }    # end foreach arg

    unless (ref($self->{client_info}) =~ /ARRAY/) {
        $self->_set_err(  "client_info must be created before using it."
                        . "Call create_client_info() to create the client_info"
                        . ref($self->{client_info}));
        return $SanUtil::FALSE;
    }

    # Make sure the client names in the client filter are valid
    foreach my $client_name (@client_names) {
        unless (scalar grep { $_->{'name'} eq $client_name }
                @{$self->{client_info}}) {
            $self->_set_err(  "-clients filter contains '$client_name' which "
                            . "was not found in {client_info} data structure."
                            . " Make sure -clients filter contains valid client"
                            . " names that were used by SanMgr");
            return $SanUtil::FALSE;
        }
    } ## end foreach my $client_name (@client_names)

    foreach my $client_info (@{$self->{client_info}}) {

        # Get Client Object
        my $client_obj = $client_info->{object};

        # Get Client Name
        my $client = $client_info->{name};

        # Apply client filter
        if (scalar(@client_names) and not(grep /^$client$/, @client_names)) {
            $self->sanlog(
                       -msg => "client $client is skipped as a -client filter '"
                         . join(', ', @client_names)
                         . "' is defined",
                       -msg_level => 2);
            next;
        }

        $self->sanlog(-msg       => "Got clientname of client " . $client,
                      -msg_level => 1);

        my @device_paths = ();

        # If SanMgr_Obj was passed
        if (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {

            # Get client device list
            my @client_dev_lists =
              $sanmgr_obj->get_client_dev_list(-client => $client);
            if ($sanmgr_obj->error()) {
                $self->_set_err("Failed to get client device list!",
                                $sanmgr_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # Get devices.
            foreach my $client_dev_list (@client_dev_lists) {
                if (defined $client_dev_list->{$client}) {
                    foreach my $cl_dev (@{$client_dev_list->{$client}}) {
                        push(@device_paths, $cl_dev->{path});
                    }
                }
            }
        } else {

            # Create a device manager object
            my $devmgrobj = $client_obj->create_devmgr();
            if ($client_obj->error()) {
                $self->_set_err("Failed to create device manager obj ",
                                $client_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # get the list of devices pathnames
            @device_paths = $devmgrobj->get_free_devices(-refresh => 1);
            if ($devmgrobj->error() or not scalar(@device_paths)) {
                $self->_set_err("Found no available device ",
                                $devmgrobj->errmsg());
                return $SanUtil::FALSE;
            }

            # For Solaris, need to convert device name to raw device name.
            if ($client_obj->hosttype() =~ /unix-solaris/) {

                foreach my $device (@device_paths) {

                    # Get the device object.
                    my $device_object = $devmgrobj->get_device($device);
                    if ($devmgrobj->error()) {
                        $self->_set_err("get_device failed",
                                        $devmgrobj->errmsg());
                        return $SanUtil::FALSE;
                    }

                    # Get the raw device name.  Note the raw device is saved
                    # into $device, which is really a reference to a location in
                    # the @device_paths array.  The @device_paths array is saved
                    # into the client_info hash before the method returns.
                    $device = $device_object->get_raw_device();
                    if ($device_object->error()) {
                        $self->_set_err("get_raw_device failed",
                                        $device_object->errmsg());
                        return $SanUtil::FALSE;
                    }

                } ## end foreach my $device (@device_paths)

            } ## end if ($client_obj->hosttype...)

            $self->sanlog(
                       -msg => "List of discovered raw device_paths includes:\n"
                         . Dumper(@device_paths),
                       -msg_level => 2);

        } ## end else [ if (ref($sanmgr_obj) =~...)]

        $self->sanlog(-msg => "Got device paths for $client: "
                        . join(", ", @device_paths),
                      -msg_level => 1);

        # Store @device_paths to the client info hash
        $client_info->{devices} = \@device_paths;
    } ## end foreach my $client_info (@{...})

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub discover_devices_client_info_clients

################################################################################
# get_ips
# Objective: Return a list of IP addresses based on SanMgr IP address range format
# Details:
#     Inputs - -ip_range (required) - Provide a range of IP addresses based on
#               SanMgr IP address range format to get an array of IP addresses.
#               Example: -ip_range => "10.228.[128-130].121"
#                        -ip_range => "10.228.128.[121-123]"
#                        -ip_range => "10.228.[128-135].[121-123]"
#     Returns On Success, an array (or reference to array) of IP addresses
#             On error, an empty array (or reference to array) and set_err
################################################################################
sub get_ips {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $ip_range;
    my (@ips) = ();

    # Parse arguments
    foreach (keys %args) {
        if (/-?ip_range$/i) {
            $ip_range = $args{$_};
            if (not length($ip_range)) {
                $self->_set_err("Invalid value for argument '-ip_range'");
                return (wantarray()) ? () : [];
            }
        }
    }    # end foreach arg

    unless ($ip_range) {
        $self->_set_err("No IP range found");
        return (wantarray()) ? () : [];
    }

    # if the ip range contains a colon, treat it as an ipv6 range
    if ($ip_range =~ /:/) {
        my @chunks = split(/:/, $ip_range);
        my $last_chunk = $chunks[-1];
        my $hex_regex = "[A-Fa-f0-9]";
        if ($last_chunk =~ /\[($hex_regex+)-($hex_regex+)\]/) {
            my $range_base = hex($1);
            my $range_max  = hex($2);

            for (my $ip_segment = $range_base; $ip_segment <= $range_max; $ip_segment++){
                my $stringified_chunk = sprintf("%x", $ip_segment);
                $chunks[-1] = $stringified_chunk;
                my $addr = join(":" , @chunks);
                push(@ips, $addr);
            }
        }
    } else {
        # 10.10.10.[start-end]
        # 10.10.[start-end].[start-end]
        # 10.10.[start-end].10
        # split ip_range into four bytes
        my ($first, $second, $third, $fourth) = split(/\./, $ip_range);

        # find the range in the third and fourth byte
        if ($third =~ /\[(\d+)-(\d+)\]/) {
            for (my $i = $1 ; $i <= $2 ; $i++) {

                # do we have the range in the forth byte as well?
                if ($fourth =~ /\[(\d+)-(\d+)\]/) {
                    for (my $j = $1 ; $j <= $2 ; $j++) {
                        push(@ips, "$first.$second.$i.$j");
                    }
                } else {

                    # range in third byte only
                    push(@ips, "$first.$second.$i.$fourth");
                }
            } ## end for (my $i = $1 ; $i <=...)
        } elsif ($fourth =~ /\[(\d+)-(\d+)\]/) {

            # range in fourth byte only
            for (my $i = $1 ; $i <= $2 ; $i++) {
                push(@ips, "$first.$second.$third.$i");
            }
        } else {
            $self->_set_err("IPv4 range does not have a valid format");
            return (wantarray()) ? () : [];
        }
    }
    return (wantarray()) ? @ips : \@ips;
} ## end sub get_ips

################################################################################
# register_eventman_resources
# Objective: Register cluster resources with EventMan which are useful for
#            starting events and monitoring activity. Resources include
#            Controller Cluster and Clients
# Details:
#     Inputs - sanmgr_obj (required)    - sanmgr object
#     Returns On Success, $SanUtil::TRUE  - all resources added successfully
#             On error, $SanUtil::FALSE - adding resource failed and set_err
################################################################################
sub register_eventman_resources {

    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    my ($sanmgr_obj);

    # Parse arguments
    foreach (keys %args) {
        if (/-?sanmgr_obj/i) {
            $sanmgr_obj = $args{$_};
            if (ref($sanmgr_obj) !~ /^San::SanMgr$/i) {
                $self->_set_err("Invalid value for argument '-sanmgr_obj'");
                return $SanUtil::FALSE;
            }
        }
    }    # end foreach arg

    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(  "Missing required argument '-sanmgr_obj'."
                        . "Must be a SanMgr object. Object: "
                        . ref($sanmgr_obj));
        return $SanUtil::FALSE;
    }

    my $eventman_obj = $self->get_eventman_obj();
    if (!$eventman_obj || $self->error()) {
        $self->_set_err(
            "Failed to find EventMan object. Make sure EventMan object is"
              . " instantiated by calling initialize_objects() before trying to"
              . " run any event.",
            $self->error());
        return $SanUtil::FALSE;
    }

    # Read the cluster object
    my @cluster_objs = $sanmgr_obj->get_objects(type => "cluster");
    if ($sanmgr_obj->error()) {
        $self->_set_err(" Retreiving cluster object from SanMgr_Obj failed",
                        $sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    if (scalar(@cluster_objs) != 1) {
        $self->_set_err("More than 1 cluster found: " . scalar(@cluster_objs));
        return $SanUtil::FALSE;
    }

    my $cluster_obj = $cluster_objs[0];
    unless (ref($cluster_obj) =~ /^San::ControllerCluster$/) {

        $self->_set_err("Cluster object not found: " . ref($cluster_obj));
        return $TCD::FAIL;
    }

    # Register the cluster with eventman
    $eventman_obj->add_resource(-name => $cluster_obj->get_name(),
                                -obj  => $cluster_obj,
                                -type => "controllerclusters");
    if ($eventman_obj->error()) {
        $self->_set_err("Failed to add resource to EventMan object",
                        $eventman_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Get all the client objects
    my @client_objs = $sanmgr_obj->get_objects(type => "client");
    if ($sanmgr_obj->error()) {
        $self->_set_err("ERROR: retreiving client objects from SanMgr",
                        $sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Register the clients with eventman
    foreach my $client_obj (@client_objs) {

        my $hostname = $client_obj->hostname();

        $eventman_obj->add_resource(-name => $hostname,
                                    -obj  => $client_obj,
                                    -type => "sanclients");
        if ($eventman_obj->error()) {
            $self->_set_err("Failed to add client $hostname to eventman",
                            $eventman_obj->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end foreach my $client_obj (@client_objs)

    # Enable SFO if not already enabled
    $cluster_obj->enable_sfo();
    if ($cluster_obj->error()) {
        $self->_set_err("Failed to enable sfo", $cluster_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub register_eventman_resources

################################################################################
# clean_eventman_events
# Objective: Reset EventMan object to start the next fresh event set.
# Details:
#     Inputs - none
#     Returns On Success, $SanUtil::TRUE  - all events removed successfully
#             On error, $SanUtil::FALSE - cleaning events failed and set_err
################################################################################
sub clean_events {

    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    unless (scalar(@{$self->{_event_objs}})) {
        $self->sanlog(-msg => "No Event to clean");
        return $SanUtil::TRUE;
    }

    # Get the eventman object
    my $eventman_obj = $self->get_eventman_obj();
    if (!$eventman_obj || $self->error()) {
        $self->_set_err(
            "Failed to find EventMan object. Make sure EventMan object is"
              . " instantiated by calling initialize_objects() before trying to"
              . " run any event.",
            $self->errmsg());
        return $SanUtil::FALSE;
    }

    # If Events are still running, return an error.
    if ($eventman_obj->is_running()) {
        $self->_set_err("Cannot kill event while the event is still running");
        return $SanUtil::FALSE;
    }

    my $event_obj;
    while (@{$self->{_event_objs}}) {    # do until event_objs is empty
        $event_obj = shift(@{$self->{_event_objs}});
        unless (ref($event_obj) =~ /^EventMan::Event$/i) {
            $self->_set_err(
                    "Failed to find Event object. Make sure Event object"
                  . " was saved by calling run_cluster_fault() or run_lif_fault() "
                  . "before trying to clean an event. Ref: "
                  . ref($event_obj));
            return $SanUtil::FALSE;
        }

        $eventman_obj->delete_event(-event_obj => $event_obj);
        if ($eventman_obj->error()) {
            $self->_set_err("Cannot delete event", $eventman_obj->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end while (@{$self->{_event_objs...}})

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub clean_events

################################################################################
# run_cluster_fault
# Objective: Run EventMan based Cluster faults
# Details:
#     Inputs - -enable_defaults (optional) - Enable default values for EventMan
#               params.
#               By default, enable_defaults is set.
#               If enable_defaults is set to 1, <eventman_add_events_params>
#               are set to default recommended values. These params can be
#               overridden by user passed values.
#               If enable_defaults is set to 0, the user wanted an entirely
#               different type of EventMan param values and it is expected that
#               the user passes all the correct EventMan params.
#
#              -events (optional) - Any cluster wide event supported by EventMan
#               Default: sfo_takeforcedgiveCm
#
#              -runid (optional) - Pass a runid to the EventMan subtest.
#               Default: EventMan_Event
#
#              <%eventman_add_events_params> (optional) - Pass any EventMan
#              specific params. See EventMan POD for details.
#
#              -async (optional) - Run Events in background or make it a waiting call
#               If set to 1, it will wait on finishing events thereby making a call
#               to this function blocking. User will not be able to run anything
#               in parallel while running events (unless it is subtest'd).
#               If set to 0, it will return immediately after starting events. User
#               will need to make sure it has the EventMan_Obj via get_eventman_obj()
#               Default is 1.
#
#     Returns On Success, $SanUtil::TRUE  - Event ran successfully
#             On error, $SanUtil::FALSE - Event failed and set_err
################################################################################
sub run_cluster_fault {

    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    my $enable_defaults = 1;
    my $async           = 1;
    my $runid           = "EventMan_Cluster_Event";
    my @events;
    my $event_list;

    # argument parsing
    foreach (keys %args) {
        if (/^-enable_defaults$/) {

            # Remove it from the args hash since it is not read by EventMan object
            $enable_defaults = delete $args{$_};
            unless ($enable_defaults =~ /^(1|0)$/) {
                $self->_set_err(
                    "Error: invalid value for \"-enable_defaults\", expected 0 or 1: $enable_defaults"
                );
                return $SanUtil::FALSE;
            }
        } elsif (/^-events$/) {

            # Remove it from the args hash since it is not read by EventMan object
            $event_list = delete $args{$_};
            if (ref($event_list) !~ /ARRAY/i) {
                $self->_set_err(
                           "Error: invalid value for \"-events\": $event_list");
                return $SanUtil::FALSE;
            }
            @events = @{$event_list};
        } elsif (/^-async$/) {

            # Remove it from the args hash since it is not read by EventMan object
            $async = delete $args{$_};
            unless ($async =~ /^(1|0)$/) {
                $self->_set_err(
                    "Error: invalid value for \"-async\", expected 0 or 1: $async"
                );
                return $SanUtil::FALSE;
            }
        } elsif (/^-runid$/) {

            # Remove it from the args hash since it is not read by EventMan object
            $runid = delete $args{$_};
            unless (length $runid) {
                $self->_set_err(
                    "Error: invalid value for \"-runid\", expected string: $runid"
                );
                return $SanUtil::FALSE;
            }
        } ## end elsif (/^-runid$/)
    } ## end foreach (keys %args)

    my $eventman_obj = $self->get_eventman_obj();
    if (!$eventman_obj || $self->error()) {
        $self->_set_err(
            "Failed to find EventMan object. Make sure EventMan object is"
              . " instantiated by calling initialize_objects() before trying to"
              . " run any event.",
            $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar(@events)) {
        @events = ("sfo_takeforcedgiveCm");
    }

    # Add default eventman hash
    if ($enable_defaults) {

        # Number of times to run an event
        unless (exists $args{'-cycle_count'}) {
            $args{'-cycle_count'} = 1;
        }

        # Used by Cluster events
        unless (exists $args{'-symmetrical_events'}) {
            $args{'-symmetrical_events'} = 1;
        }

        # Used by Cluster events
        unless (exists $args{'-sfo_timeout'}) {
            $args{'-sfo_timeout'} = 900;
        }

        # Used for Cluster events
        unless (exists $args{'-controller_down_time'}) {
            $args{'-controller_down_time'} = 300;
        }

        # Used for path validation
        unless (exists $args{'-path_validation_abort_on_error'}) {
            $args{'-path_validation_abort_on_error'} = 0;
        }

        # Used for path validation
        unless (exists $args{'-path_validation_min_wait_time'}) {
            $args{'-path_validation_min_wait_time'} = 1;
        }
    } ## end if ($enable_defaults)

    foreach my $event (@events) {
        $args{'-event_name'} = $event;

        # Add the event
        my $event_obj = $eventman_obj->add_event(%args);
        if ($eventman_obj->error()) {
            $self->_set_err("Error adding event to EventMan",
                            $eventman_obj->error());
            return $SanUtil::FALSE;
        }
        push(@{$self->{_event_objs}}, $event_obj);
    } ## end foreach my $event (@events)

    # then kick off the events
    $eventman_obj->run_events(-async          => $async,
                              -runid          => $runid,
                              -sequence_style => "iterations",
                              -sequence_count => 1);
    if ($eventman_obj->error()) {
        $self->_set_err("Failed to successfully complete the event",
                        $eventman_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub run_cluster_fault

################################################################################
# run_lif_fault
# Objective: Run EventMan based LIF fault
# Details:
#     Inputs - -enable_defaults (optional) - Enable default values for EventMan
#               params.
#               By default, enable_defaults is set.
#               If enable_defaults is set to 1, <eventman_add_events_params>
#               are set to default recommended values. These params can be
#               overridden by user passed values.
#               If enable_defaults is set to 0, the user wanted an entirely
#               different type of EventMan param values and it is expected that
#               the user passes all the correct EventMan params.
#
#              -lif_down_times (optional) - It accepts a list of LIF down/up times.
#               Default: [60, 30, 20, 10, 5, 3, 2, 1]
#
#              -runid (optional) - Pass a runid to the EventMan subtest.
#               Default: EventMan_Event
#
#              <%eventman_add_events_params> (optional) - Pass any EventMan
#              specific params. See EventMan POD for details.
#
#              -async (optional) - Run Events in background or make it a waiting call
#               If set to 1, it will wait on finishing events thereby making a call
#               to this function blocking. User will not be able to run anything
#               in parallel while running events (unless it is subtest'd).
#               If set to 0, it will return immediately after starting events. User
#               will need to make sure it has the EventMan_Obj via get_eventman_obj()
#               Default is 1
#
#     Returns On Success, $SanUtil::TRUE  - Event ran successfully
#             On error, $SanUtil::FALSE - Event failed and set_err
################################################################################
sub run_lif_fault {

    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    my $enable_defaults = 1;
    my $async           = 1;
    my $runid           = "EventMan_LIF_Event";
    my @lif_down_times  = (60, 30, 20, 10, 5, 3, 2, 1);

    # argument parsing
    foreach (keys %args) {
        if (/^-enable_defaults$/) {

            # Remove it from the args hash since it is not read by EventMan object
            $enable_defaults = delete $args{$_};
            unless ($enable_defaults =~ /^(1|0)$/) {
                $self->_set_err(
                    "Error: invalid value for \"-enable_defaults\", expected 0 or 1: $enable_defaults"
                );
                return $SanUtil::FALSE;
            }
        } elsif (/^-async$/) {

            # Remove it from the args hash since it is not read by EventMan object
            $async = delete $args{$_};
            unless ($async =~ /^(1|0)$/) {
                $self->_set_err(
                    "Error: invalid value for \"-async\", expected 0 or 1: $async"
                );
                return $SanUtil::FALSE;
            }
        } elsif (/^-runid$/) {

            # Remove it from the args hash since it is not read by EventMan object
            $runid = delete $args{$_};
            unless (length $runid) {
                $self->_set_err(
                    "Error: invalid value for \"-runid\", expected string: $runid"
                );
                return $SanUtil::FALSE;
            }
        } elsif (/^-lif_down_times$/) {

            # Remove it from the args hash since it is not read by EventMan object
            my $value = delete $args{$_};
            unless (ref($value) =~ /ARRAY/i) {
                $self->_set_err(
                    "Error: invalid value for \"-lif_down_times\", expected reference to array: "
                      . ref($value));
                return $SanUtil::FALSE;
            }
            @lif_down_times = @{$value};
        } ## end elsif (/^-lif_down_times$/)
    } ## end foreach (keys %args)

    my $eventman_obj = $self->get_eventman_obj();
    if (!$eventman_obj || $self->error()) {
        $self->_set_err(
            "Failed to find EventMan object. Make sure EventMan object is"
              . " instantiated by calling initialize_objects() before trying to"
              . " run any event.",
            $self->errmsg());
        return $SanUtil::FALSE;
    }

    if ($enable_defaults) {

        # Add default eventman hash
        unless (exists $args{'-cycle_count'}) {
            $args{'-cycle_count'} = 1;
        }
        unless (exists $args{'-collect_controller_sysstats'}) {
            $args{'-collect_controller_sysstats'} = 0;
        }
        unless (exists $args{'-collect_controller_memstats'}) {
            $args{'-collect_controller_memstats'} = 0;
        }
        unless (exists $args{'-path_validation_abort_on_error'}) {
            $args{'-path_validation_abort_on_error'} = 0;
        }

        # Wait at least 3 mins after lif fault to allow paths to settle.
        # Recommended by Windows Interop.
        unless (exists $args{'-path_validation_min_wait_time'}) {
            $args{'-path_validation_min_wait_time'} = 180;
        }

        # Do not validate paths on lif down faults.  Reason: This method doesn't
        # keep the paths down long enough for the paths to settle.
        unless (exists $args{'-path_validation_after_lif_down'}) {
            $args{'-path_validation_after_lif_down'} = 0;
        }
        unless (exists $args{'-event_name'}) {
            $args{'-event_name'} = "iscsi_lif_down_up_Cm";
        }
    } ## end if ($enable_defaults)

    my @vservers;
    my @lif_objs;

    if (ref($self->{vserver_info}) =~ /ARRAY/) {

        # Vserver loop
        foreach my $vserver_info (@{$self->{vserver_info}}) {
            push(@vservers, $vserver_info->{name});
        }
    } elsif (ref($self->{client_info}) =~ /ARRAY/) {

        # Client loop
        foreach my $client_info (@{$self->{client_info}}) {
            foreach my $client_elem_vserver (@{$client_info->{vservers}}) {
                if (grep (/^$client_elem_vserver$/, @vservers)) {
                    next;
                }
                push(@vservers, $client_elem_vserver);
            }
        }
    } else {
        $self->_set_err(
                  "Vserver_Info or Client_Info must be created before using it."
                    . "Call create_vserver_info() to create the vserver_info"
                    . " Or, create_client_info() to create the client_info");
        return $SanUtil::FALSE;
    }

    $self->sanlog(
             -msg       => "Vservers To run event on: " . join(", ", @vservers),
             -msg_level => 1);

    unless (exists $args{'-vservers'}) {
        $args{'-vservers'} = \@vservers;
    }

    foreach my $lif_down_time (@lif_down_times) {

        $args{'-lif_down_time'} = $lif_down_time;
        $args{'-lif_up_time'}   = $lif_down_time;

        my $event_obj = $eventman_obj->add_event(%args);
        if ($eventman_obj->error()) {
            $self->_set_err("Error adding event to EventMan",
                            $eventman_obj->error());
            return $SanUtil::FALSE;
        }
        push(@{$self->{_event_objs}}, $event_obj);
    } ## end foreach my $lif_down_time (...)

    # then kick off the events in the background
    $eventman_obj->run_events(-async          => $async,
                              -runid          => $runid,
                              -sequence_style => "iterations",
                              -sequence_count => 1);
    if ($eventman_obj->error()) {
        $self->_set_err("Failed to successfully complete the event",
                        $eventman_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub run_lif_fault

################################################################################
# run_port_fault
# Objective: Run Network interface (port) down/up fault
# Details:
#     Inputs - -sanmgr_obj (required) - SanMgr_Obj to get the Ports info
#
#              -cycle_count (optional) - How many times to run the fault.
#               Default: 10
#
#              -port_down_times (optional) - How long should the port be down.
#               Default: [60, 30, 20, 10, 5, 3, 2, 1]
#
#     Returns On Success, $SanUtil::TRUE  - Event ran successfully
#             On error, $SanUtil::FALSE - Event failed and set_err
################################################################################
sub run_port_fault {

    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    my @port_down_times = (60, 30, 20, 10, 5, 3, 2, 1);
    my $port_down_up_total_count = 10;
    my $sanmgr_obj;
    my @ports;
    my $value;

    # argument parsing
    foreach (keys %args) {
        if (/^-port_down_times$/) {
            $value = $args{$_};
            unless (ref($value) =~ /ARRAY/i) {
                $self->_set_err(
                    "Error: invalid value for \"-port_down_times\", expected reference to array: "
                      . ref($value));
                return $SanUtil::FALSE;
            }
            @port_down_times = @{$value};
        } elsif (/-?sanmgr_obj/i) {
            $sanmgr_obj = $args{$_};
            if (ref($sanmgr_obj) !~ /^San::SanMgr$/i) {
                $self->_set_err("Invalid value for argument '-sanmgr_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?ports/i) {
            $value = $args{$_};
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-ports', value must be a reference to an 'ARRAY'"
                );
                return $SanUtil::FALSE;
            }
            @ports = @{$value};
        } elsif (/-?cycle_count/i) {
            $value = $args{$_};
            if ($value <= 0) {
                $self->_set_err(
                    "For argument '-cycle_count', value must be greater than zero: $value"
                );
                return $SanUtil::FALSE;
            }
            $port_down_up_total_count = $value;
        } ## end elsif (/-?cycle_count/i)
    }    # end foreach arg

    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(  "Missing required argument '-sanmgr_obj'."
                        . "Must be a SanMgr object. Object: "
                        . ref($sanmgr_obj));
        return $SanUtil::FALSE;
    }

    my $eventman_obj = $self->get_eventman_obj();
    if (!$eventman_obj || $self->error()) {
        $self->_set_err(
            "Failed to find EventMan object. Make sure EventMan object is"
              . " instantiated by calling initialize_objects() before trying to"
              . " run any event.",
            $self->errmsg());
        return $SanUtil::FALSE;
    }

    my $port_up_time = 60;
    my @port_objs =
      $self->_get_port_objs(-sanmgr_obj => $sanmgr_obj, -ports => \@ports);
    if ($self->error()) {
        $self->_set_err("Failed to get Network Adapter objects for ports "
                        . join(", ", @ports));
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg       => "Event Ports: " . join(", ", @ports),
                  -msg_level => 1);

    foreach my $port_down_time (@port_down_times) {
        $self->sanlog(-msg => "Start Event for Lif-Down-Time $port_down_time",
                      -msg_level => 1);

        for (my $i = 0 ; $i < $port_down_up_total_count ; $i++) {
            $self->sanlog(
                     -msg =>
                       "Iteration '$i' for Event Lif-Down-Time $port_down_time",
                     -msg_level => 1);

            foreach my $port_obj (@port_objs) {

                my $portname = $port_obj->get_port();
                $self->sanlog(
                    -msg =>
                      "Do Port $portname 'down' in Iteration '$i' for Event Lif-Down-Time $port_down_time",
                    -msg_level => 1);

                # Modify port status
                $port_obj->modify('-up-admin' => "false");
                if ($port_obj->error()) {
                    $self->_set_err(
                        "Failed to change up-admin of network adapter '$portname' to 'false'",
                        $port_obj->errmsg());
                    return $SanUtil::FALSE;
                }

                $self->sanlog(
                    -msg =>
                      "Wait for $port_down_time for Port $portname 'down' in Iteration '$i'",
                    -msg_level => 1);

                # Sleep for port_down_time
                &SanUtil::nate_sleep($port_down_time);

                $self->sanlog(
                    -msg =>
                      "Do Port $portname 'up' in Iteration '$i' for Event Lif-Down-Time $port_down_time",
                    -msg_level => 1);

                # Modify port status
                $port_obj->modify('-up-admin' => "true");
                if ($port_obj->error()) {
                    $self->_set_err(
                        "Failed to change up-admin of network adapter '$portname' to 'true'",
                        $port_obj->errmsg());
                    return $SanUtil::FALSE;
                }

                $self->sanlog(
                    -msg =>
                      "Wait for $port_up_time for Port $portname 'down' in Iteration '$i'",
                    -msg_level => 1);

                # Sleep for port_up_time
                &SanUtil::nate_sleep($port_up_time);
            } ## end foreach my $port_obj (@port_objs)
        } ## end for (my $i = 0 ; $i < $port_down_up_total_count...)
    } ## end foreach my $port_down_time ...

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub run_port_fault

################################################################################
# get_iscsi_targets
# Objective: This function returns a array of iscsi target objects
# Details:
#     Inputs -ip_array (required) - An  array of IP addresses
#     Returns On Success, an array (or reference to array) of iscsi target objects
#             On error, an empty array (or reference to array) and set_err
################################################################################
sub get_iscsi_targets {

    my ($self, %args) = @_;

    my $iscsi_obj;
    my $filer_obj;
    my @ip_array;
    my @iscsi_targets;
    my @target_array;

    $self->_clear_err();

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    # Parse arguments
    foreach (keys %args) {
        if (/-?iscsi_obj/i) {
            $iscsi_obj = $args{$_};
            if (not length($iscsi_obj)) {
                $self->_set_err("Invalid value for argument '-iscsi_obj'");
                return (wantarray()) ? () : [];
            }
        } elsif (/-?ip_array/i) {
            my $value = $args{$_};
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-ip_array', value must be a reference to an 'ARRAY'"
                );
                return (wantarray()) ? () : [];
            }
            @ip_array = @{$value};
        } ## end elsif (/-?ip_array/i)
    }    # end foreach arg

    # Check for mandatory parameters.
    unless (scalar(@ip_array)) {
        $self->_set_err("Missing required argument '-ip_array'.");
        return (wantarray()) ? () : [];
    }
    unless (ref($iscsi_obj) =~
            /^San::ControllerCluster::Filer10M::Vserver::Iscsi$/i) {
        $self->_set_err(
                 "-iscsi_obj is a required parameter, must be an iSCSI object");
        return (wantarray()) ? () : [];
    }

    # Get the vserver object from the iscsi object
    my $vserver_obj = $iscsi_obj->get_parent_obj();
    if ($iscsi_obj->error()) {
        $self->_set_err("Failed: get_parent_obj", $iscsi_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get the filer object from the vserver object
    $filer_obj = $vserver_obj->get_parent_obj();
    if ($vserver_obj->error()) {
        $self->_set_err("Failed: get_parent_obj", $vserver_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get the name of the vserver
    my $vserver_name = $vserver_obj->get_name();
    if ($vserver_obj->error()) {
        $self->_set_err("Failed: get_name", $vserver_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get all the iscsi targets on the vserver.
    @iscsi_targets = $filer_obj->get_iscsi_targets(vserver => $vserver_name);
    if ($filer_obj->error()) {
        $self->_set_err(
            "Failed to get iscsi targets from filer for vserver $vserver_name.",
            $filer_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    $self->sanlog(-msg => "Got "
                    . scalar(@iscsi_targets)
                    . " iscsi targets for vserver "
                    . $vserver_name,
                  -msg_level => 1);

    foreach my $ip (@ip_array) {

        foreach my $target_obj (@iscsi_targets) {

            my $target_ip = $target_obj->get_ip();
            if ($target_obj->error()) {
                $self->_set_err("Failed to get iscsi target ip.",
                                $target_obj->errmsg());
                return (wantarray()) ? () : [];
            }

            if ($ip eq $target_ip) {
                push(@target_array, $target_obj);
                last;
            }

        } ## end foreach my $target_obj (@iscsi_targets)

    } ## end foreach my $ip (@ip_array)

    # If match for one or more IP's sent is not found
    # set error and return empty array.
    if (scalar(@ip_array) != scalar(@target_array)) {
        $self->_set_err("Could not find a match for one or more IP's.");
        return (wantarray()) ? () : [];
    }

    # If all the IP's sent have a match return an array
    # of corresponding iscsi target objects.
    return (wantarray()) ? @target_array : \@target_array;

} ## end sub get_iscsi_targets

################################################################################
# get_iscsi_targets_enabled
# Objective: This function scans each vserver in Vserver_Info.  It then runs
#            'iscsi interface show -enabled true' on each vserver.  It saves
#            all of the iscsi target objects and returns them.  The iscsi target
#            objects can then be used in our login/logout calls.
# Details:
#     Inputs -none
#     Returns On Success, an array (or reference to array) of iscsi target objects
#             On error, an empty array (or reference to array) and set_err
################################################################################
sub get_iscsi_targets_enabled {

    my ($self, %args) = @_;

    my @final_target_array;

    $self->_clear_err();

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    # grab vserver info
    my @vserver_info = $self->get_vserver_info();
    if ($self->error()) {
        $self->_set_err("Failed to get_vserver_info", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Scan each vserver
    foreach my $vserver (@vserver_info) {

        # Grab the vserver object and name from Vserver_Info
        my $vserver_obj  = $vserver->{object};
        my $vserver_name = $vserver->{name};

        # Grab iscsi object
        my $iscsi_obj = $vserver_obj->get_iscsi_obj();
        if ($vserver_obj->error()) {
            $self->_set_err("get_iscsi_obj failed on vserver $vserver_name",
                            $vserver_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        # Verify iscsi object was returned
        if (ref($iscsi_obj) !~ /iscsi/i) {
            $self->_set_err(
                           "Could not get iscsi obj for vserver $vserver_name");
            return (wantarray()) ? () : [];
        }

        # get list of iscsi interfaces that are enabled
        my @targets = $iscsi_obj->interface_list_info(-enabled => "true");
        if ($iscsi_obj->error()) {
            $self->_set_err(
                         "iscsi interface show failed on vserver $vserver_name",
                         $iscsi_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        my @ip_array;

        # Scan each interface returned.  Save each ip address on to an array.
        foreach my $tgt (@targets) {
            push(@ip_array, $tgt->{address});
        }

        if (!scalar(@ip_array)) {
            $self->sanlog(
                    -msg => "No iscsi targets found for vserver $vserver_name");
            next;
        }

        # get iscsi target objects
        my @iscsi_targets =
          $self->get_iscsi_targets(-ip_array  => \@ip_array,
                                   -iscsi_obj => $iscsi_obj);
        if ($self->error()) {
            $self->_set_err("Failed to get_iscsi_targets", $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # Save targets onto final array
        push(@final_target_array, @iscsi_targets);
    } ## end foreach my $vserver (@vserver_info)

    if (scalar(@final_target_array)) {
        return (wantarray()) ? @final_target_array : \@final_target_array;
    } else {
        return (wantarray()) ? () : [];
    }

} ## end sub get_iscsi_targets_enabled

################################################################################
# get_iscsi_targets_disabled
# Objective: This function scans each vserver in Vserver_Info.  It then runs
#            'iscsi interface show -enabled false' on each vserver.  It saves
#            all of the iscsi target objects and returns them.  The iscsi target
#            objects can then be used in our login/logout calls.
# Details:
#     Inputs -none
#     Returns On Success, an array (or reference to array) of iscsi target objects
#             On error, an empty array (or reference to array) and set_err
################################################################################
sub get_iscsi_targets_disabled {

    my ($self, %args) = @_;

    my @final_target_array;

    $self->_clear_err();

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    # grab vserver info
    my @vserver_info = $self->get_vserver_info();
    if ($self->error()) {
        $self->_set_err("Failed to get_vserver_info", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Scan each vserver
    foreach my $vserver (@vserver_info) {

        # Grab the vserver object and name from Vserver_Info
        my $vserver_obj  = $vserver->{object};
        my $vserver_name = $vserver->{name};

        # Grab iscsi object
        my $iscsi_obj = $vserver_obj->get_iscsi_obj();
        if ($vserver_obj->error()) {
            $self->_set_err("get_iscsi_obj failed on vserver $vserver_name",
                            $vserver_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        # Verify iscsi object was returned
        if (ref($iscsi_obj) !~ /iscsi/i) {
            $self->_set_err(
                           "Could not get iscsi obj for vserver $vserver_name");
            return (wantarray()) ? () : [];
        }

        # get list of iscsi interfaces that are enabled
        my @targets = $iscsi_obj->interface_list_info(-enabled => "false");
        if ($iscsi_obj->error()) {
            $self->_set_err(
                         "iscsi interface show failed on vserver $vserver_name",
                         $iscsi_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        my @ip_array;

        # Scan each interface returned.  Save each ip address on to an array.
        foreach my $tgt (@targets) {
            push(@ip_array, $tgt->{address});
        }

        if (!scalar(@ip_array)) {
            $self->sanlog(-msg =>
                   "No disabled iscsi targets found for vserver $vserver_name");
            next;
        }

        # get iscsi target objects
        my @iscsi_targets =
          $self->get_iscsi_targets(-ip_array  => \@ip_array,
                                   -iscsi_obj => $iscsi_obj);
        if ($self->error()) {
            $self->_set_err("Failed to get_iscsi_targets", $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # Save targets onto final array
        push(@final_target_array, @iscsi_targets);
    } ## end foreach my $vserver (@vserver_info)

    if (scalar(@final_target_array)) {
        return (wantarray()) ? @final_target_array : \@final_target_array;
    } else {
        return (wantarray()) ? () : [];
    }

} ## end sub get_iscsi_targets_disabled

###############################################################################
# Method   : verify_isns_registration
# Objective: Verifies that a vserver's information on an iSNS server matches
#            the actual information on the vserver.
# Details  : See POD documentation at the end of this file
# Inputs   : -vserver_objs - optional - Array of vserver objs to verify. Default
#                is to verify all vservers in vserver_info.
#            -isns_server_obj - required - isns server object
#            -force_isns_update - optional - Perform iscsi iSNS update before
#                attempting to collect data from iSNS server. (Default is 0)
#            -retries - optional - Number of times to retry if vserver not found
#                or some of the data is stale.
#                If force_isns_update=0, Default is 1.
#                If force_isns_update=1, the number of retries is automatically
#                set to 1.
#                With isns update, it can take up to 15 secs for info to be
#                propagated.  So retry count of 1 is required.
# Returns  : $SanUtil::TRUE - Successful execution, all vservers verified
#            $SanUtil::FALSE & set_err - Execution failed
###############################################################################
sub verify_isns_registration {
    my ($self, %args) = @_;

    $self->_clear_err();

    my (@vserver_objs, $isns_server_obj, $node_name, $isns_hash, $result);
    my $retry_count       = 1;     # Number of retries if force=0.
    my $force_isns_update = 0;     # Set default force to 0.

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_objs/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-vserver_objs', value must be a reference to an 'ARRAY'"
                );
                return $SanUtil::FALSE;
            }
            @vserver_objs = @{$value};
        } elsif ($key =~ /isns_server_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $isns_server_obj = $value;
        } elsif ($key =~ /force_isns_update/) {
            if (not length($value) or $value !~ /^(1|0)$/) {
                $self->_set_err(  "Invalid value for \"$key\": "
                                . $value
                                . ". Must be 0 or 1.");
                return $SanUtil::FALSE;
            }
            $force_isns_update = $value;
        } elsif ($key =~ /retries/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $retry_count = $value;
            if ($retry_count !~ /^\d+$/) {
                $self->_set_err(
                          "Invalid value for -retries, value must be a number");
                return $SanUtil::FALSE;
            }
        } ## end elsif ($key =~ /retries/)
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($isns_server_obj) =~ /^San::SanClient::IsnsServer/i) {
        $self->_set_err(
            "-isns_server_obj is a required parameter, must be an iSNS Server object"
        );
        return $SanUtil::FALSE;
    }

    # If force_isns_update is 1, auto set the retry count to 0.
    if ($force_isns_update) {
        $retry_count = 1;
    }

    # debug info
    $self->sanlog(-msg       => "force_isns_update: $force_isns_update",
                  -msg_level => 2);
    $self->sanlog(-msg       => "retries: $retry_count",
                  -msg_level => 2);

    # Save isns name for logging.
    my $isns_name = $isns_server_obj->hostname();
    if ($isns_server_obj->error()) {
        $self->_set_err("Failed to get hostname from isns server obj",
                        $isns_server_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # If the user passed in vserver_objs, let's verify they are all vserver
    # objects.
    if (scalar(@vserver_objs)) {

        # Scan each vserver object.
        foreach my $vserver (@vserver_objs) {

            # Verify it's a vserver
            unless (
                ref($vserver) =~ /^San::ControllerCluster::Filer10M::Vserver$/i)
            {
                $self->_set_err("-vserver_objs must be vserver objects");
                return $SanUtil::FALSE;
            }
        } ## end foreach my $vserver (@vserver_objs)
    } else {

        # Else user did not pass in a vserver object.  We will process all
        # vserver objects in vserver_info.
        my @vserver_info = $self->get_vserver_info();
        if ($self->error()) {
            $self->_set_err("Failed to get_vserver_info", $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Verify we have at least 1 vserver to process.
        if (scalar(@vserver_info) < 1) {
            $self->_set_err("vserver_info has no vservers.");
            return $SanUtil::FALSE;
        }

        # Otherwise, we have some vservers.  We need to push the objects
        # on to our vserver_objs array.
        foreach my $vserver (@vserver_info) {
            push(@vserver_objs, $vserver->{object});
        }
    } ## end else [ if (scalar(@vserver_objs...))]

    # Now we have a list of vservers to verify.  Scan each vserver.
    foreach my $vserver_obj (@vserver_objs) {

        # Get vserver name
        my $vserver_name = $vserver_obj->get_name();
        if ($vserver_obj->error()) {
            $self->_set_err("get_name failed for vserver",
                            $vserver_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Setup log label for future messages.
        my $log_label = "Vserver: $vserver_name";

        # Now use the vserver obj to get the iscsi object
        my $iscsi_obj = $vserver_obj->get_iscsi_obj();
        if ($vserver_obj->error()) {
            $self->_set_err("$log_label: get_iscsi_obj failed",
                            $vserver_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Verify iscsi object was returned. (could be empty if no iscsi)
        if (ref($iscsi_obj) !~ /iscsi/i) {
            $self->_set_err(
                          "$log_label: get_iscsi_obj did not return iscsi obj");
            return $SanUtil::FALSE;
        }

        # Get the nodename for the vserver.
        $node_name = $iscsi_obj->get_nodename(-use_cache => 0);
        if ($iscsi_obj->error()) {
            $self->_set_err("$log_label: Failed to get iscsi node name",
                            $iscsi_obj->errmsg());
            return $SanUtil::FALSE;
        }

        $self->sanlog(-msg => "Processing $log_label: nodename=$node_name",
                      -msg_level => 1);

        # If the user set force to 1, then we will call isns update.
        if ($force_isns_update) {

            # Call isns update.
            $iscsi_obj->isns_update();
            if ($iscsi_obj->error()) {
                $self->_set_err(
                          "$log_label: isns_update failed, nodename=$node_name",
                          $iscsi_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # Sleep for 16 seconds.  In C-mode, it can take up to 15 seconds
            # after you peform isns update for the update to actually occur.
            &SanUtil::nate_sleep(16);

        } ## end if ($force_isns_update)

        # Looping '$retry_count + 1' times, so there is an initial attempt
        # before start counting retries.
        for (my $i = 0 ; $i < $retry_count + 1 ; $i++) {

            # Get vserver's iSNS registration information from the iSNS server
            # by calling get_isns_info_cmode.
            $isns_hash =
              $isns_server_obj->get_isns_info_cmode(-nodename => $node_name);
            if ($isns_server_obj->error()) {
                $self->_set_err(
                              "$log_label: get_isns_info_cmode failed: "
                                . "iSNS server=$isns_name, nodename=$node_name",
                              $isns_server_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # If the node name is not registered with iSNS server.
            if (scalar(keys(%{$isns_hash})) == 0) {

                $self->sanlog(
                     -msg =>
                       "$log_label: Not registered with iSNS server $isns_name,"
                       . " nodename=$node_name",
                     -msg_level => 1);

                # If we are NOT on the final attempt, sleep for a minute  before
                # trying again.
                if ($i != $retry_count) {
                    $self->sanlog(-msg       => "Waiting for one minute...",
                                  -msg_level => 1);
                    &SanUtil::nate_sleep(60);

                    # Attempt next retry.
                    next;
                } else {

                    # Else this was the last retry.  All retries failed,
                    # so return error.
                    $self->_set_err(
                            "$log_label: verify registration with iSNS Server "
                          . "$isns_name FAILED after $retry_count retries! , nodename=$node_name"
                    );
                    return $SanUtil::FALSE;
                } ## end else [ if ($i != $retry_count)]
            } ## end if (scalar(keys(%{$isns_hash...})))

            # Call _verify_isns to compare information from iSNS server with
            # vserver's actual information.
            $result =
              $self->_verify_isns(-iscsi_obj => $iscsi_obj,
                                  -isns_hash => $isns_hash);
            if ($self->error()) {
                $self->_set_err(
                              "$log_label: _verify_isns failed: "
                                . "iSNS server=$isns_name, nodename=$node_name",
                              $self->errmsg());
                return $SanUtil::FALSE;
            }

            if ($result == $SanUtil::TRUE) {

                # If _verify_isns returns true, it means all of the
                # data matched. So we can break out of the retry loop and
                # process the next vserver.
                $self->sanlog(-msg => "$log_label: iSNS registration success"
                                . " after $i retries! "
                                . "iSNS server=$isns_name, nodename=$node_name",
                              -msg_level => 1);
                last;
            } else {

                # Else _verify_isns must return false, it means some of
                # the data is stale.
                $self->sanlog(
                    -msg => "$log_label: is registered with iSNS Server "
                      . "$isns_name, but some of the data is stale. nodename=$node_name",
                    -msg_level => 1);

                # If we are NOT on the final attempt, sleep for a minute  before
                # trying again.
                if ($i != $retry_count) {
                    $self->sanlog(-msg       => "Waiting for one minute...",
                                  -msg_level => 1);
                    &SanUtil::nate_sleep(60);
                } else {

                    # Else this was the last retry.  All retries failed,
                    # so return error.
                    $self->_set_err(
                            "$log_label: verify registration with iSNS Server "
                          . "$isns_name FAILED after $retry_count retries! , nodename=$node_name"
                    );
                    return $SanUtil::FALSE;
                } ## end else [ if ($i != $retry_count)]
            } ## end else [ if ($result == $SanUtil::TRUE)]

        } ## end for (my $i = 0 ; $i < $retry_count...)

    } ## end foreach my $vserver_obj (@vserver_objs)

    # If we get here, no vservers failed & returned FALSE.  They all passed.
    $self->sanlog(-msg       => "SUCCESS",
                  -msg_level => 1);
    return $SanUtil::TRUE;

} ## end sub verify_isns_registration

###############################################################################
# Method:    verify_isns_deregistration
# Objective: Verifies that a vserver is not registered with iSNS server.
# Details:   See POD documentation at the end of this file.
# Inputs   : -vserver_objs - optional - Array of vserver obj to verify.  Default
#                is to verify all vservers in vserver_info.
#            -isns_server_obj - required - isns server object
#            -retries - optional - Number of times to retry if vserver is still
#                registered. (Default is 0)
# Returns  : $SanUtil::TRUE - Successful execution, all vservers verified
#            $SanUtil::FALSE & set_err - Execution failed
###############################################################################
sub verify_isns_deregistration {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $retry_count = 0;    # Default value for retries.
    my (@vserver_objs, $isns_server_obj, $node_name, $isns_hash, $result);

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_objs/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-vserver_objs', value must be a reference to an 'ARRAY'"
                );
                return $SanUtil::FALSE;
            }
            @vserver_objs = @{$value};
        } elsif ($key =~ /isns_server_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $isns_server_obj = $value;
        } elsif ($key =~ /retries/) {
            unless (length($value)) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $retry_count = $value;
            if ($retry_count !~ /^\d+$/) {
                $self->_set_err(
                          "Invalid value for -retries, value must be a number");
                return $SanUtil::FALSE;
            }
        } ## end elsif ($key =~ /retries/)
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($isns_server_obj) =~ /^San::SanClient::IsnsServer/i) {
        $self->_set_err(
            "-isns_server_obj is a required parameter, must be an iSNS Server object"
        );
        return $SanUtil::FALSE;
    }

    # debug info
    $self->sanlog(-msg       => "retries: $retry_count",
                  -msg_level => 2);

    # Save isns name for logging.
    my $isns_name = $isns_server_obj->hostname();
    if ($isns_server_obj->error()) {
        $self->_set_err("Failed to get hostname from isns server obj",
                        $isns_server_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # If the user passed in vserver_objs, let's verify they are all vserver
    # objects.
    if (scalar(@vserver_objs)) {

        # Scan each vserver object.
        foreach my $vserver (@vserver_objs) {

            # Verify it's a vserver
            unless (
                ref($vserver) =~ /^San::ControllerCluster::Filer10M::Vserver$/i)
            {
                $self->_set_err("-vserver_objs must be vserver objects");
                return $SanUtil::FALSE;
            }
        } ## end foreach my $vserver (@vserver_objs)
    } else {

        # Else user did not pass in a vserver object.  We will process all
        # vserver objects in vserver_info.
        my @vserver_info = $self->get_vserver_info();
        if ($self->error()) {
            $self->_set_err("Failed to get_vserver_info", $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Verify we have at least 1 vserver to process.
        if (scalar(@vserver_info) < 1) {
            $self->_set_err("vserver_info has no vservers.");
            return $SanUtil::FALSE;
        }

        # Otherwise, we have some vservers.  We need to push the objects
        # on to our vserver_objs array.
        foreach my $vserver (@vserver_info) {
            push(@vserver_objs, $vserver->{object});
        }
    } ## end else [ if (scalar(@vserver_objs...))]

    # Now we have a list of vservers to verify.  Scan each vserver.
    foreach my $vserver_obj (@vserver_objs) {

        # Get vserver name
        my $vserver_name = $vserver_obj->get_name();
        if ($vserver_obj->error()) {
            $self->_set_err("get_name failed for vserver",
                            $vserver_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Setup log label for future messages.
        my $log_label = "Vserver: $vserver_name";

        # Now use the vserver obj to get the iscsi object
        my $iscsi_obj = $vserver_obj->get_iscsi_obj();
        if ($vserver_obj->error()) {
            $self->_set_err("$log_label: get_iscsi_obj failed",
                            $vserver_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Verify iscsi object was returned. (could be empty if no iscsi)
        if (ref($iscsi_obj) !~ /iscsi/i) {
            $self->_set_err(
                          "$log_label: get_iscsi_obj did not return iscsi obj");
            return $SanUtil::FALSE;
        }

        # Get the nodename for the vserver.
        $node_name = $iscsi_obj->get_nodename(-use_cache => 0);
        if ($iscsi_obj->error()) {
            $self->_set_err("$log_label: Failed to get iscsi node name",
                            $iscsi_obj->errmsg());
            return $SanUtil::FALSE;
        }

        $self->sanlog(-msg => "Processing $log_label: nodename=$node_name",
                      -msg_level => 1);

        # Loop for checking deregistration.
        # Looping 'retry_count + 1' times, so there is an initial attempt
        # before start counting retries.
        for (my $i = 0 ; $i < $retry_count + 1 ; $i++) {

            # Get isns information for the node name.
            $isns_hash =
              $isns_server_obj->get_isns_info_cmode(-nodename     => $node_name,
                                                    -deregistered => 1);
            if ($isns_server_obj->error) {
                $self->_set_err(
                              "$log_label: get_isns_info_cmode failed: "
                                . "iSNS server=$isns_name, nodename=$node_name",
                              $isns_server_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # if the node name is no longer registered with isns server.
            if (scalar(keys(%$isns_hash)) == 0) {

                $self->sanlog(
                        -msg =>
                          "$log_label: deregistered from isns server $isns_name"
                          . " after $i retries. nodename=$node_name",
                        -msg_level => 1);

                # Break out of retry loop to try next vserver.
                last;
            } else {

                # Else still registered
                $self->sanlog(
                    -msg =>
                      "$log_label: still registered with isns server $isns_name."
                      . "nodename=$node_name",
                    -msg_level => 1);

                # If we are NOT on the final attempt, sleep for a minute  before
                # trying again.
                if ($i != $retry_count) {
                    $self->sanlog(-msg       => "Waiting for one minute...",
                                  -msg_level => 1);
                    &SanUtil::nate_sleep(60);
                } else {

                    # Else this was the last retry.  All retries failed,
                    # so return error.
                    $self->_set_err(
                        "$log_label: verify deregistration with iSNS Server "
                          . "$isns_name FAILED after $retry_count retries! , nodename=$node_name"
                    );
                    return $SanUtil::FALSE;
                } ## end else [ if ($i != $retry_count)]
            } ## end else [ if (scalar(keys(%$isns_hash...)))]
        } ## end for (my $i = 0 ; $i < $retry_count...)

    } ## end foreach my $vserver_obj (@vserver_objs)

    # If we get here, no vservers failed & returned FALSE.  They all passed.
    $self->sanlog(-msg       => "SUCCESS",
                  -msg_level => 1);
    return $SanUtil::TRUE;
}    # End sub verify_isns_deregistration

################################################################################
# Method:    sleep_and_monitor_health
# Objective: Sleeps for the specified number of minutes.  Every 15 secounds,
#            performs cluster show and verifies health is true for each node.
# Details:
#     Inputs - sanmgr_obj (required) - sanmgr object
#            - minutes (optional) - total number of minutes to sleep
#                   (default is 3)
#     Returns $SanUtil::TRUE  - all nodes reported healthy the entire time
#             $SanUtil::FALSE - at least one node was not healthy
################################################################################
sub sleep_and_monitor_health {
    my ($self, %args) = @_;

    my $sanmgr_obj;
    my $minutes = 3;    # Number of minutes to sleep

    $self->sanlog(-msg       => "START",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /sanmgr_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $sanmgr_obj = $value;
        } elsif ($key =~ /minutes/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $minutes = $value;
            if ($minutes !~ /^\d+$/) {
                $self->_set_err(
                          "Invalid value for -minutes, value must be a number");
                return $SanUtil::FALSE;
            }
        } ## end elsif ($key =~ /minutes/)
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(
                "-sanmgr_obj is a required parameter, must be a SanMgr object");
        return $SanUtil::FALSE;
    }

    # Get the cluster object
    my @cluster_objs = $sanmgr_obj->get_objects(type => "cluster");
    if ($sanmgr_obj->error()) {
        $self->_set_err("Failed to get cluster objects", $sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Verify we got back some cluster objects.
    if (!scalar(@cluster_objs)) {
        $self->_set_err("get_objects(cluster) returned 0 objects");
        return $SanUtil::FALSE;
    }

    # We want to check health every 15 seconds, so multiply the number of minutes
    # times 4 to get a loop count.
    my $loop_count = $minutes * 4;

    foreach (1 .. $loop_count) {

        # sleep for 15 seconds
        &SanUtil::nate_sleep(15);

        # Run cluster show command.
        my @cluster_show = $cluster_objs[0]->get_state();
        if ($cluster_objs[0]->error()) {
            $self->_set_err("get_state failed", $cluster_objs[0]->errmsg());
            return $SanUtil::FALSE;
        }

        # Verify we got back some elements from cluster show.
        if (!scalar(@cluster_show)) {
            $self->_set_err("get_state returned 0 objects");
            return $SanUtil::FALSE;
        }

        # Scan each node returned and verify health is true.
        foreach my $node (@cluster_show) {
            if ($node->{health} !~ /true/i) {
                $self->_set_err(
                    "All cluster nodes are not healthy! Check for possible panic!"
                );
                return $SanUtil::FALSE;
            }

        }
    } ## end foreach (1 .. $loop_count)

    # If code reaches here, everything passed.
    $self->sanlog(-msg       => "SUCCESS!",
                  -msg_level => 1);
    return $SanUtil::TRUE;

} ## end sub sleep_and_monitor_health

################################################################################
# verify_jumbo_frames
# Objective: This function verifies that all clients in Client_Info can ping
#            each lif in each vserver in Vserver_Info.  The pings are executed
#            with a large packet size (8500) and the flag set for do NOT
#            fragment.  If the pings are successful, then jumbo frames are
#            working from client to target.
# Details:
#     Inputs -skip_unsupported_clients: 1 = skip, 0 = raise error(default)
#     Returns On Success, $SanUtil::TRUE  - all objects created successfully
#             On error, $SanUtil::FALSE - object failed and set_err
################################################################################
sub verify_jumbo_frames {
    my ($self, %args) = @_;
    my ($status, @output);
    my $skip_unsupported_clients = 0;

    $self->sanlog(-msg       => "START",
                  -msg_level => 1);

    $self->_clear_err();

    # Parse arguments
    foreach (keys %args) {
        if (/-?skip_unsupported_clients/i) {
            $skip_unsupported_clients = $args{$_};
            if (not length($skip_unsupported_clients)
                or $skip_unsupported_clients !~ /^(1|0)$/i) {
                $self->_set_err(
                    "Invalid value for argument '-skip_unsupported_clients', must be 0 or 1"
                );
                return $SanUtil::FALSE;
            }
        } ## end if (/-?skip_unsupported_clients/i)
    }    # end foreach arg

    my @client_info  = $self->get_client_info();
    my @vserver_info = $self->get_vserver_info();

    # Verify we have at least 1 vserver and 1 client, or there is nothing to
    # check.
    if ((scalar(@client_info) < 1) or (scalar(@vserver_info) < 1)) {
        $self->_set_err("Must have at least 1 vserver and 1 client, otherwise "
                        . "the method cannot verify jumbo frames.");
        return $SanUtil::FALSE;
    }

    # Need to verify each client can ping all of the lifs using jumbo frames.
    # Scan each vserver
    foreach my $vserver (@vserver_info) {

        # Grab the iscsi lifs for this vserver
        my @lifs = $vserver->{object}->lif_list_info(-data_protocol => "iscsi");
        if ($vserver->{object}->error()) {
            $self->_set_err(
                          "Failed to get lifs from vserver " . $vserver->{name},
                          $vserver->{object}->errmsg());
            return $SanUtil::FALSE;
        }

        # If there are 0 lifs on the vserver, let's raise an error because
        # we can't verify this vserver.
        if (scalar(@lifs) == 0) {
            $self->_set_err("0 lifs found on vserver " . $vserver->{name});
            return $SanUtil::FALSE;
        }

        # Scan each lif on the vserver
        foreach my $lif (@lifs) {

            # Ping from each client
            foreach my $client (@client_info) {

                my $cmd;

                if ($client->{object}->hosttype() =~ /windows/i) {

                    # check if client is using ipv6 or ipv4 and use the appropriate
                    # ping command
                    if ($client->{object}->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~
                        /ipv6/i) {

                        # -l = set packet size.  -6 = use ipv6
                        $cmd = "ping -l 8500 -6 " . $lif->{address};
                    } else {

                        # -l = set packet size.  -f = do not fragment (ipv4 only)
                        $cmd = "ping -l 8500 -f " . $lif->{address};
                    }

                    # Example: (note the first ping often fails, this is ok)
                    # C:\Users\Administrator.ssan-w2k8-64-30>ping -l 8500 -f 10.228.128.125
                    #
                    # Pinging 10.228.128.125 with 8500 bytes of data:
                    # Request timed out.
                    # Reply from 10.228.128.125: bytes=8500 time=1ms TTL=252
                    # Reply from 10.228.128.125: bytes=8500 time=1ms TTL=252
                    # Reply from 10.228.128.125: bytes=8500 time=1ms TTL=252
                    #
                    # Ping statistics for 10.228.128.125:
                    #     Packets: Sent = 4, Received = 3, Lost = 1 (25% loss),
                    # Approximate round trip times in milli-seconds:
                    #     Minimum = 1ms, Maximum = 1ms, Average = 1ms

                    # execute the command
                    ($status, @output) =
                      &ClientUtil::execute($client->{object}->{conn}, $cmd);
                    if ($status != 0 or not grep(/Reply from/i, @output)) {
                        $self->_set_err(
                                "Client: " . $client->{name} . " $cmd failed!");
                        return $SanUtil::FALSE;
                    }

                } elsif ($client->{object}->hosttype() =~ /linux/i) {

                    # check if client is using ipv6 or ipv4 and use the appropriate
                    # ping command
                    if ($client->{object}->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~
                        /ipv6/i) {

                        # -c = count. -s = packet size.  -M do = do not fragment.
                        $cmd = "ping6 -c 5 -s 8500 -M do " . $lif->{address};
                    } else {

                        # -c = count. -s = packet size.  -M do = do not fragment.
                        $cmd = "ping -c 5 -s 8500 -M do " . $lif->{address};
                    }

                    # Example
                    # [root@ssan-rhel5u4-20 ~]# ping -c 5 -s 8500 -M do 10.228.128.125
                    # PING 10.228.128.125 (10.228.128.125) 8500(8528) bytes of data.
                    # 8508 bytes from 10.228.128.125: icmp_seq=1 ttl=252 time=0.913 ms
                    # 8508 bytes from 10.228.128.125: icmp_seq=2 ttl=252 time=1.03 ms
                    # 8508 bytes from 10.228.128.125: icmp_seq=3 ttl=252 time=0.989 ms
                    # 8508 bytes from 10.228.128.125: icmp_seq=4 ttl=252 time=1.06 ms
                    # 8508 bytes from 10.228.128.125: icmp_seq=5 ttl=252 time=2.05 ms
                    #
                    # --- 10.228.128.125 ping statistics ---
                    # 5 packets transmitted, 5 received, 0% packet loss, time 4000ms
                    # rtt min/avg/max/mdev = 0.913/1.210/2.052/0.425 ms

                    # execute the command
                    ($status, @output) =
                      &ClientUtil::execute($client->{object}->{conn}, $cmd);
                    if ($status != 0
                        or not grep(/bytes from/i, @output)) {
                        $self->_set_err(
                                "Client: " . $client->{name} . " $cmd failed!");
                        return $SanUtil::FALSE;
                    }

                } elsif (!$skip_unsupported_clients) {

                    $self->_set_err(
                          "Unsupported client detected.  This method currently "
                            . "supports windows and linux only.");
                    return $SanUtil::FALSE;
                }

            } ## end foreach my $client (@client_info)
        } ## end foreach my $lif (@lifs)
    } ## end foreach my $vserver (@vserver_info)

    $self->sanlog(-msg       => "Jumbo frame connectivity verified.",
                  -msg_level => 1);

    return $SanUtil::TRUE;

} ## end sub verify_jumbo_frames

################################################################################
#                               Private Methods
################################################################################

################################################################################
# _initialize_ioman
# Objective: It will initialize IOMan library object which is useful for
#            running IO.
# Details:
#     Inputs - none
#     Returns On Success, $SanUtil::TRUE  - IOMan object created successfully
#             On error, $SanUtil::FALSE - object create failed and set_err
################################################################################
sub _initialize_ioman {

    my ($self) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    if (ref($self->{_ioman_obj}) =~ /IOMan/) {
        $self->sanlog(-msg => "IOMan object already instantiated: "
                        . ref($self->{_ioman_obj}),
                      -msg_level => 1);

        $self->sanlog(-msg => "Success", -msg_level => 1);
        return $SanUtil::TRUE;
    }

    my ($ioman_obj);

    # instantiate an IOMan object
    $ioman_obj = new IOMan();
    if ($ioman_obj->error()) {
        $self->_set_err("Failed to create IOMan obj ", $ioman_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $self->{_ioman_obj} = $ioman_obj;

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _initialize_ioman

################################################################################
# _initialize_eventman
# Objective: It will initialize EventMan library object which is useful for
#            running events (SFOs, reboots, service stop/starts etc).
# Details:
#     Inputs - none
#     Returns On Success, $SanUtil::TRUE  - EventMan object created successfully
#             On error, $SanUtil::FALSE - object create failed and set_err
################################################################################
sub _initialize_eventman {

    my ($self) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    if (ref($self->{_eventman_obj}) =~ /EventMan/) {
        $self->sanlog(-msg => "EventMan object already instantiated: "
                        . ref($self->{_eventman_obj}),
                      -msg_level => 1);

        $self->sanlog(-msg => "Success", -msg_level => 1);
        return $SanUtil::TRUE;
    }

    my ($eventman_obj);

    # instantiate an EventMan object
    $eventman_obj = EventMan->new(-sram_detection => 0);
    if ($eventman_obj->error()) {
        $self->_set_err("Failed to create EventMan obj ",
                        $eventman_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $self->{_eventman_obj} = $eventman_obj;

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _initialize_eventman

###############################################################################
# Method:    _verify_iscsi_interfaces_from_client
# Objective: Verifies target iscsi interface info collected from
#            SanClient::get_iscsi_info agrees with Data ONTAP.
# Details:
#     Inputs  -iscsi_obj(required)
#             -interfaces(required)    - array from get_iscsi_info
#             -initiator_iqn(required if isns=0) - initiator's iqn
#             -portal(optional) - 1 = use portal command instead of interface
#                               - 0 = default, use interface cmd
#             -isns(optional)   - 1 = isns discovery used, ignore acl
#                               - 0 = sendtargets used, check acls (default)
#     Returns $SanUtil::TRUE  - interfaces array verified
#             $SanUtil::FALSE - interfaces array not verified and set_err
################################################################################
sub _verify_iscsi_interfaces_from_client {
    my ($self, %args) = @_;

    my $iscsi_obj;
    my $interfaces;       # iscsi interface array to verify (as seen by client)
    my $initiator_iqn;    # initiator's iqn
    my $verify_acl;       # flag if we need to verify accesslists
    my $use_portal = 0;       # indicator to use portal command
    my @target_interfaces;    # saves interfaces from target
    my $cmd  = "interface";   # used to indicate which command we are using
    my $isns = 0;             # Flag to indicate isns discovery was used
    my @acls;                 # Array to hold accesslists

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /iscsi_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $iscsi_obj = $value;
        } elsif ($key =~ /interfaces/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $interfaces = $value;
        } elsif ($key =~ /initiator_iqn/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $initiator_iqn = $value;
        } elsif ($key =~ /portal/) {
            unless ($value =~ /^(1|0)$/) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $use_portal = $value;
        } elsif ($key =~ /isns/) {
            unless ($value =~ /^(1|0)$/) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $isns = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($iscsi_obj) =~
            /^San::ControllerCluster::Filer10M::Vserver::Iscsi$/i) {
        $self->_set_err(
                 "-iscsi_obj is a required parameter, must be an iSCSI object");
        return $SanUtil::FALSE;
    }
    unless (scalar @{$interfaces}) {
        $self->_set_err("-interfaces is a required parameter.");
        return $SanUtil::FALSE;
    }
    if (!$isns and not length($initiator_iqn)) {
        $self->_set_err("-initiator_iqn is a required parameter if isns=0");
        return $SanUtil::FALSE;
    }

    # debug info
    $self->sanlog(-msg       => "Arg: interfaces:\n" . Dumper($interfaces),
                  -msg_level => 3);

    $self->sanlog(-msg       => "Arg: initiator_iqn: $initiator_iqn",
                  -msg_level => 3);

    $self->sanlog(-msg       => "Arg: portal: $use_portal",
                  -msg_level => 1);

    $self->sanlog(-msg       => "Arg: isns: $isns",
                  -msg_level => 1);

    # If we are using portal, udpate the cmd
    if ($use_portal) {
        $cmd = "portal";
    }

    # Determine number of interfaces client sees
    my $client_interfaces = scalar @{$interfaces};

    # Grab the target's view of interfaces/portals.  Use proper command.
    # Only get interfaces that are enabled.  Disabled interfaces are not sent
    # to the client.
    if ($use_portal) {

        @target_interfaces = $iscsi_obj->portal_list_info(-enabled => "true");
        if ($iscsi_obj->error()) {
            $self->_set_err("Failed: portal_list_info", $iscsi_obj->errmsg());
            return $SanUtil::FALSE;
        }
    } else {

        @target_interfaces =
          $iscsi_obj->interface_list_info(-enabled => "true");
        if ($iscsi_obj->error()) {
            $self->_set_err("Failed: interface_list_info",
                            $iscsi_obj->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end else [ if ($use_portal) ]

    # Debug info
    $self->sanlog(
         -msg => "iscsi $cmd" . "s from target:\n" . Dumper(@target_interfaces),
         -msg_level => 3);

    # Check for discovery type.  If discovery was NOT isns (sendtargets),
    # then we need to consider acls.  If discovery was iSNS, then we skip acl
    # considerations because iSNS does not know about accesslists.
    if (!$isns) {

        # Check the number of acls for this initiator.
        @acls = $iscsi_obj->interface_accesslist_list_info(
                                            'initiator-name' => $initiator_iqn);
        if ($iscsi_obj->error()) {
            $self->_set_err("Failed: interface_accesslist_list_info",
                            $iscsi_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Debug info
        $self->sanlog(
                     -msg => "iscsi accesslists from target:\n" . Dumper(@acls),
                     -msg_level => 3);

    } ## end if (!$isns)

    # If we have acls, then the target_interfaces array is not valid.
    # We will scan the acls and construct a new targets_interfaces array
    # that represents what the client should see.
    if (scalar(@acls)) {

        my @new_target_interfaces;

        # Turn on verify flag for later
        $verify_acl = 1;

        # Scan each acl
        foreach my $acl (@acls) {

            # Take the lif name from each acl.  Search for a portal/interface
            # with this lif name that is enabled.
            my @int;
            if ($use_portal) {

                @int =
                  $iscsi_obj->portal_list_info(-enabled => "true",
                                               -lif     => $acl->{lif});
                if ($iscsi_obj->error()) {
                    $self->_set_err("Failed: portal_list_info",
                                    $iscsi_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            } else {

                @int =
                  $iscsi_obj->interface_list_info(-enabled => "true",
                                                  -lif     => $acl->{lif});
                if ($iscsi_obj->error()) {
                    $self->_set_err("Failed: interface_list_info",
                                    $iscsi_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            } ## end else [ if ($use_portal) ]

            # If we found more than 1 interface, that is bad.
            if (scalar(@int) > 1) {
                $self->_set_err("iscsi $cmd verification failed while calling "
                                . "$cmd show with arguments:\n"
                                . "-enabled => true\n"
                                . "-lif => "
                                . $acl->{lif}
                                . "\nFound "
                                . scalar(@int) . " $cmd"
                                . "s, expected 0 or 1.");
                return $SanUtil::FALSE;

            } ## end if (scalar(@int) > 1)

            # If we found an interface that was enabled and in the acl, then
            # save it to the new array.  This lif is enabled and has an acl, so
            # it should be sent to the client.
            if (scalar(@int)) {
                push(@new_target_interfaces, $int[0]);
            }
        } ## end foreach my $acl (@acls)

        # Update the target_interfaces with the new array.  Only used to
        # verify the total count the client should see.
        @target_interfaces = @new_target_interfaces;

        # Debug info
        $self->sanlog(
                 -msg =>
                   "Used accesslists to generate new target_interfaces array:\n"
                   . Dumper(@target_interfaces),
                 -msg_level => 3);

    } ## end if (scalar(@acls))

    # Make sure filer/client agree on the number of interfaces/portals
    # NOTE: the 'mismatch' text in the message is used by _verify_isns.
    # DO NOT CHANGE!
    if ($client_interfaces != scalar(@target_interfaces)) {
        $self->_set_err(  "iscsi $cmd count mismatch!\n"
                        . "Client says there are $client_interfaces $cmd"
                        . "s.\nTarget says there are "
                        . scalar(@target_interfaces) . " $cmd"
                        . "s.");
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "$cmd count verified: $client_interfaces $cmd" . "s",
                  -msg_level => 1);

    # For each interface the client sees
    foreach my $client_int (@{$interfaces}) {

        my @ints_found;

        # Use proper command - portal or interface
        if ($use_portal) {

            # Try and find portal that matches the exact parms of the client.
            @ints_found = $iscsi_obj->portal_list_info(%{$client_int});
            if ($iscsi_obj->error()) {
                $self->_set_err("Failed: portal_list_info",
                                $iscsi_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {

            # Try and find interface that matches the exact parms of the client.
            @ints_found = $iscsi_obj->interface_list_info(%{$client_int});
            if ($iscsi_obj->error()) {
                $self->_set_err("Failed: interface_list_info",
                                $iscsi_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } ## end else [ if ($use_portal) ]

        # We better find 1 interface that is an exact match.
        # NOTE: the 'mismatch' text in the message is used by _verify_isns.
        # DO NOT CHANGE!
        if (scalar(@ints_found) != 1) {
            $self->_set_err(  "Verify $cmd"
                            . "s mismatch!  Filter was expected to match 1 "
                            . "$cmd, but matched "
                            . scalar(@ints_found)
                            . "\nFilter used:\n"
                            . Dumper($client_int));
            return $SanUtil::FALSE;
        }

        # If verify_acl is on, check acl as well.
        if ($verify_acl) {

            # Now we can check the accesslist using the initiator iqn & lif name.
            # Note the lif name is pulled from the interface/portal output.
            my @acl_output =
              $iscsi_obj->interface_accesslist_list_info(
                                             'initiator-name' => $initiator_iqn,
                                             -lif => $ints_found[0]->{lif});
            if ($iscsi_obj->error()) {
                $self->_set_err("Failed: interface_accesslist_list_info",
                                $iscsi_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # Make sure we have only one match.
            if (scalar(@acl_output) != 1) {
                $self->_set_err("iscsi $cmd verification failed!\n"
                          . "iscsi interface accesslist mismatch!\n"
                          . "Expected to find 1 match on the target, but found "
                          . scalar(@acl_output)
                          . " matches.\n"
                          . "Filter used:\ninitiator-name = $initiator_iqn\n"
                          . "lif = "
                          . $ints_found[0]->{lif});
                return $SanUtil::FALSE;
            } ## end if (scalar(@acl_output...))

        } ## end if ($verify_acl)

    } ## end foreach my $client_int (@{$interfaces...})

    $self->sanlog(-msg       => "iscsi $cmd" . "s verified",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _verify_iscsi_interfaces_from_client

###############################################################################
# Method:    _verify_iscsi_tpgroups_from_client
# Objective: Verifies target iscsi tpgroup info collected from
#            SanClient::get_iscsi_info agrees with Data ONTAP.
# Details:
#     Inputs  -iscsi_obj(required)
#             -tpgroups(required)      - array from get_iscsi_info
#             -initiator_iqn(required if isns=0) - initiator's iqn
#             -isns(optional)   - 1 = isns discovery used, ignore acl
#                               - 0 = sendtargets used, check acls (default)
#     Returns $SanUtil::TRUE  - tpgroups array verified
#             $SanUtil::FALSE - tpgroups array not verified and set_err
################################################################################
sub _verify_iscsi_tpgroups_from_client {
    my ($self, %args) = @_;

    my $iscsi_obj;
    my $tpgroups;         # iscsi tpgroup array to verify (as seen by client)
    my $initiator_iqn;    # initiator's iqn
    my $verify_acl;       # flag if we need to verify accesslists
    my $isns = 0;         # Flag to indicate isns discovery was used
    my @acls;             # Array to hold accesslists

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /iscsi_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $iscsi_obj = $value;
        } elsif ($key =~ /tpgroups/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $tpgroups = $value;
        } elsif ($key =~ /initiator_iqn/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $initiator_iqn = $value;
        } elsif ($key =~ /isns/) {
            unless ($value =~ /^(1|0)$/) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $isns = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($iscsi_obj) =~
            /^San::ControllerCluster::Filer10M::Vserver::Iscsi$/i) {
        $self->_set_err(
                 "-iscsi_obj is a required parameter, must be an iSCSI object");
        return $SanUtil::FALSE;
    }
    unless (scalar @{$tpgroups}) {
        $self->_set_err("-tpgroups is a required parameter.");
        return $SanUtil::FALSE;
    }
    if (!$isns and not length($initiator_iqn)) {
        $self->_set_err("-initiator_iqn is a required parameter if isns=0");
        return $SanUtil::FALSE;
    }

    # debug info
    $self->sanlog(-msg       => "Arg: tpgroups:\n" . Dumper($tpgroups),
                  -msg_level => 3);

    $self->sanlog(-msg       => "Arg: initiator_iqn: $initiator_iqn",
                  -msg_level => 3);

    $self->sanlog(-msg       => "Arg: isns: $isns",
                  -msg_level => 1);

    # Grab the vserver object so we can translate ips to lif names.
    my $vserver = $iscsi_obj->get_parent_obj();
    if ($iscsi_obj->error()) {
        $self->_set_err("Failed: get_parent_obj", $iscsi_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Check for discovery type.  If discovery was NOT isns (sendtargets),
    # then we need to consider acls.  If discovery was iSNS, then we skip acl
    # considerations because iSNS does not know about accesslists.
    if (!$isns) {

        # Check the number of acls for this initiator.
        @acls = $iscsi_obj->interface_accesslist_list_info(
                                            'initiator-name' => $initiator_iqn);
        if ($iscsi_obj->error()) {
            $self->_set_err("Failed: interface_accesslist_list_info",
                            $iscsi_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Debug info
        $self->sanlog(
                     -msg => "iscsi accesslists from target:\n" . Dumper(@acls),
                     -msg_level => 3);

        # If acls exist, turn on the verify_acl flag.
        if (scalar(@acls)) {
            $verify_acl = 1;
        }
    } ## end if (!$isns)

    # Determine number of tpgroups client sees
    my $client_tpgroups = scalar @{$tpgroups};

    # Grab target's view of tpgroups.  Search for all tpgroups that have a lif.
    # So we filter on lif does not equal "-".
    my @target_tpgroups = $iscsi_obj->tpgroup_list_info(-lif => "!\"-\"");
    if ($iscsi_obj->error()) {
        $self->_set_err("Failed: tpgroup_list_info", $iscsi_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Debug info
    $self->sanlog(
             -msg => "iscsi tpgroups from target:\n" . Dumper(@target_tpgroups),
             -msg_level => 3);

    # We need to do some investigation to make sure our tpgroup count is
    # accurate.  ACLs and disabled iscsi interfaces can impact which tpgroups
    # the client sees.

    # A temp hash to keep track of which tpgroups the client should see
    my $new_target_tpgroups;

    # Scan each tpgroup that has a lif
    foreach my $tpgroup (@target_tpgroups) {

        # Save the tpgroup's lifs into an array
        my @lifs = split(",", $tpgroup->{lif});

        # Scan each lif of the tpgroup
        foreach my $lif (@lifs) {

            # Check if the iscsi interface of the lif is enabled
            my @ints =
              $iscsi_obj->interface_list_info(-enabled => "true",
                                              -lif     => $lif);
            if ($iscsi_obj->error()) {
                $self->_set_err("Failed: interface_list_info",
                                $iscsi_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # If we found more than 1 interface, that is bad.
            if (scalar(@ints) > 1) {
                $self->_set_err(
                              "iscsi tpgroup verification failed while calling "
                                . "iscsi interface show with arguments:\n"
                                . "-enabled => true\n"
                                . "-lif => $lif\n"
                                . "Found "
                                . scalar(@ints)
                                . " interfaces, expected 0 or 1.");
                return $SanUtil::FALSE;

            } elsif (scalar(@ints) == 0) {

                # If lif was not enabled, goto the next lif in the for loop
                next;
            }

            # Else we found 1 lif with the interface enabled.
            # If acls are present, we also need to check for acls.
            if ($verify_acl) {

                # See if the initiator has an acl for the lif in question
                @acls =
                  $iscsi_obj->interface_accesslist_list_info(
                                             'initiator-name' => $initiator_iqn,
                                             -lif             => $lif);
                if ($iscsi_obj->error()) {
                    $self->_set_err("Failed: interface_accesslist_list_info",
                                    $iscsi_obj->errmsg());
                    return $SanUtil::FALSE;
                }

                # If we found more than 1 accesslist, that is bad.
                if (scalar(@acls) > 1) {
                    $self->_set_err(
                              "iscsi tpgroup verification failed while calling "
                                . "interface accesslist show with arguments:\n"
                                . "-initiator-name => $initiator_iqn\n"
                                . "-lif => $lif\n"
                                . "Found "
                                . scalar(@acls)
                                . " accesslists, expected 0 or 1.");
                    return $SanUtil::FALSE;

                } elsif (scalar(@acls) == 0) {

                    # If we found zero, goto the next lif in the for loop
                    next;
                }

            } ## end if ($verify_acl)

            # If the code reaches here, it means we have a lif whose
            # iscsi interface is enabled.  It also means the acl was verified
            # if the initiator had acls.  So we need to save this tpgroup's
            # tag into our hash.
            # If the tpgroup tag is not currently a key in our hash, add it.
            # Keep in mind the only purpose of this hash to determine an
            # accurate count for how many tpgroups the client should see.
            if (!exists($new_target_tpgroups->{$tpgroup->{'tpgroup_tag'}})) {
                $new_target_tpgroups->{$tpgroup->{'tpgroup_tag'}} = 1;
                $self->sanlog(-msg => "saving tpgroup "
                                . $tpgroup->{'tpgroup_tag'}
                                . " into the hash.",
                              -msg_level => 3);
            }

            # We have recorded this tpgroup.  No need to process any more lifs
            # in this tpgroup.  Stop the foreach $lif for loop and goto the next
            # tpgroup.
            last;

        } ## end foreach my $lif (@lifs)
    } ## end foreach my $tpgroup (@target_tpgroups)

    # Debug info
    $self->sanlog(-msg => "List of tpgroup tags the client should see:\n"
                    . Dumper($new_target_tpgroups),
                  -msg_level => 3);

    # Make sure filer/client agree on the number of tpgroups
    # NOTE: the 'mismatch' text in the message is used by _verify_isns.
    # DO NOT CHANGE!
    if ($client_tpgroups != scalar keys %{$new_target_tpgroups}) {
        $self->_set_err(  "iscsi tpgroup count mismatch!\n"
                        . "Client says there are $client_tpgroups tpgroups.\n"
                        . "Target says there are "
                        . scalar(keys %{$new_target_tpgroups})
                        . " tpgroups.");
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Tpgroup count verified: $client_tpgroups tpgroups",
                  -msg_level => 1);

    # For each tpgroup the client found, we need to convert the ip address to
    # lif name because iscsi tpgroup show always refers to lif names.
    foreach my $tpgroup (@{$tpgroups}) {

        # String to hold lif names for each tpgroup
        my $lifs;

        # For each lif in the tpgroup
        foreach my $ip (@{$tpgroup->{lif}}) {

            # Query net interface show with the ip address to get the lif name
            my @lif_output = $vserver->lif_list_info(-address => $ip);
            if ($vserver->error()) {
                $self->_set_err("Failed: lif_list_info", $vserver->errmsg());
                return $SanUtil::FALSE;
            }

            # Make sure we found only 1 lif with the ip address.  0 or more
            # than 1 would be bad.  Note the cluster should guarantee no ip
            # conflicts.
            if (scalar(@lif_output) != 1) {
                $self->_set_err(
                    "iscsi tpgroup verification failed while converting $ip to "
                      . "a lif name.\n  net int show found "
                      . scalar(@lif_output)
                      . " lifs.\nExpected only 1.");
                return $SanUtil::FALSE;
            }

            # Ok we found one interface for the ip, save the lif name.
            # If the lif string is currently empty, add the first lif.
            if (!length($lifs)) {
                $lifs = $lif_output[0]->{lif};
            } else {

                # Else, add a comma and the next lif.
                $lifs = "$lifs," . $lif_output[0]->{lif};
            }

        } ## end foreach my $ip (@{$tpgroup->...})

        # Debug print for new @lifs
        $self->sanlog(-msg       => "New lif string:\n$lifs",
                      -msg_level => 3);

        # After we have processed all of the ips, save the lifs names back
        # into the client hash
        $tpgroup->{lif} = $lifs;

    } ## end foreach my $tpgroup (@{$tpgroups...})

    # Debug print for new tpgroup hash
    $self->sanlog(-msg       => "New tpgroup array:\n" . Dumper($tpgroups),
                  -msg_level => 3);

    # For each tpgroup the client sees
    foreach my $client_tpg (@{$tpgroups}) {

        # Try and find a tpgroup that matches the exact parms of the client.
        my $tpgroups_found =
          scalar @{$iscsi_obj->tpgroup_list_info(%{$client_tpg})};
        if ($iscsi_obj->error()) {
            $self->_set_err("Failed: tpgroup_list_info", $iscsi_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # We better find 1 tpgroup that is an exact match.
        # NOTE: the 'mismatch' text in the message is used by _verify_isns.
        # DO NOT CHANGE!
        if ($tpgroups_found != 1) {
            $self->_set_err(
                    "Verify tpgroups mismatch!  Filter was expected to match 1 "
                      . "tpgroup, but matched $tpgroups_found.\n"
                      . "Filter used:\n"
                      . Dumper($client_tpg));
            return $SanUtil::FALSE;
        }

    } ## end foreach my $client_tpg (@{$tpgroups...})

    $self->sanlog(-msg       => "iscsi tpgroups verified",
                  -msg_level => 1);

    return $SanUtil::TRUE;

} ## end sub _verify_iscsi_tpgroups_from_client

################################################################################
# Method:    _verify_iscsi_security_from_client
# Objective: Verifies target iscsi security info collected from
#            SanClient::get_iscsi_info agrees with Data ONTAP.
# Details:
#     Inputs  -iscsi_obj(required)
#             -security(required) - hash from get_iscsi_info
#     Returns $SanUtil::TRUE  - security hash verified
#             $SanUtil::FALSE - security hash not verified and set_err
################################################################################
sub _verify_iscsi_security_from_client {
    my ($self, %args) = @_;

    my $iscsi_obj;
    my $security;    # iscsi session array to verify (as seen by client)

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /iscsi_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $iscsi_obj = $value;
        } elsif ($key =~ /security/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $security = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($iscsi_obj) =~
            /^San::ControllerCluster::Filer10M::Vserver::Iscsi$/i) {
        $self->_set_err(
                 "-iscsi_obj is a required parameter, must be an iSCSI object");
        return $SanUtil::FALSE;
    }
    unless (%{$security}) {
        $self->_set_err("-security is a required parameter.");
        return $SanUtil::FALSE;
    }

    # debug info
    $self->sanlog(-msg       => "Arg: security:\n" . Dumper($security),
                  -msg_level => 3);

    # Save the initiator's nodename for easier re-use.  Return error if the
    # initNodename was not set in the session entries.
    unless (length $security->{'initiator-name'}) {
        $self->_set_err("-security hash does not contain initiator-name key.");
        return $SanUtil::FALSE;
    }
    my $client = $security->{'initiator-name'};

    # Grab target's view of security
    my @target_security =
      $iscsi_obj->initiator_auth_list_info('initiator-name' => $client);
    if ($iscsi_obj->error()) {
        $self->_set_err(
                   "Failed: initiator_auth_list_info initiator-name => $client",
                   $iscsi_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # If we found more than one entry, that is a problem.
    if (scalar(@target_security) > 1) {
        $self->_set_err("iscsi security mismatch!\n"
                   . "Expected to find 1 or 0 matches on the target, but found "
                   . scalar(@target_security)
                   . " matches.\n");
        return $SanUtil::FALSE;

    } elsif (scalar(@target_security) == 0) {

        # If we found 0 entries, change the initiator name to 'default'.
        # This means there is not a custom security entry for this initiator.
        $security->{'initiator-name'} = "default";

        $self->sanlog(
                -msg =>
                  "Custom security entry not found, checking default security.",
                -msg_level => 1);
    } ## end elsif (scalar(@target_security...))

    # Try and find a security entry that matches the exact parms of the client.
    @target_security = $iscsi_obj->initiator_auth_list_info(%{$security});
    if ($iscsi_obj->error()) {
        $self->_set_err("Failed: initiator_auth_list_info",
                        $iscsi_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # We better find 1 security that is an exact match.
    if (scalar(@target_security) != 1) {
        $self->_set_err(
                      "Verify security failed!  Filter was expected to match 1 "
                        . "security entry, but matched "
                        . scalar(@target_security) . ".\n"
                        . "Filter used:\n"
                        . Dumper($security));
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg       => "iscsi security verified",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _verify_iscsi_security_from_client

################################################################################
# _login_logout_verify_all_vserver_info_clients
# Objective: vserver_info contains info of all clients associated to each
#            vserver. This function will accept login or logout action for
#            all clients associated to each vserver from the
#            {vserver_info} data structure. It then verified whether the login
#            or logout succeeded.
# Details:
#     Inputs - -sanmgr_obj (required) - SanMgr_Obj to verify logins or logouts
#
#              -action (required) - Specify what action to perform.
#              It supports "login" and "logout" as valid inputs.
#
#              -sendtargets (optional) - Do a refresh of sendtargets or not
#              By default, sendtargets info will be refreshed.
#              Only used when (-action => "login")
#
#              -fail_on_login (optional) - When set, any attempt by the clients
#              to login to target should fail - fail_on_login=>1
#              By default, login clients should not fail - fail_on_login=>0
#              Note: A value of 1 means you expect ALL logins from ALL clients
#              to ALL vservers to FAIL.
#
#              -fail_on_login_vserver_hash (optional) - A hash of vserver names
#              which are expected to fail on login.  All other vservers should
#              succeed on login.  If fail_on_login = 1, then there is no need
#              to specify this hash.
#
#              -keep_portals (optional) - Leave the target portals on the client
#              Only used when (-action => "logout")
#
#              -isns_address (optional) - Use isns discovery to this isns
#              server address instead of sendtargets.
#
#              -num_expected_luns (optional) - An integer of number of
#              luns that each client is expected to find.  If you provide this,
#              login speed is increased.
#
#     Returns On Success, $SanUtil::TRUE  - all clients login or logout successfully
#             On error, $SanUtil::FALSE - login or logout failed and set_err
################################################################################
sub _login_logout_verify_all_vserver_info_clients {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    my ($sanmgr_obj, $fail_on_login_vserver_hash, $fail_on_login);

    ###########################################################################
    # Some args are passed down to a lower level function call for validation.
    # -action will be validated in _login_logout_verify_one_vserver_info()
    # -sendtargets is optional param for _client_config()
    # -fail_on_login is optional param for _client_setup() and _client_config()
    # -keep_portals is optional param for _client_cleanup()
    # -isns_address is optional param for _client_config
    # -num_expected_luns is an optional param for _client_discovery()
    ###########################################################################

    # Parse arguments
    foreach (keys %args) {
        if (/-?sanmgr_obj/i) {
            $sanmgr_obj = delete $args{$_};
            if (not length($sanmgr_obj)) {
                $self->_set_err("Invalid value for argument '-sanmgr_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?fail_on_login_vserver_hash/i) {
            $fail_on_login_vserver_hash = $args{$_};
            if (ref($fail_on_login_vserver_hash !~ /HASH/)
                or (not keys(%{$fail_on_login_vserver_hash}))) {
                $self->_set_err(
                     "Invalid value for argument '-fail_on_login_vserver_hash',"
                       . " must be a hash, with keys.");
                return $SanUtil::FALSE;
            }
            delete $args{$_};
        } elsif (/-?fail_on_login$/i) {
            $fail_on_login = $args{$_};
            if (not length($fail_on_login)) {
                $self->_set_err("Invalid value for argument '-fail_on_login'");
                return $SanUtil::FALSE;
            }
        }
    } ## end foreach (keys %args)

    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(  "Missing required argument '-sanmgr_obj'."
                        . "Must be a SanMgr object. Object: "
                        . ref($sanmgr_obj));
        return $SanUtil::FALSE;
    }

    unless (ref($self->{vserver_info}) =~ /ARRAY/) {
        $self->_set_err( "Vserver_Info must be created before using it."
                       . "Call create_vserver_info() to create the vserver_info"
                       . ref($self->{vserver_info}));
        return $SanUtil::FALSE;
    }

    # Make sure user did not use fail_on_login=1 and fail_on_login_vserver_hash.
    # User should only use one or the other.
    if ($fail_on_login and $fail_on_login_vserver_hash) {
        $self->_set_err(
                  "Specified fail_on_login = 1 and fail_on_login_vserver_hash. "
                    . "Must select one or the other.");
        return $SanUtil::FALSE;
    }

    # Do action per vserver
    foreach my $vserver_info (@{$self->{vserver_info}}) {

        $args{'-vserver_info_obj'} = $vserver_info;

        $self->sanlog(-msg => "Do vserver " . $vserver_info->{name} . " action",
                      -msg_level => 1);

        # If the vserver name is in the fail_on_login_vserver_hash, then
        # set fail_on_login to 1.
        if (exists($fail_on_login_vserver_hash->{$vserver_info->{name}})) {
            $args{'-fail_on_login'} = $SanUtil::TRUE;
        }

        $self->_login_logout_one_vserver_info_clients(%args);
        if ($self->error()) {
            $self->_set_err("Failed for vserver " . $vserver_info->{name},
                            $self->errmsg());
            return $SanUtil::FALSE;
        }

        $self->sanlog(-msg       => "verify all clients for all vservers",
                      -msg_level => 1);

        # delete fail_on_login only if using fail_on_login_vserver_hash.
        # (in case it was set earlier.  want to be clean for next loop
        # iteration.)  Note: if the user specified fail_on_login=>1 and wants
        # all logins to fail, we don't want to delete the fail_on_login arg.
        if (exists($fail_on_login_vserver_hash->{$vserver_info->{name}})) {
            delete $args{'-fail_on_login'};
        }

        # Do a verify for all clients
        $self->verify_iscsi_clients(-sanmgr_obj => $sanmgr_obj);
        if ($self->error()) {
            $self->_set_err("Verify for all clients failed", $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end foreach my $vserver_info (@...)

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _login_logout_verify_all_vserver_info_clients

################################################################################
# _login_logout_one_vserver_info_clients
# Objective: vserver_info contains info of all clients associated to each
#            vserver. This function will accept login or logout action for
#            all clients associated to the vserver from the vserver_info_obj.
# Details:
#     Inputs  -vserver_info_obj (required) - vserver_info object (a single
#                                array element from vserver_info array)
#
#              -action (required) - Specify what action to perform.
#              It supports "login" and "logout" as valid inputs.
#
#              -sendtargets (optional) - Do a refresh of sendtargets or not
#              By default, sendtargets info will be refreshed.
#              Only used when (-action => "login")
#
#              -fail_on_login (optional) - When set, any attempt by the clients
#              to login to target should fail - fail_on_login=>1
#              By default, login clients should not fail - fail_on_login=>0
#
#              -keep_portals (optional) - Leave the target portals on the client
#              Only used when (-action => "logout")
#
#              -isns_address (optional) - Use isns discovery to this isns
#              server address instead of sendtargets.
#
#              -num_expected_luns (optional) - An integer of number of
#              luns that each client is expected to find.  If you provide this,
#              login speed is increased.
#
#     Returns On Success, $SanUtil::TRUE  - all clients login or logout successfully
#             On error, $SanUtil::FALSE - login or logout failed and set_err
################################################################################
sub _login_logout_one_vserver_info_clients {
    my ($self, %args) = @_;

    my $vserver_info;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    my $action;

    ###########################################################################
    # Some args are passed down to a lower level function call for validation.
    # -sendtargets is optional param for _client_config()
    # -fail_on_login is optional param for _client_setup() and _client_config()
    # -keep_portals is optional param for _client_cleanup()
    # -isns_address is optional param for _client_config
    # -num_expected_luns is an optional param for _client_discovery()
    ###########################################################################

    # Parse arguments
    foreach (keys %args) {
        if (/-?vserver_info_obj/i) {
            $vserver_info = delete $args{$_};
            if (not length($vserver_info)) {
                $self->_set_err(
                              "Invalid value for argument '-vserver_info_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?action/i) {
            $action = delete $args{$_};
            if (not length($action)) {
                $self->_set_err("Invalid value for argument '-action'");
                return $SanUtil::FALSE;
            }
        }
    }    # end foreach arg

    unless ($action =~ /^(login|logout)$/) {
        $self->_set_err("Missing required argument '-action'.");
        return $SanUtil::FALSE;
    }

    # Get vserver name
    my $vserver_name = $vserver_info->{name};

    # Get vserver and filer objects
    my $vserver_obj = $vserver_info->{object};
    my $filer_obj   = $vserver_obj->get_parent_obj();
    if ($vserver_obj->error()) {
        $self->_set_err(
                     "Failed to get filer object from vserver " . $vserver_name,
                     $vserver_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $self->sanlog(-msg => "Got vserver object for vserver " . $vserver_name,
                  -msg_level => 1);

    # Get iscsi targets
    my @iscsi_targets = $filer_obj->get_iscsi_targets(vserver => $vserver_name);
    if ($filer_obj->error()) {
        $self->_set_err(
            "Failed to get iscsi targets from filer for vserver $vserver_name.",
            $filer_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $args{'-iscsi_targets'} = \@iscsi_targets;

    $self->sanlog(-msg => "Got "
                    . scalar(@iscsi_targets)
                    . " iscsi targets for vserver "
                    . $vserver_name,
                  -msg_level => 1);

    foreach my $client_obj (@{$vserver_info->{client_objs}}) {

        # Get client name
        my $hostname = $client_obj->hostname();

        $args{'-client_obj'} = $client_obj;

        $self->sanlog(
                -msg => "$action of client $hostname for vserver $vserver_name",
                -msg_level => 1);

        # Login initiator to Vserver target
        $self->_client_setup(%args) if ($action =~ /login/);

        # Set clean level to 1
        # Removed, because clean level is set to 1 as a default in _client_cleanup

        # Logout initiator from Vserver target
        $self->_client_cleanup(%args) if ($action =~ /logout/);

        # Error check for login/logout
        if ($self->error()) {
            $self->_set_err(
                "Failed to $action for client $hostname for vserver $vserver_name",
                $self->errmsg());
            return $SanUtil::FALSE;
        }

    } ## end foreach my $client_obj (@{$vserver_info...})

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _login_logout_one_vserver_info_clients

################################################################################
# _client_setup
# Objective: Perform client login to target.
# Details:
#     Inputs - -client_obj (required) - Client_Obj
#
#              -iscsi_targets (required) - Reference to iscsi_target objects arr
#
#              -sendtargets (optional) - Do a refresh of sendtargets or not
#              By default, sendtargets info will be refreshed.
#              Only used when (-action => "login")
#
#              -fail_on_login (optional) - When set, any attempt by the clients
#              to login to target should fail - fail_on_login=>1
#              By default, login clients should not fail - fail_on_login=>0
#
#              -isns_address (optional) - Use isns discovery to this isns
#              server address instead of sendtargets.
#
#              -num_expected_luns (optional) - An integer of number of
#              luns that each client is expected to find.  If you provide this,
#              login speed is increased.
#
#     Returns On Success, $SanUtil::TRUE  - all clients login successfully
#             On error, $SanUtil::FALSE - login failed and set_err
################################################################################
sub _client_setup {
    my ($self, %args) = @_;

    my $fail_on_login = 0;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    ###########################################################################
    # Some args are passed down to a lower level function call for validation.
    # -iscsi_targets will be validated in _client_config()
    # -client_obj will be validated in _client_config() and _client_discovery()
    # -sendtargets is optional param for _client_config()
    # -isns_address is optional param for _client_config
    # -num_expected_luns is an optional param for _client_discovery()
    ###########################################################################

    # Parse arguments
    foreach (keys %args) {
        if (/-?fail_on_login/i) {
            $fail_on_login = lc($args{$_});
            if (not length($fail_on_login)) {
                $self->_set_err("Invalid value for argument '-fail_on_login'");
                return $SanUtil::FALSE;
            }
        }
    }    # end foreach arg

    $self->_client_config(%args);
    if ($self->error()) {
        if ($fail_on_login) {
            $self->_set_err("client config passed when its not supposed to",
                            $self->errmsg());
        } else {
            $self->_set_err("failed to configure client", $self->errmsg());
        }
        return $SanUtil::FALSE;
    }

    # If we are doing normal login, perform device discovery. (If login failed,
    # we skip this step, there will be no devices.)
    if (!$fail_on_login) {

        $self->_client_discovery(%args);
        if ($self->error()) {
            $self->_set_err("failed to do client discovery", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _client_setup

################################################################################
# _client_config
# Objective: Configure client with target info.
# Details:
#     Inputs - -client_obj (required) - Client_Obj
#
#              -iscsi_targets (required) - Reference to iscsi_target objects arr
#
#              -sendtargets (optional) - Do a refresh of sendtargets or not
#              By default, sendtargets info will be refreshed.
#
#              -fail_on_login (optional) - When set, any attempt by the clients
#              to login to target should fail - fail_on_login=>1
#              By default, login clients should not fail - fail_on_login=>0
#
#              -isns_address (optional) - Use isns discovery to this isns
#              server address instead of sendtargets.
#
#     Returns On Success, $SanUtil::TRUE  - all clients login successfully
#             On error, $SanUtil::FALSE - login failed and set_err
################################################################################
sub _client_config {
    my ($self, %args) = @_;

    my $sendtargets = 1;
    my ($client_obj);
    my $iscsi_targets;
    my $fail_on_login = 0;
    my $isns_address;
    my $isns_address_orig;
    my $discovery_type_orig;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    # Parse arguments
    foreach (keys %args) {
        if (/-?sendtargets/i) {
            $sendtargets = lc($args{$_});
            if (not length($sendtargets)) {
                $self->_set_err("Invalid value for argument '-sendtargets'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?client_obj/i) {
            $client_obj = $args{$_};
            if (not length($client_obj)) {
                $self->_set_err("Invalid value for argument '-client_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?iscsi_targets/i) {
            $iscsi_targets = $args{$_};
            if (not length($iscsi_targets)) {
                $self->_set_err("Invalid value for argument '-iscsi_targets'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?fail_on_login/i) {
            $fail_on_login = $args{$_};
            if (not length($fail_on_login)) {
                $self->_set_err("Invalid value for argument '-fail_on_login'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?isns_address/i) {
            $isns_address = $args{$_};
            if (not length($isns_address)) {
                $self->_set_err("Invalid value for argument '-isns_address'");
                return $SanUtil::FALSE;
            }
        }
    }    # end foreach arg

    unless (ref($client_obj) =~ /^San::SanClient/i) {
        $self->_set_err(  "Missing required argument '-client_obj'."
                        . "Must be a SanClient object. Object: "
                        . ref($client_obj));
        return $SanUtil::FALSE;
    }

    unless (ref($iscsi_targets) eq "ARRAY") {
        $self->_set_err(  "Required argument '-iscsi_targets'."
                        . "must be an array of iSCSI target objects: "
                        . $iscsi_targets);
        return $SanUtil::FALSE;
    }

    # Get client name
    my $hostname = $client_obj->hostname();

    # Save Discovery setting to a temp variable
    my $discovery_orig = $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY};

    # Do iSCSI Configure by setting Discovery=1
    $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY} = 1;

    # Save Pre-cleanup setting to a temp variable
    my $precleanup_orig =
      $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP};

    if ($sendtargets) {

        # Remove Discovery information in clients if Sendtargets=1
        $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} = 1;
    } else {

        # Leave Discovery information in clients if Sendtargets=0
        $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} = 0;
    }

    # If user passed in isns_address, configure parms for isns.
    if ($isns_address) {

        # Save original parms first.
        $discovery_type_orig =
          $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE};
        $isns_address_orig = $client_obj->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER};

        $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} = "isns";
        $client_obj->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER}    = $isns_address;
    } ## end if ($isns_address)

    # Configure Client
    $self->sanlog(-msg       => "Configure the client $hostname",
                  -msg_level => 1);

    # Normal case, do not fail_on_login
    if (!$fail_on_login) {
        $client_obj->configure(-targets => $iscsi_targets);
        if ($client_obj->error()) {
            $self->_set_err("Failed to configure client $hostname",
                            $client_obj->errmsg());
        }
    } else {

        # fail_on_login case. Make sure each target fails
        foreach my $target (@{$iscsi_targets}) {
            $client_obj->configure(-targets => [$target]);

            # If login is successful, raise error.
            if (!$client_obj->error()) {
                $self->_set_err("Client $hostname configure passed for target '"
                                . $target->get_ip()
                                . "'. It was expected to fail");

                # exit the for loop.
                last;
            }
            $self->sanlog(
                -msg =>
                  "Client '$hostname' configure failed as expected for target '"
                  . $target->get_ip() . "'",
                -msg_level => 1);

            # Now let's remove the target portal.  If the target portal is left
            # behind, then mpc iscsi show targets will start to introduce long
            # delays as the lifs start to accumlate. see burt 658478 for details.
            # Make a copy of the args.
            my %args2 = %args;

            # Set targets to the current target
            $args2{'-iscsi_targets'} = [$target];

            # Call cleanup
            $self->_client_cleanup(%args2);
            if ($self->error()) {
                $self->_set_err(
                       "failed to do client deconfigure for fail_on_login case",
                       $self->errmsg());
                return $SanUtil::FALSE;
            }

        } ## end foreach my $target (@{$iscsi_targets...})

    } ## end else [ if (!$fail_on_login) ]

    # Reset Discovery setting
    $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY} = $discovery_orig;

    # Reset Precleanup setting
    $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} =
      $precleanup_orig;

    # Reset isns settings if needed.
    if ($isns_address) {

        $client_obj->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =
          $discovery_type_orig;
        $client_obj->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER} = $isns_address_orig;
    }

    # Configure Client Error Check.  set_err was already performed above.
    if ($self->error()) {
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _client_config

################################################################################
# _client_discovery
# Objective: Discover devices on the client.
# Details:
#     Inputs - -client_obj (required) - Client_Obj
#
#              -num_expected_luns (optional) - An integer of number of
#              luns that each client is expected to find.  If you provide this,
#              login speed is increased.
#
#     Returns On Success, $SanUtil::TRUE
#             On error, $SanUtil::FALSE
################################################################################
sub _client_discovery {
    my ($self, %args) = @_;

    my ($client_obj);
    my $lun_count;
    my %new_args;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # Parse arguments
    foreach (keys %args) {
        if (/-?client_obj/i) {
            $client_obj = $args{$_};
            if (not length($client_obj)) {
                $self->_set_err("Invalid value for argument '-client_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?num_expected_luns/i) {
            $lun_count = $args{$_};
            if (not length($lun_count)) {
                $self->_set_err(
                             "Invalid value for argument '-num_expected_luns'");
                return $SanUtil::FALSE;
            }
        }
    }    # end foreach arg

    unless (ref($client_obj) =~ /^San::SanClient/i) {
        $self->_set_err(  "Missing required argument '-client_obj'."
                        . "Must be a SanClient object. Object: "
                        . ref($client_obj));
        return $SanUtil::FALSE;
    }

    # Get client name
    my $hostname = $client_obj->hostname();

    # If user passed in num_expected_luns, let's add it to the args.
    if (defined $lun_count and length $lun_count) {
        $new_args{'-num_expected_luns'} = $lun_count;
    }

    # Client Discovery
    $self->sanlog(-msg       => "Discovery on client $hostname",
                  -msg_level => 1);
    $client_obj->discover(-rescan_timeout => "3",
                          %new_args);

    if ($client_obj->error()) {
        $self->_set_err("Failed to discover client $hostname for targets",
                        $client_obj->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _client_discovery

################################################################################
# _client_cleanup
# Objective: Perform client logout from target and optionally, deconfigure it
# Details:
#     Inputs - -client_obj (required) - Client_Obj
#
#              -iscsi_targets (optional) - Reference to iscsi_target objects arr
#
#              -keep_portals (optional) - Leave the target portals on the client
#              By default, Portals will be cleaned up

#              -cleanup_level (optional) - Leave the target portals on the client
#              By default, Portals will be cleaned up

#     Returns On Success, $SanUtil::TRUE  - all clients login successfully
#             On error, $SanUtil::FALSE - login failed and set_err
################################################################################
sub _client_cleanup {
    my ($self, %args) = @_;

    my ($client_obj);
    my $iscsi_targets;
    my $clean_level  = 1;
    my %cmd_args     = {};
    my $keep_portals = 0;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    # Parse arguments
    foreach (keys %args) {
        if (/-?client_obj/i) {
            $client_obj = $args{$_};
            if (not length($client_obj)) {
                $self->_set_err("Invalid value for argument '-client_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?iscsi_targets/i) {
            $iscsi_targets = $args{$_};
            if (not length($iscsi_targets)) {
                $self->_set_err("Invalid value for argument '-iscsi_targets'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?keep_portals/i) {
            $keep_portals = $args{$_};
            if (not length($keep_portals)) {
                $self->_set_err("Invalid value for argument '-keep_portals'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?clean_level/i) {
            $clean_level = $args{$_};
            if (not length($clean_level)) {
                $self->_set_err("Invalid value for argument '-clean_level'");
                return $SanUtil::FALSE;
            }
        }
    }    # end foreach arg

    unless (ref($client_obj) =~ /^San::SanClient/i) {
        $self->_set_err(  "Missing required argument '-client_obj'."
                        . "Must be a SanClient object. Object: "
                        . ref($client_obj));
        return $SanUtil::FALSE;
    }

    # Get client name
    my $hostname = $client_obj->hostname();

    # Set the args for deconfigure
    $cmd_args{'-clean_level'} = $clean_level;
    if ($iscsi_targets) {
        $cmd_args{'-targets'} = $iscsi_targets;
    }

    # Save Discovery info
    my $deconfig_orig =
      $client_obj->{iscsi}->{CONFIG_ISCSI_WIN_KEEP_TARGET_PORTALS};

    # Do not remove Discovery info if keep_portals is 1
    # Remove Discovery info if keep_portals is 0.
    $client_obj->{iscsi}->{CONFIG_ISCSI_WIN_KEEP_TARGET_PORTALS} =
      $keep_portals;

    $self->sanlog(
            -msg => "Deconfigure client $hostname with targets: $iscsi_targets",
            -msg_level => 1);

    # Deconfigure Client
    $client_obj->deconfigure(%cmd_args);

    # Reset CONFIG_ISCSI_WIN_KEEP_TARGET_PORTALS
    $client_obj->{iscsi}->{CONFIG_ISCSI_WIN_KEEP_TARGET_PORTALS} =
      $deconfig_orig;

    # Check for error after deconfigure
    if ($client_obj->error()) {
        $self->_set_err("Failed to cleanup client $hostname",
                        $client_obj->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub _client_cleanup

################################################################################
# _login_logout_verify_all_client_info_clients
# Objective: client_info contains info of all vservers associated to each
#            client. This function will accept login or logout action for
#            all vservers associated to each client from the
#            {client_info} data structure. It then verified whether the login
#            or logout succeeded.
# Details:
#     Inputs - -sanmgr_obj (required) - SanMgr_Obj to verify logins or logouts
#
#              -action (required) - Specify what action to perform.
#              It supports "login" and "logout" as valid inputs.
#
#              -sendtargets (optional) - Do a refresh of sendtargets or not
#              By default, sendtargets info will be refreshed.
#              Only used when (-action => "login")
#
#              -fail_on_login (optional) - When set, any attempt by the clients
#              to login to target should fail - fail_on_login=>1
#              By default, login clients should not fail - fail_on_login=>0
#              Note: A value of 1 means you expect ALL logins from ALL clients
#              to ALL vservers to FAIL.
#
#              -fail_on_login_vserver_hash (optional) - A hash of vserver names
#              which are expected to fail on login.  All other vservers should
#              succeed on login.  If fail_on_login = 1, then there is no need
#              to specify this hash.
#
#              -keep_portals (optional) - Leave the target portals on the client
#              Only used when (-action => "logout")
#
#              -isns_address (optional) - Use isns discovery to this isns
#              server address instead of sendtargets.
#
#              -clients (optional) -A array of client names which need to login
#              to the iscsi tagets.
#
#              -iscsi_targets (optional) -An array of iscsi target objects.
#
#              -num_expected_luns (optional) - An integer of number of
#              luns that each client is expected to find.  If you provide this,
#              login speed is increased.
#
#     Returns On Success, $SanUtil::TRUE  - all clients login or logout successfully
#             On error, $SanUtil::FALSE - login or logout failed and set_err
################################################################################
sub _login_logout_verify_all_client_info_clients {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    my ($sanmgr_obj);
    my @clients;

    ###########################################################################
    # Some args are passed down to a lower level function call for validation.
    # -action will be validated in _login_logout_verify_one_client_info()
    # -sendtargets is optional param for _client_config()
    # -fail_on_login is optional param for _client_setup() and _client_config()
    # -keep_portals is optional param for _client_cleanup()
    # -fail_on_login_vserver_hash is optional param for
    #  _login_logut_one_client_info_client()
    # -isns_address is optional param for _client(config)
    # -iscsi_targets is optional parameter for _login_logout_verify_one_client_info()
    # -num_expected_luns is an optional param for _client_discovery()
    ###########################################################################

    # Parse arguments
    foreach (keys %args) {
        if (/-?sanmgr_obj/i) {
            $sanmgr_obj = delete $args{$_};
            if (not length($sanmgr_obj)) {
                $self->_set_err("Invalid value for argument '-sanmgr_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?clients/i) {
            my $value = $args{$_};
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-clients', value must be a reference to an 'ARRAY'"
                );
                return $SanUtil::FALSE;
            }
            @clients = @{$value};
        } ## end elsif (/-?clients/i)
    } ## end foreach (keys %args)

    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(  "Missing required argument '-sanmgr_obj'."
                        . "Must be a SanMgr object. Object: "
                        . ref($sanmgr_obj));
        return $SanUtil::FALSE;
    }

    unless (ref($self->{client_info}) =~ /ARRAY/) {
        $self->_set_err(  "client_info must be created before using it."
                        . "Call create_client_info() to create the client_info"
                        . ref($self->{client_info}));
        return $SanUtil::FALSE;
    }

    if (!scalar(@clients)) {

        # If specific list of clients is not specified then
        # all clients must login to the targets.
        foreach my $client_info (@{$self->{client_info}}) {

            $args{'-client_info_obj'} = $client_info;

            $self->sanlog(
                        -msg => "Do client " . $client_info->{name} . " action",
                        -msg_level => 1);

            $self->_login_logout_one_client_info_client(%args);
            if ($self->error()) {
                $self->_set_err("Failed for client " . $client_info->{name},
                                $self->errmsg());
                return $SanUtil::FALSE;
            }

            $self->sanlog(-msg       => "verify all clients for all vservers",
                          -msg_level => 1);
        } ## end foreach my $client_info (@{...})

    } else {

        # Login only the clients that are specified in the
        # clients array.

        foreach my $client_name (@clients) {
            my $found = 0;
            foreach my $client_info (@{$self->{client_info}}) {
                $args{'-client_info_obj'} = $client_info;
                if ($client_name eq $client_info->{name}) {
                    $self->sanlog(
                        -msg => "Do client " . $client_info->{name} . " action",
                        -msg_level => 1);

                    $self->_login_logout_one_client_info_client(%args);
                    if ($self->error()) {
                        $self->_set_err(
                                    "Failed for client " . $client_info->{name},
                                    $self->errmsg());
                        return $SanUtil::FALSE;
                    }
                    $found = 1;
                } ## end if ($client_name eq $client_info...)
            } ## end foreach my $client_info (@{...})
            if (!$found) {
                $self->_set_err(
                    "The client name : $client_name does not match  any of the clients "
                );
                return $SanUtil::FALSE;
            }
        } ## end foreach my $client_name (@clients)
    } ## end else [ if (!scalar(@clients))]

    # Do a verify for all clients. Moved outside the for loop to handle the
    # case where initiators are already logged in.

    $self->sanlog(-msg       => "verify all clients for all vservers",
                  -msg_level => 1);
    $self->verify_iscsi_clients(-sanmgr_obj => $sanmgr_obj);
    if ($self->error()) {
        $self->_set_err("Verify for all clients failed", $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _login_logout_verify_all_client_info_clients

################################################################################
# _login_logout_one_client_info_client
# Objective: client_info contains info on all vservers associated to each
#            client. This function will accept login or logout action for
#            all vservers associated to the client from the client_info_obj.
# Details:
#     Inputs - -client_info_obj (required) - client_info object (a single
#                                array element from client_info array)
#
#              -action (required) - Specify what action to perform.
#              It supports "login" and "logout" as valid inputs.
#
#              -sendtargets (optional) - Do a refresh of sendtargets or not
#              By default, sendtargets info will be refreshed.
#              Only used when (-action => "login")
#
#              -fail_on_login (optional) - When set, any attempt by the clients
#              to login to target should fail - fail_on_login=>1
#              By default, login clients should not fail - fail_on_login=>0
#              Note: A value of 1 means you expect ALL logins from this client
#              to ALL vservers to FAIL.
#
#              -fail_on_login_vserver_hash (optional) - A hash of vserver names
#              which are expected to fail on login.  All other vservers should
#              succeed on login.  If fail_on_login = 1, then there is no need
#              to specify this hash.
#
#              -keep_portals (optional) - Leave the target portals on the client
#              Only used when (-action => "logout")
#
#              -isns_address (optional) - Use isns discovery to this isns
#              server address instead of sendtargets.
#
#              -iscsi_targets (optional) - Array of specific targets the
#              clients should login to.
#
#              -num_expected_luns (optional) - An integer of number of
#              luns that each client is expected to find.  If you provide this,
#              login speed is increased.
#
#     Returns On Success, $SanUtil::TRUE  - all clients login or logout successfully
#             On error, $SanUtil::FALSE - login or logout failed and set_err
################################################################################
sub _login_logout_one_client_info_client {
    my ($self, %args) = @_;

    my $client_info;
    my @all_iscsi_targets;
    my @all_iscsi_targets_that_fail;
    my $fail_on_login_vserver_hash;
    my $fail_on_login;
    my @iscsi_targets;

    $self->_clear_err();

    $self->sanlog(-msg => "Start", -msg_level => 1);

    # debug info
    $Data::Dumper::Maxdepth = 1;
    $self->sanlog(-msg       => "Args:\n" . Dumper(\%args),
                  -msg_level => 1);
    $Data::Dumper::Maxdepth = 0;

    my $action;

    ###########################################################################
    # Some args are passed down to a lower level function call for validation.
    # -sendtargets is optional param for _client_config()
    # -fail_on_login is optional param for _client_setup() and _client_config()
    # -keep_portals is optional param for _client_cleanup()
    # -isns_address is optional param for _client(config)
    # -iscsi_targets is optional param that specifies the specific iscsi tagets
    # that the host should into.
    # -num_expected_luns is an optional param for _client_discovery()
    ###########################################################################

    # Parse arguments
    foreach (keys %args) {
        if (/-?client_info_obj/i) {
            $client_info = delete $args{$_};
            if (not length($client_info)) {
                $self->_set_err(
                               "Invalid value for argument '-client_info_obj'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?action/i) {
            $action = delete $args{$_};
            if (not length($action)) {
                $self->_set_err("Invalid value for argument '-action'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?fail_on_login_vserver_hash/i) {
            $fail_on_login_vserver_hash = $args{$_};
            if (ref($fail_on_login_vserver_hash !~ /HASH/)
                or (not keys(%{$fail_on_login_vserver_hash}))) {
                $self->_set_err(
                     "Invalid value for argument '-fail_on_login_vserver_hash',"
                       . " must be a hash, with keys.");
                return $SanUtil::FALSE;
            }
            delete $args{$_};
        } elsif (/-?fail_on_login$/i) {
            $fail_on_login = $args{$_};
            if (not length($fail_on_login)) {
                $self->_set_err("Invalid value for argument '-fail_on_login'");
                return $SanUtil::FALSE;
            }
        } elsif (/-?iscsi_targets/i) {
            my $value = $args{$_};
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-iscsi_targets', value must be a reference to an 'ARRAY'"
                );
                return $SanUtil::FALSE;
            }
            @iscsi_targets = @{$value};
        } ## end elsif (/-?iscsi_targets/i)
    }    # end foreach arg

    unless ($action =~ /^(login|logout)$/) {
        $self->_set_err("Missing required argument '-action'.");
        return $SanUtil::FALSE;
    }

    # If action is logout, make sure user did not pass in
    # fail_on_login_vserver_hash.  (Insures code below will work properly for
    # logout case.)
    if ($action =~ /^logout$/i and $fail_on_login_vserver_hash) {
        $self->_set_err(
                     "Action is logout and fail_on_login_vserver_hash parm was "
                       . "specified.  You cannot use this parm with logout.");
        return $SanUtil::FALSE;
    }

    # Make sure user did not use fail_on_login and fail_on_login_vserver_hash.
    # If user passed both, fail_on_login_vserver_hash would take precedence,
    # which the user may not want.  So just raise an error.
    if ($fail_on_login and $fail_on_login_vserver_hash) {
        $self->_set_err(
                  "Specified fail_on_login = 1 and fail_on_login_vserver_hash. "
                    . "Must select one or the other.");
        return $SanUtil::FALSE;
    }

    # Get client name
    my $client_name = $client_info->{name};

    # If a specific list of targets is not passed then
    # we should retrieve and log onto all the targets.
    if (!scalar(@iscsi_targets)) {

        # Scan each vserver
        foreach my $vserver_obj (@{$client_info->{vserver_objs}}) {

            # Grab the vserver name
            my $vserver_name = $vserver_obj->get_name();
            if ($vserver_obj->error()) {
                $self->_set_err("Failed to get vserver name",
                                $vserver_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # Grab the filer object of the vserver
            my $filer_obj = $vserver_obj->get_parent_obj();
            if ($vserver_obj->error()) {
                $self->_set_err("Failed to get filer object from vserver "
                                  . $vserver_name,
                                $vserver_obj->errmsg());
                return $SanUtil::FALSE;
            }

            @iscsi_targets =
              $filer_obj->get_iscsi_targets(vserver => $vserver_name);
            if ($filer_obj->error()) {
                $self->_set_err(
                    "Failed to get iscsi targets from filer for vserver $vserver_name.",
                    $filer_obj->errmsg());
                return $SanUtil::FALSE;
            }

            $self->sanlog(-msg => "Got "
                            . scalar(@iscsi_targets)
                            . " iscsi targets for vserver "
                            . $vserver_name,
                          -msg_level => 1);

            # If the vserver name is in the fail_on_login_vserver_hash
            if (exists($fail_on_login_vserver_hash->{$vserver_name})
                || $fail_on_login) {

                # Save the iscsi targets to the fail array
                push(@all_iscsi_targets_that_fail, @iscsi_targets);

            } else {

                # Save the iscsi targets to the pass array
                push(@all_iscsi_targets, @iscsi_targets);

            }
        } ## end foreach my $vserver_obj (@{...})
    } else {

        # Else user passed in iscsi_targets.  If they passed in fail_on_login true,
        # add to failing array.
        if ($fail_on_login) {
            push(@all_iscsi_targets_that_fail, @iscsi_targets);
        } else {

            # else add to passing array
            push(@all_iscsi_targets, @iscsi_targets);
        }
    } ## end else [ if (!scalar(@iscsi_targets...))]

    # Save args
    $args{'-client_obj'} = $client_info->{object};

    $self->sanlog(-msg => "Got "
                    . scalar(@all_iscsi_targets)
                    . " total iscsi targets that should succeed for client "
                    . $client_name,
                  -msg_level => 1);

    # If we have targets that should fail
    if (scalar(@all_iscsi_targets_that_fail)) {
        $self->sanlog(-msg => "Got "
                        . scalar(@all_iscsi_targets_that_fail)
                        . " total iscsi targets that should fail for client "
                        . $client_name,
                      -msg_level => 1);
    }

    $self->sanlog(-msg       => "$action of client $client_name",
                  -msg_level => 1);

    # Login initiator to Vserver target
    if ($action =~ /login/) {

        # If we have targets that should fail.  Perform failures first, because
        # on windows, the first failure will trigger a sendtargets refresh.
        if (scalar(@all_iscsi_targets_that_fail)) {

            # Set parms
            $args{'-fail_on_login'} = $SanUtil::TRUE;
            $args{'-iscsi_targets'} = \@all_iscsi_targets_that_fail;

            # perform logins that should fail
            $self->_client_setup(%args);
            if ($self->error()) {
                $self->_set_err("Client $client_name _client_setup failed",
                                $self->errmsg());
                return $SanUtil::FALSE;
            }

            # Cleanup parm in case we have more work to do below.
            delete $args{'-fail_on_login'};

        } ## end if (scalar(@all_iscsi_targets_that_fail...))

        # If we have targets that should succeed
        if (scalar(@all_iscsi_targets)) {

            # Set parms. Delete fail_on_login in case it was set above.
            $args{'-iscsi_targets'} = \@all_iscsi_targets;

            # Perform logins that should succeed. (And device discovery)
            $self->_client_setup(%args);
            if ($self->error()) {
                $self->_set_err("Failed to $action for client $client_name",
                                $self->errmsg());
                return $SanUtil::FALSE;
            }

        } ## end if (scalar(@all_iscsi_targets...))

    } ## end if ($action =~ /login/)

    # Set clean level to 1
    # Removed, because clean level is set to 1 as a default in _client_cleanup

    # Logout initiator from Vserver target
    if ($action =~ /logout/) {
        $args{'-iscsi_targets'} = \@all_iscsi_targets;
        $self->_client_cleanup(%args);
        if ($self->error()) {
            $self->_set_err("Failed to $action for client $client_name",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Insert slight delay to allow time for sessions to be cleaned up.
        &SanUtil::nate_sleep(3);
    } ## end if ($action =~ /logout/)

    $self->sanlog(-msg => "Success", -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _login_logout_one_client_info_client

################################################################################
# Method:    _verify_expected_clients_hash
# Objective: Verifies expected iSCSI data agrees with Data ONTAP.
#            Called from verify_iscsi_expected - data comes from test case
#            (clients hash)
# Details:
#     Inputs - cserver_obj (required)
#            - sanmgr_obj  (required)
#            - \% clients_hash (required)
#                   - keys are initiator iqn
#                       - keys for session, connection, initiator
#                       - each key contains an array with specific details for
#                         all items
#     Returns $SanUtil::TRUE  - all items match
#             $SanUtil::FALSE - all items do NOT match and set_err
################################################################################
sub _verify_expected_clients_hash {
    my ($self, %args) = @_;

    my $cserver_obj;
    my $sanmgr_obj;
    my $clients_hash;    # Expected iscsi config data - clients hash

    $self->sanlog(-msg       => "START",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /cserver_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $cserver_obj = $value;
        } elsif ($key =~ /sanmgr_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $sanmgr_obj = $value;
        } elsif ($key =~ /clients_hash/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $clients_hash = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($cserver_obj) =~ /^San::ControllerCluster::Filer10M$/i) {
        $self->_set_err(
              "-cserver_obj is a required parameter, must be a cserver object");
        return $SanUtil::FALSE;
    }
    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(
                "-sanmgr_obj is a required parameter, must be a SanMgr object");
        return $SanUtil::FALSE;
    }
    unless (%{$clients_hash}) {
        $self->_set_err("-clients_hash is a required parameter.");
        return $SanUtil::FALSE;
    }

    # Scan each client in the data structure.
    foreach my $client_iqn (keys %{$clients_hash}) {

        # Scan each target under the client
        foreach my $target_iqn (keys %{$clients_hash->{$client_iqn}}) {

            # Create a label to use for messages
            my $label = "Client = $client_iqn\nTarget = $target_iqn";

            $self->sanlog(-msg       => "Processing $label",
                          -msg_level => 1);

            # Array of items to check for total flag
            my @items = ("session", "connection", "initiator");

            # Scan each array
            foreach my $item (@items) {

                # If the array has elements, then check for total flag.
                if (scalar(
                        @{  $clients_hash->{$client_iqn}->{$target_iqn}->{$item}
                         })
                  ) {

                    # Verify the total key is not used in the array:
                    my $total =
                      $self->_verify_total_flag(
                        -array => \@{
                            $clients_hash->{$client_iqn}->{$target_iqn}->{$item}
                        });

                    # If error is set, we know the user mixed using total/no
                    # total in the array elements.
                    if ($self->error()) {
                        $self->_set_err(
                               "_verify_total_flag failed: total key"
                                 . " is NOT supported under the 'clients' hash",
                               $self->errmsg());
                        return $SanUtil::FALSE;
                    }

                    # If total flag was found, return error.
                    if ($total) {
                        $self->_set_err(
                            "Cannot use the total flag within the 'clients' hash key.\n"
                              . "Please see: http://wikid/w/Scalable_SAN_Test_Engineering/CheatSheet/verify_iscsi_expected"
                        );
                        return $SanUtil::FALSE;
                    }
                } ## end if (scalar(@{$clients_hash...}))
            } ## end foreach my $item (@items)

            # Use the target nodename to find the vserver name
            my @iscsi_show =
              $cserver_obj->iscsi_list_info('target-name' => $target_iqn);
            if ($cserver_obj->error()) {
                $self->_set_err(
                           "iscsi_list_info failed for target-name $target_iqn",
                           $cserver_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # Verify we only got back 1 object.
            if (scalar(@iscsi_show) != 1) {
                $self->_set_err(
                           "Expected to find 1 match for iscsi show, but found "
                             . scalar(@iscsi_show)
                             . " matches.\n"
                             . "Filter used: target-name => $target_iqn\n");
                return $SanUtil::FALSE;
            }

            # Now use the vserver name to get the vserver obj
            my @vserver_obj =
              $sanmgr_obj->get_objects(-type  => "vserver",
                                       -names => [$iscsi_show[0]->{vserver}]);
            if ($sanmgr_obj->error()) {
                $self->_set_err("get_objects failed for vserver "
                                  . $iscsi_show[0]->{vserver},
                                $sanmgr_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # Verify we got a single vserver
            if (scalar(@vserver_obj) != 1) {
                $self->_set_err(  "get_objects returned "
                                . scalar(@vserver_obj)
                                . " vservers with name "
                                . $iscsi_show[0]->{vserver});
                return $SanUtil::FALSE;
            }

            # Check if iscsi exists on the vserver
            my $iscsi_exist = $vserver_obj[0]->iscsi_exist();
            if ($vserver_obj[0]->error()) {
                $self->_set_err("iscsi_exist failed for vserver "
                                  . $iscsi_show[0]->{vserver},
                                $vserver_obj[0]->errmsg());
                return $SanUtil::FALSE;
            }

            # If iscsi does not exist, return error.
            if (!$iscsi_exist) {
                $self->_set_err("iscsi does not exist on vserver "
                                . $iscsi_show[0]->{vserver});
                return $SanUtil::FALSE;
            }

            # Now use the vserver obj to get the iscsi object
            my $iscsi_obj = $vserver_obj[0]->get_iscsi_obj();
            if ($vserver_obj[0]->error()) {
                $self->_set_err("get_iscsi_obj failed for vserver "
                                  . $iscsi_show[0]->{vserver},
                                $vserver_obj[0]->errmsg());
                return $SanUtil::FALSE;
            }

            $self->sanlog(-msg => "Processing initiator: $client_iqn\n"
                            . "With target: $target_iqn",
                          -msg_level => 1);

            # Verify iSCSI sessions
            $self->_verify_iscsi_array(
                       -object => $iscsi_obj,
                       -type   => "session",
                       -value =>
                         $clients_hash->{$client_iqn}->{$target_iqn}->{session},
                       -initiator_iqn => $client_iqn);
            if ($self->error()) {
                $self->_set_err("Failed to verify iscsi sessions!",
                                $label, $self->errmsg());
                return $SanUtil::FALSE;
            }

            # Verify iSCSI connections
            $self->_verify_iscsi_array(
                    -object => $iscsi_obj,
                    -type   => "connection",
                    -value =>
                      $clients_hash->{$client_iqn}->{$target_iqn}->{connection},
                    -initiator_iqn => $client_iqn);
            if ($self->error()) {
                $self->_set_err("Failed to verify iscsi connections!",
                                $label, $self->errmsg());
                return $SanUtil::FALSE;
            }

            # Verify iSCSI initiators
            $self->_verify_iscsi_array(
                     -object => $iscsi_obj,
                     -type   => "initiator",
                     -value =>
                       $clients_hash->{$client_iqn}->{$target_iqn}->{initiator},
                     -initiator_iqn => $client_iqn);
            if ($self->error()) {
                $self->_set_err("Failed to verify iscsi initiators!",
                                $label, $self->errmsg());
                return $SanUtil::FALSE;
            }

            $self->sanlog(-msg       => "SUCCESS for $label",
                          -msg_level => 1);

        } ## end foreach my $target_iqn (keys...)
    } ## end foreach my $client_iqn (keys...)
} ## end sub _verify_expected_clients_hash

################################################################################
# Method:    _verify_expected_iscsi_hash
# Objective: Verifies expected iSCSI data agrees with Data ONTAP.
#            (1 client -> 1 vserver)
#            Called from verify_iscsi_expected - data comes from test case.
# Details:
#     Inputs - object (required) - iscsi or cserver object
#            - \% expected_hash (required)
#              Note this is a single vserver key from the expected hash.
#              Keys may exist for the following iscsi items:
#              accesslist - contains an array of hashes
#              alias      - contains scalar
#              connection - contains an array of hashes
#              initiator  - contains an array of hashes
#              interface  - contains an array of hashes
#              isns       - contains a hash
#              nodename   - contains scalar
#              security   - contains an array of hashes
#              session    - contains an array of hashes
#              show       - contains a hash
#              status     - contains a hash
#              tpgroup    - contains an array of hashes
#              Each hash contains keys which correspond to the c-mode cmd options
#              Details for expected hash can be found here:
#              http://wikid/w/Scalable_SAN_Test_Engineering/CheatSheet/verify_iscsi_expected
#            - vserver_name (required if object is a cserver)
#     Returns $SanUtil::TRUE  - all items match
#             $SanUtil::FALSE - all items do NOT match and set_err
################################################################################
sub _verify_expected_iscsi_hash {
    my ($self, %args) = @_;

    my $object;           # iscsi or cserver object
    my $expected_hash;    # Expected iscsi config data
    my $vserver_name;
    my $cserver = 0;      # flag to indicate if we are using cserver

    $self->sanlog(-msg       => "START",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /object/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $object = $value;
        } elsif ($key =~ /expected_hash/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $expected_hash = $value;
        } elsif ($key =~ /vserver_name/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $vserver_name = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (
        ref($object) =~
        /^San::ControllerCluster::Filer10M::Vserver::Iscsi$|^San::ControllerCluster::Filer10M$/i
      ) {
        $self->_set_err(  "-object is a required parameter, "
                        . "must be an iSCSI or cserver object");
        return $SanUtil::FALSE;
    }
    unless (%{$expected_hash}) {
        $self->_set_err("-expected_hash is a required parameter.");
        return $SanUtil::FALSE;
    }

    # If we have a cserver object, verify the user passed in the vserver name
    if (    (ref($object) =~ /^San::ControllerCluster::Filer10M$/i)
        and (not $vserver_name)) {
        $self->_set_err(  "-vserver_name is a required parameter "
                        . "since method is called with a cserver object.");
        return $SanUtil::FALSE;
    }

    # debug info
    # Note - don't dump iscsi/cserver obj, it's way too large.
    $self->sanlog(-msg => "Arg: expected_hash:\n" . Dumper($expected_hash),
                  -msg_level => 3);

    # Phase 1 - Verify target's iSCSI nodename
    # If nodename exists, verify the nodename.
    if (exists($expected_hash->{nodename})) {
        $self->_verify_iscsi_scalar(-object       => $object,
                                    -type         => "nodename",
                                    -value        => $expected_hash->{nodename},
                                    -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi nodename!",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }

    } ## end if (exists($expected_hash...))

    # Phase 2 - Verify target's iSCSI alias.
    # If alias exists, verify the alias.
    if (exists($expected_hash->{alias})) {
        $self->_verify_iscsi_scalar(-object       => $object,
                                    -type         => "alias",
                                    -value        => $expected_hash->{alias},
                                    -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi alias!", $self->errmsg());
            return $SanUtil::FALSE;
        }

    } ## end if (exists($expected_hash...))

    # Phase 3 - Verify iSCSI status
    # If status exists, verify the status data.
    if (exists($expected_hash->{status})) {
        $self->_verify_iscsi_hash(-object       => $object,
                                  -type         => "status",
                                  -value        => $expected_hash->{status},
                                  -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi status!", $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    # Phase 4 - Verify iSCSI show
    # If show exists, verify the show data.
    if (exists($expected_hash->{show})) {
        $self->_verify_iscsi_hash(-object       => $object,
                                  -type         => "show",
                                  -value        => $expected_hash->{show},
                                  -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi show!", $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    # Phase 5 - Verify iSCSI security
    # If security exists, verify the security entries.
    if (exists($expected_hash->{security})) {
        $self->_verify_iscsi_array(-object       => $object,
                                   -type         => "security",
                                   -value        => $expected_hash->{security},
                                   -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi security!",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    # Phase 6 - Verify iSCSI sessions
    # If session exists, verify the sessions.
    if (exists($expected_hash->{session})) {
        $self->_verify_iscsi_array(-object       => $object,
                                   -type         => "session",
                                   -value        => $expected_hash->{session},
                                   -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi sessions!",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    # Phase 7 - Verify iSCSI connections
    # If connection exists, verify the connections.
    if (exists($expected_hash->{connection})) {
        $self->_verify_iscsi_array(-object => $object,
                                   -type   => "connection",
                                   -value  => $expected_hash->{connection},
                                   -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi connections!",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    # Phase 8 - Verify iSCSI initiators
    # If initiator exists, verify the initiators.
    if (exists($expected_hash->{initiator})) {
        $self->_verify_iscsi_array(-object       => $object,
                                   -type         => "initiator",
                                   -value        => $expected_hash->{initiator},
                                   -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi initiators!",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    # Phase 9 - Verify iSCSI portals/interfaces.
    # If interface exists, verify using portal and interface.
    # Note: These commands are aliased.
    if (exists($expected_hash->{interface})) {
        $self->_verify_iscsi_array(-object       => $object,
                                   -type         => "interface",
                                   -value        => $expected_hash->{interface},
                                   -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi interfaces!",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Verify iSCSI portals
        $self->_verify_iscsi_array(-object       => $object,
                                   -type         => "portal",
                                   -value        => $expected_hash->{interface},
                                   -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi portals!", $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    # Phase 10 - Verify iSCSI tpgroups
    # If tpgroup exists, verify the tpgroups.
    if (exists($expected_hash->{tpgroup})) {
        $self->_verify_iscsi_array(-object       => $object,
                                   -type         => "tpgroup",
                                   -value        => $expected_hash->{tpgroup},
                                   -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi tpgroups!",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    # Phase 11 - Verify iSCSI accesslists
    # If acceslist exists, verify the accesslists.
    if (exists($expected_hash->{accesslist})) {
        $self->_verify_iscsi_array(-object => $object,
                                   -type   => "accesslist",
                                   -value  => $expected_hash->{accesslist},
                                   -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi accesslists!",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    # Phase 12 - Verify iSCSI iSNS
    # If isns exists, verify the isns data.
    if (exists($expected_hash->{isns})) {
        $self->_verify_iscsi_hash(-object       => $object,
                                  -type         => "isns",
                                  -value        => $expected_hash->{isns},
                                  -vserver_name => $vserver_name);
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi isns!", $self->errmsg());

            # Since iSNS tests have been problematic, let's do an isns show
            # here to help with debug.  No need to check for error,
            # because we are already exiting no matter what happens.
            $object->isns_list_info();

            return $SanUtil::FALSE;
        }
    } ## end if (exists($expected_hash...))

    $self->sanlog(-msg       => "SUCCESS!",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _verify_expected_iscsi_hash

################################################################################
# Method:    _verify_total_flag
# Objective: Verifies an array of hashes uses the total flag consistently.
#            Total flag should be used in all hashes or in none of them.
# Details:
#     Inputs - array (required) - array to verify
#     Returns $SanUtil::TRUE    - total used in all array elements
#             $SanUtil::FALSE   - total is NOT used in all elements
#             $SanUtil::FALSE & set_err - mixing of total and non-total elements
#                                         detected or other failure
################################################################################
sub _verify_total_flag {
    my ($self, %args) = @_;

    my $array;    # array to verify
    my $total;    # flag for total

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /array/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $array = $value;
        }
    }

    # Check mandatory parms
    unless (@{$array}) {
        $self->_set_err("-array is a required parameter.");
        return $SanUtil::FALSE;
    }

    # debug info
    $self->sanlog(-msg       => "Arg: array:\n" . Dumper($array),
                  -msg_level => 3);

    # Need to scan all array elements for consistent use of 'total' key.
    # If the array has elements, check the first element for total.
    if (scalar(@{$array}) > 0) {
        $total = exists(@{$array}[0]->{total});
    } else {

        # Else the array has 0 elements, return FALSE, total flag not set.
        return $SanUtil::FALSE;

    }

    # Now scan all elements to make sure they agree with the first element.
    foreach my $element (@{$array}) {

        # If the current element's total flag does not agree with what we found
        # in the first element, return an error.
        if ((exists($element->{total}) ne $total)) {
            $self->_set_err("Mixing of array elements with the 'total' key and "
                . "without the 'total' key is not allowed.\n"
                . "Please see: http://wikid/w/Scalable_SAN_Test_Engineering/CheatSheet/verify_iscsi_expected"
            );
            return $SanUtil::FALSE;
        }
    } ## end foreach my $element (@{$array...})

    # If we get here, all array elements agreed on the value of total.
    if ($total) {
        return $SanUtil::TRUE;
    } else {
        return $SanUtil::FALSE;
    }
} ## end sub _verify_total_flag

################################################################################
# Method:    _verify_iscsi_client_across_cluster
# Objective: Verifies iSCSI data totals across the cluster for a single client.
#            Method is called by the following:
#            - verify_iscsi_clients  - data from SanClient::get_iscsi_info
#            - verify_iscsi_expected - data from test case
#            Verifies the following:
#            connection - total number in the cluster for one client
#            initiator  - total number in the cluster for one client
#            session    - total number in the cluster for one client
# Details:
#     Inputs - cserver_obj (required)
#            - \% client_hash (optional) - client data from either:
#              - verify_iscsi_clients  - data from SanClient::get_iscsi_info
#              - verify_iscsi_expected - data from test case
#              Note this contains the entire hash returned from
#              get_iscsi_info.  It represents one cluster.  If this value is
#              null, we verify the cluster finds 0 sessions, connections, and
#              initiators for the given initiator.
#            - initiator_iqn (required) - client's iscsi nodename
#     Returns $SanUtil::TRUE  - all items match
#             $SanUtil::FALSE - all items do NOT match and set_err
################################################################################
sub _verify_iscsi_client_across_cluster {
    my ($self, %args) = @_;

    my $cserver_obj;      # (required)
    my $client_hash;      # Client hash to verify
    my $initiator_iqn;    # client's iscsi nodename (required)

    $self->sanlog(-msg       => "START",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /cserver_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $cserver_obj = $value;
        } elsif ($key =~ /client_hash/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $client_hash = $value;
        } elsif ($key =~ /initiator_iqn/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $initiator_iqn = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($cserver_obj) =~ /^San::ControllerCluster::Filer10M$/i) {
        $self->_set_err(
              "-cserver_obj is a required parameter, must be a cserver object");
        return $SanUtil::FALSE;
    }
    unless ($initiator_iqn) {
        $self->_set_err("-initiator_iqn is a required parameter.");
        return $SanUtil::FALSE;
    }

    # debug info
    # Note - don't dump cserver_obj, it's way too large.
    # Make client_hash a high debug level, we already dump it in
    # verify_iscsi_clients
    $self->sanlog(-msg       => "Arg: client_hash:\n" . Dumper($client_hash),
                  -msg_level => 3);
    $self->sanlog(-msg       => "Arg: initiator_iqn: $initiator_iqn",
                  -msg_level => 1);

    # Verify cluster wide totals for this client: session, connection, and
    # initiator
    my $client_sessions    = 0;
    my $client_connections = 0;
    my $client_initiators  = 0;

    # Scan each nodename the client is logged into, and total the values
    foreach my $nodename (keys %{$client_hash}) {

        # Make sure keys exist for session, connection, initiator.
        if (   (not exists($client_hash->{$nodename}->{session}))
            or (not exists($client_hash->{$nodename}->{connection}))
            or (not exists($client_hash->{$nodename}->{initiator}))) {
            $self->_set_err(
                 "Failed: For target nodename key '$nodename':\nall 3 keys are "
                   . "required: session, connection and initiator");
            return $SanUtil::FALSE;
        }

        $client_sessions += scalar(@{$client_hash->{$nodename}->{session}});
        $client_connections +=
          scalar(@{$client_hash->{$nodename}->{connection}});
        $client_initiators += scalar(@{$client_hash->{$nodename}->{initiator}});
    } ## end foreach my $nodename (keys ...)

    $self->sanlog(-msg => "Client $initiator_iqn sees:\n"
                    . "$client_sessions sessions\n"
                    . "$client_connections connections\n"
                    . "$client_initiators initiators",
                  -msg_level => 1);

    # Grab the cluster's view of iscsi sessions
    my @cluster_sessions =
      $cserver_obj->iscsi_session_list_info('initiator-name' => $initiator_iqn,
                                            'session-type'   => "Normal");
    if ($cserver_obj->error()) {
        $self->_set_err(
                 "Failed: iscsi_session_list_info "
                   . "initiator-name => $initiator_iqn, session-type => Normal",
                 $cserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Make sure cluster/client agree on the number of sessions
    if ($client_sessions != scalar(@cluster_sessions)) {
        $self->_set_err("Cluster iscsi session count mismatch!\n"
                      . "Client $initiator_iqn says there are $client_sessions "
                      . "sessions.\nCluster says there are "
                      . scalar(@cluster_sessions)
                      . " sessions.");
        return $SanUtil::FALSE;
    }

    $self->sanlog(
                -msg => "Client $initiator_iqn:\n"
                  . "Cluster session count verified: $client_sessions sessions",
                -msg_level => 1);

    # Grab the cluster's view of iscsi connections
    my @cluster_connections =
      $cserver_obj->iscsi_connection_list_info(
                                             'initiator-name' => $initiator_iqn,
                                             'session-type'   => "Normal");
    if ($cserver_obj->error()) {
        $self->_set_err(
                 "Failed: iscsi_connection_list_info "
                   . "initiator-name => $initiator_iqn, session-type => Normal",
                 $cserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Make sure cluster/client agree on the number of connections
    if ($client_connections != scalar(@cluster_connections)) {
        $self->_set_err("Cluster iscsi connection count mismatch!\n"
                   . "Client $initiator_iqn says there are $client_connections "
                   . "connections.\nCluster says there are "
                   . scalar(@cluster_connections)
                   . " connections.");
        return $SanUtil::FALSE;
    }

    $self->sanlog(
        -msg => "Client $initiator_iqn:\n"
          . "Cluster connection count verified: $client_connections connections",
        -msg_level => 1);

    # Grab the cluster's view of iscsi initiators
    my @cluster_initiators =
      $cserver_obj->iscsi_initiator_list_info(
                                             'initiator-name' => $initiator_iqn,
                                             'session-type'   => "Normal");
    if ($cserver_obj->error()) {
        $self->_set_err(
                 "Failed: iscsi_initiator_list_info "
                   . "initiator-name => $initiator_iqn, session-type => Normal",
                 $cserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Make sure cluster/client agree on the number of initiators
    if ($client_initiators != scalar(@cluster_initiators)) {
        $self->_set_err("Cluster iscsi initiator count mismatch!\n"
                    . "Client $initiator_iqn says there are $client_initiators "
                    . "initiators.\nCluster says there are "
                    . scalar(@cluster_initiators)
                    . " initiators.");
        return $SanUtil::FALSE;
    }

    $self->sanlog(
          -msg => "Client $initiator_iqn:\n"
            . "Cluster initiator count verified: $client_initiators initiators",
          -msg_level => 1);

    # If we get here, everything matched, return success.
    $self->sanlog(-msg       => "SUCCESS!\nClient $initiator_iqn",
                  -msg_level => 1);
    return $SanUtil::TRUE;

} ## end sub _verify_iscsi_client_across_cluster

################################################################################
# Method:    _verify_iscsi_client
# Objective: Verifies iSCSI data collected from SanClient::get_iscsi_info
#            agrees with Data ONTAP. (1 client -> 1 vserver)
#            Verifies the following:
#            alias      - contains scalar value (Solaris only)
#            connection - contains an array of hashes
#            initiator  - contains an array of hashes
#            interface  - contains an array of hashes
#            nodename   - contains scalar
#            security   - contains a hash
#            session    - contains an array of hashes
#            show       - contains a hash
#            status     - contains a hash
#            tpgroup    - contains an array of hashes
#            Each hash contains keys which correspond to the c-mode cmd options
# Details:
#     Inputs - iscsi_obj (required)
#            - \% client_hash (required) - client data from get_iscsi_info
#              Note this is a single key from the hash returned from
#              get_iscsi_info.  It represents one vserver.
#            - hosttype (required) - client's hosttype
#     Returns $SanUtil::TRUE  - all items match
#             $SanUtil::FALSE - all items do NOT match and set_err
################################################################################
sub _verify_iscsi_client {
    my ($self, %args) = @_;

    my $iscsi_obj;
    my $client_hash;    # Hash of a single target entry from
                        # SanClient::get_iscsi_info.  (required)
    my $hosttype;       # Hosttype of client OS. (required)
    my $isns = 0;       # iSNS flag

    $self->sanlog(-msg       => "START",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /iscsi_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $iscsi_obj = $value;
        } elsif ($key =~ /client_hash/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $client_hash = $value;
        } elsif ($key =~ /hosttype/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $hosttype = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($iscsi_obj) =~
            /^San::ControllerCluster::Filer10M::Vserver::Iscsi$/i) {
        $self->_set_err(
                 "-iscsi_obj is a required parameter, must be an iSCSI object");
        return $SanUtil::FALSE;
    }
    unless (%{$client_hash}) {
        $self->_set_err("-client_hash is a required parameter.");
        return $SanUtil::FALSE;
    }
    unless ($hosttype) {
        $self->_set_err("-hosttype is a required parameter.");
        return $SanUtil::FALSE;
    }

    # debug info
    # Note - don't dump iscsi_obj, it's way too large.  If we have the wrong
    # iscsi object, the very first verification (nodename) will fail.
    $self->sanlog(-msg       => "Arg: client_hash:\n" . Dumper($client_hash),
                  -msg_level => 3);

    $self->sanlog(-msg       => "Arg: hostype: $hosttype",
                  -msg_level => 3);

    # Make sure initiator name was saved at this location.
    if (!length $client_hash->{security}->{'initiator-name'}) {
        $self->_set_err("client_hash is missing {security}->{initiator-name}");
        return $SanUtil::FALSE;
    }

    # Save initiator's iqn for later use
    my $initiator_iqn = $client_hash->{security}->{'initiator-name'};

    # Create an err_label to use in case we have to set_err
    my $err_label = "Client = $initiator_iqn";

    # Check hash for discovery type.  If it is iSNS, set the flag.
    if ($client_hash->{discovery_type} =~ /isns/i) {
        $isns = 1;
    }

    $self->sanlog(-msg       => "Processing initiator: $initiator_iqn",
                  -msg_level => 1);

    # Phase 1 - Verify target's iSCSI nodename
    $self->_verify_iscsi_scalar(-object => $iscsi_obj,
                                -type   => "nodename",
                                -value  => $client_hash->{nodename});
    if ($self->error()) {
        $self->_set_err("Failed to verify iscsi nodename!",
                        $err_label, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # If the nodename check passed, add Target name to the err_label
    $err_label = $err_label . "\nTarget = " . $client_hash->{nodename};

    # Phase 2 - Verify target's iSCSI alias.
    # If alias exists, verify the alias.
    if (exists($client_hash->{alias})) {
        $self->_verify_iscsi_scalar(-object => $iscsi_obj,
                                    -type   => "alias",
                                    -value  => $client_hash->{alias});
        if ($self->error()) {
            $self->_set_err("Failed to verify iscsi alias!",
                            $err_label, $self->errmsg());
            return $SanUtil::FALSE;
        }

    } else {

        # Else, skip the alias verification.
        $self->sanlog(
              -msg => "iscsi alias verification skipped, alias was not defined",
              -msg_level => 1);
    }

    # Phase 3 - Verify iSCSI status
    $self->_verify_iscsi_hash(-object => $iscsi_obj,
                              -type   => "status",
                              -value  => $client_hash->{status});
    if ($self->error()) {
        $self->_set_err("Failed to verify iscsi status!",
                        $err_label, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Phase 4 - Verify iSCSI show
    $self->_verify_iscsi_hash(-object => $iscsi_obj,
                              -type   => "show",
                              -value  => $client_hash->{show});
    if ($self->error()) {
        $self->_set_err("Failed to verify iscsi show!",
                        $err_label, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Phase 5 - Verify iSCSI security
    $self->_verify_iscsi_security_from_client(
                                           -iscsi_obj => $iscsi_obj,
                                           -security => $client_hash->{security}
    );
    if ($self->error()) {
        $self->_set_err("Failed to verify iscsi security!",
                        $err_label, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Phase 6 - Verify iSCSI sessions
    $self->_verify_iscsi_array(-object        => $iscsi_obj,
                               -type          => "session",
                               -value         => $client_hash->{session},
                               -initiator_iqn => $initiator_iqn);
    if ($self->error()) {
        $self->_set_err("Failed to verify iscsi sessions!",
                        $err_label, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Phase 7 - Verify iSCSI connections
    $self->_verify_iscsi_array(-object        => $iscsi_obj,
                               -type          => "connection",
                               -value         => $client_hash->{connection},
                               -initiator_iqn => $initiator_iqn);
    if ($self->error()) {
        $self->_set_err("Failed to verify iscsi connections!",
                        $err_label, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Phase 8 - Verify iSCSI initiators
    $self->_verify_iscsi_array(-object        => $iscsi_obj,
                               -type          => "initiator",
                               -value         => $client_hash->{initiator},
                               -initiator_iqn => $initiator_iqn);
    if ($self->error()) {
        $self->_set_err("Failed to verify iscsi initiators!",
                        $err_label, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Phase 9 - Verify iSCSI portals.
    # The portal command is aliased to the interface command.  We use the
    # interface verification routine with the portal flag set to 1.
    # We also pass in the interface hash because get_iscsi_info only returns
    # an interface hash and not a portal hash.  (There was no need to create
    # an identical portal hash, the interface hash is sufficient for both
    # portal and interface verification.)
    $self->_verify_iscsi_interfaces_from_client(
                                       -iscsi_obj  => $iscsi_obj,
                                       -interfaces => $client_hash->{interface},
                                       -initiator_iqn => $initiator_iqn,
                                       -portal        => 1,
                                       -isns          => $isns);

    if ($self->error()) {
        $self->_set_err("Failed to verify iscsi portals!",
                        $err_label, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Phase 10 - Verify iSCSI interfaces
    $self->_verify_iscsi_interfaces_from_client(
                                       -iscsi_obj  => $iscsi_obj,
                                       -interfaces => $client_hash->{interface},
                                       -initiator_iqn => $initiator_iqn,
                                       -isns          => $isns);

    if ($self->error()) {
        $self->_set_err("Failed to verify iscsi interfaces!",
                        $err_label, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Phase 11 - Verify iSCSI tpgroups
    $self->_verify_iscsi_tpgroups_from_client(
                                           -iscsi_obj => $iscsi_obj,
                                           -tpgroups => $client_hash->{tpgroup},
                                           -initiator_iqn => $initiator_iqn,
                                           -isns          => $isns);
    if ($self->error()) {
        $self->_set_err("Failed to verify iscsi tpgroups!",
                        $err_label, $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg       => "SUCCESS!\n" . $err_label,
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _verify_iscsi_client

################################################################################
# Method:    _verify_iscsi_scalar
# Objective: Verifies a scalar value. Method is called by the following:
#            - verify_iscsi_clients - value collected from
#              SanClient::get_iscsi_info
#            - verify_iscsi_expected - value comes from test case
# Details:
#     Inputs  -object (required) - iscsi or cserver object
#             -type (required)   - indicates what we will be verifying.
#                                  (alias or nodename)
#             -value (required)  - (scalar) expected value of the type
#             -vserver_name (required if object is a cserver)
#     Returns $SanUtil::TRUE  - type/value verified
#             $SanUtil::FALSE - type/value not verified and set_err
################################################################################
sub _verify_iscsi_scalar {
    my ($self, %args) = @_;

    my $object;          # Stores the iscsi or cserver object
    my $type;            # Possible values: alias or nodename
    my $exp_value;       # The expected target value of the alias or nodename
    my $vserver_name;    # Vserver name, need if using cserver object
    my $cserver = $SanUtil::FALSE;    # flag to indicate if we are using cserver
    my $target_value;                 # Value found on the target

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters.  Note value could be empty if iscsi is not
    # created, so less strict error checking for the 'value' key.
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /object/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $object = $value;
        } elsif ($key =~ /type/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $type = $value;
        } elsif ($key =~ /value/) {
            $exp_value = $value;
        } elsif ($key =~ /vserver_name/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $vserver_name = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (
        ref($object) =~
        /^San::ControllerCluster::Filer10M::Vserver::Iscsi$|^San::ControllerCluster::Filer10M$/i
      ) {
        $self->_set_err(  "-object is a required parameter, "
                        . "must be an iSCSI object or cserver object.");
        return $SanUtil::FALSE;
    }
    unless ($type =~ /^alias$|^nodename$/i) {
        $self->_set_err("-type must be 'alias' or 'nodename'");
        return $SanUtil::FALSE;
    }
    unless (ref($exp_value) =~ /^$/i) {
        $self->_set_err("-value must be a string");
        return $SanUtil::FALSE;
    }

    # Turn on the cserver flag if needed.
    if (ref($object) =~ /^San::ControllerCluster::Filer10M$/i) {
        $cserver = $SanUtil::TRUE;
    }

    # If we have a cserver object, verify the user passed in the vserver name
    if ($cserver and not $vserver_name) {
        $self->_set_err(  "-vserver_name is a required parameter "
                        . "since method is called with a cserver object.");
        return $SanUtil::FALSE;
    }

    # debug info
    $self->sanlog(-msg       => "Arg: type: $type",
                  -msg_level => 1);
    $self->sanlog(-msg       => "Arg: value: $exp_value",
                  -msg_level => 3);
    $self->sanlog(-msg       => "Arg: vserver_name: $vserver_name",
                  -msg_level => 3);

    # If using cserver obj, use the proper method from FilerHelper to get the
    # target value.
    # Note: - FilerHelper returns array.
    #       - vserver name is added to the call
    if ($cserver == $SanUtil::TRUE) {

        # If type is alias
        if ($type =~ /^alias$/) {

            $target_value =
              $object->iscsi_alias_list_info(-vserver => $vserver_name);

        } else {

            # Else type must be nodename
            $target_value =
              $object->iscsi_nodename_list_info(-vserver => $vserver_name);
        }

        # Check for error
        if ($object->error()) {
            $self->_set_err("Failed to get $type", $object->errmsg());
            return $SanUtil::FALSE;
        }

        # If we are using the cserver object, then iscsi is not present.
        # If the user expected the value to be something other than null,
        # then return error.
        # Or, if we got back some objects from the call above, that's an
        # error because iscsi should not exist.
        if (($exp_value) or (scalar(@{$target_value}))) {
            $self->_set_err(  "Target iscsi $type mismatch!\n"
                            . "Expected $type = $exp_value\n"
                            . "Target found "
                            . scalar(@{$target_value})
                            . " $type. (0 = iscsi not created)\n"
                            . "Expected $type should be null and 0 $type "
                            . "should have been found.");
            return $SanUtil::FALSE;
        } ## end if (($exp_value) or (scalar...))

        # Update exp_value so the exiting message is user-friendly.
        $exp_value = "iscsi not created!";

    } else {

        # Else get the target_value from the iscsi object.
        # If type is alias
        if ($type =~ /^alias$/) {

            # Set use_alias_cmd to 1 to use alias cmd.
            # Note get_alias() returns a string.
            $target_value = $object->get_alias(-use_alias_cmd => 1);

        } else {

            # Else type must be nodename
            # Get nodename from target.  Don't use cache.
            # Note get_nodename() returns a string.
            $target_value = $object->get_nodename(-use_cache => 0);
        }

        # Check for error
        if ($object->error()) {
            $self->_set_err("Failed to get $type", $object->errmsg());
            return $SanUtil::FALSE;
        }

        # Compare exp_value vs target_value, they should match.
        # NOTE: the 'mismatch' text in the message is used by _verify_isns.
        # DO NOT CHANGE!
        if ($exp_value ne $target_value) {
            $self->_set_err(  "Target iscsi $type mismatch!\n"
                            . "Expected $type = $exp_value\n"
                            . "Target says the $type = $target_value");
            return $SanUtil::FALSE;
        }

    } ## end else [ if ($cserver == $SanUtil::TRUE)]

    $self->sanlog(-msg       => "SUCCESS: iscsi $type verified: $exp_value",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _verify_iscsi_scalar

################################################################################
# Method:    _verify_iscsi_hash
# Objective: Verifies a hash value. Method is called by the following:
#            - verify_iscsi_clients - value collected from
#              SanClient::get_iscsi_info
#            - verify_iscsi_expected - value comes from test case
# Details:
#     Inputs  -object (required) - iscsi or cserver object
#             -type (required)   - indicates what we will be verifying.
#                                  (isns, show, status)
#             -value (required)  - hash of the type
#             -vserver_name (required if object is a cserver)
#     Returns $SanUtil::TRUE  - type/value verified
#             $SanUtil::FALSE - type/value not verified and set_err
#     Note: An empty hash {} will be verified as 1 item.
################################################################################
sub _verify_iscsi_hash {
    my ($self, %args) = @_;

    my $object;          # Stores the iscsi or cserver object
    my $type;            # A valid type: isns, show or status
    my $exp_hash;        # The expected hash of the type
    my $vserver_name;    # Vserver name, need if using cserver object
    my $cserver = $SanUtil::FALSE;    # flag to indicate if we are using cserver
    my $target_value;                 # Value found on the target

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /object/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $object = $value;
        } elsif ($key =~ /type/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $type = $value;
        } elsif ($key =~ /value/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $exp_hash = $value;
        } elsif ($key =~ /vserver_name/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $vserver_name = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (
        ref($object) =~
        /^San::ControllerCluster::Filer10M::Vserver::Iscsi$|^San::ControllerCluster::Filer10M$/i
      ) {
        $self->_set_err(  "-object is a required parameter, "
                        . "must be an iSCSI object or cserver object.");
        return $SanUtil::FALSE;
    }
    unless ($type =~ /^show$|^status$|^isns$/i) {
        $self->_set_err("-type must be 'show', 'status', or 'isns'");
        return $SanUtil::FALSE;
    }
    unless (ref($exp_hash) =~ /^HASH$/i) {
        $self->_set_err("For -type $type, value must be a HASH");
        return $SanUtil::FALSE;
    }

    # Turn on the cserver flag if needed.
    if (ref($object) =~ /^San::ControllerCluster::Filer10M$/i) {
        $cserver = $SanUtil::TRUE;
    }

    # If we have a cserver object, verify the user passed in the vserver name
    if ($cserver and not $vserver_name) {
        $self->_set_err(  "-vserver_name is a required parameter "
                        . "since method is called with a cserver object.");
        return $SanUtil::FALSE;
    }

    # debug info
    $self->sanlog(-msg       => "Arg: type: $type",
                  -msg_level => 1);
    $self->sanlog(-msg       => "Arg: value: $exp_hash",
                  -msg_level => 3);
    $self->sanlog(-msg       => "Arg: vserver_name: $vserver_name",
                  -msg_level => 3);

    # Number of items found when we call to Data ONTAP
    my $items_found;

    # By default, we are searching for 1 element.
    my $expected = 1;

    # Make a copy of the hash in case the total key needs to be deleted.
    # We're dealing with a reference and we don't want to delete the total
    # key from the original data.
    my %arg_copy = %{$exp_hash};

    # If total flag is on, save the total and remove the key.
    if (exists($arg_copy{total})) {
        $expected = $arg_copy{total};
        delete $arg_copy{total};
    }

    # If using cserver obj, use the proper method from FilerHelper to get the
    # target value.
    # Note: - FilerHelper returns an array.
    #       - vserver name is added to the call
    if ($cserver) {

        $arg_copy{vserver} = $vserver_name;

        # If type is show
        if ($type =~ /^show$/) {

            $items_found = scalar(@{$object->iscsi_list_info(%arg_copy)});

        } elsif ($type =~ /^status$/) {

            # Else type is status
            $items_found =
              scalar(@{$object->iscsi_status_list_info(%arg_copy)});

        } else {

            # Else type must be isns
            $items_found = scalar(@{$object->iscsi_isns_list_info(%arg_copy)});
        }

    } else {

        # Call from iscsi object. Note an array will NOT be returned.  A hash
        # or null will be returned.
        if ($type =~ /^show$/) {

            $items_found = $object->get_list_info(%arg_copy);

        } elsif ($type =~ /^status$/) {

            # Else type is status
            $items_found = $object->status(%arg_copy);

        } else {

            # Else type must be isns
            $items_found = $object->isns_list_info(%arg_copy);
        }

        # Above calls will return null or a hash ref.  If they return a hash
        # ref, set items_found to 1.
        if ($items_found) {
            $items_found = 1;
        } else {
            $items_found = 0;
        }

    } ## end else [ if ($cserver) ]

    # Check for error
    if ($object->error()) {
        $self->_set_err("Failed to get $type", $object->errmsg());
        return $SanUtil::FALSE;
    }

    # We better find the expected number of items.
    # NOTE: the 'mismatch' text in the message is used by _verify_isns.
    # DO NOT CHANGE!
    if ($items_found != $expected) {
        $self->_set_err(  "iscsi $type mismatch!\n"
                        . "Expected to find $expected matches on the target, "
                        . "but found $items_found matches.\n"
                        . "Filter used:\n"
                        . Dumper(\%arg_copy));

        # Since iSNS tests have been problematic, if type is isns, let's do an
        # isns show here to help with debug.  No need to check for error, cause
        # we are already exiting no matter what happens.
        if ($type =~ /^isns$/) {
            $object->isns_list_info();
        }

        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg       => "Matched $expected iscsi $type as expected.",
                  -msg_level => 1);
    $self->sanlog(-msg       => "SUCCESS: iscsi $type verified",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _verify_iscsi_hash

################################################################################
# Method:    _verify_iscsi_array
# Objective: Verifies an array of hash value. Method is called by
#            by the following:
#            - verify_iscsi_clients - value collected from
#              SanClient::get_iscsi_info
#            - verify_iscsi_expected - value comes from test case
# Details:
#     Inputs  -object (required) - iscsi or cserver object.  (vserver object
#                                  for lif)
#             -type (required)   - indicates what we will be verifying.
#                                  (accesslist, connection, initiator,
#                                   interface, lif, portal, security, session,
#                                   tpgroup)
#             -value (required)  - array of hash of the type
#             -vserver_name (required if object is a cserver)
#             -initiator_iqn(optional) - initiator's iqn - acts as a signal to
#              add the initiator-name and session-type "Normal" to the high
#              level verification of the count. (for session, connection, and
#              initiator ONLY)
#     Returns $SanUtil::TRUE  - type/value verified
#             $SanUtil::FALSE - type/value not verified and set_err
#     Note: An empty array [] will be verified as 0 items.
################################################################################
sub _verify_iscsi_array {
    my ($self, %args) = @_;

    my $object;          # Stores the iscsi, cserver, or vserver object
    my $type;            # A valid type
    my $exp_array;       # The expected target value of the type
    my $vserver_name;    # Vserver name, need if using cserver object
    my $cserver = $SanUtil::FALSE; # flag to indicate if we are using cserver
    my $total   = $SanUtil::FALSE; # Total flag initialized to zero. Indicates
                                   # if total key is used in all array elements.
    my $initiator_iqn;    # optional parm used for session/connection/initiator
    my %new_args;         # args for function calls

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /object/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $object = $value;
        } elsif ($key =~ /type/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $type = $value;
        } elsif ($key =~ /value/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $exp_array = $value;
        } elsif ($key =~ /vserver_name/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $vserver_name = $value;
        } elsif ($key =~ /initiator_iqn/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $initiator_iqn = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (
        ref($object) =~
        /^San::ControllerCluster::Filer10M::Vserver::Iscsi$|^San::ControllerCluster::Filer10M$|^San::ControllerCluster::Filer10M::Vserver$/i
      ) {
        $self->_set_err(  "-object is a required parameter, "
                        . "must be an iSCSI, cserver, or vserver object.");
        return $SanUtil::FALSE;
    }
    unless ($type =~
        /^accesslist$|^connection$|^initiator$|^interface$|^lif$|^portal$|^security$|^session$|^tpgroup$/i
      ) {
        $self->_set_err(  "-type must be: accesslist, connection, initiator, "
                        . "interface, lif, portal, security, session, tpgroup");
        return $SanUtil::FALSE;
    }
    unless (ref($exp_array) =~ /^ARRAY$/i) {
        $self->_set_err("For -type $type, value must be an ARRAY");
        return $SanUtil::FALSE;
    }

    # If the object is vserver, and type is not lif, return error.
    if (    (ref($object) =~ /^San::ControllerCluster::Filer10M::Vserver$/)
        and ($type !~ /^lif$/)) {
        $self->_set_err("For -object(vserver object), -type must be 'lif'");
        return $SanUtil::FALSE;
    }

    # If the type is lif and the object is not a vserver, return error.
    if (    ($type =~ /^lif$/)
        and (ref($object) !~ /^San::ControllerCluster::Filer10M::Vserver$/)) {
        $self->_set_err("For -type lif, -object must be a vserver object");
        return $SanUtil::FALSE;
    }

    # Turn on the cserver flag if needed.
    if (ref($object) =~ /^San::ControllerCluster::Filer10M$/i) {
        $cserver = $SanUtil::TRUE;
    }

    # If we have a cserver object, verify the user passed in the vserver name
    if ($cserver and not $vserver_name) {
        $self->_set_err(  "-vserver_name is a required parameter "
                        . "since method is called with a cserver object.");
        return $SanUtil::FALSE;
    }

    # debug info
    $self->sanlog(-msg       => "Arg: type: $type",
                  -msg_level => 1);
    $self->sanlog(-msg       => "Arg: value: " . Dumper($exp_array),
                  -msg_level => 3);
    $self->sanlog(-msg       => "Arg: vserver_name: $vserver_name",
                  -msg_level => 3);

    # If the array has members, verify the total flag is used consistently.
    # Also sets the total flag to true or false to indicate if the total flag
    # is in use.
    if (scalar(@{$exp_array}) > 0) {
        $total = $self->_verify_total_flag(-array => $exp_array);
        if ($self->error()) {
            $self->_set_err("_verify_total_flag failed", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # If total flag is ON, skip the verification on the scalar value of
    # the array.
    if ($total) {

        $self->sanlog(
              -msg => "total key is used, skipping array count check for $type",
              -msg_level => 1);

    } else {

        # The number of array elements is the expected number of items.
        my $expected_items = scalar(@{$exp_array});

        # Container to store objects we find.
        my $items_found;

        # If the type is session/connection/initiator and initiator_iqn was
        # passed in, then we add initiator-name and session-type.
        if (($type =~ /^session$|^connection$|^initiator$/)
            and $initiator_iqn) {
            $new_args{'initiator-name'} = $initiator_iqn;
            $new_args{'session-type'}   = "Normal";
        }

        # If using cserver obj, use the proper method from FilerHelper to get the
        # target value.
        # Note: - FilerHelper returns an array.
        #       - vserver name is added to the call
        if ($cserver) {

            # Configure args
            $new_args{vserver} = $vserver_name;

            # For the proper type, determine the total number of items found.
            # We will check for error at the bottom of this structure.
            # Note lif is not listed in the cserver block, because lif will
            # always use a vserver object.
            if ($type =~ /^accesslist$/) {

                $items_found =
                  $object->iscsi_interface_accesslist_list_info(%new_args);

            } elsif ($type =~ /^connection$/) {

                $items_found = $object->iscsi_connection_list_info(%new_args);

            } elsif ($type =~ /^initiator$/) {

                $items_found = $object->iscsi_initiator_list_info(%new_args);

            } elsif ($type =~ /^interface$/) {

                $items_found = $object->iscsi_interface_list_info(%new_args);

            } elsif ($type =~ /^portal$/) {

                $items_found = $object->iscsi_portal_list_info(%new_args);

            } elsif ($type =~ /^security$/) {

                $items_found =
                  $object->iscsi_initiator_auth_list_info(%new_args);

            } elsif ($type =~ /^session$/) {

                $items_found = $object->iscsi_session_list_info(%new_args);

            } elsif ($type =~ /^tpgroup$/) {

                $items_found = $object->iscsi_tpgroup_list_info(%new_args);

            }
        } else {

            # Else use iscsi object or vserver object to get total.
            if ($type =~ /^accesslist$/) {

                $items_found =
                  $object->interface_accesslist_list_info(%new_args);

            } elsif ($type =~ /^connection$/) {

                $items_found = $object->connection_list_info(%new_args);

            } elsif ($type =~ /^initiator$/) {

                $items_found = $object->initiator_list_info(%new_args);

            } elsif ($type =~ /^interface$/) {

                $items_found = $object->interface_list_info(%new_args);

            } elsif ($type =~ /^portal$/) {

                $items_found = $object->portal_list_info(%new_args);

            } elsif ($type =~ /^security$/) {

                $items_found = $object->initiator_auth_list_info(%new_args);

            } elsif ($type =~ /^session$/) {

                $items_found = $object->session_list_info(%new_args);

            } elsif ($type =~ /^tpgroup$/) {

                $items_found = $object->tpgroup_list_info(%new_args);

            } elsif ($type =~ /^lif$/) {

                $items_found = $object->lif_list_info();

            }

        } ## end else [ if ($cserver) ]

        # Check for error
        if ($object->error()) {
            $self->_set_err("Failed to get $type", $object->errmsg());
            return $SanUtil::FALSE;
        }

        # Make sure number of items found agrees with our expectation
        if ($expected_items != scalar(@{$items_found})) {
            $self->_set_err(  "iscsi $type array count mismatch!\n"
                            . "Expected $expected_items $type(s).\n"
                            . "Target says there are "
                            . scalar(@{$items_found})
                            . " $type(s)");
            return $SanUtil::FALSE;
        }

        $self->sanlog(
                 -msg => "$type array count verified: $expected_items $type(s)",
                 -msg_level => 1);
    } ## end else [ if ($total) ]

    # For each item we expect to find
    foreach my $exp_item (@{$exp_array}) {

        # By default, we are searching for 1 element.
        my $expected_items = 1;

        # Container to store objects we find.
        my $items_found;

        # Make a copy of the hash in case the total key needs to be deleted.
        # We're dealing with a reference and we don't want to delete the total
        # key from the original data.
        %new_args = %{$exp_item};

        # If total flag is on, save the total and remove the key.
        if ($total) {
            $expected_items = $new_args{total};
            delete $new_args{total};
        }

        # If using cserver obj, use the proper method from FilerHelper to get
        # the target value.
        # Note: - FilerHelper returns an array.
        #       - vserver name is added to the call
        if ($cserver) {

            # Configure args
            $new_args{vserver} = $vserver_name;

            # For the proper type, determine the total number of items found.
            # We will check for error at the bottom of this structure.
            # Note lif is not listed in the cserver block, because lif will
            # always use a vserver object.
            if ($type =~ /^accesslist$/) {

                $items_found =
                  $object->iscsi_interface_accesslist_list_info(%new_args);

            } elsif ($type =~ /^connection$/) {

                $items_found = $object->iscsi_connection_list_info(%new_args);

            } elsif ($type =~ /^initiator$/) {

                $items_found = $object->iscsi_initiator_list_info(%new_args);

            } elsif ($type =~ /^interface$/) {

                $items_found = $object->iscsi_interface_list_info(%new_args);

            } elsif ($type =~ /^portal$/) {

                $items_found = $object->iscsi_portal_list_info(%new_args);

            } elsif ($type =~ /^security$/) {

                $items_found =
                  $object->iscsi_initiator_auth_list_info(%new_args);

            } elsif ($type =~ /^session$/) {

                $items_found = $object->iscsi_session_list_info(%new_args);

            } elsif ($type =~ /^tpgroup$/) {

                $items_found = $object->iscsi_tpgroup_list_info(%new_args);

            }
        } else {

            # Else use iscsi object or vserver object to get total.
            if ($type =~ /^accesslist$/) {

                $items_found =
                  scalar($object->interface_accesslist_list_info(%new_args));

            } elsif ($type =~ /^connection$/) {

                $items_found = $object->connection_list_info(%new_args);

            } elsif ($type =~ /^initiator$/) {

                $items_found = $object->initiator_list_info(%new_args);

            } elsif ($type =~ /^interface$/) {

                $items_found = $object->interface_list_info(%new_args);

            } elsif ($type =~ /^portal$/) {

                $items_found = $object->portal_list_info(%new_args);

            } elsif ($type =~ /^security$/) {

                $items_found = $object->initiator_auth_list_info(%new_args);

            } elsif ($type =~ /^session$/) {

                $items_found = $object->session_list_info(%new_args);

            } elsif ($type =~ /^tpgroup$/) {

                $items_found = $object->tpgroup_list_info(%new_args);

            } elsif ($type =~ /^lif$/) {

                $items_found = $object->lif_list_info(%new_args);

            }
        } ## end else [ if ($cserver) ]

        # Check for error
        if ($object->error()) {
            $self->_set_err("Failed to get $type", $object->errmsg());
            return $SanUtil::FALSE;
        }

        # Make sure the number of items found agrees with our expectation
        if ($expected_items != scalar(@{$items_found})) {
            $self->_set_err(  "iscsi $type count mismatch!\n"
                            . "Expected $expected_items $type(s).\n"
                            . "Target says there are "
                            . scalar(@{$items_found})
                            . " $type(s)");
            return $SanUtil::FALSE;
        }

        $self->sanlog(-msg => "Matched $expected_items $type(s) as expected.",
                      -msg_level => 1);

    } ## end foreach my $exp_item (@{$exp_array...})

    $self->sanlog(-msg       => "SUCCESS: iscsi $type" . "s verified",
                  -msg_level => 1);

    return $SanUtil::TRUE;

} ## end sub _verify_iscsi_array

################################################################################
# Method:    _get_port_objs
# Objective: Return an array of network adapter (port) objects
# Details:
#     Inputs - -sanmgr_obj (required) - SanMgr object
#               -ports (required) - Provide the array of port names to be modified
#               -nodes (optional) - Provide the array of node names to be modified
#     Returns On Success, an array of network adapter objects
#             On error, an empty array and set_err
################################################################################
sub _get_port_objs {

    my ($self, %args) = @_;

    my ($sanmgr_obj, $value);
    my (@ports,      @nodes);
    my @port_objs;

    $self->_clear_err();

    $self->sanlog(-msg       => "Start",
                  -msg_level => 1);

    # Parse arguments
    foreach (keys %args) {
        if (/-?sanmgr_obj/i) {
            $sanmgr_obj = $args{$_};
            if (not length($sanmgr_obj)) {
                $self->_set_err("Invalid value for argument '-sanmgr_obj'");
                return (wantarray()) ? () : [];
            }
        } elsif (/-?ports/i) {
            $value = $args{$_};
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-ports', value must be a reference to an 'ARRAY'"
                );
                return (wantarray()) ? () : [];
            }
            @ports = @{$value};
        } elsif (/^-?nodes$/i) {
            $value = $args{$_};
            if (ref($value) !~ /ARRAY/i) {
                $self->_set_err(
                    "For argument '-nodes', value must be a reference to an 'ARRAY'"
                );
                return (wantarray()) ? () : [];
            }
            @nodes = @{$value};
            unless (scalar(@nodes)) {
                $self->_set_err("No entries found in '-nodes' array");
                return (wantarray()) ? () : [];
            }
        } ## end elsif (/^-?nodes$/i)
    }    # end foreach arg

    unless (ref($sanmgr_obj) =~ /^San::SanMgr$/i) {
        $self->_set_err(  "Missing required argument '-sanmgr_obj'."
                        . "Must be sanmgr_obj: ref($sanmgr_obj)");
        return (wantarray()) ? () : [];
    }

    unless (scalar(@ports)) {
        $self->_set_err("Missing required argument '-ports'.");
        return (wantarray()) ? () : [];
    }

    # Get the cluster object
    my @cluster_objs = $sanmgr_obj->get_objects(type => "cluster");
    if ($sanmgr_obj->error()) {
        $self->_set_err("Failed to get cluster objects", $sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # Verify we got back some cluster objects.
    if (!scalar(@cluster_objs)) {
        $self->_set_err("get_objects(cluster) returned 0 objects");
        return (wantarray()) ? () : [];
    }

    # Get the cserver object
    my $cserver_obj = $cluster_objs[0]->get_cserver_obj();
    if ($cluster_objs[0]->error()) {
        $self->_set_err("Failed to get cserver object",
                        $cluster_objs[0]->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get network adapters
    $self->sanlog(-msg       => "Get all network adapters",
                  -msg_level => 1);

    my @netadap_objs = $cserver_obj->get_networkadapter_objs();
    if ($cserver_obj->error()) {
        $self->_set_err("Failed to get network adapter objects",
                        $cserver_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # network adapter loop
    foreach my $netadap_obj (@netadap_objs) {

        # Get port name
        my $portname       = $netadap_obj->get_port();
        my $ports_nodename = $netadap_obj->get_node();

        if (not grep(/^$portname$/, @ports)) {
            $self->sanlog(
                  -msg => "Skipping port '$portname' as it was not part of the "
                    . "'-ports' requested for: "
                    . join(', ', @ports),
                  -msg_level => 1);
            next;
        }

        # If the -nodes filter is specified, only accept those nodes
        if (scalar @nodes and not grep(/^$ports_nodename$/, @nodes)) {
            $self->sanlog(
                  -msg =>
                    "Skipping node '$ports_nodename' as it was not part of the "
                    . "'-nodes' requested for: "
                    . join(', ', @nodes),
                  -msg_level => 1);
            next;
        }

        $self->sanlog(
                  -msg => "Getting Port '$portname' for Node '$ports_nodename'",
                  -msg_level => 1);

        push(@port_objs, $netadap_obj);
    } ## end foreach my $netadap_obj (@netadap_objs)

    $self->sanlog(-msg => "Success!", -msg_level => 1);

    return (wantarray()) ? @port_objs : \@port_objs;
} ## end sub _get_port_objs

################################################################################
# Method:    _verify_isns
# Objective: Verifies iSCSI data collected from WinIsnsServer::get_isns_info_cmode
#            agrees with Data ONTAP.  get_isns_info_cmode collects data from a
#            single vserver only, so this method verifies a single vserver only.
#            Verifies the following:
#            alias      - contains scalar value
#            interface  - contains an array of hashes
#            isns       - contains a hash
#            nodename   - contains scalar
#            show       - contains a hash
#            status     - contains a hash
#            tpgroup    - contains an array of hashes
#            Each hash contains keys which correspond to the c-mode cmd options
# Details:
#     Inputs - iscsi_obj (required)
#            - \% isns_hash (required) - isns data from get_isns_info_cmode
#     Returns $SanUtil::TRUE  - all items match
#             $SanUtil::FALSE - all items do NOT match
#             set_err = error executing a cmd or other error
################################################################################
sub _verify_isns {
    my ($self, %args) = @_;

    my $iscsi_obj;
    my $isns_hash;    # Hash of vserver info returned from get_isns_info_cmode

    $self->sanlog(-msg       => "START",
                  -msg_level => 1);

    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /iscsi_obj/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $iscsi_obj = $value;
        } elsif ($key =~ /isns_hash/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $isns_hash = $value;
        }
    } ## end while (my ($key, $value) ...)

    # Check mandatory parms
    unless (ref($iscsi_obj) =~
            /^San::ControllerCluster::Filer10M::Vserver::Iscsi$/i) {
        $self->_set_err(
                 "-iscsi_obj is a required parameter, must be an iSCSI object");
        return $SanUtil::FALSE;
    }
    unless (%{$isns_hash}) {
        $self->_set_err("-isns_hash is a required parameter.");
        return $SanUtil::FALSE;
    }

    # debug info.  Set Purity=1 for complete dumper output.
    $Data::Dumper::Purity = 1;
    $self->sanlog(-msg       => "isns hash:\n" . Dumper($isns_hash),
                  -msg_level => 1);
    $Data::Dumper::Purity = 0;

    # Note, decided not to create an err_label with the iscsi nodename,
    # because verify_isns_registration will include the nodename in the error
    # message.

    # Phase 1 - Verify target's iSCSI nodename
    $self->_verify_iscsi_scalar(-object => $iscsi_obj,
                                -type   => "nodename",
                                -value  => $isns_hash->{nodename});
    if ($self->error()) {

        # If the err msg does not contain 'mismatch', then the error needs
        # to be passed back to the caller.  There was a major error.
        # Note if the err msg does contain 'mismatch' we return without setting
        # error.  The caller will retry again in a minute.
        if ($self->errmsg() !~ /mismatch/i) {
            $self->_set_err("Failed to verify iscsi nodename!",
                            $self->errmsg());
        } else {
            $self->sanlog(-msg       => $self->errmsg(),
                          -msg_level => 1);
            $self->_clear_err();

        }
        return $SanUtil::FALSE;
    } ## end if ($self->error())

    # Phase 2 - Verify target's iSCSI alias.
    $self->_verify_iscsi_scalar(-object => $iscsi_obj,
                                -type   => "alias",
                                -value  => $isns_hash->{alias});
    if ($self->error()) {

        # If the err msg does not contain 'mismatch', then the error needs
        # to be passed back to the caller.  There was a major error.
        # Note if the err msg does contain 'mismatch' we return without setting
        # error.  The caller will retry again in a minute.
        if ($self->errmsg() !~ /mismatch/i) {
            $self->_set_err("Failed to verify iscsi alias!", $self->errmsg());
        } else {
            $self->sanlog(-msg       => $self->errmsg(),
                          -msg_level => 1);
            $self->_clear_err();
        }
        return $SanUtil::FALSE;
    } ## end if ($self->error())

    # Phase 3 - Verify iSCSI status
    $self->_verify_iscsi_hash(-object => $iscsi_obj,
                              -type   => "status",
                              -value  => $isns_hash->{status});
    if ($self->error()) {

        # If the err msg does not contain 'mismatch', then the error needs
        # to be passed back to the caller.  There was a major error.
        # Note if the err msg does contain 'mismatch' we return without setting
        # error.  The caller will retry again in a minute.
        if ($self->errmsg() !~ /mismatch/i) {
            $self->_set_err("Failed to verify iscsi status!", $self->errmsg());
        } else {
            $self->sanlog(-msg       => $self->errmsg(),
                          -msg_level => 1);
            $self->_clear_err();
        }
        return $SanUtil::FALSE;
    } ## end if ($self->error())

    # Phase 4 - Verify iSCSI show
    $self->_verify_iscsi_hash(-object => $iscsi_obj,
                              -type   => "show",
                              -value  => $isns_hash->{show});
    if ($self->error()) {

        # If the err msg does not contain 'mismatch', then the error needs
        # to be passed back to the caller.  There was a major error.
        # Note if the err msg does contain 'mismatch' we return without setting
        # error.  The caller will retry again in a minute.
        if ($self->errmsg() !~ /mismatch/i) {
            $self->_set_err("Failed to verify iscsi show!", $self->errmsg());
        } else {
            $self->sanlog(-msg       => $self->errmsg(),
                          -msg_level => 1);
            $self->_clear_err();
        }
        return $SanUtil::FALSE;
    } ## end if ($self->error())

    # Phase 5 - Verify iSCSI portals.
    # The portal command is aliased to the interface command.  We use the
    # interface verification routine with the portal flag set to 1.
    # We also pass in the interface hash because get_isns_info_cmode only returns
    # an interface hash and not a portal hash.  (There was no need to create
    # an identical portal hash, the interface hash is sufficient for both
    # portal and interface verification.)
    $self->_verify_iscsi_interfaces_from_client(
                                         -iscsi_obj  => $iscsi_obj,
                                         -interfaces => $isns_hash->{interface},
                                         -portal     => 1,
                                         -isns       => 1);
    if ($self->error()) {

        # If the err msg does not contain 'mismatch', then the error needs
        # to be passed back to the caller.  There was a major error.
        # Note if the err msg does contain 'mismatch' we return without setting
        # error.  The caller will retry again in a minute.
        # Also, if the err msg does not contain 'invalid value for "-interfaces"',
        # then we will return error.  If the above string is returned, then
        # will just print the msg and clear the error. This error can be
        # returned when you update the nodename for a vserver that is
        # currently registered. (burt559517 may help with this.)  We've got to
        # wait for the old nodename to de-register before the new one
        # registers properly. The caller will retry again in a minute.
        if ($self->errmsg() !~ /(mismatch|invalid value for \"-interfaces\")/i)
        {
            $self->_set_err("Failed to verify iscsi portals!", $self->errmsg());
        } else {
            $self->sanlog(-msg       => $self->errmsg(),
                          -msg_level => 1);
            $self->_clear_err();
        }
        return $SanUtil::FALSE;
    } ## end if ($self->error())

    # Phase 6 - Verify iSCSI interfaces
    $self->_verify_iscsi_interfaces_from_client(
                                         -iscsi_obj  => $iscsi_obj,
                                         -interfaces => $isns_hash->{interface},
                                         -isns       => 1);
    if ($self->error()) {

        # If the err msg does not contain 'mismatch', then the error needs
        # to be passed back to the caller.  There was a major error.
        # Note if the err msg does contain 'mismatch' we return without setting
        # error.  The caller will retry again in a minute.
        # Also, if the err msg does not contain 'invalid value for "-interfaces"',
        # then we will return error.  If the above string is returned, then
        # will just print the msg and clear the error. This error can be
        # returned when you update the nodename for a vserver that is
        # currently registered. (burt559517 may help with this.)  We've got to
        # wait for the old nodename to de-register before the new one
        # registers properly. The caller will retry again in a minute.
        if ($self->errmsg() !~ /(mismatch|invalid value for \"-interfaces\")/i)
        {
            $self->_set_err("Failed to verify iscsi interfaces!",
                            $self->errmsg());
        } else {
            $self->sanlog(-msg       => $self->errmsg(),
                          -msg_level => 1);
            $self->_clear_err();
        }
        return $SanUtil::FALSE;
    } ## end if ($self->error())

    # Phase 7 - Verify iSCSI tpgroups
    $self->_verify_iscsi_tpgroups_from_client(
                                             -iscsi_obj => $iscsi_obj,
                                             -tpgroups => $isns_hash->{tpgroup},
                                             -isns     => 1);
    if ($self->error()) {

        # If the err msg does not contain 'mismatch' or 'expected to match',
        # then the error needs to be passed back to the caller.  There was a major error.
        # Note if the err msg does contain 'mismatch' we return without setting
        # error.  The caller will retry again in a minute.
        # Also, if the err msg does not contain 'invalid value for "-tpgroups"',
        # then we will return error.  If the above string is returned, then
        # will just print the msg and clear the error. This error can be
        # returned when you update the nodename for a vserver that is
        # currently registered. (burt559517 may help with this.)  We've got to
        # wait for the old nodename to de-register before the new one
        # registers properly. The caller will retry again in a minute.
        if ($self->errmsg() !~ /(mismatch|invalid value for \"-tpgroups\")/i) {
            $self->_set_err("Failed to verify iscsi tpgroups!",
                            $self->errmsg());
        } else {
            $self->sanlog(-msg       => $self->errmsg(),
                          -msg_level => 1);
            $self->_clear_err();
        }
        return $SanUtil::FALSE;
    } ## end if ($self->error())

    # Phase 8 - Verify iSNS
    # First, we need to scan all ip's returned in the address hash.  We should
    # find one that matches the value in 'isns show', or else that's an error.
    my $isns_show = $iscsi_obj->isns_list_info();
    if ($iscsi_obj->error()) {
        $self->_set_err("isns show failed!", $self->errmsg());
        return $SanUtil::FALSE;
    }

    my $found = 0;

    # Scan each hash returned in 'address'
    foreach my $nic (keys %{$isns_hash->{isns}->{address}}) {

        # Does the ip in show match ipv4 or ipv6?
        if (($isns_show->{address} eq
             $isns_hash->{isns}->{address}->{$nic}->{ipv4})
            or ($isns_show->{address} eq
                $isns_hash->{isns}->{address}->{$nic}->{ipv6})
          ) {
            $self->sanlog(-msg => "isns show address "
                            . $isns_show->{address}
                            . " was found on the iSNS server.",
                          -msg_level => 1);
            $found = 1;
        } ## end if (($isns_show->{address...}))
    } ## end foreach my $nic (keys %{$isns_hash...})

    # If we didn't find a match for the ip, raise an error.  We set_err here
    # because we are not waiting on anything to propagate from target to
    # iSNS server.
    if (!$found) {
        $self->_set_err(  "isns show address "
                        . $isns_show->{address}
                        . " was NOT found on the iSNS server!");
        return $SanUtil::FALSE;
    }

    # We now need to delete the address hash so we can verify the rest of the
    # items.
    delete($isns_hash->{isns}->{address});

    # Check the rest of the isns hash against isns show.
    $self->_verify_iscsi_hash(-object => $iscsi_obj,
                              -type   => "isns",
                              -value  => $isns_hash->{isns});
    if ($self->error()) {

        # We set_err here because we are not waiting on anything to propagate
        # from target to iSNS server.
        $self->_set_err("Failed to verify iscsi show!", $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->sanlog(-msg       => "SUCCESS!",
                  -msg_level => 1);

    return $SanUtil::TRUE;
} ## end sub _verify_isns

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

IscsiHelperCM.pm

=head1 Author

Mike Nelson (mikenel@netapp.com) / Vipul Bhulawala (vipul@netapp.com)

=head1 Description

Provides helper methods for C-mode Iscsi package.

=head1 See Also

None.

=head1 Attributes

None

=head1 B<Public Methods>

=head2 B<new()>

Class Constructor method.

=over 4

=item Inputs

=item * I<-debug_level> (optional)

Sets the debug_level on the IscsiHelperCM obj.  Default is 1.

=item Outputs

Returns an IscsiHelperCM blessed object reference.

=item Examples

   # Create an IscsiHelperCM object
   my $iscsiHelperCM_obj = new San::Task::IscsiHelperCM();
   if ($iscsiHelperCM_obj->error()) {
      $Test->nlog("Failed to instantiate IscsiHelperCM object:", $iscsiHelperCM_obj->errmsg());
      return $TCD::FAIL;
   }

   # Create an IscsiHelperCM object with debug level
   my $iscsiHelperCM_obj = new San::Task::IscsiHelperCM(-debug_level => 5);
   if ($iscsiHelperCM_obj->error()) {
      $Test->nlog("Failed to instantiate IscsiHelperCM object:", $iscsiHelperCM_obj->errmsg());
      return $TCD::FAIL;
   }

=back

=head2 B<initialize_objects()>

It will initialize other library objects which are useful for
starting other activities. Activities include running IO or inject
events.

=over 4

=item Inputs

=item Outputs

$SanUtil::TRUE - all objects created successfully

$SanUtil::FALSE and set_err - object instantiatation failed and set_err

=item Example

  $IscsiHelperCM_obj->initialize_objects();

=back

=head2 B<get_ioman_obj(%args)>

Get the IOMan object that was initialized by IscsiHelperCM object. Use
initialize_objects() before you call get_ioman_obj().

=over 4

=item Inputs

=item * None

=item Outputs

On Success, IOMan_Obj

On error, undef and set_err

=item Examples

  $ioman_obj = $IscsiHelperCM_Obj->get_ioman_obj();

=back

=head2 B<get_eventman_obj(%args)>

Get the EventMan object that was initialized by IscsiHelperCM object. Use
initialize_objects() before you call get_eventman_obj().

=over 4

=item Inputs

=item * None

=item Outputs

On Success, EventMan_Obj

On error, undef and set_err

=item Examples

  $eventman_obj = $IscsiHelperCM_Obj->get_eventman_obj();

=back

=head2 B<verify_iscsi_clients(%args)>

Verifies one or more iscsi clients by calling _verify_iscsi_client_across_cluster
and _verify_iscsi_client.

=over 4

=item Inputs

=item * I<-clients> (optional)

An array of client names you wish to verify.  If you do not pass in clients,
then all iscsi clients in the SanMgr object will be verified.

=item * -sanmgr_obj (required)

The SanMgr object.

=item Outputs

$SanUtil::TRUE - All clients successfully verified.

$SanUtil::FALSE and set_err - Clients not verified or execution error.

=item Examples

  my $rc = $iscsiHelperCM_obj->verify_iscsi_clients(-clients    => \@clients,
                                                    -sanmgr_obj => $sanmgr_obj);

=back

=head2 B<verify_iscsi_expected(%args)>

This method is used by a test case to verify the vserver/iscsi objects contain
the expected configuration.

=over 4

=item Inputs

=item * <-sanmgr_obj> (required)

The SanMgr object where the vservers exist.

=item * -expected_hash (required)

Data structure containing the items to be verified.  This data is generated
within the testcase of a thpl.  For details on this data structure, check:
http://wikid/w/Scalable_SAN_Test_Engineering/CheatSheet/verify_iscsi_expected

=item Outputs

$SanUtil::TRUE - All items successfully verified.

$SanUtil::FALSE and set_err - All items not verified or execution error.

=item Examples

 $IscsiHelperCM_Obj->verify_iscsi_expected(-expected_hash => $exp_hash,
                                           -sanmgr_obj    => $SanMgr_Obj);
 if ($IscsiHelperCM_Obj->error()) {
     $Test->nlog("verify_iscsi_expected failure",
                 $IscsiHelperCM_Obj->errmsg());
     return $TCD::FAIL;
 }

=back

=head2 B<create_vserver_info(%args)>

Creates a hash that contains data specific to vservers
It creates an array of hashes. Each hash contains the following:
name        - contains scalar value, the name of the vserver
object      - contains vserver_obj
clients     - contains an array of scalars, the names of clients
              associated with the vserver
client_objs - contains an array of client_objs

=over 4

=item Inputs

=item * -sanmgr_obj (required)

SanMgr_Obj that creates the SAN resources

=item Outputs

On Success, an array (or reference to array) of vserver_info hash

On error, an empty array (or reference to array) and set_err

=item Examples

  @vserver_info = $IscsiHelperCM_Obj->create_vserver_info (
                                                -sanmgr_obj => $SanMgr_Obj);
  if ($IscsiHelperCM_Obj->error()) {
     $Test->nlog("Vserver Info not created", $IscsiHelperCM_Obj->errmsg());
     return $TCD::CONFIG;
  }


=back

=head2 B<get_vserver_info(%args)>

Get the hash that contains data specific to vservers
It is an array of hashes. Each hash contains the following:
name        - contains scalar value, the name of the vserver
object      - contains vserver_obj
clients     - contains an array of scalars, the names of clients
              associated with the vserver
client_objs - contains an array of client_objs

=over 4

=item Inputs

=item * None

=item Outputs

On Success, an array (or reference to array) of vserver_info hash

On error, an empty array (or reference to array) and set_err

=item Examples

  @vserver_info = $IscsiHelperCM_Obj->get_vserver_info();

=back

=head2 B<login_all_vserver_info_clients(%args)>

vserver_info contains info of all clients associated to each
vserver. Login all clients associated to each vserver from the
{vserver_info} data structure. Verify the login succeeded.

=over 4

=item Inputs

=item * -sanmgr_obj (required)

The SanMgr object.

=item * I<-sendtargets> (optional)

Do a refresh of sendtargets or not. By default, sendtargets info will be
refreshed.

=item * I<-fail_on_login> (optional)

If set to 1, ALL clients should fail to login to ALL vservers. The default
is set to 0, all clients should be able to login to all vservers.

=item * I<-fail_on_login_vserver_hash> (optional)

A hash of vserver names for which all clients are expected to fail login.  Used
for tests where you would expect all clients to successfully login to some
vservers, but fail to login to other vservers.

=item * I<-isns_address> (optional)

Perform isns discovery to this ip address instead of a sendtargets.

=item * I<-num_expected_luns> (optional)

An integer of number of luns that each client is expected to find.  If you
provide this, login speed is increased.

=item Outputs

On Success, $SanUtil::TRUE  - all clients login successfully

On error, $SanUtil::FALSE - login failed and set_err

=item Examples

  $IscsiHelperCM_Obj->login_all_vserver_info_clients(-sanmgr_obj => $sanmgr_obj,
                                                     -sendtargets => 0);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients login failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->login_all_vserver_info_clients(-sanmgr_obj => $sanmgr_obj,
                                                     -sendtargets => 1);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients login failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->login_all_vserver_info_clients(-sanmgr_obj=>$sanmgr_obj);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients login failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  # How to use fail_on_login_vserver_hash
  my $hash;
  $hash->{$vserver_name} = 1;
  $IscsiHelperCM_Obj->login_all_vserver_info_clients(
           -sanmgr_obj => $SanMgr_Obj,
           -fail_on_login_vserver_hash => $hash);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("login_all_vserver_info_clients failed",
                $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  # How to perform isns discovery
  $IscsiHelperCM_Obj->login_all_vserver_info_clients(-sanmgr_obj => $sanmgr_obj,
                                                     -isns_address => "1.2.3.4");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients login failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<logout_all_vserver_info_clients(%args)>

vserver_info contains info of all clients associated to each
vserver. Logout all clients associated to each vserver from the
{vserver_info} data structure. Verify the logout succeeded.

=over 4

=item Inputs

=item * -sanmgr_obj (required)

The SanMgr object.

=item * I<-keep_portals> (optional)

If keep_portal=1, leave the target portals on the client.
By default, keep_portals=0. Portals will be cleaned up.

=item Outputs

On Success, $SanUtil::TRUE  - all clients logout successfully

On error, $SanUtil::FALSE - logout failed and set_err

=item Examples

  $IscsiHelperCM_Obj->logout_all_vserver_info_clients(-sanmgr_obj => $sanmgr_obj,
                                                     -keep_portals => 0);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients logout failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->logout_all_vserver_info_clients(-sanmgr_obj => $sanmgr_obj,
                                                     -keep_portals => 1);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients logout failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->logout_all_vserver_info_clients(-sanmgr_obj=>$sanmgr_obj);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients logout failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<do_iscsi_setup(%args)>

This is a wrapper function to perform an iscsi service action on
all vservers found in {vserver_info}. iscsi can be created,
deleted, started or stopped using this method.

=over 4

=item Inputs

=item * -action (required)

Provide the iscsi action using the iscsi_obj for all vservers in
the {vserver_info} function
The valid actions are:
.a. create
.b. delete
.c. start
.d. stop

=item * -skip_license (optional)

Flag to indicate that while doing iscsi create, if iscsi license is not present,
license should not be added before creating iscsi on the vserver.

=item Outputs

On Success, $SanUtil::TRUE  - iscsi setup successful

On error, $SanUtil::FALSE - iscsi setup failed and set_err

=item Examples

  $IscsiHelperCM_Obj->do_iscsi_setup(-action => "create");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("iscsi action failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->do_iscsi_setup(-action       => "create",
                                       -skip_license => "1");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("iscsi create failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->do_iscsi_setup(-action => "delete");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("iscsi action failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->do_iscsi_setup(-action => "start");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("iscsi action failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->do_iscsi_setup(-action => "stop");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("iscsi action failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<cluster_ports_modify(%args)>

This is a wrapper function to modify port up-admin state.
Port up-admin can be modified to 'up' or 'down'.

=over 4

=item Inputs

=item * -sanmgr_obj (required)

SanMgr object

=item * -ports (required)

Provide the array of port names to be modified

=item * -nodes (optional)

Provide the array of node names to be modified

=item * <network_adapter->modify() args

Provide at least 1 arg that is used by 'network interface modify' cmd
Examples:
'-up-admin' - Provide the correct up-admin value
 for the data port. Valid values are true or false
-mtu (optional) - Provide the correct mtu value

=item Outputs

On Success, $SanUtil::TRUE  - port modify successful

On error, $SanUtil::FALSE - port modify failed and set_err

=item Examples

  $IscsiHelperCM_Obj->cluster_ports_modify(-sanmgr_obj => $sanmgr_obj,
                                          '-up-admin'  => "true",
                                          -ports => \@ports);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("port modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->cluster_ports_modify(-sanmgr_obj => $sanmgr_obj,
                                           '-up-admin' => "down",
                                          -ports => ["e0b","e0c"]);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("port modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->cluster_ports_modify(-sanmgr_obj => $sanmgr_obj,
                                           -mtu => 9000,
                                           -nodes => ["ssan-3170-1a"],
                                           -ports => \@ports);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("port modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<vserver_info_lifs_modify(%args)>

This is a wrapper function to modify ALL lifs on ALL vservers found in
{vserver_info}.

=over 4

=item Inputs

=item *  %lif_modify_params (required)

At least one modify action. It supports the exact same params that are used by
ControllerCluster::Filer10M::Vserver::Lif::modify()

=item * -vservers (optional)

Reference to an array of vserver names to apply as a filter to the
{vserver_info} data structure

=item * -ips (optional)
since each lif requires a unique ip address, instead of using -address,
provide a reference to an array of new ip addresses to modify
existing lif ip address.

=item Outputs

On Success, $SanUtil::TRUE  - {vserver_info} lifs modify successful

On error, $SanUtil::FALSE - lifs modify failed and set_err

=item Examples

  $IscsiHelperCM_Obj->vserver_info_lifs_modify('-status-admin' => "up");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("lif modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->vserver_info_lifs_modify('-home-node' => $to_node,
                                               '-status_admin' => "down");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("lif modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->vserver_info_lifs_modify('-vservers' => \@vserver_names,
                                               '-home-port' => "e0b");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("lif modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  my @ips = ("1.1.1.2", "1.1.1.3", "1.1.1.4", "1.1.1.5");
  $IscsiHelperCM_Obj->vserver_info_lifs_modify(-ips => \@ips);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("lif modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<lifs_modify(%args)>

This is a wrapper function to modify multiple lifs

=over 4

=item Inputs

=item *  %lif_modify_params (required)

Provide at least one lif modify action. It supports the exact same params that
are used by ControllerCluster::Filer10M::Vserver::Lif::modify()

=item * -lif_objs (required)

Provide a reference to an array of lif objects to do a modify action.

=item * -ips (optional)
since each lif requires a unique ip address, instead of using -address,
provide a reference to an array of new ip addresses to modify
existing lif ip address.

=item Outputs

On Success, $SanUtil::TRUE  - lifs modify successful

On error, $SanUtil::FALSE - lifs modify failed and set_err

=item Examples

  $IscsiHelperCM_Obj->lifs_modify( -lif_objs => \@lif_objs,
                                  '-status-admin' => "up");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("lif modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->lifs_modify( -lif_objs => \@lif_objs,
                                  '-home-node' => $to_node,
                                  '-status_admin' => "down");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("lif modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->lifs_modify( -lif_objs => \@lif_objs,
                                  '-home-port' => "e0b");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("lif modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->lifs_modify( -lif_objs => \@lif_objs,
                                   -ips => \@ips);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("lif modify failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<clients_cleanup(%args)>

Cleanup all clients associated to sanmgr_obj

=over 4

=item Inputs

=item * -sanmgr_obj (required)

The SanMgr object.

=item Outputs

On Success, $SanUtil::TRUE  - clients cleanup successful

On error, $SanUtil::FALSE - clients cleanup failed and set_err

=item Examples

  $IscsiHelperCM_Obj->clients_cleanup(-sanmgr_obj => $sanmgr_obj);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("failed to cleanup clients", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<create_client_info(%args)>

Builds a data structure containing information on each client defined by SanMgr.
It creates an array of hashes. Each hash contains the following:
name         - contains scalar value, the name of the client
object       - contains client_obj
vservers     - contains an array of scalars, the names of vservers
               associated with the client
vserver_objs - contains an array of vserver_objs
nodename     - client's iscsi nodename

=over 4

=item Inputs

=item * -sanmgr_obj (required)

SanMgr_Obj that creates the SAN resources

=item Outputs

On Success, an array (or reference to array) of client_info hash

On error, an empty array (or reference to array) and set_err

=item Examples

  @client_info = $IscsiHelperCM_Obj->create_client_info (
                                                -sanmgr_obj => $SanMgr_Obj);
  if ($IscsiHelperCM_Obj->error()) {
     $Test->nlog("Client Info not created", $IscsiHelperCM_Obj->errmsg());
     return $TCD::CONFIG;
  }


=back

=head2 B<get_client_info(%args)>

Get the array that contains data specific to clients
It is an array of hashes. Each hash contains the following:
name         - contains scalar value, the name of the client
object       - contains client_obj
vservers     - contains an array of scalars, the names of vservers
               associated with the client
vserver_objs - contains an array of vserver_objs

=over 4

=item Inputs

=item * None

=item Outputs

On Success, an array (or reference to array) of client_info hash

On error, an empty array (or reference to array) and set_err

=item Examples

  @client_info = $IscsiHelperCM_Obj->get_client_info();

=back

=head2 B<login_all_client_info_clients(%args)>

client_info contains info of all vservers associated to each
client. Login all vservers associated to each client from the
{client_info} data structure. Verify the login succeeded.

=over 4

=item Inputs

=item * -sanmgr_obj (required)

The SanMgr object.

=item * I<-sendtargets> (optional)

Do a refresh of sendtargets or not. By default, sendtargets info will be
refreshed.

=item * I<-fail_on_login> (optional)

If set to 1, ALL clients should fail to login to ALL vservers. The default
is set to 0, ALL clients should be able to login to all vservers.

=item * I<-fail_on_login_vserver_hash> (optional)

A hash of vserver names for which all clients are expected to fail login.  Used
for tests where you would expect the client to successfully login to some
vservers, but fail to login to other vservers.
NOTE: You cannot specify both fail_on_login = 1 and fail_on_login_vserver_hash.

=item * I<-isns_address> (optional)

Perform isns discovery to this ip address instead of a sendtargets.

=item * I<-clients> (optional)

An array of names of client which should login to the targets.

=item * I<-iscsi_targets> (optional)

An array of iscsi target objects to which the clients should login

=item * I<-num_expected_luns> (optional)

An integer of number of luns that each client is expected to find.  If you
provide this, login speed is increased.

=item Outputs

On Success, $SanUtil::TRUE  - all clients login successfully

On error, $SanUtil::FALSE - login failed and set_err

=item Examples

  $IscsiHelperCM_Obj->login_all_client_info_clients(-sanmgr_obj => $sanmgr_obj,
                                                    -sendtargets => 0);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients login failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->login_all_client_info_clients(-sanmgr_obj => $sanmgr_obj,
                                                    -sendtargets => 1);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients login failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->login_all_client_info_clients(-sanmgr_obj=>$sanmgr_obj);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients login failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  # Example with fail_on_login_vserver_hash
  my $hash;
  $hash->{$vserver_name} = 1;
  $IscsiHelperCM_Obj->login_all_client_info_clients(
             -sanmgr_obj => $SanMgr_Obj,
             -fail_on_login_vserver_hash => $hash);
  if ($IscsiHelperCM_Obj->error()) {
      $Test->nlog("login_all_client_info_clients failed",
                  $IscsiHelperCM_Obj->errmsg());
      return $TCD::FAIL;
  }

  # How to perform isns discovery
  $IscsiHelperCM_Obj->login_all_client_info_clients(-sanmgr_obj => $sanmgr_obj,
                                                    -isns_address => "1.2.3.4");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients login failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  #How to login to specific targets from a set of hosts.
  $IscsiHelperCM_Obj->login_all_client_info_clients(
                                               -sanmgr_obj    => $SanMgr_Obj,
                                               -iscsi_targets => \@target_array,
                                               -clients => ["ssan-w2k8-64-03"]);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients login failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<logout_all_client_info_clients(%args)>

client_info contains info of all vservers associated to each
client. Logout all vservers associated to each client from the
{client_info} data structure. Verify the logout succeeded.

=over 4

=item Inputs

=item * -sanmgr_obj (required)

The SanMgr object.

=item * I<-keep_portals> (optional)

If keep_portal=1, leave the target portals on the client.
By default, keep_portals=0. Portals will be cleaned up.

=item Outputs

On Success, $SanUtil::TRUE  - all clients logout successfully

On error, $SanUtil::FALSE - logout failed and set_err

=item Examples

  $IscsiHelperCM_Obj->logout_all_client_info_clients(-sanmgr_obj => $sanmgr_obj,
                                                     -keep_portals => 0);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients logout failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->logout_all_client_info_clients(-sanmgr_obj => $sanmgr_obj,
                                                     -keep_portals => 1);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients logout failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->logout_all_client_info_clients(-sanmgr_obj=>$sanmgr_obj);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clients logout failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<_verify_iscsi_client(%args)>

Verifies the client's view of the iscsi state returned from
SanClient::get_iscsi_info matches what the vserver sees.

=over 4

=item Inputs

=item * -iscsi_obj (required)

iSCSI Object

=item * -client_hash (required)

The client hash returned from SanClient::get_iscsi_info.  Specify the key of the
hash that corresponds to the iSCSI object that you are verifying.  See the
example below.

=item * -hosttype (required)

The output of SanClient::hosttype()

=item Outputs

$SanUtil::TRUE - client and vserver agree on the iscsi data.

$SanUtil::FALSE and set_err - client and vserver do NOT agree on the iscsi data.

=item Examples

  my $rc = $iscsiHelperCM_obj->_verify_iscsi_client(-iscsi_obj   => $iscsi_obj,
                                                    -client_hash => $iscsi_info->{$nodename},
                                                    -hosttype    => $client_obj->hosttype());

=back

=head2 B<_verify_iscsi_client_across_cluster(%args)>

Verifies iSCSI data totals across the cluster for a single client.
Method is called by the following:
- verify_iscsi_clients  - data from SanClient::get_iscsi_info
- verify_iscsi_expected - data from test case
Verifies the following:
- connection - total number in the cluster for one client
- initiator  - total number in the cluster for one client
- session    - total number in the cluster for one client

=over 4

=item Inputs

=item * -cserver_obj (required)

A cserver object from your cluster.

=item * -client_hash (optional)

client data from either:
- verify_iscsi_clients  - data from SanClient::get_iscsi_info
- verify_iscsi_expected - data from test case
If the hash is empty, the method verifies there are 0 sessions, connections,
and initiators in the cluster for the client.

=item * -initiator_iqn (required)

Initiator's iscsi nodename.

=item Outputs

$SanUtil::TRUE - client and cluster agree on the iscsi data.

$SanUtil::FALSE and set_err - client and cluster do NOT agree on the iscsi data.

=item Examples

  my $rc = $iscsiHelperCM_obj->_verify_iscsi_client_across_cluster(-cserver_obj   => $iscsi_obj,
                                                                   -client_hash   => $iscsi_info,
                                                                   -initiator_iqn => $iqn);

=back

=head2 B<client_info_run_io(%args)>

client_info contains info of all vservers associated to each
client. Run IO to devices associated to each client from the
{client_info} data structure. It accepts all the params that
IOMan::create_io_records() accepts.

=over 4

=item Inputs

=item * I<-enable_defaults> (optional)
Enable default values for IOMan params.
If enable_defaults is set to 1, <ioman_create_io_records_params> are set to
default recommended values. These params can be overridden by user passed
values.
For example, blocksize is randomly selected by default but if
the user passes -bs => 1k, 1k value will be used along with the
other defaults.
If enable_defaults is set to 0, the user wanted an entirely
different type of IO workload and it is expected that user
passes all the correct IOMan params.
By default, -enable_defaults is set to 1.

=item * I<-io_tool> (optional)

load generation tool type.
Default: dt, type:string, Allowed values: dt|sio

=item * I<-bs> (optional)

Block size in bytes. By default, if no block size is passed,
it will select a random blocksize from the known list of block sizes
that's important to test with.
An additional method for use with '-io_tool'=>'dt' is to set
block size to "variable_record_<min>_<max>"
<min> and <max> values are configurable (i.e. if want to write
blocks size of 4k to 6k use "variable_record_4k_6k").
Default: random, type:string

=item * I<%ioman_create_io_records_params> (optional)

Pass any IOMan specific params. See wiki
http://wikid.netapp.com/w/Scalable_SAN_Test_Engineering/Automation/IOMan/user_guide

=item * I<-clients> (optional)

Array reference to a list of client names.
This list will be used to filter running IO to devices of only
those clients that are part of the -clients list.

=item * I<-wait_finish> (optional)

Wait on finishing IO.
If set to 1, it will wait on finishing IO thereby making a call to this function
blocking. User will not be able to run anything in parallel while running IO
(unless it is subtest'd).
If set to 0, it will return immediately after starting IO. User will need to
make sure it has the IOMan_Obj via get_ioman_obj() and then run
ioman_obj->display_results() or ioman_obj->wait_finish().
Default is 1

=item Outputs

On Success, an array (or reference to array) of io_records hash

On error, either an array (or reference to array) of io_records hash (if
io_records were created successfully) or an empty array (or reference to array) and set_err

=item Examples

  #
  # Common config
  #

  # Initialize internal objects
  $IscsiHelperCM_Obj->initialize_objects();

  # Create Client Info internal data structure
  $IscsiHelperCM_Obj->create_client_info();

  # Discover devices
  $IscsiHelperCM_Obj->discover_devices_client_info_clients();

  #
  # Scenario 1
  #
  # Run IO in foreground with implicit enable_defaults=>1
  $IscsiHelperCM_Obj->client_info_run_io(-limit => "9g");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("IO failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  #
  # Scenario 2
  #
  # Run IO in foreground with implicit enable_defaults=>1 and io-tool=>sio
  $IscsiHelperCM_Obj->client_info_run_io(  -limit => "1g"
                                         , -io_tool => "sio");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("IO failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  #
  # Scenario 3
  #
  # Run IO in foreground to 2 clients and using variable blocksize
  # from 512 to 256k
  $IscsiHelperCM_Obj->client_info_run_io(  -limit => "1g"
                                         , -clients => [$client1, $client2],
                                         , -io_tool => dt,
                                         , -bs => "variable_record_512_256k");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("IO failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  #
  # Scenario 4
  #
  # Run IO in background with a higher reads workload using the tool 'sio'
  # Useful for running IO on all clients and running other tasks in parallel
  my %args = (-limit => "1g"
            , -wait_finish => 0,
            , -io_tool => "sio",
            , -read_perc => "90");
  $IscsiHelperCM_Obj->client_info_run_io(%args);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("IO failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }
  my $ioman_obj = $IscsiHelperCM_Obj->get_ioman_obj();
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Get IOMan object failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }
  while ($ioman_obj->is_running()) {
      $IscsiHelperCM_Obj->verify_clients();
      &SanUtil::nate_sleep(30);
  }

  # Do wait_finish and display results
  $ioman_obj->display_results();
  if ($ioman_obj->error()) {
      $Test->nlog("IO display results failed", $ioman_obj->errmsg());
      return $TCD::FAIL;
  }

  # Remove IO records from IOMan object
  $ioman_obj->cleanup();
  if ($ioman_obj->error()) {
      $Test->nlog("IO records cleanup failed", $ioman_obj->errmsg());
      return $TCD::FAIL;
  }

  #
  # Scenario 5
  #
  # Run IO in background with io_records
  # Useful for re-running same IO or use part of the io_records to run IO
  # to say one of the devices or one of the clients
  my %args = (-limit => "1g"
            , -wait_finish => 0,
            , -io_tool => "sio",
            , -read_perc => "90");
  my @io_records = $IscsiHelperCM_Obj->client_info_run_io(%args);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("IO failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }
  my $ioman_obj = $IscsiHelperCM_Obj->get_ioman_obj();
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Get IOMan object failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }
  while ($ioman_obj->is_running(-io_records => \@io_records)) {
      $IscsiHelperCM_Obj->verify_clients();
      &SanUtil::nate_sleep(30);
  }

  # Do wait_finish and display results
  $ioman_obj->display_results(-io_records => \@io_records);
  if ($ioman_obj->error()) {
      $Test->nlog("IO display results failed", $ioman_obj->errmsg());
      return $TCD::FAIL;
  }

  # Remove IO records from IOMan object
  $ioman_obj->cleanup(-io_records => \@io_records);
  if ($ioman_obj->error()) {
      $Test->nlog("IO records cleanup failed", $ioman_obj->errmsg());
      return $TCD::FAIL;
  }

=back

=head2 B<discover_devices_client_info_clients(%args)>

client_info contains info of all vservers associated to each client. Discover
devices associated to each client from the {client_info} data structure and
store them in their corresponding {client_info} entry.

=over 4

=item Inputs

=item * I<-sanmgr_obj> (optional)
When SanMgr_Obj configured and setup the clients, pass SanMgr_Obj to get the
devices for the clients to run IO to.

=item * I<-clients> (optional)

Array reference to list of client names.  This list will be used to filter
discovering of devices only for clients as part of the -clients list.

=item Outputs

On Success, $SanUtil::TRUE  - all devices discovered successfully and saved in
                              {client_info} hash

On error, $SanUtil::FALSE - get device failed and set_err

=item Examples

  $IscsiHelperCM_Obj->discover_devices_client_info_clients();
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Device Discovery failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->discover_devices_client_info_clients(
        -sanmgr_obj => $sanmgr_obj);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Device Discovery failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->discover_devices_client_info_clients(
        -clients => [$client1]);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Device Discovery failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->discover_devices_client_info_clients(
        -sanmgr_obj => $sanmgr_obj,
        -clients => [$client1]);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Device Discovery failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<get_ips(%args)>

Convert SanMgr compatible IP Range to an array of IP addresses.

=over 4

=item Inputs

=item * I<-ip_range> (required)
Pass a SanMgr compatible IP address range
Valid formats are X.X.X.[A-B] or X.X.[A-B].X or X.X.[A-B].[C-D]

=item Outputs

On Success, $SanUtil::TRUE  - an array (or reference to array) of IP addresses

On error, $SanUtil::FALSE - an empty array (or reference to array) and set_err

=item Examples

  $IscsiHelperCM_Obj->get_ips(-ip_range => "10.228.[128-130].121");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Get IP addresses failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->get_ips(-ip_range => "10.228.[128-130].[121-123]");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Get IP addresses failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->get_ips(-ip_range => "10.228.128.[121-125]");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Get IP addresses failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<register_eventman_resources(%args)>

Register cluster resources with EventMan which are useful for starting events
and monitoring activity. Resources include Controller Cluster and Clients

=over 4

=item Inputs

=item * I<-sanmgr_obj> (required)
Pass SanMgr_Obj to get the resources like cluster and clients

=item Outputs

On Success, $SanUtil::TRUE  - all resources were added successfully in EventMan

On error, $SanUtil::FALSE - register resources failed and set_err

=item Examples

  $IscsiHelperCM_Obj->register_eventman_resources(-sanmgr_obj => $sanmgr_obj):
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Register Eventman Resources failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<clean_eventman_events>

Reset EventMan object to start the next fresh event set.

=over 4

=item Outputs

On Success, $SanUtil::TRUE  - all events removed successfully

On error, $SanUtil::FALSE - cleaning events failed and set_err

=item Examples

  $IscsiHelperCM_Obj->clean_eventman_events();
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Clean Eventman events failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<run_cluster_fault>

Run EventMan based Cluster faults

=over 4

=item Inputs

=item * I<-enable_defaults> (optional)
Enable default values for EventMan params.
By default, enable_defaults is set.
If enable_defaults is set to 1, <eventman_add_events_params>
are set to default recommended values. These params can be
overridden by user passed values.
If enable_defaults is set to 0, the user wanted an entirely
different type of EventMan param values and it is expected that
the user passes all the correct EventMan params.

=item * I<-events> (optional)
Any cluster wide event supported by EventMan
Default: sfo_takeforcedgiveCm

=item * I<-runid> (optional)
Pass a runid to the EventMan subtest.
Default: EventMan_Event

=item * I<%eventman_add_events_params> (optional)
Pass any EventMan specific params. See EventMan POD for details.

=item * I<-async> (optional)
Run Events in background or make it a waiting call
If set to 1, it will wait on finishing events thereby making a call
to this function blocking. User will not be able to run anything
in parallel while running events (unless it is subtest'd).
If set to 0, it will return immediately after starting events. User
will need to make sure it has the EventMan_Obj via get_eventman_obj().
Default is 1.

=item Outputs

On Success, $SanUtil::TRUE  - Event ran successfully

On error, $SanUtil::FALSE - Event failed and set_err

=item Examples

  # Only Defaults
  $IscsiHelperCM_Obj->run_cluster_fault();
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Run cluster fault failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  # Custom Run ID (Logs will be stored under this runid directory
  $IscsiHelperCM_Obj->run_cluster_fault(-runid => "iscsi_io_ha_sfo");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Run cluster fault failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  # Wait on the fault to complete.
  $IscsiHelperCM_Obj->run_cluster_fault(-async => 0);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Run cluster fault failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  $IscsiHelperCM_Obj->run_cluster_fault(-events => ["sfo_rebootforcedgiveCm", "sfo_takeforcedgiveCm"]);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Run cluster fault failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<run_lif_fault>

Run EventMan based LIF fault

=over 4

=item Inputs

=item * I<-enable_defaults> (optional)
Enable default values for EventMan params.
By default, enable_defaults is set.
If enable_defaults is set to 1, <eventman_add_events_params>
are set to default recommended values. These params can be
overridden by user passed values.
If enable_defaults is set to 0, the user wanted an entirely
different type of EventMan param values and it is expected that
the user passes all the correct EventMan params.

=item * I<-lif_down_times> (optional)
It accepts a list of LIF down/up times.
Default: [60, 30, 20, 10, 5, 3, 2, 1]

=item * I<-runid> (optional)
Pass a runid to the EventMan subtest.
Default: EventMan_Event

=item * I<%eventman_add_events_params> (optional)
Pass any EventMan specific params. See EventMan POD for details.

=item * I<-async> (optional)
Run Events in background or make it a waiting call
If set to 1, it will wait on finishing events thereby making a call
to this function blocking. User will not be able to run anything
in parallel while running events (unless it is subtest'd).
If set to 0, it will return immediately after starting events. User
will need to make sure it has the EventMan_Obj via get_eventman_obj().
Default is 1.

=item Outputs

On Success, $SanUtil::TRUE  - Event ran successfully

On error, $SanUtil::FALSE - Event failed and set_err

=item Examples

  # Only Defaults
  $IscsiHelperCM_Obj->run_lif_fault();
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Run lif fault failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  # Custom Run ID (Logs will be stored under this runid directory
  $IscsiHelperCM_Obj->run_lif_fault(-runid => "iscsi_io_ha_lif_down_up");
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Run lif fault failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  # Wait on the fault to complete.
  $IscsiHelperCM_Obj->run_lif_fault(-async => 0);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Run lif fault failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  # Only pass one lif down time of 30 secs
  $IscsiHelperCM_Obj->run_lif_fault(-lif_down_times => [30]);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Run lif fault failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<run_port_fault>

Run Network interface (port) down/up fault

=over 4

=item Inputs

=item * I<-sanmgr_obj> (required)
Pass SanMgr_Obj to get the ports

=item * I<-cycle_count> (optional)
How many times to run the port fault
Default: 10

=item * I<-port_down_times> (optional)
It accepts a list of port down/up times.
Default: [60, 30, 20, 10, 5, 3, 2, 1]

=item Outputs

On Success, $SanUtil::TRUE  - Event ran successfully

On error, $SanUtil::FALSE - Event failed and set_err

=item Examples

  $IscsiHelperCM_Obj->run_port_fault(-sanmgr_obj => $sanmgr_obj);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Run port fault failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

  # Only pass one port down time of 30 secs
  $IscsiHelperCM_Obj->run_port_fault(-sanmgr_obj => $sanmgr_obj,
                                     -port_down_times => [30]);
  if ($IscsiHelperCM_Obj->error()) {
    $Test->nlog("Run port fault failed", $IscsiHelperCM_Obj->errmsg());
    return $TCD::FAIL;
  }

=back

=head2 B<get_iscsi_targets(%args)>

Returns a reference to an array or an  array of iscsi target objetcs .

=over 4

=item Inputs

=item * -iscsi_obj (required)

iSCSI Object

=item * -ip_array (required)

An array of IP addresses.

=item Outputs

On Success an array (or reference to array) of iscsi target objects

On error an empty array (or reference to array) and set_err

=item Examples

  my @temp_array =
      $IscsiHelperCM_Obj->get_iscsi_targets(-iscsi_obj => $iscsi_obj,
                                            -ip_array  => ["10.20.30.40","1.2.3.4"]);
      if ($IscsiHelperCM_Obj->error()) {
         $Test->nlog(ERR_LABEL . $test_prefix . "get_iscsi_targets failed",
                     $IscsiHelperCM_Obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_iscsi_targets_enabled(%args)>

This function scans each vserver in Vserver_Info.  It then runs 'iscsi interface show
-enabled true' on each vserver.  It saves all of the iscsi target objects and returns
them.  The iscsi target objects can then be used in our login/logout calls.
Returns a reference to an array or an  array of iscsi target objetcs .

=over 4

=item Inputs

=item * None

=item Outputs

On Success an array (or reference to array) of iscsi target objects

If no targets are found, an empty array (or reference to array) will be returned

=item Examples

  my @temp_array =
      $IscsiHelperCM_Obj->get_iscsi_targets_enabled();
      if ($IscsiHelperCM_Obj->error()) {
         $Test->nlog(ERR_LABEL . $test_prefix . "get_iscsi_targets_enabled failed",
                     $IscsiHelperCM_Obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_iscsi_targets_disabled(%args)>

This function scans each vserver in Vserver_Info.  It then runs 'iscsi interface show
-enabled false' on each vserver.  It saves all of the iscsi target objects and returns
them.  The iscsi target objects can then be used in our login/logout calls.
Returns a reference to an array or an  array of iscsi target objetcs .

=over 4

=item Inputs

=item * None

=item Outputs

On Success an array (or reference to array) of iscsi target objects

If no targets are found, an empty array (or reference to array) will be returned

=item Examples

  my @temp_array =
      $IscsiHelperCM_Obj->get_iscsi_targets_disabled();
      if ($IscsiHelperCM_Obj->error()) {
         $Test->nlog(ERR_LABEL . $test_prefix . "get_iscsi_targets_disabled failed",
                     $IscsiHelperCM_Obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<verify_isns_registration(%args)>

Verifies that a vserver's information on an iSNS server matches the actual
information on the vserver.

=over 4

=item Inputs

=item * I<-vserver_objs> (optional)

Array of vserver objs to verify. Default is to verify all vservers in vserver_info.

=item * -isns_server_obj (required)

iSNS Server object

=item * I<-force_isns_update> (optional)

-force_isns_update = 1: Execute 'isns update' before verification.
-force_isns_update = 0: Do NOT execute 'isns update' before verification. (default)

=item * I<-retries> (optional)

Number of times to retry if vserver not found or registration information is stale.
If force=0, Default is 10. If force=1, the number of retries is automatically
set to 1. It can take up to 15 secounds after an 'isns update' for info to be
propagated to the iSNS server.  So a retry count of 1 allows for this.

=item Outputs

$SanUtil::TRUE, Vserver's registration with iSNS server is verified.

$SanUtil::FALSE & set_err, Vserver did not register properly with iSNS Server
or there was an execution error.

=item Supported Interfaces

cli, zapi

=item Examples

$IscsiHelperCM_Obj->verify_isns_registration(-vserver_objs      => \@vserver_obj_array,
                                             -isns_server_obj   => $isns_server_obj,
                                             -force_isns_update => 1);

$IscsiHelperCM_Obj->verify_isns_registration(-isns_server_obj => $isns_server_obj,
                                             -retries         => 20);

=back

=head2 B<verify_isns_deregistration(%args)>

Verifies that a vserver is not registered with iSNS server.

=over 4

=item Inputs

=item * I<-vserver_objs> (optional)

Array of vserver objs to verify. Default is to verify all vservers in vserver_info.

=item * -isns_server_obj (required)

iSNS Server object

=item * I<-retries> (optional)

Number of times to retry if vserver is still registered.  Default is 10.

=item Outputs

Returns $SanUtil::TRUE, if the vserver is NOT registered with isns server.

Returns $SanUtil::FALSE and set_err means vserver is still registered or there
was another problem.

=item Examples

$IscsiHelperCM_Obj->verify_isns_deregistration(-vserver_objs      => \@vserver_obj_array,
                                               -isns_server_obj   => $isns_server_obj);

$IscsiHelperCM_Obj->verify_isns_deregistration(-isns_server_obj => $isns_server_obj,
                                               -retries         => 20);

=back

=head2 B<sleep_and_monitor_health(%args)>

Checks the health of each node every 15 seconds to make sure the nodes report their
health as 'true'.

=over 4

=item Inputs

=item * -sanmgr_obj (required)

The SanMgr object.

=item * I<-minutes> (optional)

Number of minutes to keep checking the health. (Default is 3)

=item Outputs

Returns $SanUtil::TRUE, if all nodes report good health during the entire number
of minutes requested.

Returns $SanUtil::FALSE, if a node did not report good health, or there was another
problem.

=item Examples

$IscsiHelperCM_Obj->sleep_and_monitor_health(-sanmgr_obj => $SanMgr_Obj,
                                             -minutes    => 10);

=back

=cut
