#
## @summary Sapphire Command and Utility library
#
## @status TESTME
##
## @author anishv@netapp.ocm
##
## Added Support for cable status UNKNOWN in cablemgr_tp
#
## @pod here

=head1 NAME

Sapphire

=head1 SYNOPSIS

    use Sapphire qw(function_name);

    my $host = host($SHELF);

    my $hash = function_name(host=>$host,...);

=head1 DESCRIPTION
    This is a library module for the sasadmin command

=head1 ROUTINES

=cut
## @change 03-19-2012 rpradeep: Modified the subroutine gem_cmd to support expander diag testcase
## @change 05-21-2012 rajkuma1: Modified the subroutine phy_report to support ruckus system
## @change 07-19-2012 vickie: Modified subroutine expmgr_tp for parsing inconstent expmgr_tp output in IOM6 and IOM6E
## @change 03-04-2013 ssrinath: Modified reboot function to support CMode

package Storage::Sapphire;

use strict;
use warnings;

use Carp qw(confess croak);
use Params::Validate qw(validate validate_pos SCALAR OBJECT);
use Storage::Sasadmin qw(sasadmin_expander_cli);
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}
use Tharn qw(say script_result host logcomment logresult param);

use base qw(Exporter);
use Utils qw(make_conn);
use Storage::Misc qw(filer_cmd);
use Data::Dumper;


our  @EXPORT_OK = qw(
            phy_report
            expctrl_tp
            expmgr_tp
            exp_report
            ses_statuspage
            reboot
            gem_cmd
            phy_disable
            phy_enable
            inc_phy_counters
            reset_error_counters
            get_drive_status
            ver
            get_psu_status
            get_fan_status
            get_temp_status
            ddump_general
            ddump_drvmgr
            get_midplanevpd
            dumpdrives
            ddump_expmgr
            ddump_acpmgr
            qsfpinfo
            cablemgr_tp
            phydump
            controller_state
            watchdog_tp
            ses_report_page2
            ses_report_page8081
            ses_report_page8283
            ddump_hid_byIndex
            storage_show_diskled
            psmi_dump_regs
            );

our $MAINT_PROMPT = "\\*\>";


=head2 phy_report

    my $ret = phy_report(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine returns a hash table of information of phy_report command
It returns hash on success and croaks on failure.

Example of reference returned by the function:
$VAR1 = {

          '33' => {

                    'phyRstEr' => '0',

                    'syncEr' => '0',

                    'PhyType' => 'HOST',

                    'crcEr' => '0',

                    'LogicalPhy' => '0',

                    'State' => '3.0G',

                    'dwEr' => '0',

                    'dpEr' => '0',

                    'ChgCnt' => '3'

                  },

          '32' => {

                    'phyRstEr' => '0',

                    'syncEr' => '0',

                    'PhyType' => 'HOST',

                    'crcEr' => '0',

                    'LogicalPhy' => '0',

                    'State' => '3.0G',

                    'dwEr' => '0',

                    'dpEr' => '0',

                    'ChgCnt' => '3'

                  },

}



The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub phy_report {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "phy_report";

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }

    #print "Output is  C: $output \n";

    if($output !~ /phyRstEr/){
        croak "Unexpected phy_report output: $output";
    }

    my @output = split(/\r*\n/,$output);

    foreach my $line (@output){

        $line =~ s/^\s+//g;
        my @element = split(/\s+/,$line);
            #print "LINE: $line \n";
        if($element[0] =~ /^DRV$|^HOST$|^IM$|^SIL$|^UKWN$/){
            #print "LINE: $line \n";
            #print "Elements are : $#element \n";
            #print "phy type is : $element[0]  \n";
            $hash{$element[2]}{PhyType} = $element[0];
            $hash{$element[2]}{LogicalPhy} = $element[1];
            $hash{$element[2]}{State} = $element[3];
            $hash{$element[2]}{ChgCnt} = $element[4];
            $hash{$element[2]}{dwEr} = $element[5];
            $hash{$element[2]}{dpEr} = $element[6];
            $hash{$element[2]}{syncEr} = $element[7];
            $hash{$element[2]}{phyRstEr} = $element[8];
            $hash{$element[2]}{crcEr} = $element[9];
        }
        elsif($element[0] =~ /^DRV\d+$/){
            #print "LINE: $line \n";
            #print "Elements are : $#element \n";
            my $drv = $element[0];
            $drv =~ s/DRV//g;
            #print "Drive is $drv \n";
            $hash{$element[1]}{PhyType} = 'DRV';
            $hash{$element[1]}{LogicalPhy} = $drv;
            $hash{$element[1]}{State} = $element[2];
            $hash{$element[1]}{ChgCnt} = $element[3];
            $hash{$element[1]}{dwEr} = $element[4];
            $hash{$element[1]}{dpEr} = $element[5];
            $hash{$element[1]}{syncEr} = $element[6];
            $hash{$element[1]}{phyRstEr} = $element[7];
            $hash{$element[1]}{crcEr} = $element[8];
        }
    }
    return \%hash ;
}

#---------------------------------------------------------------------------------------------
#   phy_report function END
#---------------------------------------------------------------------------------------------

=head2 expctrl_tp

    my $ret = expctrl_tp(host =>$host, conmode=>"shelfcli",timeout => $timeout,tp_code=>"1", params=>"$phy");

This routine injects expctrl level error injection using expctrl_tp command
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub expctrl_tp {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>30,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                tp_code => {type=>SCALAR},
                params => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "expctrl_tp $opts{tp_code} $opts{params}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
    }
    else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
        say($c,"\n");
    }
    print "Output is  C: $output \n";
    if($output =~ /Usage:|Failed to/){
        croak "expctrl_tp failed with : $output";
    }
    return $output;
}

=head2 expmgr_tp


This routine executes expmgr_tp command with various parameters passed and returns the output either in a string or as a reerence to an Hash based on the options/parameters passed to expmgr_tp.
See the Usage section below for more details and exapmles.

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expmgr_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=item board

(Optional) : Board number it should be either "0" or "1". Default is " ".

=item seqNum

(Optional) : Sequence number of the respective module. Default is " ". Required only for tp_code is "4".

=item options

(Optional) : Required only when tp_code is 11 to specify the Action parameter. Default is " ".


=item Usage

Example 1:
 my $ret = expmgr_tp(host =>$host, conmode=>"shelfcli",timeout => $timeout,tp_code=>"1", params=>"$phy");

        $ret contains the output string "

Example 2:

        $ret contains a reference to a nested Hash for tp_code 8.

 my $ret = expmgr_tp(host => $host, timeout => $TIMEOUT, connmode=> "inbandcli", filermode => $FILERMODE, AorB => "A", tp_code => "11", adapter_id => "1d",board => "0", shelf_id=> "0",opions => "1",params => "256");

        OUTPUT will be converted into an HASH as shown below for "tp_code 2 and 8".

Output is  C: *** Host Phy Profile ***
 BitWind RstWind ChgCnt dwErr dpErr syncEr phyRstEr crcEr
      60      60    120   120   120     60       60   120
                    CRC   PRP   LOS    RDD      IDD   PCD
                      1     1     1      1        1     1
*** Drive Phy Profile ***
 BitWind RstWind ChgCnt dwErr dpErr syncEr phyRstEr crcEr
      60      60    120   120   120     60       60   120
                    CRC   PRP   LOS    RDD      IDD   PCD
                      1     1     1      1        1     1
*** SIL Phy Profile ***
 BitWind RstWind ChgCnt dwErr dpErr syncEr phyRstEr crcEr
      60      60    120   120   120     60       60   120
                      1     1     1      1        1     1
$VAR1 = {
          'SIL' => {
                     'phyRstEr' => '60',
                     'syncEr' => '60',
                     'dpErr' => '120',
                     'crcEr' => '120',
                     'IDD' => '1',
                     'PCD' => '1',
                     'BitWind' => '60',
                     'RstWind' => '60',
                     'CRC' => '1',
                     'dwErr' => '120',
                     'ChgCnt' => '120',
                     'PRP' => '1',
                     'LOS' => '1',
                     'RDD' => '1'
                   },
       'Drive' => {
                       'phyRstEr' => '60',
                       'syncEr' => '60',
                       'dpErr' => '120',
                       'crcEr' => '120',
                       'IDD' => '1',
                       'PCD' => '1',
                       'BitWind' => '60',
                       'RstWind' => '60',
                       'CRC' => '1',
                       'dwErr' => '120',
                       'ChgCnt' => '120',
                       'PRP' => '1',
                       'LOS' => '1',
                       'RDD' => '1'
                     },
          'Host' => {
                      'phyRstEr' => '60',
                      'syncEr' => '60',
                      'dpErr' => '120',
                      'crcEr' => '120',
                      'IDD' => '1',
                      'PCD' => '1',
                      'BitWind' => '60',
                      'RstWind' => '60',
                      'CRC' => '1',
                      'dwErr' => '120',
                      'ChgCnt' => '120',
                      'PRP' => '1',
                      'LOS' => '1',
                      'RDD' => '1'
                    }
        };

Example :3

Returns the Hash for tp_code 2 output.
Output is  C: Expander 0 Information:
         Swap: No
         Updated: Yes
         ASA: No
         CLK status: OK
         Reg Access: OK

        $VAR1 = {
          'ASA' => 'No',
          'CLK' => 'OK',
          'Information' => '0',
          'Reg' => 'OK',
          'Swap' => 'No',
          'Updated' => 'Yes'
        };




=back

=cut
sub expmgr_tp {

        my (%opts) = validate(@_,{
                                host => {type=>OBJECT,isa=>'Hostrec'},
                                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                                timeout => {type=>SCALAR, default=>30,optional=>1},
                                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                                AorB => {type=>SCALAR, default=>undef,optional=>1},
                                tp_code => {type=>SCALAR},
                                board => {type=>SCALAR, default=>" ",optional=>1},
                                params => {type=>SCALAR, default=>" ",optional=>1},
                                seqNum => {type=>SCALAR, default=>" ",optional=>1},
                                options=>{type=>SCALAR, default=>"",optional=>1},

                }
        );

        my %hash = ();
        my $output;
        my $c;
        my $cmd;
        my $id;
        my $command = "expmgr_tp $opts{tp_code} $opts{board} $opts{params}";
        if ($opts{tp_code} eq "4") { $command = "expmgr_tp $opts{tp_code} $opts{seqNum} $opts{board}";}
        elsif ($opts{tp_code} eq "9" || "11") { $command = "expmgr_tp $opts{tp_code} $opts{board} $opts{params} $opts{options} ";}

#       Make Connection and build command based on input parameters

        if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
                $opts{conntype} = "#console";
                $c = make_conn($opts{host},$opts{conntype});
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
                if(!defined($opts{gemconn})){
                        croak "GEMTEST needs to be connected before we can use gemtest connmode";
                }
                $c = $opts{gemconn};
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
                if($opts{filermode} ne "normal"){
                        $opts{conntype} = "#console";
                }
                $c = make_conn($opts{host},$opts{conntype});
                if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
                        croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
                }
                if(!defined($opts{AorB})){
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
                }
                else{
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";                }

                if($opts{filermode} ne "normal"){
                        $cmd = "sasadmin expander_cli $id \"$command\"";
                }
    else{
                        $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
                }

        }

        else {
                croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
        }

#       Issue Command
        if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
                (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
        }
        elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
                (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
        }
        else{
                ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
                say($c,"\n");
        }
        print "Output is  C: $output \n";
        if($output =~ /Usage:|Failed to/){
                croak "expmgr_tp failed with : $output";
        }


        if ($opts{tp_code} eq "2"){

         my @output = split (/\r*\n/,$output);

        my @Information = grep /Information/,@output;
        my $Information= "@Information";
        my @information= split(/\s/,$Information);
        my $INFO=$information[1];

        my @Swap = grep /Swap/,@output;
        my $Swap = "@Swap";
        my @swap = split(/:\s/,$Swap);
        $Swap = $swap[1];

    my @Updated = grep /Updated/,@output;
        my $Updated = "@Updated";
        my @updated = split(/:\s/,$Updated);
        $Updated = $updated[1];

        my @ASA = grep /ASA/,@output;
        my $ASA = "@ASA";
        my @asa = split(/:\s/,$ASA);
        $ASA = $asa[1];

        my @CLK_status = grep /CLK/,@output;
        my $CLK_status = "@CLK_status";
        my @CLK = split(/:\s/,$CLK_status);
        my $CLK = $CLK[1];

        my @Reg_access = grep /Reg/,@output;
        my $Reg_access = "@Reg_access";
        my @Reg = split(/:\s/,$Reg_access);
        my $Reg = $Reg[1];

        my %OUTPUT = ( 'Information'=>"$INFO",
                'Swap'=>"$Swap",
                'Updated'=>"$Updated",
                'ASA'=>"$ASA",
                'CLK'=>"$CLK",
                'Reg'=>"$Reg" );
        return %OUTPUT;

        }
        elsif ($opts{tp_code} eq "8" )
        {
                my %Result;
                my %hash;
                my @output = split (/\r*\n/,$output);
                for( my $i=0;$i<=$#output;$i++)
                {
                if ($output[$i]=~/Phy\sProfile/)
                {
                  my @element = split(/\s+/,$output[$i+2]);
                  shift @element;
                  my @element1 = split(/\s+/,$output[$i+4]);
                  shift @element1;
          $hash{BitWind}=$element[0];
                  if ($element[1] =~ /(m)/) {
                      $hash{RstWind}=$element[2];
                      $hash{ChgCnt}=$element[4];
                      $hash{dwErr}=$element[5];
                      $hash{dpErr}=$element[6];
                      $hash{syncEr}=$element[7];
                      $hash{phyRstEr}=$element[8];
                      $hash{crcEr}=$element[9];
                  } else {
                      $hash{RstWind}=$element[1];
                      $hash{ChgCnt}=$element[2];
                      $hash{dwErr}=$element[3];
                      $hash{dpErr}=$element[4];
                      $hash{syncEr}=$element[5];
                      $hash{phyRstEr}=$element[6];
                      $hash{crcEr}=$element[7];
                  }
                  $hash{CRC}=$element1[0];
                  $hash{PRP}=$element1[1];
                  $hash{LOS}=$element1[2];
                  $hash{RDD}=$element1[3];
                  $hash{IDD}=$element1[4];
                  $hash{PCD}=$element1[5];
                }
                if ($output[$i]=~/Host/)
                {       my %host_hash=%hash;
                        $Result{Host}=\%host_hash;}
                elsif ($output[$i]=~/Drive/)
                {       my %Drive_hash=%hash;
                        $Result{Drive}=\%Drive_hash;}
                elsif ($output[$i]=~/SIL/)
                {       my %SIL_hash=%hash;
                        $Result{SIL}=\%SIL_hash;}
                elsif ($output[$i]=~/\*\*\*\sPhy Profile\s\*\*\*/)
                {
                        if ($opts{params} eq "0")
                        {$Result {Host}=\%hash;}
                        elsif ($opts{params} eq "1")
                        {$Result {Drive}=\%hash;}
                        elsif ($opts{params} eq "2")
                        {$Result {SIL}=\%hash;}

                }
                }

                return \%Result;

        }



return $output;

}



=head2 exp_report

    my $ret = exp_report(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine returns a hash table of information of exp_report command
It returns hash on success and croaks on failure.

Example of reference returned by the function:

$VAR1 = {

          'TotalPhyCount' => '36',

          'RegAccess' => 'OK',

          'ClkStatus' => 'OK'

        };



The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub exp_report {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "exp_report";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }


    #print "Output is  C: $output \n";

    if($output !~ /Expander Information:/){
        croak "Unexpected phy_report output: $output";
    }

    my @output = split(/\r*\n/,$output);

    foreach my $line (@output){
        $line =~ s/^\s+//g;
        #my @element = split(/\s+/,$line);
        if($line =~ /CLK status:/){
            my @element = split(/CLK status:\s*/,$line);
            $hash{ClkStatus} = $element[1];
        }
        elsif($line =~ /Reg Access:/){
            my @element = split(/Reg Access:\s*/,$line);
            $hash{RegAccess} = $element[1];
        }
        elsif($line =~ /Total phy count:/){
            my @element = split(/Total phy count:\s*/,$line);
            $hash{TotalPhyCount} = $element[1];
        }
    }
    return \%hash ;
}

=head2 ses_statuspage

    my $ret = ses_statuspage(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine returns a hash table of information of exp_report command
It returns hash on success and croaks on failure.

Example of reference returned by the function:

$VAR1 = {

          '33' => '33',

          '32' => '34',

          '90' => '00',

          '63' => '00',

          '21' => '59',

};


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub ses_statuspage {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                page_num => {type=>SCALAR},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ses_statuspage $opts{page_num}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }
    print "Output is  C: $output \n";
    my @output = split(/\s+|\r*\n/,$output);
    my $byte = 0;
    my $value;
    foreach $value (@output){
        if($value =~ /^(\d|[a-f])(\d|[a-f])$/){ # To make sure that the $value is hex and filter out the rest
            print "Byte $byte value is $value \n";
            $hash{$byte} = "$value";
            $byte++;
        }
    }
    return \%hash ;

}



=head2 reboot

    my $ret = reboot(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine reboots a module
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub reboot {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>30,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                params => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "reboot $opts{params}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id '$command'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id '$command'";
        }
    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        $output = $c->execute(command => $cmd);
    }
    elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
        filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);
    }
    else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},-noprompt,-prompt,"HA Service");
        print "Output is $output \n";
        if($output =~ /failed/){
            croak "Saw failed string during boot:\n $output";
        }
        say($c,"\n",-swallow);
    }
    #print "Output is  C: $output \n";
    if($output =~ /Usage:|Failed to/){
        croak "reboot failed with : $output";
    }
    return $output;
}


=head2 gem_cmd

    my $ret = gem_cmd(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine issues a command to the Sapphire shelf fw
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub gem_cmd {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                cmd => {type=>SCALAR},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                pmc_cmd => {type=>SCALAR, default=>undef,optional=>1},
                retry => {type=>SCALAR, default=>6 ,optional=>1},
                diag_cmd => {type=>SCALAR, default=>undef,optional=>1},
                prompt_check => {type=>SCALAR, default=>undef,optional=>1}
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = $opts{cmd};
    my $PMC_PROMPT = "^*0000>";

    if(defined($opts{pmc_cmd}) && ($opts{connmode} ne "shelfcli")){
        croak "PMC commands only work with shelfcli connection mode";
    }


    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "acpcli"){ # Connection when we are directly connected to the acp console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        #if($opts{filermode} ne "normal"){
        #   $opts{conntype} = "#console";
        #}

        #$c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        #if(!defined($opts{AorB})){
        #   $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        #}
        #else{
        #   $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        #}
        $output = sasadmin_expander_cli(host=>$opts{host},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});


    }
    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli , acpcli and gemtest";
    }


    if(($opts{connmode} eq "acpcli" ) && ($cmd eq "root")){
        (undef,undef,$output) = say($c, "$cmd", '-noprompt');
    }

    elsif($opts{connmode} eq "shelfcli"){
        if(defined($opts{pmc_cmd}) && ($cmd ne "gem") && ($cmd ne "al2o3")){
                (undef,undef,$output) = say($c,"$cmd",-prompt,"$PMC_PROMPT");
        }elsif(defined($opts{diag_cmd}) && defined ($opts{prompt_check})){
                    (undef,undef,undef,$output) = say($c, "help", -prompt, '\*> |\$');
            }
         elsif((defined($opts{diag_cmd})) && !defined ($opts {prompt_check})){
                    if($cmd eq "C"){
                        (undef,undef,$output) = say($c, "$cmd","-wait",".","-noprompt");
                    }
                    elsif ($cmd eq "exit"){
                        (undef,undef,$output) = say($c, "$cmd","-wait",".",-prompt,"^*Al2O3>");
                    }
                    else{
                                (undef,undef,$output) = say($c, "$cmd",-timeout, 120 ,"-wait",".","-prompt", '^\$');
                    }

        }

        else{
            (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
        }
    }
    else{
    }
    return $output;
}


=head2 phy_disable

    my $ret = phy_disable(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0",disable_reason=>"0");

This routine disables a phy


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy to be disabled.  If we pass "all" then it will disable all the enabled phys. We can also pass list of phys separated by , or space.

=item disable_reason

(Required) : Disable reason to disable a phy

Valid options are 0 to 16

=back

=cut

sub phy_disable {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
                disable_reason => {type=>SCALAR},
        }
    );
    my %disable_reason = (
            "0" => "DIS\/MAN_SES",
            "1" => "DIS\/NOT_PRESENT",
            "2" => "DIS\/BRST_LOS",
            "3" => "DIS\/LOS",
            "4" => "DIS\/BRST_RDD",
            "5" => "DIS\/RDD",
            "6" => "DIS\/BRST_IDD",
            "7" => "DIS\/IDD",
            "8" => "DIS\/BRST_PRP",
            "9" => "DIS\/PRP",
            "10" => "DIS\/BRST_PCD",
            "11" => "DIS\/PCD",
            "12" => "DIS\/BRST_CRC",
            "13" => "DIS\/CRC",
            "14" => "DIS\/MIR",
            "15" => "DIS\/MAN_SMP",
            "16" => "DIS\/CLK_FAULT",
            );


    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my @enabled_phys;
    my $dis_reason = $opts{disable_reason};
    unless(grep(/$dis_reason/,keys(%disable_reason))){
        logresult("script","$dis_reason is not a valid disable reason. Valid Disable reasons are from 0 to 16.");
    }
    my $err = 0;
    my $phy_report;
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        my @phys = keys(%$phy_report);
        foreach my $phy (@phys){
            if($phy_report->{$phy}->{State} =~ /3\.0G|1\.5G|6\.0G/){
                if($dis_reason == 14){
                    if($phy_report->{$phy}->{PhyType} eq "DRV"){
                        push @enabled_phys,$phy;
                    }
                }
                else{
                        push @enabled_phys,$phy;
                }
            }

        }
    }
    else{
        @enabled_phys = split(/,|\s+/,$opts{phy_id});
    }
    foreach my $phy (@enabled_phys){
        logcomment("Disabling phy $phy with reason $dis_reason => $disable_reason{$dis_reason}");
        if($opts{connmode} eq "shelfcli"){
            expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>2,params=>"$phy $dis_reason");
        }
        elsif($opts{connmode} eq "inbandcli"){
            expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>2,params=>"$phy $dis_reason}");
        }
        elsif($opts{connmode} eq "acpcli"){
        }
        elsif($opts{connmode} eq "inbandacp"){
        }
        else{
            logresult("script", "Unsupported connmode");
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@enabled_phys){
        unless($phy_report->{$phy}->{State} =~ /$disable_reason{$dis_reason}/){
                logcomment("FAIL => Unable to disable phy $phy with reason $disable_reason{$dis_reason}");
                $err = 1;
            }
        }
        if($err == 1){
            logresult("fatal", "Unexpected disable state. Please check the logs above");
        }
}

=head2 phy_enable

    my $ret = phy_enable(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0",enable_reason=>"2");

This routine enables a phy
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy to be enabled.  If we pass "all" then it will enable all the disabled phys. We can also pass list of phys separated by , or space.

=item disable_reason

(Optional) : To enable a phy which is disabled because of a particular disable reason. By default it will try to enable all the disabled phys

=item enable_reason

(Required) : To enable a phy with particular enable_reason. Valid options are  0 to 6.


=back

=cut

sub phy_enable {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
                enable_reason => {type=>SCALAR},
                disable_reason => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    # Hash of supported enable reasons
    my %enable_reason = (
        "0" => "MANUAL_SES",
        "1" => "DRIVE_INSERT",
        "2" => "CABLE_INSERT",
        "3" => "PARTNER_EBOD_INSERT",
        "4" => "DRIVE_POWER_ON",
        "5" => "MIRROR",
        "6" => "MANUAL_SMP",
    );

    my %disable_reason = (
            "0" => "DIS\/MAN_SES",
            "1" => "DIS\/NOT_PRESENT",
            "2" => "DIS\/BRST_LOS",
            "3" => "DIS\/LOS",
            "4" => "DIS\/BRST_RDD",
            "5" => "DIS\/RDD",
            "6" => "DIS\/BRST_IDD",
            "7" => "DIS\/IDD",
            "8" => "DIS\/BRST_PRP",
            "9" => "DIS\/PRP",
            "10" => "DIS\/BRST_PCD",
            "11" => "DIS\/PCD",
            "12" => "DIS\/BRST_CRC",
            "13" => "DIS\/CRC",
            "14" => "DIS\/MIR",
            "15" => "DIS\/MAN_SMP",
            "16" => "DIS\/CLK_FAULT",
            );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my @disabled_phys;
    my $en_reason = $opts{enable_reason};
    unless(grep(/$en_reason/,keys(%enable_reason))){
        logresult("script","$en_reason is not a valid enable reason. Valid Disable reasons are from 0 to 6.");
    }
    my $err = 0;
    my $phy_report;
    my $phy_report1;
    my $dis_reason = "DIS";
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        my @phys = keys(%$phy_report);
        foreach my $phy (@phys){
            if(defined($opts{disable_reason})){
                    my $reason = $opts{disable_reason};
                    $dis_reason = $disable_reason{$reason};
                }
            if($phy_report->{$phy}->{State} =~ /$dis_reason/){
                if($phy_report->{$phy}->{PhyType} ne 'SIL'){
                    push @disabled_phys,$phy;
                }
            }
        }
    }
    else{
        @disabled_phys = split(/,|\s+/,$opts{phy_id});
    }
    print "DISABLED phys : @disabled_phys \n";
    sleep 30;
    foreach my $phy (@disabled_phys){

        if($en_reason =~ /1|2|3/){
            if($phy_report->{$phy}->{PhyType} eq 'DRV'){
                $en_reason = 1;
            }
            elsif($phy_report->{$phy}->{PhyType} eq 'HOST'){
                $en_reason = 2;
            }
            elsif($phy_report->{$phy}->{PhyType} eq 'SIL'){
                $en_reason = 3;
            }
        }

        logcomment("Enabling phy $phy with reason $en_reason => $enable_reason{$en_reason}");
        if($opts{connmode} eq "shelfcli"){
            eval { expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>1,params=>"$phy $en_reason");
            };
            if($@){
                logcomment("Phy enable failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "inbandcli"){
            eval {expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>1,params=>"$phy $en_reason}");
            };
            if($@){
                logcomment("Phy enable failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "acpcli"){
        }
        elsif($opts{connmode} eq "inbandacp"){
        }
        else{
            logresult("script", "Unsupported connmode");
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@disabled_phys){
        unless($phy_report1->{$phy}->{State} =~ /3\.0G|1\.5G|6\.0G/){
                logcomment("Unable to enable phy $phy with reason $enable_reason{$en_reason}");
                if($opts{phy_id} ne "all"){
                    $err = 1;
                }
            }
        }
        if($err == 1){
            logresult("fatal", "Unexpected enabled state. Please check the logs above");
        }
}

=head2 inc_phy_counters

    my $ret = inc_phy_counters(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0");

This routine increments phys counters.
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy for which the errors to be incremented.  If we pass "all" then it will increment counters of all the phys. We can also pass a list of phys separated by , or space.

=item count_type

(Optional) : Error to be incremented. Valid options are 0 to 5 and "all".

=item count

(Optional) : How many counters to increment. Default is 10.


=back

=cut

sub inc_phy_counters {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
                count_type => {type=>SCALAR},
                count => {type=>SCALAR},
        }
    );

    # Hash of supported enable reasons
    my %count_type = (
        "0" => "ChgCnt",
        "1" => "dwEr",
        "2" => "dpEr",
        "3" => "syncEr",
        "4" => "phyRstEr",
        "5" => "crcEr",
    );

    my %disable_reason = (
            "0" => "DIS\/MAN_SES",
            "1" => "DIS\/NOT_PRESENT",
            "2" => "DIS\/BRST_LOS",
            "3" => "DIS\/LOS",
            "4" => "DIS\/BRST_RDD",
            "5" => "DIS\/RDD",
            "6" => "DIS\/BRST_IDD",
            "7" => "DIS\/IDD",
            "8" => "DIS\/BRST_PRP",
            "9" => "DIS\/PRP",
            "10" => "DIS\/BRST_PCD",
            "11" => "DIS\/PCD",
            "12" => "DIS\/BRST_CRC",
            "13" => "DIS\/CRC",
            "14" => "DIS\/MIR",
            "15" => "DIS\/MAN_SMP",
            "16" => "DIS\/CLK_FAULT",
            );

    my $c;
    my $cmd;
    my $id;
    my @er_phys;
    my @count_type;
    if($opts{count_type} eq "all"){
        @count_type = keys(%count_type);
    }
    else{
        @count_type = split(/,|\s+/,$opts{count_type});
        foreach my $cnt (@count_type){
            unless(grep(/$cnt/,keys(%count_type))){
            logresult("script","$cnt is not a valid error counter. Valid Disable reasons are from 0 to 5.");
            }
        }
    }
    my $err = 0;
    my $phy_report;
    my $phy_report1;
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        @er_phys = keys(%$phy_report);
    }
    else{
        @er_phys = split(/,|\s+/,$opts{phy_id});
    }
    foreach my $phy (@er_phys){
        foreach my $count_type (@count_type){
            my $cnt_type = $count_type{$count_type};
            logcomment("Incrementing counter $count_type =>  $cnt_type for phy $phy by $opts{count}");
            if($opts{connmode} eq "shelfcli"){
                eval { expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>4,params=>"$phy $count_type $opts{count}");
                };
                if($@){
                    logcomment("Error Increment failed with error: $@");
                }
            }
            elsif($opts{connmode} eq "inbandcli"){
                eval {expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>4,params=>"$phy $count_type $opts{count}}");
                };
                if($@){
                    logcomment("Error increment failed with error: $@");
                }
            }
            elsif($opts{connmode} eq "acpcli"){
            }
            elsif($opts{connmode} eq "inbandacp"){
            }
            else{
                logresult("script", "Unsupported connmode");
            }
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@er_phys){
        foreach my $count_type (@count_type){
            my $cnt_type = $count_type{$count_type};
            my $cnt = $phy_report->{$phy}->{$cnt_type};
            my $cnt1 = $phy_report1->{$phy}->{$cnt_type};
            my $cnt2 = $cnt + $opts{count};
            if($cnt1 != $cnt2){
                logcomment("FAIL: Error $count_type => $cnt_type for phy $phy is expected $cnt2 but actual is $cnt1");
                $err = 1;
            }
        }
    }
        if($err == 1){
            print Dumper($phy_report1);
            print Dumper($phy_report);
            logresult("fatal", "Unexpected enabled state. Please check the logs above");
        }
}

=head2 reset_error_counters

    my $ret = reset_error_counters(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0");

This routine resets error counters of a phy or a set of phys
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy(s) for which the errors to be reset.  If we pass "all" then it will reset counters of all the phys. We can also pass list of phys separated by , or space.


=back

=cut
sub reset_error_counters{

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
        }
    );

    # Hash of supported enable reasons
    my %count_type = (
        "0" => "ChgCnt",
        "1" => "dwEr",
        "2" => "dpEr",
        "3" => "syncEr",
        "4" => "phyRstEr",
        "5" => "crcEr",
    );

    my $c;
    my $cmd;
    my $id;
    my @er_phys;
    my $err = 0;
    my $phy_report;
    my $phy_report1;
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        @er_phys = keys(%$phy_report);
    }
    else{
        @er_phys = split(/,|\s+/,$opts{phy_id});
    }
    foreach my $phy (@er_phys){
        logcomment("Clearing counters for phy $phy");
        if($opts{connmode} eq "shelfcli"){
            eval { expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>3,params=>"$phy");
            };
            if($@){
                logcomment("Error Increment failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "inbandcli"){
            eval {expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>3,params=>"$phy");
            };
            if($@){
                logcomment("Error increment failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "acpcli"){
        }
        elsif($opts{connmode} eq "inbandacp"){
        }
        else{
            logresult("script", "Unsupported connmode");
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@er_phys){
        foreach my $cnt_type (keys(%count_type)){
            my $cnt = $phy_report1->{$phy}->{$count_type{$cnt_type}};
            print "CNT TYPE: $cnt_type and CNT is $cnt \n";
            if(($count_type{$cnt_type} ne "ChgCnt") && ($cnt != 0)){
                logcomment("FAIL: Error counter $count_type{$cnt_type} for phy $phy is $cnt . Expected zero");
                $err = 1;
            }
        }
    }
        if($err == 1){
            logresult("fatal", "Unexpected enabled state. Please check the logs above");
        }
}

=head2 get_drive_status

    my $ret = get_drive_status(drive=><drive>,host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine retrieves a specific drive's status by issuing shelf command "getdrivestatus" and
    save in a hash table with the following keys stated:
    present, swap, SES_info_bit, SpinUpTime, drive_type, name, arrayname, wwn, faults,  faultLED, inject, pending, current

    Example of returning a specific drive's status:
    my $drivestatus = get_drive_status($drive);
    if ($drivestatus->{present} =~ /yes/) {
        print "the drive $drive is present\n";
    }

    Example of reference returned by the function:
    $VAR1 = {
          'present' => 'yes',
          'swap' => 'no',
          'SES_info_bit' => 'not set',
          'SpinUpTime' => '0+00:00:08.087',
          'drive_type' => 'SAS',
          'wwn' => '500000E0173F6792',
          'faults' => 'none',
          'faultLED' => 'OFF',
          'inject' => 'NONE',
          'pending' => 'ONLINE',
          'current' => 'ONLINE',
          'PwrCycleTime' => '0',
          'PwrInProgress' => 'NO',
          'InProgressAS' => 'NONE',
          'InProgressTS' => 'NONE',
          'ReservedCntr' => '0 (00000000)',
          'SMPPwrOnCnt' => '0',
          'SMPPwrOffCnt' => '0',
          'SMPPwrCycCnt' => '0',
          'ActiveCounter' => '0',
          'ActiveCntrAS' => 'NONE',
          'ActiveCntrTS' => 'NONE',
    };

The following arguments are accepted:

=over

=item drive

(Required) : The drive which status will be returned.

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_drive_status {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                disk_id => {type=>SCALAR, default=>undef},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "getdrivestatus $opts{disk_id}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /Drive Bay/){
        croak "Unexpected getdrivestatus output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output drive opts{disk_id} status: @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /present/) {
            my @element = split(/:/,$line);
            $hash{present} = $element[1];
        }
        elsif($line =~ /swap/) {
            my @element = split(/:/,$line);
            $hash{swap} = $element[1];
        }
        elsif($line =~ /SES_info_bit/) {
            my @element = split(/:/,$line);
            $hash{SES_info_bit} = $element[1];
        }
        elsif($line =~ /spin up time/) {
            my @element = split(/:/,$line);
            $hash{SpinUpTime} = $element[1];
        }
        elsif($line =~ /drive_type/) {
            my @element = split(/:/,$line);
            $hash{drive_type} = $element[1];
        }
        elsif($line =~ /WWN/) {
            my @element = split(/:/,$line);
            $hash{wwn} = $element[1];
        }
        elsif($line =~ /faults/) {
            my @element = split(/:/,$line);
            $hash{faults} = $element[1];
        }
        elsif($line =~ /fault LED/) {
            my @element = split(/:/,$line);
            $hash{faultLED} = $element[1];
        }
        elsif($line =~ /inject/) {
            my @element = split(/:/,$line);
            $hash{inject} = $element[1];
        }
        elsif($line =~ /pending/) {
            my @element = split(/:/,$line);
            $hash{pending} = $element[1];
        }
        elsif($line =~ /current/) {
            my @element = split(/:/,$line);
            $hash{current} = $element[1];
        }
        elsif($line =~ /Pwr Cycle Time/) {
            my @element = split(/:/,$line);
            $hash{PwrCycleTime} = $element[1];
        }
        elsif($line =~ /Pwr InProgress/) {
            my @element = split(/:/,$line);
            $hash{PwrInProgress} = $element[1];
        }
        elsif($line =~ /InProgressAS/) {
            my @element = split(/:/,$line);
            $hash{InProgressAS} = $element[1];
        }
        elsif($line =~ /InProgressTS/) {
            my @element = split(/:/,$line);
            $hash{InProgressTS} = $element[1];
        }
        elsif($line =~ /ReservedCntr/) {
            my @element = split(/:/,$line);
            $hash{ReservedCntr} = $element[1];
        }
        elsif($line =~ /SMP PwrOn cnt/) {
            my @element = split(/:/,$line);
            $hash{SMPPwrOnCnt} = $element[1];
        }
        elsif($line =~ /SMP PwrOff cnt/) {
            my @element = split(/:/,$line);
            $hash{SMPPwrOffCnt} = $element[1];
        }
        elsif($line =~ /SMP PwrCyc cnt/) {
            my @element = split(/:/,$line);
            $hash{SMPPwrCycCnt} = $element[1];
        }
        elsif($line =~ /Active Counter/) {
            my @element = split(/:/,$line);
            $hash{ActiveCounter} = $element[1];
        }
        elsif($line =~ /Active Cntr.AS/) {
            my @element = split(/:/,$line);
            $hash{ActiveCntrAS} = $element[1];
        }
        elsif($line =~ /Active Cntr.TS/) {
            my @element = split(/:/,$line);
            $hash{ActiveCntrTS} = $element[1];
        }
    }
    return \%hash;
}

=head2  ver

    my $ret = ver(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine retrieves all the components firmware versions by issuing shelf command
    "ver" and save in a hash table with following keys stated:
    CanisterFW, CanisterFWDate, CanisterBootloader, CanisterConfig CanisterVPD, CanisterVPDCRC.
    CanisterCPLD. CanisterChip. CanisterSDK, MidplaneVPD, MidplaneVPDCRC. MidplaneCPLD. PCM1FW.
    PCM2FW. PCM3FW. PCM4FW, PCM1VPD, PCM2VPD, PCM3VPD, PCM4VPD, PCM1VPDCRC, PCM2VPDCRC, PCM3VPDCRC, PCM4VPDCRC

    Example of returning version hash table:
    my $tb = ver();
    print "The current Canister Firmware is $tb->{CanisterFW}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'CanisterFW' => '0.A6',
          'CanisterFWDate' => 'Nov 26 2008 04:31:26',
          'CanisterBootloader' => '4.00',
          'CanisterConfigCRC' => 'Not present',
          'CanisterVPDStruc' => '0x06',
          'CanisterVPDCRC' => '0x0E4B5E5A',
          'CanisterCPLD' => '0x12',
          'CanisterChip' => '0x83980000',
          'CanisterSDK' => '2.03.02',
          'MidplaneVPDStruc' => '0x07',
          'MidplaneVPDCRC' => '0x3F8C62FB',
          'MidplaneCPLD' => '0x0E',
          'PCM1FW' => '2.09',
          'PCM2FW' => '2.09',
          'PCM3FW' => '2.09',
          'PCM4FW' => '2.09',
          'PCM1VPDStruc' => '0x02',
          'PCM2VPDStruc' => '0x02',
          'PCM3VPDStruc' => '0x02',
          'PCM4VPDStruc' => '0x02',
          'PCM1VPDCRC' => '0x60CFC396',
          'PCM2VPDCRC' => '0x60CFC396',
          'PCM3VPDCRC' => '0x60CFC396',
          'PCM4VPDCRC' => '0x60CFC396',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ver {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ver";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /Canister firmware/){
        croak "Unexpected ver output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Canister firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterFW} = $element[1];
        }
        elsif($line =~ /Canister firmware date\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterFWDate} = $element[1];
        }
        elsif($line =~ /Canister bootloader\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterBootloader} = $element[1];
        }
        elsif($line =~ /Canister config CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterConfigCRC} = $element[1];
        }
        elsif($line =~ /Canister VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterVPDStruc} = $element[1];
        }
        elsif($line =~ /Canister VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterVPDCRC} = $element[1];
        }
        elsif($line =~ /Canister CPLD\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterCPLD} = $element[1];
        }
        elsif($line =~ /Canister chip\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterChip} = $element[1];
        }
        elsif($line =~ /Canister SDK\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterSDK} = $element[1];
        }
        elsif($line =~ /Midplane VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{MidplaneVPDStruc} = $element[1];
        }
        elsif($line =~ /Midplane VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{MidplaneVPDCRC} = $element[1];
        }
        elsif($line =~ /Midplane CPLD\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{MidplaneCPLD} = $element[1];
        }
        elsif($line =~ /PCM 1 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM1FW} = $element[1];
        }
        elsif($line =~ /PCM 2 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM2FW} = $element[1];
        }
        elsif($line =~ /PCM 3 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM3FW} = $element[1];
        }
        elsif($line =~ /PCM 4 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM4FW} = $element[1];
        }
        elsif($line =~ /PCM 1 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM1VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 2 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM2VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 3 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM3VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 4 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM4VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 1 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM1VPDCRC} = $element[1];
        }
        elsif($line =~ /PCM 2 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM2VPDCRC} = $element[1];
        }
        elsif($line =~ /PCM 3 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM3VPDCRC} = $element[1];
        }
        elsif($line =~ /PCM 4 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM4VPDCRC} = $element[1];
        }
    }
    return \%hash;
}

=head2  get_psu_status

    my $ret = get_psu_status(psu=><psu>,host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "getpsustatus" to retrieve a specific power supply unit's
    status and save in a hash table with following keys stated:
    present, type, fruid, productName, manufacturer, partNumber, partRevision, serialNumber, firmwareVersion, powerState, fansSelfPowered, nominalPowerRating, nominalPowerRatingOutputBitmask, surgePowerRating, surgePowerRatingOutputBitmask, surgeHoldTime, ACDropoutTolerance, standbyPower, coolingPower, swap, hotswapPrivate, hotswapPublic, fault,
        output1 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent
        output2 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent
        output3 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent
    combined_power, ses_info_bit.

    Example of returning a psu's status:
        my $psustatus = get_psu_status(1);
        print "the power supply 1 is present?  $psustatus->{present}\n";
        print "the power supply 1 output2 min voltage is: $psustatus->{output2}{minVoltage}\n";

    present, type, fruid, productName, manufacturer, partNumber, partRevision, serialNumber, firmwareVersion, powerState, fansSelfPowered, nominalPowerRating, nominalPowerRatingOutputBitmask, surgePowerRating, surgePowerRatingOutputBitmask, surgeHoldTime, ACDropoutTolerance, standbyPower, coolingPower, swap, hotswapPrivate, hotswapPublic, fault,
        output1 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent

    Example of reference returned by the function:
    $VAR1 = {
          'present' => 'yes',
          'type' => 'PSMI',
          'fruid' => '0x9C',
          'productName' => 'HB-PCM-01-580AC',
          'manufacturer' => 'POWERONE',
          'partNumber' => '0082562',
          'partRevision' => '07',
          'serialNumber' => 'PMW8256200043ED',
          'firmwareVersion' => '2.09',
          'powerState' => 'on',
          'fansSelfPowered' => 'yes',
          'nominalPowerRating' => '580W',
          'nominalPowerRatingOutputBitmask' => '0x0',
          'surgePowerRating' => '580W',
          'surgePowerRatingOutputBitmask' => '0x0',
          'surgeHoldTime' => '0ms',
          'ACDropoutTolerance' => '0ms',
          'standbyPower' => '10W',
          'coolingPower' => '60W',
          'swap' => 'no',
          'hotswapPrivate' => 'yes',
          'hotswapPublic' => 'yes',
          'fault' => 'no',
          'output#' 'voltage' => '5.11V',
          'output#' 'current' => '3.43A',
          'output#' 'power' => '17.58W',
          'output#' 'minVoltage' => '4.80V',
          'output#' 'minSurgeVoltage' => '4.80V',
          'output#' 'maxVoltage' => '5.19V',
          'output#' 'maxCurrent' => '36.00A',
          'output#' 'minCurrent' => '0.00A',
          'output#' 'maxSurgeCurrent' => '42.00A',
    }

The following arguments are accepted:

=over

=item psu

(Required) : Which power supply status to be retrieved.

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_psu_status {
    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                psu => {type=>SCALAR},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "getpsustatus $opts{psu}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /PCM Bay/){
        croak "Unexpected getpsustatus output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output psu $opts{psu} status: @output \n";
    my $output_num = 0;
    my $output_full ="";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /present:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{present} = $element[1];
        }
        elsif($line =~ /type:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{type} = $element[1];
        }
        elsif($line =~ /FRUid:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{fruid} = $element[1];
        }
        elsif($line =~ /product name:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{productName} = $element[1];
        }
        elsif($line =~ /manufacturer:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{manufacturer} = $element[1];
        }
        elsif($line =~ /part number:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partNumber} = $element[1];
        }
        elsif($line =~ /part revision:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partRevision} = $element[1];
        }
        elsif($line =~ /serial number:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{serialNumber} = $element[1];
        }
        elsif($line =~ /firmware version:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{firmwareVersion} = $element[1];
        }
        elsif($line =~ /power state:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{powerState} = $element[1];
        }
        elsif($line =~ /fans self-powered:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{fansSelfPowered} = $element[1];
        }
        elsif($line =~ /nominal power rating:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nominalPowerRating} = $element[1];
        }
        elsif($line =~ /nominal power rating output bitmask:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nominalPowerRatingOutputBitmask} = $element[1];
        }
        elsif($line =~ /surge power rating:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{surgePowerRating} = $element[1];
        }
        elsif($line =~ /surge power rating output bitmask:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{surgePowerRatingOutputBitmask} = $element[1];
        }
        elsif($line =~ /surge hold time:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{surgeHoldTime} = $element[1];
        }
        elsif($line =~ /AC dropout tolerance:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{ACDropoutTolerance} = $element[1];
        }
        elsif($line =~ /standby power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{standbyPower} = $element[1];
        }
        elsif($line =~ /cooling power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{coolingPower} = $element[1];
        }
        elsif($line =~ /swap:\s+/) {
        }
        elsif($line =~ /hotSwap\(private\):\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{hotswapPrivate} = $element[1];
        }
        elsif($line =~ /hotSwap\(public\):\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{hotswapPublic} = $element[1];
        }
        elsif($line =~ /fault:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{fault} = $element[1];
        }
        elsif($line =~ /output \d/) {
            my @tmp = split(/ /,$line);
            $output_num = $tmp[1];
            $output_full = "output".$output_num;
        }
        elsif($line =~ /^voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{voltage} = $element[1];
        }
        elsif($line =~ /^current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{current} = $element[1];
        }
        elsif($line =~ /power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{power} = $element[1];
        }
        elsif($line =~ /min voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{minVoltage} = $element[1];
        }
        elsif($line =~ /min surge voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{minSurgeVoltage} = $element[1];
        }
        elsif($line =~ /max voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{maxVoltage} = $element[1];
        }
        elsif($line =~ /max current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{maxCurrent} = $element[1];
        }
        elsif($line =~ /min current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{minCurrent} = $element[1];
        }
        elsif($line =~ /max surge current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{manSurgeCurrent} = $element[1];
        }
        elsif($line =~ /combined power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{combinedPower} = $element[1];
        }
        elsif($line =~ /SES_info_bit:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{sesInfoBit} = $element[1];
        }
    }
    return \%hash;
}

=head2  get_fan_status

    my $ret = get_fan_status(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "ddump_envctrl" to retrieve all the fans' status and save in a hash table
    with the following keys stated:
    maxNumFans,
    fan number:name, currentFanSpeedRPM, currentFanSpeedLevel, faultStatesGeneratedFault, faultStatesDetectedFault, faultStatesGeneratedPredictedFail, faultStatesDetectedPredictedFail. faultStatesElementSpecificFaults

    Example of returning fan status:
    my $fan_status = get_fan_status();
    print "Fan 2's name = $fan_status->{2}{name};

    Example of reference returned by the function:
    $VAR1 = {
          'maxNumFans' => '8',
          'fan#' 'name' => 'PCM 1 Fan 1',
          'fan#' 'currentFanSpeedRPM' => '3525',
          'fan#' 'currentFanSpeedLevel' => '1',
          'fan#' 'faultStatesGeneratedFault' => '0',
          'fan#' 'faultStatesDetectedFault' => '0',
          'fan#' 'faultStatesGeneratedPredictedFail' => '0',
          'fan#' 'faultStatesDetectedPredictedFail' => '0',
          'fan#' 'faultStatesElementSpecificFaults' => '0x0',
          'PCM#fans' => 'OK',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_fan_status {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_envctrl";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /RSync ddump/){
        croak "Unexpected ddump_envctrl output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output fan status: @output \n";
    my $fan_num = 0;
    my $fan_full ="";
    my $pcm = 1;

    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /max num fans:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{maxNumFans} = $element[1];
        }
        elsif($line =~ /fan \d+$/) {
            my @tmp = split(/ /,$line);
            $fan_num = $tmp[-1];
        }
        elsif($line =~ /name\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{name} = $element[1];
        }
        elsif($line =~ /currentFanSpeedRPM\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{currentFanSpeedRPM} = $element[1];
        }
        elsif($line =~ /currentFanSpeedLevel\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{currentFanSpeedLevel} = $element[1];
        }
        elsif($line =~ /faultStates.generatedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesGeneratedFault} = $element[1];
        }
        elsif($line =~ /faultStates.detectedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesDetectedFault} = $element[1];
        }
        elsif($line =~ /faultStates.generatedPredictedFail:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesGeneratedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.detectedPredictedFail\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesDetectedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.elementSpecificFaults\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesElementSpecificFaults} = $element[1];
        }
        elsif($line =~ /PCM \d fans/) {
            my @element = split(/\s*:\s*/,$line);
            my $pcm_string = "PCM".$pcm."fans";
            $hash{$pcm_string} = $element[-1];
            $pcm++;
        }
    }
    return \%hash;
}

=head2  get_temp_status

    my $ret = get_temp_status(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "ddump_envctrl" to retrieve all the temperatures' status
    and save in a hash table with the following keys stated:
    maxNumZones,
        zone number: name, currentTemperature, faultStatesGeneratedFault, faultStatesDetectedFault,
    faultStatesGeneratedPredictedFail, faultStatesDetectedPredictedFail, faultStatesElementSpecificFaults, defaultCriticalColdTemperature, defaultCriticalHotTemperature,
    modifiedWarningColdTemperature, modifiedNormalTemperature, modifiedWarningHotTemperature
    PCM#zones

    Example of returning temperature status:
    my $temp_status = get_temp_status();
    print "Temperature 2's name = $temp_status->{2}{name}";

    Example of reference returned by the function:
    $VAR1 = {
          'maxNumZones' => '8',
          'zone#' 'name' => 'SBB Canister 1',
      'zone#' 'currentTemperature' => '37.250';
      'zone#' 'faultStatesGeneratedFault' => '0';
      'zone#' 'faultStatesDetectedFault' => '0';
      'zone#' 'faultStatesGeneratedPredictedFail' => '0';
      'zone#' 'faultStatesDetectedPredictedFail' => '0';
      'zone#' 'faultStatesElementSpecificFaults' => '0x0';
      'zone#' 'defaultCriticalColdTemperature' => '5';
      'zone#' 'defaultCriticalHotTemperature' => '60';
      'zone#' 'modifiedWarningColdTemperature' => '10';
      'zone#' 'modifiedNormalTemperature' => '42';
      'PCM#zones' => 'OK';
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_temp_status {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_envctrl";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /RSync ddump/){
        croak "Unexpected ddump_envctrl output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output zone status: @output \n";
    my $zone_num = 0;
    my $pcm;

    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /max num zones:\s+/) {
            my @element = split(/:\s*/,$line);
            $hash{maxNumZones} = $element[1];
        }
        elsif($line =~ /zone \d+$/) {
            my @tmp = split(/ /,$line);
            $zone_num = $tmp[-1];
        }
        elsif($line =~ /name\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{name} = $element[1];
        }
        elsif($line =~ /currentTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{currentTemperature} = $element[1];
        }
        elsif($line =~ /faultStates.generatedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesGeneratedFault} = $element[1];
        }
        elsif($line =~ /faultStates.detectedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesDetectedFault} = $element[1];
        }
        elsif($line =~ /faultStates.generatedPredictedFail:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesGeneratedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.detectedPredictedFail\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesDetectedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.elementSpecificFaults\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesElementSpecificFaults} = $element[1];
        }
        elsif($line =~ /defaultCriticalColdTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{defaultCriticalColdTemperature} = $element[1];
        }
        elsif($line =~ /defaultCriticalHotTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{defaultCriticalHotTemperature} = $element[1];
        }
        elsif($line =~ /modifiedWarningColdTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{modifiedWarningColdTemperature} = $element[1];
        }
        elsif($line =~ /modifiedNormalTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{modifiedNormalTemperature} = $element[1];
        }
        elsif($line =~ /modifiedWarningHotTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{modifiedWarningHotTemperature} = $element[1];
        }
        elsif($line =~ /PCM \d zones/) {
            my @tmp = split(/\s*:\s*/,$line);
            my $pcm_string = "PCM".$pcm."zones";
            $hash{$pcm_string} = $tmp[-1];
        }
    }
    return \%hash;
}

=head2  ddump_general (??)

    my $ret = ddump_general(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "ddump_general" to retrieve all the General Status
    Service Diagnostic Data and save in a hash table with the following keys stated:
        info_type(??) Canister#: enclWWN, Vendor, vendorId, ProductId, PartNum, FruId, SerialNum,
        Expander#: wwn, SoDResetType, revision, enclSerialNum,

    For Local and Remote faults:
    Local -> "I2C Bus 0"  (I2CBus0)
        totalErrorsOps, errorsInBuffer, bufferSize, intermittentThreshold, failureThreshold, softResets hardResets
    Local -> "I2C Bus 1"  (I2CBus1)
        totalErrorsOps, errorsInBuffer, bufferSize, intermittentThreshold, failureThreshold, softResets hardResets
    Local -> "I2C Bus 2"  (I2CBus2)
        totalErrorsOps, errorsInBuffer, bufferSize, intermittentThreshold, failureThreshold, softResets hardResets
        midplaneVPD0
        PCM1VPD
        PCM2VPD
        PCM3VPD
        PCM4VPD
        SBBCanisterVPD0
        SGPIO0

    Example:
    my $electstatus = ddump_general();
    print "The enclosure Serial Number is: $electstatus->{enclSerialNum}\n";
    print "The Local fault I2C Bus 2 general fault info is: $electstatus->{Local}{"I2C Bus 2"}\n";
    print "The Local fault I2C Bus 2 Failure Threshold value is: $electstatus->{Local}{I2CBus2}{failureThreshold}\n";
??? local vs Local

    Example of reference returned by the function:
    $VAR1 = {
          'Enclosure' 'enclWWN' => '50050CC10200027F',
          'Enclosure' 'Vendor' => 'NETAPP',
          'Enclosure' 'ProductId' => 'DS4243',
          'Enclosure' 'PartNum' => '0094157',
          'Enclosure' 'FruId' => '0x29',
          'Enclosure' 'SerialNum' => 'SHU94157002731E',
          'Midplane' 'ProductId' => 'DS424',
          'Midplane' 'PartNum' => '0091910-07',
          'Midplane' 'SerialNum' => 'BPS91910200027F',
          'local' 'present' => 'Present',
          'local' 'Vendor' => 'xyratex#',
          'local' 'ProductId' => 'DEFAULT',
          'local' 'PartNum' => '0094214-01',
          'local' 'FruId' => '0xBA',
          'local' 'SerialNum' => 'ims94214c14055c',
          'local' 'FirmwareVer' => '00a6',
          'local' 'CanisterCPLD' => '12',
          'local' 'SoDResetType' => 'Cold boot',
          'local' 'NumExpanders' => '1',
          'local' 'wwn' => '50050CC102100EC0',
          'Local' 'I2C_Bus_0' => 'No faults',
          'Local' 'I2CBus0' 'totalErrorsOps' => '4424/739177',
          'Local' 'I2CBus0' 'errorsInBuffer' => '0',
          'Local' 'I2CBus0' 'bufferSize' => '100',
          'Local' 'I2CBus0' 'intermittentThreshold' => '15',
          'Local' 'I2CBus0' 'failureThreshold' => '95',
          'Local' 'I2CBus0' 'softResets' => '0',
          'Local' 'I2CBus0' 'hardResets' => '0',
          'Local' 'midplaneVPD0'  => 'No faults',
          'Local' 'PCM1VPD'  => 'No faults',
          'Local' 'PCM2VPD'  => 'No faults',
          'Local' 'PCM4VPD'  => 'No faults',
          'Local' 'SBBCanisterVPD0'  => 'No faults',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_general {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_general";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /General Status Service Diagnostic Data/){
        croak "Unexpected ddump_general output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output is ddump_general: @output \n";
    my $fault;
    my @layer2;
    my $layer2_1; # to save the string without space in between in layer 2 variable
    my $info_type = "";

    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if(($line =~ /Enclosure Info:/) || ($line =~ /Midplane Info:/)) {
            my @element = split(/ /,$line);
            $info_type = $element[0];
        }
        if($line =~ /Canister \d/) {
            my @tmp = split(/ /,$line);
            my $info_num = $tmp[1];
            chop ($info_num);
            $hash{location} = $tmp[3];
            #$info_type = "Canister".$info_num;
            $info_type = $tmp[3];
            $hash{$info_type}{present} = $tmp[6];
        }
        if($line =~ /Enclosure WWN:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{enclWWN} = $element[-1];
        }
        if($line =~ /Vendor:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{Vendor} = $element[-1];
        }
        if($line =~ /Product Id:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{ProductId} = $element[1];
        }
        if($line =~ /Part Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{PartNum} = $element[1];
        }
        if($line =~ /Fru Id:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{FruId} = $element[1];
        }
        if($line =~ /Serial Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{SerialNum} = $element[1];
        }
        if($line =~ /Firmware Ver:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{FirmwareVer} = $element[1];
        }
        if($line =~ /Canister CPLD:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{CanisterCPLD} = $element[1];
        }
        elsif($line =~ /Revision:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{revision} = $element[1];
        }
        elsif($line =~ /SoD reset type:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{SoDResetType} = $element[1];
        }
        elsif($line =~ /Num Expanders:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{numExpanders} = $element[1];
        }
        elsif($line =~ /Expander \d WWN:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{wwn} = $element[1];
        }
        elsif($line =~ /CID Family ID/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$info_type}{CIDFamilyID} = $element[1];
        }
        elsif($line =~ /CID Family Member ID/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$info_type}{CIDFamilyMemberID} = $element[1];
        }
        elsif(($line =~ /Local faults:/) || ($line =~ /Remote faults:/)) {
            my @element = split(/ /,$line);
            $fault = $element[0];
        }
        elsif($line =~ /remote board not present/) {
            $hash{$fault} = "yes";
        }
        elsif(($line =~ /I2C Bus \d:/) || ($line =~ /SGPIO 0/)) {
            @layer2 = split(/:\s*/,$line);
            my @layer2_tmp = split(/ /, $layer2[1]);
            $layer2_1 = join("", @layer2_tmp);
            my $layer2_1_t = join("_", @layer2_tmp);
            $hash{$fault}{$layer2_1_t} = "$layer2[2]";
        }
        elsif($line =~ /Total Errors\/Ops/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{totalErrorsOps} = $tmp[-1];
        }
        elsif($line =~ /Errors in Buffer/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{errorsInBuffer} = $tmp[-1];
        }
        elsif($line =~ /Buffer Size/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{bufferSize} = $tmp[-1];
        }
        elsif($line =~ /Intermittent Threshold/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{intermittentThreshold} = $tmp[-1];
        }
        elsif($line =~ /Failure Threshold/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{failureThreshold} = $tmp[-1];
        }
        elsif($line =~ /Soft resets/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{softResets} = $tmp[-1];
        }
        elsif($line =~ /Hard resets/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{hardResets} = $tmp[-1];
        }
        elsif($line =~ /Midplane VPD 0:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"midplaneVPD0"} = $tmp[-1];
        }
        elsif($line =~ /PCM 1 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM1VPD"} = $tmp[-1];
        }
        elsif($line =~ /PCM 2 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM2VPD"} = $tmp[-1];
        }
        elsif($line =~ /PCM 3 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM3VPD"} = $tmp[-1];
        }
        elsif($line =~ /PCM 4 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM4VPD"} = $tmp[-1];
        }
        elsif($line =~ /SBB Canister VPD 0:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"SBBCanisterVPD0"} = $tmp[-1];
        }
    }
    return \%hash;
}

=head2  ddump_drvmgr

    my $ret = ddump_drvmgr(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the Drive Manager diagnostic data and save in a hash table
    with the following keys stated for the Drive Manager and for each Device.
        HAMode. driveSpinUp, driveBays, allowedDrives, drivePowerControl, deviceRole,
        instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS
    For each device bay: present, swap, SESInfoBit, spinUpTime, driveType, name, arrayName,
    wwn, faults, faultLED, inject, pending, current.

        Example of returning ddump_drvmgr info:
        $info = ddump_drvmgr();
        print "the fault LED status for drive bay 9 is: $info->{9}{faultLED}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'HAMode'  => 'master',
          'driveSpinUp'  => '0',
          'driveBays'  => '24',
          'driveIndexBase'  => '0',
          'allowedDrives'  => 'SAS OR SATA',
          'drivePowerControl'  => 'supported',
          'deviceRole' '' => 'MASTER,
          'instanceRunState'  => 'RUNNING',
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x63E8',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0x40 (64)',
          'transPoolFreeCount'  => '0x40 (64)',
          'numConACKS'  => '0x48 (72)',
          '0' 'present' => 'yes',
          '0' 'swap' => 'no',
          '0' 'SESInfoBit' => 'not set',
          '0' 'spinUpTime' => '0+00:00:08.087',
          '0' 'driveType' => 'SAS',
          '0' 'wwn' => '500000E0173F6792',
          '0' 'faults' => 'none',
          '0' 'faultLED' => 'OFF',
          '0' 'inject' => 'NONE',
          '0' 'pending' => 'ONLINE',
          '0' 'current' => 'ONLINE',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_drvmgr {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_drvmgr";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /Drive Manager diagnostic dump/){
        croak "Unexpected ddump_drvmgr output: $output";
    }

    my $bay;
    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /HA mode:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{HAMode} = $element[1];
        }
        elsif($line =~ /Drives spinning up:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{driveSpinUp} = $element[1];
        }
        elsif($line =~ /Drive bays:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{driveBays} = $element[1];
        }
        elsif($line =~ /Allowed drives:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{allowedDrives} = $element[1];
        }
        elsif($line =~ /Drive power control:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{drivePowerControl} = $element[1];
        }
        elsif($line =~ /Device role:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{deviceRole} = $element[1];
        }
        elsif($line =~ /Instance run state:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{instanceRunState} = $element[1];
        }
        elsif($line =~ /Sync to Slave status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{syncToSlaveStatus} = $element[1];
        }
        elsif($line =~ /This instance's next UID will be:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nextUID} = $element[1];
        }
        elsif($line =~ /Total expanded transactions:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalExpandTrans} = $element[1];
        }
        elsif($line =~ /Transaction pool capacity:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolCap} = $element[1];
        }
        elsif($line =~ /Transaction pool free count:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolFreeCount} = $element[1];
        }
        elsif($line =~ /Num concurrent ACKS:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{numConACKS} = $element[1];
        }
        elsif($line =~ /Drive Bay \d* status/) {
            my @tmp = split(/ /,$line);
            $bay = $tmp[3];
        }
        elsif($line =~ /present\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{present} = $element[1];
        }
        elsif($line =~ /swap\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{swap} = $element[1];
        }
        elsif($line =~ /SES_info_bit\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{SESInfoBit} = $element[1];
        }
        elsif($line =~ /spin up time\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{spinUpTime} = $element[2].".".$element[3];
        }
        elsif($line =~ /drive_type\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{driveType} = $element[1];
        }
        elsif($line =~ /name\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{name} = $element[1];
        }
        elsif($line =~ /arrayname\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{arrayName} = $element[1];
        }
        elsif($line =~ /WWN\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{wwn} = $element[1];
        }
        elsif($line =~ /faults\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{faults} = $element[1];
        }
        elsif($line =~ /fault LED\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{faultLED} = $element[1];
        }
        elsif($line =~ /inject\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{inject} = $element[1];
        }
        elsif($line =~ /pending\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{pending} = $element[1];
        }
        elsif($line =~ /current\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{current} = $element[1];
        }
    }
    return \%hash;
}

=head2  get_midplanevpd

    my $ret = get_midplane(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the midplane vpd types data and save in a hash table with the following keys.
    For bus 1: mfDateTime_1, enclWWN_1, chasSerialNum_1, mdplnSerialNum_1, chasPartNum_1, mdplnPartNum_1, enclVendID_1, enclProdID_1, enclID_1
    For bus 2: mfDateTime_2, enclWWN_2, chasSerialNum_2, mdplnSerialNum_2, chasPartNum_2, mdplnPartNum_2, enclVendID_2, enclProdID_2, enclID_2

        Example of returning enclosure vendor id:
        $info = get_midplanevpd();
        print "the enclosure vendor IDs from bus 1 and bus 2 are: $info->{enclVendID_1} and $info->{enclVendID_2}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'mfDateTime_1'  => '96 38 62',  # 1:manufacturing datetime
          'mfDateTime_2'  => '96 38 62',
          'enclWWN_1'  => '50050cc10200027f', # 2:enclosure wwn
          'enclWWN_2'  => '50050cc10200027f',
          'chasSerialNum_1'  => 'SHU94157002731E', # 3:chassis serialnumber
          'chasSerialNum_2'  => 'SHU94157002731E',
          'mdplnSerialNum_1'  => 'BPS91910200027F', # 4:midplane serialnumber
          'mdplnSerialNum_2'  => 'BPS91910200027F',
          'chasPartNum_1'  => '0094157-04', # 5:chassis partnumber
          'chasPartNum_2'  => '0094157-04',
          'mdplnPartNum_1'  => '0091910-07', # 6:midplane partnumber
          'mdplnPartNum_2'  => '0091910-07',
          'enclVendID_1'  => 'NETAPP  ', # 7:enclosure vendorID
          'enclVendID_2'  => 'NETAPP  ',
          'enclProdID_1'  => 'DS424   ', # 8:enclosure productID
          'enclProdID_2'  => 'DS424   ',
          'enclID_1'  => '000', # 9:enclosureID ???
          'enclID_2'  => '000',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_midplanevpd {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
#   Make Connection and build command based on input parameters

    for (my $vpd_type=1; $vpd_type<10; $vpd_type++)
    {
        my $command = "set_midplanevpd $vpd_type";

        # Look for manufacture Date Time
        if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
            $opts{conntype} = "#console";
            $c = make_conn($opts{host},$opts{conntype});
            $cmd = "$command";
        }
        elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
            if(!defined($opts{gemconn})){
                croak "GEMTEST needs to be connected before we can use gemtest connmode";
            }
            $c = $opts{gemconn};
            $cmd = "$command";
        }
        elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
            if($opts{filermode} ne "normal"){
                $opts{conntype} = "#console";
            }

            $c = make_conn($opts{host},$opts{conntype});
            if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
                croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
            }
            if(!defined($opts{AorB})){
                $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
            }
            else{
                $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
            }

            if($opts{filermode} ne "normal"){

                $cmd = "sasadmin expander_cli $id \"$command\"";
            }
            else{
                $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
            }

        }
        else {
            croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
        }

        #   Issue Command

        if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
            (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
        }
        else{
            (undef,undef,$output) = say($c,"$cmd");
        }

        if($output !~ /TWI bus/){
            croak "Unexpected set_midplanevpd output: $output";
        }

        my @output = split(/\r*\n/,$output);

        print "Output is : @output \n";
        if ($vpd_type eq "1") {
        # Look for manufacturing datetime
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mfDateTime_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mfDateTime_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "2") {
        # Look for enclosure wwn
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    my @new = split(/ /, $element[1]);
                    my $new_element = join("", @new);
                    $hash{enclWWN_1} = $new_element;
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    my @new = split(/ /, $element[1]);
                    my $new_element = join("", @new);
                    $hash{enclWWN_2} = $new_element;
                }
            }
        }
        if ($vpd_type eq "3") {
        # Look for chassis serial number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasSerialNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasSerialNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "4") {
        # Look for midplane serial number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnSerialNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnSerialNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "5") {
        # Look for chassis part number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasPartNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasPartNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "6") {
        # Look for midplane part number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnPartNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnPartNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "7") {
        # Look for enclosure vendor id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclVendID_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclVendID_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "8") {
        # Look for enclosure product id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclProdID_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclProdID_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "9") {
        # Look for enclosure id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{enclID_1} = $element[-1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{enclID_2} = $element[-1];
                }
            }
        }
        if ($vpd_type eq "13") {
        # Look for Branding Id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{BrandingID_1} = $element[-1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{BrandingID_2} = $element[-1];
                }
            }
        }
    }
    return \%hash;
}


=head2  dumpdrives

    my $ret = dumpdrives(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the drives brief online/empty status and save in a hash
    table with the drive number as key index.

        Example of returning ddump_drvmgr info:
        $info = dumpdrives();
        print "the first drive current status is: $info->{1}\n";

    Example of reference returned by the function:
    $VAR1 = {
          '0'  => 'online',
          '5'  => 'off',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub dumpdrives {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_type => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "dumpdrives";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output =~ /Usage:|Unknown/){
        croak "dumpdrives failed with : $output";
    }

    my @output = split(/\r*\n/,$output);
    my $i;
    my $j = 0;

    print "Output is : @output \n";
    if ($opts{shelf_type} =~ /DS42/) {
        foreach my $line (@output) {
            $line =~ s/^\s+//g;
            if($line =~ /\d+/) {
                my @element = split(/\]\[/,$line);
                for($i = 0; $i <4; $i++) {
                    my @new = split(/ /,$element[$i]);
                    if ($j < 10) {
                        $hash{$j} = $new[3];
                    } else {
                        $hash{$j} = $new[2];
                    }
                    $j++;
                }
            }
        }
    } elsif ($opts{shelf_type} =~ /DS22/) {
        my $i = 0;
        foreach my $line (@output) {
            $line =~ s/^\[//g;
            if($line =~ /\d+/) {
                my @e1 = split(/\s+\[\s+/,$line);
                my @element = split(/\]/,$e1[0]);
                my @new = split(/\s+/,$element[0]);
                $hash{$i} = $new[2];
                $i++;
            }
        }
    }
    return \%hash;
}

=head2  ddump_expmgr

    my $ret = ddump_drvmgr(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine retrieves all the Drive Manager diagnostic data and save in a hash table
    with the following keys stated for the Drive Manager and for each Device.
        deviceRole, instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS
    For each phy type (host, drive, sil): BitWind, RstWind, ChgCnt, dwErr, dpErr, syncEr, phyRstEr, crcEr
                    or (??): PCD, IDD, RDD, LOS, PRP, CRC
    For each expander id: swap, updated, ASA, regAccess

        Example of returning ddump_drvmgr info:
        $info = ddump_drvmgr();
        print "the CRC error for drive 8 is: $info->{local}{drive}{crcEr}\n";??

    Example of reference returned by the function:
    $VAR1 = {
          'deviceRole'  => 'MASTER',
          'instanceRunState'  => 'RUNNING'.
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x199 (slot=0 val=409)',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0xA (10)',
          'transPoolFreeCount'  => '0x6 (6)',
          'numConACKS'  => '0x12 (18)',
          'HAmode'  => 'master',
          'local' 'mirror'  => 'disabled', # local phy profiles
          'local' 'host' 'BitWind' => '60',
          'remote' 'host' 'RstWind' => '60',
          'local' 'host' 'CRC' => '1',
          'local' 'host' 'PCD' => '1',
          'local' 'drive' 'ChgCnt' => '120',
          'remote' 'drive' 'dwErr' => '120',
          'local' 'drive' 'PRP' => '1',
          'local' 'drive' 'LOS' => '1',
          'local' 'sil' 'phyRstEr' => '60',
          'remote' 'sil' 'crcEr' => '120',
          'local' 'sil' 'RDD' => '1',
          'local' 'sil' 'IDD' => '1',
          'expander0' 'swap'  => 'No',
          'expander0' 'updated'  => 'No',
          'expander1' 'clkStatus'  => 'OK',
          'expander1' 'regAccess'  => 'OK',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_expmgr {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_expmgr";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /Expander Manager diagnostic dump/){
        croak "Unexpected ddump_expmgr output: $output";
    }

    my $expander_id = 0;
    my $location;
    my $phy_type;
    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Device role:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{deviceRole} = $element[1];
        }
        elsif($line =~ /Instance run state:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{instanceRunState} = $element[1];
        }
        elsif($line =~ /Sync to Slave status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{syncToSlaveStatus} = $element[1];
        }
        elsif($line =~ /This instance's next UID will be:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nextUID} = $element[1];
        }
        elsif($line =~ /Total expanded transactions:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalExpandTrans} = $element[1];
        }
        elsif($line =~ /Transaction pool capacity:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolCap} = $element[1];
        }
        elsif($line =~ /Transaction pool free count:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolFreeCount} = $element[1];
        }
        elsif($line =~ /Num concurrent ACKS:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{numConACKS} = $element[1];
        }
        elsif($line =~ /Local\s+Phy Profiles/) {
            $location = "local";
        }
        elsif($line =~ /Remote\s+Phy Profiles/) {
            $location = "remote";
        }
        elsif($line =~ /Mirroring:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$location}{mirror} = $element[1];
        }
        elsif($line =~ /Host Phy Profile:/) {
            $phy_type = "host";
        }
        elsif($line =~ /Drive Phy Profile:/) {
            $phy_type = "drive";
        }
        elsif($line =~ /Sil Phy Profile:/) {
            $phy_type = "sil";
        }
        elsif($line =~ /\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s*/) {
            my @element = split(/\s+/,$line);
            $hash{$location}{$phy_type}{BitWind} = $element[0];
            $hash{$location}{$phy_type}{RstWind} = $element[1];
            $hash{$location}{$phy_type}{ChgCnt} = $element[2];
            $hash{$location}{$phy_type}{dwErr} = $element[3];
            $hash{$location}{$phy_type}{dpErr} = $element[4];
            $hash{$location}{$phy_type}{syncEr} = $element[5];
            $hash{$location}{$phy_type}{phyRstEr} = $element[6];
            $hash{$location}{$phy_type}{crcEr} = $element[7];
        }
        elsif($line =~ /\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s*/) {
            my @element = split(/\s+/,$line);
            $hash{$location}{$phy_type}{PCD} = $element[1];
            $hash{$location}{$phy_type}{IDD} = $element[2];
            $hash{$location}{$phy_type}{RDD} = $element[3];
            $hash{$location}{$phy_type}{LOS} = $element[4];
            $hash{$location}{$phy_type}{PRP} = $element[5];
            $hash{$location}{$phy_type}{CRC} = $element[6];
        }
        elsif($line =~ /Expander \d Information:/) {
            my @tmp = split(/ /,$line);
            $expander_id = "expander".$tmp[1];
        }
        elsif($line =~ /Swap:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{swap} = $element[-1];
        }
        elsif($line =~ /Updated:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{updated} = $element[-1];
        }
        elsif($line =~ /ASA:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{ASA} = $element[-1];
        }
        elsif($line =~ /CLK status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{clkStatus} = $element[-1];
        }
        elsif($line =~ /Reg Access:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{regAccess} = $element[-1];
            $expander_id++;
        }
    }
    return \%hash;
}

=head2  ddump_acpmgr

    my $ret = ddump_acpmgr(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the ACP Processor status and save in a hash table
    with the following keys stated for the ACP (APC) Manager:
        deviceRole, instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS, totalACP, HAMode, partnerPresent, partnerRemoved,
    For each ACP processor: present, swap, SESInfoBit, heartBeatFail, resetInProgress, macAddr.

        Example of returning ddump_drvmgr info:
        $info = ddump_drvmgr();
        print "the fault LED status for drive bay 9 is: $info->{$bay_number}{faultLED}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'deviceRole'  => 'MASTER',
          'instanceRunState'  => 'RUNNING'.
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x199 (slot=0 val=409)',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0xA (10)',
          'transPoolFreeCount'  => '0x6 (6)',
          'numConACKS'  => '0xC (12)
          'totalACP'  => '2',
          'HAmode'  => 'master',
          'partnerPresent'  => 'Yes',
          'partnerRemoved'  => 'No',
          'acp0' 'present'  => 'Yes',
          'acp0' 'swap'  => 'No',
          'acp0' 'SESInfoBit'  => 'Not Set', 'acp1' 'heartBeatFail'  => 'Yes',
          'acp1' 'macAddr'  => '98:98:98:98:98:98',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_acpmgr {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $id;
    my $command = "ddump_acpmgr";

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }

    if($output !~ /ACP Manager Service Diagnostic Data/){
        croak "Unexpected ddump_acpmgr output: $output";
    }

    my $bay;
    my $acp_proc;
    my @output = split(/\r*\n/,$output);

    #print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Device role:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{deviceRole} = $element[1];
        }
        elsif($line =~ /Instance run state:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{instanceRunState} = $element[1];
        }
        elsif($line =~ /Sync to Slave status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{syncToSlaveStatus} = $element[1];
        }
        elsif($line =~ /This instance's next UID will be:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nextUID} = $element[1];
        }
        elsif($line =~ /Total expanded transactions:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalExpandTrans} = $element[1];
        }
        elsif($line =~ /Transaction pool capacity:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolCap} = $element[1];
        }
        elsif($line =~ /Transaction pool free count:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolFreeCount} = $element[1];
        }
        elsif($line =~ /Num concurrent ACKS:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{numConACKS} = $element[1];
        }
        elsif($line =~ /Total ACPs*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalACP} = $element[1];
        }
        elsif($line =~ /HA mode*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{HAMode} = $element[1];
        }
        elsif($line =~ /Partner Present*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partnerPresent} = $element[1];
        }
        elsif($line =~ /Partner*removed*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partnerRemoved} = $element[1];
        }
        elsif($line =~ /ACP Processor \[\d*\] status/) {
            my @tmp1 = split(/ /,$line);
            my @tmp = split(/\[/,$tmp1[3]);
            chop($tmp[1]);
            $acp_proc = "acp".$tmp[1];
        }
        elsif($line =~ /Present\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{present} = $element[1];
        }
        elsif($line =~ /Swap\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{swap} = $element[1];
        }
        elsif($line =~ /SES_info_bit\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{SESInfoBit} = $element[1];
        }
        elsif($line =~ /Heart Beat Failed\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{heartBeatFail} = $element[1];
        }
        elsif($line =~ /Reset In Progress\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{resetInProgress} = $element[1];
        }
        elsif($line =~ /MAC Address\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{macAddr} = $element[1];
        }
    }
    return \%hash;
}

=head2  qsfpinfo

    my $ret = qsfpinfo(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the qsfp data and save in a hash table
    with the following keys stated:
    slot# - present or not_present
          id, len, powerClass, vedorSpecific, CLEI, vendorName, partNum, serialNum, dateCode, revision, vendorOUI, deviceTech

        Example of returning qsfp info:
        $info = qsfpinfo();
        print "the Serial Number for cable0 is: $info->{0}{SerialNum}\n";

    Example of reference returned by the function:
    $VAR1 = {
          '1'  => 'not_present',
          '0' 'id' => '0x0c'.
          '0' 'len' => '2'.
          '0' 'powerClass' => '0x00'.
          '0' 'vendorSpecific' => '0x00'.
          '1' 'CLEI' => 'TRUE'.
          '1' 'vendorName' => 'Amphenol'.
          '1' 'partNum' => '112-00177'.
          '1' 'serialNum' => 'APL08311770387'.
          '1' 'dateCode' => '080801'.
          '0' 'revision' => '10'.
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub qsfpinfo {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "qsfp_info";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /QSFP Slot/){
        croak "Unexpected qsfp_info output: $output";
    }

    my @output = split(/\r*\n/,$output);
    my $slot = "";

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /QSFP Slot \d/) {
            my @tmp = split(/ /,$line);
            $slot = $tmp[-1];
        }
        elsif($line =~ /No module present/) {
            $hash{$slot} = "not_present";
        }
        elsif($line =~ /Identifier:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{id} = $element[1];
        }
        elsif($line =~ /Length:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{len} = $element[1];
        }
        elsif($line =~ /Power Class:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{powerClass} = $element[1];
        }
        elsif($line =~ /Vendor Specific:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{vendorSpecific} = $element[1];
        }
        elsif($line =~ /CLEI Present:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{CLEI} = $element[1];
        }
        elsif($line =~ /Vendor Name:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{vendorName} = $element[1];
        }
        elsif($line =~ /Part Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{partNum} = $element[1];
        }
        elsif($line =~ /Serial Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{serialNum} = $element[1];
        }
        elsif($line =~ /Date code:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{dateCode} = $element[1];
        }
        elsif($line =~ /Revision:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{revision} = $element[1];
        }
        elsif($line =~ /Vendor OUI:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{vendorOUI} = $element[1];
        }
        elsif($line =~ /Device Technology:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{deviceTech} = $element[1];
        }
    }
    return \%hash;
}

=head2  cablemgr_tp

    my $ret = cablemgr_tp(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the Cable Manager data and save in a hash table
    with the following keys stated:
    For each board# in slot#: status, cableType, swap, updated, cableCommsFail, cableCRCErr, serialNum, vendorName, id, cableLen, devieTech, cableEnd

        Example of returning cablemgr info:??
        $info = cablemgr_tp();
        print "the Cable Length for cable 1 in board 0 is: $info->{0}{1}{cableLen}\n";

    Example of reference returned by the function:
    $hash{$board}{$slot}{$VAR1}
    $VAR1 = {
          'status'  => 'Present',
          'cableType'  => 'QSFP 4x',
          'swap'  => 'No',
          'updated'  => 'No',
          'cableCommsFail'  => 'No',
          'cableCRCErr'  => 'No',
          'serialNum'  => 'APL08311770387',
          'vendorName'  => 'Amphenol',
          'id'  => '0x0c',
          'cableLen'  => '0x02',
          'deviceTech'  => '0x80',
          'cableEnd'  => '0x00',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub cablemgr_tp {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                board => {type=>SCALAR, default=>undef,optional=>1},
                slot => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "cablemgr_tp 1 $opts{board} $opts{slot}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output =~ /Usage:/){
        croak "cablemgr_tp failed with : $output";
    }


    my @output = split(/\r*\n/,$output);

    my $board = $opts{board};
    my $slot = $opts{slot};

    #print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Cable \d on board \d/) {
            if($line =~ /NOT PRESENT/) {
                $hash{$board}{$slot}{status} = "NOT PRESENT";
            }
            elsif($line =~ /UNKNOWN/) {
                $hash{$board}{$slot}{status} = "UNKNOWN";
            }
             else {
                $hash{$board}{$slot}{status} = "PRESENT";
            }
        }
        elsif($line =~ /Cable Type:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{cableType} = $element[1];
        }
        elsif($line =~ /Swap:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{swap} = $element[1];
        }
        elsif($line =~ /Updated:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{updated} = $element[1];
        }
        elsif($line =~ /Cable Comms failure:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{cableCommsFail} = $element[1];
        }
        elsif($line =~ /Cable CRC Error:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{cableCRCErr} = $element[1];
        }
        elsif($line =~ /Serial Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{serialNum} = $element[1];
        }
        elsif($line =~ /Vendor Name:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{vendorName} = $element[1];
        }
        elsif($line =~ /Part Number:/) {
                        my @element = split(/:\s*/,$line);
                        $hash{$board}{$slot}{partNum} = $element[1];
                }
        elsif($line =~ /Identifier:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{id} = $tmp[-1];
        }
        elsif($line =~ /Cable Length:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{cableLen} = $tmp[-1];
        }
        elsif($line =~ /Device Tech:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{deviceTech} = $tmp[-1];
        }
        elsif($line =~ /Cable End:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{cableEnd} = $tmp[-1];
        }
    }
    return \%hash;
}

=head2  phydump

    my $ret = phydump(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the phy data and save in a hash table
    with the following keys stated:
    For each phy: type, index, state, speed, protocol, wwn

        Example of returning qsfp info:
        $info = phydump();
        print "PHY 8 WWN is: $info->{$phy}{wwn}\n";

    Example of reference returned by the function:
    $VAR1 = {
          '0x00' 'type'  => 'Port',
          '0x04' 'index'  => '1',
          '0x0a' 'state'  => 'E L S',
          '0x1a' 'speed'  => '3.0Gbps',
          '0x23' 'protocol'  => 'SAS',
          '0x24' 'wwn'  => '5000c500016487ad',
    }
The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub phydump {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                board => {type=>SCALAR, default=>undef,optional=>1},
                slot => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "phydump";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /State:/){
        croak "Unexpected phydump output: $output";
    }

    my @output = split(/\r*\n/,$output);

    my $board = $opts{board};
    my $slot = $opts{slot};

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        my @element = split(/\s*\|/,$line);
        if($element[1] =~ /Port|Drive|SIL/){
            print "element0 : $element[0] \n";
            print "element2 : $element[2] \n";
            print "element3 : $element[3] \n";
            print "phy type is : $element[1]  \n";
            my @phy = split ("0x", $element[0]);
            $hash{$phy[1]}{type} = $element[1];
            $hash{$phy[1]}{index} = $element[2];
            $hash{$phy[1]}{state} = $element[4];
            $hash{$phy[1]}{speed} = $element[5];
            $hash{$phy[1]}{protocol} = $element[6];
            $hash{$phy[1]}{wwn} = $element[7];
        }
    }
    return \%hash;
}

=head2 controller_state

    my $ret = controller_state(
        host     => $host,
        connmode => "shelfcli",
        timeout  => $timeout
    );

This routine checks the master/slave state of the module
It returns 'master'|'slave' on success and croaks on failure.

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : Connection type.
Defaults:
If connmode is 'inband' then "#rsh".
If connmode is 'shelfcli' or 'gemtest' then "#console"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode.
Possible values are:
"shelfcli" for direct console connection to shelf
"gemtest" for using gemtest Windows utility
"inband" for using ontap inband connections
Default: "shelfcli"

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub controller_state {

    my (%opts) = validate(
        @_,
        {
            host => {type=>OBJECT,isa=>'Hostrec'},
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
            filermode => {type=>SCALAR,default=>"normal",optional=>1},
            gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
            timeout => {type=>SCALAR, default=>30,optional=>1},
            adapter_id => {type=>SCALAR, default=>undef,optional=>1},
            shelf_id => {type=>SCALAR, default=>undef,optional=>1},
            AorB => {type=>SCALAR, default=>undef,optional=>1},
            params => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_ha"; # output includes "Device role: SLAVE|MASTER

    # Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){
        # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    } elsif($opts{connmode} eq "gemtest") {
        # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    } elsif($opts{connmode} eq "inbandcli"){
        # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    } elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
    } else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
        say($c,"\n");
    }

    if ($output =~ /:\s+(\w+)/) {
        if ($1 eq 'Slave') {
            return 'slave';
        } elsif ($1 eq 'Master') {
            return 'master';
        } else {
            croak "Device Mode of $1 not recognized";
        }
    }
    croak "Didn't find Device Mode in $output";
}

=head2 watchdog_tp

    my $ret = controller_state(
        host     => $host,
        connmode => "shelfcli",
        timeout  => $timeout,
    tp_code     => $tp_code
    );

This routine calls firmware internal command watchdog_tp to inject different type of failures: =1 assertion; =2 exception; =3 deadlock; =4 busyloop

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : Connection type.
Defaults:
If connmode is 'inband' then "#rsh".
If connmode is 'shelfcli' or 'gemtest' then "#console"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode.
Possible values are:
"shelfcli" for direct console connection to shelf
"gemtest" for using gemtest Windows utility
"inband" for using ontap inband connections
Default: "shelfcli"

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 4

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub watchdog_tp {

    my (%opts) = validate(
        @_,
        {
            host => {type=>OBJECT,isa=>'Hostrec'},
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
            filermode => {type=>SCALAR,default=>"normal",optional=>1},
            gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
            timeout => {type=>SCALAR, default=>30,optional=>1},
            adapter_id => {type=>SCALAR, default=>undef,optional=>1},
            shelf_id => {type=>SCALAR, default=>undef,optional=>1},
            AorB => {type=>SCALAR, default=>undef,optional=>1},
            tp_code => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "watchdog_tp $opts{tp_code}";

    # Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){
        # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    } elsif($opts{connmode} eq "gemtest") {
        # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    } elsif($opts{connmode} eq "inbandcli"){
        # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    } elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
    } else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
        say($c,"\n");
    }
    print "Output is  C: $output \n";
    if($output =~ /Usage:|Invalid/){
        croak "watchdog_tp failed with : $output";
    }
    return $output;
}

=head2  ses_report_page2

    my $ret = ses_report_page2(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the output data from firmware command "ses_report_page 2" and save in a hash table
    with the following keys stated:
    len (ses page length),
    For each element (Device, Psu, Fan, Temperature, EncElectronics, Voltage, Current, SASConncetor, Expander, ACP):
        status_code, status, status1, status2, status3, status4  (4 bytes status for each element)

        Example of returning ses_report_page2 info:
        $info = ses_report_page2();
        print "Fan 2 overall status is: $info->{Fan}{2}{status}\n";
        print "ACP 1 status 1 is: $info->{ACP}{1}{status1}\n";

    Sample output from ses_report_page 2:
0+00:00:26.187 S0 GEM> ses_report_page 2
SES Page 02h (344 Bytes)
-----------------------------
Page Length: 01 54h
Generation Code: 00 00 00 00h

01h:Overall Device                   00h:UNSUPPORTED    :(hex)00 00 00 00
  01h:Device                         01h:OK             :(hex)01 00 00 00
02h:Overall Psu                      00h:UNSUPPORTED    :(hex)00 00 00 00
  02h:Psu                            01h:OK             :(hex)01 00 00 a0
03h:Overall Fan                      00h:UNSUPPORTED    :(hex)00 00 00 00
  03h:Fan                            01h:OK             :(hex)01 01 51 a1
04h:Overall Temperature Sensor       00h:UNSUPPORTED    :(hex)00 00 00 00
  04h:Temperature Sensor             01h:OK             :(hex)01 00 30 00

    Example of reference returned by the function:
    $VAR1 = {
          'Device' '1' 'status_code'  => '01h',
          'Psu' '2' 'status'  => 'OK',
          'Fan' '3' 'status1'  => '01',
          'Temperature' '4' 'status2'  => '00',
          'EncElectronics' '1' 'status3'  => '01',
          'Voltage' '2' 'status4'  => '49',
          'Current' '3' 'status_code'  => '01h',
          'SASConnector' '1' 'status'  => '05',
          'Expander' '1' 'status2'  => '00',
          'ACP' '0' 'status3'  => '00',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ses_report_page2 {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ses_report_page 2";

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }



    if($output !~ /SES Page 02h/){
        croak "Unexpected ses_report_page 2 output: $output";
    }

    my $component = "";
    my $index = 0;
    my $comp_before = "Device";
    my @output = split(/\r*\n/,$output);

    #print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        my @element = "";
        if($line =~ /Page Length:/) {
            @element = split(/:s*/,$line);
            @element = split(/ /,$element[-1]);
            chop($element[2]);
            $hash{len} = $element[1].$element[2];
        }
        elsif($line =~ /01h:Device|02h:Psu|03h:Fan|04h:Temperature|07h:Enc Electronics|12h:Voltage|13h:Current|19h:SAS Connector|83h:Expander|85h:ACP Processor/) {
            @element = split(/:/,$line);
            chop($element[0]);
            my $comp_id = $element[0];
            my @tmp1 = split(/\s+/,$element[1]);
            my $component = $tmp1[0];
            if ($component =~ /Enc|SAS/) {
                $component = $tmp1[0].$tmp1[1];
            }

            # reset component index with new component
            if ($comp_before ne $component) {
                if ($component =~ /Psu/) {
                    $index = 1;
                } else {
                    $index = 0;
                }
            }

            chop($tmp1[-1]);
            my $status_code = $tmp1[-1];
            my @tmp2 = split(/\s+/,$element[2]);
            my $status = $tmp2[0];
            my @tmp3 = split(/\s+/,$element[3]);
            my @tmp4 = split(/\)/, $tmp3[0]);
            my $status1 = $tmp4[-1];
            my $status2 = $tmp3[1];
            my $status3 = $tmp3[2];
            my $status4 = $tmp3[3];
            $hash{$component}{$index}{status_code} = $status_code;
            $hash{$component}{$index}{status} = $status;
            $hash{$component}{$index}{status1} = $status1;
            $hash{$component}{$index}{status2} = $status2;
            $hash{$component}{$index}{status3} = $status3;
            $hash{$component}{$index}{status4} = $status4;
            $comp_before = $component;
            $index++;
        }
    }
    return \%hash;
}

=head2  ses_report_page8081

    my $ret = ses_report_page8081(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the SES page 80/81 data and save in a hash table
    with the following keys.
    expPageSeq, phyCount, logPhyId
    For each PHY:
        phyStatus, negPhyLinkRate, partPathTimeout, invDwordCount, runDispCount, lossDwordSyncCount, phyResetCount, crcErrCount, phyChangeCount

        Example of returning ses_report_page8081 info:
        $info = ses_report_page8081();
        print "PHY 2 has Running Disparity Count: $info->{$phy}{RunDispCount}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'expPageSeq'  => '87h',
          'phyCount' ''  => '24h',
          'logPhyId'   => '00h',
          '0' 'phyStatus'  => '00h',
          '1' 'negPhyLinkRate'  => '09h',
          '2' 'partPathTimeout'  => '07h',
          '3' 'invDwordCount'  => '00000000h',
          '4' 'runDispCount'  => '00000000h',
          '5' 'lossDwordSyncCount'  => '00000000h',
          '6' 'phyResetCount'  => '00000000h',
          '7' 'crcErrCount'  => '00000000h',
          '8' 'phyChangeCount'  => '00000002h',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ses_report_page8081 {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                page => {type=>SCALAR},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ses_report_page $opts{page}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /SES Page/){
        croak "Unexpected ses_report_page output: $output";
    }

    my $phy;
    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Expander Status Page Sequence/) {
            my @element = split(/:\s*/,$line);
            $hash{expPageSeq} = $element[1];
        }
        elsif($line =~ /PHY Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{phyCount} = $element[1];
        }
        elsif($line =~ /Logical PHY Identifier/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{logPhyId} = $element[1];
        }
        elsif($line =~ /PHY \d+$/) {
            my @tmp = split(/ /,$line);
            $phy = $tmp[1];
        }
        elsif($line =~ /PHY Status/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{phyStatus} = $element[1];
        }
        elsif($line =~ /Negotiated Physical Link Rate/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{negPhyLinkRate} = $element[1];
        }
        elsif($line =~ /Partial Pathway Timeout Value/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{partPathTimeout} = $element[1];
        }
        elsif($line =~ /Invalid Dword Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{invDwordCount} = $element[1];
        }
        elsif($line =~ /Running Disparity Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{runDispCount} = $element[1];
        }
        elsif($line =~ /Loss of Dword Synchronization Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{lossDwordSyncCount} = $element[1];
        }
        elsif($line =~ /PHY Reset Problem Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{phyResetCount} = $element[1];
        }
        elsif($line =~ /CRC Error Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{crcErrCount} = $element[1];
        }
        elsif($line =~ /PHY Change Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{phyChangeCount} = $element[1];
        }
    }
    return \%hash;
}

=head2  ses_report_page8283

        my $ret = ses_report_page8283(host=>$host, conmode=>"shelfcli",timeout => $timeout);
        This routine retrieves all the SES page 82/83 data and save in a hash table
        with the following keys.

        Host,Drive,SIL,GlobalModuleControl,PHYResetTimeout,AWTBias.

dd        For Keys Host,Drive and SIL:
          Internally the Nested Hash keys are

        "Profile,BitErrorRateSampleWindow,ResetProblemSampleWindow,RunningDisparityThreshold,InvalidDwordThreshold,LossofDWORDSyncThreshold,PHYCHANGEThreshold,CRCErrorThreshold,PHYResetProblemThreshold."


        Example of returning ses_report_page8283 info:
                $info = ses_report_page8283();

        $VAR1 = \{
            'SIL' => {
                       'Profile' => '3f00h',
                       'BitErrorRateSampleWindow' => '60 min',
                       'ResetProblemSampleWindow' => '60 min',
                       'RunningDisparityThreshold' => '0078h',
                       'InvalidDwordThreshold' => '0078h',
                       'LossofDWORDSyncThreshold' => '003ch',
                       'PHYCHANGEThreshold' => '78h',
                       'CRCErrorThreshold' => '0078h',
                       'PHYResetProblemThreshold' => '003ch'
                     },
            'PHYResetTimeout' => '00 per 100 mS',
            'GlobalModuleControl' => '00h',
            'Drive' => {
                         'Profile' => '3f00h',
                         'BitErrorRateSampleWindow' => '60 min',
                         'ResetProblemSampleWindow' => '60 min',
                         'RunningDisparityThreshold' => '0078h',
                         'InvalidDwordThreshold' => '0078h',
                         'LossofDWORDSyncThreshold' => '003ch',
                         'PHYCHANGEThreshold' => '78h',
                         'CRCErrorThreshold' => '0078h',
                         'PHYResetProblemThreshold' => '003ch'
                       },
       'Host' => {
                        'Profile' => '3f00h',
                        'BitErrorRateSampleWindow' => '60 min',
                        'ResetProblemSampleWindow' => '60 min',
                        'RunningDisparityThreshold' => '0078h',
                        'InvalidDwordThreshold' => '0078h',
                        'LossofDWORDSyncThreshold' => '003ch',
                        'PHYCHANGEThreshold' => '78h',
                        'CRCErrorThreshold' => '0078h',
                        'PHYResetProblemThreshold' => '003ch'
                      },
            'AWTBias' => '0000 uS'
          };






The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub ses_report_page8283 {

        my (%opts) = validate(@_,{
                                host => {type=>OBJECT,isa=>'Hostrec'},
                                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                                timeout => {type=>SCALAR, default=>300,optional=>1},
                                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                                AorB => {type=>SCALAR, default=>undef,optional=>1},
                                page => {type=>SCALAR},
                }
        );

        my %hash = ();
        my $output;
        my $c;
        my $cmd;
        my $id;
        my $command = "ses_report_page $opts{page}";

#       Make Connection and build command based on input parameters
    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
                $opts{conntype} = "#console";
                $c = make_conn($opts{host},$opts{conntype});
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
                if(!defined($opts{gemconn})){
                        croak "GEMTEST needs to be connected before we can use gemtest connmode";
                }
                $c = $opts{gemconn};
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
                if($opts{filermode} ne "normal"){
                        $opts{conntype} = "#console";
                }

                $c = make_conn($opts{host},$opts{conntype});
                if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
                        croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
                }
                if(!defined($opts{AorB})){
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
                }
                else{
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
                }

                if($opts{filermode} ne "normal"){

                        $cmd = "sasadmin expander_cli $id \"$command\"";
                }
                else{
                        $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
                }

        }

    else {
                croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
        }

#       Issue Command

        if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
                (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
        }
        else{
                (undef,undef,$output) = say($c,"$cmd");
        }


        if($output !~ /SES Page/){
                croak "Unexpected ses_report_page output: $output";
        }

        my $phy;
        my @output = split(/\r*\n/,$output);

        print "Output is : @output \n";
        foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /Global\sModule\sControl/) {
                        my @element = split(/:\s*/,$line);
                        $hash{GlobalModuleControl} = $element[1];
                }
                elsif($line =~ /PHY\sReset\sTimeout/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{PHYResetTimeout} = $element[1];
                }
                elsif($line =~ /AWT\sBias/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{AWTBias} = $element[1];
                }
                elsif($line =~ /PHYs/) {
                        my @tmp = split(/ /,$line);
                        $phy = $tmp[0];
                }
        elsif($line =~ /Profile/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{Profile} = $element[1];
                }
                elsif($line =~ /Bit\sError\sRate\sSample\sWindow/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{BitErrorRateSampleWindow} = $element[1];
                }
                elsif($line =~ /Reset Problem Sample Window/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{ResetProblemSampleWindow} = $element[1];
                }
                elsif($line =~ /PHY CHANGE Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{PHYCHANGEThreshold} = $element[1];
                }
                elsif($line =~ /Invalid Dword Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{InvalidDwordThreshold} = $element[1];
                }
                elsif($line =~ /Running Disparity Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{RunningDisparityThreshold} = $element[1];
                }
                elsif($line =~ /Loss of DWORD Sync Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{LossofDWORDSyncThreshold} = $element[1];
                }
                elsif($line =~ /PHY Reset Problem Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{PHYResetProblemThreshold} = $element[1];
                }
                elsif($line =~ /CRC Error Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{CRCErrorThreshold} = $element[1];
                }
        }
        return \%hash;
}

=head2  ddump_hid_byIndex

    my $ret = ddump_hid_byIndex(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the LID data and save in a hash table
    with the following keys stated for the Drive Manager and for each Device.
        HAMode. driveSpinUp, driveBays, allowedDrives, drivePowerControl, deviceRole,
        instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS
    For each device bay: present, swap, SESInfoBit, spinUpTime, driveType, name, arrayName,
    wwn, faults, faultLED, inject, pending, current.

        Example of returning ddump_hid info:
        $info = ddump_hid();
        print "the fault LED status for drive bay 9 is: $info->{9}{faultLED}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'HAMode'  => 'master',
          'driveSpinUp'  => '0',
          'driveBays'  => '24',
          'driveIndexBase'  => '0',
          'allowedDrives'  => 'SAS OR SATA',
          'drivePowerControl'  => 'supported',
          'deviceRole' '' => 'MASTER,
          'instanceRunState'  => 'RUNNING',
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x63E8',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0x40 (64)',
          'transPoolFreeCount'  => '0x40 (64)',
          'numConACKS'  => '0x48 (72)',
          '0' 'present' => 'yes',
          '0' 'swap' => 'no',
          '0' 'SESInfoBit' => 'not set',
          '0' 'spinUpTime' => '0+00:00:08.087',
          '0' 'driveType' => 'SAS',
          '0' 'wwn' => '500000E0173F6792',
          '0' 'faults' => 'none',
          '0' 'faultLED' => 'OFF',
          '0' 'inject' => 'NONE',
          '0' 'pending' => 'ONLINE',
          '0' 'current' => 'ONLINE',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_hid_byIndex {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_hid";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }

#   Issue Command

    if($output !~ /HID diagnostic dump/){
        croak "Unexpected ddump_hid output: $output";
    }

    my %hash = ();
    my $bay;
    my @output = split(/\r*\n/,$output);
    my $index = "";

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Total number of LEDs/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{total_leds} = $element[1];
        }
        elsif($line =~ /^Index           \s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $index = $element[1];
        }
        elsif($line =~ /Purpose\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Purpose} = $element[1];
        }
        elsif($line =~ /Type\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Type} = $element[1];
        }
        elsif($line =~ /Supported Colours\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Supported_Colours} = $element[1];
        }
        elsif($line =~ /User Info\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{User_Info} = $element[1];
        }
        elsif($line =~ /Current Status\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Status} = $element[1];
        }
    }
    return \%hash;
}

=head2 storage_show_diskled

        my $ret = storage_show_diskled(host=>$host, adapter_name=>$adapter_name,timeout => $timeout);

This routine returns a hash table of information for disk led status.

It returns 1 on success and croaks on failure.
The following named arguments are accepted:

=over

=item host

(Required) : Hostspec to filer.

=item adapter_name

(Optional) : Specific channel for which the information is
 required. Default is all

=item timeout

(Optional) : timeout value. Default is 300 secs.

=back
=cut

sub storage_show_diskled($@) {
        my (%opts) = validate(@_,{
                                host => {type=>OBJECT, isa=>'Hostrec'},
                                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                                timeout => {type=>SCALAR, default=>300,optional=>1},
                                filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                }
        );
        my $adapter = undef ;
    my $output;

        my %hash =();
        if ($opts{filermode} ne "normal") {
        $opts{conntype} = "#console";
    }

        my $conn = make_conn($opts{host},$opts{conntype});
        if ($opts{filermode} ne "normal") {
                my $cmd = "storage show disk_led -a";
                (undef, undef, $output) = say($conn, "$cmd", -prompt, "$MAINT_PROMPT", "-timeout", $opts{timeout});
        }
        else{
                my $cmd = "priv set -q diag;storage show disk_led -a";
                (undef, undef, $output) = say($conn, "$cmd","-timeout", $opts{timeout});
        }
        if($output !~ /Disk name/){
                croak $output;
        }
        my $hba;

        my @lines = split(/\r*\n/, $output);

    foreach my $line (@lines) {
        $line =~ s/^\s+//g;
        if (($line !~ /Disk name/) && ($line !~ /----/)) {
            my @disk_t1 = split(/\s+/,$line);
            my @disks = split(/\./,$disk_t1[0]);
            my $disk = $disks[1].".".$disks[2];
            $hash{$disk} = $disk_t1[1];
        }
    }
        return \%hash;
}


sub psmi_dump_regs {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "psmi_dump_regs";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }
    print "Output is  C: $output \n";
    my @output = split(/\r*\n/,$output);
    my $register = 0;
    my $count = 0;

    my $value;
    foreach my $line (@output){
        if ($line =~ /: /) {
            my @line_temp = split(/: /,$line);
            my @output_temp = split(//,$line_temp[1]);
            foreach $value (@output_temp){
                if($value =~ /\d|[a-f]/){ # To make sure that the $value is hex and filter out the rest
                    if ($count == 0) {
                        $hash{$register} = "$value";
                    } elsif ($count < 4) {
                        $hash{$register} = $hash{$register}."$value";
                    }
                    $count = $count + 1;
                    if ($count > 3) {
                        #print "value in dump psmi Register $register is value $value\n";
                        $register = $register + 1;
                        $count = 0;
                    }

                }
            }
        }
    }
    return \%hash ;
}


=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    Anish Vashist (anishv@netapp.com)
    Last revised $Date: 2013/09/30 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;
## ==== THEIRS //depot/prod/test/Rsierranevada/lib/Storage/Sapphire.pm#3
#$Id: //depot/prod/test/Rsierranevada/lib/Storage/Sapphire.pm#6 $
#
## @summary Sapphire Command and Utility library
#
## @status TESTME
##
## @author anishv@netapp.ocm
##
## Added Support for cable status UNKNOWN in cablemgr_tp
#
## @pod here

=head1 NAME

Sapphire

=head1 SYNOPSIS

    use Sapphire qw(function_name);

    my $host = host($SHELF);

    my $hash = function_name(host=>$host,...);

=head1 DESCRIPTION
    This is a library module for the sasadmin command

=head1 ROUTINES

=cut
## @change 03-19-2012 rpradeep: Modified the subroutine gem_cmd to support expander diag testcase
## @change 05-21-2012 rajkuma1: Modified the subroutine phy_report to support ruckus system
## @change 07-19-2012 vickie: Modified subroutine expmgr_tp for parsing inconstent expmgr_tp output in IOM6 and IOM6E
## @change 03-04-2013 ssrinath: Modified reboot function to support CMode

package Storage::Sapphire;

use strict;
use warnings;
use Carp qw(confess croak);
use Params::Validate qw(validate validate_pos SCALAR OBJECT HASHREF);
use Storage::Sasadmin qw(sasadmin_expander_cli);
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}
use Tharn qw(say script_result host logcomment logresult param);

use base qw(Exporter);
use Utils qw(make_conn);
use Storage::Misc qw(filer_cmd);
use Data::Dumper;


our  @EXPORT_OK = qw(
            phy_report
            expctrl_tp
            expmgr_tp
            exp_report
            ses_statuspage
            reboot
            gem_cmd
            phy_disable
            phy_enable
            inc_phy_counters
            reset_error_counters
            get_drive_status
            ver
            get_psu_status
            get_fan_status
            get_temp_status
            ddump_general
            ddump_drvmgr
            get_midplanevpd
            dumpdrives
            ddump_expmgr
            ddump_acpmgr
            qsfpinfo
            cablemgr_tp
            phydump
            controller_state
            watchdog_tp
            ses_report_page2
            ses_report_page8081
            ses_report_page8283
            ddump_hid_byIndex
            storage_show_diskled
            psmi_dump_regs
            );

our $MAINT_PROMPT = "\\*\>";


=head2 phy_report

    my $ret = phy_report(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine returns a hash table of information of phy_report command
It returns hash on success and croaks on failure.

Example of reference returned by the function:
$VAR1 = {

          '33' => {

                    'phyRstEr' => '0',

                    'syncEr' => '0',

                    'PhyType' => 'HOST',

                    'crcEr' => '0',

                    'LogicalPhy' => '0',

                    'State' => '3.0G',

                    'dwEr' => '0',

                    'dpEr' => '0',

                    'ChgCnt' => '3'

                  },

          '32' => {

                    'phyRstEr' => '0',

                    'syncEr' => '0',

                    'PhyType' => 'HOST',

                    'crcEr' => '0',

                    'LogicalPhy' => '0',

                    'State' => '3.0G',

                    'dwEr' => '0',

                    'dpEr' => '0',

                    'ChgCnt' => '3'

                  },

}



The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub phy_report {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "phy_report";

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }

    #print "Output is  C: $output \n";

    if($output !~ /phyRstEr/){
        croak "Unexpected phy_report output: $output";
    }

    my @output = split(/\r*\n/,$output);

    foreach my $line (@output){

        $line =~ s/^\s+//g;
        my @element = split(/\s+/,$line);
            #print "LINE: $line \n";
        if($element[0] =~ /^DRV$|^HOST$|^IM$|^SIL$|^UKWN$/){
            #print "LINE: $line \n";
            #print "Elements are : $#element \n";
            #print "phy type is : $element[0]  \n";
            $hash{$element[2]}{PhyType} = $element[0];
            $hash{$element[2]}{LogicalPhy} = $element[1];
            $hash{$element[2]}{State} = $element[3];
            $hash{$element[2]}{ChgCnt} = $element[4];
            $hash{$element[2]}{dwEr} = $element[5];
            $hash{$element[2]}{dpEr} = $element[6];
            $hash{$element[2]}{syncEr} = $element[7];
            $hash{$element[2]}{phyRstEr} = $element[8];
            $hash{$element[2]}{crcEr} = $element[9];
        }
        elsif($element[0] =~ /^DRV\d+$/){
            #print "LINE: $line \n";
            #print "Elements are : $#element \n";
            my $drv = $element[0];
            $drv =~ s/DRV//g;
            #print "Drive is $drv \n";
            $hash{$element[1]}{PhyType} = 'DRV';
            $hash{$element[1]}{LogicalPhy} = $drv;
            $hash{$element[1]}{State} = $element[2];
            $hash{$element[1]}{ChgCnt} = $element[3];
            $hash{$element[1]}{dwEr} = $element[4];
            $hash{$element[1]}{dpEr} = $element[5];
            $hash{$element[1]}{syncEr} = $element[6];
            $hash{$element[1]}{phyRstEr} = $element[7];
            $hash{$element[1]}{crcEr} = $element[8];
        }
    }
    return \%hash ;
}

#---------------------------------------------------------------------------------------------
#   phy_report function END
#---------------------------------------------------------------------------------------------

=head2 expctrl_tp

    my $ret = expctrl_tp(host =>$host, conmode=>"shelfcli",timeout => $timeout,tp_code=>"1", params=>"$phy");

This routine injects expctrl level error injection using expctrl_tp command
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub expctrl_tp {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>30,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                tp_code => {type=>SCALAR},
                params => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "expctrl_tp $opts{tp_code} $opts{params}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }
    else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
        say($c,"\n");
    }
    print "Output is  C: $output \n";
    if($output =~ /Usage:|Failed to/){
        croak "expctrl_tp failed with : $output";
    }
    return $output;
}

=head2 expmgr_tp


This routine executes expmgr_tp command with various parameters passed and returns the output either in a string or as a reerence to an Hash based on the options/parameters passed to expmgr_tp.
See the Usage section below for more details and exapmles.

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expmgr_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=item board

(Optional) : Board number it should be either "0" or "1". Default is " ".

=item seqNum

(Optional) : Sequence number of the respective module. Default is " ". Required only for tp_code is "4".

=item options

(Optional) : Required only when tp_code is 11 to specify the Action parameter. Default is " ".


=item Usage

Example 1:
 my $ret = expmgr_tp(host =>$host, conmode=>"shelfcli",timeout => $timeout,tp_code=>"1", params=>"$phy");

        $ret contains the output string "

Example 2:

        $ret contains a reference to a nested Hash for tp_code 8.

 my $ret = expmgr_tp(host => $host, timeout => $TIMEOUT, connmode=> "inbandcli", filermode => $FILERMODE, AorB => "A", tp_code => "11", adapter_id => "1d",board => "0", shelf_id=> "0",opions => "1",params => "256");

        OUTPUT will be converted into an HASH as shown below for "tp_code 2 and 8".

Output is  C: *** Host Phy Profile ***
 BitWind RstWind ChgCnt dwErr dpErr syncEr phyRstEr crcEr
      60      60    120   120   120     60       60   120
                    CRC   PRP   LOS    RDD      IDD   PCD
                      1     1     1      1        1     1
*** Drive Phy Profile ***
 BitWind RstWind ChgCnt dwErr dpErr syncEr phyRstEr crcEr
      60      60    120   120   120     60       60   120
                    CRC   PRP   LOS    RDD      IDD   PCD
                      1     1     1      1        1     1
*** SIL Phy Profile ***
 BitWind RstWind ChgCnt dwErr dpErr syncEr phyRstEr crcEr
      60      60    120   120   120     60       60   120
                      1     1     1      1        1     1
$VAR1 = {
          'SIL' => {
                     'phyRstEr' => '60',
                     'syncEr' => '60',
                     'dpErr' => '120',
                     'crcEr' => '120',
                     'IDD' => '1',
                     'PCD' => '1',
                     'BitWind' => '60',
                     'RstWind' => '60',
                     'CRC' => '1',
                     'dwErr' => '120',
                     'ChgCnt' => '120',
                     'PRP' => '1',
                     'LOS' => '1',
                     'RDD' => '1'
                   },
       'Drive' => {
                       'phyRstEr' => '60',
                       'syncEr' => '60',
                       'dpErr' => '120',
                       'crcEr' => '120',
                       'IDD' => '1',
                       'PCD' => '1',
                       'BitWind' => '60',
                       'RstWind' => '60',
                       'CRC' => '1',
                       'dwErr' => '120',
                       'ChgCnt' => '120',
                       'PRP' => '1',
                       'LOS' => '1',
                       'RDD' => '1'
                     },
          'Host' => {
                      'phyRstEr' => '60',
                      'syncEr' => '60',
                      'dpErr' => '120',
                      'crcEr' => '120',
                      'IDD' => '1',
                      'PCD' => '1',
                      'BitWind' => '60',
                      'RstWind' => '60',
                      'CRC' => '1',
                      'dwErr' => '120',
                      'ChgCnt' => '120',
                      'PRP' => '1',
                      'LOS' => '1',
                      'RDD' => '1'
                    }
        };

Example :3

Returns the Hash for tp_code 2 output.
Output is  C: Expander 0 Information:
         Swap: No
         Updated: Yes
         ASA: No
         CLK status: OK
         Reg Access: OK

        $VAR1 = {
          'ASA' => 'No',
          'CLK' => 'OK',
          'Information' => '0',
          'Reg' => 'OK',
          'Swap' => 'No',
          'Updated' => 'Yes'
        };




=back

=cut
sub expmgr_tp {

        my (%opts) = validate(@_,{
                                host => {type=>OBJECT,isa=>'Hostrec'},
                                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                                timeout => {type=>SCALAR, default=>30,optional=>1},
                                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                                AorB => {type=>SCALAR, default=>undef,optional=>1},
                                tp_code => {type=>SCALAR},
                                board => {type=>SCALAR, default=>" ",optional=>1},
                                params => {type=>SCALAR, default=>" ",optional=>1},
                                seqNum => {type=>SCALAR, default=>" ",optional=>1},
                                options=>{type=>SCALAR, default=>"",optional=>1},

                }
        );

        my %hash = ();
        my $output;
        my $c;
        my $cmd;
        my $id;
        my $command = "expmgr_tp $opts{tp_code} $opts{board} $opts{params}";
        if ($opts{tp_code} eq "4") { $command = "expmgr_tp $opts{tp_code} $opts{seqNum} $opts{board}";}
        elsif ($opts{tp_code} eq "9" || "11") { $command = "expmgr_tp $opts{tp_code} $opts{board} $opts{params} $opts{options} ";}

#       Make Connection and build command based on input parameters

        if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
                $opts{conntype} = "#console";
                $c = make_conn($opts{host},$opts{conntype});
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
                if(!defined($opts{gemconn})){
                        croak "GEMTEST needs to be connected before we can use gemtest connmode";
                }
                $c = $opts{gemconn};
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
                if($opts{filermode} ne "normal"){
                        $opts{conntype} = "#console";
                }
                $c = make_conn($opts{host},$opts{conntype});
                if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
                        croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
                }
                if(!defined($opts{AorB})){
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
                }
                else{
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";                }

                if($opts{filermode} ne "normal"){
                        $cmd = "sasadmin expander_cli $id \'$command\'";
                }
    else{
                        $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
                }

        }

        else {
                croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
        }

#       Issue Command
        if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
                (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
        }
        elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
        }
        else{
                ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
                say($c,"\n");
        }
        print "Output is  C: $output \n";
        if($output =~ /Usage:|Failed to/){
                croak "expmgr_tp failed with : $output";
        }


        if ($opts{tp_code} eq "2"){

         my @output = split (/\r*\n/,$output);

        my @Information = grep /Information/,@output;
        my $Information= "@Information";
        my @information= split(/\s/,$Information);
        my $INFO=$information[1];

        my @Swap = grep /Swap/,@output;
        my $Swap = "@Swap";
        my @swap = split(/:\s/,$Swap);
        $Swap = $swap[1];

    my @Updated = grep /Updated/,@output;
        my $Updated = "@Updated";
        my @updated = split(/:\s/,$Updated);
        $Updated = $updated[1];

        my @ASA = grep /ASA/,@output;
        my $ASA = "@ASA";
        my @asa = split(/:\s/,$ASA);
        $ASA = $asa[1];

        my @CLK_status = grep /CLK/,@output;
        my $CLK_status = "@CLK_status";
        my @CLK = split(/:\s/,$CLK_status);
        my $CLK = $CLK[1];

        my @Reg_access = grep /Reg/,@output;
        my $Reg_access = "@Reg_access";
        my @Reg = split(/:\s/,$Reg_access);
        my $Reg = $Reg[1];

        my %OUTPUT = ( 'Information'=>"$INFO",
                'Swap'=>"$Swap",
                'Updated'=>"$Updated",
                'ASA'=>"$ASA",
                'CLK'=>"$CLK",
                'Reg'=>"$Reg" );
        return %OUTPUT;

        }
        elsif ($opts{tp_code} eq "8" )
        {
                my %Result;
                my %hash;
                my @output = split (/\r*\n/,$output);
                for( my $i=0;$i<=$#output;$i++)
                {
                if ($output[$i]=~/Phy\sProfile/)
                {
                  my @element = split(/\s+/,$output[$i+2]);
                  shift @element;
                  my @element1 = split(/\s+/,$output[$i+4]);
                  shift @element1;
          $hash{BitWind}=$element[0];
                  if ($element[1] =~ /(m)/) {
                      $hash{RstWind}=$element[2];
                      $hash{ChgCnt}=$element[4];
                      $hash{dwErr}=$element[5];
                      $hash{dpErr}=$element[6];
                      $hash{syncEr}=$element[7];
                      $hash{phyRstEr}=$element[8];
                      $hash{crcEr}=$element[9];
                  } else {
                      $hash{RstWind}=$element[1];
                      $hash{ChgCnt}=$element[2];
                      $hash{dwErr}=$element[3];
                      $hash{dpErr}=$element[4];
                      $hash{syncEr}=$element[5];
                      $hash{phyRstEr}=$element[6];
                      $hash{crcEr}=$element[7];
                  }
                  $hash{CRC}=$element1[0];
                  $hash{PRP}=$element1[1];
                  $hash{LOS}=$element1[2];
                  $hash{RDD}=$element1[3];
                  $hash{IDD}=$element1[4];
                  $hash{PCD}=$element1[5];
                }
                if ($output[$i]=~/Host/)
                {       my %host_hash=%hash;
                        $Result{Host}=\%host_hash;}
                elsif ($output[$i]=~/Drive/)
                {       my %Drive_hash=%hash;
                        $Result{Drive}=\%Drive_hash;}
                elsif ($output[$i]=~/SIL/)
                {       my %SIL_hash=%hash;
                        $Result{SIL}=\%SIL_hash;}
                elsif ($output[$i]=~/\*\*\*\sPhy Profile\s\*\*\*/)
                {
                        if ($opts{params} eq "0")
                        {$Result {Host}=\%hash;}
                        elsif ($opts{params} eq "1")
                        {$Result {Drive}=\%hash;}
                        elsif ($opts{params} eq "2")
                        {$Result {SIL}=\%hash;}

                }
                }

                return \%Result;

        }



return $output;

}



=head2 exp_report

    my $ret = exp_report(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine returns a hash table of information of exp_report command
It returns hash on success and croaks on failure.

Example of reference returned by the function:

$VAR1 = {

          'TotalPhyCount' => '36',

          'RegAccess' => 'OK',

          'ClkStatus' => 'OK'

        };



The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub exp_report {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "exp_report";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }


    #print "Output is  C: $output \n";

    if($output !~ /Expander Information:/){
        croak "Unexpected phy_report output: $output";
    }

    my @output = split(/\r*\n/,$output);

    foreach my $line (@output){
        $line =~ s/^\s+//g;
        #my @element = split(/\s+/,$line);
        if($line =~ /CLK status:/){
            my @element = split(/CLK status:\s*/,$line);
            $hash{ClkStatus} = $element[1];
        }
        elsif($line =~ /Reg Access:/){
            my @element = split(/Reg Access:\s*/,$line);
            $hash{RegAccess} = $element[1];
        }
        elsif($line =~ /Total phy count:/){
            my @element = split(/Total phy count:\s*/,$line);
            $hash{TotalPhyCount} = $element[1];
        }
    }
    return \%hash ;
}

=head2 ses_statuspage

    my $ret = ses_statuspage(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine returns a hash table of information of exp_report command
It returns hash on success and croaks on failure.

Example of reference returned by the function:

$VAR1 = {

          '33' => '33',

          '32' => '34',

          '90' => '00',

          '63' => '00',

          '21' => '59',

};


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub ses_statuspage {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                page_num => {type=>SCALAR},
                retry => {type=>SCALAR, default=>5 ,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ses_statuspage $opts{page_num}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = sasadmin_expander_cli(host=>$opts{host},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});

    }

    print "Output is  C: $output \n";
    my @output = split(/\s+|\r*\n/,$output);
    my $byte = 0;
    my $value;
    foreach $value (@output){
        if($value =~ /^(\d|[a-f])(\d|[a-f])$/){ # To make sure that the $value is hex and filter out the rest
            print "Byte $byte value is $value \n";
            $hash{$byte} = "$value";
            $byte++;
        }
    }
    return \%hash ;

}



=head2 reboot

    my $ret = reboot(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine reboots a module
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub reboot {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>30,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                params => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "reboot $opts{params}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id '$command'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id '$command'";
        }
    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        $output = $c->execute(command => $cmd);
    }
    elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
        filer_cmd(host=>$opts{host},filermode=>$opts{filermode},conntype=>$opts{conntype},timeout=>$opts{timeout},cmd=>$cmd);
    }
    else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},-noprompt,-prompt,"HA Service");
        print "Output is $output \n";
        if($output =~ /failed/){
            croak "Saw failed string during boot:\n $output";
        }
        say($c,"\n",-swallow);
    }
    #print "Output is  C: $output \n";
    if($output =~ /Usage:|Failed to/){
        croak "reboot failed with : $output";
    }
    return $output;
}


=head2 gem_cmd

    my $ret = gem_cmd(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine issues a command to the Sapphire shelf fw
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub gem_cmd {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                                node => {type=>HASHREF, default=>undef, optional=>1},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                cmd => {type=>SCALAR},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                pmc_cmd => {type=>SCALAR, default=>undef,optional=>1},
                retry => {type=>SCALAR, default=>5 ,optional=>1},
                diag_cmd => {type=>SCALAR, default=>undef,optional=>1},
                prompt_check => {type=>SCALAR, default=>undef,optional=>1}
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = $opts{cmd};
    my $PMC_PROMPT = "^*0000>";

    if(defined($opts{pmc_cmd}) && ($opts{connmode} ne "shelfcli")){
        croak "PMC commands only work with shelfcli connection mode";
    }


    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "acpcli"){ # Connection when we are directly connected to the acp console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        #if($opts{filermode} ne "normal"){
        #   $opts{conntype} = "#console";
        #}

        #$c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        #if(!defined($opts{AorB})){
        #   $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        #}
        #else{
        #   $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        #}
                if (defined $opts{node}) {
                    $output = sasadmin_expander_cli(host=>$opts{host},node=>$opts{node},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});
                } else {
                    $output = sasadmin_expander_cli(host=>$opts{host},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});
                }
    }
    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli , acpcli and gemtest";
    }


    if(($opts{connmode} eq "acpcli" ) && ($cmd eq "root")){
        (undef,undef,$output) = say($c, "$cmd", '-noprompt');
    }

    elsif($opts{connmode} eq "shelfcli"){
        if(defined($opts{pmc_cmd}) && ($cmd ne "gem") && ($cmd ne "al2o3")){
                (undef,undef,$output) = say($c,"$cmd",-prompt,"$PMC_PROMPT");
        }elsif(defined($opts{diag_cmd}) && defined ($opts{prompt_check})){
                    (undef,undef,undef,$output) = say($c, "help", -prompt, '\*> |\$');
            }
         elsif((defined($opts{diag_cmd})) && !defined ($opts {prompt_check})){
                    if($cmd eq "C"){
                        (undef,undef,$output) = say($c, "$cmd","-wait",".","-noprompt");
                    }
                    elsif ($cmd eq "exit"){
                        (undef,undef,$output) = say($c, "$cmd","-wait",".",-prompt,"^*Al2O3>");
                    }
                    else{
                                (undef,undef,$output) = say($c, "$cmd",-timeout, 120 ,"-wait",".","-prompt", '^\$');
                    }

        }

        else{
            (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
        }
    }
    else{
    }
    return $output;
}


=head2 phy_disable

    my $ret = phy_disable(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0",disable_reason=>"0");

This routine disables a phy


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy to be disabled.  If we pass "all" then it will disable all the enabled phys. We can also pass list of phys separated by , or space.

=item disable_reason

(Required) : Disable reason to disable a phy

Valid options are 0 to 16

=back

=cut

sub phy_disable {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
                disable_reason => {type=>SCALAR},
        }
    );
    my %disable_reason = (
            "0" => "DIS\/MAN_SES",
            "1" => "DIS\/NOT_PRESENT",
            "2" => "DIS\/BRST_LOS",
            "3" => "DIS\/LOS",
            "4" => "DIS\/BRST_RDD",
            "5" => "DIS\/RDD",
            "6" => "DIS\/BRST_IDD",
            "7" => "DIS\/IDD",
            "8" => "DIS\/BRST_PRP",
            "9" => "DIS\/PRP",
            "10" => "DIS\/BRST_PCD",
            "11" => "DIS\/PCD",
            "12" => "DIS\/BRST_CRC",
            "13" => "DIS\/CRC",
            "14" => "DIS\/MIR",
            "15" => "DIS\/MAN_SMP",
            "16" => "DIS\/CLK_FAULT",
            );


    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my @enabled_phys;
    my $dis_reason = $opts{disable_reason};
    unless(grep(/$dis_reason/,keys(%disable_reason))){
        logresult("script","$dis_reason is not a valid disable reason. Valid Disable reasons are from 0 to 16.");
    }
    my $err = 0;
    my $phy_report;
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        my @phys = keys(%$phy_report);
        foreach my $phy (@phys){
            if($phy_report->{$phy}->{State} =~ /3\.0G|1\.5G|6\.0G/){
                if($dis_reason == 14){
                    if($phy_report->{$phy}->{PhyType} eq "DRV"){
                        push @enabled_phys,$phy;
                    }
                }
                else{
                        push @enabled_phys,$phy;
                }
            }

        }
    }
    else{
        @enabled_phys = split(/,|\s+/,$opts{phy_id});
    }
    foreach my $phy (@enabled_phys){
        logcomment("Disabling phy $phy with reason $dis_reason => $disable_reason{$dis_reason}");
        if($opts{connmode} eq "shelfcli"){
            expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>2,params=>"$phy $dis_reason");
        }
        elsif($opts{connmode} eq "inbandcli"){
            expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>2,params=>"$phy $dis_reason}");
        }
        elsif($opts{connmode} eq "acpcli"){
        }
        elsif($opts{connmode} eq "inbandacp"){
        }
        else{
            logresult("script", "Unsupported connmode");
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@enabled_phys){
        unless($phy_report->{$phy}->{State} =~ /$disable_reason{$dis_reason}/){
                logcomment("FAIL => Unable to disable phy $phy with reason $disable_reason{$dis_reason}");
                $err = 1;
            }
        }
        if($err == 1){
            logresult("fatal", "Unexpected disable state. Please check the logs above");
        }
}

=head2 phy_enable

    my $ret = phy_enable(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0",enable_reason=>"2");

This routine enables a phy
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy to be enabled.  If we pass "all" then it will enable all the disabled phys. We can also pass list of phys separated by , or space.

=item disable_reason

(Optional) : To enable a phy which is disabled because of a particular disable reason. By default it will try to enable all the disabled phys

=item enable_reason

(Required) : To enable a phy with particular enable_reason. Valid options are  0 to 6.


=back

=cut

sub phy_enable {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
                enable_reason => {type=>SCALAR},
                disable_reason => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    # Hash of supported enable reasons
    my %enable_reason = (
        "0" => "MANUAL_SES",
        "1" => "DRIVE_INSERT",
        "2" => "CABLE_INSERT",
        "3" => "PARTNER_EBOD_INSERT",
        "4" => "DRIVE_POWER_ON",
        "5" => "MIRROR",
        "6" => "MANUAL_SMP",
    );

    my %disable_reason = (
            "0" => "DIS\/MAN_SES",
            "1" => "DIS\/NOT_PRESENT",
            "2" => "DIS\/BRST_LOS",
            "3" => "DIS\/LOS",
            "4" => "DIS\/BRST_RDD",
            "5" => "DIS\/RDD",
            "6" => "DIS\/BRST_IDD",
            "7" => "DIS\/IDD",
            "8" => "DIS\/BRST_PRP",
            "9" => "DIS\/PRP",
            "10" => "DIS\/BRST_PCD",
            "11" => "DIS\/PCD",
            "12" => "DIS\/BRST_CRC",
            "13" => "DIS\/CRC",
            "14" => "DIS\/MIR",
            "15" => "DIS\/MAN_SMP",
            "16" => "DIS\/CLK_FAULT",
            );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my @disabled_phys;
    my $en_reason = $opts{enable_reason};
    unless(grep(/$en_reason/,keys(%enable_reason))){
        logresult("script","$en_reason is not a valid enable reason. Valid Disable reasons are from 0 to 6.");
    }
    my $err = 0;
    my $phy_report;
    my $phy_report1;
    my $dis_reason = "DIS";
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        my @phys = keys(%$phy_report);
        foreach my $phy (@phys){
            if(defined($opts{disable_reason})){
                    my $reason = $opts{disable_reason};
                    $dis_reason = $disable_reason{$reason};
                }
            if($phy_report->{$phy}->{State} =~ /$dis_reason/){
                if($phy_report->{$phy}->{PhyType} ne 'SIL'){
                    push @disabled_phys,$phy;
                }
            }
        }
    }
    else{
        @disabled_phys = split(/,|\s+/,$opts{phy_id});
    }
    print "DISABLED phys : @disabled_phys \n";
    sleep 30;
    foreach my $phy (@disabled_phys){

        if($en_reason =~ /1|2|3/){
            if($phy_report->{$phy}->{PhyType} eq 'DRV'){
                $en_reason = 1;
            }
            elsif($phy_report->{$phy}->{PhyType} eq 'HOST'){
                $en_reason = 2;
            }
            elsif($phy_report->{$phy}->{PhyType} eq 'SIL'){
                $en_reason = 3;
            }
        }

        logcomment("Enabling phy $phy with reason $en_reason => $enable_reason{$en_reason}");
        if($opts{connmode} eq "shelfcli"){
            eval { expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>1,params=>"$phy $en_reason");
            };
            if($@){
                logcomment("Phy enable failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "inbandcli"){
            eval {expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>1,params=>"$phy $en_reason}");
            };
            if($@){
                logcomment("Phy enable failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "acpcli"){
        }
        elsif($opts{connmode} eq "inbandacp"){
        }
        else{
            logresult("script", "Unsupported connmode");
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@disabled_phys){
        unless($phy_report1->{$phy}->{State} =~ /3\.0G|1\.5G|6\.0G/){
                logcomment("Unable to enable phy $phy with reason $enable_reason{$en_reason}");
                if($opts{phy_id} ne "all"){
                    $err = 1;
                }
            }
        }
        if($err == 1){
            logresult("fatal", "Unexpected enabled state. Please check the logs above");
        }
}

=head2 inc_phy_counters

    my $ret = inc_phy_counters(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0");

This routine increments phys counters.
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy for which the errors to be incremented.  If we pass "all" then it will increment counters of all the phys. We can also pass a list of phys separated by , or space.

=item count_type

(Optional) : Error to be incremented. Valid options are 0 to 5 and "all".

=item count

(Optional) : How many counters to increment. Default is 10.


=back

=cut

sub inc_phy_counters {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
                count_type => {type=>SCALAR},
                count => {type=>SCALAR},
        }
    );

    # Hash of supported enable reasons
    my %count_type = (
        "0" => "ChgCnt",
        "1" => "dwEr",
        "2" => "dpEr",
        "3" => "syncEr",
        "4" => "phyRstEr",
        "5" => "crcEr",
    );

    my %disable_reason = (
            "0" => "DIS\/MAN_SES",
            "1" => "DIS\/NOT_PRESENT",
            "2" => "DIS\/BRST_LOS",
            "3" => "DIS\/LOS",
            "4" => "DIS\/BRST_RDD",
            "5" => "DIS\/RDD",
            "6" => "DIS\/BRST_IDD",
            "7" => "DIS\/IDD",
            "8" => "DIS\/BRST_PRP",
            "9" => "DIS\/PRP",
            "10" => "DIS\/BRST_PCD",
            "11" => "DIS\/PCD",
            "12" => "DIS\/BRST_CRC",
            "13" => "DIS\/CRC",
            "14" => "DIS\/MIR",
            "15" => "DIS\/MAN_SMP",
            "16" => "DIS\/CLK_FAULT",
            );

    my $c;
    my $cmd;
    my $id;
    my @er_phys;
    my @count_type;
    if($opts{count_type} eq "all"){
        @count_type = keys(%count_type);
    }
    else{
        @count_type = split(/,|\s+/,$opts{count_type});
        foreach my $cnt (@count_type){
            unless(grep(/$cnt/,keys(%count_type))){
            logresult("script","$cnt is not a valid error counter. Valid Disable reasons are from 0 to 5.");
            }
        }
    }
    my $err = 0;
    my $phy_report;
    my $phy_report1;
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        @er_phys = keys(%$phy_report);
    }
    else{
        @er_phys = split(/,|\s+/,$opts{phy_id});
    }
    foreach my $phy (@er_phys){
        foreach my $count_type (@count_type){
            my $cnt_type = $count_type{$count_type};
            logcomment("Incrementing counter $count_type =>  $cnt_type for phy $phy by $opts{count}");
            if($opts{connmode} eq "shelfcli"){
                eval { expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>4,params=>"$phy $count_type $opts{count}");
                };
                if($@){
                    logcomment("Error Increment failed with error: $@");
                }
            }
            elsif($opts{connmode} eq "inbandcli"){
                eval {expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>4,params=>"$phy $count_type $opts{count}}");
                };
                if($@){
                    logcomment("Error increment failed with error: $@");
                }
            }
            elsif($opts{connmode} eq "acpcli"){
            }
            elsif($opts{connmode} eq "inbandacp"){
            }
            else{
                logresult("script", "Unsupported connmode");
            }
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@er_phys){
        foreach my $count_type (@count_type){
            my $cnt_type = $count_type{$count_type};
            my $cnt = $phy_report->{$phy}->{$cnt_type};
            my $cnt1 = $phy_report1->{$phy}->{$cnt_type};
            my $cnt2 = $cnt + $opts{count};
            if($cnt1 != $cnt2){
                logcomment("FAIL: Error $count_type => $cnt_type for phy $phy is expected $cnt2 but actual is $cnt1");
                $err = 1;
            }
        }
    }
        if($err == 1){
            print Dumper($phy_report1);
            print Dumper($phy_report);
            logresult("fatal", "Unexpected enabled state. Please check the logs above");
        }
}

=head2 reset_error_counters

    my $ret = reset_error_counters(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0");

This routine resets error counters of a phy or a set of phys
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy(s) for which the errors to be reset.  If we pass "all" then it will reset counters of all the phys. We can also pass list of phys separated by , or space.


=back

=cut
sub reset_error_counters{

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
        }
    );

    # Hash of supported enable reasons
    my %count_type = (
        "0" => "ChgCnt",
        "1" => "dwEr",
        "2" => "dpEr",
        "3" => "syncEr",
        "4" => "phyRstEr",
        "5" => "crcEr",
    );

    my $c;
    my $cmd;
    my $id;
    my @er_phys;
    my $err = 0;
    my $phy_report;
    my $phy_report1;
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        @er_phys = keys(%$phy_report);
    }
    else{
        @er_phys = split(/,|\s+/,$opts{phy_id});
    }
    foreach my $phy (@er_phys){
        logcomment("Clearing counters for phy $phy");
        if($opts{connmode} eq "shelfcli"){
            eval { expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>3,params=>"$phy");
            };
            if($@){
                logcomment("Error Increment failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "inbandcli"){
            eval {expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>3,params=>"$phy");
            };
            if($@){
                logcomment("Error increment failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "acpcli"){
        }
        elsif($opts{connmode} eq "inbandacp"){
        }
        else{
            logresult("script", "Unsupported connmode");
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@er_phys){
        foreach my $cnt_type (keys(%count_type)){
            my $cnt = $phy_report1->{$phy}->{$count_type{$cnt_type}};
            print "CNT TYPE: $cnt_type and CNT is $cnt \n";
            if(($count_type{$cnt_type} ne "ChgCnt") && ($cnt != 0)){
                logcomment("FAIL: Error counter $count_type{$cnt_type} for phy $phy is $cnt . Expected zero");
                $err = 1;
            }
        }
    }
        if($err == 1){
            logresult("fatal", "Unexpected enabled state. Please check the logs above");
        }
}

=head2 get_drive_status

    my $ret = get_drive_status(drive=><drive>,host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine retrieves a specific drive's status by issuing shelf command "getdrivestatus" and
    save in a hash table with the following keys stated:
    present, swap, SES_info_bit, SpinUpTime, drive_type, name, arrayname, wwn, faults,  faultLED, inject, pending, current

    Example of returning a specific drive's status:
    my $drivestatus = get_drive_status($drive);
    if ($drivestatus->{present} =~ /yes/) {
        print "the drive $drive is present\n";
    }

    Example of reference returned by the function:
    $VAR1 = {
          'present' => 'yes',
          'swap' => 'no',
          'SES_info_bit' => 'not set',
          'SpinUpTime' => '0+00:00:08.087',
          'drive_type' => 'SAS',
          'wwn' => '500000E0173F6792',
          'faults' => 'none',
          'faultLED' => 'OFF',
          'inject' => 'NONE',
          'pending' => 'ONLINE',
          'current' => 'ONLINE',
          'PwrCycleTime' => '0',
          'PwrInProgress' => 'NO',
          'InProgressAS' => 'NONE',
          'InProgressTS' => 'NONE',
          'ReservedCntr' => '0 (00000000)',
          'SMPPwrOnCnt' => '0',
          'SMPPwrOffCnt' => '0',
          'SMPPwrCycCnt' => '0',
          'ActiveCounter' => '0',
          'ActiveCntrAS' => 'NONE',
          'ActiveCntrTS' => 'NONE',
    };

The following arguments are accepted:

=over

=item drive

(Required) : The drive which status will be returned.

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_drive_status {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                disk_id => {type=>SCALAR, default=>undef},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "getdrivestatus $opts{disk_id}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }

    if($output !~ /Drive Bay/){
        croak "Unexpected getdrivestatus output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output drive opts{disk_id} status: @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /present/) {
            my @element = split(/:/,$line);
            $hash{present} = $element[1];
        }
        elsif($line =~ /swap/) {
            my @element = split(/:/,$line);
            $hash{swap} = $element[1];
        }
        elsif($line =~ /SES_info_bit/) {
            my @element = split(/:/,$line);
            $hash{SES_info_bit} = $element[1];
        }
        elsif($line =~ /spin up time/) {
            my @element = split(/:/,$line);
            $hash{SpinUpTime} = $element[1];
        }
        elsif($line =~ /drive_type/) {
            my @element = split(/:/,$line);
            $hash{drive_type} = $element[1];
        }
        elsif($line =~ /WWN/) {
            my @element = split(/:/,$line);
            $hash{wwn} = $element[1];
        }
        elsif($line =~ /faults/) {
            my @element = split(/:/,$line);
            $hash{faults} = $element[1];
        }
        elsif($line =~ /fault LED/) {
            my @element = split(/:/,$line);
            $hash{faultLED} = $element[1];
        }
        elsif($line =~ /inject/) {
            my @element = split(/:/,$line);
            $hash{inject} = $element[1];
        }
        elsif($line =~ /pending/) {
            my @element = split(/:/,$line);
            $hash{pending} = $element[1];
        }
        elsif($line =~ /current/) {
            my @element = split(/:/,$line);
            $hash{current} = $element[1];
        }
        elsif($line =~ /Pwr Cycle Time/) {
            my @element = split(/:/,$line);
            $hash{PwrCycleTime} = $element[1];
        }
        elsif($line =~ /Pwr InProgress/) {
            my @element = split(/:/,$line);
            $hash{PwrInProgress} = $element[1];
        }
        elsif($line =~ /InProgressAS/) {
            my @element = split(/:/,$line);
            $hash{InProgressAS} = $element[1];
        }
        elsif($line =~ /InProgressTS/) {
            my @element = split(/:/,$line);
            $hash{InProgressTS} = $element[1];
        }
        elsif($line =~ /ReservedCntr/) {
            my @element = split(/:/,$line);
            $hash{ReservedCntr} = $element[1];
        }
        elsif($line =~ /SMP PwrOn cnt/) {
            my @element = split(/:/,$line);
            $hash{SMPPwrOnCnt} = $element[1];
        }
        elsif($line =~ /SMP PwrOff cnt/) {
            my @element = split(/:/,$line);
            $hash{SMPPwrOffCnt} = $element[1];
        }
        elsif($line =~ /SMP PwrCyc cnt/) {
            my @element = split(/:/,$line);
            $hash{SMPPwrCycCnt} = $element[1];
        }
        elsif($line =~ /Active Counter/) {
            my @element = split(/:/,$line);
            $hash{ActiveCounter} = $element[1];
        }
        elsif($line =~ /Active Cntr.AS/) {
            my @element = split(/:/,$line);
            $hash{ActiveCntrAS} = $element[1];
        }
        elsif($line =~ /Active Cntr.TS/) {
            my @element = split(/:/,$line);
            $hash{ActiveCntrTS} = $element[1];
        }
    }
    return \%hash;
}

=head2  ver

    my $ret = ver(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine retrieves all the components firmware versions by issuing shelf command
    "ver" and save in a hash table with following keys stated:
    CanisterFW, CanisterFWDate, CanisterBootloader, CanisterConfig CanisterVPD, CanisterVPDCRC.
    CanisterCPLD. CanisterChip. CanisterSDK, MidplaneVPD, MidplaneVPDCRC. MidplaneCPLD. PCM1FW.
    PCM2FW. PCM3FW. PCM4FW, PCM1VPD, PCM2VPD, PCM3VPD, PCM4VPD, PCM1VPDCRC, PCM2VPDCRC, PCM3VPDCRC, PCM4VPDCRC

    Example of returning version hash table:
    my $tb = ver();
    print "The current Canister Firmware is $tb->{CanisterFW}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'CanisterFW' => '0.A6',
          'CanisterFWDate' => 'Nov 26 2008 04:31:26',
          'CanisterBootloader' => '4.00',
          'CanisterConfigCRC' => 'Not present',
          'CanisterVPDStruc' => '0x06',
          'CanisterVPDCRC' => '0x0E4B5E5A',
          'CanisterCPLD' => '0x12',
          'CanisterChip' => '0x83980000',
          'CanisterSDK' => '2.03.02',
          'MidplaneVPDStruc' => '0x07',
          'MidplaneVPDCRC' => '0x3F8C62FB',
          'MidplaneCPLD' => '0x0E',
          'PCM1FW' => '2.09',
          'PCM2FW' => '2.09',
          'PCM3FW' => '2.09',
          'PCM4FW' => '2.09',
          'PCM1VPDStruc' => '0x02',
          'PCM2VPDStruc' => '0x02',
          'PCM3VPDStruc' => '0x02',
          'PCM4VPDStruc' => '0x02',
          'PCM1VPDCRC' => '0x60CFC396',
          'PCM2VPDCRC' => '0x60CFC396',
          'PCM3VPDCRC' => '0x60CFC396',
          'PCM4VPDCRC' => '0x60CFC396',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ver {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ver";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }

    if($output !~ /Canister firmware/){
        croak "Unexpected ver output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Canister firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterFW} = $element[1];
        }
        elsif($line =~ /Canister firmware date\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterFWDate} = $element[1];
        }
        elsif($line =~ /Canister bootloader\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterBootloader} = $element[1];
        }
        elsif($line =~ /Canister config CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterConfigCRC} = $element[1];
        }
        elsif($line =~ /Canister VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterVPDStruc} = $element[1];
        }
        elsif($line =~ /Canister VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterVPDCRC} = $element[1];
        }
        elsif($line =~ /Canister CPLD\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterCPLD} = $element[1];
        }
        elsif($line =~ /Canister chip\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterChip} = $element[1];
        }
        elsif($line =~ /Canister SDK\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterSDK} = $element[1];
        }
        elsif($line =~ /Midplane VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{MidplaneVPDStruc} = $element[1];
        }
        elsif($line =~ /Midplane VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{MidplaneVPDCRC} = $element[1];
        }
        elsif($line =~ /Midplane CPLD\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{MidplaneCPLD} = $element[1];
        }
        elsif($line =~ /PCM 1 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM1FW} = $element[1];
        }
        elsif($line =~ /PCM 2 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM2FW} = $element[1];
        }
        elsif($line =~ /PCM 3 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM3FW} = $element[1];
        }
        elsif($line =~ /PCM 4 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM4FW} = $element[1];
        }
        elsif($line =~ /PCM 1 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM1VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 2 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM2VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 3 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM3VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 4 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM4VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 1 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM1VPDCRC} = $element[1];
        }
        elsif($line =~ /PCM 2 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM2VPDCRC} = $element[1];
        }
        elsif($line =~ /PCM 3 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM3VPDCRC} = $element[1];
        }
        elsif($line =~ /PCM 4 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM4VPDCRC} = $element[1];
        }
    }
    return \%hash;
}

=head2  get_psu_status

    my $ret = get_psu_status(psu=><psu>,host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "getpsustatus" to retrieve a specific power supply unit's
    status and save in a hash table with following keys stated:
    present, type, fruid, productName, manufacturer, partNumber, partRevision, serialNumber, firmwareVersion, powerState, fansSelfPowered, nominalPowerRating, nominalPowerRatingOutputBitmask, surgePowerRating, surgePowerRatingOutputBitmask, surgeHoldTime, ACDropoutTolerance, standbyPower, coolingPower, swap, hotswapPrivate, hotswapPublic, fault,
        output1 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent
        output2 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent
        output3 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent
    combined_power, ses_info_bit.

    Example of returning a psu's status:
        my $psustatus = get_psu_status(1);
        print "the power supply 1 is present?  $psustatus->{present}\n";
        print "the power supply 1 output2 min voltage is: $psustatus->{output2}{minVoltage}\n";

    present, type, fruid, productName, manufacturer, partNumber, partRevision, serialNumber, firmwareVersion, powerState, fansSelfPowered, nominalPowerRating, nominalPowerRatingOutputBitmask, surgePowerRating, surgePowerRatingOutputBitmask, surgeHoldTime, ACDropoutTolerance, standbyPower, coolingPower, swap, hotswapPrivate, hotswapPublic, fault,
        output1 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent

    Example of reference returned by the function:
    $VAR1 = {
          'present' => 'yes',
          'type' => 'PSMI',
          'fruid' => '0x9C',
          'productName' => 'HB-PCM-01-580AC',
          'manufacturer' => 'POWERONE',
          'partNumber' => '0082562',
          'partRevision' => '07',
          'serialNumber' => 'PMW8256200043ED',
          'firmwareVersion' => '2.09',
          'powerState' => 'on',
          'fansSelfPowered' => 'yes',
          'nominalPowerRating' => '580W',
          'nominalPowerRatingOutputBitmask' => '0x0',
          'surgePowerRating' => '580W',
          'surgePowerRatingOutputBitmask' => '0x0',
          'surgeHoldTime' => '0ms',
          'ACDropoutTolerance' => '0ms',
          'standbyPower' => '10W',
          'coolingPower' => '60W',
          'swap' => 'no',
          'hotswapPrivate' => 'yes',
          'hotswapPublic' => 'yes',
          'fault' => 'no',
          'output#' 'voltage' => '5.11V',
          'output#' 'current' => '3.43A',
          'output#' 'power' => '17.58W',
          'output#' 'minVoltage' => '4.80V',
          'output#' 'minSurgeVoltage' => '4.80V',
          'output#' 'maxVoltage' => '5.19V',
          'output#' 'maxCurrent' => '36.00A',
          'output#' 'minCurrent' => '0.00A',
          'output#' 'maxSurgeCurrent' => '42.00A',
    }

The following arguments are accepted:

=over

=item psu

(Required) : Which power supply status to be retrieved.

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_psu_status {
    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                psu => {type=>SCALAR},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "getpsustatus $opts{psu}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /PCM Bay/){
        croak "Unexpected getpsustatus output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output psu $opts{psu} status: @output \n";
    my $output_num = 0;
    my $output_full ="";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /present:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{present} = $element[1];
        }
        elsif($line =~ /type:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{type} = $element[1];
        }
        elsif($line =~ /FRUid:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{fruid} = $element[1];
        }
        elsif($line =~ /product name:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{productName} = $element[1];
        }
        elsif($line =~ /manufacturer:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{manufacturer} = $element[1];
        }
        elsif($line =~ /part number:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partNumber} = $element[1];
        }
        elsif($line =~ /part revision:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partRevision} = $element[1];
        }
        elsif($line =~ /serial number:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{serialNumber} = $element[1];
        }
        elsif($line =~ /firmware version:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{firmwareVersion} = $element[1];
        }
        elsif($line =~ /power state:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{powerState} = $element[1];
        }
        elsif($line =~ /fans self-powered:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{fansSelfPowered} = $element[1];
        }
        elsif($line =~ /nominal power rating:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nominalPowerRating} = $element[1];
        }
        elsif($line =~ /nominal power rating output bitmask:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nominalPowerRatingOutputBitmask} = $element[1];
        }
        elsif($line =~ /surge power rating:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{surgePowerRating} = $element[1];
        }
        elsif($line =~ /surge power rating output bitmask:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{surgePowerRatingOutputBitmask} = $element[1];
        }
        elsif($line =~ /surge hold time:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{surgeHoldTime} = $element[1];
        }
        elsif($line =~ /AC dropout tolerance:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{ACDropoutTolerance} = $element[1];
        }
        elsif($line =~ /standby power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{standbyPower} = $element[1];
        }
        elsif($line =~ /cooling power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{coolingPower} = $element[1];
        }
        elsif($line =~ /swap:\s+/) {
        }
        elsif($line =~ /hotSwap\(private\):\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{hotswapPrivate} = $element[1];
        }
        elsif($line =~ /hotSwap\(public\):\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{hotswapPublic} = $element[1];
        }
        elsif($line =~ /fault:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{fault} = $element[1];
        }
        elsif($line =~ /output \d/) {
            my @tmp = split(/ /,$line);
            $output_num = $tmp[1];
            $output_full = "output".$output_num;
        }
        elsif($line =~ /^voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{voltage} = $element[1];
        }
        elsif($line =~ /^current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{current} = $element[1];
        }
        elsif($line =~ /power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{power} = $element[1];
        }
        elsif($line =~ /min voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{minVoltage} = $element[1];
        }
        elsif($line =~ /min surge voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{minSurgeVoltage} = $element[1];
        }
        elsif($line =~ /max voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{maxVoltage} = $element[1];
        }
        elsif($line =~ /max current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{maxCurrent} = $element[1];
        }
        elsif($line =~ /min current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{minCurrent} = $element[1];
        }
        elsif($line =~ /max surge current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{manSurgeCurrent} = $element[1];
        }
        elsif($line =~ /combined power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{combinedPower} = $element[1];
        }
        elsif($line =~ /SES_info_bit:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{sesInfoBit} = $element[1];
        }
    }
    return \%hash;
}

=head2  get_fan_status

    my $ret = get_fan_status(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "ddump_envctrl" to retrieve all the fans' status and save in a hash table
    with the following keys stated:
    maxNumFans,
    fan number:name, currentFanSpeedRPM, currentFanSpeedLevel, faultStatesGeneratedFault, faultStatesDetectedFault, faultStatesGeneratedPredictedFail, faultStatesDetectedPredictedFail. faultStatesElementSpecificFaults

    Example of returning fan status:
    my $fan_status = get_fan_status();
    print "Fan 2's name = $fan_status->{2}{name};

    Example of reference returned by the function:
    $VAR1 = {
          'maxNumFans' => '8',
          'fan#' 'name' => 'PCM 1 Fan 1',
          'fan#' 'currentFanSpeedRPM' => '3525',
          'fan#' 'currentFanSpeedLevel' => '1',
          'fan#' 'faultStatesGeneratedFault' => '0',
          'fan#' 'faultStatesDetectedFault' => '0',
          'fan#' 'faultStatesGeneratedPredictedFail' => '0',
          'fan#' 'faultStatesDetectedPredictedFail' => '0',
          'fan#' 'faultStatesElementSpecificFaults' => '0x0',
          'PCM#fans' => 'OK',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_fan_status {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_envctrl";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /RSync ddump/){
        croak "Unexpected ddump_envctrl output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output fan status: @output \n";
    my $fan_num = 0;
    my $fan_full ="";
    my $pcm = 1;

    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /max num fans:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{maxNumFans} = $element[1];
        }
        elsif($line =~ /fan \d+$/) {
            my @tmp = split(/ /,$line);
            $fan_num = $tmp[-1];
        }
        elsif($line =~ /name\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{name} = $element[1];
        }
        elsif($line =~ /currentFanSpeedRPM\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{currentFanSpeedRPM} = $element[1];
        }
        elsif($line =~ /currentFanSpeedLevel\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{currentFanSpeedLevel} = $element[1];
        }
        elsif($line =~ /faultStates.generatedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesGeneratedFault} = $element[1];
        }
        elsif($line =~ /faultStates.detectedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesDetectedFault} = $element[1];
        }
        elsif($line =~ /faultStates.generatedPredictedFail:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesGeneratedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.detectedPredictedFail\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesDetectedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.elementSpecificFaults\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesElementSpecificFaults} = $element[1];
        }
        elsif($line =~ /PCM \d fans/) {
            my @element = split(/\s*:\s*/,$line);
            my $pcm_string = "PCM".$pcm."fans";
            $hash{$pcm_string} = $element[-1];
            $pcm++;
        }
    }
    return \%hash;
}

=head2  get_temp_status

    my $ret = get_temp_status(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "ddump_envctrl" to retrieve all the temperatures' status
    and save in a hash table with the following keys stated:
    maxNumZones,
        zone number: name, currentTemperature, faultStatesGeneratedFault, faultStatesDetectedFault,
    faultStatesGeneratedPredictedFail, faultStatesDetectedPredictedFail, faultStatesElementSpecificFaults, defaultCriticalColdTemperature, defaultCriticalHotTemperature,
    modifiedWarningColdTemperature, modifiedNormalTemperature, modifiedWarningHotTemperature
    PCM#zones

    Example of returning temperature status:
    my $temp_status = get_temp_status();
    print "Temperature 2's name = $temp_status->{2}{name}";

    Example of reference returned by the function:
    $VAR1 = {
          'maxNumZones' => '8',
          'zone#' 'name' => 'SBB Canister 1',
      'zone#' 'currentTemperature' => '37.250';
      'zone#' 'faultStatesGeneratedFault' => '0';
      'zone#' 'faultStatesDetectedFault' => '0';
      'zone#' 'faultStatesGeneratedPredictedFail' => '0';
      'zone#' 'faultStatesDetectedPredictedFail' => '0';
      'zone#' 'faultStatesElementSpecificFaults' => '0x0';
      'zone#' 'defaultCriticalColdTemperature' => '5';
      'zone#' 'defaultCriticalHotTemperature' => '60';
      'zone#' 'modifiedWarningColdTemperature' => '10';
      'zone#' 'modifiedNormalTemperature' => '42';
      'PCM#zones' => 'OK';
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_temp_status {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                                node => {type=>HASHREF, default=>undef, optional=>1},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                                retry => {type=>SCALAR,default=>5,optional=>1 },
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_envctrl";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
#       (undef,undef,$output) = say($c,"$cmd");
                if (defined $opts{node}) {
                    $output = sasadmin_expander_cli(host=>$opts{host},node=>$opts{node},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});
                } else {
                    $output = sasadmin_expander_cli(host=>$opts{host},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});
                }
    }

    if($output !~ /RSync ddump/){
        croak "Unexpected ddump_envctrl output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output zone status: @output \n";
    my $zone_num = 0;
    my $pcm;

    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /max num zones:\s+/) {
            my @element = split(/:\s*/,$line);
            $hash{maxNumZones} = $element[1];
        }
        elsif($line =~ /zone \d+$/) {
            my @tmp = split(/ /,$line);
            $zone_num = $tmp[-1];
        }
        elsif($line =~ /name\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{name} = $element[1];
        }
        elsif($line =~ /currentTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{currentTemperature} = $element[1];
        }
        elsif($line =~ /faultStates.generatedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesGeneratedFault} = $element[1];
        }
        elsif($line =~ /faultStates.detectedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesDetectedFault} = $element[1];
        }
        elsif($line =~ /faultStates.generatedPredictedFail:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesGeneratedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.detectedPredictedFail\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesDetectedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.elementSpecificFaults\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesElementSpecificFaults} = $element[1];
        }
        elsif($line =~ /defaultCriticalColdTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{defaultCriticalColdTemperature} = $element[1];
        }
        elsif($line =~ /defaultCriticalHotTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{defaultCriticalHotTemperature} = $element[1];
        }
        elsif($line =~ /modifiedWarningColdTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{modifiedWarningColdTemperature} = $element[1];
        }
        elsif($line =~ /modifiedNormalTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{modifiedNormalTemperature} = $element[1];
        }
        elsif($line =~ /modifiedWarningHotTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{modifiedWarningHotTemperature} = $element[1];
        }
        elsif($line =~ /PCM \d zones/) {
            my @tmp = split(/\s*:\s*/,$line);
            my $pcm_string = "PCM".$pcm."zones";
            $hash{$pcm_string} = $tmp[-1];
        }
    }
    return \%hash;
}

=head2  ddump_general (??)

    my $ret = ddump_general(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "ddump_general" to retrieve all the General Status
    Service Diagnostic Data and save in a hash table with the following keys stated:
        info_type(??) Canister#: enclWWN, Vendor, vendorId, ProductId, PartNum, FruId, SerialNum,
        Expander#: wwn, SoDResetType, revision, enclSerialNum,

    For Local and Remote faults:
    Local -> "I2C Bus 0"  (I2CBus0)
        totalErrorsOps, errorsInBuffer, bufferSize, intermittentThreshold, failureThreshold, softResets hardResets
    Local -> "I2C Bus 1"  (I2CBus1)
        totalErrorsOps, errorsInBuffer, bufferSize, intermittentThreshold, failureThreshold, softResets hardResets
    Local -> "I2C Bus 2"  (I2CBus2)
        totalErrorsOps, errorsInBuffer, bufferSize, intermittentThreshold, failureThreshold, softResets hardResets
        midplaneVPD0
        PCM1VPD
        PCM2VPD
        PCM3VPD
        PCM4VPD
        SBBCanisterVPD0
        SGPIO0

    Example:
    my $electstatus = ddump_general();
    print "The enclosure Serial Number is: $electstatus->{enclSerialNum}\n";
    print "The Local fault I2C Bus 2 general fault info is: $electstatus->{Local}{"I2C Bus 2"}\n";
    print "The Local fault I2C Bus 2 Failure Threshold value is: $electstatus->{Local}{I2CBus2}{failureThreshold}\n";
??? local vs Local

    Example of reference returned by the function:
    $VAR1 = {
          'Enclosure' 'enclWWN' => '50050CC10200027F',
          'Enclosure' 'Vendor' => 'NETAPP',
          'Enclosure' 'ProductId' => 'DS4243',
          'Enclosure' 'PartNum' => '0094157',
          'Enclosure' 'FruId' => '0x29',
          'Enclosure' 'SerialNum' => 'SHU94157002731E',
          'Midplane' 'ProductId' => 'DS424',
          'Midplane' 'PartNum' => '0091910-07',
          'Midplane' 'SerialNum' => 'BPS91910200027F',
          'local' 'present' => 'Present',
          'local' 'Vendor' => 'xyratex#',
          'local' 'ProductId' => 'DEFAULT',
          'local' 'PartNum' => '0094214-01',
          'local' 'FruId' => '0xBA',
          'local' 'SerialNum' => 'ims94214c14055c',
          'local' 'FirmwareVer' => '00a6',
          'local' 'CanisterCPLD' => '12',
          'local' 'SoDResetType' => 'Cold boot',
          'local' 'NumExpanders' => '1',
          'local' 'wwn' => '50050CC102100EC0',
          'Local' 'I2C_Bus_0' => 'No faults',
          'Local' 'I2CBus0' 'totalErrorsOps' => '4424/739177',
          'Local' 'I2CBus0' 'errorsInBuffer' => '0',
          'Local' 'I2CBus0' 'bufferSize' => '100',
          'Local' 'I2CBus0' 'intermittentThreshold' => '15',
          'Local' 'I2CBus0' 'failureThreshold' => '95',
          'Local' 'I2CBus0' 'softResets' => '0',
          'Local' 'I2CBus0' 'hardResets' => '0',
          'Local' 'midplaneVPD0'  => 'No faults',
          'Local' 'PCM1VPD'  => 'No faults',
          'Local' 'PCM2VPD'  => 'No faults',
          'Local' 'PCM4VPD'  => 'No faults',
          'Local' 'SBBCanisterVPD0'  => 'No faults',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_general {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_general";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }

    if($output !~ /General Status Service Diagnostic Data/){
        croak "Unexpected ddump_general output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output is ddump_general: @output \n";
    my $fault;
    my @layer2;
    my $layer2_1; # to save the string without space in between in layer 2 variable
    my $info_type = "";

    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if(($line =~ /Enclosure Info:/) || ($line =~ /Midplane Info:/)) {
            my @element = split(/ /,$line);
            $info_type = $element[0];
        }
        if($line =~ /Canister \d/) {
            my @tmp = split(/ /,$line);
            my $info_num = $tmp[1];
            chop ($info_num);
            $hash{location} = $tmp[3];
            #$info_type = "Canister".$info_num;
            $info_type = $tmp[3];
            $hash{$info_type}{present} = $tmp[6];
        }
        if($line =~ /Enclosure WWN:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{enclWWN} = $element[-1];
        }
        if($line =~ /Vendor:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{Vendor} = $element[-1];
        }
        if($line =~ /Product Id:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{ProductId} = $element[1];
        }
        if($line =~ /Part Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{PartNum} = $element[1];
        }
        if($line =~ /Fru Id:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{FruId} = $element[1];
        }
        if($line =~ /Serial Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{SerialNum} = $element[1];
        }
        if($line =~ /Firmware Ver:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{FirmwareVer} = $element[1];
        }
        if($line =~ /Canister CPLD:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{CanisterCPLD} = $element[1];
        }
        elsif($line =~ /Revision:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{revision} = $element[1];
        }
        elsif($line =~ /SoD reset type:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{SoDResetType} = $element[1];
        }
        elsif($line =~ /Num Expanders:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{numExpanders} = $element[1];
        }
        elsif($line =~ /Expander \d WWN:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{wwn} = $element[1];
        }
        elsif($line =~ /CID Family ID/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$info_type}{CIDFamilyID} = $element[1];
        }
        elsif($line =~ /CID Family Member ID/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$info_type}{CIDFamilyMemberID} = $element[1];
        }
        elsif(($line =~ /Local faults:/) || ($line =~ /Remote faults:/)) {
            my @element = split(/ /,$line);
            $fault = $element[0];
        }
        elsif($line =~ /remote board not present/) {
            $hash{$fault} = "yes";
        }
        elsif(($line =~ /I2C Bus \d:/) || ($line =~ /SGPIO 0/)) {
            @layer2 = split(/:\s*/,$line);
            my @layer2_tmp = split(/ /, $layer2[1]);
            $layer2_1 = join("", @layer2_tmp);
            my $layer2_1_t = join("_", @layer2_tmp);
            $hash{$fault}{$layer2_1_t} = "$layer2[2]";
        }
        elsif($line =~ /Total Errors\/Ops/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{totalErrorsOps} = $tmp[-1];
        }
        elsif($line =~ /Errors in Buffer/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{errorsInBuffer} = $tmp[-1];
        }
        elsif($line =~ /Buffer Size/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{bufferSize} = $tmp[-1];
        }
        elsif($line =~ /Intermittent Threshold/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{intermittentThreshold} = $tmp[-1];
        }
        elsif($line =~ /Failure Threshold/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{failureThreshold} = $tmp[-1];
        }
        elsif($line =~ /Soft resets/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{softResets} = $tmp[-1];
        }
        elsif($line =~ /Hard resets/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{hardResets} = $tmp[-1];
        }
        elsif($line =~ /Midplane VPD 0:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"midplaneVPD0"} = $tmp[-1];
        }
        elsif($line =~ /PCM 1 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM1VPD"} = $tmp[-1];
        }
        elsif($line =~ /PCM 2 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM2VPD"} = $tmp[-1];
        }
        elsif($line =~ /PCM 3 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM3VPD"} = $tmp[-1];
        }
        elsif($line =~ /PCM 4 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM4VPD"} = $tmp[-1];
        }
        elsif($line =~ /SBB Canister VPD 0:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"SBBCanisterVPD0"} = $tmp[-1];
        }
    }
    return \%hash;
}

=head2  ddump_drvmgr

    my $ret = ddump_drvmgr(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the Drive Manager diagnostic data and save in a hash table
    with the following keys stated for the Drive Manager and for each Device.
        HAMode. driveSpinUp, driveBays, allowedDrives, drivePowerControl, deviceRole,
        instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS
    For each device bay: present, swap, SESInfoBit, spinUpTime, driveType, name, arrayName,
    wwn, faults, faultLED, inject, pending, current.

        Example of returning ddump_drvmgr info:
        $info = ddump_drvmgr();
        print "the fault LED status for drive bay 9 is: $info->{9}{faultLED}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'HAMode'  => 'master',
          'driveSpinUp'  => '0',
          'driveBays'  => '24',
          'driveIndexBase'  => '0',
          'allowedDrives'  => 'SAS OR SATA',
          'drivePowerControl'  => 'supported',
          'deviceRole' '' => 'MASTER,
          'instanceRunState'  => 'RUNNING',
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x63E8',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0x40 (64)',
          'transPoolFreeCount'  => '0x40 (64)',
          'numConACKS'  => '0x48 (72)',
          '0' 'present' => 'yes',
          '0' 'swap' => 'no',
          '0' 'SESInfoBit' => 'not set',
          '0' 'spinUpTime' => '0+00:00:08.087',
          '0' 'driveType' => 'SAS',
          '0' 'wwn' => '500000E0173F6792',
          '0' 'faults' => 'none',
          '0' 'faultLED' => 'OFF',
          '0' 'inject' => 'NONE',
          '0' 'pending' => 'ONLINE',
          '0' 'current' => 'ONLINE',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_drvmgr {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_drvmgr";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }

    if($output !~ /Drive Manager diagnostic dump/){
        croak "Unexpected ddump_drvmgr output: $output";
    }

    my $bay;
    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /HA mode:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{HAMode} = $element[1];
        }
        elsif($line =~ /Drives spinning up:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{driveSpinUp} = $element[1];
        }
        elsif($line =~ /Drive bays:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{driveBays} = $element[1];
        }
        elsif($line =~ /Allowed drives:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{allowedDrives} = $element[1];
        }
        elsif($line =~ /Drive power control:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{drivePowerControl} = $element[1];
        }
        elsif($line =~ /Device role:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{deviceRole} = $element[1];
        }
        elsif($line =~ /Instance run state:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{instanceRunState} = $element[1];
        }
        elsif($line =~ /Sync to Slave status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{syncToSlaveStatus} = $element[1];
        }
        elsif($line =~ /This instance's next UID will be:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nextUID} = $element[1];
        }
        elsif($line =~ /Total expanded transactions:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalExpandTrans} = $element[1];
        }
        elsif($line =~ /Transaction pool capacity:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolCap} = $element[1];
        }
        elsif($line =~ /Transaction pool free count:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolFreeCount} = $element[1];
        }
        elsif($line =~ /Num concurrent ACKS:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{numConACKS} = $element[1];
        }
        elsif($line =~ /Drive Bay \d* status/) {
            my @tmp = split(/ /,$line);
            $bay = $tmp[3];
        }
        elsif($line =~ /present\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{present} = $element[1];
        }
        elsif($line =~ /swap\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{swap} = $element[1];
        }
        elsif($line =~ /SES_info_bit\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{SESInfoBit} = $element[1];
        }
        elsif($line =~ /spin up time\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{spinUpTime} = $element[2].".".$element[3];
        }
        elsif($line =~ /drive_type\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{driveType} = $element[1];
        }
        elsif($line =~ /name\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{name} = $element[1];
        }
        elsif($line =~ /arrayname\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{arrayName} = $element[1];
        }
        elsif($line =~ /WWN\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{wwn} = $element[1];
        }
        elsif($line =~ /faults\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{faults} = $element[1];
        }
        elsif($line =~ /fault LED\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{faultLED} = $element[1];
        }
        elsif($line =~ /inject\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{inject} = $element[1];
        }
        elsif($line =~ /pending\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{pending} = $element[1];
        }
        elsif($line =~ /current\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{current} = $element[1];
        }
    }
    return \%hash;
}

=head2  get_midplanevpd

    my $ret = get_midplane(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the midplane vpd types data and save in a hash table with the following keys.
    For bus 1: mfDateTime_1, enclWWN_1, chasSerialNum_1, mdplnSerialNum_1, chasPartNum_1, mdplnPartNum_1, enclVendID_1, enclProdID_1, enclID_1
    For bus 2: mfDateTime_2, enclWWN_2, chasSerialNum_2, mdplnSerialNum_2, chasPartNum_2, mdplnPartNum_2, enclVendID_2, enclProdID_2, enclID_2

        Example of returning enclosure vendor id:
        $info = get_midplanevpd();
        print "the enclosure vendor IDs from bus 1 and bus 2 are: $info->{enclVendID_1} and $info->{enclVendID_2}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'mfDateTime_1'  => '96 38 62',  # 1:manufacturing datetime
          'mfDateTime_2'  => '96 38 62',
          'enclWWN_1'  => '50050cc10200027f', # 2:enclosure wwn
          'enclWWN_2'  => '50050cc10200027f',
          'chasSerialNum_1'  => 'SHU94157002731E', # 3:chassis serialnumber
          'chasSerialNum_2'  => 'SHU94157002731E',
          'mdplnSerialNum_1'  => 'BPS91910200027F', # 4:midplane serialnumber
          'mdplnSerialNum_2'  => 'BPS91910200027F',
          'chasPartNum_1'  => '0094157-04', # 5:chassis partnumber
          'chasPartNum_2'  => '0094157-04',
          'mdplnPartNum_1'  => '0091910-07', # 6:midplane partnumber
          'mdplnPartNum_2'  => '0091910-07',
          'enclVendID_1'  => 'NETAPP  ', # 7:enclosure vendorID
          'enclVendID_2'  => 'NETAPP  ',
          'enclProdID_1'  => 'DS424   ', # 8:enclosure productID
          'enclProdID_2'  => 'DS424   ',
          'enclID_1'  => '000', # 9:enclosureID ???
          'enclID_2'  => '000',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_midplanevpd {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
#   Make Connection and build command based on input parameters

    for (my $vpd_type=1; $vpd_type<10; $vpd_type++)
    {
        my $command = "set_midplanevpd $vpd_type";

        # Look for manufacture Date Time
        if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
            $opts{conntype} = "#console";
            $c = make_conn($opts{host},$opts{conntype});
            $cmd = "$command";
        }
        elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
            if(!defined($opts{gemconn})){
                croak "GEMTEST needs to be connected before we can use gemtest connmode";
            }
            $c = $opts{gemconn};
            $cmd = "$command";
        }
        elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
            if($opts{filermode} ne "normal"){
                $opts{conntype} = "#console";
            }

            $c = make_conn($opts{host},$opts{conntype});
            if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
                croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
            }
            if(!defined($opts{AorB})){
                $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
            }
            else{
                $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
            }

            if($opts{filermode} ne "normal"){

                $cmd = "sasadmin expander_cli $id \'$command\'";
            }
            else{
                $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
            }

        }
        else {
            croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
        }

        #   Issue Command

        if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
            (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
        }
        else{
                         $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
        }

        if($output !~ /TWI bus/){
            croak "Unexpected set_midplanevpd output: $output";
        }

        my @output = split(/\r*\n/,$output);

        print "Output is : @output \n";
        if ($vpd_type eq "1") {
        # Look for manufacturing datetime
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mfDateTime_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mfDateTime_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "2") {
        # Look for enclosure wwn
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    my @new = split(/ /, $element[1]);
                    my $new_element = join("", @new);
                    $hash{enclWWN_1} = $new_element;
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    my @new = split(/ /, $element[1]);
                    my $new_element = join("", @new);
                    $hash{enclWWN_2} = $new_element;
                }
            }
        }
        if ($vpd_type eq "3") {
        # Look for chassis serial number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasSerialNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasSerialNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "4") {
        # Look for midplane serial number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnSerialNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnSerialNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "5") {
        # Look for chassis part number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasPartNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasPartNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "6") {
        # Look for midplane part number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnPartNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnPartNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "7") {
        # Look for enclosure vendor id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclVendID_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclVendID_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "8") {
        # Look for enclosure product id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclProdID_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclProdID_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "9") {
        # Look for enclosure id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{enclID_1} = $element[-1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{enclID_2} = $element[-1];
                }
            }
        }
        if ($vpd_type eq "13") {
        # Look for Branding Id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{BrandingID_1} = $element[-1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{BrandingID_2} = $element[-1];
                }
            }
        }
    }
    return \%hash;
}


=head2  dumpdrives

    my $ret = dumpdrives(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the drives brief online/empty status and save in a hash
    table with the drive number as key index.

        Example of returning ddump_drvmgr info:
        $info = dumpdrives();
        print "the first drive current status is: $info->{1}\n";

    Example of reference returned by the function:
    $VAR1 = {
          '0'  => 'online',
          '5'  => 'off',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub dumpdrives {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_type => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "dumpdrives";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }

    if($output =~ /Usage:|Unknown/){
        croak "dumpdrives failed with : $output";
    }

    my @output = split(/\r*\n/,$output);
    my $i;
    my $j = 0;

    print "Output is : @output \n";
    if ($opts{shelf_type} =~ /DS42/) {
        foreach my $line (@output) {
            $line =~ s/^\s+//g;
            if($line =~ /\d+/) {
                my @element = split(/\]\[/,$line);
                for($i = 0; $i <4; $i++) {
                    my @new = split(/ /,$element[$i]);
                    if ($j < 10) {
                        $hash{$j} = $new[3];
                    } else {
                        $hash{$j} = $new[2];
                    }
                    $j++;
                }
            }
        }
    } elsif ($opts{shelf_type} =~ /DS22/) {
        my $i = 0;
        foreach my $line (@output) {
            $line =~ s/^\[//g;
            if($line =~ /\d+/) {
                my @e1 = split(/\s+\[\s+/,$line);
                my @element = split(/\]/,$e1[0]);
                my @new = split(/\s+/,$element[0]);
                $hash{$i} = $new[2];
                $i++;
            }
        }
    }
    return \%hash;
}

=head2  ddump_expmgr

    my $ret = ddump_drvmgr(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine retrieves all the Drive Manager diagnostic data and save in a hash table
    with the following keys stated for the Drive Manager and for each Device.
        deviceRole, instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS
    For each phy type (host, drive, sil): BitWind, RstWind, ChgCnt, dwErr, dpErr, syncEr, phyRstEr, crcEr
                    or (??): PCD, IDD, RDD, LOS, PRP, CRC
    For each expander id: swap, updated, ASA, regAccess

        Example of returning ddump_drvmgr info:
        $info = ddump_drvmgr();
        print "the CRC error for drive 8 is: $info->{local}{drive}{crcEr}\n";??

    Example of reference returned by the function:
    $VAR1 = {
          'deviceRole'  => 'MASTER',
          'instanceRunState'  => 'RUNNING'.
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x199 (slot=0 val=409)',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0xA (10)',
          'transPoolFreeCount'  => '0x6 (6)',
          'numConACKS'  => '0x12 (18)',
          'HAmode'  => 'master',
          'local' 'mirror'  => 'disabled', # local phy profiles
          'local' 'host' 'BitWind' => '60',
          'remote' 'host' 'RstWind' => '60',
          'local' 'host' 'CRC' => '1',
          'local' 'host' 'PCD' => '1',
          'local' 'drive' 'ChgCnt' => '120',
          'remote' 'drive' 'dwErr' => '120',
          'local' 'drive' 'PRP' => '1',
          'local' 'drive' 'LOS' => '1',
          'local' 'sil' 'phyRstEr' => '60',
          'remote' 'sil' 'crcEr' => '120',
          'local' 'sil' 'RDD' => '1',
          'local' 'sil' 'IDD' => '1',
          'expander0' 'swap'  => 'No',
          'expander0' 'updated'  => 'No',
          'expander1' 'clkStatus'  => 'OK',
          'expander1' 'regAccess'  => 'OK',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_expmgr {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_expmgr";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }

    if($output !~ /Expander Manager diagnostic dump/){
        croak "Unexpected ddump_expmgr output: $output";
    }

    my $expander_id = 0;
    my $location;
    my $phy_type;
    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Device role:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{deviceRole} = $element[1];
        }
        elsif($line =~ /Instance run state:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{instanceRunState} = $element[1];
        }
        elsif($line =~ /Sync to Slave status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{syncToSlaveStatus} = $element[1];
        }
        elsif($line =~ /This instance's next UID will be:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nextUID} = $element[1];
        }
        elsif($line =~ /Total expanded transactions:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalExpandTrans} = $element[1];
        }
        elsif($line =~ /Transaction pool capacity:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolCap} = $element[1];
        }
        elsif($line =~ /Transaction pool free count:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolFreeCount} = $element[1];
        }
        elsif($line =~ /Num concurrent ACKS:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{numConACKS} = $element[1];
        }
        elsif($line =~ /Local\s+Phy Profiles/) {
            $location = "local";
        }
        elsif($line =~ /Remote\s+Phy Profiles/) {
            $location = "remote";
        }
        elsif($line =~ /Mirroring:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$location}{mirror} = $element[1];
        }
        elsif($line =~ /Host Phy Profile:/) {
            $phy_type = "host";
        }
        elsif($line =~ /Drive Phy Profile:/) {
            $phy_type = "drive";
        }
        elsif($line =~ /Sil Phy Profile:/) {
            $phy_type = "sil";
        }
        elsif($line =~ /\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s*/) {
            my @element = split(/\s+/,$line);
            $hash{$location}{$phy_type}{BitWind} = $element[0];
            $hash{$location}{$phy_type}{RstWind} = $element[1];
            $hash{$location}{$phy_type}{ChgCnt} = $element[2];
            $hash{$location}{$phy_type}{dwErr} = $element[3];
            $hash{$location}{$phy_type}{dpErr} = $element[4];
            $hash{$location}{$phy_type}{syncEr} = $element[5];
            $hash{$location}{$phy_type}{phyRstEr} = $element[6];
            $hash{$location}{$phy_type}{crcEr} = $element[7];
        }
        elsif($line =~ /\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s*/) {
            my @element = split(/\s+/,$line);
            $hash{$location}{$phy_type}{PCD} = $element[1];
            $hash{$location}{$phy_type}{IDD} = $element[2];
            $hash{$location}{$phy_type}{RDD} = $element[3];
            $hash{$location}{$phy_type}{LOS} = $element[4];
            $hash{$location}{$phy_type}{PRP} = $element[5];
            $hash{$location}{$phy_type}{CRC} = $element[6];
        }
        elsif($line =~ /Expander \d Information:/) {
            my @tmp = split(/ /,$line);
            $expander_id = "expander".$tmp[1];
        }
        elsif($line =~ /Swap:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{swap} = $element[-1];
        }
        elsif($line =~ /Updated:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{updated} = $element[-1];
        }
        elsif($line =~ /ASA:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{ASA} = $element[-1];
        }
        elsif($line =~ /CLK status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{clkStatus} = $element[-1];
        }
        elsif($line =~ /Reg Access:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{regAccess} = $element[-1];
            $expander_id++;
        }
    }
    return \%hash;
}

=head2  ddump_acpmgr

    my $ret = ddump_acpmgr(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the ACP Processor status and save in a hash table
    with the following keys stated for the ACP (APC) Manager:
        deviceRole, instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS, totalACP, HAMode, partnerPresent, partnerRemoved,
    For each ACP processor: present, swap, SESInfoBit, heartBeatFail, resetInProgress, macAddr.

        Example of returning ddump_drvmgr info:
        $info = ddump_drvmgr();
        print "the fault LED status for drive bay 9 is: $info->{$bay_number}{faultLED}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'deviceRole'  => 'MASTER',
          'instanceRunState'  => 'RUNNING'.
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x199 (slot=0 val=409)',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0xA (10)',
          'transPoolFreeCount'  => '0x6 (6)',
          'numConACKS'  => '0xC (12)
          'totalACP'  => '2',
          'HAmode'  => 'master',
          'partnerPresent'  => 'Yes',
          'partnerRemoved'  => 'No',
          'acp0' 'present'  => 'Yes',
          'acp0' 'swap'  => 'No',
          'acp0' 'SESInfoBit'  => 'Not Set', 'acp1' 'heartBeatFail'  => 'Yes',
          'acp1' 'macAddr'  => '98:98:98:98:98:98',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_acpmgr {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $id;
    my $command = "ddump_acpmgr";

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }

    if($output !~ /ACP Manager Service Diagnostic Data/){
        croak "Unexpected ddump_acpmgr output: $output";
    }

    my $bay;
    my $acp_proc;
    my @output = split(/\r*\n/,$output);

    #print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Device role:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{deviceRole} = $element[1];
        }
        elsif($line =~ /Instance run state:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{instanceRunState} = $element[1];
        }
        elsif($line =~ /Sync to Slave status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{syncToSlaveStatus} = $element[1];
        }
        elsif($line =~ /This instance's next UID will be:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nextUID} = $element[1];
        }
        elsif($line =~ /Total expanded transactions:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalExpandTrans} = $element[1];
        }
        elsif($line =~ /Transaction pool capacity:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolCap} = $element[1];
        }
        elsif($line =~ /Transaction pool free count:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolFreeCount} = $element[1];
        }
        elsif($line =~ /Num concurrent ACKS:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{numConACKS} = $element[1];
        }
        elsif($line =~ /Total ACPs*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalACP} = $element[1];
        }
        elsif($line =~ /HA mode*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{HAMode} = $element[1];
        }
        elsif($line =~ /Partner Present*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partnerPresent} = $element[1];
        }
        elsif($line =~ /Partner*removed*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partnerRemoved} = $element[1];
        }
        elsif($line =~ /ACP Processor \[\d*\] status/) {
            my @tmp1 = split(/ /,$line);
            my @tmp = split(/\[/,$tmp1[3]);
            chop($tmp[1]);
            $acp_proc = "acp".$tmp[1];
        }
        elsif($line =~ /Present\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{present} = $element[1];
        }
        elsif($line =~ /Swap\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{swap} = $element[1];
        }
        elsif($line =~ /SES_info_bit\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{SESInfoBit} = $element[1];
        }
        elsif($line =~ /Heart Beat Failed\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{heartBeatFail} = $element[1];
        }
        elsif($line =~ /Reset In Progress\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{resetInProgress} = $element[1];
        }
        elsif($line =~ /MAC Address\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{macAddr} = $element[1];
        }
    }
    return \%hash;
}

=head2  qsfpinfo

    my $ret = qsfpinfo(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the qsfp data and save in a hash table
    with the following keys stated:
    slot# - present or not_present
          id, len, powerClass, vedorSpecific, CLEI, vendorName, partNum, serialNum, dateCode, revision, vendorOUI, deviceTech

        Example of returning qsfp info:
        $info = qsfpinfo();
        print "the Serial Number for cable0 is: $info->{0}{SerialNum}\n";

    Example of reference returned by the function:
    $VAR1 = {
          '1'  => 'not_present',
          '0' 'id' => '0x0c'.
          '0' 'len' => '2'.
          '0' 'powerClass' => '0x00'.
          '0' 'vendorSpecific' => '0x00'.
          '1' 'CLEI' => 'TRUE'.
          '1' 'vendorName' => 'Amphenol'.
          '1' 'partNum' => '112-00177'.
          '1' 'serialNum' => 'APL08311770387'.
          '1' 'dateCode' => '080801'.
          '0' 'revision' => '10'.
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub qsfpinfo {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "qsfp_info";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }

    if($output !~ /QSFP Slot/){
        croak "Unexpected qsfp_info output: $output";
    }

    my @output = split(/\r*\n/,$output);
    my $slot = "";

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /QSFP Slot \d/) {
            my @tmp = split(/ /,$line);
            $slot = $tmp[-1];
        }
        elsif($line =~ /No module present/) {
            $hash{$slot} = "not_present";
        }
        elsif($line =~ /Identifier:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{id} = $element[1];
        }
        elsif($line =~ /Length:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{len} = $element[1];
        }
        elsif($line =~ /Power Class:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{powerClass} = $element[1];
        }
        elsif($line =~ /Vendor Specific:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{vendorSpecific} = $element[1];
        }
        elsif($line =~ /CLEI Present:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{CLEI} = $element[1];
        }
        elsif($line =~ /Vendor Name:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{vendorName} = $element[1];
        }
        elsif($line =~ /Part Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{partNum} = $element[1];
        }
        elsif($line =~ /Serial Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{serialNum} = $element[1];
        }
        elsif($line =~ /Date code:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{dateCode} = $element[1];
        }
        elsif($line =~ /Revision:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{revision} = $element[1];
        }
        elsif($line =~ /Vendor OUI:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{vendorOUI} = $element[1];
        }
        elsif($line =~ /Device Technology:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{deviceTech} = $element[1];
        }
    }
    return \%hash;
}

=head2  cablemgr_tp

    my $ret = cablemgr_tp(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the Cable Manager data and save in a hash table
    with the following keys stated:
    For each board# in slot#: status, cableType, swap, updated, cableCommsFail, cableCRCErr, serialNum, vendorName, id, cableLen, devieTech, cableEnd

        Example of returning cablemgr info:??
        $info = cablemgr_tp();
        print "the Cable Length for cable 1 in board 0 is: $info->{0}{1}{cableLen}\n";

    Example of reference returned by the function:
    $hash{$board}{$slot}{$VAR1}
    $VAR1 = {
          'status'  => 'Present',
          'cableType'  => 'QSFP 4x',
          'swap'  => 'No',
          'updated'  => 'No',
          'cableCommsFail'  => 'No',
          'cableCRCErr'  => 'No',
          'serialNum'  => 'APL08311770387',
          'vendorName'  => 'Amphenol',
          'id'  => '0x0c',
          'cableLen'  => '0x02',
          'deviceTech'  => '0x80',
          'cableEnd'  => '0x00',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub cablemgr_tp {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                board => {type=>SCALAR, default=>undef,optional=>1},
                slot => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "cablemgr_tp 1 $opts{board} $opts{slot}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }

    if($output =~ /Usage:/){
        croak "cablemgr_tp failed with : $output";
    }


    my @output = split(/\r*\n/,$output);

    my $board = $opts{board};
    my $slot = $opts{slot};

    #print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Cable \d on board \d/) {
            if($line =~ /NOT PRESENT/) {
                $hash{$board}{$slot}{status} = "NOT PRESENT";
            }
            elsif($line =~ /UNKNOWN/) {
                $hash{$board}{$slot}{status} = "UNKNOWN";
            }
             else {
                $hash{$board}{$slot}{status} = "PRESENT";
            }
        }
        elsif($line =~ /Cable Type:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{cableType} = $element[1];
        }
        elsif($line =~ /Swap:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{swap} = $element[1];
        }
        elsif($line =~ /Updated:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{updated} = $element[1];
        }
        elsif($line =~ /Cable Comms failure:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{cableCommsFail} = $element[1];
        }
        elsif($line =~ /Cable CRC Error:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{cableCRCErr} = $element[1];
        }
        elsif($line =~ /Serial Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{serialNum} = $element[1];
        }
        elsif($line =~ /Vendor Name:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{vendorName} = $element[1];
        }
        elsif($line =~ /Part Number:/) {
                        my @element = split(/:\s*/,$line);
                        $hash{$board}{$slot}{partNum} = $element[1];
                }
        elsif($line =~ /Identifier:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{id} = $tmp[-1];
        }
        elsif($line =~ /Cable Length:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{cableLen} = $tmp[-1];
        }
        elsif($line =~ /Device Tech:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{deviceTech} = $tmp[-1];
        }
        elsif($line =~ /Cable End:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{cableEnd} = $tmp[-1];
        }
    }
    return \%hash;
}

=head2  phydump

    my $ret = phydump(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the phy data and save in a hash table
    with the following keys stated:
    For each phy: type, index, state, speed, protocol, wwn

        Example of returning qsfp info:
        $info = phydump();
        print "PHY 8 WWN is: $info->{$phy}{wwn}\n";

    Example of reference returned by the function:
    $VAR1 = {
          '0x00' 'type'  => 'Port',
          '0x04' 'index'  => '1',
          '0x0a' 'state'  => 'E L S',
          '0x1a' 'speed'  => '3.0Gbps',
          '0x23' 'protocol'  => 'SAS',
          '0x24' 'wwn'  => '5000c500016487ad',
    }
The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub phydump {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                board => {type=>SCALAR, default=>undef,optional=>1},
                slot => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "phydump";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }

    if($output !~ /State:/){
        croak "Unexpected phydump output: $output";
    }

    my @output = split(/\r*\n/,$output);

    my $board = $opts{board};
    my $slot = $opts{slot};

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        my @element = split(/\s*\|/,$line);
        if($element[1] =~ /Port|Drive|SIL/){
            print "element0 : $element[0] \n";
            print "element2 : $element[2] \n";
            print "element3 : $element[3] \n";
            print "phy type is : $element[1]  \n";
            my @phy = split ("0x", $element[0]);
            $hash{$phy[1]}{type} = $element[1];
            $hash{$phy[1]}{index} = $element[2];
            $hash{$phy[1]}{state} = $element[4];
            $hash{$phy[1]}{speed} = $element[5];
            $hash{$phy[1]}{protocol} = $element[6];
            $hash{$phy[1]}{wwn} = $element[7];
        }
    }
    return \%hash;
}

=head2 controller_state

    my $ret = controller_state(
        host     => $host,
        connmode => "shelfcli",
        timeout  => $timeout
    );

This routine checks the master/slave state of the module
It returns 'master'|'slave' on success and croaks on failure.

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : Connection type.
Defaults:
If connmode is 'inband' then "#rsh".
If connmode is 'shelfcli' or 'gemtest' then "#console"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode.
Possible values are:
"shelfcli" for direct console connection to shelf
"gemtest" for using gemtest Windows utility
"inband" for using ontap inband connections
Default: "shelfcli"

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub controller_state {

    my (%opts) = validate(
        @_,
        {
            host => {type=>OBJECT,isa=>'Hostrec'},
            node => {type=>HASHREF, default=>undef, optional=>1},
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
            filermode => {type=>SCALAR,default=>"normal",optional=>1},
            gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
            timeout => {type=>SCALAR, default=>30,optional=>1},
            adapter_id => {type=>SCALAR, default=>undef,optional=>1},
            shelf_id => {type=>SCALAR, default=>undef,optional=>1},
            AorB => {type=>SCALAR, default=>undef,optional=>1},
            params => {type=>SCALAR, default=>" ",optional=>1},
            retry => {type=>SCALAR, default=>5 ,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_ha"; # output includes "Device role: SLAVE|MASTER

    # Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){
        # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    } elsif($opts{connmode} eq "gemtest") {
        # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    } elsif($opts{connmode} eq "inbandcli"){
        # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    } elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
#                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
                if (defined $opts{node}) {
                    $output = sasadmin_expander_cli(host=>$opts{host},node=>$opts{node},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});
                } else {
                    $output = sasadmin_expander_cli(host=>$opts{host},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});
                }
    } else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
        say($c,"\n");
    }

    if ($output =~ /:\s+(\w+)/) {
        if ($1 eq 'Slave') {
            return 'slave';
        } elsif ($1 eq 'Master') {
            return 'master';
        } else {
            croak "Device Mode of $1 not recognized";
        }
    }
    croak "Didn't find Device Mode in $output";
}

=head2 watchdog_tp

    my $ret = controller_state(
        host     => $host,
        connmode => "shelfcli",
        timeout  => $timeout,
    tp_code     => $tp_code
    );

This routine calls firmware internal command watchdog_tp to inject different type of failures: =1 assertion; =2 exception; =3 deadlock; =4 busyloop

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : Connection type.
Defaults:
If connmode is 'inband' then "#rsh".
If connmode is 'shelfcli' or 'gemtest' then "#console"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode.
Possible values are:
"shelfcli" for direct console connection to shelf
"gemtest" for using gemtest Windows utility
"inband" for using ontap inband connections
Default: "shelfcli"

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 4

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub watchdog_tp {

    my (%opts) = validate(
        @_,
        {
            host => {type=>OBJECT,isa=>'Hostrec'},
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
            filermode => {type=>SCALAR,default=>"normal",optional=>1},
            gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
            timeout => {type=>SCALAR, default=>30,optional=>1},
            adapter_id => {type=>SCALAR, default=>undef,optional=>1},
            shelf_id => {type=>SCALAR, default=>undef,optional=>1},
            AorB => {type=>SCALAR, default=>undef,optional=>1},
            tp_code => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "watchdog_tp $opts{tp_code}";

    # Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){
        # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    } elsif($opts{connmode} eq "gemtest") {
        # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    } elsif($opts{connmode} eq "inbandcli"){
        # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    } elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    } else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
        say($c,"\n");
    }
    print "Output is  C: $output \n";
    if($output =~ /Usage:|Invalid/){
        croak "watchdog_tp failed with : $output";
    }
    return $output;
}

=head2  ses_report_page2

    my $ret = ses_report_page2(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the output data from firmware command "ses_report_page 2" and save in a hash table
    with the following keys stated:
    len (ses page length),
    For each element (Device, Psu, Fan, Temperature, EncElectronics, Voltage, Current, SASConncetor, Expander, ACP):
        status_code, status, status1, status2, status3, status4  (4 bytes status for each element)

        Example of returning ses_report_page2 info:
        $info = ses_report_page2();
        print "Fan 2 overall status is: $info->{Fan}{2}{status}\n";
        print "ACP 1 status 1 is: $info->{ACP}{1}{status1}\n";

    Sample output from ses_report_page 2:
0+00:00:26.187 S0 GEM> ses_report_page 2
SES Page 02h (344 Bytes)
-----------------------------
Page Length: 01 54h
Generation Code: 00 00 00 00h

01h:Overall Device                   00h:UNSUPPORTED    :(hex)00 00 00 00
  01h:Device                         01h:OK             :(hex)01 00 00 00
02h:Overall Psu                      00h:UNSUPPORTED    :(hex)00 00 00 00
  02h:Psu                            01h:OK             :(hex)01 00 00 a0
03h:Overall Fan                      00h:UNSUPPORTED    :(hex)00 00 00 00
  03h:Fan                            01h:OK             :(hex)01 01 51 a1
04h:Overall Temperature Sensor       00h:UNSUPPORTED    :(hex)00 00 00 00
  04h:Temperature Sensor             01h:OK             :(hex)01 00 30 00

    Example of reference returned by the function:
    $VAR1 = {
          'Device' '1' 'status_code'  => '01h',
          'Psu' '2' 'status'  => 'OK',
          'Fan' '3' 'status1'  => '01',
          'Temperature' '4' 'status2'  => '00',
          'EncElectronics' '1' 'status3'  => '01',
          'Voltage' '2' 'status4'  => '49',
          'Current' '3' 'status_code'  => '01h',
          'SASConnector' '1' 'status'  => '05',
          'Expander' '1' 'status2'  => '00',
          'ACP' '0' 'status3'  => '00',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ses_report_page2 {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                                node => {type=>HASHREF, default=>undef, optional=>1},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                retry => {type=>SCALAR, default=>5 ,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ses_report_page 2";

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        if (defined $opts{node}) {
            $output = sasadmin_expander_cli(host=>$opts{host},node=>$opts{node},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});
        } else {
            $output = sasadmin_expander_cli(host=>$opts{host},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});
        }
    }

    if($output !~ /SES Page 02h/){
        croak "Unexpected ses_report_page 2 output: $output";
    }

    my $component = "";
    my $index = 0;
    my $comp_before = "Device";
    my @output = split(/\r*\n/,$output);

    #print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        my @element = "";
        if($line =~ /Page Length:/) {
            @element = split(/:s*/,$line);
            @element = split(/ /,$element[-1]);
            chop($element[2]);
            $hash{len} = $element[1].$element[2];
        }
        elsif($line =~ /01h:Device|02h:Psu|03h:Fan|04h:Temperature|07h:Enc Electronics|12h:Voltage|13h:Current|19h:SAS Connector|83h:Expander|85h:ACP Processor/) {
            @element = split(/:/,$line);
            chop($element[0]);
            my $comp_id = $element[0];
            my @tmp1 = split(/\s+/,$element[1]);
            my $component = $tmp1[0];
            if ($component =~ /Enc|SAS/) {
                $component = $tmp1[0].$tmp1[1];
            }

            # reset component index with new component
            if ($comp_before ne $component) {
                if ($component =~ /Psu/) {
                    $index = 1;
                } else {
                    $index = 0;
                }
            }

            chop($tmp1[-1]);
            my $status_code = $tmp1[-1];
            my @tmp2 = split(/\s+/,$element[2]);
            my $status = $tmp2[0];
            my @tmp3 = split(/\s+/,$element[3]);
            my @tmp4 = split(/\)/, $tmp3[0]);
            my $status1 = $tmp4[-1];
            my $status2 = $tmp3[1];
            my $status3 = $tmp3[2];
            my $status4 = $tmp3[3];
            $hash{$component}{$index}{status_code} = $status_code;
            $hash{$component}{$index}{status} = $status;
            $hash{$component}{$index}{status1} = $status1;
            $hash{$component}{$index}{status2} = $status2;
            $hash{$component}{$index}{status3} = $status3;
            $hash{$component}{$index}{status4} = $status4;
            $comp_before = $component;
            $index++;
        }
    }
    return \%hash;
}

=head2  ses_report_page8081

    my $ret = ses_report_page8081(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the SES page 80/81 data and save in a hash table
    with the following keys.
    expPageSeq, phyCount, logPhyId
    For each PHY:
        phyStatus, negPhyLinkRate, partPathTimeout, invDwordCount, runDispCount, lossDwordSyncCount, phyResetCount, crcErrCount, phyChangeCount

        Example of returning ses_report_page8081 info:
        $info = ses_report_page8081();
        print "PHY 2 has Running Disparity Count: $info->{$phy}{RunDispCount}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'expPageSeq'  => '87h',
          'phyCount' ''  => '24h',
          'logPhyId'   => '00h',
          '0' 'phyStatus'  => '00h',
          '1' 'negPhyLinkRate'  => '09h',
          '2' 'partPathTimeout'  => '07h',
          '3' 'invDwordCount'  => '00000000h',
          '4' 'runDispCount'  => '00000000h',
          '5' 'lossDwordSyncCount'  => '00000000h',
          '6' 'phyResetCount'  => '00000000h',
          '7' 'crcErrCount'  => '00000000h',
          '8' 'phyChangeCount'  => '00000002h',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ses_report_page8081 {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                page => {type=>SCALAR},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ses_report_page $opts{page}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
    }

    if($output !~ /SES Page/){
        croak "Unexpected ses_report_page output: $output";
    }

    my $phy;
    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Expander Status Page Sequence/) {
            my @element = split(/:\s*/,$line);
            $hash{expPageSeq} = $element[1];
        }
        elsif($line =~ /PHY Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{phyCount} = $element[1];
        }
        elsif($line =~ /Logical PHY Identifier/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{logPhyId} = $element[1];
        }
        elsif($line =~ /PHY \d+$/) {
            my @tmp = split(/ /,$line);
            $phy = $tmp[1];
        }
        elsif($line =~ /PHY Status/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{phyStatus} = $element[1];
        }
        elsif($line =~ /Negotiated Physical Link Rate/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{negPhyLinkRate} = $element[1];
        }
        elsif($line =~ /Partial Pathway Timeout Value/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{partPathTimeout} = $element[1];
        }
        elsif($line =~ /Invalid Dword Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{invDwordCount} = $element[1];
        }
        elsif($line =~ /Running Disparity Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{runDispCount} = $element[1];
        }
        elsif($line =~ /Loss of Dword Synchronization Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{lossDwordSyncCount} = $element[1];
        }
        elsif($line =~ /PHY Reset Problem Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{phyResetCount} = $element[1];
        }
        elsif($line =~ /CRC Error Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{crcErrCount} = $element[1];
        }
        elsif($line =~ /PHY Change Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{phyChangeCount} = $element[1];
        }
    }
    return \%hash;
}

=head2  ses_report_page8283

        my $ret = ses_report_page8283(host=>$host, conmode=>"shelfcli",timeout => $timeout);
        This routine retrieves all the SES page 82/83 data and save in a hash table
        with the following keys.

        Host,Drive,SIL,GlobalModuleControl,PHYResetTimeout,AWTBias.

dd        For Keys Host,Drive and SIL:
          Internally the Nested Hash keys are

        "Profile,BitErrorRateSampleWindow,ResetProblemSampleWindow,RunningDisparityThreshold,InvalidDwordThreshold,LossofDWORDSyncThreshold,PHYCHANGEThreshold,CRCErrorThreshold,PHYResetProblemThreshold."


        Example of returning ses_report_page8283 info:
                $info = ses_report_page8283();

        $VAR1 = \{
            'SIL' => {
                       'Profile' => '3f00h',
                       'BitErrorRateSampleWindow' => '60 min',
                       'ResetProblemSampleWindow' => '60 min',
                       'RunningDisparityThreshold' => '0078h',
                       'InvalidDwordThreshold' => '0078h',
                       'LossofDWORDSyncThreshold' => '003ch',
                       'PHYCHANGEThreshold' => '78h',
                       'CRCErrorThreshold' => '0078h',
                       'PHYResetProblemThreshold' => '003ch'
                     },
            'PHYResetTimeout' => '00 per 100 mS',
            'GlobalModuleControl' => '00h',
            'Drive' => {
                         'Profile' => '3f00h',
                         'BitErrorRateSampleWindow' => '60 min',
                         'ResetProblemSampleWindow' => '60 min',
                         'RunningDisparityThreshold' => '0078h',
                         'InvalidDwordThreshold' => '0078h',
                         'LossofDWORDSyncThreshold' => '003ch',
                         'PHYCHANGEThreshold' => '78h',
                         'CRCErrorThreshold' => '0078h',
                         'PHYResetProblemThreshold' => '003ch'
                       },
       'Host' => {
                        'Profile' => '3f00h',
                        'BitErrorRateSampleWindow' => '60 min',
                        'ResetProblemSampleWindow' => '60 min',
                        'RunningDisparityThreshold' => '0078h',
                        'InvalidDwordThreshold' => '0078h',
                        'LossofDWORDSyncThreshold' => '003ch',
                        'PHYCHANGEThreshold' => '78h',
                        'CRCErrorThreshold' => '0078h',
                        'PHYResetProblemThreshold' => '003ch'
                      },
            'AWTBias' => '0000 uS'
          };






The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub ses_report_page8283 {

        my (%opts) = validate(@_,{
                                host => {type=>OBJECT,isa=>'Hostrec'},
                                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                                timeout => {type=>SCALAR, default=>300,optional=>1},
                                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                                AorB => {type=>SCALAR, default=>undef,optional=>1},
                                page => {type=>SCALAR},
                }
        );

        my %hash = ();
        my $output;
        my $c;
        my $cmd;
        my $id;
        my $command = "ses_report_page $opts{page}";

#       Make Connection and build command based on input parameters
    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
                $opts{conntype} = "#console";
                $c = make_conn($opts{host},$opts{conntype});
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
                if(!defined($opts{gemconn})){
                        croak "GEMTEST needs to be connected before we can use gemtest connmode";
                }
                $c = $opts{gemconn};
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
                if($opts{filermode} ne "normal"){
                        $opts{conntype} = "#console";
                }

                $c = make_conn($opts{host},$opts{conntype});
                if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
                        croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
                }
                if(!defined($opts{AorB})){
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
                }
                else{
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
                }

                if($opts{filermode} ne "normal"){

                        $cmd = "sasadmin expander_cli $id \'$command\'";
                }
                else{
                        $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
                }

        }

    else {
                croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
        }

#       Issue Command

        if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
                (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
        }
        else{
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
        }


        if($output !~ /SES Page/){
                croak "Unexpected ses_report_page output: $output";
        }

        my $phy;
        my @output = split(/\r*\n/,$output);

        print "Output is : @output \n";
        foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /Global\sModule\sControl/) {
                        my @element = split(/:\s*/,$line);
                        $hash{GlobalModuleControl} = $element[1];
                }
                elsif($line =~ /PHY\sReset\sTimeout/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{PHYResetTimeout} = $element[1];
                }
                elsif($line =~ /AWT\sBias/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{AWTBias} = $element[1];
                }
                elsif($line =~ /PHYs/) {
                        my @tmp = split(/ /,$line);
                        $phy = $tmp[0];
                }
        elsif($line =~ /Profile/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{Profile} = $element[1];
                }
                elsif($line =~ /Bit\sError\sRate\sSample\sWindow/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{BitErrorRateSampleWindow} = $element[1];
                }
                elsif($line =~ /Reset Problem Sample Window/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{ResetProblemSampleWindow} = $element[1];
                }
                elsif($line =~ /PHY CHANGE Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{PHYCHANGEThreshold} = $element[1];
                }
                elsif($line =~ /Invalid Dword Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{InvalidDwordThreshold} = $element[1];
                }
                elsif($line =~ /Running Disparity Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{RunningDisparityThreshold} = $element[1];
                }
                elsif($line =~ /Loss of DWORD Sync Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{LossofDWORDSyncThreshold} = $element[1];
                }
                elsif($line =~ /PHY Reset Problem Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{PHYResetProblemThreshold} = $element[1];
                }
                elsif($line =~ /CRC Error Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{CRCErrorThreshold} = $element[1];
                }
        }
        return \%hash;
}

=head2  ddump_hid_byIndex

    my $ret = ddump_hid_byIndex(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the LID data and save in a hash table
    with the following keys stated for the Drive Manager and for each Device.
        HAMode. driveSpinUp, driveBays, allowedDrives, drivePowerControl, deviceRole,
        instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS
    For each device bay: present, swap, SESInfoBit, spinUpTime, driveType, name, arrayName,
    wwn, faults, faultLED, inject, pending, current.

        Example of returning ddump_hid info:
        $info = ddump_hid();
        print "the fault LED status for drive bay 9 is: $info->{9}{faultLED}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'HAMode'  => 'master',
          'driveSpinUp'  => '0',
          'driveBays'  => '24',
          'driveIndexBase'  => '0',
          'allowedDrives'  => 'SAS OR SATA',
          'drivePowerControl'  => 'supported',
          'deviceRole' '' => 'MASTER,
          'instanceRunState'  => 'RUNNING',
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x63E8',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0x40 (64)',
          'transPoolFreeCount'  => '0x40 (64)',
          'numConACKS'  => '0x48 (72)',
          '0' 'present' => 'yes',
          '0' 'swap' => 'no',
          '0' 'SESInfoBit' => 'not set',
          '0' 'spinUpTime' => '0+00:00:08.087',
          '0' 'driveType' => 'SAS',
          '0' 'wwn' => '500000E0173F6792',
          '0' 'faults' => 'none',
          '0' 'faultLED' => 'OFF',
          '0' 'inject' => 'NONE',
          '0' 'pending' => 'ONLINE',
          '0' 'current' => 'ONLINE',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_hid_byIndex {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_hid";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }

#   Issue Command

    if($output !~ /HID diagnostic dump/){
        croak "Unexpected ddump_hid output: $output";
    }

    my %hash = ();
    my $bay;
    my @output = split(/\r*\n/,$output);
    my $index = "";

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Total number of LEDs/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{total_leds} = $element[1];
        }
        elsif($line =~ /^Index           \s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $index = $element[1];
        }
        elsif($line =~ /Purpose\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Purpose} = $element[1];
        }
        elsif($line =~ /Type\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Type} = $element[1];
        }
        elsif($line =~ /Supported Colours\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Supported_Colours} = $element[1];
        }
        elsif($line =~ /User Info\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{User_Info} = $element[1];
        }
        elsif($line =~ /Current Status\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Status} = $element[1];
        }
    }
    return \%hash;
}

=head2 storage_show_diskled

        my $ret = storage_show_diskled(host=>$host, adapter_name=>$adapter_name,timeout => $timeout);

This routine returns a hash table of information for disk led status.

It returns 1 on success and croaks on failure.
The following named arguments are accepted:

=over

=item host

(Required) : Hostspec to filer.

=item adapter_name

(Optional) : Specific channel for which the information is
 required. Default is all

=item timeout

(Optional) : timeout value. Default is 300 secs.

=back
=cut

sub storage_show_diskled($@) {
        my (%opts) = validate(@_,{
                                host => {type=>OBJECT, isa=>'Hostrec'},
                                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                                timeout => {type=>SCALAR, default=>300,optional=>1},
                                filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                }
        );
        my $adapter = undef ;
    my $output;

        my %hash =();
        if ($opts{filermode} ne "normal") {
        $opts{conntype} = "#console";
    }

        my $conn = make_conn($opts{host},$opts{conntype});
        if ($opts{filermode} ne "normal") {
                my $cmd = "storage show disk_led -a";
                (undef, undef, $output) = say($conn, "$cmd", -prompt, "$MAINT_PROMPT", "-timeout", $opts{timeout});
         } else {
                my $cmd = "priv set -q diag;storage show disk_led -a";
                $output = filer_cmd(host=>$opts{host},cmd=>$cmd,timeout=>$opts{timeout});
        }

        if($output !~ /Disk name/){
                croak $output;
        }
        my $hba;

        my @lines = split(/\r*\n/, $output);

    foreach my $line (@lines) {
        $line =~ s/^\s+//g;
        if (($line !~ /Disk name/) && ($line !~ /----/)) {
            my @disk_t1 = split(/\s+/,$line);
            my @disks = split(/\./,$disk_t1[0]);
            my $disk = $disks[1].".".$disks[2];
            $hash{$disk} = $disk_t1[1];
        }
    }
        return \%hash;
}


sub psmi_dump_regs {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                                psu_no => {type=>SCALAR, default=>1,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
        my $command = "psmi_dump_regs $opts{psu_no}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
                $c = host($opts{host});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \'$command\'";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \'$command\'";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    } elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
            $output = filer_cmd(host=>$c, cmd=>$cmd, timeout=>$opts{timeout});
        } else{
        (undef,undef,$output) = say($c,"$cmd");
    }
    print "Output is  C: $output \n";
    my @output = split(/\r*\n/,$output);
    my $register = 0;
    my $count = 0;

    my $value;
    foreach my $line (@output){
        if ($line =~ /: /) {
            my @line_temp = split(/: /,$line);
            my @output_temp = split(//,$line_temp[1]);
            foreach $value (@output_temp){
                if($value =~ /\d|[a-f]/){ # To make sure that the $value is hex and filter out the rest
                    if ($count == 0) {
                        $hash{$register} = "$value";
                    } elsif ($count < 4) {
                        $hash{$register} = $hash{$register}."$value";
                    }
                    $count = $count + 1;
                    if ($count > 3) {
                        #print "value in dump psmi Register $register is value $value\n";
                        $register = $register + 1;
                        $count = 0;
                    }

                }
            }
        }
    }
    return \%hash ;
}


=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    Anish Vashist (anishv@netapp.com)
    Last revised $Date: 2013/09/30 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;
## ==== YOURS //vickie:DOT_dev/test/lib/Storage/Sapphire.pm
#$Id: //depot/prod/test/Rsierranevada/lib/Storage/Sapphire.pm#6 $
#
## @summary Sapphire Command and Utility library
#
## @status TESTME
##
## @author anishv@netapp.ocm
##
## Added Support for cable status UNKNOWN in cablemgr_tp
#
## @pod here

=head1 NAME

Sapphire

=head1 SYNOPSIS

    use Sapphire qw(function_name);

    my $host = host($SHELF);

    my $hash = function_name(host=>$host,...);

=head1 DESCRIPTION
    This is a library module for the sasadmin command

=head1 ROUTINES

=cut
## @change 03-19-2012 rpradeep: Modified the subroutine gem_cmd to support expander diag testcase
## @change 05-21-2012 rajkuma1: Modified the subroutine phy_report to support ruckus system
## @change 07-19-2012 vickie: Modified subroutine expmgr_tp for parsing inconstent expmgr_tp output in IOM6 and IOM6E

package Storage::Sapphire;

use strict;
use warnings;

use Carp qw(confess croak);
use Params::Validate qw(validate validate_pos SCALAR OBJECT);
use Storage::Sasadmin qw(sasadmin_expander_cli);
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}
use Tharn qw(say script_result host logcomment logresult param);

use base qw(Exporter);
use Utils qw(make_conn);
use Data::Dumper;


our  @EXPORT_OK = qw(
            phy_report
            expctrl_tp
            expmgr_tp
            exp_report
            ses_statuspage
            reboot
            gem_cmd
            phy_disable
            phy_enable
            inc_phy_counters
            reset_error_counters
            get_drive_status
            ver
            get_psu_status
            get_fan_status
            get_temp_status
            ddump_general
            ddump_drvmgr
            get_midplanevpd
            dumpdrives
            ddump_expmgr
            ddump_acpmgr
            qsfpinfo
            cablemgr_tp
            phydump
            controller_state
            watchdog_tp
            ses_report_page2
            ses_report_page8081
            ses_report_page8283
            ddump_hid_byIndex
            storage_show_diskled
            psmi_dump_regs
            );

our $MAINT_PROMPT = "\\*\>";


=head2 phy_report

    my $ret = phy_report(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine returns a hash table of information of phy_report command
It returns hash on success and croaks on failure.

Example of reference returned by the function:
$VAR1 = {

          '33' => {

                    'phyRstEr' => '0',

                    'syncEr' => '0',

                    'PhyType' => 'HOST',

                    'crcEr' => '0',

                    'LogicalPhy' => '0',

                    'State' => '3.0G',

                    'dwEr' => '0',

                    'dpEr' => '0',

                    'ChgCnt' => '3'

                  },

          '32' => {

                    'phyRstEr' => '0',

                    'syncEr' => '0',

                    'PhyType' => 'HOST',

                    'crcEr' => '0',

                    'LogicalPhy' => '0',

                    'State' => '3.0G',

                    'dwEr' => '0',

                    'dpEr' => '0',

                    'ChgCnt' => '3'

                  },

}



The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub phy_report {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "phy_report";

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }

    #print "Output is  C: $output \n";

    if($output !~ /phyRstEr/){
        croak "Unexpected phy_report output: $output";
    }

    my @output = split(/\r*\n/,$output);

    foreach my $line (@output){

        $line =~ s/^\s+//g;
        my @element = split(/\s+/,$line);
            #print "LINE: $line \n";
        if($element[0] =~ /^DRV$|^HOST$|^IM$|^SIL$|^UKWN$/){
            #print "LINE: $line \n";
            #print "Elements are : $#element \n";
            #print "phy type is : $element[0]  \n";
            $hash{$element[2]}{PhyType} = $element[0];
            $hash{$element[2]}{LogicalPhy} = $element[1];
            $hash{$element[2]}{State} = $element[3];
            $hash{$element[2]}{ChgCnt} = $element[4];
            $hash{$element[2]}{dwEr} = $element[5];
            $hash{$element[2]}{dpEr} = $element[6];
            $hash{$element[2]}{syncEr} = $element[7];
            $hash{$element[2]}{phyRstEr} = $element[8];
            $hash{$element[2]}{crcEr} = $element[9];
        }
        elsif($element[0] =~ /^DRV\d+$/){
            #print "LINE: $line \n";
            #print "Elements are : $#element \n";
            my $drv = $element[0];
            $drv =~ s/DRV//g;
            #print "Drive is $drv \n";
            $hash{$element[1]}{PhyType} = 'DRV';
            $hash{$element[1]}{LogicalPhy} = $drv;
            $hash{$element[1]}{State} = $element[2];
            $hash{$element[1]}{ChgCnt} = $element[3];
            $hash{$element[1]}{dwEr} = $element[4];
            $hash{$element[1]}{dpEr} = $element[5];
            $hash{$element[1]}{syncEr} = $element[6];
            $hash{$element[1]}{phyRstEr} = $element[7];
            $hash{$element[1]}{crcEr} = $element[8];
        }
    }
    return \%hash ;
}

#---------------------------------------------------------------------------------------------
#   phy_report function END
#---------------------------------------------------------------------------------------------

=head2 expctrl_tp

    my $ret = expctrl_tp(host =>$host, conmode=>"shelfcli",timeout => $timeout,tp_code=>"1", params=>"$phy");

This routine injects expctrl level error injection using expctrl_tp command
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub expctrl_tp {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>30,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                tp_code => {type=>SCALAR},
                params => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "expctrl_tp $opts{tp_code} $opts{params}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
    }
    else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
        say($c,"\n");
    }
    print "Output is  C: $output \n";
    if($output =~ /Usage:|Failed to/){
        croak "expctrl_tp failed with : $output";
    }
    return $output;
}

=head2 expmgr_tp


This routine executes expmgr_tp command with various parameters passed and returns the output either in a string or as a reerence to an Hash based on the options/parameters passed to expmgr_tp.
See the Usage section below for more details and exapmles.

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expmgr_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=item board

(Optional) : Board number it should be either "0" or "1". Default is " ".

=item seqNum

(Optional) : Sequence number of the respective module. Default is " ". Required only for tp_code is "4".

=item options

(Optional) : Required only when tp_code is 11 to specify the Action parameter. Default is " ".


=item Usage

Example 1:
 my $ret = expmgr_tp(host =>$host, conmode=>"shelfcli",timeout => $timeout,tp_code=>"1", params=>"$phy");

        $ret contains the output string "

Example 2:

        $ret contains a reference to a nested Hash for tp_code 8.

 my $ret = expmgr_tp(host => $host, timeout => $TIMEOUT, connmode=> "inbandcli", filermode => $FILERMODE, AorB => "A", tp_code => "11", adapter_id => "1d",board => "0", shelf_id=> "0",opions => "1",params => "256");

        OUTPUT will be converted into an HASH as shown below for "tp_code 2 and 8".

Output is  C: *** Host Phy Profile ***
 BitWind RstWind ChgCnt dwErr dpErr syncEr phyRstEr crcEr
      60      60    120   120   120     60       60   120
                    CRC   PRP   LOS    RDD      IDD   PCD
                      1     1     1      1        1     1
*** Drive Phy Profile ***
 BitWind RstWind ChgCnt dwErr dpErr syncEr phyRstEr crcEr
      60      60    120   120   120     60       60   120
                    CRC   PRP   LOS    RDD      IDD   PCD
                      1     1     1      1        1     1
*** SIL Phy Profile ***
 BitWind RstWind ChgCnt dwErr dpErr syncEr phyRstEr crcEr
      60      60    120   120   120     60       60   120
                      1     1     1      1        1     1
$VAR1 = {
          'SIL' => {
                     'phyRstEr' => '60',
                     'syncEr' => '60',
                     'dpErr' => '120',
                     'crcEr' => '120',
                     'IDD' => '1',
                     'PCD' => '1',
                     'BitWind' => '60',
                     'RstWind' => '60',
                     'CRC' => '1',
                     'dwErr' => '120',
                     'ChgCnt' => '120',
                     'PRP' => '1',
                     'LOS' => '1',
                     'RDD' => '1'
                   },
       'Drive' => {
                       'phyRstEr' => '60',
                       'syncEr' => '60',
                       'dpErr' => '120',
                       'crcEr' => '120',
                       'IDD' => '1',
                       'PCD' => '1',
                       'BitWind' => '60',
                       'RstWind' => '60',
                       'CRC' => '1',
                       'dwErr' => '120',
                       'ChgCnt' => '120',
                       'PRP' => '1',
                       'LOS' => '1',
                       'RDD' => '1'
                     },
          'Host' => {
                      'phyRstEr' => '60',
                      'syncEr' => '60',
                      'dpErr' => '120',
                      'crcEr' => '120',
                      'IDD' => '1',
                      'PCD' => '1',
                      'BitWind' => '60',
                      'RstWind' => '60',
                      'CRC' => '1',
                      'dwErr' => '120',
                      'ChgCnt' => '120',
                      'PRP' => '1',
                      'LOS' => '1',
                      'RDD' => '1'
                    }
        };

Example :3

Returns the Hash for tp_code 2 output.
Output is  C: Expander 0 Information:
         Swap: No
         Updated: Yes
         ASA: No
         CLK status: OK
         Reg Access: OK

        $VAR1 = {
          'ASA' => 'No',
          'CLK' => 'OK',
          'Information' => '0',
          'Reg' => 'OK',
          'Swap' => 'No',
          'Updated' => 'Yes'
        };




=back

=cut
sub expmgr_tp {

        my (%opts) = validate(@_,{
                                host => {type=>OBJECT,isa=>'Hostrec'},
                                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                                timeout => {type=>SCALAR, default=>30,optional=>1},
                                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                                AorB => {type=>SCALAR, default=>undef,optional=>1},
                                tp_code => {type=>SCALAR},
                                board => {type=>SCALAR, default=>" ",optional=>1},
                                params => {type=>SCALAR, default=>" ",optional=>1},
                                seqNum => {type=>SCALAR, default=>" ",optional=>1},
                                options=>{type=>SCALAR, default=>"",optional=>1},

                }
        );

        my %hash = ();
        my $output;
        my $c;
        my $cmd;
        my $id;
        my $command = "expmgr_tp $opts{tp_code} $opts{board} $opts{params}";
        if ($opts{tp_code} eq "4") { $command = "expmgr_tp $opts{tp_code} $opts{seqNum} $opts{board}";}
        elsif ($opts{tp_code} eq "9" || "11") { $command = "expmgr_tp $opts{tp_code} $opts{board} $opts{params} $opts{options} ";}

#       Make Connection and build command based on input parameters

        if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
                $opts{conntype} = "#console";
                $c = make_conn($opts{host},$opts{conntype});
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
                if(!defined($opts{gemconn})){
                        croak "GEMTEST needs to be connected before we can use gemtest connmode";
                }
                $c = $opts{gemconn};
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
                if($opts{filermode} ne "normal"){
                        $opts{conntype} = "#console";
                }
                $c = make_conn($opts{host},$opts{conntype});
                if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
                        croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
                }
                if(!defined($opts{AorB})){
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
                }
                else{
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";                }

                if($opts{filermode} ne "normal"){
                        $cmd = "sasadmin expander_cli $id \"$command\"";
                }
    else{
                        $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
                }

        }

        else {
                croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
        }

#       Issue Command
        if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
                (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
        }
        elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
                (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
        }
        else{
                ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
                say($c,"\n");
        }
        print "Output is  C: $output \n";
        if($output =~ /Usage:|Failed to/){
                croak "expmgr_tp failed with : $output";
        }


        if ($opts{tp_code} eq "2"){

         my @output = split (/\r*\n/,$output);

        my @Information = grep /Information/,@output;
        my $Information= "@Information";
        my @information= split(/\s/,$Information);
        my $INFO=$information[1];

        my @Swap = grep /Swap/,@output;
        my $Swap = "@Swap";
        my @swap = split(/:\s/,$Swap);
        $Swap = $swap[1];

    my @Updated = grep /Updated/,@output;
        my $Updated = "@Updated";
        my @updated = split(/:\s/,$Updated);
        $Updated = $updated[1];

        my @ASA = grep /ASA/,@output;
        my $ASA = "@ASA";
        my @asa = split(/:\s/,$ASA);
        $ASA = $asa[1];

        my @CLK_status = grep /CLK/,@output;
        my $CLK_status = "@CLK_status";
        my @CLK = split(/:\s/,$CLK_status);
        my $CLK = $CLK[1];

        my @Reg_access = grep /Reg/,@output;
        my $Reg_access = "@Reg_access";
        my @Reg = split(/:\s/,$Reg_access);
        my $Reg = $Reg[1];

        my %OUTPUT = ( 'Information'=>"$INFO",
                'Swap'=>"$Swap",
                'Updated'=>"$Updated",
                'ASA'=>"$ASA",
                'CLK'=>"$CLK",
                'Reg'=>"$Reg" );
        return %OUTPUT;

        }
        elsif ($opts{tp_code} eq "8" )
        {
                my %Result;
                my %hash;
                my @output = split (/\r*\n/,$output);
                for( my $i=0;$i<=$#output;$i++)
                {
                if ($output[$i]=~/Phy\sProfile/)
                {
                  my @element = split(/\s+/,$output[$i+2]);
                  shift @element;
                  my @element1 = split(/\s+/,$output[$i+4]);
                  shift @element1;
          $hash{BitWind}=$element[0];
                  if ($element[1] =~ /(m)/) {
                      $hash{RstWind}=$element[2];
                      $hash{ChgCnt}=$element[4];
                      $hash{dwErr}=$element[5];
                      $hash{dpErr}=$element[6];
                      $hash{syncEr}=$element[7];
                      $hash{phyRstEr}=$element[8];
                      $hash{crcEr}=$element[9];
                  } else {
                      $hash{RstWind}=$element[1];
                      $hash{ChgCnt}=$element[2];
                      $hash{dwErr}=$element[3];
                      $hash{dpErr}=$element[4];
                      $hash{syncEr}=$element[5];
                      $hash{phyRstEr}=$element[6];
                      $hash{crcEr}=$element[7];
                  }
                  $hash{CRC}=$element1[0];
                  $hash{PRP}=$element1[1];
                  $hash{LOS}=$element1[2];
                  $hash{RDD}=$element1[3];
                  $hash{IDD}=$element1[4];
                  $hash{PCD}=$element1[5];
                }
                if ($output[$i]=~/Host/)
                {       my %host_hash=%hash;
                        $Result{Host}=\%host_hash;}
                elsif ($output[$i]=~/Drive/)
                {       my %Drive_hash=%hash;
                        $Result{Drive}=\%Drive_hash;}
                elsif ($output[$i]=~/SIL/)
                {       my %SIL_hash=%hash;
                        $Result{SIL}=\%SIL_hash;}
                elsif ($output[$i]=~/\*\*\*\sPhy Profile\s\*\*\*/)
                {
                        if ($opts{params} eq "0")
                        {$Result {Host}=\%hash;}
                        elsif ($opts{params} eq "1")
                        {$Result {Drive}=\%hash;}
                        elsif ($opts{params} eq "2")
                        {$Result {SIL}=\%hash;}

                }
                }

                return \%Result;

        }



return $output;

}



=head2 exp_report

    my $ret = exp_report(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine returns a hash table of information of exp_report command
It returns hash on success and croaks on failure.

Example of reference returned by the function:

$VAR1 = {

          'TotalPhyCount' => '36',

          'RegAccess' => 'OK',

          'ClkStatus' => 'OK'

        };



The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub exp_report {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "exp_report";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }


    #print "Output is  C: $output \n";

    if($output !~ /Expander Information:/){
        croak "Unexpected phy_report output: $output";
    }

    my @output = split(/\r*\n/,$output);

    foreach my $line (@output){
        $line =~ s/^\s+//g;
        #my @element = split(/\s+/,$line);
        if($line =~ /CLK status:/){
            my @element = split(/CLK status:\s*/,$line);
            $hash{ClkStatus} = $element[1];
        }
        elsif($line =~ /Reg Access:/){
            my @element = split(/Reg Access:\s*/,$line);
            $hash{RegAccess} = $element[1];
        }
        elsif($line =~ /Total phy count:/){
            my @element = split(/Total phy count:\s*/,$line);
            $hash{TotalPhyCount} = $element[1];
        }
    }
    return \%hash ;
}

=head2 ses_statuspage

    my $ret = ses_statuspage(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine returns a hash table of information of exp_report command
It returns hash on success and croaks on failure.

Example of reference returned by the function:

$VAR1 = {

          '33' => '33',

          '32' => '34',

          '90' => '00',

          '63' => '00',

          '21' => '59',

};


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub ses_statuspage {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                page_num => {type=>SCALAR},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ses_statuspage $opts{page_num}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }
    print "Output is  C: $output \n";
    my @output = split(/\s+|\r*\n/,$output);
    my $byte = 0;
    my $value;
    foreach $value (@output){
        if($value =~ /^(\d|[a-f])(\d|[a-f])$/){ # To make sure that the $value is hex and filter out the rest
            print "Byte $byte value is $value \n";
            $hash{$byte} = "$value";
            $byte++;
        }
    }
    return \%hash ;

}



=head2 reboot

    my $ret = reboot(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine reboots a module
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub reboot {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>30,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                params => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "reboot $opts{params}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
    }
    else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},-noprompt,-prompt,"HA Service");
        print "Output is $output \n";
        if($output =~ /failed/){
            croak "Saw failed string during boot:\n $output";
        }
        say($c,"\n",-swallow);
    }
    #print "Output is  C: $output \n";
    if($output =~ /Usage:|Failed to/){
        croak "reboot failed with : $output";
    }
    return $output;
}


=head2 gem_cmd

    my $ret = gem_cmd(host =>$host, conmode=>"shelfcli",timeout => $timeout);

This routine issues a command to the Sapphire shelf fw
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub gem_cmd {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                cmd => {type=>SCALAR},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                pmc_cmd => {type=>SCALAR, default=>undef,optional=>1},
                retry => {type=>SCALAR, default=>6 ,optional=>1},
                diag_cmd => {type=>SCALAR, default=>undef,optional=>1},
                prompt_check => {type=>SCALAR, default=>undef,optional=>1}
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = $opts{cmd};
    my $PMC_PROMPT = "^*0000>";

    if(defined($opts{pmc_cmd}) && ($opts{connmode} ne "shelfcli")){
        croak "PMC commands only work with shelfcli connection mode";
    }


    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "acpcli"){ # Connection when we are directly connected to the acp console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        #if($opts{filermode} ne "normal"){
        #   $opts{conntype} = "#console";
        #}

        #$c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        #if(!defined($opts{AorB})){
        #   $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        #}
        #else{
        #   $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        #}
        $output = sasadmin_expander_cli(host=>$opts{host},conntype=>$opts{conntype},command=>$command,adapter_name=>$opts{adapter_id},shelf_id=>$opts{shelf_id},timeout=>$opts{timeout},filermode=>$opts{filermode},return_data=>1,retry=>$opts{retry});


    }
    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli , acpcli and gemtest";
    }


    if(($opts{connmode} eq "acpcli" ) && ($cmd eq "root")){
        (undef,undef,$output) = say($c, "$cmd", '-noprompt');
    }

    elsif($opts{connmode} eq "shelfcli"){
        if(defined($opts{pmc_cmd}) && ($cmd ne "gem") && ($cmd ne "al2o3")){
                (undef,undef,$output) = say($c,"$cmd",-prompt,"$PMC_PROMPT");
        }elsif(defined($opts{diag_cmd}) && defined ($opts{prompt_check})){
                    (undef,undef,undef,$output) = say($c, "help", -prompt, '\*> |\$');
            }
         elsif((defined($opts{diag_cmd})) && !defined ($opts {prompt_check})){
                    if($cmd eq "C"){
                        (undef,undef,$output) = say($c, "$cmd","-wait",".","-noprompt");
                    }
                    elsif ($cmd eq "exit"){
                        (undef,undef,$output) = say($c, "$cmd","-wait",".",-prompt,"^*Al2O3>");
                    }
                    else{
                                (undef,undef,$output) = say($c, "$cmd",-timeout, 120 ,"-wait",".","-prompt", '^\$');
                    }

        }

        else{
            (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
        }
    }
    else{
    }
    return $output;
}


=head2 phy_disable

    my $ret = phy_disable(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0",disable_reason=>"0");

This routine disables a phy


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy to be disabled.  If we pass "all" then it will disable all the enabled phys. We can also pass list of phys separated by , or space.

=item disable_reason

(Required) : Disable reason to disable a phy

Valid options are 0 to 16

=back

=cut

sub phy_disable {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
                disable_reason => {type=>SCALAR},
        }
    );
    my %disable_reason = (
            "0" => "DIS\/MAN_SES",
            "1" => "DIS\/NOT_PRESENT",
            "2" => "DIS\/BRST_LOS",
            "3" => "DIS\/LOS",
            "4" => "DIS\/BRST_RDD",
            "5" => "DIS\/RDD",
            "6" => "DIS\/BRST_IDD",
            "7" => "DIS\/IDD",
            "8" => "DIS\/BRST_PRP",
            "9" => "DIS\/PRP",
            "10" => "DIS\/BRST_PCD",
            "11" => "DIS\/PCD",
            "12" => "DIS\/BRST_CRC",
            "13" => "DIS\/CRC",
            "14" => "DIS\/MIR",
            "15" => "DIS\/MAN_SMP",
            "16" => "DIS\/CLK_FAULT",
            );


    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my @enabled_phys;
    my $dis_reason = $opts{disable_reason};
    unless(grep(/$dis_reason/,keys(%disable_reason))){
        logresult("script","$dis_reason is not a valid disable reason. Valid Disable reasons are from 0 to 16.");
    }
    my $err = 0;
    my $phy_report;
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        my @phys = keys(%$phy_report);
        foreach my $phy (@phys){
            if($phy_report->{$phy}->{State} =~ /3\.0G|1\.5G|6\.0G/){
                if($dis_reason == 14){
                    if($phy_report->{$phy}->{PhyType} eq "DRV"){
                        push @enabled_phys,$phy;
                    }
                }
                else{
                        push @enabled_phys,$phy;
                }
            }

        }
    }
    else{
        @enabled_phys = split(/,|\s+/,$opts{phy_id});
    }
    foreach my $phy (@enabled_phys){
        logcomment("Disabling phy $phy with reason $dis_reason => $disable_reason{$dis_reason}");
        if($opts{connmode} eq "shelfcli"){
            expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>2,params=>"$phy $dis_reason");
        }
        elsif($opts{connmode} eq "inbandcli"){
            expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>2,params=>"$phy $dis_reason}");
        }
        elsif($opts{connmode} eq "acpcli"){
        }
        elsif($opts{connmode} eq "inbandacp"){
        }
        else{
            logresult("script", "Unsupported connmode");
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@enabled_phys){
        unless($phy_report->{$phy}->{State} =~ /$disable_reason{$dis_reason}/){
                logcomment("FAIL => Unable to disable phy $phy with reason $disable_reason{$dis_reason}");
                $err = 1;
            }
        }
        if($err == 1){
            logresult("fatal", "Unexpected disable state. Please check the logs above");
        }
}

=head2 phy_enable

    my $ret = phy_enable(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0",enable_reason=>"2");

This routine enables a phy
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy to be enabled.  If we pass "all" then it will enable all the disabled phys. We can also pass list of phys separated by , or space.

=item disable_reason

(Optional) : To enable a phy which is disabled because of a particular disable reason. By default it will try to enable all the disabled phys

=item enable_reason

(Required) : To enable a phy with particular enable_reason. Valid options are  0 to 6.


=back

=cut

sub phy_enable {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
                enable_reason => {type=>SCALAR},
                disable_reason => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    # Hash of supported enable reasons
    my %enable_reason = (
        "0" => "MANUAL_SES",
        "1" => "DRIVE_INSERT",
        "2" => "CABLE_INSERT",
        "3" => "PARTNER_EBOD_INSERT",
        "4" => "DRIVE_POWER_ON",
        "5" => "MIRROR",
        "6" => "MANUAL_SMP",
    );

    my %disable_reason = (
            "0" => "DIS\/MAN_SES",
            "1" => "DIS\/NOT_PRESENT",
            "2" => "DIS\/BRST_LOS",
            "3" => "DIS\/LOS",
            "4" => "DIS\/BRST_RDD",
            "5" => "DIS\/RDD",
            "6" => "DIS\/BRST_IDD",
            "7" => "DIS\/IDD",
            "8" => "DIS\/BRST_PRP",
            "9" => "DIS\/PRP",
            "10" => "DIS\/BRST_PCD",
            "11" => "DIS\/PCD",
            "12" => "DIS\/BRST_CRC",
            "13" => "DIS\/CRC",
            "14" => "DIS\/MIR",
            "15" => "DIS\/MAN_SMP",
            "16" => "DIS\/CLK_FAULT",
            );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my @disabled_phys;
    my $en_reason = $opts{enable_reason};
    unless(grep(/$en_reason/,keys(%enable_reason))){
        logresult("script","$en_reason is not a valid enable reason. Valid Disable reasons are from 0 to 6.");
    }
    my $err = 0;
    my $phy_report;
    my $phy_report1;
    my $dis_reason = "DIS";
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        my @phys = keys(%$phy_report);
        foreach my $phy (@phys){
            if(defined($opts{disable_reason})){
                    my $reason = $opts{disable_reason};
                    $dis_reason = $disable_reason{$reason};
                }
            if($phy_report->{$phy}->{State} =~ /$dis_reason/){
                if($phy_report->{$phy}->{PhyType} ne 'SIL'){
                    push @disabled_phys,$phy;
                }
            }
        }
    }
    else{
        @disabled_phys = split(/,|\s+/,$opts{phy_id});
    }
    print "DISABLED phys : @disabled_phys \n";
    sleep 30;
    foreach my $phy (@disabled_phys){

        if($en_reason =~ /1|2|3/){
            if($phy_report->{$phy}->{PhyType} eq 'DRV'){
                $en_reason = 1;
            }
            elsif($phy_report->{$phy}->{PhyType} eq 'HOST'){
                $en_reason = 2;
            }
            elsif($phy_report->{$phy}->{PhyType} eq 'SIL'){
                $en_reason = 3;
            }
        }

        logcomment("Enabling phy $phy with reason $en_reason => $enable_reason{$en_reason}");
        if($opts{connmode} eq "shelfcli"){
            eval { expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>1,params=>"$phy $en_reason");
            };
            if($@){
                logcomment("Phy enable failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "inbandcli"){
            eval {expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>1,params=>"$phy $en_reason}");
            };
            if($@){
                logcomment("Phy enable failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "acpcli"){
        }
        elsif($opts{connmode} eq "inbandacp"){
        }
        else{
            logresult("script", "Unsupported connmode");
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@disabled_phys){
        unless($phy_report1->{$phy}->{State} =~ /3\.0G|1\.5G|6\.0G/){
                logcomment("Unable to enable phy $phy with reason $enable_reason{$en_reason}");
                if($opts{phy_id} ne "all"){
                    $err = 1;
                }
            }
        }
        if($err == 1){
            logresult("fatal", "Unexpected enabled state. Please check the logs above");
        }
}

=head2 inc_phy_counters

    my $ret = inc_phy_counters(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0");

This routine increments phys counters.
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy for which the errors to be incremented.  If we pass "all" then it will increment counters of all the phys. We can also pass a list of phys separated by , or space.

=item count_type

(Optional) : Error to be incremented. Valid options are 0 to 5 and "all".

=item count

(Optional) : How many counters to increment. Default is 10.


=back

=cut

sub inc_phy_counters {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
                count_type => {type=>SCALAR},
                count => {type=>SCALAR},
        }
    );

    # Hash of supported enable reasons
    my %count_type = (
        "0" => "ChgCnt",
        "1" => "dwEr",
        "2" => "dpEr",
        "3" => "syncEr",
        "4" => "phyRstEr",
        "5" => "crcEr",
    );

    my %disable_reason = (
            "0" => "DIS\/MAN_SES",
            "1" => "DIS\/NOT_PRESENT",
            "2" => "DIS\/BRST_LOS",
            "3" => "DIS\/LOS",
            "4" => "DIS\/BRST_RDD",
            "5" => "DIS\/RDD",
            "6" => "DIS\/BRST_IDD",
            "7" => "DIS\/IDD",
            "8" => "DIS\/BRST_PRP",
            "9" => "DIS\/PRP",
            "10" => "DIS\/BRST_PCD",
            "11" => "DIS\/PCD",
            "12" => "DIS\/BRST_CRC",
            "13" => "DIS\/CRC",
            "14" => "DIS\/MIR",
            "15" => "DIS\/MAN_SMP",
            "16" => "DIS\/CLK_FAULT",
            );

    my $c;
    my $cmd;
    my $id;
    my @er_phys;
    my @count_type;
    if($opts{count_type} eq "all"){
        @count_type = keys(%count_type);
    }
    else{
        @count_type = split(/,|\s+/,$opts{count_type});
        foreach my $cnt (@count_type){
            unless(grep(/$cnt/,keys(%count_type))){
            logresult("script","$cnt is not a valid error counter. Valid Disable reasons are from 0 to 5.");
            }
        }
    }
    my $err = 0;
    my $phy_report;
    my $phy_report1;
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        @er_phys = keys(%$phy_report);
    }
    else{
        @er_phys = split(/,|\s+/,$opts{phy_id});
    }
    foreach my $phy (@er_phys){
        foreach my $count_type (@count_type){
            my $cnt_type = $count_type{$count_type};
            logcomment("Incrementing counter $count_type =>  $cnt_type for phy $phy by $opts{count}");
            if($opts{connmode} eq "shelfcli"){
                eval { expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>4,params=>"$phy $count_type $opts{count}");
                };
                if($@){
                    logcomment("Error Increment failed with error: $@");
                }
            }
            elsif($opts{connmode} eq "inbandcli"){
                eval {expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>4,params=>"$phy $count_type $opts{count}}");
                };
                if($@){
                    logcomment("Error increment failed with error: $@");
                }
            }
            elsif($opts{connmode} eq "acpcli"){
            }
            elsif($opts{connmode} eq "inbandacp"){
            }
            else{
                logresult("script", "Unsupported connmode");
            }
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@er_phys){
        foreach my $count_type (@count_type){
            my $cnt_type = $count_type{$count_type};
            my $cnt = $phy_report->{$phy}->{$cnt_type};
            my $cnt1 = $phy_report1->{$phy}->{$cnt_type};
            my $cnt2 = $cnt + $opts{count};
            if($cnt1 != $cnt2){
                logcomment("FAIL: Error $count_type => $cnt_type for phy $phy is expected $cnt2 but actual is $cnt1");
                $err = 1;
            }
        }
    }
        if($err == 1){
            print Dumper($phy_report1);
            print Dumper($phy_report);
            logresult("fatal", "Unexpected enabled state. Please check the logs above");
        }
}

=head2 reset_error_counters

    my $ret = reset_error_counters(host =>$host, conmode=>"shelfcli",timeout => $timeout,phy_id=>"0");

This routine resets error counters of a phy or a set of phys
It returns output on success and croaks on failure.


The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item cmd

(Required) : GEM command to be issued to the controller

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gemtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item phy_id

(Required) : Phy(s) for which the errors to be reset.  If we pass "all" then it will reset counters of all the phys. We can also pass list of phys separated by , or space.


=back

=cut
sub reset_error_counters{

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                phy_id => {type=>SCALAR},
        }
    );

    # Hash of supported enable reasons
    my %count_type = (
        "0" => "ChgCnt",
        "1" => "dwEr",
        "2" => "dpEr",
        "3" => "syncEr",
        "4" => "phyRstEr",
        "5" => "crcEr",
    );

    my $c;
    my $cmd;
    my $id;
    my @er_phys;
    my $err = 0;
    my $phy_report;
    my $phy_report1;
    if($opts{connmode} eq "shelfcli"){
         $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }
    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }

    if($opts{phy_id} eq "all"){
        @er_phys = keys(%$phy_report);
    }
    else{
        @er_phys = split(/,|\s+/,$opts{phy_id});
    }
    foreach my $phy (@er_phys){
        logcomment("Clearing counters for phy $phy");
        if($opts{connmode} eq "shelfcli"){
            eval { expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},tp_code=>3,params=>"$phy");
            };
            if($@){
                logcomment("Error Increment failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "inbandcli"){
            eval {expctrl_tp(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},tp_code=>3,params=>"$phy");
            };
            if($@){
                logcomment("Error increment failed with error: $@");
            }
        }
        elsif($opts{connmode} eq "acpcli"){
        }
        elsif($opts{connmode} eq "inbandacp"){
        }
        else{
            logresult("script", "Unsupported connmode");
        }
    }
    if($opts{connmode} eq "shelfcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode});
    }
    elsif($opts{connmode} eq "inbandcli"){
        $phy_report1 = phy_report(host=>$opts{host},timeout=>$opts{timeout},conntype=>$opts{conntype},connmode=>$opts{connmode},filermode=>$opts{filermode},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id});
    }

    elsif($opts{connmode} eq "acpcli"){
    }
    elsif($opts{connmode} eq "inbandacp"){
    }
    else{
        logresult("script", "Unsupported connmode");
    }
    foreach my $phy (@er_phys){
        foreach my $cnt_type (keys(%count_type)){
            my $cnt = $phy_report1->{$phy}->{$count_type{$cnt_type}};
            print "CNT TYPE: $cnt_type and CNT is $cnt \n";
            if(($count_type{$cnt_type} ne "ChgCnt") && ($cnt != 0)){
                logcomment("FAIL: Error counter $count_type{$cnt_type} for phy $phy is $cnt . Expected zero");
                $err = 1;
            }
        }
    }
        if($err == 1){
            logresult("fatal", "Unexpected enabled state. Please check the logs above");
        }
}

=head2 get_drive_status

    my $ret = get_drive_status(drive=><drive>,host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine retrieves a specific drive's status by issuing shelf command "getdrivestatus" and
    save in a hash table with the following keys stated:
    present, swap, SES_info_bit, SpinUpTime, drive_type, name, arrayname, wwn, faults,  faultLED, inject, pending, current

    Example of returning a specific drive's status:
    my $drivestatus = get_drive_status($drive);
    if ($drivestatus->{present} =~ /yes/) {
        print "the drive $drive is present\n";
    }

    Example of reference returned by the function:
    $VAR1 = {
          'present' => 'yes',
          'swap' => 'no',
          'SES_info_bit' => 'not set',
          'SpinUpTime' => '0+00:00:08.087',
          'drive_type' => 'SAS',
          'wwn' => '500000E0173F6792',
          'faults' => 'none',
          'faultLED' => 'OFF',
          'inject' => 'NONE',
          'pending' => 'ONLINE',
          'current' => 'ONLINE',
          'PwrCycleTime' => '0',
          'PwrInProgress' => 'NO',
          'InProgressAS' => 'NONE',
          'InProgressTS' => 'NONE',
          'ReservedCntr' => '0 (00000000)',
          'SMPPwrOnCnt' => '0',
          'SMPPwrOffCnt' => '0',
          'SMPPwrCycCnt' => '0',
          'ActiveCounter' => '0',
          'ActiveCntrAS' => 'NONE',
          'ActiveCntrTS' => 'NONE',
    };

The following arguments are accepted:

=over

=item drive

(Required) : The drive which status will be returned.

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_drive_status {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                disk_id => {type=>SCALAR, default=>undef},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "getdrivestatus $opts{disk_id}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /Drive Bay/){
        croak "Unexpected getdrivestatus output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output drive opts{disk_id} status: @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /present/) {
            my @element = split(/:/,$line);
            $hash{present} = $element[1];
        }
        elsif($line =~ /swap/) {
            my @element = split(/:/,$line);
            $hash{swap} = $element[1];
        }
        elsif($line =~ /SES_info_bit/) {
            my @element = split(/:/,$line);
            $hash{SES_info_bit} = $element[1];
        }
        elsif($line =~ /spin up time/) {
            my @element = split(/:/,$line);
            $hash{SpinUpTime} = $element[1];
        }
        elsif($line =~ /drive_type/) {
            my @element = split(/:/,$line);
            $hash{drive_type} = $element[1];
        }
        elsif($line =~ /WWN/) {
            my @element = split(/:/,$line);
            $hash{wwn} = $element[1];
        }
        elsif($line =~ /faults/) {
            my @element = split(/:/,$line);
            $hash{faults} = $element[1];
        }
        elsif($line =~ /fault LED/) {
            my @element = split(/:/,$line);
            $hash{faultLED} = $element[1];
        }
        elsif($line =~ /inject/) {
            my @element = split(/:/,$line);
            $hash{inject} = $element[1];
        }
        elsif($line =~ /pending/) {
            my @element = split(/:/,$line);
            $hash{pending} = $element[1];
        }
        elsif($line =~ /current/) {
            my @element = split(/:/,$line);
            $hash{current} = $element[1];
        }
        elsif($line =~ /Pwr Cycle Time/) {
            my @element = split(/:/,$line);
            $hash{PwrCycleTime} = $element[1];
        }
        elsif($line =~ /Pwr InProgress/) {
            my @element = split(/:/,$line);
            $hash{PwrInProgress} = $element[1];
        }
        elsif($line =~ /InProgressAS/) {
            my @element = split(/:/,$line);
            $hash{InProgressAS} = $element[1];
        }
        elsif($line =~ /InProgressTS/) {
            my @element = split(/:/,$line);
            $hash{InProgressTS} = $element[1];
        }
        elsif($line =~ /ReservedCntr/) {
            my @element = split(/:/,$line);
            $hash{ReservedCntr} = $element[1];
        }
        elsif($line =~ /SMP PwrOn cnt/) {
            my @element = split(/:/,$line);
            $hash{SMPPwrOnCnt} = $element[1];
        }
        elsif($line =~ /SMP PwrOff cnt/) {
            my @element = split(/:/,$line);
            $hash{SMPPwrOffCnt} = $element[1];
        }
        elsif($line =~ /SMP PwrCyc cnt/) {
            my @element = split(/:/,$line);
            $hash{SMPPwrCycCnt} = $element[1];
        }
        elsif($line =~ /Active Counter/) {
            my @element = split(/:/,$line);
            $hash{ActiveCounter} = $element[1];
        }
        elsif($line =~ /Active Cntr.AS/) {
            my @element = split(/:/,$line);
            $hash{ActiveCntrAS} = $element[1];
        }
        elsif($line =~ /Active Cntr.TS/) {
            my @element = split(/:/,$line);
            $hash{ActiveCntrTS} = $element[1];
        }
    }
    return \%hash;
}

=head2  ver

    my $ret = ver(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine retrieves all the components firmware versions by issuing shelf command
    "ver" and save in a hash table with following keys stated:
    CanisterFW, CanisterFWDate, CanisterBootloader, CanisterConfig CanisterVPD, CanisterVPDCRC.
    CanisterCPLD. CanisterChip. CanisterSDK, MidplaneVPD, MidplaneVPDCRC. MidplaneCPLD. PCM1FW.
    PCM2FW. PCM3FW. PCM4FW, PCM1VPD, PCM2VPD, PCM3VPD, PCM4VPD, PCM1VPDCRC, PCM2VPDCRC, PCM3VPDCRC, PCM4VPDCRC

    Example of returning version hash table:
    my $tb = ver();
    print "The current Canister Firmware is $tb->{CanisterFW}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'CanisterFW' => '0.A6',
          'CanisterFWDate' => 'Nov 26 2008 04:31:26',
          'CanisterBootloader' => '4.00',
          'CanisterConfigCRC' => 'Not present',
          'CanisterVPDStruc' => '0x06',
          'CanisterVPDCRC' => '0x0E4B5E5A',
          'CanisterCPLD' => '0x12',
          'CanisterChip' => '0x83980000',
          'CanisterSDK' => '2.03.02',
          'MidplaneVPDStruc' => '0x07',
          'MidplaneVPDCRC' => '0x3F8C62FB',
          'MidplaneCPLD' => '0x0E',
          'PCM1FW' => '2.09',
          'PCM2FW' => '2.09',
          'PCM3FW' => '2.09',
          'PCM4FW' => '2.09',
          'PCM1VPDStruc' => '0x02',
          'PCM2VPDStruc' => '0x02',
          'PCM3VPDStruc' => '0x02',
          'PCM4VPDStruc' => '0x02',
          'PCM1VPDCRC' => '0x60CFC396',
          'PCM2VPDCRC' => '0x60CFC396',
          'PCM3VPDCRC' => '0x60CFC396',
          'PCM4VPDCRC' => '0x60CFC396',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ver {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ver";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /Canister firmware/){
        croak "Unexpected ver output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Canister firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterFW} = $element[1];
        }
        elsif($line =~ /Canister firmware date\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterFWDate} = $element[1];
        }
        elsif($line =~ /Canister bootloader\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterBootloader} = $element[1];
        }
        elsif($line =~ /Canister config CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterConfigCRC} = $element[1];
        }
        elsif($line =~ /Canister VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterVPDStruc} = $element[1];
        }
        elsif($line =~ /Canister VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterVPDCRC} = $element[1];
        }
        elsif($line =~ /Canister CPLD\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterCPLD} = $element[1];
        }
        elsif($line =~ /Canister chip\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterChip} = $element[1];
        }
        elsif($line =~ /Canister SDK\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{CanisterSDK} = $element[1];
        }
        elsif($line =~ /Midplane VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{MidplaneVPDStruc} = $element[1];
        }
        elsif($line =~ /Midplane VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{MidplaneVPDCRC} = $element[1];
        }
        elsif($line =~ /Midplane CPLD\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{MidplaneCPLD} = $element[1];
        }
        elsif($line =~ /PCM 1 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM1FW} = $element[1];
        }
        elsif($line =~ /PCM 2 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM2FW} = $element[1];
        }
        elsif($line =~ /PCM 3 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM3FW} = $element[1];
        }
        elsif($line =~ /PCM 4 firmware\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM4FW} = $element[1];
        }
        elsif($line =~ /PCM 1 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM1VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 2 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM2VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 3 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM3VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 4 VPD structure\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM4VPDStruc} = $element[1];
        }
        elsif($line =~ /PCM 1 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM1VPDCRC} = $element[1];
        }
        elsif($line =~ /PCM 2 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM2VPDCRC} = $element[1];
        }
        elsif($line =~ /PCM 3 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM3VPDCRC} = $element[1];
        }
        elsif($line =~ /PCM 4 VPD CRC\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{PCM4VPDCRC} = $element[1];
        }
    }
    return \%hash;
}

=head2  get_psu_status

    my $ret = get_psu_status(psu=><psu>,host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "getpsustatus" to retrieve a specific power supply unit's
    status and save in a hash table with following keys stated:
    present, type, fruid, productName, manufacturer, partNumber, partRevision, serialNumber, firmwareVersion, powerState, fansSelfPowered, nominalPowerRating, nominalPowerRatingOutputBitmask, surgePowerRating, surgePowerRatingOutputBitmask, surgeHoldTime, ACDropoutTolerance, standbyPower, coolingPower, swap, hotswapPrivate, hotswapPublic, fault,
        output1 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent
        output2 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent
        output3 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent
    combined_power, ses_info_bit.

    Example of returning a psu's status:
        my $psustatus = get_psu_status(1);
        print "the power supply 1 is present?  $psustatus->{present}\n";
        print "the power supply 1 output2 min voltage is: $psustatus->{output2}{minVoltage}\n";

    present, type, fruid, productName, manufacturer, partNumber, partRevision, serialNumber, firmwareVersion, powerState, fansSelfPowered, nominalPowerRating, nominalPowerRatingOutputBitmask, surgePowerRating, surgePowerRatingOutputBitmask, surgeHoldTime, ACDropoutTolerance, standbyPower, coolingPower, swap, hotswapPrivate, hotswapPublic, fault,
        output1 - voltage, current, power, minVoltage, minSurgeVoltage, maxVoltage, maxCurrent, minCurrent, maxSurgeCurrent

    Example of reference returned by the function:
    $VAR1 = {
          'present' => 'yes',
          'type' => 'PSMI',
          'fruid' => '0x9C',
          'productName' => 'HB-PCM-01-580AC',
          'manufacturer' => 'POWERONE',
          'partNumber' => '0082562',
          'partRevision' => '07',
          'serialNumber' => 'PMW8256200043ED',
          'firmwareVersion' => '2.09',
          'powerState' => 'on',
          'fansSelfPowered' => 'yes',
          'nominalPowerRating' => '580W',
          'nominalPowerRatingOutputBitmask' => '0x0',
          'surgePowerRating' => '580W',
          'surgePowerRatingOutputBitmask' => '0x0',
          'surgeHoldTime' => '0ms',
          'ACDropoutTolerance' => '0ms',
          'standbyPower' => '10W',
          'coolingPower' => '60W',
          'swap' => 'no',
          'hotswapPrivate' => 'yes',
          'hotswapPublic' => 'yes',
          'fault' => 'no',
          'output#' 'voltage' => '5.11V',
          'output#' 'current' => '3.43A',
          'output#' 'power' => '17.58W',
          'output#' 'minVoltage' => '4.80V',
          'output#' 'minSurgeVoltage' => '4.80V',
          'output#' 'maxVoltage' => '5.19V',
          'output#' 'maxCurrent' => '36.00A',
          'output#' 'minCurrent' => '0.00A',
          'output#' 'maxSurgeCurrent' => '42.00A',
    }

The following arguments are accepted:

=over

=item psu

(Required) : Which power supply status to be retrieved.

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_psu_status {
    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                psu => {type=>SCALAR},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "getpsustatus $opts{psu}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /PCM Bay/){
        croak "Unexpected getpsustatus output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output psu $opts{psu} status: @output \n";
    my $output_num = 0;
    my $output_full ="";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /present:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{present} = $element[1];
        }
        elsif($line =~ /type:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{type} = $element[1];
        }
        elsif($line =~ /FRUid:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{fruid} = $element[1];
        }
        elsif($line =~ /product name:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{productName} = $element[1];
        }
        elsif($line =~ /manufacturer:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{manufacturer} = $element[1];
        }
        elsif($line =~ /part number:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partNumber} = $element[1];
        }
        elsif($line =~ /part revision:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partRevision} = $element[1];
        }
        elsif($line =~ /serial number:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{serialNumber} = $element[1];
        }
        elsif($line =~ /firmware version:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{firmwareVersion} = $element[1];
        }
        elsif($line =~ /power state:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{powerState} = $element[1];
        }
        elsif($line =~ /fans self-powered:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{fansSelfPowered} = $element[1];
        }
        elsif($line =~ /nominal power rating:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nominalPowerRating} = $element[1];
        }
        elsif($line =~ /nominal power rating output bitmask:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nominalPowerRatingOutputBitmask} = $element[1];
        }
        elsif($line =~ /surge power rating:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{surgePowerRating} = $element[1];
        }
        elsif($line =~ /surge power rating output bitmask:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{surgePowerRatingOutputBitmask} = $element[1];
        }
        elsif($line =~ /surge hold time:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{surgeHoldTime} = $element[1];
        }
        elsif($line =~ /AC dropout tolerance:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{ACDropoutTolerance} = $element[1];
        }
        elsif($line =~ /standby power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{standbyPower} = $element[1];
        }
        elsif($line =~ /cooling power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{coolingPower} = $element[1];
        }
        elsif($line =~ /swap:\s+/) {
        }
        elsif($line =~ /hotSwap\(private\):\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{hotswapPrivate} = $element[1];
        }
        elsif($line =~ /hotSwap\(public\):\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{hotswapPublic} = $element[1];
        }
        elsif($line =~ /fault:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{fault} = $element[1];
        }
        elsif($line =~ /output \d/) {
            my @tmp = split(/ /,$line);
            $output_num = $tmp[1];
            $output_full = "output".$output_num;
        }
        elsif($line =~ /^voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{voltage} = $element[1];
        }
        elsif($line =~ /^current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{current} = $element[1];
        }
        elsif($line =~ /power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{power} = $element[1];
        }
        elsif($line =~ /min voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{minVoltage} = $element[1];
        }
        elsif($line =~ /min surge voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{minSurgeVoltage} = $element[1];
        }
        elsif($line =~ /max voltage:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{maxVoltage} = $element[1];
        }
        elsif($line =~ /max current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{maxCurrent} = $element[1];
        }
        elsif($line =~ /min current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{minCurrent} = $element[1];
        }
        elsif($line =~ /max surge current:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$output_full}{manSurgeCurrent} = $element[1];
        }
        elsif($line =~ /combined power:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{combinedPower} = $element[1];
        }
        elsif($line =~ /SES_info_bit:\s+/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{sesInfoBit} = $element[1];
        }
    }
    return \%hash;
}

=head2  get_fan_status

    my $ret = get_fan_status(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "ddump_envctrl" to retrieve all the fans' status and save in a hash table
    with the following keys stated:
    maxNumFans,
    fan number:name, currentFanSpeedRPM, currentFanSpeedLevel, faultStatesGeneratedFault, faultStatesDetectedFault, faultStatesGeneratedPredictedFail, faultStatesDetectedPredictedFail. faultStatesElementSpecificFaults

    Example of returning fan status:
    my $fan_status = get_fan_status();
    print "Fan 2's name = $fan_status->{2}{name};

    Example of reference returned by the function:
    $VAR1 = {
          'maxNumFans' => '8',
          'fan#' 'name' => 'PCM 1 Fan 1',
          'fan#' 'currentFanSpeedRPM' => '3525',
          'fan#' 'currentFanSpeedLevel' => '1',
          'fan#' 'faultStatesGeneratedFault' => '0',
          'fan#' 'faultStatesDetectedFault' => '0',
          'fan#' 'faultStatesGeneratedPredictedFail' => '0',
          'fan#' 'faultStatesDetectedPredictedFail' => '0',
          'fan#' 'faultStatesElementSpecificFaults' => '0x0',
          'PCM#fans' => 'OK',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_fan_status {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_envctrl";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /RSync ddump/){
        croak "Unexpected ddump_envctrl output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output fan status: @output \n";
    my $fan_num = 0;
    my $fan_full ="";
    my $pcm = 1;

    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /max num fans:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{maxNumFans} = $element[1];
        }
        elsif($line =~ /fan \d+$/) {
            my @tmp = split(/ /,$line);
            $fan_num = $tmp[-1];
        }
        elsif($line =~ /name\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{name} = $element[1];
        }
        elsif($line =~ /currentFanSpeedRPM\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{currentFanSpeedRPM} = $element[1];
        }
        elsif($line =~ /currentFanSpeedLevel\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{currentFanSpeedLevel} = $element[1];
        }
        elsif($line =~ /faultStates.generatedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesGeneratedFault} = $element[1];
        }
        elsif($line =~ /faultStates.detectedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesDetectedFault} = $element[1];
        }
        elsif($line =~ /faultStates.generatedPredictedFail:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesGeneratedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.detectedPredictedFail\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesDetectedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.elementSpecificFaults\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$fan_num}{faultStatesElementSpecificFaults} = $element[1];
        }
        elsif($line =~ /PCM \d fans/) {
            my @element = split(/\s*:\s*/,$line);
            my $pcm_string = "PCM".$pcm."fans";
            $hash{$pcm_string} = $element[-1];
            $pcm++;
        }
    }
    return \%hash;
}

=head2  get_temp_status

    my $ret = get_temp_status(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "ddump_envctrl" to retrieve all the temperatures' status
    and save in a hash table with the following keys stated:
    maxNumZones,
        zone number: name, currentTemperature, faultStatesGeneratedFault, faultStatesDetectedFault,
    faultStatesGeneratedPredictedFail, faultStatesDetectedPredictedFail, faultStatesElementSpecificFaults, defaultCriticalColdTemperature, defaultCriticalHotTemperature,
    modifiedWarningColdTemperature, modifiedNormalTemperature, modifiedWarningHotTemperature
    PCM#zones

    Example of returning temperature status:
    my $temp_status = get_temp_status();
    print "Temperature 2's name = $temp_status->{2}{name}";

    Example of reference returned by the function:
    $VAR1 = {
          'maxNumZones' => '8',
          'zone#' 'name' => 'SBB Canister 1',
      'zone#' 'currentTemperature' => '37.250';
      'zone#' 'faultStatesGeneratedFault' => '0';
      'zone#' 'faultStatesDetectedFault' => '0';
      'zone#' 'faultStatesGeneratedPredictedFail' => '0';
      'zone#' 'faultStatesDetectedPredictedFail' => '0';
      'zone#' 'faultStatesElementSpecificFaults' => '0x0';
      'zone#' 'defaultCriticalColdTemperature' => '5';
      'zone#' 'defaultCriticalHotTemperature' => '60';
      'zone#' 'modifiedWarningColdTemperature' => '10';
      'zone#' 'modifiedNormalTemperature' => '42';
      'PCM#zones' => 'OK';
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_temp_status {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_envctrl";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /RSync ddump/){
        croak "Unexpected ddump_envctrl output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output zone status: @output \n";
    my $zone_num = 0;
    my $pcm;

    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /max num zones:\s+/) {
            my @element = split(/:\s*/,$line);
            $hash{maxNumZones} = $element[1];
        }
        elsif($line =~ /zone \d+$/) {
            my @tmp = split(/ /,$line);
            $zone_num = $tmp[-1];
        }
        elsif($line =~ /name\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{name} = $element[1];
        }
        elsif($line =~ /currentTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{currentTemperature} = $element[1];
        }
        elsif($line =~ /faultStates.generatedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesGeneratedFault} = $element[1];
        }
        elsif($line =~ /faultStates.detectedFault\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesDetectedFault} = $element[1];
        }
        elsif($line =~ /faultStates.generatedPredictedFail:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesGeneratedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.detectedPredictedFail\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesDetectedPredictedFail} = $element[1];
        }
        elsif($line =~ /faultStates.elementSpecificFaults\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{faultStatesElementSpecificFaults} = $element[1];
        }
        elsif($line =~ /defaultCriticalColdTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{defaultCriticalColdTemperature} = $element[1];
        }
        elsif($line =~ /defaultCriticalHotTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{defaultCriticalHotTemperature} = $element[1];
        }
        elsif($line =~ /modifiedWarningColdTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{modifiedWarningColdTemperature} = $element[1];
        }
        elsif($line =~ /modifiedNormalTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{modifiedNormalTemperature} = $element[1];
        }
        elsif($line =~ /modifiedWarningHotTemperature\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$zone_num}{modifiedWarningHotTemperature} = $element[1];
        }
        elsif($line =~ /PCM \d zones/) {
            my @tmp = split(/\s*:\s*/,$line);
            my $pcm_string = "PCM".$pcm."zones";
            $hash{$pcm_string} = $tmp[-1];
        }
    }
    return \%hash;
}

=head2  ddump_general (??)

    my $ret = ddump_general(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine runs shelf command "ddump_general" to retrieve all the General Status
    Service Diagnostic Data and save in a hash table with the following keys stated:
        info_type(??) Canister#: enclWWN, Vendor, vendorId, ProductId, PartNum, FruId, SerialNum,
        Expander#: wwn, SoDResetType, revision, enclSerialNum,

    For Local and Remote faults:
    Local -> "I2C Bus 0"  (I2CBus0)
        totalErrorsOps, errorsInBuffer, bufferSize, intermittentThreshold, failureThreshold, softResets hardResets
    Local -> "I2C Bus 1"  (I2CBus1)
        totalErrorsOps, errorsInBuffer, bufferSize, intermittentThreshold, failureThreshold, softResets hardResets
    Local -> "I2C Bus 2"  (I2CBus2)
        totalErrorsOps, errorsInBuffer, bufferSize, intermittentThreshold, failureThreshold, softResets hardResets
        midplaneVPD0
        PCM1VPD
        PCM2VPD
        PCM3VPD
        PCM4VPD
        SBBCanisterVPD0
        SGPIO0

    Example:
    my $electstatus = ddump_general();
    print "The enclosure Serial Number is: $electstatus->{enclSerialNum}\n";
    print "The Local fault I2C Bus 2 general fault info is: $electstatus->{Local}{"I2C Bus 2"}\n";
    print "The Local fault I2C Bus 2 Failure Threshold value is: $electstatus->{Local}{I2CBus2}{failureThreshold}\n";
??? local vs Local

    Example of reference returned by the function:
    $VAR1 = {
          'Enclosure' 'enclWWN' => '50050CC10200027F',
          'Enclosure' 'Vendor' => 'NETAPP',
          'Enclosure' 'ProductId' => 'DS4243',
          'Enclosure' 'PartNum' => '0094157',
          'Enclosure' 'FruId' => '0x29',
          'Enclosure' 'SerialNum' => 'SHU94157002731E',
          'Midplane' 'ProductId' => 'DS424',
          'Midplane' 'PartNum' => '0091910-07',
          'Midplane' 'SerialNum' => 'BPS91910200027F',
          'local' 'present' => 'Present',
          'local' 'Vendor' => 'xyratex#',
          'local' 'ProductId' => 'DEFAULT',
          'local' 'PartNum' => '0094214-01',
          'local' 'FruId' => '0xBA',
          'local' 'SerialNum' => 'ims94214c14055c',
          'local' 'FirmwareVer' => '00a6',
          'local' 'CanisterCPLD' => '12',
          'local' 'SoDResetType' => 'Cold boot',
          'local' 'NumExpanders' => '1',
          'local' 'wwn' => '50050CC102100EC0',
          'Local' 'I2C_Bus_0' => 'No faults',
          'Local' 'I2CBus0' 'totalErrorsOps' => '4424/739177',
          'Local' 'I2CBus0' 'errorsInBuffer' => '0',
          'Local' 'I2CBus0' 'bufferSize' => '100',
          'Local' 'I2CBus0' 'intermittentThreshold' => '15',
          'Local' 'I2CBus0' 'failureThreshold' => '95',
          'Local' 'I2CBus0' 'softResets' => '0',
          'Local' 'I2CBus0' 'hardResets' => '0',
          'Local' 'midplaneVPD0'  => 'No faults',
          'Local' 'PCM1VPD'  => 'No faults',
          'Local' 'PCM2VPD'  => 'No faults',
          'Local' 'PCM4VPD'  => 'No faults',
          'Local' 'SBBCanisterVPD0'  => 'No faults',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_general {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_general";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /General Status Service Diagnostic Data/){
        croak "Unexpected ddump_general output: $output";
    }

    my @output = split(/\r*\n/,$output);

    print "Output is ddump_general: @output \n";
    my $fault;
    my @layer2;
    my $layer2_1; # to save the string without space in between in layer 2 variable
    my $info_type = "";

    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if(($line =~ /Enclosure Info:/) || ($line =~ /Midplane Info:/)) {
            my @element = split(/ /,$line);
            $info_type = $element[0];
        }
        if($line =~ /Canister \d/) {
            my @tmp = split(/ /,$line);
            my $info_num = $tmp[1];
            chop ($info_num);
            $hash{location} = $tmp[3];
            #$info_type = "Canister".$info_num;
            $info_type = $tmp[3];
            $hash{$info_type}{present} = $tmp[6];
        }
        if($line =~ /Enclosure WWN:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{enclWWN} = $element[-1];
        }
        if($line =~ /Vendor:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{Vendor} = $element[-1];
        }
        if($line =~ /Product Id:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{ProductId} = $element[1];
        }
        if($line =~ /Part Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{PartNum} = $element[1];
        }
        if($line =~ /Fru Id:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{FruId} = $element[1];
        }
        if($line =~ /Serial Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{SerialNum} = $element[1];
        }
        if($line =~ /Firmware Ver:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{FirmwareVer} = $element[1];
        }
        if($line =~ /Canister CPLD:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{CanisterCPLD} = $element[1];
        }
        elsif($line =~ /Revision:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{revision} = $element[1];
        }
        elsif($line =~ /SoD reset type:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{SoDResetType} = $element[1];
        }
        elsif($line =~ /Num Expanders:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{numExpanders} = $element[1];
        }
        elsif($line =~ /Expander \d WWN:/) {
            my @element = split(/:\s*/,$line);
            $hash{$info_type}{wwn} = $element[1];
        }
        elsif($line =~ /CID Family ID/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$info_type}{CIDFamilyID} = $element[1];
        }
        elsif($line =~ /CID Family Member ID/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$info_type}{CIDFamilyMemberID} = $element[1];
        }
        elsif(($line =~ /Local faults:/) || ($line =~ /Remote faults:/)) {
            my @element = split(/ /,$line);
            $fault = $element[0];
        }
        elsif($line =~ /remote board not present/) {
            $hash{$fault} = "yes";
        }
        elsif(($line =~ /I2C Bus \d:/) || ($line =~ /SGPIO 0/)) {
            @layer2 = split(/:\s*/,$line);
            my @layer2_tmp = split(/ /, $layer2[1]);
            $layer2_1 = join("", @layer2_tmp);
            my $layer2_1_t = join("_", @layer2_tmp);
            $hash{$fault}{$layer2_1_t} = "$layer2[2]";
        }
        elsif($line =~ /Total Errors\/Ops/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{totalErrorsOps} = $tmp[-1];
        }
        elsif($line =~ /Errors in Buffer/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{errorsInBuffer} = $tmp[-1];
        }
        elsif($line =~ /Buffer Size/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{bufferSize} = $tmp[-1];
        }
        elsif($line =~ /Intermittent Threshold/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{intermittentThreshold} = $tmp[-1];
        }
        elsif($line =~ /Failure Threshold/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{failureThreshold} = $tmp[-1];
        }
        elsif($line =~ /Soft resets/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{softResets} = $tmp[-1];
        }
        elsif($line =~ /Hard resets/) {
            my @tmp = split(/ /,$line);
            $hash{$fault}{$layer2_1}{hardResets} = $tmp[-1];
        }
        elsif($line =~ /Midplane VPD 0:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"midplaneVPD0"} = $tmp[-1];
        }
        elsif($line =~ /PCM 1 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM1VPD"} = $tmp[-1];
        }
        elsif($line =~ /PCM 2 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM2VPD"} = $tmp[-1];
        }
        elsif($line =~ /PCM 3 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM3VPD"} = $tmp[-1];
        }
        elsif($line =~ /PCM 4 VPD:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"PCM4VPD"} = $tmp[-1];
        }
        elsif($line =~ /SBB Canister VPD 0:/) {
            my @tmp = split(/:\s*/,$line);
            $hash{$fault}{"SBBCanisterVPD0"} = $tmp[-1];
        }
    }
    return \%hash;
}

=head2  ddump_drvmgr

    my $ret = ddump_drvmgr(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the Drive Manager diagnostic data and save in a hash table
    with the following keys stated for the Drive Manager and for each Device.
        HAMode. driveSpinUp, driveBays, allowedDrives, drivePowerControl, deviceRole,
        instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS
    For each device bay: present, swap, SESInfoBit, spinUpTime, driveType, name, arrayName,
    wwn, faults, faultLED, inject, pending, current.

        Example of returning ddump_drvmgr info:
        $info = ddump_drvmgr();
        print "the fault LED status for drive bay 9 is: $info->{9}{faultLED}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'HAMode'  => 'master',
          'driveSpinUp'  => '0',
          'driveBays'  => '24',
          'driveIndexBase'  => '0',
          'allowedDrives'  => 'SAS OR SATA',
          'drivePowerControl'  => 'supported',
          'deviceRole' '' => 'MASTER,
          'instanceRunState'  => 'RUNNING',
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x63E8',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0x40 (64)',
          'transPoolFreeCount'  => '0x40 (64)',
          'numConACKS'  => '0x48 (72)',
          '0' 'present' => 'yes',
          '0' 'swap' => 'no',
          '0' 'SESInfoBit' => 'not set',
          '0' 'spinUpTime' => '0+00:00:08.087',
          '0' 'driveType' => 'SAS',
          '0' 'wwn' => '500000E0173F6792',
          '0' 'faults' => 'none',
          '0' 'faultLED' => 'OFF',
          '0' 'inject' => 'NONE',
          '0' 'pending' => 'ONLINE',
          '0' 'current' => 'ONLINE',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_drvmgr {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_drvmgr";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /Drive Manager diagnostic dump/){
        croak "Unexpected ddump_drvmgr output: $output";
    }

    my $bay;
    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /HA mode:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{HAMode} = $element[1];
        }
        elsif($line =~ /Drives spinning up:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{driveSpinUp} = $element[1];
        }
        elsif($line =~ /Drive bays:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{driveBays} = $element[1];
        }
        elsif($line =~ /Allowed drives:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{allowedDrives} = $element[1];
        }
        elsif($line =~ /Drive power control:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{drivePowerControl} = $element[1];
        }
        elsif($line =~ /Device role:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{deviceRole} = $element[1];
        }
        elsif($line =~ /Instance run state:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{instanceRunState} = $element[1];
        }
        elsif($line =~ /Sync to Slave status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{syncToSlaveStatus} = $element[1];
        }
        elsif($line =~ /This instance's next UID will be:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nextUID} = $element[1];
        }
        elsif($line =~ /Total expanded transactions:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalExpandTrans} = $element[1];
        }
        elsif($line =~ /Transaction pool capacity:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolCap} = $element[1];
        }
        elsif($line =~ /Transaction pool free count:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolFreeCount} = $element[1];
        }
        elsif($line =~ /Num concurrent ACKS:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{numConACKS} = $element[1];
        }
        elsif($line =~ /Drive Bay \d* status/) {
            my @tmp = split(/ /,$line);
            $bay = $tmp[3];
        }
        elsif($line =~ /present\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{present} = $element[1];
        }
        elsif($line =~ /swap\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{swap} = $element[1];
        }
        elsif($line =~ /SES_info_bit\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{SESInfoBit} = $element[1];
        }
        elsif($line =~ /spin up time\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{spinUpTime} = $element[2].".".$element[3];
        }
        elsif($line =~ /drive_type\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{driveType} = $element[1];
        }
        elsif($line =~ /name\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{name} = $element[1];
        }
        elsif($line =~ /arrayname\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{arrayName} = $element[1];
        }
        elsif($line =~ /WWN\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{wwn} = $element[1];
        }
        elsif($line =~ /faults\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{faults} = $element[1];
        }
        elsif($line =~ /fault LED\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{faultLED} = $element[1];
        }
        elsif($line =~ /inject\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{inject} = $element[1];
        }
        elsif($line =~ /pending\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{pending} = $element[1];
        }
        elsif($line =~ /current\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$bay}{current} = $element[1];
        }
    }
    return \%hash;
}

=head2  get_midplanevpd

    my $ret = get_midplane(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the midplane vpd types data and save in a hash table with the following keys.
    For bus 1: mfDateTime_1, enclWWN_1, chasSerialNum_1, mdplnSerialNum_1, chasPartNum_1, mdplnPartNum_1, enclVendID_1, enclProdID_1, enclID_1
    For bus 2: mfDateTime_2, enclWWN_2, chasSerialNum_2, mdplnSerialNum_2, chasPartNum_2, mdplnPartNum_2, enclVendID_2, enclProdID_2, enclID_2

        Example of returning enclosure vendor id:
        $info = get_midplanevpd();
        print "the enclosure vendor IDs from bus 1 and bus 2 are: $info->{enclVendID_1} and $info->{enclVendID_2}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'mfDateTime_1'  => '96 38 62',  # 1:manufacturing datetime
          'mfDateTime_2'  => '96 38 62',
          'enclWWN_1'  => '50050cc10200027f', # 2:enclosure wwn
          'enclWWN_2'  => '50050cc10200027f',
          'chasSerialNum_1'  => 'SHU94157002731E', # 3:chassis serialnumber
          'chasSerialNum_2'  => 'SHU94157002731E',
          'mdplnSerialNum_1'  => 'BPS91910200027F', # 4:midplane serialnumber
          'mdplnSerialNum_2'  => 'BPS91910200027F',
          'chasPartNum_1'  => '0094157-04', # 5:chassis partnumber
          'chasPartNum_2'  => '0094157-04',
          'mdplnPartNum_1'  => '0091910-07', # 6:midplane partnumber
          'mdplnPartNum_2'  => '0091910-07',
          'enclVendID_1'  => 'NETAPP  ', # 7:enclosure vendorID
          'enclVendID_2'  => 'NETAPP  ',
          'enclProdID_1'  => 'DS424   ', # 8:enclosure productID
          'enclProdID_2'  => 'DS424   ',
          'enclID_1'  => '000', # 9:enclosureID ???
          'enclID_2'  => '000',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub get_midplanevpd {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
#   Make Connection and build command based on input parameters

    for (my $vpd_type=1; $vpd_type<10; $vpd_type++)
    {
        my $command = "set_midplanevpd $vpd_type";

        # Look for manufacture Date Time
        if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
            $opts{conntype} = "#console";
            $c = make_conn($opts{host},$opts{conntype});
            $cmd = "$command";
        }
        elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
            if(!defined($opts{gemconn})){
                croak "GEMTEST needs to be connected before we can use gemtest connmode";
            }
            $c = $opts{gemconn};
            $cmd = "$command";
        }
        elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
            if($opts{filermode} ne "normal"){
                $opts{conntype} = "#console";
            }

            $c = make_conn($opts{host},$opts{conntype});
            if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
                croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
            }
            if(!defined($opts{AorB})){
                $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
            }
            else{
                $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
            }

            if($opts{filermode} ne "normal"){

                $cmd = "sasadmin expander_cli $id \"$command\"";
            }
            else{
                $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
            }

        }
        else {
            croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
        }

        #   Issue Command

        if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
            (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
        }
        else{
            (undef,undef,$output) = say($c,"$cmd");
        }

        if($output !~ /TWI bus/){
            croak "Unexpected set_midplanevpd output: $output";
        }

        my @output = split(/\r*\n/,$output);

        print "Output is : @output \n";
        if ($vpd_type eq "1") {
        # Look for manufacturing datetime
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mfDateTime_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mfDateTime_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "2") {
        # Look for enclosure wwn
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    my @new = split(/ /, $element[1]);
                    my $new_element = join("", @new);
                    $hash{enclWWN_1} = $new_element;
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    my @new = split(/ /, $element[1]);
                    my $new_element = join("", @new);
                    $hash{enclWWN_2} = $new_element;
                }
            }
        }
        if ($vpd_type eq "3") {
        # Look for chassis serial number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasSerialNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasSerialNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "4") {
        # Look for midplane serial number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnSerialNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnSerialNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "5") {
        # Look for chassis part number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasPartNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{chasPartNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "6") {
        # Look for midplane part number
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnPartNum_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{mdplnPartNum_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "7") {
        # Look for enclosure vendor id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclVendID_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclVendID_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "8") {
        # Look for enclosure product id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclProdID_1} = $element[1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/\[/,$line);
                    chop($element[1]);
                    $hash{enclProdID_2} = $element[1];
                }
            }
        }
        if ($vpd_type eq "9") {
        # Look for enclosure id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{enclID_1} = $element[-1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{enclID_2} = $element[-1];
                }
            }
        }
        if ($vpd_type eq "13") {
        # Look for Branding Id
            foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /bus 1/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{BrandingID_1} = $element[-1];
                }
                elsif($line =~ /bus 2/) {
                    my @element = split(/=/,$line);
                    $element[-1] =~ s/^ \d//;
                    $hash{BrandingID_2} = $element[-1];
                }
            }
        }
    }
    return \%hash;
}


=head2  dumpdrives

    my $ret = dumpdrives(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the drives brief online/empty status and save in a hash
    table with the drive number as key index.

        Example of returning ddump_drvmgr info:
        $info = dumpdrives();
        print "the first drive current status is: $info->{1}\n";

    Example of reference returned by the function:
    $VAR1 = {
          '0'  => 'online',
          '5'  => 'off',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub dumpdrives {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_type => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "dumpdrives";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output =~ /Usage:|Unknown/){
        croak "dumpdrives failed with : $output";
    }

    my @output = split(/\r*\n/,$output);
    my $i;
    my $j = 0;

    print "Output is : @output \n";
    if ($opts{shelf_type} =~ /DS42/) {
        foreach my $line (@output) {
            $line =~ s/^\s+//g;
            if($line =~ /\d+/) {
                my @element = split(/\]\[/,$line);
                for($i = 0; $i <4; $i++) {
                    my @new = split(/ /,$element[$i]);
                    if ($j < 10) {
                        $hash{$j} = $new[3];
                    } else {
                        $hash{$j} = $new[2];
                    }
                    $j++;
                }
            }
        }
    } elsif ($opts{shelf_type} =~ /DS22/) {
        my $i = 0;
        foreach my $line (@output) {
            $line =~ s/^\[//g;
            if($line =~ /\d+/) {
                my @e1 = split(/\s+\[\s+/,$line);
                my @element = split(/\]/,$e1[0]);
                my @new = split(/\s+/,$element[0]);
                $hash{$i} = $new[2];
                $i++;
            }
        }
    }
    return \%hash;
}

=head2  ddump_expmgr

    my $ret = ddump_drvmgr(host=>$host, conmode=>"shelfcli",timeout => $timeout);

    This routine retrieves all the Drive Manager diagnostic data and save in a hash table
    with the following keys stated for the Drive Manager and for each Device.
        deviceRole, instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS
    For each phy type (host, drive, sil): BitWind, RstWind, ChgCnt, dwErr, dpErr, syncEr, phyRstEr, crcEr
                    or (??): PCD, IDD, RDD, LOS, PRP, CRC
    For each expander id: swap, updated, ASA, regAccess

        Example of returning ddump_drvmgr info:
        $info = ddump_drvmgr();
        print "the CRC error for drive 8 is: $info->{local}{drive}{crcEr}\n";??

    Example of reference returned by the function:
    $VAR1 = {
          'deviceRole'  => 'MASTER',
          'instanceRunState'  => 'RUNNING'.
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x199 (slot=0 val=409)',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0xA (10)',
          'transPoolFreeCount'  => '0x6 (6)',
          'numConACKS'  => '0x12 (18)',
          'HAmode'  => 'master',
          'local' 'mirror'  => 'disabled', # local phy profiles
          'local' 'host' 'BitWind' => '60',
          'remote' 'host' 'RstWind' => '60',
          'local' 'host' 'CRC' => '1',
          'local' 'host' 'PCD' => '1',
          'local' 'drive' 'ChgCnt' => '120',
          'remote' 'drive' 'dwErr' => '120',
          'local' 'drive' 'PRP' => '1',
          'local' 'drive' 'LOS' => '1',
          'local' 'sil' 'phyRstEr' => '60',
          'remote' 'sil' 'crcEr' => '120',
          'local' 'sil' 'RDD' => '1',
          'local' 'sil' 'IDD' => '1',
          'expander0' 'swap'  => 'No',
          'expander0' 'updated'  => 'No',
          'expander1' 'clkStatus'  => 'OK',
          'expander1' 'regAccess'  => 'OK',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_expmgr {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_expmgr";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /Expander Manager diagnostic dump/){
        croak "Unexpected ddump_expmgr output: $output";
    }

    my $expander_id = 0;
    my $location;
    my $phy_type;
    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Device role:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{deviceRole} = $element[1];
        }
        elsif($line =~ /Instance run state:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{instanceRunState} = $element[1];
        }
        elsif($line =~ /Sync to Slave status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{syncToSlaveStatus} = $element[1];
        }
        elsif($line =~ /This instance's next UID will be:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nextUID} = $element[1];
        }
        elsif($line =~ /Total expanded transactions:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalExpandTrans} = $element[1];
        }
        elsif($line =~ /Transaction pool capacity:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolCap} = $element[1];
        }
        elsif($line =~ /Transaction pool free count:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolFreeCount} = $element[1];
        }
        elsif($line =~ /Num concurrent ACKS:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{numConACKS} = $element[1];
        }
        elsif($line =~ /Local\s+Phy Profiles/) {
            $location = "local";
        }
        elsif($line =~ /Remote\s+Phy Profiles/) {
            $location = "remote";
        }
        elsif($line =~ /Mirroring:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$location}{mirror} = $element[1];
        }
        elsif($line =~ /Host Phy Profile:/) {
            $phy_type = "host";
        }
        elsif($line =~ /Drive Phy Profile:/) {
            $phy_type = "drive";
        }
        elsif($line =~ /Sil Phy Profile:/) {
            $phy_type = "sil";
        }
        elsif($line =~ /\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s*/) {
            my @element = split(/\s+/,$line);
            $hash{$location}{$phy_type}{BitWind} = $element[0];
            $hash{$location}{$phy_type}{RstWind} = $element[1];
            $hash{$location}{$phy_type}{ChgCnt} = $element[2];
            $hash{$location}{$phy_type}{dwErr} = $element[3];
            $hash{$location}{$phy_type}{dpErr} = $element[4];
            $hash{$location}{$phy_type}{syncEr} = $element[5];
            $hash{$location}{$phy_type}{phyRstEr} = $element[6];
            $hash{$location}{$phy_type}{crcEr} = $element[7];
        }
        elsif($line =~ /\d+\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s*/) {
            my @element = split(/\s+/,$line);
            $hash{$location}{$phy_type}{PCD} = $element[1];
            $hash{$location}{$phy_type}{IDD} = $element[2];
            $hash{$location}{$phy_type}{RDD} = $element[3];
            $hash{$location}{$phy_type}{LOS} = $element[4];
            $hash{$location}{$phy_type}{PRP} = $element[5];
            $hash{$location}{$phy_type}{CRC} = $element[6];
        }
        elsif($line =~ /Expander \d Information:/) {
            my @tmp = split(/ /,$line);
            $expander_id = "expander".$tmp[1];
        }
        elsif($line =~ /Swap:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{swap} = $element[-1];
        }
        elsif($line =~ /Updated:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{updated} = $element[-1];
        }
        elsif($line =~ /ASA:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{ASA} = $element[-1];
        }
        elsif($line =~ /CLK status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{clkStatus} = $element[-1];
        }
        elsif($line =~ /Reg Access:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$expander_id}{regAccess} = $element[-1];
            $expander_id++;
        }
    }
    return \%hash;
}

=head2  ddump_acpmgr

    my $ret = ddump_acpmgr(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the ACP Processor status and save in a hash table
    with the following keys stated for the ACP (APC) Manager:
        deviceRole, instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS, totalACP, HAMode, partnerPresent, partnerRemoved,
    For each ACP processor: present, swap, SESInfoBit, heartBeatFail, resetInProgress, macAddr.

        Example of returning ddump_drvmgr info:
        $info = ddump_drvmgr();
        print "the fault LED status for drive bay 9 is: $info->{$bay_number}{faultLED}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'deviceRole'  => 'MASTER',
          'instanceRunState'  => 'RUNNING'.
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x199 (slot=0 val=409)',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0xA (10)',
          'transPoolFreeCount'  => '0x6 (6)',
          'numConACKS'  => '0xC (12)
          'totalACP'  => '2',
          'HAmode'  => 'master',
          'partnerPresent'  => 'Yes',
          'partnerRemoved'  => 'No',
          'acp0' 'present'  => 'Yes',
          'acp0' 'swap'  => 'No',
          'acp0' 'SESInfoBit'  => 'Not Set', 'acp1' 'heartBeatFail'  => 'Yes',
          'acp1' 'macAddr'  => '98:98:98:98:98:98',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_acpmgr {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $id;
    my $command = "ddump_acpmgr";

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }

    if($output !~ /ACP Manager Service Diagnostic Data/){
        croak "Unexpected ddump_acpmgr output: $output";
    }

    my $bay;
    my $acp_proc;
    my @output = split(/\r*\n/,$output);

    #print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Device role:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{deviceRole} = $element[1];
        }
        elsif($line =~ /Instance run state:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{instanceRunState} = $element[1];
        }
        elsif($line =~ /Sync to Slave status:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{syncToSlaveStatus} = $element[1];
        }
        elsif($line =~ /This instance's next UID will be:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{nextUID} = $element[1];
        }
        elsif($line =~ /Total expanded transactions:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalExpandTrans} = $element[1];
        }
        elsif($line =~ /Transaction pool capacity:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolCap} = $element[1];
        }
        elsif($line =~ /Transaction pool free count:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{transPoolFreeCount} = $element[1];
        }
        elsif($line =~ /Num concurrent ACKS:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{numConACKS} = $element[1];
        }
        elsif($line =~ /Total ACPs*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{totalACP} = $element[1];
        }
        elsif($line =~ /HA mode*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{HAMode} = $element[1];
        }
        elsif($line =~ /Partner Present*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partnerPresent} = $element[1];
        }
        elsif($line =~ /Partner*removed*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{partnerRemoved} = $element[1];
        }
        elsif($line =~ /ACP Processor \[\d*\] status/) {
            my @tmp1 = split(/ /,$line);
            my @tmp = split(/\[/,$tmp1[3]);
            chop($tmp[1]);
            $acp_proc = "acp".$tmp[1];
        }
        elsif($line =~ /Present\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{present} = $element[1];
        }
        elsif($line =~ /Swap\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{swap} = $element[1];
        }
        elsif($line =~ /SES_info_bit\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{SESInfoBit} = $element[1];
        }
        elsif($line =~ /Heart Beat Failed\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{heartBeatFail} = $element[1];
        }
        elsif($line =~ /Reset In Progress\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{resetInProgress} = $element[1];
        }
        elsif($line =~ /MAC Address\s+:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$acp_proc}{macAddr} = $element[1];
        }
    }
    return \%hash;
}

=head2  qsfpinfo

    my $ret = qsfpinfo(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the qsfp data and save in a hash table
    with the following keys stated:
    slot# - present or not_present
          id, len, powerClass, vedorSpecific, CLEI, vendorName, partNum, serialNum, dateCode, revision, vendorOUI, deviceTech

        Example of returning qsfp info:
        $info = qsfpinfo();
        print "the Serial Number for cable0 is: $info->{0}{SerialNum}\n";

    Example of reference returned by the function:
    $VAR1 = {
          '1'  => 'not_present',
          '0' 'id' => '0x0c'.
          '0' 'len' => '2'.
          '0' 'powerClass' => '0x00'.
          '0' 'vendorSpecific' => '0x00'.
          '1' 'CLEI' => 'TRUE'.
          '1' 'vendorName' => 'Amphenol'.
          '1' 'partNum' => '112-00177'.
          '1' 'serialNum' => 'APL08311770387'.
          '1' 'dateCode' => '080801'.
          '0' 'revision' => '10'.
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub qsfpinfo {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "qsfp_info";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /QSFP Slot/){
        croak "Unexpected qsfp_info output: $output";
    }

    my @output = split(/\r*\n/,$output);
    my $slot = "";

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /QSFP Slot \d/) {
            my @tmp = split(/ /,$line);
            $slot = $tmp[-1];
        }
        elsif($line =~ /No module present/) {
            $hash{$slot} = "not_present";
        }
        elsif($line =~ /Identifier:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{id} = $element[1];
        }
        elsif($line =~ /Length:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{len} = $element[1];
        }
        elsif($line =~ /Power Class:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{powerClass} = $element[1];
        }
        elsif($line =~ /Vendor Specific:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{vendorSpecific} = $element[1];
        }
        elsif($line =~ /CLEI Present:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{CLEI} = $element[1];
        }
        elsif($line =~ /Vendor Name:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{vendorName} = $element[1];
        }
        elsif($line =~ /Part Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{partNum} = $element[1];
        }
        elsif($line =~ /Serial Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{serialNum} = $element[1];
        }
        elsif($line =~ /Date code:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{dateCode} = $element[1];
        }
        elsif($line =~ /Revision:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{revision} = $element[1];
        }
        elsif($line =~ /Vendor OUI:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{vendorOUI} = $element[1];
        }
        elsif($line =~ /Device Technology:/) {
            my @element = split(/:\s*/,$line);
            $hash{$slot}{deviceTech} = $element[1];
        }
    }
    return \%hash;
}

=head2  cablemgr_tp

    my $ret = cablemgr_tp(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the Cable Manager data and save in a hash table
    with the following keys stated:
    For each board# in slot#: status, cableType, swap, updated, cableCommsFail, cableCRCErr, serialNum, vendorName, id, cableLen, devieTech, cableEnd

        Example of returning cablemgr info:??
        $info = cablemgr_tp();
        print "the Cable Length for cable 1 in board 0 is: $info->{0}{1}{cableLen}\n";

    Example of reference returned by the function:
    $hash{$board}{$slot}{$VAR1}
    $VAR1 = {
          'status'  => 'Present',
          'cableType'  => 'QSFP 4x',
          'swap'  => 'No',
          'updated'  => 'No',
          'cableCommsFail'  => 'No',
          'cableCRCErr'  => 'No',
          'serialNum'  => 'APL08311770387',
          'vendorName'  => 'Amphenol',
          'id'  => '0x0c',
          'cableLen'  => '0x02',
          'deviceTech'  => '0x80',
          'cableEnd'  => '0x00',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub cablemgr_tp {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                board => {type=>SCALAR, default=>undef,optional=>1},
                slot => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "cablemgr_tp 1 $opts{board} $opts{slot}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output =~ /Usage:/){
        croak "cablemgr_tp failed with : $output";
    }


    my @output = split(/\r*\n/,$output);

    my $board = $opts{board};
    my $slot = $opts{slot};

    #print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Cable \d on board \d/) {
            if($line =~ /NOT PRESENT/) {
                $hash{$board}{$slot}{status} = "NOT PRESENT";
            }
            elsif($line =~ /UNKNOWN/) {
                $hash{$board}{$slot}{status} = "UNKNOWN";
            }
             else {
                $hash{$board}{$slot}{status} = "PRESENT";
            }
        }
        elsif($line =~ /Cable Type:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{cableType} = $element[1];
        }
        elsif($line =~ /Swap:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{swap} = $element[1];
        }
        elsif($line =~ /Updated:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{updated} = $element[1];
        }
        elsif($line =~ /Cable Comms failure:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{cableCommsFail} = $element[1];
        }
        elsif($line =~ /Cable CRC Error:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{cableCRCErr} = $element[1];
        }
        elsif($line =~ /Serial Number:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{serialNum} = $element[1];
        }
        elsif($line =~ /Vendor Name:/) {
            my @element = split(/:\s*/,$line);
            $hash{$board}{$slot}{vendorName} = $element[1];
        }
        elsif($line =~ /Part Number:/) {
                        my @element = split(/:\s*/,$line);
                        $hash{$board}{$slot}{partNum} = $element[1];
                }
        elsif($line =~ /Identifier:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{id} = $tmp[-1];
        }
        elsif($line =~ /Cable Length:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{cableLen} = $tmp[-1];
        }
        elsif($line =~ /Device Tech:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{deviceTech} = $tmp[-1];
        }
        elsif($line =~ /Cable End:/) {
            my @element = split(/:\s*/,$line);
            my @tmp = split(/x/,$element[1]);
            $hash{$board}{$slot}{cableEnd} = $tmp[-1];
        }
    }
    return \%hash;
}

=head2  phydump

    my $ret = phydump(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the phy data and save in a hash table
    with the following keys stated:
    For each phy: type, index, state, speed, protocol, wwn

        Example of returning qsfp info:
        $info = phydump();
        print "PHY 8 WWN is: $info->{$phy}{wwn}\n";

    Example of reference returned by the function:
    $VAR1 = {
          '0x00' 'type'  => 'Port',
          '0x04' 'index'  => '1',
          '0x0a' 'state'  => 'E L S',
          '0x1a' 'speed'  => '3.0Gbps',
          '0x23' 'protocol'  => 'SAS',
          '0x24' 'wwn'  => '5000c500016487ad',
    }
The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub phydump {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                board => {type=>SCALAR, default=>undef,optional=>1},
                slot => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "phydump";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /State:/){
        croak "Unexpected phydump output: $output";
    }

    my @output = split(/\r*\n/,$output);

    my $board = $opts{board};
    my $slot = $opts{slot};

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        my @element = split(/\s*\|/,$line);
        if($element[1] =~ /Port|Drive|SIL/){
            print "element0 : $element[0] \n";
            print "element2 : $element[2] \n";
            print "element3 : $element[3] \n";
            print "phy type is : $element[1]  \n";
            my @phy = split ("0x", $element[0]);
            $hash{$phy[1]}{type} = $element[1];
            $hash{$phy[1]}{index} = $element[2];
            $hash{$phy[1]}{state} = $element[4];
            $hash{$phy[1]}{speed} = $element[5];
            $hash{$phy[1]}{protocol} = $element[6];
            $hash{$phy[1]}{wwn} = $element[7];
        }
    }
    return \%hash;
}

=head2 controller_state

    my $ret = controller_state(
        host     => $host,
        connmode => "shelfcli",
        timeout  => $timeout
    );

This routine checks the master/slave state of the module
It returns 'master'|'slave' on success and croaks on failure.

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : Connection type.
Defaults:
If connmode is 'inband' then "#rsh".
If connmode is 'shelfcli' or 'gemtest' then "#console"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode.
Possible values are:
"shelfcli" for direct console connection to shelf
"gemtest" for using gemtest Windows utility
"inband" for using ontap inband connections
Default: "shelfcli"

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 11

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub controller_state {

    my (%opts) = validate(
        @_,
        {
            host => {type=>OBJECT,isa=>'Hostrec'},
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
            filermode => {type=>SCALAR,default=>"normal",optional=>1},
            gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
            timeout => {type=>SCALAR, default=>30,optional=>1},
            adapter_id => {type=>SCALAR, default=>undef,optional=>1},
            shelf_id => {type=>SCALAR, default=>undef,optional=>1},
            AorB => {type=>SCALAR, default=>undef,optional=>1},
            params => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_ha"; # output includes "Device role: SLAVE|MASTER

    # Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){
        # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    } elsif($opts{connmode} eq "gemtest") {
        # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    } elsif($opts{connmode} eq "inbandcli"){
        # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    } elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
    } else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
        say($c,"\n");
    }

    if ($output =~ /:\s+(\w+)/) {
        if ($1 eq 'Slave') {
            return 'slave';
        } elsif ($1 eq 'Master') {
            return 'master';
        } else {
            croak "Device Mode of $1 not recognized";
        }
    }
    croak "Didn't find Device Mode in $output";
}

=head2 watchdog_tp

    my $ret = controller_state(
        host     => $host,
        connmode => "shelfcli",
        timeout  => $timeout,
    tp_code     => $tp_code
    );

This routine calls firmware internal command watchdog_tp to inject different type of failures: =1 assertion; =2 exception; =3 deadlock; =4 busyloop

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : Connection type.
Defaults:
If connmode is 'inband' then "#rsh".
If connmode is 'shelfcli' or 'gemtest' then "#console"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode.
Possible values are:
"shelfcli" for direct console connection to shelf
"gemtest" for using gemtest Windows utility
"inband" for using ontap inband connections
Default: "shelfcli"

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=item tp_code

(Required) : Error type to be injected. First argument to the expctrl_tp command. Valid values are: 1 to 4

=item params

(Required) : Params to the Error type to be injected.

=back

=cut



sub watchdog_tp {

    my (%opts) = validate(
        @_,
        {
            host => {type=>OBJECT,isa=>'Hostrec'},
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
            filermode => {type=>SCALAR,default=>"normal",optional=>1},
            gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
            timeout => {type=>SCALAR, default=>30,optional=>1},
            adapter_id => {type=>SCALAR, default=>undef,optional=>1},
            shelf_id => {type=>SCALAR, default=>undef,optional=>1},
            AorB => {type=>SCALAR, default=>undef,optional=>1},
            tp_code => {type=>SCALAR, default=>" ",optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "watchdog_tp $opts{tp_code}";

    # Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){
        # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    } elsif($opts{connmode} eq "gemtest") {
        # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    } elsif($opts{connmode} eq "inbandcli"){
        # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }
        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){
            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command
    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    } elsif(($opts{filermode} eq "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-timeout,$opts{timeout});
    } else{
        ($a,$b,$output) = say($c,"$cmd",-timeout,$opts{timeout},"-ok");
        say($c,"\n");
    }
    print "Output is  C: $output \n";
    if($output =~ /Usage:|Invalid/){
        croak "watchdog_tp failed with : $output";
    }
    return $output;
}

=head2  ses_report_page2

    my $ret = ses_report_page2(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the output data from firmware command "ses_report_page 2" and save in a hash table
    with the following keys stated:
    len (ses page length),
    For each element (Device, Psu, Fan, Temperature, EncElectronics, Voltage, Current, SASConncetor, Expander, ACP):
        status_code, status, status1, status2, status3, status4  (4 bytes status for each element)

        Example of returning ses_report_page2 info:
        $info = ses_report_page2();
        print "Fan 2 overall status is: $info->{Fan}{2}{status}\n";
        print "ACP 1 status 1 is: $info->{ACP}{1}{status1}\n";

    Sample output from ses_report_page 2:
0+00:00:26.187 S0 GEM> ses_report_page 2
SES Page 02h (344 Bytes)
-----------------------------
Page Length: 01 54h
Generation Code: 00 00 00 00h

01h:Overall Device                   00h:UNSUPPORTED    :(hex)00 00 00 00
  01h:Device                         01h:OK             :(hex)01 00 00 00
02h:Overall Psu                      00h:UNSUPPORTED    :(hex)00 00 00 00
  02h:Psu                            01h:OK             :(hex)01 00 00 a0
03h:Overall Fan                      00h:UNSUPPORTED    :(hex)00 00 00 00
  03h:Fan                            01h:OK             :(hex)01 01 51 a1
04h:Overall Temperature Sensor       00h:UNSUPPORTED    :(hex)00 00 00 00
  04h:Temperature Sensor             01h:OK             :(hex)01 00 30 00

    Example of reference returned by the function:
    $VAR1 = {
          'Device' '1' 'status_code'  => '01h',
          'Psu' '2' 'status'  => 'OK',
          'Fan' '3' 'status1'  => '01',
          'Temperature' '4' 'status2'  => '00',
          'EncElectronics' '1' 'status3'  => '01',
          'Voltage' '2' 'status4'  => '49',
          'Current' '3' 'status_code'  => '01h',
          'SASConnector' '1' 'status'  => '05',
          'Expander' '1' 'status2'  => '00',
          'ACP' '0' 'status3'  => '00',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ses_report_page2 {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ses_report_page 2";

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }



    if($output !~ /SES Page 02h/){
        croak "Unexpected ses_report_page 2 output: $output";
    }

    my $component = "";
    my $index = 0;
    my $comp_before = "Device";
    my @output = split(/\r*\n/,$output);

    #print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        my @element = "";
        if($line =~ /Page Length:/) {
            @element = split(/:s*/,$line);
            @element = split(/ /,$element[-1]);
            chop($element[2]);
            $hash{len} = $element[1].$element[2];
        }
        elsif($line =~ /01h:Device|02h:Psu|03h:Fan|04h:Temperature|07h:Enc Electronics|12h:Voltage|13h:Current|19h:SAS Connector|83h:Expander|85h:ACP Processor/) {
            @element = split(/:/,$line);
            chop($element[0]);
            my $comp_id = $element[0];
            my @tmp1 = split(/\s+/,$element[1]);
            my $component = $tmp1[0];
            if ($component =~ /Enc|SAS/) {
                $component = $tmp1[0].$tmp1[1];
            }

            # reset component index with new component
            if ($comp_before ne $component) {
                if ($component =~ /Psu/) {
                    $index = 1;
                } else {
                    $index = 0;
                }
            }

            chop($tmp1[-1]);
            my $status_code = $tmp1[-1];
            my @tmp2 = split(/\s+/,$element[2]);
            my $status = $tmp2[0];
            my @tmp3 = split(/\s+/,$element[3]);
            my @tmp4 = split(/\)/, $tmp3[0]);
            my $status1 = $tmp4[-1];
            my $status2 = $tmp3[1];
            my $status3 = $tmp3[2];
            my $status4 = $tmp3[3];
            $hash{$component}{$index}{status_code} = $status_code;
            $hash{$component}{$index}{status} = $status;
            $hash{$component}{$index}{status1} = $status1;
            $hash{$component}{$index}{status2} = $status2;
            $hash{$component}{$index}{status3} = $status3;
            $hash{$component}{$index}{status4} = $status4;
            $comp_before = $component;
            $index++;
        }
    }
    return \%hash;
}

=head2  ses_report_page8081

    my $ret = ses_report_page8081(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the SES page 80/81 data and save in a hash table
    with the following keys.
    expPageSeq, phyCount, logPhyId
    For each PHY:
        phyStatus, negPhyLinkRate, partPathTimeout, invDwordCount, runDispCount, lossDwordSyncCount, phyResetCount, crcErrCount, phyChangeCount

        Example of returning ses_report_page8081 info:
        $info = ses_report_page8081();
        print "PHY 2 has Running Disparity Count: $info->{$phy}{RunDispCount}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'expPageSeq'  => '87h',
          'phyCount' ''  => '24h',
          'logPhyId'   => '00h',
          '0' 'phyStatus'  => '00h',
          '1' 'negPhyLinkRate'  => '09h',
          '2' 'partPathTimeout'  => '07h',
          '3' 'invDwordCount'  => '00000000h',
          '4' 'runDispCount'  => '00000000h',
          '5' 'lossDwordSyncCount'  => '00000000h',
          '6' 'phyResetCount'  => '00000000h',
          '7' 'crcErrCount'  => '00000000h',
          '8' 'phyChangeCount'  => '00000002h',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ses_report_page8081 {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
                page => {type=>SCALAR},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ses_report_page $opts{page}";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }

    if($output !~ /SES Page/){
        croak "Unexpected ses_report_page output: $output";
    }

    my $phy;
    my @output = split(/\r*\n/,$output);

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Expander Status Page Sequence/) {
            my @element = split(/:\s*/,$line);
            $hash{expPageSeq} = $element[1];
        }
        elsif($line =~ /PHY Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{phyCount} = $element[1];
        }
        elsif($line =~ /Logical PHY Identifier/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{logPhyId} = $element[1];
        }
        elsif($line =~ /PHY \d+$/) {
            my @tmp = split(/ /,$line);
            $phy = $tmp[1];
        }
        elsif($line =~ /PHY Status/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{phyStatus} = $element[1];
        }
        elsif($line =~ /Negotiated Physical Link Rate/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{negPhyLinkRate} = $element[1];
        }
        elsif($line =~ /Partial Pathway Timeout Value/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{partPathTimeout} = $element[1];
        }
        elsif($line =~ /Invalid Dword Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{invDwordCount} = $element[1];
        }
        elsif($line =~ /Running Disparity Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{runDispCount} = $element[1];
        }
        elsif($line =~ /Loss of Dword Synchronization Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{lossDwordSyncCount} = $element[1];
        }
        elsif($line =~ /PHY Reset Problem Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{phyResetCount} = $element[1];
        }
        elsif($line =~ /CRC Error Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{crcErrCount} = $element[1];
        }
        elsif($line =~ /PHY Change Count/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$phy}{phyChangeCount} = $element[1];
        }
    }
    return \%hash;
}

=head2  ses_report_page8283

        my $ret = ses_report_page8283(host=>$host, conmode=>"shelfcli",timeout => $timeout);
        This routine retrieves all the SES page 82/83 data and save in a hash table
        with the following keys.

        Host,Drive,SIL,GlobalModuleControl,PHYResetTimeout,AWTBias.

dd        For Keys Host,Drive and SIL:
          Internally the Nested Hash keys are

        "Profile,BitErrorRateSampleWindow,ResetProblemSampleWindow,RunningDisparityThreshold,InvalidDwordThreshold,LossofDWORDSyncThreshold,PHYCHANGEThreshold,CRCErrorThreshold,PHYResetProblemThreshold."


        Example of returning ses_report_page8283 info:
                $info = ses_report_page8283();

        $VAR1 = \{
            'SIL' => {
                       'Profile' => '3f00h',
                       'BitErrorRateSampleWindow' => '60 min',
                       'ResetProblemSampleWindow' => '60 min',
                       'RunningDisparityThreshold' => '0078h',
                       'InvalidDwordThreshold' => '0078h',
                       'LossofDWORDSyncThreshold' => '003ch',
                       'PHYCHANGEThreshold' => '78h',
                       'CRCErrorThreshold' => '0078h',
                       'PHYResetProblemThreshold' => '003ch'
                     },
            'PHYResetTimeout' => '00 per 100 mS',
            'GlobalModuleControl' => '00h',
            'Drive' => {
                         'Profile' => '3f00h',
                         'BitErrorRateSampleWindow' => '60 min',
                         'ResetProblemSampleWindow' => '60 min',
                         'RunningDisparityThreshold' => '0078h',
                         'InvalidDwordThreshold' => '0078h',
                         'LossofDWORDSyncThreshold' => '003ch',
                         'PHYCHANGEThreshold' => '78h',
                         'CRCErrorThreshold' => '0078h',
                         'PHYResetProblemThreshold' => '003ch'
                       },
       'Host' => {
                        'Profile' => '3f00h',
                        'BitErrorRateSampleWindow' => '60 min',
                        'ResetProblemSampleWindow' => '60 min',
                        'RunningDisparityThreshold' => '0078h',
                        'InvalidDwordThreshold' => '0078h',
                        'LossofDWORDSyncThreshold' => '003ch',
                        'PHYCHANGEThreshold' => '78h',
                        'CRCErrorThreshold' => '0078h',
                        'PHYResetProblemThreshold' => '003ch'
                      },
            'AWTBias' => '0000 uS'
          };






The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut


sub ses_report_page8283 {

        my (%opts) = validate(@_,{
                                host => {type=>OBJECT,isa=>'Hostrec'},
                                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                                timeout => {type=>SCALAR, default=>300,optional=>1},
                                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                                AorB => {type=>SCALAR, default=>undef,optional=>1},
                                page => {type=>SCALAR},
                }
        );

        my %hash = ();
        my $output;
        my $c;
        my $cmd;
        my $id;
        my $command = "ses_report_page $opts{page}";

#       Make Connection and build command based on input parameters
    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
                $opts{conntype} = "#console";
                $c = make_conn($opts{host},$opts{conntype});
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
                if(!defined($opts{gemconn})){
                        croak "GEMTEST needs to be connected before we can use gemtest connmode";
                }
                $c = $opts{gemconn};
                $cmd = "$command";
        }
        elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
                if($opts{filermode} ne "normal"){
                        $opts{conntype} = "#console";
                }

                $c = make_conn($opts{host},$opts{conntype});
                if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
                        croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
                }
                if(!defined($opts{AorB})){
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
                }
                else{
                        $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
                }

                if($opts{filermode} ne "normal"){

                        $cmd = "sasadmin expander_cli $id \"$command\"";
                }
                else{
                        $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
                }

        }

    else {
                croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
        }

#       Issue Command

        if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
                (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
        }
        else{
                (undef,undef,$output) = say($c,"$cmd");
        }


        if($output !~ /SES Page/){
                croak "Unexpected ses_report_page output: $output";
        }

        my $phy;
        my @output = split(/\r*\n/,$output);

        print "Output is : @output \n";
        foreach my $line (@output) {
                $line =~ s/^\s+//g;
                if($line =~ /Global\sModule\sControl/) {
                        my @element = split(/:\s*/,$line);
                        $hash{GlobalModuleControl} = $element[1];
                }
                elsif($line =~ /PHY\sReset\sTimeout/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{PHYResetTimeout} = $element[1];
                }
                elsif($line =~ /AWT\sBias/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{AWTBias} = $element[1];
                }
                elsif($line =~ /PHYs/) {
                        my @tmp = split(/ /,$line);
                        $phy = $tmp[0];
                }
        elsif($line =~ /Profile/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{Profile} = $element[1];
                }
                elsif($line =~ /Bit\sError\sRate\sSample\sWindow/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{BitErrorRateSampleWindow} = $element[1];
                }
                elsif($line =~ /Reset Problem Sample Window/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{ResetProblemSampleWindow} = $element[1];
                }
                elsif($line =~ /PHY CHANGE Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{PHYCHANGEThreshold} = $element[1];
                }
                elsif($line =~ /Invalid Dword Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{InvalidDwordThreshold} = $element[1];
                }
                elsif($line =~ /Running Disparity Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{RunningDisparityThreshold} = $element[1];
                }
                elsif($line =~ /Loss of DWORD Sync Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{LossofDWORDSyncThreshold} = $element[1];
                }
                elsif($line =~ /PHY Reset Problem Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{PHYResetProblemThreshold} = $element[1];
                }
                elsif($line =~ /CRC Error Threshold/) {
                        my @element = split(/\s*:\s*/,$line);
                        $hash{$phy}{CRCErrorThreshold} = $element[1];
                }
        }
        return \%hash;
}

=head2  ddump_hid_byIndex

    my $ret = ddump_hid_byIndex(host=>$host, conmode=>"shelfcli",timeout => $timeout);
    This routine retrieves all the LID data and save in a hash table
    with the following keys stated for the Drive Manager and for each Device.
        HAMode. driveSpinUp, driveBays, allowedDrives, drivePowerControl, deviceRole,
        instanceRunState, syncToSlaveStatus, nextUID, totalExpandTrans, transPoolCap,
        transPoolFreeCount, numConACKS
    For each device bay: present, swap, SESInfoBit, spinUpTime, driveType, name, arrayName,
    wwn, faults, faultLED, inject, pending, current.

        Example of returning ddump_hid info:
        $info = ddump_hid();
        print "the fault LED status for drive bay 9 is: $info->{9}{faultLED}\n";

    Example of reference returned by the function:
    $VAR1 = {
          'HAMode'  => 'master',
          'driveSpinUp'  => '0',
          'driveBays'  => '24',
          'driveIndexBase'  => '0',
          'allowedDrives'  => 'SAS OR SATA',
          'drivePowerControl'  => 'supported',
          'deviceRole' '' => 'MASTER,
          'instanceRunState'  => 'RUNNING',
          'syncToSlaveStatus'  => 'None in progress',
          'nextUID'  => '0x63E8',
          'totalExpandTrans'  => '0x0 (0)',
          'transPoolCap'  => '0x40 (64)',
          'transPoolFreeCount'  => '0x40 (64)',
          'numConACKS'  => '0x48 (72)',
          '0' 'present' => 'yes',
          '0' 'swap' => 'no',
          '0' 'SESInfoBit' => 'not set',
          '0' 'spinUpTime' => '0+00:00:08.087',
          '0' 'driveType' => 'SAS',
          '0' 'wwn' => '500000E0173F6792',
          '0' 'faults' => 'none',
          '0' 'faultLED' => 'OFF',
          '0' 'inject' => 'NONE',
          '0' 'pending' => 'ONLINE',
          '0' 'current' => 'ONLINE',
    }

The following arguments are accepted:

=over

=item host

(Required) : Hostspec to shelf or filer.

=item conntype

(Optional) : connection type. Default "#rsh" if connmode is "inband" and "#console" if connmode is "shelfcli" or "gemtest"

=item timeout

(Optional) : timeout value. Default is 300 secs.

=item connmode

(Optional) : Connection Mode. "shelfcli" if we are connecting directly through shelf cli. "inband" if we are issuing inband command via ontap. "gemtest" if we are using "gremtest" windows utility. Default is "shelfcli".

=item filermode

(Optional) : Filermode. Maintenance Mode or Normal Mode. Default is "normal".

=item adapter_id

(Optional) : Adapter id to which the shelf is attached to. Required only when connmode=>"inband"

=item shelf_id

(Optional) : shelf id to which we want to issue command to. Required only when connmode=>"inband"

=item AorB

(Optional) : Controller A or B of the shelf. Required only when connmode=>"inband"

=back

=cut

sub ddump_hid_byIndex {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "ddump_hid";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},cmd=>$command);
    }
    elsif($opts{connmode} eq "gemtest"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},gemconn=>$opts{gemconn},cmd=>$command);
    }
    elsif($opts{connmode} eq "inbandcli"){
        $output = gem_cmd(host=>$opts{host},conntype=>$opts{conntype},connmode=>$opts{connmode},timeout=>$opts{timeout},adapter_id=>$opts{adapter_id},shelf_id=>$opts{shelf_id},filermode=>$opts{filermode},cmd=>$command);
    }

#   Issue Command

    if($output !~ /HID diagnostic dump/){
        croak "Unexpected ddump_hid output: $output";
    }

    my %hash = ();
    my $bay;
    my @output = split(/\r*\n/,$output);
    my $index = "";

    print "Output is : @output \n";
    foreach my $line (@output) {
        $line =~ s/^\s+//g;
        if($line =~ /Total number of LEDs/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{total_leds} = $element[1];
        }
        elsif($line =~ /^Index           \s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $index = $element[1];
        }
        elsif($line =~ /Purpose\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Purpose} = $element[1];
        }
        elsif($line =~ /Type\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Type} = $element[1];
        }
        elsif($line =~ /Supported Colours\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Supported_Colours} = $element[1];
        }
        elsif($line =~ /User Info\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{User_Info} = $element[1];
        }
        elsif($line =~ /Current Status\s*:/) {
            my @element = split(/\s*:\s*/,$line);
            $hash{$index}{Status} = $element[1];
        }
    }
    return \%hash;
}

=head2 storage_show_diskled

        my $ret = storage_show_diskled(host=>$host, adapter_name=>$adapter_name,timeout => $timeout);

This routine returns a hash table of information for disk led status.

It returns 1 on success and croaks on failure.
The following named arguments are accepted:

=over

=item host

(Required) : Hostspec to filer.

=item adapter_name

(Optional) : Specific channel for which the information is
 required. Default is all

=item timeout

(Optional) : timeout value. Default is 300 secs.

=back
=cut

sub storage_show_diskled($@) {
        my (%opts) = validate(@_,{
                                host => {type=>OBJECT, isa=>'Hostrec'},
                                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                                timeout => {type=>SCALAR, default=>300,optional=>1},
                                filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                }
        );
        my $adapter = undef ;
    my $output;

        my %hash =();
        if ($opts{filermode} ne "normal") {
        $opts{conntype} = "#console";
    }

        my $conn = make_conn($opts{host},$opts{conntype});
        if ($opts{filermode} ne "normal") {
                my $cmd = "storage show disk_led -a";
                (undef, undef, $output) = say($conn, "$cmd", -prompt, "$MAINT_PROMPT", "-timeout", $opts{timeout});
        }
        else{
                my $cmd = "priv set -q diag;storage show disk_led -a";
                (undef, undef, $output) = say($conn, "$cmd","-timeout", $opts{timeout});
        }
        if($output !~ /Disk name/){
                croak $output;
        }
        my $hba;

        my @lines = split(/\r*\n/, $output);

    foreach my $line (@lines) {
        $line =~ s/^\s+//g;
        if (($line !~ /Disk name/) && ($line !~ /----/)) {
            my @disk_t1 = split(/\s+/,$line);
            my @disks = split(/\./,$disk_t1[0]);
            my $disk = $disks[1].".".$disks[2];
            $hash{$disk} = $disk_t1[1];
        }
    }
        return \%hash;
}


sub psmi_dump_regs {

    my (%opts) = validate(@_,{
                host => {type=>OBJECT,isa=>'Hostrec'},
                conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
                connmode => {type=>SCALAR,default=>"shelfcli",optional=>1},
                filermode => {type=>SCALAR,default=>"normal",optional=>1},
                gemconn => {type=>OBJECT,default=>undef,isa=>'Connectrec',optional=>1},
                timeout => {type=>SCALAR, default=>300,optional=>1},
                adapter_id => {type=>SCALAR, default=>undef,optional=>1},
                shelf_id => {type=>SCALAR, default=>undef,optional=>1},
                AorB => {type=>SCALAR, default=>undef,optional=>1},
        }
    );

    my %hash = ();
    my $output;
    my $c;
    my $cmd;
    my $id;
    my $command = "psmi_dump_regs";

#   Make Connection and build command based on input parameters

    if($opts{connmode} eq "shelfcli"){ # Connection when we are directly connected to the shelf console
        $opts{conntype} = "#console";
        $c = make_conn($opts{host},$opts{conntype});
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "gemtest"){ # Connection through gemtest
        if(!defined($opts{gemconn})){
            croak "GEMTEST needs to be connected before we can use gemtest connmode";
        }
        $c = $opts{gemconn};
        $cmd = "$command";
    }
    elsif($opts{connmode} eq "inbandcli"){ # Connection through the filer
        if($opts{filermode} ne "normal"){
            $opts{conntype} = "#console";
        }

        $c = make_conn($opts{host},$opts{conntype});
        if(!defined($opts{adapter_id}) || !defined($opts{shelf_id})){
            croak ("For inband cli through filer, adapter_id parameter and shelf_id parameter needs to be specified");
        }
        if(!defined($opts{AorB})){
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" ;
        }
        else{
            $id = "$opts{adapter_id}" . "." . "$opts{shelf_id}" . "." . "$opts{AorB}";
        }

        if($opts{filermode} ne "normal"){

            $cmd = "sasadmin expander_cli $id \"$command\"";
        }
        else{
            $cmd = "priv set -q diag;sasadmin expander_cli $id \"$command\"";
        }

    }

    else {
        croak "Unsupported connection mode $opts{connmode}. Only supported connection modes are : shelfcli, inbandcli and gemtest";
    }

#   Issue Command

    if(($opts{filermode} ne "normal") && ($opts{connmode} eq "inbandcli")){
        (undef,undef,$output) = say($c,"$cmd",-prompt,"$MAINT_PROMPT");
    }
    else{
        (undef,undef,$output) = say($c,"$cmd");
    }
    print "Output is  C: $output \n";
    my @output = split(/\r*\n/,$output);
    my $register = 0;
    my $count = 0;

    my $value;
    foreach my $line (@output){
        if ($line =~ /: /) {
            my @line_temp = split(/: /,$line);
            my @output_temp = split(//,$line_temp[1]);
            foreach $value (@output_temp){
                if($value =~ /\d|[a-f]/){ # To make sure that the $value is hex and filter out the rest
                    if ($count == 0) {
                        $hash{$register} = "$value";
                    } elsif ($count < 4) {
                        $hash{$register} = $hash{$register}."$value";
                    }
                    $count = $count + 1;
                    if ($count > 3) {
                        #print "value in dump psmi Register $register is value $value\n";
                        $register = $register + 1;
                        $count = 0;
                    }

                }
            }
        }
    }
    return \%hash ;
}


=head1 SEE ALSO

    Where additional information can be found.

=head1 BUGS

    Description of bugs or what the library does not do and why

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    Anish Vashist (anishv@netapp.com)
    Last revised $Date: 2013/09/30 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

1;
