#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     CIFS Topology Library
##
## @author      bellizzi@netapp.com
##
## @pod here

=head1 NAME

Topology::Config::7M - CIFS Topology library for 7Mode ONTAP

=head1 SYNOPSIS

    See Topology.pm

=head1 DESCRIPTION

=head1 TODO

=head1 SEE ALSO

Options Topology

=head1 BUGS

=head1 STATUS

  Shared

=head1 AUTHOR

  David Bellizzi (bellizzi@netapp.com)
    
=head1 COPYRIGHT

Copyright 2012, Network Appliance., All Rights Reserved

=cut

package Topology::Config::7M;

use strict;

use Tharn qw($Log connect config_result script_result host say snooze);
$Log->comment("Starting " . __FILE__);
use base qw(Topology::NAS::CIFS);
use Module::Common qw(validate_mixed);
use Params::Validate
    qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Data::Dumper qw(Dumper);
use Options;
use Topology::Config::Parameters;
use File::Basename;
use Utils::DNS qw(get_name_by_ip);
use CIFS::Client;
use Net::IP qw(:PROC);
use CIFS::Setup qw(cifs_setup);
use Sort::Key::Natural qw(natsort);
use filer qw(filer_vfiler_command filer_vfiler_context_simulate);
use POSIX qw(strftime);
use NetAddr::IP::Lite;
use Utils::HostManager::Helper;
use NACL::APISet;
use NACL::C::Volume;
use NACL::C::Node;
use NACL::Transit;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::FeatureNotLicensedException qw(:try);
use CLI::FileIO qw(wrfile rdfile);
use Limits;

use if ($^O ne "MSWin32"), 'DBM::Deep';

$Data::Dumper::Sortkeys = 1;

my $Options = new Options;
$Options->process();

sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;
    my $self     = {@_,};
    $Log->enter();
    bless($self, $class);
    $self->{reserve_clients} = 0;
    $Log->exit();
    return $self;
} ## end sub new

my ($Options_hash, %apiset, $Response, %conn, $topology_database);
my @ontap_shares = ('C$', 'ETC$', 'ONTAP_ADMIN$', 'IPC$');
my $volume_basename = 'base';

sub define_topology {
    $Log->enter();

    my $self = shift;

    # Execute the one from the parent
    $self->SUPER::define_topology(@_);

    $Options_hash = $Options->get_hash();
    $Log->comment(Dumper($Options_hash));

    # Find calling script's name
    my $script = fileparse($0, qr/\.[^.]*/);

    $self->{'script'} = $script;
    my $topo       = $Options_hash->{'TOPO'};
    my $top_logdir = $Options_hash->{'TOP_LOGDIR'};

    $topology_database = DBM::Deep->new("$top_logdir/$topo.db");
    $self->{'topology_database'} = $topology_database;

    # Wait for other scripts to start if needed
    if ($Options_hash->{'WAIT_FOR_FILE'}) {
        $Log->comment("Waiting for $Options_hash->{'WAIT_FOR_FILE'}");
        $self->wait_for_file('file' => $Options_hash->{'TOP_LOGDIR'} . '/'
                . $Options_hash->{'WAIT_FOR_FILE'});
    }

    $self->{'clients'} = 0;

    # Retrieve the clients from the client file or ndispd
    my @clients
        = retrieve_clients('script' => $script, 'options' => $Options_hash);
    $self->{'clients'} = \@clients;

    # Make 7Mode and ZAPI APISet connections for all nodes and store
    # for reuse
    foreach my $node (@{$Options_hash->{'NODES'}}) {

        $Log->comment("Creating CLI connection object for $node");
        my $Host   = host($node);
        my %fields = $Host->param_fields();

        # if model is vsa, dedupe is not supported
        if (($fields{'TYPE'} =~ /VSA/i)
            && (  $Options_hash->{'DEDUPE_COMPRESSION_PERCENT'}
                + $Options_hash->{'DEDUPE_INLINE_PERCENT'} > 0)
            ) {
            die config_result(
                'VSAs do not support dedup with compression: set DEDUPE_COMPRESSION_PERCENT and DEDUPE_INLINE_PERCENT to 0'
            );
        }
        if (   ($fields{'TYPE'} =~ /VSA/i)
            && ($Options_hash->{'FLEX_CACHE_PERCENTAGE'} > 0)) {
            die config_result(
                'VSAs do not support FLEX CACHE: set FLEX_CACHE_PERCENTAGE to 0'
            );
        }
        $apiset{'cli'}{$node} = NACL::APISet->new(
            hostobj   => $Host,
            category  => "Node",
            interface => "CLI",
            set       => "7Mode",
            cdef_path => $Options_hash->{'CDEF_PATH'}
        );

        $apiset{'cli'}{$node}
            ->set_timeout(timeout => $Options_hash->{'APISET_TIMEOUT'});

        $Log->comment("Creating ZAPI connection object for $node");
        $apiset{'zapi'}{$node} = NACL::APISet->new(
            hostobj   => $Host,
            category  => "Node",
            interface => "ZAPI",
            set       => "7Mode",
            cdef_path => $Options_hash->{'CDEF_PATH'}
        );

        $apiset{'zapi'}{$node}
            ->set_timeout(timeout => $Options_hash->{'APISET_TIMEOUT'});

        $Log->comment("Creating Tharn connection for $node");
        $conn{$node} = connect('#rsh', '-nologin', -timeout, 300);

        $topology_database->put($node, {});

    }

    my $version_mgr = $apiset{'cli'}{${$Options_hash->{'NODES'}}[0]}
        ->get_version_manager();

    # Attribute can be any of branch; release; build_options; build_path;
    # change_number; wafl_version; node_version
    my $release
        = $version_mgr->get_version_attribute('attribute' => 'release');
    $self->{'release'}
        = $version_mgr->map_release_to_codename(release => $release);

    my @targets = (
        'aggregates',    'cifs',   'interfaces', 'licenses',
        'qtrees',        'quotas', 'shares',     'vfilers',
        'vfiler_limits', 'volumes'
    );

    if (   $Options_hash->{'PRE_TEST_CLEANUP'}
        && $Options_hash->{'PRE_TEST_CLEANUP'} eq 'discovered') {

        $self->filer_cleanup_discovered();
        snooze 1;
    }

    $self->discover_topology('targets' => \@targets);

    if (!$Options_hash->{'AGGREGATES'}) {

        #$self->find_disks();
        die script_result('AGGREGATES is a required parameter');
    }

    if (!@{$Options_hash->{'DATA_INTERFACES'}}) {
        die script_result('DATA_INTERFACES is a required parameter');
    }

    if ($Options_hash->{'CIFS_ENABLED'}) {
        my @clients;
        try {
            @clients = grep { host($_)->hosttype() =~ /win/ }
                @{$self->{'clients'}};
        }
        otherwise {
            my $Exception = shift;
            $Log->comment($Exception->text());
        };
        $Log->comment(@clients);
        if (!@clients) {
            die script_result(
                'In order to use CIFS_ENABLED, windows clients must be provided in CLIENTFILE for shares'
            );
        }
    }
    if (!$Options_hash->{'NODES'}) {
        die config_result('NODES(FILER) is a required parameter ');
    }

    if (   !$Options_hash->{'VOLUME_COUNT'}
        && !$Options_hash->{'VOLUME_COUNT_PERCENTAGE'}) {
        die config_result(
            'VOLUME_COUNT or VOLUME_COUNT_PERCENTAGE is a required parameter'
        );
    }

# if Volume Count not defined but Volume Count Percentage is, find max vols and calc percent
    if (  !$Options_hash->{'VOLUME_COUNT'}
        && $Options_hash->{'VOLUME_COUNT_PERCENTAGE'}) {
        if ($Options_hash->{'VOLUME_COUNT_PERCENTAGE'} > 100) {
            die config_result(
                'VOLUME_COUNT_PERCENTAGE must be less than 100');
        }
        my $max_volumes = 0;
        foreach my $node (@{$Options_hash->{'NODES'}}) {
            my $volume_max;
            try {
                my $Limits = Limits->new(host => $node);
                $volume_max = $Limits->getvalue(
                    limit     => "max_flexvols",
                    subsystem => "wafl"
                );
                $Log->comment(
                    "Found max_flexvols for $node to be $volume_max from Limits Finder"
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            }
            otherwise {
                my $Zapi        = $apiset{'zapi'}{$node};
                my $Result      = $Zapi->system_get_info();
                my $system_info = $Result->get_parsed_output();
                $volume_max = $system_info->[0]->{'system-info'}->[0]
                    ->{'maximum-flexible-volume-count'};
                $Log->comment(
                    "Found max_flexvols for $node to be $volume_max from System Info"
                );
            };
            $max_volumes += $volume_max;
        }
        $self->{'volume_count'} = int(
            (         $max_volumes 
                    * $Options_hash->{'VOLUME_COUNT_PERCENTAGE'} / 100
            )
        );
        $Log->comment(
            "Calculated the volume count to be " . $self->{'volume_count'});
    } else {

        # Store the current volume count as the total
        $self->{'volume_count'} = $Options_hash->{'VOLUME_COUNT'};
    }

    if (   $Options_hash->{'VFILER_COUNT'}
        && ($Options_hash->{'VFILER_COUNT'} * 2) > $self->{'volume_count'}
        && !$Options_hash->{'SHARED_VFILERS'}) {
        die config_result('Not enough volumes specified for all vfilers');
    }

    if ((     $Options_hash->{'FLEX_CACHE_PERCENTAGE'}
            + $Options_hash->{'FLEX_CLONE_PERCENTAGE'}
        ) > 50
        ) {
        die config_result(
            'FLEX_CACHE_PERCENTAGE + FLEX_CLONE_PERCENTAGE must be less than 50%'
        );
    }

    # Store the relevant info in the topo
    $self->{'vfiler_count'}    = $Options_hash->{'VFILER_COUNT'};
    $self->{'aggregates'}      = $Options_hash->{'AGGREGATES'};
    $self->{'data_interfaces'} = $Options_hash->{'DATA_INTERFACES'};

    # Split the volume count per node
    $self->{'volumes_per_node'}
        = int($self->{'volume_count'} / @{$Options_hash->{'NODES'}});

    if ($self->{'volumes_per_node'} < 1) {
        $self->{'volumes_per_node'} = 1;
    }

    # Force creation of the vfiler0 key
    foreach my $node (@{$Options_hash->{'NODES'}}) {
        $self->{'defined_topology'}->{$node}->{'vfilers'}->{'vfiler0'} = {};
    }

    # If vfilers, spread the vfilers across the nodes
    $self->{'vfilers_per_node'}   = 0;
    $self->{'volumes_per_vfiler'} = 0;

#calculate the number of cache and clone vols to create and reduce the number of datavols accordingly
    $self->{'flex_cache_count'}
        = int($self->{'volume_count'} 
            * $Options_hash->{'FLEX_CACHE_PERCENTAGE'}
            / 100);
    $self->{'flex_clone_count'}
        = int($self->{'volume_count'} 
            * $Options_hash->{'FLEX_CLONE_PERCENTAGE'}
            / 100);
    $self->{'data_volume_count'}
        = $self->{'volume_count'} 
        - $self->{'flex_cache_count'}
        - $self->{'flex_clone_count'};
    $self->{'extent_volumes'}
        = int($self->{'volume_count'} 
            * $Options_hash->{'VOLUME_EXTENT_PERCENT'}
            / 100);

    foreach my $node (@{$Options_hash->{'NODES'}}) {

        #calculate the number of volumes dedupe and dedupe+compression
        $self->{'defined_topology'}->{$node}->{'dedupe_vol_count'}
            = int($self->{'data_volume_count'} 
                * $Options_hash->{'DEDUPE_PERCENT'}
                / 100);
        $self->{'defined_topology'}->{$node}->{'compress_vol_count'}
            = int($self->{'data_volume_count'}
                * $Options_hash->{'DEDUPE_COMPRESSION_PERCENT'} 
                / 100);
        $self->{'defined_topology'}->{$node}->{'inline_vol_count'}
            = int($self->{'data_volume_count'}
                * $Options_hash->{'DEDUPE_INLINE_PERCENT'} 
                / 100);

        $self->{'defined_topology'}->{$node}->{'dedupe_clone_count'}
            = int($self->{'flex_clone_count'} 
                * $Options_hash->{'DEDUPE_PERCENT'}
                / 100);
        $self->{'defined_topology'}->{$node}->{'compress_clone_count'}
            = int($self->{'flex_clone_count'}
                * $Options_hash->{'DEDUPE_COMPRESSION_PERCENT'} 
                / 100);
        $self->{'defined_topology'}->{$node}->{'inline_clone_count'}
            = int($self->{'flex_clone_count'}
                * $Options_hash->{'DEDUPE_INLINE_PERCENT'} 
                / 100);

        #calculate the number of volumes to create qtrees on
        $self->{'defined_topology'}->{$node}->{'qtree_count'}
            = int($self->{'volume_count'} 
                * $Options_hash->{'QTREE_VOL_PERCENT'}
                / 100);

        #calculate the number of volumes to enable quotas on
        $self->{'defined_topology'}->{$node}->{'quota_vol_count'}
            = int($self->{'volume_count'} 
                * $Options_hash->{'QUOTA_VOLUME_PERCENT'}
                / 100);
    }

    if ($Options_hash->{'VFILER_COUNT'}) {

        my @nodes = @{$Options_hash->{'NODES'}};
        $self->{'vfilers_per_node'} = int(
            $Options_hash->{'VFILER_COUNT'} / @{$Options_hash->{'NODES'}});
        foreach my $node (@{$Options_hash->{'NODES'}}) {
            my $discovered_vfilers
                = $self->{'discovered_topology'}->{$node}->{'vfilers'};
            my $discovered_volumes
                = $self->{'discovered_topology'}->{$node}->{'volumes'};
            my @vfilers;
            my $vfiler_index = 1;
            foreach my $v (1 ... $self->{'vfilers_per_node'}) {

                # Create the vfiler keys
                my $vfiler = $node . '-vf' . $vfiler_index;
                push(@vfilers, $vfiler);
                $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                    = {};
                $vfiler_index++;
            }

            if (!$Options_hash->{'SHARED_VFILERS'}) {
                $self->assign_vfiler_root_volumes(
                    'node'    => $node,
                    'size'    => 100,
                    'vfilers' => \@vfilers,
                );
            } else {

                foreach my $v (1 ... $self->{'vfilers_per_node'}) {

                    # Create the vfiler keys
                    my $vfiler = $node . '-vf' . $v;

               # Populate the root volume entries in the defined_topology hash
                    my $volume
                        = ${$discovered_vfilers->{$vfiler}->{'path'}}[0];
                    $volume =~ s/^.+\/(.+)$/$1/;

                    my $aggregate = $discovered_volumes->{$volume}
                        ->{'containing-aggregate'};
                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'aggregate'} = $aggregate;

                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'vfiler'} = $vfiler;

                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'share_count'} = 0;

                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'root_volume'} = 1;

                    # Assign root volume for vfiler
                    push(
                        @{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                                ->{$vfiler}->{'volumes'}
                            },
                        $volume
                    );

                }
            }
        }

        # Determine the max number of volumes each vfiler gets and if the
        # assignment method. rnd = random sym = symmetrical
        if (!$Options_hash->{'SHARED_VFILERS'}) {
            $self->{'data_volume_count'} -= $Options_hash->{'VFILER_COUNT'};
        }
        if ($Options_hash->{'VOLUMES_PER_VFILER'} =~ /(\w+):(\d+)/) {
            $self->{'volumes_per_vfiler_assignment'} = $1;
            $self->{'volumes_per_vfiler'}            = $2;
        } elsif ($Options_hash->{'VOLUMES_PER_VFILER'} =~ /(\w+)/) {
            $self->{'volumes_per_vfiler_assignment'} = $1;

            # Account for the root volume
            if (!$Options_hash->{'SHARED_VFILERS'}) {
                $self->{'volumes_per_vfiler'} = 1;
            }

            # Make sure all vfilers have a data volume
            my $data_volumes_per_vfiler = int($self->{'data_volume_count'}
                    / ($self->{'vfiler_count'} + @nodes));
            if (!$data_volumes_per_vfiler) { $data_volumes_per_vfiler = 1; }
            $self->{'volumes_per_vfiler'} += $data_volumes_per_vfiler;
        } else {
            $self->{'volumes_per_vfiler_assignment'} = 'sym';
            $self->{'volumes_per_vfiler'}
                = $Options_hash->{'VOLUMES_PER_VFILER'};
        }

        my $vfiler_volumes
            = $self->{'volumes_per_vfiler'} * $self->{'vfiler_count'};
        my $required_volumes = $vfiler_volumes + @nodes;
        if ($required_volumes > $self->{'volume_count'}) {
            my $vc = $self->{'volume_count'};
            die script_result("The number of required volumes "
                    . "($required_volumes) exceeds VOLUME_COUNT"
                    . "($vc) Please increase VOLUME_COUNT.");
        }

        if (  !$Options_hash->{'SHARED_VFILERS'}
            && $self->{'volumes_per_vfiler'} < 2) {
            my $required_volumes
                = ($Options_hash->{'VFILER_COUNT'} * 2) + @nodes;
            die script_result("Not enough volumes for all vfilers. "
                    . "$Options_hash->{'VOLUME_COUNT'} was specified but "
                    . "$required_volumes are required. There must be at least "
                    . "2 volumes per vfiler. Please increase VOLUME_COUNT");
        }
    }

    if ($Options_hash->{'LOCAL_GROUP_COUNT'}) {
        $self->{'local_group_count'} = $Options_hash->{'LOCAL_GROUP_COUNT'};

        # Split the localgroup count per node
        $self->{'local_groups_per_node'}
            = int($self->{'local_group_count'} / @{$Options_hash->{'NODES'}});

        # Split the localgroup count per vfiler
        if ($self->{'vfilers_per_node'} > 0) {
            $self->{'local_groups_per_vfiler'}
                = int($self->{'local_groups_per_node'}
                    / $self->{'vfilers_per_node'});
        } else {
            $self->{'local_groups_per_vfiler'}
                = $self->{'local_groups_per_node'};
        }
    }

    my $size;
    foreach my $node (@{$Options_hash->{'NODES'}}) {
        if ($Options_hash->{'VOLUME_SIZE'}) {
            $Log->comment("using volume size "
                    . convert_size($Options_hash->{'VOLUME_SIZE'})
                    . "Mb");
            $size = convert_size($Options_hash->{'VOLUME_SIZE'});
            $self->{'discovered_topology'}->{$node}->{'vol_size'} = $size;

#if size is not defined, calculate based on available aggr space and # of vols
        } else {
            my $aggregate_space = 0;

            my @discovered_aggregates = natsort(
                keys(
                    %{  $self->{'discovered_topology'}->{$node}
                            ->{'aggregates'}
                        }
                )
            );

           # loop through the discovered aggregates and create an array of our
           # defined aggregates
            my @defined_aggregates;
            foreach my $aggregate (@discovered_aggregates) {
                if (grep { $_ eq $aggregate }
                    split(/,/, $Options_hash->{'AGGREGATES'})
                    ) {
                    push(@defined_aggregates, $aggregate);
                }
            }

            # get total available aggr space
            foreach my $aggregate (@defined_aggregates) {
                $self->{'discovered_topology'}->{$node}->{'aggregates'}
                    ->{$aggregate}->{'total'} =~ /(\d+)M/;
                $aggregate_space += int(
                    $1 * ($Options_hash->{'MAX_AGGREGATE_USAGE'} / 100));
            }
            $Log->comment("$node aggregate_space = $aggregate_space");
            $size = int($aggregate_space / $self->{'volume_count'});
            $self->{'discovered_topology'}->{$node}->{'vol_size'} = $size;
            $Log->comment("Calculated volume size for $node to be $size");
            if (convert_size($size) < 40) {
                die config_result('Volume size must be larger than 40m');
            }
        }
    }

    # Assign data volumes
    $self->assign_data_volumes(
        'count'    => $self->{'data_volume_count'},
        'basename' => $volume_basename,
        'type'     => 'flexvol'
    );

    #assign clone volumes
    $self->assign_data_volumes(
        'count'    => $self->{'flex_clone_count'},
        'basename' => $volume_basename . '_clone',
        'type'     => 'clone',
    ) if ($self->{'flex_clone_count'} > 0);

    $Log->comment(Dumper($self->{'defined_topology'}));

    # Parse address ranges
    if (!$Options_hash->{'DESTIP'}) {
        die config_result('DESTIP is not defined');
    }

    $Log->comment("Parsing DESTIP");
    foreach my $spec (split(/,/, $Options_hash->{'DESTIP'})) {
        if ($spec =~ /^(\d+);([^;]+);(.+)/) {

            my $index     = $1;
            my $node      = ${$Options_hash->{'NODES'}}[$index];
            my $interface = $2;
            my $ranges    = $3;
            foreach my $range (split(/:/, $ranges)) {
                $Log->comment(
                    "Found IP range $range for $interface on $node");
                my $Host             = host($range);
                my %host_file_params = $Host->param_fields();
                $self->{'defined_topology'}->{$node}->{'addresses'}
                    ->{$interface}->{$range} = \%host_file_params;
            }
        } else {
            die config_result("DESTIP must be in the format of\n"
                    . "<node index>;<interface>;<range>[:<range>],"
                    . "<node index>;<interface>;<range>[:<range>]\n"
                    . "Where range is a file in your nate_hosts "
                    . "dir that contains at least\n"
                    . "STARTIP\n"
                    . "GATEWAY\n"
                    . "NETMASK\n"
                    . "IPCOUNT");
        }
    }

    # Spread the addresses across vfilers
    foreach my $node (@{$Options_hash->{'NODES'}}) {

        my $discovered_interfaces
            = $self->{'discovered_topology'}->{$node}->{'interfaces'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_addresses
            = $self->{'defined_topology'}->{$node}->{'addresses'};
        my @interfaces = keys(%{$defined_addresses});
        if (!@interfaces) {
            die script_result(
                "No interfaces defined in DESTIP for node $node");
        }
        my %addresses;
        my %ranges;
        foreach my $interface (@interfaces) {

            foreach
                my $range (natsort(keys(%{$defined_addresses->{$interface}})))
            {
                my $start_ip
                    = $defined_addresses->{$interface}->{$range}->{'STARTIP'};
                my $ip_count
                    = $defined_addresses->{$interface}->{$range}->{'IPCOUNT'};
                my $ip_obj = new Net::IP("$start_ip  \+  " . ($ip_count - 1))
                    || die config_result("Failed to create ip object: $@");
                do {
                    my $ip = $ip_obj->ip();
                    push(@{$addresses{$interface}{$range}}, $ip);
                    $ranges{$ip} = $range;
                } while (++$ip_obj);
            }

        }

        $Log->comment("Adding addresses to node $node");

        # Assign addresses to vfiler0 data interfaces if needed
        my @assigned_ranges;

        # loop through each interface
        foreach my $interface (keys(%{$discovered_interfaces})) {
            $Log->comment(
                "Adding addresses to node node $node interface $interface");

            # Check both v4 and v6 addresses
            foreach my $version ('ipv4_info', 'ipv6_info') {
                my $address;

                # Are there addresses in this versions address list?
                if (exists(
                        $discovered_interfaces->{$interface}->{$version}->[0]
                    )
                    ) {

                    # Get the list of addresses for this interface
                    foreach my $info (
                        @{$discovered_interfaces->{$interface}->{$version}}) {

                        $address = $info->{'ip'};
                        $Log->comment(
                            "processing discovered interface $interface with address $address"
                        );
                        my $autoconf = $info->{'autoconf'};

                        # Don't use autoconf IPV6 addresses
                        if ($version =~ /ipv6/i && $autoconf) {
                            $Log->comment(
                                "Skipping IPV6 autoconf address $address");
                            next;
                        }

                        my $version_name = uc($version);
                        $version_name =~ s/_INFO//g;
                        $Log->comment("Adding discovered $version_name "
                                . "address $address to node $node "
                                . "vfiler vfiler0 interface $interface");

                        if ($version =~ /ipv6/i) {
                            $address = ip_expand_address($address, 6);
                        }

                        if ($ranges{$address}) {
                            my $range = $ranges{$address};
                            my $index = $self->get_array_index(
                                'array' => \@{$addresses{$interface}{$range}},
                                'value' => $address
                            );
                            if (${$addresses{$interface}{$range}}[$index]) {
                                $Log->comment(
                                    "Removing $address from the list of available "
                                        . "addresses for interface $interface"
                                );
                                splice(@{$addresses{$interface}{$range}},
                                    $index, 1);
                            }
                        }
                        if ($address) {
                            push(
                                @{  $self->{'defined_topology'}->{$node}
                                        ->{'vfilers'}->{'vfiler0'}
                                        ->{'interfaces'}->{$interface}
                                    },
                                {   'address' => $address,
                                    'range'   => $ranges{$address},
                                    'mask'    => $info->{'mask'}
                                }
                            );
                            push(@assigned_ranges, $ranges{$address});
                            last;

                        }
                    }
                }
            }
        }

        foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
            foreach my $range (keys(%{$defined_addresses->{$interface}})) {
                if (grep { $_ eq $range } @assigned_ranges) {
                    next;
                }

#if(!defined($self->{'defined_topology'}->{$node}->{'vfilers'}
#                ->{'vfiler0'}->{'interfaces'}->{$interface}->[0]->{'address'})) {
                my $address = shift(@{$addresses{$interface}{$range}});
                $Log->comment("Adding defined address $address to node $node "
                        . "vfiler vfiler0 interface $interface");

                push(
                    @{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                            ->{'vfiler0'}->{'interfaces'}->{$interface}
                        },
                    {   'address' => $address,
                        'range'   => $range
                    }    ## end

                );

                #}
            }
        }

        # If we are doing vfilers assign the addresses to interfaces
        if ($self->{'vfilers_per_node'}) {

            my @vfilers = keys(%{$defined_vfilers});
            @vfilers = natsort(@vfilers);
            foreach my $vfiler (@vfilers) {
                if ($vfiler eq 'vfiler0') { next; }
                foreach my $interface (@interfaces) {
                    foreach
                        my $range (keys(%{$defined_addresses->{$interface}}))
                    {

                        my $address
                            = shift(@{$addresses{$interface}{$range}});
                        if (!$address) {
                            die config_result(
                                "No more IP addresses for interface $interface"
                            );
                        }
                        $Log->comment(
                            "Adding addresses to node $node vfiler $vfiler "
                                . "interface $interface address $address "
                                . "range $range");

                        push(
                            @{  $self->{'defined_topology'}->{$node}
                                    ->{'vfilers'}->{$vfiler}->{'interfaces'}
                                    ->{$interface}
                                },
                            {'address' => $address, 'range' => $range}
                        );

                    }
                }
            }
        }
    }

    # Create the CIFS server names
    $Log->comment("Creating the CIFS server name for all vfilers");
    foreach my $node (@{$Options_hash->{'NODES'}}) {
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $cifs_server = $node;
        $cifs_server =~ s/(-|fas)//g;
        $self->{'defined_topology'}->{$node}->{'vfilers'}->{'vfiler0'}
            ->{'cifs_server'} = $cifs_server;
        if ($self->{'vfilers_per_node'}) {
            my @vfilers = keys(%{$defined_vfilers});
            @vfilers = natsort(@vfilers);
            foreach my $vfiler (@vfilers) {

                if ($vfiler eq 'vfiler0') { next; }
                $cifs_server = $vfiler;
                $cifs_server =~ s/(-|fas)//g;
                $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                    ->{'cifs_server'} = $cifs_server;
            }
        }
    }

    $Log->comment("Adding volume shares");
    foreach my $node (@{$Options_hash->{'NODES'}}) {

        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};

        foreach my $volume (natsort(keys(%{$defined_volumes}))) {

            my $share  = 's_' . $volume;
            my $vfiler = $defined_volumes->{$volume}->{'vfiler'};

            $Log->comment(
                "Adding $share for node $node vfiler $vfiler volume $volume");

            # Store the the volume name with the share
            $topology_database->{$node}->{'shares'}->{$share}->{'volume'}
                = $volume;

            # Store the share path with the share
            $topology_database->{$node}->{'shares'}->{$share}->{'path'}
                = '/vol/' . $volume;

            # Store the owning vfiler with the share
            $topology_database->{$node}->{'shares'}->{$share}->{'vfiler'}
                = $vfiler;

            # Configure the filer_setup flag that determines if the
            # share is created in filer_setup or some other way
            $topology_database->{$node}->{'shares'}->{$share}->{'filer_setup'}
                = 1;

            push(
                @{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                        ->{$vfiler}->{'shares'}
                    },
                $share
            );
        }
    }

    if ($Options_hash->{'CIFS_ENABLED'}) {
        my $share_index = 1;

        $self->define_shares(
            'script'      => $self->{'testname'},
            'share_index' => $share_index,
            'recipe'      => 'swifttest'
        );
    }

    #assign cache volumes
    $self->assign_data_volumes(
        'count'    => $self->{'flex_cache_count'},
        'basename' => $volume_basename . '_cache',
        'type'     => 'cache',
    ) if ($self->{'flex_cache_count'} > 0);

    # Configure resource agitation
    if ($Options_hash->{'RA'}) {
        $self->get_ra_config();
    }

    # if option CLEANUP_ONLY is set, kill script before starting setup
    if ($Options_hash->{'CLEANUP_ONLY'}) {
        if ($Options_hash->{'PRE_TEST_CLEANUP'} eq 'defined') {
            $self->filer_cleanup_defined();
        }
        $Log->comment(Dumper($self));
        die script_result('death');
    }
    $Log->exit();
} ## end sub define_topology

sub compare_topology() {
    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    $self->define_topology();

    foreach my $node (@{$Options_hash->{'NODES'}}) {
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};

        my @defined_vfilers
            = keys(%{$self->{'defined_topology'}->{$node}->{'vfilers'}});
        my @discovered_vfilers
            = keys(%{$self->{'discovered_topology'}->{$node}->{'vfilers'}});

        foreach my $volume (keys(%{$discovered_volumes})) {
            if (grep(/^$volume$/, keys(%{$defined_volumes}))) {

                my $discovered_reserve
                    = $discovered_volumes->{$volume}->{'space-reserve'};
                my $defined_reserve
                    = $defined_volumes->{$volume}->{'guar_type'};
                if ($discovered_reserve ne $defined_reserve) {
                    $Log->warn(
                        "$volume space guarantee of $discovered_reserve does not match defined guarantee of $defined_reserve"
                    );
                }
                my $discovered_aggr = $discovered_volumes->{$volume}
                    ->{'containing-aggregate'};
                my $defined_aggr = $defined_volumes->{$volume}->{'aggregate'};
                if ($discovered_aggr ne $defined_aggr) {
                    $Log->warn(
                        "$volume containing aggregate $discovered_aggr does not match defined aggregate $defined_aggr"
                    );
                }

                my $discovered_vfiler
                    = $discovered_volumes->{$volume}->{'owning-vfiler'};
                my $defined_vfiler = $defined_volumes->{$volume}->{'vfiler'};
                if ($discovered_vfiler ne $defined_vfiler) {
                    $Log->warn(
                        "$volume owning vfiler $discovered_vfiler does not match defined vfiler $defined_vfiler"
                    );
                }

            } else {
                $Log->warn("Volume $volume was discovered but not defined");
            }
        }

        foreach my $vfiler (@discovered_vfilers) {
            if (grep(/^$vfiler$/, @defined_vfilers)) {
                next;
            } else {
                $Log->warn("Vfiler $vfiler was discovered but not defined");
            }
        }

    }

} ## end sub compare_topology()

sub filer_setup {

    $Log->enter();

    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::filer_setup(%options);

    $Options_hash = $Options->get_hash();

    # Make sure the multistore feature is licensed
    if ($Options_hash->{'VFILER_COUNT'}) {
        foreach my $node (@{$Options_hash->{'NODES'}}) {

            my $Cli  = $apiset{'cli'}{$node};
            my $Zapi = $apiset{'zapi'}{$node};

            # Check the discovered status of the multistore license
            my $multistore_license = 'not licensed';
            if (keys(
                    %{  $self->{'discovered_topology'}->{$node}->{'licenses'}
                        }
                )
                ) {
                $multistore_license
                    = $self->{'discovered_topology'}->{$node}->{'licenses'}
                    ->{'multistore'};
            }

            # License the multistore option
            if ($multistore_license =~ /not licensed/i) {
                $Log->comment("Adding multistore license");
                eval { $Cli->license_add("code" => 'ESSRXFF'); };
            }

            # Check the discovered status of the multistore enable option
            my $discovered_options
                = $self->{'discovered_topology'}->{$node}->{'options'};
            if (grep { $_ eq 'licensed_feature.multistore.enable' }
                keys(%{$discovered_options})
                ) {

                my $enable = 0;
                my $option = $discovered_options->{
                    'licensed_feature.multistore.enable'};

            # Is the licensed_feature option valid for this realease/platform?
                my $option_valid = $self->is_option_valid(
                    'node'   => $node,
                    'option' => 'licensed_feature.multistore.enable'
                );

                if ($option_valid && $option ne 'on') {

                    eval {
                        $Zapi->options_set(
                            'name'  => 'licensed_feature.multistore.enable',
                            'value' => 'on'
                        );
                    };
                }
            }
            if ($Options_hash->{'BOOTARGS'}) {
                $self->enable_bootargs('node' => $node);
            }

            # check vfiler limit, set it if needed and reboot
            my $vfiler_limit = $Options_hash->{'VFILER_COUNT'} + 1;
            if ($vfiler_limit
                > $self->{'discovered_topology'}->{$node}->{'vfiler_limits'}
                ->{'current_limit'}) {
                try {
                    $Response = $Cli->execute_raw_command(
                        command => "vfiler limit $vfiler_limit");
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
                if ($Options_hash->{'REBOOT'}) {
                    $self->reboot_node('node' => $node);
                }
            }

        }
    }

    if (   $Options_hash->{'PRE_TEST_CLEANUP'}
        && $Options_hash->{'PRE_TEST_CLEANUP'} eq 'defined') {
        $self->filer_cleanup_defined();
        my @targets = (
            'aggregates',    'cifs',   'interfaces', 'licenses',
            'qtrees',        'quotas', 'shares',     'vfilers',
            'vfiler_limits', 'volumes'
        );
        $self->discover_topology('targets' => \@targets);
    }
    foreach my $node (@{$Options_hash->{'NODES'}}) {
        eval { $self->time_setup('node' => $node); };
    }

    #TODO if syncmirror is enabled, create it
    if ($Options_hash->{'AGGR_SYNCMIRROR'}) {
        my @aggregates = split(/,/, $Options_hash->{'AGGREGATES'});
        $self->create_syncmirror('aggregate' => $aggregates[0]);
    }
    foreach my $node (@{$Options_hash->{'NODES'}}) {
        $Log->comment("Configuring $node");

        my $Cli  = $apiset{'cli'}{$node};
        my $Zapi = $apiset{'zapi'}{$node};

        if ($Options_hash->{'FLEX_CLONE_PERCENTAGE'}
            && ($self->{'discovered_topology'}->{$node}->{'licenses'}
                ->{'flexclone'} ne 'licensed')
            ) {
            $Log->comment("Adding flexclone license");
            if ($self->{'release'} =~ /sierranevada/i) {
                eval {
                    $Cli->license_add(
                        "code" => 'GCKYSWDUSAFDAAAAAAAAAAAAAAAA');
                };
            } else {
                eval { $Cli->license_add("code" => 'QNFYRVH'); };
            }
            $self->{'discovered_topology'}->{$node}->{'licenses'}
                ->{'flexclone'} = 'licensed';
        }

        if ($Options_hash->{'FLEX_CACHE_PERCENTAGE'}
            && ($self->{'discovered_topology'}->{$node}->{'licenses'}
                ->{'flexcache'} ne 'licensed')
            ) {
            $Log->comment("Adding flexcache license");
            if ($self->{'release'} !~ /sierranevada/i) {
                eval { $Cli->license_add("code" => 'OSRAREM'); };
            }
            $self->{'discovered_topology'}->{$node}->{'licenses'}
                ->{'flexcache'} = 'licensed';
        }

        my $defined_addresses
            = $self->{'defined_topology'}->{$node}->{'addresses'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};
        my $discovered_interfaces
            = $self->{'discovered_topology'}->{$node}->{'interfaces'};

        if ($Options_hash->{'AUTOSUPPORT'}) {
            $self->enable_autosupport('node' => $node);

        }

        my $vfiler = 'vfiler0';
        if ($Options_hash->{'DESTIP_VERSION'} eq 6) {
            $Log->comment("Enabling IPv6 on $vfiler");

            # Set the vfiler context for the next ZAPI calls
            try {
                $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
            try {

                $Response = $Zapi->options_set(
                    'name'  => 'ip.v6.enable',
                    'value' => 'on'
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
            try {

                $Response = $Zapi->options_set(
                    'name'  => 'ip.v6.ra_enable',
                    'value' => 'on'
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }

        $Log->comment("Configuring data interfaces for $vfiler");
        my @interfaces = keys(%{$defined_vfilers->{$vfiler}->{'interfaces'}});
        foreach my $interface (@interfaces) {
            if (!grep { $_ eq $interface }
                @{$Options_hash->{'DATA_INTERFACES'}}) {
                next;
            }

            $Log->comment("Configuring interface $interface");

            my $defined_addresses_by_interface
                = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};
            foreach
                my $address_information (@{$defined_addresses_by_interface}) {

                my $address = $address_information->{'address'};
                if (!$address) { next; }
                my $range = $address_information->{'range'};
                my $netmask;
                if (!defined($range)) {
                    $netmask = $address_information->{'mask'};
                } else {
                    $netmask = $defined_addresses->{$interface}->{$range}
                        ->{'NETMASK'};
                }

                my ($version, $regex);
                if ($address =~ /\./) {
                    $version = 'ipv4_info';
                    $regex   = $address;
                    $regex =~ s/(\.)/\\$1/g;
                } else {
                    $version = 'ipv6_info';
                    my $ip = new Net::IP($address);
                    $address = $ip->short();
                    $regex   = $address;
                }

                my $created;
                my $retries = 10;
                while (!$created && $retries) {

                    # Check if the interface is configured
                    my $Ifconfig = $self->ifconfig(
                        'node'    => $node,
                        'retries' => 5,
                        'wait'    => 60
                    );

                    # Wait for #$%^&*($#%^& ONTAP
                    snooze 1;

                    foreach my $i (@{$Ifconfig}) {
                        if ($interface eq $i->{'name'}) {

                            #if (grep(/$regex/, @{$i->{$version}})) {
                            if (grep { $_->{'ip'} eq $address }
                                @{$i->{$version}}) {
                                $created = 1;
                                last;
                            } else {
                                $Log->debug("$address not found, retrying\n"
                                        . Dumper($Ifconfig));
                                $retries--;

                                # Wait for #$%^&*($#%^& ONTAP
                                snooze 30;

                            }
                        }
                    }
                    if (!$created) {
                        $self->ifconfig_address(
                            'node'      => $node,
                            'interface' => $interface,
                            'address'   => $address,
                            'netmask'   => $netmask,
                            'retries'   => 5,
                            'wait'      => 60
                        );
                    }
                }
                if (!$retries) {
                    die config_result(
                        "Unable to create alias $address on $interface");
                }

                my $rc_data = $self->read_file(
                    'path' => '/vol/vol0/etc/rc',
                    'node' => $node
                );

                # Remove any aliases using this address
                my $new_rc_data;
                my $found = 0;
                foreach my $line (split(/\n/, $rc_data)) {
                    chomp($line);
                    if ($line !~ /ifconfig $interface alias $regex/) {
                        $new_rc_data .= $line . "\n";
                    }
                }

                # Checking for an existing ifconfig command
                $rc_data = $new_rc_data;
                $found   = 0;
                foreach my $line (split(/\n/, $rc_data)) {
                    chomp($line);
                    if ($line =~ /ifconfig $interface $regex/) {
                        $found = 1;
                        last;
                    }
                }

                # Add the ifconfig command if required
                if (!$found) {
                    $new_rc_data = '';
                    foreach my $line (split(/\n/, $rc_data)) {
                        chomp($line);
                        if ($line =~ /ifconfig/ && $line !~ /alias/) {
                            $new_rc_data .= $line . "\n";
                            if ($address !~ /:/) {
                                $new_rc_data
                                    .= "ifconfig $interface $address netmask $netmask\n";
                            } else {
                                $new_rc_data
                                    .= "ifconfig $interface $address\n";
                            }
                        } else {
                            $new_rc_data .= $line . "\n";
                        }
                    }
                }
                $Log->comment(Dumper($new_rc_data));
                $self->write_file(
                    'ascii' => $new_rc_data,
                    'node'  => $node,
                    'path'  => '/vol/vol0/etc/rc'
                );
                snooze 1;
            }
        }

        if ($Options_hash->{'WAFL_DEFAULT_SECURITY_STYLE'}) {

            # Set the vfiler context for the next ZAPI calls
            try {
                $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };

            try {

                $Response = $Zapi->options_set(
                    'name'  => 'wafl.default_security_style',
                    'value' => $Options_hash->{'WAFL_DEFAULT_SECURITY_STYLE'}
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }

        # Set the vfiler context for the next ZAPI calls
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        # Create the volumes that don't already exist
        foreach my $volume (natsort(keys(%{$defined_volumes}))) {
            if (grep(/^$volume$/, keys(%{$discovered_volumes}))) {
                next;
            }

            my $aggregate = $defined_volumes->{$volume}->{'aggregate'};
            if (!$aggregate) {
                die script_result("Aggregate is undef for for volume "
                        . "$volume in defined_topology\n"
                        . Dumper($defined_volumes->{$volume}));
            }
            my $size = $defined_volumes->{$volume}->{'size'} . 'm';
            if (!$size) {
                die script_result("Size is undef for for volume "
                        . "$volume in defined_topology\n"
                        . Dumper($defined_volumes->{$volume}));
            }
            my $guar_type = $defined_volumes->{$volume}->{'guar_type'};
            if (!$guar_type) {
                die script_result("Guarantee Type is undef for for volume "
                        . "$volume in defined_topology\n"
                        . Dumper($defined_volumes->{$volume}));
            }
            my $type = $defined_volumes->{$volume}->{'type'};
            if (!$type) {
                die script_result("Volume Type is undef for for volume "
                        . "$volume in defined_topology\n"
                        . Dumper($defined_volumes->{$volume}));
            }
            my $extent = $defined_volumes->{$volume}->{'extent'};
            if ($type eq 'flexvol') {
                try {
                    $Log->comment("Creating volume $volume");
                    $Response = $Zapi->volume_create(
                        'volume'               => $volume,
                        'containing-aggr-name' => $aggregate,
                        'size'                 => $size,
                        'space-reserve'        => $guar_type
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            } elsif ($type eq 'clone') {
                my $parent_vol = $defined_volumes->{$volume}->{'parent_vol'};
                if (!$parent_vol) {
                    die script_result("Parent Vol is undef for clone vol "
                            . "$volume in defined_topology\n"
                            . Dumper($defined_volumes->{$volume}));
                }
                try {

                    $Log->comment(
                        "Creating clone volume $volume from parent vol $parent_vol"
                    );
                    $Response = $Zapi->volume_clone_create(
                        'volume'        => $volume,
                        'parent-volume' => $parent_vol,
                        'space-reserve' => $guar_type
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }
            if ($extent =~ /on/i) {
                $Log->comment("Enabling extent on vol $volume");
                try {
                    $Response = $Zapi->volume_set_option(
                        'volume'       => $volume,
                        'option-name'  => 'extent',
                        'option-value' => 'on'
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }

            #create qtrees if needed
            if ($defined_volumes->{$volume}->{'qtree_count'} > 0) {
                $self->create_qtrees(
                    'volume' => $volume,
                    'node'   => $node
                );
            }
        }

        # Create the vfilers if they don't already exist
        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            if ($vfiler eq 'vfiler0') { next; }

            if (!grep { $_ eq $vfiler } keys(%{$discovered_vfilers})) {

                # Create the vfiler
                $self->create_vfiler(
                    'node'    => $node,
                    'vfiler'  => $vfiler,
                    'retries' => 5,
                    'wait'    => 60
                );
                snooze 12;

            } else {
                $Log->comment("Skipping creation of vfiler $vfiler "
                        . "because it already exists");
                $Log->comment("Adding storage to vfiler $vfiler");

                # Add volumes to the vfiler becasue it's already created
                my $storage;
                my @volumes = @{$defined_vfilers->{$vfiler}->{'volumes'}};
                @volumes = natsort(@volumes);
                foreach my $volume (@volumes) {
                    $Log->debug(
                        "Checking if volume $volume is owned by vfiler $vfiler"
                            . Dumper($discovered_vfilers->{$vfiler}->{'path'})
                    );
                    if (grep { $_ eq "/vol/$volume" }
                        @{$discovered_vfilers->{$vfiler}->{'path'}}) {
                        next;
                    }
                    $storage .= "/vol/$volume ";
                }

                if ($storage) {
                    $Log->comment("Adding volumes to vfiler $vfiler");
                    $Log->debug("Volumes: $storage");
                    $self->vfiler_add(
                        'node'    => $node,
                        'vfiler'  => $vfiler,
                        'storage' => $storage,
                        'retries' => 5,
                        'wait'    => 60
                    );

                }
            }
        }

        #create flexcache volumes after origins have been added to vfilers
        if ($self->{'flex_cache_count'} > 0) {
            $self->enable_vfiler_flexcache('node' => $node);

            foreach my $volume (keys(%{$defined_volumes})) {
                if (grep(/^$volume$/, keys(%{$discovered_volumes}))) {
                    next;
                }

                my $type = $defined_volumes->{$volume}->{'type'};
                if (!$type) {
                    die script_result("Volume Type is undef for for volume "
                            . "$volume in defined_topology\n"
                            . Dumper($defined_volumes->{$volume}));
                }
                if ($type eq 'cache') {
                    my $origin_vol
                        = $defined_volumes->{$volume}->{'origin_vol'};
                    if (!$origin_vol) {
                        die script_result("Origin Vol is undef for cache vol "
                                . "$volume in defined_topology\n"
                                . Dumper($defined_volumes->{$volume}));
                    }
                    my $aggregate
                        = $defined_volumes->{$volume}->{'aggregate'};
                    if (!$aggregate) {
                        die script_result("Aggregate is undef for for volume "
                                . "$volume in defined_topology\n"
                                . Dumper($defined_volumes->{$volume}));
                    }
                    my $size = $defined_volumes->{$volume}->{'size'} . 'm';
                    if (!$size) {
                        die script_result("Size is undef for for volume "
                                . "$volume in defined_topology\n"
                                . Dumper($defined_volumes->{$volume}));
                    }
                    my $guar_type
                        = $defined_volumes->{$volume}->{'guar_type'};
                    if (!$guar_type) {
                        die script_result(
                                  "Guarantee Type is undef for for volume "
                                . "$volume in defined_topology\n"
                                . Dumper($defined_volumes->{$volume}));
                    }
                    try {

                        $Log->comment(
                            "Creating cache volume $volume from origin vol $origin_vol"
                        );
                        $Response = $Zapi->volume_create(
                            'volume'               => $volume,
                            'containing-aggr-name' => $aggregate,
                            'remote-location'      => $origin_vol,
                            'size'                 => $size,
                        );
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }
            }
        }

        # setup thin provisioning if desired
        if ($Options_hash->{'THIN_PROVISIONING'} =~ /true/i) {
            foreach my $volume (keys(%{$defined_volumes})) {
                my $type = $defined_volumes->{$volume}->{'type'};
                if ($type eq 'cache') {

                    #if volume is a flexcache, enable flexcache_autogrow
                    $Log->comment(
                        "Enabling flexcache autogrow volume $volume");
                    try {
                        $Response = $Cli->execute_raw_command(command =>
                                "vol options $volume flexcache_autogrow on");
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };

                } else {

        #if volume is not a flexcache, enable snapshot autodelete and autosize
                    try {
                        $Log->comment(
                            "Enabling snapshot autodelete and autosize on volume $volume"
                        );
                        $Response = $Zapi->snapshot_autodelete_set_option(
                            'volume'       => $volume,
                            'option-name'  => 'state',
                            'option-value' => 'on',
                        );
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                    try {
                        if ($self->{'release'} =~ /sierranevada/i) {
                            $Response
                                = $Cli->execute_raw_command(
                                command => "vol autosize $volume "
                                    . " -grow-threshold-percent 80 -shrink-threshold-percent 20 grow_shrink"
                                );
                        } else {
                            $Response = $Cli->execute_raw_command(
                                command => "vol autosize $volume on");
                        }

                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }
            }
        }

        # enable dedupe and compression if specified
        if (   ($Options_hash->{'DEDUPE_PERCENT'} > 0)
            || ($Options_hash->{'DEDUPE_COMPRESSION_PERCENT'} > 0)
            || ($Options_hash->{'DEDUPE_INLINE_PERCENT'} > 0)) {
            foreach my $volume (keys(%{$defined_volumes})) {
                my $sis_mode = $defined_volumes->{$volume}->{'sis'};
                if ($sis_mode =~ /dedupe|compress|inline/i) {
                    my $path = "/vol/" . $volume;
                    try {
                        $Log->comment("Enabling dedupe on volume $volume");
                        $Response = $Zapi->sis_enable('path' => $path,);
                        $Response = $Zapi->sis_set_config(
                            'path'     => $path,
                            'schedule' => '0-23@sun-sat',
                        );
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }
                if ($sis_mode =~ /inline/i) {
                    my $path = "/vol/" . $volume;
                    try {
                        $Log->comment(
                            "Enabling dedupe+ inline compress on volume $volume"
                        );
                        if ($self->{'release'} =~ /sierranevada/i) {
                            $Response = $Zapi->sis_set_config(
                                'path'                      => $path,
                                'enable-compression'        => 'true',
                                'enable-idd'                => 'true',
                                'enable-inline-compression' => 'true'
                            );
                        } else {
                            $Response = $Zapi->sis_set_config(
                                'path'                      => $path,
                                'enable-compression'        => 'true',
                                'enable-inline-compression' => 'true'
                            );
                        }
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }
                if ($sis_mode =~ /compress/i) {
                    my $path = "/vol/" . $volume;
                    try {
                        $Log->comment(
                            "Enabling dedupe+compress on volume $volume");
                        $Response = $Zapi->sis_set_config(
                            'path'               => $path,
                            'enable-compression' => 'true',
                        );
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }
            }
        }

        #enable wafl scanners if specified
        if (($Options_hash->{'WAFL_SCANNER_ENABLE'} =~ /true/i)) {
            $Log->comment("Enabling disk scrub on all aggrs");
            try {
                $Response = $Zapi->aggr_scrub_start();
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
            my $discovered_aggrs
                = $self->{'discovered_topology'}->{$node}->{'aggregates'};
            foreach my $aggr (keys(%{$discovered_aggrs})) {
                $Log->comment("Enabling reallocate on aggr $aggr");
                try {
                    $Response = $Zapi->reallocate_start(
                        'path'         => $aggr,
                        'interval'     => '5m',
                        'is-aggregate' => 'true'
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    if ($Exception->text() =~ /scan already exists/i) {
                        $Log->comment("reallocate is already configured");
                    } else {
                        die config_result($Exception->text());
                    }
                };
                $Log->comment("Enabling free space reallocate on aggr $aggr");
                try {
                    $Response = $Zapi->aggr_set_option(
                        'aggregate'    => $aggr,
                        'option-name'  => 'free_space_realloc',
                        'option-value' => 'on'
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    if ($Exception->text() =~ /scan already exists/i) {
                        $Log->comment("reallocate is already configured");
                    } else {
                        die config_result($Exception->text());
                    }
                };
            }
        }

        my @nodes = ($node);
        my @targets = ('vfilers', 'licenses');
        $self->discover_topology(
            'nodes'   => \@nodes,
            'targets' => \@targets
        );

        # Configure cifs.ipv6_enable
        $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        if ($self->{'release'} !~ /boilermaker/) {

            foreach my $vfiler (@vfilers) {

                # Get vfiler status
                $Log->comment("Configuring cifs.ipv6.enable for $vfiler");
                my $status = $discovered_vfilers->{$vfiler}->{'status'};
                $Log->debug("Status: $status\n"
                        . Dumper($discovered_vfilers->{$vfiler}));

                if ($status ne 'running') {
                    $Log->comment(
                        "Skipping vfiler $vfiler becasue it is not running");
                    next;
                }

                # Set the vfiler context for the next ZAPI calls
                try {
                    $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                my $cifs_ipv6_enabled = 'off';
                if ($Options_hash->{'DESTIP_VERSION'} eq 6) {
                    $cifs_ipv6_enabled = 'on';
                }

                $Log->comment("Setting IPv6 for CIFS to $cifs_ipv6_enabled "
                        . "for vfiler $vfiler");
                try {

                    $Response = $Zapi->options_set(
                        'name'  => 'cifs.ipv6.enable',
                        'value' => $cifs_ipv6_enabled
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }
        }

        @nodes = ($node);
        @targets = ('vfilers', 'licenses');
        $self->discover_topology(
            'nodes'   => \@nodes,
            'targets' => \@targets
        );

        # Configure the vfiler if running
        my %mapped;
        my $status;
        foreach my $vfiler (@vfilers) {

            # Get the current vfilers status
            my $discovered_vfilers
                = $self->{'discovered_topology'}->{$node}->{'vfilers'};
            $status = $discovered_vfilers->{$vfiler}->{'status'};

            # Get the current vfilers path to etc
            my $path
                = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
            if ($path eq '/etc') { $path = '/vol/vol0/etc'; }

            if ($status eq 'running' || $vfiler eq 'vfiler0') {

                # Check for resolv.conf
                eval {

                    $Response = $Zapi->file_get_file_info(
                        'path' => "$path/resolv.conf");
                };

                # if resolv.conf exists DNS is configured
                if ($@) {
                    $self->configure_dns(
                        'node'    => $node,
                        'vfiler'  => $vfiler,
                        'retries' => 5,
                        'wait'    => 60
                    );
                    snooze 1;
                }

                if (   $Options_hash->{'NIS_DOMAINNAME'}
                    && $Options_hash->{'NIS_SERVERS'}) {
                    $self->configure_nis(
                        'node'   => $node,
                        'vfiler' => $vfiler,
                    );
                    snooze 1;
                }

                # If CIFS option specified, enable it
                if ($Options_hash->{'CIFS_ENABLED'}) {
                    eval {

                        $Response = $Zapi->file_get_file_info(
                            'path' => "$path/cifsconfig_share.cfg");
                    };

                    # if cifsconfig_share.cfg exists CIFS is configured
                    if ($@) {
                        $self->configure_cifs(
                            'node'    => $node,
                            'vfiler'  => $vfiler,
                            'retries' => 5,
                            'wait'    => 60
                        );
                        snooze 5;
                    } else {

                        my $Status;
                        try {
                            $Response = undef;
                            $Response = $Zapi->cifs_status();
                            $Status   = $Response->get_parsed_output();
                        }
                        catch NACL::APISet::Exceptions::ResponseException
                            with {
                            my $Exception = shift;
                            die config_result($Exception->text());
                            };
                        if (grep(/stopped/, @{$Status->[0]->{'status'}})) {

                            # Start CIFS
                            try {

                                $Response = $Zapi->cifs_start();
                                $Status   = $Response->get_parsed_output();

                            }
                            catch NACL::APISet::Exceptions::ResponseException
                                with {
                                my $Exception = shift;
                                die config_result($Exception->text());
                                };
                            snooze 5;
                        }
                    }

                #                    if ($vfiler ne 'vfiler0') {
                #                        my $regex = "^s_v_$vfiler";
                #                        $regex =~ s/-/_/g;
                #
                #                        # Create the root volume shares first
                #                        $self->create_shares(
                #                            'node'   => $node,
                #                            'vfiler' => $vfiler,
                #                            'regex'  => $regex,
                #                            'mapped' => \%mapped
                #                        );
                #                    }
                #
                #                    # Create the data volume shares first
                #                    my $regex = '^s_v_\d+';
                # Create the volume shares
                    my $regex = '^s_' . $volume_basename . '.+';
                    $self->create_shares(
                        'node'   => $node,
                        'vfiler' => $vfiler,
                        'regex'  => $regex,
                        'mapped' => \%mapped
                    );
                }

            }    # end if ($status eq 'running' || $vfiler eq 'vfiler0')
        }    #end foreach my $vfiler (@vfilers)

        snooze 60;

        # Create any additional shares (volume shares required)
        if ($Options_hash->{'CIFS_ENABLED'}) {
            foreach my $vfiler (@vfilers) {

                my $regex = '^s_.+';
                $self->create_shares(
                    'node'   => $node,
                    'vfiler' => $vfiler,
                    'regex'  => $regex,
                    'mapped' => \%mapped
                );
                snooze 1;
            }

            # Populate local groups if specified
            if ($Options_hash->{'LOCAL_GROUP_COUNT'}) {
                my $group         = $Options_hash->{'TEST_GROUP_BASENAME'};
                my $gids          = $self->{'local_groups_per_vfiler'};
                my $group_members = $Options_hash->{'LOCAL_GROUP_MEMBERS'};

                foreach my $vfiler (@vfilers) {
                    $self->create_local_groups(
                        'node'                 => $node,
                        'vfiler'               => $vfiler,
                        'group_basename'       => $group,
                        'starting_group_index' => 1,
                        'ending_group_index'   => $gids,
                        'group_members'        => $group_members
                    );
                }
            }
        }

        # Create the volume quotas
        if ($Options_hash->{'QUOTAS'}) {
            foreach my $vfiler (@vfilers) {
                $self->configure_quotas(
                    'node'   => $node,
                    'vfiler' => $vfiler
                );
            }
        }

        # Setup nfs exports of the volumes in the vfilers.
        if ($Options_hash->{'NFS_SETUP'}) {
            foreach my $vfiler (@vfilers) {
                $self->configure_nfs(
                    'node'    => $node,
                    'vfiler'  => $vfiler,
                    'retries' => 5,
                    'wait'    => 30
                );
            }
        }

        # Set the vfiler context for the next ZAPI calls
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        #Configure SMB 2 if needed.
        #We need to do this for before we turn on branchcache.
        if ($Options_hash->{'SMB2'}) {
            foreach my $vfiler (@vfilers) {
                $self->enable_smb2(
                    'node'   => $node,
                    'vfiler' => $vfiler
                );
            }
        }

        #Configure Branch Cache if needed
        if ($Options_hash->{'BRANCH_CACHE'}) {
            foreach my $vfiler (@vfilers) {
                $self->enable_branch_cache(
                    'node'   => $node,
                    'vfiler' => $vfiler
                );
            }
        }

        # Create the spn for the cifs server.
        if ($Options_hash->{'KERBEROS'}) {
            if (   $Options_hash->{'PRE_TEST_CLEANUP'}
                && $Options_hash->{'PRE_TEST_CLEANUP'} ne 'disabled'
                && $Options_hash->{'CIFS_SERVER_ACCESS_MODE'} ne 'ip') {
                $self->create_spn();

            }
        }
    }

    $Log->exit();

} ## end sub filer_setup

sub filer_cleanup {

    $Log->enter();

    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::filer_cleanup(%options);

    $Options_hash = $Options->get_hash();

    if (!$Options_hash->{'POST_FAILURE_CLEANUP'}) {
        $Log->comment(
            "Skipping filer_cleanup STAGE because POST_FAILURE_CLEANUP=0");
        return;
    }

    $self->filer_cleanup_defined();

    $Log->exit();
} ## end sub filer_cleanup

sub DESTROY {

    $Log->exit();

    my $self = shift;

    local ($@);

    $self->client_cleanup();

    $Log->exit();
} ## end sub DESTROY

sub discover_topology {

    $Log->enter();

    my ($self, %options) = @_;
    my @targets = @{$options{'targets'}};
    my @nodes = @{$options{'nodes'}} if $options{'nodes'};

    # Discover all existing vfilers and volumes
    foreach my $node (@{$Options_hash->{'NODES'}}) {

        if (@nodes && !grep { $_ eq $node } @nodes) {
            next;
        }

        my $Cli  = $apiset{'cli'}{$node};
        my $Zapi = $apiset{'zapi'}{$node};

        my $Host = host($node);

        # check to ensure Zapi is enabled, if not, try to enable it
        try {
            $Response = $Zapi->system_get_version();
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;

            # if zapi connect fails to find socket, attempt to enable it
            if ($Exception->text() =~ /cannot connect to socket/) {
                $Log->comment("Enabling Zapi");
                $Response = $Cli->options('httpd.admin.enable' => 'on');
            } else {
                die config_result($Exception->text());
            }
        };

        if (grep { $_ eq 'licenses' } @targets) {
            $Log->comment("Discovering licenses on $node");
            delete($self->{'discovered_topology'}->{$node}->{'licenses'});
            try {
                if ($self->{'release'} =~ /sierranevada/i) {
                    $Response = $Zapi->license_v2_list_info();
                    my $Licenses     = $Response->get_parsed_output();
                    my $License_list = $Licenses->[0]->{'licenses'}->[0]
                        ->{'license-v2-info'};
                    foreach my $lic (@$License_list) {
                        $self->{'discovered_topology'}->{$node}->{'licenses'}
                            ->{lc($lic->{'package'})} = 'Licensed';
                    }
                    if (!exists(
                            $self->{'discovered_topology'}->{$node}
                                ->{'licenses'}->{'cifs'}
                        )
                        ) {
                        $self->{'discovered_topology'}->{$node}->{'licenses'}
                            ->{'cifs'} = "Not Licensed";
                    }
                } else {
                    $Response = $Cli->license();
                    my $Licenses = $Response->get_parsed_output();
                    foreach my $license (@{$Licenses}) {
                        foreach my $key (keys(%{$license})) {
                            $self->{'discovered_topology'}->{$node}
                                ->{'licenses'}->{$key} = $license->{$key};
                        }
                    }
                }
                snooze 1;

            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                my $Response  = $Exception->get_response_object();
                $Log->warn(
                    $Exception->text() . ": " . $Response->get_raw_output());
            }

        }

        my $discover_multistore = 1;

        # Get the multistore license status
        my $multistore_license = 'not licensed';
        if (keys(%{$self->{'discovered_topology'}->{$node}->{'licenses'}})) {
            $multistore_license
                = $self->{'discovered_topology'}->{$node}->{'licenses'}
                ->{'multistore'};
        }

        if (($multistore_license =~ /not licensed/i)
            || !exists(
                $self->{'discovered_topology'}->{$node}->{'licenses'}
                    ->{'multistore'}
            )
            ) {
            $Log->comment("Adding vfiler license");
            try {

                #multistore doesnt need license in SN
                if ($self->{'release'} !~ /sierranevada/) {
                    $Response = $Cli->license_add("code" => 'SPQLGBN');
                }
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                my $Response  = $Exception->get_response_object();
                die config_result(
                    $Exception->text() . ": " . $Response->get_raw_output());
            };
            $self->{'discovered_topology'}->{$node}->{'licenses'}
                ->{'multistore'} = "licensed";
            snooze 1;
        }

        # Get licensed_feature.multistore.enable
        my $multistore_enable;

        # Set the vfiler context for the next ZAPI calls
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        eval {
            $Response
                = $Zapi->options_get(
                'name' => 'licensed_feature.multistore.enable');
            my $Return = $Response->get_parsed_output();

            # Store its value
            $self->{'discovered_topology'}->{$node}->{'options'}
                ->{'licensed_feature.multistore.enable'}
                = $Return->[0]->{'value'};
            $multistore_enable = $Return->[0]->{'value'};
            $Log->comment(
                "licensed_feature.multistore.enable: $multistore_enable");

        };

        # Is the licensed_feature.multistore.enable option is valid for this
        # realease/platform?
        my $option_valid = $self->is_option_valid(
            'node'   => $node,
            'option' => 'licensed_feature.multistore.enable'
        );

     # If multistore isnt enabled, enable it so that discovery works correctly
        if ($option_valid && $multistore_enable ne 'on') {
            $Log->comment("enabling multistore");
            try {
                $Response = $Zapi->options_set(
                    'name'  => 'licensed_feature.multistore.enable',
                    'value' => 'on'
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                my $Response  = $Exception->get_response_object();
                $Log->warn(
                    $Exception->text() . ": " . $Response->get_raw_output());
            };
        }

        if ($discover_multistore && grep { $_ eq 'vfiler_limits' } @targets) {
            $Log->comment("Discovering vfiler limits on $node");
            try {

                $Response = $Cli->vfiler_limit();
                snooze 1;
                my $Vfiler_limits = $Response->get_parsed_output();
                foreach my $limit (natsort(keys(%{$Vfiler_limits->[0]}))) {
                    $self->{'discovered_topology'}->{$node}->{'vfiler_limits'}
                        ->{$limit} = $Vfiler_limits->[0]->{$limit};
                }
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                my $Response  = $Exception->get_response_object();
                $Log->warn(
                    $Exception->text() . ": " . $Response->get_raw_output());
            };
        }

        if ($discover_multistore && grep { $_ eq 'vfilers' } @targets) {
            delete($self->{'discovered_topology'}->{$node}->{'vfilers'});
            my $retry   = 1;
            my $retries = 5;
            while ($retry) {
                try {
                    $Log->comment("Discovering vfilers on $node");
                    $Response = $Cli->vfiler_status("all-details" => 1);
                    snooze 1;
                    my $Vfilers = $Response->get_parsed_output();
                    foreach my $vfiler (@{$Vfilers}) {
                        $self->{'discovered_topology'}->{$node}->{'vfilers'}
                            ->{$vfiler->{'vfiler'}} = $vfiler;
                    }
                    $retry = 0;
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    my $Response  = $Exception->get_response_object();
                    my $output    = $Response->get_raw_output();
                    if ($output =~ /ifconfig cannot be run/) {
                        $retry++;
                        snooze 60;
                    } else {
                        die config_result(
                            $Exception->text() . ": " . $output);
                    }
                };
                if ($retry == $retries) {
                    die config_result(
                        "Unable to get ifconfig information after $retries tries"
                    );
                }
            }
        } else {
            $self->{'discovered_topology'}->{$node}->{'vfilers'} = {};
        }

        if (!grep(/vfiler0/,
                keys(
                    %{  $self->{'discovered_topology'}->{$node}->{'vfilers'}
                        }
                    ))
            ) {
            $self->{'discovered_topology'}->{$node}->{'vfilers'}->{'vfiler0'}
                = {};
        }

        if (grep { $_ eq 'interfaces' } @targets) {

            # Get the current interface config
            $Log->comment("Discovering interfaces on $node");
            delete($self->{'discovered_topology'}->{$node}->{'interfaces'});

            my $retry   = 1;
            my $retries = 5;
            while ($retry) {
                try {

                    $Response = $Cli->ifconfig(all => 1);
                    my $Ifconfig = $Response->get_parsed_output();
                    foreach my $int (@{$Ifconfig}) {
                        $self->{'discovered_topology'}->{$node}
                            ->{'interfaces'}->{$int->{'name'}} = $int;
                        $Log->comment("found $int->{'name'}");
                    }
                    $retry = 0;
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    my $Response  = $Exception->get_response_object();
                    my $output    = $Response->get_raw_output();
                    if ($output =~ /ifconfig cannot be run/) {
                        $retry++;
                        snooze 60;
                    } else {
                        die config_result(
                            $Exception->text() . ": " . $output);
                    }
                };
                if ($retry == $retries) {
                    die config_result(
                        "Unable to get ifconfig information after $retries tries"
                    );
                }
            }
        }

        if (grep { $_ eq 'aggregates' } @targets) {
            $Log->comment("Discovering aggregates on $node");
            delete($self->{'discovered_topology'}->{$node}->{'aggregates'});
            my @fields = ('total', 'used', 'avail', 'capacity');
            my $Aggregates;
            try {

                $Response   = $Cli->aggr_status();
                $Aggregates = $Response->get_parsed_output();
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                my $Response  = $Exception->get_response_object();
                die config_result(
                    $Exception->text() . ": " . $Response->get_raw_output());
            };

            if (@{$Aggregates}) {
                try {

                    $Response = $Cli->df(
                        aggregate         => 1,
                        'scale-megabytes' => 1
                    );
                    snooze 1;
                    my @Output = @{$Response->get_parsed_output()};
                    my @fields = ('total', 'used', 'avail', 'capacity');
                    foreach my $hash (@Output) {
                        for (my $i = 0; $i < @{$Aggregates}; $i++) {
                            my $aggregate = ${$Aggregates}[$i];
                            if ($hash->{'aggregate'} eq
                                $aggregate->{'aggregate'}) {
                                foreach my $field (@fields) {
                                    $aggregate->{$field} = $hash->{$field};
                                }
                                @{$Aggregates}[$i] = $aggregate;
                                last;
                            }
                        }
                    }
                    for (my $i = 0; $i < @{$Aggregates}; $i++) {
                        my $aggregate = ${$Aggregates}[$i];
                        $self->{'discovered_topology'}->{$node}
                            ->{'aggregates'}->{$aggregate->{'aggregate'}}
                            = $aggregate;
                    }
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    my $Response  = $Exception->get_response_object();
                    die config_result($Exception->text() . ": "
                            . $Response->get_raw_output());
                };
            }
        }
        if (grep { $_ eq 'volumes' } @targets) {
            $Log->comment("Discovering volumes on $node");
            delete($self->{'discovered_topology'}->{$node}->{'volumes'});
            try {

                $Response = $Zapi->volume_list_info_iter_start();
                my $Return  = $Response->get_parsed_output();
                my $records = ${$Return}[0]->{'records'};
                my $tag     = ${$Return}[0]->{'tag'};
                $Log->comment("$records volumes discovered");

                $Response = $Zapi->volume_list_info_iter_next(
                    'tag'     => $tag,
                    'maximum' => $records
                );
                my $Volumes = $Response->get_parsed_output('tag' => $tag);
                my @volume_info
                    = @{$Volumes->[0]->{'volumes'}->[0]->{'volume-info'}};

                for (my $i = 0; $i < @volume_info; $i++) {
                    my $volume = $volume_info[$i];
                    $self->{'discovered_topology'}->{$node}->{'volumes'}
                        ->{$volume->{'name'}} = $volume;
                }

                $Response = $Zapi->volume_list_info_iter_end('tag' => $tag);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }
        if (grep { $_ eq 'qtrees' } @targets) {

            # Get the current qtree config
            $Log->comment("Discovering qtrees on $node");
            delete($self->{'discovered_topology'}->{$node}->{'qtrees'});
            try {

                $Response = $Zapi->qtree_list_iter_start();
                my $Return  = $Response->get_parsed_output();
                my $records = ${$Return}[0]->{'records'};
                my $tag     = ${$Return}[0]->{'tag'};

                $Response = $Zapi->qtree_list_iter_next(
                    'tag'     => $tag,
                    'maximum' => $records
                );
                my $Qtrees = $Response->get_parsed_output();
                my @qtree_info
                    = @{$Qtrees->[0]->{'qtrees'}->[0]->{'qtree-info'}};

                for (my $i = 0; $i < @qtree_info; $i++) {
                    my $qtree = $qtree_info[$i];
                    if ($qtree->{'id'} > 0) {
                        $self->{'discovered_topology'}->{$node}->{'qtrees'}
                            ->{$qtree->{'qtree'}} = $qtree;
                    }
                }

                $Response = $Zapi->qtree_list_iter_end('tag' => $tag);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }

        if (grep { $_ eq 'quotas' } @targets) {
            snooze 1;
            $Log->comment("Discovering quotas on $node");
            my $discovered_vfilers
                = $self->{'discovered_topology'}->{$node}->{'vfilers'};
            my $discovered_volumes
                = $self->{'discovered_topology'}->{$node}->{'volumes'};

            foreach my $vfiler (keys(%{$discovered_vfilers})) {

                $Log->comment("Discovering quotas owned by vfiler $vfiler");

                # Don't discover information on stopped vfilers
                if ($discovered_vfilers->{$vfiler}->{'status'} ne 'running') {
                    next;
                }

                # Set the vfiler context for the next ZAPI calls
                try {
                    $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                my $path
                    = "$discovered_vfilers->{$vfiler}->{'configuration_directory'}";
                if ($path eq '/etc') { $path = '/vol/vol0/etc'; }
                $path .= '/quotas';

                eval {
                    $Response = $Zapi->file_get_file_info('path' => $path);
                };

                if ($@) {
                    $self->write_file(
                        'ascii' => "\n",
                        'node'  => $node,
                        'path'  => $path
                    );
                    snooze 1;

                }

                # Get the current quota config
                my ($Return, $records, $tag);

                try {
                    $Response = $Zapi->quota_list_entries_iter_start();
                    $Return   = $Response->get_parsed_output();
                    $records  = ${$Return}[0]->{'records'};
                    $tag      = ${$Return}[0]->{'tag'};
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                if ($records) {
                    try {
                        $Response = $Zapi->quota_list_entries_iter_next(
                            'tag'     => $tag,
                            'maximum' => $records
                        );
                        my $Quotas
                            = $Response->get_parsed_output('tag' => $tag);
                        if ($Quotas->[0]->{'records'}) {

                            my @quota_info
                                = @{$Quotas->[0]->{'quota-entries'}->[0]
                                    ->{'quota-entry'}};

                            for (my $i = 0; $i < @quota_info; $i++) {
                                my $quota = $quota_info[$i];
                                push(
                                    @{  $self->{'discovered_topology'}
                                            ->{$node}->{'volumes'}
                                            ->{$quota->{'volume'}}->{'quotas'}
                                        },
                                    $quota
                                );

                            }
                        }
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };

                    try {

                        $Response = $Zapi->quota_list_entries_iter_end(
                            'tag' => $tag);
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }
            }
        }

        if (grep { $_ eq 'cifs' } @targets) {
            $Log->comment("Discovering CIFS on $node");
            delete($self->{'discovered_topology'}->{$node}->{'shares'});
            my $discovered_vfilers
                = $self->{'discovered_topology'}->{$node}->{'vfilers'};

            foreach my $vfiler (keys(%{$discovered_vfilers})) {
                $Log->comment("Discovering CIFS on vfiler $vfiler");

                # Don't discover information on stopped vfilers
                if ($discovered_vfilers->{$vfiler}->{'status'} ne 'running') {
                    next;
                }

                # Set the vfiler context for the next ZAPI calls

                try {
                    $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                # Check if CIFS is configured
                eval { $Response = $Zapi->cifs_list_config(); };

                if (!$@) {

                    # Set the vfiler context for the next ZAPI calls

                    my $original_cifs_status;
                    my $Status;
                    try {
                        $Response = undef;
                        $Response = $Zapi->cifs_status();
                        $Status   = $Response->get_parsed_output();
                        $original_cifs_status = $Status;

                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };

                    if (grep(/stopped/, @{$Status->[0]->{'status'}})) {

                        # Start CIFS
                        try {

                            $Response = $Zapi->cifs_start();
                            $Status   = $Response->get_parsed_output();

                        }
                        catch NACL::APISet::Exceptions::ResponseException
                            with {
                            my $Exception = shift;
                            die config_result($Exception->text());
                            };
                        snooze 5;

                        # get the CIFS status
                        try {

                            $Response = $Zapi->cifs_status();
                            $Status   = $Response->get_parsed_output();

                        }
                        catch NACL::APISet::Exceptions::ResponseException
                            with {
                            my $Exception = shift;
                            die config_result($Exception->text());
                            };
                    }

                    if (!grep(/started/, @{$Status->[0]->{'status'}})) {
                        next;
                    }

                    if (grep { $_ eq 'shares' } @targets) {
                        $Log->comment(
                            "Discovering CIFS shares on node $node vfiler $vfiler"
                        );
                        my $discovered_vfilers
                            = $self->{'discovered_topology'}->{$node}
                            ->{'vfilers'};
                        my $discovered_volumes
                            = $self->{'discovered_topology'}->{$node}
                            ->{'volumes'};
                        my @vfilers = keys(%{$discovered_vfilers});
                        @vfilers = natsort(@vfilers);

                        # Get the current share config
                        my ($Return, $records, $tag);
                        try {

                            $Response = $Zapi->cifs_share_list_iter_start(
                                'zapi-timeout' => 3600);
                            $Return  = $Response->get_parsed_output();
                            $records = ${$Return}[0]->{'records'};
                            $tag     = ${$Return}[0]->{'tag'};
                        }
                        catch NACL::APISet::Exceptions::ResponseException
                            with {
                            my $Exception = shift;
                            die config_result($Exception->text());
                            };

                        my $max_records_per_call = 10000;
                        while ($records) {

                            my $records_this_call;
                            if ($max_records_per_call >= $records) {
                                $records_this_call = $records;
                            } else {
                                $records_this_call = $max_records_per_call;

                            }
                            $records -= $records_this_call;

                            my $Shares;
                            try {
                                $Response = $Zapi->cifs_share_list_iter_next(
                                    'tag'     => $tag,
                                    'maximum' => $records_this_call
                                );
                                $Shares = $Response->get_parsed_output(
                                    'tag' => $tag);
                            }
                            catch NACL::APISet::Exceptions::ResponseException
                                with {
                                my $Exception = shift;
                                die config_result($Exception->text());
                                };

                            if ($Shares->[0]->{'records'}) {
                                my @share_info
                                    = @{$Shares->[0]->{'cifs-shares'}->[0]
                                        ->{'cifs-share-info'}};

                                for (my $i = 0; $i < @share_info; $i++) {
                                    my $share = $share_info[$i];
                                    if (grep { $_ eq $share->{'share-name'} }
                                        @ontap_shares) {
                                        next;
                                    }
                                    $self->{'discovered_topology'}->{$node}
                                        ->{'shares'}->{$share->{'share-name'}}
                                        = $share;

                                    # Get the owning vfiler
                                    if ($share->{'mount-point'}
                                        =~ /\/vol\/([^\/]+)/) {
                                        my $volume = $1;
                                        if (grep { $_ eq $volume }
                                            keys(%{$discovered_volumes})
                                            ) {
                                            my $vfiler
                                                = $discovered_volumes
                                                ->{$volume}
                                                ->{'owning-vfiler'};
                                            $self->{'discovered_topology'}
                                                ->{$node}->{'shares'}
                                                ->{$share->{'share-name'}}
                                                ->{'owning-vfiler'} = $vfiler;
                                        }

                                    }

                                }
                            }
                            snooze 10;
                        }

                        try {

                            $Response = $Zapi->cifs_share_list_iter_end(
                                'tag' => $tag);
                        }
                        catch NACL::APISet::Exceptions::ResponseException
                            with {
                            my $Exception = shift;
                            die config_result($Exception->text());
                            };

                        if (!grep(/started/,
                                @{$original_cifs_status->[0]->{'status'}})
                            ) {

                            # Stop CIFS
                            try {

                                $Response = $Zapi->cifs_stop();
                                $Status   = $Response->get_parsed_output();

                            }
                            catch NACL::APISet::Exceptions::ResponseException
                                with {
                                my $Exception = shift;
                                die config_result($Exception->text());
                                };
                        }
                    }
                }
            }
        }
    }

    $Log->exit();
} ## end sub discover_topology

sub create_vfiler {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $defined_addresses
        = $self->{'defined_topology'}->{$node}->{'addresses'};
    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $discovered_interfaces
        = $self->{'discovered_topology'}->{$node}->{'interfaces'};

    my @interfaces = keys(%{$defined_vfilers->{$vfiler}->{'interfaces'}});
    $Log->comment("Removing aliases for $vfiler");
    foreach my $interface (@interfaces) {
        my $defined_addresses_by_interface
            = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};
        foreach my $address_information (@{$defined_addresses_by_interface}) {

            my $address = $address_information->{'address'};
            if (!$address) { next; }
            my $range = $address_information->{'range'};
            my $netmask
                = $defined_addresses->{$interface}->{$range}->{'NETMASK'};

            my $address_regex = $address;
            $address_regex =~ s/(\.)/\\$1/g;

            # Delete the alias if needed
            if (grep { $_->{'ip'} eq $address }
                @{$discovered_interfaces->{$interface}->{'ipv4_info'}}) {

                $self->ifconfig_remove_alias(
                    'node'      => $node,
                    'interface' => $interface,
                    'address'   => $address,
                    'retries'   => 5,
                    'wait'      => 60
                );

                # Wait for #$%^&*($#%^& ONTAP
                snooze 1;
            }

        }
    }

    # Create a list of storage devices for the vfiler
    my $storage;
    foreach my $volume (@{$defined_vfilers->{$vfiler}->{'volumes'}}) {
        $storage .= "/vol/$volume ";
    }

    # Create the list of IP addresses
    my $addresses;
    my @ipv4_addresses;
    my @ipv6_addresses;
    foreach my $interface (@interfaces) {
        my $defined_addresses_by_interface
            = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};
        foreach my $address_information (@{$defined_addresses_by_interface}) {
            my $address = $address_information->{'address'};
            if ($address =~ /\./) {
                push(@ipv4_addresses, $address);
            } else {
                push(@ipv6_addresses, $address);
            }
        }
    }

    if (@ipv4_addresses) {
        $addresses .= '-i ' . join(' -i ', @ipv4_addresses);
        $addresses .= ' ';
    }

    if (@ipv6_addresses) {
        $addresses .= '-i ' . join(' -i ', @ipv6_addresses);

    }

    $Log->comment("Creating $vfiler");
    $self->vfiler_create(
        'node'      => $node,
        'vfiler'    => $vfiler,
        'addresses' => $addresses,
        'storage'   => $storage,
        'retries'   => 5,
        'wait'      => 60
    );

    # Wait for #$%^&*($#%^& ONTAP
    snooze 1;

    $Log->comment("Creating aliases for $vfiler");
    foreach my $interface (@interfaces) {
        my $defined_addresses_by_interface
            = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};
        foreach my $address_information (@{$defined_addresses_by_interface}) {

            my $address = $address_information->{'address'};
            my $range   = $address_information->{'range'};
            my $netmask
                = $defined_addresses->{$interface}->{$range}->{'NETMASK'};
            my %ifconfig_options = (
                'node'      => $node,
                'interface' => $interface,
                'address'   => $address,
                'retries'   => 5,
                'wait'      => 60
            );

            my ($version, $regex);
            if ($address =~ /\./) {
                $version = 'ipv4_info';
                $regex   = $address;
                $regex =~ s/(\.)/\\$1/g;
                $ifconfig_options{'netmask'} = $netmask;
            } else {
                $version = 'ipv6_info';
                my $ip = new Net::IP($address);
                $address = $ip->short();
                $regex   = $address;
            }

            my $created;
            my $retries = 10;
            while (!$created && $retries) {
                $self->ifconfig_alias(%ifconfig_options);

                # Wait for #$%^&*($#%^& ONTAP
                snooze 1;

                # Check that the aliases were created per burt 433887
                my $Ifconfig = $self->ifconfig(
                    'node'    => $node,
                    'retries' => 5,
                    'wait'    => 60
                );

                # Wait for #$%^&*($#%^& ONTAP
                snooze 1;

                foreach my $i (@{$Ifconfig}) {
                    if ($interface eq $i->{'name'}) {
                        if (grep { $_->{'ip'} eq $address } @{$i->{$version}})
                        {
                            $created = 1;
                            last;
                        } else {
                            $Log->debug("$address not found, sleeping");
                            $retries--;

                            # Wait for #$%^&*($#%^& ONTAP
                            snooze 60;

                        }
                    }
                }
            }
            if (!$retries) {
                die config_result(
                    "Unable to create alias $address on $interface");
            }

            my $rc_data = $self->read_file(
                'path' => '/vol/vol0/etc/rc',
                'node' => $node
            );

            if ($rc_data !~ /$regex/) {
                if ($address !~ /:/) {
                    $rc_data
                        .= "ifconfig $interface alias $address netmask $netmask\n";
                } else {
                    $rc_data .= "ifconfig $interface alias $address\n";
                }

                $self->write_file(
                    'ascii' => $rc_data,
                    'node'  => $node,
                    'path'  => '/vol/vol0/etc/rc'
                );
                snooze 1;
            }
        }
    }
    $Log->exit();
} ## end sub create_vfiler

sub create_shares {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};
    my $regex  = $options{'regex'};
    my $mapped = $options{'mapped'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $defined_shares  = $topology_database->{$node}->{'shares'};
    my $discovered_qtrees
        = $self->{'discovered_topology'}->{$node}->{'qtrees'};
    my $discovered_shares
        = $self->{'discovered_topology'}->{$node}->{'shares'};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Creating shares on vfiler $vfiler matching $regex");

    my $conn_obj;
    my @clients
        = grep { host($_)->hosttype() =~ /win/ } @{$self->{'clients'}};
    if (@clients) {

        host($clients[0]);
        $conn_obj = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
            '-nologin', -timeout, 300);
        eval { check_connectivity('conn' => $conn_obj); };
        my $error = $@;
        if ($error) { die config_result($error); }
    }

    $Log->debug("vfiler $vfiler\n" . Dumper($defined_vfilers->{$vfiler}));
    foreach my $share (@{$defined_vfilers->{$vfiler}->{'shares'}}) {

        $Log->debug(
            "Checking share $share\n" . Dumper($defined_shares->{$share}));

        # If the share exists skip it
        if (grep { $_ eq $share } keys(%{$discovered_shares})) {
            $Log->debug("Skipping share $share because it allways exists\n"
                    . Dumper($discovered_shares->{$share}));
            next;
        }

        # If filer_setup is not set then we skip share creation in
        # filer_setup
        if (!$defined_shares->{$share}->{'filer_setup'}) {
            $Log->debug("Skipping share $share because filer_setup is not 1");
            next;
        }

        # If this shares volume does not belong to the current vfiler
        # skip it
        my $volume = $defined_shares->{$share}->{'volume'};
        if (!grep(/^$volume$/, @{$defined_vfilers->{$vfiler}->{'volumes'}})) {
            $Log->debug(
                "Skipping share $share because it does not belong to the current vfiler\n"
                    . Dumper($defined_vfilers->{$vfiler}));
            next;
        }

        # These are not the droids we are looking for, move along...
        if ($share !~ /$options{'regex'}/) { next; }

        # If the qtree exists skip its creation
        my $qtree  = $defined_shares->{$share}->{'qtree'};
        my $vfiler = $defined_shares->{$share}->{'vfiler'};
        if (   $qtree
            && !grep(/^$share$/, keys(%{$discovered_qtrees}))
            && $share ne "s_$volume") {
            $Log->comment("Creating qtree $share on $volume");

            try {

                $Response = $Zapi->qtree_create(
                    'qtree'  => $share,
                    'volume' => $volume
                );
                $self->{'defined_topology'}->{$node}->{'qtrees'}->{$share}
                    = {'volume' => $volume, 'vfiler' => $vfiler};
            }

            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
            snooze 1;

        } elsif (!$qtree && $share ne "s_$volume") {

            if (!@clients) {
                die config_result("There are no windows clients "
                        . "to create share directories");
            }

            my @interfaces
                = keys(%{$defined_vfilers->{$vfiler}->{'interfaces'}});
            my $interface_found = 0;
            my %address_indexes;
            foreach my $i (@interfaces) {
                if (grep { $_ eq $i } @{$Options_hash->{'DATA_INTERFACES'}}) {
                    $interface_found = 1;
                    $address_indexes{$i} = 0;
                }
            }

            if (!$interface_found) {
                die config_result("Unable to find an interface for $vfiler "
                        . Dumper($defined_vfilers->{$vfiler}->{'interfaces'})
                );
            }

            my ($server, undef) = $self->assign_cifs_server(
                'node'            => $node,
                'vfiler'          => $vfiler,
                'address_indexes' => \%address_indexes,
                'interface_index' => 0
            );

            if (!${$mapped}{$volume}) {
                map_share(
                    'conn'     => $conn_obj,
                    'share'    => "\\\\$server\\s_$volume",
                    'drive'    => '',
                    'username' => $Options_hash->{'CIFS_USER'},
                    'password' => $Options_hash->{'CIFS_PASSWD'},
                    'timeout'  => 300,
                    'reuse'    => 1,
                    'delay'    => 60,
                    'attempts' => 3,
                );
                ${$mapped}{$volume} = 1;
            }

            my $path = "\\\\$server\\s_$volume\\$share";
            my $cmd  = 'cmd.exe /c if not exist ' . $path . ' mkdir ' . $path;

            my $output;
            eval {
                (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 300);
            };
            my $error = $@;
            if ($output || $error) {
                die config_result("Output:$output\nError:$error");
            }
        }

        my $path = $defined_shares->{$share}->{'path'};

        $Log->comment(
            "Creating share $share on node $node vfiler $vfiler Path:$path");

        # Set the vfiler context for the next ZAPI calls
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        try {

            $Response = $Zapi->cifs_share_add(
                'share-name' => $share,
                'path'       => $path
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result(
                $Exception->text() . "\n" . Dumper($defined_shares));
        };
        snooze 1;

        # Disable the filer_setup flag for this share so multiple
        # calls to create share don't fail
        $topology_database->{$node}->{'shares'}->{$share}->{'filer_setup'}
            = 0;
    }
    $Log->exit();
} ## end sub create_shares

sub configure_cifs {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Configuring CIFS on $vfiler");

    my $cifs_license
        = $self->{'discovered_topology'}->{$node}->{'licenses'}->{'cifs'};
    if ($cifs_license =~ /not licensed/i) {
        $Log->comment("Adding cifs license");
        try {
            if ($self->{'release'} =~ /sierranevada/i) {
                $Response = $Cli->license_add(
                    "code" => 'MJXMPWDUSAFDAAAAAAAAAAAAAAAA');
            } else {
                $Response = $Cli->license_add("code" => 'GTLYRVH');
            }

        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        snooze 1;
    }

    my $Status;
    try {

        $Response = $Zapi->cifs_status();
        $Status   = $Response->get_parsed_output();
        if (grep(/started/, @{$Status->[0]->{'status'}})) {
            $Zapi->cifs_stop();
        }

    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Response = $Zapi->cifs_setup_verify_passwd_and_group();
    $Status   = $Response->get_parsed_output();

    if (!grep(/passed/, @{$Status->[0]->{'status'}})) {
        try {
            $Zapi->cifs_setup_create_group_file();
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
        try {
            $Zapi->cifs_setup_create_passwd_file(
                'default-root-password' => 'netapp1!');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    my $cifs_server
        = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
        ->{'cifs_server'};

    my $user;
    if ($Options_hash->{'CIFS_USER'} =~ /\w+\\(.*)/) {
        $user = $Options_hash->{'CIFS_USER'};
    } else {
        $user = $Options_hash->{'CIFS_USER'} . '\\'
            . $Options_hash->{'CIFS_USER'};
    }

    # Run cifs setup on the current vfiler
    my %args;
    $args{'auth-type'}        = 'ad';
    $args{'cifs_pass'}        = 'netapp1!';
    $args{'create_admin'}     = 'yes';
    $args{'domain'}           = $Options_hash->{'CIFS_DOMAIN'};
    $args{'domain_user'}      = $Options_hash->{'CIFS_USER'};
    $args{'domain_pass'}      = $Options_hash->{'CIFS_PASSWD'};
    $args{'host'}             = host($node);
    $args{'passwd_group'}     = 'yes';
    $args{'select_site_name'} = 'yes';
    $args{'server-name'}      = $cifs_server;
    $args{'local_admins'}     = 'ssqa\\Domain Admins';
    if ($Options_hash->{'CIFS_SECURITY_STYLE'}) {
        $args{'security-style'} = $Options_hash->{'CIFS_SECURITY_STYLE'};
    }

    if ($vfiler ne 'vfiler0') { $args{'vfiler'} = $vfiler; }
    $args{'wins_option'} = 0;

    $Log->debug("cifs_setup args:\n" . Dumper(\%args));
    my $max_cifs_setup_attempts = 5;
    for (my $i = 1; $i <= $max_cifs_setup_attempts; $i++) {
        eval { cifs_setup(%args); };

        if ($@) {
            $Log->comment("Error detected in cifs setupi: $@");
            if ($i == $max_cifs_setup_attempts) {
                die config_result("Error configuring CIFS on vfiler $vfiler "
                        . "after $max_cifs_setup_attempts attempts");
            } else {
                $Log->comment(
                    "Waiting 60 seconds before retrying cifs setup");
                snooze 60;
            }
        } elsif (!$@) {
            last;
        }
    }
    $Log->exit();
} ## end sub configure_cifs

sub configure_nfs {
    $Log->enter();

    my ($self, %options) = @_;

    my $node    = $options{'node'};
    my $vfiler  = $options{'vfiler'};
    my $volumes = $options{'volumes'};
    my $retries = $options{'retries'};
    my $wait    = $options{'wait'};
    my $retry   = 1;

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $defined_volumes = $self->{'defined_topology'}->{$node}->{'vfilers'};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    # Get the path to the /vol/<vfiler>/etc/exports file for the given vfiler
    my $root_volume;
    my $exportvol;
    if ($vfiler eq 'vfiler0') {
        $exportvol = '/vol/vol0/etc/exports';
    } else {
        $root_volume
            = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
            ->{'volumes'}->[0];
        $exportvol = "/vol/$root_volume/etc/exports";
    }

  # Get the status of the exports file it may get created depending on the use
  # of the -n option at the time the vfiler is created.
    eval {

        $Response = $Zapi->file_get_file_info('path' => $exportvol);
    };

    # If the exports file does not exist for a given vfiler create it
    if ($@) {

        # Create the file
        $Log->comment("Creating the $exportvol file for vfiler $vfiler");
        $Cli->wrfile(
            'append'  => 1,
            'content' => "'# Generated by " . $self->{'script'} . ".thpl'\n",
            'file'    => $exportvol
        );
    }

# Because we are running with default security as ntfs we need a usermap. Set the path
    my $usermapcfg;
    if ($vfiler eq 'vfiler0') {
        $usermapcfg = '/vol/vol0/etc/usermap.cfg';
    } else {
        $usermapcfg = "/vol/$root_volume/etc/usermap.cfg";
    }

# The usermap.cfg exists when the vfiler is created, append the user mapping to it.
    $Log->comment(
        "Creating the $usermapcfg file to map unix root to windows administrator for vfiler $vfiler"
    );
    $Cli->wrfile(
        'append'  => 1,
        'content' => "administrator <= root",
        'file'    => $usermapcfg,
    );

# Get a list of all the volumes for a given vfiler for appending to ../etc/exports
    if ($vfiler ne 'vfiler0') {
        foreach my $volumes (@{$defined_volumes->{$vfiler}->{'volumes'}}) {
            $Log->comment(
                "Creating the export entry for /vol/$volumes in $exportvol for vfiler $vfiler"
            );
            $Cli->wrfile(
                'append'  => 1,
                'content' => "/vol/$volumes -sec=sys,rw,nosuid,anon=0",
                'file'    => "$exportvol",
            );
        }
    }

    # Now we have to run exportfs -a for each vfiler
    while ($retry) {
        try {
            $Log->comment("Running exportfs -a for vfiler $vfiler");
            my $Response = $Cli->vfiler_run(
                vfiler         => "$vfiler",
                command        => "exportfs",
                'command-args' => "-a",
            );
            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze 30;
            } else {
                $retry = 0;
            }

        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result(
                "Unable to get exportfs -a info for vfiler $vfiler after $retries tries"
            );
        }
    }

    $Log->exit();

} ## end sub configure_nfs

sub configure_dns {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Log->comment("Configuring DNS on $vfiler");

    # Create the resolv.conf file
    my $ascii
        = 'search ' . join(' ', @{$Options_hash->{'DNS_SEARCH'}}) . "\n";
    foreach my $ns (split(/,/, $Options_hash->{'DNS_SERVERS'})) {
        $ascii .= "nameserver $ns\n";
    }

    my $path = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
    if ($path eq '/etc') { $path = '/vol/vol0/etc'; }

    $self->write_file(
        'ascii' => $ascii,
        'node'  => $node,
        'path'  => "$path/resolv.conf",
    );

    snooze 1;

    my $rc_data;
    eval {
        $rc_data = $self->read_file(
            'path' => "$path/rc",
            'node' => $node
        );
    };

    my %data;
    $data{'dns.domainname'} = $Options_hash->{'CIFS_DOMAIN'};
    $data{'dns.enable'}     = 'on';
    foreach my $key (keys(%data)) {
        my $regex = $key;
        $regex =~ s/(\.)/\\$1/g;
        my $found = 0;
        my $new_rc_data;
        foreach my $line (split(/\n/, $rc_data)) {
            chomp($line);
            if ($line =~ /$regex/) {
                $new_rc_data .= "options $key $data{$key}\n";
                $found = 1;
            } else {
                $new_rc_data .= $line . "\n";
            }
        }
        if (!$found) {
            $new_rc_data .= "options $key $data{$key}\n";
        }
        $rc_data = $new_rc_data;
    }

    $Log->comment(Dumper($rc_data));
    $self->write_file(
        'ascii' => $rc_data,
        'node'  => $node,
        'path'  => "$path/rc"
    );

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    foreach my $key (keys(%data)) {
        try {

            $Response = $Zapi->options_set(
                'name'  => $key,
                'value' => $data{$key}
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    $Log->exit();
} ## end sub configure_dns

sub configure_nis {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Log->comment("Configuring NIS on $vfiler");

    my $path = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
    if ($path eq '/etc') { $path = '/vol/vol0/etc'; }

    my $rc_data;
    eval {
        $rc_data = $self->read_file(
            'path' => "$path/rc",
            'node' => $node
        );
    };

    my %data;
    if ($Options_hash->{'NIS_DOMAINNAME'} && $Options_hash->{'NIS_SERVERS'}) {
        $data{'nis.domainname'} = $Options_hash->{'NIS_DOMAINNAME'};
        $data{'nis.servers'}    = $Options_hash->{'NIS_SERVERS'};
        $data{'nis.enable'}     = 'on';
    }

    foreach my $key (keys(%data)) {
        my $regex = $key;
        $regex =~ s/(\.)/\\$1/g;
        my $found = 0;
        my $new_rc_data;
        foreach my $line (split(/\n/, $rc_data)) {
            chomp($line);
            if ($line =~ /$regex/) {
                $new_rc_data .= "options $key $data{$key}\n";
                $found = 1;
            } else {
                $new_rc_data .= $line . "\n";
            }
        }
        if (!$found) {
            $new_rc_data .= "options $key $data{$key}\n";
        }
        $rc_data = $new_rc_data;
    }

    $Log->comment(Dumper($rc_data));

    $self->write_file(
        'ascii' => $rc_data,
        'node'  => $node,
        'path'  => "$path/rc"
    );

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    foreach my $key (keys(%data)) {
        if ($key eq 'nis.enable') { next; }
        $Log->comment("Configuring options $key to $data{$key} on $vfiler");
        try {

            $Response = $Zapi->options_set(
                'name'  => $key,
                'value' => $data{$key}
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    try {

        $Response = $Zapi->options_set(
            'name'  => 'nis.enable',
            'value' => 'on'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->exit();
} ## end sub configure_nis

sub destroy_vfiler {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $defined_addresses
        = $self->{'defined_topology'}->{$node}->{'addresses'};
    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $discovered_interfaces
        = $self->{'discovered_topology'}->{$node}->{'interfaces'};
    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Log->comment("$vfiler is $discovered_vfilers->{$vfiler}->{'status'}");
    if (defined($discovered_vfilers->{$vfiler}->{'status'})
        && $discovered_vfilers->{$vfiler}->{'status'} =~ /stopped/i) {
        $Log->comment("Destroying $vfiler");
        $self->vfiler_destroy(
            'node'    => $node,
            'vfiler'  => $vfiler,
            'retries' => 5,
            'wait'    => 120
        );
        snooze 1;
    }

    foreach
        my $interface (keys(%{$defined_vfilers->{$vfiler}->{'interfaces'}})) {
        my $defined_addresses_by_interface
            = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};
        foreach my $address_information (@{$defined_addresses_by_interface}) {

            my $address = $address_information->{'address'};
            my $range   = $address_information->{'range'};
            my $netmask
                = $defined_addresses->{$interface}->{$range}->{'NETMASK'};

            my $address_regex;
            if ($address =~ /\./) {
                $address_regex = $address;
                $address_regex =~ s/(\.)/\\$1/g;
            } else {
                my $ip = new Net::IP($address);
                $address       = $ip->short();
                $address_regex = $address;

            }

            # Delete the alias if needed
            if (grep { $_->{'ip'} eq $address }
                @{$discovered_interfaces->{$interface}->{'ipv4_info'}}) {
                $self->ifconfig_remove_alias(
                    'node'      => $node,
                    'interface' => $interface,
                    'address'   => $address,
                    'retries'   => 5,
                    'wait'      => 60
                );
                snooze 1;
            }

            my $rc_data = $self->read_file(
                'path' => '/vol/vol0/etc/rc',
                'node' => $node
            );
            my $new_rc_data;
            foreach my $line (split(/\n/, $rc_data)) {
                chomp($line);
                if ($line !~ /$address_regex/) {
                    $new_rc_data .= $line . "\n";
                }
            }
            $self->write_file(
                'ascii' => $new_rc_data,
                'node'  => $node,
                'path'  => '/vol/vol0/etc/rc'
            );

        }
    }

    $Log->exit();
} ## end sub destroy_vfiler

sub delete_shares {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $defined_shares  = $topology_database->{$node}->{'shares'};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Deleting shares on $vfiler");

    while (my $share = each %{$defined_shares}) {
        if ($vfiler ne $defined_shares->{$share}->{'vfiler'}) { next; }

        my $volume = $defined_shares->{$share}->{'volume'};
        my $path   = $defined_shares->{$share}->{'path'};
        $Log->comment("Deleting share $share on $volume");

        try {

            $Response = $Zapi->cifs_share_delete('share-name' => $share);
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    $Log->exit();
} ## end sub delete_shares

sub terminate_cifs {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Terminating CIFS on $vfiler");

    my $Status;
    try {

        $Response = $Zapi->cifs_status();
        $Status   = $Response->get_parsed_output();
        if (grep(/started/, @{$Status->[0]->{'status'}})) {
            $Zapi->cifs_stop();
        }

    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->exit();
} ## end sub terminate_cifs

# Add to the vfiler
sub vfiler_add {

    $Log->enter();

    my ($self, %options) = @_;
    my $node      = $options{'node'};
    my $vfiler    = $options{'vfiler'};
    my $addresses = $options{'addresses'};
    my $storage   = $options{'storage'};
    my $retries   = $options{'retries'};
    my $wait      = $options{'wait'};
    my $retry     = 1;
    my $Cli       = $apiset{'cli'}{$node};
    my %args;
    $args{'vfiler'} = $vfiler;

    if ($addresses) {
        $args{'ip-addresses'} = $addresses;
    }
    if ($storage) {
        $args{'storage-units'} = $storage;
    }
    while ($retry) {
        try {

            $Response = $Cli->vfiler_add(%args);
            $retry    = 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            my $output    = $Response->get_raw_output();
            my $text      = $Exception->text();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                die config_result("$text: $output");
            }
        };
        if ($retry == $retries) {
            die config_result(
                "Unable to create vfiler $vfiler after $retries tries");
        }
    }
    $Log->exit();
} ## end sub vfiler_add

# Stop the vfiler
sub vfiler_stop {

    $Log->enter();

    my ($self, %options) = @_;
    my $vfiler  = $options{'vfiler'};
    my $node    = $options{'node'};
    my $retries = $options{'retries'};
    my $wait    = $options{'wait'};
    my $retry   = 1;
    my $Cli     = $apiset{'cli'}{$node};
    while ($retry) {
        try {

            $Response = $Cli->vfiler_stop(vfiler => $vfiler);
            $retry = 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            my $output    = $Response->get_raw_output();
            my $text      = $Exception->text();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                die config_result("$text: $output");
            }
        };
        if ($retry == $retries) {
            $Log->warn("Unable to stop vfiler $vfiler after $retries tries");
            last;
        }
    }

    $Log->exit();
} ## end sub vfiler_stop

# Destroy the vfiler
sub vfiler_destroy {

    $Log->enter();

    my ($self, %options) = @_;
    my $vfiler  = $options{'vfiler'};
    my $node    = $options{'node'};
    my $retries = $options{'retries'};
    my $wait    = $options{'wait'};
    my $Cli     = $apiset{'cli'}{$node};
    my $retry   = 1;
    while ($retry) {
        try {

            $Response = $Cli->vfiler_destroy(
                vfiler => $vfiler,
                force  => 1
            );
            $retry = 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            my $output    = $Response->get_raw_output();
            my $text      = $Exception->text();
            if ($output =~ /cannot be run|active command in progress/) {
                $retry++;
                snooze $wait;
            } else {
                die config_result("$text: $output");
            }
        };
        if ($retry == $retries) {
            $Log->warn(
                "Unable to destroy vfiler $vfiler after $retries tries");
            last;
        }
    }

    $Log->exit();
} ## end sub vfiler_destroy

# Create the vfiler
sub vfiler_create {

    $Log->enter();

    my ($self, %options) = @_;
    my $node      = $options{'node'};
    my $vfiler    = $options{'vfiler'};
    my $addresses = $options{'addresses'};
    my $storage   = $options{'storage'};
    my $retries   = $options{'retries'};
    my $wait      = $options{'wait'};
    my $retry     = 1;
    my $Cli       = $apiset{'cli'}{$node};
    while ($retry) {
        try {

            $Response = $Cli->vfiler_create(
                vfiler          => $vfiler,
                "bypass-setup"  => 1,
                "ip-addresses"  => $addresses,
                "storage-units" => $storage
            );
            $retry = 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            my $output    = $Response->get_raw_output();
            my $text      = $Exception->text();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                die config_result("$text: $output");
            }
        };
        if ($retry == $retries) {
            die config_result(
                "Unable to create vfiler $vfiler after $retries tries");
        }
    }
    $Log->exit();
} ## end sub vfiler_create

sub ifconfig_remove_alias {

    $Log->enter();

    my ($self, %options) = @_;
    my $node      = $options{'node'};
    my $interface = $options{'interface'};
    my $address   = $options{'address'};
    my $retries   = $options{'retries'};
    my $wait      = $options{'wait'};
    my $retry     = 1;
    my $Cli       = $apiset{'cli'}{$node};
    while ($retry) {
        try {

            $Response = $Cli->ifconfig(
                'interface' => $interface,
                'unalias'   => $address
            );
            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                $retry = 0;
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result("Unable to remove alias $address on "
                    . "interface $interface after $retries tries");
        }
    }

    $Log->exit();
} ## end sub ifconfig_remove_alias

sub ifconfig_alias {

    $Log->enter();

    my ($self, %options) = @_;
    my $node             = $options{'node'};
    my $interface        = $options{'interface'};
    my $address          = $options{'address'};
    my $netmask          = $options{'netmask'} || undef;
    my $retries          = $options{'retries'};
    my $wait             = $options{'wait'};
    my $retry            = 1;
    my $Cli              = $apiset{'cli'}{$node};
    my %ifconfig_options = ('interface' => $interface, 'alias' => $address);
    if ($address =~ /\./ && $netmask) {
        $ifconfig_options{'netmask'} = $netmask;
    }
    while ($retry) {
        try {

            $Response = $Cli->ifconfig(%ifconfig_options);
            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                $retry = 0;
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result("Unable to alias $address on interface "
                    . "$interface after $retries tries");
        }
    }

    $Log->exit();
} ## end sub ifconfig_alias

sub ifconfig_remove_address {

    $Log->enter();

    my ($self, %options) = @_;
    my $node      = $options{'node'};
    my $interface = $options{'interface'};
    my $address   = $options{'address'};
    my $netmask   = $options{'netmask'} || undef;
    my $retries   = $options{'retries'};
    my $wait      = $options{'wait'};
    my $retry     = 1;
    my $Cli       = $apiset{'cli'}{$node};
    my %ifconfig_options
        = ('interface' => $interface, 'address' => $address, 'down' => 1);

    if ($address =~ /\./ && $netmask) {
        $ifconfig_options{'netmask'} = $netmask;
    }
    while ($retry) {
        try {

            $Response = $Cli->ifconfig(%ifconfig_options);
            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                $retry = 0;
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result("Unable to remove address $address on "
                    . "interface $interface after $retries tries");
        }
    }

    $Log->exit();
} ## end sub ifconfig_remove_address

sub ifconfig_address {

    $Log->enter();

    my ($self, %options) = @_;
    my $node             = $options{'node'};
    my $interface        = $options{'interface'};
    my $address          = $options{'address'};
    my $netmask          = $options{'netmask'} || undef;
    my $retries          = $options{'retries'};
    my $wait             = $options{'wait'};
    my $retry            = 1;
    my $Cli              = $apiset{'cli'}{$node};
    my %ifconfig_options = ('interface' => $interface, 'address' => $address);

    if ($address =~ /\./ && $netmask) {
        $ifconfig_options{'netmask'} = $netmask;
    }

    while ($retry) {
        try {

            $Response = $Cli->ifconfig(%ifconfig_options);

            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                $retry = 0;
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result("Unable to alias $address on interface "
                    . "$interface after $retries tries");
        }
    }

    $Log->exit();
} ## end sub ifconfig_address

sub ifconfig {

    $Log->enter();

    my ($self, %options) = @_;
    my $node    = $options{'node'};
    my $retries = $options{'retries'};
    my $wait    = $options{'wait'};
    my $retry   = 1;
    my $Cli     = $apiset{'cli'}{$node};
    my $Ifconfig;

    while ($retry) {
        try {

            $Response = $Cli->ifconfig(all => 1);
            $Ifconfig = $Response->get_parsed_output();
            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                $retry = 0;
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result("Unable to get ifconfig information "
                    . " after $retries tries");
        }
    }

    $Log->exit();

    return $Ifconfig;
} ## end sub ifconfig

sub assign_cifs_server {

    $Log->enter();

    my ($self, %options) = @_;

    my $share           = $options{'share'};
    my $node            = $options{'node'};
    my $vfiler          = $options{'vfiler'};
    my $interface_index = $options{'interface_index'};
    my $address_indexes = $options{'address_indexes'};

    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    # Get the vfiler address from the vfiler entry
    my $interface = ${$Options_hash->{'DATA_INTERFACES'}}[$interface_index];
    if (!$interface) {
        die script_result("Unable to get the interface from DATA_INTERFACES\n"
                . Dumper($Options_hash->{'DATA_INTERFACES'}));
    }

    # Increment or reset the interface index
    $interface_index++;
    if ($interface_index > $#{$Options_hash->{'DATA_INTERFACES'}}) {
        $interface_index = 0;
    }

    # Get the ip address from the defined vfilers interface
    $Log->debug("Getting an address from interface $interface "
            . "address index $$address_indexes{$interface} for "
            . "for vfiler $vfiler");

    my $defined_addresses_this_interface
        = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};

    if ($$address_indexes{$interface} > $#{$defined_addresses_this_interface})
    {
        $$address_indexes{$interface} = 0;
    }
    my $ip
        = ${$defined_addresses_this_interface}[$$address_indexes{$interface}]
        ->{'address'};

    if (!$ip) {
        die script_result("Could not get IP address from interface "
                . "$interface on node $node\n"
                . "Defined_ addresses this interface\n"
                . Dumper($defined_addresses_this_interface) . "\n"
                . "Address indexes\n"
                . Dumper($address_indexes)
                . "Defined vfiler\n"
                . Dumper($defined_vfilers->{$vfiler})
                . "Discovered vfiler\n"
                . Dumper($discovered_vfilers->{$vfiler}));
    }

    my $net_ip                = new Net::IP($ip);
    my $version               = $Options_hash->{'DESTIP_VERSION'};
    my $current_address_index = $$address_indexes{$interface};

    while ($version != $net_ip->version()) {
        $Log->debug("The version of $ip does not match DESTIP_VERSION. "
                . "Getting next address on interface $interface");

        $$address_indexes{$interface}++;
        if ($$address_indexes{$interface}
            > $#{$defined_addresses_this_interface}) {
            $$address_indexes{$interface} = 0;
        }
        if ($$address_indexes{$interface} == $current_address_index) {
            last;
        }
        $ip = ${$defined_addresses_this_interface}
            [$$address_indexes{$interface}]->{'address'};
        if (!$ip) { last; }
        $net_ip = new Net::IP($ip);
    }

    if (!$ip) {

        my $defined_addresses
            = $self->{'defined_topology'}->{$node}->{'addresses'};
        my $defined_vfiler = $defined_vfilers->{$vfiler};
        my $discovered_interfaces
            = $self->{'discovered_topology'}->{$node}->{'interfaces'};

        my $msg
            = sprintf("Unable to retrieve an IP address for vfiler "
                . "%s\n using index %d of interface %s\'s "
                . "defined addresses\n",
            $vfiler, $$address_indexes{$interface}, $interface);
        die config_result($msg
                . "\nDiscovered Interfaces on $node:\n"
                . Dumper($discovered_interfaces)
                . "\nDefined vfiler $vfiler on $node:\n"
                . Dumper($defined_vfiler)
                . "\nDefined address ranges:\n"
                . Dumper($defined_addresses));
    }

    # If there is more than one address range specified
    # for a data interface we should switch back and forth
    # to spread the load
    $$address_indexes{$interface}++;
    if ($$address_indexes{$interface} > $#{$defined_addresses_this_interface})
    {
        $$address_indexes{$interface} = 0;
    }

    my $mode = $Options_hash->{'CIFS_SERVER_ACCESS_MODE'};
    my $server = $self->get_cifs_server_name('ip' => $ip, 'mode' => $mode);

    $Log->exit();
    return $server;

} ## end sub assign_cifs_server

sub filer_cleanup_defined {

    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    foreach my $node (@{$Options_hash->{'NODES'}}) {
        $Log->comment("Cleaning up node $node");

        my $Zapi          = $apiset{'zapi'}{$node};
        my $definedshares = $self->{'defined_topology'}->{$node}->{'shares'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};

        my @targets = ('vfilers');
        my @nodes   = ($node);
        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        # Stop vfilers
        $Log->comment("Stopping vfilers");
        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            if (!grep { $_ eq $vfiler } keys(%{$discovered_vfilers})) {
                next;
            }
            my $status = $discovered_vfilers->{$vfiler}->{'status'};
            if ($vfiler eq 'vfiler0' || $status =~ /stopped/) { next; }
            $self->vfiler_stop(
                'node'    => $node,
                'vfiler'  => $vfiler,
                'retries' => 5,
                'wait'    => 60
            );
            snooze 1;

        }

        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        # Destroy vfilers
        $Log->comment("Destroying vfilers");
        foreach my $vfiler (@vfilers) {
            if (!grep { $_ eq $vfiler } keys(%{$discovered_vfilers})) {
                next;
            }
            if ($vfiler eq 'vfiler0') { next; }
            $self->destroy_vfiler(
                'node'    => $node,
                'vfiler'  => $vfiler,
                'retries' => 5,
                'wait'    => 60
            );
            snooze 1;
        }

        # Unconfigure vfiler0's data interfaces
        foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
            my $address
                = $defined_vfilers->{'vfiler0'}->{'interfaces'}->{$interface}
                ->{'address'};
            if ($address =~ /\./) {

                $self->ifconfig_remove_address(
                    'node'      => $node,
                    'interface' => $interface,
                    'address'   => '0.0.0.0',
                    'netmask'   => '255.255.255.255',
                    'wait'      => 60,
                    'retries'   => 5
                );
                snooze 1;
            } else {
                eval {
                    $self->ifconfig_remove_address(
                        'node'      => $node,
                        'interface' => $interface,
                        'address'   => '0:0:0:0:0:0:0:0',
                        'wait'      => 60,
                        'retries'   => 5
                    );
                };
                snooze 1;
            }
        }

        my $discovered_licenses
            = $self->{'discovered_topology'}->{$node}->{'licenses'};
        if ($discovered_licenses->{'cifs'} !~ /not licensed/i) {
            $self->terminate_cifs('node' => $node, 'vfiler' => 'vfiler0');

            $self->cleanup_cifsconfig_share_file(
                'node'   => $node,
                'vfiler' => 'vfiler0'
            );

            $self->cleanup_lcl_groups_file(
                'node'   => $node,
                'vfiler' => 'vfiler0'
            );
        }

        @targets = ('volumes');
        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        my $defined = keys(%{$self->{'defined_topology'}->{'volumes'}});
        $Log->comment("Destroying $defined volumes");

        # Set the vfiler context for the next ZAPI call
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
        my @skipped_volumes;
        foreach my $volume (keys(%{$defined_volumes})) {
            if (!grep { $_ eq $volume } keys(%{$discovered_volumes})) {
                next;
            }

            $Log->comment("Checking quota status on $volume");

            my $Status;
            try {
                $Response = $Zapi->quota_status('volume' => $volume);
                $Status = $Response->get_parsed_output();
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;

                #dont die if flexcache vol
                if ($Exception->text()
                    =~ /sparse volume that does not support quotas/) {
                    $Log->comment($Exception->text());
                } elsif ($Exception->text() =~ /is offline/) {
                    $Log->comment($Exception->text());
                } else {
                    die config_result($Exception->text());
                }
            };

            if (grep { $_ eq 'on' } @{$Status->[0]->{'status'}}) {

                $Log->comment("Stopping quotas on $volume");

                try {

                    $Response = $Zapi->quota_off('volume' => $volume);
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }

            foreach my $quota (@{$defined_volumes->{$volume}->{'quotas'}}) {
                $quota =~ s/CIFS_USER/$Options_hash->{'CIFS_USER'}/g;
                my @fields = split(/:/, $quota);
                eval {
                    $Zapi->quota_delete_entry(
                        'quota-target' => $fields[0],
                        'quota-type'   => $fields[1],
                        'volume'       => $volume,
                        'qtree'        => ''
                    );
                };
            }

            if ($discovered_volumes->{$volume}->{'is-online'} ne 'false') {

                try {

                    $Response = $Zapi->volume_offline(
                        'name'       => $volume,
                        'cifs-delay' => 0
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    if ($Exception->text() =~ /is already offline/) {
                        $Log->comment($Exception->text());
                    } else {
                        die config_result($Exception->text());
                    }
                };
            }

            snooze 1;

            try {

                $Response = $Zapi->volume_destroy(
                    'name'  => $volume,
                    'force' => 1
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                if ($Exception->text()
                    =~ /cannot be destroyed because it has clone volumes/) {
                    push(@skipped_volumes, $volume);
                    $Log->comment(
                        "skipping $volume Reason: " . $Exception->text());
                } else {
                    die config_result($Exception->text());
                }
            };

        }
        foreach my $volume (@skipped_volumes) {
            $Log->comment("Destroying $volume");
            try {

                $Response = $Zapi->volume_destroy(
                    'name'  => $volume,
                    'force' => 1
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }
    }

    $Log->exit();
} ## end sub filer_cleanup_defined

sub filer_cleanup_discovered {

    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    $Log->comment("Cleaning up discovered configuration");

    foreach my $node (@{$Options_hash->{'NODES'}}) {
        my $Cli = $apiset{'cli'}{$node};
        $Log->comment("Cleaning up node $node");

        my $Zapi = $apiset{'zapi'}{$node};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my ($discovered_vfilers, $discovered_interfaces,
            $discovered_volumes, $discovered_shares,
            $discovered
        );

        try {
            $Response = $Cli->setral(
                'reset'           => 1,
                'privilege-level' => "diag"
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            if ($Exception->text() =~ /No such command/i) {
                $Log->comment("Setral not found");
            } else {
                die config_result($Exception->text());
            }
        };

        my @targets = ('vfilers', 'licenses', 'volumes', 'qtrees');
        my @nodes = ($node);

        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};
        $discovered = keys(%{$discovered_vfilers});
        $Log->comment("Cleaning up $discovered vfilers");

        my @vfilers = keys(%{$discovered_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {

            # Set the vfiler context for the next ZAPI call
            try {
                $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };

            # Destroy all discovered volumes quotas
            $Log->comment(
                "Disabling quotas on volumes owned by vfiler $vfiler");
            foreach my $volume (keys(%{$discovered_volumes})) {
                if (   $volume eq 'vol0'
                    || $discovered_volumes->{$volume}->{'owning-vfiler'} ne
                    $vfiler) {
                    next;
                }

                if ($discovered_volumes->{$volume}->{'state'} eq 'offline') {
                    next;
                }

                $Log->comment("Checking quota status on $volume");

                my $Status;
                try {
                    $Response = $Zapi->quota_status('volume' => $volume);
                    $Status = $Response->get_parsed_output();
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;

                    #dont die if flexcache vol
                    if ($Exception->text() =~ /does not support quotas/) {
                        $Log->comment($Exception->text());
                    } else {
                        die config_result($Exception->text());
                    }
                };

                if (grep { $_ eq 'on' } @{$Status->[0]->{'status'}}) {
                    $Log->comment("Stopping quotas on $volume");

                    try {

                        $Response = $Zapi->quota_off('volume' => $volume);
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }
            }

            my $path
                = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
            if ($path eq '/etc') { $path = '/vol/vol0/etc'; }
            if ($path !~ /\/vol/) {
                $Log->comment("Cannot determine the path to the configuration"
                        . "directory of vfiler $vfiler\n"
                        . Dumper($discovered_vfilers->{$vfiler}));
                next;
            }

            $Log->comment("Cleaning quota file $path");
            $self->truncate_file(
                'node'   => $node,
                'vfiler' => $vfiler,
                'path'   => "$path/quotas"
            );

        }

        # Stop vfilers
        $Log->comment("Stopping vfilers");
        if (keys(%{$discovered_vfilers}) == 0) {
            $Log->comment(
                "No vfilers discovered " . Dumper($discovered_vfilers));
        }

        foreach my $vfiler (@vfilers) {
            my $status = $discovered_vfilers->{$vfiler}->{'status'};
            if ($vfiler eq 'vfiler0' || $status =~ /stopped/) { next; }
            $self->vfiler_stop(
                'node'    => $node,
                'vfiler'  => $vfiler,
                'retries' => 5,
                'wait'    => 60
            );
            snooze 1;

        }

        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        # Destroy all vfilers
        $Log->comment("Destroying vfilers");
        foreach my $vfiler (@vfilers) {
            if ($vfiler eq 'vfiler0') { next; }
            $self->destroy_vfiler(
                'node'    => $node,
                'vfiler'  => $vfiler,
                'retries' => 5,
                'wait'    => 60
            );

            snooze 5;
        }

        # Unconfigure vfiler0's data interfaces
        @targets = ('interfaces');
        @nodes   = ($node);

        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        $discovered_interfaces
            = $self->{'discovered_topology'}->{$node}->{'interfaces'};

        foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
            if (!grep { $_ eq $interface }
                natsort(keys(%{$discovered_interfaces}))) {
                $Log->comment("$interface not found on $node, skipping");
                next;
            }

            if (grep { $_->{'ip'} =~ /\./ }
                @{$discovered_interfaces->{$interface}->{'ipv4_info'}}) {

                $self->ifconfig_remove_address(
                    'node'      => $node,
                    'interface' => $interface,
                    'address'   => '0.0.0.0',
                    'netmask'   => '255.255.255.255',
                    'wait'      => 60,
                    'retries'   => 5
                );
                snooze 1;
            }

            if (grep { $_->{'ip'} =~ /:/ }
                @{$discovered_interfaces->{$interface}->{'ipv6_info'}}) {

                $self->ifconfig_remove_address(
                    'node'      => $node,
                    'interface' => $interface,
                    'address'   => '0:0:0:0:0:0:0:0',
                    'wait'      => 60,
                    'retries'   => 5
                );
                snooze 1;
            }
        }

        @targets = ('vfilers', 'licenses', 'volumes', 'qtrees');
        @nodes = ($node);

        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );
        $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        my $discovered_licenses
            = $self->{'discovered_topology'}->{$node}->{'licenses'};
        if ($discovered_licenses->{'cifs'} !~ /not licensed/i) {
            $self->terminate_cifs('node' => $node, 'vfiler' => 'vfiler0');

            $self->delete_cifsconfig_share_file(
                'node'   => $node,
                'vfiler' => 'vfiler0'
            );

            $self->cleanup_lcl_groups_file(
                'node'   => $node,
                'vfiler' => 'vfiler0'
            );
        }

        @targets = ('volumes');
        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};

        $discovered = keys(%{$discovered_volumes});
        if ($discovered > 1) { $discovered--; }
        $Log->comment("Destroying $discovered volumes");

        # Set the vfiler context for the next ZAPI call
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
        my @skipped_volumes;

        # Destroy all discovered volume
        foreach my $volume (keys(%{$discovered_volumes})) {
            if ($volume eq 'vol0') { next; }
            if (!grep(/^$volume$/, keys(%{$discovered_volumes}))) {
                next;
            }

            if ($discovered_volumes->{$volume}->{'state'} ne 'offline') {
                $Log->comment("Offlining $volume");

                try {

                    $Response = $Zapi->volume_offline(
                        'name'       => $volume,
                        'cifs-delay' => 0
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }

            snooze 1;

            $Log->comment("Destroying $volume");
            try {

                $Response = $Zapi->volume_destroy(
                    'name'  => $volume,
                    'force' => 1
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                if ($Exception->text()
                    =~ /cannot be destroyed because it has clone volumes/) {
                    push(@skipped_volumes, $volume);
                    $Log->comment(
                        "skipping $volume Reason: " . $Exception->text());
                } else {
                    die config_result($Exception->text());
                }
            };
        }
        foreach my $volume (@skipped_volumes) {
            $Log->comment("Destroying $volume");
            try {

                $Response = $Zapi->volume_destroy(
                    'name'  => $volume,
                    'force' => 1
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }

    }

    $Log->exit();
} ## end sub filer_cleanup_discovered

sub cleanup_cifsconfig_share_file {

    $Log->enter();

    my ($self, %options) = @_;
    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};
    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Options_hash = $Options->get_hash();

    $Log->comment("Reseting cifsconfig_share.cfg on vfiler $vfiler");

    #<<<
    my $ascii = 'cifs shares -add "ETC$" "/etc" -comment "Remote Administration"' . "\n";
    $ascii   .= 'cifs access "ETC$" S-1-5-32-544 Full Control' . "\n";
    $ascii   .= 'cifs shares -add "C$" "/" -comment "Remote Administration"' . "\n";
    $ascii   .= 'cifs access "C$" S-1-5-32-544 Full Control' . "\n";
    #>>>
    my $path = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
    if ($path eq '/etc') { $path = '/vol/vol0/etc'; }
    $path .= '/cifsconfig_share.cfg';

    # Set the vfiler context for the next ZAPI calls
    my $Zapi = $apiset{'zapi'}{$node};

    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $self->write_file(
        'ascii' => $ascii,
        'node'  => $node,
        'path'  => $path
    );

    $Log->exit();

} ## end sub cleanup_cifsconfig_share_file

sub delete_cifsconfig_share_file {

    $Log->enter();

    my ($self, %options) = @_;
    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};
    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Options_hash = $Options->get_hash();

    $Log->comment("Delete cifsconfig_share.cfg on vfiler $vfiler");

    my $path = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
    if ($path eq '/etc') { $path = '/vol/vol0/etc'; }
    $path .= '/cifsconfig_share.cfg';

    # Set the vfiler context for the next ZAPI calls
    my $Zapi = $apiset{'zapi'}{$node};

    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    eval {

        $Response = $Zapi->file_delete_file('path' => $path);
    };

    $Log->exit();

} ## end sub delete_cifsconfig_share_file

sub cleanup_lcl_groups_file {

    $Log->enter();

    my ($self, %options) = @_;
    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};
    my $Zapi   = $apiset{'zapi'}{$node};
    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Options_hash = $Options->get_hash();

    $Log->comment("Reseting lclgroups.cfg on vfiler $vfiler");

    #<<<
    my $ascii = '[ "Replicators"  552 ( "not supported" ) ]' . "\n";
      $ascii .= '[ "Backup Operators"  551 ( "Members can bypass file security to backup files" ) ]' . "\n";
      $ascii .= '[ "Power Users"  547 ( "Members that can share directories" ) ]' . "\n";
      $ascii .= '[ "Guests"  546 ( "Users granted Guest Access" ) ]' . "\n";
      $ascii .= '[ "Users"  545 ( "Ordinary Users" ) ]' . "\n";
      $ascii .= '[ "Administrators"  544 ( "Members can fully administer the filer" ) ]' . "\n";
    #>>>
    my $path = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
    if ($path eq '/etc') { $path = '/vol/vol0/etc'; }
    $path .= '/lclgroups.cfg';

    # Set the vfiler context for the next ZAPI calls
    $Zapi = $apiset{'zapi'}{$node};
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $self->write_file(
        'ascii' => $ascii,
        'node'  => $node,
        'path'  => $path
    );

    $Log->exit();

} ## end sub cleanup_lcl_groups_file

sub truncate_file {

    $Log->enter();

    my ($self, %options) = @_;
    my $node = $options{'node'};
    my $path = $options{'path'};
    my $Zapi = $apiset{'zapi'}{$node};

    $Options_hash = $Options->get_hash();

    $Log->comment("Truncating $path");

    try {

        $Response = $Zapi->file_truncate_file('path' => $path);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        if ($Exception->text() !~ /No such file or directory/) {
            die config_result($Exception->text());
        }
    };

    $Log->exit();

} ## end sub truncate_file

sub time_setup {

    $Log->enter();

    my ($self, %options) = @_;
    my $node = $options{'node'};
    my $Zapi = $apiset{'zapi'}{$node};

    host($Options_hash->{'CIFS_DC'});
    my $conn_obj = connect('#rsh', '-nologin', -timeout, 300);

    $Log->comment("Getting date from $Options_hash->{'CIFS_DC'}");
    my ($date, $output);
    my $cmd = "cmd /c date /t";
    eval { (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 300); };

    foreach my $line (split(/\n/, $output)) {
        chomp($line);

        # Tue 07/17/2007
        if ($line =~ /\w+\s+(.+)/) {
            $date = $1;
        }
    }

    $Log->comment("Getting time from $Options_hash->{'CIFS_DC'}");
    $cmd = "cmd /c time /t";
    eval { (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 300); };

    foreach my $line (split(/\n/, $output)) {
        chomp($line);

        # 07:00 PM
        if ($line =~ /^(\d+):(\d+)\s+(\w+)/) {
            my $hour     = $1;
            my $minutes  = $2;
            my $day_time = $3;
            if ($day_time =~ /AM/i && $hour == 12) { $hour = '00'; }
            if ($day_time =~ /PM/i && $hour != 12) { $hour = $hour + 12; }
            $date .= " $hour:$minutes:00";
        }
    }

    $Log->comment("Configuring time on $node");
    $Log->comment("Disabling NTP");
    try {

        $Response = $Zapi->options_set(
            'name'  => 'timed.enable',
            'value' => 'off'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Setting time servers");
    try {

        $Response = $Zapi->options_set(
            'name'  => 'timed.servers',
            'value' => $Options_hash->{'TIME_SERVER'}
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Setting time protocol");
    try {

        $Response = $Zapi->options_set(
            'name'  => 'timed.proto',
            'value' => 'ntp'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Setting timezone");
    try {

        $Response = $Zapi->clock_set_timezone(
            'timezone' => $Options_hash->{'TIME_ZONE'});
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    my $time = $self->get_epoch_time('date' => $date);
    $Log->comment("Setting time to epoch time $time");
    try {

        $Response = $Zapi->clock_set_clock(
            'is-utc-clock' => 0,
            'time'         => $time
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    try {

        $Response = $Zapi->options_set(
            'name'  => 'timed.enable',
            'value' => 'on'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->exit();
} ## end sub time_setup

sub read_file {

    $Log->enter();

    my ($self, %options) = @_;

    my $node = $options{'node'};
    my $path = $options{'path'};

    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Getting file info for $path");
    my $size;
    try {

        $Response = $Zapi->file_get_file_info('path' => $path);
        my $file_info = $Response->get_parsed_output();
        $size = $file_info->[0]->{'file-info'}->[0]->{'file-size'};
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Reading $path");
    my $data;
    try {

        $Response = $Zapi->file_read_file(
            'length' => $size,
            'offset' => 0,
            'path'   => $path
        );
        my $file_info = $Response->get_parsed_output();
        $data = $file_info->[0]->{'data'};
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->exit();
    return pack("H*", $data);

} ## end sub read_file

sub write_file {

    $Log->enter();

    my ($self, %options) = @_;

    my $ascii     = $options{'ascii'};
    my $node      = $options{'node'};
    my $path      = $options{'path'};
    my $offset    = $options{'offset'} || 0;
    my $overwrite = $options{'overwrite'} || 'true';

    my $Zapi = $apiset{'zapi'}{$node};
    my $Cli  = $apiset{'cli'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    try {
        if ($overwrite eq 'true') {
            $self->truncate_file(
                'node'   => $node,
                'vfiler' => 'vfiler0',
                'path'   => $path
            );
            snooze 1;
        }

    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Writing file $path");

    try {

        $Response = $Zapi->file_write_file(
            'data'      => unpack("H*", $ascii),
            'offset'    => $offset,
            'overwrite' => $overwrite,
            'path'      => $path
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        eval {
            foreach my $line (split(/\n/, $ascii)) {
                $Cli->wrfile(
                    append    => 1,
                    'content' => "$line",
                    'file'    => $path,
                );
            }
        };
        if ($@) { die config_result($@); }
    };

    $Log->exit();
} ## end sub write_file

sub list_shares {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Listing shares on vfiler $options{'vfiler'}");

    filer_vfiler_context_simulate($options{'conn'}, $options{'vfiler'});

    my $output;
    my $cmd = filer_vfiler_command($options{'conn'}, 'cifs shares');
    eval {
        (undef, undef, $output) = say($options{'conn'}, $cmd, -timeout, 300);
    };

    if ($@) {
        $Log->comment("$cmd\n $output");
        die "Error listing cifs shares\n";
    } else {
        $Log->debug("Command output: \n$output");
    }

    # Search the commands output
    my (%shares, @share_info);
    foreach my $line (split(/\n/, $output)) {

        #Name         Mount Point                       Description
        #----         -----------                       -----------
        #ETC$         /etc                              Remote Administration
        #                        BUILTIN\Administrators / Full Control
        #C$           /                                 Remote Administration
        #			     BUILTIN\Administrators / Full Control

        if ($line =~ /^(\S+)\s+(\/\S+)\s+/) {
            $shares{'share_name'} = $1;
            $shares{'path'}       = $2;
            if (   $options{'filter'}
                && $shares{'share_name'} =~ /$options{'filter'}/) {
                push(@share_info, \%shares);
            } elsif (!$options{'filter'}) {
                push(@share_info, \%shares);
            }
        }
    }

    $Log->exit();
    return \@share_info;
} ## end sub list_shares

sub ra_setup {

    $Log->enter();

    my ($self, %options) = @_;

    foreach my $node (@{$Options_hash->{'NODES'}}) {
        $Log->comment(
            "Configuring resource agitation on node $options{'node'}");

        filer_vfiler_context_simulate($conn{$node}, 'vfiler0');

        my $Cli = $apiset{'cli'}{$node};

        foreach my $r (keys(%{$options{'ra_config'}})) {
            if (!grep(/$r/, @{$Options_hash->{'RA_IGNORE'}})) {

                try {
                    $Response = $Cli->setral(
                        $r                => ${$options{'ra_config'}}{$r},
                        'privilege-level' => "diag"
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                if (0) {
                    my $output;
                    my $cmd = filer_vfiler_command($options{'conn'},
                              'priv set advanced; setral ' 
                            . $r . ' '
                            . ${$options{'ra_config'}}->{$r});
                    eval {
                        (undef, undef, $output)
                            = say($conn{$node}, $cmd, -timeout, 300);
                    };

                    if ($@) {
                        $Log->comment("$cmd\n $output");
                        die "Error configuring resource agitation\n";
                    } else {
                        $Log->debug("Command output: \n$output");
                    }
                }
            }
        }
    }

    $Log->exit();
} ## end sub ra_setup

sub create_local_groups {

    $Log->enter();

    my ($self, %options) = @_;

    my $node                 = $options{'node'};
    my $vfiler               = $options{'vfiler'};
    my $group_basename       = $options{'group_basename'};
    my $starting_group_index = $options{'starting_group_index'};
    my $ending_group_index   = $options{'ending_group_index'};
    my $group_members        = $options{'group_members'};

    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $defined_shares  = $topology_database->{$node}->{'shares'};
    my $discovered_qtrees
        = $self->{'discovered_topology'}->{$node}->{'qtrees'};
    my $discovered_shares
        = $self->{'discovered_topology'}->{$node}->{'shares'};

    $Log->comment("Creating "
            . ($ending_group_index - $starting_group_index + 1)
            . " local groups on node $node vfiler $vfiler");

    my @interfaces = keys(%{$defined_vfilers->{$vfiler}->{'interfaces'}});
    my $interface;
    foreach my $i (@interfaces) {
        if (grep { $_ eq $i } @{$Options_hash->{'DATA_INTERFACES'}}) {
            $interface = $i;
            last;
        }
    }

    if (!$interface) {
        die config_result("Unable to find an interface for $vfiler "
                . Dumper($defined_vfilers->{$vfiler}->{'interfaces'}));
    }

    my $ip = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface}->[0]
        ->{'address'};

    my $server;
    if ($ip =~ /:/) {
        $server = $self->get_cifs_server_name('ip' => $ip, 'mode' => 'name');
    } else {
        $server = $ip;
    }

    my @clients
        = grep { host($_)->hosttype() =~ /win/ } @{$self->{'clients'}};
    if (!@clients) {
        die config_result(
            "There are no windows clients to create local groups");
    }

    host($clients[0]);
    my $conn_obj = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
        '-nologin', -timeout, 300);
    eval { check_connectivity('conn' => $conn_obj); };
    my $error = $@;
    if ($error) { die config_result($error); }

    map_share(
        'conn'     => $conn_obj,
        'share'    => "\\\\$server\\c\$",
        'drive'    => '',
        'username' => $Options_hash->{'CIFS_USER'},
        'password' => $Options_hash->{'CIFS_PASSWD'},
        'timeout'  => 300,
        'reuse'    => 1,
        'delay'    => 60,
        'attempts' => 3,
        'tests'    => 'r'
    );

    my $cmd = 'cmd /c del \\lclgroups.txt';

    my $output;

    eval { (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 60); };

    $error = $@;

    if ($error) {
        die config_result(
            "Error creating add users input file:\n$output\n$error");
    }

    snooze 1;

    $cmd
        = 'cscript x:\\agnostic\\tools\\windows\\vbs\\createAddUsersFile.vbs '
        . '\\lclgroups.txt '
        . "$group_basename  $starting_group_index $ending_group_index "
        . "$group_members";

    eval { (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 60); };

    $error = $@;

    if ($error) {
        die config_result(
            "Error creating add users input file:\n$output\n$error");
    }

    snooze 1;

    $cmd = 'x:\\bin\\addusers \\\\' . $server . ' /c \\lclgroups.txt';

    eval { (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 3600); };

    $error = $@;

    if ($error) {
        die config_result(
            "Error creating add users input file:\n$output\n$error");
    }

    $Log->exit();
} ## end sub create_local_groups

sub is_option_valid {

    $Log->enter();

    my ($self, %options) = @_;
    my $Host   = host($options{'node'});
    my %fields = $Host->param_fields();

    $Log->comment(
        "Checking option $options{'option'} against release $self->{'release'}"
    );

    if ($options{'option'} eq 'licensed_feature.multistore.enable') {
        if (   $self->{'release'} !~ /boilermaker/i
            && $fields{'TYPE'} =~ /VSA/i) {
            $Log->comment("$options{'option'} is valid on this paltform");
            $Log->exit();
            return 1;
        } elsif ($self->{'release'} =~ /sierranevada/i) {
            $Log->comment("$options{'option'} is valid on this paltform");
            $Log->exit();
            return 1;
        }
    } elsif ($options{'option'} eq 'licensed_feature.flexcache_nfs.enable') {
        if ($self->{'release'} =~ /sierranevada|rollingrock/i) {
            $Log->comment("$options{'option'} is valid on this paltform");
            $Log->exit();
            return 1;
        }
    }
    $Log->exit();
    return 0;
} ## end sub is_option_valid

sub configure_quotas {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Zapi = $apiset{'zapi'}{$node};
    $Log->comment("Configuring quotas for vfiler $vfiler node $node");
    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $path;
    if ($vfiler eq 'vfiler0') {
        $path = '/vol/vol0/etc/quotas';
    } else {
        my $volume
            = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
            ->{'volumes'}->[0];
        $path = "/vol/$volume/etc/quotas";
    }

    my $file_info;
    eval {

        $Response = $Zapi->file_get_file_info('path' => $path);
    };

    # if the file does not exist create it
    if ($@) {

        # Create the file
        $self->write_file(
            'ascii' => "'# Generated by " . $self->{'script'} . ".thpl'\n",
            'node'  => $node,
            'path'  => $path
        );

    }

    snooze 1;

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $defined_volumes = $self->{'defined_topology'}->{$node}->{'volumes'};
    my $defined_qtrees  = $self->{'defined_topology'}->{$node}->{'qtrees'};
    foreach my $volume (keys(%{$defined_volumes})) {

        my $owning_vfiler = $defined_volumes->{$volume}->{'vfiler'};
        if ($vfiler ne $owning_vfiler || $volume eq "v_$vfiler") { next; }

        $Log->comment("Creating quota entries for $volume");

        #target:type:disk:files:thold:sdisk:sfile
        #CIFS_USER:user@volume:-:-:700M:700M:5K

        foreach my $quota (split(/;/, $Options_hash->{'QUOTAS'})) {
            if ($quota !~ /\@volume/) { next; }
            $quota =~ s/CIFS_USER/$Options_hash->{'CIFS_USER'}/g;
            my @fields = split(/:/, $quota);
            $fields[1] =~ s/\@volume//g;
            my $Entry;
            eval {
                $Response = $Zapi->quota_get_entry(
                    'quota-target' => $fields[0],
                    'quota-type'   => $fields[1],
                    'volume'       => $volume,
                    'qtree'        => ''
                );
                $Entry = $Response->get_parsed_output();

            };

            if (!$Entry) {
                $Log->comment("Adding quota to $volume\n");
                try {

                    $Response = $Zapi->quota_add_entry(
                        'quota-target'    => $fields[0],
                        'quota-type'      => $fields[1],
                        'disk-limit'      => $fields[2],
                        'file-limit'      => $fields[3],
                        'threshold '      => $fields[4],
                        'soft-disk-limit' => $fields[5],
                        'soft-file-limit' => $fields[6],
                        'volume'          => $volume,
                        'qtree'           => ''
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }

            push(
                @{  $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'quotas'}
                    },
                $quota
            );

        }

        foreach my $qtree (keys(%{$defined_qtrees})) {

            my $owning_volume = $defined_qtrees->{$qtree}->{'volume'};
            if ($volume ne $owning_volume) { next; }

            #target:type:disk:files:thold:sdisk:sfile
            #CIFS_USER:user@qtree:-:-:700M:700M:5K

            foreach my $quota (split(/;/, $Options_hash->{'QUOTAS'})) {
                if ($quota !~ /\@qtree/) { next; }
                $quota =~ s/CIFS_USER/$Options_hash->{'CIFS_USER'}/g;
                my @fields = split(/:/, $quota);
                $fields[1] =~ s/\@qtree//g;
                my $Entry;
                eval {
                    $Response = $Zapi->quota_get_entry(
                        'quota-target' => $fields[0],
                        'quota-type'   => $fields[1],
                        'volume'       => $volume,
                        'qtree'        => $qtree
                    );
                    $Entry = $Response->get_parsed_output();

                };

                if (!$Entry) {
                    $Log->comment("Adding quota to qtree $qtree on $volume");
                    try {

                        $Response = $Zapi->quota_add_entry(
                            'quota-target'    => $fields[0],
                            'quota-type'      => $fields[1],
                            'disk-limit'      => $fields[2],
                            'file-limit'      => $fields[3],
                            'threshold '      => $fields[4],
                            'soft-disk-limit' => $fields[5],
                            'soft-file-limit' => $fields[6],
                            'volume'          => $volume,
                            'qtree'           => $qtree
                        );
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }

                push(
                    @{  $self->{'defined_topology'}->{$node}->{'volumes'}
                            ->{$volume}->{'quotas'}
                        },
                    $quota
                );

            }
        }

        # turn on quotas until the number of vols for quotas is hit
        if ($self->{'defined_topology'}->{$node}->{'quota_vol_count'} > 0) {
            try {

                $Response = $Zapi->quota_on('volume' => $volume,);
            }
            catch NACL::APISet::Exceptions::ResponseException with {

                #my $Exception = shift;
                #die config_result($Exception->text());
            };
            $self->{'defined_topology'}->{$node}->{'quota_vol_count'}--;
        }
        snooze 1;
    }
} ## end sub configure_quotas

sub assign_data_volumes {
    $Log->enter();

    my ($self, %options) = @_;
    my $basename = $options{'basename'};
    my $count    = $options{'count'};
    my $type     = $options{'type'};

    my @nodes                    = @{$Options_hash->{'NODES'}};
    my $count_per_node           = int($count / @nodes);
    my $count_per_node_remainder = $count % @nodes;

    # Spread the volumes acrosss the specified aggregates
    $Log->comment("Adding $count $type volumes to aggregates");

    my $volume_index = 1;
    my @volumes;
    foreach my $node (@nodes) {

        my $size = $self->{'discovered_topology'}->{$node}->{'vol_size'};
        if ($type =~ /clone/) {
            $size = 0;
        }
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $aggregate_index  = 0;
        my $volume_usage     = 0;
        my $volumes_per_node = $count_per_node;
        if ($node eq $nodes[0]) {
            $volumes_per_node += $count_per_node_remainder;
        }

        # Get an array of the discovered aggregates
        my @discovered_aggregates = natsort(
            keys(%{$self->{'discovered_topology'}->{$node}->{'aggregates'}}));

        # loop through the discovered aggregates and create an array of our
        # defined aggregates
        my @defined_aggregates;
        foreach my $aggregate (@discovered_aggregates) {
            if (grep { $_ eq $aggregate }
                split(/,/, $Options_hash->{'AGGREGATES'})
                ) {
                push(@defined_aggregates, $aggregate);
            }
        }

        # Set the current aggregate and it's size
        my $aggregate = $defined_aggregates[$aggregate_index];
        $self->{'discovered_topology'}->{$node}->{'aggregates'}->{$aggregate}
            ->{'total'} =~ /(\d+)M/;
        my $aggregate_size
            = int($1 * ($Options_hash->{'MAX_AGGREGATE_USAGE'} / 100));
        $self->{'discovered_topology'}->{$node}->{'aggregates'}->{$aggregate}
            ->{'volume_usage'} = 0;

        #array for guarantee type
        my @guarantee_types = qw(volume file none);

        #default guarantee type is volume
        my $guar_type = 'volume';

        # Loop though each volume assigning it to the current aggregate
        foreach my $i (1 ... $volumes_per_node) {

            # Create the volume name
            my $volume = $basename . '_' . $volume_index;

            # Check to make sure we don't exceed the current aggregates space.
            $volume_usage
                = $self->{'discovered_topology'}->{$node}->{'aggregates'}
                ->{$aggregate}->{'volume_usage'};
            if ($volume_usage + $size > $aggregate_size) {

                # Switch to the next aggregate and grab it's size
                if ($defined_aggregates[$aggregate_index + 1]) {
                    $aggregate_index++;

                    #$volume_usage=0;
                    $aggregate = $defined_aggregates[$aggregate_index];
                    $self->{'discovered_topology'}->{$node}->{'aggregates'}
                        ->{$aggregate}->{'volume_usage'} = 0;
                    $Log->comment(
                        Dumper(
                            $self->{'discovered_topology'}->{$node}
                                ->{'aggregates'}
                        )
                    );
                    $self->{'discovered_topology'}->{$node}->{'aggregates'}
                        ->{$aggregate}->{'total'} =~ /(\d+)M/;
                    $aggregate_size = int(
                        $1 * ($Options_hash->{'MAX_AGGREGATE_USAGE'} / 100));
                } else {
                    die config_result("Not enough aggregate space on "
                            . "$node:$aggregate when trying to calculate "
                            . "space for data volume $volume\n"
                            . "Allocated space . . . $volume_usage\n"
                            . "Data volume size  . . $size\n"
                            . "Available space . . . $aggregate_size");
                }
            }

#if option to use random guarantee style, set it based on $i so types change on each vol
            if ($Options_hash->{'SPACE_GUARANTEE_STYLE'} =~ /random/i) {
                $guar_type
                    = $guarantee_types[($i % scalar(@guarantee_types))];
            } else {
                $guar_type = $Options_hash->{'SPACE_GUARANTEE_STYLE'};
            }

            $Log->comment("Adding data volume $volume "
                    . "to node $node aggregate $aggregate "
                    . "of guarantee type $guar_type");

            # Populate the volume entries in the defined_topology hash
            $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                ->{'size'} = $size;

            $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                ->{'aggregate'} = $aggregate;

            $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                ->{'root_volume'} = 0;

            $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                ->{'guar_type'} = $guar_type;

            $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                ->{'type'} = $type;

            #assign volumes to be dedupe or dedupe+compress if necessary
            # dedupe not supported on cache vols, so skip those
            if ($type eq 'flexvol') {
                if ($self->{'defined_topology'}->{$node}->{'dedupe_vol_count'}
                    > 0) {
                    $Log->comment("Assigning $volume to use dedupe");
                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'sis'} = 'dedupe';
                    $self->{'defined_topology'}->{$node}->{'dedupe_vol_count'}
                        --;
                } elsif ($self->{'defined_topology'}->{$node}
                    ->{'compress_vol_count'} > 0) {
                    $Log->comment(
                        "Assigning $volume to use dedupe+compression");
                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'sis'} = 'compress';
                    $self->{'defined_topology'}->{$node}
                        ->{'compress_vol_count'}--;
                } elsif (
                    $self->{'defined_topology'}->{$node}->{'inline_vol_count'}
                    > 0) {
                    $Log->comment(
                        "Assigning $volume to use dedupe+ inline compression"
                    );
                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'sis'} = 'inline';
                    $self->{'defined_topology'}->{$node}->{'inline_vol_count'}
                        --;
                }

                # flag volumes to hold QTREE_NUMBER of qtrees
                if ($self->{'defined_topology'}->{$node}->{'qtree_count'} > 0)
                {
                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'qtree_count'}
                        = $Options_hash->{'QTREE_NUMBER'};
                    $Log->comment(
                        "assigning qtrees: $Options_hash->{'QTREE_NUMBER'} to $volume"
                    );
                    $self->{'defined_topology'}->{$node}->{'qtree_count'}--;
                }
            }

            # find parent volume to use for cloning, if vol type is clone
            if ($type eq 'clone') {

                if ($self->{'defined_topology'}->{$node}
                    ->{'dedupe_clone_count'} > 0) {
                    $Log->comment("Assigning $volume to use dedupe");
                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'sis'} = 'dedupe';
                    $self->{'defined_topology'}->{$node}
                        ->{'dedupe_clone_count'}--;
                } elsif ($self->{'defined_topology'}->{$node}
                    ->{'compress_clone_count'} > 0) {
                    $Log->comment(
                        "Assigning $volume to use dedupe+compression");
                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'sis'} = 'compress';
                    $self->{'defined_topology'}->{$node}
                        ->{'compress_clone_count'}--;
                } elsif ($self->{'defined_topology'}->{$node}
                    ->{'inline_clone_count'} > 0) {
                    $Log->comment(
                        "Assigning $volume to use dedupe+ inline compression"
                    );
                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'sis'} = 'inline';
                    $self->{'defined_topology'}->{$node}
                        ->{'inline_clone_count'}--;
                }

                #get list of existing non-root volumes
                my @parent_vols;
                foreach my $vol (keys(%{$defined_volumes})) {
                    if (   ($defined_volumes->{$vol}->{'type'} eq 'flexvol')
                        && ($defined_volumes->{$vol}->{'root_volume'} == 0)) {
                        push(@parent_vols, $vol);
                    }
                }
                my $parent_vol = $parent_vols[$i % scalar(@parent_vols)];
                $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                    ->{'parent_vol'} = $parent_vol;
                $Log->comment("Adding clone of $parent_vol");
            }

            # find origin volume for cache if vol type is cache
            if ($type eq 'cache') {

                #get list of existing volumes
                my @origin_vols;
                foreach my $vol (keys(%{$defined_volumes})) {

      # only use flexvol and clone vols, and dont use root volumes for origins
                    if ((   $defined_volumes->{$vol}->{'type'}
                            =~ /flexvol|clone/i
                        )
                        && ($defined_volumes->{$vol}->{'root_volume'} == 0)
                        ) {
                        push(@origin_vols, $vol);
                    }
                }
                my $origin_vol = $origin_vols[$i % scalar(@origin_vols)];
                my $origin_vfiler
                    = $defined_volumes->{$origin_vol}->{'vfiler'};

                #find the vfiler interface to use in creating flexcache
                my @interfaces = keys(
                    %{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                            ->{$origin_vfiler}->{'interfaces'}
                        }
                );

                # find the first interfaces IP
                my $origin_vfiler_ip
                    = $self->{'defined_topology'}->{$node}->{'vfilers'}
                    ->{$origin_vfiler}->{'interfaces'}->{$interfaces[0]}->[0]
                    ->{'address'};

   # create the flexcache as origin_Vfiler_IP:volname (192.168.1.14:originvol)
                $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                    ->{'origin_vol'} = $origin_vfiler_ip . ":" . $origin_vol;

# flexcache space gaurantee type is partial, so force it for comparing topology
                $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                    ->{'guar_type'} = 'partial';
                $Log->comment(
                    "Adding cache of $origin_vol on $origin_vfiler");
            }
            if ($self->{'extent_volumes'} > 0) {
                $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                    ->{'extent'} = 'on';
                $self->{'extent_volumes'}--;
            } else {
                $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                    ->{'extent'} = 'off';
            }
            $volume_index++;
            $self->{'discovered_topology'}->{$node}->{'aggregates'}
                ->{$aggregate}->{'volume_usage'} += $size;
        }
    }

    # Spread the volumes across vfilers
    foreach my $node (@{$Options_hash->{'NODES'}}) {

        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};

        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);

        # Get the data volumes we just assigned to this nodes aggregates
        my @volumes = grep(/^$basename/, keys(%{$defined_volumes}));
        @volumes = natsort(@volumes);

        # If we are doing vfilers assign the volumes and interfaces to them
        if ($self->{'vfilers_per_node'} && ($type eq 'flexvol')) {
            foreach my $vfiler (@vfilers) {
                if ($vfiler eq 'vfiler0') { next; }
                $Log->comment("Assigning volumes to vfiler $vfiler");

                # If volumes_per_vfiler_assignment eq rnd get a randomized
                # value of volumes_per_vfiler
                #
                my $volumes_per_vfiler = $self->{'volumes_per_vfiler'};
                if (!$Options_hash->{'SHARED_VFILERS'}) {
                    $volumes_per_vfiler--;
                }

                if ($self->{'volumes_per_vfiler_assignment'} eq 'rnd') {
                    $volumes_per_vfiler = int(rand($volumes_per_vfiler));
                }
                $Log->debug("Volumes per vfiler: $volumes_per_vfiler");

                # Assign the data volumes to the defined vfiler
                foreach (1 ... $volumes_per_vfiler) {
                    my $volume = shift(@volumes);
                    if (!$volume) {
                        die config_result("Not enough volumes specified to "
                                . "create $volumes_per_vfiler volumes "
                                . "on vfiler $vfiler. Please increase "
                                . "VOLUME_COUNT");
                    }
                    $Log->comment(
                        "Adding $volume to node $node vfiler $vfiler");
                    push(
                        @{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                                ->{$vfiler}->{'volumes'}
                            },
                        $volume
                    );

                    # Update the defined volumes with the owning vfiler
                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'vfiler'} = $vfiler;

                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'share_count'} = 0;

                }
            }
        }

        # If we have left over volumes assign them to vfiler0
        if (@volumes) {
            my $defined_vfiler
                = $self->{'defined_topology'}->{$node}->{'vfilers'}
                ->{'vfiler0'};
            foreach my $volume (@volumes) {
                $Log->comment("Adding $volume to node $node vfiler vfiler0");
                push(@{$defined_vfiler->{'volumes'}}, $volume);

                # Update the defined volumes with the owning vfiler
                $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                    ->{'vfiler'} = 'vfiler0';
                $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                    ->{'share_count'} = 0;
            }

        }
    }
    $Log->exit();
} ## end sub assign_data_volumes

sub assign_vfiler_root_volumes {
    $Log->enter();

    my ($self, %options) = @_;
    my $size    = $options{'size'};
    my $vfilers = $options{'vfilers'};
    my $node    = $options{'node'};

    # Spread the volumes acrosss the specified aggregates
    $Log->comment("Adding root volumes to aggregates");

    my $aggregate_index = 0;
    my $volume_usage    = 0;

    # Get an array of the discovered aggregates
    my @discovered_aggregates = natsort(
        keys(%{$self->{'discovered_topology'}->{$node}->{'aggregates'}}));

    # loop through the discovered aggregates and create an array of our
    # defined aggregates
    my @defined_aggregates;
    foreach my $aggregate (@discovered_aggregates) {
        if (grep { $_ eq $aggregate }
            split(/,/, $Options_hash->{'AGGREGATES'})
            ) {
            push(@defined_aggregates, $aggregate);
        }
    }

    # Set the current aggregate and it's size
    my $aggregate = $defined_aggregates[$aggregate_index];
    $self->{'discovered_topology'}->{$node}->{'aggregates'}->{$aggregate}
        ->{'total'} =~ /(\d+)M/;
    my $aggregate_size
        = int($1 * ($Options_hash->{'MAX_AGGREGATE_USAGE'} / 100));

    # Loop though each vfiler assigning it's root volume to the current
    # aggregate
    foreach my $vfiler (@{$vfilers}) {

        # Check to make sure we don't exceed the current aggregates space.
        if ($volume_usage + $size > $aggregate_size) {

            # Switch to the next aggregate and grab it's size
            if ($defined_aggregates[$aggregate_index + 1]) {
                $aggregate_index++;
                $aggregate    = $defined_aggregates[$aggregate_index];
                $volume_usage = 0;
                $self->{'discovered_topology'}->{$node}->{'aggregates'}
                    ->{$aggregate}->{'total'} =~ /(\d+)M/;
                $aggregate_size = int(
                    $1 * ($Options_hash->{'MAX_AGGREGATE_USAGE'} / 100));
            } else {
                die config_result("Not enough aggregate space on "
                        . "$node:$aggregate when trying to calculate"
                        . "space for $vfiler\'s root volume\n"
                        . "Allocated space . . . $volume_usage\n"
                        . "Root volume size  . . $size\n"
                        . "Available space . . . $aggregate_size");
            }
        }

        # Create the volume name
        my $volume = 'v_' . $vfiler;
        $volume =~ s/-/_/g;
        $Log->comment("Adding $volume to node $node aggregate $aggregate");

        # Populate the volume entries in the defined_topology hash
        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}->{'size'}
            = $size;
        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
            ->{'aggregate'} = $aggregate;

        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
            ->{'vfiler'} = $vfiler;

        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
            ->{'share_count'} = 0;

        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
            ->{'root_volume'} = 1;

        #use volume guarantee for root volumes
        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
            ->{'guar_type'} = 'volume';

        #set root vol type to volume
        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}->{'type'}
            = 'flexvol';

        # Assign root volume for vfiler
        push(
            @{  $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                    ->{'volumes'}
                },
            $volume
        );
        $volume_usage += $size;
    }
    $Log->exit();
} ## end sub assign_vfiler_root_volumes

sub enable_branch_cache {
    $Log->enter();

    my ($self, %options) = @_;

    my $vfiler = $options{'vfiler'};
    my $node   = $options{'node'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    # Enable branch cache
    $Log->comment("Enabling branch cache on vfiler $vfiler on node $node");

    my %data;
    $data{'cifs.smb2_1.branch_cache.enable'} = 'on';

    foreach my $key (keys(%data)) {
        try {

            $Response = $Zapi->options_set(
                'name'  => $key,
                'value' => $data{$key}
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    $Log->comment("Setting the cifs server secret");

    # Set the secret

    try {
        $Response = $Zapi->cifs_branchcache_set_key(
            "server-secret" => "timtebowrocks211");
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->exit();

} ## end sub enable_branch_cache

sub enable_smb2 {
    $Log->enter();

    my ($self, %options) = @_;

    my $vfiler = $options{'vfiler'};
    my $node   = $options{'node'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    my %data;
    $data{'cifs.smb2.enable'} = 'on';

    foreach my $key (keys(%data)) {
        try {

            $Response = $Zapi->options_set(
                'name'  => $key,
                'value' => $data{$key}
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    $Log->exit();

} ## end sub enable_smb2

sub create_spn {
    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    my $vfiler = $options{'vfiler'};
    my $node   = $options{'node'};

    #Get the actual cifs name of the vfiler in question.

    my $cifs_server
        = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
        ->{'cifs_server'};
    $Log->comment("Cifs server is $cifs_server");

    my @vfiler_interfaces = keys(
        %{  $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                ->{'interfaces'}
            }
    );
    $Log->comment("Interfaces @vfiler_interfaces");

    foreach my $interface (@vfiler_interfaces) {
        my $defined_address
            = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
            ->{'interfaces'}->{$interface};
        foreach my $ip_info (@{$defined_address}) {
            my $vfiler_ip = $ip_info->{'address'};
            $Log->comment("vfiler IP is $vfiler_ip");
            my $resolved_address = get_name_by_ip('ip' => $vfiler_ip);

            if ($resolved_address->{'ERROR'}) {
                die config_result(
                    "IP address $resolved_address does not resolve to a name"
                );
            }

            $resolved_address->{'PTR'} =~ /([^\.]+)\.(.+)/;

            my $spn_name = $1;
            $spn_name .= ".$Options_hash->{'CIFS_DOMAIN'}";

            $self->set_spn(
                'cifs_server' => $cifs_server,
                'spn_name'    => $spn_name,
            );
        }
    }

    $Log->exit();
} ## end sub create_spn

sub enable_vfiler_flexcache {
    $Log->enter();

    my ($self, %options) = @_;
    my $node = $options{'node'};

#make sure flexcache options are enabled licensed_feature.flexcache_nfs.enable

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};
    my $Response;
    my $flexcache_enable;

    eval {
        $Response
            = $Zapi->options_get(
            'name' => 'licensed_feature.flexcache_nfs.enable');
        my $Return = $Response->get_parsed_output();

        # Store its value
        $flexcache_enable = $Return->[0]->{'value'};
        $self->{'discovered_topology'}->{$node}->{'options'}
            ->{'licensed_feature.flexcache_nfs.enable'} = $flexcache_enable;

        $Log->comment(
            "licensed_feature.flexcache_nfs.enable: $flexcache_enable");
    };

    # check that licensed_feature.flexcache... is valid for current build
    my $option_valid = $self->is_option_valid(
        'node'   => $node,
        'option' => 'licensed_feature.flexcache_nfs.enable'
    );

    if ($option_valid && $flexcache_enable ne 'on') {
        eval {
            $Zapi->options_set(
                'name'  => 'licensed_feature.flexcache_nfs.enable',
                'value' => 'on'
            );
        };
    }

    my $discovered_vfilers
        = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $discovered = keys(%{$discovered_vfilers});
    $Log->comment("Enabling flexcache on $discovered vfilers");

    my @vfilers = keys(%{$discovered_vfilers});
    @vfilers = natsort(@vfilers);

    # enable flexcache and flexcache.access on each vfiler
    foreach my $vfiler (@vfilers) {

        # Set the vfiler context for the next ZAPI call
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
        try {
            $Zapi->options_set(
                'name'  => 'flexcache.enable',
                'value' => 'on'
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
        try {
            $Zapi->options_set(
                'name'  => 'flexcache.access',
                'value' => 'all'
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    # Return to vfiler0 context
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };
}    #end sub enable_vfiler_flexcache

sub enable_autosupport() {
    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();
    my $node = $options{'node'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my %data;
    $data{'autosupport.enable'}            = 'on';
    $data{'autosupport.support.enable'}    = 'on';
    $data{'autosupport.support.transport'} = 'https';

    foreach my $key (keys(%data)) {
        try {

            $Response = $Zapi->options_set(
                'name'  => $key,
                'value' => $data{$key}
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            $Log->warn($Exception->text());
        };
    }

}    #end sub enable_autosupport

sub enable_bootargs() {

    # enable the wafl boot args
    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $Host   = host($node);
    my %fields = $Host->param_fields();
    my $Cli    = $apiset{'cli'}{$node};
    my $Zapi   = $apiset{'zapi'}{$node};

    # if filer has 4 or less processors, nthreads will be 1
    my $nthreads = 1;

    #query the system to find number of processors
    my $Result      = $Zapi->system_get_info();
    my $system_info = $Result->get_parsed_output();
    my $num_processors
        = $system_info->[0]->{'system-info'}->[0]->{'number-of-processors'};
    $Log->comment("Found $num_processors processors");

    # if filer has more then 4 processors, set nthreads to 20
    if ($num_processors >= 4) {
        $nthreads = 20;
    }
    if ($num_processors >= 8) {
        $nthreads = 20;
    }
    my @commands;

    # if model is vsa, set nthreads to 1
    if ($fields{'TYPE'} =~ /VSA/i) {
        $Log->comment("Model is VSA so unsetting raidio_nthreads");
        push(@commands, "priv set -q diag;bootargs unset raidio_nthreads");
        push(@commands, "priv set -q diag;bootargs unset bigio_nthreads");
    } else {
        push(@commands,
            "priv set -q diag;bootargs set raidio_nthreads $nthreads");
        push(@commands,
            "priv set -q diag;bootargs set bigio_nthreads $nthreads");
    }
    push(@commands,
        "priv set -q diag;bootargs set wafl-enable-incremental-cksum? true");
    push(@commands,
        "priv set -q diag;bootargs set wafl-metadata-page-protect? true");
    push(@commands,
        "priv set -q diag;bootargs set wafl-disable-metadata-consistency-check? false"
    );
    foreach my $command (@commands) {
        try {
            $Response = $Cli->execute_raw_command(command => $command);
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

}    #end sub enable_autosupport

sub create_syncmirror() {

    # create an aggr sync mirror
    $Log->enter();

    my ($self, %options) = @_;
    my $aggregate = $options{'aggregate'};

    my $node = $options{'node'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    try {
        $Response = $Zapi->aggr_mirror('aggregate' => $aggregate);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

}    #end sub create_syncmirror

sub create_qtrees() {

    # create # of qtrees based on the defined qtree_count for the vol
    $Log->enter();

    my ($self, %options) = @_;
    my $volume = $options{'volume'};

    my $node = $options{'node'};
    my $qtree_number
        = $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
        ->{'qtree_count'};

    my $Zapi = $apiset{'zapi'}{$node};
    $Log->comment("Creating qtrees\ncount: $qtree_number volume: $volume");
    for (my $i = 0; $i < $qtree_number; $i++) {
        my $qtree_name = 'qtree_' . $i;
        try {
            $Response = $Zapi->qtree_create(
                'volume' => $volume,
                'qtree'  => $qtree_name
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

}    #end sub create_syncmirror

# reboot the node
sub reboot_node() {
    $Log->enter();

    my ($self, %options) = @_;
    my $node = $options{'node'};

    my $node_obj = NACL::C::Node->new(node => $node);
    try {
        $node_obj->reboot();
    }
    catch NATE::BaseException with {
        my $exception = shift;
        my $text      = $exception->text();
        $Log->comment("Exceptions is: $text");
        die config_result($text);
    };

    my $Transit_obj = NACL::Transit->new(name => $node_obj->hostrec());
    my $retries     = 24;
    my $to          = $Transit_obj->get_state(timeout => 300);
    $Log->comment("Original state is: $to");
    while (
        !"NACL::Transit::StateDefinitions::$to"->isa(
            "NACL::Transit::StateDefinitions::UP")
        and $retries > 0
        ) {
        $to = $Transit_obj->get_state();
        $retries--;
        sleep(5);
    }

    # If autoboot set to YES, The machine will go to loader prompt
    # Using transit to bring UP the filer
    if ($retries == 0) {
        $to = $Transit_obj->change_state(to => "CLI");
    }
    if (!"NACL::Transit::StateDefinitions::$to"->isa(
            "NACL::Transit::StateDefinitions::UP")
        ) {
        die config_result("Reboot failed");
    }
}    #end sub reboot_node

sub find_disks() {

#    $Log->enter();
#
#    my ($self, %options) = @_;
#    my $node = $options{'node'};
#
#    $Options_hash = $Options->get_hash();
#    my @spare_disks;
#    my @zeroed_disks;
#    my $Zapi = $apiset{'zapi'}{$node};
#    my $Result = $Zapi->disk_list_info();
#    my $disk_info = $Result->get_parsed_output();
#    my @disks = @{$disk_info->[0]->{'disk-details'}->[0]->{'disk-detail-info'}};
#    foreach my $disk (@disks){
#        if ($disk->{'node'} => $$node) {
#            if ($disk->{'raid-state'}) {
#                $Log->comment("$disk->{'name'} is spare");
#                push ( @spare_disks, $disk->{'name'});
#            }
#            if ($disk->{'is-zeroed'}) {
#                $Log->comment("$disk->{'name'} is zeroed");
#                push ( @zeroed_disks, $disk->{'name'});
#            }
#        }
#    }

}    #end sub find_disks

sub convert_size($) {

    #returns input size in megabytes
    my $input_size = shift;
    my ($size, $modifier) = $input_size =~ /(\d+)(\w?)/;
    $Log->comment("converting size: $size mod: $modifier");
    my $size_conversion = 1;
    $size_conversion = 1024 * 1024 if $modifier =~ /t/i;
    $size_conversion = 1024        if $modifier =~ /g/i;
    $size_conversion = 1           if $modifier =~ /m/i;
    $size_conversion = 1 / 1024    if $modifier =~ /k/i;
    return $size_conversion * $size;
} ## end sub convert_size($)

sub define_shares {

    $Log->enter();

    my ($self, %options) = @_;
    my $script      = $options{'script'};
    my $share_index = $options{'share_index'};
    my $recipe      = $options{'recipe'};
    my $server;
    my @nodes
        = grep(/$Options_hash->{'NODE_REGEX'}/, @{$Options_hash->{'NODES'}});
    if (!@nodes) {
        die config_result("Unable to find any nodes using the "
                . "NODE_REGEX $Options_hash->{'NODE_REGEX'} against "
                . Dumper($Options_hash->{'NODES'}));
    }

    my %shares;
    my $vfiler_count = 0;
    foreach my $node (@nodes) {
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_shares = $topology_database->{$node}->{'shares'};
        my $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};
        my $discovered_shares
            = $self->{'discovered_topology'}->{$node}->{'shares'};

        # Count the existing ontap shares
        my @vfilers = keys(%{$discovered_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            $shares{$node}{$vfiler} = @ontap_shares;
        }

        # Count the existing discovered shares
        my @ds = keys(%{$discovered_shares});
        @ds = natsort(@ds);
        foreach my $s (@ds) {
            if (!$s) { next; }
            my $owner = $discovered_shares->{$s}->{'owning-vfiler'};
            if (!$owner) {
                die config_result(
                    "Unable to get owning vfiler from discovered share $s\n"
                        . Dumper($discovered_shares->{$s}));
            }

            # Ignore shares we created from previous runs. The existing
            # shares will be redfined but filer_setup will be set to 0
            # so the share is not recreated. This way if the test fail
            # for some reason it can pick up whre it left off.
            if ($s =~ /^s_$script/) { next; }

            # Add the count to the owning vfiler
            $shares{$node}{$owner} += 1;
        }

        # Count the to be created shares
        foreach my $vfiler (keys(%{$defined_vfilers})) {

            $vfiler_count++;

            # Count the to be created ontap shares
            $shares{$node}{$vfiler} = @ontap_shares;

            # Count all shares that belong to the current vfiler
            # and are not already created and not a data share
            # This is basically to pickup the volume shares
            while (my $s = each %{$defined_shares}) {
                my $owner = $defined_shares->{$s}->{'vfiler'};
                if (!$owner) {
                    die config_result(
                        "Unable to get owning vfiler from defined share\n"
                            . Dumper($defined_shares->{$s}));
                } elsif ($vfiler ne $owner) {
                    next;
                } elsif (
                    grep {
                        $_ eq $s
                    }
                    keys(%{$discovered_shares->{$node}->{'shares'}})
                    ) {
                    next;
                } elsif ($s =~ /^s_$script/) {
                    next;
                }
                $shares{$node}{$vfiler} += 1;
            }
        }
    }

    # Total all the shares found
    my $total_existing_shares = 0;
    foreach my $n (keys(%shares)) {
        foreach my $v (keys(%{$shares{$n}})) {
            $total_existing_shares += $shares{$n}{$v};
        }
    }

    if ($total_existing_shares >= $Options_hash->{'SHARE_COUNT'}) {
        die config_result("Existing shares ($total_existing_shares) is "
                . "greater or equal to the specified SHARE_COUNT "
                . "($Options_hash->{'SHARE_COUNT'}) ");
    }

    my $new_shares = $Options_hash->{'SHARE_COUNT'} - $total_existing_shares;
    my $shares_per_vfiler           = int($new_shares / $vfiler_count);
    my $shares_per_vfiler_remainder = $new_shares % $vfiler_count;
    my $shares_per_client           = 0;
    my $shares_per_client_remainder = 0;

    $self->{'existing_shares'}            = $total_existing_shares;
    $self->{'new_shares'}                 = $new_shares;
    $self->{'shares_per_vfiler'}          = $shares_per_vfiler;
    $self->{'shares_per_vfiler_reminder'} = $shares_per_vfiler_remainder;

    my $shares_per_volume = 0;
    if ($Options_hash->{'SHARES_PER_VOLUME'}) {
        $shares_per_volume = $Options_hash->{'SHARES_PER_VOLUME'};
    } else {
        $shares_per_volume
            = int($new_shares / $Options_hash->{'VOLUME_COUNT'});
        if ($new_shares / $Options_hash->{'VOLUME_COUNT'}) {
            $shares_per_volume++;
        }
    }

    $self->{'shares_per_volume'} = $shares_per_volume;

    #<<<
    my $summary = "Requested shares . . . . . . . . . . . . $Options_hash->{'SHARE_COUNT'}\n";
    $summary .= "Target nodes . . . . . . . . . . . . . . " . @nodes . "\n";
    $summary .= "Target vfilers   . . . . . . . . . . . . $vfiler_count\n";
    $summary .= "Existing shares  . . . . . . . . . . . . $total_existing_shares\n";
    $summary .= "Shares to create   . . . . . . . . . . . $new_shares\n";
    $summary .= "Shares per vfiler  . . . . . . . . . . . $shares_per_vfiler\n";
    $summary .= "Shares per vfiler remainder  . . . . . . $shares_per_vfiler_remainder\n";
    $summary .= "Shares per volume  . . . . . . . . . . . $shares_per_volume\n";

    $Log->comment($summary);

    # Spread the shares across the nodes, vfilers and volumes
    $Log->comment("Adding shares to vfilers");
    foreach my $node (@nodes) {
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $discovered_shares
            = $self->{'discovered_topology'}->{$node}->{'shares'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};

        # Initialize the index for for the addresses for each vfiler's data
        # interface
        my $interface_index = 0;
        my %address_indexes;
        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
                $address_indexes{$interface} = 0;
            }
        }

        foreach my $vfiler (@vfilers) {
            if (!$vfiler) {
                die script_result(
                    "Blank vfiler detected in defined vfilers hash");
            }

            my $current_shares_per_vfiler = $shares_per_vfiler;
            if ($vfiler eq 'vfiler0' && $node eq $nodes[0]) {
                $current_shares_per_vfiler += $shares_per_vfiler_remainder;
            }

            my $vfiler_volumes
                = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                ->{'volumes'};
            my @volumes;
            my $volume;
            my $share_count = 0;
            $Log->comment(
                "Adding $current_shares_per_vfiler shares to vfiler $vfiler on node $node"
            );
            foreach (1 ... $current_shares_per_vfiler) {

                # Create the share name
                my $share = 's_' . $script . '_' . $share_index;
                $share_index++;

                my $filer_setup = 1;
                my $path;
                my $vf;
                my $vol;

                # If the share exists use it otherwise define a new share
                if (grep { $_ eq $share } keys(%{$discovered_shares})) {
                    $Log->debug("Found existing share $share\n"
                            . Dumper($discovered_shares->{$share}));
                    $path = $discovered_shares->{$share}->{'mount-point'};
                    $path =~ /^\/vol\/([^\/]+)\/.+/;
                    $vol = $1;
                    if (!$vol) {
                        die script_result("Unable to determine the volume "
                                . "from the share path $path");
                    }
                    $vf = $discovered_shares->{$share}->{'owning-vfiler'};
                    $filer_setup = 0;
                    delete($discovered_shares->{$share});

                } else {

                    # Get the volume name. If there are no more volumes in the
                    # array repopulate the array
                    if (!@volumes) {
                        foreach my $v (natsort(@{$vfiler_volumes})) {
                            my $root_volume
                                = $defined_volumes->{$v}->{'root_volume'};
                            if ($root_volume) { next; }
                            push(@volumes, $v);
                        }
                    }
                    if (!$volume) {
                        $volume = shift(@volumes);
                        if (!$volume) {
                            die script_result(
                                      "No available data volumes found on "
                                    . "node $node for vfiler $vfiler "
                                    . "Try increasing VOLUME_COUNT");
                        }
                    }
                    $vf  = $vfiler;
                    $vol = $volume;
                    if ($Options_hash->{'SHARE_PATH_CONFIG'} eq 'single') {
                        $path = '/vol/' . $vol;
                    } elsif (
                        $Options_hash->{'SHARE_PATH_CONFIG'} eq 'multiple') {
                        $path = '/vol/' . $vol . '/' . $share;
                    }
                    $Log->debug(
                        "Defining new share $share on vfiler $vf with a path of $path"
                    );
                }

                if (!$vf) {
                    die script_result("Unable to determine the vfiler for "
                            . "share $share on volume $vol\n"
                            . Dumper($discovered_shares->{$share}));
                }

                # Store the the volume name with the share
                $topology_database->{$node}->{'shares'}->{$share}->{'volume'}
                    = $vol;

                # Store the share path
                $topology_database->{$node}->{'shares'}->{$share}->{'path'}
                    = $path;

                # Store the vfiler name with the share
                $topology_database->{$node}->{'shares'}->{$share}->{'vfiler'}
                    = $vf;

                # Decide when we create the share
                $topology_database->{$node}->{'shares'}->{$share}
                    ->{'filer_setup'} = $filer_setup;

                # Don't use qtrees
                $topology_database->{$node}->{'shares'}->{$share}->{'qtree'}
                    = 0;

                push(
                    @{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                            ->{$vf}->{'shares'}
                        },
                    $share
                );

                $topology_database->{$node}->{'shares'}->{$share}->{'node'}
                    = $node;
                ($server, undef) = $self->assign_cifs_server(
                    'node'            => $node,
                    'vfiler'          => $vf,
                    'address_indexes' => \%address_indexes,
                    'interface_index' => 0,
                );

                $Log->comment("Assigning CIFS server $server for "
                        . "share $share on vfiler $vfiler node "
                        . "$node");

                $topology_database->{$node}->{'shares'}->{$share}->{'server'}
                    = $server;

                $share_count++;
                if ($share_count == $current_shares_per_vfiler) {
                    $volume = undef;
                }
            }
        }
    }

    $Log->exit();
    return $share_index;
} ## end sub define_shares

1;
