#Hey emacs, this is basically -*- perl -*-
# $Id: $

## @summary     CIFS Topology Library
##
## @author      bellizzi@netapp.com
##
## @pod here

=head1 NAME

Topology::NAS::CIFS:7M - CIFS Stress Scalability test Topology library for 7Mode ONTAP

=head1 SYNOPSIS

    See Topology.pm

=head1 DESCRIPTION

More detailed documentation will be available at
http://wikid.netapp.com/w/Stress-Scalability_QA/TestManagerLite
http://wikid.netapp.com/w/Stress-Scalability_QA/Topology

=head1 TODO

=head1 SEE ALSO

Options TestManagerLite Topology

=head1 BUGS

Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  David Bellizzi (bellizzi@netapp.com)
    
=head1 COPYRIGHT

Copyright 2011, Network Appliance., All Rights Reserved

=cut

package Topology::NAS::CIFS::7M;

use strict;

use Tharn qw($Log connect config_result script_result host say snooze);
$Log->comment("Starting " . __FILE__);
use base qw(Topology::NAS::CIFS);
use Module::Common qw(validate_mixed);
use Params::Validate
    qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Data::Dumper qw(Dumper);
use Options;
use File::Basename;
use Utils::DNS qw(get_name_by_ip);
use CIFS::Client;
use Net::IP qw(:PROC);
use CIFS::Setup qw(cifs_setup);
use Sort::Key::Natural qw(natsort);
use filer qw(filer_vfiler_command filer_vfiler_context_simulate);
use POSIX qw(strftime ceil);
use NetAddr::IP::Lite;
use Utils::HostManager::Helper;
use NACL::APISet;
use NACL::C::Volume;
use NACL::C::Node;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::FeatureNotLicensedException qw(:try);

use if ($^O ne "MSWin32"), 'DBM::Deep';

$Data::Dumper::Sortkeys = 1;

#Global variables. Yeah I know they are bad but so what.
our (@users, @userShares, $homedir_per_volume, @userDefinedShares);

my $Options = new Options;
$Options->process();

sub new {
    my $invocant = shift;
    my $class    = ref($invocant) || $invocant;
    my $self     = {@_,};
    $Log->enter();
    bless($self, $class);
    $self->{reserve_clients} = 0;
    $Log->exit();
    return $self;
} ## end sub new

my ($Options_hash, %apiset, $Response, %conn, $topology_database);
my @ontap_shares = ('C$', 'ETC$', 'ONTAP_ADMIN$', 'IPC$');

sub define_topology {
    $Log->enter();

    my $self = shift;

    # Execute the one from the parent
    $self->SUPER::define_topology(@_);

    $Options_hash = $Options->get_hash();
    $Log->comment(Dumper($Options_hash));

    if ($Options_hash->{'CLIENT_ASSIGNMENT'} eq 'io'
        && (   !$Options_hash->{'TEST_MAIN_DURATION'}
            || !$Options_hash->{'TEST_MAIN_SLEEP'})
        ) {
        die script_result("TEST_MAIN_DURATION and/or "
                . "TEST_MAIN_SLEEP are required\n"
                . Dumper($Options_hash));
    }

    # Find calling script's name
    my $script = fileparse($0, qr/\.[^.]*/);
    if ($Options_hash->{'SCRIPT_INDEX'}) {
        $script .= "_$Options_hash->{'SCRIPT_INDEX'}";
    }
    $self->{'testname'} = $script;

    $topology_database
        = DBM::Deep->new("$Options_hash->{'TOP_LOGDIR'}/$script.db");
    $self->{'topology_database'} = $topology_database;

    # Wait for other scripts to start if needed
    if ($Options_hash->{'WAIT_FOR_FILE'}) {
        $Log->comment("Waiting for $Options_hash->{'WAIT_FOR_FILE'}");
        $self->wait_for_file('file' => $Options_hash->{'TOP_LOGDIR'} . '/'
                . $Options_hash->{'WAIT_FOR_FILE'});
    }

    $self->{'clients'} = 0;

    # Retrieve the clients from the client file or ndispd
    my @clients
        = retrieve_clients('script' => $script, 'options' => $Options_hash);
    $self->{'clients'} = \@clients;

    # Calculate MAX_CLIENTS
    $self->{'max_clients'} = get_max_clients();

    # Make 7Mode and ZAPI APISet connections for all nodes and store
    # for reuse
    foreach my $node (@{$Options_hash->{'NODES'}}) {

        $Log->comment("Creating CLI connection object for $node");
        my $Host = host($node);
        $apiset{'cli'}{$node} = NACL::APISet->new(
            hostobj   => $Host,
            category  => "Node",
            interface => "CLI",
            set       => "7Mode",
            cdef_path => $Options_hash->{'CDEF_PATH'}
        );

        $apiset{'cli'}{$node}
            ->set_timeout(timeout => $Options_hash->{'APISET_TIMEOUT'});

        $Log->comment("Creating ZAPI connection object for $node");
        $apiset{'zapi'}{$node} = NACL::APISet->new(
            hostobj   => $Host,
            category  => "Node",
            interface => "ZAPI",
            set       => "7Mode",
            cdef_path => $Options_hash->{'CDEF_PATH'}
        );

        $apiset{'zapi'}{$node}
            ->set_timeout(timeout => $Options_hash->{'APISET_TIMEOUT'});

        $Log->comment("Creating Tharn connection for $node");
        $conn{$node} = connect('#rsh', '-nologin', -timeout, 300);

        $topology_database->put($node, {});

    }

    my $version_mgr = $apiset{'cli'}{${$Options_hash->{'NODES'}}[0]}
        ->get_version_manager();

    # Attribute can be any of branch; release; build_options; build_path;
    # change_number; wafl_version; node_version
    my $release
        = $version_mgr->get_version_attribute('attribute' => 'release');
    $self->{'release'}
        = $version_mgr->map_release_to_codename(release => $release);

    my @targets = (
        'aggregates',    'cifs',   'interfaces', 'licenses',
        'qtrees',        'quotas', 'shares',     'vfilers',
        'vfiler_limits', 'volumes'
    );

    if (   $Options_hash->{'PRE_TEST_CLEANUP'}
        && $Options_hash->{'PRE_TEST_CLEANUP'} eq 'discovered') {

        $self->filer_cleanup_discovered();
        snooze 1;
    }

    $self->discover_topology('targets' => \@targets);

    if ($Options_hash->{'VFILER_COUNT'}
        && ($Options_hash->{'VFILER_COUNT'} * 2)
        > $Options_hash->{'VOLUME_COUNT'}
        && !$Options_hash->{'SHARED_VFILERS'}) {
        die config_result('Not enough volumes specified for all vfilers');
    }

    if (!$Options_hash->{'AGGREGATES'}) {
        die script_result('AGGREGATES is a required parameter');
    }

    if (!@{$Options_hash->{'DATA_INTERFACES'}}) {
        die script_result('DATA_INTERFACES is a required parameter');
    }

    if (!$Options_hash->{'VOLUME_COUNT'}) {
        die config_result('VOLUME_COUNT is a required parameter');
    }

    if (!$Options_hash->{'NODES'}) {
        die config_result('NODES(FILER) is a required parameter ');
    }

    # Get the IO configs and store them in $self
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'io'
        && (   $Options_hash->{'IO_LOAD_CONFIG'}
            || $Options_hash->{'IO_TOOLS_CONFIG'})
        ) {
        $self->get_io_configs();
    }

    # Store the relevant info in the topo
    $self->{'vfiler_count'}    = $Options_hash->{'VFILER_COUNT'};
    $self->{'aggregates'}      = $Options_hash->{'AGGREGATES'};
    $self->{'data_interfaces'} = $Options_hash->{'DATA_INTERFACES'};

    # Store the current volume count as the total
    $self->{'volume_count'} = $Options_hash->{'VOLUME_COUNT'};

    # Split the volume count per node
    $self->{'volumes_per_node'}
        = int($self->{'volume_count'} / @{$Options_hash->{'NODES'}});

    if ($self->{'volumes_per_node'} < 1) {
        $self->{'volumes_per_node'} = 1;
    }

    # Force creation of the vfiler0 key
    foreach my $node (@{$Options_hash->{'NODES'}}) {
        $self->{'defined_topology'}->{$node}->{'vfilers'}->{'vfiler0'} = {};
    }

    # If vfilers, spread the vfilers across the nodes
    $self->{'vfilers_per_node'}   = 0;
    $self->{'volumes_per_vfiler'} = 0;
    $self->{'data_volume_count'}  = $Options_hash->{'VOLUME_COUNT'};
    if ($Options_hash->{'VFILER_COUNT'}) {

        my @nodes = @{$Options_hash->{'NODES'}};
        $self->{'vfilers_per_node'} = int(
            $Options_hash->{'VFILER_COUNT'} / @{$Options_hash->{'NODES'}});
        foreach my $node (@{$Options_hash->{'NODES'}}) {
            my @vfilers;
            my $vfiler_index = 1;
            foreach my $v (1 ... $self->{'vfilers_per_node'}) {

                # Create the vfiler keys
                my $vfiler = $node . '-vf' . $vfiler_index;
                push(@vfilers, $vfiler);
                $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                    = {};
                $vfiler_index++;
            }

            $self->assign_vfiler_root_volumes(
                'node'    => $node,
                'size'    => 100,
                'vfilers' => \@vfilers,
            );

        }

        # Determine the max number of volumes each vfiler gets and if the
        # assignment method. rnd = random sym = symmetrical

        $self->{'data_volume_count'} = $Options_hash->{'VOLUME_COUNT'}
            - $Options_hash->{'VFILER_COUNT'};
        if ($Options_hash->{'VOLUMES_PER_VFILER'} =~ /(\w+):(\d+)/) {
            $self->{'volumes_per_vfiler_assignment'} = $1;
            $self->{'volumes_per_vfiler'}            = $2;
        } elsif ($Options_hash->{'VOLUMES_PER_VFILER'} =~ /(\w+)/) {
            $self->{'volumes_per_vfiler_assignment'} = $1;

            # Account for the root volume
            $self->{'volumes_per_vfiler'} = 1;

            # Make sure all vfilers have a data volume
            my $data_volumes_per_vfiler = int($self->{'data_volume_count'}
                    / ($self->{'vfiler_count'} + @nodes));
            if (!$data_volumes_per_vfiler) { $data_volumes_per_vfiler = 1; }
            $self->{'volumes_per_vfiler'} += $data_volumes_per_vfiler;
        } else {
            $self->{'volumes_per_vfiler_assignment'} = 'sym';
            $self->{'volumes_per_vfiler'}
                = $Options_hash->{'VOLUMES_PER_VFILER'};
        }

        my $vfiler_volumes
            = $self->{'volumes_per_vfiler'} * $self->{'vfiler_count'};
        my $required_volumes = $vfiler_volumes + @nodes;
        if ($required_volumes > $self->{'volume_count'}) {
            die script_result("The number of required volumes "
                    . "($required_volumes) exceeds VOLUME_COUNT. "
                    . "Please increase VOLUME_COUNT.");
        }

        if ($self->{'volumes_per_vfiler'} < 2) {
            my $required_volumes
                = ($Options_hash->{'VFILER_COUNT'} * 2) + @nodes;
            die script_result("Not enough volumes for all vfilers. "
                    . "$Options_hash->{'VOLUME_COUNT'} was specified but "
                    . "$required_volumes are required. There must be at least "
                    . "2 volumes per vfiler. Please increase VOLUME_COUNT");
        }
    }

    if ($Options_hash->{'LOCAL_GROUP_COUNT'}) {
        $self->{'local_group_count'} = $Options_hash->{'LOCAL_GROUP_COUNT'};

        # Split the localgroup count per node
        $self->{'local_groups_per_node'}
            = int($self->{'local_group_count'} / @{$Options_hash->{'NODES'}});

        # Split the localgroup count per vfiler
        if ($self->{'vfilers_per_node'} > 0) {
            $self->{'local_groups_per_vfiler'}
                = int($self->{'local_groups_per_node'}
                    / $self->{'vfilers_per_node'});
        } else {
            $self->{'local_groups_per_vfiler'}
                = $self->{'local_groups_per_node'};
        }
    }

    my $basename = 'v_' . $self->{'testname'};
    my $size     = $Options_hash->{'VOLUME_SIZE'};

    $self->assign_data_volumes(
        'count'    => $self->{'data_volume_count'},
        'basename' => $basename,
        'size'     => $size
    );

    # Parse address ranges
    if (!$Options_hash->{'DESTIP'}) {
        die config_result('DESTIP is not defined');
    }

    $Log->comment("Parsing DESTIP");
    foreach my $spec (split(/,/, $Options_hash->{'DESTIP'})) {
        if ($spec =~ /^(\d+);([^;]+);(.+)/) {

            my $index     = $1;
            my $node      = ${$Options_hash->{'NODES'}}[$index];
            my $interface = $2;
            my $ranges    = $3;
            foreach my $range (split(/:/, $ranges)) {
                $Log->comment(
                    "Found IP range $range for $interface on $node");
                my $Host             = host($range);
                my %host_file_params = $Host->param_fields();
                $self->{'defined_topology'}->{$node}->{'addresses'}
                    ->{$interface}->{$range} = \%host_file_params;
            }
        } else {
            die config_result("DESTIP must be in the format of\n"
                    . "<node index>;<interface>;<range>[:<range>],"
                    . "<node index>;<interface>;<range>[:<range>]\n"
                    . "Where range is a file in your nate_hosts "
                    . "dir that contains at least\n"
                    . "STARTIP\n"
                    . "GATEWAY\n"
                    . "NETMASK\n"
                    . "IPCOUNT");
        }
    }

    # Spread the addresses across vfilers
    foreach my $node (@{$Options_hash->{'NODES'}}) {

        my $discovered_interfaces
            = $self->{'discovered_topology'}->{$node}->{'interfaces'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_addresses
            = $self->{'defined_topology'}->{$node}->{'addresses'};
        my @interfaces = keys(%{$defined_addresses});
        if (!@interfaces) {
            die script_result(
                "No interfaces defined in DESTIP for node $node");
        }
        my %addresses;
        my %ranges;
        foreach my $interface (@interfaces) {

            foreach
                my $range (natsort(keys(%{$defined_addresses->{$interface}})))
            {
                my $start_ip
                    = $defined_addresses->{$interface}->{$range}->{'STARTIP'};
                my $ip_count
                    = $defined_addresses->{$interface}->{$range}->{'IPCOUNT'};
                my $ip_obj = new Net::IP("$start_ip  \+  " . ($ip_count - 1))
                    || die config_result("Failed to create ip object: $@");
                do {
                    my $ip = $ip_obj->ip();
                    push(@{$addresses{$interface}{$range}}, $ip);
                    $ranges{$ip} = $range;
                } while (++$ip_obj);
            }

        }

        $Log->comment("Adding addresses to node $node");

        # Assign addresses to vfiler0 data interfaces if needed
        my @assigned_ranges;

        # loop through each interface
        foreach my $interface (keys(%{$discovered_interfaces})) {
            $Log->comment(
                "Adding addresses to node node $node interface $interface");

            # Check both v4 and v6 addresses
            foreach my $version ('ipv4_info', 'ipv6_info') {
                my $address;

                # Are there addresses in this versions address list?
                if (grep { $_ eq $version }
                    @{$discovered_interfaces->{$interface}->{$version}}) {

                    # Get the list of addresses for this interface
                    foreach my $info (
                        @{$discovered_interfaces->{$interface}->{$version}}) {

                        $address = $info->{'ip'};

                        my $autoconf = $info->{'autoconf'};

                        # Don't use autoconf IPV6 addresses
                        if ($version =~ /ipv6/ && $autoconf) {
                            $Log->comment(
                                "Skipping IPV6 autoconf address $address");
                            next;
                        }

                        my $version_name = uc($version);
                        $version_name =~ s/_INFO//g;
                        $Log->comment("Adding discovered $version_name "
                                . "  address $address to node $node "
                                . "vfiler vfiler0 interface $interface");

                        if ($version =~ /ipv6/) {
                            $address = ip_expand_address($address, 6);
                        }

                        if ($ranges{$address}) {
                            my $range = $ranges{$address};
                            my $index = $self->get_array_index(
                                'array' => \@{$addresses{$interface}{$range}},
                                'value' => $address
                            );
                            if (${$addresses{$interface}{$range}}[$index]) {
                                $Log->comment(
                                    "Removing $address from the list of available "
                                        . "addresses for interface $interface"
                                );
                                splice(@{$addresses{$interface}{$range}},
                                    $index, 1);
                            }
                        }
                        if ($address) {
                            push(
                                @{  $self->{'defined_topology'}->{$node}
                                        ->{'vfilers'}->{'vfiler0'}
                                        ->{'interfaces'}->{$interface}
                                    },
                                {   'address' => $address,
                                    'range'   => $ranges{$address}
                                }
                            );
                            push(@assigned_ranges, $ranges{$address});
                            last;

                        }
                    }
                }
            }
        }

        foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
            foreach my $range (keys(%{$defined_addresses->{$interface}})) {
                if (grep { $_ eq $range } @assigned_ranges) {
                    next;
                }

                my $address = shift(@{$addresses{$interface}{$range}});
                $Log->comment("Adding defined address $address to node $node "
                        . "vfiler vfiler0 interface $interface");

                push(
                    @{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                            ->{'vfiler0'}->{'interfaces'}->{$interface}
                        },
                    {'address' => $address, 'range' => $range}
                );

            }
        }

        # If we are doing vfilers assign the addresses to interfaces
        if ($self->{'vfilers_per_node'}) {

            my @vfilers = keys(%{$defined_vfilers});
            @vfilers = natsort(@vfilers);
            foreach my $vfiler (@vfilers) {
                if ($vfiler eq 'vfiler0') { next; }
                foreach my $interface (@interfaces) {
                    foreach
                        my $range (keys(%{$defined_addresses->{$interface}}))
                    {

                        my $address
                            = shift(@{$addresses{$interface}{$range}});
                        if (!$address) {
                            die config_result(
                                "No more IP addresses for interface $interface"
                            );
                        }
                        $Log->comment(
                            "Adding addresses to node $node vfiler $vfiler "
                                . "interface $interface address $address "
                                . "range $range");

                        push(
                            @{  $self->{'defined_topology'}->{$node}
                                    ->{'vfilers'}->{$vfiler}->{'interfaces'}
                                    ->{$interface}
                                },
                            {'address' => $address, 'range' => $range}
                        );

                    }
                }
            }
        }
    }

    # Create the CIFS server names
    $Log->comment("Creating the CIFS server name for all vfilers");
    foreach my $node (@{$Options_hash->{'NODES'}}) {
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $cifs_server = $node;
        $cifs_server =~ s/(-|fas)//g;
        $self->{'defined_topology'}->{$node}->{'vfilers'}->{'vfiler0'}
            ->{'cifs_server'} = $cifs_server;
        if ($self->{'vfilers_per_node'}) {
            my @vfilers = keys(%{$defined_vfilers});
            @vfilers = natsort(@vfilers);
            foreach my $vfiler (@vfilers) {

                if ($vfiler eq 'vfiler0') { next; }
                $cifs_server = $vfiler;
                $cifs_server =~ s/(-|fas)//g;
                $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                    ->{'cifs_server'} = $cifs_server;
            }
        }
    }

    $Log->comment("Adding volume shares");
    foreach my $node (@{$Options_hash->{'NODES'}}) {

        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};

        foreach my $volume (natsort(keys(%{$defined_volumes}))) {

            my $share  = 's_' . $volume;
            my $vfiler = $defined_volumes->{$volume}->{'vfiler'};

            $Log->comment(
                "Adding $share for node $node vfiler $vfiler volume $volume");

            # Store the the volume name with the share
            $topology_database->{$node}->{'shares'}->{$share}->{'volume'}
                = $volume;

            # Store the share path with the share
            $topology_database->{$node}->{'shares'}->{$share}->{'path'}
                = '/vol/' . $volume;

            # Store the owning vfiler with the share
            $topology_database->{$node}->{'shares'}->{$share}->{'vfiler'}
                = $vfiler;

            # Configure the filer_setup flag that determines if the
            # share is created in filer_setup or some other way
            $topology_database->{$node}->{'shares'}->{$share}->{'filer_setup'}
                = 1;

            push(
                @{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                        ->{$vfiler}->{'shares'}
                    },
                $share
            );

        }
    }

    my $share_index = 1;

    # If I/O loads are specified parse each load and get the max number of
    # shares per client per load and assign the clients to each load
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'io') {
        $self->parse_io_configs('script' => $self->{'testname'});
        $share_index = $self->assign_io_clients(
            'script'      => $self->{'testname'},
            'clients'     => $self->{'clients'},
            'share_index' => $share_index
        );
        if (keys(%{$self->{'test_options'}}) == 0) {
            die config_result('No test options defined');
        }
    }

    # Define shares and assign them to clients to be use with the
    # mapstress.exe win32 tool for share limits testing
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'shares') {
        $self->define_shares(
            'script'      => $self->{'testname'},
            'share_index' => $share_index,
            'recipe'      => 'clients'
        );

        $share_index = $self->assign_share_verification_clients(
            'script'      => $self->{'testname'},
            'clients'     => $self->{'clients'},
            'share_index' => $share_index
        );
        if (keys(%{$self->{'test_options'}}) == 0) {
            die config_result('No test options defined');
        }
    }

    # Assign clients for NTFS streams testing
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'streams') {

        $share_index = $self->assign_stream_clients(
            'script'      => $self->{'testname'},
            'clients'     => $self->{'clients'},
            'share_index' => $share_index
        );
        if (keys(%{$self->{'test_options'}}) == 0) {
            die config_result('No test options defined ');
        }
    }

    # Define shares and assign them to clients to be use with the
    # mapstress.exe win32 tool for share limits testing
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'swifttest') {

        $Log->comment("Creating SwiftTest user parameter data");
        $topology_database->put('swifttest', {});

        $Log->comment("Creating domain");
        my $domain = $Options_hash->{'CIFS_DOMAIN'};
        $domain =~ s/^([^\.]+).+/$1/;
        push(
            @{  $topology_database->{'swifttest'}->{'user_parameters'}
                    ->{'domain'}
                },
            $domain
        );

        $Log->comment("Creating password");
        push(
            @{  $topology_database->{'swifttest'}->{'user_parameters'}
                    ->{'password'}
                },
            $Options_hash->{'CIFS_PASSWD'}
        );

        # Create the specified test users
        if ($Options_hash->{'TEST_USER_COUNT'}) {
            $Log->comment("Creating user");
            my $test_user_basename = $Options_hash->{'TEST_USER_BASENAME'};
            my $test_user_count    = $Options_hash->{'TEST_USER_COUNT'};
            for (my $row = 0; $row < $test_user_count; $row++) {

                # Create unique username and file
                my $user = $test_user_basename . ($row + 1);
                my $path = $self->{'testname'} . '_' . ($row + 1);

                push(
                    @{  $topology_database->{'swifttest'}->{'user_parameters'}
                            ->{'user'}
                        },
                    $user
                );
                push(@users, $user);

            }

            #natsort the global arrays.
            #@users             = natsort(@users);
            @userDefinedShares = @users;
            @userShares        = @users;

        }

        # Create the specified test files
        if ($Options_hash->{'TEST_FILE_COUNT'}) {
            $Log->comment("Creating files");
            my $test_file_basename = $Options_hash->{'TEST_FILE_BASENAME'};
            my $test_file_count    = $Options_hash->{'TEST_FILE_COUNT'};
            for (my $row = 0; $row < $test_file_count; $row++) {

                # Create unique file
                my $file = $test_file_basename . ($row + 1);

                push(
                    @{  $topology_database->{'swifttest'}->{'user_parameters'}
                            ->{'file'}
                        },
                    $file
                );

            }
        }
        if ($Options_hash->{'CIFS_HOMEDIR'}) {

    # Get the test user count and divide it by the volume count so that we can
    # properly assign I/O to the volumes.

            my $test_user_count = $Options_hash->{'TEST_USER_COUNT'};
            my $volume_count    = $Options_hash->{'VOLUME_COUNT'};

            $homedir_per_volume = ceil($test_user_count / $volume_count);

            $Log->comment(
                "We will have $homedir_per_volume users per volume");

            $self->define_homedirs(
                'script'      => $self->{'testname'},
                'share_index' => $share_index,
                'recipe'      => 'swifttest'
            );

        } else {
            $self->define_shares(
                'script'      => $self->{'testname'},
                'share_index' => $share_index,
                'recipe'      => 'swifttest'
            );
        }

    # Get the test user count and divide it by the volume count so that we can
    # properly assign I/O to the volumes.

        my $test_user_count = $Options_hash->{'TEST_USER_COUNT'};
        my $volume_count    = $Options_hash->{'VOLUME_COUNT'};

        my $users_per_volume = ceil($test_user_count / $volume_count);

        # Loop through each node and get the server and share info
        $Log->comment("Creating server and share");
        foreach my $node (@{$Options_hash->{'NODES'}}) {
            my $defined_shares = $topology_database->{$node}->{'shares'};
            my $defined_vfilers
                = $self->{'defined_topology'}->{$node}->{'vfilers'};
            my @vfilers = keys(%{$defined_vfilers});
            @vfilers = natsort(@vfilers);

            foreach my $vfiler (@vfilers) {
                if ($vfiler eq '') {
                    die script_result("Undefined vfiler key detected");
                }

                # Loop through each defined share
                foreach my $share (@{$defined_vfilers->{$vfiler}->{'shares'}})
                {
                    if ($Options_hash->{'CIFS_HOMEDIR'}) {

                        foreach my $user (@userShares) {
                            if ($user eq "") {
                                next;
                            }
                            if ($user eq $share) {

                               #Get the server name from the topology database
                                my $server
                                    = $topology_database->{$node}->{'shares'}
                                    ->{$share}->{'server'};
                                $Log->comment(
                                    "Using server $server and share $share");
                                push(
                                    @{  $topology_database->{'swifttest'}
                                            ->{'user_parameters'}->{'share'}
                                        },
                                    $share
                                );
                                push(
                                    @{  $topology_database->{'swifttest'}
                                            ->{'user_parameters'}->{'server'}
                                        },
                                    $server
                                );
                            } else {
                                next;
                            }
                        }
                    } else {

                        # Only get the shares we want
                        my $regex = '^s_' . $self->{'testname'};
                        if ($share !~ /$regex/) {
                            next;
                        }

                        # Get the server name from the defined share
                        my $server = $defined_shares->{$share}->{'server'};

                        push(
                            @{  $topology_database->{'swifttest'}
                                    ->{'user_parameters'}->{'share'}
                                },
                            $share
                        );
                        push(
                            @{  $topology_database->{'swifttest'}
                                    ->{'user_parameters'}->{'server'}
                                },
                            $server
                        );
                    }
                }
            }
        }

    }

    # Assign clients for TML testing
    if (   $Options_hash->{'CLIENT_ASSIGNMENT'}
        && $Options_hash->{'CLIENT_ASSIGNMENT'} eq 'test') {

        $self->assign_test_clients();

        if (keys(%{$self->{'test_options'}}) == 0) {
            die config_result('No test options defined ');
        }
    }

    # Configure resource agitation
    if ($Options_hash->{'RA'}) {
        $self->get_ra_config();
    }

    $Log->exit();
} ## end sub define_topology

sub filer_setup {

    $Log->enter();

    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::filer_setup(%options);

    $Options_hash = $Options->get_hash();

    # Make sure the multistore feature is licensed
    if ($Options_hash->{'VFILER_COUNT'}) {
        foreach my $node (@{$Options_hash->{'NODES'}}) {

            my $Cli  = $apiset{'cli'}{$node};
            my $Zapi = $apiset{'zapi'}{$node};

            # Check the discovered status of the multistore license
            my $multistore_license = 'not licensed';
            if (keys(
                    %{  $self->{'discovered_topology'}->{$node}->{'licenses'}
                        }
                )
                ) {
                $multistore_license
                    = $self->{'discovered_topology'}->{$node}->{'licenses'}
                    ->{'multistore'};
            }

            # License the multistore option
            if ($multistore_license eq "not licensed") {
                $Log->comment("Adding multistore license");
                eval { $Cli->license_add("code" => 'ESSRXFF'); };

                try {

                    $Response = $Zapi->options_set(
                        'name'  => 'licensed_feature.multistore.enable',
                        'value' => 'on'
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }

            # Check the discovered status of the multistore enable option
            my $discovered_options
                = $self->{'discovered_topology'}->{$node}->{'options'};
            if (grep { $_ eq 'licensed_feature.multistore.enable' }
                keys(%{$discovered_options})
                ) {

                my $enable = 0;
                my $option = $discovered_options->{
                    'licensed_feature.multistore.enable'};

            # Is the licensed_feature option valid for this realease/platform?
                my $option_valid = $self->is_option_valid(
                    'node'   => $node,
                    'option' => 'licensed_feature.multistore.enable'
                );

                if ($option_valid && $option ne 'on') {

                    eval {
                        $Zapi->options_set(
                            'name'  => 'licensed_feature.multistore.enable',
                            'value' => 'on'
                        );
                    };
                }
            }
        }
    }

    if (   $Options_hash->{'PRE_TEST_CLEANUP'}
        && $Options_hash->{'PRE_TEST_CLEANUP'} eq 'defined') {
        $self->filer_cleanup_defined();
        my @targets = (
            'aggregates',    'cifs',   'interfaces', 'licenses',
            'qtrees',        'quotas', 'shares',     'vfilers',
            'vfiler_limits', 'volumes'
        );
        $self->discover_topology('targets' => \@targets);
    }
    foreach my $node (@{$Options_hash->{'NODES'}}) {
        eval { $self->time_setup('node' => $node); };
    }

    foreach my $node (@{$Options_hash->{'NODES'}}) {
        $Log->comment("Configuring $node");

        my $Cli  = $apiset{'cli'}{$node};
        my $Zapi = $apiset{'zapi'}{$node};

        my $defined_addresses
            = $self->{'defined_topology'}->{$node}->{'addresses'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};
        my $discovered_interfaces
            = $self->{'discovered_topology'}->{$node}->{'interfaces'};

        my $vfiler = 'vfiler0';
        if ($Options_hash->{'DESTIP_VERSION'} eq 6) {
            $Log->comment("Enabling IPv6 on $vfiler");

            # Set the vfiler context for the next ZAPI calls
            try {
                $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
            try {

                $Response = $Zapi->options_set(
                    'name'  => 'ip.v6.enable',
                    'value' => 'on'
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
            try {

                $Response = $Zapi->options_set(
                    'name'  => 'ip.v6.ra_enable',
                    'value' => 'on'
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }

        $Log->comment("Configuring data interfaces for $vfiler");
        my @interfaces = keys(%{$defined_vfilers->{$vfiler}->{'interfaces'}});
        foreach my $interface (@interfaces) {
            if (!grep { $_ eq $interface }
                @{$Options_hash->{'DATA_INTERFACES'}}) {
                next;
            }

            $Log->comment("Configuring interface $interface");

            my $defined_addresses_by_interface
                = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};
            foreach
                my $address_information (@{$defined_addresses_by_interface}) {

                my $address = $address_information->{'address'};
                if (!$address) { next; }
                my $range = $address_information->{'range'};
                my $netmask
                    = $defined_addresses->{$interface}->{$range}->{'NETMASK'};

                my ($version, $regex);
                if ($address =~ /\./) {
                    $version = 'ipv4_info';
                    $regex   = $address;
                    $regex =~ s/(\.)/\\$1/g;
                } else {
                    $version = 'ipv6_info';
                    my $ip = new Net::IP($address);
                    $address = $ip->short();
                    $regex   = $address;
                }

                my $created;
                my $retries = 10;
                while (!$created && $retries) {
                    $self->ifconfig_address(
                        'node'      => $node,
                        'interface' => $interface,
                        'address'   => $address,
                        'netmask'   => $netmask,
                        'retries'   => 5,
                        'wait'      => 60
                    );

                    # Wait for #$%^&*($#%^& ONTAP
                    snooze 1;

                    # Check that the interface was configured
                    my $Ifconfig = $self->ifconfig(
                        'node'    => $node,
                        'retries' => 5,
                        'wait'    => 60
                    );

                    # Wait for #$%^&*($#%^& ONTAP
                    snooze 1;

                    foreach my $i (@{$Ifconfig}) {
                        if ($interface eq $i->{'name'}) {

                            #if (grep(/$regex/, @{$i->{$version}})) {
                            if (grep { $_->{'ip'} eq $address }
                                @{$i->{$version}}) {
                                $created = 1;
                                last;
                            } else {
                                $Log->debug("$address not found, retrying\n"
                                        . Dumper($Ifconfig));
                                $retries--;

                                # Wait for #$%^&*($#%^& ONTAP
                                snooze 60;

                            }
                        }
                    }
                }
                if (!$retries) {
                    die config_result(
                        "Unable to create alias $address on $interface");
                }

                my $rc_data = $self->read_file(
                    'path' => '/vol/vol0/etc/rc',
                    'node' => $node
                );

                # Remove any aliases using this address
                my $new_rc_data;
                my $found = 0;
                foreach my $line (split(/\n/, $rc_data)) {
                    chomp($line);
                    if ($line !~ /ifconfig $interface alias $regex/) {
                        $new_rc_data .= $line . "\n";
                    }
                }

                # Checking for an existing ifconfig command
                $rc_data = $new_rc_data;
                my $found = 0;
                foreach my $line (split(/\n/, $rc_data)) {
                    chomp($line);
                    if ($line =~ /ifconfig $interface $regex/) {
                        $found = 1;
                        last;
                    }
                }

                # Add the ifconfig command if required
                if (!$found) {
                    $new_rc_data = '';
                    foreach my $line (split(/\n/, $rc_data)) {
                        chomp($line);
                        if ($line =~ /ifconfig/ && $line !~ /alias/) {
                            $new_rc_data .= $line . "\n";
                            if ($address !~ /:/) {
                                $new_rc_data
                                    .= "ifconfig $interface $address netmask $netmask\n";
                            } else {
                                $new_rc_data
                                    .= "ifconfig $interface $address\n";
                            }
                        } else {
                            $new_rc_data .= $line . "\n";
                        }
                    }
                }
                $Log->comment(Dumper($new_rc_data));
                $self->write_file(
                    'ascii' => $new_rc_data,
                    'node'  => $node,
                    'path'  => '/vol/vol0/etc/rc'
                );
                snooze 1;
            }
        }

        if ($Options_hash->{'WAFL_DEFAULT_SECURITY_STYLE'}) {

            # Set the vfiler context for the next ZAPI calls
            try {
                $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };

            try {

                $Response = $Zapi->options_set(
                    'name'  => 'wafl.default_security_style',
                    'value' => $Options_hash->{'WAFL_DEFAULT_SECURITY_STYLE'}
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }

        # Set the vfiler context for the next ZAPI calls
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        # Create the volumes that don't already exist
        foreach my $volume (keys(%{$defined_volumes})) {
            if (grep(/^$volume$/, keys(%{$discovered_volumes}))) {
                next;
            }

            my $aggregate = $defined_volumes->{$volume}->{'aggregate'};
            if (!$aggregate) {
                die script_result("Aggregate is undef for for volume "
                        . "$volume in defined_topology\n"
                        . Dumper($defined_volumes->{$volume}));
            }
            my $size = $defined_volumes->{$volume}->{'size'} . 'm';
            if (!$size) {
                die script_result("Size is undef for for volume "
                        . "$volume in defined_topology\n"
                        . Dumper($defined_volumes->{$volume}));
            }
            try {

                $Log->comment("Creating volume $volume");

                $Response = $Zapi->volume_create(
                    'volume'               => $volume,
                    'containing-aggr-name' => $aggregate,
                    'size'                 => $size
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };

        }

        # Create the vfilers if they don't already exist
        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            if ($vfiler eq 'vfiler0') { next; }

            if (!grep { $_ eq $vfiler } keys(%{$discovered_vfilers})) {

                # Create the vfiler
                $self->create_vfiler(
                    'node'    => $node,
                    'vfiler'  => $vfiler,
                    'retries' => 5,
                    'wait'    => 60
                );
                snooze 12;

            } else {
                $Log->comment("Skipping creation of vfiler $vfiler "
                        . "because it already exists");
                $Log->comment("Adding storage to vfiler $vfiler");

                # Add volumes to the vfiler becasue it's already created
                my $storage;
                my @volumes = @{$defined_vfilers->{$vfiler}->{'volumes'}};
                @volumes = natsort(@volumes);
                foreach my $volume (@volumes) {
                    $Log->debug(
                        "Checking if volume $volume is owned by vfiler $vfiler"
                            . Dumper($discovered_vfilers->{$vfiler}->{'path'})
                    );
                    if (grep { $_ eq "/vol/$volume" }
                        @{$discovered_vfilers->{$vfiler}->{'path'}}) {
                        next;
                    }
                    $storage .= "/vol/$volume ";
                }

                if ($storage) {
                    $Log->comment("Adding volumes to vfiler $vfiler");
                    $Log->debug("Volumes: $storage");
                    $self->vfiler_add(
                        'node'    => $node,
                        'vfiler'  => $vfiler,
                        'storage' => $storage,
                        'retries' => 5,
                        'wait'    => 60
                    );

                }
            }
        }

        my @nodes = ($node);
        my @targets = ('vfilers', 'licenses');
        $self->discover_topology(
            'nodes'   => \@nodes,
            'targets' => \@targets
        );

        # Configure cifs.ipv6_enable
        my $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        if ($self->{'release'} !~ /boilermaker/) {

            foreach my $vfiler (@vfilers) {

                # Get vfiler status
                $Log->comment("Configuring cifs.ipv6.enable for $vfiler");
                my $status = $discovered_vfilers->{$vfiler}->{'status'};
                $Log->debug("Status: $status\n"
                        . Dumper($discovered_vfilers->{$vfiler}));

                if ($status ne 'running') {
                    $Log->comment(
                        "Skipping vfiler $vfiler becasue it is not running");
                    next;
                }

                # Set the vfiler context for the next ZAPI calls
                try {
                    $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                my $cifs_ipv6_enabled = 'off';
                if ($Options_hash->{'DESTIP_VERSION'} eq 6) {
                    $cifs_ipv6_enabled = 'on';
                }

                $Log->comment("Setting IPv6 for CIFS to $cifs_ipv6_enabled "
                        . "for vfiler $vfiler");
                try {

                    $Response = $Zapi->options_set(
                        'name'  => 'cifs.ipv6.enable',
                        'value' => $cifs_ipv6_enabled
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }
        }

        my @nodes = ($node);
        my @targets = ('vfilers', 'licenses');
        $self->discover_topology(
            'nodes'   => \@nodes,
            'targets' => \@targets
        );

        # Configure the vfiler if running
        my %mapped;
        my $status;
        foreach my $vfiler (@vfilers) {

            # Get the current vfilers status
            my $discovered_vfilers
                = $self->{'discovered_topology'}->{$node}->{'vfilers'};
            $status = $discovered_vfilers->{$vfiler}->{'status'};

            # Get the current vfilers path to etc
            my $path
                = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
            if ($path eq '/etc') { $path = '/vol/vol0/etc'; }

            if ($status eq 'running' || $vfiler eq 'vfiler0') {

                # Check for resolv.conf
                eval {

                    $Response = $Zapi->file_get_file_info(
                        'path' => "$path/resolv.conf");
                };

                # if resolv.conf exists DNS is configured
                if ($@) {
                    $self->configure_dns(
                        'node'    => $node,
                        'vfiler'  => $vfiler,
                        'retries' => 5,
                        'wait'    => 60
                    );
                    snooze 1;
                }

                if (   $Options_hash->{'NIS_DOMAINNAME'}
                    && $Options_hash->{'NIS_SERVERS'}) {
                    $self->configure_nis(
                        'node'   => $node,
                        'vfiler' => $vfiler,
                    );
                    snooze 1;
                }

                # Check for cifsconfig_share.cfg
                eval {

                    $Response = $Zapi->file_get_file_info(
                        'path' => "$path/cifsconfig_share.cfg");
                };

                # if cifsconfig_share.cfg exists CIFS is configured
                if ($@) {
                    $self->configure_cifs(
                        'node'    => $node,
                        'vfiler'  => $vfiler,
                        'retries' => 5,
                        'wait'    => 60
                    );
                    snooze 5;
                } else {

                    my $Status;
                    try {
                        $Response = undef;
                        $Response = $Zapi->cifs_status();
                        $Status   = $Response->get_parsed_output();
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                    if (grep(/stopped/, @{$Status->[0]->{'status'}})) {

                        # Start CIFS
                        try {

                            $Response = $Zapi->cifs_start();
                            $Status   = $Response->get_parsed_output();

                        }
                        catch NACL::APISet::Exceptions::ResponseException
                            with {
                            my $Exception = shift;
                            die config_result($Exception->text());
                            };
                        snooze 5;
                    }
                }

                if ($vfiler ne 'vfiler0') {
                    my $regex = "^s_v_$vfiler";
                    $regex =~ s/-/_/g;

                    # Create the root volume shares first
                    $self->create_shares(
                        'node'   => $node,
                        'vfiler' => $vfiler,
                        'regex'  => $regex,
                        'mapped' => \%mapped
                    );
                }

                # Create the data volume shares first
                my $regex = '^s_v_' . $self->{'testname'};
                $self->create_shares(
                    'node'   => $node,
                    'vfiler' => $vfiler,
                    'regex'  => $regex,
                    'mapped' => \%mapped
                );

            }
        }

        snooze 60;

        # Create the data shares (volume shares required)
        foreach my $vfiler (@vfilers) {

            #Load the Homedirs if needed.
            if ($Options_hash->{'CIFS_HOMEDIR'}) {

                $self->create_homedir(
                    'node'   => $node,
                    'vfiler' => $vfiler
                );

            } else {

                my $regex = '^s_' . $self->{'testname'};
                $self->create_shares(
                    'node'   => $node,
                    'vfiler' => $vfiler,
                    'regex'  => $regex,
                    'mapped' => \%mapped
                );
                snooze 1;
            }
        }

        # Populate local groups if specified
        if ($Options_hash->{'LOCAL_GROUP_COUNT'}) {
            my $group         = $Options_hash->{'TEST_GROUP_BASENAME'};
            my $gids          = $self->{'local_groups_per_vfiler'};
            my $group_members = $Options_hash->{'LOCAL_GROUP_MEMBERS'};

            foreach my $vfiler (@vfilers) {
                $self->create_local_groups(
                    'node'                 => $node,
                    'vfiler'               => $vfiler,
                    'group_basename'       => $group,
                    'starting_group_index' => 1,
                    'ending_group_index'   => $gids,
                    'group_members'        => $group_members
                );
            }
        }

        # Create the volume quotas
        if ($Options_hash->{'QUOTAS'}) {
            foreach my $vfiler (@vfilers) {
                $self->configure_quotas(
                    'node'   => $node,
                    'vfiler' => $vfiler
                );
            }
        }

        # Set the vfiler context for the next ZAPI calls
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        #Configure SMB 2 if needed.
        #We need to do this for before we turn on branchcache.
        if ($Options_hash->{'SMB2'}) {
            foreach my $vfiler (@vfilers) {
                $self->enable_smb2(
                    'node'   => $node,
                    'vfiler' => $vfiler
                );
            }
        }

        #Configure SMB signing if needed.
        if ($Options_hash->{'SMB_SIGNING'}) {
            foreach my $vfiler (@vfilers) {
                $self->smb_signing(
                    'node'   => $node,
                    'vfiler' => $vfiler
                );
            }
        }

        #Configure Branch Cache if needed
        if ($Options_hash->{'BRANCH_CACHE'}) {
            foreach my $vfiler (@vfilers) {
                $self->enable_branch_cache(
                    'node'   => $node,
                    'vfiler' => $vfiler
                );
            }
        }

        # Create the spn for the cifs server.
        if ($Options_hash->{'KERBEROS'}) {
            if (   $Options_hash->{'PRE_TEST_CLEANUP'}
                && $Options_hash->{'PRE_TEST_CLEANUP'} ne 'disabled'
                && $Options_hash->{'CIFS_SERVER_ACCESS_MODE'} ne 'ip') {
                $self->create_spn();

            }
        }

    }

    # Configure resource agitation if needed
    if ($Options_hash->{'RA'}) {
        my $defined_ra_config = $self->{'defined_topology'}->{'ra_config'};
        $self->ra_setup('ra_config' => $defined_ra_config);
    }

    $Log->comment(Dumper($self));

    $Log->exit();

} ## end sub filer_setup

sub filer_cleanup {

    $Log->enter();

    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::filer_cleanup(%options);

    $Options_hash = $Options->get_hash();

    if (!$Options_hash->{'POST_FAILURE_CLEANUP'}) {
        $Log->comment(
            "Skipping filer_cleanup STAGE because POST_FAILURE_CLEANUP=0");
        return;
    }

    $self->filer_cleanup_defined();

    $Log->exit();
} ## end sub filer_cleanup

sub client_setup {
    $Log->enter();

    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::client_setup(%options);

    $Options_hash = $Options->get_hash();

    #Check for ets/hosts Clean
    if ($Options_hash->{'CLIENT_ETC_HOSTS_CLEAN_UP'}) {
        my ($error, $client_conn_obj, $cmd, $output, $client_name,
            $os_version, %fields);

        host($Options_hash->{'CLIENTNAME'});

        $client_name = $Options_hash->{'CLIENTNAME'};

        %fields = host($client_name)->param_fields();

        $Log->comment("Cleaning etc\\hosts file on $client_name");

        #-------------------------------
        # Connect to the client
        #-------------------------------
        $client_conn_obj = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
            '-nologin', -timeout, 300);

        #-------------------------------
        # Verify connectivity to the client
        #-------------------------------
        eval { check_connectivity('conn' => $client_conn_obj); };
        $error = $@;
        if ($error) { die config_result($error); }

        # Clean etc\hosts file
        $cmd  = "cmd.exe /c echo 127.0.0.1 localhost \> "
            . "\%SYSTEMROOT\%\\system32\\drivers\\etc\\hosts";
        eval {
            (undef, undef, $output)
                = say($client_conn_obj, $cmd, -timeout, 60);
        };
        $error = $@;
        if ($error) { die config_result($error); }

        $cmd = "cmd.exe /c echo ::1 localhost \>\> "
            . "\%SYSTEMROOT\%\\system32\\drivers\\etc\\hosts";
        eval {
            (undef, undef, $output)
                = say($client_conn_obj, $cmd, -timeout, 60);
        };
        $error = $@;
        if ($error) { die config_result($error); }

        # Verify etc\hosts file is clean
        $cmd =
            "cmd.exe /c type \%SYSTEMROOT\%\\system32\\drivers\\etc\\hosts";
        eval {
            (undef, undef, $output)
                = say($client_conn_obj, $cmd, -timeout, 60);
        };
        $error = $@;
        if ($error) { die config_result($error); }

        $Log->comment("Client etc\\hosts file on $client_name:\n $output");
    }

    #Check for Branch Cache
    if ($Options_hash->{'BRANCH_CACHE'}) {
        my ($error, $client_conn_obj, $cmd, $output, $client_name,
            $os_version, %fields);

        host($Options_hash->{'CLIENTNAME'});

        $client_name = $Options_hash->{'CLIENTNAME'};

        %fields = host($client_name)->param_fields();

        $Log->comment("Attempting to enable BranchCache on $client_name");

        #-------------------------------
        # Connect to the client
        #-------------------------------
        $client_conn_obj = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
            '-nologin', -timeout, 300);

        #-------------------------------
        # Verify connectivity to the client
        #-------------------------------
        eval { check_connectivity('conn' => $client_conn_obj); };
        $error = $@;
        if ($error) { die config_result($error); }

        $cmd = "netsh branchcache set service mode=DISTRIBUTED";

        eval {
            (undef, undef, $output)
                = say($client_conn_obj, $cmd, -timeout, 60);
        };
        $error = $@;
        if ($error) { die config_result($error); }

        sleep(5);

        my $verify_cmd = "netsh branchcache show status detail=ALL";

        eval {
            (undef, undef, $output)
                = say($client_conn_obj, $verify_cmd, -timeout, 60);
        };
        $error = $@;
        if ($error) { die config_result($error); }
    }

    $Log->exit();
} ## end sub client_setup

sub DESTROY {

    $Log->exit();

    my $self = shift;

    local ($@);

    $self->client_cleanup();

    $Log->exit();
} ## end sub DESTROY

sub client_cleanup {
    $Log->enter();

    my ($self, %options) = @_;

    # Execute the one from the parent
    $self->SUPER::client_cleanup(%options);

    #Check for Branch Cache and if it present disable it on the client.
    my ($Options_hash, $client_conn_obj, $cmd, $output, $error, $client_name);

    if ($Options_hash->{'BRANCH_CACHE'}) {
        host($Options_hash->{'CLIENTNAME'});

        $client_name = $Options_hash->{'CLIENTNAME'};

        $Log->comment("Attempting to disable branch cache on $client_name");

        #----------------------------
        # Connect to the client
        #----------------------------
        $client_conn_obj = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
            '-nologin', -timeout, 300);

        #----------------------------
        # Verify connectivity to the client
        #----------------------------
        eval { check_connectivity('conn' => $client_conn_obj); };
        $error = $@;
        if ($error) { die config_result($error); }

        $cmd = "netsh branchcache set service DISABLED";

        eval {
            (undef, undef, $output)
                = say($client_conn_obj, $cmd, -timeout, 60);
        };
        $error = $@;
        if ($error) { die config_result($error); }

        sleep(5);

        my $verify_cmd = "netsh branchcache show status detail=ALL";

        eval {
            (undef, undef, $output)
                = say($client_conn_obj, $verify_cmd, -timeout, 60);
        };
        $error = $@;
        if ($error) { die config_result($error); }
    }

    $Log->exit();
} ## end sub client_cleanup

sub discover_topology {

    $Log->enter();

    my ($self, %options) = @_;
    my @targets = @{$options{'targets'}};
    my @nodes = @{$options{'nodes'}} if $options{'nodes'};

    # Discover all existing vfilers and volumes
    foreach my $node (@{$Options_hash->{'NODES'}}) {

        if (@nodes && !grep { $_ eq $node } @nodes) {
            next;
        }

        my $Cli  = $apiset{'cli'}{$node};
        my $Zapi = $apiset{'zapi'}{$node};

        my $Host = host($node);

        if (grep { $_ eq 'licenses' } @targets) {
            $Log->comment("Discovering licenses on $node");
            delete($self->{'discovered_topology'}->{$node}->{'licenses'});
            try {

                $Response = $Cli->license();
                snooze 1;
                my $Licenses = $Response->get_parsed_output();
                foreach my $license (@{$Licenses}) {
                    foreach my $key (keys(%{$license})) {
                        $self->{'discovered_topology'}->{$node}->{'licenses'}
                            ->{$key} = $license->{$key};
                    }
                }
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                my $Response  = $Exception->get_response_object();
                $Log->warn(
                    $Exception->text() . ": " . $Response->get_raw_output());
            }

        }

        my $discover_multistore = 1;

        # Get the multistore license status
        my $multistore_license = 'not licensed';
        if (keys(%{$self->{'discovered_topology'}->{$node}->{'licenses'}})) {
            $multistore_license
                = $self->{'discovered_topology'}->{$node}->{'licenses'}
                ->{'multistore'};
        }

        if ($multistore_license eq "not licensed") {
            $Log->comment(
                "Skipping vfiler discovery because multistore is not licensed"
            );
            $discover_multistore = 0;
        }

        # Get licensed_feature.multistore.enable
        my $multistore_enable;

        # Set the vfiler context for the next ZAPI calls
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        eval {
            $Response
                = $Zapi->options_get(
                'name' => 'licensed_feature.multistore.enable');
            my $Return = $Response->get_parsed_output();

            # Store its value
            $self->{'discovered_topology'}->{$node}->{'options'}
                ->{'licensed_feature.multistore.enable'}
                = $Return->[0]->{'value'};
            $multistore_enable = $Return->[0]->{'value'};
            $Log->comment(
                "licensed_feature.multistore.enable: $multistore_enable");

        };

        # Is the licensed_feature.multistore.enable option is valid for this
        # realease/platform?
        my $option_valid = $self->is_option_valid(
            'node'   => $node,
            'option' => 'licensed_feature.multistore.enable'
        );

        # Don't discover vfilers if it is off
        if ($option_valid && $multistore_enable ne 'on') {
            $discover_multistore = 0;
            die config_result("Unable to discover vfilers because "
                    . "licensed_feature.multistore.enable is "
                    . "off and that matters on this platform"
                    . "and release");
        }

        if ($discover_multistore && grep { $_ eq 'vfiler_limits' } @targets) {
            $Log->comment("Discovering vfiler limits on $node");
            try {

                $Response = $Cli->vfiler_limit();
                snooze 1;
                my $Vfiler_limits = $Response->get_parsed_output();
                foreach my $limit (natsort(keys(%{$Vfiler_limits->[0]}))) {
                    $self->{'discovered_topology'}->{$node}->{'vfiler_limits'}
                        ->{$limit} = $Vfiler_limits->[0]->{$limit};
                }
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                my $Response  = $Exception->get_response_object();
                $Log->warn(
                    $Exception->text() . ": " . $Response->get_raw_output());
            };
        }

        if ($discover_multistore && grep { $_ eq 'vfilers' } @targets) {
            delete($self->{'discovered_topology'}->{$node}->{'vfilers'});
            my $retry   = 1;
            my $retries = 5;
            while ($retry) {
                try {
                    $Log->comment("Discovering vfilers on $node");
                    $Response = $Cli->vfiler_status("all-details" => 1);
                    snooze 1;
                    my $Vfilers = $Response->get_parsed_output();
                    foreach my $vfiler (@{$Vfilers}) {
                        $self->{'discovered_topology'}->{$node}->{'vfilers'}
                            ->{$vfiler->{'vfiler'}} = $vfiler;
                    }
                    $retry = 0;
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    my $Response  = $Exception->get_response_object();
                    my $output    = $Response->get_raw_output();
                    if ($output =~ /ifconfig cannot be run/) {
                        $retry++;
                        snooze 60;
                    } else {
                        die config_result(
                            $Exception->text() . ": " . $output);
                    }
                };
                if ($retry == $retries) {
                    die config_result(
                        "Unable to get ifconfig information after $retries tries"
                    );
                }
            }
        } else {
            $self->{'discovered_topology'}->{$node}->{'vfilers'} = {};
        }

        if (!grep(/vfiler0/,
                keys(
                    %{  $self->{'discovered_topology'}->{$node}->{'vfilers'}
                        }
                    ))
            ) {
            $self->{'discovered_topology'}->{$node}->{'vfilers'}->{'vfiler0'}
                = {};
        }

        if (grep { $_ eq 'interfaces' } @targets) {

            # Get the current interface config
            $Log->comment("Discovering interfaces on $node");
            delete($self->{'discovered_topology'}->{$node}->{'interfaces'});

            my $retry   = 1;
            my $retries = 5;
            while ($retry) {
                try {

                    $Response = $Cli->ifconfig(all => 1);
                    my $Ifconfig = $Response->get_parsed_output();
                    foreach my $int (@{$Ifconfig}) {
                        $self->{'discovered_topology'}->{$node}
                            ->{'interfaces'}->{$int->{'name'}} = $int;
                    }
                    $retry = 0;
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    my $Response  = $Exception->get_response_object();
                    my $output    = $Response->get_raw_output();
                    if ($output =~ /ifconfig cannot be run/) {
                        $retry++;
                        snooze 60;
                    } else {
                        die config_result(
                            $Exception->text() . ": " . $output);
                    }
                };
                if ($retry == $retries) {
                    die config_result(
                        "Unable to get ifconfig information after $retries tries"
                    );
                }
            }
        }

        if (grep { $_ eq 'aggregates' } @targets) {
            $Log->comment("Discovering aggregates on $node");
            delete($self->{'discovered_topology'}->{$node}->{'aggregates'});
            my @fields = ('total', 'used', 'avail', 'capacity');
            my $Aggregates;
            try {

                $Response   = $Cli->aggr_status();
                $Aggregates = $Response->get_parsed_output();
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                my $Response  = $Exception->get_response_object();
                die config_result(
                    $Exception->text() . ": " . $Response->get_raw_output());
            };

            if (@{$Aggregates}) {
                try {

                    $Response = $Cli->df(
                        aggregate         => 1,
                        'scale-megabytes' => 1
                    );
                    snooze 1;
                    my @Output = @{$Response->get_parsed_output()};
                    my @fields = ('total', 'used', 'avail', 'capacity');
                    foreach my $hash (@Output) {
                        for (my $i = 0; $i < @{$Aggregates}; $i++) {
                            my $aggregate = ${$Aggregates}[$i];
                            if ($hash->{'aggregate'} eq
                                $aggregate->{'aggregate'}) {
                                foreach my $field (@fields) {
                                    $aggregate->{$field} = $hash->{$field};
                                }
                                @{$Aggregates}[$i] = $aggregate;
                                last;
                            }
                        }
                    }
                    for (my $i = 0; $i < @{$Aggregates}; $i++) {
                        my $aggregate = ${$Aggregates}[$i];
                        $self->{'discovered_topology'}->{$node}
                            ->{'aggregates'}->{$aggregate->{'aggregate'}}
                            = $aggregate;
                    }
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    my $Response  = $Exception->get_response_object();
                    die config_result($Exception->text() . ": "
                            . $Response->get_raw_output());
                };
            }
        }
        if (grep { $_ eq 'volumes' } @targets) {
            $Log->comment("Discovering volumes on $node");
            delete($self->{'discovered_topology'}->{$node}->{'volumes'});
            try {

                $Response = $Zapi->volume_list_info_iter_start();
                my $Return  = $Response->get_parsed_output();
                my $records = ${$Return}[0]->{'records'};
                my $tag     = ${$Return}[0]->{'tag'};
                $Log->comment("$records volumes discovered");

                $Response = $Zapi->volume_list_info_iter_next(
                    'tag'     => $tag,
                    'maximum' => $records
                );
                my $Volumes = $Response->get_parsed_output('tag' => $tag);
                my @volume_info
                    = @{$Volumes->[0]->{'volumes'}->[0]->{'volume-info'}};

                for (my $i = 0; $i < @volume_info; $i++) {
                    my $volume = $volume_info[$i];
                    $self->{'discovered_topology'}->{$node}->{'volumes'}
                        ->{$volume->{'name'}} = $volume;
                }

                $Response = $Zapi->volume_list_info_iter_end('tag' => $tag);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }
        if (grep { $_ eq 'qtrees' } @targets) {

            # Get the current qtree config
            $Log->comment("Discovering qtrees on $node");
            delete($self->{'discovered_topology'}->{$node}->{'qtrees'});
            try {

                $Response = $Zapi->qtree_list_iter_start();
                my $Return  = $Response->get_parsed_output();
                my $records = ${$Return}[0]->{'records'};
                my $tag     = ${$Return}[0]->{'tag'};

                $Response = $Zapi->qtree_list_iter_next(
                    'tag'     => $tag,
                    'maximum' => $records
                );
                my $Qtrees = $Response->get_parsed_output();
                my @qtree_info
                    = @{$Qtrees->[0]->{'qtrees'}->[0]->{'qtree-info'}};

                for (my $i = 0; $i < @qtree_info; $i++) {
                    my $qtree = $qtree_info[$i];
                    if ($qtree->{'id'} > 0) {
                        $self->{'discovered_topology'}->{$node}->{'qtrees'}
                            ->{$qtree->{'qtree'}} = $qtree;
                    }
                }

                $Response = $Zapi->qtree_list_iter_end('tag' => $tag);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }

        if (grep { $_ eq 'quotas' } @targets) {
            snooze 1;
            $Log->comment("Discovering quotas on $node");
            my $discovered_vfilers
                = $self->{'discovered_topology'}->{$node}->{'vfilers'};
            my $discovered_volumes
                = $self->{'discovered_topology'}->{$node}->{'volumes'};

            foreach my $vfiler (keys(%{$discovered_vfilers})) {

                $Log->comment("Discovering quotas owned by vfiler $vfiler");

                # Don't discover information on stopped vfilers
                if ($discovered_vfilers->{$vfiler}->{'status'} ne 'running') {
                    next;
                }

                # Set the vfiler context for the next ZAPI calls
                try {
                    $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                my $path
                    = "$discovered_vfilers->{$vfiler}->{'configuration_directory'}";
                if ($path eq '/etc') { $path = '/vol/vol0/etc'; }
                $path .= '/quotas';

                eval {
                    $Response = $Zapi->file_get_file_info('path' => $path);
                };

                if ($@) {
                    $self->write_file(
                        'ascii' => "\n",
                        'node'  => $node,
                        'path'  => $path
                    );
                    snooze 1;

                }

                # Get the current quota config
                my ($Return, $records, $tag);

                try {
                    $Response = $Zapi->quota_list_entries_iter_start();
                    $Return   = $Response->get_parsed_output();
                    $records  = ${$Return}[0]->{'records'};
                    $tag      = ${$Return}[0]->{'tag'};
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                if ($records) {
                    try {
                        $Response = $Zapi->quota_list_entries_iter_next(
                            'tag'     => $tag,
                            'maximum' => $records
                        );
                        my $Quotas
                            = $Response->get_parsed_output('tag' => $tag);
                        if ($Quotas->[0]->{'records'}) {

                            my @quota_info
                                = @{$Quotas->[0]->{'quota-entries'}->[0]
                                    ->{'quota-entry'}};

                            for (my $i = 0; $i < @quota_info; $i++) {
                                my $quota = $quota_info[$i];
                                push(
                                    @{  $self->{'discovered_topology'}
                                            ->{$node}->{'volumes'}
                                            ->{$quota->{'volume'}}->{'quotas'}
                                        },
                                    $quota
                                );

                            }
                        }
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };

                    try {

                        $Response = $Zapi->quota_list_entries_iter_end(
                            'tag' => $tag);
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }
            }
        }

        if (grep { $_ eq 'cifs' } @targets) {
            $Log->comment("Discovering CIFS on $node");
            delete($self->{'discovered_topology'}->{$node}->{'shares'});
            my $discovered_vfilers
                = $self->{'discovered_topology'}->{$node}->{'vfilers'};

            foreach my $vfiler (keys(%{$discovered_vfilers})) {
                $Log->comment("Discovering CIFS on vfiler $vfiler");

                # Don't discover information on stopped vfilers
                if ($discovered_vfilers->{$vfiler}->{'status'} ne 'running') {
                    next;
                }

                # Set the vfiler context for the next ZAPI calls

                try {
                    $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                # Check if CIFS is configured
                eval { $Response = $Zapi->cifs_list_config(); };

                if (!$@) {

                    # Set the vfiler context for the next ZAPI calls

                    my $original_cifs_status;
                    my $Status;
                    try {
                        $Response = undef;
                        $Response = $Zapi->cifs_status();
                        $Status   = $Response->get_parsed_output();
                        $original_cifs_status = $Status;

                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };

                    if (grep(/stopped/, @{$Status->[0]->{'status'}})) {

                        # Start CIFS
                        try {

                            $Response = $Zapi->cifs_start();
                            $Status   = $Response->get_parsed_output();

                        }
                        catch NACL::APISet::Exceptions::ResponseException
                            with {
                            my $Exception = shift;
                            die config_result($Exception->text());
                            };
                        snooze 5;

                        # get the CIFS status
                        try {

                            $Response = $Zapi->cifs_status();
                            $Status   = $Response->get_parsed_output();

                        }
                        catch NACL::APISet::Exceptions::ResponseException
                            with {
                            my $Exception = shift;
                            die config_result($Exception->text());
                            };
                    }

                    if (!grep(/started/, @{$Status->[0]->{'status'}})) {
                        next;
                    }

                    if (grep { $_ eq 'shares' } @targets) {
                        $Log->comment(
                            "Discovering CIFS shares on node $node vfiler $vfiler"
                        );
                        my $discovered_vfilers
                            = $self->{'discovered_topology'}->{$node}
                            ->{'vfilers'};
                        my $discovered_volumes
                            = $self->{'discovered_topology'}->{$node}
                            ->{'volumes'};
                        my @vfilers = keys(%{$discovered_vfilers});
                        @vfilers = natsort(@vfilers);

                        # Get the current share config
                        my ($Return, $records, $tag);
                        try {

                            $Response = $Zapi->cifs_share_list_iter_start(
                                'zapi-timeout' => 3600);
                            $Return  = $Response->get_parsed_output();
                            $records = ${$Return}[0]->{'records'};
                            $tag     = ${$Return}[0]->{'tag'};
                        }
                        catch NACL::APISet::Exceptions::ResponseException
                            with {
                            my $Exception = shift;
                            die config_result($Exception->text());
                            };

                        my $max_records_per_call = 10000;
                        while ($records) {

                            my $records_this_call;
                            if ($max_records_per_call >= $records) {
                                $records_this_call = $records;
                            } else {
                                $records_this_call = $max_records_per_call;

                            }
                            $records -= $records_this_call;

                            my $Shares;
                            try {
                                $Response = $Zapi->cifs_share_list_iter_next(
                                    'tag'     => $tag,
                                    'maximum' => $records_this_call
                                );
                                $Shares = $Response->get_parsed_output(
                                    'tag' => $tag);
                            }
                            catch NACL::APISet::Exceptions::ResponseException
                                with {
                                my $Exception = shift;
                                die config_result($Exception->text());
                                };

                            if ($Shares->[0]->{'records'}) {
                                my @share_info
                                    = @{$Shares->[0]->{'cifs-shares'}->[0]
                                        ->{'cifs-share-info'}};

                                for (my $i = 0; $i < @share_info; $i++) {
                                    my $share = $share_info[$i];
                                    if (grep { $_ eq $share->{'share-name'} }
                                        @ontap_shares) {
                                        next;
                                    }
                                    $self->{'discovered_topology'}->{$node}
                                        ->{'shares'}->{$share->{'share-name'}}
                                        = $share;

                                    # Get the owning vfiler
                                    if ($share->{'mount-point'}
                                        =~ /\/vol\/([^\/]+)/) {
                                        my $volume = $1;
                                        if (grep { $_ eq $volume }
                                            keys(%{$discovered_volumes})
                                            ) {
                                            my $vfiler
                                                = $discovered_volumes
                                                ->{$volume}
                                                ->{'owning-vfiler'};
                                            $self->{'discovered_topology'}
                                                ->{$node}->{'shares'}
                                                ->{$share->{'share-name'}}
                                                ->{'owning-vfiler'} = $vfiler;
                                        }

                                    }

                                }
                            }
                            snooze 10;
                        }

                        try {

                            $Response = $Zapi->cifs_share_list_iter_end(
                                'tag' => $tag);
                        }
                        catch NACL::APISet::Exceptions::ResponseException
                            with {
                            my $Exception = shift;
                            die config_result($Exception->text());
                            };

                        if (!grep(/started/,
                                @{$original_cifs_status->[0]->{'status'}})
                            ) {

                            # Stop CIFS
                            try {

                                $Response = $Zapi->cifs_stop();
                                $Status   = $Response->get_parsed_output();

                            }
                            catch NACL::APISet::Exceptions::ResponseException
                                with {
                                my $Exception = shift;
                                die config_result($Exception->text());
                                };
                        }
                    }
                }
            }
        }
    }

    $Log->exit();
} ## end sub discover_topology

sub assign_io_clients {

    $Log->enter();

    my ($self, %options) = @_;

    my $Options_hash = $Options->get_hash();
    my $share_index  = $options{'share_index'};
    my $volume_count = 1;
    my $client_count = 1;
    my $max_shares_per_occurrence;
    my $max_occurrences_per_client;
    my $io_config;
    my $io_config_index = 0;
    my @nodes;
    my $node_index      = 0;
    my $interface_index = 0;
    my %address_indexes;
    my @io_configs = keys(%{$self->{'defined_topology'}->{'io_configs'}});

    # Populate an array of volumes for each node and initialize the index for
    # for the addresses for each vfiler's data interface
    foreach my $node (@{$Options_hash->{'NODES'}}) {
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
                $address_indexes{$interface} = 0;
            }
        }
    }

    my @node_regex = split(/,/, $Options_hash->{'NODE_REGEX'});

    my @shares_per_volume;
    if ($Options_hash->{'SHARES_PER_VOLUME'}) {
        @shares_per_volume = split(/,/, $Options_hash->{'SHARES_PER_VOLUME'});
    }

    foreach my $i (@io_configs) {

        # Create a pointer to the current loads config
        $io_config = $self->{'defined_topology'}->{'io_configs'}->{$i};

        if (@shares_per_volume) {
            $io_config->{'shares_per_volume'} = $shares_per_volume[$i];
            next;
        }

        # Get the clients used for this load
        my $clients = 0;
        if ($io_config->{'max_clients'}) {
            $clients = $io_config->{'max_clients'};
        } else {
            $clients = @{$self->{'clients'}};
            $io_config->{'max_clients'} = $clients;
        }

        # Get the max number of shares each client can possibly use
        my $shares = $io_config->{'max_shares_per_client'} * $clients;
        $self->{'defined_topology'}->{'io_configs'}->{$i}->{'total_shares'}
            = $shares;

        # Die if we can't match any nodes
        @nodes = grep(/$node_regex[$i]/, @{$Options_hash->{'NODES'}});
        if (!@nodes) {
            die config_result("No nodes found by NODE_REGEX $node_regex[$i]");
        }

        # Search for all volumes whos aggregate is defined in AGGREGATES
        my $volumes = 0;
        foreach my $node (@nodes) {
            $Log->comment(
                "Populating target_volumes from defined volumes on $node");
            my $defined_volumes
                = $self->{'defined_topology'}->{$node}->{'volumes'};
            my @volumes = keys(%{$defined_volumes});
            @volumes = natsort(@volumes);
            foreach my $volume (@volumes) {
                my $aggregate = $defined_volumes->{$volume}->{'aggregate'};
                my $root_volume
                    = $defined_volumes->{$volume}->{'root_volume'};
                if (!$root_volume
                    && grep { $_ eq $aggregate }
                    split(/,/, $Options_hash->{'AGGREGATES'})
                    ) {
                    $volumes++;
                    push(@{$io_config->{'target_volumes'}}, $volume);
                }
            }
        }

        $io_config->{'target_volume_count'} = $volumes;

    }

    if (!@shares_per_volume) {
        my $shares_per_volume = 0;
        my $required_shares   = 0;
        my @target_volumes;
        foreach my $i (@io_configs) {

            # Create a pointer to the current loads config
            $io_config = $self->{'defined_topology'}->{'io_configs'}->{$i};
            $required_shares += $io_config->{'total_shares'};
            foreach my $volume (@{$io_config->{'target_volumes'}}) {
                if (!grep { $_ eq $volume } @target_volumes) {
                    push(@target_volumes, $volume);
                }
            }

        }

        if ($required_shares > @target_volumes) {
            $shares_per_volume = int($required_shares / @target_volumes);
            if ($required_shares % @target_volumes) {
                $shares_per_volume++;
                $Log->comment("Setting the number shares per volume to "
                        . "$shares_per_volume "
                        . "because the required number of shares ("
                        . $required_shares
                        . ")\ndoes not fit on all volumes evenly");
            } else {
                $Log->comment("Setting the number of shares per volume to "
                        . $shares_per_volume
                        . " because the required "
                        . "number of shares ("
                        . $required_shares
                        . ")\ndoes not fit on all target volumes ("
                        . @target_volumes
                        . ")");
            }
        } else {
            $shares_per_volume = 1;
        }

        foreach my $i (@io_configs) {

            # Create a pointer to the current loads config
            $io_config = $self->{'defined_topology'}->{'io_configs'}->{$i};

            # Setting shares per volume for the current config
            $io_config->{'shares_per_volume'} = $shares_per_volume;
        }
    }

    # Loop through each client and assign the correct info to it. Since the
    # tools themselves define the share needs we create share names and
    # allocate them to clients here.
    my @volumes;
    for my $client (@{$self->{'clients'}}) {

        $Log->comment("\nClient: $client");

        # Set the client state to reserved
        $self->{'clients_status'}->{$client}->{'status'} = 'reserved';

        # Get the I/O configuration
        $io_config
            = $self->{'defined_topology'}->{'io_configs'}->{$io_config_index};

        @nodes = grep(/$node_regex[$io_config_index]/,
            @{$Options_hash->{'NODES'}});
        if (!@nodes) {
            die config_result(
                "No nodes found by NODE_REGEX $node_regex[$io_config_index]");
        }

        if (host($client)->hosttype()
            !~ /$io_config->{'CLIENT_HOSTTYPE_REGEX'}/) {
            $Log->comment(
                      "Skipping $client becase of hosttype does not match "
                    . "CLIENT_HOSTTYPE_REGEX $io_config->{'CLIENT_HOSTTYPE_REGEX'}\n"
                    . Dumper($io_config));
            next;
        }

        # Added instance loop for future multiple client instances right now
        # only one instance is supported
        for (
            my $instance = 0;
            $instance < $Options_hash->{'INSTANCES_PER_CLIENT'};
            $instance++
            ) {

            $Log->comment("\nInstance: $instance");

            # Set current client to assigned so supervisor2startnext knows to
            # start it
            $self->{'clients_status'}->{$client}->{'status'} = 'assigned';

            # Configure how we connect to the client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'CLIENT_CONNECT_METHOD'}
                = $Options_hash->{'CLIENT_CONNECT_METHOD'};

            # Configure the desired hosttype for all clients in case this
            # script uses multiple client types. This is here for historical
            # purposes
            $self->{'test_options'}->{$client}->{$instance}
                ->{'CLIENT_HOSTTYPE_REGEX'}
                = $io_config->{'CLIENT_HOSTTYPE_REGEX'};

            # Set the length of time test_main executes on the current client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'TEST_MAIN_DURATION'} = $io_config->{'TEST_MAIN_DURATION'};

            # Set the length of time test_main sleeps on the current client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'TEST_MAIN_SLEEP'} = $io_config->{'TEST_MAIN_SLEEP'};

            # Provide the correct path the the tools config file
            if (host($client)->hosttype() =~ /win/) {
                $self->{'test_options'}->{$client}->{$instance}
                    ->{'IO_TOOLS_CONFIG'}
                    = "$Options_hash->{'TOP_LOGDIR_WINDOWS'}\\$io_config->{'io_tools_config'}";
            } else {
                $self->{'test_options'}->{$client}->{$instance}
                    ->{'IO_TOOLS_CONFIG'}
                    = "$Options_hash->{'TOP_LOGDIR'}/$io_config->{'io_tools_config'}";
            }

            # Set the tools regex for the current client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'IO_TOOLS_REGEX'} = $io_config->{'IO_TOOLS_REGEX'};

            # Set the tools overrides for the current client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'IO_TOOLS_OVERRIDES'} = $io_config->{'IO_TOOLS_OVERRIDES'};

            # Set the flag whether to map shares for the current client
            $self->{'test_options'}->{$client}->{$instance}->{'IO_MAP_SHARES'}
                = $io_config->{'IO_MAP_SHARES'};

            # Set the  wether to map shares for the current client
            $self->{'test_options'}->{$client}->{$instance}
                ->{'IO_REMOVE_DATA'} = $io_config->{'IO_REMOVE_DATA'};

            # Determine the max shares per client for this load
            $max_occurrences_per_client
                = $io_config->{'max_occurrences_per_client'};
            $max_shares_per_occurrence
                = $io_config->{'max_shares_per_occurrence'};

            # Loop though the max number of shares per client for this load
            # This is where the share info is populated in defined_topology
            # and linked to volumes and vfilers
            for (my $o = 0; $o < $max_occurrences_per_client; $o++) {
                my $volume;
                for (my $s = 1; $s <= $max_shares_per_occurrence; $s++) {

                    $Log->comment("\nOccurence: $o");

                    # Get the current node name
                    my $node = $nodes[$node_index];
                    $Log->comment("Node: $node");

                    # Create refs to the defined_topology hash
                    my $defined_shares
                        = $topology_database->{$node}->{'shares'};
                    my $defined_vfilers
                        = $self->{'defined_topology'}->{$node}->{'vfilers'};
                    my $defined_volumes
                        = $self->{'defined_topology'}->{$node}->{'volumes'};

                    # Get the volume name
                    if (!@volumes) {

                        # Get the defined volumes
                        my @defined_volumes = keys(%{$defined_volumes});
                        @defined_volumes = natsort(@defined_volumes);

                        # Get the defined target volumes
                        my $volume_regex   = 'v_' . $self->{'testname'};
                        my @target_volumes = grep(/$volume_regex/,
                            @{$io_config->{'target_volumes'}});
                        @target_volumes = natsort(@target_volumes);

                        # Use only volumes on the current node
                        foreach my $volume (@defined_volumes) {
                            if (grep { $_ eq $volume } @target_volumes) {
                                push(@volumes, $volume);
                            }
                        }

                        if (!@volumes) {
                            die config_result("No available target volumes\n"
                                    . "\@defined_volumes\n"
                                    . Dumper(\@defined_volumes)
                                    . "\@target_volumes\n"
                                    . Dumper(\@target_volumes));
                        }
                    }

                    if (!$volume) {
                        $volume = shift(@volumes);
                    }

                    $Log->comment("Volume: $volume");

                    # Get the vfiler name from the volume entry
                    my $vfiler = $defined_volumes->{$volume}->{'vfiler'};
                    if (!$vfiler) {
                        die config_result(
                                  "Unable to retrieve vfiler name from "
                                . "defined volume $volume on node $node\n"
                                . "Defined volume $volume\n"
                                . Dumper($defined_volumes->{$volume}));
                    }
                    $Log->comment("Vfiler: $vfiler");

                    # Get the vfiler address from the vfiler entry
                    my $interface = ${$Options_hash->{'DATA_INTERFACES'}}
                        [$interface_index];
                    $Log->comment("Interface: $interface");

                    # Get the ip address from the defined vfilers interface
                    $Log->debug(
                        "Getting an address from interface $interface");
                    my $defined_addresses_this_interface
                        = $defined_vfilers->{$vfiler}->{'interfaces'}
                        ->{$interface};

                    if ($address_indexes{$interface}
                        > $#{$defined_addresses_this_interface}) {
                        $address_indexes{$interface} = 0;
                    }

                    my $ip = ${$defined_addresses_this_interface}
                        [$address_indexes{$interface}]->{'address'};

                    my $net_ip  = new Net::IP($ip);
                    my $version = $Options_hash->{'DESTIP_VERSION'};
                    my $current_address_index = $address_indexes{$interface};

                    $Log->debug("Verify the version of $ip "
                            . "matches DESTIP_VERSION ($version)");
                    while ($version != $net_ip->version()) {
                        $Log->debug(
                            "The version of $ip does not match DESTIP_VERSION. "
                                . "Getting next address on interface $interface"
                        );

                        $address_indexes{$interface}++;
                        if ($address_indexes{$interface}
                            > $#{$defined_addresses_this_interface}) {
                            $address_indexes{$interface} = 0;
                        }
                        if ($address_indexes{$interface}
                            == $current_address_index) {
                            last;
                        }
                        $ip = ${$defined_addresses_this_interface}
                            [$address_indexes{$interface}]->{'address'};
                        $Log->debug("Verify the version of $ip "
                                . "matches DESTIP_VERSION ($version)");
                        if (!$ip) { last; }
                        $net_ip = new Net::IP($ip);
                    }

                    if (!$ip) {

                        my $defined_addresses
                            = $self->{'defined_topology'}->{$node}
                            ->{'addresses'};
                        my $defined_vfiler = $defined_vfilers->{$vfiler};
                        my $discovered_interfaces
                            = $self->{'discovered_topology'}->{$node}
                            ->{'interfaces'};

                        my $msg
                            = sprintf(
                            "Unable to retrieve an IP address for vfiler "
                                . "%s\n using index %d of interface %s\'s "
                                . "defined addresses\n",
                            $vfiler, $address_indexes{$interface},
                            $interface);
                        die config_result($msg
                                . "\nDiscovered Interfaces on $node:\n"
                                . Dumper($discovered_interfaces)
                                . "\nDefined vfiler $vfiler on $node:\n"
                                . Dumper($defined_vfiler)
                                . "\nDefined address ranges:\n"
                                . Dumper($defined_addresses));
                    }

                    $Log->comment("IP Address: $ip");

                    my $mode = $io_config->{'CIFS_SERVER_ACCESS_MODE'};
                    if (!$mode) {
                        $mode = $Options_hash->{'CIFS_SERVER_ACCESS_MODE'};
                    }

                    my $server = $self->get_cifs_server_name(
                        'ip'   => $ip,
                        'mode' => $mode
                    );
                    $Log->comment("Server: $server");

                    # Defined the share name
                    my $share
                        = 's_' . $options{'script'} . '_' . $share_index;
                    $Log->comment("Share: $share");

                    push(
                        @{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                                ->{$vfiler}->{'shares'}
                            },
                        $share
                    );

                    # Store the the volume name with the share
                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'volume'} = $volume;

                    # Store the share path with the share
                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'path'} = '/vol/' . $volume . '/' . $share;

                    # Store the owning vfiler with the share
                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'vfiler'} = $vfiler;

                    # Configure the filer_setup flag that determines if the
                    # share is created in filer_setup or some other way
                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'filer_setup'} = 1;

                    # Use qtrees
                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'qtree'} = 1;

                    $Log->comment('Updating test options');
                    $self->update_test_options(
                        'client'   => $client,
                        'instance' => $instance,
                        'node'     => $node,
                        'vfiler'   => $vfiler,
                        'server'   => $server,
                        'volume'   => $volume,
                        'share'    => $share,
                        'destip'   => $ip
                    );

                    # Increment the share index for the current volume
                    $defined_volumes->{$volume}->{'share_count'}++;

                    # Get the avavilable target volumes
                    my @target_volumes = @{$io_config->{'target_volumes'}};
                    my $available_target_volumes = 0;
                    my $shares_per_volume = $io_config->{'shares_per_volume'};

                    # Find if volumes have available shares
                    foreach my $n (@{$Options_hash->{'NODES'}}) {
                        my $dv
                            = $self->{'defined_topology'}->{$n}->{'volumes'};
                        foreach my $v (keys(%{$dv})) {
                            if (!grep { $_ eq $v } @target_volumes) {
                                next;
                            }
                            my $share_count = $dv->{$v}->{'share_count'};
                            $Log->debug("$v:$share_count");

                            # Get the count of volumes that are not full of
                            # shares
                            if ($share_count < $shares_per_volume) {
                                $available_target_volumes++;
                            }
                        }
                    }

                    # If we still have shares but there are no more target
                    # volumes left then die
                    $Log->comment("\$s:$s "
                            . "\$max_shares_per_occurrence:$max_shares_per_occurrence "
                            . "\$available_target_volumes:$available_target_volumes "
                            . "\$io_config_index:$io_config_index ");

                    if ($s < $max_shares_per_occurrence
                        && !$available_target_volumes) {
                        die config_result("Not enough volumes to honor "
                                . "SHARES_PER_VOLUME $shares_per_volume");
                    }

                    # Increment the share counter
                    $share_index++;

                    # If there is more than one node specified we should
                    # switch back and forth to spread the load
                    if ($node_index == $#nodes) {
                        $node_index = 0;
                    } else {
                        $node_index++;
                    }
                    @volumes = ();

                    # If there is more than one address range specified
                    # for a data interface we should switch back and forth
                    # to spread the load
                    $address_indexes{$interface}++;
                    if ($address_indexes{$interface}
                        > $#{$defined_addresses_this_interface}) {
                        $address_indexes{$interface} = 0;
                    }

                    # Increment or reset the data interfaces index
                    $interface_index++;
                    if ($interface_index
                        >= $#{$Options_hash->{'DATA_INTERFACES'}}) {
                        $interface_index = 0;
                    }

                    # undef $volume so we get a new one
                    $volume = undef;
                }
            }
        }

        # Increment the client count and switch to the next load if needed
        if ($client_count == $io_config->{'max_clients'}) {
            if (   $Options_hash->{'IO_LOAD_CONFIG'}
                || $Options_hash->{'IO_TOOLS_CONFIG'}) {
                if ($io_config_index == $#io_configs) {

                    # No more loads so bail
                    last;
                } else {

                    # Reset the client count
                    $client_count = 1;

                    # Switch to the next load
                    $Log->comment("Switching to the next load");
                    $io_config_index++;
                    $Log->debug("\$io_config_index:$io_config_index");

                    # Get the I/O configuration
                    $io_config = $self->{'defined_topology'}->{'io_configs'}
                        ->{$io_config_index};

                    @volumes = ();

                    # Get the filtered array of nodes for this load if a regex
                    # has been specified otherwise use all nodes
                    $Log->comment("Getting the nodes for this load");
                    @nodes = grep(/$node_regex[$io_config_index]/,
                        @{$Options_hash->{'NODES'}});
                    if (!@nodes) {
                        die config_result(
                            "No nodes found by NODE_REGEX $node_regex[$io_config_index]"
                        );
                    }
                }
            }
        } else {
            $client_count++;
        }

    }

    $Log->exit();

    return $share_index;
} ## end sub assign_io_clients

sub assign_share_verification_clients {

    $Log->enter();

    my ($self, %options) = @_;
    my $Options_hash = $Options->get_hash();
    my $instance     = 0;
    my $share_index;
    my @shares;
    my %nodes;
    my @nodes
        = grep(/$Options_hash->{'NODE_REGEX'}/, @{$Options_hash->{'NODES'}});

    if (!@nodes) {
        die config_result("Unable to find any nodes using the "
                . "NODE_REGEX $Options_hash->{'NODE_REGEX'} against "
                . Dumper($Options_hash->{'NODES'}));
    }

    foreach my $node (@nodes) {
        my $defined_shares = $topology_database->{$node}->{'shares'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            foreach my $share (@{$defined_vfilers->{$vfiler}->{'shares'}}) {
                if ($share =~ /^s_v_/) { next; }
                push(@shares, $share);
                $nodes{$share} = $node;
            }
        }
    }
    @shares = natsort(@shares);

    # Loop through each client and assign the correct info to it.
    my $client_count = 0;
    $Log->comment("Shares to verify: " . @shares);
    for my $client (@{$self->{'clients'}}) {

        if (host($client)->hosttype()
            !~ /$Options_hash->{'CLIENT_HOSTTYPE_REGEX'}/) {
            next;
        }

        # Set the client state to reserved

        if (!@shares) {
            $self->{'clients_status'}->{$client}->{'status'} = 'reserved';
            next;
        }

        $Log->comment("Client: $client");

        $self->{'clients_status'}->{$client}->{'status'} = 'assigned';
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_CONNECT_METHOD'}
            = $Options_hash->{'CLIENT_CONNECT_METHOD'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_HOSTTYPE_REGEX'}
            = $Options_hash->{'CLIENT_HOSTTYPE_REGEX'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'TEST_MAIN_DURATION'} = $Options_hash->{'TEST_MAIN_DURATION'};
        $self->{'test_options'}->{$client}->{$instance}->{'TEST_MAIN_SLEEP'}
            = $Options_hash->{'TEST_MAIN_SLEEP'};

        # Determine the number of shares each client will create
        my $shares_per_client = 0;
        if (   $client eq ${$self->{'clients'}}[0]
            && $self->{'shares_per_client_remainder'}) {
            $shares_per_client = $self->{'shares_per_client'}
                + $self->{'shares_per_client_remainder'};
        } else {
            $shares_per_client = $self->{'shares_per_client'};
        }

        $Log->comment("Shares: $shares_per_client");

        # Loop through each share and if any of the related info
        # changes save it as a new set of test options
        my $starting_share_index;
        foreach my $share_count (1 ... $shares_per_client) {

            my ($next_node, $next_vfiler, $next_server, $next_volume,
                $next_share, $share_basename, $share);

            # Get the next share
            if (@shares) {
                $share = shift(@shares);
            } else {
                last;
            }

            if ($share =~ /^(.+)_(\d+)$/) {
                $share_basename = $1;
                $share_index    = $2;
            }

            if (!$starting_share_index) {
                $starting_share_index = $share_index;
            }

            # Get the related information
            my $node           = $nodes{$share};
            my $defined_shares = $topology_database->{$node}->{'shares'};
            my $vfiler         = $defined_shares->{$share}->{'vfiler'};
            my $server         = $defined_shares->{$share}->{'server'};
            my $volume         = $defined_shares->{$share}->{'volume'};

            if (@shares) {
                $next_share  = $shares[0];
                $next_node   = $nodes{$next_share};
                $next_vfiler = $defined_shares->{$next_share}->{'vfiler'};
                $next_server = $defined_shares->{$next_share}->{'server'};
                $next_volume = $defined_shares->{$next_share}->{'volume'};
            }

            # If any of the information has changed then save
            # the set the next values and update the test
            # options
            if (  !@shares
                || $next_node   ne $node
                || $next_vfiler ne $vfiler
                || $next_volume ne $volume
                || $next_server ne $server
                || $share_count == $shares_per_client) {

                my $share_info;
                if ($shares_per_client == 1) {
                    $share_info = $share;
                } else {
                    $share_info
                        = $share_basename . '_;'
                        . $starting_share_index . '-'
                        . $share_index;
                }

                $Log->debug("Updating test options");
                $self->update_test_options(
                    'client'   => $client,
                    'instance' => $instance,
                    'node'     => $node,
                    'vfiler'   => $vfiler,
                    'server'   => $server,
                    'volume'   => $volume,
                    'share'    => $share_info
                );

                $starting_share_index = 0;
            }
        }
        $client_count++;
        if ($client_count == $self->{'max_clients'}) { last; }
    }

    $Log->exit();
    return $share_index++;
} ## end sub assign_share_verification_clients

sub create_vfiler {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $defined_addresses
        = $self->{'defined_topology'}->{$node}->{'addresses'};
    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $discovered_interfaces
        = $self->{'discovered_topology'}->{$node}->{'interfaces'};

    my @interfaces = keys(%{$defined_vfilers->{$vfiler}->{'interfaces'}});
    $Log->comment("Removing aliases for $vfiler");
    foreach my $interface (@interfaces) {
        my $defined_addresses_by_interface
            = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};
        foreach my $address_information (@{$defined_addresses_by_interface}) {

            my $address = $address_information->{'address'};
            if (!$address) { next; }
            my $range = $address_information->{'range'};
            my $netmask
                = $defined_addresses->{$interface}->{$range}->{'NETMASK'};

            my $address_regex = $address;
            $address_regex =~ s/(\.)/\\$1/g;

            # Delete the alias if needed
            if (grep { $_->{'ip'} eq $address }
                @{$discovered_interfaces->{$interface}->{'ipv4_info'}}) {

                $self->ifconfig_remove_alias(
                    'node'      => $node,
                    'interface' => $interface,
                    'address'   => $address,
                    'retries'   => 5,
                    'wait'      => 60
                );

                # Wait for #$%^&*($#%^& ONTAP
                snooze 1;
            }

        }
    }

    # Create a list of storage devices for the vfiler
    my $storage;
    foreach my $volume (@{$defined_vfilers->{$vfiler}->{'volumes'}}) {
        $storage .= "/vol/$volume ";
    }

    # Create the list of IP addresses
    my $addresses;
    my @ipv4_addresses;
    my @ipv6_addresses;
    foreach my $interface (@interfaces) {
        my $defined_addresses_by_interface
            = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};
        foreach my $address_information (@{$defined_addresses_by_interface}) {
            my $address = $address_information->{'address'};
            if ($address =~ /\./) {
                push(@ipv4_addresses, $address);
            } else {
                push(@ipv6_addresses, $address);
            }
        }
    }

    if (@ipv4_addresses) {
        $addresses .= '-i ' . join(' -i ', @ipv4_addresses);
        $addresses .= ' ';
    }

    if (@ipv6_addresses) {
        $addresses .= '-i ' . join(' -i ', @ipv6_addresses);

    }

    $Log->comment("Creating $vfiler");
    $self->vfiler_create(
        'node'      => $node,
        'vfiler'    => $vfiler,
        'addresses' => $addresses,
        'storage'   => $storage,
        'retries'   => 5,
        'wait'      => 60
    );

    # Wait for #$%^&*($#%^& ONTAP
    snooze 1;

    $Log->comment("Creating aliases for $vfiler");
    foreach my $interface (@interfaces) {
        my $defined_addresses_by_interface
            = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};
        foreach my $address_information (@{$defined_addresses_by_interface}) {

            my $address = $address_information->{'address'};
            my $range   = $address_information->{'range'};
            my $netmask
                = $defined_addresses->{$interface}->{$range}->{'NETMASK'};
            my %ifconfig_options = (
                'node'      => $node,
                'interface' => $interface,
                'address'   => $address,
                'retries'   => 5,
                'wait'      => 60
            );

            my ($version, $regex);
            if ($address =~ /\./) {
                $version = 'ipv4_info';
                $regex   = $address;
                $regex =~ s/(\.)/\\$1/g;
                $ifconfig_options{'netmask'} = $netmask;
            } else {
                $version = 'ipv6_info';
                my $ip = new Net::IP($address);
                $address = $ip->short();
                $regex   = $address;
            }

            my $created;
            my $retries = 10;
            while (!$created && $retries) {
                $self->ifconfig_alias(%ifconfig_options);

                # Wait for #$%^&*($#%^& ONTAP
                snooze 1;

                # Check that the aliases were created per burt 433887
                my $Ifconfig = $self->ifconfig(
                    'node'    => $node,
                    'retries' => 5,
                    'wait'    => 60
                );

                # Wait for #$%^&*($#%^& ONTAP
                snooze 1;

                foreach my $i (@{$Ifconfig}) {
                    if ($interface eq $i->{'name'}) {
                        if (grep { $_->{'ip'} eq $address } @{$i->{$version}})
                        {
                            $created = 1;
                            last;
                        } else {
                            $Log->debug("$address not found, sleeping");
                            $retries--;

                            # Wait for #$%^&*($#%^& ONTAP
                            snooze 60;

                        }
                    }
                }
            }
            if (!$retries) {
                die config_result(
                    "Unable to create alias $address on $interface");
            }

            my $rc_data = $self->read_file(
                'path' => '/vol/vol0/etc/rc',
                'node' => $node
            );

            if ($rc_data !~ /$regex/) {
                if ($address !~ /:/) {
                    $rc_data
                        .= "ifconfig $interface alias $address netmask $netmask\n";
                } else {
                    $rc_data .= "ifconfig $interface alias $address\n";
                }

                $self->write_file(
                    'ascii' => $rc_data,
                    'node'  => $node,
                    'path'  => '/vol/vol0/etc/rc'
                );
                snooze 1;
            }
        }
    }
    $Log->exit();
} ## end sub create_vfiler

sub create_shares {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};
    my $regex  = $options{'regex'};
    my $mapped = $options{'mapped'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $defined_shares  = $topology_database->{$node}->{'shares'};
    my $discovered_qtrees
        = $self->{'discovered_topology'}->{$node}->{'qtrees'};
    my $discovered_shares
        = $self->{'discovered_topology'}->{$node}->{'shares'};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Creating shares on vfiler $vfiler matching $regex");

    my $conn_obj;
    my @share_creation_clients
        = grep { host($_)->hosttype() =~ /win/ } @{$self->{'clients'}};
    if (@share_creation_clients) {

        host($share_creation_clients[0]);
        $conn_obj = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
            '-nologin', -timeout, 300);
        eval { check_connectivity('conn' => $conn_obj); };
        my $error = $@;
        if ($error) { die config_result($error); }
    }

    $Log->debug("vfiler $vfiler\n" . Dumper($defined_vfilers->{$vfiler}));
    foreach my $share (@{$defined_vfilers->{$vfiler}->{'shares'}}) {

        $Log->debug(
            "Checking share $share\n" . Dumper($defined_shares->{$share}));

        # If the share exists skip it
        if (grep { $_ eq $share } keys(%{$discovered_shares})) {
            $Log->debug("Skipping share $share because it allways exists\n"
                    . Dumper($discovered_shares->{$share}));
            next;
        }

        # If filer_setup is not set then we skip share creation in
        # filer_setup
        if (!$defined_shares->{$share}->{'filer_setup'}) {
            $Log->debug("Skipping share $share because filer_setup is not 1");
            next;
        }

        # If this shares volume does not belong to the current vfiler
        # skip it
        my $volume = $defined_shares->{$share}->{'volume'};
        if (!grep(/^$volume$/, @{$defined_vfilers->{$vfiler}->{'volumes'}})) {
            $Log->debug(
                "Skipping share $share because it does not belong to the current vfiler\n"
                    . Dumper($defined_vfilers->{$vfiler}));
            next;
        }

        # These are not the droids we are looking for, move along...
        if ($share !~ /$options{'regex'}/) { next; }

        # If the qtree exists skip its creation
        my $qtree  = $defined_shares->{$share}->{'qtree'};
        my $vfiler = $defined_shares->{$share}->{'vfiler'};
        if (   $qtree
            && !grep(/^$share$/, keys(%{$discovered_qtrees}))
            && $share ne "s_$volume") {
            $Log->comment("Creating qtree $share on $volume");

            try {

                $Response = $Zapi->qtree_create(
                    'qtree'  => $share,
                    'volume' => $volume
                );
                $self->{'defined_topology'}->{$node}->{'qtrees'}->{$share}
                    = {'volume' => $volume, 'vfiler' => $vfiler};
            }

            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
            snooze 1;

        } elsif (!$qtree && $share ne "s_$volume") {

            if (!@share_creation_clients) {
                die config_result("There are no windows clients "
                        . "to create share directories");
            }

            my @interfaces
                = keys(%{$defined_vfilers->{$vfiler}->{'interfaces'}});
            my $interface_found = 0;
            my %address_indexes;
            foreach my $i (@interfaces) {
                if (grep { $_ eq $i } @{$Options_hash->{'DATA_INTERFACES'}}) {
                    $interface_found = 1;
                    $address_indexes{$i} = 0;
                }
            }

            if (!$interface_found) {
                die config_result("Unable to find an interface for $vfiler "
                        . Dumper($defined_vfilers->{$vfiler}->{'interfaces'})
                );
            }

            my ($server, undef) = $self->assign_cifs_server(
                'node'            => $node,
                'vfiler'          => $vfiler,
                'address_indexes' => \%address_indexes,
                'interface_index' => 0
            );

            if (!${$mapped}{$volume}) {
                map_share(
                    'conn'     => $conn_obj,
                    'share'    => "\\\\$server\\s_$volume",
                    'drive'    => '',
                    'username' => $Options_hash->{'CIFS_USER'},
                    'password' => $Options_hash->{'CIFS_PASSWD'},
                    'timeout'  => 300,
                    'reuse'    => 1,
                    'delay'    => 60,
                    'attempts' => 3,
                );
                ${$mapped}{$volume} = 1;
            }

            my $path = "\\\\$server\\s_$volume\\$share";
            my $cmd  = 'cmd.exe /c if not exist ' . $path . ' mkdir ' . $path;

            my $output;
            eval {
                (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 300);
            };
            my $error = $@;
            if ($output || $error) {
                die config_result("Output:$output\nError:$error");
            }
        }

        my $path = $defined_shares->{$share}->{'path'};

        $Log->comment(
            "Creating share $share on node $node vfiler $vfiler Path:$path");

        # Set the vfiler context for the next ZAPI calls
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        try {

            $Response = $Zapi->cifs_share_add(
                'share-name' => $share,
                'path'       => $path
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result(
                $Exception->text() . "\n" . Dumper($defined_shares));
        };
        snooze 1;
    }

    $Log->exit();
} ## end sub create_shares

sub configure_cifs {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Configuring CIFS on $vfiler");

    my $cifs_license
        = $self->{'discovered_topology'}->{$node}->{'licenses'}->{'cifs'};
    if ($cifs_license eq 'not licensed') {
        $Log->comment("Adding cifs license");
        try {

            $Response = $Cli->license_add("code" => 'GTLYRVH');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        snooze 1;
    }

    my $Status;
    try {

        $Response = $Zapi->cifs_status();
        $Status   = $Response->get_parsed_output();
        if (grep(/started/, @{$Status->[0]->{'status'}})) {
            $Zapi->cifs_stop();
        }

    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Response = $Zapi->cifs_setup_verify_passwd_and_group();
    $Status   = $Response->get_parsed_output();

    if (!grep(/passed/, @{$Status->[0]->{'status'}})) {
        try {
            $Zapi->cifs_setup_create_group_file();
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
        try {
            $Zapi->cifs_setup_create_passwd_file(
                'default-root-password' => 'netapp1!');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    my $cifs_server
        = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
        ->{'cifs_server'};

    my $user;
    if ($Options_hash->{'CIFS_USER'} =~ /\w+\\(.*)/) {
        $user = $Options_hash->{'CIFS_USER'};
    } else {
        $user = $Options_hash->{'CIFS_USER'} . '\\'
            . $Options_hash->{'CIFS_USER'};
    }

    # Run cifs setup on the current vfiler
    my %args;
    $args{'auth-type'}        = 'ad';
    $args{'cifs_pass'}        = 'netapp1!';
    $args{'create_admin'}     = 'yes';
    $args{'domain'}           = $Options_hash->{'CIFS_DOMAIN'};
    $args{'domain_user'}      = $Options_hash->{'CIFS_USER'};
    $args{'domain_pass'}      = $Options_hash->{'CIFS_PASSWD'};
    $args{'host'}             = host($node);
    $args{'passwd_group'}     = 'yes';
    $args{'select_site_name'} = 'yes';
    $args{'server-name'}      = $cifs_server;
    $args{'local_admins'}     = 'ssqa\\Domain Admins';
    if ($Options_hash->{'CIFS_SECURITY_STYLE'}) {
        $args{'security-style'} = $Options_hash->{'CIFS_SECURITY_STYLE'};
    }

    if ($vfiler ne 'vfiler0') { $args{'vfiler'} = $vfiler; }
    $args{'wins_option'} = 0;

    $Log->debug("cifs_setup args:\n" . Dumper(\%args));
    my $max_cifs_setup_attempts = 5;
    for (my $i = 1; $i <= $max_cifs_setup_attempts; $i++) {
        eval { cifs_setup(%args); };

        if ($@) {
            $Log->comment("Error detected in cifs setup");
            if ($i == $max_cifs_setup_attempts) {
                die config_result("Error configuring CIFS on vfiler $vfiler "
                        . "after $max_cifs_setup_attempts attempts");
            } else {
                $Log->comment(
                    "Waiting 60 seconds before retrying cifs setup");
                snooze 60;
            }
        } elsif (!$@) {
            last;
        }
    }
    $Log->exit();
} ## end sub configure_cifs

sub configure_dns {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Log->comment("Configuring DNS on $vfiler");

    # Create the resolv.conf file
    my $ascii
        = 'search ' . join(' ', @{$Options_hash->{'DNS_SEARCH'}}) . "\n";
    foreach my $ns (split(/,/, $Options_hash->{'DNS_SERVERS'})) {
        $ascii .= "nameserver $ns\n";
    }

    my $path = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
    if ($path eq '/etc') { $path = '/vol/vol0/etc'; }

    $self->write_file(
        'ascii' => $ascii,
        'node'  => $node,
        'path'  => "$path/resolv.conf",
    );

    snooze 1;

    my $rc_data;
    eval {
        $rc_data = $self->read_file(
            'path' => "$path/rc",
            'node' => $node
        );
    };

    my %data;
    $data{'dns.domainname'} = $Options_hash->{'CIFS_DOMAIN'};
    $data{'dns.enable'}     = 'on';
    foreach my $key (keys(%data)) {
        my $regex = $key;
        $regex =~ s/(\.)/\\$1/g;
        my $found = 0;
        my $new_rc_data;
        foreach my $line (split(/\n/, $rc_data)) {
            chomp($line);
            if ($line =~ /$regex/) {
                $new_rc_data .= "options $key $data{$key}\n";
                $found = 1;
            } else {
                $new_rc_data .= $line . "\n";
            }
        }
        if (!$found) {
            $new_rc_data .= "options $key $data{$key}\n";
        }
        $rc_data = $new_rc_data;
    }

    $Log->comment(Dumper($rc_data));
    $self->write_file(
        'ascii' => $rc_data,
        'node'  => $node,
        'path'  => "$path/rc"
    );

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    foreach my $key (keys(%data)) {
        try {

            $Response = $Zapi->options_set(
                'name'  => $key,
                'value' => $data{$key}
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    $Log->exit();
} ## end sub configure_dns

sub configure_nis {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Log->comment("Configuring NIS on $vfiler");

    my $path = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
    if ($path eq '/etc') { $path = '/vol/vol0/etc'; }

    my $rc_data;
    eval {
        $rc_data = $self->read_file(
            'path' => "$path/rc",
            'node' => $node
        );
    };

    my %data;
    if ($Options_hash->{'NIS_DOMAINNAME'} && $Options_hash->{'NIS_SERVERS'}) {
        $data{'nis.domainname'} = $Options_hash->{'NIS_DOMAINNAME'};
        $data{'nis.servers'}    = $Options_hash->{'NIS_SERVERS'};
        $data{'nis.enable'}     = 'on';
    }

    foreach my $key (keys(%data)) {
        my $regex = $key;
        $regex =~ s/(\.)/\\$1/g;
        my $found = 0;
        my $new_rc_data;
        foreach my $line (split(/\n/, $rc_data)) {
            chomp($line);
            if ($line =~ /$regex/) {
                $new_rc_data .= "options $key $data{$key}\n";
                $found = 1;
            } else {
                $new_rc_data .= $line . "\n";
            }
        }
        if (!$found) {
            $new_rc_data .= "options $key $data{$key}\n";
        }
        $rc_data = $new_rc_data;
    }

    $Log->comment(Dumper($rc_data));

    $self->write_file(
        'ascii' => $rc_data,
        'node'  => $node,
        'path'  => "$path/rc"
    );

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    foreach my $key (keys(%data)) {
        if ($key eq 'nis.enable') { next; }
        $Log->comment("Configuring options $key to $data{$key} on $vfiler");
        try {

            $Response = $Zapi->options_set(
                'name'  => $key,
                'value' => $data{$key}
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    try {

        $Response = $Zapi->options_set(
            'name'  => 'nis.enable',
            'value' => 'on'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->exit();
} ## end sub configure_nis

sub destroy_vfiler {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $defined_addresses
        = $self->{'defined_topology'}->{$node}->{'addresses'};
    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $discovered_interfaces
        = $self->{'discovered_topology'}->{$node}->{'interfaces'};
    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Log->comment("$vfiler is $discovered_vfilers->{$vfiler}->{'status'}");
    if (defined($discovered_vfilers->{$vfiler}->{'status'})
        && $discovered_vfilers->{$vfiler}->{'status'} =~ /stopped/) {
        $Log->comment("Destroying $vfiler");
        $self->vfiler_destroy(
            'node'    => $node,
            'vfiler'  => $vfiler,
            'retries' => 5,
            'wait'    => 120
        );
        snooze 1;
    }

    foreach
        my $interface (keys(%{$defined_vfilers->{$vfiler}->{'interfaces'}})) {
        my $defined_addresses_by_interface
            = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};
        foreach my $address_information (@{$defined_addresses_by_interface}) {

            my $address = $address_information->{'address'};
            my $range   = $address_information->{'range'};
            my $netmask
                = $defined_addresses->{$interface}->{$range}->{'NETMASK'};

            my $address_regex;
            if ($address =~ /\./) {
                $address_regex = $address;
                $address_regex =~ s/(\.)/\\$1/g;
            } else {
                my $ip = new Net::IP($address);
                $address       = $ip->short();
                $address_regex = $address;

            }

            # Delete the alias if needed
            if (grep { $_->{'ip'} eq $address }
                @{$discovered_interfaces->{$interface}->{'ipv4_info'}}) {
                $self->ifconfig_remove_alias(
                    'node'      => $node,
                    'interface' => $interface,
                    'address'   => $address,
                    'retries'   => 5,
                    'wait'      => 60
                );
                snooze 1;
            }

            my $rc_data = $self->read_file(
                'path' => '/vol/vol0/etc/rc',
                'node' => $node
            );
            my $new_rc_data;
            foreach my $line (split(/\n/, $rc_data)) {
                chomp($line);
                if ($line !~ /$address_regex/) {
                    $new_rc_data .= $line . "\n";
                }
            }
            $self->write_file(
                'ascii' => $new_rc_data,
                'node'  => $node,
                'path'  => '/vol/vol0/etc/rc'
            );

        }
    }

    $Log->exit();
} ## end sub destroy_vfiler

sub delete_shares {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $defined_shares  = $topology_database->{$node}->{'shares'};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Deleting shares on $vfiler");

    while (my $share = each %{$defined_shares}) {
        if ($vfiler ne $defined_shares->{$share}->{'vfiler'}) { next; }

        my $volume = $defined_shares->{$share}->{'volume'};
        my $path   = $defined_shares->{$share}->{'path'};
        $Log->comment("Deleting share $share on $volume");

        try {

            $Response = $Zapi->cifs_share_delete('share-name' => $share);
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    $Log->exit();
} ## end sub delete_shares

sub terminate_cifs {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Terminating CIFS on $vfiler");

    my $Status;
    try {

        $Response = $Zapi->cifs_status();
        $Status   = $Response->get_parsed_output();
        if (grep(/started/, @{$Status->[0]->{'status'}})) {
            $Zapi->cifs_stop();
        }

    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->exit();
} ## end sub terminate_cifs

# Add to the vfiler
sub vfiler_add {

    $Log->enter();

    my ($self, %options) = @_;
    my $node      = $options{'node'};
    my $vfiler    = $options{'vfiler'};
    my $addresses = $options{'addresses'};
    my $storage   = $options{'storage'};
    my $retries   = $options{'retries'};
    my $wait      = $options{'wait'};
    my $retry     = 1;
    my $Cli       = $apiset{'cli'}{$node};
    my %args;
    $args{'vfiler'} = $vfiler;

    if ($addresses) {
        $args{'ip-addresses'} = $addresses;
    }
    if ($storage) {
        $args{'storage-units'} = $storage;
    }
    while ($retry) {
        try {

            $Response = $Cli->vfiler_add(%args);
            $retry    = 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            my $output    = $Response->get_raw_output();
            my $text      = $Exception->text();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                die config_result("$text: $output");
            }
        };
        if ($retry == $retries) {
            die config_result(
                "Unable to create vfiler $vfiler after $retries tries");
        }
    }
    $Log->exit();
} ## end sub vfiler_add

# Stop the vfiler
sub vfiler_stop {

    $Log->enter();

    my ($self, %options) = @_;
    my $vfiler  = $options{'vfiler'};
    my $node    = $options{'node'};
    my $retries = $options{'retries'};
    my $wait    = $options{'wait'};
    my $retry   = 1;
    my $Cli     = $apiset{'cli'}{$node};
    while ($retry) {
        try {

            $Response = $Cli->vfiler_stop(vfiler => $vfiler);
            $retry = 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            my $output    = $Response->get_raw_output();
            my $text      = $Exception->text();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                die config_result("$text: $output");
            }
        };
        if ($retry == $retries) {
            $Log->warn("Unable to stop vfiler $vfiler after $retries tries");
            last;
        }
    }

    $Log->exit();
} ## end sub vfiler_stop

# Destroy the vfiler
sub vfiler_destroy {

    $Log->enter();

    my ($self, %options) = @_;
    my $vfiler  = $options{'vfiler'};
    my $node    = $options{'node'};
    my $retries = $options{'retries'};
    my $wait    = $options{'wait'};
    my $Cli     = $apiset{'cli'}{$node};
    my $retry   = 1;
    while ($retry) {
        try {

            $Response = $Cli->vfiler_destroy(
                vfiler => $vfiler,
                force  => 1
            );
            $retry = 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            my $output    = $Response->get_raw_output();
            my $text      = $Exception->text();
            if ($output =~ /cannot be run|active command in progress/) {
                $retry++;
                snooze $wait;
            } else {
                die config_result("$text: $output");
            }
        };
        if ($retry == $retries) {
            $Log->warn(
                "Unable to destroy vfiler $vfiler after $retries tries");
            last;
        }
    }

    $Log->exit();
} ## end sub vfiler_destroy

# Create the vfiler
sub vfiler_create {

    $Log->enter();

    my ($self, %options) = @_;
    my $node      = $options{'node'};
    my $vfiler    = $options{'vfiler'};
    my $addresses = $options{'addresses'};
    my $storage   = $options{'storage'};
    my $retries   = $options{'retries'};
    my $wait      = $options{'wait'};
    my $retry     = 1;
    my $Cli       = $apiset{'cli'}{$node};
    while ($retry) {
        try {

            $Response = $Cli->vfiler_create(
                vfiler          => $vfiler,
                "bypass-setup"  => 1,
                "ip-addresses"  => $addresses,
                "storage-units" => $storage
            );
            $retry = 0;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            my $output    = $Response->get_raw_output();
            my $text      = $Exception->text();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                die config_result("$text: $output");
            }
        };
        if ($retry == $retries) {
            die config_result(
                "Unable to create vfiler $vfiler after $retries tries");
        }
    }
    $Log->exit();
} ## end sub vfiler_create

sub ifconfig_remove_alias {

    $Log->enter();

    my ($self, %options) = @_;
    my $node      = $options{'node'};
    my $interface = $options{'interface'};
    my $address   = $options{'address'};
    my $retries   = $options{'retries'};
    my $wait      = $options{'wait'};
    my $retry     = 1;
    my $Cli       = $apiset{'cli'}{$node};
    while ($retry) {
        try {

            $Response = $Cli->ifconfig(
                'interface' => $interface,
                'unalias'   => $address
            );
            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                $retry = 0;
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result("Unable to remove alias $address on "
                    . "interface $interface after $retries tries");
        }
    }

    $Log->exit();
} ## end sub ifconfig_remove_alias

sub ifconfig_alias {

    $Log->enter();

    my ($self, %options) = @_;
    my $node             = $options{'node'};
    my $interface        = $options{'interface'};
    my $address          = $options{'address'};
    my $netmask          = $options{'netmask'} || undef;
    my $retries          = $options{'retries'};
    my $wait             = $options{'wait'};
    my $retry            = 1;
    my $Cli              = $apiset{'cli'}{$node};
    my %ifconfig_options = ('interface' => $interface, 'alias' => $address);
    if ($address =~ /\./ && $netmask) {
        $ifconfig_options{'netmask'} = $netmask;
    }
    while ($retry) {
        try {

            $Response = $Cli->ifconfig(%ifconfig_options);
            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                $retry = 0;
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result("Unable to alias $address on interface "
                    . "$interface after $retries tries");
        }
    }

    $Log->exit();
} ## end sub ifconfig_alias

sub ifconfig_remove_address {

    $Log->enter();

    my ($self, %options) = @_;
    my $node      = $options{'node'};
    my $interface = $options{'interface'};
    my $address   = $options{'address'};
    my $netmask   = $options{'netmask'} || undef;
    my $retries   = $options{'retries'};
    my $wait      = $options{'wait'};
    my $retry     = 1;
    my $Cli       = $apiset{'cli'}{$node};
    my %ifconfig_options
        = ('interface' => $interface, 'address' => $address, 'down' => 1);

    if ($address =~ /\./ && $netmask) {
        $ifconfig_options{'netmask'} = $netmask;
    }
    while ($retry) {
        try {

            $Response = $Cli->ifconfig(%ifconfig_options);
            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                $retry = 0;
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result("Unable to remove address $address on "
                    . "interface $interface after $retries tries");
        }
    }

    $Log->exit();
} ## end sub ifconfig_remove_address

sub ifconfig_address {

    $Log->enter();

    my ($self, %options) = @_;
    my $node             = $options{'node'};
    my $interface        = $options{'interface'};
    my $address          = $options{'address'};
    my $netmask          = $options{'netmask'} || undef;
    my $retries          = $options{'retries'};
    my $wait             = $options{'wait'};
    my $retry            = 1;
    my $Cli              = $apiset{'cli'}{$node};
    my %ifconfig_options = ('interface' => $interface, 'address' => $address);

    if ($address =~ /\./ && $netmask) {
        $ifconfig_options{'netmask'} = $netmask;
    }

    while ($retry) {
        try {

            $Response = $Cli->ifconfig(%ifconfig_options);

            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                $retry = 0;
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result("Unable to alias $address on interface "
                    . "$interface after $retries tries");
        }
    }

    $Log->exit();
} ## end sub ifconfig_address

sub ifconfig {

    $Log->enter();

    my ($self, %options) = @_;
    my $node    = $options{'node'};
    my $retries = $options{'retries'};
    my $wait    = $options{'wait'};
    my $retry   = 1;
    my $Cli     = $apiset{'cli'}{$node};
    my $Ifconfig;

    while ($retry) {
        try {

            $Response = $Cli->ifconfig(all => 1);
            $Ifconfig = $Response->get_parsed_output();
            my $output = $Response->get_raw_output();
            if ($output =~ /cannot be run/) {
                $retry++;
                snooze $wait;
            } else {
                $retry = 0;
            }
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            my $Response  = $Exception->get_response_object();
            die config_result(
                $Exception->text() . ": " . $Response->get_raw_output());
        };
        if ($retry == $retries) {
            die config_result("Unable to get ifconfig information "
                    . " after $retries tries");
        }
    }

    $Log->exit();

    return $Ifconfig;
} ## end sub ifconfig

sub define_shares {

    $Log->enter();

    my ($self, %options) = @_;
    my $script      = $options{'script'};
    my $share_index = $options{'share_index'};
    my $recipe      = $options{'recipe'};
    my $server;
    my @nodes
        = grep(/$Options_hash->{'NODE_REGEX'}/, @{$Options_hash->{'NODES'}});
    if (!@nodes) {
        die config_result("Unable to find any nodes using the "
                . "NODE_REGEX $Options_hash->{'NODE_REGEX'} against "
                . Dumper($Options_hash->{'NODES'}));
    }

    my %shares;
    my $vfiler_count = 0;
    foreach my $node (@nodes) {
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_shares = $topology_database->{$node}->{'shares'};
        my $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};
        my $discovered_shares
            = $self->{'discovered_topology'}->{$node}->{'shares'};

        # Count the existing ontap shares
        my @vfilers = keys(%{$discovered_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            $shares{$node}{$vfiler} = @ontap_shares;
        }

        # Count the existing discovered shares
        my @ds = keys(%{$discovered_shares});
        @ds = natsort(@ds);
        foreach my $s (@ds) {
            if (!$s) { next; }
            my $owner = $discovered_shares->{$s}->{'owning-vfiler'};
            if (!$owner) {
                die config_result(
                    "Unable to get owning vfiler from discovered share $s\n"
                        . Dumper($discovered_shares->{$s}));
            }

            # Ignore shares we created from previous runs. The existing
            # shares will be redfined but filer_setup will be set to 0
            # so the share is not recreated. This way if the test fail
            # for some reason it can pick up whre it left off.
            if ($s =~ /^s_$script/) { next; }

            # Add the count to the owning vfiler
            $shares{$node}{$owner} += 1;
        }

        # Count the to be created shares
        foreach my $vfiler (keys(%{$defined_vfilers})) {

            $vfiler_count++;

            # Count the to be created ontap shares
            $shares{$node}{$vfiler} = @ontap_shares;

            # Count all shares that belong to the current vfiler
            # and are not already created and not a data share
            # This is basically to pickup the volume shares
            while (my $s = each %{$defined_shares}) {
                my $owner = $defined_shares->{$s}->{'vfiler'};
                if (!$owner) {
                    die config_result(
                        "Unable to get owning vfiler from defined share\n"
                            . Dumper($defined_shares->{$s}));
                } elsif ($vfiler ne $owner) {
                    next;
                } elsif (
                    grep {
                        $_ eq $s
                    }
                    keys(%{$discovered_shares->{$node}->{'shares'}})
                    ) {
                    next;
                } elsif ($s =~ /^s_$script/) {
                    next;
                }
                $shares{$node}{$vfiler} += 1;
            }
        }
    }

    # Total all the shares found
    my $total_existing_shares = 0;
    foreach my $n (keys(%shares)) {
        foreach my $v (keys(%{$shares{$n}})) {
            $total_existing_shares += $shares{$n}{$v};
        }
    }

    if ($total_existing_shares >= $Options_hash->{'SHARE_COUNT'}) {
        die config_result("Existing shares ($total_existing_shares) is "
                . "greater or equal to the specified SHARE_COUNT "
                . "($Options_hash->{'SHARE_COUNT'}) ");
    }

    my $new_shares = $Options_hash->{'SHARE_COUNT'} - $total_existing_shares;
    my $shares_per_vfiler           = int($new_shares / $vfiler_count);
    my $shares_per_vfiler_remainder = $new_shares % $vfiler_count;
    my $shares_per_client           = 0;
    my $shares_per_client_remainder = 0;

    if ($recipe eq 'clients') {
        my @share_creation_clients
            = grep { host($_)->hosttype() =~ /win/ } @{$self->{'clients'}};

        my $max_clients = $self->{'max_clients'};
        if (!$max_clients) {
            $max_clients = @share_creation_clients;
        } elsif ($max_clients < @share_creation_clients) {
            splice(@share_creation_clients, $max_clients);
        }

        if (@share_creation_clients > $new_shares) {

            $shares_per_client = 1;
            splice(@share_creation_clients, $new_shares);
            $self->{'clients'} = \@share_creation_clients;

        } else {
            $shares_per_client = int($new_shares / @share_creation_clients);
            $shares_per_client_remainder
                = $new_shares % @share_creation_clients;
        }

        $self->{'shares_per_client'}           = $shares_per_client;
        $self->{'shares_per_client_remainder'} = $shares_per_client_remainder;

    }

    $self->{'existing_shares'}            = $total_existing_shares;
    $self->{'new_shares'}                 = $new_shares;
    $self->{'shares_per_vfiler'}          = $shares_per_vfiler;
    $self->{'shares_per_vfiler_reminder'} = $shares_per_vfiler_remainder;

    my $shares_per_volume = 0;
    if ($Options_hash->{'SHARES_PER_VOLUME'}) {
        $shares_per_volume = $Options_hash->{'SHARES_PER_VOLUME'};
    } else {
        $shares_per_volume
            = int($new_shares / $Options_hash->{'VOLUME_COUNT'});
        if ($new_shares / $Options_hash->{'VOLUME_COUNT'}) {
            $shares_per_volume++;
        }
    }

    $self->{'shares_per_volume'} = $shares_per_volume;

    #<<<
    my $summary = "Requested shares . . . . . . . . . . . . $Options_hash->{'SHARE_COUNT'}\n";
    $summary .= "Target nodes . . . . . . . . . . . . . . " . @nodes . "\n";
    $summary .= "Target vfilers   . . . . . . . . . . . . $vfiler_count\n";
    $summary .= "Existing shares  . . . . . . . . . . . . $total_existing_shares\n";
    $summary .= "Shares to create   . . . . . . . . . . . $new_shares\n";
    $summary .= "Shares per vfiler  . . . . . . . . . . . $shares_per_vfiler\n";
    $summary .= "Shares per vfiler remainder  . . . . . . $shares_per_vfiler_remainder\n";
    $summary .= "Shares per volume  . . . . . . . . . . . $shares_per_volume\n";
    if ($recipe eq 'clients') {
        $summary .= "Share creation clients . . . . . . . . . " . @{$self->{'clients'}} . "\n";
        $summary .= "Shares 1st client creates  . . . . . . . " . ($shares_per_client + $shares_per_client_remainder) . "\n";
        $summary .= "Shares created by remaining clients  . . $shares_per_client\n";
    }
    #>>>
    $Log->comment($summary);

    # Spread the shares across the nodes, vfilers and volumes
    $Log->comment("Adding shares to vfilers");
    foreach my $node (@nodes) {
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $discovered_shares
            = $self->{'discovered_topology'}->{$node}->{'shares'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};

        # Initialize the index for for the addresses for each vfiler's data
        # interface
        my $interface_index = 0;
        my %address_indexes;
        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
                $address_indexes{$interface} = 0;
            }
        }

        foreach my $vfiler (@vfilers) {
            if (!$vfiler) {
                die script_result(
                    "Blank vfiler detected in defined vfilers hash");
            }

            my $current_shares_per_vfiler = $shares_per_vfiler;
            if ($vfiler eq 'vfiler0' && $node eq $nodes[0]) {
                $current_shares_per_vfiler += $shares_per_vfiler_remainder;
            }

            my $vfiler_volumes
                = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                ->{'volumes'};
            my @volumes;
            my $volume;
            my $share_count = 0;
            $Log->comment(
                "Adding $current_shares_per_vfiler shares to vfiler $vfiler on node $node"
            );
            foreach (1 ... $current_shares_per_vfiler) {

                # Create the share name
                my $share = 's_' . $script . '_' . $share_index;
                $share_index++;

                my $filer_setup = 1;
                my $path;
                my $vf;
                my $vol;

                # If the share exists use it otherwise define a new share
                if (grep { $_ eq $share } keys(%{$discovered_shares})) {
                    $Log->debug("Found existing share $share\n"
                            . Dumper($discovered_shares->{$share}));
                    $path = $discovered_shares->{$share}->{'mount-point'};
                    $path =~ /^\/vol\/([^\/]+)\/.+/;
                    $vol = $1;
                    if (!$vol) {
                        die script_result("Unable to determine the volume "
                                . "from the share path $path");
                    }
                    $vf = $discovered_shares->{$share}->{'owning-vfiler'};
                    $filer_setup = 0;
                    delete($discovered_shares->{$share});

                } else {

                    # Get the volume name. If there are no more volumes in the
                    # array repopulate the array
                    if (!@volumes) {
                        foreach my $v (natsort(@{$vfiler_volumes})) {
                            my $root_volume
                                = $defined_volumes->{$v}->{'root_volume'};
                            if ($root_volume) { next; }
                            push(@volumes, $v);
                        }
                    }
                    if (!$volume) {
                        $volume = shift(@volumes);
                        if (!$volume) {
                            die script_result(
                                      "No available data volumes found on "
                                    . "node $node for vfiler $vfiler "
                                    . "Try increasing VOLUME_COUNT");
                        }
                    }
                    $vf  = $vfiler;
                    $vol = $volume;
                    if ($Options_hash->{'SHARE_PATH_CONFIG'} eq 'single') {
                        $path = '/vol/' . $vol;
                    } elsif (
                        $Options_hash->{'SHARE_PATH_CONFIG'} eq 'multiple') {
                        $path = '/vol/' . $vol . '/' . $share;
                    }
                    $Log->debug(
                        "Defining new share $share on vfiler $vf with a path of $path"
                    );
                }

                if (!$vf) {
                    die script_result("Unable to determine the vfiler for "
                            . "share $share on volume $vol\n"
                            . Dumper($discovered_shares->{$share}));
                }

                # Store the the volume name with the share
                $topology_database->{$node}->{'shares'}->{$share}->{'volume'}
                    = $vol;

                # Store the share path
                $topology_database->{$node}->{'shares'}->{$share}->{'path'}
                    = $path;

                # Store the vfiler name with the share
                $topology_database->{$node}->{'shares'}->{$share}->{'vfiler'}
                    = $vf;

                # Decide when we create the share
                $topology_database->{$node}->{'shares'}->{$share}
                    ->{'filer_setup'} = $filer_setup;

                # Don't use qtrees
                $topology_database->{$node}->{'shares'}->{$share}->{'qtree'}
                    = 0;

                push(
                    @{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                            ->{$vf}->{'shares'}
                        },
                    $share
                );

                $topology_database->{$node}->{'shares'}->{$share}->{'node'}
                    = $node;
                ($server, $interface_index) = $self->assign_cifs_server(
                    'node'            => $node,
                    'vfiler'          => $vf,
                    'address_indexes' => \%address_indexes,
                    'interface_index' => $interface_index
                );

                $Log->comment("Assigning CIFS server $server for "
                        . "share $share on vfiler $vfiler node "
                        . "$node");

                $topology_database->{$node}->{'shares'}->{$share}->{'server'}
                    = $server;

                $share_count++;
                if ($share_count == $current_shares_per_vfiler) {
                    $volume = undef;
                }
            }
        }
    }

    $Log->exit();
    return $share_index;
} ## end sub define_shares

sub assign_cifs_server {

    $Log->enter();

    my ($self, %options) = @_;

    my $share           = $options{'share'};
    my $node            = $options{'node'};
    my $vfiler          = $options{'vfiler'};
    my $interface_index = $options{'interface_index'};
    my $address_indexes = $options{'address_indexes'};

    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    # Get the vfiler address from the vfiler entry
    my $interface = ${$Options_hash->{'DATA_INTERFACES'}}[$interface_index];
    if (!$interface) {
        die script_result("Unable to get the interface from DATA_INTERFACES\n"
                . Dumper($Options_hash->{'DATA_INTERFACES'}));
    }

    # Increment or reset the interface index
    $interface_index++;
    if ($interface_index > $#{$Options_hash->{'DATA_INTERFACES'}}) {
        $interface_index = 0;
    }

    # Get the ip address from the defined vfilers interface
    $Log->debug("Getting an address from interface $interface "
            . "address index $$address_indexes{$interface} for "
            . "for vfiler $vfiler");

    my $defined_addresses_this_interface
        = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};

    if ($$address_indexes{$interface} > $#{$defined_addresses_this_interface})
    {
        $$address_indexes{$interface} = 0;
    }
    my $ip
        = ${$defined_addresses_this_interface}[$$address_indexes{$interface}]
        ->{'address'};

    if (!$ip) {
        die script_result("Could not get IP address from interface "
                . "$interface on node $node\n"
                . "Defined_ addresses this interface\n"
                . Dumper($defined_addresses_this_interface) . "\n"
                . "Address indexes\n"
                . Dumper($address_indexes)
                . "Defined vfiler\n"
                . Dumper($defined_vfilers->{$vfiler})
                . "Discovered vfiler\n"
                . Dumper($discovered_vfilers->{$vfiler}));
    }

    my $net_ip                = new Net::IP($ip);
    my $version               = $Options_hash->{'DESTIP_VERSION'};
    my $current_address_index = $$address_indexes{$interface};

    while ($version != $net_ip->version()) {
        $Log->debug("The version of $ip does not match DESTIP_VERSION. "
                . "Getting next address on interface $interface");

        $$address_indexes{$interface}++;
        if ($$address_indexes{$interface}
            > $#{$defined_addresses_this_interface}) {
            $$address_indexes{$interface} = 0;
        }
        if ($$address_indexes{$interface} == $current_address_index) {
            last;
        }
        $ip = ${$defined_addresses_this_interface}
            [$$address_indexes{$interface}]->{'address'};
        if (!$ip) { last; }
        $net_ip = new Net::IP($ip);
    }

    if (!$ip) {

        my $defined_addresses
            = $self->{'defined_topology'}->{$node}->{'addresses'};
        my $defined_vfiler = $defined_vfilers->{$vfiler};
        my $discovered_interfaces
            = $self->{'discovered_topology'}->{$node}->{'interfaces'};

        my $msg
            = sprintf("Unable to retrieve an IP address for vfiler "
                . "%s\n using index %d of interface %s\'s "
                . "defined addresses\n",
            $vfiler, $$address_indexes{$interface}, $interface);
        die config_result($msg
                . "\nDiscovered Interfaces on $node:\n"
                . Dumper($discovered_interfaces)
                . "\nDefined vfiler $vfiler on $node:\n"
                . Dumper($defined_vfiler)
                . "\nDefined address ranges:\n"
                . Dumper($defined_addresses));
    }

    # If there is more than one address range specified
    # for a data interface we should switch back and forth
    # to spread the load
    $$address_indexes{$interface}++;
    if ($$address_indexes{$interface} > $#{$defined_addresses_this_interface})
    {
        $$address_indexes{$interface} = 0;
    }

    my $mode = $Options_hash->{'CIFS_SERVER_ACCESS_MODE'};
    my $server = $self->get_cifs_server_name('ip' => $ip, 'mode' => $mode);

    $Log->exit();
    return $server;

} ## end sub assign_cifs_server

sub filer_cleanup_defined {

    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    foreach my $node (@{$Options_hash->{'NODES'}}) {
        $Log->comment("Cleaning up node $node");

        my $Zapi          = $apiset{'zapi'}{$node};
        my $definedshares = $self->{'defined_topology'}->{$node}->{'shares'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};

        my @targets = ('vfilers');
        my @nodes   = ($node);
        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        # Stop vfilers
        $Log->comment("Stopping vfilers");
        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            if (!grep { $_ eq $vfiler } keys(%{$discovered_vfilers})) {
                next;
            }
            my $status = $discovered_vfilers->{$vfiler}->{'status'};
            if ($vfiler eq 'vfiler0' || $status =~ /stopped/) { next; }
            $self->vfiler_stop(
                'node'    => $node,
                'vfiler'  => $vfiler,
                'retries' => 5,
                'wait'    => 60
            );
            snooze 1;

        }

        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        # Destroy vfilers
        $Log->comment("Destroying vfilers");
        foreach my $vfiler (@vfilers) {
            if (!grep { $_ eq $vfiler } keys(%{$discovered_vfilers})) {
                next;
            }
            if ($vfiler eq 'vfiler0') { next; }
            $self->destroy_vfiler(
                'node'    => $node,
                'vfiler'  => $vfiler,
                'retries' => 5,
                'wait'    => 60
            );

            snooze 1;
        }

        # Unconfigure vfiler0's data interfaces
        foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
            my $address
                = $defined_vfilers->{'vfiler0'}->{'interfaces'}->{$interface}
                ->{'address'};
            if ($address =~ /\./) {

                $self->ifconfig_remove_address(
                    'node'      => $node,
                    'interface' => $interface,
                    'address'   => '0.0.0.0',
                    'netmask'   => '55.255.255.255',
                    'wait'      => 60,
                    'retries'   => 5
                );
                snooze 1;
            } else {
                eval {
                    $self->ifconfig_remove_address(
                        'node'      => $node,
                        'interface' => $interface,
                        'address'   => '0:0:0:0:0:0:0:0',
                        'wait'      => 60,
                        'retries'   => 5
                    );
                };
                snooze 1;
            }
        }

        my $discovered_licenses
            = $self->{'discovered_topology'}->{$node}->{'licenses'};
        if ($discovered_licenses->{'cifs'} ne 'Not licensed') {
            $self->terminate_cifs('node' => $node, 'vfiler' => 'vfiler0');

            $self->cleanup_cifsconfig_share_file(
                'node'   => $node,
                'vfiler' => 'vfiler0'
            );

            $self->cleanup_lcl_groups_file(
                'node'   => $node,
                'vfiler' => 'vfiler0'
            );
        }

        @targets = ('volumes');
        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        my $defined = keys(%{$self->{'defined_topology'}->{'volumes'}});
        $Log->comment("Destroying $defined volumes");

        # Set the vfiler context for the next ZAPI call
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        foreach my $volume (keys(%{$defined_volumes})) {
            if (!grep { $_ eq $volume } keys(%{$discovered_volumes})) {
                next;
            }

            $Log->comment("Checking quota status on $volume");

            my $Status;
            try {
                $Response = $Zapi->quota_status('volume' => $volume);
                $Status = $Response->get_parsed_output();
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };

            if (grep { $_ eq 'on' } @{$Status->[0]->{'status'}}) {

                $Log->comment("Stopping quotas on $volume");

                try {

                    $Response = $Zapi->quota_off('volume' => $volume);
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }

            foreach my $quota (@{$defined_volumes->{$volume}->{'quotas'}}) {
                $quota =~ s/CIFS_USER/$Options_hash->{'CIFS_USER'}/g;
                my @fields = split(/:/, $quota);
                eval {
                    $Zapi->quota_delete_entry(
                        'quota-target' => $fields[0],
                        'quota-type'   => $fields[1],
                        'volume'       => $volume,
                        'qtree'        => ''
                    );
                };
            }

            if ($discovered_volumes->{$volume}->{'is-online'} ne 'false') {

                try {

                    $Response = $Zapi->volume_offline(
                        'name'       => $volume,
                        'cifs-delay' => 0
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }

            snooze 1;

            try {

                $Response = $Zapi->volume_destroy(
                    'name'  => $volume,
                    'force' => 1
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }
    }

    $Log->exit();
} ## end sub filer_cleanup_defined

sub filer_cleanup_discovered {

    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    $Log->comment("Cleaning up discovered configuration");

    foreach my $node (@{$Options_hash->{'NODES'}}) {
        my $Cli = $apiset{'cli'}{$node};
        $Log->comment("Cleaning up node $node");

        my $Zapi = $apiset{'zapi'}{$node};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my ($discovered_vfilers, $discovered_interfaces,
            $discovered_volumes, $discovered_shares,
            $discovered
        );

        try {
            $Response = $Cli->setral(
                'reset'           => 1,
                'privilege-level' => "diag"
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        my @targets = ('vfilers', 'licenses', 'volumes', 'qtrees');
        my @nodes = ($node);

        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};
        $discovered = keys(%{$discovered_vfilers});
        $Log->comment("Cleaning up $discovered vfilers");

        my @vfilers = keys(%{$discovered_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {

            # Set the vfiler context for the next ZAPI call
            try {
                $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };

            # Destroy all discovered volumes quotas
            $Log->comment(
                "Disabling quotas on volumes owned by vfiler $vfiler");
            foreach my $volume (keys(%{$discovered_volumes})) {
                if (   $volume eq 'vol0'
                    || $discovered_volumes->{$volume}->{'owning-vfiler'} ne
                    $vfiler) {
                    next;
                }

                if ($discovered_volumes->{$volume}->{'state'} eq 'offline') {
                    next;
                }

                $Log->comment("Checking quota status on $volume");

                my $Status;
                try {
                    $Response = $Zapi->quota_status('volume' => $volume);
                    $Status = $Response->get_parsed_output();
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                if (grep { $_ eq 'on' } @{$Status->[0]->{'status'}}) {
                    $Log->comment("Stopping quotas on $volume");

                    try {

                        $Response = $Zapi->quota_off('volume' => $volume);
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }
            }

            my $path
                = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
            if ($path eq '/etc') { $path = '/vol/vol0/etc'; }
            if ($path !~ /\/vol/) {
                die script_result(
                          "Cannot determine the path to the configuration"
                        . "directory of vfiler $vfiler\n"
                        . Dumper($discovered_vfilers->{$vfiler}));
            }

            $Log->comment("Cleaning quota file $path");
            $self->truncate_file(
                'node'   => $node,
                'vfiler' => $vfiler,
                'path'   => "$path/quotas"
            );

        }

        # Stop vfilers
        $Log->comment("Stopping vfilers");
        if (keys(%{$discovered_vfilers}) == 0) {
            $Log->comment(
                "No vfilers discovered " . Dumper($discovered_vfilers));
        }

        foreach my $vfiler (@vfilers) {
            my $status = $discovered_vfilers->{$vfiler}->{'status'};
            if ($vfiler eq 'vfiler0' || $status =~ /stopped/) { next; }
            $self->vfiler_stop(
                'node'    => $node,
                'vfiler'  => $vfiler,
                'retries' => 5,
                'wait'    => 60
            );
            snooze 1;

        }

        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        # Destroy all vfilers
        $Log->comment("Destroying vfilers");
        foreach my $vfiler (@vfilers) {
            if ($vfiler eq 'vfiler0') { next; }
            $self->destroy_vfiler(
                'node'    => $node,
                'vfiler'  => $vfiler,
                'retries' => 5,
                'wait'    => 60
            );

            snooze 5;
        }

        # Unconfigure vfiler0's data interfaces
        @targets = ('interfaces');
        @nodes   = ($node);

        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        $discovered_interfaces
            = $self->{'discovered_topology'}->{$node}->{'interfaces'};

        foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
            if (!grep { $_ eq $interface }
                natsort(keys(%{$discovered_interfaces}))) {
                $Log->comment("$interface not found on $node, skipping");
                next;
            }

            if (grep { $_->{'ip'} =~ /\./ }
                @{$discovered_interfaces->{$interface}->{'ipv4_info'}}) {

                $self->ifconfig_remove_address(
                    'node'      => $node,
                    'interface' => $interface,
                    'address'   => '0.0.0.0',
                    'netmask'   => '55.255.255.255',
                    'wait'      => 60,
                    'retries'   => 5
                );
                snooze 1;
            }

            if (grep { $_->{'ip'} =~ /:/ }
                @{$discovered_interfaces->{$interface}->{'ipv6_info'}}) {

                $self->ifconfig_remove_address(
                    'node'      => $node,
                    'interface' => $interface,
                    'address'   => '0:0:0:0:0:0:0:0',
                    'wait'      => 60,
                    'retries'   => 5
                );
                snooze 1;
            }
        }

        my @targets = ('vfilers', 'licenses', 'volumes', 'qtrees');
        my @nodes = ($node);

        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );
        $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        my $discovered_licenses
            = $self->{'discovered_topology'}->{$node}->{'licenses'};
        if ($discovered_licenses->{'cifs'} ne 'Not licensed') {
            $self->terminate_cifs('node' => $node, 'vfiler' => 'vfiler0');

            $self->delete_cifsconfig_share_file(
                'node'   => $node,
                'vfiler' => 'vfiler0'
            );

            $self->cleanup_lcl_groups_file(
                'node'   => $node,
                'vfiler' => 'vfiler0'
            );
        }

        @targets = ('volumes');
        $self->discover_topology(
            'targets' => \@targets,
            'nodes'   => \@nodes
        );

        $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};

        $discovered = keys(%{$discovered_volumes});
        if ($discovered > 1) { $discovered--; }
        $Log->comment("Destroying $discovered volumes");

        # Set the vfiler context for the next ZAPI call
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        # Destroy all discovered volume
        foreach my $volume (keys(%{$discovered_volumes})) {
            if ($volume eq 'vol0') { next; }
            if (!grep(/^$volume$/, keys(%{$discovered_volumes}))) {
                next;
            }

            if ($discovered_volumes->{$volume}->{'state'} ne 'offline') {
                $Log->comment("Offlining $volume");

                try {

                    $Response = $Zapi->volume_offline(
                        'name'       => $volume,
                        'cifs-delay' => 0
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }

            snooze 1;

            $Log->comment("Destroying $volume");
            try {

                $Response = $Zapi->volume_destroy(
                    'name'  => $volume,
                    'force' => 1
                );
            }
            catch NACL::APISet::Exceptions::ResponseException with {
                my $Exception = shift;
                die config_result($Exception->text());
            };
        }

    }

    $Log->exit();
} ## end sub filer_cleanup_discovered

sub cleanup_cifsconfig_share_file {

    $Log->enter();

    my ($self, %options) = @_;
    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};
    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Options_hash = $Options->get_hash();

    $Log->comment("Reseting cifsconfig_share.cfg on vfiler $vfiler");

    #<<<
    my $ascii = 'cifs shares -add "ETC$" "/etc" -comment "Remote Administration"' . "\n";
    $ascii   .= 'cifs access "ETC$" S-1-5-32-544 Full Control' . "\n";
    $ascii   .= 'cifs shares -add "C$" "/" -comment "Remote Administration"' . "\n";
    $ascii   .= 'cifs access "C$" S-1-5-32-544 Full Control' . "\n";
    #>>>
    my $path = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
    if ($path eq '/etc') { $path = '/vol/vol0/etc'; }
    $path .= '/cifsconfig_share.cfg';

    # Set the vfiler context for the next ZAPI calls
    my $Zapi = $apiset{'zapi'}{$node};

    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $self->write_file(
        'ascii' => $ascii,
        'node'  => $node,
        'path'  => $path
    );

    $Log->exit();

} ## end sub cleanup_cifsconfig_share_file

sub delete_cifsconfig_share_file {

    $Log->enter();

    my ($self, %options) = @_;
    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};
    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Options_hash = $Options->get_hash();

    $Log->comment("Delete cifsconfig_share.cfg on vfiler $vfiler");

    my $path = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
    if ($path eq '/etc') { $path = '/vol/vol0/etc'; }
    $path .= '/cifsconfig_share.cfg';

    # Set the vfiler context for the next ZAPI calls
    my $Zapi = $apiset{'zapi'}{$node};

    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    eval {

        $Response = $Zapi->file_delete_file('path' => $path);
    };

    $Log->exit();

} ## end sub delete_cifsconfig_share_file

sub cleanup_lcl_groups_file {

    $Log->enter();

    my ($self, %options) = @_;
    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};
    my $Zapi   = $apiset{'zapi'}{$node};
    my $discovered_vfilers
        = $self->{'discovered_topology'}->{$node}->{'vfilers'};

    $Options_hash = $Options->get_hash();

    $Log->comment("Reseting lclgroups.cfg on vfiler $vfiler");

    #<<<
    my $ascii = '[ "Replicators"  552 ( "not supported" ) ]' . "\n";
      $ascii .= '[ "Backup Operators"  551 ( "Members can bypass file security to backup files" ) ]' . "\n";
      $ascii .= '[ "Power Users"  547 ( "Members that can share directories" ) ]' . "\n";
      $ascii .= '[ "Guests"  546 ( "Users granted Guest Access" ) ]' . "\n";
      $ascii .= '[ "Users"  545 ( "Ordinary Users" ) ]' . "\n";
      $ascii .= '[ "Administrators"  544 ( "Members can fully administer the filer" ) ]' . "\n";
    #>>>
    my $path = $discovered_vfilers->{$vfiler}->{'configuration_directory'};
    if ($path eq '/etc') { $path = '/vol/vol0/etc'; }
    $path .= '/lclgroups.cfg';

    # Set the vfiler context for the next ZAPI calls
    $Zapi = $apiset{'zapi'}{$node};
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $self->write_file(
        'ascii' => $ascii,
        'node'  => $node,
        'path'  => $path
    );

    $Log->exit();

} ## end sub cleanup_lcl_groups_file

sub truncate_file {

    $Log->enter();

    my ($self, %options) = @_;
    my $node = $options{'node'};
    my $path = $options{'path'};
    my $Zapi = $apiset{'zapi'}{$node};

    $Options_hash = $Options->get_hash();

    $Log->comment("Truncating $path");

    try {

        $Response = $Zapi->file_truncate_file('path' => $path);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        if ($Exception->text() !~ /No such file or directory/) {
            die config_result($Exception->text());
        }
    };

    $Log->exit();

} ## end sub truncate_file

sub time_setup {

    $Log->enter();

    my ($self, %options) = @_;
    my $node = $options{'node'};
    my $Zapi = $apiset{'zapi'}{$node};

    host($Options_hash->{'CIFS_DC'});
    my $conn_obj = connect('#rsh', '-nologin', -timeout, 300);

    $Log->comment("Getting date from $Options_hash->{'CIFS_DC'}");
    my ($date, $output);
    my $cmd = "cmd /c date /t";
    eval { (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 300); };

    foreach my $line (split(/\n/, $output)) {
        chomp($line);

        # Tue 07/17/2007
        if ($line =~ /\w+\s+(.+)/) {
            $date = $1;
        }
    }

    $Log->comment("Getting time from $Options_hash->{'CIFS_DC'}");
    $cmd = "cmd /c time /t";
    eval { (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 300); };

    foreach my $line (split(/\n/, $output)) {
        chomp($line);

        # 07:00 PM
        if ($line =~ /^(\d+):(\d+)\s+(\w+)/) {
            my $hour     = $1;
            my $minutes  = $2;
            my $day_time = $3;
            if ($day_time =~ /AM/i && $hour == 12) { $hour = '00'; }
            if ($day_time =~ /PM/i && $hour != 12) { $hour = $hour + 12; }
            $date .= " $hour:$minutes:00";
        }
    }

    $Log->comment("Configuring time on $node");
    $Log->comment("Disabling NTP");
    try {

        $Response = $Zapi->options_set(
            'name'  => 'timed.enable',
            'value' => 'off'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Setting time servers");
    try {

        $Response = $Zapi->options_set(
            'name'  => 'timed.servers',
            'value' => $Options_hash->{'TIME_SERVER'}
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Setting time protocol");
    try {

        $Response = $Zapi->options_set(
            'name'  => 'timed.proto',
            'value' => 'ntp'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Setting timezone");
    try {

        $Response = $Zapi->clock_set_timezone(
            'timezone' => $Options_hash->{'TIME_ZONE'});
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    my $time = $self->get_epoch_time('date' => $date);
    $Log->comment("Setting time to epoch time $time");
    try {

        $Response = $Zapi->clock_set_clock(
            'is-utc-clock' => 0,
            'time'         => $time
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    try {

        $Response = $Zapi->options_set(
            'name'  => 'timed.enable',
            'value' => 'on'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->exit();
} ## end sub time_setup

sub read_file {

    $Log->enter();

    my ($self, %options) = @_;

    my $node = $options{'node'};
    my $path = $options{'path'};

    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Getting file info for $path");
    my $size;
    try {

        $Response = $Zapi->file_get_file_info('path' => $path);
        my $file_info = $Response->get_parsed_output();
        $size = $file_info->[0]->{'file-info'}->[0]->{'file-size'};
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Reading $path");
    my $data;
    try {

        $Response = $Zapi->file_read_file(
            'length' => $size,
            'offset' => 0,
            'path'   => $path
        );
        my $file_info = $Response->get_parsed_output();
        $data = $file_info->[0]->{'data'};
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->exit();
    return pack("H*", $data);

} ## end sub read_file

sub write_file {

    $Log->enter();

    my ($self, %options) = @_;

    my $ascii     = $options{'ascii'};
    my $node      = $options{'node'};
    my $path      = $options{'path'};
    my $offset    = $options{'offset'} || 0;
    my $overwrite = $options{'overwrite'} || 'true';

    my $Zapi = $apiset{'zapi'}{$node};
    my $Cli  = $apiset{'cli'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    try {
        if ($overwrite eq 'true') {
            $self->truncate_file(
                'node'   => $node,
                'vfiler' => 'vfiler0',
                'path'   => $path
            );
            snooze 1;
        }

    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->comment("Writing file $path");

    try {

        $Response = $Zapi->file_write_file(
            'data'      => unpack("H*", $ascii),
            'offset'    => $offset,
            'overwrite' => $overwrite,
            'path'      => $path
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        eval {
            foreach my $line (split(/\n/, $ascii)) {
                $Cli->wrfile(
                    append    => 1,
                    'content' => "$line",
                    'file'    => $path,
                );
            }
        };
        if ($@) { die config_result($@); }
    };

    $Log->exit();
} ## end sub write_file

sub assign_share_acl_clients {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Assigning test options for share ACLs");

    $Options_hash = $Options->get_hash();
    my $instance     = 0;
    my $client_count = 0;

    for my $client (@{$self->{'clients'}}) {

        if (host($client)->hosttype()
            !~ /$Options_hash->{'CLIENT_HOSTTYPE_REGEX'}/) {
            next;
        }
        $client_count++;
        $Log->comment("Client: $client");
        $Log->comment("Instance: $instance");

        $self->{'clients_status'}->{$client}->{'status'} = 'assigned';
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_CONNECT_METHOD'}
            = $Options_hash->{'CLIENT_CONNECT_METHOD'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_HOSTTYPE_REGEX'}
            = $Options_hash->{'CLIENT_HOSTTYPE_REGEX'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'TEST_MAIN_DURATION'} = $Options_hash->{'TEST_MAIN_DURATION'};
        $self->{'test_options'}->{$client}->{$instance}->{'TEST_MAIN_SLEEP'}
            = $Options_hash->{'TEST_MAIN_SLEEP'};

        if ($client_count == $self->{'max_clients'}) {
            last;
        }
    }

    $Log->exit();
} ## end sub assign_share_acl_clients

sub list_shares {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Listing shares on vfiler $options{'vfiler'}");

    filer_vfiler_context_simulate($options{'conn'}, $options{'vfiler'});

    my $output;
    my $cmd = filer_vfiler_command($options{'conn'}, 'cifs shares');
    eval {
        (undef, undef, $output) = say($options{'conn'}, $cmd, -timeout, 300);
    };

    if ($@) {
        $Log->comment("$cmd\n $output");
        die "Error listing cifs shares\n";
    } else {
        $Log->debug("Command output: \n$output");
    }

    # Search the commands output
    my (%shares, @share_info);
    foreach my $line (split(/\n/, $output)) {

        #Name         Mount Point                       Description
        #----         -----------                       -----------
        #ETC$         /etc                              Remote Administration
        #                        BUILTIN\Administrators / Full Control
        #C$           /                                 Remote Administration
        #			     BUILTIN\Administrators / Full Control

        if ($line =~ /^(\S+)\s+(\/\S+)\s+/) {
            $shares{'share_name'} = $1;
            $shares{'path'}       = $2;
            if (   $options{'filter'}
                && $shares{'share_name'} =~ /$options{'filter'}/) {
                push(@share_info, \%shares);
            } elsif (!$options{'filter'}) {
                push(@share_info, \%shares);
            }
        }
    }

    $Log->exit();
    return \@share_info;
} ## end sub list_shares

sub list_share_acls {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment(
        "Listing share ACLs for $options{'share'} on vfiler $options{'vfiler'}"
    );

    filer_vfiler_context_simulate($options{'conn'}, $options{'vfiler'});

    my $output;
    my $cmd = filer_vfiler_command($options{'conn'},
        "cifs shares $options{'share'}");
    eval {
        (undef, undef, $output) = say($options{'conn'}, $cmd, -timeout, 300);
    };

    if ($@) {
        $Log->comment("$cmd\n $output");
        die "Error listing cifs share ACLS\n";
    } else {
        $Log->debug("Command output: \n$output");
    }

    # Search the commands output
    my @share_info;
    foreach my $line (split(/\n/, $output)) {

        #Name         Mount Point                       Description
        #----         -----------                       -----------
        #C$           /                                 Remote Administration
        #			     BUILTIN\Administrators / Full Control

        if ($line =~ /^\s+(\S+)\s+\/\s+(.*)/) {
            my %shares;
            $shares{'share_name'}    = $options{'share'};
            $shares{'user_or_group'} = $1;
            $shares{'permission'}    = $2;
            push(@share_info, \%shares);
        }
    }

    $Log->exit();
    return \@share_info;
    $Log->exit();
} ## end sub list_share_acls

sub create_share_acl {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Creating share ACL $options{'permission'} for "
            . "group $options{'user_or_group'} on $options{'share'}");

    filer_vfiler_context_simulate($options{'conn'}, $options{'vfiler'});

    my $output;
    my $cmd = filer_vfiler_command($options{'conn'},
              'cifs access '
            . $options{'share'} . ' "'
            . $options{'user_or_group'} . '" ' . '"'
            . $options{'permission'}
            . '"');
    eval {
        (undef, undef, $output) = say($options{'conn'}, $cmd, -timeout, 300);
    };

    if ($@) {
        $Log->comment("$cmd\n $output");
        die "Error creating share ACL\n";
    } else {
        $Log->debug("Command output: \n$output");
    }

    $Log->exit();
} ## end sub create_share_acl

sub delete_share_acl {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Deleting share ACL for user or group "
            . "$options{'user_or_group'} on $options{'share'}");

    filer_vfiler_context_simulate($options{'conn'}, $options{'vfiler'});

    my $output;
    my $cmd = filer_vfiler_command($options{'conn'},
              'cifs access '
            . $options{'share'} . ' "'
            . $options{'user_or_group'} . '" ' . '"'
            . $options{'permission'}
            . '"');
    eval {
        (undef, undef, $output) = say($options{'conn'}, $cmd, -timeout, 300);
    };

    if ($@) {
        $Log->comment("$cmd\n $output");
        die "Error creating share ACL\n";
    } else {
        $Log->debug("Command output: \n$output");
    }

    $Log->exit();
} ## end sub delete_share_acl

sub modify_share_acl {

    $Log->enter();

    my ($self, %options) = @_;

    $Log->comment("Modifying share ACL $options{'permission'} for "
            . "group $options{'user_or_group'} on $options{'share'}");

    filer_vfiler_context_simulate($options{'conn'}, $options{'vfiler'});

    my $output;
    my $cmd = filer_vfiler_command($options{'conn'},
              'cifs access '
            . $options{'share'} . ' "'
            . $options{'user_or_group'} . '" ' . '"'
            . $options{'permission'}
            . '"');
    eval {
        (undef, undef, $output) = say($options{'conn'}, $cmd, -timeout, 300);
    };

    if ($@) {
        $Log->comment("$cmd\n $output");
        die "Error modifying share ACL\n";
    } else {
        $Log->debug("Command output: \n$output");
    }

    $Log->exit();
} ## end sub modify_share_acl

sub ra_setup {

    $Log->enter();

    my ($self, %options) = @_;

    foreach my $node (@{$Options_hash->{'NODES'}}) {
        $Log->comment(
            "Configuring resource agitation on node $options{'node'}");

        filer_vfiler_context_simulate($conn{$node}, 'vfiler0');

        my $Cli = $apiset{'cli'}{$node};

        foreach my $r (keys(%{$options{'ra_config'}})) {
            if (!grep(/$r/, @{$Options_hash->{'RA_IGNORE'}})) {

                try {
                    $Response = $Cli->setral(
                        $r                => ${$options{'ra_config'}}{$r},
                        'privilege-level' => "diag"
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                if (0) {
                    my $output;
                    my $cmd = filer_vfiler_command($options{'conn'},
                              'priv set advanced; setral ' 
                            . $r . ' '
                            . ${$options{'ra_config'}}->{$r});
                    eval {
                        (undef, undef, $output)
                            = say($conn{$node}, $cmd, -timeout, 300);
                    };

                    if ($@) {
                        $Log->comment("$cmd\n $output");
                        die "Error configuring resource agitation\n";
                    } else {
                        $Log->debug("Command output: \n$output");
                    }
                }
            }
        }
    }

    $Log->exit();
} ## end sub ra_setup

sub create_local_groups {

    $Log->enter();

    my ($self, %options) = @_;

    my $node                 = $options{'node'};
    my $vfiler               = $options{'vfiler'};
    my $group_basename       = $options{'group_basename'};
    my $starting_group_index = $options{'starting_group_index'};
    my $ending_group_index   = $options{'ending_group_index'};
    my $group_members        = $options{'group_members'};

    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $defined_shares  = $topology_database->{$node}->{'shares'};
    my $discovered_qtrees
        = $self->{'discovered_topology'}->{$node}->{'qtrees'};
    my $discovered_shares
        = $self->{'discovered_topology'}->{$node}->{'shares'};

    $Log->comment("Creating "
            . ($ending_group_index - $starting_group_index + 1)
            . " local groups on node $node vfiler $vfiler");

    my @interfaces = keys(%{$defined_vfilers->{$vfiler}->{'interfaces'}});
    my $interface;
    foreach my $i (@interfaces) {
        if (grep { $_ eq $i } @{$Options_hash->{'DATA_INTERFACES'}}) {
            $interface = $i;
            last;
        }
    }

    if (!$interface) {
        die config_result("Unable to find an interface for $vfiler "
                . Dumper($defined_vfilers->{$vfiler}->{'interfaces'}));
    }

    my $ip = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface}->[0]
        ->{'address'};

    my $server;
    if ($ip =~ /:/) {
        $server = $self->get_cifs_server_name('ip' => $ip, 'mode' => 'name');
    } else {
        $server = $ip;
    }

    my @clients
        = grep { host($_)->hosttype() =~ /win/ } @{$self->{'clients'}};
    if (!@clients) {
        die config_result(
            "There are no windows clients to create local groups");
    }

    host($clients[0]);
    my $conn_obj = connect($Options_hash->{'CLIENT_CONNECT_METHOD'},
        '-nologin', -timeout, 300);
    eval { check_connectivity('conn' => $conn_obj); };
    my $error = $@;
    if ($error) { die config_result($error); }

    map_share(
        'conn'     => $conn_obj,
        'share'    => "\\\\$server\\c\$",
        'drive'    => '',
        'username' => $Options_hash->{'CIFS_USER'},
        'password' => $Options_hash->{'CIFS_PASSWD'},
        'timeout'  => 300,
        'reuse'    => 1,
        'delay'    => 60,
        'attempts' => 3,
        'tests'    => 'r'
    );

    my $cmd = 'cmd /c del \\lclgroups.txt';

    my $output;

    eval { (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 60); };

    $error = $@;

    if ($error) {
        die config_result(
            "Error creating add users input file:\n$output\n$error");
    }

    snooze 1;

    $cmd
        = 'cscript x:\\agnostic\\tools\\windows\\vbs\\createAddUsersFile.vbs '
        . '\\lclgroups.txt '
        . "$group_basename  $starting_group_index $ending_group_index "
        . "$group_members";

    eval { (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 60); };

    $error = $@;

    if ($error) {
        die config_result(
            "Error creating add users input file:\n$output\n$error");
    }

    snooze 1;

    $cmd = 'x:\\bin\\addusers \\\\' . $server . ' /c \\lclgroups.txt';

    eval { (undef, undef, $output) = say($conn_obj, $cmd, -timeout, 3600); };

    $error = $@;

    if ($error) {
        die config_result(
            "Error creating add users input file:\n$output\n$error");
    }

    $Log->exit();
} ## end sub create_local_groups

sub assign_stream_clients {

    $Log->enter();

    my ($self, %options) = @_;
    my $instance = 0;
    my $share_index;
    my @volumes;
    my %nodes;
    my %address_indexes;
    my @target_nodes
        = grep(/$Options_hash->{'NODE_REGEX'}/, @{$Options_hash->{'NODES'}});

    if (!@target_nodes) {
        die config_result("Unable to find any nodes using the NODE_REGEX "
                . $Options_hash->{'NODE_REGEX'}
                . " against "
                . Dumper($Options_hash->{'NODES'}));
    }

    foreach my $node (@target_nodes) {
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        foreach my $volume (natsort(keys(%{$defined_volumes}))) {
            my $root_volume = $defined_volumes->{$volume}->{'root_volume'};
            if ($root_volume) { next; }
            push(@volumes, $volume);
            $nodes{$volume} = $node;
        }
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
                $address_indexes{$interface} = 0;
            }
        }
    }

    # Loop through each client and assign the correct info to it.
    my $client_count    = 0;
    my $interface_index = 0;
    for my $client (@{$self->{'clients'}}) {

        if (host($client)->hosttype()
            !~ /$Options_hash->{'CLIENT_HOSTTYPE_REGEX'}/) {
            next;
        }

        $self->{'clients_status'}->{$client}->{'status'} = 'assigned';
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_CONNECT_METHOD'}
            = $Options_hash->{'CLIENT_CONNECT_METHOD'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'CLIENT_HOSTTYPE_REGEX'}
            = $Options_hash->{'CLIENT_HOSTTYPE_REGEX'};
        $self->{'test_options'}->{$client}->{$instance}
            ->{'TEST_MAIN_DURATION'} = $Options_hash->{'TEST_MAIN_DURATION'};
        $self->{'test_options'}->{$client}->{$instance}->{'TEST_MAIN_SLEEP'}
            = $Options_hash->{'TEST_MAIN_SLEEP'};

        # Get the next volume
        my $volume;
        if (@volumes) {
            $volume = shift(@volumes);
        } else {
            die script_result(
                      "Not enough volumes for $client_count clients,\n"
                    . "please specify a higher VOLUME_COUNT or change\n"
                    . "your NODE_REGEX to include more nodes or\n"
                    . "decrease the number of clients");
        }

        # Get the related information
        my $share = 's_' . $volume;
        my $node  = $nodes{$volume};
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $vfiler = $defined_volumes->{$volume}->{'vfiler'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};

        # Get the vfiler address from the vfiler entry
        my $interface
            = ${$Options_hash->{'DATA_INTERFACES'}}[$interface_index];

        # Increment or reset the interface index
        $interface_index++;
        if ($interface_index >= $#{$Options_hash->{'DATA_INTERFACES'}}) {
            $interface_index = 0;
        }

        # Get the ip address from the defined vfilers interface
        my $defined_addresses_this_interface
            = $defined_vfilers->{$vfiler}->{'interfaces'}->{$interface};

        if ($address_indexes{$interface}
            > $#{$defined_addresses_this_interface}) {
            $address_indexes{$interface} = 0;
        }

        my $ip = ${$defined_addresses_this_interface}
            [$address_indexes{$interface}]->{'address'};
        if (!$ip) {

            my $defined_addresses
                = $self->{'defined_topology'}->{$node}->{'addresses'};
            my $defined_vfiler = $defined_vfilers->{$vfiler};
            my $discovered_interfaces
                = $self->{'discovered_topology'}->{$node}->{'interfaces'};

            my $msg
                = sprintf("Unable to retrieve an IP address for vfiler "
                    . "%s\n using index %d of interface %s\'s "
                    . "defined addresses\n",
                $vfiler, $address_indexes{$interface}, $interface);
            die config_result($msg
                    . "\nDiscovered Interfaces on $node:\n"
                    . Dumper($discovered_interfaces)
                    . "\nDefined vfiler $vfiler on $node:\n"
                    . Dumper($defined_vfiler)
                    . "\nDefined address ranges:\n"
                    . Dumper($defined_addresses));
        }

        # If there is more than one address range specified
        # for a data interface we should switch back and forth
        # to spread the load
        $address_indexes{$interface}++;
        if ($address_indexes{$interface}
            > $#{$defined_addresses_this_interface}) {
            $address_indexes{$interface} = 0;
        }

        my $mode = $Options_hash->{'CIFS_SERVER_ACCESS_MODE'};
        my $server
            = $self->get_cifs_server_name('ip' => $ip, 'mode' => $mode);

        $Log->comment("Client: $client");
        $Log->comment("Volume: $volume");
        $Log->comment("Vfiler: $vfiler");
        $Log->comment("Share: $share");
        $Log->comment("Node: $node");
        $Log->comment("IP: $ip");
        $Log->comment("Server: $server");

        $self->update_test_options(
            'client'   => $client,
            'instance' => $instance,
            'node'     => $node,
            'vfiler'   => $vfiler,
            'server'   => $server,
            'volume'   => $volume,
            'share'    => $share
        );

        $client_count++;
        if ($client_count == $self->{'max_clients'}) { last; }
    }

    $Log->exit();
    return $share_index++;
} ## end sub assign_stream_clients

sub is_option_valid {

    $Log->enter();

    my ($self, %options) = @_;
    my $Host   = host($options{'node'});
    my %fields = $Host->param_fields();

    $Log->comment(
        "Checking option $options{'option'} against release $self->{'release'}"
    );

    if ($options{'option'} eq 'licensed_feature.multistore.enable') {
        if (   $self->{'release'} !~ /boilermaker/
            && $fields{'MODEL'} =~ /VSA/) {
            $Log->comment("$options{'option'} is valid on this paltform");
            $Log->exit();
            return 1;
        }
    }
    $Log->exit();
    return 0;
} ## end sub is_option_valid

sub configure_quotas {

    $Log->enter();

    my ($self, %options) = @_;

    my $node   = $options{'node'};
    my $vfiler = $options{'vfiler'};

    my $Zapi = $apiset{'zapi'}{$node};
    $Log->comment("Configuring quotas for vfiler $vfiler node $node");
    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $path;
    if ($vfiler eq 'vfiler0') {
        $path = '/vol/vol0/etc/quotas';
    } else {
        my $volume
            = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
            ->{'volumes'}->[0];
        $path = "/vol/$volume/etc/quotas";
    }

    my $file_info;
    eval {

        $Response = $Zapi->file_get_file_info('path' => $path);
    };

    # if the file does not exist create it
    if ($@) {

        # Create the file
        $self->write_file(
            'ascii' => "'# Generated by " . $self->{'testname'} . "'\n",
            'node'  => $node,
            'path'  => $path
        );

    }

    snooze 1;

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $defined_volumes = $self->{'defined_topology'}->{$node}->{'volumes'};
    my $defined_qtrees  = $self->{'defined_topology'}->{$node}->{'qtrees'};
    foreach my $volume (keys(%{$defined_volumes})) {

        my $owning_vfiler = $defined_volumes->{$volume}->{'vfiler'};
        if ($vfiler ne $owning_vfiler || $volume eq "v_$vfiler") { next; }

        $Log->comment("Creating quota entries for $volume");

        #target:type:disk:files:thold:sdisk:sfile
        #CIFS_USER:user@volume:-:-:700M:700M:5K

        foreach my $quota (split(/;/, $Options_hash->{'QUOTAS'})) {
            if ($quota !~ /\@volume/) { next; }
            $quota =~ s/CIFS_USER/$Options_hash->{'CIFS_USER'}/g;
            my @fields = split(/:/, $quota);
            $fields[1] =~ s/\@volume//g;
            my $Entry;
            eval {
                $Response = $Zapi->quota_get_entry(
                    'quota-target' => $fields[0],
                    'quota-type'   => $fields[1],
                    'volume'       => $volume,
                    'qtree'        => ''
                );
                $Entry = $Response->get_parsed_output();

            };

            if (!$Entry) {
                $Log->comment("Adding quota to $volume\n");
                try {

                    $Response = $Zapi->quota_add_entry(
                        'quota-target'    => $fields[0],
                        'quota-type'      => $fields[1],
                        'disk-limit'      => $fields[2],
                        'file-limit'      => $fields[3],
                        'threshold '      => $fields[4],
                        'soft-disk-limit' => $fields[5],
                        'soft-file-limit' => $fields[6],
                        'volume'          => $volume,
                        'qtree'           => ''
                    );
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
            }

            push(
                @{  $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'quotas'}
                    },
                $quota
            );

        }

        foreach my $qtree (keys(%{$defined_qtrees})) {

            my $owning_volume = $defined_qtrees->{$qtree}->{'volume'};
            if ($volume ne $owning_volume) { next; }

            #target:type:disk:files:thold:sdisk:sfile
            #CIFS_USER:user@qtree:-:-:700M:700M:5K

            foreach my $quota (split(/;/, $Options_hash->{'QUOTAS'})) {
                if ($quota !~ /\@qtree/) { next; }
                $quota =~ s/CIFS_USER/$Options_hash->{'CIFS_USER'}/g;
                my @fields = split(/:/, $quota);
                $fields[1] =~ s/\@qtree//g;
                my $Entry;
                eval {
                    $Response = $Zapi->quota_get_entry(
                        'quota-target' => $fields[0],
                        'quota-type'   => $fields[1],
                        'volume'       => $volume,
                        'qtree'        => $qtree
                    );
                    $Entry = $Response->get_parsed_output();

                };

                if (!$Entry) {
                    $Log->comment("Adding quota to qtree $qtree on $volume");
                    try {

                        $Response = $Zapi->quota_add_entry(
                            'quota-target'    => $fields[0],
                            'quota-type'      => $fields[1],
                            'disk-limit'      => $fields[2],
                            'file-limit'      => $fields[3],
                            'threshold '      => $fields[4],
                            'soft-disk-limit' => $fields[5],
                            'soft-file-limit' => $fields[6],
                            'volume'          => $volume,
                            'qtree'           => $qtree
                        );
                    }
                    catch NACL::APISet::Exceptions::ResponseException with {
                        my $Exception = shift;
                        die config_result($Exception->text());
                    };
                }

                push(
                    @{  $self->{'defined_topology'}->{$node}->{'volumes'}
                            ->{$volume}->{'quotas'}
                        },
                    $quota
                );

            }
        }

        try {

            $Response = $Zapi->quota_on('volume' => $volume,);
        }
        catch NACL::APISet::Exceptions::ResponseException with {

            #my $Exception = shift;
            #die config_result($Exception->text());
        };
        snooze 1;
    }
} ## end sub configure_quotas

sub assign_data_volumes {
    $Log->enter();

    my ($self, %options) = @_;
    my $size     = $options{'size'};
    my $basename = $options{'basename'};
    my $count    = $options{'count'};
    my $type     = $options{'type'};

    my @nodes                    = @{$Options_hash->{'NODES'}};
    my $count_per_node           = int($count / @nodes);
    my $count_per_node_remainder = $count % @nodes;

    # Spread the volumes acrosss the specified aggregates
    $Log->comment("Adding $count data volumes to aggregates");

    my $volume_index = 1;
    my @volumes;
    foreach my $node (@nodes) {

        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $aggregate_index  = 0;
        my $volume_usage     = 0;
        my $volumes_per_node = $count_per_node;
        if ($node eq $nodes[0]) {
            $volumes_per_node += $count_per_node_remainder;
        }

        # Get an array of the discovered aggregates
        my @discovered_aggregates = natsort(
            keys(%{$self->{'discovered_topology'}->{$node}->{'aggregates'}}));

        # loop through the discovered aggregates and create an array of our
        # defined aggregates
        my @defined_aggregates;
        foreach my $aggregate (@discovered_aggregates) {
            if (grep { $_ eq $aggregate }
                split(/,/, $Options_hash->{'AGGREGATES'})
                ) {
                push(@defined_aggregates, $aggregate);
            }
        }

        # Set the current aggregate and it's size
        my $aggregate = $defined_aggregates[$aggregate_index];
        $self->{'discovered_topology'}->{$node}->{'aggregates'}->{$aggregate}
            ->{'total'} =~ /(\d+)M/;
        my $aggregate_size
            = int($1 * ($Options_hash->{'MAX_AGGREGATE_USAGE'} / 100));

        # Loop though each volume assigning it to the current aggregate
        foreach my $i (1 ... $volumes_per_node) {

            # Check to make sure we don't exceed the current aggregates space.
            if ($volume_usage + $size > $aggregate_size) {

                # Switch to the next aggregate and grab it's size
                if ($defined_aggregates[$aggregate_index + 1]) {
                    $aggregate_index++;
                    $aggregate = $defined_aggregates[$aggregate_index];
                    $self->{'discovered_topology'}->{$node}->{'aggregates'}
                        ->{$aggregate}->{'total'} =~ /(\d+)M/;
                    $aggregate_size = int(
                        $1 * ($Options_hash->{'MAX_AGGREGATE_USAGE'} / 100));
                } else {
                    die config_result(
                        "Not enough aggregate space on $node"
                            . Dumper(
                            $self->{'discovered_topology'}->{$node}
                                ->{'aggregates'}->{$aggregate}
                            )
                    );
                }
            }

            # Create the volume name
            my $volume = $basename . '_' . $volume_index;
            $Log->comment("Adding data volume $volume "
                    . "to node $node aggregate $aggregate");

            # Populate the volume entries in the defined_topology hash
            $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                ->{'size'} = $size;

            $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                ->{'aggregate'} = $aggregate;

            $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                ->{'root_volume'} = 0;

            $volume_index++;
            $volume_usage += $size;
        }
    }

    # Spread the volumes across vfilers
    foreach my $node (@{$Options_hash->{'NODES'}}) {

        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};

        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);

        # Get the data volumes we just assigned to this nodes aggregates
        my @volumes = grep(/^$basename/, keys(%{$defined_volumes}));
        @volumes = natsort(@volumes);

        # If we are doing vfilers assign the volumes and interfaces to them
        if ($self->{'vfilers_per_node'}) {
            foreach my $vfiler (@vfilers) {
                if ($vfiler eq 'vfiler0') { next; }
                $Log->debug("Assigning volumes to vfiler $vfiler");

                # If volumes_per_vfiler_assignment eq rnd get a randomized
                # value of volumes_per_vfiler
                my $volumes_per_vfiler = $self->{'volumes_per_vfiler'} - 1;
                if ($self->{'volumes_per_vfiler_assignment'} eq 'rnd') {
                    $volumes_per_vfiler = int(rand($volumes_per_vfiler));
                }
                $Log->debug("Volumes per vfiler: $volumes_per_vfiler");

                # Assign the data volumes to the defined vfiler
                foreach (1 ... $volumes_per_vfiler) {
                    my $volume = shift(@volumes);
                    if (!$volume) {
                        die config_result("Not enouugh volumes specified to "
                                . "create $volumes_per_vfiler volumes "
                                . "on vfiler $vfiler. Please increase "
                                . "VOLUME_COUNT");
                    }
                    $Log->comment(
                        "Adding $volume to node $node vfiler $vfiler");
                    push(@{$defined_vfilers->{$vfiler}->{'volumes'}},
                        $volume);

                    # Update the defined volumes with the owning vfiler
                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'vfiler'} = $vfiler;

                    $self->{'defined_topology'}->{$node}->{'volumes'}
                        ->{$volume}->{'share_count'} = 0;

                }
            }
        }

        # If we have left over volumes assign them to vfiler0
        if (@volumes) {
            my $defined_vfiler
                = $self->{'defined_topology'}->{$node}->{'vfilers'}
                ->{'vfiler0'};
            foreach my $volume (@volumes) {
                $Log->comment("Adding $volume to node $node vfiler vfiler0");
                push(@{$defined_vfiler->{'volumes'}}, $volume);

                # Update the defined volumes with the owning vfiler
                $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                    ->{'vfiler'} = 'vfiler0';
                $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
                    ->{'share_count'} = 0;
            }

        }
    }
    $Log->exit();
} ## end sub assign_data_volumes

sub assign_vfiler_root_volumes {
    $Log->enter();

    my ($self, %options) = @_;
    my $size    = $options{'size'};
    my $vfilers = $options{'vfilers'};
    my $node    = $options{'node'};

    # Spread the volumes acrosss the specified aggregates
    $Log->comment("Adding root volumes to aggregates");

    my $aggregate_index = 0;
    my $volume_usage    = 0;

    # Get an array of the discovered aggregates
    my @discovered_aggregates = natsort(
        keys(%{$self->{'discovered_topology'}->{$node}->{'aggregates'}}));

    # loop through the discovered aggregates and create an array of our
    # defined aggregates
    my @defined_aggregates;
    foreach my $aggregate (@discovered_aggregates) {
        if (grep { $_ eq $aggregate }
            split(/,/, $Options_hash->{'AGGREGATES'})
            ) {
            push(@defined_aggregates, $aggregate);
        }
    }

    # Set the current aggregate and it's size
    my $aggregate = $defined_aggregates[$aggregate_index];
    $self->{'discovered_topology'}->{$node}->{'aggregates'}->{$aggregate}
        ->{'total'} =~ /(\d+)M/;
    my $aggregate_size
        = int($1 * ($Options_hash->{'MAX_AGGREGATE_USAGE'} / 100));

    # Loop though each vfiler assigning it's root volume to the current
    # aggregate
    foreach my $vfiler (@{$vfilers}) {

        # Check to make sure we don't exceed the current aggregates space.
        if ($volume_usage + $size > $aggregate_size) {

            # Switch to the next aggregate and grab it's size
            if ($defined_aggregates[$aggregate_index + 1]) {
                $aggregate_index++;
                $aggregate = $defined_aggregates[$aggregate_index];
                $self->{'discovered_topology'}->{$node}->{'aggregates'}
                    ->{$aggregate}->{'total'} =~ /(\d+)M/;
                $aggregate_size = int(
                    $1 * ($Options_hash->{'MAX_AGGREGATE_USAGE'} / 100));
            } else {
                die config_result(
                    "Not enough aggregate space on $node"
                        . Dumper(
                        $self->{'discovered_topology'}->{$node}
                            ->{'aggregates'}->{$aggregate}
                        )
                );
            }
        }

        # Create the volume name
        my $volume = 'v_' . $vfiler;
        $volume =~ s/-/_/g;
        $Log->comment("Adding $volume to node $node aggregate $aggregate");

        # Populate the volume entries in the defined_topology hash
        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}->{'size'}
            = $size;
        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
            ->{'aggregate'} = $aggregate;

        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
            ->{'vfiler'} = $vfiler;

        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
            ->{'share_count'} = 0;

        $self->{'defined_topology'}->{$node}->{'volumes'}->{$volume}
            ->{'root_volume'} = 1;

        # Assign root volume for vfiler
        push(
            @{  $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                    ->{'volumes'}
                },
            $volume
        );
        $volume_usage += $size;
    }
    $Log->exit();
} ## end sub assign_vfiler_root_volumes

sub enable_branch_cache {
    $Log->enter();

    my ($self, %options) = @_;

    my $vfiler = $options{'vfiler'};
    my $node   = $options{'node'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    # Enable branch cache
    $Log->comment("Enabling branch cache on vfiler $vfiler on node $node");

    my %data;
    $data{'cifs.smb2_1.branch_cache.enable'} = 'on';

    foreach my $key (keys(%data)) {
        try {

            $Response = $Zapi->options_set(
                'name'  => $key,
                'value' => $data{$key}
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    $Log->comment("Setting the cifs server secret");

    # Set the secret

    try {
        $Response = $Zapi->cifs_branchcache_set_key(
            "server-secret" => "timtebowrocks211");
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    $Log->exit();

} ## end sub enable_branch_cache

sub enable_smb2 {
    $Log->enter();

    my ($self, %options) = @_;

    my $vfiler = $options{'vfiler'};
    my $node   = $options{'node'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    my %data;
    $data{'cifs.smb2.enable'} = 'on';

    foreach my $key (keys(%data)) {
        try {

            $Response = $Zapi->options_set(
                'name'  => $key,
                'value' => $data{$key}
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    $Log->exit();

} ## end sub enable_smb2

sub smb_signing {
    $Log->enter();

    my ($self, %options) = @_;

    my $vfiler = $options{'vfiler'};
    my $node   = $options{'node'};

    my $Cli  = $apiset{'cli'}{$node};
    my $Zapi = $apiset{'zapi'}{$node};

    # Set the vfiler context for the next ZAPI calls
    try {
        $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $Exception = shift;
        die config_result($Exception->text());
    };

    my %data;
    $data{'cifs.signing.enable'} = 'on';

    #Check if smb2 is enabled and require smb signing for it.

    if ($Options_hash->{'SMB2'}) {
        $data{'cifs.smb2.signing.required'} = 'on';
    }
    foreach my $key (keys(%data)) {
        try {

            $Response = $Zapi->options_set(
                'name'  => $key,
                'value' => $data{$key}
            );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    $Log->exit();
} ## end sub smb_signing

sub create_spn {
    $Log->enter();

    my ($self, %options) = @_;

    $Options_hash = $Options->get_hash();

    my $vfiler = $options{'vfiler'};
    my $node   = $options{'node'};

    #Get the actual cifs name of the vfiler in question.

    my $cifs_server
        = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
        ->{'cifs_server'};
    $Log->comment("Cifs server is $cifs_server");

    my @vfiler_interfaces = keys(
        %{  $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                ->{'interfaces'}
            }
    );
    $Log->comment("Interfaces @vfiler_interfaces");

    foreach my $interface (@vfiler_interfaces) {
        my $defined_address
            = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
            ->{'interfaces'}->{$interface};
        foreach my $ip_info (@{$defined_address}) {
            my $vfiler_ip = $ip_info->{'address'};
            $Log->comment("vfiler IP is $vfiler_ip");
            my $resolved_address = get_name_by_ip('ip' => $vfiler_ip);

            if ($resolved_address->{'ERROR'}) {
                die config_result(
                    "IP address $resolved_address does not resolve to a name"
                );
            }

            $resolved_address->{'PTR'} =~ /([^\.]+)\.(.+)/;

            my $spn_name = $1;
            $spn_name .= ".$Options_hash->{'CIFS_DOMAIN'}";

            $self->set_spn(
                'cifs_server' => $cifs_server,
                'spn_name'    => $spn_name,
            );
        }
    }

    $Log->exit();
} ## end sub create_spn

sub create_homedir() {
    $Log->enter();

    my ($self,        %options) = @_;
    my (@qtree_paths, $qtrees)  = undef;

    my $vfiler = $options{'vfiler'};
    my $node   = $options{'node'};

    my $user_qtree      = 'users';
    my $defined_vfilers = $self->{'defined_topology'}->{$node}->{'vfilers'};
    my $defined_volumes = $self->{'defined_topology'}->{$node}->{'volumes'};

    my $Zapi = $apiset{'zapi'}{$node};
    my $Cli  = $apiset{'cli'}{$node};

    $Log->comment("Configuring homedirs for vfiler $vfiler");

    # Create the qtrees first before we do anything else.
    my @volumes
        = @{$self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
            ->{'volumes'}};
    foreach my $volume (@volumes) {
        my $qtree_path;
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        # Create the qtree for the users.
        try {
            $Response = $Zapi->qtree_create(
                'qtree'  => $user_qtree,
                'volume' => $volume
            );
            $qtree_path = "/vol/$volume/users";
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
        push(@qtree_paths, $qtree_path);

        snooze 1;

    }

#Loop through each of the qtree paths and then add the path to the actual file.

    foreach my $qtree (@qtree_paths) {

        $qtrees .= "$qtree \n";
    }

    my $path;
    if ($vfiler eq 'vfiler0') {
        $path = '/vol/vol0/etc/cifs_homedir.cfg';
    } else {
        my $volume
            = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
            ->{'volumes'}->[0];
        $path = "/vol/$volume/etc/cifs_homedir.cfg";
    }
    my $offset = -1;

    # Create the file
    $self->write_file(
        'ascii'  => "$qtrees \n",
        'node'   => $node,
        'path'   => $path,
        'offset' => $offset,
    );

#Get the volume and then in each volume create the number of homedirs that equals the user_per_volume count
#keep track of this with a counter variable and if that counter is met then  go to the next volume.

    foreach my $share (
        @{  $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                ->{'shares'}
        },
        ) {

        my $counter;
        

        my $perms = "0755";

        foreach my $user (@users) {

            #$Log->comment("Printing user $user");

            #Check for a blank user.
            if ($user eq "") {
                next;
            }
            if ($user eq $share) {

                my $volume = $topology_database->{$node}->{'shares'}->{$share}
                    ->{'volume'};
                my $userPath      = "/vol/$volume/users/$user";
                my $qtree_path    = "/vol/$volume/users";
                my $owning_vfiler = $defined_volumes->{$volume}->{'vfiler'};
                if ($vfiler ne $owning_vfiler || $volume eq "v_$vfiler") {
                    next;
                }

                $Log->comment(
                    "Creating homedir path on vfiler $vfiler on volume $volume for user $user on share $share."
                );

                try {
                    $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };

                try {
                    $Response = $Zapi->file_create_directory(
                        'path' => $userPath,
                        'perm' => $perms
                    );
                    $counter++;
                }
                catch NACL::APISet::Exceptions::ResponseException with {
                    my $Exception = shift;
                    die config_result($Exception->text());
                };
                snooze 1;
            } else {
                next;
            }
        }

    }

# Load the homedirs For some reason if we run this in any other vfiler than vfiler0
# the command is not honored.
    if ($vfiler eq 'vfiler0') {
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => $vfiler);
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        try {
            $Response
                = $Cli->execute_raw_command(command => "cifs homedir load");
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    } else {
        try {
            $Response = $Zapi->set_vfiler('vfiler-name' => 'vfiler0');
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };

        try {
            $Response = $Cli->execute_raw_command(
                command => "vfiler run $vfiler cifs homedir load");
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $Exception = shift;
            die config_result($Exception->text());
        };
    }

    #die config_result("Fail");

    $Log->exit();

} ## end sub create_homedir()

sub define_homedirs() {
    $Log->enter();
    my ($self, %options) = @_;
    my $script      = $options{'script'};
    my $share_index = $options{'share_index'};
    my $recipe      = $options{'recipe'};

    my $server;
    my @nodes
        = grep(/$Options_hash->{'NODE_REGEX'}/, @{$Options_hash->{'NODES'}});
    if (!@nodes) {
        die config_result("Unable to find any nodes using the "
                . "NODE_REGEX $Options_hash->{'NODE_REGEX'} against "
                . Dumper($Options_hash->{'NODES'}));
    }

    my %shares;
    my $vfiler_count = 0;
    foreach my $node (@nodes) {
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_shares = $topology_database->{$node}->{'shares'};
        my $discovered_vfilers
            = $self->{'discovered_topology'}->{$node}->{'vfilers'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};
        my $discovered_shares
            = $self->{'discovered_topology'}->{$node}->{'shares'};

        # Count the existing ontap shares
        my @vfilers = keys(%{$discovered_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            $shares{$node}{$vfiler} = @ontap_shares;
        }

        # Count the existing discovered shares
        my @ds = keys(%{$discovered_shares});
        @ds = natsort(@ds);
        foreach my $s (@ds) {
            if (!$s) { next; }
            my $owner = $discovered_shares->{$s}->{'owning-vfiler'};
            if (!$owner) {
                die config_result(
                    "Unable to get owning vfiler from discovered share $s\n"
                        . Dumper($discovered_shares->{$s}));
            }

            # Ignore shares we created from previous runs. The existing
            # shares will be redfined but filer_setup will be set to 0
            # so the share is not recreated. This way if the test fail
            # for some reason it can pick up whre it left off.
            if ($s =~ /^s_$script/) { next; }

            # Add the count to the owning vfiler
            $shares{$node}{$owner} += 1;
        }

        # Count the to be created shares
        foreach my $vfiler (keys(%{$defined_vfilers})) {

            $vfiler_count++;

            # Count the to be created ontap shares
            $shares{$node}{$vfiler} = @ontap_shares;

            # Count all shares that belong to the current vfiler
            # and are not already created and not a data share
            # This is basically to pickup the volume shares
            while (my $s = each %{$defined_shares}) {
                my $owner = $defined_shares->{$s}->{'vfiler'};
                if (!$owner) {
                    die config_result(
                        "Unable to get owning vfiler from defined share\n"
                            . Dumper($defined_shares->{$s}));
                } elsif ($vfiler ne $owner) {
                    next;
                } elsif (
                    grep {
                        $_ eq $s
                    }
                    keys(%{$discovered_shares->{$node}->{'shares'}})
                    ) {
                    next;
                } elsif ($s =~ /^s_$script/) {
                    next;
                }
                $shares{$node}{$vfiler} += 1;
            }
        }
    }

    # Total all the shares found
    my $total_existing_shares = 0;
    foreach my $n (keys(%shares)) {
        foreach my $v (keys(%{$shares{$n}})) {
            $total_existing_shares += $shares{$n}{$v};
        }
    }

    my $new_shares = $Options_hash->{'SHARE_COUNT'} - $total_existing_shares;
    my $shares_per_vfiler           = int($new_shares / $vfiler_count);
    my $shares_per_vfiler_remainder = $new_shares % $vfiler_count;

    $Log->comment("We will have $shares_per_vfiler share for each vfiler");

    # Spread the shares across the nodes, vfilers and volumes
    $Log->comment("Adding shares to vfilers");

    foreach my $node (@nodes) {
        my $defined_vfilers
            = $self->{'defined_topology'}->{$node}->{'vfilers'};
        my $defined_volumes
            = $self->{'defined_topology'}->{$node}->{'volumes'};
        my $discovered_shares
            = $self->{'discovered_topology'}->{$node}->{'shares'};
        my $discovered_volumes
            = $self->{'discovered_topology'}->{$node}->{'volumes'};

        my $Zapi = $apiset{'zapi'}{$node};
        my $Cli  = $apiset{'cli'}{$node};

        # Initialize the index for for the addresses for each vfiler's data
        # interface
        my $interface_index = 0;
        my %address_indexes;
        my @vfilers = keys(%{$defined_vfilers});
        @vfilers = natsort(@vfilers);
        foreach my $vfiler (@vfilers) {
            foreach my $interface (@{$Options_hash->{'DATA_INTERFACES'}}) {
                $address_indexes{$interface} = 0;
            }
        }

        foreach my $vfiler (@vfilers) {
            if (!$vfiler) {
                die script_result(
                    "Blank vfiler detected in defined vfilers hash");
            }

            my $current_shares_per_vfiler = $shares_per_vfiler;
            if ($vfiler eq 'vfiler0' && $node eq $nodes[0]) {
                $current_shares_per_vfiler += $shares_per_vfiler_remainder;
            }

            my $vfiler_volumes
                = $self->{'defined_topology'}->{$node}->{'vfilers'}->{$vfiler}
                ->{'volumes'};
            my @volumes;
            my $volume;
            my $share_count = 0;

#$Log->comment(
#    "Adding $current_shares_per_vfiler shares to vfiler $vfiler on node $node"
#);

            for (1 ... $current_shares_per_vfiler) {

                my $filer_setup = 1;

                my $vf = $vfiler;

                foreach my $volume (@{$vfiler_volumes}) {

                    my $vol   = $volume;
                    my $share = shift(@userDefinedShares);

                    my $owning_vfiler
                        = $defined_volumes->{$volume}->{'vfiler'};
                    if ($vfiler ne $owning_vfiler || $volume eq "v_$vfiler") {
                        next;
                    }

                    #Check for a blank user.
                    if ($share eq "") {
                        next;
                    }

                    my $path = "/vol/$volume/$share";

                    #Check to see if this volume is in the vserver

                    # Store the the volume name with the share
                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'volume'} = $vol;

                    # Store the share path
                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'path'} = $path;

                    # Store the vfiler name with the share
                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'vfiler'} = $vf;

                    # Decide when we create the share
                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'filer_setup'} = $filer_setup;

                    # Don't use qtrees
                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'qtree'} = 0;

                    push(
                        @{  $self->{'defined_topology'}->{$node}->{'vfilers'}
                                ->{$vf}->{'shares'}
                            },
                        $share
                    );

                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'node'} = $node;
                    ($server, $interface_index) = $self->assign_cifs_server(
                        'node'            => $node,
                        'vfiler'          => $vf,
                        'address_indexes' => \%address_indexes,
                        'interface_index' => $interface_index
                    );

                    #$Log->comment("Assigning CIFS server $server for "
                    #       . "share $share on vfiler $vfiler node "
                    #      . "$node");

                    $topology_database->{$node}->{'shares'}->{$share}
                        ->{'server'} = $server;

                }
            }

        }

    }

    $Log->exit();
    return $share_index;
} ## end sub define_homedirs()
1;
