# $Id: //depot/prod/test/Rsierranevada/lib/VSeries/Clustered/filer_convenience.pm#3 $
# -*- perl -*-
#
## @summary Additional VSeries functionality for Server.pm
##
## @description
##  These test functions are needed for the VSeries test suites for testing
##  BoilerMaker 10-Mode and beyond.  These methods are all intended to either
##  be additional convenience routines for Server.pm or are temporary ports of
##  pre-existing 7G to 7-Mode functionality.  This may be because the needed
##  information or control isn't implemented in BoilerMaker 10-Mode cli yet
##  (as of March 2008).  When equivalent functionality is available in the
##  10-Mode cli these methods will either be depricated or modified to use
##  such functionality.
##
## @usage
##   use VSeries::Clustered::filer_convenience;
##   use VSeries::Clustered::filer_convenience qw(...);
##
## @keywords  util test library
##
## @dependencies
## This pm is not object oriented because it is intended to augment Server.pm
## for communication with filers.  Most methods will require a Server Object be
## passed in.
##
## @author cattel@netapp.com
##
## @status testme
##
##

package VSeries::Clustered::filer_convenience;

use strict;
use Tharn;
use Getopt::Long;           
use Data::Dumper;
use Time::localtime;
use Text::ParseWords;	# So many textual utils in here.
use VSeries::utils;
use NACL::C::Node;
use NACL::STask::Cluster;

use vars ('@ISA','@EXPORT');
require Exporter;
@ISA = ('Exporter');
@EXPORT = ('get_disks_by_traits', 'remove_luns', 'two_path_check', 'get_SFO_partners',
           'vseries_hba_port_disable_enable', 'check_cf_state', 'poll_hba_interfaces',
           'get_vseries_storage_port_count', 'get_vseries_storage_port_associate',
           'get_vseries_storage_port_find', 'compare_disk_configs', 'check_sfo_status',
           'get_all_unowned_storage', 'create_10mode_vol', 'get_coredump_count',
           'showfh', 'dumpblock', 'dd', 'run_wafliron', 'storage_array_modify',
           'storage_array_remove_port', 'storage_array_show', 'storage_array_show_ports',
           'storage_array_show_luns', 'storage_array_show_config', 'storage_array_purge_database',
           'storage_array_remove', 'fcadmin_link_state', 'fcadmin_offline_online', 'storage_load_balance',
           'storage_adapter_enable_disable', 'ems_event_search', 'system_health_check', 'cleanup_components',
           'scsi_fitp_block','scsi_fitp_swap','scsi_fitp_clear','scsi_fitp_change', 'modify_volume_junction_path',
           'compare_storage_disk_show_path_equivalency','version_check', 'get_active_inactive_config',
           'assign_luns_and_create_test_aggr', 'setup_io_components','delete_io_components',
           'create_desired_topology', 'check_panic', 'get_ontap_version', 'check_subtest_result', 'get_checksum', 
           'connection_error_timeout_retry', 'get_array_name_by_ports');


# Global variables for testing.
my $GOOD_STATUS = 0;
my $ERR_STATUS = 1;
my $TRUE = "true";
my $FALSE = "false";

my $BR_STR = "BR.0";
my $RR_STR = "RR.0";
my $ERR_STR = "ERROR:";

my $NTAP_DISK = '^\w{2}\.\d+';               # Regex for OnTap disks.



## @name ::get_disks_by_traits($%;%)
##
## @description
## Takes in a server object to communicate with filers and runs
## storage_disk_show on the filer that is set as the execution server.  It then
## walks each disks hash and matches any key value pairs that were passed into
## the method in the include hash.  If any key/value pair passed in doesn't
## match then the disk isn't considered a match.  Then it looks to see if
## the disk matches any of they key/value pairs passed into the method in the
## exclude hash.  If any of those match then the disk is rejected.  If any
## disks, match all the include key/value pairs and none of the exclude
## key/value pairs then it's hash is pushed onto the array of hash tables that
## get returned as a reference.  An reference to an empty array will be
## returned if no disks match.
##
##
## @synopsis
## $matched_disks_array_ref = VSeries::Clustered::filer_convenience::get_disks_by_traits($server_obj, \%include, \%exclude);
##
## @arg $server Required: Server object to communicate to the filer through
##
## @arg %include_hash Required: Hash table containing key value pairs of
##                              attributes to match disks based on.
##
## @arg %exclude_hash Optional: Hash table containing key value pairs of
##                              attributes that disks shouldn't match on.
##
sub get_disks_by_traits($%;%){
    my ($server, $include_ref, $exclude_ref) = @_;
    my %include = %$include_ref;
    my %exclude = ();

    # This will be the array of successfully matched disks/LUNs returned to the
    # caller.
    my @selected_disks = ();

    # If we were passed in an exclude hash then use that during matching.
    if (defined($exclude_ref)) {
        %exclude = %$exclude_ref;
    }

    # Make the call to the server object and get the output to sift through.
    # The output will be an array of hash tables where the key value pairs are
    # the fields and values returned by the "storage disk show -instance" on
    # the CLI of a filer.
    my $disks = undef;
    my $output = "";
    eval{
        $disks = $server->storage_disk_show("-Timeout" => 1200, "output" => \$output);
    };
    if($@){
        Tharn::logresult('FATAL', $@);
    }
    if(!defined($disks)){
        Tharn::logresult('FATAL', "disk ref was undefined in return call from server.pm  raw output is:\n$output");
    }
    my @my_disks = @$disks;


    # We'll loop over each disks/LUNs hash table matching on all the attributes
    # in the include hash and if any of the attributes of the exclude hash match
    # then we discard the disk/LUN and move on.
    foreach my $disk_output (@my_disks) {

        my %disk_hash = %$disk_output;

        # Set the flag to true we'll loop through looking for any reason to
        # exclude a disk, either by not matching an attribute we want or by
        # matching an attribute we didn't want.
        my $selected = 1;
        foreach my $key (keys(%disk_hash)) {

            if (defined($include{$key})) {
                if ($include{$key} ne $disk_hash{$key}) {
                    $selected = 0;
                    last;
                }
            }
            if (defined($exclude{$key})) {
                if($exclude{$key} eq $disk_hash{$key}){
                    $selected = 0;
                    last;
                }
            }
        }

        # If selected is still true then we'll return the disk/LUN to the
        # caller.
        if ($selected == 1) {
            push (@selected_disks, \%disk_hash);
        }

    }

    return \@selected_disks;
}

## @name ::get_all_unowned_storage ($;$)
##
## @description
## Obtains the serial numbers of all unowned disks and luns on the filer
## specified by the Server Object and determines its associated storage type.
## To get the storage type we need to run 7 mode command storage show disk
## -T -x  because the 10 mode command only shows storage type of owned disks
## or luns.
## Return value will be a hash reference where the key is the 10 mode
## formatted disk name and the value is the storage type.
##
## @synopsis
## $unonwed_disks_hash_ref = VSeries::Clustered::filer_convenience::get_all_unowned_storage($server_obj);
##
## @arg $server  Required: Server object to communicate to the filer through
## @arg $node_scope_disk_names Optional: If it set to '1' then we will get the node_scope disk names
##                                       irrespective of ONTAP version. Otherwise it gets the output as
##                                       per the version.
## @arg $array Optional: If defined, then only storage from the specified array
##                       will be returned.
##
sub get_all_unowned_storage ($;$$) {

    my ($server, $node_scope_disk_names, $array) = @_;

    my %unowned_disk_storage_types = ();

    # The new arg '$node_scope_disk_names' has been introduced due to the following reason:
    # The disk_names might vary between CMode and Node_Scope command outputs.
    # Getting the list from CMode command might cause the disk_assign operation to fail because
    # of the disk_name that does not match with the node scope output.
    # Example :-
    # -------
    # CMode Output :-
    # ------------
    #
    # vgv3070f19a-vgv3070f19b::*> storage disk show -owner - -nodelist vgv3070f19a
    #                      Usable           Container
    # Disk                   Size Shelf Bay Type        Position   Aggregate Owner
    # ---------------- ---------- ----- --- ----------- ---------- --------- --------
    # vgv3070f19a:vgci9124s38:1-23.126L2
    #                           -     -   - unassigned  present    -         -
    # vgv3070f19a:vgci9124s38:1-23.126L3
    #                           -     -   - unassigned  present    -         -
    # vgv3070f19a:vgci9124s38:1-23.126L4
    #                           -     -   - unassigned  present    -         -
    #
    # vgv3070f19a-vgv3070f19b::*> storage disk show -owner - -nodelist vgv3070f19b
    #                      Usable           Container
    # Disk                   Size Shelf Bay Type        Position   Aggregate Owner
    # ---------------- ---------- ----- --- ----------- ---------- --------- --------
    # vgv3070f19a:vgci9124s38:1-23.126L2
    #                           -     -   - unassigned  present    -         -
    # vgv3070f19a:vgci9124s38:1-23.126L3
    #                           -     -   - unassigned  present    -         -
    # vgv3070f19a:vgci9124s38:1-23.126L4
    #                           -     -   - unassigned  present    -         -
    # 
    # NodeScope Output :-
    # ----------------
    #
    # vgv3070f19a> storage show disk -T -x
    # DISK                    SHELF BAY SERIAL       VENDOR   MODEL           REV TYPE
    # ----------------------- --------- ------------ -------- --------------- ---- -----
    # vgci9124s38:1-23.126L2    -    -  600508B400010D2700029000111E0000 COMPAQ   HSV111(C)COMPAQ 4100 LUN
    # vgci9124s38:1-23.126L3    -    -  600508B400010D27000290000FC70000 COMPAQ   HSV111(C)COMPAQ 4100 LUN
    # vgci9124s38:1-23.126L4    -    -  600508B400010D27000290000FCE0000 COMPAQ   HSV111(C)COMPAQ 4100 LUN
    # 
    # vgv3070f19b> storage show disk -T -x
    # DISK                    SHELF BAY SERIAL       VENDOR   MODEL           REV TYPE
    # ----------------------- --------- ------------ -------- --------------- ---- -----
    # vgci9124s38:1-24.126L2    -    -  600508B400010D2700029000111E0000 COMPAQ   HSV111(C)COMPAQ 4100 LUN
    # vgci9124s38:1-24.126L3    -    -  600508B400010D27000290000FC70000 COMPAQ   HSV111(C)COMPAQ 4100 LUN
    # vgci9124s38:1-24.126L4    -    -  600508B400010D27000290000FCE0000 COMPAQ   HSV111(C)COMPAQ 4100 LUN
    #

    my $version = &version_check($server);

    # Get the hostname of the server, we'll need it later
    my $current_server = $server->get_execution_server();

    # Get a list of all unowned disks and luns attached to the filer.
    my $unowned_disks_arr_ref = 0;
    my %include = ();

    if (defined($array)) {
        %include = ('owner-str' => "\"Not Owned\"", 'owner' => "-", 'array-name' => $array);
    } else {
        %include = ('owner-str' => "\"Not Owned\"", 'owner' => "-");
    }

    eval{
        $unowned_disks_arr_ref = get_disks_by_traits($server, \%include);
    };
    if($@){
        Tharn::logresult('FATAL', $@);
    }

    # Define our array of unowned disk hash references
    my @unowned_disks = @$unowned_disks_arr_ref;

    # Added a condition in such a way that the existing logic doesn't get disturbed.
    # 'If' block gets executed when the version matches with RR and the arg 'node_scope_disk_names'
    # is undefined.
    # If the arg 'node_scope_disk_names' is defined then irrespective of RR and BR version, we will 
    # get the disk names from the nodescope command.
    if ($version eq $RR_STR && $node_scope_disk_names eq q{}) {
        foreach my $unowned_disk_hash_ref (@unowned_disks) {
            my %unowned_disk = %$unowned_disk_hash_ref;
            my $disk_name = $unowned_disk{disk};
            $unowned_disk_storage_types{$disk_name} = $unowned_disk{type};
        }

        return \%unowned_disk_storage_types;

    } elsif ($version eq $BR_STR || $node_scope_disk_names == 1) {
        my $output = "";
        eval{
            $output = $server->dblade('cmd' => "storage show disk -T -x");
        };
        if($@){
            Tharn::logresult('FATAL', $@);
        }    
        
        # Command parsing assumptions:
        # Disk and Storage type will never be null
        #
        #vra-v3050-2b::> system node run -node vra-v3050-2b -command storage show disk -T -x
        #DISK                   SHELF BAY SERIAL                           VENDOR   MODEL  REV TYPE
        #---------------------- --------- -------------------------------- -------- ------ ---- -----
        #vrs-br200e-3:15.126L0    -    -  600E000000A800000066700700000000 FUJITSU  E3000  0000 LUN 
        #vrs-br200e-3:15.126L1    -    -  600E000000A800000066700700010000 FUJITSU  E3000  0000 LUN 
        #vrs-br200e-4:15.126L2    -    -  600E000000A800000066700700020000 FUJITSU  E3000  0000 LUN 
        #vnci9506s42:1-2.122L1692   -    -  461432C97000 EMC      SYMMETRIX        0000 LUN 
        #0f.16                      1    0  9QG39F8N     NETAPP   X267_SGLXY500SSX AQJZ ATA
           
        my @storage_output = split(/\n/, $output);
                    
        # Chop off the header
        @storage_output = @storage_output[2.. $#storage_output];
        my @line_tokens = ();
    
        # Prepare to walk the output of storage show disk -T -x (7 mode) and        
        # all storage devices that are unowned to a hash. The key is the disk
        # name while the value is the storage type.
        foreach my $line (@storage_output) {
           $line =~ s/\r//;
           # Check for an expected line of command output and save the 7 mode
           # disk name and storage type to memory locations.
           if ($line =~  /^([^ ]+)\s+.*\s+(\w+) *$/ ) {
              if ($node_scope_disk_names == 1) {
                 my $disk_name = $1;
                 my $storage_type = $2;
                 foreach my $unowned_disk_hash_ref (@unowned_disks) {
                    my %unowned_disk = %$unowned_disk_hash_ref;

                    # If the line matches with the unowned disk serial_number then
                    # it will be pushed into the unowned_disk_storage_types hash.
                    if ($line =~ m/$unowned_disk{serialnumber}/) {
                       $unowned_disk_storage_types{$disk_name} = $storage_type;
                       last;
                    }
                 }
              } else {
                 # Create 10 mode disk name by appending current server name to the
                 # beginning of the 7 mode disk name
                 my $disk_name_10mode = $current_server . ":" . $1;
                 foreach my $unowned_disk_hash_ref (@unowned_disks) {
                    my %unowned_disk = %$unowned_disk_hash_ref;
 
                    if ($unowned_disk{disk} eq $disk_name_10mode) {
                       $unowned_disk_storage_types{$disk_name_10mode} = $2;
                       last;
                    }
                 }
              }
           }
        }
        return \%unowned_disk_storage_types;
    }
}        

## @name ::create_10mode_vol ($%)
##
## @description
## Create a 10 mode test volume (along with an associated aggr and vserver if
## they do not alreaady exist). Optionally create the volume using specific
## storage and checksum types. Returns an error status along with an error
## array ref if errors were detected.
##
##
## @synopsis 
## ($status, \@errors) = VSeries::Clustered::filer_convenience::create_10mode_vol($server, \%arg_hash);
##
## @arg $server Required:   Server object to communicate to the filer through
## @arg  %arg_hash  Required   hash of various 10 mode command options
##    %arg_hash keys:
##    aggregate - aggr name (if value is not defined, will return error)
##    vserver - vserver name (if value is not defined, will return error)
##    rootvolume  - rootvol name (if value is not defined, will return error)
##    volume - volume name (if value is not defined, will return error)
##    size - volume size (if value is not defined, default is 1G)
##    state - volume state when created (ie. online or offline)
##                  (if value is not defined, default is online)
##    num_devices - amount of storage devices that will make up a new aggr
##                  (if value is not defined, default is 3)
##    disklist   - list of disks to be used for aggregate creation
##                  (if value is not defined, considers any unowned disks
##                   based on num_devices)
##    storage_type - storage device type (ie. FCAL, LUN)
##                  (if value is not defined, default is LUN)
##    checksum_type - LUN checksum type (block or zoned)
##                  (if value is not defined, default is block))
##    max_aggr_retries - waiting for aggr to come online
## 
## NOTES: 
## This method is designed to create a new volume entitiy but does
## a quick initial check for an existing vol/aggr/vserver combo. It is
## not a requirement of this method to verify all parameters required
## for a new volume must also be exactly configured on any existing volume
## components. The calling program should be aware of such params on existing
## volume components before calling this method.
##
sub create_10mode_vol ($%) {

    my $server = shift;
    my $arg_hash_ref = shift;   

    my %arg_hash = %$arg_hash_ref;
    my @errors = ();
    my $output = "";

    # Check that we have the required keys defined in the %arg_hash
    if ( ! defined $arg_hash{aggregate} || ! defined $arg_hash{vserver} 
        || ! defined $arg_hash{rootvolume} || ! defined $arg_hash{volume} ) {
        push @errors, '%arg_hash is missing 1 or more required hash values';
        return ($ERR_STATUS, \@errors);
    }
    
    # Create defaults for any undefined non-required %arg_hash values   
    if ( ! defined $arg_hash{size} ) {
        $arg_hash{size} = "1G";
    }
    if ( ! defined $arg_hash{state} ) {
        $arg_hash{state} = "online";
    }
    if ( ! defined $arg_hash{storage_type} ) {
        $arg_hash{storage_type} = "LUN";
    }
    if ( ! defined $arg_hash{checksum_type} ) {
        $arg_hash{checksum_type} = "block";
    }

    if ( ! defined $arg_hash{disklist}) {
        if ( ! defined $arg_hash{num_devices} ) {
            $arg_hash{num_devices} = 3;
        }
    }

    my $disk_list = "";
    my $create_aggr = 0; # flag to determine whether to create aggr
    my $create_vserver = 0; # flag to determine whether to create vserver
    
    # Get the hostname of the server, we'll need it later
    my $server_name = $server->get_execution_server();
 
    # Check whether our user specified aggregate already exists
    eval {
        $server->storage_aggregate_show('aggregate' => $arg_hash{aggregate}, 'output' => \$output);
    };
    if ( $output =~ m/There are no entries matching your query/ || $output =~ m/Error: show failed: Aggregate .* does not exist/i ) {
        # The aggregate does not exist so we will create it
        $create_aggr = 1;
    } elsif ( $@ && $output !~ m/There are no entries matching your query/ ) {
	# Aggregate exists then we use it
	$create_aggr = 0;
    }

    # Check whether our user specified vserver already exists
    my @vserver_show_info = ();
    eval {
        @vserver_show_info = $server->vserver_show('vserver' => $arg_hash{vserver}, 'output' => \$output);
    };

    if ( $output =~ m/There are no entries matching your query/ ) {
        $create_vserver = 1;
    } elsif ( $@ && $output !~ m/There are no entries matching your query/ ) {
        push @errors, $@;
        return ($ERR_STATUS, \@errors);
    } elsif ( ! $@ && not exists $arg_hash{vserver_exists}) {
        # 'vserver_exists' arg is helpful while creating multiple volumes of different aggrs on
        # a single Vserver.
        # vserver show command passed, user specified vserver exists
        # but we don't yet know if it's suitable to use for the test vol.
        # Check the command output to see if the following conditions are 
        # met:
        # 1. existing vserver's rootvol matches user specified rootvol
        # 2. existing vserver's aggr matches user specified aggr_name
        # 3. existing vserver's type is Cluster      
        # If any of these conditions aren't  met then fail because the
        # caller specified illegal conditions or combinations.
        # There cannot be duplicate vserver names so the array will have
        # only one hash ref.
        my %vserver_hash = %{$vserver_show_info[0]};
        if ( $vserver_hash{type} ne "cluster" ) {
            push @errors, "ERROR: Existing vserver $arg_hash{vserver} to be used for test is not of type cluster as expected, it is type $vserver_hash{type}";
            return ($ERR_STATUS, \@errors);
        }
        if ( $vserver_hash{aggregate} ne $arg_hash{aggregate} ) {
            push @errors, "ERROR: Existing vserver $arg_hash{vserver} to be used for test is not part of expected test aggregate: $arg_hash{aggregate}, it uses: $vserver_hash{aggregate}";
            return ($ERR_STATUS, \@errors);
        }
        if ( $vserver_hash{rootvolume} ne  $arg_hash{rootvolume} ) {
            push @errors, "ERROR: Existing vserver $arg_hash{vserver} to be used for test does not contain expected rootvol: $arg_hash{rootvolume}, it contains: $vserver_hash{rootvolume}";
            return ($ERR_STATUS, \@errors);
        }
    }
        
    # Check if volume already exists. It must belong to vserver
    # arg_hash{vserver_name} which contains rootvol arg_hash{vserver_rootvol}
    # and belongs to aggr arg_hash{aggr_name}. We can do one stop shopping
    # with the vol show command.
    
    # Params for volume show commands output:
    my $vol_stat_cmd_ar = 0;
    my %vol_hash = ();

 
    eval {
        $vol_stat_cmd_ar = $server->volume_show('volume' => $arg_hash{volume}, 'output' => \$output);
    };
    
    # Check volume show command output. If the command passed, the volume
    # exists so check if it's associated with the user specified aggr and 
    # vserver names.  We may have a case of multiple volume names used by
    # different vservers. If this is true  then we can't extract lone volume
    # info by querying the volume with its vserver (BURT337186). This is why
    # we just queried only using the vol name. Now we will go through the array
    # to pull out the volume info associated with the right vserver
    # ($arg_hash{vserver_name}).
     
    if ( ! $@ ) { # vol exists

        my @vol_stat_aoh = @$vol_stat_cmd_ar;
        foreach my $vol_stat_hr (@vol_stat_aoh) {
            %vol_hash = %$vol_stat_hr;
            if ( $vol_hash{vserver} eq $arg_hash{vserver}
                && $vol_hash{aggregate} eq $arg_hash{aggregate} ) {
                # The vol already exists, we're done here
                return ($GOOD_STATUS);
            }
        }
    } elsif ($@ && $output !~ m/There are no entries matching your query/ ) {
        # Command failed but it was an error other than the volume not
        # existing so abort
        push @errors, $@;
        return ($ERR_STATUS, \@errors);     
    }
    
    if ( $create_aggr ) {
        # Create our test aggr
        # Ensure we have enough storage devices of a given storage type
        # to create the test aggr.
        if ( ! defined $arg_hash{'disklist'}) {
        
            # Get a hash of all unowned storage devices
            # and their corresponding storage types.
            my $unowned_disks_hash_ref = 0;

            eval{
                $unowned_disks_hash_ref = VSeries::Clustered::filer_convenience::get_all_unowned_storage($server);
            };
            if($@){
                push @errors, $@;
                return ($ERR_STATUS, \@errors);
            }
            my %unowned_disk_storage_types = %$unowned_disks_hash_ref;

   
            my @unowned_disk_list = ();
            my $disk_name = ""; # 10 mode formatted disk name
            my $num_unowned_disks_found = 0;

            # Extract the required amount of the selected devices (of a given
            # storage type) needed for aggr creation.
            foreach $disk_name (keys %unowned_disk_storage_types) {
                if ( $unowned_disk_storage_types{$disk_name} eq "$arg_hash{storage_type}" ) {
                    push @unowned_disk_list, $disk_name;
                }
            
                $num_unowned_disks_found = scalar(@unowned_disk_list);
            
                # Break out of loop once we have the required amount of unowned devices
                if ( $num_unowned_disks_found == $arg_hash{num_devices} ) { last; }
            }
  
            # Ensure we found the required minimum number of devices to assign
            if ( $num_unowned_disks_found < $arg_hash{num_devices} ) {
                push @errors, "ERROR: did not find enough devices of type $arg_hash{storage_type} to assign. Needed $arg_hash{num_devices}, found $num_unowned_disks_found";
                return ($ERR_STATUS, \@errors);
            } 
 
            # If we are creating a lun aggregate and we care  about setting the
            # checksum for our test aggr then we have to do it via 7 mode disk
            # assign command. Otherwise we assign the device in 10 mode via storage
            # disk modify
            if ( $arg_hash{storage_type} eq "LUN" ) {
                foreach my $lun_name (@unowned_disk_list) {
                   # Need to strip off leading node name from disk name because
                   # we're running 7 mode commmand
                    my $lun_name_7mode = $lun_name;
                    $lun_name_7mode =~ s/^$server_name://;

                    eval{
                        $output = $server->dblade('cmd' => "disk assign $lun_name_7mode -c $arg_hash{checksum_type}");
                    };
                    if($@){
                        push @errors, $@;
                        return ($ERR_STATUS, \@errors);
                    }

                    # Add lun to a formatted disk list for upcoming create aggr
                    $disk_list .= "$lun_name,";
                }
            } else {  # else we assign the device via 10 mode
    
                foreach my $disk (@unowned_disk_list) {
    
                    eval {
                        $server->storage_disk_modify('disk' => $disk, 'owner' => $server_name);
                    };
                    if($@){
                        push @errors, $@;
                        return ($ERR_STATUS, \@errors);
                    }
                   
                    # Add disk to a formatted disk list for upcoming create aggr
                    $disk_list .= "$disk,";
                }
            }
            chop $disk_list;
        }
        else {
            my @disk_list_array;
            my @disk_uid_array = ();
            $disk_list = $arg_hash{'disklist'};
            
            if($disk_list =~ /,/) {
                @disk_list_array = split(/,/, $disk_list);
            }
            else {
                $disk_list_array[0] = $disk_list;
            }

            foreach my $disk (@disk_list_array) {

                # get the disk uid for each disk in @disk_list_array before ownership change, then
                # use the disk uid retrieve the disk name after the ownership change of the disks
                my @disk_array = ();

                my $disk_name_c_mode = $disk;
                # Add the server name to the disk, if it's not existing.
                # Regex '^(\w+\.){2}\d+' is used to match NNDs/Boab disks. For example it works for the
                # following format of disks: 
                # If disk name is 0c.00.10 or 0c.00.10L1 then the server name will be added to it i.e.
                # vgv3070f19a:0c.00.10 or vgv3070f19a:0c.00.10L1.
                # The regex doesn't match for the disk names like vgv3070f19a:0c.00.10 or vgv3070f19a:0c.00.10L1, 
                # where server name is already existing.
                # Regex '((\w+\.){2}|(.*?:){2})' is used to match only LUNs.
                # If the disk name matches with vgbr300s32:7.126L1 then the server name will be added to it i.e.
                # it will become as vgv3070f19a:vgbr300s32:7.126L1. Otherwise there will not be any change
                # i.e. when disk name matches with vgv3070f19a:vgbr300s32:7.126L1.
                # If we use only (.*?:){2} then in some conditions it will fail. 
                # Let's take the following example:
                # If disk name matches with vgv3070f19a:0c.00.10 then the first regex i.e. '^(\w+\.){2}\d+' doesn't
                # apply and if we use only '(.*?:){2})' then it matches with it and again adds the server name to it i.e.
                # it will become as vgv3070f19a:vgv3070f19a:0c.00.10.
                # To avoid such scenarios we have used '((\w+\.){2}|(.*?:){2})' which will skip NNDs/Boab disks.
                $disk_name_c_mode = "$server_name:$disk" if($disk =~ m/^(\w+\.){2}\d+/ || $disk !~ m/((\w+\.){2}|(.*?:){2})/);
                eval {
                    @disk_array = $server->storage_disk_show('disk' => $disk_name_c_mode);
                };
                if($@){
                    push @errors, $@;
                    return ($ERR_STATUS, \@errors);
                }
                my $disk_hash = $disk_array[0];
                push @disk_uid_array, $disk_hash->{'uid'};

                eval{
                   # Need to strip off leading node name from disk name because
                   # we're running 7 mode commmand
                   $disk_name_c_mode =~ s/.*?://;

                   my $checksum = "-c $arg_hash{checksum_type}";
                   # As it is not allowed to change the checksum type of NNDs or 
                   # BOAB disks, we are setting the variable 'checksum' to undef.
                   $checksum = undef if ($disk_name_c_mode =~ m/(\w+\.){2}\d+/);

                   $server->dblade('cmd' => "disk assign $disk_name_c_mode $checksum");
                };
                if($@){
                   push @errors, $@;
                   return ($ERR_STATUS, \@errors);
                }
            }
            
            # now use disk uid to get disk name which might been changed after the ownership change, then
            # re-populate @disk_list_array and $disk_list with the new disk names 

            $disk_list = ""; 
            @disk_list_array = ();
            foreach my $disk_id (@disk_uid_array) {
                my @disk_array = ();
                eval {
                    @disk_array = $server->storage_disk_show('uid' => $disk_id);
                };
                if($@){
                    push @errors, $@;
                    return ($ERR_STATUS, \@errors);
                }

                push @disk_list_array, $disk_array[0]->{"disk"};
            }

            my $disk_array_length = scalar(@disk_list_array);
            if($disk_array_length == 1) {
                $disk_list = $disk_list_array[0];
            } else {
                for (my $i=0; $i<$disk_array_length; $i++) {
                    $disk_list .= $disk_list_array[$i] . ",";
                }
                chop $disk_list;
            }

        }
 
        # Create our test aggregate
        eval{
           my @aggr_params = ('aggregate' => "$arg_hash{aggregate}", 'disklist' => $disk_list, 'volume-type' => "flex");
           push(@aggr_params, 'block-type' => "$arg_hash{block_type}") if (exists $arg_hash{block_type});
           push(@aggr_params, 'raidtype' => "$arg_hash{raidtype}") if (exists $arg_hash{raidtype});
           $server->storage_aggregate_new(@aggr_params);
        };
        if ($@) {

            # If the command to create new aggr failed, pool on aggr status until
            # the aggregate is discovered, else we error on timeout

            # Loop as long as the verify command is failing and we haven't
            # timed out
            my $timeout = time() + 120;
            my $verified = 0;
            
            while(time() < $timeout){
                # Attempt to get new aggr instance data, if we fail then sleep 
                # and retry command.
                eval{
                    $server->storage_aggregate_show('aggregate' => $arg_hash{aggregate}, 'output' => \$output );
                };
                if($@){
                    sleep 5;
                }
                else{
                    # New aggregate discovered
                    $verified = 1;
                    last;
                }
            }
        
            if ($verified == 0) {
                push @errors, "Aggregate $arg_hash{aggregate} wasn't able to be verified after create and exhausting verify retries";
                return ($ERR_STATUS, \@errors);
            }
        }

        # A new aggregate may not immediately go into an online state. We need
        # to poll and retry to confirm it does before we move on to create the
        # vserver.
        my @aggr_show = ();
        my %aggr_hash = ();
        my $retries = 0;
                
        # Wait up to 60 minutes waiting for aggr to come online
        my $max_aggr_show_retries = 120;
        # NNDs takes more time than LUNs. So provided an option to specify the max_aggr_retries
        $max_aggr_show_retries = $arg_hash{max_aggr_retries} if (exists $arg_hash{max_aggr_retries});

        # Poll until the aggregate comes online
        while (1) {
            eval{
                @aggr_show = $server->storage_aggregate_show('aggregate' => $arg_hash{aggregate});
            };
            if($@){
                push @errors, "Couldn't obtain status of aggregate $arg_hash{aggregate}";
                return ($ERR_STATUS, \@errors);
            }
 
            %aggr_hash = %{$aggr_show[0]};
                        
            if ( $aggr_hash{'state'} eq "online" ) { 
                last;
            } 
            elsif ( $aggr_hash{'state'} ne "online" && $retries >= $max_aggr_show_retries) { 
                push @errors, "Timed out waiting for aggregate $arg_hash{aggregate} to come online";
                return ($ERR_STATUS, \@errors);
            }
            elsif ( $aggr_hash{'state'} ne "online" ) {
                $retries++;
                sleep 30;
            }
       }
    } # end if ( $create_aggr )
    
    if ( $create_vserver ) {
        # Create our vserver
        eval{
            $server->vserver_create("-Timeout" => 120, 'vserver' => "$arg_hash{vserver}", 'rootvolume' => "$arg_hash{rootvolume}", 'ns-switch'
 => 'file', 'aggregate' => "$arg_hash{aggregate}", 'rootvolume-security-style' => 'mixed');
        };
        if($@){
            push @errors, $@;
            return ($ERR_STATUS, \@errors);
        }   
    } # end if ( $create_vserver )

    # Create our test volume
    eval{
        $server->volume_create("-Timeout" => 120, 'volume' => "$arg_hash{volume}", 'vserver' => "$arg_hash{vserver}", 'aggregate' => "$arg_hash{aggregate}",
 'size' => "$arg_hash{size}", 'state' => "$arg_hash{state}");
    };
    if($@){
        push @errors, $@;
        return ($ERR_STATUS, \@errors);
    }

    # Verify volume and its related components were created successfully.
    eval{
        $vol_stat_cmd_ar= $server->volume_show('volume' => "$arg_hash{volume}", 'vserver' => "$arg_hash{vserver}", 'output' => \$output );
    };
    if($@){
        push @errors, $@;
        return ($ERR_STATUS, \@errors);
    }    
    
    # Parse vol show command output. Unlike the previous check for existing
    # volumes, this time we know that there is only one existing volume using
    # this name so we won't hit BURT337186, we can query using the vserver and
    # know the array ref has only one hash of volume info to parse. 
    my @vol_cmd_aoh = @$vol_stat_cmd_ar;
    %vol_hash = %{$vol_cmd_aoh[0]};

    # Verify the names of newly created aggr, vserver and volume match user
    # specified parameters
    if ( $vol_hash{volume} ne $arg_hash{volume} || $vol_hash{aggregate} ne $arg_hash{aggregate}
        || $vol_hash{vserver} ne $arg_hash{vserver} ) {
        push @errors, "volume created but attributes do not match expected values: $output";
        return ($ERR_STATUS, \@errors);
    }   

    return ($GOOD_STATUS);
}


## @name ::remove_luns($@)
##
## @description
## Takes in a server object to communicate with filers and an array reference
## containing hash tables of each disk to be removed.  The hash table is in the
## form of what is returned by most CLI commands but really just needs to
## contain the "disk" key with a value that is a known LUN on the given filer
## that the server object is configured to talk to.  This will then use system
## node run to execute a 'disk remove -w <disk name>' call in the 7-Mode
## context.  This is a temporary shim because as of March 2008 there isn't
## support to remove LUNs in the BoilerMaker 10-Mode context and this is needed
## for VSeries Automation.  This will return an array ref with any errors
## encountered during the LUN removal.  If the array is empty it is assumed the
## removal was successful.
##
## @synopsis
## $error_msg_array_ref = VSeries::Clustered::filer_convenience::remove_luns($server_obj, \@luns_to_remove);
##
## @arg $server Required: Server object to communicate to the filer through
##
## @arg @luns_to_remove Required: Array containing hash tables where 'disk' key
##                                has the value of the LUN name to be removed.
##
##
sub remove_luns($@){
    my ($server, $luns_to_remove_ref) = @_;
    my @luns_to_remove = @$luns_to_remove_ref;
    my @error_array = ();

    # Loop over each hash table for the LUNs passed in pulling out the name and
    # using system node run to call 'disk remove -w <$lun_name>'.  Catch any
    # errors and send them back to the user.
    foreach my $lun_ref (@luns_to_remove) {
        my %lun_hash = %$lun_ref;
        my $lun_name = $lun_hash{'disk'};
        my @name_tokens = ();

        # Here's a sample name that we'd be splitting here
        # vra-v3050-2a:vrs-br200e-3:15.126L24
        @name_tokens = split(/\:/, $lun_name);
        $lun_name = $name_tokens[1] . ":" . $name_tokens[2];

        # This is the command that should be used to remove LUNs in Boilermaker
        # 10-Mode.  Leaving it here as a breadcrumb for when this works.
        # $server->storage_disk_setstate('disk' => $lun_hash{'disk'}, 'action' => 'remove');

        # If the removal of the LUN fails trap the error and send it back to
        # the caller but continue to try to remove the other luns in the list.
        eval{
            my $output = $server->dblade('cmd' => "disk remove -w $lun_name");
        };
        if($@){
            push(@error_array, $@);
        }
    }

    return \@error_array;

}


## @name ::two_path_check($)
##
## @description
## Takes in a server object to communicate with the filer and does a "storage
## disk show -instance".  It then loops through the output weeding out the
## entries of the output that is from the other filers perspective if
## clustered.  This determination is made by checking if the value for the
## owner field shows up as part of the value for the disk field.  The disk
## field is the disk/LUN name and convention is to include the owner as the
## start of the name for easier distinction in BoilerMaker 10-Mode.  It then
## checks if any of the disks that are owned have a dash '-' for the disk or
## secondary name fields, any disks/LUNs that do have a dash are pushed onto
## an array and returned to the caller, if the returned array is empty then
## that means all owned disks have two paths reported.
##
## @synopsis
## $single_pathed_array_ref = VSeries::Clustered::filer_convenience::two_path_check($server_obj);
##
## @arg $server Required: Server object to communicate to the filer through
##
##
sub two_path_check($){
    my ($server_object) = @_;


    # Get the output of storage disk show from the filer that the server object
    # is set to talk to.
    my $disks = "";
    eval{
        $disks = $server_object->storage_disk_show("-Timeout" => 600);
    };
    if(@$){
        Tharn::logresult('FATAL', $@);
    }
    my @my_disks = @$disks;
    my @disks_with_single_path = ();

    # Now loop through the output.  Any entries that don't have an owner or
    # the owner doesn't show up as part of the disk field are skipped over.
    # Disks that do have an owner that shows up in the disk field are
    # added to the singled pathed disk if the secondaryname field is populated
    # with a dash '-'
    foreach my $disk_ref (@my_disks){
        my %disk_hash = %$disk_ref;
        if ($disk_hash{owner} ne '-' && $disk_hash{disk} !~ m/$disk_hash{owner}:/) {
            next;
        }
        if ($disk_hash{owner} ne '-' && ($disk_hash{disk} eq '-' || $disk_hash{secondaryname} eq '-')) {
            push(@disks_with_single_path, \%disk_hash);
        }

    }

    return \@disks_with_single_path;
}


## @name ::get_SFO_partners(@)
##
## @description
## Takes in an array ref of filer host names and proceeds to get the hostspec
## for each of the names passed.  From there it looks through the hostspec for
## the SFOPAIR entry and tries to match that with any other name passed in.
## The SFOPAIR entry should be unique to each filer pair that are SFO
## partners for redundancy.  Not that this is different than just saying that a
## bunch of filers participate in a cluster together since that can be much more
## than two filers in BoilerMaker 10-Mode and beyond.
##
## @synopsis
## $sfo_pair_hash_ref = VSeries::Clustered::filer_convenience::get_SFO_partners(\@host_names);
##
## @arg @host_names Required: Elements are names of filers to look up their
##                            their tharnhost files to search for the SFOPAIR
##                            value.
##
sub get_SFO_partners(@){
    my ($host_names_ref) = @_;
    my @host_names = @$host_names_ref;
    my %sfo_pairs = ();

    # Loop through the names passed in matching up SFOPAIR values in the
    # hostspecs for each filer.  If they match add them to the sfo_pairs
    # hash that will be returned.
    foreach my $name (@host_names) {

        # The if statement here just ensures that we haven't already
        # found this name's sfo pair
        if (defined($sfo_pairs{$name})) {next;}
        my $unique = 1;

        # This for loop makes sure that the current name hasn't already been
        # found as an sfo pair of a name already seen in the list.
        foreach my $found_name (keys %sfo_pairs) {
            if ($sfo_pairs{$found_name} eq $name) {
                $unique = 0;
                last;
            }
        }

        # If the name has been found move on to next name since we already
        # know its sfo pair.
        if ($unique == 0) {next;}

        # Otherwise call host and look for its sfo pair value and match it
        # with any of the names passed into the method.
        my $host_rec = Tharn::host($name);
        foreach my $second_name (@host_names){
            my $second_host_rec = host($second_name);
            if ($host_rec->{SFOPAIR} eq $second_host_rec->{SFOPAIR} && $name ne $second_name) {
                $sfo_pairs{$name} = $second_name;
                last;
            }
        }

    }

    return \%sfo_pairs;

}



## @name ::vseries_hba_port_disable_enable
##
## @description
## Disables the list of hba ports for the given filer passed in,
## then enables them.  This is ported from VSeries.pm because
## the commands to disable/enable hba ports are not present in
## BoilerMaker 10-Mode as of March 2008.  So we need to
## talk to the 7-Mode interface and want to user Server.pm to
## talk to filers.  This method assumes that all storage has two paths when
## called and will expect that all storage returns to two paths when the
## enabling of the hba is complete.
##
## @synopsis
## $returned_value = VSeries::Clustered::filer_convenience::vseries_hba_port_disable_enable($server_obj, \@hba_portlist);
##
## @arg $server Required: Server object to communicate to the filer through
## @arg \@hba_portlist Required: Array of hba ports.
## @arg $max_timeout Optional: Timeout value to wait on a port either being
##                         disabled or enabled again.  Default is 120 seconds.
##
## @author silva@netapp.com, ported to BoilerMaker 10-Mode by cattel@netapp.com
##
sub vseries_hba_port_disable_enable($@;$) {

    my ($server_object, $hba_list_ref, $max_timeout) = @_;
    my @hba_list = @$hba_list_ref;

    if (!defined($max_timeout)) {
        $max_timeout = 120;
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # If we find zero hba_ports exit with return code 1
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    if (@hba_list == 0){
        return($ERR_STATUS);
    }

    my $after = "";
    my $output = "";
    my $timeout = 0;

    foreach my $hba_port (@hba_list) {

        $timeout = time() + $max_timeout;
        $output = "";

        while ($output !~ /disable succeeded/) {

            $output = $server_object->dblade('cmd' => "storage disable adapter $hba_port");

            # -  - - - - - - - - - - - - - - - - -  - - - - - -
            # Check if time expired, if so return($ERR_STATUS) for error.
            # -  - - - - - - - - - - - - - - - - -  - - - - - -

            if ((time() > $timeout) && ($output !~ /disable succeeded/)) {
                return($ERR_STATUS);
            }
            sleep(5);
        }


        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Sleep for 35 seconds in between disable enable so the filer acknowlages the command.
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        sleep(35);   # need 35 seconds here see burt 242330


        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Enable the hba port..
        # -  - - - - - - - - - - - - - - - - -  - - - - - -


        $server_object->dblade('cmd' => "storage enable adapter $hba_port");

        my $secondary = '-';

        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Reset the timeout.
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        $timeout = time() + $max_timeout;  # this way we have plenty of time to come back up.

        my @temp = ();
        my $line = '';

        my @single_pathed_disks = ();

        # We're calling two path check here to ensure that all storage has been returned to
        # having two paths after re-enabling the hba.  We'll loop and poll for up to the
        # timeout value to make sure things have settled.
        do{

            my $two_path_check_array_ref = VSeries::Clustered::filer_convenience::two_path_check($server_object);
            @single_pathed_disks = @$two_path_check_array_ref;

            # -  - - - - - - - - - - - - - - - - -  - - - - - -
            # Check if time expired, if so return(1) for error.
            # -  - - - - - - - - - - - - - - - - -  - - - - - -

            if ((time() > $timeout) && scalar(@single_pathed_disks) != 0) {
                return($ERR_STATUS);
            }
            elsif (scalar(@single_pathed_disks) != 0) {
                sleep(5);
            }
        }while (scalar(@single_pathed_disks) != 0)
    }
    return($GOOD_STATUS);
}


## @name check_cf_state($;$)
##
## @description
## State will either be "UP", "TAKEOVER", or "DEAD" (a subset of the
## strings that "cf monitor" will print) or "UNREACH" if we can't seem
## to talk to a filer to find out its state, or "DISABLED" if
## clustering is disabled.
##
## If a takeover is in progress or a giveback is in progress, this
## command will keep retrying.
##
## About what the states mean: We don't accept "DOWN" and "REBOOTING"
## as states from cf monitor, for some reason.  This makes sense to me
## for REBOOTING (we'll keep trying, and it sounds like a transitory
## state) but not DOWN, but perhaps 'cf monitor' can never print DOWN
## since maybe DOWN implies the filer is halted and not taken over and
## hence neither filer has that IP address to be able to reply to rsh
## of "cf monitor".  "DEAD" might be more appropriately termed "UNDEAD"
## or "TAKEN OVER": someone at this filer's IP address was alive enough
## to tell us that this filer is dead, so it must be that the partner
## has taken over.
##
## @synopsis
## $returned_state = VSeries::Clustered::filer_convenience::check_cf_state($server_obj, $rate);
##
## @arg $server Required: Server object to communicate to the filer through
##
## @arg $rate Optional: rate to poll while waiting to determine the
##                      cf state of the filer.
##
## - Note from silva@netapp.com - needed to add support to catch the
##   case where $retstr =~ /has taken over/ b/c ignoring this causes
##   some of our .thpl's to enter infinite loops.
##
sub check_cf_state($;$){
    my ($server_object,$rate) = @_;
    $rate = $rate || 10;


    my $cur_state;

    while (1) {

        my $retstr;
        eval{
            $retstr = $server_object->dblade('cmd' => "cf status");
        };
        return 'UNREACH' if ($@);
        return 'DISABLED' if ($retstr =~ /Cluster disabled/);
        return 'DISABLED' if ($retstr =~ /takeover disabled/);
        return 'DISABLED' if ($retstr =~ /has disabled takeover by/);

        if ($retstr =~ /\s+has taken over\s+/) { last; }

        last if (length($retstr) &&
            ($retstr !~ /attempting takeover/) &&
            ($retstr !~ /giveback in progress/));

        sleep $rate;

    } # while (1)

    my $out;
    do {
        eval{
            $out = $server_object->dblade('cmd' => "cf monitor");
        };
        if(@$){
            return 'UNREACH';
        }
        if ($out =~ /(UP|TAKEOVER|DEAD)/) {
            $cur_state = $1;
        }
    } while ($out =~ /Try again/ && !defined($cur_state));

    if (!defined($cur_state)) {
        return 'UNREACH';
    }
    else {
        return $cur_state;
    }
}


## @name ::fcadmin_link_state($)
##
## @description
## Takes in a server object to communicate with the filer and does a
## "fcadmin link_state".  It then splits the output and tokenizes on whitespace
## creating an array.  From there it creates a hash table with the key being
## the interface name (0a,0b,0c,etc) and the value being a hash table that
## contains attributes for each piece of information on the line.  See sample
## output and notes around it for examples.
##
## @synopsis
## $fcadmin_hash_ref = VSeries::Clustered::filer_convenience::fcadmin_link_state($server_obj);
##
## @arg $server Required: Server object to communicate to the filer through
##
##
## SAMPLE OUTPUT:
##system node run -node vra-v3050-2a -command "priv set advanced; fcadmin link_state"
##Warning: These advanced commands are potentially dangerous; use
##         them only when directed to do so by Network Appliance
##         personnel.
##Link State for adapter channel 0a: UP - N-port --> Port 0 on switch 'vrs-br200e-3' (100000051e03ba6d)
##Link State for adapter channel 0b: OFFLINE (physical)
##Link State for adapter channel 0c: UP - N-port --> Port 0 on switch 'vrs-br200e-4' (100000051e03b684)
##Link State for adapter channel 0d: OFFLINE (physical)
##
## So if you take the line with 0a as the interface the hash that gets returned
## will have key value pairs of
## slot -> 0a
## state -> UP
## port_type -> N-port
## switch_port -> 0
## switch -> vrs-br200e-3
## port_wwn -> 100000051e03ba6d

sub fcadmin_link_state($){
    my ($server) = @_;
    my %return_hash = ();


    # This is the call to the 7-Mode CLI.
    my $output = $server->dblade('cmd' => "priv set advanced; fcadmin link_state");

    # Split the raw output on new lines.
    my @temp = split(/\n/, $output);

    # for each line of output we only care about lines that begin with
    # Link State.  From there we tokenize on whitespace and take
    # specific tokens and put them into the returned hash for a given
    # interface.
    foreach my $line (@temp) {

        if ($line =~ m/^Link State/) {
            my @tokens = split(/\s+/, $line);
            my %tmp_hash = ();

            my $current_slot = "";
            if (defined($tokens[5])) {
                chop($tokens[5]);
                $current_slot = $tokens[5];
            }
            $tmp_hash{'slot'} = $current_slot;
            if (defined($tokens[6])) {
                $tmp_hash{'state'} = $tokens[6];
            }
            if (defined($tokens[8])) {
                $tmp_hash{'port_type'} = $tokens[8];
            }
            if (defined($tokens[11])) {
                $tmp_hash{'switch_port'} = $tokens[11];
            }
            if (defined($tokens[14])) {
                $tmp_hash{'switch'} = substr($tokens[14], 1, -1);
            }
            if (defined($tokens[15])) {
                $tmp_hash{'port_wwn'} = substr($tokens[15], 1, -1);
            }

            $return_hash{$current_slot} = {%tmp_hash};

        }
    }

    return \%return_hash;

}

## @name ::poll_hba_interfaces($$$@)
##
## @description
## This is a convenience routine to poll on hba interfaces to detect
## transitions of state.  It is common practice in VSeries testing
## to manipulate the backend san through switches by doing
## port disable/enable, switch reboots, switch renaming, etc.  During these
## tests we want to poll on the hba interfaces that the filer sees the given
## switch through.  This method will poll on the given interfaces and loop
## until all interfaces hit the desired state or a timeout occurs.  The method
## will call fcadmin_link_state though that can be swapped out once an
## equivalent command is available in 10-Mode.
##
## @synopsis
## my $status = VSeries::Clustered::filer_convenience::poll_hba_interfaces($server_obj, $state, $timeout, \@interfaces);
##
## @arg $server Required: Server object to communicate to the filer through
##
## @arg $state Required: The expected resulting state you want the interfaces
##                       to transition to.
##
## @arg $timeout Required: The maximum timeout to poll for the interfaces to
##                         transition to the end state.
##
## @arg @interfaces Required: Array of interface names to poll on the given
##                            filer that the server object is configured to
##                            talk to.
##
sub poll_hba_interfaces($$$@){
    my ($server, $state, $timeout, $interfaces_ref) = @_;
    my @interfaces = @$interfaces_ref;

    # Timeout we'll use to control looping for polling.
    my $polling_timeout = time() + $timeout;

    # Fix for BURT-753658. Adjusting timeout to a multiple of sleep time of 3 seconds.
    $polling_timeout = time() + 9 if ($timeout == 10);

    # This is the flag that will be used to control if we should keep polling
    my $continue_polling = 1;

    while (time <= $polling_timeout && $continue_polling == 1) {
        $continue_polling = 0;
        my $fcadmin_poll_ref = VSeries::Clustered::filer_convenience::fcadmin_link_state($server);
        my %fcadmin_poll_hash = %$fcadmin_poll_ref;

        foreach my $port (keys(%fcadmin_poll_hash)) {
            foreach my $port_to_poll (@interfaces) {
                if ($port ne $port_to_poll) {
                    next;
                }
                elsif ($fcadmin_poll_hash{$port}{'state'} ne $state) {
                    $continue_polling = 1;
                    last;
                }

            }
            # This is for efficiency, if there's a reason to keep polling for
            # any reason there's no need to keep looking, just break out and
            # poll again.
            if ($continue_polling == 1) {
                last;
            }
        }

        # Sleep and re-poll.
        if ($continue_polling == 1) {
            sleep(3);
        }

    }

    # If continue polling is set, we must have timed out so return $ERR_STATUS
    if ($continue_polling == 1) {

        # Rechecking the link states one last time just in case the link has come up after 9 seconds
        # timeout. This fix is for BURT-753658.
        if ($timeout == 10) {

            my $fcadmin_poll_ref = VSeries::Clustered::filer_convenience::fcadmin_link_state($server);
            my %fcadmin_poll_hash = %$fcadmin_poll_ref;

            foreach my $port (keys(%fcadmin_poll_hash)) {
                foreach my $port_to_poll (@interfaces) {
                    if ($port ne $port_to_poll) {
                        next;
                    }
                    elsif ($fcadmin_poll_hash{$port}{'state'} ne $state) {
                        return $ERR_STATUS;
                    }
                }
            }
        } #end of fix for BURT-753658

        return $ERR_STATUS if ($timeout != 10);# If time-out is not 10, then we are failing as we did before. 

    }

    # Otherwise return good status.
    return $GOOD_STATUS;

}


## @name ::get_vseries_storage_port_count($@)
##
## @description
## Returns the number of storage ports for all filers in the
## array of filer names passed on by looping through the server object
## calling storage disk show and finding all the switch names and ports
## listed in primary or secondary paths for the luns.
## It ignores Native Disks which are Netapp Disks that typically look like:
## 0d.27 or 0a.18.  These are not associated with storage ports because
## they are not part of some FC switch connected to some 3rd party array.
##
## @synopsis
## my $sp_count = VSeries::Clustered::filer_convenience::get_vseries_storage_port_count($server, \@filer_names);
##
##
## @arg $server_obj Required: a server object used to talk to the filers.
##
## @arg \@filer_node_array Required: Array of vseries head DNS names in the cluster
##
##
sub get_vseries_storage_port_count($@) {

    my ($server_obj, $filer_node_ref) = @_;

    # Names of each filer in the cluster to iterate over.
    my @filer_node_array = @$filer_node_ref;

    # Will hold the actual storage ports seen.
    my @storage_port_list = ();

    # Will hold all output of storage disk show on each filer.
    my @all_disks_array = ();

    # Port count that will be returned to caller.
    my $port_count = 0;

    my %seen = ();                        # Hash used to eliminate duplicates
    my @uniq = ();                        # in @storage_port_list which becomes
                                          # @uniq.


    my %storage_port_table = %{VSeries::Clustered::filer_convenience::get_vseries_storage_port_associate($server_obj, \@filer_node_array)};

    foreach my $filer (keys(%storage_port_table)) {
        my $port_array_ref = $storage_port_table{$filer};
        my @port_array = @$port_array_ref;
        foreach my $port (@port_array){
            push(@storage_port_list, $port);
        }
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now lets convert @storage_port_list into a unique list of
    # switch:storageports.  Walk the through all entrys in
    # @switch_port_list and eliminate duplicates.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $item (@storage_port_list) {
        push (@uniq, $item) unless $seen{$item}++;  #(Recipe 4.6 page 102 Perl Cookbook)
    }

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # port_count is equal to the number of elements in @uniq
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    $port_count = scalar(@uniq);

    return ($port_count);
}


## @name ::get_vseries_storage_port_associate($@)
##
## @description
## Returns a reference to a hash of filer names and their associated
## switch:port's.
##
## @synopsis
## my $hash_ref = VSeries::Clustered::filer_convenience::get_vseries_storage_port_associate($server, \@filer_names);
##
## @arg $server_obj Required: A Server Object to talk to the filers with.
##
## @arg \@filer_node_array Required: Array of filer names.
##
sub get_vseries_storage_port_associate($@) {

    my ($server_obj, $filer_node_ref) = @_;

    # Names of each filer in the cluster to iterate over.
    my @filer_node_array = @$filer_node_ref;

    my %storage_port_list = (); # Hash used to track associated storage ports
    my @filer_lun_paths = ();   # List of all paths to luns
    my @output = ();
    my $sw_port = '';           # Temp holder for switch ports used in loops
    my %seen_on_filer = ();     # Hash used to help build an array of uniq
    my @uniq_on_filer = ();     # Storage ports in filera '@uniq_in_filera'


    foreach my $node (@filer_node_array) {
        $server_obj->set_execution_server('-Server' => $node);

        # Make the call to the server object and get the output to sift through.
        # The output will be an array of hash tables where the key value pairs are
        # the fields and values returned by the same call on the CLI of a filer.
        my $disks = $server_obj->storage_disk_show("-Timeout" => 600);
        my @my_disks = @$disks;

        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Now @my_disks is a list of all the primary/secondary
        # LUNS combined for each specific filer.
        #
        # Example:
        #
        #                         Disk: vra-v3050-2a:vrs-br200e-3:15.126L1
        #                      UUID: -
        #        Physical Size (MB): -
        #            Used Size (MB): -
        #                      Type: -
        #                    Vendor: -
        #             Serial Number: 600E000000A800000066700700010000
        #                     Model: -
        #         Firmware Revision: -
        #                       RPM: -
        #          Bytes Per Sector: -
        #    Checksum Compatibility: -
        #                     Shelf: -
        #                       Bay: -
        #                     State: -
        #              Detail State: -
        #                 Raid Type: -
        #                Raid Group: -
        #                      Plex: -
        #                 Aggregate: -
        #                      Home: vra-v3050-2b
        #                     Owner: vra-v3050-2b
        #                   Home ID: 101194430
        #                  Owner ID: 101194430
        #       Reservation Host ID: 101194430
        #                 Prefailed: -
        #            Being Replaced: -
        #                    Zeroed: -
        #          Copy Destination: -
        #           Copy Percentage: -
        #        Zeroing Percentage: -
        # Reconstruction Percentage: -
        #              Host Adapter: -
        #            Secondary Port: -
        #              Primary Port: -
        #            Secondary Path: -
        #          Hours Powered On: -
        #             # Blocks Read: -
        #          # Blocks Written: -
        #Seconds Since SHM Err Counting: -
        #       Grown Defect List Count: -
        #  Minutes Since Scrub Done: -
        #               Scrub Times: -
        #    Dynamically Qualified?: -
        #        Number Raw Sectors: -
        # Number of physical blocks: -
        #
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        # If the secondaryname (Secondary Path) field is populated with
        # meaningful info, meaning something other than a dash '-' then we care
        # about those too.
        foreach my $disk_ref (@my_disks) {
            my %disk_hash = %$disk_ref;
            push @filer_lun_paths, $disk_hash{'disk'};
            if ($disk_hash{'secondaryname'} ne '-') {
                push @filer_lun_paths, $disk_hash{'secondaryname'};
            }
        }


        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Splitting on \. splits the switch:storage_port.lun_number
        # so that you are left with two pieces, switch:storage_port
        # and lun_number.  Lets do that for both vseries heads.
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        foreach my $tmp3 (@filer_lun_paths) {
            ($sw_port, undef) = split(/\./, $tmp3);

            # This pulls out the filer name since we want the switch and port info.
            # Effectively the filer name will be the first token, then the switch
            # and the port will be part of the third token and need to be pulled
            # apart by splitting on \.
            my @tokens = split(/:/, $sw_port);
            my $switch = $tokens[1];
            my $port_num = (split(/\./, $tokens[2]))[0];

            my $switch_and_port_num = $switch . ":" . $port_num;

            push @output, $switch_and_port_num;
        }

        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Now we need to eliminate all duplicates in @output_filera
        # and @output_filerb.  We will use %seen_on_filer to do this,
        # then send the uniq elements to @uniq_on_filer
        #
        # The foreach blocks ahead are explained here:
        # (Recipe 4.6 page 102 Perl Cookbook)
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        foreach my $item1 (@output) {
            push (@uniq_on_filer, $item1) unless $seen_on_filer{$item1}++;
        }

        # -  - - - - - - - - - - - - - - - - -  - - - - - -
        # Now construct the %storage_port_list hash and return a reference to it.
        # -  - - - - - - - - - - - - - - - - -  - - - - - -

        %storage_port_list = (
            $node => [@uniq_on_filer],
        );
    }

    return (\%storage_port_list);
}

## @name ::get_vseries_storage_port_find($$$)
##
## @description Returns $ERR_STATUS if the given storage port (switchname and
## port_num) can be seen currently from a given vseries
##
## @synopsis
## my $return_val = VSeries::Clustered::filer_convenience::get_vseries_storage_port_find($server_obj, $switch, $port);
##
## @arg $server_obj Required: Server Object used to talk to the filer.
## @arg $switch Required: the switchname
## @arg $port Required: the port_num on the given switch
##
##
sub get_vseries_storage_port_find ($$$) {
    my ($server_obj, $switch, $port) = @_;
    my (%disklist, $path, $lun, $switch_and_port_num) = "";
    my $version = &version_check;

    if ($version eq $BR_STR) {
        # Get all the disks seen by the filer.
        my $disks_ref = $server_obj->storage_disk_show("-Timeout" => 600);
        my @disks = @$disks_ref;

		# Now for each of them look at the 'disk' field which is the primary path
		# and if the secondary path is known (not a dash) look at that as well (for
		# RR builds and later we'll consider "partner" status rather than
		# "secondary" path status). Pull out from that just the switch name and port
		# if at any point those are matched return $GOOD_STATUS otherwise return 
		# $ERR_STATUS. Format for the primary or secondary path will look like this:
		# vra-v3050-2a:vrs-br200e-3:15.126L1
		foreach my $disk_ref (@disks) {
			my %disk_hash = %$disk_ref;
			my $secondary = "";
			my @tokens = ();
        
			if($disk_hash{'type'} ne 'LUN'){next;}
	
			my $primary = $disk_hash{'disk'};
			@tokens = split(/:/, $primary);
			my $switch_name = $tokens[1];
			my $port_num = (split(/\./, $tokens[2]))[0];
	
			$switch_and_port_num = $switch_name . ":" . $port_num;

			if ($switch_and_port_num eq ("$switch" . ":" . "$port")) {
				 return $GOOD_STATUS; #found the storage port
			} 
			
			if ($disk_hash{'secondaryname'} ne '-') {

				$secondary = $disk_hash{'secondaryname'};
			}

			
			@tokens = split(/:/, $secondary);
			$switch_name = $tokens[1];
			$port_num = (split(/\./, $tokens[2]))[0];
	
			$switch_and_port_num = $switch_name . ":" . $port_num;
            
			if ($switch_and_port_num eq ("$switch" . ":" . "$port")) {
				return $GOOD_STATUS; #found the storage port
			}
        }
    } else {
        
        # For RR+, we'll look at the output of "storage array config show".
        my $array_config_ref = $server_obj->storage_array_config_show("-Timeout" => 1200);
        my @array_config = @$array_config_ref;
        
        foreach my $array_ref (@array_config) {
            my %array_config_hash = %$array_ref;
            if ($array_config_hash{'target-side-switch-port'} eq ("$switch" . ":" . "$port")) {
                return $GOOD_STATUS;
            }
        }
    }   
        
    return ($ERR_STATUS);
}

## @name ::compare_disk_configs(@@)
##
## @description
## Takes in two arrays that should be the output of
## the storage_disk_show command returned to Server.pm  Typically this would
## occur when you run the command then perform some action like a
## takeover/giveback and then run the command again.  It then
## figures out which array is larger, if they are equal the first
## array is treated as larger for looping purposes.  It then loops
## over the larger config, taking each hash and seeing if it's in the smaller
## hash if it isn't then that's one type of error and the serial number
## is put in the error_disks array.  If the hash does exist in the smaller
## array, it then walks all the key/value pairs of the hashes to see if there's
## a miscompare.  If there is, that's also another type of error and the serial
## number gets pushed onto error_disks.  When it's done it returns the
## error_disks array.  If the array is empty that means the configs are
## identical.
##
## @synopsis
## my @err_disks = @{VSeries::Clustered::filer_convenience::compare_disk_configs(\@start_disks, \@end_disks)};
##
## @arg @config1 Required: Array of hash tables returned from server.pm
##                         commands to a filer.
##
## @arg @config2 Required: Array of hash tables returned from server.pm
##                         commands to a filer.
##
sub compare_disk_configs(@@){
    my ($config1_ref, $config2_ref) = @_;
    my @config1 = @$config1_ref;
    my @config2 = @$config2_ref;
    my ($key1, $value1, $key2, $value2) = "";

    # Flag used to track if we have a miscompare of information
    # within a given disks hash.
    my $mismatched_disk = 0;

    my @larger_config = ();
    my @smaller_config = ();


    my @error_disks = ();

    # Figure out which array is the larger config and use that as the outer
    # loop.
    if (scalar(@config1) >= scalar(@config2)) {
        @larger_config = @config1;
        @smaller_config = @config2;
    }
    else{
        @larger_config = @config2;
        @smaller_config = @config1;
    }

    # This will be used to track when an element from the larger config is
    # found in the smaller config.  This way when we're done looping through
    # the larger config, we don't then have to loop through the smaller config
    # to see if the larger config is a superset of the smaller config or if
    # the smaller config has elements that aren't in the larger.
    my @smaller_config_matched_array = ();

    # Initialize the elements of the array to "n" and as they are found in
    # looping we'll switch them to "y".
    for (my $i = 0; $i < scalar(@smaller_config); $i++) { $smaller_config_matched_array[$i] = 'n';}


    # For each hash in the larger config, loop over the smaller config looking
    # for the same disk, if the disks aren't the same, move on to the next one
    # until we find the matching disk, if we don't that's an error_disk and we
    # add it's serial number to the error_disks array.  If we have matching
    # disks loop over each key/value pair checking they they exist in both
    # hashes and that their values are equal.
    foreach my $hash1_ref (@larger_config) {
        my %larger_config_disk = %$hash1_ref;
        my $disk_found = 0;
        for (my $i = 0; $i < scalar(@smaller_config); $i++) {
            my %smaller_config_disk = %{$smaller_config[$i]};

            if ($larger_config_disk{'disk'} ne $smaller_config_disk{'disk'}) {next;}
            else { $smaller_config_matched_array[$i] = 'y'; }

            # Use the flags to check if we have a missing disk, or that a
            # given disks info isn't identical.
            $mismatched_disk = 0;
            $disk_found = 1;

            # We'll loop over the larger config and compare with the
            # smaller config.  We replace any tabs that turn up in
            # the output with whitespace because some irregularities
            # showed up during some failover testing that shouldn't
            # result in mismatches.
            while (($key1, $value1) = each(%larger_config_disk)) {
                if (!defined($smaller_config_disk{$key1}) || $value1 ne $smaller_config_disk{$key1}) {
                    my $mod_val1 = $value1;
                    $mod_val1 =~ s/\t/ /g;
                    my $mod_val2 = $smaller_config_disk{$key1};
                    $mod_val2 =~ s/\t/ /g;

                    if($mod_val1 ne $mod_val2){
                        $mismatched_disk = 1;
                        last;
                    }
                }
            }
            # If the two values didn't match this is one type of error we'll
            # report.
            if ($mismatched_disk == 1) {
                push(@error_disks, $larger_config_disk{'serialnumber'});
            }
        }
        # If we couldn't find the disk in the other config this is another
        # error we'll return a disk for.
        if ($disk_found == 0) {
            push(@error_disks, $larger_config_disk{'serialnumber'});
        }
    }

    # Now loop over the array that kept track of finding elements in the
    # smaller config.  If the value wasn't changed to a 'y' then we add that
    # disk's serial number to the error array.
    for (my $i = 0; $i < scalar(@smaller_config_matched_array); $i++) {
        if ($smaller_config_matched_array[$i] eq 'n') {
        my %disk_hash = %{$smaller_config[$i]};
        push(@error_disks, $disk_hash{'serialnumber'});

    }
}



    return \@error_disks;

    }


## @name ::check_sfo_status($$$)
##
## @description
## Takes in a server object and two filers names.  It then runs the storage
## failover show command on the server object.  It takes that output and
## checks that various pieces of the two filer passed in are in synch for
## sfo status.  It will check that each filer believes the other is it's
## sfo partner.  It also polls for each filer till sfo is enabled, the
## state is connected and takeover is possible.  It will return $GOOD_STATUS
## and an empty error message string if all those things are true and
## $ERR_STATUS and a meaningful error message otherwise.
##
## @synopsis
## my ($sfo_status, $status_msg) = @{VSeries::Clustered::filer_convenience::check_sfo_status($server_obj, $primary, $partner)};
##
## @arg $server_object Required: Server object to talk to the filers
##
## @arg $primary Required: filer name that you want to check sfo status of.
##
## @arg $partner Required: partner of the primary to check the sfo status.
##
##
sub check_sfo_status($$$){
    my ($server_object, $primary, $partner) = @_;

    # Flags to keep track that we saw the primary and partner filers in the
    # output of storage failover show
    my $primary_seen = 0;
    my $partner_seen = 0;
    my $fail_str = undef;
    # Define timeout to wait for 'storage failover show' output to be UP
    my $wait_time = time() + 1200;
    # Hash to track status of primary and partner. A value of '1'
    # for a node indicates, that node is connected, enabled and
    # takeover-possible.
    my %node_status = ();
    my @sfo_output = ();
    # Poll till failover status is normal for primary and partner node
    while(time() < $wait_time){
        my $failover_enabled = 1;
        my $failover_connected_state = 1;
        my $failover_possible = 1;
        # Run the command and capture the output, should be an array of hashes.
        eval{
            @sfo_output = @{$server_object->storage_failover_show("-Timeout" => 1200)};
        };
        if($@){
            return $ERR_STATUS, $@;
        }
        
        # If there's only one hash then obviously there wasn't output for both
        # filers that were passed in, so report the an error.
        if(scalar(@sfo_output) < 2){
            my $filer_ref = $sfo_output[0];
            my %filer_sfo_hash = $filer_ref;
            return $ERR_STATUS, "$filer_sfo_hash{node} appears to be the only filer listed in storage_failover_show please triage...";

        }
        
        # Otherwise we'll loop through the output looking for either the primary
        # or partners output
        foreach my $ref (@sfo_output) {
            my %filer_hash = %$ref;

            # If the node isn't one of the filer names passed in then move on.
            if($filer_hash{"node"} ne $primary && $filer_hash{"node"} ne $partner){next;}

            # If the node matches the primary then check that it's partner is the
            # filer name passed in, if not then that's an error.  Else, if the name
            # matches the partner, check that it's partner is the primary filer
            # name passed in.  If it's not that's also an error.
            if($filer_hash{"node"} eq $primary){
                $primary_seen = 1;
                if($filer_hash{"partner-name"} ne $partner){
                    return $ERR_STATUS, $filer_hash{"node"} . "doesn't have it's partner set to $partner, it is set to " . $filer_hash{"partner-name"} . ", please triage...";
                }
            }
            elsif($filer_hash{"node"} eq $partner){
                $partner_seen = 1;
                if($filer_hash{"partner-name"} ne $primary){
                    return $ERR_STATUS, $filer_hash{"node"} . " doesn't have it's partner set to $primary, it is set to " . $filer_hash{"partner-name"} . ", please triage...";
                }
            }
            # Initialize node_status hash
            $node_status{$filer_hash{'node'}} = 0;

            # Now for each of the filers, either primary or partner, check that the
            # enabled status isn't false or blank.  Then check if the state is not
            # equal to connected, if that's true that's an error.  Then check if
            # the value of possible isn't true, if that isn't true then it means
            # takeover isn't possible and that's an error.
            
            if($filer_hash{"enabled"} eq "false" || $filer_hash{"enabled"} eq "-"){
                $failover_enabled = 0;
                $fail_str = $filer_hash{"node"} . " has a storage failover state of " . $filer_hash{"state"} . ", please triage...";
                sleep(15);
            }
            elsif(defined($filer_hash{"state"}) && $filer_hash{"state"} ne "connected" ){
                $failover_connected_state = 0;
                $fail_str = $filer_hash{"node"} . " has a storage failover state of " . $filer_hash{"state"} . ", please triage...";
                sleep(15);
            }
            elsif(defined($filer_hash{"state-description"}) && $filer_hash{"state-description"} !~ m/connected/i){
                $failover_connected_state = 0;
                $fail_str = $filer_hash{"node"} . " has a storage failover state of " . $filer_hash{"state-description"} . ", please triage...";
                sleep(15);            
            }  
            elsif(!defined($filer_hash{"state-description"}) && !defined($filer_hash{"state"})){
                $failover_connected_state = 0;
                $fail_str = $filer_hash{"node"} . " had storage failover show output that did not contain the key state or state-description cannot determine status, please triage...";
                sleep(15);   
            }          
            elsif($filer_hash{"possible"} ne "true"){
                $failover_possible = 0;
                $fail_str = "Failover for " . $filer_hash{"node"} . " isn't possible (" . $filer_hash{"possible"} . ") due to " . $filer_hash{"reason"} . ", partner reason is " . $filer_hash{"partner-reason"} . ", please triage...";
                sleep(15);
            }
            # Mark 'all-cool' for current node if the three fields are 'true'.
            $node_status{$filer_hash{'node'}} = 1 if( $failover_enabled == 1 && $failover_connected_state == 1 &&  $failover_possible == 1 );
        }
        last if( $node_status{$primary} == 1 && $node_status{$partner} == 1 );
    } # while

    # Return error if failover-status didn't come UP with-in required timeout
    return $ERR_STATUS, $fail_str if( $node_status{$primary} != 1 or $node_status{$partner} != 1 );

    # Now that we've looped over every node, check to make sure that we saw
    # both the primary and partner names in the output, if not then report the
    # error.
    if($primary_seen == 0){
        return $ERR_STATUS, "$primary wasn't found in storage failover show output, please triage...";
    }

    if($partner_seen == 0){
        return $ERR_STATUS, "$partner wasn't found in storage failover show output, please triage...";
    }

    # If we get this far then everything checks out and return good status.
    return $GOOD_STATUS, "";

}


## @name ::get_coredump_count($$)
##
## @description
## This is a convenience routine to check the number of core dumps listed in
## the system node coredump show command.  Since the command may fail
## non-gracefully if the table is empty it seemed easier to catch that here
## so that at the thpl layer the code looks cleaner.  It returns a hash with
## a key of "coredump_count" and it's value.  This is the first, and limited
## fuctionality around coredumps in 10-Mode so I left this flexible by using
## a hash for the return value.  If there's an error there will be a key
## "error" with the value being the output of the command when the error
## happened.
##
## @synopsis
## my $hash_ref = VSeries::Clustered::filer_convenience::get_coredump_count($server_obj, $filer_name);
##
## @arg $server_object Required: Server object to talk to the filers
##
## @arg $filer_name Required: filer name that you want to get the coredump
##                            count for.
##
##
sub get_coredump_count($$){
    my ($server_object, $filer_name) = @_;

    # Variables used to get the coredump count.
    my @coredumps = ();
    my $coredump_count = 0;
    my $output = "";
    my %return_hash = ();

    # Run the necessary command and check if the command failed due to an
    # empty table.  If it did fail but not as expected return an error.
    eval{
        @coredumps = @{$server_object->system_node_coredump_show("-Timeout" => 1200, 'output' => \$output)};
    };
    if($@){
        if($output =~ m/This table is currently empty/i){
            $coredump_count = 0;
            $return_hash{"saving"} = 0;
        }
        elsif ($output =~ m/saving(.*)?core/i) {
            $return_hash{"saving"} = 1;
        }
        else {
            $return_hash{"error"} = "Command failed to show coredumps on filer with message:\n$output"
        }

    # Here we catch RPC timeouts for coredump show. It won't be caught above
    # because error status doesn't get returned for that condition.
    } elsif ($output =~ m/(Unable to list entries on node .* RPC: Timed out)/i) {
        $return_hash{"error"} = $1;
    }

    # Otherwise now we'll look at each entry to of the array of hashes
    # returned by server.pm and look at the 'panic-node' attribute
    # if it matches the filer_name passed in, increment the count
    else{
        foreach my $coredump_ref (@coredumps){
            my %coredump = %$coredump_ref;
            if($coredump{'panic-node'} ne $filer_name) {
              next;
            }
            else {
              $coredump_count++;
              $return_hash{"saving"} = 0;
            }
        }
    }

    # Put the coredump count into the return hash and send it back to the
    # caller.
    $return_hash{"coredump_count"} = $coredump_count;

    return \%return_hash;

}

## @name ::showfh($$)
##
## @description
## Gets the filehandle information for $file and places that info in a hash
## called %fh_hash.  The keys are the whitespace separated param labels that
## appear in the output.
## We return a reference to that hash on success.  We return an error
## status along with an errormsg buffer if errors are detected.
##
## @synopsis
## ($status, $ref) = $server_obj>showfh($$);
##
## @arg $server_obj Required: Server object
## @arg $file       Required: File to retrieve filehandle from. The filename
##                            must be contained in an absolute path
##
sub showfh($$) {
    my ($server_obj, $filepath) = @_;
    
    my $output  = '';                  # Raw output from dblade call
    my @fh_vals = ();                  # Used fo splitting the return of showfh
    my %fh_hash = ();                  # Hash that will be returned with fh info
    my $param   = '';                  # Different params in showfh output
    my @errors  = ();                  # Error message buffer

    eval{
        $output = $server_obj->dblade('cmd' => "priv set -q diag; showfh $filepath");
    };
    if($@){
        push @errors, $@;
        return ($ERR_STATUS, \@errors);
    }

    # Check for command errors which can include:
    # 
    # showfh: Invalid parameter. 
    # usage: showfh filename
    # <path>: No such file or directory
    #

    if ($output =~ /^.*:\s+\w+/) {
        push @errors, "showfh error detected: $output";
        return ($ERR_STATUS, \@errors);
    }

    # Parse command output
    #
    # Sample outputs:
    #
    # vra-v3050-2a*> showfh /etc/messages                  
    # flags=0x00 snapid=0 fileid=0x0003f1 gen=0x000a0e fsid=0x90bc5d
    # vra-v3050-2a*> showfh /vol/csum_test_vol/csum_file_4k
    # flags=0x00 snapid=0 fileid=0x000060 gen=0x66984c fsid=0x6a5b8ba dsid=0x00000000000419 msid=0x00000080000419
    #

    @fh_vals = split(/\s+/, $output);
    
    foreach $param (@fh_vals) {
    
        # Do a further split on each line parameter. The identifier to the left
        # of the "=" is our hash key, the value to the right of "=" is hash
        # value
        my @param_arr = split ("=", $param);
        $fh_hash{"$param_arr[0]"} = $param_arr[1];

    } # end foreach $param


    return($GOOD_STATUS, \%fh_hash);
}


#--------------------------------------------------------------------------------

## @name ::dumpblock($$;$$$$$$$$$)
##
## @description
## Print all sorts of useful information about blocks or files such as
## disk, dbn, vbn, block_contents, status, pvbn, pvbn_from_parent,
## pvbn_of_parent, fbn_of_inofile, pvbn_of_inofile, pvbn_of_inofile_from_parent
## ckse_dbn, ckse_checksum, ckse_buftreeid, ckse_fbn, and ckse_cp_count.
## All of this information is returned in a hash ref which contains the keys
## mentioned above.  Additionally there is one key called 'parse_error' which
## stores any lines that didn't match a regex for debugging purposes.
## We return a reference to hash on success.  We return an error
## status along with an error msg buffer if errors are detected.
##


## *** THE FOLLOWING IS ONTAP USAGE ***
## vnv3050f9a> priv set -q diag
## vnv3050f9a*> dumpblock
## usage:
## dumpblock ifile [-A | -V] <vol-name> <fileid>
##   - dump block of inofile containing that inode
## dumpblock file [-A | -V] <vol-name> <fileid> <blockno>
##   - dump data block <blockno> from file inode # <fileid>
## dumpblock vbn [-A | -V] <vol-name> <vbn>
##   - dump block with the specified vbn on the specified volume/aggregate
## dumpblock dbn [-c] <disk_name> <dbn>
##   - dump block with the specified dbn on the specified disk
## dumpblock info [-A | -V] <vol-name> <fileid> <blockno>
##   - dump information about the specified block
## vnv3050f9a*>
## ***------------------------------***
##
## @synopsis
## ($status, $ref) = $filer_obj->dumpblock($server_obj, "ifile", "-vol_name=vol0", "-file_id=0x0d6a16");

##
## @arg $server_object   Required: Server Object
## @arg $type      Required: Dumpblock type <ifile|file|vbn|dbn|info>
## @arg $vol_name  Optional: A volume name
## @arg $file_id   Optional: A file id number (0x342d)
## @arg $A         Optional: Aggregate specifier
## @arg $V         Optional: Volume specifier
## @arg $block_no  Optional: A block number (physical)
## @arg $vbn       Optional: A block number (virtual)
## @arg $c         Optional: Disk specifier
## @arg $dbn       Optional: A disk block number
## @arg $disk_name Optional: A disk name
##
##
sub dumpblock($$;$$$$$$$$$) {
    my $server_obj = shift;               # Server object
    my $type       = shift;               # DB type <ifile|file|vbn|dbn|info>    
    my $db_options = '';                  # Holds concatenated @ARGV
    my @temp       = ();                  # Used to hold values while parsing
    my %db_hash    = ();                  # Hash returned with db info
    my $output     = '';                  # dblade command raw output
    my $line       = '';                  # Holds loop lines while iterating
    @ARGV          = @_;                  # Setup @ARGV for GetOptions
    my @errors     = ();                  # Error message buffer

    # Initialize option vars
    my($vol_name,
       $file_id,
       $A,
       $V,
       $block_no,
       $vbn,
       $c,
       $dbn,
       $disk_name
       ) = '';

    # Grab our options
    GetOptions( "vol_name:s"    => \$vol_name,
                "file_id:s"     => \$file_id,
                "A"             => \$A,
                "V"             => \$V,
                "block_no:s"    => \$block_no,
                "vbn:s"         => \$vbn,
                "c"             => \$c,
                "dbn:s"         => \$dbn,
                "disk_name:s"   => \$disk_name
              );

    # - - - - - - -  - - - - - - - - - - - - - -
    # dumpblock ifile [-A | -V] <vol-name> <fileid>
    #
    #vnv3050f9a*> dumpblock ifile vol0 0x0d6a16
    #disk 0d.25, dbn 7527790, vbn 7522542
    #0x0000: 0x4946424d 0x0005fdfe 0x0000a387 0x00000000 <MBFI............>
    #0x0010: 0x00000000 0x00000000 0x00000000 0x00000000 <................>
    #(More of the same, which we'll capture)
    # - - - - - - -  - - - - - - - - - - - - - -

    # - - - - - - -  - - - - - - - - - - - - - -
    # dumpblock file [-A | -V] <vol-name> <fileid> <blockno>
    #
    #vnv3050f9a*> dumpblock file -A vol0 0x0d6a16 0
    #disk 0d.25, dbn 7285674, vbn 7280426
    #0x0000: 0x206e7553 0x20727041 0x30203620 0x30303a30 <Sun Apr  6 00:00>
    #0x0010: 0x2030303a 0x20544d47 0x766e765b 0x30353033 <:00 GMT [vnv3050>
    #(More of the same, which we'll capture)
    # - - - - - - -  - - - - - - - - - - - - - -

    # - - - - - - -  - - - - - - - - - - - - - -
    #dumpblock vbn [-A | -V] <vol-name> <vbn>
    #
    #vnv3050f9a*> dumpblock vbn vol0 7280426
    #VBN 7280426, disk 0d.25, dbn 7285674
    #0x0000: 0x206e7553 0x20727041 0x30203620 0x30303a30 <Sun Apr  6 00:00>
    #0x0010: 0x2030303a 0x20544d47 0x766e765b 0x30353033 <:00 GMT [vnv3050>
    #0x0020: 0x3a613966 0x6f6e7520 0x64656e77 0x7369642e <f9a: unowned.dis>
    #(More of the same, which we'll capture)
    # - - - - - - -  - - - - - - - - - - - - - -

    # - - - - - - -  - - - - - - - - - - - - - -
    # dumpblock dbn [-c] <disk_name> <dbn>
    #
    #vnv3050f9a*> dumpblock dbn 0d.25 7285674
    #disk 0d.25, dbn 7285674
    #0x0000: 0x206e7553 0x20727041 0x30203620 0x30303a30 <Sun Apr  6 00:00>
    #0x0010: 0x2030303a 0x20544d47 0x766e765b 0x30353033 <:00 GMT [vnv3050>
    #0x0020: 0x3a613966 0x6f6e7520 0x64656e77 0x7369642e <f9a: unowned.dis>
    #(More of the same, which we'll capture)
    # - - - - - - -  - - - - - - - - - - - - - -

    # - - - - - - -  - - - - - - - - - - - - - -
    # dumpblock info [-A | -V] <vol-name> <fileid> <blockno>
    #
    #vra-v3050-2a*> dumpblock  info csum_test_vol 0x000060 0
    #ok
    #disk vrs-br200e-3:15.126L23
    #dbn 21761
    #pvbn 16513
    #pvbn-from-parent 16513
    #pvbn-from-container 16513
    #vvbn 12325
    #fbn-of-inofile 4
    #pvbn-of-inofile 16612
    #pvbn-of-inofile-from-parent 16613
    #pvbn-of-inofile-from-container 16612
    #vvbn-of-inofile 16384
    #fileid-of-container-l1 10931
    #fileid-of_container-l1-is-private 1
    #fbn-of-container-l1 16384
    #ckse-dbn 21761
    #ckse-checksum 0x570747ae
    #ckse-buftreeid 0x2abe
    #ckse-fbn 12325
    #ckse-cp-count 36
    #ckse-notflexvol-flag 0
    # - - - - - - -  - - - - - - - - - - - - - -

    # Validate options foreach type

    if ($type =~ /^(ifile|file|vbn|info)$/) {
        # We can't have $A & $V, we need a $vol_name
        if (($A && $V) || (!$vol_name)){
            push @errors, "Error: filer::dumpblock(): Incorrect parameters given to dumpblock.  Must have optionally A or V and vol_name.";
            return ($ERR_STATUS, \@errors);
        }
        if ($type =~ /^(ifile|file|info)$/) {
            # ifile file and info need a $file_id specified.
            if (!$file_id) {
                push @errors, "Error: filer::dumpblock(): Incorrect parameters given to dumpblock.  Must have a file_id.";
                return ($ERR_STATUS, \@errors);
            }
        }
        if (($type =~ /^(file|info)$/) && ($block_no !~ /\d+/)) {
            # file & info type needs to have a $block_no   
            push @errors, "Error: filer::dumpblock(): Incorrect parameters given to dumpblock.  Must have a block_no.";
            return ($ERR_STATUS, \@errors);
        }
   
        if (($type =~ /^vbn$/) && ($vbn !~ /\d+/)) {
            # vbn needs a $vbn which should be some virtual block number
            push @errors,"Error: filer::dumpblock(): Incorrect parameters given to dumpblock vbn.  Must have a vbn.";
            return ($ERR_STATUS, \@errors);
        }
    }
    elsif ($type =~ /^dbn$/  ) {
        # We need a $disk_name & $dbn.
        # To complicate further, $dbn is a number and it must exist,
        # Therefore to validate we ensure it matches >= 1 sequential digits.
        if ((!$disk_name) || ($dbn !~ /\d+/)){
            push @errors, "Error: filer::dumpblock(): Incorrect parameters given to dumpblock dbn.  Must have disk_name and dbn.";
            return ($ERR_STATUS, \@errors);
        }
    } else {
        # User did not specify a type supported by the dumpblock command.
        push @errors, "Error: filer::dumpblock(): Incorrect dumpblock type, type specified: $type.  Must be info, dbn, vbn, file or ifile.";
        return ($ERR_STATUS, \@errors);
    }

    # All options/arguments have been validated.
    # We can continue on and build our $db_options string by appending
    # options that are defined.  Strip any whitespace or commas globally.

    foreach my $option (($A, $V, $c, $vol_name, $disk_name, $file_id, $block_no, $vbn, $dbn)) {
        # If option is true, or is some number, its valid.
        if (($option) || ($option =~ /\d+/)) {
            $option =~ s/(\s+|,)//g;
            $db_options .= "$option ";
        }
    }

    # Strip the last space off of the $db_options string.
    chop($db_options);

    # Send cmd to OnTap then populate the dumpblock hash (%db_hash)
    eval{
        $output = $server_obj->dblade('cmd' => "priv set -q diag; dumpblock $type $db_options");
    };
    if($@){
        push @errors, $@;
        return ($ERR_STATUS, \@errors);
    }

    $output =~ s/\r//;
    @temp = split("\n", $output);
    
    # Check that our first line of output is not the echoed command line.
    # If it is then remove it from the temp array
    if ( $temp[0] =~ m/dumpblock/ ) {
        @temp =  @temp[1 .. $#temp];
    }
  
    # Handle the 'dumpblock info' option which will be a separate line
    # for each attribute. All lines except the first consist of a label
    # and a value. All labels found become %db_hash keys.   
    if ($type eq "info") {
        foreach $line (@temp) {
            # Parse for exception: single entity 'ok' line  
            if  ($line =~ m/^ok/) { 
                $db_hash{'status'} = "ok";
            } elsif ( $line =~ /^([^ ]+)\s+([^ ]+)$/ ) {
                $db_hash{$1} = $2;
            } else {
                push @errors, "found unexpected formatted dumpblock info line: $line";
                return ($ERR_STATUS, \@errors);
            }
        }
    } else {
        # Type is ifile|file|vbn|dbn
        # Strip the first line of output and push the rest
        # into $db_options{'block_contents'}
        # Type is ifile|file|vbn|dbn
        # Strip the first line of output and push the rest
        # into $db_options{'block_contents'}
        
        foreach my $x (split(/\,/, $temp[0])) {
            if ($x =~ /vbn\s+/i) {
                # Strip whitespace, VBN, or vbn and then store just the data.
                $x =~ s/(\s+|VBN|vbn)//g;
                $db_hash{'vbn'} = $x;
            }
            elsif ($x =~ /disk\s+/) {
                # Strip whitespace and disk tag and then store just the data.
                $x =~ s/(\s+|disk)//g;
                $db_hash{'disk'} = $x;
            }
            elsif ($x =~ /dbn\s+/) {
                # Strip whitespace and dbn tag and then store just the data.
                $x =~ s/(\s+|dbn)//g;
                $db_hash{'dbn'} = $x;
            }
        }
        @temp = @temp[1 .. $#temp];
        foreach $line (@temp) {$db_hash{'block_contents'} .= $line;}

        # Get  the cksum disk, dbn and cksum_dbn values
        # sample output
        # 0x0fe0: 0x00000000 0x00000000 0x00000000 0x00000000     <................>
        # 0x0ff0: 0x00000000 0x00000000 0x00000000 0x00000000     <................>
        # cksum_data: disk vgci9124s38:1-23.126L33, dbn 40266, cksum_dbn 40288
        # 0x0000: 0x00000100 0x02000001 0x00009d4a 0x4d053881     <........J....8.M>
        if ($output =~ /cksum_data\:\sdisk\s(.*?)\,\sdbn\s(\d+)\,\scksum_dbn\s(\d+)/) {
            Tharn::logcomment("line is $output");
            $db_hash{'chkdata_disk'} = $1;
            $db_hash{'chkdata_dbn'} = $2;
            $db_hash{'chkdata_cksum_dbn'} = $3;
        }
    }
    
    # Strip any endlines off of the values in %db_hash
    # unless its the block_contents key.

    foreach my $element (keys %db_hash) {
        unless ($element eq 'block_contents') {
            $db_hash{$element} =~ s/\n//g;
        }
    }

    return($GOOD_STATUS, \%db_hash);
}

## @name ::dd($;$$$$$$$)
##
## @description Method that allows for OnTap's 'dd' utility
## which copies blocks of data.
##
## dd [ [if= file ] | [ din= disknum bin= blocknum ] ]
##    [ [of= file ] | [ dout= disknum bout= blocknum ] ]
## count= number_of_blocks
##
## dd  copies  the  specified number of blocks from source to
## destination.  Source and destination may be  specified  as
## either  a file that is a fully qualified pathname, or as a
## starting block on a disk. The parameter disknum may  range
## from  zero  to the maximum number reported by na_sysconfig
## (1) -r. In the latter form of specifying source or  desti-
## nation,  both  disknum  and blocknum must be specified.
## If the number of blocks exceeds the  size  of  the
## file, copying stops upon reaching EOF.
##
## Unlike the OnTap version of this utility you always need to
## specify a source and destination.  This will not substitute
## those voids with STDIN or STDOUT.
##
## Vseries will be placed into 'priv set diag' mode during this
## methods use, but will be returned to the default 'priv set admin'
## at its completion.
##
## @synopsis
## ($status, $ref) = $server_obj->dd($server_obj,"-if=/vol/test_vol/test_file",
##                          "-dout=0f.28", "-bout=13491");
##
## @arg $server_obj  Required:  Server Object
## @arg $if     Optional: Input file
## @arg $of     Optional: Output file
## @arg $din    Optional: Disk in
## @arg $bin    Optional: Block in
## @arg $dout   Optional: Disk out
## @arg $bout   Optional: Block out
## @arg $count  Optional: Count (of blocks)
##
sub dd($;$$$$$$$) {

    my $server_obj = shift;                # Server object
    my @argv_copy  = @_;                   # Copy ARGV before we process
    my $dd_options = '';                   # Holds concatenated @ARGV
    my $output     = '';                   # dblade command raw output
    @ARGV = @_;                            # Setup @ARGV for GetOptions
    my @errors     = ();                   # Error message buffer


    # These are all false by default.
    my ($if, $of, $din, $bin, $dout, $bout, $count) = '';

    GetOptions( "if:s"    => \$if,
                "of:s"    => \$of,
                "din:s"   => \$din,
                "bin:i"   => \$bin,
                "dout:s"  => \$dout,
                "bout:i"  => \$bout,
                "count:i" => \$count);

    # Let's validate our options.

    # Can't have din w/out bin & vice versa
    if (($din && !$bin) or ($bin && !$din)) {
        push @errors, "Error: filer::dd(): din and bin must not be mutually exclusive.";
        return ($ERR_STATUS, \@errors);
    }
    # Can't have dout w/out bout & vice versa
    if (($dout && !$bout) or ($bout && !$dout)) {
        push @errors, "Error: filer::dd(): dout and bout must not be mutually exclusive.";
        return ($ERR_STATUS, \@errors);
    }
    # Can't have any combination of 'of' with dout or bout
    if ($of && ($dout|$bout)) {
        push @errors, "Error: filer::dd(): of together with dout or bout are mutually exlusive options.";
        return ($ERR_STATUS, \@errors);
    }
    # Can't ONLY have count|din|bin|dout|bout|if|of & you need a src and dest
    if (scalar @argv_copy < 2) {
        push @errors, "Error: filer::dd(): must specify a source and destination.";
        return ($ERR_STATUS, \@errors);
    }
    # If the user doesn't get the arguments right, and it gets passed
    # the wall above we'll just have to give them a general purpose error
    # if things don't end up successful.  We'll set the error flag in the
    # filer object.

    # Create the options string to be passed via say().  Strip any white-
    # space in any of the options. Get rid of any leading dash in an argument.
    # We do not delete all dashes because disk arg may contain them

    foreach my $opt (@argv_copy) {
        $opt =~ s/^\s+$//g;
        $opt =~ s/^-//g;
        $dd_options .= $opt . " ";
    }

    # We're not checking dblade output here for any errors as we should have
    # caught all issues with command composition by now.  Furthermore, I wanted
    # to simplify here as it was recently discovered that the raw output from
    # the dd() can actually be seen now.  That means virtually any text could
    # be in an input file or block, making screen scrubbing non-ideal here.
    eval{
        $server_obj->dblade('cmd' => "priv set -q diag; dd $dd_options");
    };
    if($@){
        push @errors, $@;
        return ($ERR_STATUS, \@errors);
    }

    return($GOOD_STATUS);
}


## @name ::run_wafliron($$)
##
## @description
## Run wafliron on an aggregate and poll to verify when wafliron completes
##
## @synopsis
## ($status, $ref) = $server_obj>run_wafliron($$);
##
## @arg $server_obj Required: Server object
## @arg $arg_hash_ref Required: hash ref of wafliron command args 
##    %arg_hash keys:
##    aggregate - aggr name (this is the only required key)
##    max_timeout - number of seconds we wait for wafliron to finish
##                  if not defined, default is 600 sec 
sub run_wafliron($$) {
    my ($server_obj, $arg_hash_ref) = @_;

    my @errors     = ();             # Error message buffer
    my %arg_hash = %$arg_hash_ref;   # wafliron command opts
    my $max_timeout = 600;

    # Check that our arg hash minimally defines the aggr name
    if ( ! defined $arg_hash{'aggregate'} ) {
        push @errors, "ERROR: aggregate name not specified in arguments hash";
        return ($ERR_STATUS, \@errors);
    }

    # Set max_retries to user specified value if specified
    if ( defined $arg_hash{'max_timeout'} ) {
        $max_timeout = $arg_hash{'max_timeout'};
    }

    # Kick off wafliron
    eval{
        $server_obj->storage_aggregate_wafliron_start('aggregate' => $arg_hash{'aggregate'});
    };
    if ($@){
        push @errors, $@;
        return ($ERR_STATUS, \@errors);
    }   

    my $retries = 0;
    my @wafliron_show = ();
    my $start_time = time();
    my $elapsed_time = 0;
    
    # Wait for wafliron to finish
    while ( $elapsed_time < $max_timeout ) {
        eval {
            @wafliron_show = $server_obj->storage_aggregate_wafliron_show('aggregate' => $arg_hash{'aggregate'});
        };
        if($@){
            push @errors, $@;
            return ($ERR_STATUS, \@errors);
        }       
        my %wafliron_hash = %{$wafliron_show[0]};

        # Poll on the expected online aggr state, wafliron completion and 
        # success 
        if ( $wafliron_hash{"state"} eq "online" &&  $wafliron_hash{"iron-last-start-error-info"} eq "Successful." &&
            $wafliron_hash{"iron-status"} eq "not_running" ) {
            # wafliron completed
            last;
        } else {
            # sleep for 15 seconds before we check wafliron status again
            sleep 15;
            $elapsed_time = time() - $start_time;
        }
    }

    # If we exceeded the timeout waiting for wafliron to finish, flag an error
    if ( $elapsed_time >= $max_timeout ) {
            push @errors, "ERROR: exceeded $max_timeout sec timeout waiting for wafliron to finish";
            return ($ERR_STATUS, \@errors);
    }

    return ($GOOD_STATUS);
}

## @name ::storage_array_modify($$$;$$$$)
##
## @author Warren Crossfield (warrenc@netapp.com)
##         James Silva (silva@netapp.com) - Just added getops stuff.
##                                        - ported for 10mode use.
##
## @description
## Modifies an existing array profile and returns $GOOD_STATUS or $ERR_STATUS
## on success/failure. On error conditions, we'll also return an array
## reference with the error conditions we've encountered.
## Be cautious about sending your options into this
## method, as they will get truncated if they fall outside the bounds of
## the defined option formats.  i.e. $prefix cannot be longer than 4 chars.
## The following call would modify 2 of the 5 options, in this case the
## options and model fields.  Note you NEED to give atleast one option
## to modify, hence the required $$$ in the prototype.  The problem is
## that any of those options are 'optional' so pick one to modify & that
## becomes your required param.
##
## @synopsis
## ($status, $err_array) = VSeries::Clustered::filer_convenience::storage_array_modify($server, $name, "-options=jimbo", "-model=EMC2");
##
## @arg $server      Required: Server object to communicate with.
##
## @arg $name        Required: Array profile instance to modify.
##
## @arg $name_change Optional: Change existing array profile name to this.
##                   Note: max string length of 29 bytes.
##
## @arg $model       Optional: Change the model of the $name array profile.
##                   Note: max string length of 16 bytes.
##
## @arg $vendor      Optional: Change the vendor of the $name array profile.
##                   Note: max string length of 8 bytes.
##
## @arg $prefix      Optional: Change the prefix of the $name array profile.
##                   Note: max string length of 4 bytes.
##
## @arg $options     Optional: Change the options of the $name array profile.
##                   Note: max string length of 128 bytes.
##
sub storage_array_modify($$$;$$$$) {
    # We need a $server, $name & at least one option
    my $server  = shift @_;    # The filer object to shift.
    my $name    = shift @_;    # The name of the array profile to modify.
    my $after   = '';          # Holds raw output from Tharn::say().
    my @err     = ();          # Error array used when we return $ERR_STATUS.
    @ARGV       = @_;          # Copy @_ into @ARGV for getops processing.

    # These are all false by default.
    my($name_change,   # See @arg's section for info on these.
       $model,
       $vendor,
       $prefix,
       $options,
       ) = '';

    # storage_array modify <array-name> [-m <model>] [-n <new_name>] [-v vendor] [-p <prefix>] [-o options]

    # bash-2.05b$ rsh -l root vnv3050f9a "storage array show"
    # Prefix                        Name    Vendor             Model  Options
    # XXXX  XXXXXXXXXXXXXXXXXXXXXXXXXXXX   XXXXXXX   XXXXXXXXXXXXXXX  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX


    my $cmd = 'storage array modify';
    $cmd .= " $name";

    # Grab our options
    GetOptions( "name_change:s"    => \$name_change,  # Optional string
                "model:s"          => \$model,       # Optional string
                "vendor:s"         => \$vendor,      # Optional string
                "prefix:s"         => \$prefix,      # Optional string
                "options:s"        => \$options,     # Optional string
              );

    # Take in the string values and trim them to legal sizes
    # The sizes were all tested empirically. The 29 bytes for name bugs me.
    # Everything else seems to be a logical size. If we were using catch-
    # throw here I could also throw a warning if any of these values are
    # passed in too long. Alas, we do not use it. :( I wrapped these all in
    # if's in case we ever wanted to do other operations on them before
    # "saying" them.
    if($model) {
        my $truncated_model = $model;
        $cmd .= " -m $truncated_model";
    }
    if($name_change) {
        my $truncated_name = $name_change;
        $cmd .= " -n $truncated_name";
    }
    if($vendor) {
        my $truncated_vendor = $vendor;
        $cmd .= " -v $truncated_vendor";
    }
    if($prefix) {
        my $truncated_prefix = $prefix;
        $cmd .= " -p $truncated_prefix";
    }
    if($options) {
        my $truncated_options = $options;
        $cmd .= " -o $truncated_options";
    }
    # Send the command via the dblade.
    $after = $server->dblade('cmd' => $cmd);

    # Did it complete properly? If not, toss an error back.
    if($after =~ m/Array profile updated/) {
        return($GOOD_STATUS);
    } else {
        push(@err, "Error: filer_convenience::storage_array_modify() failure: $after");
        return($ERR_STATUS, \@err);
    }
}


## @name ::storage_array_remove_port($$$)
##
## @author Mark LaBonte (labonte@netapp.com)
##         James Silva  (silva@netapp.com)
##                      - Ported for 10-mode use.
##
## @description
## Deletes the specified WWPN associated with the specified array.
## Returns $GOOD_SATATUS or $ERR_STATUS & on error conditions,
## we'll also return an array reference with those errors.
##
## @synopsis
## ($status, $err_array) = VSeries::Clustered::filer_convenience::storage_array_remove_port($server, "EMC_SYMMETRIX_1", "50060482cb1bce3d");
##
## @arg $server       Required: Server object to communicate with.
##
## @arg $array_name   Required: The array associated with the WWPN to delete.
##
## @arg $port_name    Required: The WWPN to delete.
##
##
sub storage_array_remove_port($$$) {

    #storage array remove-port <array-name> -p <WWPN>

    my ($server, $array_name, $port_name) = @_;
    my $after = '';
    my @err = ();   # Error array used when we return $ERR_STATUS.

    my $cmd = "storage array remove-port $array_name -p $port_name";

    $after = $server->dblade('cmd' => $cmd);

    if ($after =~ /Port $port_name removed from Array Profile $array_name/) {
        return ($GOOD_STATUS);
    } else {
        push(@err, "filer_convenience::Failed to remove port $port_name: $after");
        return($ERR_STATUS, \@err);
    }

}


## @name ::storage_array_show($;$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Captures the information shown by the 'storage array show' command:
## vnv3040f8a> storage array show
## Prefix         Name     Vendor    Model           Options
##          IBM_1722_1        IBM     1722
##    123   IBM_1722_2        IBM     1722   ou812great2hear
## This information will be stored and returned as a hash reference on 
## $GOOD_STATUS. The format for the hash will have the 'Name' of the
## array as the root key. Each Name will have a respective Prefix,
## Vendor, Model and Options. On error we return $ERR_STATUS and the error
## array reference. On sucess we return $GOOD_STATUS and a storage
## array show hash ref. Lastly if you specify an $array_name you will
## only be returned info for that $array_name.
##
## @synopsis
## ($status, $ref) = VSeries::Clustered::filer_convenience::storage_array_show($server);
##
## @arg $server       Required: Server object to communicate with.
##
## @arg $array_name    Optional: An array profile name that we can capture
##                               information specifically for.
##
sub storage_array_show($;$) {
    my ($server, $array_name) = @_;
    my $after      = ''; # Holds raw output from Tharn::say().
    my @ap_builder = (); # Contains split up raw output from $after.
    my %ap_hash    = (); # Hash to contain array profiles.
    my @err        = (); # Error handling, to be returned as a ref.

    #  $cmd is the base command, we'll add $array_name, if specified.
    my $cmd = 'storage array show';
    $cmd   .= " $array_name" if ($array_name);

    # $after contains the raw output from our sent command.
    $after = $server->dblade('cmd' => $cmd);

    # Did it complete properly? If not, toss an error back.
    if($after =~ /No Array Profile with the name/) {
        push(@err, "Error: filer_convenience::storage_array_show() failure: $after");
        return($ERR_STATUS, \@err);
    }

    # $after looks something like this:
    # Prefix         Name     Vendor    Model           Options
    #          IBM_1722_1        IBM     1722

    # Split the returned output on endlines, then strip off the header.
    # Use @ap_builder as an initial container.
    @ap_builder = split(/\n/, $after);
    @ap_builder = @ap_builder[1..$#ap_builder];

    # Did we even have an array profile?
    if (scalar @ap_builder == 0) {
        push(@err, "Error: filer_convenience::storage_array_show(): Reports the presence of zero array profiles.\nRaw output: $after");
        return($ERR_STATUS, \@err);
    }

    # Now we need to iterate over each line of the @ap_builder array
    # which at this point is 1 or more array profiles.
    foreach my $line (@ap_builder) {

        # I know where all the fields have to live
        # due to the formatting of the header of the 'storage array show' cmd:
        # char *header_str = "%6.6s  %28.28s  %8.8s  %16.16s  %s\n";
        #                     prefix name     vendor model    options
        my $name = substr($line, 8,31);

        # Strip out any whitespace from $name:
        $name =~ s/\s+//g;

        # Now build a temporary hash which will be our $name key's value.
        my %temp = (
            'prefix'  => substr($line, 0, 8),
            'vendor'  => substr($line, 39, 10),
            'model'   => substr($line, 49, 17),
            'options' => substr($line, 66)
        );

        # Now insert the %temp hash into the $name key, making %ap_hash a HoH.
        # If we detect that a hash with this name already exists, we'll need
        # to return $ERR_STATUS.  Why, because this means we are showing
        # duplicate array names, which is on ONTAP bug.  
        if (exists $ap_hash{$name}) {
            push(@err, "Error: filer_convenience::storage_array_show(): Reports the presence of duplicate array names.\nRaw output: $after");
            return($ERR_STATUS, \@err);            
        }
        else { %{$ap_hash{$name}} = %temp; }

        # Get rid of any whitespace in the hash values:
        foreach my $element (keys %{$ap_hash{$name}}) {
            $ap_hash{$name}{$element} =~ s/\s+//g;
        }
    }

    return($GOOD_STATUS, \%ap_hash);
}


## @name ::storage_array_show_ports($;$)
##
## @author Mark LaBonte (labonte@netapp.com)
##         James Silva  (silva@netapp.com)
##                      - Ported for 10-mode use.
##
## @description
## Run storage array show-ports on a V-Series.  Optionally takes in
## an array name if the caller wishes to obtain specific information
## associated with a given array name (array profile).  This will
## return $GOOD_STATUS along with a hash reference on success.
## The hash has the following format:  The keys are the unique
## wwpn's and they have the following values 'array_name' & 'wwnn'.
## i.e. $ports{'50060482cb1bce3d'}{'array_name'} = "EMC_SYMMETRIX_1";
##      $ports{'50060482cb1bce3d'}{'wwnn'} = "50060482cb1bce3d";
## On error, it will return $ERR_STATUS along with an error array reference.
##
## @arg $server         Required: The server object to use.
##
## @arg $array_name     Optional: Array profile you wish to get
##                                specific port info for.
##
## @synopsis
## ($status, $ref) = VSeries::Clustered::filer_convenience::storage_array_show_ports($server);
##
sub storage_array_show_ports($;$) {
    my ($server, $array_name) = @_;
    my $after = '';
    my %port_hash = ();
    my $wwpn = '';
    my $wwnn = '';
    my @err  = ();  # The a stack of errors returned on $ERR_STATUS. 

    # $cmd is the base command, we'll add $array_name, if specified.
    my $cmd = 'storage array show-ports';
    $cmd .= " $array_name" if ($array_name);

    # This is an example of the raw output from this command.
    #vnv3050f9a*> storage array show-ports
    #                  Array Name                    WWPN                    WWNN
    #             EMC_SYMMETRIX_1        50060482cb1bce3d        50060482cb1bce3d
    #vnv3050f9a*>

    # $after contains the raw output from our sent command.
    $after = $server->dblade('cmd' => $cmd);

    # Split the command on new lines.
    my @output = split(/\n/, $after);

    my $regex = '\s+(\S+)\s+(\w{16})\s+(\w{16})';

    foreach my $item (@output[1..$#output]) {    # Index 0 is the column headers.
        if ($item =~ /$regex/) {
            ($array_name, $wwpn, $wwnn) = ($1, $2, $3);
        } else {
            # This will catch ALL errors, be it usage or array profiles
            # not found.  We know this due to the $item not containing
            # the three elements we need per line.
            push(@err, "filer_convenience::storage_array_show_ports: Found a line missing one or more of the following: array_name, wwpn, or wwnn while parsing output from $cmd.\n Output is: $item");
            return($ERR_STATUS, \@err);
        }

        # Now build a temporary hash which will be our $wwpn key's value.
        my %temp = (
            'array_name' => $array_name,
            'wwnn' => $wwnn,
        );

        # Now insert the %temp hash into the $name key, making %port_hash a HoH.
        %{$port_hash{$wwpn}} = %temp;

    }
    return($GOOD_STATUS, \%port_hash);
}


## @name ::storage_array_show_luns($$;$)
##
## @author Warren Crossfield (warrenc@netapp.com)
##         James Silva  (silva@netapp.com)
##                      - Ported for 10-mode use.
##
## @description
## Runs the storage array show-luns command and takes in a server, array_name
## and optionally some options. On success this will return $GOOD_STATUS and
## a hash reference. On error, it will return $ERR_STATUS and an error array
## reference.
##
## @synopsis
## ($status, $ref) = VSeries::Clustered::filer_convenience::storage_array_show_luns($server, "EMC_SYMMETRIX_1");
## ($status, $ref) = VSeries::Clustered::filer_convenience::storage_array_show_luns($server, "EMC_SYMMETRIX_1","-p 50:06:04:8a:cb:1b:ce:2c");
## ($status, $ref) = VSeries::Clustered::filer_convenience::storage_array_show_luns($server, "EMC_SYMMETRIX_1","-p 50:06:04:8a:cb:1b:ce:2c -a");
## ($status, $ref) = VSeries::Clustered::filer_convenience::storage_array_show_luns($server, "EMC_SYMMETRIX_1", "-a");
##
## @arg $server           Required: Server object to communicate with.
## @arg $array_descriptor Required: Array profile name.
## @arg $options          Optional: Options to append to the base command
##                                 'storage array show-luns $array_descriptor'
##
## _Ontap_ Usage for command is as follows (helps user in defining $options):
## storage array show-luns <array-name> [-a] [-p <WWPN>]
## Which means $options should only be populated with stuff in []'s.
##
## Sample %ap_hash follows..
##
##  $VAR1 = 'vnbr3800s46:3.126L3';
##              $VAR2 = [
##          '50:06:04:82:cb:1b:ce:2c',
##          '50:06:04:82:cb:1b:ce:3c'
##                      ];
##  $VAR3 = ...;
##  $VAR4 = [...];
sub storage_array_show_luns($$;$) {
    my ($server, $array_descriptor, $options) = @_;
    my $after      = ''; # Holds raw output from Tharn::say().
    my @ap_builder = (); # Contains split up raw output from $after.
    my %ap_hash    = (); # Hash to contain array profiles.
    my $i          = 0;  # A generic iterator. Iterate baby.
    my @err        = (); # An error array used when returning $ERR_STATUS.

    # storage_array show-luns <array-name> [-p <WWPN>] [-a]
    # $cmd is the base command. We'll add $options if specified,
    # but $array_descriptor is required.
    my $cmd = "storage array show-luns $array_descriptor ";

    # Did the user specify any options?
    if(defined $options) {
        # Append the options.
        $cmd .= "$options";
    }

    # Send @cmd via the standard connection, no need for console as this
    # isn't available when the machine is down.
    $after = $server->dblade('cmd' => $cmd);

    # Check for errors, they are all listed in the or'd regex below.
    if ($after =~ m/(No Array Profile with the name|usage:|Please use the following format:|No array found with the name|Unable to obtain storage configuration information)/) {
        push(@err, "filer_convenience::storage_array_show_luns: Error found while executing command $cmd\n$after");
        return($ERR_STATUS, \@err);
    }

    # Split the output on end lines.
    my @first_split = split(/\n/, $after);

    # Loop through the array and remove the headers.
    foreach my $element (@first_split) {
        if($element =~ m/Name(\s*)WWPNs/) {
            delete @first_split[$i];
            $i++;
        } else {
            # Convert the comma delimited WS to single spaces, no comma.
            $element =~ s/,\s/ /;
            # Convert spaces > 3 to just a single space
            $element =~ s/\s*/ /;
            # Get rid of leading spaces in the string
            $element =~ s/^\s*//;
            # Now split the string on those spaces
            my @values_array = &quotewords('\s+', 0, $element);
            # Get the size of the values array
            my $values_array_size = @values_array;
            $values_array_size--;
            # Now go loopy and push shit around.
            for($i=0;$i<=$values_array_size;$i++)
            {
                if($i != 0) {
                    push @{ $ap_hash{$values_array[0]} }, $values_array[$i];
                }
            }
        }
    }
    return($GOOD_STATUS, \%ap_hash);
}


## @name ::storage_array_show_config($;$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Executes the storage-array-show-config and returns all the relevent data
## in an array of hashes along with $GOOD_STATUS on success.
## Note there are cases where we don't have a lun group, hence you may have
## a line item not preceeded by 'Group n'. For these scenarios, we'll add
## 'item n' as the value to the group key. The keys for each hash will be
## group, lun_count, name, atp, switch_port, & initiator. Optionally takes
## in $opt which will be appended to the 'storage array show-config'
## base command. On error, we'll return $ERR_STATUS and return an array ref
## which will contain error logs.
##
## @synopsis
## ($status, $ref) = VSeries::Clustered::filer_convenience::storage_array_show_config($server);
##
## Sample CLI output:
## LUN Group        Array Name       Array Target Ports       Switch Port  Initiator
## Group 0 (6 LUNS) IBM_1722_1       20:04:00:a0:b8:21:a9:f9  vnci9124s52:1-3     0a
##                                   20:05:00:a0:b8:21:a9:f9  vnbr3800s1:2        0c
## Group 1 (4 LUNS) IBM_1722_1       20:05:00:a0:b8:21:a9:fa  vnbr3800s1:3        0d
##                                   20:04:00:a0:b8:21:a9:fa  vnci9124s52:1-4     0b
##         (7 LUNS) EMC_SYMMETRIX_1  50:06:04:82:cb:1b:ce:3d  vnbr3800s46:2       0c
##
## A direct attached config would look like this:
## Group 1 (4 LUNS) IBM_1722_1       20:05:00:a0:b8:21:a9:fa  N/A                 0d
##                                   20:04:00:a0:b8:21:a9:fa  N/A                 0b
## @arg $server Required: Server object to communicate with.
## @arg $opt    Optional: A string of options to append to the 'storage array
##                        show-config' base command.
##
sub storage_array_show_config($;$) {
    my $server   = shift;
    my $opt      = shift;
    my $cmd      = "storage array show-config";
    my @aoh      = (); # Array of Hashes that will store our config output.
    my $after    = ""; # Init the value that stores our Tharn::say() output.
    my @output   = (); # Used to hold the split up output from the cmd.
    my $count    = 0;  # Used as the main loop counter.
    my @err      = (); # Error array used when we return $ERR_STATUS.

    # Did the user specify any options, if so append to base.
    if (defined $opt) { $cmd .= " " . $opt; }

    # The following are most of the values for the keys in our return hash.
    my ($group, $lun_count, $name, $array_tp, $switch_port, $initiator) = ();

    # Send the command via node scope.
    $after = $server->dblade('cmd' => $cmd);    

    # Split $after on endlines and throw it in an array called @output.
    @output = split(/\n/, $after);

    # This will serve as our array index counter, which help us understand
    # where the the lun group data will be indexed.  Everytime we detect a
    # new lun group, we'll increment this count.
    my $i = -1;

    # Skip the header, as it won't be needed for anything.
    # Then begin to walk our output placing values where they belong.
    # There are 3 types of lines as seen in the header.
    foreach my $line (@output[1..$#output]) {
        
        # Clear any whitespace from the beginning of the line.
        $line =~ s/^\s+//;

        if ($line =~ /\d+ LUNS/) {
            # Increment our array index counter.
            $i++;

            # Increment the new record counter, records being new LUN groups
            # or items.
            $count += 1;

            # Parse up the current line if LUNS are preceeded by
            # Group, do it this way:
            # $1      $2        $3              $4                      $5              $6
            # Group 0 (11 LUNS) EMC_SYMMETRIX_1 50:06:04:82:cb:1b:ce:2d vnci9124s51:1-1 0a
            if ($line =~ /^(Group\s+\d+)\s+\((\d+)\s+LUNS\)\s+(\w+)\s+(\S{23})\s+(\S+)\s+(\S+)/) {
                $group          = $1;
                $lun_count      = $2;
                $name           = $3;
                $array_tp       = $4;
                $switch_port    = $5;
                $initiator      = $6;

                # Start filling the %temp hash with a group.
                $aoh[$i]->{'group'} = $group;
            }
            # This line would not have a Group number.
            # $1        $2              $3                      $4              $5
            # (11 LUNS) EMC_SYMMETRIX_1 50:06:04:82:cb:1b:ce:2d vnci9124s51:1-1 0a
            elsif ($line =~ /\((\d+)\s+LUNS\)\s+(\w+)\s+(\S{23})\s+(\S+)\s+(\S+)/) {
                $lun_count      = $1;
                $name           = $2;
                $array_tp       = $3;
                $switch_port    = $4;
                $initiator      = $5;

                # Start filling the %temp hash with an item.
                # Note the $count doesn't need to be sequential - just needs to
                # be different each time we find a line that contains LUNS and
                # doesn't contain Group - we've covered that.
                $aoh[$i]->{'group'} = "item $count";
            }

            # Continue to populate the %temp hash.
            $aoh[$i]->{'name'}      = $name;
            $aoh[$i]->{'lun_count'} = $lun_count;

            push(@{$aoh[$i]->{'atp'}},          $array_tp);
            push(@{$aoh[$i]->{'switch_port'}},  $switch_port);
            push(@{$aoh[$i]->{'initiator'}},    $initiator);
        }
        # Maybe this line just contains ATP's, SP's and Initiators.
        # $1                      $2            $3
        # 50:06:04:82:cb:1b:ce:3d vnbr3800s46:2 0c
        elsif ($line =~ /(\S{23})\s+(\S+)\s+(\S+)/) {
                $array_tp       = $1;
                $switch_port    = $2;
                $initiator      = $3;
                
                # Push these values onto the three common arrays found
                # in all lun groups.
                push(@{$aoh[$i]->{'atp'}},          $array_tp);
                push(@{$aoh[$i]->{'switch_port'}},  $switch_port);
                push(@{$aoh[$i]->{'initiator'}},    $initiator);                
        }
        else {
            # This line wouldn't match anything, so let's fail.
            push(@err, "filer_convenience::storage_array_show_config(): Failed to parse a line of output from $cmd.\nLine was $line");
            return($ERR_STATUS, \@err);
        }
    }

    # Before we return \@aoh we need to sort values within the lun groups
    # so compares will work even if the primary/secondary paths flipped.
    # This can create a scenario where the 'atp', 'switch_port' and 'initiator'
    # arrays per lun group may have different ordering.  i.e. during a compare
    # of group1 pretest vsersus posttest 'initiator' may be (0a, 0c) versus
    # (0c, 0a).  This compare really should not fail the test as the data is
    # still accurate, but the array profiles ONTAP code built the tables
    # differently after paths flipped.  As a result, we will preserve the
    # relationships of each initiator and its associated 'switch_port' and
    # 'atp' but we'll sort based on 'initiator' so that in theory our compares
    # will always be sorting data in the correct order.  Thank you ONTAP
    # for making life difficult :)
    foreach my $lg_ref (@aoh) {

        # The populate our temporary containers.
        my @array1 = @{$lg_ref->{'atp'}};
        my @array2 = @{$lg_ref->{'initiator'}};
        my @array3 = @{$lg_ref->{'switch_port'}};

        # Use "map" to setup the relationships
        my @relationships = map { [ $array1[$_], $array2[$_], $array3[$_] ] } (0..$#array1);

        # Sort and rearrange the data
        foreach my $r ( sort { $a->[1] cmp $b->[1] } @relationships ) {
            # Push the reordered data onto the bottom
            # of the "stack" and shift off the old data from the top of
            # the stack. You end up with the the orig arrays, reordered.
            push(@array1, @$r[0]);  # <-- atp's
            shift(@array1);

            push(@array2, @$r[1]);  # <-- initiators
            shift(@array2);

            push(@array3, @$r[2]);  # <-- switch ports
            shift(@array3);
        }

        # Replace old with new
        @{$lg_ref->{'atp'}}         = @array1;
        @{$lg_ref->{'initiator'}}   = @array2;
        @{$lg_ref->{'switch_port'}} = @array3;

    } #$lg_ref

    return($GOOD_STATUS, \@aoh);
}


## @name ::storage_array_purge_database($)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Purge the array profiles database which deletes
## /net/vnv3040f8a/vol/vol0/etc/arraymon forcing a rescan that will
## recreate the arraymon file. On success we return $GOOD_STATUS.
## On error, we'll return $ERR_STATUS and an error array ref.
##
## @synopsis
## ($status, $ref) = VSeries::Clustered::filer_convenience::storage_array_purge_database($server);
##
sub storage_array_purge_database($) {
    my $server  = shift;  # The filer object to shift in.
    my $after = '';     # Holds raw output returned from Tharn::say().
    my @err = ();      # Used as a list of errors for failure cases.

    # Send the command via node scope.
    $after = $server->dblade('cmd' => "storage array purge-database");    

    # On success we are told the database has been cleared.
    if ($after !~ /Array Profile database has been cleared/) {
        push(@err, "Error: filer_convenience::storage_array_purge_database(): May not have cleared the array profiles database located at \/vol\/vol0\/etc\/arraymon.\nReturned output: $after");
        return($ERR_STATUS, \@err);
    }
    return($GOOD_STATUS);
}


## @name ::storage_array_remove($$;$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Removes an array profile instance.  These instances are names of arrays
## that have the following format: vendor_catmodel_instance.  The concatenated
## model name comes from the ONTAP code trashing any remaining data that is
## present to the right of a dash '-'.  Instance exists because their may be
## more than one of the same array out on the back end.  Since the ONTAP
## code suggests doing so, an option is provided to purge the array profiles
## database once the array profile instance has been removed.  We may not
## always want to purge, so this is an optional parameter.  Lastly when pro-
## viding values for 'storage-array' commands please ensure they are within
## the bounds of those variables, i.e. array-name can't be greater than 28
## characters, otherwise ONTAP complains with 'usage:' statements.
## On success we return $GOOD_STATUS.  On error we return $ERR_STATUS
## along with an error array ref.
##
## Note: This works on BR-> user beware.
##
## @synopsis
## ($status, $ref) = VSeries::Clustered::filer_convenience::storage_array_remove($server, $array_name);
##
## @arg $server        Required: Server object to communicate with.
##
## @arg $array_name    Required: Array profile instance to remove.
##
## @arg $purge_flag    Optional: Set the flag to purge, undef by default.
##
sub storage_array_remove($$;$) {
    my ($server, $array_name, $purge_flag) = @_;
    my $after = '';    # Holds raw output from Tharn::say()
    my @err   = ();    # Holds the errors returned with $ERR_STATUS.
    my $stat  = '';    # Holds status from call to purge_database.
    my $e_ref = '';    # An error reference.

    # Take the third array slice from Tharn::say() and place it within
    # $after, it contains the raw output from our sent command.
    $after = $server->dblade('cmd' => "storage array remove $array_name");        

    # This is the positive case, all others are negative.
    if ($after =~ /Note: This array profile.*still present on the fabric/) {
        # If the purge flag was set, purge the database.
        if ($purge_flag) {
            ($stat, $e_ref) = storage_array_purge_database($server);
            if ($stat == $ERR_STATUS) {
                push(@err, "Error: filer_convenience When storage_array_remove() called storage_array_purge_database()\nMay not have cleared the array profiles database located at \/vol\/vol0\/etc\/arraymon.\nReturned output: $after");
                return($ERR_STATUS, \@err);                
            }

        }
    } else {
        push(@err, "Error: filer_convenience::storage_array_remove(): Could not remove the array profile $array_name.\n$after");
        return($ERR_STATUS, \@err);                
    }

    return($GOOD_STATUS);
}

## @name ::fcadmin_offline_online($$$$)
##
## @author Mark LaBonte (labonte@netapp.com)
##         James Silva  (silva@netapp.com) <<-- Just ported to 10-mode.
##
## @modified by Hari Prasad M (Hari.M@NetApp.com)
##
## @description
## Offlines or onlines the supplied adapter, and returns $GOOD_STATUS
## or $ERR_STATUS followed by the appropriate error ref on error.
##
## @synopsis
## ($status, $ref) = VSeries::Clustered::fcadmin_offline_online($server, "offline", "0a");
##
## @arg $server  Required: Server object to communicate through.
##
## @arg $state   Required: State to set adapter to (offline|online).
##
## @arg $adapter Required: Adapter to change the state of.
##
## @arg $node    Optional: Node to modify adapter on.
##
sub fcadmin_offline_online($$$;$) {
    my ($server, $state, $adapter, $node) = @_;
    my $cmd1 = "\"priv set -q diag;fcadmin $state $adapter\"";
    my $cmd2 = "\"priv set -q diag;fcadmin link_state\"";
    my $after = '';
    my $timeout_flag = '';
    my @err = ();

    if (!defined($node)) { $node = "local"; }

    # If we are setting the state to ONLINE, then sleep for 30 seconds
    # first to allow for port "debounce logic".
    if ($state eq "online") {sleep 30;}

    # Issue our command.
    $after = $server->system_node_run('command' => $cmd1, 'node'  => $node);

    # Make sure we weren't passed a non-existent adapter somehow.
    if ($after =~ /fcadmin offline: No such channel/) {
        push(@err, "Error: filer_convenience::fcadmin_offline_online(): $after");
        return($ERR_STATUS, \@err);                
    }

    # Set a timeout and check for the specified adapter state. FAIL if the
    # adapter has not changed to the specified state within the timeout period.
    my $timeout = time() + 120;
    while (time() < $timeout) {
        $after = $server->system_node_run('command' => $cmd2, 'node'  => $node);

        # Sleep briefly to prevent too much output in the log.
        sleep (3);
        if ($state eq "offline") {
            if ($after =~ /Link State for adapter channel $adapter: OFFLINE/) {
                $timeout_flag = 0;
                return ($GOOD_STATUS);
                last;
            } else {
                $timeout_flag = 1;
                next;
            }
        } elsif ($state eq "online") {
            if ($after =~ /Link State for adapter channel $adapter: UP/) {
                $timeout_flag = 0;

                # Sleep briefly to give devices on this adapter time to come
                # online.
                sleep (3);
                return ($GOOD_STATUS);
                last;
            } else {
                $timeout_flag = 1;
                next;
            }
        }
    }
    if ($timeout_flag == 1) {
        push(@err, "Error: filer_convenience::fcadmin_offline_online(): Failed to $state adapter $adapter.");
        return($ERR_STATUS, \@err);                
    }
}


## @name ::storage_load_balance($;$)
##
## @author Mathew Cattel (Mathew.Cattel@netapp.com)
##
## @description
## This method kicks off the storage load balance command via node scoping
## which does an ra_recast of path weight for load balancing in VSeries.  It
## will either verify the load balance completes or can be run with a no-verify
## flag.
##
## @arg $server  Required: Server object to communicate through.
##
## @arg no_verify Optional: Default is undef, if a scalar is passed in to the
##                          call the command will not verify if a load balance
##                          was done.
##
## @synopsis
## $status = VSeries::Clustered::filer_convenience::storage_load_balance($server);
##
sub storage_load_balance($;$) {

    # Shift in variables.
    my ($server, $no_verify) = @_;
    my @err_msgs = ();


    my $after = '';     # Holds raw output returned from Tharn::say().
    my $poll_time = "";
    my $date = "";

    # If we want to verify that the command actually kicked off a rebalance
    # then we'll get the time off the filer and store it for checking after we
    # issue the load balancing command.
    if(!defined($no_verify)){
        $date = $server->dblade('cmd' => "date");
        # we sleep for 2 seconds here because it's possible to issue the load
        # balance command in less than a second after and then the dates are
        # the same.  This sleep makes sure that we know for a fact if the
        # command did kick off a rebalance.
        sleep 2;
    }

    # Take the third array slice from Tharn::say() and place it within
    # $after, it contains the raw output from our sent command.
    $after = $server->dblade('cmd' => "storage load balance");

    # If we're to verify that the command actually kicked off a rebalance
    # we'll enter this block.
    if(!defined($no_verify)){

        my $last_balanced_date = "";

        # We're going to issue a storage load show here because at the end of
        # the output is the timestamp for the last time a load balance was done
        # The line looks like this:
        # I/O was last balanced at: Mon Feb  2 14:26:24 EST 2009
        my $output = $server->dblade('cmd' => "storage load show");

        # Split on newlines so each element in the array is a line of output.
        my @tmp_output = split(/\n/, $output);
        foreach my $line (@tmp_output){

            # Next until we find the line we're looking for.
            if($line !~ m/balanced at\:/){next;}

            # Split out the date to the right of the 'at:'
            $last_balanced_date = (split(/at\: /, $line))[1];
            last;
        }

        # This is where our error checking is performed, if $last_balanced_date
        # is before $date or if it's still the empty string (unexpected error)
        # the ret_val here for compare_dates will be something other than 2
        # and we'll error accordingly.  A return value of 2 means the second
        # date is later than the first date.
        my $ret_val = VSeries::utils::compare_dates($date, $last_balanced_date);

        if($ret_val != 2){
            # If we've entered this code path it's because we expected Ontap
            # to have issued a load balance and it did not.  The only tolerable
            # case where this can happen is if Ontap had aynchronously
            # issue a load balance that had completed within 10 seconds of
            # when we issued our command.  So we'll check for that case by
            # taking the timestamp we took right before issuing the command
            # and subtracting 10 seconds and seeing if that is before
            # the date Ontap is saying is when the last load balance
            # completed.  Any other cases of why a load balance
            # didn't complete when we expected one is still treated as a
            # failure case.


            # Split out the date string into tokens
            my @date_tokens = split(/ +/, $date);

            # These are the units that the date string will have.
            my ($seconds, $minutes, $hours, $days, $months, $years) = 0;

            # It's easier to assign numerical values to the months for
            # comparison.
            my %month_vals = ('JAN' => 1, 'FEB' => 2, 'MAR' => 3, 'APR' => 4, 'MAY' => 5, 'JUN' => 6,
                              'JUL' => 7, 'AUG' => 8, 'SEP' => 9, 'OCT' => 10, 'NOV' => 11, 'DEC' => 12);

            # Eventually at the end we'll want to re-map from month values to
            # strings for months again.
            my %reversed_months = reverse %month_vals;

            # For reference here's a date string and then the index into the
            # token array so that the code below is easier to follow.
            # Mon Feb  2  14:26:24 EST 2009
            # (0) (1) (2)  (3)     (4)  (5)
            #

            # Convert months into a numerical value for manipulation.
            $months = $month_vals{uc($date_tokens[1])};

            # Should be obvious based on sample above.
            $days = $date_tokens[2];
            $years = $date_tokens[5];

            # This teases out the values because they're all part of one token.
            $date_tokens[3] =~ m/(\d+):(\d+):(\d+)/;
            $hours = $1;
            $minutes = $2;
            $seconds = $3;


            # Ok so now we want to subtract 10 seconds from the seconds value,
            # this is easy unless seconds is under 10 seconds then we begin
            # to wrap and affect other units as we roll time backwards.

            # This is the easy case.
            if($seconds >= 10){$seconds = $seconds - 10;}

            # This is where seconds wraps but minutes doesn't since it's
            # greater than 1.
            elsif($minutes != 0){
                $minutes = $minutes - 1;
                $seconds = 60 - (10 - $seconds);
            }

            # This is where seconds and minutes wrap but hours does not
            # since it's greater than 0
            elsif($hours != 0){
                $hours = $hours - 1;
                # if minutes wraps it'll always be the 59th minute of the
                # previous hour.
                $minutes = 59;
                $seconds = 60 - (10 - $seconds);
            }
            # This is where seconds, minutes, and hours wrap but days does not
            # because it's greater than 1.
            elsif($days != 1){

                $days = $days - 1;
                # If days wrap the hours will always be the 23rd hour of the
                # previous day.
                $hours = 23;

                # if minutes wraps it'll always be the 59th minute of the
                # previous hour.
                $minutes = 59;
                $seconds = 60 - (10 - $seconds);
            }
            # This is where seconds, minutes, hours, days and months rollback
            # but not years since months was greater than 1 (Jan).  This does
            # get tricky because the last day of the preceding month isn't
            # always the same value, in some cases its 31 or 30 or 28, or
            # even 29 in a leap year.
            elsif($months != 1){
                $months = $months -1;

                # Ok so let's rollback the days looking at the month value
                # to determine what the days should be set as.
                # Feb, leap year
                if($months == 2 && $years % 4 == 0){$days = 29;}
                # Feb, non-leap year
                elsif($months == 2){$days = 28;}
                # Months that have 30 days.
                elsif($months == 4 || $months == 6 || $months == 9 || $months == 11){$days = 30;}
                # All other months have 31.
                else{$days = 31;}
                $hours = 23;
                $minutes = 59;
                $seconds = 60 - (10 - $seconds);
            }
            # The last case is where everything wraps including the year
            # so basically its Dec 31 23:59: and some seconds!
            else{ 
                $years = $years - 1; 
                $months = 12;
                $days = 31;
                $hours = 23;
                $minutes = 59;
                $seconds = 60 - (10 - $seconds);
            }

            # For formatting these units are padded with a leading zero if
            # they are single digits.
            if($seconds < 10){$seconds = "0" . $seconds;}
            if($minutes < 10){$minutes = "0" . $minutes;}
            if($hours < 10){$hours = "0" . $hours;}

            # Build the new date string to look like the old and recompare
            my $new_date = $date_tokens[0] . " " . $reversed_months{$months} . " " . $days . " " . $hours . ":" . $minutes . ":" . $seconds . " " . $date_tokens[4] . " " . $years;

            # Recompare the dates anything but a ret_val of 2 is a failure.
            # If it's a 2 that means by subracting 10 seconds from date
            # new_date was now "earlier" than the last balanced date and thus
            # within our tolerance.
            $ret_val = VSeries::utils::compare_dates($new_date, $last_balanced_date);		
		
            if($ret_val != 2){
                push(@err_msgs, "Error: filer_convenience::storage_load_balance: Unable to verify storage load balance completion.  Filer date: $date Last Load Balance Date: $last_balanced_date");
                return ($ERR_STATUS, \@err_msgs);
            } 
        }
    }


    return($GOOD_STATUS, \@err_msgs);
}

## @name ::storage_adapter_enable_disable($$$)
##
## @author Mark Bozen (mbozen@netapp.com)
##
## @description
## Sets a filer's hba adapter state to either enabled or disabled depending on
## specified input. Verifies the state change via fcadmin link_state
## On success we return $GOOD_STATUS.  On error we return $ERR_STATUS
## along with an error array ref.
##
## @synopsis
## ($status, $ref) = VSeries::Clustered::storage_adapter_enable_disable($server, $state, $port);
##
## @arg $server  Required: Server object to communicate through.
## @arg $state   Required: state of the hba port to change to 
##              (either 'enable' or 'disable')
## @arg $port    Required: filer hba port 
##
sub storage_adapter_enable_disable($$$) {
    my ($server, $state, $port) = @_;
    my $after = '';
    my @err = ();
    my $output = "";
    my $timeout = time() + 120;
	
    $after = $server->dblade('cmd' => "priv set -q diag; storage $state adapter $port");

    # Check for command errors
    if ($after =~ /usage:/) {
        push(@err, "Error: command execution error detected: $after");
        return($ERR_STATUS, \@err);                
    }

    # Polling loop to verify hba port has transitioned to expected state
    while (1) {
	
        my %hba_states = %{VSeries::Clustered::filer_convenience::fcadmin_link_state($server)};

        if ( ($state eq "enable" && $hba_states{$port}{'state'} eq "UP") ||
            ($state eq "disable" && $hba_states{$port}{'state'} eq "OFFLINE") ) {
            # Adapter is at expected state
            last;
        } elsif ( time() > $timeout ) {
            # hba never transitioned to expected state, so abort
            push(@err, "Error: timeout waiting for hba port $port to transition to $state, current state is $hba_states{$port}{'state'}");
            return($ERR_STATUS, \@err);
        } else {
            # sleep and re-check port state
            sleep 5;
        }            
    }
	
    return $GOOD_STATUS;	
}			


## @name ::ems_event_search($%$;$$)
##
## @author Mark Bozen (mbozen@netapp.com)
##
## @description
## Runs event log show command to detect logging of specific EMS events.
## Takes in a hash of EMS event keys to look for and returns the hash 
## with updated values consisting of number of occurences of that event
## within the specificed time range
##
## @synopsis
## $ref = VSeries::Clustered::ems_event_search($server, $event_href, $start_time, $interval, $node);
##
## @arg $server  Required: Server object
## @arg $event_href   Required: hash ref where keys are the event strings that
##                              will be searched in event logger.
##                              key: event message (ie.  cf.fm.partner)
##                              value: 0 (will be updated by this method)
## @arg $start_time   Required: start time timestamp that we begin looking for
##                              logged events (format is the dblade date
##                              command: Tue Jun 23 18:25:31 GMT 2009)
## @arg $offset  Optional: offset added to $start_time to determine end time to
##                         search for for events.
##                         (default: infinite, all ems events found on or 
##                          after the $start_time are considered within 
##                          time range)
## @arg $node    Optional: If the node value is passed then the event_log_show
##                         will list the events on that particular node. Otherwise
##                         it will be set to local.
sub ems_event_search($%$;$$) {
    my ($server, $event_href, $start_time, $offset, $node) = @_;

    my %events = %$event_href;
    my @events_found = ();
    my $end_time = "";

    # If the node is not mentioned then the value will be set to 'local'.
    $node = "local" if ($node eq q{});
    
    #
    # As we are hitting the time synchronization issue consistently with
    # the event_log_show command, we are using the NACL code to perform
    # the synchronization and this fix is wrt the BURT 737486.
    #
    my @nodes = NACL::C::Node->find();
    NACL::STask::Cluster->sync_clocks(command_interface => $nodes[0]);
  
  
    # For date conversion
    my @month_vals = qw / filler Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec /;
					  
    # First let's calculate our end time to search for the EMS events
    # only if offset is defined
    if ( defined $offset ) {
        $end_time = VSeries::utils::date_offset($start_time, $offset);
    }

    # Strip out any leading whitespaces from $start_time string before we split
    $start_time =~ s/^ +//;

    # Extract the timezone in the $start_time. We'll need it later for
    # date comparison
    my $time_zone = (split (/ +/, $start_time))[4];

    # Find all occurences of specified events that have occured in the current
    # EMS log
    foreach my $event (sort keys %events) {

        eval{
            @events_found = $server->event_log_show('node' => $node, 'messagename' => $event);
        };
        # Do not fatally fail if no log entries were found
        if($@ && $@ !~ m/There are no entries matching your query/ ) {
            Tharn::logresult('FATAL', $@);
        }

        # We have gathered all the events found, now determine if they are
        # within the specified time range 
        foreach my $ref (@events_found) {
            my %event_occurences = %$ref;
			
            # Get the time of the event, strip out the timestamp and covert it
            # to unix date command output format
            # Sample output of event log show for time: "6/23/2009 21:00:00"
			
            my $occurence_day = (split (/ +/, $event_occurences{'time'}))[0];
            my $occurence_time = (split (/ +/, $event_occurences{'time'}))[1];

            # Remove leading " from $occurence_day
            $occurence_day =~ s/^"//;			
            # Remove trailing " from $occurence_time
            chop $occurence_time;			

            # Split occurence day into month, day, year
            my $month_num  = (split ("/", $occurence_day))[0];
            my $day = (split ("/", $occurence_day))[1];
            my $year = (split ("/", $occurence_day))[2];

            # Convert month number to equivalent name
            my $month = $month_vals[$month_num];
 
            # Assemble our date/time stamp to send to compare_dates which must
            # be in this format: Tue Jun 23 18:25:31 GMT 2009)
            # compare_dates() does not parse name of day but a field needs to 
            # be there for token positioning so we add a junk 'DAY' token to 
            # handle this
            my $event_time = "DAY " . "$month " . "$day " . "$occurence_time " . "$time_zone " . "$year";

            my $date_check_begin = 0;
            my $date_check_end= 0;
			
            # Now check whether the event timestamp falls within the range
            $date_check_begin = VSeries::utils::compare_dates($event_time, $start_time);
            if ( defined $offset ) {
                $date_check_end = VSeries::utils::compare_dates($event_time, $end_time);
            }

            # Check if EMS event timestamp matched or came after the 
            # $start_time
            if ($date_check_begin == 1 || $date_check_begin == 0) {
            
                # Check if the EMS event timestamp matched or came
                # before the $end_time. If no offset was defined then the event
                # is automatically considered within $end_time since we
                # count any events that occur after $start_time
                if ( (defined $offset && ($date_check_end == 2 || $date_check_end == 0)) ||
                    ! defined $offset ) {			
                    $events{$event}++;
                }
            }	
        }
    }
 
    return \%events;		
}




## @name ::system_health_check($$%$)
##
## @author Mathew Cattel (Mathew.Cattel@netapp.com)
##
## @description
## This routine is intented to be a general collection of checks of
## functionality that are commonly expected for a VSeries System to be
## considered in a nominal state.  The keys map to specific common
## commands and should any more be added to this method as test
## and Ontap functionality evolves they should be labelled in such a way that
## they are as clear as possible in terms of what the test checks for.
## These are passed in through the %checks_hash and the keys are the checks.
## The checks can optionally have values with them.  Each of these checks and
## their anticipated values if any should be spelled out here in the header.
##
## Checks:
## ha-sfo - true/false  Calls "storage failover show" and checks the "possible"
##                      attribute of the instance to check if storage failover
##                      is possible for the filer.
##
## ha-cluster - true/false  Calls "cluster ha show" and checks the "configured"
##                          attribute to check the cluster failover is possible
##                          for the filer.
##
## lun-path-check - integer - Calls "storage array lun show" and checks that
##                            the number of paths for each lun equals the
##                            number passed in.
##
## lun-misconfig - A flag that when present will call "storage array show
##                 errors" and return an error if any errors are reported.
##
## vol-state - online/offline - Calls "volume show" and checks the "state"
##                              attribute for each volume on the system.  If
##                              any volumes do not match the state passed in
##                              then ERR_STATUS is returned.
## aggr-state - online/offline - Calls "storage aggregate show" and checks the
##                               "state" attribute for each aggregate so that
##                               each aggregate is in the expected state, if
##                               not then returns ERR_STATUS.
##
## @synopsis
## %returned_hash = %{VSeries::Clustered::system_health_check($server, $node, $checks_hash_ref, $retry_time)};
##
## @arg $server  Required: Server object to communicate through.
## @arg $node    Required: Name of filer in cluster to check for, filters out
##                         cluster view commands down to a specific filer.
## @arg $checks_hash_ref Required: Hash of key/value pairs that are the checks
##                                 to be performed.
## @arg $retry_time Required: time in seconds for filer to retry if any check
##                            should fail so that there's tolerance in the
##                            call for potential minor transient issues.
##
##
sub system_health_check($$%$){

    # Take passed in parameters.
    my ($server, $node, $checks_hash_ref, $retry_time) = @_;

    my %checks_hash = %$checks_hash_ref;

    # Initialize return hash.
    my %return_hash = ();

    # If the caller passed in a small timeout bump it so that
    # every check will be run at least once.
    if($retry_time < 2){
        $retry_time = 2;
    }


    # Put the keys onto @ARGV so that we can use GetOps, flags without values
    # just get pushed on, flags with values get concatenated with an equal sign
    foreach my $key (keys (%checks_hash)){
        if($key =~ m/lun-misconfig/){
            push(@ARGV, $key);
            next;
        }
        push(@ARGV, $key . "=" . $checks_hash{$key});
    }


    # These are all false by default, '' evaluates to false.
    my($ha_sfo,   # See @arg's section for info on these.
       $ha_cluster,
       $lun_path_check,
       $lun_misconfig,
       $vol_state,
       $aggr_state,
       ) = '';

    # Grab our options
    GetOptions( "ha-sfo:s"                 => \$ha_sfo,
                "ha-cluster:s"             => \$ha_cluster,
                "lun-path-check:s"         => \$lun_path_check,
                "lun-misconfig"            => \$lun_misconfig,
                "vol-state:s"              => \$vol_state,
                "aggr-state:s"             => \$aggr_state,
                );



    # Set the execution server so that we're talking to the filer we want
    # to perform the checks on.
    $server->set_execution_server("-Server" => $node);

    # If the variable is defined perform the proper check.
    if( $lun_path_check ){
    
    
        # Take the defined value that was passed in for this check.
        my $num_expected_lun_paths = $checks_hash{"-lun-path-check"};

        # Initialize timeout value for polling.
        my $timeout = time() + $retry_time;

        
        # In the event we fail the test we'll retry until we exceed a timeout
        # passed in by the caller.
        while(time() <= $timeout){

            # Temp array of active hba's discovered
            my @active_hbas = ();
        
            # Determine the total number of active hbas discovered. 
            # Each @sasc_arr element represents an active hba
            my @sasc_arr = @{$server->storage_array_config_show("node" => $node)};
            foreach my $sasc_hash_ref (@sasc_arr) {
                my %sasc = %$sasc_hash_ref;

                my $initiator = $sasc{'initiator'};
                # Populate array with active hbas discovered
                push @active_hbas, $initiator;
            }

            if ( scalar @active_hbas == $num_expected_lun_paths ) {
                $return_hash{"lun-path-check"} = $GOOD_STATUS;
                last;
            } else {
                $return_hash{"lun-path-check"} = $ERR_STATUS;
                sleep 10;
            } 
        }                   
    }

    # If the variable is defined perform the proper check.
    if($ha_sfo){

        # Initialize timeout value for polling.
        my $timeout = time() + $retry_time;

        # In the event we fail the test we'll retry until we exceed a timeout
        # passed in by the caller.
        while(time() <= $timeout){


            my $cmd_output = $server->storage_failover_show("node" => $node);

            # Take the defined value that was passed in for this check.
            my $ha_status = $checks_hash{"-ha-sfo"};

            my @output_array = @$cmd_output;

            # We assume the first element of array of hashes, since there should
            # only be one is the hash we want to check.
            my %output_hash = %{$output_array[0]};

            # If the passed in value matches the command output return GOOD_STATUS
            # otherwise return ERR_STATUS for this check.
            if($ha_status eq $output_hash{"possible"}){
                $return_hash{"ha-sfo"} = $GOOD_STATUS;
                last;
            }
            else{
                $return_hash{"ha-sfo"} = $ERR_STATUS;
                sleep(10);
            }
        }
    }

    # If the variable is defined perform the proper check.
    if($ha_cluster){

        # Initialize timeout value for polling.
        my $timeout = time() + $retry_time;

        # In the event we fail the test we'll retry until we exceed a timeout
        # passed in by the caller.
        while(time() <= $timeout){

            my $cmd_output = $server->cluster_ha_show();

            # Take the defined value that was passed in for this check.
            my $cluster_status = $checks_hash{"-ha-cluster"};


            my @output_array = @$cmd_output;

            # We assume the first element of array of hashes, since there should
            # only be one is the hash we want to check.
            my %output_hash = %{$output_array[0]};

            # If the passed in value matches the command output return GOOD_STATUS
            # otherwise return ERR_STATUS for this check.
            if($cluster_status eq $output_hash{'configured'}){
                $return_hash{"ha-cluster"} = $GOOD_STATUS;
                last;
            }
            else{
                $return_hash{"ha-cluster"} = $ERR_STATUS;
                sleep(10);
            }
        }
    }

    # If the variable is defined perform the proper check.
    if($vol_state){

        # Initialize timeout value for polling.
        my $timeout = time() + $retry_time;


        # In the event we fail the test we'll retry until we exceed a timeout
        # passed in by the caller.
        while(time() <= $timeout){

            my $cmd_output = $server->volume_show("node" => $node);

            # Take the defined value that was passed in for this check.
            my $vol_status = $checks_hash{'-vol-state'};

            my @output_array = @$cmd_output;


            # For this check we assume all volumes will pass until a mismatch,
            # then we'd set this value to ERR_STATUS and return.
            $return_hash{"vol-state"} = $GOOD_STATUS;

            # Loop through each volume checking it's state against the passed
            # in value for this check.
            foreach my $hash_ref (@output_array){
                my %vol_hash = %$hash_ref;
                if($vol_hash{'state'} ne $vol_status){
                    $return_hash{"vol-state"} = $ERR_STATUS;
                    last;
                }
            }

            # If we pass the check then break out of the loop, otherwise
            # sleep and retry if we haven't exceeded the timeout.
            if($return_hash{"vol-state"} == $GOOD_STATUS){ last;}
            else{ sleep 10;}

        }
    }

    # If the variable is defined perform the proper check.
    if($aggr_state){

        # Initialize timeout value for polling.
        my $timeout = time() + $retry_time;

        # In the event we fail the test we'll retry until we exceed a timeout
        # passed in by the caller.
        while(time() <= $timeout){

            my $cmd_output = $server->storage_aggregate_show("node" => $node);

            # Take the defined value that was passed in for this check.
            my $aggr_status = $checks_hash{'-aggr-state'};

            my @output_array = @$cmd_output;

            # For this check we assume all aggregates will pass until a mismatch,
            # then we'd set this value to ERR_STATUS and return.
            $return_hash{"aggr-state"} = $GOOD_STATUS;


            # Loop through each aggregate checking its state against the passed
            # in value for this check.
            foreach my $hash_ref (@output_array){
                my %aggr_hash = %$hash_ref;
                if($aggr_hash{'state'} ne $aggr_status){
                    $return_hash{"aggr-state"} = $ERR_STATUS;
                    last;
                }
            }

            # If we pass the check then break out of the loop, otherwise
            # sleep and retry if we haven't exceeded the timeout.
            if($return_hash{"aggr-state"} == $GOOD_STATUS){ last;}
            else{ sleep 10;}
        }
    }

    # If the variable is defined perform the proper check.
    if($lun_misconfig){

        # Initialize timeout value for polling.
        my $timeout = time() + $retry_time;

        # In the event we fail the test we'll retry until we exceed a timeout
        # passed in by the caller.
        while(time() <= $timeout){


            # In the event that there are no errors then we need to catch on the
            # case where the table is empty and check that that is in fact the
            # error case, this is how Server.pm handles empty tables.
            eval{
                $server->storage_errors_show("node" => $node);
            };


            # Catch for errors in Server.pm and check for an empty table. If it
            # empty then return GOOD_STATUS otherwise return ERR_STATUS for
            # unanticipated errors.            
            if( $@ && $@ =~ m/(This table is currently empty|There are no entries matching your query)/i ){
                    $return_hash{"lun-misconfig"} = $GOOD_STATUS;
                    last;
            } else {
                $return_hash{"lun-misconfig"} = $ERR_STATUS;
                sleep(10);
            }
        }
    }

    return \%return_hash;

}

## @name ::cleanup_components($$)
##
## @author Mark LaBonte (labonte@netapp.com)
##
## @description
## Deletes the supplied components and returns a hash, each value being a
## pointer to an array of hash pointers representing the components which
## encountered errors during deletion. Also returns error status and error
## messages within the hash.
##
## Here's the expected command (CDEF) output to be passed in for each component:
##
## "aggregates": storage aggregate show
## "volumes"   : volume show
## "vservers"  : vserver show
## "disks"     : disk show
##
## Here's an example of what the passed in hash should look like (where each
## hash value is a pointer to an array of hash pointers, as returned by the
## relevant CDEF call):
##
## %config_to_delete = ("aggregates" => $delete_aggrs,
##                      "volumes"    => $delete_vols,
##                      "vservers"   => $delete_vservers,
##                      "disks"      => $delete_disks,
##                      );
##
##
## @synopsis
## %return_hash = VSeries::Clustered::filer_convenience::cleanup_components($Server_Object, \%config_to_delete);
## @arg $server          Required: Server object to communicate through.
## @arg $components_ref  Required: Hash pointer containing elements to delete.
##
##
sub cleanup_components($$) {

    my ($server, $components_ref) = @_;

    my %components = %$components_ref;
	# These logics define components to be deleted
    my $delete_aggrs = 0;       # Indicates if we delete aggrs
    my $delete_vols = 0;        # Indicates if we delete vols                                     
    my $delete_vservers = 0;    # Indicates if we delete vservers
    my $delete_disks = 0;       # Indicates if we unassign disks

    # Mark components to be deleted    
    if ( defined $components{'aggregates'} ) {
        $delete_aggrs = 1;
   	}
   	if ( defined $components{'volumes'} ) {
        $delete_vols = 1;
    }
    if ( defined $components{'vservers'} ) {
        $delete_vservers = 1;
    }
    if ( defined $components{'disks'} ) {
        $delete_disks = 1;
    }
 
    my %return_hash = ();
    
    my @error_messages = ();
    my @vserver_errors = ();
    my @volume_errors = ();
    my @aggregate_errors = ();
    my @disk_errors = ();
    
    my $rootvol_flag = 0;
    my @aggr_disk_errors_by_name = ();
    my $aggr_disk_error_flag = 0;
    my $aggr_recheck_vols_ref_array_ref = "";

    my @vservers_remaining = ();
    my @vserver_rootvols_by_name = ();
    my @vserver_rootvol_ref_array = ();
    my $vserver_recheck_rootvol_ref_array_ref = "";

    # Delete the passed in vservers if possible, otherwise store them for later
    # deletion or flag the error in the event of a failed deletion.
    my $vserver_ref_array_ref = "";
    my @vserver_ref_array = ();
            
    if ( $delete_vservers == 1){

	$vserver_ref_array_ref = $components{'vservers'};
	@vserver_ref_array = @$vserver_ref_array_ref;
	foreach my $current_vserver_ref (@vserver_ref_array) {
            my %current_vserver_hash = %$current_vserver_ref;
                
            # Just delete it if there's no root volume
	    if ($current_vserver_hash{'rootvolume'} eq "-") {
                eval {
        	    $server->vserver_delete('vserver' => $current_vserver_hash{'vserver'});
                };
                if($@){
                    $return_hash{"status"} = $ERR_STATUS;
                    push @vserver_errors, $current_vserver_ref;
                    push @error_messages, $@;
                }      
	    } else {
        	push @vserver_rootvols_by_name, $current_vserver_hash{'rootvolume'};
    	           
	        # Store the vserver hash for later deletion.
                push @vservers_remaining, $current_vserver_ref;
            }
        }
    }
    
    # Delete the passed in volumes which are not vserver rootvols, or flag the 
    # error in the event of a failed deletion.
    my $vol_ref_array_ref = "";
    my @vol_ref_array = ();
    
    if( $delete_vols == 1){

        $vol_ref_array_ref = $components{'volumes'};
        @vol_ref_array = @$vol_ref_array_ref;

    	foreach my $current_vol_ref (@vol_ref_array) {		
            my %current_vol_hash = %$current_vol_ref;
            $rootvol_flag = 0;

            # Delete the volumes which are not rootvols
            foreach my $vserver_rootvol (@vserver_rootvols_by_name) {           
            	if ($vserver_rootvol eq $current_vol_hash{'volume'}) {
                	 push @vserver_rootvol_ref_array, $current_vol_ref;
                 	$rootvol_flag = 1;
                 	last;
            	}
            }

            if ($rootvol_flag == 0) {
            	if ($current_vol_hash{'junction-path'} ne "-") {
                    eval {
                    	$server->volume_unmount('vserver' => $current_vol_hash{'vserver'}, 'volume' => $current_vol_hash{'volume'});
                    };
                    if($@){
                    	$return_hash{"status"} = $ERR_STATUS;
                    	push @volume_errors, $current_vol_ref;
                    	push @error_messages, $@;
                    }
            	}
            	eval {
                    $server->volume_modify('volume' => $current_vol_hash{'volume'}, 'vserver' => $current_vol_hash{'vserver'}, 'state' => "offline");
                    $server->volume_delete('volume' => $current_vol_hash{'volume'}, 'vserver' => $current_vol_hash{'vserver'});
            	};
            	if($@){
                    $return_hash{"status"} = $ERR_STATUS;
                    push @volume_errors, $current_vol_ref;
                    push @error_messages, $@;
            	}
            }   
    	}
    }
    
    # Delete the vserver rootvols, or flag the error in the event of a failed
    # deletion.
    if ( $delete_vservers == 1){
            
    	foreach my $current_vol_ref (@vserver_rootvol_ref_array) {
            my %current_vol_hash = %$current_vol_ref;
	    eval {
    	        $server->volume_modify('volume' => $current_vol_hash{'volume'}, 'vserver' => $current_vol_hash{'vserver'}, 'state' => "offline");
        	$server->volume_delete('volume' => $current_vol_hash{'volume'}, 'vserver' => $current_vol_hash{'vserver'});
	    };
    	    if($@){
        	$return_hash{"status"} = $ERR_STATUS;
            	push @error_messages, $@;
	        push @volume_errors, $current_vol_ref;
    	    }
	}
      
    	# Now delete the remaining vservers, or flag the error in the event of a
	# failed deletion.
    	foreach my $vserver_remaining_ref (@vservers_remaining) {
            my %vserver_remaining_hash = %$vserver_remaining_ref;

	    # First re-check to see if any vserver rootvols still exist. If so, let's
    	    # not bother trying to delete the corresponding vserver - just place it
            # on the return hash instead.
	    eval {
        	$vserver_recheck_rootvol_ref_array_ref = $server->vserver_show('vserver' => $vserver_remaining_hash{'vserver'});
    	    };
	    if($@){
            	$return_hash{"status"} = $ERR_STATUS;
        	push @vserver_errors, $vserver_remaining_ref;
    	        push @error_messages, $@;
	    }
        
            my @vserver_recheck_rootvol_ref_array = @$vserver_recheck_rootvol_ref_array_ref;
    	    my $rechecked_vserver_ref = $vserver_recheck_rootvol_ref_array[0];
	    my %vserver_recheck_rootvol_hash = %$rechecked_vserver_ref;
        
            if ($vserver_recheck_rootvol_hash{'rootvolume'} ne "-") { 
    	        $return_hash{"status"} = $ERR_STATUS;
	        push @error_messages, "The vserver $vserver_recheck_rootvol_hash{'vserver'} cannot be deleted because its root volume, $vserver_recheck_rootvol_hash{'rootvolume'}, still exists.";
                push @vserver_errors, $vserver_recheck_rootvol_ref_array_ref;
    	    } else {
	        eval {
          	    $server->vserver_delete('vserver' => $vserver_recheck_rootvol_hash{'vserver'});
                };
    	        if($@){
	            $return_hash{"status"} = $ERR_STATUS;
                    push @vserver_errors, $vserver_recheck_rootvol_ref_array_ref;
            	    push @error_messages, $@;
        	}
    	    }
	}
    }
    
    # Delete the passed in aggregates, or flag the error in the event of a
    # failed deletion.
	my $aggr_ref_array_ref = "";
	my @aggr_ref_array = ();

	if( $delete_aggrs == 1 ) {       
            
	    $aggr_ref_array_ref = $components{'aggregates'};
            @aggr_ref_array = @$aggr_ref_array_ref;
    
	    foreach my $current_aggr_ref (@aggr_ref_array) {
                my %current_aggr_hash = %$current_aggr_ref;
        
                # Re-check the volume count, since we may have deleted volumes
                # previously.
                eval {
                    $aggr_recheck_vols_ref_array_ref = $server->storage_aggregate_show('aggregate' => $current_aggr_hash{'aggregate'});
                };
                if($@){
                    $return_hash{"status"} = $ERR_STATUS;
                    push @aggregate_errors, $current_aggr_ref;
                    push @error_messages, $@;
                }
        
                my @aggr_recheck_vols_ref_array = @$aggr_recheck_vols_ref_array_ref;
                my $rechecked_aggr_ref = $aggr_recheck_vols_ref_array[0];
                my %aggr_recheck_vols_hash = %$rechecked_aggr_ref;
        
                # Attempt to delete the aggregate if no volumes exist. If volumes still
                # exist, flag an error but don't bother trying to delete the aggregate. 
                # Flag an error in the event of a failed deletion.
                if ($aggr_recheck_vols_hash{'volcount'} != 0) {
                    $return_hash{"status"} = $ERR_STATUS;
                    push @error_messages, "The aggregate $aggr_recheck_vols_hash{'aggregate'} could not be deleted because one or more volumes still exist.";
                    push @aggregate_errors, $aggr_recheck_vols_ref_array_ref;
                    push @aggr_disk_errors_by_name, $aggr_recheck_vols_hash{'aggregate'};
                } else {
	        eval {
            	    $server->storage_aggregate_delete('aggregate' => $aggr_recheck_vols_hash{'aggregate'});
                };
    	        if($@){
	            $return_hash{"status"} = $ERR_STATUS;
                    push @error_messages, $@;
                    push @aggregate_errors, $aggr_recheck_vols_ref_array_ref;
            	    push @aggr_disk_errors_by_name, $aggr_recheck_vols_hash{'aggregate'};
        	}
    	    }
	}
    }
    
    # Remove ownership of the passed in disks.
    my $disk_ref_array_ref = ();
    my @disk_ref_array = ();
    
    if ($delete_disks == 1){
        $disk_ref_array_ref = $components{'disks'};
        @disk_ref_array = @$disk_ref_array_ref;
        foreach my $current_disk_ref (@disk_ref_array) {
    	    my %current_disk_hash = %$current_disk_ref;
            $aggr_disk_error_flag = 0;

	    # First check to see if any aggregate delete failures occured. If so,
    	    # let's not bother trying to unassign the corresponding disks - just
            # it on the return hash instead.
	    foreach my $aggr_disk_error_by_name (@aggr_disk_errors_by_name) {
    	        if ($aggr_disk_error_by_name eq $current_disk_hash{'aggregate'}) {
        	    $return_hash{"status"} = $ERR_STATUS;
            	    push @error_messages, "The disk $current_disk_hash{'disk'} could not be unassigned because its aggregate, $current_disk_hash{'aggregate'}, still exists.";
                    push @disk_errors, $current_disk_ref;
	            $aggr_disk_error_flag = 1;
    	        }
            }
	    if ($aggr_disk_error_flag == 0) {
    	        eval {
        	    $server->storage_disk_removeowner('disk' => $current_disk_hash{'disk'});
            	};
	        if($@){
    	            $return_hash{"status"} = $ERR_STATUS;
        	    push @disk_errors, $current_disk_ref;
            	    push @error_messages, $@;
	        }
    	    }
    	}
    }
    
    # Build the return hash.
    $return_hash{"error_messages"} = \@error_messages;
    $return_hash{"vserver_errors"} = \@vserver_errors;
    $return_hash{"volume_errors"} = \@volume_errors;
    $return_hash{"aggregate_errors"} = \@aggregate_errors;
    $return_hash{"disk_errors"} = \@disk_errors;
    
    return (\%return_hash);
    
}

## @name ::scsi_fitp_block($$;$$$$$$)
##
## @author Mathew Cattel (Mathew.Cattel@netapp.com)
##         Mark Bozen (Mark.Bozen@netapp.com)
##
## @description
## Wrapper for scsi fitp block command
##
## @synopsis
## $status = VSeries::Clustered:scsi_fitp_block($server, $node, $first_range, $initiator, $wwpn, $change_inventory, $persist, $defer);
##
## @arg $server      Required: Server object to communicate through
## @arg $node        Required: filer hostname
## @arg $first_range Optional: lun or comma separated lun range to be blocked
## @arg $initiator   Optional: filer initiator associated with blocking lun
## @arg $wwpn        Optional: array target port associated with blocking lun
## @arg $change_inventory  Optional: -c change_inventory option
## @arg $persist     Optional: -n persist option
## @arg $defer       Optional: -d defer option
##
## NOTE : Some 'optional' command args are required for command to run
## successfully so the user must ensure the proper command args are 
## specified in the calling program's invocation
##
sub scsi_fitp_block($$;$$$$$$){
    my $server = shift;
    my $node = shift;
    @ARGV = @_;         # Populate @ARGV for getops usage.
    my ($first_range,
        $initiator,
        $wwpn,
        $change_inventory,
        $persist,
        $defer
        ) = "";


    # Grab our options
    GetOptions( "r:s"             => \$first_range, 
                "i:s"             => \$initiator,
                "p:s"             => \$wwpn,
                "c"               => \$change_inventory,    # Flag
                "n"               => \$persist,    # Flag
                "d"               => \$defer,      # Flag
              );

  
    # Add colons to form octets if they're missing from wwpn
    if ( $wwpn !~ m/:/ ) {
        $wwpn =~ s/(\w{2})/$1\:/g;
        chop $wwpn;
    }
    
  
    my $cmd = "\"priv set test; scsi fitp block " . "-r " . $first_range . " " . "-i " . $initiator . " " . "-p " . $wwpn . " ";

    if($change_inventory){
        $cmd .= "-c ";
    }
    if($persist){
        $cmd .= "-n ";
    }
    if($defer){
        $cmd .= "-d ";
    }

    # Add trailing " to command
    $cmd .= "\"";

    # Issue our fitp command
    $server->system_node_run('command' => $cmd, 'node' => $node);
 
    return $GOOD_STATUS;

}


## @name ::scsi_fitp_swap($$;$$$$$$$)
##
## @author Mathew Cattel (Mathew.Cattel@netapp.com)
##         Mark Bozen (Mark.Bozen@netapp.com)
##
## @description
## Wrapper for scsi fitp swap command
##
## @synopsis
## $status = VSeries::Clustered:scsi_fitp_swap($server, $node, $first_range, $secondary_range, $initiator, $wwpn, $change_inventory, $persist, $defer);
##
## @arg $server      Required: Server object to communicate through
## @arg $node        Required: filer hostname
## @arg $first_range Optional: lun or comma separated lun starting range to be
##                             swapped
## @arg $secondary_range Optional: lun or comma separated lun end range to be
##                                 swapped
## @arg $initiator   Optional: filer initiator associated with swapping lun
## @arg $wwpn        Optional: array target port associated with  lun
## @arg $change_inventory  Optional: -c change_inventory option
## @arg $persist     Optional: -n persist option
## @arg $defer       Optional: -d defer option
##
## NOTE : Some 'optional' command args are required for command to run
## successfully so the user must ensure the proper command args are 
## specified in the calling program's invocation
##
sub scsi_fitp_swap($$;$$$$$$$){
    my $server = shift;
    my $node = shift;
    @ARGV = @_;         # Populate @ARGV for getops usage.
    my ($first_range,
        $secondary_range,
        $initiator,
        $wwpn,
        $change_inventory,
        $persist,
        $defer
        ) = "";


    # Grab our options
    GetOptions( "r:s"             => \$first_range,
                "s:s"             => \$secondary_range,
                "i:s"             => \$initiator,
                "p:s"             => \$wwpn,
                "c"               => \$change_inventory,    # Flag
                "n"               => \$persist,    # Flag
                "d"               => \$defer,      # Flag
              );

    # Add colons to form octets if they're missing from wwpn
    if ( $wwpn !~ m/:/ ) {
        $wwpn =~ s/(\w{2})/$1\:/g;
        chop $wwpn;
    }
    

    my $cmd = "\"priv set test; scsi fitp swap " . "-r " . $first_range . " " . "-s " . $secondary_range . " " . "-i " . $initiator . " " . "-p " . $wwpn . " ";

    if($change_inventory){
        $cmd .= "-c ";
    }
    if($persist){
        $cmd .= "-n ";
    }
    if($defer){
        $cmd .= "-d ";
    }

    # Add trailing " to command
    $cmd .= "\"";

    # Issue our fitp command
    $server->system_node_run('command' => $cmd, 'node' => $node);
 
    return $GOOD_STATUS;
    
}

## @name ::scsi_fitp_clear($$;$$$$)
##
## @author Mathew Cattel (Mathew.Cattel@netapp.com)
##         Mark Bozen (Mark.Bozen@netapp.com)
##
## @description
## Wrapper for scsi fitp clear command
##
## @synopsis
## $status = VSeries::Clustered:scsi_fitp_clear($server, $node, $initiator, $wwpn, $change_inventory, $defer);
##
## @arg $server      Required: Server object to communicate through
## @arg $node        Required: filer hostname
## @arg $initiator   Optional: filer initiator to clear events on
## @arg $wwpn        Optional: array target port to clear events on
## @arg $change_inventory  Optional: -c change_inventory option
## @arg $defer       Optional: -d defer option
##
## NOTE: if no optional params are specified. All fitp events will be cleared
##
sub scsi_fitp_clear($$;$$$$){
    my $server = shift;
    my $node = shift;
    @ARGV = @_;         # Populate @ARGV for getops usage.
    my ($initiator,
        $wwpn,
        $change_inventory,
        $defer
        ) = "";


    # Grab our options
    GetOptions( "i:s"             => \$initiator,
                "p:s"             => \$wwpn,
                "c"               => \$change_inventory,    # Flag
                "d"               => \$defer,      # Flag
              );


    # Add colons to form octets if they're missing from wwpn
    if ( $wwpn !~ m/:/ ) {
        $wwpn =~ s/(\w{2})/$1\:/g;
        chop $wwpn;
    }
    

    my $cmd = "\"priv set test; scsi fitp clear ";

    if($initiator ne ""){
        $cmd .= "-i " . $initiator . " ";
    }

    if($wwpn ne ""){
        $cmd .= "-p " . $wwpn . " ";
    }
    if($change_inventory){
        $cmd .= "-c ";
    }
    if($defer){
        $cmd .= "-d ";
    }

    # Add trailing " to command
    $cmd .= "\"";

    # Issue our fitp command
    $server->system_node_run('command' => $cmd, 'node' => $node);
 
    return $GOOD_STATUS;
}


## @name ::scsi_fitp_change($$;$$$$$$$)
##
## @author Mathew Cattel (Mathew.Cattel@netapp.com)
##         Mark Bozen (Mark.Bozen@netapp.com)
##
## @description
## Wrapper for scsi fitp swap command
##
## @synopsis
## $status = VSeries::Clustered:scsi_fitp_change($server, $node, $first_range, $secondary_range, $initiator, $wwpn, $change_inventory, $persist, $defer);
##
## @arg $server      Required: Server object to communicate through
## @arg $node        Required: filer hostname
## @arg $first_range Optional: lun or comma separated lun starting range to be
##                             changed
## @arg $secondary_range Optional: lun or comma separated lun end range to be
##                                 changed
## @arg $initiator   Optional: filer initiator associated with changing luns
## @arg $wwpn        Optional: array target port associated with changing luns
## @arg $change_inventory  Optional: -c change_inventory option
## @arg $persist     Optional: -n persist option
## @arg $defer       Optional: -d defer option
##
## NOTE : Some 'optional' command args are required for command to run
## successfully so the user must ensure the proper command args are 
## specified in the calling program's invocation
##
sub scsi_fitp_change($$;$$$$$$){
    my $server = shift;
    my $node = shift;
    @ARGV = @_;         # Populate @ARGV for getops usage.
    my ($first_range,
        $secondary_range,
        $initiator,
        $wwpn,
        $change_inventory,
        $persist,
        $defer
        ) = "";


    # Grab our options
    GetOptions( "r:s"             => \$first_range,
                "s:s"             => \$secondary_range,
                "i:s"             => \$initiator,
                "p:s"             => \$wwpn,
                "c"               => \$change_inventory,    # Flag
                "n"               => \$persist,    # Flag
                "d"               => \$defer,      # Flag
              );


    # Add colons to form octets if they're missing from wwpn
    if ( $wwpn !~ m/:/ ) {
        $wwpn =~ s/(\w{2})/$1\:/g;
        chop $wwpn;
    }
    

    my $cmd = "\"priv set test; scsi fitp change " . "-r " . $first_range . " " . "-s " . $secondary_range . " " . "-i " . $initiator . " " . "-p " . $wwpn . " ";

    if($change_inventory){
        $cmd .= "-c ";
    }
    if($persist){
        $cmd .= "-n ";
    }
    if($defer){
        $cmd .= "-d ";
    }

    # Add trailing " to command
    $cmd .= "\"";

    # Issue our fitp command
    $server->system_node_run('command' => $cmd, 'node' => $node);
 
    return $GOOD_STATUS;
    
}

## @name ::modify_volume_junction_path($$$$)
##
## @author Matt Cattel (Matthew.Cattel@netapp.com)
##
## @description
## This convenience routine takes in a server object, a volume name, and a 
## junction-path that the caller wants the volume to be mounted at.  It does
## a volume show to verify the volume exists, then unmounts the volume and 
## mounts the volume at the desired junction-path.  If any error is encountered
## the method returns two values, ERR_STATUS and the error message.  In the 
## positive case it returns GOOD_STATUS and the empty string.
##
## @synopsis
## ($status, $ref) = VSeries::Clustered::modify_volume_junction_path($server, $volume, $junction_path, $vserver);
##
## @arg $server  Required: Server object to communicate through.
## @arg $volume  Required: The name of the volume to modify 
## @arg $junction-path  Required: desired junction-path for volume 
## @arg $vserver Required: name of the vserver which has volume
## 
sub modify_volume_junction_path($$$$){
    my ($server, $volume, $junction_path, $vserver) = @_;
    
    eval{
        $server->volume_show('volume' => $volume);
        
        $server->volume_unmount('volume' => $volume, 'vserver' => $vserver);
        
        $server->volume_mount('volume' => $volume, 'junction-path' => $junction_path, 'vserver' => $vserver);
    
    };
    if($@){
        return($ERR_STATUS, $@);
    }
    
    return ($GOOD_STATUS, "");
    
}


## @name ::compare_storage_disk_show_path_equivalency($$)
##
## @author Matt Cattel (Matthew.Cattel@netapp.com)
##
## @description
## In the RollingRock release, the format to the storage disk show command
## was changed to support greater than two path topologies.  This means
## the notion of primary and secondary paths were replaced with a disk/lun
## having a primary and N number of non-primary paths.  This also made checking
## these attributes about a disk/LUN before and after a test event difficult.
## This method attempts to compre sets of paths to a LUN from two different
## invocations of 'storage disk show' for equivalency.  The attributes that
## define a path are 'initiator', 'initiator-side-switch-port', 
## 'target-side-switch-port', and 'lun'.  If the set of paths and each path
## is present in both output, then the method will return GOOD_STATUS
## otherwise it returns ERR_STATUS.
##
## @synopsis
## my $ret_val = VSeries::Clustered::filer_convenience::compare_storage_disk_show_path_equivalency(\%output1, \%output2);
##
## @arg $output1  Required: First invocation of storage disk show to compare.
## @arg $output2  Required: Second invocation of storage disk show to compare. 
##
sub compare_storage_disk_show_path_equivalency($$){
    my ($output1_ref, $output2_ref) = @_;
    my %output1 = %$output1_ref;
    my %output2 = %$output2_ref;
    
    # Path specific attributes for each piece of output.
    my @output1_initiator = ();
    my @output2_initiator = ();
    my @output1_initiator_side_switch_port = ();
    my @output2_initiator_side_switch_port = ();
    my @output1_target_side_switch_port = ();
    my @output2_target_side_switch_port = ();
    my @output1_target_wwpn = ();
    my @output2_target_wwpn = ();
    my @output1_lun = ();
    my @output2_lun = ();
    
    # For each of these attributes, cast them into arrays
    # or if they are single elements, push them onto the
    # array so that down below we're only handling arrays.
    if(ref($output1{'initiator'}) eq "ARRAY"){
        @output1_initiator = @{$output1{'initiator'}};
    }
    else{
    	push(@output1_initiator, $output1{'initiator'});
    }    

    if(ref($output2{'initiator'}) eq "ARRAY"){
        @output2_initiator = @{$output2{'initiator'}};
    }
    else{
    	push(@output2_initiator, $output2{'initiator'});
    }   
    
    if(ref($output1{'initiator-side-switch-port'}) eq "ARRAY"){
        @output1_initiator_side_switch_port = @{$output1{'initiator-side-switch-port'}};
    }
    else{
    	push(@output1_initiator_side_switch_port, $output1{'initiator-side-switch-port'});
    }    

    if(ref($output2{'initiator-side-switch-port'}) eq "ARRAY"){
        @output2_initiator_side_switch_port = @{$output2{'initiator-side-switch-port'}};
    }
    else{
    	push(@output2_initiator_side_switch_port, $output2{'initiator-side-switch-port'});
    }            

    if(ref($output1{'target-side-switch-port'}) eq "ARRAY"){
        @output1_target_side_switch_port = @{$output1{'target-side-switch-port'}};
    }
    else{
    	push(@output1_target_side_switch_port, $output1{'target-side-switch-port'});
    }    

    if(ref($output2{'target-side-switch-port'}) eq "ARRAY"){
        @output2_target_side_switch_port = @{$output2{'target-side-switch-port'}};
    }
    else{
    	push(@output2_target_side_switch_port, $output2{'target-side-switch-port'});
    }            

    if(ref($output1{'target-wwpn'}) eq "ARRAY"){
        @output1_target_wwpn = @{$output1{'target-wwpn'}};
    }
    else{
    	push(@output1_target_wwpn, $output1{'target-wwpn'});
    }    

    if(ref($output2{'target-wwpn'}) eq "ARRAY"){
        @output2_target_wwpn = @{$output2{'target-wwpn'}};
    }
    else{
    	push(@output2_target_wwpn, $output2{'target-wwpn'});
    } 

    if(ref($output1{'lun'}) eq "ARRAY"){
        @output1_lun = @{$output1{'lun'}};
    }
    else{
    	push(@output1_lun, $output1{'lun'});
    }    

    if(ref($output2{'lun'}) eq "ARRAY"){
        @output2_lun = @{$output2{'lun'}};
    }
    else{
    	push(@output2_lun, $output2{'lun'});
    }   
    
    # Since all these attributes are associated with the same set of paths
    # we check for consistency that each attribute has the same number
    # of elements in each set of output.
    if(scalar(@output1_initiator) != scalar(@output1_initiator_side_switch_port) || scalar(@output1_initiator) != scalar(@output1_target_side_switch_port)
        || scalar(@output1_initiator) != scalar(@output1_target_wwpn) || scalar(@output1_initiator) != scalar(@output1_lun)){ return $ERR_STATUS;}
        
    if(scalar(@output2_initiator) != scalar(@output2_initiator_side_switch_port) || scalar(@output2_initiator) != scalar(@output2_target_side_switch_port)
        || scalar(@output2_initiator) != scalar(@output2_target_wwpn) || scalar(@output2_initiator) != scalar(@output2_lun)){ return $ERR_STATUS;}
        
    # Using the associative property we can just compare a single
    # attribute between the two for equality.
    if(scalar(@output1_initiator) != scalar(@output2_initiator)){ return $ERR_STATUS;}
    
    # Now that we know all arrays are of the same size we can loop over one
    # and use that index for all attributes.
        
    for(my $i = 0; $i < scalar(@output1_initiator); $i++){
        # Flag to track if the path matched.
        my $path_matched = 0;
        
        # Now loop over the second set of output checking if each attribute
        # of a given path is equivalent, this way even if not in the same
        # order we can find the same path between sets of output.
        for(my $j = 0; $j < scalar(@output2_initiator); $j++){
            if($output1_initiator[$i] eq $output2_initiator[$j] && $output1_initiator_side_switch_port[$i] eq $output2_initiator_side_switch_port[$j]
                && $output1_target_side_switch_port[$i] eq $output2_target_side_switch_port[$j] && $output1_target_wwpn[$i] eq $output2_target_wwpn[$j]
                && $output1_lun[$i] eq $output2_lun[$j]){
                $path_matched = 1;
                last;
            }
        }
    
        # If we didn't find an equivalent path then return an error.
        if($path_matched == 0){return $ERR_STATUS;}

    }
            
    return $GOOD_STATUS;

}

## @name ::version_check($)
##
## @author Mark LaBonte (labonte@netapp.com)
##
## @description
## This method checks, in a functional manner, whether we are testing against
## BR or RR (or later). This is accomplished by issuing "storage disk show" and 
## checking for the 'TPGN' attribute, which exists in RR, but not BR.
## Returns a string ($BR_STR or $RR_STR, or $ERR_STR in the event of an error).
## Takes in a server object only.
##
## @synopsis
## my $ret_val = VSeries::Clustered::filer_convenience::version_check($Server_Object);
##
## @arg $server  Required: Server object to communicate through.
sub version_check($){
    my $server = shift;
    my $disks_ref = undef;
    my $output = "";
    my $version = "";
    
    eval{
        $disks_ref = $server->storage_disk_show("-Timeout" => 600, "output" => \$output);
    };
    
    if($@){
        return($@);
    }
    if(!defined($disks_ref)){
        return($ERR_STR . "disk ref was undefined in return call from server.pm  raw output is:\n$output");
    }
    
    my @disks = @$disks_ref;
    my $first_disk_ref = $disks[0];
    my %first_disk = %$first_disk_ref;
    
    if (defined($first_disk{'tpgn'})) {
        return ($RR_STR);
    } else {
        return ($BR_STR);
    }
}


## @name ::get_active_inactive_config($$$$)
##
## @author Mathew Cattel (Matthew.Cattel@netapp.com)
##
## @description
## This subroutine handles taking in the information about
## connections and paths to storage specified by the user
## and makes appropriate calls to storage array config show
## and get_seen_wwpns on the switches to determine how
## all the paths are constituted at the start of the test.
## It's easy to figure out what is connected, but it's harder
## to know what things aren't connected and this method puts
## all those things together in the %Paths data structure
## which is keyed by the filer hbas.  So this structure
## doesn't expect that you can have fanout where one filer
## hba is connected to multiple sets of storage in multiple
## zones.  It returns GOOD_STATUS or ERR_STATUS to the caller.
## If the status was GOOD_STATUS it also returns the paths data 
## structure otherwise it returns an error message as well.
##
## @arg $server  Required: Server object to communicate through.
##
## @arg $switch_obj_array_ref Required: reference to an array of switch objects
##                                      common in most test thpls.
##
## @arg $filer_hbas Required: Comma separated list of filer hbas
##                            to be considered in this topology.
##                            "0a,0b,0c,..."
##
## @arg $array_ports Required: Comma separated list of array target
##                             ports wwpns to be considered in the
##                             topology.
##
## @synopsis
## my $paths_ref = &get_active_inactive_config($server, \@switch_obj_array, $filer_hbas, $array_ports);
##
sub get_active_inactive_config($$$$) {

    my ($server, $switch_object_array_ref, $filer_hbas, $array_ports) = @_;
    
    my @switch_obj_array = @$switch_object_array_ref;  

    # Initialize data structure to be returned to caller.
     my %paths = ();
    
    my $sacs_ref = "";
    eval{
        $sacs_ref = $server->storage_array_config_show('node' => $server->get_execution_server());
    };
    if($@){
        return($ERR_STATUS, $@);
    }

    my @active_hbas = ();

    # Temp hashes that will help us fill up sub-hash for %paths
    my %array_target_port = ();
    my %switch_names = ();


    # Get output of storage array config show and put elements into
    # temporary hashes using initiator as key. This will help us
    # populate the %path_hash
    foreach my $sacs_hash_ref (@{$sacs_ref}) {
        my %sacs = %$sacs_hash_ref;
        my $initiator = $sacs{'initiator'};
        # Populate temp array with active hbas
        push @active_hbas, $initiator;

        # Get array wwpn associated with initiator and put into a hash
        $array_target_port{$initiator} = $sacs{'target-wwpn'};

        # Get the switch name associated with initiator
        $switch_names{$initiator} = $sacs{'target-side-switch-port'};
    }

    my @all_hbas = split (',', $filer_hbas);

    # Define our hba name keys for %paths
    foreach my $hba (@all_hbas) {

        # Define new hash key for each initiator
        $paths{$hba} = 0;
        # Initialize sub-hash for %paths
        my %path_hash = ();
        # Flag to indicate active initiator
        my $active = 0;

        # Fill in the current state of the initiator by checking if
        # it is in temp array @active_hbas
        foreach my $hba_act (@active_hbas) {
            if ( $hba_act eq $hba ) {
                $path_hash{'state'} = "active";
                $active = 1;
                last;
            }
        }
        if ( ! $active ) {
            $path_hash{'state'} = "inactive";
        }

        # Fill in the array wwpn associated with active initiator from
        # temp hash
        if ( defined $array_target_port{"$hba"} ) {
            $path_hash{'wwpn'} = $array_target_port{"$hba"};
            if($path_hash{'wwpn'} !~ /:/){
                $path_hash{'wwpn'} =~ s/(\w{2})/$1\:/g;
                chop $path_hash{'wwpn'};
            }
        }

        # Fill switch name associated with active initiator from temp hash
        if ( defined $switch_names{"$hba"} ) {
            my $switch = $switch_names{"$hba"};

            # Strip out port number from switch name
            $switch =~ s/:(.*)$//;
            my $port = $1;
            $path_hash{'switch'} = $switch;
            $path_hash{'switch-storage-port'} = $port;
        }

        # Get our filer hba wwpn from fcadmin channels
        my $hba_wwpn_raw =  $server->system_node_run('node' => $server->get_execution_server(), 'command' => "\"priv set -q test; fcadmin channels $hba\"");

        my $hba_wwpn = (split(/\s+/, $hba_wwpn_raw))[0];

        $hba_wwpn =~ s/(\w{2})/$1\:/g;
        chop $hba_wwpn;

        # Fill in wwpn of the associated initiator
        $path_hash{'hba_wwpn'} = $hba_wwpn;

        # Attach populated %path_hash to %paths
        $paths{"$hba"} = \%path_hash;
    }
  
    my %return_hash = %{VSeries::Clustered::filer_convenience::fcadmin_link_state($server)};
    
    foreach my $hba (keys %return_hash){
        if(defined($paths{$hba})){
            $paths{$hba}{'switch-initiator-port'} = $return_hash{$hba}{'switch_port'};
        }
    }


    # Now loop through and based on what paths were seen as active
    # find the inactive paths and pair up the hba_wwpns and array
    # port WWNs this will serve as a useful reference in later
    # test cases when zoning needs to be modified.
    my @array_of_ports = split(/,/, $array_ports);
  
    foreach my $array_port (@array_of_ports){
        my $found_port = 0;
        my $found_switch = "";
        my @seen_wwpns = ();


        # If the target-wwpn is defined and matches the current 
        # array port than this wwn was already paired up on an
        # active path up above and there isn't much more 
        # needed to be done with it.
        foreach my $hba (keys %paths){
            if(defined($paths{$hba}{'wwpn'}) && $paths{$hba}{'wwpn'} eq $array_port){
                $found_port = 1;
                last;
            }
        }
  
        # If you get here it means the array port hasn't been matched
        # up yet. We'll loop through each switch object calling get
        # seen wwns and see if this array port is seen on that
        # switch.  If so then we find the inactive path connected
        # to this switch and fill in the array wwn and switch
        # value.
        if($found_port == 0){
            foreach my $switch_obj (@switch_obj_array){
                @seen_wwpns = @{$switch_obj->get_seen_wwns()};
                foreach my $seen_wwpn (@seen_wwpns){
                    if($seen_wwpn =~ m/^$array_port/i){
                        $found_switch = $switch_obj->get_name();
                        last;
                    }
                }
                if($found_switch ne ""){ last;}
            }

            # If this is not blank than we matched on the current
            # switch.  Then we'll loop and find the inactive
            # path to this switch and fill in the appropriate
            # values.
            if($found_switch ne ""){
                my $resolved_information = 0;
            
                foreach my $hba (keys %paths){
                    # Skip over hba's that are active or array
                    # ports that have already been defined                    
                    if($paths{$hba}{'state'} =~ m/^active/ ||
                        defined $paths{$hba}{'wwpn'}) {
                        next;
                    }
  
                    foreach my $wwpn (@seen_wwpns){
                        if($paths{$hba}{'hba_wwpn'} =~ m/^$wwpn/i){
                            $paths{$hba}{'wwpn'} = $array_port;
                            $paths{$hba}{'wwpn'} =~ tr/A-Z/a-z/;
                            $paths{$hba}{'switch'} = $found_switch;
                            $resolved_information = 1;
                            last;
                        }
                    }
                    
                    if($resolved_information == 1){ last;}
                }
            } # end if found_switch
        } # end if found_port
    } # end foreach array_port
    
    # Ok so lastly we need to know the switch initiator ports and 
    # target initiator ports by calling get_port_wwn_seen_on for
    # both the hba_wwpn and wwpn (array wwn) for each path 
    # connected to the given switch.
    foreach my $switch_obj (@switch_obj_array){
        foreach my $hba (keys %paths){
            if($paths{$hba}{'switch'} eq $switch_obj->get_name()){
                $paths{$hba}{'switch-initiator-port'} = $switch_obj->get_port_wwn_seen_on($paths{$hba}{'hba_wwpn'});
                $paths{$hba}{'switch-storage-port'} = $switch_obj->get_port_wwn_seen_on($paths{$hba}{'wwpn'});
            }
        }
    }  
  
    return ($GOOD_STATUS, \%paths);
}


## @name ::assign_luns_and_create_test_aggr($$$;$)
##
## @author Mathew Cattel (Matthew.Cattel@netapp.com)
##
## @description
## This subroutine walks the @test_lun array and assigns each of
## the luns seen there.  Once they're all assigned it creates the
## test aggregate comprising of those luns.  This is common in many
## test cases requiring IO to be run to the misconfigured luns.  It
## returns to the caller a status and message.
##
## @arg $server  Required: Server object to communicate through.
##
## @arg $test_luns_ref Required: Array of luns to create the test aggr with,
##                               this is expected to be in the form of storage
##                               disk show output.
##
## @arg $aggr_name Required: desired name for the test aggr to be created.
##
## @arg $chksum    Optional: Chksum type for assigning lun
##
## @synopsis
## my ($status, $message) = VSeries::Clustered::filer_convenience::assign_luns_and_create_test_aggr($server_object, \@test_luns, $aggr_name);
##
sub assign_luns_and_create_test_aggr($$$;$){

    my ($server, $test_luns_ref, $aggr_name, $chksum) = @_;
    my $chksum_flag = 1;
    if (!defined $chksum) {
        $chksum_flag = 0;
    }

    my @test_luns = @$test_luns_ref;

    my $disk_list = "";

    # This will hold an updated list of LUNs to be passed to the
    # aggregate create command should it become necessary to retry
    # due to changed LUN names.
    my $alt_disk_list = "";

    my @sn_array = ();
       
    foreach my $ref (@test_luns){
    
        my %disk_hash = %$ref;

        # These serial numbers will be needed in the event that aggregate
        # creation fails initially because one or more LUN names changed.
        push @sn_array, $disk_hash{'serialnumber'};

        if ($chksum_flag == 0) {
            eval{
                $server->storage_disk_modify('disk' => $disk_hash{'disk'}, 'owner' => $server->get_execution_server());
            };

            # If we got an error, make sure we got the explicit error we were
            # expecting otherwise the test is still a failure.
            if($@){
                return ($ERR_STATUS,"Unable to take ownership of LUN " . $disk_hash{'disk'} . " please triage..." . $@); 
            }
        }
        else {
            eval{
                $server->storage_disk_assign('disk' => $disk_hash{'disk'}, 'owner' => $server->get_execution_server(), 'checksum' => $chksum );
            };

            # If we got an error, make sure we got the explicit error we were
            # expecting otherwise the test is still a failure.
            if($@){
                return ($ERR_STATUS,"Unable to take ownership of LUN " . $disk_hash{'disk'} . " please triage..." . $@);
            }
        }
        
        # Add name of disk to disk_list attribute for hash creation in
        # the next step.
        $disk_list .= $disk_hash{'disk'} . ",";        

    }

    # Remove dangling comma
    chop($disk_list);

    eval{
        $server->storage_aggregate_create("-Timeout" => 600, 'aggregate' => $aggr_name, 'disklist' => $disk_list, 'owner' => $server->get_execution_server());
    };

    # If we got an error, make sure we got the explicit error we were
    # expecting otherwise the test is still a failure.
    if($@){
    
        my %job_hash = ();
        
        Tharn::logcomment("Here's the error thrown while trying to create $aggr_name. " . Dumper($@) );
    
        if($@ =~ m/Failed to get information on the job (\d+):/i){
            
            # We presume a job will match since we are pulling this
            # from an error string about a job failing and the number
            # is included in the message.  Plus we don't want to squash
            # what's in $@ from the aggr create.
            my @job_array = @{$server->job_history_show("id" => $1)};
            %job_hash = %{$job_array[0]};
        }

        # It's possible that one or more LUN names have changed since the LUN
        # assignment. Here we will check for this condition and if necessary
        # we'll loop through our SN array and store the current LUN names for 
        # those devices.
        if ($@ =~ /disk .*does not exist/i || $job_hash{'completion'} =~ /disk .*does not exist/i) {

            Tharn::logcomment("Retrying aggregate creation with current LUN names.");

            # Let's wait for things to settle a bit.
            sleep 30;

            foreach my $serial (@sn_array) {

                my $current_disk_array_ref = $server->storage_disk_show('serialnumber' => $serial);
                my $current_disk_hash_ref = @$current_disk_array_ref[0];
                my %current_disk_hash = %$current_disk_hash_ref;

                $alt_disk_list .= $current_disk_hash{'disk'} . ",";
            }

            chop($alt_disk_list);

            # Reissue the aggregate create command using the new disklist,
            # which should now contain the current LUN names.
            eval{
                $server->storage_aggregate_create("-Timeout" => 600, 'aggregate' => $aggr_name, 'disklist' => $alt_disk_list, 'owner' => $server->get_execution_server());
            };
        }
        eval{
            $server->storage_aggregate_show("aggregate" => $aggr_name);
        
        };
        if($@){
            return ($ERR_STATUS, "Unable to create aggregate $aggr_name please triage..." . $@);
        }
    }
    
    my $timeout = time() + 180;
    my @aggregate_status = ();
    my %aggr_hash = ();
    
    while(time() < $timeout){
        eval{
            @aggregate_status = @{$server->storage_aggregate_show("aggregate" => $aggr_name)};
        };
        if($@){
            return( $ERR_STATUS, "ERROR: During check for test aggr to come online aggr show failed.\n" . $@);
        }
        
        %aggr_hash = %{$aggregate_status[0]};
        
        if($aggr_hash{'state'} =~ m/online/i){
            last;
        }
        sleep(10);           
    }
    
    if($aggr_hash{'state'} !~ m/online/i){
        return($ERR_STATUS, "Test Aggregate $aggr_name did not come online in 180 seconds, aborting test please triage...");
    }   
    else{
        return($GOOD_STATUS, "Test Aggregate $aggr_name successfully created and online");
    }
    
}

## @name ::setup_io_components($$)
##
## @author Mathew Cattel (Matthew.Cattel@netapp.com)
##
## @description
## This method handles setting up network ip components necessary to
## run IO to a volume.  It presumes that the vserver and other components
## already exist.  These are pretty atomic commands either the creates
## succeed or the command will fail and server.pm will take exception.  
## It takes in a server object and a hash with keys being the required 
## parameters for all the values necessary.  Will return either GOOD_STATUS
## or ERR_STATUS to the caller and a message as the second value returned.
##
## @arg $server  Required: Server object to communicate through.
##
## @arg $options_hash_ref: Hash table of options containing kesy for all
##                         the required parameters.  Those are: vserver,
##                         lif, home-node, home-port, address, netmask and
##                         gateway.
##
## @synopsis
## my ($status, $message) = VSeries::Clustered::filer_convenience::setup_io_components($server_object, \%options);
##
sub setup_io_components($$){

    my ($server, $options_hash_ref) = @_;
    
    my %options_hash = %$options_hash_ref;
    
    if(!defined($options_hash{'vserver'}) || !defined($options_hash{'lif'}) ||
       !defined($options_hash{'home-node'}) || !defined($options_hash{'home-port'}) ||
       !defined($options_hash{'address'}) || !defined($options_hash{'netmask'})){
       
        return($ERR_STATUS, "Options hash has an undefined value for a required param, please triage...");
    }
    
    
    eval{
        $server->network_interface_create('vserver' => $options_hash{'vserver'}, 'lif' => $options_hash{'lif'}, 'role' => "data",
                                         'home-node' => $options_hash{'home-node'}, 'home-port' => $options_hash{'home-port'},
                                         'address' => $options_hash{'address'}, 'netmask' => $options_hash{'netmask'});

        if (defined($options_hash{'gateway'})) {
           my @output = @{$server->network_routing_groups_show('vserver' => $options_hash{'vserver'})};

           $server->network_routing_groups_route_create('vserver' => $options_hash{'vserver'}, 'routing-group' => $output[0]->{'routing-group'}, 'gateway' => $options_hash{'gateway'});
        }
                                         
        $server->vserver_export_policy_rule_create('vserver' => $options_hash{'vserver'}, 'policyname' => 'default', 'clientmatch' => "0.0.0.0/0", 'rorule' => 'any', 'rwrule' => 'any', 'allow-duid' => 'false', 'allow-dev' => 'true', 'superuser' => 'any', 'anon' => "0");
   
        $server->vserver_nfs_create('vserver' => $options_hash{'vserver'}, 'access' => 'true', 'v3' => 'enabled', 'udp' => 'enabled', 'tcp' => 'enabled');
    };
    if($@){
        return ($ERR_STATUS, $@);
    }
    else{
        return ($GOOD_STATUS, "Network Interface, Routing Groups, export policy, and nfs rule created successfully for VServer " . $options_hash{'vserver'}) if (defined($options_hash{'gateway'}));
        return ($GOOD_STATUS, "Network Interface, export policy, and nfs rule created successfully for VServer " . $options_hash{'vserver'});
    }

}

## @name ::delete_io_components($$)
##
## @author Mathew Cattel (Matthew.Cattel@netapp.com)
##
## @description
## Method to conveniently clean up the network interface and routing group
## so that vservers can be deleted successfully.  Returns either GOOD_STATUS
## or ERR_STATUS and a status message.
##
## @arg $server  Required: Server object to communicate through.
##
## @arg $options_hash_ref  Required: hash table of options containing the keys
##                                   for the 'vserver' the components are on 
##                                   and the 'lif' to be deleted.
##
## @synopsis
## my ($status, $message) = VSeries::Clustered::filer_convenience::delete_io_components($server_object, \%options);
##
sub delete_io_components($$){

    my ($server, $options_hash_ref) = @_;
    
    my %options_hash = %$options_hash_ref;
    
    if(!defined($options_hash{'vserver'}) || !defined($options_hash{'lif'})){
       
        return($ERR_STATUS, "Options hash has an undefined value for a required param, please triage...");
    }

    
    eval{
        $server->network_interface_delete('vserver' => $options_hash{'vserver'}, 'lif' => $options_hash{'lif'});
    };
    
    # Check if error was defined and if it was catch on the case where 
    # the error is anything besides no entries found or in some cases the
    # error string was blank.
    # This is the case for each eval block here and separated out because
    # if everything was in a common eval block and somehow one of these
    # pieces got deleted as part of a vserver delete then the whole block
    # would exit out so we handle each of these separately and can continue
    # if one of the deletes hits this condition.
    if(defined($@) && ($@ !~ m/There are no entries/i && $@ !~ m/\s*/)){
        Tharn::logcomment(Dumper($@));
        return ($ERR_STATUS, $@);
    }        
    
    eval{
        $server->network_routing_groups_route_delete('vserver' => $options_hash{'vserver'}, 'routing-group' => "*");
    };
    if(defined($@) && ($@ !~ m/There are no entries/i && $@ !~ m/\s*/)){
        Tharn::logcomment(Dumper($@));
        return ($ERR_STATUS, $@);
    }    
    eval{    
        $server->network_routing_groups_delete('vserver' => $options_hash{'vserver'}, 'routing-group' => "*");
    };
    if(defined($@) && ($@ !~ m/There are no entries/i && $@ !~ m/\s*/)){
        Tharn::logcomment(Dumper($@));
        return ($ERR_STATUS, $@);
    }
    
    
    return ($GOOD_STATUS, "Network Interface and Routing Group deleted successfully for VServer " . $options_hash{'vserver'});
    
}

## @name ::check_panic($$)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## Method to be used in subtests to look for PANICs in console logs.
## This method when called in a subtest will return log results with FATAL
## severity and return with an error status on finding a PANIC in the log files.
## In case a matching log file does not exist when the subroutine is called, it will
## still loop until the subtest is killed. This is to handle cases where a log
## file could be created at a later point.
## The results returned will not directly kill the parent script that calls this
## sub routine. But the FATAL results can be handled in an exit handling routine in the 
## parent script. Exit handling routines can be defined by using -onexit in the
## subtest call.
##
## @arg $logDir  Required: Path to log dir that has the console logs.
##
## @arg $fileTypes_ref  Required: A reference to an array that contains the
##                                list of log files to be checked.
##
## @synopsis
## my $status = VSeries::Clustered::filer_convenience::check_panic($logDir, \@fileTypes);
##
## my $subtest = Tharn::subtest(\&VSeries::Clustered::filer_convenience::check_panic, '-bg', 
##                               '-onexit', \&check_subtest_result, '--', $logDir, \@fileTypes);
##
##
sub check_panic ($$) {
    my $logDir = shift;
    my $fileTypes_ref = shift;

    my @fileTypes = @{$fileTypes_ref};
    my $file_prefix = join('|',@fileTypes);
    
    #Start an infinite while loop
    # Check for panic in the console logs every 15 seconds
    # If there is a panic string, then return with an error status.
    my $dh = undef;
    my $file = undef;
    my $fh = undef;
    my @logFile = ();
    my $result = undef;

    while (1) {
        #Open the log directory and read the files in it.
        if(!opendir($dh, $logDir)) {
            # As this subroutine is intended to be called only in sutests, the logresult
            # below will not kill the parent. The log result is only used to sends status of
            # of the subtest back to parent.
            Tharn::logresult('FATAL', "Unable to open dir $logDir");
            return $ERR_STATUS;
        }
        while(defined ($file = readdir $dh)) {
            # If the file is a console log, open it and check for PANIC string in it.
            # If the user passes @fileTypes = ('console','ems');
            # Then $file_prefix will be console|ems. So all console* and ems* files
            # will be checked for PANIC string.
            if ($file =~ /^$file_prefix/) {

                #Check for PANIC in the logfile and return a FATAL result if found.
                $result = system("grep PANIC $logDir/$file");
                if ($result == 0) {
                    closedir $dh;
                    # As this subroutine is intended to be called only in sutests, the logresult
                    # below will not kill the parent. The log result is only used to send status of
                    # of the subtest back to parent.
                    Tharn::logresult('FATAL', "PANIC found in logfile $file");
                    return $ERR_STATUS;
                }
            }
        }
        closedir $dh;
        # Sleep for 60 seconds and continue.
        sleep 60;
    }
}

## @name ::create_desired_topology($$$)
##
## @author Mathew Cattel (Matthew.Cattel@netapp.com)
##
## @description
## This is essentially a easy common routine that allows the caller to take
## the system from the start state, which should be a 2 path, 1 per switch
## configuration to any of the other required test topologies.  By looking into
## the %paths data structure and based on what the user passed in for a desired
## topology (four_path_straight_through, four_path_fan_in, four_path_fan_out, 
## and eight_path_chaos) will set up the proper zoning on each switch.    
## 
## Each of these configs will be set up as such.
##
## four_path_straight_through - in addition to the starting cfg zoning, will add
## a zone on each switch bringing up the initiator and array target port that
## have the "inactive" state.
##
## four_path_fan_in - in addition to the starting cfg zoning, will add a zone that
## pairs together the inactive initiator connected to the switch and that 
## active path's array port.  Creating two paths per switch that go into the 
## same array target port.
##
## four_path_fan_out - in addition to the starting cfg zoning, will add a zone 
## that pairs together the inactive array target port connected to the switch 
## and that active path's initiator port.  Creating two paths per switch that
## start with the same initiator and go to two separate array target ports.
## 
## eight_path_chaos - in addition to the starting cfg zoning, adds in the straight
## through, fan in, and fan out paths so that each switch has 4 paths where
## 2 initiators have paths to each of 2 array target ports.
##
##
## Returns to the caller either GOOD_STATUS or ERR_STATUS.  If the result was
## GOOD_STATUS then the second value is the modified config containing the cfg 
## and zoning created so that it can be passed to a cleanup routine later.  If
## the result is ERR_STATUS then the second result is an error message.
##
## @arg $desired_topolgy Required: string containing the desired topology 
##                                 Supported values already stated above.
##
## @arg $paths_ref Required: reference to the paths data structure returned
##                           by calling get_active_inactive_config.
##
## @arg $switch_obj_array_ref Required: reference to an array of switch objects
##                                      common in most test thpls.
##
##
## @synopsis
## my ($result, $err_msg_or_switch_configs_ref) = VSeries::Clustered::filer_convenience::create_desired_topology($$$);
##
sub create_desired_topology($$$){

    my ($desired_topology, $paths_ref, $switch_obj_array_ref) = @_;
    
    my %paths = %$paths_ref;
    my @switch_obj_array = @$switch_obj_array_ref;
    
    Tharn::logcomment("Here's desired topology: $desired_topology");

    # We want the name of the zones created to indicate which test case is
    # being run so set that according to if $owned is defined or not.
    my $zone_name = "";
    my $cfg_name = $desired_topology . "_cfg";

    # Keep track of the modified switch configs that we create for cleanup
    # later.
    my %modified_switch_cfgs = ();


    # Loop through the switch objects until we have the one with the
    # name that's the same as $switch_to_modify then creating the
    # zoning we want.  For this test case we want to bring up the
    # inactive path on this switch which is contained in the path
    # data structure with the key of $other_adapter.
    foreach my $switch_obj (@switch_obj_array) {

        # Determine active switch config
        my %active_cfg = %{$switch_obj->active_config_and_zone_get()};
        # Make a copy of the active config that we can start manipulating.
        my %modified_cfg = %active_cfg;
        
                
        # Now modify names of all the pre-existing zones so that we
        # make equivalent zones rather than reusing existing zones.
        # We also rename the overal config so do that here too.
        # We need a tmp_hash to help convert the key names.
        my %tmp_hash = ();
          
        foreach my $key (keys %modified_cfg){
            if($key =~ m/config_name/){
                $tmp_hash{$key} = $cfg_name;
            }
            else{
                $tmp_hash{$key . "_start"} = $modified_cfg{$key};
            }
        }
        
        # Now copy the tmp_hash back to modified_cfg since it's in the form
        # we wanted.
        %modified_cfg = %tmp_hash;

        # Check for desired topology, if so then simply find the inactive
        # paths and create zoning.
        if($desired_topology eq "four_path_straight_through"){
            
            $zone_name = "second_path_on_switch";
	        
            # Loop through until we have the right switch and the info for
            # the inactive path and put that zone in place.
            foreach my $adapter (keys %paths){
                if($paths{$adapter}{'switch'} ne $switch_obj->get_name()){ next;}
                if($paths{$adapter}{'state'} =~ m/^active/){ next;}


                # Pull together the proper elements for the new zone
                # we want to create and put that into the modified config.
                my @new_zone_arr = ();
                push @new_zone_arr, $paths{$adapter}{'wwpn'};
                push @new_zone_arr, $paths{$adapter}{'hba_wwpn'};

                $modified_cfg{$zone_name} = [@new_zone_arr];

                $switch_obj->zone_config_create_enable(\%modified_cfg);
                if ( $switch_obj->get_error_state() == $ERR_STATUS) {
                    return ($ERR_STATUS, "Error: Could not create or enable switch config:\n" . (Dumper($switch_obj->get_error_msg())));
                }
            
                $modified_switch_cfgs{$switch_obj->get_name()} = \%modified_cfg; 
            }
        }
        
        # Check for desired topology, if so then simply find the inactive
        # paths and create zoning pairing up the active initiator and
        # inactive array port.
        if($desired_topology eq "four_path_fan_in"){
        
            $zone_name = "fan_in_path_on_switch";
        
            # Will contain the new zone we want to create's info.
            my @new_zone_arr = ();

            # Ok now for each switch we need to pull the proper info together
            # to create the soft zones we want, we'll add this info to the
            # modified config structure and pass it into zone_config_create_enable.
            foreach my $adapter (keys %paths){

                # Look for the adapters that are connected to the current switch.
                if($paths{$adapter}{'switch'} ne $switch_obj->get_name()){ next;}

                # If we have the active path to this switch we need the array
                # wwpn for the zoning we want for this test case.  If we have
                # the passive path we want the filer hba info.
                if($paths{$adapter}{'state'} =~ m/^active/){
                    push @new_zone_arr, $paths{$adapter}{'wwpn'};
                }
                else{
                    push @new_zone_arr, $paths{$adapter}{'hba_wwpn'};
                }
            }

            # Add the zone to the overall config.
            $modified_cfg{$zone_name} = [@new_zone_arr];
        
            $switch_obj->zone_config_create_enable(\%modified_cfg);
            if ( $switch_obj->get_error_state() == $ERR_STATUS) {
                return( $ERR_STATUS, "Error: Could not create or enable switch config interim cfg:\n" . (Dumper($switch_obj->get_error_msg())));
            }
         
            $modified_switch_cfgs{$switch_obj->get_name()} = \%modified_cfg; 
        
        }        

        # Check for desired topology, if so then simply find the inactive
        # paths and create zoning pairing up the active array port and
        # inactive initiator.
        if($desired_topology eq "four_path_fan_out"){
        
            $zone_name = "fan_out_path_on_switch";
        
            # Will contain the new zone we want to create's info.
            my @new_zone_arr = ();

            # Ok now for each switch we need to pull the proper info together
            # to create the soft zones we want, we'll add this info to the
            # modified config structure and pass it into zone_config_create_enable.
            foreach my $adapter (keys %paths){

                # Look for the adapters that are connected to the current switch.
                if($paths{$adapter}{'switch'} ne $switch_obj->get_name()){ next;}

                # If we have the active path to this switch we need the array
                # wwpn for the zoning we want for this test case.  If we have
                # the passive path we want the filer hba info.
                if($paths{$adapter}{'state'} =~ m/^active/){
                    push @new_zone_arr, $paths{$adapter}{'hba_wwpn'};
                }
                else{
                    push @new_zone_arr, $paths{$adapter}{'wwpn'};
                }
            }

            # Add the zone to the overall config.
            $modified_cfg{$zone_name} = [@new_zone_arr];
            $switch_obj->zone_config_create_enable(\%modified_cfg);
            if ( $switch_obj->get_error_state() == $ERR_STATUS) {
                return($ERR_STATUS, "Error: Could not create or enable switch config interim cfg:\n" . (Dumper($switch_obj->get_error_msg())));
            }
         
            $modified_switch_cfgs{$switch_obj->get_name()} = \%modified_cfg; 
        
        }  
        
        # This case basically pulls the other 3 cases together.  Due to how 
        # zone config create enable only takes a single configuration in
        # this method resuses the zone creation code from each case rather
        # than calling this method from inside itself and implementing
        # each of the zones.  This also enables cleanup to be consistent
        # with the rest of the use cases here.
        if($desired_topology eq "eight_path_chaos"){

            # Create the straight through path.
            $zone_name = "second_path_on_switch";
	        
            # Loop through until we have the right switch and the info for
            # the inactive path and put that zone in place.
            foreach my $adapter (keys %paths){
                if($paths{$adapter}{'switch'} ne $switch_obj->get_name()){ next;}
                if($paths{$adapter}{'state'} =~ m/^active/){ next;}


                # Pull together the proper elements for the new zone
                # we want to create and put that into the modified config.
                my @new_zone_arr = ();
                push @new_zone_arr, $paths{$adapter}{'wwpn'};
                push @new_zone_arr, $paths{$adapter}{'hba_wwpn'};

                $modified_cfg{$zone_name} = [@new_zone_arr];
            }  
            
            # Now create the fan in path.
            $zone_name = "fan_in_path_on_switch";
        
            # Will contain the new zone we want to create's info.
            my @new_zone_arr = ();

            # Ok now for each switch we need to pull the proper info together
            # to create the soft zones we want, we'll add this info to the
            # modified config structure and pass it into zone_config_create_enable.
            foreach my $adapter (keys %paths){

                # Look for the adapters that are connected to the current switch.
                if($paths{$adapter}{'switch'} ne $switch_obj->get_name()){ next;}

                # If we have the active path to this switch we need the array
                # wwpn for the zoning we want for this test case.  If we have
                # the passive path we want the filer hba info.
                if($paths{$adapter}{'state'} =~ m/^active/){
                    push @new_zone_arr, $paths{$adapter}{'wwpn'};
                }
                else{
                    push @new_zone_arr, $paths{$adapter}{'hba_wwpn'};
                }
            }

            # Add the zone to the overall config.
            $modified_cfg{$zone_name} = [@new_zone_arr];            

            # Now create the fan out path.
            $zone_name = "fan_out_path_on_switch";
        
            # Will contain the new zone we want to create's info.
            @new_zone_arr = ();

            # Ok now for each switch we need to pull the proper info together
            # to create the soft zones we want, we'll add this info to the
            # modified config structure and pass it into zone_config_create_enable.
            foreach my $adapter (keys %paths){

                # Look for the adapters that are connected to the current switch.
                if($paths{$adapter}{'switch'} ne $switch_obj->get_name()){ next;}

                # If we have the active path to this switch we need the array
                # wwpn for the zoning we want for this test case.  If we have
                # the passive path we want the filer hba info.
                if($paths{$adapter}{'state'} =~ m/^active/){
                    push @new_zone_arr, $paths{$adapter}{'hba_wwpn'};
                }
                else{
                    push @new_zone_arr, $paths{$adapter}{'wwpn'};
                }
            }

            # Add the zone to the overall config.
            $modified_cfg{$zone_name} = [@new_zone_arr];
        
            $switch_obj->zone_config_create_enable(\%modified_cfg);
            if ( $switch_obj->get_error_state() == $ERR_STATUS) {
                return($ERR_STATUS, "Error: Could not create or enable switch config interim cfg:\n" . (Dumper($switch_obj->get_error_msg())));
            }
         
            $modified_switch_cfgs{$switch_obj->get_name()} = \%modified_cfg; 
                            
        }
        
    }    
    return ($GOOD_STATUS, \%modified_switch_cfgs);

}

## @name ::get_ontap_version($;$)
##
## @author Arvind Chandrasekaran
##
## @description
## Method to get the Ontap image versions available on the node.
##
## @arg $server        Required: Server object to communicate through.
##
## @arg $options_ref   Optional: reference to hash with options
##
## @synopsis
## my ($status, $output_ref) = VSeries::Clustered::filer_convenience::get_ontap_version($server_object, $options_ref);
##
sub get_ontap_version($;$){
    my ($server, $options_ref) = @_;
    my %options = ();
    my $output_ref = undef;
    if (defined $options_ref) {
        %options = %{$options_ref};
    }
    Tharn::logcomment("options". Dumper (%options));
    eval {
        $output_ref = $server->system_node_image_show(%options);
    };
    if ($@) {
        return ($ERR_STATUS, $@);
    }
    else {
        return ($GOOD_STATUS, $output_ref);
    }
}

## @name ::check_subtest_result($)
##
## @author Arvind Chandrasekaran (arvind.chandrasekaran@netapp.com)
##
## @description
## Checks the result of subtest used to look for PANIC in console logs.
##
## @arg $subtest Subtest object passed by default.
##
## @synopsis
## VSeries::Clustered::filer_convenience::check_subtest_result($subtest);
##
sub check_subtest_result($) {
    my ($subtest) = @_;
    my @result_msg = ();

    # Check the own_results() of the subtest and look for any FATAL messages returned by the subtest.
    if (scalar($subtest->own_results()) && (NATE::Result::worst($subtest->own_results())->severity() >= NATE::Result::Fail::severity())) {
        foreach my $rslt ($subtest->own_results()) {
            if (UNIVERSAL::isa($rslt, 'NATE::Result')) {
                push(@result_msg, $rslt->messages());
            }
            else{
                push(@result_msg, $rslt);
            }
        }
        # Fail the test with the appropriate message returned by the subtest.
        Tharn::logresult('FATAL', "Failures detected:\n".join("\n", @result_msg));
    }
}

## @name ::get_checksum($)
##
## @author Hariprasad M
##
## @description
## This sub routine gets the checksum details of all the luns seen.
##
## @arg $server        Required: Server object to communicate through.
##
## @synopsis
## my $output_ref = VSeries::Clustered::filer_convenience::get_checksum($server_object);
##
sub get_checksum($) {
    my $server = shift;
    my %checksum_hash = undef;

    foreach my $disk (@{ $server->storage_disk_show( ) }) {
        $checksum_hash{$disk->{'serial-number'}} = $disk->{'checksum'} unless $checksum_hash{$disk->{'serial-number'}};
    }

    return \%checksum_hash;
}

## @name ::connection_error_timeout_retry($$$$)
##
## @author Mark LaBonte
##
## @description
## This subroutine retries the establishment of a given connection type.
##
## Returns 
##
## @arg $server             Required: Server object to communicate through.
##
## @arg $connection_type    Required: Type of connection to attempt.
##
## @arg $retry_count        Required: Number of times to retry.
##
## @arg $sleep_time         Required: Seconds to sleep between retries.
##
## @synopsis
## my ($status, $error) = VSeries::Clustered::filer_convenience::connection_error_timeout_retry($server_object, $error);
##
sub connection_error_timeout_retry($$$$) {
    my ($server, $connection_type, $retry_count, $sleep_time) = @_;
    my $flag = 0;
    my $count = 0;

    while ($count < $retry_count) {
        eval {
            $server->handle_connections(open => [$connection_type]);
        };
        if (($@) && ($@ =~ m/Timed out/i)) {
            sleep ($sleep_time);
            $count++;
            next;
        } elsif ($@) {
            Tharn::logresult('FATAL', "$@");
        } else {
            $flag = 1;
            last; 
        }
    }
    
    if ($flag == 0) {
        my $error_string = "Failed to establish $connection_type after $retry_count retries.";
        return ($ERR_STATUS, $error_string);
    } else { return $GOOD_STATUS; }
}

## @name ::get_array_name_by_ports($)
##
## @author Mark LaBonte
##
## @description
## This sub routine gets the name of an array based on a list of array WWPNs.
##
## @arg $array_ports        Required: Comma-separated list of array WWPNs.
##
## @arg $sacs_out           Required: Reference to storage array config show
##                                    output.
##
## @synopsis
## my ($status, $info) = VSeries::Clustered::filer_convenience::get_array_name_by_ports($array_ports, $sacs_out);
##
sub get_array_name_by_ports($$) {
    my ($array_ports, $sacs_out) = @_;

    my $error_string = "";
    my $output = "";
    my $array_name = undef;
    my @array_ports = split(/,/, $array_ports);

    foreach my $array_port (@array_ports) {

        $array_port =~ s/://g;

        foreach my $output_hash_ref (@{$sacs_out}) {
    
            my $found_array_port = $output_hash_ref->{'target-wwpn'};
            my $found_array_name = $output_hash_ref->{'array-name'};
    
            if ($found_array_port ne $array_port) { next; }
    
            if (defined($array_name) && ($found_array_name ne $array_name)) {
    
                $error_string = "ERROR: All ARRAY_PORTS specified must be from the same storage array.";
                return ($ERR_STATUS, $error_string);
    
            }
            $array_name = $found_array_name;
        }

    }

    return ($GOOD_STATUS, $array_name);

}

1;  # Needed for *.pm's




