# -*- perl -*-
# $Id$
#
## @summary filer setup/cleanup library functions to be used by VSeries tests
##
## @description
##   These test functions are needed for the VSeries test suites
## @usage
##   use filer;
##   use filer qw(...);
##
## @keywords  filer test library
##
## @dependencies
## use config_base
## use Tharn
##
## @author Matt Cattel(cattel@netapp.com), dl-vseries-automation
## @status testme
## @change 2006-11-26 clairek: created, based on style of VSeries.pm as well
## as the OO style of the host side libraries
##

package VSeries::filer;

use strict;
use Tharn;
use VSeries::cluster;
use VSeries::utils;
use Expect;
use config_base;
use Switch;
use Time::localtime;
use Getopt::Long;           # Helps with methods that take lots of options.
use File::Basename;         # So we can parse up files/dirs easily.
use Text::ParseWords;       # So many textual utils in here.
use ZapiTest;               # An rsh lib to get ontapi from filer.
use XML::Simple;            # This chews on XML and returns them out as hashes.
use Data::Dumper;

use vars ('@ISA','@EXPORT');
require Exporter;
@ISA = ('Exporter');
@EXPORT = ('new','set_internal_variables','connect_to_filer','create_aggr',
       'create_vols','create_luns','delete_aggr','delete_vols','delete_luns',
       'boot_into_maintenance','add_options','add_licenses', 'auto_entries',
       'get_disks_by_owner', 'get_sysID', 'discover_sysID', 'get_error_state',
       'get_error_msg', 'disk_assign', 'disk_unassign', 'get_name', 'check_cf_state',
       'get_use_aggr', 'discover_disks', 'discover_disk_paths', 'two_path_check',
       'get_disk_info', 'vol_status', 'aggr_status', 'get_aggr_size', 'get_vol_size',
       'aggr_show_space', 'clear_error', 'aggr_add', 'vol_add', 'set_aggr_option',
       'set_vol_option', 'set_snap_reserve', 'filer_fabric_count_get', 'filer_lun_count_get',
       'lun_scsi_inquiry', 'storage_show_disk', 'wwn_adapter_get', 'filer_env_variable_set',
       'cf_enable','cf_disable', 'send_filer_cmd', 'reboot_filer', 'go_to_cfe',
       'go_to_maintenance_mode', 'go_to_special_menu', 'go_to_boot_sequence', 'disk_remove_ownership',
       'software_install', 'download_software', 'get_printenv', 'get_version', 'configure_filer',
       'get_licenses', 'poll_filer', 'get_sysconfig', 'discover_ip_interfaces', 'get_zeroed_disks',
       'disk_zero_spares', 'revision_checker', 'discover_storage_type', 'exportfs', 'get_disk_type_counts',
       'get_switch_portlist', 'get_storage_portlist', 'aggr_mirror', 'cf_forcetakeover', 'cf_forcegiveback',
       'asis_sis_on', 'asis_sis_off', 'asis_sis_start', 'asis_sis_status', 'get_df_params', 'get_cf_monitor',
       'offline_vols', 'set_snap_sched', 'get_fcvi_config', 'get_nvram_config', 'get_cluster_type',
       'dd', 'mv', 'showfh', 'dumpblock', 'dump', 'get_priv', 'rdfile', 'rm', 'wafl_iron_start',
       'get_flex_vol_container', 'aggr_rename', 'vol_rename',  'storage_array_remove', 'storage_array_modify',
       'storage_array_remove_port', 'storage_array_show', 'storage_array_show_ports', 'storage_array_show_luns',
       'storage_array_show_config', 'storage_array_purge_database', 'discover_lun_group_count', 'get_lun_group_count',
       'fcadmin_offline_online', 'systemshell_login_logout', 'ontapi_cmd', 'storage_port_associate', 'storage_port_find',
       'fcadmin_link_state', 'poll_hba_interfaces', 'storage_load_balance', 'online_aggrs', 'online_vols',
       'offline_aggrs', 'get_storage_show_disk', 'check_messages', 'get_disk_serial', 'wait_aggr_grow',
       'wait_for_zeroing_to_complete', 'wait_aggr_initialize', 'get_license_version', 'add_licenses_v2');

my $GOOD_STATUS = 0;
my $ERR_STATUS = 1;
my $INFO_STATUS = 2;
my @VSERIES_TEST_LICENSE_SUBSET = ("vseries", "nfs", "cifs", "cluster", "http", "fcp", "iscsi", "syncmirror_local",
                                   "cluster_remote", "nearstore_option", "a_sis", "snaplock", "snaplock_enterprise");

my @VSERIES_TEST_LICENSE_SUBSET_V2 = ("v_storageattach", "nfs", "cifs", "fcp", "iscsi", "snaplock", "snaplock_enterprise", "snapmirror");

# This is an array of the known licenses for a filer.  In cases where one
# license is a substring of another license the longer license name is placed
# first intentionally so that if you iterate over the array you'll do the more
# specific matching first.
my @KNOWN_LICENSES = ("a_sis", "cifs", "cluster_remote", "cluster", "disk_sanitization", "fcp",
                      "flex_cache", "flexcache_nfs", "flex_clone", "flex_scale", "gateway_hitachi", "http", "iscsi",
                      "multistore", "nearstore_option", "nfs", "smdomino", "smsql", "snaplock_enterprise",
                      "snaplock", "snapmanagerexchange", "snapmirror_sync", "snapmirror",
                      "snapmover", "snaprestore", "snapvalidator", "sv_linux_pri", "sv_nbu_sec",
                      "sv_ontap_pri", "sv_ontap_sec", "sv_unix_pri", "sv_windows_ofm_pri",
                      "sv_windows_pri", "syncmirror_local", "v-series", "vld", "cf");

my @KNOWN_LICENSES_V2= ("v_storageattach", "nfs", "cifs", "fcp", "iscsi", "snaplock_enterprise", "snaplock", "snapmirror", "snaprestore", "snapvault", "insight_balance", "flexclone");



# Constants
my $FORCEDTAKEOVER_TIME=240;
my $FORCEDGIVEBACK_TIME=240;
my $AGGR_MIRROR_REJOIN_TIME=300;
my $TRUE = "true";
my $FALSE = "false";
my $INITIALIZATION_MINUTES = 60;

## @name ::new()
##
## @description
## Initializes a new Filer object and it's internal variables.
##
## @synopsis
## $filer_obj=VSeries::filer->new();
##
sub new ()
{
    my ($class, %args) = @_;
    my ($self);

    # create a new object only if necessary
    if (defined $args{-self}) {
        $self=$args{-self};
    } else  {
        $self={};
        bless $self,$class;
    }

    # declare and initialize the constructs

    # filer_index is used for root vol creation, even numbers will get a trad
    # vol for the root vol, odd numbers will create a flex vol.
    $self->{filer_index} = -1;
    $self->{conn}= [];
    $self->{connection}= [];
    $self->{name} = "";
    $self->{user} = "";
    $self->{password} = "";
    $self->{timeout} = "";
    $self->{clients} = [];
    $self->{switches} = [];
    @{$self->{vol_list}} = [];
    $self->{vol_number} = 0;
    $self->{vol_disks} = 0;
    $self->{use_flexvol} = 0;
    $self->{flexvol_size} = 0;
    $self->{aggr} = "";
    $self->{use_aggr} = 0;
    $self->{aggr_disks} = 0;
    @{$self->{lun_list}} = [];
    $self->{lun_number} = 0;
    $self->{lun_size} = "";
    $self->{lun_type} = "";
    $self->{error} = 0;
    $self->{errmsg} = "";
    $self->{debug_level} = 0;
    # Stores non-fatal but important info that may have resulted in a method.
    @{$self->{info_msg}} = ();

    # Flags used to check if san and metrocluster licenses should be install
    # can be bypassed by adding "san_setup" and "metro_setup" strings to the
    # array passed into add_licenses()
    $self->{san_setup} = 0;
    $self->{metro_setup} = 0;

    $self->{dns_server} = "";
    $self->{domain_name} = "";
    # ip address info
    $self->{gateway} = "";
    # primary
    $self->{ip_addr_1} = "";
    $self->{ip_mask_1} = "";
    $self->{ip_int_1} = "";
    $self->{ip_dup_1} = "";
    # secondary
    $self->{ip_addr_2} = "";
    $self->{ip_mask_2} = "";
    $self->{ip_int_2} = "";
    $self->{ip_dup_2} = "";
    # components specifically for the from-scratch setup
    $self->{sys_id} = undef;

    # Contains information about the disks a filer sees.  Primarily gets
    # populated by calling discover_disks()
    %{$self->{disks}} = ();

    # Contains information for the configuration of IP interfaces for the filer
    %{$self->{ip_interfaces}} = ();

    # The system model for the filer.
    $self->{model} = "";

    # Flag used to have communication with the filer in most methods prefer
    # using the console connection.  Other commands that can only be run over
    # the console because of the state of the filer, will automatically do so
    $self->{use_console} = "false";

    # Physical location string of the filer used in the original configuration
    # of the filer.
    $self->{location} = " ";

    # Hash table of performance metrics, this is initially being created to
    # start recording length of time it takes to create aggregates or volumes
    # when disks haven't been zeroed.  All creation times should be stored
    # underneath the attribute {creation} and then the type of entity being
    # created.  Currently those are {aggregates},{traditional_volumes},
    # {vol_mirror},{aggr_mirror}, {corals}
    # Thus aggregate create times will be {creation}{aggregates}.
    %{$self->{performance_metrics}} = ();

    # These are array profile specific attributes should only be used in BR.
    # Array Profiles are related to the 'storage array' command suite.
    $self->{lun_group_count} = "";    # Holds the number of lun groups.
                                      # Updated using  'discover_lun_group_count'.
                                      # Accessed using 'get_lun_group_count'.

    return($self);
}


## @name ::set_internal_variables()
##
## @description
## Parameters are all optional.  Defined as such in order to introduce
## readability and flexibility moving forward (i.e. you should be able to call
## this from anywhere with parameters missing and/or in the wrong order.
## Errors will occur if a parameter is not labeled correctly.
##
## @synopsis
## $filer_obj=VSeries::filer->set_internal_variables();
##
## @arg %args Required: The actual key value pairs that you wish to change
## in the $filer_obj.
##
sub set_internal_variables(\%)
{
    # pick up our object
    my $self = shift;

    # catch the optional function parameters here
    my %args = @_;

    # parse function parameters that are present, putting into data members
    # could just say "$self->{$arg} = $val", but no check
    while (my ($arg, $val) = each %args)
    {
        if ($arg eq "name")
        {
            $self->{name} = $val;
        }
        elsif ($arg eq "filer_index")
        {
            $self->{filer_index} = $val;
        }
        elsif ($arg eq "user")
        {
            $self->{user} = $val;
        }
        elsif ($arg eq "password")
        {
            $self->{password} = $val;
        }
        elsif ($arg eq "timeout")
        {
            $self->{timeout} = $val;
        }
        elsif ($arg eq "clients")
        {
            $self->{clients} = $val;
        }
        elsif ($arg eq "switches")
        {
            $self->{switches} = $val;
        }
        elsif ($arg eq "vol_list")
        {
            @{$self->{vol_list}} = split(/,/, $val);
        }
        elsif ($arg eq "vol_disks")
        {
            $self->{vol_disks} = $val;
        }
        elsif ($arg eq "use_flexvol")
        {
            $self->{use_flexvol} = $val;
        }
        elsif ($arg eq "flexvol_size")
        {
            $self->{flexvol_size} = $val;
        }
        elsif ($arg eq "aggr")
        {
            $self->{aggr} = $val;
        }
        elsif ($arg eq "use_aggr")
        {
            $self->{use_aggr} = $val;
        }
        elsif ($arg eq "aggr_disks")
        {
            $self->{aggr_disks} = $val;
        }
        elsif ($arg eq "lun_list")
        {
            @{$self->{lun_list}} = split(/,/, $val);
        }
        elsif ($arg eq "lun_size")
        {
            $self->{lun_size} = $val;
        }
        elsif ($arg eq "lun_type")
        {
            $self->{lun_type} = $val;
        }
        elsif ($arg eq "ip_addr_1")
        {
            $self->{ip_addr_1} = $val;
        }
        elsif ($arg eq "ip_mask_1")
        {
            $self->{ip_mask_1} = $val;
        }
        elsif ($arg eq "ip_int_1")
        {
            $self->{ip_int_1} = $val;
        }
        elsif ($arg eq "ip_dup_1")
        {
            $self->{ip_dup_1} = $val;
        }
        elsif ($arg eq "ip_addr_2")
        {
            $self->{ip_addr_2} = $val;
        }
        elsif ($arg eq "ip_mask_2")
        {
            $self->{ip_mask_2} = $val;
        }
        elsif ($arg eq "ip_int_2")
        {
            $self->{ip_int_2} = $val;
        }
        elsif ($arg eq "ip_dup_2")
        {
            $self->{ip_dup_2} = $val;
        }
        elsif ($arg eq "gateway")
        {
            $self->{gateway} = $val;
        }
        elsif ($arg eq "domain_name")
        {
            $self->{domain_name} = $val;
        }
        elsif ($arg eq "dns_server")
        {
            $self->{dns_server} = $val;
        }
        elsif ($arg eq "san_setup")
        {
            $self->{san_setup} = $val;
        }
        elsif ($arg eq "metro_setup")
        {
            $self->{metro_setup} = $val;
        }
        elsif ($arg eq "model")
        {
            $self->{model} = $val;
        }
        elsif ($arg eq "use_console")
        {
            if($val =~ m/true/i){
                $self->{use_console} = "true";
            }
            else{
                $self->{use_console} = "false";
            }

        }
        elsif ($arg eq "ip_interfaces")
        {
            %{$self->{ip_interfaces}} = %$val;
        }
        elsif ($arg eq "location"){
            $self->{location} = $val;
        }
        else
        {
            $self->{error} = 1;
            $self->{errmsg} = "Invalid parameter: " . $arg . " => " . $val;
        }
    }
}


## @name ::connect_to_filer()
##
## @description
## Make the TCP/IP connection to the filer and store it
##
## @synopsis
## $filer_obj=VSeries::filer->connect_to_filer();
##
##
sub connect_to_filer()
{
    my $self = shift;

    $self->{conn} = Tharn::host($self->{name});
    $self->{connection} = VSeries::cluster::make_connection($self->{conn});
    $self->{console_connection} = connect("#console");
}

## @name ::create_aggr()
##
## @description
## Create an aggregate as specified by user.
## This method thus far supports two general ways of going about creating
## an aggregate.  The first is by using a named disk and the second is by
## specifying a count of disks/luns where an aggregate creation uses avail-
## able disks/luns to create your aggregate.
##
## 1.) CREATE AGGREGATE WITH A DISK/LUN LIST & NO OPTIONS.
## You should have populated the
## $filer_obj->{aggr_disks} with a space separated disk/lun list preceeded
## by "-d " by now. Also, the $filer_obj->{use_aggr} flag should be set.
## Also the aggregate stored in $filer_obj->{aggr} should be the aggregate
## that you are trying to create, NOT an aggregate that already exists.
##
## 2.) CREATE AGGREGATE WITH A DISK/LUN COUNT AND/OR OTHER OPTIONS.
## $filer_obj->{use_aggr} flag should be set.  The aggregate stored in
## $filer_obj->{aggr} should be the aggregate that you are trying to create,
## NOT an aggregate that already exists. You should have passed into the
## \%options_ref hash key value pairs, where your keys are the option names
## and values are what you'd like set for each respective option.  It is
## expected that when the caller wishes to use an option which has no value,
## i,e, simply the "-m" option, the key should be "-m" and the value should
## be "" an empty string.
##
## *Note it is up to the caller to pass in the appropriate options for
##  the createion of an aggregate.  We depend on the OnTap CLI for intelli-
##  gence.  If the CLI has an error, we will too.
##
## *Note for best performance, the caller should zero any disks that will
##  be used for aggregate creation.
##
## @synopsis
## $create_status = VSeries::filer->create_aggr();
##
## @arg \%options Optional:  The 'aggr create' options to use for creating
##                          an aggregate.
##
##
sub create_aggr(;%)
{
    my ($self, $options_ref) = @_;
    my %options = %$options_ref if (defined $options_ref);
    my ($match, $after, $cmd, $prompt) = undef;
    my @output_array = ();    # aggr status output
    my %status_hash = ();     # parse_aggr_status output
    my $str = "";

    # return to caller if we have not been told by the user to do this
    if (!$self->{use_aggr})
    {
        return $ERR_STATUS;
    }

    # Get a list of existing aggregate names
    @output_array = @{$self->aggr_status()};
    %status_hash = %{VSeries::utils::parse_aggr_status("", @output_array)};

    # Check if the aggr name we wish to create already exists
    foreach my $existing_aggr_name (keys %status_hash) {
        if ($self->{aggr} eq $existing_aggr_name ) {
            $match = 1;
            last;
        }
    }

    if ($match)
    {
        Tharn::logcomment("aggregate has been previously created");
    } else {

        # - - - - - - -  - - - - - - - - - - - - - -
        # Here we need to append any options that exist in the %options
        # hash to the aggr create command stored in $cmd.  This will also
        # be influenced by the -n option because if that exists, we don't
        # need to use the $self->{aggr_disks} data member as a source of disks.
        # - - - - - - -  - - - - - - - - - - - - - -

        if (keys %options >= 1) {
            # do the aggr create specifying any options provided
            $cmd = "aggr create " . $self->{aggr} . " ";

            foreach my $aggr_option (keys %options) {

                if ($options{$aggr_option} eq "") {
                    $cmd .= $aggr_option . " ";
                } else {
                    $cmd .= $aggr_option . " " . $options{$aggr_option} . " " unless $aggr_option eq "-n";
                }
            }
            # Get rid of the trailing space at the end of the command.
            chop($cmd) if($cmd =~ / $/);

            # Does -n option exist? It should if you've specified command line
            # options in the call.  If so (-n exists), it MUST be added last.
            if (exists $options{-n}) { $cmd .= " -n " . $options{-n}; }

            # - - - - - - -  - - - - - - - - - - - - - -
            # Next, we'll pass this command to OnTap & it will send us back
            # a string that we can then enter to create our aggregate with
            # "n" number of disks or luns.
            # - - - - - - -  - - - - - - - - - - - - - -

            if ($self->{use_console} eq "true") {
                (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow);
            } else {
                (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow);
            }

            # - - - - - - -  - - - - - - - - - - - - - -
            # If OnTap likes what we sent it, they'll send us the appropriate
            # command to create what we would like.  It will always start off
            # with "aggr create $self->{aggr}".  If it doesn't we know we've
            # encountered an error.
            #
            # vsa-v3050-auto-1*> aggr create fcal_aggr2 -T FCAL -t raid_dp -R 10000 -n 3
            # aggr create fcal_aggr2 -t raid_dp -d 0d.29 0d.28 0d.27
            # - - - - - - -  - - - - - - - - - - - - - -

            if ($after !~ /aggr create $self->{aggr}/) {
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "Malformed aggr create command.\n" . $after;
                return $ERR_STATUS;
            }

            # This fixes the case where if the -n flag was used with the
            # command sent over the console it shows up in $after with the
            # translated command and messes up the parsing.  So we stringently
            # check for that here and remove the original form of the command
            # with the -n.
            if((exists $options{-n}) && $self->{use_console} eq "true" && $after =~ m/$cmd/){
                $after =~ s/$cmd//;
            }

            # - - - - - - -  - - - - - - - - - - - - - -
            # Now we need to parse the raw output contained within $after so
            # we may send back to OnTap a correct command to create the aggr.
            # - - - - - - -  - - - - - - - - - - - - - -

            my @after_array = split (/\s/, $after);

            # - - - - - - -  - - - - - - - - - - - - - -
            # Our command starts with "aggr" lets make sure the array starts
            # with that.  If it does not, shift the garbage off.
            # - - - - - - -  - - - - - - - - - - - - - -

            foreach my $element (@after_array) {
                if ($element eq "aggr") {
                    last;
                } else {
                    shift @after_array;
                }
            }

            # - - - - - - -  - - - - - - - - - - - - - -
            # At this point we're sure our array starts with "aggr".
            # continue to build the command:
            #
            # aggr create $self->{aggr} ((-\w)|(\w+))* (.*)\.(\d+.*)
            #
            # Options match this format: (-\w)|(\w+))
            # Examples:
            # -t raid_dp
            # -m
            #
            # Disks and LUNS match this format: (.*)\.(\d+.*)
            # Examples:
            # 0d.25
            # vss_b3800_7:2.126L34
            # vss-b3800-7:2.126L33
            # vns-c9124-3:1-1.126L12
            # - - - - - - -  - - - - - - - - - - - - - -

            $cmd = "";
            my @temp = ();

            foreach my $element (@after_array) {
                if (($element eq "aggr")             ||
                    ($element eq "create")           ||
                    ($element eq $self->{aggr})      ||
                    ($element =~ /((-\w)|(\w+))/)    ||
                    ($element =~ /(.*)\.(\d+.*)/)) {
                    $cmd .= "$element" . " ";
                } else {
                    push(@temp, $element);
                }
            }

            # Get that last space off of the command.
            chop($cmd);

            # - - - - - - -  - - - - - - - - - - - - - -
            # Send the command back to OnTap and make sure the create began.
            # Note, you may be prompted with the following question:
            #
            # vsa-v3050-auto-1*> aggr create fcal_aggr2 -t raid_dp -d 0d.29 0d.28 0d.27
            # WARNING! Continuing with aggr create will result in a
            # low spares disk condition for one or more raid groups.
            # Are you sure you want to continue with aggr create? yes
            # - - - - - - -  - - - - - - - - - - - - - -

            if ($self->{use_console} eq "true"){
                (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow);
            } else {
                (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow,);
            }

            if ($after =~ /Are you sure you want to continue with aggr create/) {

                if ($self->{use_console} eq "true"){
                    (undef, undef, $after) = Tharn::say($self->{console_connection}, "yes", -swallow);
                } else {
                    (undef, undef, $after) = Tharn::say($self->{connection}, "yes", -swallow);
                }
            }

        } else {

            # - - - - - - -  - - - - - - - - - - - - - -
            # No options were specified, so the user must have wanted to
            # use a disk list which doesn't allow for any options.  The disk
            # list should be contained within $self->{aggr_disks} and should
            # look like "-d disk1 diskn diskn+1".  $self->{aggr_disks} is
            # usually populated by "filer::set_internal_variables".
            # - - - - - - -  - - - - - - - - - - - - - -

            $cmd = "aggr create " . $self->{aggr} . " " . $self->{aggr_disks};

            if ($self->{use_console} eq "true"){
                (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow);
            } else {
                (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow,);
            }

            if ($after =~ /Are you sure you want to continue with aggr create/) {

                if ($self->{use_console} eq "true"){
                    (undef, undef, $after) = Tharn::say($self->{console_connection}, "yes", -swallow, "-wait", "initiated", "-ok");
                } else {
                    (undef, undef, $after) = Tharn::say($self->{connection}, "yes", -swallow, "-wait", "initiated", "-ok");
                }
            }
        }

        if ($after !~ /initiated/) {
            # adding this check for the freak aggr that completes super fast
            # once in a while (disks already were initialized, perhaps)
            $cmd = "aggr status " . $self->{aggr};
            my $exit = undef;
            if($self->{use_console} eq "true"){
                $exit = Tharn::say($self->{console_connection}, $cmd, "-wait", "No aggregate", "-ok");
            } else {
                $exit = Tharn::say($self->{connection}, $cmd, "-wait", "No aggregate", "-ok");
            }

            if ($exit)
            {
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "Aggr create failure.\n" . $after;
                return $ERR_STATUS;
            }
        }

        $str = "online";

        $cmd = "aggr status -r " . $self->{aggr};

        my $status = undef;

        do {
            if($self->{use_console} eq "true"){
                $status = Tharn::say($self->{console_connection}, $cmd, "-wait", $str, "-ok");
            } else {
                $status = Tharn::say($self->{connection}, $cmd, "-wait", $str, "-ok");
            }
            sleep(30);
        }  while (!$status);
        
        # - - - - - - -  - - - - - - - - - - - - - -
        # Now we need to go into a wait loop until aggr status
        # is not initializing.
        # - - - - - - -  - - - - - - - - - - - - - -

        my $cmd = "aggr status " . $self->{aggr};
        my $str = "initializing.+\n";
        my $match = 0;

        # - - - - - - -  - - - - - - - - - - - - - -
        # Restricting the wait time to be 60 minutes to avoid going
        # into an infinite loop. By experience volume initialization
        # completes in about 45 minutes if it is NND.
        # - - - - - - -  - - - - - - - - - - - - - -
        my $initialization_timeout = $INITIALIZATION_MINUTES * 60;
        my $initialization_end_time = Tharn::timeout2time($initialization_timeout);

        Tharn::logcomment("Going to check if aggregate is still in initialization state, if yes we will wait for $INITIALIZATION_MINUTES minutes and fail on timeout");

        do {
            if($self->{use_console} eq "true"){
                $match = Tharn::say($self->{console_connection}, $cmd, "-wait", $str, "-ok");
            }
            else{
                $match = Tharn::say($self->{connection}, $cmd, "-wait", $str, "-ok");
            }
            sleep(60);
        }  while ($match && time() <= $initialization_end_time);

        if ($match)
        {
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "$self->{aggr} is still in initialization state";
            return ($ERR_STATUS);
        }
    } # The else meaning this aggregate wasn't already created.

    return $GOOD_STATUS;
}



## @name ::create_vols()
##
## @description
## Create new volume(s) on the filer.  A volume create will not be attempted
## if a volume with the same name already exists.  Otherwise, if
## $self->{use_flexvol} has been set by the user, a flex vol(s) will be created
## on top of a previously created aggregate ($self->{aggr}), if not, a
## traditional vol will be created with a certain number of disks as specified
## in $self->{vol_disks}.
##
## $self->{vol_number} is a count stored within the filer object for number of
## volumes.  This will be incremented by one each time we successfully create
## a flexible or traditional volume.
##
##
## TRADITIONAL VOLUMES HOW-TO
## - Populate @{$self->{vol_list}} with a single volume name that you'd like
##   to create.  Even if you push more on, we're only going to attempt to
##   create one by using ${$self->{vol_list}}[0] which is the first element
##   of the volume list array.  If a volume in vol_list has already been
##   created, we will not attempt to create it, rather push onto the @info_msg
##   array an informative string stating that the volume $vol already exists.
##   We will then return $INFO_STATUS.  An already created volume will not
##   increment the $self->{vol_number} counter.
## - Unset $self->{use_flexvol}, i.e. $self->{use_flexvol} = 0;
## - Populate $self->{vol_disks} with the disks/luns you wish to use to create
##   the traditional volume.  This should be a space separated list of
##   disks/luns prepended with -d.  i.e. $self->{vol_disks} = "-d disk1 disk2".
##   Make sure that disk list contains all the same type of storage, i.e. LUNS
##   versus FCAL disks versus SATA disks.
##
##
## FLEX VOLUMES HOW-TO
## - $self->{aggr} the name of the aggregate you'd like to use as a source for
##   your flex volume(s).  Make sure the aggr is big enough for your needs.
## - $self->{flexvol_size} is set to size[k|m|g|t] for kilobytes, megabytes,
##   gigabytes, and terabytes respectively.  i.e. $self->{flexvol_size} = 500m.
## - Populate $self->{vol_list} with one or more volume names that you'd like
##   to create.
## - If a volume in vol_list has already been created, we will not attempt to
##   create it, rather push onto the @info_msg array an informative string
##   stating that the volume $vol already exists.  We will eventually return
##   $INFO_STATUS once we've tried to create the rest of the volumes that may
##   exist in @{$self->{vol_list}}.  An already created volume will not
##   increment the $self->{vol_number} counter.


## @synopsis
## $return_status = VSeries::filer->create_vols();
##
##
sub create_vols()
{
    my ($self) = @_;
    my ($cmd, $str, $match) = "";
    my $vol_exists_flag = $FALSE;
    my @output_array = ();    # vol status output
    my %status_hash = ();     # parse_vol_status output
    $str = "";


    # TRADITIONAL VOLUME CREATION
    if (!$self->{use_flexvol})
    {
        # - - - - - - -  - - - - - - - - - - - - - -
        # Take the first element of the vol list array as described above.
        # - - - - - - -  - - - - - - - - - - - - - -

        my $vol = ${$self->{vol_list}}[0];

        # - - - - - - -  - - - - - - - - - - - - - -
        # Does this volume already exist?
        # - - - - - - -  - - - - - - - - - - - - - -
        @output_array = @{$self->vol_status()};
        %status_hash = %{VSeries::utils::parse_vol_status("", @output_array)};

        # Check if the tradvol name we wish to create already exists
        foreach my $existing_tradvol_name (keys %status_hash) {
            if ( $vol eq $existing_tradvol_name ) {
                $match = 1;
                last;
            }
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # If $match is true, it means the volume already exists. As a result
        # we need to push onto the @info_msg array an informative message for
        # the calling .thpl and also return $INFO_STATUS.
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($match)
        {
            push (@{$self->{info_msg}}, "$vol volume has been previously created");
            $self->{error} = $INFO_STATUS;
            return ($INFO_STATUS);
        }

        Tharn::logcomment("Traditional volume has not been created yet, please be prepared to wait");
        $cmd = "vol create " . $vol . " " . $self->{vol_disks};

        if($self->{use_console} eq "true"){
            $match = Tharn::say($self->{console_connection}, $cmd, "-wait", $vol, "-ok");
        }
        else{
            $match = Tharn::say($self->{connection}, $cmd, "-wait", $vol, "-ok");
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Now we need to go into a wait loop until volume is created.
        # This could take a while especially if the disks chosen have
        # not been zeroed.
        # - - - - - - -  - - - - - - - - - - - - - -

        $cmd = "vol status " . $vol;
        $str = $vol . " creating.+\n";
        do {
            Tharn::logcomment("Checking to see if volume has been created ...");
            if($self->{use_console} eq "true"){
                $match = Tharn::say($self->{console_connection}, $cmd, "-wait", $str, "-ok");
            }
            else{
                $match = Tharn::say($self->{connection}, $cmd, "-wait", $str, "-ok");
            }
            sleep(30);
        }  while ($match);
        
        # - - - - - - -  - - - - - - - - - - - - - -
        # Now we need to go into a wait loop until volume is online.
        # - - - - - - -  - - - - - - - - - - - - - -

        $cmd = "vol status " . $vol;
        $str = $vol . " online.+\n";
            
        # - - - - - - -  - - - - - - - - - - - - - -
        # Restricting the wait time to be 90 seconds to avoid going
        # into an infinite loop.
        # - - - - - - -  - - - - - - - - - - - - - -
        my $timeout = 90;
        my $end_time = Tharn::timeout2time($timeout);
            
        do {
            
            if($self->{use_console} eq "true"){
                $match = Tharn::say($self->{console_connection}, $cmd, "-wait", $str, "-ok");
            }
            else{
                $match = Tharn::say($self->{connection}, $cmd, "-wait", $str, "-ok");
            }
            sleep(15);
        }  while (!$match && time() <= $end_time);
            
        if (!$match)
        {
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "$vol is not in online state";
            return ($ERR_STATUS);
        }
        
        # - - - - - - -  - - - - - - - - - - - - - -
        # Now we need to go into a wait loop until trad vol status
        # is not initializing.
        # - - - - - - -  - - - - - - - - - - - - - -

        my $cmd = "vol status " . $vol;
        my $str = "initializing.+\n";
        my $match = 0;

        # - - - - - - -  - - - - - - - - - - - - - -
        # Restricting the wait time to be 60 minutes to avoid going
        # into an infinite loop. By experience volume initialization
        # completes in about 45 minutes if it is NND.
        # - - - - - - -  - - - - - - - - - - - - - -
        my $initialization_timeout = $INITIALIZATION_MINUTES * 60;
        my $initialization_end_time = Tharn::timeout2time($initialization_timeout);

        Tharn::logcomment("Going to check if tradvol is still in initialization state, if yes we will wait for $INITIALIZATION_MINUTES minutes and fail on timeout");

        do {
            if($self->{use_console} eq "true"){
                $match = Tharn::say($self->{console_connection}, $cmd, "-wait", $str, "-ok");
            }
            else{
                $match = Tharn::say($self->{connection}, $cmd, "-wait", $str, "-ok");
            }
            sleep(60);
        }  while ($match && time() <= $initialization_end_time);

        if ($match)
        {
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "$vol is still in initialization state";
            return ($ERR_STATUS);
        }
        
        # - - - - - - -  - - - - - - - - - - - - - -
        # We added a traditional volume, thus our total volume
        # count must be incremented by one.
        # - - - - - - -  - - - - - - - - - - - - - -

        $self->{vol_number}++;

        return ($GOOD_STATUS);

    } # if (!$self->{use_flexvol})

    # FLEX VOLUME CREATION
    else {

        # - - - - - - -  - - - - - - - - - - - - - -
        # Make a flex vol(s) using $self->{aggr} as a source.
        # - - - - - - -  - - - - - - - - - - - - - -

        foreach my $vol (@{$self->{vol_list}}) {

            # - - - - - - -  - - - - - - - - - - - - - -
            # Does this volume already exist?
            # - - - - - - -  - - - - - - - - - - - - - -

            @output_array = @{$self->vol_status()};
            %status_hash = %{VSeries::utils::parse_vol_status("", @output_array)};

            $match = 0;
            # Check if the flexvol name we wish to create already exists
            foreach my $existing_flexvol_name (keys %status_hash) {
                if ( $vol eq $existing_flexvol_name ) {
                    $match = 1;
                    last;
                }
            }

            # - - - - - - -  - - - - - - - - - - - - - -
            # If $match is true, it means the volume already exists.  As a result
            # we need to push onto the @info_msg array an informative message for
            # the calling .thpl and also set a flag to remind us that we need to
            # set the return status to #INFO_STATUS once we attempt to create the
            # rest of the possible volumes within @{$self->{vol_list}}
            # - - - - - - -  - - - - - - - - - - - - - -

            if ($match)
            {
                $vol_exists_flag = $TRUE;
                push (@{$self->{info_msg}}, "$vol volume has been previously created");
                next;
            }

            $cmd = "vol create " . $vol . " " . $self->{aggr} . " " . $self->{flexvol_size};
            $str = " has completed.+\n";

            if($self->{use_console} eq "true"){
                $match = Tharn::say($self->{console_connection}, $cmd, "-wait", $str, "-ok", -timeout, 120);
            }
            else {
                $match = Tharn::say($self->{connection}, $cmd, "-wait", $str, "-ok", -timeout, 120);
            }
            if (!$match)
            {
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "Unable to create vol $vol";
                return ($ERR_STATUS);
            }
            
            # - - - - - - -  - - - - - - - - - - - - - -
            # Now we need to go into a wait loop until volume is online.
            # - - - - - - -  - - - - - - - - - - - - - -

            $cmd = "vol status " . $vol;
            $str = $vol . " online.+\n";
            
            # - - - - - - -  - - - - - - - - - - - - - -
            # Restricting the wait time to be 90 seconds to avoid going
            # into an infinite loop.
            # - - - - - - -  - - - - - - - - - - - - - -
            my $timeout = 90;
            my $end_time = Tharn::timeout2time($timeout);
            
            do {
                
                if($self->{use_console} eq "true"){
                    $match = Tharn::say($self->{console_connection}, $cmd, "-wait", $str, "-ok");
                }
                else{
                    $match = Tharn::say($self->{connection}, $cmd, "-wait", $str, "-ok");
                }
                sleep(15);
            }  while (!$match && time() <= $end_time);
            
            if (!$match)
            {
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "$vol is not in online state";
                return ($ERR_STATUS);
            }

	    # - - - - - - -  - - - - - - - - - - - - - -
            # Now we need to go into a wait loop until volume status
            # is not initializing.
            # - - - - - - -  - - - - - - - - - - - - - -

            $cmd = "vol status " . $vol;
            $str = "initializing.+\n";
            $match = 0;

            # - - - - - - -  - - - - - - - - - - - - - -
            # Restricting the wait time to be 60 minutes to avoid going
            # into an infinite loop. By experience volume initialization
            # completes in about 45 minutes if it is NND.
            # - - - - - - -  - - - - - - - - - - - - - -
            my $initialization_timeout = $INITIALIZATION_MINUTES * 60;
            my $initialization_end_time = Tharn::timeout2time($initialization_timeout);

            Tharn::logcomment("Going to check if volume is still in initialization state, if yes we will wait for $INITIALIZATION_MINUTES minutes and fail otherwise");

            do {
                if($self->{use_console} eq "true"){
                    $match = Tharn::say($self->{console_connection}, $cmd, "-wait", $str, "-ok");
                }
                else{
                    $match = Tharn::say($self->{connection}, $cmd, "-wait", $str, "-ok");
                }
                sleep(60);
            }  while ($match && time() <= $initialization_end_time);

            if ($match)
            {
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "$vol is still in initialization state";
                return ($ERR_STATUS);
            }

            # - - - - - - -  - - - - - - - - - - - - - -
            # We added a flexible volume, thus our total volume
            # count must be incremented by one.
            # - - - - - - -  - - - - - - - - - - - - - -
            $self->{vol_number}++;
        } # foreach my $vol (@{$self->{vol_list}}) {
    } # else {

    if ($vol_exists_flag) {
        $self->{error} = $INFO_STATUS;
        return ($INFO_STATUS);
    } else {
        return ($GOOD_STATUS);
    }
}


## @name ::create_luns()
##
## @description
## Creates luns on a filer.  Uses internal variables to determine what to create.
## If a lun has already been created on a filer it will log a comment about that
## and continue creating other luns.
##
## @synopsis
## $filer_obj=VSeries::filer->create_luns();
##
##
sub create_luns()
{
    my $self = shift;
    my $str = "";
    my $cmd = "";
    my $match = undef;
    # assume we need to create luns for now
    my $create_lun = 1;
    # if tally > 0, then we already have a lun count
    my $tally = $self->{lun_number};

    foreach my $vol (@{$self->{vol_list}})
    {
        foreach my $lun (@{$self->{lun_list}})
        {
            # tally check
            if ($tally == 0)
            {
                $self->{lun_number}++;
            }

            # create luns is false until proven true
            $create_lun = 0;
            $cmd = "lun show /vol/" . $vol . "/" . $lun;
            $str = $lun . ".+\n";
            if($self->{use_console} eq "true"){
                $match = Tharn::say($self->{console_connection}, $cmd, "-wait", $str, "-ok");
            }
            else{
                $match = Tharn::say($self->{connection}, $cmd, "-wait", $str, "-ok");
            }
            if ($match)
            {
                Tharn::logcomment("lun has been previously created");
            }
            else
            {
                Tharn::logcomment("lun has not been created yet");

                # need to worry about ostype for the lun creation
                # need to worry about size for the lun creation
                $cmd = "lun create -s " . $self->{lun_size} . " -t " . $self->{lun_type} . " /vol/" . $vol . "/" . $lun;
                if($self->{use_console} eq "true"){
                    Tharn::say($self->{console_connection}, $cmd, "-wait", $lun, "-ok");
                }
                else{
                    Tharn::say($self->{connection}, $cmd, "-wait", $lun, "-ok");
                }
            }
        }
    }
}


## @name ::delete_aggr()
##
## @description
## Deletes the aggr whose name exists in the internal variable aggr.
##
## @synopsis
## $filer_obj=VSeries::filer->delete_aggr();
##
##
sub delete_aggr()
{
    my $self = shift;
    my $cmd = "";
    my $status = "";
    my $output_array_ref = undef;
    my @output_array = ();
    my $status_hash_ref = undef;
    my %status_hash = ();
    my $counter = 0;
    my $aggr_name = $self->{aggr};

    do{
        # In the event that the first time through here the offline failed,
        # delay for 10 seconds and retry in case it was a transient condition
        if($counter > 0){
            sleep(10);
        }

        $cmd = "aggr offline " . $aggr_name;
        if($self->{use_console} eq "true"){
            $status = Tharn::say($self->{console_connection}, $cmd, "-wait", $self->{aggr}, "-ok");
        }
        else{
            $status = Tharn::say($self->{connection}, $cmd, "-wait", $self->{aggr}, "-ok");
        }

        $output_array_ref = $self->aggr_status();
        @output_array = @$output_array_ref;
        $status_hash_ref = VSeries::utils::parse_aggr_status("", @output_array);
        %status_hash = %$status_hash_ref;
        $counter++;

	# Bumping the counter value from old 3 to new 6 to wait for more time
        # for the unmounting state to go away.
        if($counter > 6){
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "Unable to offline aggr $aggr_name after multiple tries";
            return $ERR_STATUS;
        }


    }while($status_hash{$aggr_name}{state} ne "offline");



    $cmd = "aggr destroy " . $self->{aggr} . " -f";
    if($self->{use_console} eq "true"){
        $status = Tharn::say($self->{console_connection}, $cmd, "-wait", $self->{aggr}, "-ok");
    }
    else{
        $status = Tharn::say($self->{connection}, $cmd, "-wait", $self->{aggr}, "-ok");
    }

    return $GOOD_STATUS;
}



## @name ::delete_vols(;$)
##
## @description
## Walks a filer object's internal volume list offlining then deleting them.
## Will return an error after 3 tries to offline a given volume.  Waits 10
## second between retries.
##
## @synopsis
## $filer_obj=VSeries::filer->delete_vols($timeout);
##
## @arg $timeout    Optional: Timeout value in seconds for volume to go to
##                            offline state.
##                            Default (180 seconds)
##
##
sub delete_vols(;$)
{
    my $self = shift;
    my $timeout = shift;
    if (!defined $timeout) {
        $timeout = 180;
    }
    my $cmd = "";
    my $status = "";
    my $output_array_ref = undef;
    my @output_array = ();
    my $status_hash_ref = undef;
    my %status_hash = ();
    my $iter = 0;

    foreach my $vol (@{$self->{vol_list}})
    {
        # Initialize sleep_ctr to keep track of the amount of time
        # that has passed. This will be used to ensure that
        # we error out when vol status does not go to offline
        # within $timeout seconds
        my $sleep_ctr=0;
        do{
            # In the event that the first time through here the offline failed,
            # delay for 10 seconds and retry in case it was a transient
            # condition
            if ($iter > 0) {
                sleep 10;
                $sleep_ctr+=10;
            }

            $cmd = "vol offline " . $vol;
            if($self->{use_console} eq "true"){
                $status = Tharn::say($self->{console_connection}, $cmd, "-wait", $vol, "-ok");
            }
            else{
                $status = Tharn::say($self->{connection}, $cmd, "-wait", $vol, "-ok");
            }
            # Wait for a few seconds before checking status
            sleep 10;
            $sleep_ctr+=10;

            $output_array_ref = $self->vol_status();
            @output_array = @$output_array_ref;

            # Check if volume is in a busy state
            # If it is busy (offlining) then wait a little longer
            if (grep(/\'$vol\' is temporarily busy/, @output_array)) {
                sleep 30;
                $sleep_ctr+=30;
                $status_hash{$vol}{state} = "busy";
            }
            else {
                $status_hash_ref = VSeries::utils::parse_vol_status("", @output_array);
                %status_hash = %$status_hash_ref;
            }
            $iter++;

        }while($status_hash{$vol}{state} ne "offline" && $sleep_ctr <= $timeout);

        if($status_hash{$vol}{state} ne "offline"){
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "Unable to offline volume $vol after multiple tries";
            return $ERR_STATUS;
        }

        $cmd = "vol destroy " . $vol . " -f";
        if($self->{use_console} eq "true"){
            $status = Tharn::say($self->{console_connection}, $cmd, "-wait", $vol, "-ok");
        }
        else{
            $status = Tharn::say($self->{connection}, $cmd, "-wait", $vol, "-ok");
        }
    }

    return $GOOD_STATUS;
}


## @name ::delete_luns()
##
## @description
## Walks a filer object's internal volume list and attempts to delete the
## corresponding lun in a brute force way.  It tries all lun numbers for a vol
## thus it doesn't check for error messages.
##
## @synopsis
## $filer_obj=VSeries::filer->delete_luns();
##
##
sub delete_luns()
{
    my $self = shift;

    foreach my $vol (@{$self->{vol_list}})
    {
        foreach my $lun (@{$self->{lun_list}})
        {
            my $cmd = "lun destroy -f /vol/" . $vol . "/" . $lun;
            my $status = "";
            if($self->{use_console} eq "true"){
                $status = Tharn::say($self->{console_connection}, $cmd, "-wait", $lun, "-ok");
            }
            else{
                $status = Tharn::say($self->{connection}, $cmd, "-wait", $lun, "-ok");
            }
        }
    }
}

## @name ::auto_entries()
##
## @description
## Sets flow control variables if 'auto' is desired for IP duplex
##
## @synopsis
## $filer_obj=VSeries::filer->auto_entries();
##
##
sub auto_entries()
{
    my $self = shift;

    # since we entered auto, it wants flow control here
    Tharn::say($self->{console_connection}, "full", "-wait", "jumbo", "-ok");
    # since we entered auto, it wants jumbo support here
    Tharn::say($self->{console_connection}, "n", "-wait", "web", "-ok");

    # small sleep since waitstring is different for different interfaces
    # (which is also why "-ok" was used)
    sleep(5);
}


## @name ::configure_filer()
##
## @description
## Handles booting a filer from scratch up to a base level configuration with
## a root vol, IP address and such.
##
## @synopsis
## $filer_obj->configure_filer();
##
##
sub configure_filer()
{
    my $self = shift;
    my $SMALL_TIMEOUT = 30;
    my $LARGE_TIMEOUT = 240;
    my %filer_ips = %{$self->{ip_interfaces}};
    my @setup_prompts = ('\[.*\]:? +$', '\(.*\)\? +$');
    my $any_line_prompt = '^.*\n';
    my $h = $self->{conn};
    my @norm_prompts = @{$h->prompt_norm()};

    $self->go_to_special_menu();
    my ($match, $before, $after, $prompt) = "";



    # here we should see the numbered menu
    # we want to send 4 and 4a (one trad root vol, one flex root vol)
    my $type = "4";
    if ($self->{filer_index} % 2)
    {
        $type = "4a";
    }
    # root volume type
    Tharn::say($self->{console_connection}, $type, "-wait", "system?", "-ok");
    # prompt for new file system
    Tharn::say($self->{console_connection}, "y", "-wait", "sure?", "-ok");
    # prompt to make sure wiping the file system out is ok
    # this is where we want the large timeout in case root vol creation
    # takes a long time
    ($match, $before, $after, $prompt) = Tharn::say($self->{console_connection}, "y", "-noprompt", "-timeout", $LARGE_TIMEOUT, "-prompt", $any_line_prompt);
    # here we wait for some ...'s, vol creation, ifconfig, DHCP search
    # prompt for host name

    # Sit and listen to console output till we get prompted to enter ctrl-c to
    # skip DHCP.  Adding a max breakout time if we don't see the prompt for
    # entering ctrl-c
    my $breaktime = timeout2time(300);
    my $matched = 1;

    while($prompt !~ m/Ctrl-C to skip DHCP/ && $matched == 1){
        ($match, $before, $after, $prompt) = Tharn::say($self->{console_connection}, "", "-noprompt", "-timeout", $SMALL_TIMEOUT, "-prompt", $any_line_prompt);

        if(time() > $breaktime){
            $self->{error} = $ERR_STATUS;
            $self->{err_msg} = "Timed out waiting for DHCP prompt, aborting test...";
            return $ERR_STATUS;
        }

    }



    # Enter the ctrl-c to skip DHCP and wait for prompt for hostname.
    Tharn::say($self->{console_connection}, $h->{irptchar}, "-noprompt", "-timeout", $LARGE_TIMEOUT, "-wait", "Please enter the new hostname");


    # Hostname prompt.
    Tharn::say($self->{console_connection}, $self->{name}, "-wait", "virtual", "-ok");
    # vifs prompt
    # We get a little funky here.  Because the number of IP interfaces per
    # platform can vary we need a way to dynamically loop through the set of
    # questions. What's started here is a way to dynamically work though all
    # prompting questions of the configuration prompts regardless of order but
    # for now it works for IP interface questions.
    my $send_string = "n";
    my $current_interface = undef;
    $matched = 0;
    my $break = 0;
    my $interface = "";
    do{

        ($match, $before, $after, $prompt) = Tharn::say($self->{console_connection}, $send_string, "-timeout", $SMALL_TIMEOUT, "-swallow", "-noprompt", "-prompt", @setup_prompts);

        switch($after){
            # Prompt for IP Address, look up corresponding IP interface
            # if the interface isn't in our internal IP hash set it to
            # zeros.  We will keep track of the current interface for
            # subsequent IP interface questions.
            case /Please enter the IP address for Network Interface/{
                    (undef, $interface) = split(/Network Interface /, $after);
                    $interface =~ s/\s//g;
                    foreach my $interface_name (keys(%filer_ips)){
                        if($interface_name eq $interface && (exists $filer_ips{$interface_name}{ip_addr})){
                            $send_string = $filer_ips{$interface_name}{ip_addr};
                            $current_interface = $interface;
                            $matched = 1;
                            last;
                        }
                    }
                    if($matched == 0){
                        $send_string = "0.0.0.0";
                        $current_interface = undef;
                    }
            }

            # Netmask prompt, look up current interface in IP hash, otherwise
            # let it be the default which is usually 255.0.0.0
            case /Please enter the netmask/{
                if(!($current_interface)){
                    $send_string = " ";
                }
                elsif(exists $filer_ips{$current_interface}{netmask}) {
                    $send_string = $filer_ips{$current_interface}{netmask};
                }
            }

            # IP Interface takeover prompt, if this is not in the IP hash then
            # the interface will NOT takeover for a partner's IP.
            case /take over a partner IP address/{
                if(!($current_interface)){
                    $send_string = "n";
                }
                elsif(exists($filer_ips{$current_interface}{takeover_interface})){
                    $send_string = "yes";
                }
                else{
                    $send_string = "n";
                }
            }

            # In the case that the interface will take over a partner's
            # then enter the IP or interface that this interface will take over
            # for.
            case /Please enter the IP address or interface name to be taken over/{
                if(!($current_interface)){
                    $send_string = " ";
                }
                elsif($filer_ips{$current_interface}{takeover_interface}){
                    $send_string = $filer_ips{$current_interface}{takeover_interface};
                }
            }

            # Prompt for media type
            case /Please enter media type/{
                if(!($current_interface)){
                    $send_string = " ";
                }
                elsif($filer_ips{$current_interface}{media_type}){
                    $send_string = $filer_ips{$current_interface}{media_type};
                }
            }

            # Prompt for flow control
            case /Please enter flow control/{
                if(!($current_interface)){
                    $send_string = " ";
                }
                elsif($filer_ips{$current_interface}{flow_control}){
                    $send_string = $filer_ips{$current_interface}{flow_control};
                }

            }

            # Prompt for jumbo frames
            case /support jumbo frames/{
                if(!($current_interface)){
                    $send_string = " ";
                }
                elsif($filer_ips{$current_interface}{jumbo_frames}){
                    $send_string = $filer_ips{$current_interface}{jumbo_frames};
                }
            }

            # If we didn't match on an IP interface question then we assume
            # we're done with this and break out of this loop.  The next prompt
            # will be to continue configuration through the web interface and
            # we know the answer to that is no
            else{
                $send_string = "n";
                $break = 1;
            }

        }
        $matched = 0;


    }while($break == 0);

    # Setting up for next block of configuration questions.  Separated these
    # out into two loops for sake fo handling ip interfaces independent of the
    # other configuration questions.  Adding additional accepted prompts for
    # questions that don't have a [] or () at the end of them.
    $break = 0;
    push(@setup_prompts, 'host:');
    push(@setup_prompts, 'gateway:');
    push(@setup_prompts, 'continue\.\r');

    do{

        ($match, $before, $after, $prompt) = Tharn::say($self->{console_connection}, $send_string, "-timeout", $SMALL_TIMEOUT, "-swallow", "-prompt", @setup_prompts);

        switch($after){

            # Prompt for web interface setup, we never want to say yes here
            case/Would you like to continue setup through the web interface/{
                $send_string = "n";
            }

            # Default Gateway prompt
            case/Please enter the name or IP address of the default/{
                $send_string = $self->{gateway};

            }

            # Administration host IP, currently we've chosen to leave this
            # blank.
            case/Please enter the name or IP address of the administration/{
                $send_string = " ";
            }

            # Time zone question, again we've chosen to default this to EST.
            case/Please enter timezone/{
                $send_string = "EST";
            }

            # Filer location prompt, this can be passed in by
            # set_internal_variables
            case/Where is the filer located/{
                $send_string = $self->{location};
            }

            # Multi-protocol file language prompt, again leave this blank.
            case/What language will be used for multi-protocol files/{
                $send_string = " ";
            }

            # HTTP root directory prompt, again leaving blank for now.
            case/Enter the root directory for HTTP files/{
                $send_string = " ";
            }

            # If a DNS Domain is defined then we want to run the
            # DNS resolver and enter a DNS server ip.
            case/Do you want to run DNS resolver/{
                if($self->{domain_name} ne ""){
                    $send_string = "y";
                }
                else{
                    $send_string = "n";
                }
            }

            # We will only get this question if we said yes to running the DNS
            # resolver, and we would have only said yes if domain_name was
            # defined as something other than an empty string so we can just
            # enter it here.
            case/Please enter DNS domain name/{
                $send_string = $self->{domain_name};
            }

            # Again, if dns_server is an empty string this will be treated like
            # the user didn't mean to run the dns resolver so we'll send a
            # whitespace and new line so that it appears we just "hit enter".
            case/Please enter the IP address for first nameserver/{
                if($self->{dns_server} ne ""){
                    $send_string = $self->{dns_server};
                }
                else{
                    $send_string = " ";
                }
            }

            # Currently only supporting entering one DNS server, in the future
            # the data structure for dns_servers IPs may change and we'll be
            # able to dynamically add up to three.  For now, it's just one.
            case/Do you want another nameserver/{
                $send_string = "n";
            }

            # Again for now we'll just decide to not run the NIS client.
            case/Do you want to run NIS client/{
                $send_string = "n";
            }

            # This handles the prompt after all configuration is done and it
            # tells you to hit return to continue.
            case/Press the return key to/{
                $send_string = " ";
            }

            # This is acting like a catch all so that if we don't match
            # any case above, just send an enter command back to try to flush
            # the cli and move along.  This is useful when console messages
            # come up while we're in this loop near the end and want to just
            # continue to the filer's normal prompt.
            else{
                $send_string = " ";
            }

            # We break out of the loop once we see a normal prompt, indicating
            # configuration is complete.
            foreach my $norm_prompt (@norm_prompts){
                if($prompt =~ m/$norm_prompt/){
                    $break = 1;
                }
            }

        }


    } while($break == 0);

    return $GOOD_STATUS;

}

## @name ::boot_into_maintenance()
##
## @description
## Handles booting a filer from scratch up to a base level configuration with
## a root vol, IP address and such.
##
## @synopsis
## $filer_obj=VSeries::filer->boot_into_maintenance();
##
##
sub boot_into_maintenance()
{
    my $self = shift;
    my $SMALL_TIMEOUT = 30;
    my $LARGE_TIMEOUT = 180;

    # halt the filer, then bring it up into maintenance mode
    # ideally, we want to bring up the first one with traditional root volume,
    # and the second with flex root volume
    Tharn::say($self->{console_connection}, 'halt -f', '-eval_error', -timeout, 60);
    # in case of prompt, this must be immediate
    Tharn::say($self->{console_connection}, '0', '-blind');
    sleep(60);

    # make sure we're getting a fw prompt
        my @prompts = ('^CFE> ', '^LOADER\\*?>','^ok ', '^LOADER-(A|B)\*?>');

    my (undef,undef,$after,$prompt) =
            say($self->{console_connection},"",'-nocr','-timeout',120,'-maxidle',120,
                '-noprompt','-prompt',@prompts);

    # hit 'bye' to bring filer up into maintenance mode
    # The parse takes so long to recognize the string here that the window
    # for the ctrl-c closes, need to find a new angle here
    Tharn::say($self->{console_connection}, 'bye', "-blind");
    sleep(20);

    # hit <ctrl-c> to bring filer up into maintenance mode
    # for now, we will do this a few times to make sure, since the parse takes
    # too long for the timing window
    my $h = $self->{conn};
    for (my $i = 0; $i < 2; $i++)
    {
        Tharn::say($self->{console_connection}, $h->{irptchar},-nocr, "-blind");
        sleep(3);
    }

    Tharn::say($self->{console_connection}, $h->{irptchar},-nocr, "-wait", "Maintenance", "-ok");

    # here we should see the numbered menu
    # we want to send 4 and 4a (one trad root vol, one flex root vol)
    my $type = "4";
    if ($self->{filer_index} % 2)
    {
        $type = "4a";
    }
    # root volume type
    Tharn::say($self->{console_connection}, $type, "-wait", "system?", "-ok");
    # prompt for new file system
    Tharn::say($self->{console_connection}, "y", "-wait", "sure?", "-ok");
    # prompt to make sure wiping the file system out is ok

    # this is where we want the large timeout in case root vol creation
    # takes a long time
    Tharn::say($self->{console_connection}, "y", "-noprompt", "-timeout", $LARGE_TIMEOUT, "-wait", "Please enter the new hostname");
    # here we wait for some ...'s, vol creation, ifconfig, DHCP search
    # prompt for host name
    Tharn::say($self->{console_connection}, $self->{name}, "-wait", "virtual", "-ok");
    # vifs prompt
    Tharn::say($self->{console_connection}, "n", "-timeout", $SMALL_TIMEOUT, "-wait", "Network", "-swallow", "-noprompt");

    # prompt for primary ip address
    Tharn::say($self->{console_connection}, $self->{ip_addr_1}, "-timeout", $SMALL_TIMEOUT, "-wait", "netmask", "-swallow", "-noprompt");
    # prompt for primary subnet mask
    Tharn::say($self->{console_connection}, $self->{ip_mask_1}, "-timeout", $SMALL_TIMEOUT, "-wait", "partner", "-swallow", "-noprompt");
    # prompt for whether or not to take over partner interface during failover
    Tharn::say($self->{console_connection}, "y", "-timeout", $SMALL_TIMEOUT, "-wait", "taken over", "-swallow", "-noprompt");
    # prompt for primary interface name
    Tharn::say($self->{console_connection}, $self->{ip_int_1}, "-timeout", $SMALL_TIMEOUT, "-wait", "media type", "-swallow", "-noprompt");
    # prompt for primary interface speed/duplex
    Tharn::say($self->{console_connection}, $self->{ip_dup_1}, "-timeout", $SMALL_TIMEOUT, "-wait", "Network", "-swallow", "-noprompt");

    # check for follow up questions if this parameter is "auto"
    if ($self->{ip_dup_1} eq "auto")
    {
        $self->auto_entries();
    }

    # prompt for secondary ip address
    Tharn::say($self->{console_connection}, $self->{ip_addr_2}, "-timeout", $SMALL_TIMEOUT, "-wait", "netmask", "-swallow", "-noprompt");
    # prompt for secondary subnet mask
    Tharn::say($self->{console_connection}, $self->{ip_mask_2}, "-timeout", $SMALL_TIMEOUT, "-wait", "partner", "-swallow", "-noprompt");
    # prompt for whether or not to take over partner interface during failover
    Tharn::say($self->{console_connection}, "y", "-timeout", $SMALL_TIMEOUT, "-wait", "taken over", "-swallow", "-noprompt");
    # prompt for secondary interface name
    Tharn::say($self->{console_connection}, $self->{ip_int_2}, "-timeout", $SMALL_TIMEOUT, "-wait", "media type", "-swallow", "-noprompt");
    # prompt for secondary interface speed/duplex
    Tharn::say($self->{console_connection}, $self->{ip_dup_2}, "-timeout", $SMALL_TIMEOUT, "-wait", "flow", "-swallow", "-noprompt");

    # check for follow up questions if this parameter is "auto"
    if ($self->{ip_dup_2} eq "auto")
    {
        $self->auto_entries();
    }

    # now it wants to know whether or not to continue via web interface
    Tharn::say($self->{console_connection}, "n", "-timeout", $SMALL_TIMEOUT, "-wait", "gateway", "-swallow", "-noprompt");

    # prompt for default gateway
    my ($undef,$undef,$after,$prompt) = Tharn::say($self->{console_connection}, $self->{gateway}, "-timeout", $SMALL_TIMEOUT, "-wait", "IP", "-swallow", "-noprompt");

    # entering return here, end of prompt is as follows:
    # To allow /etc root access
    # to all NFS clients enter RETURN below.
    # Please enter the name or IP address of the administration host:
    my ($undef,$undef,$after,$prompt) = Tharn::say($self->{console_connection}, "", "-timeout", $SMALL_TIMEOUT, "-wait", "administration", "-noprompt");

    # Where is the filer located? []:
    my ($undef,$undef,$after,$prompt) = Tharn::say($self->{console_connection}, "\r", "-wait", "resolver", "-ok");

    # prompt for dns
    Tharn::say($self->{console_connection}, "y", "-timeout", $SMALL_TIMEOUT, "-wait", "domain", "-swallow", "-noprompt");
    # prompt for domain name
    Tharn::say($self->{console_connection}, $self->{domain_name}, "-timeout", $SMALL_TIMEOUT, "-wait", "server", "-swallow", "-noprompt");
    # prompt for dns server
    Tharn::say($self->{console_connection}, $self->{dns_server}, "-timeout", $SMALL_TIMEOUT, "-wait", "another", "-swallow", "-noprompt");
    # prompt for additional dns servers
    Tharn::say($self->{console_connection}, "n", "-timeout", $SMALL_TIMEOUT, "-wait", "NIS", "-swallow", "-noprompt");
    # prompt for nis
    Tharn::say($self->{console_connection}, "n", "-timeout", $SMALL_TIMEOUT, "-wait", "autosupport", "-swallow", "-noprompt");
    # This system will send event messages and weekly reports to Network
    # Appliance Technical Support. To disable this feature, enter "options
    # autosupport.support.enable off" within 24 hours. Enabling Autosupport
    # can significantly speed problem determination and resolution should a
    # problem occur on your system. For further information on Autosupport,
    # please see: http://now.netapp.com/autosupport/
    #    Press the return key to continue.
    Tharn::say($self->{console_connection}, "", "-timeout", $SMALL_TIMEOUT, "-wait", "return", "-noprompt");
    Tharn::say($self->{console_connection}, "", "-timeout", $SMALL_TIMEOUT, "-wait", "options", "-ok");
    Tharn::say($self->{console_connection}, "", "-blind");
}






## @name ::add_options()
##
## @description
## Funtion for adding options regarding autosuppor and cf takeover/giveback
##
## @synopsis
## $filer_obj=VSeries::filer->add_options();
##
##
sub add_options()
{
    my $self = shift;

    Tharn::say($self->{console_connection}, "options autosupport.support.enable off", "-blind");
    sleep(2);
    Tharn::say($self->{console_connection}, "options cf.giveback.auto.cifs.terminate.minutes 0", "-blind");
    sleep(2);
    Tharn::say($self->{console_connection}, "options cf.giveback.auto.enable off", "-blind");
    sleep(2);
    Tharn::say($self->{console_connection}, "options cf.giveback.auto.terminate.bigjobs on", "-blind");
    sleep(2);
    Tharn::say($self->{console_connection}, "options cf.giveback.auto.check.partner on", "-blind");
    sleep(2);
    Tharn::say($self->{console_connection}, "options cf.takeover.detection.seconds 15", "-blind");
    sleep(2);
    Tharn::say($self->{console_connection}, "options cf.takeover.on_disk_shelf_miscompare off", "-blind");
    sleep(2);
    Tharn::say($self->{console_connection}, "options cf.takeover.on_failure on", "-blind");
    sleep(2);
    Tharn::say($self->{console_connection}, "options cf.takeover.on_short_uptime on", "-blind");
    sleep(2);
}

## @name ::get_license_version()
##
## @description
## Determines which licensing scheme Ontap is using.
##
## Currently returns either "V1" or "V2".
##
## @synopsis
## $filer->get_license_version();
##
##
sub get_license_version() {

    my $self = shift;

    my $cmd = "license add ?";
    my $output = "";
    
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    if ($output =~ m/license codes must be 7 letters/i) {
        return "V1";
    } elsif ($output =~ m/license keys must be 28 or 42 letters in length/i) {
        return "V2";
    } else {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Unable to determine licensing scheme.";
        return $ERR_STATUS;
    }

}

## @name ::add_licenses()
##
## @description
## Function that enables necessary licenses for VSeries filers.  Licenses were
## taken from well known licenses used by the VSeries team.  Optionally takes
## in an array of strings that correspond to licenses.  Strings are listed
## right below and are fairly obvious in what they are for.
##
## @synopsis
## $filer->add_licenses(\@);
##
##
sub add_licenses(;\@)
{
    my $self = shift;
    my ($args_ref) = @_;
    my @args = ();
    if (defined($args_ref)) { @args = @$args_ref; }
    my $after = "";
    my $ret_v2 = undef;

    # Check which licensing scheme we're dealing with.
    my $license_version = $self->get_license_version();

    # If args is undefined we add all licenses to keep legacy code working
	if ((!(defined(@args))) && ($license_version eq "V1")) {
		@args = @VSERIES_TEST_LICENSE_SUBSET;
	} elsif ($license_version eq "V2") {
		if (!(defined(@args))) { @args = @VSERIES_TEST_LICENSE_SUBSET_V2; }
        $ret_v2 = $self->add_licenses_v2(\@args);
        return $ret_v2;
	} 

    foreach my $license (@args){
        switch($license){
            case /vseries/{
                # add vseries license first
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add ZQSHUVC");
                sleep(2);
            }
            case /nfs/{
                # add nfs license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add NMTWHKM");
                sleep(2);
            }
            case /cifs/{
                # add cifs license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add JKTYDIG");
                sleep(2);
            }
            case /^cluster$/{
                # add cluster license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add LLSJTEH");
                sleep(2);
            }
            case /http/{
                # add http license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add RSBDZOE");
                sleep(2);
            }
            case /fcp/{
                # add fcp license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add NQTHXGN");
                sleep(2);
            }
            case /isci/{
                # add iscsi license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add DNVUIBI");
                sleep(2);
            }
            case /syncmirror_local/{
                # add syncmirror_local license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add DFFHGCB");
                sleep(2);
            }
            case /cluster_remote/{
                # add cluster_remote license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add TTEOOLF");
                sleep(2);
            }
            case /^snaplock$/{
                # add snaplock license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add LLYCLVC");
                sleep(2);
            }
            case /^snaplock_enterprise$/{
                # add snaplock_enterprise license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add NWEDLVC");
                sleep(2);
            }
            case /nearstore_option/{
                # add nearstore license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add MFOWTQK");
                sleep(2);
            }
            case /a_sis/{
                # add a_sis license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add GAGSYJM");
                sleep(2);
            }
            else{
                return "Badly formed license name $license";
            }
        }

        # Check that none of the say calls returned an error. We parse on the success string:
        # "A ${license} [site] license has been installed"
        my $license_failed = 1;
        my @license_cmd_out = split (/\n/, $after);
        foreach my $line (@license_cmd_out) {
           #
           # We have been using 'cluster_remote' license in the older releases and
           # with the newer releases it's been identified as 'cf_remote'.
           # To make the fix generic, we are using a regex to handle both
           # cluster_remote and cf_remote.
           # Fixed it as part of BURT 694250.
           #
           $license = '(cluster_remote|cf_remote)' if ($license =~ m/cluster_remote/);

           if ( $line =~ /^A $license (site )?license has been installed/ ) {
              $license_failed = 0;
              last;
           }
        }
        if ( $license_failed ) {
           return "Failed to install license";
        }
    }

    # pause for 5 seconds
    sleep(5);
    return "";


}

## @name ::add_licenses_v2()
##
## @description
## Function that enables necessary V2 licenses for VSeries filers. Licenses were
## taken from well known licenses used by the VSeries team.  Optionally takes
## in an array of strings that correspond to licenses.  Strings are listed
## right below and are fairly obvious in what they are for.
##
## @synopsis
## $filer->add_licenses_v2(;\@);
##
##
sub add_licenses_v2(;\@) {

    my $self = shift;
    my ($args_ref) = @_;
    my @args = ();
    if (defined($args_ref)) { @args = @$args_ref; }
    my $after = "";

    # If args is undefined we add all licenses.
    if(!(defined(@args))){
        @args = @VSERIES_TEST_LICENSE_SUBSET_V2;
    }

    foreach my $license (@args){
        switch($license){
            case /v_storageattach|vseries/{
                # add vseries license first 
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add AXFTWDBBVAAAAAAAAAAAAAAAAAAA");
                sleep(2);
            }
            case /nfs/{
                # add nfs license 
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add APZXOWDUSAFDAAAAAAAAAAAAAAAA");
                sleep(2);
            }
            case /cifs/{
                # add cifs license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add MJXMPWDUSAFDAAAAAAAAAAAAAAAA");
                sleep(2);
            }
            case /fcp/{
                # add fcp license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add KYSQQWDUSAFDAAAAAAAAAAAAAAAA");
                sleep(2);
            }
            case /iscsi/{
                # add scsi license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add YDVBQWDUSAFDAAAAAAAAAAAAAAAA");
                sleep(2);
            }
            case /^snaplock$/{
                # add snaplock license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add ERFCUWDUSAFDAAAAAAAAAAAAAAAA");
                sleep(2);
            }
            case /^snaplock_enterprise$/{
                # add snaplock_enterprise license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add MRDIXDBBVAAAAAAAAAAAAAAAAAAA");
                sleep(2);
            }
            case /^snapmirror$/{
                # add snapmirror license
                (undef, undef, $after) = Tharn::say($self->{console_connection}, "license add UHMJSWDUSAFDAAAAAAAAAAAAAAAA");
                sleep(2);
            }
            else{
                return "Badly formed license name $license";
            }
        }

        # Check that none of the say calls returned an error. We parse on the success string:
        # "license add: successfully added license key "<license_key>""
        my $license_failed = 1;
        my @license_cmd_out = split (/\n/, $after);
        foreach my $line (@license_cmd_out) {
           if ( $line =~ /successfully added license key/ ) {
              $license_failed = 0;
              last;
           }
        }
        if ( $license_failed ) {
           return "Failed to install license";
        }
    }

    # pause for 5 seconds
    sleep(5);
    return "";

}

## @name ::reboot_filer()
##
## @description
## Method to handle the rebooting of a filer.  Expects the filer to be in a
## booted and properly configured state.  If the filer fails to boot in the
## specified time, panics, or restarts the boot process then an error will be
## reported.
##
## @synopsis
## $filer_obj->reboot_filer($$)
##
## @arg $max_time Required: Maximum time expected to reboot.
##
## @arg $wait_string Required: String you want to poll for in output.
##
##
sub reboot_filer($;$){
    my $self = shift;
    my ($max_time, $cifs_shutdown) = @_;
    my $timeout = 0;
    my $seen = 0;
    my $booted = 0;
    my $cmd = "reboot -t";
    my $any_line_prompt = '^.*\n';
    my ($match, $before, $after, $prompt) = "";
    my @norm_prompts = @{$self->{conn}->prompt_norm()};

    # Check to see if a cifs_shutdown value was passed in and that it's valid.
    if(defined($cifs_shutdown) && ($cifs_shutdown >= 0 || $cifs_shutdown <= 5)){
        $cmd = $cmd . $cifs_shutdown;
    }
    else{
        $cmd = $cmd . "0";
    }

    my $timedout_time = timeout2time($max_time);

    # Send command to begin rebooting
    if($self->{use_console} eq "true"){
        Tharn::say($self->{console_connection}, $cmd, "-swallow", "-blind");
    }
    else{
        Tharn::say($self->{connection}, $cmd, "-swallow", "-blind");
    }

    # Loop through listening to the output on the console.  If we see a panic
    # or the start of the boot sequence more than once report errors.  We break
    # out if we see the filer is fully booted or
    while($timeout == 0 && $booted == 0){
        ($match, $before, $after, $prompt) = Tharn::say($self->{console_connection}, "", "-noprompt", "-prompt", $any_line_prompt, '-maxidle', 300, "-timeout", 300);

        # Keep track of the amount of times we see the autoboot prompt, if we
        # see it more than once that's cause to fail.
        if($prompt =~ m/Starting AUTOBOOT press (.*) to abort/){
            $seen++;
            if($seen > 1){
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "Filer appears to be booting for a second time, this was not expected, aborting";
                return $ERR_STATUS;
            }

        }
        # This is the last piece of output that signals the end of the booting
        # process.  1st string applies to pre BR & 2nd applies to BR->.
        
	elsif($seen>0 && $prompt =~ m/(System rebooted after a reboot command|root logged in from console)/){
            $booted = 1;
        }

        # If at any point we see a panic that's cause to fail.
        elsif($prompt =~ m/PANIC:/){
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "FILER PANIC DURING BOOT PROCESS!  Please check logs to diagnose.";
            return $ERR_STATUS;
        }

        # Add a condition to check for normal prompts, because depending
        # on the OnTap release, the strings mentioned above do not apply.
        # This will serve as a catch all for booted & up filers.
        foreach my $norm_p (@norm_prompts) {
            if ($prompt =~ m/^$norm_p/) {$booted = 1;}
        }

        # Check to see if we've timed out during any loop.
        if(time() >= $timedout_time){
            $timeout = 1;
        }

    }

    # Once out of the loop check if we timed out, if so that's cause to fail.
    if($timeout >= 1){
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Filer did not successfully boot in expected time ($max_time), aborting";
        return $ERR_STATUS;
    }
	
    # The end of output from the booting process doesn't actually display a
    # normal prompt on the screen, this last say sends an newline across
    # which will cause the filer to print a prompt thus leaving the filer in
    # a state where it's ready for susequent commands.
    
    # Leaving this here as a breadcrumb.  This line creates BURT 362510 so
    # I hope we can leave it commented out, or at some point remove it.
    #($match, $before, $after, $prompt) = Tharn::say($self->{console_connection}, " ", "-swallow");

}


## @name ::get_disks_by_owner($)
##
## @description
## Takes in a System ID and returns all luns seen as owned by the passed in
## System ID on this filer.  Note that there are cases where we want the output
## of this command but don't want to use this filers system ID  Returns a hash
## with the keys being the serial numbers of the disks owned by the passed in
## system ID.  The values are a subsequent hash of pertinent disk info like
## primary and secondary paths,
##
## @synopsis
## $filer_obj=VSeries::filer->get_disks_by_owner($system_id);
##
## @arg $system_ID Required: The System of a given filer that you want to see
## what the filer get_disks_by_owner sees as owned by the given system ID.
## There are cases where you may not want to see what's owned by the filer but
## perhaps it's partner or "Not Owned".
##
##
sub get_disks_by_owner($)
{

    my $self = shift;
    my ($system_id) = @_;
    my %owned_disks = ();
    my %internal_disks = %{$self->{disks}};

    # had to add an extra check here for a leading zero in the sytem ID.
    foreach my $serial_number (keys(%internal_disks)){
        my $tmpstring = $internal_disks{$serial_number}->{owner};

        if($system_id =~ m/0*$tmpstring/){

          $owned_disks{$serial_number} = \$internal_disks{$serial_number};

        }

    }

    return (\%owned_disks);

}

## @name ::get_sysID()
##
## @description
## Function that returns the system ID of a given filer. The system ID of a
## filer is it's unique ID set in manufacturing.  If the case that no system
## ID is present, or discover_sysID hasn't been called yet then sys_id will
## be set to undef.
##
## @synopsis
## $filer_obj=VSeries::filer->get_sysID();
##
##
sub get_sysID()
{
    my $self = shift;
    return ($self->{sys_id});


}



## @name ::discover_sysID()
##
## @description
## Function that will run sysconfig on a given filer and pull out the system ID
## The internal variable sys_id will get set with the value.  The system ID of
## a filer is it's unique ID that is set in manufacturing.
##
## @synopsis
## $filer_obj=VSeries::filer->discover_sysID();
##
##
sub discover_sysID(){

    my $self = shift;
    my (undef, undef, $status) = "";
    if($self->{use_console} eq "true"){
        (undef, undef, $status) = Tharn::say($self->{console_connection}, "sysconfig", "System ID: ", -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $status) = Tharn::say($self->{connection}, "sysconfig", "System ID: ", -swallow, "-timeout", 60);
    }

    ($status, undef) = split(/ /, $status);
    $self->{sys_id} = $status;

}


## @name ::get_error_state()
##
## @description
## Returns the value of the error flag
##
## @synopsis
## $filer_obj=VSeries::filer->get_error_state();
##
##
sub get_error_state(){
    my $self = shift;
    return $self->{error};
}


## @name ::get_error_msg()
##
## @description
## Returns the error message for a given filer.  This is useful for reporting
## failures or unexpected behavior for a given test.
##
## @synopsis
## $filer_obj=VSeries::filer->get_error_msg();
##
##
sub get_error_msg(){
    my $self = shift;
    return $self->{errmsg};
}

## @name ::get_info_msg()
##
## @description
## Returns the info message array for a given filer.  This is useful for
## reporting info or mentionable behavior for a given test.
##
## @synopsis
## $info_msg_ref = VSeries::filer->get_info_msg();
##
##
sub get_info_msg(){
    my $self = shift;
    return \@{$self->{info_msg}};
}


## @name ::disk_assign($;@)
##
## @description
## Method that takes a serial number of a given disk, converts the serial to a
## primary path for this specific filer and attempts to take ownership of the
## disk.  Will return an error string with failure specific information,
## either from the command output from the filer, or created here in the case
## that we can't find a disk with a given serial number on this filer.  Returns
## the empty string in success cases.  This method does not actually query the
## filer to determine if the disk is seen, it expects discover_disks() to have
## already been called for the given filer.  It is worth noting here that if
## the -n flag is intended on being used then the value for $disk_serial_param
## should be set to the empty string "".
##
## @synopsis
## $filer_obj=VSeries::filer->disk_assign($disk_serial, @option_flags);
##
## @arg $disk_serial Required: Serial number of the disk you want the filer to
## take ownership of.
## @arg \@option_flags Optional: Flags that can be used for the disk assign
## command.  Caller is expected to know what flags can be used alone or
## together.
##
##
sub disk_assign($;@){
    my $self = shift;
    my ($disk_serial_param, @optional_flags) = @_;
    my $disk_name = "";
    my $status = "";
    my $verify = "";
    my $success = 0;


    # This set of variables are used in conjunction with command line flags for
    # disk assignment that may be passed in through @optional_flags.  These
    # will be used in the verification process to make sure the command behaved
    # as expected.

    # This corresponds to the -T flag for storage types
    my $verify_storage_type = "";

    # This corresponds to the -n flag for selecting 1 or more disks at a time.
    my $verify_count = 1;

    # This corresponds to the -c flag
    my $verify_checksum = "";

    my $verify_cmd = "disk show -v";


    # Check to see if a serial number was passed in or not, it may be blank
    # intentionally if the -n flag is intended to be used here.
    if($disk_serial_param ne ""){
        foreach my $disk_serial (keys(%{$self->{disks}})){
            if($disk_serial eq $disk_serial_param){
                $disk_name = $self->{disks}->{$disk_serial}->{primary};
            }
        }

        # If the name is blank then we didn't find the serial number we were
        # looking for and will return an error.
        if($disk_name eq ""){
            $status = "Could not find disk with Serial Number " . $disk_serial_param . " on filer " . $self->{name};
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = $status;
            return $ERR_STATUS;
        }
    }


    my $cmd = "disk assign $disk_name ";

    # This is where we break apart the optional flags, add them to the command
    # string and populate the correct variables for verification after the
    # command is run.
    foreach my $option(@optional_flags){
        $cmd = $cmd . " " . $option;

        switch($option){
            case /-T/{
                $verify_storage_type = $option;
                $verify_storage_type =~ s/-T +//;
            }
            case /-n/{
                $verify_count = $option;
                $verify_count =~ s/-n +//;
            }
            case /-c/{
                $verify_checksum = $option;
                $verify_checksum =~ s/-c +//;
            }
        }
    }

    # Depending on if we use the console for the command or not we also add the
    # prompt to handle the maint mode prompt over the console since normal IP
    # connectivity isn't up while you're in maint mode.
    if($self->{use_console} eq "true"){
        (undef, undef, $status) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $status) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    # For verification purposes we want to remember what the filer object saw
    # before the disk assign, then discover what it currently sees and compare.
    my %old_disks = %{$self->get_disks_by_owner($self->{sys_id})};

    $self->discover_disks();

    if($self->get_error_state()){
        return $ERR_STATUS;
    }

    my %new_disks = %{$self->get_disks_by_owner($self->{sys_id})};

    # We'll start by comparing their serial numbers which are returned
    # from the get_disks_by_owner call in hash form.
    my @old_disk_serials = keys(%old_disks);
    my @new_disk_serials = keys(%new_disks);

    # This array will contain the serial numbers of disks that were
    # not owned at the start of this disk assign call but are now
    # owned.
    my @verify_disk_serials = ();

    # This array will contain all the serials that are in both the
    # new and old disk serial array, this should be equal to all the old
    # disk serials to prove what we owned before the assign is still
    # owned now.
    my @verify_old_disk_serials = ();

    my $matched = 0;


    # Loop through the serial arrays looking for what disks are
    # newly owned.
    for(my $i = 0; $i < scalar(@new_disk_serials); $i++){
        $matched = 0;
        for(my $j = 0; $j < scalar(@old_disk_serials); $j++){
            if($new_disk_serials[$i] eq $old_disk_serials[$j]){
                $matched = 1;
                push(@verify_old_disk_serials, $new_disk_serials[$i]);

            }
        }
        if($matched == 0){
            push(@verify_disk_serials, $new_disk_serials[$i]);

        }

    }


    # This begins the verification part of the method.  We'll start by checking
    # if the number of newly owned serials matches the amount we expected from
    # above.  This defaults to 1 unless -n was specified as an optional flag.
    if(scalar(@verify_disk_serials) != $verify_count){
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Unable to verify success of disk assign mismatch on verify count($verify_count) and number of new disks owned (" . scalar(@verify_disk_serials) . ")";
        return $ERR_STATUS;

    }

    # Check to see if a lun serial was specified and in that case the count in
    # the @verify_disk_serials should be 1 and that element should be equal to
    # the serial number we were passed in.
    if($disk_serial_param ne ""){
        if($disk_serial_param ne $verify_disk_serials[0]){
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "Mismatch in serial number attempted to assign and actual serial number that got assigned. Attempted: $disk_serial_param Assigned: $verify_disk_serials[0]";
            return $ERR_STATUS;
        }
    }


    # Now we'll check that all the disks we owned before are also still owned by
    # the filer.
    @verify_old_disk_serials = sort(@verify_old_disk_serials);
    @old_disk_serials = sort(@old_disk_serials);

    my $result = VSeries::utils::compare_arrays(\@verify_old_disk_serials, \@old_disk_serials);
    if($result == $ERR_STATUS){

        my $triage_string = "PREVIOUSLY OWNED DISKS BEFORE ASSIGN           PREVIOUSLY OWNED DISKS AFTER ASSIGN\n";
        my @bigger_array = ();
        my @smaller_array = ();
        if(scalar(@verify_old_disk_serials) >= scalar(@old_disk_serials)){
            @bigger_array = @verify_old_disk_serials;
            @smaller_array = @old_disk_serials;
        }
        else{
            @bigger_array = @old_disk_serials;
            @smaller_array = @verify_old_disk_serials;
        }

        for(my $i = 0; $i < scalar(@bigger_array); $i++){
            if($i < scalar(@smaller_array)){

                $triage_string = $triage_string . $bigger_array[$i] . "         " . $smaller_array[$i] . "\n";
            }
            else{
                $triage_string = $triage_string . $bigger_array[$i] . "\n";
            }

        }

        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Disk(s) previously owned are not showing as owned after disk assign\n" . $triage_string;
        return $ERR_STATUS;

    }

    # Now we loop through getting the disk specific info and verify the pieces
    # of information we care about based on optional flags passed in.
    foreach my $serial (@verify_disk_serials){
        my %new_disks_info = %{$self->get_disk_info($serial)};

        # If the -T flag was passed in we'll verify the type here.
        if($verify_storage_type ne ""){
            if($new_disks_info{storage_type} ne $verify_storage_type){
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "Unable to verify success of disk assign mismatch on verify storage type($verify_storage_type) and disk $serial (" . $new_disks_info{storage_type} . ")";
                return $ERR_STATUS;

            }

        }

        # If -c was passed in here we'll check it if the storage type is LUN
        # otherwise the cli actually ignores the -c flag.
        if($verify_checksum ne "" && $verify_storage_type eq "LUN"){
            if($new_disks_info{chksum} !~ m/$verify_checksum/i){
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "Unable to verify success of disk assign mismatch on verify checksum($verify_checksum) and disk $serial (" . $new_disks_info{chksum} . ")";
                return $ERR_STATUS;
            }
        }


    }

    return $GOOD_STATUS;

}






## @name ::disk_unassign($)
##
## @description
## Takes in a serial number for a disk, looks up the corresponding name on this
## given filer and removes it.  If successfull it will return $GOOD_STATUS,
## otherwise it will return $ERR_STATUS and set the internal error flag.
## It does call discover_disks() pre & post lun unassign to update the Filer
## object internal disks hash "$self->{disks}".
##
## @synopsis
## $status = VSeries::filer->disk_unassign($disk_serial);
##
## @arg $disk_serial Required: Serial number of the disk you want the filer to
## remove ownership of.
##
sub disk_unassign($){
    my $self = shift;           # Filer object shifted in.
    my ($lun_serial) = @_;      # Lun serial passed in.
    my $lun_name = "";          # A lun name.
    my $status = "";            # Holds say() output.
    my $flag    = 0;            # A generic flag.
    my @not_owned_serials = (); # Hopefully just one unowned serial.

    # Ensure we know the system ID.
    $self->discover_sysID();

    # Learn all disk ownership information before the unassign.
    $self->discover_disks();

    # Errors from past call?
    if ($self->get_error_state() == $ERR_STATUS) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = $self->get_error_msg();
        return($ERR_STATUS);
    }

    # Start walking the internal disks hash
    foreach my $disk_serial (keys(%{$self->{disks}})){
        if($disk_serial eq $lun_serial){
            $lun_name = $self->{disks}->{$disk_serial}->{primary};
        }
    }

    # This means the lun_serial was not found in the internal disks hash.
    if($lun_name eq ""){
        $status = "Could not find disk with Serial Number " . $lun_serial . " on filer " . $self->{name};
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = $status;
        return $ERR_STATUS;
    }

    # This is the command needed to remove luns.
    my $cmd = "disk remove -w $lun_name";

    # Either send the cmd over the console or standard connection (ususally rsh).
    if($self->{use_console} eq "true"){
        (undef, undef, $status) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $status) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    # For verification purposes we want to remember what the filer object saw
    # before the disk remove, then discover what it currently sees and compare.
    my %disks_pre_unassign = %{$self->get_disks_by_owner($self->{sys_id})};

    # Learn all disk ownership information post unassign.
    $self->discover_disks();

    # Errors from past call?
    if ($self->get_error_state() == $ERR_STATUS) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = $self->get_error_msg();
        return($ERR_STATUS);
    }

    # Learn current state of internal disks post unassign.
    my %disks_post_unassign = %{$self->get_disks_by_owner($self->{sys_id})};

    # Verify our disk info is now accurate:
    foreach my $pre_serial (keys %disks_pre_unassign) {
        foreach my $post_serial (keys % disks_post_unassign) {
            if ($pre_serial eq $post_serial) {
                # Set a flag
                $flag = 1;
            }
        }
        if ($flag == 0) {
            # This means the serial used to be owned by this filer
            # but no longer is, so store it and compare it later.
            push(@not_owned_serials, $pre_serial);
        } else {
            # Make sure we reinit the flag for the next iteration:
            $flag = 0;
            }
            }

    # Are there more than one serial numbers in our array?
    if (scalar @not_owned_serials > 1) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "disk_unassign: expected to unassign 1 disk, but found several.";
        return($ERR_STATUS);
    }

    # Ensure our serial we found is the same as the one passed in.
    if ($not_owned_serials[0] ne $lun_serial) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "disk_unassign: serial unassigned $not_owned_serials[0] is not equal to the serial passed in $lun_serial.";
        return($ERR_STATUS);
    }

    return($GOOD_STATUS);
}


## @name ::get_name()
##
## @description
## Returns the name of a given filer.
##
## @synopsis
## $filer_obj=VSeries::filer->get_name();
##
##
sub get_name(){
    my $self = shift;
    return($self->{name});
}


## @name ::check_cf_state()
##
## @description
## Simple method to check the CF state of a partner filer.  This is useful
## to make sure a filer doesn't have adverse affects from various actions.  It
## is effectively pass through to VSeries::cluster but we wanted a way to
## interface with filer for this functionality without having to scrub
## legacy code to have to use filer.pm at this point.  Return values are
## consistent with those of VSeries::cluster::get_cf_state()
##
## @synopsis
## $filer_obj=VSeries::filer->get_cf_state();
##
##
sub check_cf_state(){
    my $self = shift;
    my $state = VSeries::cluster::get_cf_state($self->{connection});
    return $state;
}


## @name ::get_use_aggr()
##
## @description
## Returns the use_aggr flag to determine if a filer used aggregates or not.
##
## @synopsis
## $filer_obj=VSeries::filer->get_use_aggr();
##
##
sub get_use_aggr(){
    my $self = shift;
    return $self->{use_aggr};
}


## @name ::discover_disks()
##
## @description
## Method that discovers/updates the filer objects internal structures keeping
## track of disk information.  Relies on the "disk show -v" command.  Creates
## an internal data structure of a hash of hashes to keep track of the disks.
## The key to the first hash is the disk's serial number.  The keys to the
## corresponding hash is the attributes of a given disk.
##
## @synopsis
## $filer_obj=VSeries::filer->discover_disks();
##
##
sub discover_disks(;@){
    my $self = shift;
    my (@faked_input) = @_;
    my $output = undef;
    # Get the output of disk show -v and split into separate lines
    my $cmd = "disk show -v";
    my $returned_prompt = "";
    my $system_id = "";
    if($self->{use_console} eq "true"){
        (undef, undef, $output, $returned_prompt) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }


    if(defined($output)){
        my @disk_output = ();
        if(defined(@faked_input)){
            @disk_output = @faked_input;
        }
        else{
            @disk_output = split(/\n/, $output);
        }
        my @disk_serials = '';

        # Needed to add this here due to the output for disk show -v having the
        # extra two lines of output if run from maint mode.  One of the lines
        # contains the system ID which is actually very important to grab
        # if you're booting a filer for the first time and creating a root vol.
        my $header_size = 2;
        foreach my $line (@disk_output){
            # if we have the system ID in the header store it and then adjust
            # the size of the header we'll crop off the output to then parse
            # for actual disk info.
            if($line =~ m/^LOCAL SYSTEM ID:/i){
                (undef, $self->{sys_id}) = split(/: +/, $line);
                $self->{sys_id} =~ s/[^0-9]//g;
                $header_size = 4;
            }

        }

        # Chop off the header of the output
        @disk_output = @disk_output[$header_size..$#disk_output];


        my ($disk_name, $disk_serial, $disk_pool, $disk_owner, $disk_chksum, $disk_home) = '';

        # Hash that will contain the information about a specific disk
        my %disk_info = ();
        # Hash that will contain the hashes of information about each specific
        # disk.  The Key is the serial number of the disk.
        my %internal_disks = ();

        # Loop through each line of output tokenizing each piece of disk
        # information.
        foreach my $output_line (@disk_output){

            # Adding catch for case where serial number and checksum aren't
            # separated by a whitespace
            if($output_line =~ m/(Block|Zone)/i){
                $output_line =~ s/(Block|Zone)/ \1/i;
            }
            
            # Determine if there is a space between the hostname and the system-id
            # in the Owner column, this space can occur if the hostname is shorter
            # than 9 characters
            # This behavior has been observed in the pNATE customer environment.
            # >>lando2#rsh CMD 'disk show -v'
            # <<lando2#rsh|  DISK       OWNER                  POOL   SERIAL NUMBER           HOME                     CHKSUM\n
            # <<lando2#rsh|------------ -------------          -----  -------------           -------------            ------\n
            # <<lando2#rsh|svcswb2:18.126L1 lando1    (1573753848)    Pool0  7804E34006C           lando1    (1573753848)  Block\n
            # <<lando2#rsh|svcswb1:18.126L2 lando1    (1573753848)    Pool0  7804E34006B           lando1    (1573753848)  Block\n
            # <<lando2#rsh|svcswb1:18.126L3 lando1    (1573753848)    Pool0  7804E34006A           lando1    (1573753848)  Block\n

            my $separated_system_id = 0;
            if($output_line =~ /\w+\s+\(/i) {
                $separated_system_id = 1;
            }

            # I decided to split on white space so I needed to handle the case
            # where the owner is "Not Owned" separately.
            my @output_tokens = split(/ +/, $output_line);
            my $token_counter = 0;

            # The first token will always be the name of the disk.
            $disk_name = $output_tokens[$token_counter++];

            # Now we need to look ahead since if the disk isn't owned it will
            # show up as "Not Owned" which has a whitespace in it, resulting
            # in more tokens and breaking the rest of the parsing.  If the
            # current token is "Not" and the next token is "Owned" then the
            # disk must be unowned and we record that information properly and
            # move on.  Otherwise we simpley record the owner information.
            if($output_tokens[$token_counter] =~ m/not/i && $output_tokens[$token_counter + 1] =~ m/owned/i){
                $disk_owner = "Not Owned";
                $token_counter = $token_counter + 2
            }
            elsif($output_tokens[$token_counter] =~ m/(NONE|FAILED)/i || $output_tokens[$token_counter] =~ m/^pool/i){
                $disk_owner = "-";
            }
            elsif($separated_system_id) {
                $disk_owner = $output_tokens[++$token_counter];
                $token_counter++;
            }
            else{
                $disk_owner = $output_tokens[$token_counter++];
            }


            # Here's where we work around the presence of the "HOME" column
            # column being added after the "OWNER" column in BoilerMaker 7-mode
            # if our current token matches the known values for pool
            # information, either "NONE" or "FAILED" or "PoolX" where X is a
            # number, then we infer that we must have the HOME token and reord
            # it and the the pool token next.  Otherwise we must have the pool
            # token already and record that in its proper place.  This way we
            # stay flexible between BoilerMaker and previous releases.
            if($output_tokens[$token_counter] !~ m/(NONE|FAILED)/i && $output_tokens[$token_counter] !~ m/^pool/i ){
                $disk_home = $output_tokens[$token_counter++];
            }

            if($output_tokens[$token_counter] =~ m/(NONE|FAILED)/i || $output_tokens[$token_counter] =~ m/^pool/i ){
                $disk_pool = $output_tokens[$token_counter++];
            }
            else{
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "ERROR in discover_disks.  Expected to encounter a token for pool information.\nToken is: $output_tokens[$token_counter]\nLine is: $output_line";
            }

            # There are cases where we might get a space in the serial number (or) checksum might have got
            # clubbed with the serial number (or) a space between serial number and checksum field with '-'.
            # Also, we might get 'DR Home' or 'HOME' fields in between serial number and checksum fields.
            # To validate all the above conditions we have modified the logic.
            # 1) If disk_show output has 'DR Home' or 'HOME' fields then we will look out for a opening and
            #    closing brace in the next token_counter i.e. $token_counter+1. Currently token_counter will
            #    hold the index value of serial number. If the next counter matches with the 'DR Home' or
            #    'HOME' fields (the value of the fields will look like 'vgv6240f100b(1873775257)') then we
            #    can assume that the serial number is a single string without any space and we will
            #    initialize disk_serial with the value of the token_counter.
            #    Say for example, output_tokens array has the following values
            #    vgci9148s171:1-13.126L36 vgv6240f100b(1873775257) Pool0 500FF1D0024 vgv6240f100b(1873775257) Block
            #    As stated above, 'DR Home' or 'HOME' field is existing between serial number and
            #    checksum fields and as a result of it 'token_counter + 1' matches with the opening
            #    and closing brace i.e. if block gets executed and disk_serial will be populated with
            #    the value of '500FF1D0024'.
            # 2) If the serial number has a space in it then the if block will fail because the next counter
            #    will not match either with opening/closing brace or with the checksum value i.e.
            #    block|zoned|-.
            #    Else block will get executed and the disk_serial will be populated with the values of
            #    current token_counter and the incremented index value of the token_counter.
            #    Say for example, output_tokens array has the following values
            #    vgci9148s171:1-13.126L36 vgv6240f100b(1873775257) Pool0 50 0FF1D0024 Block
            #    As stated above, serial number has a space in it and so else block will get executed and
            #    disk_serial will be populated with the value of '50 0FF1D0024' i.e.
            #    $output_tokens[$token_counter] => holds '50' and
            #    $output_tokens[$token_counter + 1] => holds '0FF1D0024'.
            # 3) If the checksum field has a value of '-' without any space in it and if the next
            #    token_counter of serial number is matching with the '-' then we can assume that the
            #    serial number is a single string without any space and we will initialize disk_serial with
            #    the value of the token_counter i.e. if block gets executed.
            #    Say for example, output_tokens array has the following values
            #    12c.02.11 vgv6240f100a(1873760986) Pool0 PDWBUHKE -
            #    As stated above, checksum field has a value of '-' without any space in it and so if block
            #    will get executed and disk_serial will be populated with the value of 'PDWBUHKE' i.e.
            #    $output_tokens[$token_counter] => holds 'PDWBUHKE'.
            # 4) If there is a space between the hostname and the system-id in the Owner column then
            #    'separated_system_id' flag will be set to '1' and in this case we need to increment
            #    the 'token_counter' to '2' to match the opening and closing brace of the system-id.
            #    Say for example, output_tokens array has the following values
            #    svcswb2:18.126L1 lando1 (1573753848) Pool0 7804E34006C lando1 (1573753848) Block
            #    As stated above, there is a space between the hostname and the system-id in the
            #    Owner column and as a result of it 'token_counter + 2' matches with the opening
            #    and closing brace i.e. if block gets executed and disk_serial will be populated with
            #    the value of '7804E34006C'.

            if (($output_tokens[$token_counter+1] =~ m/\(/ && $output_tokens[$token_counter+1] =~ m/\)/) ||
                ($separated_system_id == 1 && $output_tokens[$token_counter+2] =~ m/\(/ && $output_tokens[$token_counter+2] =~ m/\)/) ||
                ($output_tokens[$token_counter+1] =~ /block|zoned|\-/i)) {
                $disk_serial = $output_tokens[$token_counter++];
            }
            else{
                $disk_serial = $output_tokens[$token_counter] . " " . $output_tokens[$token_counter + 1];
                $token_counter = $token_counter + 2;
            }

            # If there is no space between serial number and checksum fields then we will retrieve
            # the serial number alone by using the following regex: '(.*?)(Block|Zoned)'.
            if ($disk_serial =~ m/(.*?)(Block|Zoned)/i) {
                $disk_serial = $1;
            }

            if ($disk_serial =~ /\t/) {
                my @temp = split(/\t/, $disk_serial);
                $disk_serial = $temp[0];
            }

            # Lastly we record the checksum information.
            # Fix for BURT 562063
            if ($output_line =~ /block/i) {
                $disk_chksum = "Block";
            } elsif ($output_line =~ /zoned/i) {
                $disk_chksum = "Zoned";
            } else {
                $disk_chksum = "-";
            }

            # Match on open and closed parenthesis, then remove the actual
            # parenthesis and keep the system ID of the disks owner.
            if($disk_owner =~ m/\(.*\)/){
                $disk_owner = substr($&, 1, -1);
            }

            # Build the internal hash of hashes where each top level is keyed
            # with a disk serial number and the value is the hash with the
            # information about the disk.
            $internal_disks{$disk_serial}{primary} = $disk_name;
            $internal_disks{$disk_serial}{secondary} = "-";
            $internal_disks{$disk_serial}{pool} = $disk_pool;
            $internal_disks{$disk_serial}{owner} = $disk_owner;
            $internal_disks{$disk_serial}{chksum} = $disk_chksum;
            $internal_disks{$disk_serial}{adapter} = "-";

        }

        # Place current version of disk info as our latest internal hash and
        # then make the call to fill in the secondary paths.
        $self->{disks} = {%internal_disks};

        $self->discover_disk_paths();

        # In versions of ontap prior to IronCity (IC) we didn't suppor the -T
        # flag in the storage show disk command.  That's what
        # discover_storage_type uses and adds to the storage type to the disk
        # hash.  This prevents us from calling that method improperly on
        # earlier releases.  The chronological order of releases from oldest to
        # newest is AnchorSteam(AS), TsingTao(TT), GordonBiersch(GB),
        # IronCity(IC), BoilerMaker(BM), RollingRock(RR)
        if($self->revision_checker("AS") != $GOOD_STATUS && $self->revision_checker("TT") != $GOOD_STATUS && $self->revision_checker("GB") != $GOOD_STATUS ){
            $self->discover_storage_type();

        }

    }
    else{
        # If Tharn:;say returned a null value that means the command timed
        # out and we will record that information in the error structures
        # so that a test calling this can determine if it should continue or
        # not.
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Filer did not send a response to discovering disks";
        return ($ERR_STATUS);
    }


}


## @name ::discover_disk_paths()
##
## @description
## Method to discover the secondary path of disks.  Since this information
## doesn't come from disk show -v and may be usefull in other places it is its
## own routine.  Searches the internal disk hash created by discover_disks()
## and adds the information to the proper disk's hash.
##
## @synopsis
## $filer_obj=VSeries::filer->discover_disk_paths();
##
##
sub discover_disk_paths(){
    my $self = shift;
    my $cmd = "storage show disk -p all";
    my $output = undef;
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }
    if(defined($output)){
        my @storage_output = split(/\n/, $output);
        # Chop off the header
        @storage_output = @storage_output[2.. $#storage_output];
        my ($primary, $secondary, $adapter);

        foreach my $output_line(@storage_output){
        ($primary, undef, $secondary, undef, undef, undef, $adapter) = split(/ +/, $output_line);
            foreach my $serial (keys (%{$self->{disks}})){
                if($self->{disks}->{$serial}->{primary} eq $primary){
                    $self->{disks}->{$serial}->{secondary} = $secondary;
                    $self->{disks}->{$serial}->{adapter} = $adapter;
                }
            }
        }
    }
    else{
        $self->{error} = 1;
        $self->{errmsg} = "Filer did not send a response in discover disk paths";
    }
    return 0;

}

## @name ::discover_storage_type()
##
## @description
## Method to discover the storage type of disks.  Since this information
## doesn't come from disk show -v and may be usefull in other places it is its
## own routine.  Searches the internal disk hash created by discover_disks()
## and adds the information to the proper disk's hash.
## Attributes added to the disk hash are {shelf},{bay},{vendor},{model},{rev},
## and {storage_type}.
##
## @synopsis
## $filer_obj=VSeries::filer->discover_storage_type();
##
## @arg @optional_input Optional: an array of optional output to override the
##                                returned output of storage show disk -T -x
##                                used as a way to fake various erroneous
##                                output scenarios seen on filers.
##
##
sub discover_storage_type(;@){
    my $self = shift;
    my (@optional_input) = @_;
    my $cmd = "storage show disk -T -x";
    my $after = undef;

    # Remove this outer IF statement once filers are available to test with
    if(@optional_input == undef){
        if($self->{use_console} eq "true"){
            (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
        }
        else{
            (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
        }
    }
    if(defined($after) || defined(@optional_input)){
        my @storage_output = undef;

        # If optional input wasn't passed in then we'll have say output here so
        # split that into storage output otherwise set storage output to be
        # the optional input passed in.
        if(defined($after)){
            @storage_output = split(/\n/, $after);
        }
        else{
            @storage_output = @optional_input;

        }
        # Chop off the header
        @storage_output = @storage_output[2.. $#storage_output];
        my @line_tokens = ();

        # The output for the storage show disk -T -x command looks like this:
        # DISK                  SHELF BAY SERIAL           VENDOR   MODEL      REV TYPE
        # --------------------- --------- ---------------- -------- ---------- ---- ------
        # 0a.16                   1    0  V5Y6GGAA         NETAPP   X274_HPYTA NA03 FCAL
        # vss_b3800_2:0.126L0     -    -  751373110FE      IBM      2107900    0000 LUN
        #
        # Note the two different types of storage, disks and luns.  One thing
        # that can happen is serial numbers sometimes can contain whitespace
        # so the parsing of a given line needs to potentially factor that in.
        # So rather than purely splitting on whitespace and assuming all things
        # match up.  We tokenize the line on whitespace and then if the column
        # that corresponds to serial numbers doesn't match against a known
        # serial number we then try concatenating it with the next token to
        # see if that matches, if it does then we resolved the whitespace issue
        # and put the rest of the columns in their proper variables.  If
        # neither of those attempts match we throw an error for an unseen disk
        # serial number.

        foreach my $output_line(@storage_output){
            @line_tokens = split(/ +/, $output_line);
            my $token_index = 0;

            foreach my $serial (keys (%{$self->{disks}})){
                # Due to potential whitespace in serial number we match on
                # either primary or secondary paths.  Populate shelf and bay
                # and then check the serial number and resolve the whitespace
                # if we have to.
                if($self->{disks}->{$serial}->{primary} eq $line_tokens[$token_index] || $self->{disks}->{$serial}->{secondary} eq $line_tokens[$token_index]){
                    $self->{disks}->{$serial}->{shelf} = $line_tokens[++$token_index];
                    $self->{disks}->{$serial}->{bay} = $line_tokens[++$token_index];

                    if($serial eq $line_tokens[++$token_index]){
                        $self->{disks}->{$serial}->{vendor} = $line_tokens[++$token_index];
                        $self->{disks}->{$serial}->{model} = $line_tokens[++$token_index];
                        $self->{disks}->{$serial}->{rev} = $line_tokens[++$token_index];
                        $self->{disks}->{$serial}->{storage_type} = $line_tokens[++$token_index];

                    }
                    elsif($serial eq $line_tokens[$token_index] . " " . $line_tokens[++$token_index]){
                        $self->{disks}->{$serial}->{vendor} = $line_tokens[++$token_index];
                        $self->{disks}->{$serial}->{model} = $line_tokens[++$token_index];
                        $self->{disks}->{$serial}->{rev} = $line_tokens[++$token_index];
                        $self->{disks}->{$serial}->{storage_type} = $line_tokens[++$token_index];
                    }
                    else{
                        $self->{error} = $ERR_STATUS;
                        $self->{errmsg} = "ERROR: Disk Serial Number ($serial) and line $output_line didn't resolve in tokenizing in discover storage type";
                        return $ERR_STATUS;
                    }

                }
            }
        }
    }
    else{
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Filer did not send a response in discover storage type";
        return $ERR_STATUS;
    }
    return $GOOD_STATUS;

}


## @name ::two_path_check()
##
## @description
## Method that checks to make sure that if a disk is owned we know about both
## it's paths.  Meaning if a disks owner is something besides "Not Owned" then
## the secondary path has to be something besides "-".  Returns $GOOD_STATUS if
## that's the case and returns $ERR_STATUS otherwise.
##
## @synopsis
## $filer_obj=VSeries::filer->two_path_check();
##
## @arg $filer_name Optional: A filer name that if specified, the caller only
##                            wants the two path check to be performed on
##                            disks/LUNs owned by that filer.
##
##
sub two_path_check(;$){
    my $self = shift;
    my ($filer_name) = @_;

    if(!defined($filer_name)){
    foreach my $serial (keys (%{$self->{disks}})){
        if($self->{disks}->{$serial}->{owner} ne "Not Owned" && $self->{disks}->{$serial}->{secondary} eq "-"){
            return $ERR_STATUS;
        }
    }
    }
    else{
        foreach my $serial (keys (%{$self->{disks}})){
            if($self->{disks}->{$serial}->{owner} eq $filer_name && $self->{disks}->{$serial}->{secondary} eq "-"){
                return $ERR_STATUS;
            }
        }
    }

    return $GOOD_STATUS;

}


## @name ::get_disk_info($)
##
## @description
## Method that takes in a disks serial number, and returns a reference to the
## hash of that disk's information.  If it cannot find the disk for that serial
## number it returns -1
##
## @synopsis
## $filer_obj=VSeries::filer->get_disk_info($disk_serial);
##
## @arg $disk_serial Required: Serial number of the disk you want the hash of.
##
##
sub get_disk_info($){
    my $self = shift;
    my ($serial_number) = @_;

    if(! exists $$self{disks}->{$serial_number}){
        return -1;
    }

    return $$self{disks}->{$serial_number};

}



## @name ::vol_status(;$)
##
## @description
## Method that is effectively a pass through to run the "vol status" command
## with an optional flag.  Since the output can vary greatly parsers for the
## various output exist in VSeries::utils to keep this method flexible and
## simple.
##
## @synopsis
## $filer_obj=VSeries::filer->vol_status($command_flag);
##
## @arg $command_flag Optional: flag for the command that the caller wants
##
##
sub vol_status(;$){
    my $self = shift;
    my ($cmd_flag) = @_;
    my $cmd = "vol status";
    if(defined($cmd_flag)){
        $cmd = $cmd . " " . $cmd_flag;
    }
    my $output = undef;
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    if(defined($output)){
        # strip out all \t's because they cause funky irregularities in output.
        $output =~ s/\t/ /g;
        my @output_array = split(/\n/, $output);
        return \@output_array;
    }
    else{
        $self->{error} = 1;
        $self->{errmsg} = "Connection timeout during vol status";
        return 0;
    }


}


## @name ::aggr_status(;$)
##
## @description
## Method that is effectively a pass through to run the "aggr status" command
## with an optional flag.  Since the output can vary greatly parsers for the
## various output exist in VSeries::utils to keep this method flexible and
## simple.
##
## @synopsis
## $filer_obj=VSeries::filer->aggr_status($command_flag);
##
## @arg $command_flag Optional: flag for the command that the caller wants
##
##
sub aggr_status(;$){
    my $self = shift;
    my ($cmd_flag) = @_;
    my $cmd = "aggr status";
    if(defined($cmd_flag)){
        $cmd = $cmd . " " . $cmd_flag;
    }
    my $output = undef;
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    if(defined($output)){
        # strip out all \t's because they cause funky irregularities in output.
        $output =~ s/\t/ /g;
        my @output_array = split(/\n/, $output);
        return \@output_array;
    }
    else{
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Connection timeout during aggr status";
        return $ERR_STATUS;
    }

}


## @name ::get_aggr_size($)
##
## @description
## Method that returns the size of a given aggregate in MB.  This seems to be
## the most common case.  If future tests want some other unit they can add it
## here as needed with an optional param.
##
## @synopsis
## $filer_obj=VSeries::filer->get_aggr_size($aggregate_name);
##
## @arg $aggregate_name Required: Name of the aggregate the caller wants the
## size of.
##
##
sub get_aggr_size($){
    my $self = shift;
    my ($aggr_name) = @_;
    my $cmd = "df -Am";
    my $aggr_size = 0;
    my $output = undef;
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }
    if(defined($output)){
        # strip out all \t's because they cause funky irregularities in output.
        $output =~ s/\t/ /g;
        my @output_array = split(/\n/, $output);
        foreach my $line (@output_array){
            if($line =~ m/$aggr_name /){
                (undef, $aggr_size, undef) = split(/ +/, $line);
            }
        }
        $aggr_size =~ s/MB//;
        return $aggr_size;

    }
    else{
        $self->{error} = 1;
        $self->{errmsg} = "Connection timeout during get aggr size";
        return ();
    }

}


## @name ::get_vol_size($)
##
## @description
## Method that returns the size of a given volume in MB.  This seems to be the
## most common case.  If future tests want some other unit they can add it here
## as needed with an optional param.
##
## @synopsis
## $filer_obj=VSeries::filer->get_vol_size($vol_name);
##
## @arg $vol_name Required: Name of the vol the caller wants the
## size of.
##
##
sub get_vol_size($){
    my $self = shift;
    my($vol_name) = @_;
    my $cmd = "df -m";
    my $vol_size = 0;
    my $output = undef;
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }
    if(defined($output)){
        # strip out all \t's because they cause funky irregularities in output.
        $output =~ s/\t/ /g;
        my @output_array = split(/\n/, $output);
        foreach my $line (@output_array){
            if($line =~ m/$vol_name\/ /){
                (undef, $vol_size, undef) = split(/ +/, $line);
            }
        }
        $vol_size =~ s/MB//;
        return $vol_size;

    }
    else{
        $self->{error} = 1;
        $self->{errmsg} = "Connection timeout during get vol size";
        return;
    }

}






## @name ::aggr_show_space(;$)
##
## @description
## Method to run aggr show_space on a given filer.  Optionally takes the name
## of a given aggr otherwise will show the space for all aggregates on a given
## filer.
##
## @synopsis
## $filer_obj=VSeries::filer->aggr_show_space();
##
## @arg $aggr_name Optional: Name of the aggr the caller wants the
## available free size of.
##
##
sub aggr_show_space(;$){
    my $self = shift;
    my ($aggr_name) = @_;
    my $cmd = "aggr show_space " . $aggr_name;
    my $output = undef;
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    if(defined($output)){
        $output =~ s/\t/ /g;
        my @output_array = split(/\n/, $output);

        return \@output_array;
    }
    else{
        $self->{error} = 1;
        $self->{errmsg} = "Connection timeout during aggr show_space";
        return ();
    }
}


## @name ::clear_error()
##
## @description
## Method to clear the error state of a filer.  In some negative tests, it's
## expected that an error is generated.  This allows for gracefull continuation
## after an error.
##
## @synopsis
## $filer_obj=VSeries::filer->clear_error();
##
##
sub clear_error(){
    my $self = shift;
    $self->{error} = 0;
    $self->{errmsg} = "";

}



## @name ::aggr_add($)
##
## @description
## Method to add an array of disks to a given aggregate.  This will set the
## error flag and appropriate message if the disks were not seen or fails to
## add to the aggr.
##
## @synopsis
## $filer_obj=VSeries::filer->aggr_add($array_ref);
##
## @arg $array_reference Required: Reference point to the array of disk names
## the caller wants to add to the internal filer's aggregate.
##
##
sub aggr_add($){
    my $self = shift;
    my ($disks_to_add_ref) = @_;
    my @disks_to_add = @$disks_to_add_ref;
    my $cmd = "aggr add " . $self->{aggr} . " -d ";
    foreach my $disk (@disks_to_add){
        $cmd = $cmd . $disk . " ";
    }

    # Check if the disks couldn't be added.
    my $output = "";
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }
    if ($output =~ m/aggr add: (Couldn't add disks:|The specified operation could not be completed as the aggregate is currently busy)/){
        $self->{error} = 1;
        $self->{errmsg} = $output;
        return $ERR_STATUS;
    }

    my $pending = 0;
    my @output_array = ();
    # In the event that the disks could be added, do not return until the no
    # disk is pending.
    do{
        $pending = 0;
        $cmd = "aggr status ". $self->{aggr} . " -r";
        my $output = "";
        if($self->{use_console} eq "true"){
            (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
        }
        else{
            (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
        }
        $output =~ s/\t/ /g;
        @output_array = split(/\n/, $output);
        foreach my $line (@output_array){
            if($line =~ m/^ *Aggregate $self->{aggr} .*growing/){
                $pending = 1;
            }
        }
        # Wait for sometime before checking status again
        sleep(180);
    } while ($pending == 1);

    # Make sure that even though the disks aren't pending anymore, they did in
    # fact get added to the aggr.
    my $aggr_status_ref = VSeries::utils::parse_aggr_status("-r", @output_array);
    my %aggr_status_hash = %$aggr_status_ref;

    foreach my $disk (@disks_to_add){
        my $disk_primary_name = undef;
        my $disk_secondary_name = undef;
        # Get both primary and secondary path
        foreach my $serial (keys (%{$self->{disks}})){
            if($self->{disks}->{$serial}->{primary} eq $disk || $self->{disks}->{$serial}->{secondary} eq $disk){
                $disk_primary_name = $self->{disks}->{$serial}->{primary};
                $disk_secondary_name = $self->{disks}->{$serial}->{secondary};
                last;
            }
        }

        if(! exists $aggr_status_hash{$disk_primary_name} && ! exists $aggr_status_hash{$disk_secondary_name}){
            $self->{error} = 1;
            $self->{errmsg} = "Disk $disk did no show up as part of aggregate after expansion\nThe disks internally that appear to be part of the aggr are:";
            foreach my $err_disk (keys(%aggr_status_hash)){
                $self->{errmsg} = $self->{errmsg} . "\n" . $err_disk;
            }
            return $ERR_STATUS;
        }
    }


}


## @name ::vol_add($)
##
## @description
## Method to add an array of disks to a given volume.  This will set the error
## flag and appropriate message if the disks were not seen or fails to add to
## the aggr.
##
## @synopsis
## $filer_obj=VSeries::filer->vol_add($array_ref);
##
## @arg $array_reference Required: Reference point to the array of disk names
## the caller wants to add to the internal filer's aggregate.
##
##
sub vol_add($){
    my $self = shift;
    my ($disks_to_add_ref) = @_;
    my @disks_to_add = @$disks_to_add_ref;
    my @vol_list = @{$self->{vol_list}};
    my $cmd = "vol add " . $vol_list[0] . " -d ";
    foreach my $disk (@disks_to_add){
        $cmd = $cmd . $disk . " ";
    }

    # Check if the disks couldn't be added.
    my $output = "";
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }
    if ($output =~ m/vol add: (Couldn't add disks:|The specified operation could not be completed as the volume is currently busy)/){
        $self->{error} = 1;
        $self->{errmsg} = $output;
        return $ERR_STATUS;
    }

    my $pending = 0;
    my @output_array = ();

    # In the event that the disks could be added, do not return until no
    # disk is pending.
    do{
        $pending = 0;
        $cmd = "vol status ". $vol_list[0] . " -r";
        my $output = "";
        if($self->{use_console} eq "true"){
            (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
        }
        else{
            (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
        }
        $output =~ s/\t/ /g;
        @output_array = split(/\n/, $output);
        foreach my $line (@output_array){
            if($line =~ m/^ *Volume $vol_list[0] .*growing/){
                $pending = 1;
                sleep(5);
            }
        }
    } while ($pending == 1);

    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }
    $output =~ s/\t/ /g;
    @output_array = split(/\n/, $output);
    
    # Make sure that even though the disks aren't pending anymore, they did in
    # fact get added to the volume.
    my $vol_status_ref = VSeries::utils::parse_vol_status("-r", @output_array);
    my %vol_status_hash = %$vol_status_ref;
    foreach my $disk (@disks_to_add){
        my $disk_primary_name = undef;
        my $disk_secondary_name = undef;
        # Get both primary and secondary path
        foreach my $serial (keys (%{$self->{disks}})){
            if($self->{disks}->{$serial}->{primary} eq $disk || $self->{disks}->{$serial}->{secondary} eq $disk){
                $disk_primary_name = $self->{disks}->{$serial}->{primary};
                $disk_secondary_name = $self->{disks}->{$serial}->{secondary};
                last;
            }
        }

        if(! exists $vol_status_hash{$disk_primary_name} && ! exists $vol_status_hash{$disk_secondary_name}){
            $self->{error} = 1;
            $self->{errmsg} = "Disk $disk did no show up as part of volume after expansion\nThe disks internally that appear to be part of the volume are:";
            foreach my $err_disk (keys(%vol_status_hash)){
                $self->{errmsg} = $self->{errmsg} . "\n" . $err_disk;
            }
            return $ERR_STATUS;
        }
    }

}



## @name ::set_aggr_option($$$)
##
## @description
## Method to set an option on an aggr.  This method assumes the option you are
## setting is legal and valid.  If it isn't it will return as an error.
##
## @synopsis
## $filer_obj=VSeries::filer->set_aggr_option($aggr_name, $aggr_option, $option_value);
##
## @arg $aggr_name Required: Name of the aggregate the caller wants to set the
## option on.
##
## @arg $aggr_option Required: Name of the option the caller wants to set on
## the given aggregate
##
## @arg $option_value Required: The value the caller wants to set he given
## option on the aggregate to.
##
sub set_aggr_option($$$){
    my $self = shift;
    my ($aggr_name, $aggr_option, $option_value) = @_;
    my $cmd = "aggr options $aggr_name $aggr_option $option_value";
    my $output = "";
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }
    my $output_ref = $self->aggr_status();
    my @output_array = @$output_ref;
    my $hash_ref = VSeries::utils::parse_aggr_status("", @output_array);
    my %result_hash = %$hash_ref;
    if(! $result_hash{$aggr_name}{options} =~ m/$aggr_option/){
       $self->{error} = 1;
       $self->{errmsg} = "Option did no show up as set on aggregate";
    }
    return 0;

}

## @name ::set_vol_option($$$)
##
## @description
## Method to set an option on a volume.  This method assumes the option you are
## setting is legal and valid.  If it isn't it will return as an error.
##
## @synopsis
## $filer_obj=VSeries::filer->set_vol_option($vol_name, $vol_option, $option_value);
##
## @arg $vol_name Required: Name of the volume the caller wants to set the
## option on.
##
## @arg $vol_option Required: Name of the option the caller wants to set on
## the given volume
##
## @arg $option_value Required: The value the caller wants to set he given
## option on the volume to.
##
sub set_vol_option($$$){
    my $self = shift;
    my ($vol_name, $vol_option, $option_value) = @_;
    my $cmd = "vol options $vol_name $vol_option $option_value";
    my $output = "";
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }
    my $output_ref = $self->vol_status();
    my @output_array = @$output_ref;
    my $hash_ref = VSeries::utils::parse_vol_status("", @output_array);
    my %result_hash = %$hash_ref;
    if(! $result_hash{$vol_name}{options} =~ m/$vol_option/){
       $self->{error} = 1;
       $self->{errmsg} = "Option did no show up as set on volume";
    }
    return 0;

}


## @name ::set_snap_reserve($$$)
##
## @description
## Method to set the snapshot reserve value on an aggr or volume.  Verifies the
## correct value after setting it.  The parameters are the type flag (aggr/vol),
## the name of the vol/aggr and the value to set the snap reserve to.
##
## @synopsis
## $filer_obj=VSeries::filer->set_snap_reserve($type_flag, $name, $value);
##
## @arg $type_flag Required: Either volume or aggregate.
##
## @arg $name Required: Name of the vol/aggr the caller wants to set the snap
## reserve value on.
##
## @arg $value Required: Value that the caller wants the snap reserve
## percentage to be.  Expected to be in a raw number form, no need for the '%'
##
##
sub set_snap_reserve($$$){
    my $self = shift;
    my ($type_flag, $name, $value) = @_;
    my $cmd = "snap reserve $type_flag $name $value";
    my $output = "";
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    my $verify_cmd = "snap reserve $type_flag $name";
    my $percent = $value . "%";
    my $output = "";
    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $verify_cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $verify_cmd, -swallow, "-timeout", 60);
    }
    if($output !~ m/ $percent /){
        $self->{error} = 1;
        $self->{errmsg} = "Snapshot reserve was not set to requested value of $percent.\n" . $output;
    }
    return 0;

}

#-------------------------------------------------------------------------------

## @name ::filer_fabric_count_get()
##
## @author James Silva (silva@netapp.com)
##
## @description Returns number of fabrics a filer object is connected to.
##
## @synopsis
## $filer_obj->filer_fabric_count_get()
##
sub filer_fabric_count_get() {

    my $self = shift;
    my $after = '';
    my @fabric_list_array = ();

    (undef, undef, $after) = Tharn::say($self->{connection}, "storage show fabric", -swallow, '-timeout', 60, '-prompt', "$self->{name}\*>");

    @fabric_list_array = split(/\n/, $after);
    my $fabric_count = 0;

    foreach (@fabric_list_array){
        if ($_ =~ /Fabric:/){ ++$fabric_count; }
    }

    return($fabric_count);
}

#-------------------------------------------------------------------------------

## @name ::filer_lun_count_get()
##
## @author James Silva (silva@netapp.com)
##
## @description Returns number of LUNs owned by a given filer object.  It is
## assumed that the current state of the filer is that it is at its normal
## prompt.  i.e. not CFE, ok, maint. mode, etc.
##
## @synopsis
## $filer_obj->filer_lun_count_get()
##
sub filer_lun_count_get() {

    my $self = shift;
    my $after = '';
    my @lun_list_array = ();
    my @temp = ();

    (undef, undef, $after) = Tharn::say($self->{connection}, "disk show -o $self->{name}",
                             -swallow, '-timeout', 60, '-prompt', "$self->{name}\*>");

    @lun_list_array = split(/\n/, $after);

    # - - - - - - -  - - - - - - - - - - - - - -
    # We only want elements 2-onward of the array.
    # - - - - - - -  - - - - - - - - - - - - - -

    @lun_list_array = @lun_list_array[2 .. $#lun_list_array];

    # - - - - - - -  - - - - - - - - - - - - - -
    # Now for some additional checking of the elements in @lun_list_array.
    # we want to make sure that we can handle cases where there may be
    # noisy output on the connection and we may mistakenly bring that noise
    # into @lun_list_array.  This could possible create a situation where
    # we get an inaccurate number of LUNS that we own.  To prevent this,
    # we are only going to add luns to our list if they have the format of
    # a LUN which looks something like this:  brcdqa6:2.126L2 or in the case of
    # a Cisco switch a LUN may look like this: vns-c9124-2:1-2.126L1
    # - - - - - - -  - - - - - - - - - - - - - -

    foreach my $disk (@lun_list_array) {

        if ($disk =~ /.+:\d+\.\d+L\d+/ || $disk =~ /.+:\d+-\d+\.\d+L\d+/) {
            push @temp, $disk;
        }
        else { next; }

    } # foreach my $disk (@lun_list_array) {

    my $filer_lun_count = scalar(@temp);

    return($filer_lun_count);

}

#-------------------------------------------------------------------------------

## @name ::lun_scsi_inquiry($device_name)
##
## @author James Silva (silva@netapp.com)
##
## @description
## 'Pings' the scsi device_name passed in (a LUN) by doing a scsi inquiry.  If
## the device_name is unreachable returns $ERR_STATUS, otherwise returns
## $GOOD_STATUS.
##
## @synopsis
## $filer_obj->lun_scsi_inquiry($device_name)
##
## @arg $device_name Required: The scsi device_name to do an inquiry on.
##
sub lun_scsi_inquiry($){

    my ($self, $device_name) = @_;

    # - - - - - - -  - - - - - - - - - - - - - -
    # When a device_name is reachable it looks like this:
    # -----------------------------------------------
    # vsa-v3050-auto-2*> scsi inquiry brcdqa6:2.126L2
    # type 0 support 4
    # vendor EMC
    # product SYMMETRIX
    # revision 5671
    # vendor unique 32150000Q
    # -----------------------------------------------
    #
    # When a device_name is unreachable it looks like this:
    # -----------------------------------------------
    # vsa-v3050-auto-2*> scsi inquiry brcdqa6:2.126L2
    # device brcdqa6:2.126L2 not found
    # - - - - - - -  - - - - - - - - - - - - - -

    my $status = '';

    (undef, undef, $status) = Tharn::say($self->{connection}, "priv set advanced; scsi inquiry $device_name", -swallow, '-timeout', 30, '-prompt', "$self->{name}\*>");

    # - - - - - - -  - - - - - - - - - - - - - -
    # return $ERR_STATUS on error and return $GOOD_STATUS for pass
    # - - - - - - -  - - - - - - - - - - - - - -

    if ($status =~ /device $device_name not found/) {
        return($ERR_STATUS);

    } else { return($GOOD_STATUS); }

}

#-------------------------------------------------------------------------------

## @name ::storage_show_disk(;$$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## With no arguments passed in it will display to just the results of the
## 'storage show disk' ontap CLI command.  You can alter this output by
## passing optional arguments in this format:
## storage show disk [ -a | -x | -p ] [ <name> ]
##
## @synopsis
##
sub storage_show_disk(;$$){
    my ($self, $arg1, $arg2) = @_;
    my $after = '';
    my $cmd = "storage show disk ";

    if (defined $arg1) { $cmd .= $arg1 . " "; }
    if (defined $arg2) { $cmd .= $arg2; }

    (undef, undef, $after) = Tharn::say($self->{connection}, "$cmd", -swallow, '-timeout', 30, '-prompt', "$self->{name}\*>");

     return($GOOD_STATUS);
}

#-------------------------------------------------------------------------------

## @name ::wwn_adapter_get($adapter)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Returns the wwn of $adapter in this format: 50:0a:09:80:00:01:43:88.
## On error, it sets $self->{error} to 1 and populates $self->{errmsg}
## with an error message.
## returns the wwn
##
## @synopsis
## $filer_obj->wwn_adapter_get($adapter)
##
## @arg $adapter Required: The adapter you wish to seek a wwn from.
##
sub wwn_adapter_get($){

    my ($self, $adapter) = @_;
    my @temp = ();
    my ($current_adapter, $wwn, $after) = '';
    my $error_flag = 1;
    my $count = 0;

    (undef, undef, $after) = Tharn::say($self->{connection}, "priv set -q advanced;fcadmin channels", -swallow, '-timeout', 60, '-prompt', "$self->{name}\*>");

    @temp = split(/\n/, $after);

    foreach my $line (@temp){
        ($wwn, $current_adapter) = split(/\s+/, $line);

        if ($current_adapter eq $adapter) {$error_flag = 0; last;}
    }

    # - - - - - - -  - - - - - - - - - - - - - -
    # Time to place the $wwn in the appropriate form, i.e.
    # transform a wwn that looks like this: 500a098000014388
    # to something that looks like this: 50:0a:09:80:00:01:43:88
    # - - - - - - -  - - - - - - - - - - - - - -

    @temp = ();
    @temp = split(/(\w{2})/, $wwn);


    # - - - - - - -  - - - - - - - - - - - - - -
    # If we didn't find our adapter, we had an error so return it.
    # - - - - - - -  - - - - - - - - - - - - - -

    if ($error_flag){
        $self->{error}  = 1;
        $self->{errmsg} = "Could not find adapter $adapter on FILER " . $self->get_name() .".";
        return(-1);
    }

    # - - - - - - -  - - - - - - - - - - - - - -
    # I only want elements 1 onward (1st is a blank).
    # - - - - - - -  - - - - - - - - - - - - - -

    @temp = @temp[1 .. $#temp];

    # - - - - - - -  - - - - - - - - - - - - - -
    # Insert : in the blanks.
    # - - - - - - -  - - - - - - - - - - - - - -

    foreach (@temp){

        if ($_ eq ''){$temp[$count] = ":";}
        $count++;
    }

    # - - - - - - -  - - - - - - - - - - - - - -
    # Reconstruct $wwn from elements of @temp
    # - - - - - - -  - - - - - - - - - - - - - -

    $wwn = '';

    foreach (@temp){ $wwn .= $_; }

    return($wwn);

}

#-------------------------------------------------------------------------------

## @name ::filer_env_varible_set()
##
## @author James Silva (silva@netapp.com)
##
## @description Uses the filers console connection to do a 'cf disable',
## then halts, then sets the passed in environment variable to the value
## specified then boots ontap for all filer objects passed in.  Then it will
## issue 'cf enable' to bring the filers back into their previous cf state.
##
## @synopsis
## filer_env_varible_set(\@filer_obj_array, $env_var, $value)
##
## @arg $env_var Required: The environment variable to set.
## @arg $value Required: The value you wish to set your $env_var to.
##

sub filer_env_variable_set($$) {
    my ($self, $env_var, $value) = @_;
    my ($match_string, $temp) = '';
my @accepted_prompts = ('^CFE> ', '^LOADER\\*?>','^ok ', '^LOADER-(A|B)\*?>');
    my @normal_prompts = ("$self->{name}\*>", "$self->{name}>");
    my @environment = ();

    # - - - - - - -  - - - - - - - - - - - - - -
    # Halt the filer then make sure we get a fw prompt.
    # - - - - - - -  - - - - - - - - - - - - - -

    Tharn::say($self->{console_connection}, 'halt -f', '-blind');

    # - - - - - - -  - - - - - - - - - - - - - -
    # In the case of a CIFS connection.
    # - - - - - - -  - - - - - - - - - - - - - -

    Tharn::say($self->{console_connection}, '0', '-blind');

    my (undef,undef,$after,$prompt) = Tharn::say($self->{console_connection},
                                   "",'-nocr','-timeout',180,'-maxidle',120,
                                   '-noprompt','-prompt', @accepted_prompts);

    Tharn::say($self->{console_connection}, "setenv $env_var $value", '-timeout', 60,
                                   '-maxidle', 30, '-noprompt', '-prompt', @accepted_prompts);

    (undef, undef, $temp) = Tharn::say($self->{console_connection}, 'printenv', '-timeout', 60,
                                   '-maxidle', 30, '-noprompt', '-prompt', @accepted_prompts);

    @environment = split(/(\n)/, $temp); # don't lose the \n delimiter

    # - - - - - - -  - - - - - - - - - - - - - -
    # Ensure we have set the environment variable to the value passed in.
    # - - - - - - -  - - - - - - - - - - - - - -

    my $error_flag = 1;

    foreach (@environment){
        if (($_ =~ /$env_var/) && ($_ =~ /$value\r/)){ $error_flag = 0; last; }
    }

    if ($error_flag) { return($ERR_STATUS); }

    # - - - - - - -  - - - - - - - - - - - - - -
    # Boot ontap and wait until you get filer prompts.
    # - - - - - - -  - - - - - - - - - - - - - -

    (undef,undef,$after,$prompt) = Tharn::say($self->{console_connection},
                                   "boot_ontap",'-timeout',300,'-maxidle',120,
                                   '-noprompt','-prompt', @normal_prompts);

    # - - - - - - -  - - - - - - - - - - - - - -
    # If we got this far, we passed.
    # - - - - - - -  - - - - - - - - - - - - - -

    return($GOOD_STATUS);

}

#-------------------------------------------------------------------------------

## @name ::cf_enable()
##
## @author James Silva (silva@netapp.com)
##
## @description
## A pass through for cluster.pm's 'cf_enable();
##
## @synopsis
## $filer_obj->cf_enable();
##
sub cf_enable($) {
    my $self = shift;

    return(VSeries::cluster::cf_enable($self->{conn}));
}

#-------------------------------------------------------------------------------

## @name ::cf_disable()
##
## @author James Silva (silva@netapp.com)
##
## @description
## A pass through for cluster.pm's 'cf_disable()'
##
## @synopsis
## $filer_obj->cf_disable();
##
sub cf_disable($) {
    my $self = shift;

    return(VSeries::cluster::cf_disable($self->{conn}));
}

#-------------------------------------------------------------------------------

## @name ::send_filer_cmd($;$$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## A pass through for a say with limited functionality.  Used mostly for
## subroutines that may exist within a .thpl and need a call to say.  Since
## we don't use say() in thpl code.
##
## @synopsis
## $filer_obj->send_filer_cmd($;$$)
##
## @arg $cmd_string Required: A string you wish to pass via the say() routine.
## @arg $conn  Optional: Uses rsh as a default.
## @arg $timeout    Optional: Timeout for the command
##

sub send_filer_cmd($;$$) {
    my ($self, $cmd_string, $conn, $timeout) = @_;
    my @say_result = ();
    my $connection = '';
    my @CFE_prompts = ('^CFE>', '^LOADER\\*?>','^ok ', '^LOADER-(A|B)\*?>');

    if (! defined $conn) {
        $connection = $self->{connection};
    } else { $connection = $self->{$conn}};
    if (!defined $timeout) {
        $timeout = 60;
    }

    @say_result = Tharn::say($connection, "$cmd_string", -swallow, '-timeout', $timeout, "-prompt", @CFE_prompts);

    return (@say_result);
}

#--------------------------------------------------------------------------------

## @name ::dd()
##
## @author James Silva (silva@netapp.com)
##
## @description Method that allows for OnTap's 'dd' utility
## which copies blocks of data.
##
## dd [ [if= file ] | [ din= disknum bin= blocknum ] ]
##    [ [of= file ] | [ dout= disknum bout= blocknum ] ]
## count= number_of_blocks
##
## dd  copies  the  specified number of blocks from source to
## destination.  Source and destination may be  specified  as
## either  a file that is a fully qualified pathname, or as a
## starting block on a disk. The parameter disknum may  range
## from  zero  to the maximum number reported by na_sysconfig
## (1) -r. In the latter form of specifying source or  desti-
## nation,  both  disknum  and blocknum must be specified.
## If the number of blocks exceeds the  size  of  the
## file, copying stops upon reaching EOF.
##
## Unlike the OnTap version of this utility you always need to
## specify a source and destination.  This will not substitute
## those voids with STDIN or STDOUT.
##
## Vseries will be placed into 'priv set diag' mode during this
## methods use, but will be returned to the default 'priv set admin'
## at its completion.
##
## This method does not support the console_connection, only the
## default connection stored in the objects tharnhost file will
## be used, i.e. $self->{connection}.  This is usually rsh for
## filers.
##
## The prototype takes a single scalar only to insure that the user
## has supplied input to the method.  The real argument parsing
## takes place using GetOptions for parsing @ARGV.
##
## @synopsis
## $status = $filer_obj->dd("-if=/vol/test_vol/test_file",
##                          "-dout=0f.28", "-bout=13491");
##
## @arg $if     Optional: Input file
## @arg $of     Optional: Output file
## @arg $din    Optional: Disk in
## @arg $bin    Optional: Block in
## @arg $dout   Optional: Disk out
## @arg $bout   Optional: Block out
## @arg $count  Optional: Count (of blocks)
## @arg $checsum_copy Optional: Flag for checsum aware copy (zoned v2)
##
sub dd($) {

    my $self       = shift;                # Filer object
    my $priv       = "priv set -q diag;";  # Cmd to change priv's on filer
    my $admin      = "priv set admin;";    # Returns privs to admin mode
    my @argv_copy  = @_;                   # Copy ARGV before we process
    my $dd_options = '';                   # Holds concatenated @ARGV
    my $after      = '';                   # Holds raw output from say()
    @ARGV = @_;                            # Setup @ARGV for GetOptions

    # These are all false by default.
    my ($if, $of, $din, $bin, $dout, $bout, $count, $checksum_copy) = '';

    GetOptions( "if:s"    => \$if,
                "of:s"    => \$of,
                "din:s"   => \$din,
                "bin:i"   => \$bin,
                "dout:s"  => \$dout,
                "bout:i"  => \$bout,
                "count:i" => \$count,
                "c"       => \$checksum_copy,   # Flag
                );

    # Let's validate our options.

    # Can't have din w/out bin & vice versa
    if (($din && !$bin) or ($bin && !$din)) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::dd(): din and bin must not be mutually exclusive.";
        return($ERR_STATUS);
    }
    # Can't have dout w/out bout & vice versa
    if (($dout && !$bout) or ($bout && !$dout)) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::dd(): dout and bout must not be mutually exclusive.";
        return($ERR_STATUS);
    }
    # Can't have any combination of 'of' with dout or bout
    if ($of && ($dout|$bout)) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::dd(): of together with dout or bout are mutually exlusive options.";
        return($ERR_STATUS);
    }
    # Can't ONLY have count|din|bin|dout|bout|if|of & you need a src and dest
    if (scalar @argv_copy < 2) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::dd(): must specify a source and destination.";
        return($ERR_STATUS);
    }
    # If the user doesn't get the arguments right, and it gets passed
    # the wall above we'll just have to give them a general purpose error
    # if things don't end up successful.  We'll set the error flag in the
    # filer object.

    # Create the options string to be passed via say().  Strip any white-
    # space in any of the options plus any dashes.

    foreach my $opt (@argv_copy) {
        $opt =~ s/^\s+$//g;
        $opt =~ s/^-//;
        $dd_options .= $opt . " ";
    }

    if($checksum_copy) {
        $dd_options = "-c " . $dd_options;   
    }
    
    # We're not checking $after here for any errors as we should have caught
    # all issues with command composition by now.  Furthermore, I wanted
    # to simplify here as it was recently discovered that the raw output from
    # the dd() can actually be seen now.  That means virtually any text could
    # be in an input file or block, making screen scrubbing non-ideal here.
    # Using "-noprompt" to work around BURT 380390.
    Tharn::say($self->{connection}, "$priv dd $dd_options", -swallow, '-timeout', 60, '-noprompt');
    
    # On some systems, the dd command doesn't return a prompt as expected (BURT
    # 380390), so this is a little poke.
    $self->send_filer_cmd(" ");

    # Return prompt to admin mode.
    Tharn::say($self->{connection}, "$admin", -swallow, '-timeout', 30, '-prompt', "$self->{name}\*>");

    return($GOOD_STATUS);
}

#--------------------------------------------------------------------------------

## @name ::mv($$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Takes $src (some source file) and renames it to $dest (destination file).
## For this to work you must specify the correct path to a particlular file
## such as '/vol/vol0/etc/messages'
##
## This method does not support the console_connection, only the
## default connection stored in the objects tharnhost file will
## be used, i.e. $self->{connection}.  This is usually rsh for
## filers.
##
## Vseries will be placed into 'priv set diag' mode during this
## methods use, but will be returned to the default 'priv set admin'
## at its completion.
##
## @synopsis
## $status = $filer_obj->mv($src, $dest);
##
## @arg $src    Required: Input file
## @arg $dest   Required: Output file
##
sub mv($$) {

    my $self         = shift;               # Filer object
    my ($src, $dest) = @_;                  # Source and destination files
    my $priv         = "priv set -q diag;"; # Cmd to change priv's on filer
    my $admin        = "priv set admin;";   # Returns privs to admin mode
    my $after        = '';                  # Holds the raw output from Tharn::say()

    (undef, undef, $after) = Tharn::say($self->{connection}, "mv $src $dest", -swallow, '-timeout', 30);

    # All errors are prefaced with 'rename:'
    # vnv3050f9a> mv /vlba /fkl.txt
    # rename: No such file or directory
    # vnv3050f9a>

    if ($after =~ /^rename:/i) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::mv(): Rename did not complete successfully.  Ensure file paths are correct.\n $after";
        return($ERR_STATUS);
    }

    # Further validate the move success.
    my ($filename, $directories, $suffix) = File::Basename::fileparse($dest);
    $filename .= $suffix;

    # Do an ls on the dir provided to get dir contents.
    (undef, undef, $after) = Tharn::say($self->{connection}, "$priv ls $directories", -swallow, '-timeout', 30);
    my @temp = split (/\n/, $after);
    my $found_file_flag = 0;

    # Make sure our destination file exists.
    foreach (@temp) {$_ =~ /^$filename/ ? $found_file_flag = 1 : next;}

    if (!$found_file_flag) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::mv(): Unable to find $filename in $directories.\n $after";

        # Change the privs back to admin mode.
        Tharn::say($self->{connection}, "$admin", -swallow, '-timeout', 30, '-prompt', "$self->{name}\*>");
        return($ERR_STATUS);
    }

    # Change the privs back to admin mode.
    Tharn::say($self->{connection}, "$admin", -swallow, '-timeout', 30, '-prompt', "$self->{name}\*>");

    return($GOOD_STATUS);

}

#--------------------------------------------------------------------------------

## @name ::showfh($)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Gets the filehandle information for $file and places that info in a hash
## called %fh_hash.  The keys are filename, flags, snapid, fileid, gen & fsid.
## We return a reference to that hash on success and failure.  If a failure
## does exist we set the filer_obj's internal error flag.  You can also
## validate the hash from the .thpl layer by counting hash keys (6).
## Note: its suggested to give the entire file path i.e.:
## /vol/vol0/etc/messages
##
## Vseries will be placed into 'priv set diag' mode during this
## methods use, but will be returned to the default 'priv set admin'
## at its completion.
##
## This method does not support the console_connection, only the
## default connection stored in the objects tharnhost file will
## be used, i.e. $self->{connection}.  This is usually rsh for
## filers.
##
## @synopsis
## $fh_hash_ref = $filer_obj->showfh($file);
##
## @arg $file Required: File to retrieve filehandle from.
##
sub showfh($) {
    my $self    = shift;               # Filer object
    my ($file)  = @_;                  # File to get filehandle info from.
    my $after   = '';                  # Holds the raw output from Tharn::say()
    my $priv    = "priv set -q diag;"; # Cmd to change priv's on filer
    my $admin   = "priv set admin;";   # Returns privs to admin mode
    my @fh_vals = ();                  # Used fo splitting the return of showfh
    my %fh_hash = ();                  # Hash that will be returned with fh info
    my $param   = '';                  # Different params in showfh output

    (undef, undef, $after) = Tharn::say($self->{connection}, "$priv showfh $file", -swallow, '-timeout', 30);

    # When errors occur it will match ^.*:\s+\w+
    # vnv3050f9a*> showfh
    # showfh: Invalid parameter.
    # usage: showfh filename
    # vnv3050f9a*>

    if ($after =~ /^.*:\s+\w+/) {

        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::showfh(): Unable to get filehandle for $file.";

        # Return prompt to admin mode.
        Tharn::say($self->{connection}, "$admin", -swallow, '-timeout', 30, '-prompt', "$self->{name}\*>");

        return(\%fh_hash);

    } else {

        # On success you get:
        # vnv3050f9a*> showfh /etc/messages
        # flags=0x00 snapid=0 fileid=0x0f6392 gen=0x4cd11b3 fsid=0x390809a5
        # vnv3050f9a*>

        @fh_vals = split(/\s+/, $after);

        foreach $param (@fh_vals) {
            if ($param =~ /^flags/) {
                $param =~ s/^flags=//;
                $fh_hash{'flags'} = $param;
            }
            elsif ($param =~ /^snapid/) {
                $param =~ s/^snapid=//;
                $fh_hash{'snapid'} = $param;
            }
            elsif ($param =~ /^fileid/) {
                $param =~ s/^fileid=//;
                $fh_hash{'fileid'} = $param;
            }
            elsif ($param =~ /^gen/) {
                $param =~ s/^gen=//;
                $fh_hash{'gen'} = $param;
            }
            elsif ($param =~ /^fsid/) {
                $param =~ s/^fsid=//;
                $fh_hash{'fsid'} = $param;
            }
            elsif ($param =~ /^filename/) {
                $param =~ s/^filename=//;
                $fh_hash{'filename'} = $param;
            }
        } # end foreach $param

        # Return prompt to admin mode.
        Tharn::say($self->{connection}, "$admin", -swallow, '-timeout', 30, '-prompt', "$self->{name}\*>");

        return(\%fh_hash);
    }
}
#--------------------------------------------------------------------------------

## @name ::dumpblock($)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Print all sorts of useful information about blocks or files such as
## disk, dbn, vbn, block_contents, status, pvbn, pvbn_from_parent,
## pvbn_of_parent, fbn_of_inofile, pvbn_of_inofile, pvbn_of_inofile_from_parent
## ckse_dbn, ckse_checksum, ckse_buftreeid, ckse_fbn, and ckse_cp_count.
## All of this information is returned in a hash ref which contains the keys
## mentioned above.  Additionally there is one key called 'parse_error' which
## stores any lines that didn't match a regex for debugging purposes.
##
## *** THE FOLLOWING IS ONTAP USAGE ***
## vnv3050f9a> priv set -q diag
## vnv3050f9a*> dumpblock
## usage:
## dumpblock ifile [-A | -V] <vol-name> <fileid>
##   - dump block of inofile containing that inode
## dumpblock file [-A | -V] <vol-name> <fileid> <blockno>
##   - dump data block <blockno> from file inode # <fileid>
## dumpblock vbn [-A | -V] <vol-name> <vbn>
##   - dump block with the specified vbn on the specified volume/aggregate
## dumpblock dbn [-c] <disk_name> <dbn>
##   - dump block with the specified dbn on the specified disk
## dumpblock info [-A | -V] <vol-name> <fileid> <blockno>
##   - dump information about the specified block
## vnv3050f9a*>
## ***------------------------------***
##
## Vseries will be placed into 'priv set diag' mode during this
## method's use, but will be returned to the mode it was in before
## this method was called by calling get_priv() and storing it.
##
## This method does not support the console_connection, only the
## default connection stored in the objects tharnhost file will
## be used, i.e. $self->{connection}.  This is usually rsh for
## filers.
##
## This was tested in 7.3 and 7-mode (OnTap 8.0)
##
## @synopsis
## $db_hash_ref = $filer_obj->dumpblock("ifile", "-V", "-vol_name=vol0", "-file_id=0x0d6a16");
##
## @arg $type      Required: Dumpblock type <ifile|file|vbn|dbn|info>
## @arg $vol_name  Optional: A volume name
## @arg $file_id   Optional: A file id number (0x342d)
## @arg $A         Optional: Flag for aggregate
## @arg $V         Optional: Flag for volume
## @arg $block_no  Optional: A block number (physical)
## @arg $vbn       Optional: A block number (virtual)
## @arg $c         Optional: A flag for something...
## @arg $dbn       Optional: A disk block number
## @arg $disk_name Optional: A disk name
## @arg $level     Optional: A level or sorts...
##
sub dumpblock($) {
    my $self       = shift;               # Filer object
    my $type       = shift;               # DB type <ifile|file|vbn|dbn|info>
    my $db_options = '';                  # Holds concatenated @ARGV
    my @temp       = ();                  # Used to hold values while parsing
    my $after      = '';                  # Holds raw output from say()
    my $priv       = "priv set -q diag;"; # Cmd to change priv's on filer
    my %db_hash    = ();                  # Hash returned with db info
    my $line       = '';                  # Holds loop lines while iterating
    @ARGV          = @_;                  # Setup @ARGV for GetOptions

    # These are all false by default.
    my($vol_name,                         # A volume name
       $file_id,                          # A file id number
       $A,                                # Flag for aggregate
       $V,                                # Flag for volume
       $block_no,                         # A block number (physical)
       $vbn,                              # A block number (virtual)
       $c,                                # A flag for something...
       $dbn,                              # A disk block number
       $disk_name                         # A disk name
       ) = '';

    # Init all keys even though some will not be populated based on
    # dumpblock type specified.
    my %db_hash = ( 'disk'                        => '',
                    'dbn'                         => '',
                    'vbn'                         => '',
                    'block_contents'              => '',
                    'status'                      => '',
                    'pvbn'                        => '',
                    'pvbn_from_parent'            => '',
                    'pvbn_of_parent'              => '',
                    'fbn_of_inofile'              => '',
                    'pvbn_of_inofile'             => '',
                    'pvbn_of_inofile_from_parent' => '',
                    'ckse_dbn'                    => '',
                    'ckse_checksum'               => '',
                    'ckse_buftreeid'              => '',
                    'ckse_fbn'                    => '',
                    'ckse_cp_count'               => '',
                    'parse_error'                 => ''
    );

    # Store our priv mode so we can restore it later.
    my $priv_mode = $self->get_priv();

    # Grab our options
    GetOptions( "vol_name:s"    => \$vol_name,  # Optional string
                "file_id:s"     => \$file_id,   # Optional string
                "A"             => \$A,         # Flag
                "V"             => \$V,         # Flag
                "block_no:i"    => \$block_no,  # Optional integer
                "vbn:i"         => \$vbn,       # Optional integer
                "c"             => \$c,         # Flag
                "dbn:i"         => \$dbn,       # Optional integer
                "disk_name:s"   => \$disk_name  # Optional string
              );

    if ($c) {
       $c = "-c";
    }
    if ($A) {
       $A = "-A";
    } 
    if ($V) {
       $V = "-V";
    }

    # - - - - - - -  - - - - - - - - - - - - - -
    # dumpblock ifile [-A | -V] <vol-name> <fileid>
    #
    #vnv3050f9a*> dumpblock ifile vol0 0x0d6a16
    #disk 0d.25, dbn 7527790, vbn 7522542
    #0x0000: 0x4946424d 0x0005fdfe 0x0000a387 0x00000000 <MBFI............>
    #0x0010: 0x00000000 0x00000000 0x00000000 0x00000000 <................>
    #(More of the same, which we'll capture)
    # - - - - - - -  - - - - - - - - - - - - - -

    # - - - - - - -  - - - - - - - - - - - - - -
    # dumpblock file [-A | -V] <vol-name> <fileid> <blockno>
    #
    #vnv3050f9a*> dumpblock file -A vol0 0x0d6a16 0
    #disk 0d.25, dbn 7285674, vbn 7280426
    #0x0000: 0x206e7553 0x20727041 0x30203620 0x30303a30 <Sun Apr  6 00:00>
    #0x0010: 0x2030303a 0x20544d47 0x766e765b 0x30353033 <:00 GMT [vnv3050>
    #(More of the same, which we'll capture)
    # - - - - - - -  - - - - - - - - - - - - - -

    # - - - - - - -  - - - - - - - - - - - - - -
    #dumpblock vbn [-A | -V] <vol-name> <vbn>
    #
    #vnv3050f9a*> dumpblock vbn vol0 7280426
    #VBN 7280426, disk 0d.25, dbn 7285674
    #0x0000: 0x206e7553 0x20727041 0x30203620 0x30303a30 <Sun Apr  6 00:00>
    #0x0010: 0x2030303a 0x20544d47 0x766e765b 0x30353033 <:00 GMT [vnv3050>
    #0x0020: 0x3a613966 0x6f6e7520 0x64656e77 0x7369642e <f9a: unowned.dis>
    #(More of the same, which we'll capture)
    # - - - - - - -  - - - - - - - - - - - - - -

    # - - - - - - -  - - - - - - - - - - - - - -
    # dumpblock dbn [-c] <disk_name> <dbn>
    #
    #vnv3050f9a*> dumpblock dbn 0d.25 7285674
    #disk 0d.25, dbn 7285674
    #0x0000: 0x206e7553 0x20727041 0x30203620 0x30303a30 <Sun Apr  6 00:00>
    #0x0010: 0x2030303a 0x20544d47 0x766e765b 0x30353033 <:00 GMT [vnv3050>
    #0x0020: 0x3a613966 0x6f6e7520 0x64656e77 0x7369642e <f9a: unowned.dis>
    #(More of the same, which we'll capture)
    # - - - - - - -  - - - - - - - - - - - - - -

    # - - - - - - -  - - - - - - - - - - - - - -
    # dumpblock info [-A | -V] <vol-name> <fileid> <blockno>
    #
    #vnv3050f9a*> dumpblock info vol0 0x0d6a16 0
    #ok
    #disk 0d.25
    #dbn 7285674
    #pvbn 7280426
    #pvbn-from-parent 7280426
    #fbn-of-inofile 41863
    #pvbn-of-inofile 7524365
    #pvbn-of-inofile-from-parent 7524371
    #ckse-dbn 7285674
    #ckse-checksum 0x44dcffad
    #ckse-buftreeid 0x132d45
    #ckse-fbn 0
    #ckse-cp-count 379077
    #vnv3050f9a*>
    # - - - - - - -  - - - - - - - - - - - - - -

    # Validate options foreach type

    if ($type =~ /^(ifile|file|vbn|info)$/) {
        # We can't have $A & $V, we need a $vol_name
        if (($A && $V) || (!$vol_name)){
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "Error: filer::dumpblock(): Incorrect parameters given to dumpblock.  Must have optionally A or V and vol_name.";
            return(\%db_hash);
        }
        if ($type =~ /^(ifile|file|info)$/) {
            # ifile file and info need a $file_id specified.
             if (!$file_id) {
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "Error: filer::dumpblock(): Incorrect parameters given to dumpblock.  Must have a file_id.";
                return(\%db_hash);
            }
            if (($type =~ /^(file|info)$/) && ($block_no !~ /\d+/)) {
                # file & info type needs to have a $block_no
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "Error: filer::dumpblock(): Incorrect parameters given to dumpblock.  Must have a block_no.";
                return(\%db_hash);
            }
        }
        if (($type =~ /^vbn$/) && ($vbn !~ /\d+/)) {
            # vbn needs a $vbn which should be some virtual block number
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "Error: filer::dumpblock(): Incorrect parameters given to dumpblock vbn.  Must have a vbn.";
            return(\%db_hash);
        }
    }
    elsif ($type =~ /^dbn$/  ) {
        # We need a $disk_name & $dbn.
        # To complicate further, $dbn is a number and it must exist,
        # Therefore to validate we ensure it matches >= 1 sequential digits.
        if ((!$disk_name) || ($dbn !~ /\d+/)){
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "Error: filer::dumpblock(): Incorrect parameters given to dumpblock dbn.  Must have disk_name and dbn.";
            return(\%db_hash);
        }
    } else {
        # User did not specify a type supported by the dumpblock command.
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::dumpblock(): Incorrect dumpblock type, type specified: $type.  Must be info, dbn, vbn, file or ifile.";
        return(\%db_hash);
    }

    # All options/arguments have been validated.
    # We can continue on and build our $db_options string by appending
    # options that are defined.  Strip any whitespace or commas globally.

    foreach my $option (($A, $V, $c, $vol_name, $disk_name, $file_id, $block_no, $vbn, $dbn)) {
        # If option is true, or is some number, its valid.
        if (($option) || ($option =~ /\d+/)) {
            $option =~ s/(\s+|,)//g;
            $db_options .= "$option ";
        }
    }

    # Strip the last space off of the $db_options string.
    chop($db_options);

    # Send cmd to OnTap then populate the dumpblock hash (%db_hash).
    (undef, undef, $after) = Tharn::say($self->{connection}, "$priv dumpblock $type $db_options", -swallow, '-timeout', 60);
    @temp = split(/\n/, $after);

    # We'll parse types ifile, file, vbn, & dbn one way, and info
    # another way.  Hence we will parse depending on type.

    if ($type eq "info") {
        foreach $line (@temp) {
            if    ($line =~ /^ok/)                         {(       $db_hash{'status'})                      = split(/\s+/, $line);}
            elsif ($line =~ /^disk/)                       {(undef, $db_hash{'disk'})                        = split(/\s+/, $line);}
            elsif ($line =~ /^dbn/)                        {(undef, $db_hash{'dbn'})                         = split(/\s+/, $line);}
            elsif ($line =~ /^pvbn\s+\d+/)                 {(undef, $db_hash{'pvbn'})                        = split(/\s+/, $line);}
            elsif ($line =~ /^pvbn-from-parent/)           {(undef, $db_hash{'pvbn_from_parent'})            = split(/\s+/, $line);}
            elsif ($line =~ /^fbn-of-inofile/)             {(undef, $db_hash{'fbn_of_inofile'})              = split(/\s+/, $line);}
            elsif ($line =~ /^pvbn-of-inofile/)            {(undef, $db_hash{'pvbn_of_inofile'})             = split(/\s+/, $line);}
            elsif ($line =~ /^pvbn-of-inofile-from-parent/){(undef, $db_hash{'pvbn_of_inofile_from_parent'}) = split(/\s+/, $line);}
            elsif ($line =~ /^pvbn-of-parent/)             {(undef, $db_hash{'pvbn_of_parent'})              = split(/\s+/, $line);}
            elsif ($line =~ /^ckse-dbn/)                   {(undef, $db_hash{'ckse_dbn'})                    = split(/\s+/, $line);}
            elsif ($line =~ /^ckse-checksum/)              {(undef, $db_hash{'ckse_checksum'})               = split(/\s+/, $line);}
            elsif ($line =~ /^ckse-buftreeid/)             {(undef, $db_hash{'ckse_buftreeid'})              = split(/\s+/, $line);}
            elsif ($line =~ /^ckse-fbn/)                   {(undef, $db_hash{'ckse_fbn'})                    = split(/\s+/, $line);}
            elsif ($line =~ /^ckse-cp-count/)              {(undef, $db_hash{'ckse_cp_count'})               = split(/\s+/, $line);}
            # This will hold anything that didn't match.
            else                                           {        $db_hash{'parse_error'} .= $line }
        }
    } else {
        # Type is ifile|file|vbn|dbn
        # Strip the first line of output and push the rest
        # into $db_options{'block_contents'}

        foreach my $x (split(/\,/, $temp[0])) {
            if ($x =~ /vbn\s/i) {
                # Strip whitespace, VBN, or vbn and then store just the data.
                $x =~ s/(\s+|VBN|vbn)//g;
                $db_hash{'vbn'} = $x;
            }
            elsif ($x =~ /disk\s/) {
                # Strip whitespace and disk tag and then store just the data.
                $x =~ s/(\s+|disk)//g;
                $db_hash{'disk'} = $x;
            }
            elsif ($x =~ /dbn\s/) {
                # Strip whitespace and dbn tag and then store just the data.
                $x =~ s/(\s+|dbn)//g;
                $db_hash{'dbn'} = $x;
            }
        }
        @temp = @temp[1 .. $#temp] if ($temp[0] !~ m/cksum_data/i);
        foreach $line (@temp) {$db_hash{'block_contents'} .= $line;}
    }

    # Strip any endlines off of the values in %db_hash
    # unless its the block_contents key.

    foreach my $element (keys %db_hash) {
        unless ($element eq 'block_contents') {
            $db_hash{$element} =~ s/\n//g;
        }
    }

    # Return prompt to original mode.
    Tharn::say($self->{connection}, "priv set $priv_mode", -swallow, '-timeout', 30);

    return(\%db_hash);
}

#-------------------------------------------------------------------------------

## @name ::dump($)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Note OnTap supports more options and arguments then I do for this port.
## If you need support for other options, please build it in as the design
## is flexible enough to allow for that.  Below are the only options that
## this method will be supporting for now.
##
## The dump command examines files in tree and writes to tape the files
## that need to be backed up. The Data ONTAP dump command differs
## slightly from the standard UNIX dump, but the output format is
## compatible with Solaris ufsrestore.
##
## The method will return $ERR_STATUS on failure and $GOOD_STATUS
## on a successful dump.  Please check the filer object's {errmsg}
## datamember for error messages.
##
## Below is a subset of OnTap's usage.  Ours will be shown via the
## synopsis and arguments.
##
## *** THE FOLLOWING IS ONTAP USAGE ***
## vnv3050f9a> dump
## usage:
##  dump [ <options> [ <arguments> ... ] ] <path>
##
##  <options> are a string of characters.  Option flags include:
##  {0123456789} f l n u v
##
##  The following flags require an argument:  f n
##  Arguments must be ordered to correspond with flag order.
##
##  <path> is the name of the tree to be dumped.
## ***------------------------------***
##
##
## This method must use the console_connection if and only if the
## "l" multi-subtree dump is used due to the fact that we have to wait
## for a certain prompt and provide input via STDIN.  Otherwise we will
## use the filers default connection stored in the objects tharnhost file
## i.e. $self->{connection}.  This is usually rsh for filers.
##
## The QA engineer running a test that calls on this method should be
## informed to look at a few files in order to make observations that
## are affected by filer::dump().  These are:
## - path from $f i.e. /vol/vol0/etc/log/<backup_file_name>
## - /etc/dumpdates <-- holds records of dumpdates when '$u' is true
## - /vol/vol0/etc/log/backup & look for the $n dumpname
##
## This was tested in 7.3 and 7-mode (OnTap 8.0)
##
## @synopsis
## $status = $filer_obj->dump("-d_lev=0", "-u", "-f=null", "-n=vseries_test",
##                 "-l=orig_file1.1,orig_file1.24,orig_file1.32,orig_file1.60",
##                 "-path=/vol/vol0/etc/log");
##
## @arg $path  Required: The root path that you would like dumped.  If 'l' is
##                       specified you will need to specify a comma separated
##                       list of files rooted at path.
## @arg $f     Optional: Write backup to specified file.  I suggest
##                       /vol/vol0/etc/log/<backup_file_name> or null.
## @arg $l     Optional: Specifies this is a multi-subtree dump.  With option
##                       'l' you MUST specify option n.  Because of option 'l'
##                       we will be prompted to list additional files rooted at
##                       <path> supplied via STDIN and terminated with a blank
##                       line.  You must supply those files rooted at path as a
##                       comma separated list "l=orig_fil1.1,orig_file1.24".
## @arg $n     Optional: Dumpname for a multi-subtree dump (by using 'l' flag).
##                       This name will be denoted in the
##                       /vol/vol0/etc/log/backup file prefixing each line with
##                       dmp date <dumpname>(dumplevel (0-9)) Start
##                       (Level <level_no>) like this:
##                       dmp Wed Apr 23 11:38:27 GMT 12_bkup(0) Start (Level 0)
##                       where 12_bkup is the dumpname and the dump level is
##                       zero.  Please see the bottom of this method for an
##                       example of /vol/vol0/etc/log/backup output.
## @arg $u     Optional: Update the /etc/dumpdates file AFTER a successful
##                       dump. This file just lists individual dumps that took
##                       place on the particular system with time stamps and
##                       subtree/increment level data.
## @arg $v     Optional: Verbose mode, dump will print out detailed info during
##                       the dump.
## @arg $d_lev Optional: The dump level (0-9) zero indicates a full level dump.
##                       The other 1-9 dump levels vary in depth of the nature
##                       of what types of things get dumped.  See the man pages
##                       for help on that.
##
sub dump($) {
    my $self             = shift; # Filer object
    my ($after, $prompt) = '';    # Variables used by Tharn::say()
    my @options          = ();    # Array used to hold ordered options
    my @arguments        = ();    # Array used to hold ordered arguments
    my $std_in           = '';    # A string of filenames from $l
    my $dump_name        = '';    # The dumpname from $n
    my $file_name        = '';    # The file name from $f flag
    my $cmd              = '';    # Final constructed cmd sent to OnTap
    my $dmp_time_out     = 600;   # Time to wait for a dump to end
    my $any_line       = '^.*\n'; # Regex used to match any line of output.
    my $std_in_prompt    = 'DUMP: Reading file names from standard input';

    # Put @_ in @ARGV as GetOptions uses @ARGV for processing options.
    @ARGV = @_;

    # These are all false by default.
    my($path,   # See @arg's section for info on these.
       $f,
       $l,
       $n,
       $u,
       $v,
       $d_lev
       ) = '';

    # Grab our options that were passed into this method.
    GetOptions( "path=s"    => \$path,    # Mandatory string
                "f:s"       => \$f,       # Optional string
                "l:s"       => \$l,       # Optional string
                "n:s"       => \$n,       # Optional string
                "u"         => \$u,       # Optional flag
                "v"         => \$v,       # Optional flag
                "d_lev:i"   => \$d_lev    # Optional integer
              );

    # Further evaluate our passed in parameters.
    # We can't have $l without $n
    if ($l && !$n){
        $self->{error}  = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::dump(): A multi subtree dump requires a dump name.";
        return($ERR_STATUS);
    }

    # More argument processing for $l.  Need to ensure its a comma separated
    # list of files. $l now looks something like this:
    # 'orig_fil1.1,orig_file1.24,orig_file1.32,orig_fil1.60' & we need to convert
    # that into the following format:
    # orig_file1.1\n
    # orig_file1.24\n
    # \n <<------------ This blank end line denotes the end of STDIN.
    if ($l) {
        foreach (split(/\,/, $l)){
            $std_in .= "$_\n";
        }
        # Now append the final '\n' carrige return to terminate STDIN.
        $std_in .= "\n";
        push(@options, 'l');
    }

    # Capture the dumpname if the $n option exists.  Also the $n option
    # requires an argument listed in the same order that the option was listed.
    if ($n) {
        $dump_name = $n;
        push(@options, 'n');
        push(@arguments, $dump_name);
    }

    # If $f is specified capture the filename where the dump will be sent.
    if ($f) {
        $file_name = $f;
        push(@options, 'f');
        push(@arguments, $file_name);
    }

    # Should we update /etc/dumpdates?
    if ($u) {
        push(@options, 'u');
    }

    # Should the logging be verbose?
    if ($v) {
        push(@options, 'v');
    }

    # What dump level did the user specify?  If none, OnTap defaults
    # to level 0 (full level dump).
    if ($d_lev =~ /\d/) {
        push(@options, $d_lev);
    }

    # Now we are to build our complete dump command by appending to
    # $cmd the appropriate options and arguments in order.

    # First append the options.
    $cmd .= "dump ";
    foreach (@options) {
        $cmd .= $_;
    }

    # Now append the arguments.
    $cmd .= " ";
    foreach (@arguments) {
        $cmd .= "$_ ";
    }

    # Chop the last space off from the arguments, we do this because you
    # don't need to specify a dump command that takes arguments in all cases.
    chop($cmd);

    # Finally append the path
    $cmd .= " $path";

    # Now we will automate something similar to what is seen below:
    #vnv3050f9a> dump 0ufnl null 14_bkup /vol/vol0/etc/log
    #DUMP: creating "/vol/vol0/../snapshot_for_backup.9" snapshot.
    #DUMP: Using Partial Volume Dump of selected subtrees
    #DUMP: Dumping tape file 1 on null
    #DUMP: Date of this level 0 dump: Wed Apr 23 12:06:53 2008.
    #DUMP: Date of last level 0 dump: the epoch.
    #DUMP: Dumping /vol/vol0/etc/log to null
    #DUMP: mapping (Pass I)[regular files]
    #DUMP: Reading file names from standard input
    #orig_file1.1
    #orig_file1.24
    #orig_file1.32
    #orig_file1.60
    # << ------------------------------------------Notice here, we end STDIN w/ a CR (\n)
    #DUMP: mapping (Pass II)[directories]
    #DUMP: estimated 477 KB.
    #DUMP: dumping (Pass III) [directories]
    #DUMP: dumping (Pass IV) [regular files]
    #DUMP: dumping (Pass V) [ACLs]
    #DUMP: 2076 KB
    #DUMP: DUMP IS DONE <<-- This line means we had a successful dump.
    #DUMP: Deleting "/vol/vol0/../snapshot_for_backup.9" snapshot.
    #vnv3050f9a>


    # Set a max timeout for our while loop to 30 minutes.
    my $end_time = time() + 1800;

    # Set a flag to serve as a condition for exiting the loop.
    my $exit_flag = 0;

    # Initialize a counter $i to zero for loop control.
    my $i = 0;

    while ((time() < $end_time) && (!$exit_flag)) {

        # If multisubtree, we must use the console connection.
        if ($l && $i == 0) {
            (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, "$cmd", -swallow, "-prompt", "$std_in_prompt", "-timeout", $dmp_time_out);

            # If true, we need to provide STDIN w/ files from $l.
            if ($prompt =~ /$std_in_prompt/) {

                # Send this blindly so we can quickly move on and listen.
                Tharn::say($self->{console_connection}, "$std_in", -swallow, "-blind");
            }

            # Increment the counter so we don't send the dump again.
            $i++;
        }
        elsif ($i == 0) {

            # Send this blindly so we can quickly move on and listen.
            Tharn::say($self->{connection}, "$cmd", -swallow, "-blind");

            # Increment the counter so we don't send the dump again.
            $i++;
        }

        # Now we need to listen on the console for various prompts.
        (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, "", "-prompt", $any_line, "-timeout", $dmp_time_out, "-maxidle", $dmp_time_out);

        if ($prompt =~ /^DUMP: DUMP IS DONE/){

            # Mark the $exit_flag accordingly, as this is good news.
            $exit_flag = 1;

            # Just listen again until we get the normal prompt.  At this point
            # we're just waiting for the filer to delete a snapshot it created
            # during the DUMP processs.  This won't take long...
            (undef, undef, $after) = Tharn::say($self->{console_connection}, "", "-timeout", 60);

            # A successful dump.
            return ($GOOD_STATUS);
        }
        # What the heck! -- Ontap will panic w/ lowercase at times and
        # it doesn't need to be anchored to the beginning of a new line.
        elsif ($prompt =~ /panic:/i) {

            $self->{error}  = $ERR_STATUS;
            $self->{errmsg} = "Error: filer::dump(): Encountered a PANIC! during backup.\n$prompt";

            # We had a PANIC - that can be bad or expected.
            return ($ERR_STATUS);
        }
        elsif ($prompt =~ /^usage:/) {

            $self->{error}  = $ERR_STATUS;
            $self->{errmsg} = "Error: filer::dump(): Encountered a usage issue.\n$prompt";

            # Read the header.
            return ($ERR_STATUS);
        }
    } # while ((time() < $end_time) && (!$exit_flag))

    # Cover some unlikely scenarios:

    if (time() > $end_time) {

        $self->{error}  = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::dump(): DUMP took longer than 30 minutes to complete and didn't finish.";
        return ($ERR_STATUS);
    } else {
        return ($GOOD_STATUS);
    }

    # When we vi /vol/vol0/etc/log/backup it looks like this:
    #dmp Wed Apr 23 11:38:27 GMT 12_bkup(0) Start (Level 0)
    #dmp Wed Apr 23 11:38:27 GMT 12_bkup(0) Options (b=63, u, l)
    #dmp Wed Apr 23 11:38:27 GMT 12_bkup(0) Snapshot (snapshot_for_backup.2, Wed Apr 23 11:38:26 GMT)
    #dmp Wed Apr 23 11:38:27 GMT 12_bkup(0) Error (Standard out is not a socket.)
    #dmp Wed Apr 23 11:38:59 GMT 12_bkup(0) Abort (0 MB)
    #dmp Wed Apr 23 11:40:26 GMT 12_bkup(0) Start (Level 0)
    #dmp Wed Apr 23 11:40:26 GMT 12_bkup(0) Options (b=63, u, l)
    #dmp Wed Apr 23 11:40:26 GMT 12_bkup(0) Snapshot (snapshot_for_backup.3, Wed Apr 23 11:40:24 GMT)
    #dmp Wed Apr 23 11:40:26 GMT 12_bkup(0) Tape_open (file)
    #dmp Wed Apr 23 11:40:26 GMT 12_bkup(0) Phase_change (I)
    #dmp Wed Apr 23 11:41:31 GMT 12_bkup(0) Phase_change (II)
    #dmp Wed Apr 23 11:41:33 GMT 12_bkup(0) Phase_change (III)
    #dmp Wed Apr 23 11:41:34 GMT 12_bkup(0) Phase_change (IV)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Phase_change (V)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Tape_close (file)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) End (2 MB)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (reg inodes: 4 other inodes: 0 dirs: 1 nt dirs: 0 nt inodes: 0 acls: 0)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 1 time: 65287)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 3: directories dumped: 1)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 3: wafl directory blocks read: 1)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 3: average wafl directory blocks per inode: 1)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 3: average tape blocks per inode: 1)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 3 throughput (MB sec): read 0 write 0)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Percent of phase3 time spent for: reading inos 0% dumping ino 0%)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Percent of phase3 dump time spent for: convert-wafl-dirs 0% lev0-ra 0%)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 3 averages (usec): wafl load buf time 0 level 0 ra time 0)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 4: inodes dumped: 4)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 4: wafl data blocks read: 117)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 4: average wafl data blocks per inode: 29)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 4: average tape data blocks per inode: 117)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Phase 4 throughput (MB sec): read 1 write 1)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Percent of phase4 time spent for: reading inos 0% dumping inos 9%)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Tape write times (msec): average: 0 max: 1)
    #dmp Wed Apr 23 11:41:35 GMT 12_bkup(0) Log_msg (Tape changes: 1)
    # Note the "12_bkup(0)" is the dumpname which comes from the $n option.
}

#-------------------------------------------------------------------------------

## @name ::get_priv()
##
## @author James Silva (silva@netapp.com)
##
## @description
## Returns the priv mode that the filer is currently in.
##
## @synopsis
## $priv = $filer_obj->get_priv();
##
sub get_priv(){
    my $self  = shift;  # Filer object to shift in
    my $after = '';     # Returned output from Tharn::say()

    # Send 'priv' to the filer, it will respond with the current priv mode:
    # vnv3050f9b*> priv
    # advanced
    # vnv3050f9b*>
    (undef, undef, $after) = Tharn::say($self->{connection}, "priv", -swallow, "-timeout", 60);

    # Strip any whitespace out of our returned output since we desire solely
    # the mode, and nothing else.  i.e 'priv' instead of ' priv\n' or some-
    # thing of that nature.
    $after =~ s/(\s+|\n|\r)//g;

    # $after now holds your priv mode, so return it.
    return ($after);
}

#-------------------------------------------------------------------------------

## @name ::rdfile($)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Returns a string containing any raw output from 'rdfile $file'.
##
## @synopsis
## $output = $filer_obj->rdfile($file);
##
sub rdfile($){
    my $self  = shift;  # Filer object to shift in
    my $file  = shift;
    my $after = '';     # Returned output from Tharn::say()

    # Send 'rdfile $file' to the filer, it will return the raw output:
    # vnv3050f9b> rdfile /etc/hosts
    # Auto-generated by setup Fri Feb 22 04:02:45 GMT 2008
    # 127.0.0.1 localhost
    # 10.97.120.18    vnv3050f9b vnv3050f9b-e0a
    # vnv3050f9b>

    (undef, undef, $after) = Tharn::say($self->{connection}, "rdfile $file", -swallow, "-timeout", 60);

    # $after now holds raw output.
    return ($after);
}

#-------------------------------------------------------------------------------

## @name ::rm($)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Removes the file located at the absolute path: $file
## For this to work you must specify the absolute path to a particular file
## such as /vol/vol0/etc/messages.
##
## Vseries will be placed into 'priv set diag' mode during this
## methods use, but will be returned to the mode it found the shell at.
##
## @synopsis
## $status = $filer_obj->rm($file);
##
## @arg $file    Required: Absolute path to a file that you wish to remove.
##
sub rm($) {

    my $self         = shift;               # Filer object
    my $file         = shift;               # File to remove.
    my $after        = '';                  # Holds the raw output from Tharn::say()
    my $diag         = 'priv set -q diag;'; # String to change into diag priv mode.
    my $priv         = $self->get_priv();   # Store the current priv mode.

    (undef, undef, $after) = Tharn::say($self->{connection}, "$diag rm $file", -swallow, '-timeout', 30);

    # All errors are prefaced with the format /^.*:/
    # But, silence is success, i.e. $after eq null:

    if ($after ne '') {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::rm(): Remove did not complete successfully. Ensure file path is correct.\n $after";

        # Change the privs back to original mode.
        Tharn::say($self->{connection}, "priv set $priv", -swallow, '-timeout', 30);

        return($ERR_STATUS);
    }

    # Validate the remove success.
    my ($filename, $directories, $suffix) = File::Basename::fileparse($file);
    $filename .= $suffix;

    # Do an ls on the dir provided to get dir contents.
    (undef, undef, $after) = Tharn::say($self->{connection}, "$diag ls $directories", -swallow, '-timeout', 30);
    my @temp = split (/\n/, $after);
    my $found_file_flag = 0;

    # Make sure our file no longer exists.
    foreach my $file_entry (@temp) {$file_entry =~ /^$filename$/ ? $found_file_flag = 1 : next;}

    if ($found_file_flag) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::rm(): Unable to remove $filename in $directories.";

        # Change the privs back to original mode.
        Tharn::say($self->{connection}, "priv set $priv", -swallow, '-timeout', 30);
        return($ERR_STATUS);
    }

    # Change the privs back to the original mode.
    Tharn::say($self->{connection}, "priv set $priv", -swallow, '-timeout', 30);

    return($GOOD_STATUS);
}

#--------------------------------------------------------------------------------

## @name ::wafl_iron_start($%)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Start wafliron on the tradvol/aggr supplied.  It always returns
## either $GOOD_STATUS or $BAD_STATUS.  Depending on the options within
## the $op_ref hash this method will start wafliron in different manners.
## If you'd like more options, add code to support them.
## If you want to wafliron a flex vol, we'll iron the containing aggr.
##
## Supported Options:
## $op_ref->{'wait'} = 1; We'll wait until wafliron completes.
##                   = 0; We'll start wafliron blindly and exit.
##                        Note depending on the size of the vol/aggr this may
##                        take a long time to complete.  User beware.
##                        We'll wait until wafliron completes.
##
## $op_ref->{'poll_time'} = 86400;
##                        Defaults to one day in seconds but user
##                        may set to whatever they wish.
##
## Vseries will be placed into 'priv set diag' mode during this
## methods use, but will be returned to the mode it found the shell at.
##
## @synopsis
## $status = $filer_obj->wafl_iron_start($vol, $op_ref);
##
## @arg $vol       Required: Volume or aggr to run wafliron on.
##                           Note: don't need /vol/, just vol name will do.
## @arg $op_ref    Optional: A hash reference with options.
##
sub wafl_iron_start($%) {
    my $self      = shift;               # Filer object to use.
    my $vol       = shift;               # The vol/aggr to run wafliron on.
    my $op_ref    = shift;               # Options hash ref.
    my %options   = %$op_ref;            # Dereferenced hash.
    my $priv      = $self->get_priv();   # Store the current priv mode.
    my $container = '';                  # Container of a possible flex vol.
    my $cmd       = '';                  # Holds the cmd that we'll build.

    # Start building the $cmd.
    $cmd = "priv set -q diag; aggr wafliron start ";

    # Need to know if this is a flex vol, if so we need to iron its container.
    # Call 'vol status' to get the volume list.
    my $vol_list_ref = $self->vol_status();

    # Check for errors resulting from the VSeries::filer::vol_status() method.
    if ($self->get_error_state() == $ERR_STATUS) {
        $self->{error}  = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::wafl_iron_start(): A failure occured in filer::vol_status().";
        return ($ERR_STATUS);
    }

    # Dereference the volume list.
    my @vol_list = @$vol_list_ref;

    # Send this @vol_list out to be parsed.
    my $vol_hash_ref = VSeries::utils::parse_vol_status("", @vol_list);

    # Dereference our new volume hash.
    my %vol_hash = %$vol_hash_ref;

    # Is $vol a flex?  $vol can be an aggr too so when a flex vol is
    # down (b/c it is marked inconsistent) this will still work, as the
    # user should pass in the containing aggr name.
    if ($vol_hash{$vol}{status} =~ /flex/i) {
        # Get the containing volume.
        $container = $self->get_flex_vol_container($vol);

        # Complete $cmd:
        $cmd .= $container;
    } else {
        # This means the $vol was a trad vol or aggr.
        $cmd .= $vol;
    }

    # Kick off wafliron blindly, if told to wait, we'll listen on the console.
    Tharn::say($self->{connection}, "$cmd", -blind);

    # Check if we are to wait until wafliron completes to return.
    if (defined $options{'wait'} && $options{'wait'} > 0) {

        # Need to poll the filer until we see 'wafliron completed'.
        # We use 1 day as a poll time by default, but a user may have
        # specified another value for poll_time.
        my $poll_time = 86400;

        if (defined $options{'poll_time'} && $options{'poll_time'} > 0 ) {
            $poll_time = $options{'poll_time'};
        }

        # Wafliron is done when we see this message:
        my $poll_pattern = "(wafliron completed|wafl\.scan\.iron\.done|Aggregate.*\'$container\'.*is.*now.*online|Volume.*\'$vol\'.*is.*now.*online)";

        # Change to console mode:
        $self->{use_console} = "true";

        $self->poll_filer($poll_time, $poll_pattern);

        # Revert to console mode:
        $self->{use_console} = "false";

        # Change priv back on filer.
        Tharn::say($self->{connection}, "priv set $priv", -swallow, '-timeout', 30);

        # Check for errors:
        if ($self->get_error_state()) {
            return($ERR_STATUS);
        }
    }

    # Change priv back on filer.
    Tharn::say($self->{connection}, "priv set $priv", -swallow, '-timeout', 30);

    return($GOOD_STATUS);
}

#--------------------------------------------------------------------------------

## @name ::get_flex_vol_container($)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Return the container name for the passed in flex volume.
## Doesn't check if the passed in volume is a flex, so the user
## should take care when using this method.
##
## @synopsis
## $container = $filer_obj->get_flex_vol_container($flex_vol);
##
## @arg $flex_vol  Required: Flex vol to retrieve a container from.
##
sub get_flex_vol_container($) {
    my $self = shift;
    my ($flex_vol) = @_;

    # Get the containing volume.
    my $container = (Tharn::say($self->{connection}, "vol container $flex_vol", -swallow, '-timeout', 30))[2];

    # container now looks like this:
    #Volume 'jimmy_flex' is contained in aggregate 'jimmy_aggr'

    # We want the last element of this split:
    $container = (split(/\s+/, $container))[-1];

    # Strip the single quotes left: 'jimmy_aggr'
    $container =~ s/\'//g;

    # Return the $container.
    return($container);
}

#--------------------------------------------------------------------------------

## @name ::disk_remove_ownership($)
##
## @description
## Takes in a serial number for a disk, looks up the corresponding name on this
## given filer and removes it using disk remove_ownership.  If successfull it
## will return the empty string, otherwise it will return whatever the error was.
## This method does not actually query the filer to determine if the disk is seen,
## it expects discover_disks() to have already been called for the given filer
##
## @synopsis
## $filer_obj->disk_remove_ownership($disk_serial);
##
## @arg $disk_serial Required: Serial number of the disk you want the filer to
## remove ownership of.
##
##
sub disk_remove_ownership($){
    my $self = shift;
    my ($lun_serial) = @_;
    my $lun_name = "";
    my $status = "";
    my $verify = "";
    my $success = 0;

    # Look up the serial number for the passed in disk, if we can find it then
    # we'll use the primary path for the remove ownership command, if we can't
    # find the disk serial number then we return an error.
    foreach my $disk_serial (keys(%{$self->{disks}})){
        if($disk_serial eq $lun_serial){
            $lun_name = $self->{disks}->{$disk_serial}->{primary};
            last;
        }
    }

    # If the name is still blank it means we didn't find the serial in our disk
    # hash so return an error.
    if($lun_name eq ""){
        $status = "Could not find disk with Serial Number " . $lun_serial . " on filer " . $self->{name};
        $self->{error} = 1;
        $self->{errmsg} = $status;
        return $ERR_STATUS;
    }

    my $cmd = "disk remove_ownership $lun_name";
    my $confirm_cmd = "yes";
    my $verify_cmd = "disk show -v";

    if($self->{use_console} eq "true"){
        (undef, undef, $status) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-noprompt", "-prompt", "Volumes must be taken offline.", "-timeout", 60);
        (undef, undef, $status) = Tharn::say($self->{console_connection}, $confirm_cmd, "-prompt", "^\*?>", "-timeout", 60);
        (undef, undef, $verify) = Tharn::say($self->{console_connection}, $verify_cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else{
        (undef, undef, $status) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
        (undef, undef, $status) = Tharn::say($self->{connection}, $confirm_cmd, -swallow, "-timeout", 60);
        (undef, undef, $verify) = Tharn::say($self->{connection}, $verify_cmd, -swallow, "-timeout", 60);
    }


    my @output_array = split(/\n/, $verify);
    # Chop off the header of the output
    @output_array = @output_array[2..$#output_array];
    my ($disk_name, $disk_owner1, $disk_owner2, $disk_pool, $disk_serial, $disk_chksum) = '';
    foreach my $line (@output_array){
        # Adding catch for case where serial number and checksum aren't
        # separated by a whitespace
        if($line =~ m/(Block|Zone)/i){
            $line =~ s/(Block|Zone)/ \1/i;
        }

        # Split out the line so that we can validate the removal of ownership
        # for the disk.
        ($disk_name, $disk_owner1, $disk_owner2, $disk_pool, $disk_serial, $disk_chksum) = split(/ +/, $line);
        if($disk_name eq $lun_name){
            if(($disk_owner1 . " " . $disk_owner2) eq "Not Owned"){
               $success = 1;
            }
            else{
               $self->{error} = 1;
               $self->{errmsg} = "Unable to verify success of disk remove_ownership $lun_name \n $status";
               return;
            }
            last;

        }
    }
}



## @name ::go_to_cfe()
##
## @description
## Takes a filer from whatever point it currently is at and brings it to the
## CFE/LOADER prompt.  It is assumed that the filer is properly cf disabled
## otherwise subsequent booting and such will be affected. It will return the
## matched prompt in case the caller wants that information.
##
## @synopsis
## $returned_prompt=$filer_obj->go_to_cfe();
##
##
##
sub go_to_cfe(){
    my $self = shift;

    # Use all valid prompts as a starting point, smaller sets of prompts
    # will be used throughout to make sure the filer gets to where it is
    # supposed to
    my @start_prompts = ('^CFE>', '^LOADER\\*?>','^ok ', '^\*>', 'Selection (.*)\?', '(^\r\n$)*^Please', '^.*\n$', '^LOADER-(A|B)\*?>');
    my @CFE_prompts = ('^CFE>', '^LOADER\\*?>','^ok ', '^LOADER-(A|B)\*?>');
    my $maint_prompt = '^>? *\*>';
    my @norm_prompts = @{$self->{conn}->prompt_norm()};
    my $returned_prompt = undef;
    my $matched_prompt = 0;

    # Keeping these here for future debugging when necessary.
    my ($foo, $bar, $baz) = "";
    # First we have to figure out where we are
    ($foo, $bar, $baz, $returned_prompt) = Tharn::say($self->{console_connection}, " ", -swallow, "-prompt", @start_prompts);


    # Based on where we're at, make decisions that will get the filer to the CFE/LOADER prompt
    if($returned_prompt =~ m/^CFE>/ || $returned_prompt =~ m/^LOADER\\*?>/ || $returned_prompt =~ m/^ok/ || $returned_prompt =~ m/^LOADER-(A|B)\*?>/){
        Tharn::logcomment("Already at CFE/LOADER prompt");
    }
    elsif($returned_prompt =~ m/^\*?>/ ){
        Tharn::logcomment("At Maint Mode prompt, halting to CFE/LOADER prompt");
        (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "halt", "-noprompt", "-prompt", @CFE_prompts);
    }
    elsif($returned_prompt =~ m/Autoboot of PRIMARY image aborted/){
        Tharn::logcomment("Connecting to console resulted in aborting AUTOBOOT, booting to CFE prompt");
        (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "\n", "-noprompt", "-prompt", @CFE_prompts);
    }
    elsif($returned_prompt =~ m/Selection \(.*\)\?/ || $returned_prompt =~ m/(^\r\n$)*^Please/){
        Tharn::logcomment("At Special Menu, continuing to maintenance mode and halting..");
        Tharn::say($self->{console_connection}, "5", "-noprompt", "-prompt", "Continue");
        Tharn::say($self->{console_connection}, "yes", "-noprompt", "-prompt", $maint_prompt);
        (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "halt", "-noprompt", "-prompt", @CFE_prompts);

    }
    else{

        # Handle the case where we are either explicitly booted or in the boot
        # sequence.  Since the format for the normal prompts could be a few
        # different ones it has to be a looped check.  If it's not that then
        # the last possibility is that it's output from the boot sequence.
        foreach my $prompt (@norm_prompts){
            if($returned_prompt =~ m/$prompt/){
                $matched_prompt = 1;
                (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "halt", "-noprompt", "-prompt", @CFE_prompts);
                last;
            }

        }

        if($returned_prompt =~ m/.*\n/ && $matched_prompt == 0){
            Tharn::logcomment($self->get_name() . " appears to be in boot sequence, this may be due to lack of a root vol, booting to CFE/LOADER prompt");
            $self->go_to_special_menu();
            Tharn::say($self->{console_connection}, "5", "-noprompt", "-prompt", "Continue");
            Tharn::say($self->{console_connection}, "yes", "-noprompt", "-prompt", $maint_prompt);
            (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "halt", "-noprompt", "-prompt", @CFE_prompts);
        }
    }

    return $returned_prompt;


}


## @name ::go_to_special_menu()
##
## @description
## Takes a filer from whatever point it currently is at and brings it to the
## special menu.  It is assumed that the filer is properly cf disabled
## otherwise subsequent booting and such will be affected. It will return the
## special menu in case the caller wants that information to make subsequent
## calls.
##
## @synopsis
## @returned_menu=$filer_obj->go_to_special_menu();
##
##
##
sub go_to_special_menu(){
    my $self = shift;
    my $tmpprompt = ".*\n";
    my @start_prompts = ('^CFE>', '^LOADER\\*?>','^ok ', '^\*>', 'Selection (.*)\?', '(^\r\n$)*^Please', '^.*\n$', '^LOADER-(A|B)\*?>');
    my $maint_prompt = '^>? *\*>';
    my @CFE_prompts = ('^CFE>', '^LOADER\\*?>','^ok ', '\[\d+\]\:', '^LOADER-(A|B)\*?>');
    my $CIFS_prompt = '\[\d+\]\:';
    my $returned_prompt = undef;
    my @norm_prompts = @{$self->{conn}->prompt_norm()};
    my $h = $self->{conn};
    my $output = "";
    my @menu_array = ();
    my $matched_prompt = 0;
    my $at_special_menu = 0;

    # First we have to figure out where we are
    (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, " ", -swallow, "-noprompt", "-prompt", @start_prompts);

    # Based on where we're at, make decisions that will get the filer to the CFE/LOADER prompt
        if($returned_prompt =~ m/^CFE>/ || $returned_prompt =~ m/^LOADER\\*?>/ || $returned_prompt =~ m/^ok/ || $returned_prompt =~ m/^LOADER-(A|B)\*?>/){
        Tharn::logcomment("At CFE/LOADER prompt, entering bye to start boot sequence");
        Tharn::say($self->{console_connection}, "bye", -blind);
    }
    elsif($returned_prompt =~ m/^\*?>/ ){
        Tharn::logcomment("At Maint Mode prompt, halting and starting boot sequence");
        (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "halt", "-noprompt", "-prompt", @CFE_prompts);
        Tharn::say($self->{console_connection}, "bye", -blind);
    }
    elsif($returned_prompt =~ m/Autoboot of PRIMARY image aborted/){
        Tharn::logcomment("Connecting to console resulted in aborting AUTOBOOT, booting to CFE prompt then continuing");
        Tharn::say($self->{console_connection}, "bye", -blind);

    }
    elsif($returned_prompt =~ m/Selection \(.*\)\?/ || $returned_prompt =~ m/(^\r\n$)*^Please/){
        Tharn::logcomment("Already at Special Menu, continuing on..");
        $at_special_menu = 1;
    }
    else{

        # Handle the case where we are either explicitly booted or in the boot
        # sequence.  Since the format for the normal prompts could be a few
        # different ones it has to be a looped check.  If it's not that then
        # the last possibility is that it's output from the boot sequence.
        foreach my $prompt (@norm_prompts){
            if($returned_prompt =~ m/$prompt/){
                $matched_prompt = 1;
                (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "halt", "-noprompt", "-prompt", @CFE_prompts);
                if($returned_prompt =~ m/$CIFS_prompt/){
                    (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, '0', '-noprompt', "-prompt", @CFE_prompts);
                }
                last;
            }

        }

        if($returned_prompt =~ m/.*\n/ && $matched_prompt == 0){
            Tharn::logcomment($self->get_name() . " appears to be in the boot sequence, this may be due to the lack of a root vol, waiting to enter ctrl-c");
        }
    }

    # If we're not already at the special menu, wait during the boot sequence
    # to enter ctrl-c and bring up the special menu.
    while($at_special_menu != 1){
        (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "", "-noprompt", "-prompt", $tmpprompt);
        if($returned_prompt =~ m/special boot menu/){
            Tharn::say($self->{console_connection}, $h->{irptchar},-nocr, "-blind");
            Tharn::logcomment("Matched prompt, entered ctrl-c \n");
            last;
        }

    }

    # We know we're in the boot sequence and waiting for the special menu to
    # come up so listen on the connection and return the menu to the caller.
    (undef, $output, undef, undef) = Tharn::say($self->{console_connection}, "",-nocr, "-wait", "Selection (.*)\?", "-ok");
    @menu_array = split(/\n/, $output);
    @menu_array = @menu_array[-10 .. -1];
    return \@menu_array;

}



## @name ::go_to_maintenance_mode()
##
## @description
## Takes a filer from whatever point it currently is at and brings it to
## maintenance mode.  It is assumed that the filer is properly cf disabled
## otherwise subsequent booting and such will be affected. It will return the
## matched prompt in case the caller wants that information.
##
## @synopsis
## $returned_prompt=$filer_obj->go_to_maintenance_mode();
##
##
##
sub go_to_maintenance_mode(){
    my $self = shift;
    my @start_prompts = ('^CFE>', '^LOADER\\*?>','^ok ', '^\*>', 'Selection (.*)\?', '(^\r\n$)*^Please', '^.*\n$', '^LOADER-(A|B)\*?>');
    my @CFE_prompts = ('^CFE>', '^LOADER\\*?>','^ok ', '^LOADER-(A|B)\*?>');
    my $maint_prompt = '^>? *\*>';
    my @norm_prompts = @{$self->{conn}->prompt_norm()};
    my $returned_prompt = undef;
    my ($foo, $bar, $baz) = "";
    my $matched_prompt = 0;

    # First we have to figure out where we are
    ($foo, $bar, $baz, $returned_prompt) = Tharn::say($self->{console_connection}, " ", -swallow, "-noprompt", "-prompt", @start_prompts);

    # Based on where we're at, make decisions that will get the filer to the CFE/LOADER prompt
    if($returned_prompt =~ m/^CFE>/ || $returned_prompt =~ m/^LOADER\\*?>/ || $returned_prompt =~ m/^ok/ || $returned_prompt =~ m/^LOADER-(A|B)\*?>/){
        Tharn::logcomment("At CFE/LOADER prompt, booting to maint mode");
        $self->go_to_special_menu();
        Tharn::say($self->{console_connection}, "5", -blind);
        (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "yes", "-noprompt", "-prompt", $maint_prompt);

    }
    elsif($returned_prompt =~ m/^\*?>/ ){
        Tharn::logcomment("At Maint Mode prompt already");
        (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "\n", "-noprompt", "-prompt", $maint_prompt);

    }
    elsif($returned_prompt =~ m/Autoboot of PRIMARY image aborted/){
        Tharn::logcomment("Connecting to console resulted in aborting AUTOBOOT, booting to CFE prompt");
        $self->go_to_special_menu();
        Tharn::say($self->{console_connection}, "5", "-noprompt", "-prompt", "Continue");
        (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "yes", "-noprompt", "-prompt", $maint_prompt);
    }
    elsif($returned_prompt =~ m/Selection \(.*\)\?/ || $returned_prompt =~ m/(^\r\n$)*^Please/){
        Tharn::logcomment("Already at Special Menu, continuing on..");
        Tharn::say($self->{console_connection}, "5", "-noprompt", "-prompt", "Continue");
        (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "yes", "-noprompt", "-prompt", $maint_prompt);
    }
    else{

        # Handle the case where we are either explicitly booted or in the boot
        # sequence.  Since the format for the normal prompts could be a few
        # different ones it has to be a looped check.  If it's not that then
        # the last possibility is that it's output from the boot sequence.
        foreach my $prompt (@norm_prompts){
           if($returned_prompt =~ m/$prompt/){
                $matched_prompt = 1;
                (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "halt", "-noprompt", "-prompt", @CFE_prompts);
                $self->go_to_special_menu();
                Tharn::say($self->{console_connection}, "5", "-noprompt", "-prompt", "Continue");
                (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "yes", "-noprompt", "-prompt", $maint_prompt);
                last;
            }

        }


        if($returned_prompt =~ m/.*\n/ && $matched_prompt == 0){
            Tharn::logcomment($self->get_name() . " appears to be in the boot sequence, this may be due to lack of a root vol, booting to CFE/LOADER prompt");
            $self->go_to_special_menu();
            Tharn::say($self->{console_connection}, "5", "-noprompt", "-prompt", "Continue");
            (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "yes", "-noprompt", "-prompt", $maint_prompt);

        }

    }

    return $returned_prompt;


}

## @name ::go_to_boot_sequence()
##
## @description
## Takes a filer from whatever point it currently is at and brings it to
## the boot sequence.  It is assumed that the filer is properly cf disabled
## otherwise not being able to grab the root vol will affect booting and
## result in booting hanging.  This method will kick off the boot sequence
## blindly and exit rather than return a prompt/menu/output
##
## @synopsis
## $filer_obj->go_to_boot_sequence();
##
##
##
sub go_to_boot_sequence(){
    my $self = shift;
    my @start_prompts = ('^CFE>', '^LOADER\\*?>','^ok ', '^\*>', 'Selection (.*)\?', '(^\r\n$)*^Please', '^.*\n$', '^LOADER-(A|B)\*?>');
    my @CFE_prompts = ('^CFE>', '^LOADER\\*?>','^ok ', '^LOADER-(A|B)\*?>');
    my $maint_prompt = '^>? *\*>';
    my $returned_prompt = undef;
    my ($foo, $bar, $baz) = "";
    my @norm_prompts = @{$self->{conn}->prompt_norm()};
    my $matched_prompt = 0;

    # First we have to figure out where we are
    ($foo, $bar, $baz, $returned_prompt) = Tharn::say($self->{console_connection}, " ", -swallow, "-noprompt", "-prompt", @start_prompts);

    # Based on where we're at, make decisions that will get the filer to the CFE/LOADER prompt
    if($returned_prompt =~ m/^CFE>/ || $returned_prompt =~ m/^LOADER\\*?>/ || $returned_prompt =~ m/^ok/ || $returned_prompt =~ m/^LOADER-(A|B)\*?>/){
        Tharn::logcomment("At CFE/LOADER prompt, entering bye...");
        Tharn::say($self->{console_connection}, "bye", -blind);

    }
    elsif($returned_prompt =~ m/^\*?>/ ){
        Tharn::logcomment("At Maint Mode prompt halting to CFE and entering boot sequence");
        Tharn::say($self->{console_connection}, "halt", "-noprompt", "-prompt", @CFE_prompts);
        Tharn::say($self->{console_connection}, "bye", -blind);

    }
    elsif($returned_prompt =~ m/Autoboot of PRIMARY image aborted/){
       Tharn::logcomment("Connecting to console resulted in aborting AUTOBOOT, booting to CFE prompt and restarting boot sequence");
       Tharn::say($self->{console_connection}, "bye", -blind);
    }
    elsif($returned_prompt =~ m/Selection \(.*\)\?/ || $returned_prompt =~ m/(^\r\n$)*^Please/){
        Tharn::logcomment("Already at Special Menu, going to maint mode, then halting to restart boot sequence..");
        Tharn::say($self->{console_connection}, "5", "-noprompt", "-prompt", "Continue");
        Tharn::say($self->{console_connection}, "yes", "-noprompt", "-prompt", $maint_prompt);
        Tharn::say($self->{console_connection}, "halt", "-noprompt", "-prompt", @CFE_prompts);
        Tharn::say($self->{console_connection}, "bye", -blind);
    }
    else{
        # Handle the case where we are either explicitly booted or in the boot
        # sequence.  Since the format for the normal prompts could be a few
        # different ones it has to be a looped check.  If it's not that then
        # the last possibility is the it's output from the boot sequence.
        foreach my $prompt (@norm_prompts){
            if($returned_prompt =~ m/$prompt/){
                $matched_prompt = 1;
                (undef, undef, undef, $returned_prompt) = Tharn::say($self->{console_connection}, "halt", "-noprompt", "-prompt", @CFE_prompts);
                Tharn::say($self->{console_connection}, "bye", -blind);
                last;
            }

        }

        if($returned_prompt =~ m/.*\n/ && $matched_prompt == 0){
            Tharn::logcomment($self->get_name() . " already appears to be in boot sequence");

        }

    }

    return 0;


}

## @name ::software_install($;$)
##
## @description
## Expects a filer to be booted up to a normal prompt where this command is
## available.  Will perform a software installation takes in a url/location
## of the software image to install.  Will poll and monitor the progress while
## catching timeout issues.  Errors will be reported in the filer object's
## error flag and message.  Takes an optional flag that is the flag for the
## software install command itself on a filer.
##
## @synopsis
## $returned_status=$filer_obj->software_install($;$);
##
##
##
sub software_install($;$){
    my $self = shift;
    my ($software_url, $flag) = @_;
    my $exit = 1;

    # Bump the timeout because there may be intervals where polling to the
    # console may lag a little.
    my $LARGE_TIMEOUT = 240;

    # Prompt used in matching for progress on software installation
    my $listen_prompt = 'software: \d?\d?\d% file read from location.\r';

    # Prompt used to match the beginning of software installation
    my $install_prompt = 'software: installing software, this could take a few minutes...\r\r\n';

    # Seeded returned prompt with a silly value to make debug in the future
    # obvious for cases where this doesn't get populated.
    my $returned_prompt = "HOLA!!!!\n";

    # Create array of possible prompts for both positive and negative cases
    my @prompts = ($listen_prompt, "software: HTTP Status .*\n", "Unable to open connection .*\n");

    # Build the command that we'll send for software installation, add the
    # optional flag if provided
    my $cmd = "software install $software_url";
    if(defined($flag)){
        $cmd = $cmd . " " . $flag;

    }
    my ($match, $before, $after) = "";

    # Kick off the software installation and listen for either an error or
    # progress.
    ($match, $before, $after, $returned_prompt) = Tharn::say($self->{console_connection}, $cmd, "-noprompt", "-prompt", @prompts);

    # If we got an error prompt returned then set the error flag and message
    # and exit, otherwise continue and poll progress.
    if($returned_prompt =~ m/HTTP Status/ || $returned_prompt =~ m/Unable to open connection/){
        $self->{error} = 1;
        $self->{errmsg} = "Error during software install: $returned_prompt";
        return $ERR_STATUS;
    }

    # Poll software installation progress, breaking out once we see 100%
    while($exit == 1){
        ($match, $before, $after, $returned_prompt) = Tharn::say($self->{console_connection}, "", "-prompt", $listen_prompt);
        if($returned_prompt =~ m/software: 100%/){
            $exit = 0;

        }

    }

    # Now sit and listen for the prompt where software installation may take a
    # few minutes.
    ($match, $before, $after, $returned_prompt) = Tharn::say($self->{console_connection}, "", "-prompt", $install_prompt);

    # Now wait for software installation to complete and confirm its completion
    # LARGE_TIMEOUT is used since this can take a while with no actual output
    ($match, $before, $after, $returned_prompt) = Tharn::say($self->{console_connection}, "", "-timeout", $LARGE_TIMEOUT);

    # Confirm software installation completed successfully, otherwise report an
    # error.
    if($after !~ m/software: installation .*completed/){
        $self->{error} = 1;
        $self->{errmsg} = "ERROR DURING SOFTWARE INSTALLATION: $after";
        return $ERR_STATUS;
    }

    # If we got here, software installation was a success.
    return $GOOD_STATUS;





}


## @name ::download_software(;$)
##
## @description
## Expects a filer to be booted up to a normal prompt where this command is
## available.  Will perform a software download, and takes in an optional flag
## Will poll and monitor the progress while catching timeout issues.  Errors
## will be reported in the filer object's error flag and message.
##
## @synopsis
## $returned_status=$filer_obj->download_software(;$);
##
##
##
sub download_software(;$){
    my $self = shift;
    my ($flag) = @_;
    my ($match, $before, $after, $returned_prompt) = "";
    my $cmd = "download ";
    my @norm_prompts = @{$self->{conn}->prompt_norm()};
    my $break = 0;

    # Used to confirm download was successfull
    my $completed_prompt = "Operator requested download completed";

    # If an optional flag was passed in, append to the command we'll send over
    if(defined($flag)){
        $cmd = $cmd . $flag;
    }

    # Starting prompt in the download process.
    my @prompts = ("download: Downloading boot device");

    ($match, $before, $after, $returned_prompt) = Tharn::say($self->{console_connection}, $cmd, "-prompt", @prompts);

    # Add to the accepted prompts the prompt for completion and also the . that
    # is printed to the console to indicate progress.  With these in place,
    # poll until the download is completed.
    push(@prompts, '\.');
    push(@prompts, $completed_prompt);
    while($break == 0){
        ($match, $before, $after, $returned_prompt) = Tharn::say($self->{console_connection}, "", "-prompt", @prompts);

        # These checks effectively poll and if we get something other than the
        # '.' (dot) back then it completed or we get a normal prompt.
        if($returned_prompt =~ m/$completed_prompt/){
            $break = 1;
        }
        foreach my $prompt (@norm_prompts){
            if($returned_prompt =~ m/$prompt/){
                $break = 1;
                last;
            }

        }
    }

    return $GOOD_STATUS;


}


## @name ::get_printenv()
##
## @description
## Takes a filer from any point, down to the CFE/LOADER prompt and enters the
## printenv command, captures the output in an array with each line of output
## being an entry in the array, kicks off the boot process again and returns
## the output of printenv to the caller.
##
## @synopsis
## $printent_ref=$filer_obj->get_printenv();
##
##
##
sub get_printenv(){
    my $self = shift;
my @CFE_prompts = ('^CFE>', '^LOADER\\*?>','^ok ', '^LOADER-(A|B)\*?>');
    $self->go_to_cfe();
    my ($match, $before, $after, $returned_prompt) = "";
    my $cmd = "printenv";

    ($match, $before, $after, $returned_prompt) = Tharn::say($self->{console_connection}, $cmd, "-prompt", @CFE_prompts);
    my @output = ();
    @output = split(/\n/, $after);


    $self->go_to_boot_sequence();

    return \@output;


}


## @name ::get_version()
##
## @description
## A pass through command to get the version from a filer and pass it back to
## the caller.
##
## @synopsis
## $version=$filer_obj->get_version();
##
##
##
sub get_version(){
    my $self = shift;
    my $cmd = "version";
    my $output = "";

    if($self->{use_console} eq "true"){
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    return $output;

}

## @name ::get_licenses()
##
## @description
## A pass through command to get the current licenses from a filer and pass a
## hash ref back to the caller, where the hash contains the license info.
##
## @synopsis
## $license_hash_ref=$filer_obj->get_licenses();
##
##
##
sub get_licenses(){
    my $self = shift;
    my $cmd = "license";
    my $after = "";
    my @output_array = ();
    my @tmp_array = ();
    my %license_hash = ();

    my $license_version = $self->get_license_version();

    if($self->{use_console} eq "true"){
        (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    @output_array = split(/\n/, $after);

    if ($license_version eq "V1") {

        foreach my $line (@output_array){
            @tmp_array = ();
            # Eliminate dynamic amount of whitespace from start of line to actual
            # start of a licensce name.
            $line =~ s/^ *//;
    
            @tmp_array = split(/ /, $line);
    
            # Adding more stringent parsing of license output to ensure that garbage
            # doesn't end up in the license hash.  Will look to see if the first
            # piece of the split line matches a known license name, if so we'll add
            # it to the license hash, otherwise we'll just move on to the next line.
            foreach my $license (@KNOWN_LICENSES){
                if($tmp_array[0] =~ m/$license/){
                    if($line =~ m/not licensed/){
                        $license_hash{$tmp_array[0]} = "not licensed";
                    }
                    # adding this check because there are times where the type of
                    # license is blank and doesn't show up as "site" or some other
                    # value and can trip up the parsing here.
                    elsif(scalar(@tmp_array) == 2){
                        $license_hash{$tmp_array[0]} = $tmp_array[1];
                    }
                    else{
                        $license_hash{$tmp_array[0]} = $tmp_array[2];
                    }
                    last;
                }
            }
        }

    } elsif ($license_version eq "V2") {

        # For V2 licensing, the license will only appear in the output array
        # if it has actually been installed.

        # vgv3070f37a*> license
        # Serial Number: 1076968
        # Owner: vgv3070f37a
        # Package           Type    Description           Expiration
        # ----------------- ------- --------------------- --------------------
        # NFS               site    NFS License           -                   
        # CIFS              site    CIFS License          -                   
        # iSCSI             site    iSCSI License         -                   
        # FCP               site    FCP License           -                   
        # SnapMirror        site    SnapMirror License    -                   
        # SnapLock          site    SnapLock License      -                   
        # V_Storageattach   site    V-Series Storage License -                   
        # -                 site    Converged Licensing   -                   
        # vgv3070f37a*>
		
        foreach my $line (@output_array){
            if ($line =~ m/^serial number:|^owner:/i) { next; }
            elsif ($line =~ m/^no licenses available/i) { last; }
            elsif ($line =~ m/^package|^-/i ) { next; }
            else {
                foreach my $license (@KNOWN_LICENSES_V2){
                    if ($line =~ m/($license)/i) {
                        $license_hash{$1} = undef;
                    }
                }

            }
        }
    }

    return ($license_version, \%license_hash);

}

## @name ::get_sysconfig()
##
## @description
## A pass through command to get the output of sysconfig from a filer and pass
## an array ref back to the caller, where the array is the command's output.
##
## @synopsis
## $sysconfig_ref=$filer_obj->get_sysconfig($);
##
## @arg $cmd_flag Optional: optional flag that sysconfig takes on a filer.
##
sub get_sysconfig(;$){
    my $self = shift;
    my ($cmd_flag) = @_;
    my $after = "";


    my $cmd = "sysconfig ";
    if(defined($cmd_flag)){
        $cmd = $cmd . $cmd_flag;
    }

    if($self->{use_console} eq "true"){
        (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    my @output_array = split(/\n/, $after);

    return \@output_array;



}

## @name ::poll_filer()
##
## @description
## A mechanism to poll the filer till a given string is seen in the output.
## This is a passive mechanism so it will only poll properly when the filer is
## generating output, ie booting, printing messages, or command output.  If the
## filler is not generating this output this method, due to say's behavior will
## timeout and ultimately die.  Please call correctly.
##
## @synopsis
## $poll_time=$filer_obj->poll_filer($$;$$);
##
## @arg $max_time    Required: Maximum time you want to poll.
## @arg $wait_string Required: String you want to poll for in output.
## @arg $timeout     Optional: Timeout value for say().
## @arg $maxidle     Optional: Max idle value for say().
##
sub poll_filer($$;$$){
    my $self = shift;
    my ($max_time, $wait_string, $timeout, $maxidle) = @_;
    my ($match, $before, $after, $prompt) = "";
    my $any_line_prompt = '^.*\n';
    my $progress_prompt = '^\.+$';  # This prompt handles progress dots.
    my $matched_time = -1;
    my @poll_prompts = ($any_line_prompt,
                        $progress_prompt
    );

    # Initialize finish time to use as loop control.
    my $finish_time = timeout2time($max_time);

    # Record start time so we can report how long the polling took.
    my $start_time = time();

    if (! defined $timeout) { $timeout = 120; }  
    if (! defined $maxidle) { $maxidle = 120; }

    # Polling loop, listen and see if we saw the match_string.
    while(time() < $finish_time){
        if($self->{use_console} eq "true"){
            ($match, $before, $after, $prompt) = Tharn::say($self->{console_connection}, "", "-prompt", @poll_prompts, -timeout, $timeout, -maxidle, $maxidle);
        }
        else{
            ($match, $before, $after, $prompt) = Tharn::say($self->{connection}, "", "-prompt", @poll_prompts);
        }

        if($prompt =~ m/$wait_string/){
            $matched_time = time();
            last;
        }
    }

    # If matched_time is -1 then we didn't match on the string so report error.
    # Otherwise return the time it took to poll.
    if($matched_time == -1){
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Timeout while polling filer for: $wait_string";
        return $ERR_STATUS;
    }
    else{
        my $total_time = $matched_time - $start_time;
        return $total_time;
    }
}


## @name ::get_switch_portlist()
##
## @description
## A passthrough to the VSeries package method vseries_host_portlist_get.
## Returns an array of hash (key = "switchname" value = port_num) having
## info about the switch ports connected to the storage adapters of the filer
## we connect to.
##
## NOTE: only a telnet connection is supported, no support for console connection
##
## @synopsis
## $portlist_ref=$filer_obj->get_portlist();
##
##

sub get_switch_portlist(){

    my $self = shift;

    # execute command
    my @output_array = VSeries::VSeries::vseries_host_portlist_get($self->{connection});

    return \@output_array;
}

## @name ::get_storage_portlist()
##
## @description
## A passthrough to the VSeries package method vseries_storage_portlist_get.
## Returns an array of hash (key = "switchname" value = port_num) having
## info about the switch ports connected to the storage arrays
##
##
## NOTE: only a telnet connection is supported, no support for console connection
##
## @synopsis
## $portlist_ref=$filer_obj->get_storage_portlist();
##
##
sub get_storage_portlist(){

    my $self = shift;

    # execute command
    my @output_array = VSeries::VSeries::vseries_storage_portlist_get($self->{connection});

    return \@output_array;
}

## @name ::aggr_mirror($$;%)
##
## @description
## Run the "aggr mirror" command to rejoin broken mirrors.
##
## @synopsis
## $takeover_cmd_out=$filer_obj->aggr_mirror($$;%);
##
## @arg $good_mirror: (required) aggregate of the currently active (good) half of the split mirror
## @arg $failed_mirror: (required) aggregate of the failed half of the split mirror
## @arg %cmdopts (optional): a hash of command line options (key = option, value = any required parameter(s)
##    for the option)
##

sub aggr_mirror($$;%){
    my $self = shift;
    my ($good_mirror, $failed_mirror, %cmdopts) = @_;
    my $after="";
    my $prompt="";
    my $output_array_ref=();
    my @output_array=();
    my $status_hash_ref=0;
    my $current_status="";
    my $current_state="";
    my %status_hash=();
    my $status_hash_ref;
    my $cmdopt="";
    my $cmd="";
    my $wait_for_resync=0;
    my $loopcount=0;
    my $line="";
    my $MAX_PASSES_FOR_AGGR_RESYNC=20;
    my $WAIT_AGGR_STAT_RESYNCING=15;

    if ( $good_mirror eq "" || $failed_mirror eq "" ) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "missing one or more input parameters";
        return ($ERR_STATUS);
    }


    # parse optional command line options to form appropriate command syntax
    foreach  $cmdopt ( sort keys %cmdopts ) {
       if ( $cmdopt =~ / *-f */ ) {
          $cmd = "aggr mirror " . $good_mirror . " -f -v " . $failed_mirror;
       # TBD: handle other command line options here
       } else {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "ERROR: unrecognized command options: $cmdopt";
        return $ERR_STATUS;
       }
    }
    # execute command
    if ($self->{use_console} eq "true"){
         (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "> ?", "-timeout", $AGGR_MIRROR_REJOIN_TIME);
       }
       else {
         (undef, undef, $after, $prompt) = Tharn::say($self->{connection}, $cmd, -swallow, "-prompt", "> ?", "-timeout", $AGGR_MIRROR_REJOIN_TIME);
    }

    # check the aggr mirror command output for resync errors
    if ( $after =~ /Can\'t find common resync snapshot/ ) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "ERROR aggr_mirrror: Can't find common resync snapshot";
        return $ERR_STATUS;
    } # TBD parse for other error options as new option support is added in

    # Check that the mirrors have begun resyncing. We spin until the aggr status on the aggr shows 'resyncing'
    # We don't wait for the mirrors to finish syncing, that is the job of the calling program.
    while ( ! $wait_for_resync ) {
       $output_array_ref = $self->aggr_status($good_mirror);
       @output_array = @$output_array_ref;

       foreach $line (@output_array) {
          if ( $line =~ /^ *resyncing / || $line =~ /^ *mirrored / ) {
            $wait_for_resync=1;
            last;
          }
       }
       if ( ! $wait_for_resync ) {
             @output_array=();
             $loopcount++;
             if ( $loopcount > $MAX_PASSES_FOR_AGGR_RESYNC ) {
               $self->{error} = $ERR_STATUS;
               $self->{errmsg} = "ERROR aggr_mirrror: aggr mirror resync has not started or status is in unexpected state ";
               return $ERR_STATUS;
             }
             sleep $WAIT_AGGR_STAT_RESYNCING;
       }
    }

    return $GOOD_STATUS;
}


## @name ::cf_forcetakeover(;$)
##
## @description
## Issues a cf forcetakeover command on the filer
##
## NOTE: only console connection is supported
##
## @synopsis
## $forcetakeover_cmd_out=$filer_obj->cf_forcetakeover(;$);
##
## @arg $takeover_options (optional): command line options (ie. -d,-f)
##
## sample command output:
## vsa-v6030-1a> cf forcetakeover -d
## Following the command, mirrored volumes will be split and
## clients of the partner filer will be required to remount.
## After the giveback, remirroring the volumes will be necessary.
## Prior to issuing this command, the partner filer should be powered off.
## If the cluster partner is operational or if it becomes operational at any time
## while this filer is running in takeover mode, your filesystems may be destroyed.
## Do you wish to continue [y/n] ??

sub cf_forcetakeover(;$){
    my $self = shift;
    my ($takeover_options) = @_;
    my $after="";
    my $prompt="";

    my $cmd = "cf forcetakeover " . $takeover_options;

    if ($self->{use_console} eq "true"){
         # handle case of no command options specified
         if ( $takeover_options eq "" ) {
            (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "> ?", "-timeout", $FORCEDTAKEOVER_TIME);
         }
         elsif ( $takeover_options =~ / *-d */ ) {
             # issue the cf forcetakeover -d  command, wait for followup interactive confirmation query and respond with 'y'
             (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "Do you wish to continue", "-timeout", $FORCEDTAKEOVER_TIME);
             # answer yes to followup query that you want to force takeover
             (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, "y", -swallow, "-prompt", "> ?", "-timeout", $FORCEDTAKEOVER_TIME);
         }
    } else {
         # handle case of no command options specified
         if ( $takeover_options eq "" ) {
            (undef, undef, $after, $prompt) = Tharn::say($self->{connection}, $cmd, -swallow, "-prompt", "> ?", "-timeout", $FORCEDTAKEOVER_TIME);
         }
         elsif ( $takeover_options =~ / *-d */  ) {
             # issue the cf forcetakeover -d  command, wait for followup interactive confirmation query and respond with 'y'
             (undef, undef, $after, $prompt) = Tharn::say($self->{connection}, $cmd, -swallow, "-prompt", "Do you wish to continue", "-timeout", $FORCEDTAKEOVER_TIME);
             # answer yes to followup query that you want to force takeover
             (undef, undef, $after, $prompt) = Tharn::say($self->{connection}, "y", -swallow, "-prompt", "> ?", "-timeout", $FORCEDTAKEOVER_TIME);
         }
    }

    my @output_array = split(/\n/, $after);

    return \@output_array;
}

## @name ::cf_forcegiveback(;$)
##
## @description
## Issues a cf forcetakegiveback command on the filer
##
## @synopsis
## $forcetakeover_cmd_out=$filer_obj->cf_forcegiveback(;$);
##
## @arg $giveback_options (optional): command line options
##
## sample command output:
## vsa-v6030-1a(takeover)> cf forcegiveback
## cf forcegiveback may lead to data corruption; 'cf giveback -f' may be a
## better choice.  Really force a giveback?
## vsa-v6030-1a(takeover)> cf forcegiveback -f
## please make sure you have rejoined your aggr before giveback.
## Do you wish to continue [y/n] ??

sub cf_forcegiveback(;$){
    my $self = shift;
    my ($giveback_options) = @_;
    my $after="";
    my $prompt="";

    my $cmd = "cf forcegiveback -f" . $giveback_options;

    if ($self->{use_console} eq "true"){
         # issue the cf forcegiveback and wait for followup interactive query
         (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "Do you wish to continue", "-timeout", $FORCEDGIVEBACK_TIME);
         # determine if we have a rejoined mirror giveback - this case will issue a followup
         # confirmation query that we need to respond to with 'y'
         if ( $prompt =~ /Do you wish to continue/ ) {
            (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, "y", -swallow, "-prompt", "> ?", "-timeout", $FORCEDGIVEBACK_TIME);
         }
       }
       else {
         (undef, undef, $after, $prompt) = Tharn::say($self->{connection}, $cmd, -swallow, "-prompt", "Do you wish to continue", "-timeout", $FORCEDGIVEBACK_TIME);
         # determine if we have a rejoined mirror giveback - this case will issue a followup
         # confirmation query that we need to respond to with 'y'
         if ( $prompt =~ /Do you wish to continue/ ) {
            (undef, undef, $after, $prompt) = Tharn::say($self->{connection}, "y", -swallow, "-prompt", "> ?", "-timeout", $FORCEDGIVEBACK_TIME);
         }
    }

    my @output_array = split(/\n/, $after);

    return \@output_array;
}




## @name ::revision_checker()
##
## @description
## A mechanism to check the revision of ontap on the filer.  This is used
## to check in different places where certain ontap functionality may
## differ from one version to the next.  Different commands availaiblity
## for prompt strings etc...
##
## @synopsis
## $version_check=$filer_obj->revision_checker($);
##
## @arg $release_name Required: Two letter abbreviations given to releases of
##                              ontap.  IC,GB,TT,AS,etc...
##
## @arg $test_string Optional: Optional param that can be passed in to override
##                             the result of the call to get_version() to check
##                             what release a given filer is running.  This is
##                             a way to handle sandbox builds that may be a
##                             given relase of ontap but not follow the
##                             standard format.
##
sub revision_checker($;$){
    my $self = shift;
    my ($release_name, $test_string) = @_;
    my %release_hash = ();

    # Add to the release hash the regular expressions that will match on all
    # well known releases of ontap.  This will make looping to match a given
    # release easier.
    my @ic_regexes = ("7[\.]3", "ironcity");
    $release_hash{IC} = [@ic_regexes];

    my @gb_regexes = ("7[\.]2", "gordonbiersch");
    $release_hash{GB} = [@gb_regexes];

    my @tt_regexes = ("7[\.]1", "tsingtao");
    $release_hash{TT} = [@tt_regexes];

    my @as_regexes = ("7[\.]0", "anchorsteam");
    $release_hash{AS} = [@as_regexes];

    # Override the version of the filer if one was passed in.
    my $version = $self->get_version();
    if(defined($test_string)){
        $version = $test_string;
    }

    # Loop through the release hash matching on any well known release.  The
    # assumption here is that one version of ontap is not a substring of
    # of another.  For example we do NOT have "Bud" and "Bud Light"
    foreach my $key (keys(%release_hash)){
        if($release_name eq $key){
            foreach my $regex (@{$release_hash{$release_name}}){
                if($version =~ m/$regex/){
                    return $GOOD_STATUS;
                }
            }
        }

    }

    # If we did not match on a well known version of ontap return an error
    return $ERR_STATUS;

}


## @name ::disk_zero_spares()
##
## @description
## A mechanism to proactively kick off the zeroing of all spares on a filer as
## a background task.  This is a normal command in ontap and not a special
## creation of automation.  There is a small window after the command is issued
## but before 1% has been zeroed where the "not zeroed" status will go away
## from the line with the disk info on it but the status of the zeroing hasn't
## been updated and added to the line.  The advice to the caller is to allow
## this delta to pass before calling get_zeroed_disks by simply delaying
## the subsequent call to get_zeroed_disk in the thpl layer.
##
## @synopsis
## $cmd_result = $filer_obj->disk_zero_spares();
##
##
sub disk_zero_spares(){
    my $self = shift;
    my $after = "";

    my $cmd = "disk zero spares";

    if($self->{use_console} eq "true"){
        (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    return $after;

}

## @name ::get_zeroed_disks()
##
## @description
## A mechanism to check the spare disks owned by a system by running the
## aggr status -s command and parsing the result.  In the resulting hash
## there is an attribute called "zeroed" and it's a flag to determine if
## a given spare is zeroed or not.  All spares that are zeroed are returned
## to the caller in an array containing the primary disk names.
##
## @synopsis
## @zeroed_disks = @{$filer_obj->get_zeroed_disks()};
##
##
sub get_zeroed_disks(){
    my $self = shift;

    # Get the result of aggr status -s which shows all spares on the system
    my $output_ref = $self->aggr_status("-s");
    my @output = @$output_ref;

    # Call the parser to turn the output into a workable hash form, see
    # parser for the attributes of the hash, right here we care about
    # the names and the zeroed flag.
    my $hash_ref = VSeries::utils::parse_aggr_status("-s", @output);
    my %disks_hash = %$hash_ref;
    my @zeroed_disks = ();

    # Loop through each disk in the returned hash checking for zeroed spares.
    foreach my $disk_name (keys(%disks_hash)){
        if($disks_hash{$disk_name}{zeroed} eq $TRUE){
            push(@zeroed_disks, $disk_name);
        }

    }

    return \@zeroed_disks;


}


## @name ::discover_ip_interfaces()
##
## @description
## Issues "ifconfig -a" on the filer via the console connection and then parses
## the output and creates the IP hash used for IP connectivity.  Since say will
## timeout if the filer is unreachable and the command doesn't really know what
## good values for the IP interfaces really is, this method won't return an
## err status and it's up to the caller to get the IP information and make
## determination on whether the information returned indicates good things or
## bad things.
##
## @synopsis
## $filer_obj->discover_ip_interfaces()
##
##
sub discover_ip_interfaces(){
    my $self = shift;
    my $cmd = "ifconfig -a";
    my ($match, $before, $after, $prompt) = "";
    my %ip_hash = ();
    my $current_interface = "";

    ($match, $before, $after, $prompt) = Tharn::say($self->{console_connection}, $cmd, -swallow);
    # Remove unnecessary carriage return and new lines.  In the event of new
    # lines we actually don't want to munge together what's on the next line
    # so we'll actually replace those with simple whitespace for splitting.
    $after =~ s/\r/ /g;
    $after =~ s/\n/ /g;
    my @tokens = split(/ +/, $after);
    for(my $i = 0; $i <=  scalar(@tokens); $i++){
        my $token = $tokens[$i];

        # This switch statement will handle the tokens of information that
        # exist in the output of the ifconfig -a command.  Some values are
        # pairs based on a key word and then a value after it some things
        # are just stuffed in pieces of info.  Here's an example
        #
        #  e0a: flags=948043<UP,BROADCAST,RUNNING,MULTICAST,TCPCKSUM> mtu 1500
        #  inet 10.97.103.85 netmask 0xffffff00 broadcast 10.97.103.255
        #  partner inet 10.97.103.76 (not in use)
        #  ether 00:a0:98:04:c2:8a (auto-100tx-up) flowcontrol none
        #  e0b: flags=108042<BROADCAST,RUNNING,MULTICAST,TCPCKSUM> mtu 1500
        #  ether 00:a0:98:04:c2:8b (auto-unknown-cfg_down) flowcontrol full

        switch($token){

            # This matches the e0a: token or other interface names all
            # subsequent information will be associated with this interface
            # until another interface name is seen.
            case /^\w+:$/{
                $current_interface = $token;
            }

            # This matches the flags piece, not fully sure when or where this
            # will be usefull but we'll stuff it in here just in case it is
            case /flags/{
                my @tmp_array = split(/=/, $token);
                $ip_hash{$current_interface}{flags} = $tmp_array[1];
            }

            # This precedes the token that is the actual ip address of the
            # interface.
            case /inet/{
                $ip_hash{$current_interface}{ip_addr} = $tokens[++$i];
            }

            # This precedes the token that is the netmask of the interface
            case /netmask/{
                my $converted_mask = VSeries::utils::netmask_hex_to_decimal($tokens[++$i]);
                $ip_hash{$current_interface}{netmask} = $converted_mask;
            }

            # This precedes the token that is the mac address of the interface
            case /ether/{
                $ip_hash{$current_interface}{mac_addr} = $tokens[++$i];
            }

            # This precedes the token that is the setting for flow control on
            # the interface
            case /flowcontrol/{
                $ip_hash{$current_interface}{flow_control} = $tokens[++$i];
            }

            # This precedes the token that is the ip addr for the broadcast
            # on the given subnet.
            case /^broadcast$/{
                $ip_hash{$current_interface}{broadcast} = $tokens[++$i];
            }

            # This is the format for the token that has the format of
            # (auto-100tx-up) what we really care about is the first
            # piece of info before the hyphen so in this case 'auto'
            # so we need to carve this up a bit to get the value we want.
            case /\(.*\)/{

                # first split on the hyphen then remove the leading open
                # parenthesis
                my @tmp_array = split(/\-/, $token);
                $tmp_array[0] =~ s/\(//g;
                $ip_hash{$current_interface}{media_type} = $tmp_array[0];
                $ip_hash{$current_interface}{transmission_speed} = $tmp_array[1];
                $ip_hash{$current_interface}{phys_link} = $tmp_array[2];
            }

            # This token precedes the MTU setting for the interface.  This has
            # to do with jumbo frames on the interface.
            case /mtu/{
                $ip_hash{$current_interface}{mtu} = $tokens[++$i];
            }

            # This token is actually followed by inet but when paired like this
            # is actually the ip addr of the partner interface that this
            # interface is configured to take over on failover.  Otherwise the
            # word partner is followed by the interface name ie: e0a.
            case /partner/{
                if($tokens[$i + 1] =~ m/inet/){
                    $i++;
                    $ip_hash{$current_interface}{takeover_interface} = $tokens[++$i];
                }
                else{
                    $ip_hash{$current_interface}{takeover_interface} = $tokens[++$i];
                }

            }
        } #end switch
    } #end for

    # Set internal IP hash to be what we've discovered off the filer.
    %{$self->{ip_interfaces}} = %ip_hash;

    return $GOOD_STATUS;

}


## @name ::exportfs(;$$)
##
## @description
## Executes the exportfs command to retrieve or set attributes. Returns a hash
## reference with key = volume path and name,  value = export attributes
## Very limited command line parsing is performed. This method can be used
## with no agruments to get current export attributes of all existing
## volumes or with options and volume name specified to get or set
## information on a specific volume depending on attributes
##
## @synopsis
## $exportfs_output_hash_ref=$filer_obj->exportfs(;$$);
##
## NOTE: This method currently does not support any other command options
## save for the standalone command with no options or -o to set attributes

sub exportfs(;$$){

   my $self = shift;
   my ($exportfs_options, $volume) = @_;
   my %exportfs_hash=();
   my $cmd = "exportfs $exportfs_options $volume";
   my $line = "";
   my $after = "";
   my $i = 0;
   my $attrs = "";

   my @non_supported_options = ("-a", "-b", "-c", "-f", "-q", "-r", "-s", "-u", "-v", "-w", "-z", "-h");

   # Ensure that $exportfs_options contains only supported fields
   foreach my $nonsupported (@non_supported_options) {
      if ($exportfs_options =~ / *$nonsupported */ ) {
         $self->{error} = $ERR_STATUS;
         $self->{errmsg} = "ERROR: found nonsupported exportfs syntax '$nonsupported'";
         return $ERR_STATUS;
      }
   }

   if ( (defined($exportfs_options) && !defined($volume)) || (defined($volume) && !defined($exportfs_options)) ) {
      $self->{error} = $ERR_STATUS;
      $self->{errmsg} = "ERROR: insufficient method parameters";
      return $ERR_STATUS;
   }

   if($self->{use_console} eq "true"){
      (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
   } else {
      (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
   }

   # Check for other syntax errors or bogus options
   if ( $after =~ /Extra arguments supplied after/ || $after =~ /usage:/ || $after =~ /invalid option/ ) {
      $self->{error} = $ERR_STATUS;
      $self->{errmsg} = "ERROR: bad syntax or parameters. String parsed: $after";
      return $ERR_STATUS;
   }

   # Check return value of exportfs command with specified options and volume.
   # If Tharn::say returned no output, then the command was successful.
   if ( $after eq "" && defined($exportfs_options) && defined($volume) ) {
      return $GOOD_STATUS;
   }

   # If we get here then we issued the exportfs command with no arguments
   # to list all the exported volumes and attributes, which we now parse

   # Put say output into array, remove carriage returns
   $after =~ s/\r//g;
   my @output_array = split(/\n/, $after);

   # Further split each line of the @output_array into a new two item array
   # consisting of the exported filesystem and its attributes. First element
   # is the exported volume which will be the key of %exportfs_hash. The second
   # element is the string of parameters which will be the hash value.
   # Sample output:
   #/vol/vol0/home  -sec=sys,rw,nosuid
   #/vol/vol0       -sec=sys,rw,anon=0,nosuid
   #/vol/vol1       -sec=sys,rw,anon=0
   #/vol/vol2       -sec=sys,rw,anon=0

   foreach $line (@output_array) {
      # Parse only the lines with a leading volume name
      if ($line =~ /^\/vol\// ) {
         $volume = (split (/\s+/, $line))[0];
         $attrs =  (split (/\s+/, $line))[1];
         $exportfs_hash{$volume} = $attrs;
      }
   }

   return \%exportfs_hash;
}

## @name get_disk_type_counts($)
##
## @description
## A simple method that takes in the ID of a system to check on the
## filer the counts of the various disk types that filer thinks the
## owning system ID has.  This can be the actual filer's system ID, its
## partners, or Not Owned to see what's unowned.  It returns a hash with
## the key value pairs being the disks storage type and the number of disks
## of that type seen.
##
## @synopsis
## $return_hash_ref = get_disk_type_counts($system_id);
##
##
sub get_disk_type_counts($){
    my $self = shift;
    my $system_id = shift;

    # Take a look at what disks the first filer in filer object array sees
    # as owned by the system ID passed in.
    my %owned_disks = %{$self->get_disks_by_owner($system_id)};

    # Initialize the hash that we'll use to keep track of the disk counts
    my %type_counts = (
                      FCAL       => '0',
                      LUN        => '0',
                      ATA        => '0',
                      EATA       => '0',
                      SAS        => '0',
                      SATA       => '0',
                      SCSI       => '0',
                      XATA       => '0',
                      XSAS       => '0',
                      BSAS       => '0',
                      SSD        => '0',
                      MSATA      => '0',
                      FSAS       => '0',
    );

    # Now that each type of disk is intitialized and the count set to zero
    # loop over the output of get_disks_by_owner and when we match a disk's
    # storage type within the type counts hash then increment the counter
    # for that type.  Return the hash when we're done.
    foreach my $disk_serial (keys(%owned_disks)){
        my %tmp_hash = %{$self->get_disk_info($disk_serial)};
        foreach my $disk_type (keys(%type_counts)){
            if($disk_type eq $tmp_hash{storage_type}){
                $type_counts{$disk_type} = $type_counts{$disk_type} + 1;
            }
        }


    }

    return \%type_counts;
}


## @name ::asis_sis_on($)
##
## @description
## Executes the asis command 'sis on' to enable a volume for de-duplication.
## Errors will be reported in the filer object's error flag and message.
##
## @arg $volume (required): volume name to be enabled for de-duplicated
##
## @synopsis
## $status = $filer_obj->asis_sis_on($);

sub asis_sis_on ($) {

    my $self = shift;
    my ($volume) = @_;
    my $after = "";

    # Check if volume name leads with /vol path. If not then append it
    if ( $volume !~ /^\/vol\// ) {
        $volume = "/vol/" . $volume;
    }

    my $cmd = "sis on $volume";

    if($self->{use_console} eq "true"){
        (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow);
    } else {
        (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow);
    }

    # Check for two possible expected conditions: either we're enabling sis for
    # the first time or else it's already enabled
    if ($after =~ /SIS for \"$volume\" is enabled/ || $after =~ /Operation has already been enabled: *$volume/) {
        return $GOOD_STATUS;
    } else {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "ERROR: sis on command error: $after";
        return $ERR_STATUS;
    }
}

## @name ::asis_sis_off($)
##
## @description
## Executes the asis command 'sis off' to disable volume from de-duplication.
## Errors will be reported in the filer object's error flag and message.
##
## @arg $volume (required): volume name to be disabled from de-duplication
##
## @synopsis
## $status = $filer_obj->asis_sis_off($);

sub asis_sis_off ($) {

    my $self = shift;
    my ($volume) = @_;
    my $after = "";

    # Check if volume name leads with /vol path. If not then append it
    if ( $volume !~ /^\/vol\// ) {
        $volume = "/vol/" . $volume;
    }

    my $cmd = "sis off $volume";

    if($self->{use_console} eq "true"){
        (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow);
    } else {
        (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow);
    }

    # Check for two possible expected conditions: either we're disabling sis for
    # the first time or else it's already disabled
    if ($after =~ /SIS for \"$volume\" is disabled/ || $after =~ /Operation has already been disabled: *$volume/) {
        return $GOOD_STATUS;
    } else {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "ERROR: sis off command error: $after";
        return $ERR_STATUS;
    }
}

## @name ::asis_sis_start($)
##
## @description
## Executes the asis command 'sis_start -s' to begin volume de-duplication.
## Errors will be reported in the filer object's error flag and message.
##
## @arg $volume (required): volume name to be de-duplicated
##
## @synopsis
## $status = $filer_obj->asis_sis_start($);

sub asis_sis_start ($) {

    my $self = shift;
    my ($volume) = @_;
    my $after = "";
    my $prompt = "";

    # check if volume name leads with /vol path. If not then append it
    if ( $volume !~ /^\/vol\// ) {
        $volume = "/vol/" . $volume;
    }

    my $cmd = "sis start -s $volume";

    # Issue the command and the "y" response to the followup query
    if($self->{use_console} eq "true"){
        (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "Are you sure you want to proceed");

        # Check if any errors occured (bad syntax)
        if ( $after =~ /Invalid path specification/ || $after =~ /Volume does not exist/
            || $after =~ /Path must be supplied as/ ) {
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "ERROR: sis start command error: $after";
            return $ERR_STATUS;
        }

        if ( $prompt =~ /Are you sure you want to proceed/ ) {
            (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, "y", -swallow);
        } else {
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "ERROR: unexpected sis start command prompt: $prompt";
            return $ERR_STATUS;
        }
    } else {
        (undef, undef, $after, $prompt) = Tharn::say($self->{connection}, $cmd, -swallow, "-prompt", "Are you sure you want to proceed");

        # Check if any errors occured (bad syntax)
        if ( $after =~ /Invalid path specification/ || $after =~ /Volume does not exist/
            || $after =~ /Path must be supplied as/ ) {
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "ERROR: sis start command error: $after";
            return $ERR_STATUS;
        }

        if ( $prompt =~ /Are you sure you want to proceed/ ) {
            (undef, undef, $after, $prompt) = Tharn::say($self->{connection}, "y", -swallow);
        } else {
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "ERROR: unexpected sis start command prompt: $prompt";
            return $ERR_STATUS;
        }
    }

    return $GOOD_STATUS;

}


## @name ::asis_sis_status(;$)
##
## @description
## Executes the asis command 'sis status -l' to determine current status
## and statistics of an executing de-duplictaion process. We parse
## "sis status -l" vs "sis status" because it displays the complete
## set of parameters.
##
## @arg $volume (optional): specific volume to get status of
##
## @synopsis
## $sis_status_array_ref = $filer_obj->asis_sis_status($);
##
## Output will be an array of hashes, one hash for each volume (Path). The
## hash keys are the keyword(s) to the left of the first semicolon on each
## line. The hash value will be the remainder of the line to the right of
## the first semicolon (see example below).  Errors will be reported in
## the filer object's error flag and message.

# Sample output:
#vra-v3050-1a> sis status -l
#
#Path:                    /vol/asis_test_vol
#State:                   Enabled
#Status:                  Idle
#Progress:                Idle for 04:38:32
#Type:                    Regular
#Schedule:                sun-sat@0
#Last Operation Begin:    Wed Oct 24 10:46:45 EST 2007
#Last Operation End:      Wed Oct 24 11:11:41 EST 2007
#Last Operation Size:     10 GB
#Last Operation Error:    -
#
#Path:                    /vol/mbozen_asis
#State:                   Enabled
#Status:                  Idle
#Progress:                Idle for 00:00:12
#Type:                    Regular
#Schedule:                sun-sat@0
#Last Operation Begin:    Wed Oct 24 15:50:01 EST 2007
#Last Operation End:      Wed Oct 24 15:50:01 EST 2007
#Last Operation Size:     0 KB
#Last Operation Error:    -

# Depending on ASIS settings and when status command is executed,
# more fields may appear in command output.
#Path:                    /vol/asis_test_vol_vnv3050f9a
#State:                   Enabled
#Status:                  Idle
#Progress:                Idle for 00:00:31
#Type:                    Regular
#Schedule:                sun-sat@0
#Last Operation Begin:    Mon Apr 14 17:02:19 GMT 2008
#Last Operation End:      Mon Apr 14 17:09:03 GMT 2008
#Last Operation Size:     2559 MB
#Last Operation Error:    -
#Changelog Usage:         0%
#Stage:                   Done
#Fingerprints Gathered:   655350
#Gathering Begin:         Mon Apr 14 17:02:19 GMT 2008
#Fingerprints Sorted:     655350
#Duplicate Blocks Found:  655348
#Sorting Begin:           Mon Apr 14 17:05:04 GMT 2008
#Blocks De-duplicated:    653203
#De-duping Begin:         Mon Apr 14 17:05:09 GMT 2008
#Fingerprints Deleted:    0
#Checking Begin:          -

sub asis_sis_status (;$) {

    my $self = shift;
    my ($volume) = @_;
    my $after = "";
    my @sis_status = ();
    my @sis_status_output = ();
    my $line = "";
    my %temp_hash = ();
    my $cmd = "";

    if (! defined($volume)) {
        $cmd = "sis status -l";
    } else {
        # check if volume name leads with /vol path. If not then append it
        if ( $volume !~ /^\/vol\// ) {
            $volume = "/vol/" . $volume;
        }
        $cmd = "sis status -l $volume";
    }

    if ($self->{use_console} eq "true"){
        (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow);
    } else {
        (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow);
    }

    # Check for an unexpected errors: either sis is not on or else
    # there's a syntax error
    if ($after =~ /Volume does not exist/ || $after =~ /No status entry found/
        || $after =~ /Invalid path specification/ ) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "ERROR: sis status -l command error: $after";
        return $ERR_STATUS;
    }

    # Get rid of any carriage return chars in command output
    $after =~ s/\r//g;

    # Put commnand output into array
    @sis_status = split (/\n/, $after);

    my %sis_output_hash = ();

    # Parse output and put fields into a hash
    for (my $i = 0; $i < scalar @sis_status; $i++) {

        # Skip over blank lines
        if ( $sis_status[$i] eq "" ) { next; }

        # Parse lines of output with a semi-colon
        if ( $sis_status[$i] =~ /^(.*):\s+(.*)$/ ) {

            # Each occurence of "Path: " in the output will clear
            # the hash to (re)populate it
            if ( $1 eq "Path" ) {
               %sis_output_hash = ();
            }

            # Hash key: string to the left of the semicolon
            # Hash value: string to the right of the semicolon
            $sis_output_hash{$1} = $2;
        }

        # Check if we come to an end of a set of data for a volume
        # This is indicated either by the occurence of no more lines
        # or the next occurence of "Path: ". When we come to either
        # either of these conditions, push the sis_output_hash into
        # an array
        if ( ! defined $sis_status[$i+1] || $sis_status[$i+1] =~ /^Path: / ) {
            # Create new container for the hash to push into array
            my %temp_hash = %sis_output_hash;
            push @sis_status_output, \%temp_hash;
        }
    }

    return (\@sis_status_output);
}


## @name ::get_df_params($;$)
##
## @description
## Run the 'df' command on the filer to return aggregate and volume stats.
## Returns a reference to an arrays of hashes. Each line of data will make
## up a separate hash. Hash keys are white space separated column headers
## read from the first line of output. Values are whitespace separated
## columnar data fields in subsequent lines. The parsing method executes on
## premise that there is an equal number of header keywords and data columns
## in the command output. Exceptions to this are handled and treated as
## special cases (see NOTES below for details).
## This method uses standard non-console connections to filers.
##
## This method supports all df command options.
##
## NOTES:
## 1) $df_params command options must be bundled together
##   (ie.  df -AkS)
## 2) We do not parse up front for a valid aggr or vol name. If it's
##    bogus, the command will fail and we will set the filer object's
##    error flag.
## 3) None of the command output parsed and put into hash keys or values
##    is modified with two exceptions, which modify data to ensure there is
##    an equal number of header keywords and data columns.
##     a) The 'Mounted on' header for volumes is shortened to 'Mounted'
##     for the hash key
##     b) If the Filesystem name is "snap reserve", this string is combined
##     into one word with whitespace removed for the hash value
## 4) Illegal syntax and/or violations of mutually exclusive options
##    will be caught and will set the filer object's error flag.
## 5) This method will not verify the version of ontap to determine
##    supported or no-supported options. If a non-supported option is
##    specified, this will be caught and set the filer object's error
##    flag.
##
## @synopsis
## $df_array_ref = $filer_obj->get_df_params($;$);
##
## @arg $device     Required:   Aggregate or volume name
## @arg $df_params  Optional:   df command options
##

# Various command options and output:

#vnv3040f5a> df vol0
#Filesystem              kbytes       used      avail capacity  Mounted on
#/vol/vol0/            26144320     788632   25355688       3%  /vol/vol0/
#/vol/vol0/.snapshot    6536080      22460    6513620       0%  /vol/vol0/.snapshot

#vnv3040f5a> df /vol/vol0
#Filesystem              kbytes       used      avail capacity  Mounted on
#/vol/vol0/            26144320     788632   25355688       3%  /vol/vol0/
#/vol/vol0/.snapshot    6536080      22460    6513620       0%  /vol/vol0/.snapshot

#vnv3040f5a> df /vol/vol0/.snapshot
#Filesystem              kbytes       used      avail capacity  Mounted on
#/vol/vol0/            26144320     788632   25355688       3%  /vol/vol0/
#/vol/vol0/.snapshot    6536080      22460    6513620       0%  /vol/vol0/.snapshot

#vnv3040f5a> df -k vol0
#Filesystem               total       used      avail capacity  Mounted on
#/vol/vol0/          26144320KB   788632KB 25355688KB       3%  /vol/vol0/
#/vol/vol0/.snapshot  6536080KB    22460KB  6513620KB       0%  /vol/vol0/.snapshot

#vnv3040f5b> df /vol/vol2/.snapshot
#Filesystem              kbytes       used      avail capacity  Mounted on
#/vol/vol2/            31457280    5262976   26194304      17%  /vol/vol2/
#snap reserve                 0          0          0     ---%  /vol/vol2/..

#vra-v3050-1a> df -A aggr0
#Aggregate               kbytes       used      avail capacity
#aggr0                 23280544   22221004    1059540      95%
#aggr0/.snapshot        1225288     394364     830924      32%

#vra-v3050-1a> df -S vol0
#Filesystem                used       compressed         a-sis     %saved
#/vol/vol0/              332836                0             0         0%

#vra-v3050-1a> df -AS aggr0
#Aggregate                 used       compressed         a-sis     %saved
#aggr0                 22220972                0             0         0%

#vra-v3050-1a> df -Ag aggr0
#Aggregate                total       used      avail capacity
#aggr0                     22GB       21GB        1GB      95%
#aggr0/.snapshot            1GB        0GB        0GB      32%

#vnv3040f5a> df -s vol0
#Filesystem                used      saved       %saved
#/vol/vol0/              788632          0           0%

#vnv3040f5a> df -s /vol/vol0/.snapshot
#Filesystem                used      saved       %saved
#/vol/vol0/              788632          0           0%

#vnv3040f5a> df -sm vol0
#Filesystem                used      saved       %saved
#/vol/vol0/               770MB        0MB           0%

sub get_df_params ($;$) {

    my $self = shift;
    my ($device, $df_params ) = @_;
    my @return_arr = ();
    my $after = "";   # say() output

    # Verify the df_params command options syntax contains a leading dash
    if (defined $df_params && $df_params !~ /^-/ ) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "ERROR: malformed command option syntax: $df_params";
        return \@return_arr;
    }

    my $cmd = "df $df_params $device";
    (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow);

    $after =~ s/\r//g;
    my @df_out = split (/\n/, $after);

    # Check for command execution errors
    foreach my $line (@df_out) {
        if ($line =~ /^df:.*No such file or directory/ ||
            $line =~ /^df:.*Invalid option/ ||
            $line =~ /^df:.*unrecognized argument/ ||
            $line =~ /^df:.*options are mutually exclusive/ ||
            $line =~ /^Invalid name for/) {

            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "ERROR: command execution error: @df_out";
            return \@return_arr;
        }
    }

    my @keys = ();

    # Parse command output
    for ( my $i = 0; $i < scalar (@df_out); $i++ ) {

        # If this is the header line then extract the header
        # strings into an array to be used for hash keys
        if ( $i == 0 ) {
            @keys = split (/ +/, $df_out[0]);
            # We check for an exception case where a header line
            # ends with "Mounted on". If so then pop the last
            # word "on" off the array so that the hash key for
            # this column will be shortened to "Mounted".
           if ( ( $keys[-1] =~ m/on/ ) && ( $keys[-2] =~ m/Mounted/ ) ) {
                pop @keys;
            }
            next;
        }

        # Parse followup non-header lines of output and extract
        # the whitespace separated fields to be used as hash values
        my @tokens = split (/ +/, $df_out[$i]);

        # We check for another exception case where the Filesystem value
        # is 'snap reserve'. If so then combine the two words into
        # one token and resize the token array so that the number of tokens
        # matches the number of header entries
        if ( ( $tokens[0] =~ m/snap/ ) && ( $tokens[1] =~ m/reserve/ ) ) {
            $tokens[1] = $tokens[0] . " " . $tokens[1];
            @tokens = @tokens[1..$#tokens];
        }

        # Ensure that after parsing lines of data and handling exception
        # cases we have an equal number of header keywords (keys) and data
        # (values)
        if ( scalar (@keys) != scalar (@tokens) ) {
            my $num_keys = scalar (@keys);
            my $num_tokens = scalar (@tokens);
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "ERROR: in df output, mismatch between numbers of columns ($num_keys) and data fields ($num_tokens)\n@df_out";
            return \@return_arr;
        }

        # At this point we have two ordered arrays of header keywords and
        # a corresponding line of data values. Create a new hash for the line
        # of data from the two arrays.
        my %tmp_hash = ();
        for ( my $j = 0; $j < scalar (@keys); $j++ ) {
            $tmp_hash{"$keys[$j]"} = $tokens[$j];
        }

        # Add populated hash to the array
        push @return_arr, \%tmp_hash;
    }

    return \@return_arr;
}



## @name ::get_cf_monitor()
##
## @description
## Runs and parses cf monitor command
##
## @synopsis
## $cf_mon_hash_ref = $filer_obj->get_cf_monitor();
##
## Assumptions:
##  1. Method does not care if say was successful or what it returns. If
##  filer was down or unaccessible, method will return a hash ref
##  of initialized null values.
##  2. Method returns exact strings and fields as they appear in command
##  output. Any reformatting of dates, times must be done by the calling
##  program.
#
# Sample output, various states:
#  current time: 31Jan2008 15:33:42
#  UP 13+23:11:56, partner 'vnv3050f9a', cluster monitor enabled
#  VIA Interconnect is up (link 0 up, link 1 up), takeover capability on-line
#  partner update TAKEOVER_ENABLED (31Jan2008 15:33:42)
#
#  current time: 01Feb2008 15:53:37
#  UP 52+12:49:14, partner 'vnv6030f22b', cluster monitor disabled
#
#  current time: 14Jan2008 21:09:10
#  DEAD 00:01:49, partner 'vnv3050f9a', cluster monitor enabled
#
#  current time: 14Jan2008 21:04:19
#  TAKEOVER 00:01:12, partner 'vnv3050f9b', cluster monitor enabled
#
#  current time: 12Feb2008 13:48:41
#  UP 04:47:21, partner 'vnv3070f16b', cluster monitor enabled
#  VIA Interconnect is down (link 0 down, link 1 up), takeover capability off-line (unsynchronized log)
#  takeover by partner off-line (unsynchronized log)
#  partner update TAKEOVER_DISABLED (12Feb2008 13:48:41)

sub get_cf_monitor () {

    my $self = shift;
    my @cf_mon_out= ();
    my %cf_mon_hash = ();
    my $line = "";
    my $after = "";

    my $cmd = "cf monitor";

    if ($self->{use_console} eq "true"){
        (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow);
    } else {
        (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow);
    }

    # Get rid of any carriage return chars in command output
    $after =~ s/\r//g;

    # Put command output into array
    @cf_mon_out = split (/\n/, $after);

    # Initialize all hash keys because number of lines of command output
    # can vary widely and we must return a hash with all keys defined.
    $cf_mon_hash{date} = "";
    $cf_mon_hash{time} = "";
    $cf_mon_hash{state} = "";
    $cf_mon_hash{state_time} = "";
    $cf_mon_hash{partner} = "";
    $cf_mon_hash{clu_mon_state} = "";
    $cf_mon_hash{via_interconn_state} = "";
    $cf_mon_hash{link0_state} = "";
    $cf_mon_hash{link1_state} = "";
    $cf_mon_hash{takeover_cap} = "";
    $cf_mon_hash{partner_update_state} = "";
    $cf_mon_hash{partner_update_date} = "";
    $cf_mon_hash{partner_update_time} = "";

    # Parse output and put fields into a hash
    foreach $line (@cf_mon_out) {

        if ( $line =~ /^\s+current time:\s+(\w+)\s+(\d+:\d+:\d+)$/ ) {
            $cf_mon_hash{date} = $1;
            $cf_mon_hash{time} = $2;
            next;
        }

        if ( $line =~ /^\s+(\w+)\s+(\d+?\+?\d+:\d+:\d+),\s+partner\s+\'(\w+)\',\s+cluster monitor\s+(\w+)$/ ) {
            $cf_mon_hash{state} = $1;
            $cf_mon_hash{state_time} = $2;
            $cf_mon_hash{partner} = $3;
            $cf_mon_hash{clu_mon_state} = $4;
            next;
        }

        if ( $line =~ /^\s+VIA Interconnect is (\w+)\s+\(link 0 (\w+),\s+link 1 (\w+)\),\s+takeover capability\s+(\w+\-\w+)/ ) {
            $cf_mon_hash{via_interconn_state} = $1;
            $cf_mon_hash{link0_state} = $2;
            $cf_mon_hash{link1_state} = $3;
            $cf_mon_hash{takeover_cap} = $4;
            next;
        }

        if ( $line =~ /^\s+partner update\s+(\w+)\s+\((\w+)\s+(\d+:\d+:\d+)\)$/ ) {
            $cf_mon_hash{partner_update_state} = $1;
            $cf_mon_hash{partner_update_date} = $2;
            $cf_mon_hash{partner_update_time} = $3;
        }
    }

    return \%cf_mon_hash;
}


## @name ::offline_vols()
##
## @description
## Walks a filer object's internal volume list offlining them
## Will return an error after 3 tries to offline a given volume.  Waits 10
## second between retries.
##
## @synopsis
## $filer_obj=VSeries::filer->offline_vols();
##
##
sub offline_vols()
{
    my $self = shift;
    my $cmd = "";
    my $status = "";
    my $output_array_ref = undef;
    my @output_array = ();
    my $status_hash_ref = undef;
    my %status_hash = ();
    my $counter = 0;
    my $sleep_time = 10;
    foreach my $vol (@{$self->{vol_list}})
    {
        do{
            $cmd = "vol offline " . $vol;
            if($self->{use_console} eq "true"){
                $status = Tharn::say($self->{console_connection}, $cmd, "-wait", $vol, "-ok");
            }
            else{
                $status = Tharn::say($self->{connection}, $cmd, "-wait", $vol, "-ok");
            }
            # Wait a few seconds before checking vol status.
            sleep 10;
            $output_array_ref = $self->vol_status();
            @output_array = @$output_array_ref;

            # Check if volume is in a busy state
            # If it is busy (offlining) then wait a little longer
            if (grep(/\'$vol\' is temporarily busy/, @output_array)) {
                sleep 30;
                $status_hash{$vol}{state} = "busy";
            }
            else {
                $status_hash_ref = VSeries::utils::parse_vol_status("", @output_array);
                %status_hash = %$status_hash_ref;
                sleep 10;
            }
            $counter++;

        }while($status_hash{$vol}{state} ne "offline" && $counter <= 3);

        if($counter >= 3){
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} .= "ERROR: Unable to offline volume $vol after multiple tries";
            return ($ERR_STATUS);
        }
    }

    return $GOOD_STATUS;
}

## @name ::set_snap_sched($$$)
##
## @description
## Method to set the snapshot schedule on an aggr or volume.  Verifies the
## correct value after setting it.  The parameters are the type flag
## ("-A" for aggr, "-V" for vol), the name of the volume or aggregate
## and the value string to set the schedule to. There can be 1 to 3 white
## space separated values in the value string to correspond with
## weekly/daily/hourly  command parameters
## (ie. snap sched -V vol0 0 5 6@8,12,16,20)
## Note the third value can contain comma separated entities, this is
## allowable for the hourly parameter only
##
## @synopsis
## $status=VSeries::filer->set_snap_sched($type_flag, $name, $value);
##
## @arg $type_flag Required: Either volume or aggregate
## @arg $name Required: Name of the vol/aggr the caller wants to set the snap
## sched value on (name only, do not append full pathnames to this field)
## @arg $values Required: string of schedule settings. There must be at least
## one value in this string, it cannot be null. This method will check that
## there are no more than 3 white space separated settings specified in the
## command but will not check validity or syntax of any settings
##
##
sub set_snap_sched($$$){
    my $self = shift;
    my ($type_flag, $name, $values) = @_;
    my $cmd = "snap sched $type_flag $name $values";
    my $after = "";
    my $j = 0;

    # Check that we have the proper type_flag (-A or -V)
    if ( defined $type_flag && ( $type_flag ne "-A" && $type_flag ne "-V" ) ) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "illegal; type_flag: $type_flag";
        return ($ERR_STATUS);
    }

    # Check that we have proper syntax and allowable range of
    # command parameters in $value. We can have 1-3 parameters.
    # The 1st and 2nd params must be digits but the third
    # parameter can be digits, "@" or "," chars
    if ( $values !~ /^\d+((\s+\d+)?|(\s+\d+\s+[\d|@|,]+)?)\s*$/ ) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "syntax error with snap schedule parameters";
        return ($ERR_STATUS);
    }

    if($self->{use_console} eq "true"){
        (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    # Verify that the schedule settings were modified. If there are less than
    # 3 settings specified, we only verify the specified settings were changed
    my $verify_cmd = "snap sched $type_flag $name";

    if($self->{use_console} eq "true"){
        (undef, undef, $after) = Tharn::say($self->{console_connection}, $verify_cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $after) = Tharn::say($self->{connection}, $verify_cmd, -swallow, "-timeout", 60);
    }

    # Get rid of any carriage return chars in command output
    $after =~ s/\r//g;
    my @sched_cmd_out = split (/\n/, $after);

    # Strip out the output line we're looking for to verify that
    # the schedule settings have been updated
    # Sample comamnd output:
    # Volume vol0: 0 2 6@8,12,16,20

    foreach my $current_values (@sched_cmd_out) {
        if ( ($type_flag eq "-A" && $current_values =~ /^Aggregate $name:/ ) ||
            ( $type_flag eq "-V" && $current_values =~ /^Volume $name:/ ) ) {
            # Strip out all chars preceeding the schedule settings
            $current_values =~ s/^.*: *//;

            # Verify the current values match the initial values
            if ( $current_values ne $values ) {
                    $self->{error} = $ERR_STATUS;
                    $self->{errmsg} = "updated snap sched values: $current_values did not match command line values: $values";
                    return ($ERR_STATUS);
            }
            last;
        }
    }
    return $GOOD_STATUS;
}

## @name ::get_fcvi_config()
##
## @description
## Runs configuration information of a Metrocluster attached FCVI cluster interconnect
## card. An array reference is returned. An empty array simply means an FCVI card isn't
## configured so this method can be run without problems on a non-Metrocluster filer.
##
## @synopsis
## $array_ref=VSeries::filer->get_fcvi_config();
##

# Example: a Qlogic 2312 configured in a vanilla FCVI Metrocluster config
#       slot 2: FCVI Host Adapter 2a (Dual-channel, QLogic 2312(2352) rev. 2, 64-bit, F-port, <UP>)
#                Firmware rev:   3.1.140
#                Host Port WWN:  20:00:00:e0:8b:09:3c:59
#                Host Loop Id:   0x81    Host Port Id:   0xe0900
#                Cacheline size: 8       FC Packet size: 2048
#                SRAM parity:    Yes     External GBIC:  No
#                Link Data Rate: 2 Gbit
#                Switch Port:    vnci9124s50:16842752
#        slot 2: FCVI Host Adapter 2b (Dual-channel, QLogic 2312(2352) rev. 2, 64-bit, F-port, <UP>)
#                Firmware rev:   3.1.140
#                Host Port WWN:  20:01:00:e0:8b:29:3c:59
#                Host Loop Id:   0x81    Host Port Id:   0x20500
#                Cacheline size: 8       FC Packet size: 2048
#                SRAM parity:    Yes     External GBIC:  No
#                Link Data Rate: 2 Gbit
#                Switch Port:    vnbr200es23:5
#
# Example: a Qlogic 2432 configured in a Spectre stretch Metrocluster config
#        slot 1: FCVI Host Adapter 1a (QLogic 2432(2462) rev. 2, L-port, <UP>)
#                Physical link: UP
#                FC Node Name: 21:00:00:1b:32:01:ab:49
#                Firmware rev: 4.0.140 Serial No: GFC0729A84395
#                Host Loop Id: 0x0
#                Host Port Id: 0xef
#                Cacheline size: 16 FC Packet size: 2048
#                SFF Vendor: FINISAR CORP.
#                SFF Part Number: FTLF8524E2KNL
#                SFF Serial Number: PBM3PA3
#                SFF Capabilities: 1, 2 or 4 Gbit
#                Link Data Rate: 4 Gbit
#                     I/O base 0x0, size 0x0
#                     memory mapped I/O base 0xc0800000, size 0x4000
#        slot 1: FCVI Host Adapter 1b (QLogic 2432(2462) rev. 2, L-port, <UP>)
#                Physical link: UP
#                FC Node Name: 21:01:00:1b:32:21:ab:49
#                Firmware rev: 4.0.140 Serial No: GFC0729A84395
#                Host Loop Id: 0x1
#                Host Port Id: 0xe8
#                Cacheline size: 16 FC Packet size: 2048
#                SFF Vendor: FINISAR CORP.
#                SFF Part Number: FTLF8524E2KNL
#                SFF Serial Number: PBM3S3Q
#                SFF Capabilities: 1, 2 or 4 Gbit
#                Link Data Rate: 4 Gbit
#                     I/O base 0x0, size 0x0
#                     memory mapped I/O base 0xc0804000, size 0x4000

sub get_fcvi_config() {

    my $self = shift;
    # Array of adapter config info hashes
    my @fcvi_config_arr = ();
    # Config info for each hash
    my %fcvi_config = ();

    # Run sysconfig with no args to get quick summary info.
    # We look for the first FCVI entry just to extract
    # its bus slot number. We issue a followup sysconfig -av
    # command with that bus slot number to extract the full
    # information of the FCVI card config. This is much
    # faster than running sysconfig -a
    my $sysconfig_out = $self->get_sysconfig();
    my $fcvi_slot = 0; # FCVI slot number

    # Flag to indicate if FCVI card is configured
    my $fcvi_configured = 0;


    foreach my $line (@$sysconfig_out) {
        if ( $line =~  /^\s+slot (\d+):\s+FCVI Host Adapter/ ) {
            $fcvi_slot = $1;
            $fcvi_configured = 1;
            last;
        }
    }

    # If FCVI card is not configured then return immediately
    # with an empty array reference.
    if ( ! $fcvi_configured ) {
        return \@fcvi_config_arr;
    }

    # Now run the sysconfig command specifically on the
    # FCVI slot. Only FCVI info will be displayed.
    my $fcvi_out_ref = $self->get_sysconfig("-av $fcvi_slot");
    my @fcvi_out = @$fcvi_out_ref;

    my @token_arr = (); # Line of tokenized commmand output
    my $tmp_str = "";   # Temp string for parsing
    # Hash key index to create unique key names for informational messages
    my $idx = 0;

    for (my $i = 0; $i < scalar @fcvi_out; $i++) {

        # An "FCVI Host Adapter" line identifies a new block of data for an
        # adapter. We tokenize this line via comma separated fields to extract
        # keys/value pairs. For now all we support is Qlogic FCVI adapters so
        # we base our tokenizing and parsing for the variants from this lone
        # manufacturer.

        # Sample output line format differences we need to deal with:
        #        slot 1: FCVI Host Adapter 1b (QLogic 2432(2462) rev. 2, L-port, <UP>)
        #        slot 2: FCVI Host Adapter 2a (Dual-channel, QLogic 2312(2352) rev. 2, 64-bit, F-port, <UP>)

        if ( $fcvi_out[$i] =~ /^\s+slot \d+: *FCVI Host Adapter (\w+)/ ) {

            # We're parsing new adapter info so clear the hash
            %fcvi_config = ();

            # Extract the slot adapter (a,b)
            $fcvi_config{adapter} = $1;

            # Save all chars between inner and outermost parentheses on
            # this line
            $tmp_str = $fcvi_out[$i];
            $tmp_str =~ s/^ .* \(//g;
            # Remove outermost parentheses char
            chop $tmp_str;

            # Tokenize comma separated entities
            @token_arr = split (/\,/, $tmp_str);

            # Since the number of tokens and token order is different for
            # each adapter, we need to do some parsing for specific keywords
            foreach my $token (@token_arr) {

                # Dual-channel is a token that may or may not appear first.
                # Since it may only be specific to one type of adapter,
                # it will go into a miscellaneous "attributes" key.
                if ( $token =~ /Dual-channel/ ) {
                    $fcvi_config{attributes} = $token;

                } elsif ( $token =~ /QLogic (\d+)\((\d+)\)\s+rev\.\s+(\d+)/ ) {
                    # If this is the Qlogic token, then extract model numbers,
                    # and rev. This format is consistent across the two
                    # adapter types currently supported
                    $fcvi_config{manufacturer} = "QLogic";
                    $fcvi_config{model_num} = $1;
                    $fcvi_config{model_sub_num} = $2;
                    $fcvi_config{rev} = $3;

                } elsif ( $token =~ /(\d+\-bit)/ ) {
                    # Extract the port type: 32,64 bit etc
                    $fcvi_config{architecture} = $1;

                } elsif ( $token =~/(\w+\-port)/ ) {
                    # Extract the port type: L-port, F-port
                    $fcvi_config{port_type} = $1;

                } elsif ( $token =~ /\<(\w+)\>/ ) {
                    # Extract the current adapter link state: UP, DOWN
                    $fcvi_config{state} = $1;
                }
            }
            next;

        } # Parsing completed for initial "FCVI Host Adapter" line

        # Parse followup lines for this adapter. We tokenize key/value pairs
        # with a pattern of a semicolon followed by one or more whitespaces.
        # The string(s) to the left of the semicolon is the key, values are
        # the remainder of the line to the right to the semicolon. We need to
        # handle special cases where there may be more than one key/value pair
        # on a line.
        # As the sysconfig output has been modified, we have changed the regex
        # from '/^(\s+\w+)*:\s+/' to '/^(\s+\w+)*\s*:\s+/' and it's been tracked
        # as part of the BURT 708048. Please refer the BURT for more information.

        if ( $fcvi_out[$i] =~ /^(\s+\w+)*\s*:\s+/ ) {

            # Remove leading whitespaces
            $fcvi_out[$i] =~ s/^\s+//;

            # Tokenize semicolon separated entities
            @token_arr = split (/\s*:\s+/, $fcvi_out[$i]);

            # Do we have only one key/value pair on this line?
            if ( scalar @token_arr == 2) {
                $fcvi_config{$token_arr[0]} = $token_arr[1];

                # If we parse the "Switch Port", we create two additional
                # keys because there are two values embedded on this line,
                # one for switch name, the other for switch port
                if ( $token_arr[0] eq "Switch Port" ) {
                    my @tmp_arr = split ( /:/, $token_arr[1]);
                    $fcvi_config{sw_name} = $tmp_arr[0];
                    $fcvi_config{sw_port} = $tmp_arr[1];
                }
            } elsif ( scalar @token_arr > 2 ) {
                # We have more than one key/value pair on this line to parse
                for ( my $j = 0; $j < scalar @token_arr; $j++ ) {

                    if ( defined $token_arr[$j+1] ) {
                        # Lookahead to the array element following the
                        # semicolon. We will split this element into a
                        # whitespace separated array @tmp_arr and extract its
                        # first element as the hash value
                        my @tmp_arr = split ( / +/, $token_arr[$j+1]);
                        $fcvi_config{$token_arr[$j]} = $tmp_arr[0];

                        # We remove the substring we just parsed from the
                        # lookahead element so that on the next loop pass,
                        # the modified array element will be interpreted as
                        # the next hash key on the line
                        $token_arr[$j+1] =~ s/^\S+\s+//;
                    }
                }
            } else {
                # We have a line of code that has no token. It is considered
                # an informational message and will have a key of the form
                # "info_${idx}" where idx is a simple index to create a unique
                # key. The hash value is the entire line of output.
                $fcvi_config{"info_$idx"} = $fcvi_out[$i];
                $idx++;
            }
        }

        # We've parsed a line of output at this point. Now we need to check
        # if next line is the last line of output or the start of the next
        # adapter's block of info. If either of these conditions is true,
        # then we're ready to add an adapter's populated hash into an array.
        if ( ! defined $fcvi_out[$i+1] || $fcvi_out[$i+1] =~ /^\s+slot \d+: *FCVI Host Adapter/ ) {
            my %fcvi_adapter = %fcvi_config;
            push @fcvi_config_arr, \%fcvi_adapter;
        }

    }

    return \@fcvi_config_arr;

}

## @name  ::get_nvram_config()
##
## @description
## Obtains NVRAM and cluster interconnect information.
##
## @synopsis
## $hash_ref=VSeries::filer->get_nvram_config();
##
#Example sysconfig (no args) brief output:
#        slot 3: NVRAM
#                Memory Size:        512 MB
#
#Examples sysconfig -av output:
#        slot 3: NVRAM (NetApp NVRAM V)
#                Revision:           H0
#                Serial Number:      514318
#                Memory Size:        512 MB
#                Battery Status:     Battery fully charged (4122 mV)
#                Charger Status:     Recharging complete
#                Running Firmware:   7 (3.5.0)
#                Cluster Interconnect Port 1: 4x fiber
#                Cluster Interconnect Port 2: 4x fiber
#
#       slot 3: NVRAM (NetApp NVRAM VI)
#                Revision:           G0
#                Serial Number:      601162
#                Memory Size:        512 MB
#                Battery 1 Status:   OK (3996 mV)
#                Charger 1 Status:   OFF
#                Running Firmware:   8 (4.7.600)
#                Cluster Interconnect Port 1: 4x copper
#                Cluster Interconnect Port 2: 4x copper
#
#Example Spectre MC:
#        slot 0: NVRAM (NetApp NVRAM VII)
#                Revision:           0
#                Serial Number:      700954
#                Memory Size:        2048 MB
#                Battery Status:     OK (3960 mV)
#                Charger Status:     OFF
#                Running Firmware:   2 (4.7.600)
#                Cluster Interconnect Port 1:  copper
#                Cluster Interconnect Port 2:  copper

sub get_nvram_config() {

    my $self = shift;
    my %nvram_config = ();
    my $sysconfig_ref = 0; # sysconfig output array reference

    # sysconfig with no args obtains only summary info and runs
    # quickly. We run it to extract only the bus slot of NVRAM.
    # We issue a followup sysconfig -av command to extract
    # the full information of the NVRAM config.
    $sysconfig_ref = $self->get_sysconfig();

    # Flag to indicate if NVRAM card is configured
    my $nvram_configured = 0;

    # Extract the NVRAM bus slot number from the 'brief'
    # sysconfig command output
    my $line = "";

    foreach $line (@$sysconfig_ref) {

        if ( $line =~  /^\s+slot\s+(\d+):\s+NVRAM/ ) {
            $nvram_config{slot} = $1;
            $nvram_configured = 1;
            last;
        }
    }

    # If NVRAM card is not configured, return immediately with empty
    # hash reference
    if ( ! $nvram_configured ) {
        return \%nvram_config;
    }

    # Now run the sysconfig command specifically on the NVRAM slot
    $sysconfig_ref = $self->get_sysconfig("-av $nvram_config{slot}");

    # If the NVRAM is part of the motherboard (ie. slot 0), then sysconfig
    # -av 0 will list all motherboard devices. In this case we will need to
    # extract only the NVRAM info into an array.
    if ( $nvram_config{slot} == 0 ) {

        # Create new array for sysconfig slot 0 output parsing
        my @sysconfig_out = @$sysconfig_ref;
        # Temp array contains NVRAM data extracted from sysconfig -av 0
        my @tmp_arr = ();
        # Flag to indicate we've discovered the 'header' line with NVRAM slot
        # info
        my $nvram_found = 0;

        for ( my $n = 0; $n < scalar (@sysconfig_out); $n++ ) {

            # If we find the NVRAM slot info line, save it to an array and set
            # a flag so that followup lines will also be saved into array
            if ( $sysconfig_out[$n] =~ /^\s+slot\s+\d+:\s+NVRAM/ ) {
                $nvram_found = 1;
                push @tmp_arr, $sysconfig_out[$n];
                next;
            }

            # If the current line is slot info for the next device then we're
            # done parsing the NVRAM data
            if ( $sysconfig_out[$n] =~ /^\s+slot \d+:/  && $nvram_found) {
                last;
            } elsif ( $nvram_found ) {
                # If we've discovered the NVRAM header, then this line is
                # sysconfig NVRAM data info, save it to the array
                push @tmp_arr, $sysconfig_out[$n];
            }
        }

        $sysconfig_ref = \@tmp_arr;
    }

    foreach $line (@$sysconfig_ref) {

        # Remove leading white spaces from line of output
        $line =~ s/^\s+//;

        # Extract the model number from the slot info line
        if ( $line =~  /^\s+slot \d+:\s+NVRAM\s+\((.*)\)/ ) {
            $nvram_config{model} = $1;
            next;
        }

        # The remaining line we put info key/value pairs separated by a
        # pattern of a semicolon followed by one or more whitespaces
        my @token_arr = split (/:\s+/, $line);
        $nvram_config{$token_arr[0]} = $token_arr[1];
    }

    return \%nvram_config;
}


## @name ::get_cluster_type()
##
## @description
## Method to determine a cluster configuration type. This is a general purpose
## method to determine standard Vseries cluster and Metrocluster configurations.
## It will call get_fcvi_config and get_nvram_config methods to determine the
## config type. It will return a string containing one of the following
## values:
## "VSeries": Standard Vseries cluster (non-Metrocluster)
## "MC-FCVI": FCVI switched cluster interconnect Metrocluster
## "MC-FCVI-Stretch-Spectre": Spectre FCVI non-switched direct connect
##                       'stretch' Metrocluster
## "MC-Stretch":  direct connect stretch Metrocluster
##
##
## @synopsis
## $cluster_config_type=VSeries::filer->get_cluster_type();
##

sub get_cluster_type() {
    my $self = shift;
    my $config_type = "unknown";
    my %fcvi_adapter = ();
    my $fcvi_adapter_ref = 0;


    # Get the NVRAM sysconfig info
    my %nvram_conf = %{$self->get_nvram_config()};

    # Get the FCVI sysconfig info: there will be an array
    # element for each configured FCVI adapter found
    my @fcvi_conf_arr = @{$self->get_fcvi_config()};

    # VSeries (non-Metro)cluster requirements:
    # 1) NVRAM is Nx copper, where N is the connector speed multiple (ie. 4)
    # 2) FCVI card is not configured.

    if ( $nvram_conf{"Cluster Interconnect Port 1"} =~  /\d+x copper/i &&
         $nvram_conf{"Cluster Interconnect Port 2"} =~  /\d+x copper/i &&
         scalar @fcvi_conf_arr == 0 ) {

        $config_type = "Clustered";
        return ($config_type);
    }

    # FCVI MC requirements:
    # 1)FCVI port types = F-port
    # 2)FCVI switch and port names must be defined

    my $fcvi_config_ok = 0;

    if ( scalar @fcvi_conf_arr > 0 ) {
        foreach $fcvi_adapter_ref ( @fcvi_conf_arr ) {

            %fcvi_adapter = %$fcvi_adapter_ref;
            # If an FCVI adapter meets all FCVI config requirements, increment
            # a counter

            if ( $fcvi_adapter{port_type} eq "F-port"
                 && $fcvi_adapter{sw_name} ne ""
                 && $fcvi_adapter{sw_port} ne "" ) {

                $fcvi_config_ok++;
            }
        }

        # Ensure that all FCVI adapters have the same settings.
        # If this is true than the $fcvi_config_ok count
        # will match the number of (hashes) adapters
        # in @fcvi_conf_arr
        if ( scalar @fcvi_conf_arr == $fcvi_config_ok ) {
            $config_type = "MC-FCVI";
            return ($config_type);
        }
    }

    # Spectre stretch MC requirements:
    # 1)NVRAM is copper
    # 2)FCVI port types = L-port
    # 3)FCVI switch and port names must not be defined

    $fcvi_config_ok = 0;
    if ( $nvram_conf{"Cluster Interconnect Port 1"} =~ /copper/i &&
         $nvram_conf{"Cluster Interconnect Port 2"} =~ /copper/i &&
         scalar @fcvi_conf_arr > 0 ) {

        foreach $fcvi_adapter_ref ( @fcvi_conf_arr ) {

            %fcvi_adapter = %$fcvi_adapter_ref;
            # If an FCVI adapter meets all FCVI config requirements, increment
            # a counter
            if ( $fcvi_adapter{port_type} eq "L-port" &&
                 $fcvi_adapter{sw_name} eq "" &&
                 $fcvi_adapter{sw_port} eq "" ) {

                $fcvi_config_ok++;
            }
        }

        # Ensure that all FCVI adapters have the same settings.
        # If this is true than the $fcvi_config_ok count
        # will match the number of (hashes) adapters
        # in @fcvi_conf_arr
        if ( scalar @fcvi_conf_arr == $fcvi_config_ok ) {
            $config_type = "MC-FCVI-Stretch-Spectre";
            return ($config_type);
        }
    }

    # Stretch MC requirements:
    # 1) NVRAM is Fibre, indicated by 'Fibre' or 'Nx fiber' where N
    #    is the connector speed multiple (ie. 4)
    # 2) FCVI is not configured

    if ( $nvram_conf{"Cluster Interconnect Port 1"} =~ /(Fibre|fiber)/i &&
         $nvram_conf{"Cluster Interconnect Port 2"} =~ /(Fibre|fiber)/i &&
         scalar @fcvi_conf_arr == 0) {

        $config_type = "MC-Stretch";
        return ($config_type);

    # Stretch MC requirements for Carnegie/Absolut:
    # 1) FCVI is configured as L-port or N-port
    # 2) FCVI switch and port names must not be defined

    } elsif (scalar @fcvi_conf_arr > 0) {
        foreach $fcvi_adapter_ref ( @fcvi_conf_arr ) {

        %fcvi_adapter = %$fcvi_adapter_ref;
        # If an FCVI adapter meets all FCVI config requirements, increment
        # a counter
        if (( $fcvi_adapter{port_type} eq "L-port" || $fcvi_adapter{port_type} eq "N-port") &&
             $fcvi_adapter{sw_name} eq "" &&
             $fcvi_adapter{sw_port} eq "" ) {

            $fcvi_config_ok++;
            }
        }

        # Ensure that all FCVI adapters have the same settings.
        # If this is true than the $fcvi_config_ok count
        # will match the number of (hashes) adapters
        # in @fcvi_conf_arr
        if ( scalar @fcvi_conf_arr == $fcvi_config_ok ) {
            $config_type = "MC-Stretch";
            return ($config_type);
        }
    }

    # If we get here then we haven't detected any of the expected config types
    # so return the default 'unknown' type
    return ($config_type);
}

## @name ::aggr_rename($$)
##
## @description
## Rename an aggregate and verify renaming was successful via aggr status.
## Return error status if rename not successful.
##
## @synopsis
## $status=VSeries::filer->aggr_rename($aggr_current_name, $aggr_new_name);
##
## @arg $aggr_current_name Required: Name of existing aggregate to be renamed
##
## @arg $aggr_new_name Required: New name for aggregate
##
sub aggr_rename($$) {
    my $self = shift;
    my ($aggr_current_name, $aggr_new_name) = @_;
    my $cmd = "aggr rename $aggr_current_name $aggr_new_name";
    my $output = "";

    if($self->{use_console} eq "true") {
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else {
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    # Verify aggr has been renamed
    $cmd = "aggr status $aggr_new_name";
    if($self->{use_console} eq "true") {
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else {
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    if ( $output =~ /^aggr status: No aggregate named \'$aggr_new_name\' exists/  ) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "ERROR: could not rename $aggr_current_name to $aggr_new_name: $output";
        return $ERR_STATUS;
    }

    return $GOOD_STATUS;

}

## @name ::vol_rename($$)
##
## @description
## Rename a volume and verify renaming was successful via vol status.
## Return error status if rename not successful.
##
## @synopsis
## $status=VSeries::filer->vol_rename($vol_current_name, $vol_new_name);
##
## @arg $vol_current_name Required: Name of existing volume to be renamed
##
## @arg $vol_new_name Required: New name for volume
##
sub vol_rename($$) {
    my $self = shift;
    my ($vol_current_name, $vol_new_name) = @_;
    my $cmd = "vol rename $vol_current_name $vol_new_name";
    my $output = "";

    if($self->{use_console} eq "true") {
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else {
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }
    # Verify volume has been renamed
    $cmd = "vol status $vol_new_name";
    if($self->{use_console} eq "true") {
        (undef, undef, $output) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-prompt", "^\*?>", "-timeout", 60);
    }
    else {
        (undef, undef, $output) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    if ( $output =~ /^vol status: No volume named \'$vol_new_name\' exists/  ) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "ERROR: could not rename $vol_current_name to $vol_new_name: $output";
        return $ERR_STATUS;
    }

    return $GOOD_STATUS;

}
#--------------------------------------------------------------------------------

## @name ::storage_array_modify()
##
## @author Warren Crossfield (warrenc@netapp.com)
##         James Silva (silva@netapp.com) - Just added getops stuff.
##
## @description
## Modifies an existing array profile and returns $GOOD_STATUS or $ERR_STATUS
## on success/failure.  Be cautious about sending your options into this
## method, as they will get truncated if they fall outside the bounds of
## the defined option formats.  i.e. $prefix cannot be longer than 4 chars.
## The following call would modify 2 of the 5 options, in this case the
## options and model fields.
##
## @synopsis
## $status = $filer_obj->storage_array_modify($name, "-options=jimbo", "-model=EMC2");
##
## @arg $name        Required: Array profile instance to modify.
##
## @arg $name_change Optional: Change existing array profile name to this.
##                   Note: max string length of 29 bytes.
##
## @arg $model       Optional: Change the model of the $name array profile.
##                   Note: max string length of 16 bytes.
##
## @arg $vendor      Optional: Change the vendor of the $name array profile.
##                   Note: max string length of 8 bytes.
##
## @arg $prefix      Optional: Change the prefix of the $name array profile.
##                   Note: max string length of 4 bytes.
##
## @arg $options     Optional: Change the options of the $name array profile.
##                   Note: max string length of 128 bytes.
##
sub storage_array_modify($$) {
    # We need a $name & at least one option
    my $self  = shift @_;   # The filer object to shift.
    my $name  = shift @_;   # The name of the array profile to modify.
    my $after = '';         # Holds raw output from Tharn::say().
    @ARGV    = @_;          # Copy @_ into @ARGV for getops processing.

    # These are all false by default.
    my($name_change,   # See @arg's section for info on these.
       $model,
       $vendor,
       $prefix,
       $options,
       ) = '';

    # storage_array modify <array-name> [-m <model>] [-n <new_name>] [-v vendor] [-p <prefix>] [-o options]

    # bash-2.05b$ rsh -l root vnv3050f9a "storage array show"
    # Prefix                        Name    Vendor             Model  Options
    # XXXX  XXXXXXXXXXXXXXXXXXXXXXXXXXXX   XXXXXXX   XXXXXXXXXXXXXXX  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX


    my $cmd = 'storage array modify';
    $cmd .= " $name";

    # Grab our options
    GetOptions( "name_change:s"    => \$name_change,  # Optional string
                "model:s"          => \$model,       # Optional string
                "vendor:s"         => \$vendor,      # Optional string
                "prefix:s"         => \$prefix,      # Optional string
                "options:s"        => \$options,     # Optional string
              );

    # Take in the string values and trim them to legal sizes
    # The sizes were all tested empirically. The 29 bytes for name bugs me.
    # Everything else seems to be a logical size. If we were using catch-
    # throw here I could also throw a warning if any of these values are
    # passed in too long. Alas, we do not use it. :( I wrapped these all in
    # if's in case we ever wanted to do other operations on them before
    # "saying" them.
    if($model) {
        my $truncated_model = $model;
        $cmd .= " -m $truncated_model";
    }
    if($name_change) {
        my $truncated_name = $name_change;
        $cmd .= " -n $truncated_name";
    }
    if($vendor) {
        my $truncated_vendor = $vendor;
        $cmd .= " -v $truncated_vendor";
    }
    if($prefix) {
        my $truncated_prefix = $prefix;
        $cmd .= " -p $truncated_prefix";
    }
    if($options) {
        my $truncated_options = $options;
        $cmd .= " -o $truncated_options";
    }
    # Send @cmd via the standard connection, no need for console as this
    # isn't available when the machine isn't up.
    (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow);


    # Did it complete properly? If not, toss an error back.
    if($after =~ m/Array profile updated/) {
        return($GOOD_STATUS);
    } else {
        $self->{error}  = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::storage_array_modify() failure: $after";
        return($ERR_STATUS);
    }
}

#--------------------------------------------------------------------------------

## @name ::storage_array_remove_port()
##
## @author Mark LaBonte (labonte@netapp.com)
##
## @description
## Deletes the specified WWPN associated with the specified array.
## Sets the filer object's error & errmsg attributes if errors are encountered.
##
## @synopsis
## $status = $filer_obj->storage_array_remove_port("EMC_SYMMETRIX_1", "50060482cb1bce3d");
##
## @arg $array_name   Required: The array associated with the WWPN to delete.
##
## @arg $port_name    Required: The WWPN to delete.
##
##
sub storage_array_remove_port($$) {

    #storage array remove-port <array-name> -p <WWPN>

    my ($self, $array_name, $port_name) = @_;
    my $after = '';

    my $cmd = "storage array remove-port $array_name -p $port_name";

    (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow, '-timeout', 30);

    if ($after =~ /Port $port_name removed from Array Profile $array_name/) {
        return ($GOOD_STATUS);
    } else {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Failed to remove port $port_name: $after";
        return ($ERR_STATUS);
    }

}

#--------------------------------------------------------------------------------

## @name ::storage_array_show(;$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Captures the information shown by the 'storage array show' command:
## vnv3040f8a> storage array show
## Prefix         Name     Vendor    Model           Options
##          IBM_1722_1        IBM     1722
##    123   IBM_1722_2        IBM     1722   ou812great2hear
## This information will be stored and returned as a hash reference.  The
## format for the hash will have the 'Name' of the array as the root key.
## Each Name will have a respective Prefix, Vendor, Model and Options.
## On error, we'll set the error flag and populate the inf0_msg array, but
## we _ONLY_ return a hash reference no matter what.  Check the filer object
## error flag for failures, then read the info_msg.  Lastly if you specify
## an $array_name you will only be returned info for that $array_name.
##
## @synopsis
## $ap_ref = $filer_obj->storage_array_show();
##
## @arg $array_name    Optional: An array profile name that we can capture
##                               information specifically for.
##
sub storage_array_show(;$) {
    my ($self, $array_name) = @_;
    my $after      = ''; # Holds raw output from Tharn::say().
    my @ap_builder = (); # Contains split up raw output from $after.
    my %ap_hash    = (); # Hash to contain array profiles.

    #  $cmd is the base command, we'll add $array_name, if specified.
    my $cmd = 'storage array show';
    $cmd   .= " $array_name" if ($array_name);

    # Take the third array slice from Tharn::say() and place it within
    # $after, it contains the raw output from our sent command.
    $after = (Tharn::say($self->{connection}, "$cmd", -swallow, '-timeout', 30))[2];

    # Did it complete properly? If not, toss an error back.
    if($after =~ /No Array Profile with the name/) {
        $self->{error}  = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::storage_array_show() failure: $after";
        return($ERR_STATUS);
    }

    # $after looks something like this:
    # Prefix         Name     Vendor    Model           Options
    #          IBM_1722_1        IBM     1722

    # Split the returned output on endlines, then strip off the header.
    # Use @ap_builder as an initial container.
    @ap_builder = split(/\n/, $after);
    @ap_builder = @ap_builder[1..$#ap_builder];

    # Did we even have an array profile?
    if (scalar @ap_builder == 0) {
        $self->{error} = $INFO_STATUS;
        push (@{$self->{info_msg}}, "Info: filer::storage_array_show(): Reports the presence of zero array profiles.\nRaw output: $after");
        return(\%ap_hash);
    }

    # Now we need to iterate over each line of the @ap_builder array
    # which at this point is 1 or more array profiles.
    
    # >>'storage array show'
    # Prefix                          Name    Vendor             Model  Options
    # abc                  3PARdata_VV_1  3PARdata                VV
    #                      3PARdata_VV_2  3PARdata                VV


    foreach my $line (@ap_builder) {
        my $is_prefix = 0;
	
	# The maximum allowed length for the prefix field is 5 characters.
	# So if the first 6 characters in an output record are all spaces,
	# it implies that a prefix isn't present. The rest of the record 
	# is then processed based on whether prefix is present or not.
	$is_prefix = 1 if ( $line !~ m/^\s{6,}/);

	my ($arr_name, $prefix, $vendor, $model, $options) = ('') x 5;
        $line =~ s/^\s+//;
        my @values = split(/\s+/, $line);
	
	if ( $is_prefix ){
	    $options = $values[4] if(scalar(@values) > 4);
	    ($prefix, $arr_name, $vendor, $model) = ( $values[0], $values[1],$values[2],$values[3] );
	 } else {
	    $options = $values[3] if(scalar(@values) > 3);
	    ($arr_name, $vendor, $model) = ( $values[0],$values[1], ,$values[2] );
	}
	# Now build a temporary hash which will be our $name key's value.
        my %temp = (
            'prefix'  => $prefix,
            'vendor'  => $vendor,
            'model'   => $model,
            'options' => $options
        );
        #
        # Now insert the %temp hash into the $name key, making %ap_hash a HoH.
        if (exists $ap_hash{$arr_name}) {
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "Error: filer::storage_array_show(): Reports the presence of duplicate array names.\nRaw output: $after";
        }
        else { %{$ap_hash{$arr_name}} = %temp; }
        #
        # Get rid of any whitespace in the hash values:
        foreach my $element (keys %{$ap_hash{$arr_name}}) {
            $ap_hash{$arr_name}{$element} =~ s/\s+//g;
        }
    }

    return(\%ap_hash);
}

#--------------------------------------------------------------------------------

## @name ::storage_array_show_ports(;$)
##
## @author Mark LaBonte (labonte@netapp.com)
##
## @description
## Run storage array show-ports on a V-Series.  Optionally takes in
## an array name if the caller wishes to obtain specific information
## associated with a given array name (array profile).  This will
## only return a hash reference.  It will however set the filer
## object's error & errmsg attributes if errors are encountered.
## The hash has the following format:  The keys are the unique
## wwpn's and they have the following values 'array_name' & 'wwnn'.
## i.e. $ports{'50060482cb1bce3d'}{'array_name'} = "EMC_SYMMETRIX_1";
##      $ports{'50060482cb1bce3d'}{'wwnn'} = "50060482cb1bce3d";
##
## @arg $array_name     Optional: Array profile you wish to get
##                                specific port info for.
##
## %ports_hash = %{$filer_obj->storage_array_show_ports()};
##
sub storage_array_show_ports(;$) {
    my ($self, $array_name) = @_;
    my $after = '';
    my %port_hash = ();
    my $wwpn = '';
    my $wwnn = '';

    # $cmd is the base command, we'll add $array_name, if specified.
    my $cmd = 'storage array show-ports';
    $cmd .= " $array_name" if ($array_name);

    # This is an example of the raw output from this command.
    #vnv3050f9a*> storage array show-ports
    #                  Array Name                    WWPN                    WWNN
    #             EMC_SYMMETRIX_1        50060482cb1bce3d        50060482cb1bce3d
    #vnv3050f9a*>

    # $after contains the raw output from our sent command.
    (undef, undef, $after) = Tharn::say($self->{connection}, "$cmd", '-timeout', 30);

    # Split the command on new lines.
    my @output = split(/\n/, $after);

    my $regex = '\s+(\S+)\s+(\w{16})\s+(\w{16})';

    foreach my $item (@output[1..$#output]) {    # Index 0 is the column headers.
        if ($item =~ /$regex/) {
            ($array_name, $wwpn, $wwnn) = ($1, $2, $3);
        } else {
            # This will catch ALL errors, be it usage or array profiles
            # not found.  We know this due to the $item not containing
            # the three elements we need per line.
            $self->{error}  = $ERR_STATUS;
            $self->{errmsg} = "storage_array_show_ports: Found a line missing one or more of the following: array_name, wwpn, or wwnn while parsing output from $cmd.\n Output is: $item";
        }

        # Now build a temporary hash which will be our $wwpn key's value.
        my %temp = (
            'array_name' => $array_name,
            'wwnn' => $wwnn,
        );

        # Now insert the %temp hash into the $name key, making %port_hash a HoH.
        %{$port_hash{$wwpn}} = %temp;

    }
    return(\%port_hash);
}


#--------------------------------------------------------------------------------

## @name ::storage_array_show_luns($;$)
##
## @author Warren Crossfield (warrenc@netapp.com)
##
## @description
## Runs the storage array show-luns command and takes in an array_name
## and optionally some options.  This will only return a hash reference, so
## error checking requires the caller to check the error attribute
## of the filer object post call.
##
## @synopsis
## %ap_hash1 = %{$filer_obj->storage_array_show_luns("EMC_SYMMETRIX_1")};
## %ap_hash2 = %{$filer_obj->storage_array_show_luns("EMC_SYMMETRIX_1",
##                                           "-p 50:06:04:8a:cb:1b:ce:2c")};
## %ap_hash3 = %{$filer_obj->storage_array_show_luns("EMC_SYMMETRIX_1",
##                                           "-p 50:06:04:8a:cb:1b:ce:2c -a")};
## %ap_hash4 = %{$filer_obj->storage_array_show_luns("EMC_SYMMETRIX_1", "-a")};
##
## @arg $array_descriptor Required: Array profile name.
## @arg $options          Optional: Options to append to the base command
##                                 'storage array show-luns $array_descriptor'
##
## Ontap Usage for command is as follows (helps user in defining $options):
## storage array show-luns <array-name> [-a] [-p <WWPN>]
## Which means $options should only be populated with stuff in []'s.
##
## Sample %ap_hash follows..
##
##  $VAR1 = 'vnbr3800s46:3.126L3';
##              $VAR2 = [
##          '50:06:04:82:cb:1b:ce:2c',
##          '50:06:04:82:cb:1b:ce:3c'
##                      ];
##  $VAR3 = ...;
##  $VAR4 = [...];
sub storage_array_show_luns($;$) {
    my ($self, $array_descriptor, $options) = @_;
    my $after      = ''; # Holds raw output from Tharn::say().
    my @ap_builder = (); # Contains split up raw output from $after.
    my %ap_hash    = (); # Hash to contain array profiles.
    my $i          = 0;  # A generic iterator. Iterate baby.

    # storage_array show-luns <array-name> [-p <WWPN>] [-a]
    # $cmd is the base command. We'll add $options if specified,
    # but $array_descriptor is required.
    my $cmd = "storage array show-luns $array_descriptor ";

    # Did the user specify any options?
    if(defined $options) {
        # Append the options.
        $cmd .= "$options";
    }

    # Send @cmd via the standard connection, no need for console as this
    # isn't available when the machine is down.
    (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow);

    # Check for errors, they are all listed in the or'd regex below.
    if ($after =~ m/(No Array Profile with the name|usage:|Please use the following format:|No array found with the name|Unable to obtain storage configuration information)/) {
        $self->{error}  = $ERR_STATUS;
        $self->{errmsg} = "storage_array_show_luns: Error found while executing command $cmd\n$after";
    }

    # Split the output on end lines.
    my @first_split = split(/\n/, $after);

    # Loop thru this stupid thing and take out the punk-assed marque headers.
    # Why do we loop and not just delete the first array element. <shrug> Dunno.
    foreach my $element (@first_split) {
        if($element =~ m/Name(\s*)WWPNs/) {
            delete @first_split[$i];
            $i++;
        } else {
            # Convert the comma delimited WS to single spaces, no comma.
            $element =~ s/,\s/ /;
            # Convert spaces > 3 to just a single space
            $element =~ s/\s*/ /;
            # Get rid of leading spaces in the string
            $element =~ s/^\s*//;
            # Now split the string on those spaces
            my @values_array = &quotewords('\s+', 0, $element);
            # Get the size of the values array
            my $values_array_size = @values_array;
            $values_array_size--;
            # Now go loopy and push shit around.
            for($i=0;$i<=$values_array_size;$i++)
            {
                if($i != 0) {
                    push @{ $ap_hash{$values_array[0]} }, $values_array[$i];
                }
            }
        }
    }
    return(\%ap_hash);
}

#--------------------------------------------------------------------------------

## @name ::storage_array_show_config(;$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Executes the storage-array-show-config and returns all the relevent data
## in an array of hashes. Note there are cases where we don't have a
## lun group, hence you may have a line item not preceeded by 'Group n'.
## For these scenarios, we'll add 'item n' as the value to the group key.
## The keys for each hash will be group, lun_count, name, atp, switch_port, & initiator.
## Optionally takes in $opt which will be appended to the 'storage array show-config'
## base command.
##
## @synopsis
## @config = @{$filer_obj->storage_array_show_config()};
##
## Sample CLI output:
## LUN Group        Array Name       Array Target Ports       Switch Port  Initiator
## Group 0 (6 LUNS) IBM_1722_1       20:04:00:a0:b8:21:a9:f9  vnci9124s52:1-3     0a
##                                   20:05:00:a0:b8:21:a9:f9  vnbr3800s1:2        0c
## Group 1 (4 LUNS) IBM_1722_1       20:05:00:a0:b8:21:a9:fa  vnbr3800s1:3        0d
##                                   20:04:00:a0:b8:21:a9:fa  vnci9124s52:1-4     0b
##         (7 LUNS) EMC_SYMMETRIX_1  50:06:04:82:cb:1b:ce:3d  vnbr3800s46:2       0c
##
## A direct attached config would look like this:
## Group 1 (4 LUNS) IBM_1722_1       20:05:00:a0:b8:21:a9:fa  N/A                 0d
##                                   20:04:00:a0:b8:21:a9:fa  N/A                 0b
## @arg $opt Optional:  A string of options to append to the 'storage array
##                      show-config' base command.
##
sub storage_array_show_config(;$) {
    my $self     = shift;
    my $opt      = shift;
    my $cmd      = "storage array show-config";
    my @aoh      = (); # Array of Hashes that will store our config output.
    my $after    = ""; # Init the value that stores our Tharn::say() output.
    my @output   = (); # Used to hold the split up output from the cmd.
    my $count    = 0;  # Used as the main loop counter.


    # Did the user specify any options, if so append to base.
    if (defined $opt) { $cmd .= " " . $opt; }

    # The following are most of the values for the keys in our return hash.
    my ($group, $lun_count, $name, $array_tp, $switch_port, $initiator) = ();

    # Return the 2nd slice from the cmd output (which is $after).
    ($after) = (Tharn::say($self->{connection}, $cmd, -swallow))[2];

    # Split $after on endlines and throw it in an array called @output.
    @output = split(/\n/, $after);

    # This will serve as our array index counter, which help us understand
    # where the the lun group data will be indexed.  Everytime we detect a
    # new lun group, we'll increment this count.
    my $i = -1;

    # Skip the header, as it won't be needed for anything.
    # Then begin to walk our output placing values where they belong.
    # There are 3 types of lines as seen in the header.
    foreach my $line (@output[1..$#output]) {
        
        # Clear any whitespace from the beginning of the line.
        $line =~ s/^\s+//;

        if ($line =~ /\d+ LUN[S]?/) {
            # Increment our array index counter.
            $i++;

            # Increment the new record counter, records being new LUN groups
            # or items.
            $count += 1;

            # Parse up the current line if LUNS are preceeded by
            # Group, do it this way:
            # $1      $2        $3              $4                      $5              $6
            # Group 0 (11 LUNS) EMC_SYMMETRIX_1 50:06:04:82:cb:1b:ce:2d vnci9124s51:1-1 0a
            if ($line =~ /^(Group\s+\d+)\s+\((\d+)\s+LUN[S]?\)\s+(\w+)\s+(\S{23})\s+(\S+)\s+(\S+)/) {
                $group          = $1;
                $lun_count      = $2;
                $name           = $3;
                $array_tp       = $4;
                $switch_port    = $5;
                $initiator      = $6;

                # Start filling the %temp hash with a group.
                $aoh[$i]->{'group'} = $group;
            }
            # This line would not have a Group number.
            # $1        $2              $3                      $4              $5
            # (11 LUNS) EMC_SYMMETRIX_1 50:06:04:82:cb:1b:ce:2d vnci9124s51:1-1 0a
            elsif ($line =~ /\((\d+)\s+LUN[S]?\)\s+(\w+)\s+(\S{23})\s+(\S+)\s+(\S+)/) {
                $lun_count      = $1;
                $name           = $2;
                $array_tp       = $3;
                $switch_port    = $4;
                $initiator      = $5;

                # Start filling the %temp hash with an item.
                # Note the $count doesn't need to be sequential - just needs to
                # be different each time we find a line that contains LUNS and
                # doesn't contain Group - we've covered that.
                $aoh[$i]->{'group'} = "item $count";
            }

            # Continue to populate the %temp hash.
            $aoh[$i]->{'name'}      = $name;
            $aoh[$i]->{'lun_count'} = $lun_count;

            push(@{$aoh[$i]->{'atp'}},          $array_tp);
            push(@{$aoh[$i]->{'switch_port'}},  $switch_port);
            push(@{$aoh[$i]->{'initiator'}},    $initiator);
        }
        # Maybe this line just contains ATP's, SP's and Initiators.
        # $1                      $2            $3
        # 50:06:04:82:cb:1b:ce:3d vnbr3800s46:2 0c
        elsif ($line =~ /(\S{23})\s+(\S+)\s+(\S+)/) {
                $array_tp       = $1;
                $switch_port    = $2;
                $initiator      = $3;
                
                # Push these values onto the three common arrays found
                # in all lun groups.
                push(@{$aoh[$i]->{'atp'}},          $array_tp);
                push(@{$aoh[$i]->{'switch_port'}},  $switch_port);
                push(@{$aoh[$i]->{'initiator'}},    $initiator);                
        }
        else {
            # This line wouldn't match anything, so let's fail.
            $self->{error}  = $ERR_STATUS;
            $self->{errmsg} = "storage_array_show_config(): Failed to parse a line of output from $cmd.\nLine was $line";
        }
    }

    # Before we return \@aoh we need to sort values within the lun groups
    # so compares will work even if the primary/secondary paths flipped.
    # This can create a scenario where the 'atp', 'switch_port' and 'initiator'
    # arrays per lun group may have different ordering.  i.e. during a compare
    # of group1 pretest vsersus posttest 'initiator' may be (0a, 0c) versus
    # (0c, 0a).  This compare really should not fail the test as the data is
    # still accurate, but the array profiles ONTAP code built the tables
    # differently after paths flipped.  As a result, we will preserve the
    # relationships of each initiator and its associated 'switch_port' and
    # 'atp' but we'll sort based on 'initiator' so that in theory our compares
    # will always be sorting data in the correct order.  Thank you ONTAP
    # for making life difficult :)
    foreach my $lg_ref (@aoh) {

        # The populate our temporary containers.
        my @array1 = @{$lg_ref->{'atp'}};
        my @array2 = @{$lg_ref->{'initiator'}};
        my @array3 = @{$lg_ref->{'switch_port'}};

        # Use "map" to setup the relationships
        my @relationships = map { [ $array1[$_], $array2[$_], $array3[$_] ] } (0..$#array1);

        # Sort and rearrange the data
        foreach my $r ( sort { $a->[1] cmp $b->[1] } @relationships ) {
            # Push the reordered data onto the bottom
            # of the "stack" and shift off the old data from the top of
            # the stack. You end up with the the orig arrays, reordered.
            push(@array1, @$r[0]);  # <-- atp's
            shift(@array1);

            push(@array2, @$r[1]);  # <-- initiators
            shift(@array2);

            push(@array3, @$r[2]);  # <-- switch ports
            shift(@array3);
        }

        # Replace old with new
        @{$lg_ref->{'atp'}}         = @array1;
        @{$lg_ref->{'initiator'}}   = @array2;
        @{$lg_ref->{'switch_port'}} = @array3;

    } #$lg_ref

    return(\@aoh);
}

#--------------------------------------------------------------------------------

## @name ::storage_array_purge_database()
##
## @author James Silva (silva@netapp.com)
##
## @description
## Purge the array profiles database which deletes
## /net/vnv3040f8a/vol/vol0/etc/arraymon forcing a rescan that will
## recreate the arraymon file.
##
## @synopsis
## $status = $filer_obj->storage_array_purge_database();
##
sub storage_array_purge_database() {
    my $self  = shift;  # The filer object to shift in.
    my $after = '';     # Holds raw output returned from Tharn::say().

    # Take the third array slice from Tharn::say() and place it within
    # $after, it contains the raw output from our sent command.
    ($after) = (Tharn::say($self->{connection}, "storage array purge-database", -swallow, '-timeout', 30))[2];

    # On success we are told the database has been cleared.
    if ($after !~ /Array Profile database has been cleared/) {
        $self->{error}  = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::storage_array_purge_database(): May not have cleared the array profiles database located at \/vol\/vol0\/etc\/arraymon.\nReturned output: $after";
        return($ERR_STATUS);
    }

    return($GOOD_STATUS);
}
#--------------------------------------------------------------------------------

## @name ::storage_array_remove($;$$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Removes an array profile instance.  These instances are names of arrays
## that have the following format: vendor_catmodel_instance.  The concatenated
## model name comes from the ONTAP code trashing any remaining data that is
## present to the right of a dash '-'.  Instance exists because their may be
## more than one of the same array out on the back end.  Since the ONTAP
## code suggests doing so, an option is provided to purge the array profiles
## database once the array profile instance has been removed.  We may not
## always want to purge, so this is an optional parameter.  Lastly when pro-
## viding values for 'storage-array' commands please ensure they are within
## the bounds of those variables, i.e. array-name can't be greater than 28
## characters, otherwise ONTAP complains with 'usage:' statements.
##
## For SN the success message will be 'Array Profile x has been removed'.
## For others it will be 'This array profile.*still present on the fabric'.
##
## Note: This works on BR-> user beware.
## Purge database is not required in SN.
##
## @synopsis
## $status = $filer_obj->storage_array_remove($array_name, $purge_flag);
##
## @arg $array_name    Required: Array profile instance to remove.
##
## @arg $purge_flag    Optional: Set the flag to purge, undef by default.
##
## @arg $ontap_version Optional: Ontap version on the filer (IC, BR, RR, SN, etc)
##
sub storage_array_remove($;$$) {
    my ($self, $array_name, $purge_flag, $ontap_version) = @_;

    if (!defined $ontap_version) {
        $ontap_version = "unknown";
    }

    my $after = '';    # Holds raw output from Tharn::say()

    # Take the third array slice from Tharn::say() and place it within
    # $after, it contains the raw output from our sent command.
    ($after) = (Tharn::say($self->{connection}, "storage array remove $array_name", -swallow, '-timeout', 30))[2];
    if ($ontap_version eq "SN") {
        # This is the positive case, all others are negative.
        if ($after !~ /Array Profile.*has been removed/) {
            $self->{error}  = $ERR_STATUS;
            $self->{errmsg} = "Error: filer::storage_array_remove(): Could not remove the array profile $array_name.\n$after";
            return($ERR_STATUS);
        }
    }
    else {
        # This is the positive case, all others are negative.
        if ($after =~ /Note: This array profile.*still present on the fabric/) {
            # If the purge flag was set, purge the database.
            if ($purge_flag) {
                return($ERR_STATUS) if ($self->storage_array_purge_database() == $ERR_STATUS);
            }
        } else {
            $self->{error}  = $ERR_STATUS;
            $self->{errmsg} = "Error: filer::storage_array_remove(): Could not remove the array profile $array_name.\n$after";
            return($ERR_STATUS);
        }
    }
    return($GOOD_STATUS);
}

#--------------------------------------------------------------------------------

## @name ::discover_lun_group_count()
##
## @author James Silva (silva@netapp.com)
##
## @description
## Calls storage_array_show_config method which returns a data structure
## containing lun group information.  From that, discover_lun_group_count will
## figure out how many lun groups we have, and set the 'lun_group_count'
## filer object attribute.  If a hash in @aoh exists where the key 'name' has
## a value of 'item \d+' i.e. not a group, we won't increment our counter.
## That's because in order to be considered a lun group, the luns within it
## must be multipathed.  We call single pathed groups items.
##
## @synopsis
## $status = $filer_obj->discover_lun_group_count();
##
sub discover_lun_group_count() {
    my $self = shift;
    my @aoh = ();           # Stores the array profile config.
    my $num_groups = 0;     # Number of lun groups we find.
    my %temp = ();          # Used as temp holder for each elemnt of @aoh.

    # Call storage_array_show_config which will return to us an aoh.
    @aoh = @{$self->storage_array_show_config("-a")};

    # Check for errors from the call...
    if ($self->get_error_state() == $ERR_STATUS) {
        return($ERR_STATUS);
    }

    # Walk the aoh to determine how many groups we have. We do this because
    # not all hashes in the @aoh are lun groups, some of them are what we
    # call items, which are a group of single pathed luns.
    foreach my $hash_ref (@aoh) {
        %temp = %{$hash_ref};

        # Determine if the 'group' key matches 'Group n'.
        if ($temp{'group'} =~ m/Group\s+\d+/) { $num_groups++; }
    }

    # Set the lun_group_count attribute.
    $self->{'lun_group_count'} = $num_groups;

    return($GOOD_STATUS);
}


#--------------------------------------------------------------------------------

## @name ::get_lun_group_count()
##
## @author James Silva (silva@netapp.com)
##
## @description
## Returns the value held in the 'lun_group_count' filer object attribute.
## This value is the number of lun groups found as reported by the
## 'storage array show-config' ontap BR cli command.
##
## @synopsis
## $count = $filer_obj->get_lun_group_count();
##
sub get_lun_group_count() {
    my $self = shift;
    return ($self->{lun_group_count});
}

#--------------------------------------------------------------------------------

## @name ::fcadmin_offline_online($$)
##
## @author Mark LaBonte (labonte@netapp.com)
##
## @description
## Offlines or onlines the supplied adapter, and returns $GOOD_STATUS
## or $ERR_STATUS.
##
## @synopsis
## $result = $filer_obj->fcadmin_offline_online("offline", "0a");
##
## @arg $state   Required: State to set adapter to (offline|online).
##
## @arg $adapter Required: Adapter to change the state of.
##
sub fcadmin_offline_online($$) {
    my ($self, $state, $adapter) = @_;
    my $cmd1 = "priv set -q diag;fcadmin $state $adapter";
    my $cmd2 = "priv set -q diag;fcadmin link_state";
    my $after = '';
    my $timeout_flag = '';

    # If we are setting the state to ONLINE, then sleep for 30 seconds
    # first to allow for port "debounce logic".
    if ($state eq "online") {sleep 30;}

    # Issue our command.
    (undef, undef, $after) = Tharn::say($self->{connection}, $cmd1, -swallow);

    # Make sure we weren't passed a non-existent adapter somehow.
    if ($after =~ /fcadmin offline: No such channel/) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "$after";
        return ($ERR_STATUS);
    }

    # Set a timeout and check for the specified adapter state. FAIL if the
    # adapter has not changed to the specified state within the timeout period.
    my $timeout = time() + 120;
    while (time() < $timeout) {
        (undef, undef, $after) = Tharn::say($self->{connection}, $cmd2, -swallow);

        # Sleep briefly to prevent too much output in the log.
        sleep (3);
        if ($state eq "offline") {
            if ($after =~ /Link State for adapter channel $adapter: OFFLINE/) {
                $timeout_flag = 0;
                return ($GOOD_STATUS);
                last;
            } else {
                $timeout_flag = 1;
                next;
            }
        } elsif ($state eq "online") {
            if ($after =~ /Link State for adapter channel $adapter: UP/) {
                $timeout_flag = 0;

                # Sleep briefly to give devices on this adapter time to come
                # online.
                sleep (3);
                return ($GOOD_STATUS);
                last;
            } else {
                $timeout_flag = 1;
                next;
            }
        }
    }
    if ($timeout_flag == 1) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Failed to $state adapter $adapter.";
        return ($ERR_STATUS);
    }
}

#--------------------------------------------------------------------------------

## @name ::systemshell_login_logout($;$$)
##
## @author Mark LaBonte (labonte@netapp.com)
##
## @description
## Logs in or out of the system shell and returns $GOOD_STATUS
## or $ERR_STATUS. The node must be up and running before this method is called,
## and since console connections are used, it is strongly recommended that you
## log out of the system shell (see $action) when finished with it, or else
## Bad Things Will Happen Further Down The Line.
##
## @synopsis
## $result = $filer_obj->systemshell_login_logout("login", "root", "admin");
##
## @arg $action Required: Tells the method whether to log in ("login") or log out ("logout").
##
## @arg $user   Optional: User to log in as.
##
## @arg $pass   Optional: Password for $user.
##
sub systemshell_login_logout($;$$) {

    my ($self, $action, $user, $pass) = @_;
    my $change_to_systemshell = "priv set diag;systemshell";
    my $login_prompt = "login: ";
    my $pass_prompt = "Password:";
    my $systemshell_prompt = "$self->{name}\#";
    my $after;
    my $test_systemshell_prompt;
    my $test_normal_prompt;

    # If we're logging in, change to the appropriate privilege set, issue "systemshell",
    # look for the expected prompts, supply name and password, and check for the expected login string.
    if ($action eq "login") {

        # Test to see if we're already at the system shell prompt. If so, we're done here.
        $test_systemshell_prompt = (Tharn::say($self->{console_connection}, " "))[3];
        if ("$test_systemshell_prompt" eq "$systemshell_prompt") {
            return $GOOD_STATUS;
        }

        # Issue our systemshell command, then wait for the login prompt.
        Tharn::say($self->{console_connection}, $change_to_systemshell, -noprompt, "-prompt", $login_prompt);

        # Supply the username when prompted.
        Tharn::say($self->{console_connection}, $user, "-prompt", $pass_prompt);

        # Supply the password when prompted.
        $after = (Tharn::say($self->{console_connection}, $pass, "-prompt", $systemshell_prompt))[2];

        # Look for evidence of a successful login, and return the appropriate status.
        if ($after =~ /Last login:/) {
            return $GOOD_STATUS;
        } else {
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "System shell login failed.";
            return $ERR_STATUS;
        }

    # If we're logging out, issue the "exit" command and look for the expected logout string.
    } elsif ($action eq "logout") {

        # Test to see if we've already exited the system shell. If so, we're done here.
        $test_normal_prompt = (Tharn::say($self->{console_connection}, " "))[3];
        if ("$test_normal_prompt" ne "$systemshell_prompt") {
            return $GOOD_STATUS;
        }

        # Exit from the system shell.
        $after = (Tharn::say($self->{console_connection}, "exit"))[2];

        # Look for evidence of a successful logout, and return the appropriate status.
        if ($after =~ /cmd_systemshell: statusCode=1/) {
            return $GOOD_STATUS;
        } else {
           $self->{error} = $ERR_STATUS;
           $self->{errmsg} = "System shell logout failed.";
           return $ERR_STATUS;
        }
    }
}

#--------------------------------------------------------------------------------

## @name ::ontapi_cmd($;$)
##
## @author Warren Crossfield (warrenc@netapp.com)
##
## @description
## Issues your ontapi command to a filer and passes back a hash
## containing the parsed XML data
##
## @synopsis
## $result = $filer_obj->ontapi_cmd("ontapi_cmd");
##
## example;
##
## my @command = ("array-name", "EMC_SYMMETRIX_1");
## $result = $Filer_Obj->ontapi_cmd("storage-array-luns-list-info", @command);
##
##
## @arg $api_command Required: The ontapi command to run.
## @arg $options Optional: The ontapi options to run.
##
sub ontapi_cmd ($;$) {

    my $self = shift;
    my $api_command = shift;
    my @options = @_;
    my $result = '';


    # Invoke the server to run zapi commands at this filer.
    my $server = new ZapiTest($self->get_name());

    # Make the call to the zapi api and check that the result is ok.  If there
    # was an error return it to the caller via the proper mechanism.
    if(@options) {
        $result = $server->invoke($api_command, @options);
    }
    else {
        $result = $server->invoke($api_command);
    }
    # Error check it beyotch-o-rama.
    if (! $result->results_ok() ) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = $result->results_reason();
        return $ERR_STATUS;
    }

    # The NaElement data object is a bit clunky so let's turn it to a string
    # and parse the xml.
    my $string = $result->toString();

    # In the event of an error this produces a nice tabbed human readable
    # form of the xml for printing in logs or to the console.
    my $xml_string = $result->sprint_pretty();

    # Use the XML parser to break this into a hash
    my $xml = new XML::Simple;
    my $data = $xml->XMLin($xml_string);

    return $data;
}


## @name ::vseries_storage_port_associate
## @description Returns a reference to a hash of filer names and their
## associated switch:port's.
##
## @synopsis
## vseries_storage_port_associate {\@filer_list}
##
##
## @author = silva@netapp.com
##
sub storage_port_associate() {
    my $self = shift;


    my %storage_port_list = ();           # Hash used to track associated storage ports
    my @output_filera;                    # Output of storage show disk -p to filera

    my @filera_lun_paths = ();            # List of all paths to luns
    my $after = "";
    my $sw_port = '';                     # Temporary holder for switch ports used in loops
    my %seen_in_filera = ();              # Hash used to help build an array of uniq
    my @uniq_in_filera = ();              # Storage ports in filera '@uniq_in_filera'



    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Set @output_filera with the proper output
    # for future processing.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -


    my $cmd ="storage show disk -p";


    if($self->{use_console} eq "true"){
        (undef, undef, $after) = Tharn::say($self->{console_connection}, $cmd, -swallow, "-timeout", 60);
    }
    else{
        (undef, undef, $after) = Tharn::say($self->{connection}, $cmd, -swallow, "-timeout", 60);
    }

    @output_filera = split(/\n/, $after);



    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now eliminate the first 2 lines of output produced from the
    # 'storage show disk -p' command.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    @output_filera = @output_filera[2 .. $#output_filera]; # (Programming Perl


    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now @output_filera is a list of all the primary/secondary
    # LUNS combined for each specific filer.
    #
    # Example:
    #
    # vss_b4100_1:7.126L8     -    vss-m6140-1:3.127L8     -     -    -
    # vss_b4100_1:7.126L16    -    vss-m6140-1:3.127L16    -     -    -
    # vss_b4100_1:7.126L24    -    vss-m6140-1:3.127L24    -     -    -
    # vss_b4100_1:7.126L36    -    vss-m6140-1:3.127L36    -     -    -
    #
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now populate @filera_lun_paths
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    my ($p_lun, $s_lun) = '';

    foreach my $tmp1 (@output_filera){
        ($p_lun, undef, $s_lun, undef, undef, undef) = split(/\s+/, $tmp1);
        push @filera_lun_paths, $p_lun;
        push @filera_lun_paths, $s_lun;
    }



    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # @filera_lun_paths now contains
    # all of the primary and secondary luns from the result
    # of the 'storage show disk -p' command.  Now we need to
    # further carve up each lun path down to the
    # switch_name:storage_port_number level.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Lets clear @output_filera so we may reuse it and not create additional
    # arrays.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    @output_filera = ();


    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Splitting on \. splits the switch:storage_port.lun_number
    # so that you are left with two pieces, switch:storage_port
    # and lun_number.  Lets do that for both vseries heads.
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $tmp3 (@filera_lun_paths){
        ($sw_port, undef) = split(/\./, $tmp3);
        push @output_filera, $sw_port;
    }



    # -  - - - - - - - - - - - - - - - - -  - - - - - -
    # Now we need to eliminate all duplicates in @output_filera
    # and @output_filerb.  We will use %seen_in_filera/b to do this,
    # then send the uniq elements to @uniq_in_filera/b
    #
    # The foreach blocks ahead are explained here:
    # (Recipe 4.6 page 102 Perl Cookbook)
    # -  - - - - - - - - - - - - - - - - -  - - - - - -

    foreach my $item1 (@output_filera){
        push (@uniq_in_filera, $item1) unless $seen_in_filera{$item1}++;
    }


    return (\@uniq_in_filera);
}


## @name ::storage_port_find($)
##
## @description takes in a switch and port value where the expected format
## is the same as what's seen in disk path names for vseries which is the
## <switch_name>:<port>.  The method will walk the internal disk hash of
## the object looking any disk that hash either it's primary or secondary
## path going through the switch and port value.  If it finds a match it
## returns GOOD_STATUS and if there is no match it returns $ERR_STATUS.
##
## @synopsis
## my $return_status = $filer_obj->storage_port_find($switch_and_port_value);
##
##
## @author  (James Silva) silva@netapp.com, (Matt Cattel) cattel@netapp.com
##
sub storage_port_find($){
    my $self = shift;
    my ($switch_and_port_num) = @_;

    # Get the objects internal disk hash for walking.
    my %seen_disks = %{$self->{disks}};

    # For each disk we want to look at both the primary and secondary paths
    # so let's pull those out.
    foreach my $serial (keys %seen_disks){
        my $primary = $self->{'disks'}->{$serial}->{'primary'};
        my $secondary = $self->{'disks'}->{$serial}->{'secondary'};

        # This will separate the switch:port value from the lun info about the
        # disk.
        my $switch_port = (split(/\./, $primary))[0];

        # If the primary path is a match return otherwise we'll check the
        # secondary path
        if($switch_port eq $switch_and_port_num){
            return $GOOD_STATUS
        }

        # If the secondary path isn't a dash (meaning there's no secondary
        # path) then lets do the same split to pull out the switch:port
        # value and compare.
        elsif($secondary ne "-"){
            $switch_port = (split(/\./, $secondary))[0];

            # If we have a match here return GOOD_STATUS.
            if($switch_port eq $switch_and_port_num){
                return $GOOD_STATUS
            }
        }

    }

    # If we get here then we didn't find a match in the loop above so return
    # ERR_STATUS to indicate a match wasn't found.
    return $ERR_STATUS;



}


## @name ::fcadmin_link_state($)
##
## @description
## Takes in a server object to communicate with the filer and does a
## "fcadmin link_state".  It then splits the output and tokenizes on whitespace
## creating an array.  From there it creates a hash table with the key being
## the interface name (0a,0b,0c,etc) and the value being a hash table that
## contains attributes for each piece of information on the line.  See sample
## output and notes around it for examples.
##
## @synopsis
## $fcadmin_hash_ref = $filer_obj->fcadmin_link_state();
##
##
##
## SAMPLE OUTPUT:
##"priv set advanced; fcadmin link_state"
##Warning: These advanced commands are potentially dangerous; use
##         them only when directed to do so by Network Appliance
##         personnel.
##Link State for adapter channel 0a: UP - N-port --> Port 0 on switch 'vrs-br200e-3' (100000051e03ba6d)
##Link State for adapter channel 0b: OFFLINE (physical)
##Link State for adapter channel 0c: UP - N-port --> Port 0 on switch 'vrs-br200e-4' (100000051e03b684)
##Link State for adapter channel 0d: OFFLINE (physical)
##
## So if you take the line with 0a as the interface the hash that gets returned
## will have key value pairs of
## slot -> 0a
## state -> UP
## port_type -> N-port
## switch_port -> 0
## switch -> vrs-br200e-3
## port_wwn -> 100000051e03ba6d

sub fcadmin_link_state(){
    my $self = shift;
    my %return_hash = ();

    # The priviledge level for this command is advanced so lets see what
    # priv level the filer is at so we can return it to that after the
    # the command.
    my $priv = $self->get_priv();

    # Set the priv level to advanced over the console so that it's
    # persistent.
    $self->send_filer_cmd("priv set advanced", "console_connection");

    # variables used to hold returned output from say calls.
    my ($after, $prompt) = "";

    (undef, undef, $after, $prompt) = Tharn::say($self->{console_connection}, "fcadmin link_state", -swallow);

    # Split the raw output on new lines.
    my @temp = split(/\n/, $after);

    # for each line of output we only care about lines that begin with
    # Link State.  From there we tokenize on whitespace and take
    # specific tokens and put them into the returned hash for a given
    # interface.  See header for how the line gets tokenized and what
    # tokens end up in what variables, should be straight forward.
    foreach my $line (@temp) {

        if ($line =~ m/^Link State/) {
            my @tokens = split(/\s+/, $line);
            my %tmp_hash = ();

            my $current_slot = "";
            if (defined($tokens[5])) {
                chop($tokens[5]);
                $current_slot = $tokens[5];
            }
            $tmp_hash{'slot'} = $current_slot;
            if (defined($tokens[6])) {
                $tmp_hash{'state'} = $tokens[6];
            }
            if (defined($tokens[8])) {
                $tmp_hash{'port_type'} = $tokens[8];
            }
            if (defined($tokens[11])) {
                $tmp_hash{'switch_port'} = $tokens[11];
            }
            if (defined($tokens[14])) {
                $tmp_hash{'switch'} = substr($tokens[14], 1, -1);
            }
            if (defined($tokens[15])) {
                $tmp_hash{'port_wwn'} = substr($tokens[15], 1, -1);
            }

            $return_hash{$current_slot} = {%tmp_hash};

        }
    }

    # Return the console connection back to the priv level it was at before the
    # call was made.
    $self->send_filer_cmd("priv set $priv", "console_connection");

    # Return the parsed output in hash form.
    return \%return_hash;

}




## @name ::poll_hba_interfaces($$$@)
##
## @description
## This is a convenience routine to poll on hba interfaces to detect
## transitions of state.  It is common practice in VSeries testing
## to manipulate the backend san through switches by doing
## port disable/enable, switch reboots, switch renaming, etc.  During these
## tests we want to poll on the hba interfaces that the filer sees the given
## switch through.  This method will poll on the given interfaces and loop
## until all interfaces hit the desired state or a timeout occurs.  The method
## will call fcadmin_link_state though that can be swapped out once an
## equivalent command is available in 10-Mode.
##
## @synopsis
## my $status = $filer_obj->poll_hba_interfaces($state, $timeout, \@interfaces);
##
##
## @arg $state Required: The expected resulting state you want the interfaces
##                       to transition to.
##
## @arg $timeout Required: The maximum timeout to poll for the interfaces to
##                         transition to the end state.
##
## @arg @interfaces Required: Array of interface names to poll on the given
##                            filer that the server object is configured to
##                            talk to.
##
sub poll_hba_interfaces($$@){
    my $self = shift;
    my ($state, $timeout, $interfaces_ref) = @_;
    my @interfaces = @$interfaces_ref;

    # Timeout we'll use to control looping for polling.
    my $polling_timeout = time() + $timeout;

    # This is the flag that will be used to control if we should keep polling
    my $continue_polling = 1;

    # Keep polling until we hit the timeout or the flag is set to stop.
    while (time <= $polling_timeout && $continue_polling == 1) {
        $continue_polling = 0;

        # Call fcadmin linkstate to pull out the state of the interfaces
        # we're polling on.
        my $fcadmin_poll_ref = $self->fcadmin_link_state();
        my %fcadmin_poll_hash = %$fcadmin_poll_ref;

        # Loop through the hash which has keys of the hba interfaces
        # find the interfaces we're supposed to poll on and check
        # their state to see if it matches the expected state.
        foreach my $port (keys(%fcadmin_poll_hash)) {
            foreach my $port_to_poll (@interfaces) {
                # Move onto the next port if this isn't a port we're
                # polling on.
                if ($port ne $port_to_poll) {
                    next;
                }
                # If this is a port we're polling on check that the
                # state is what we expect, if it isn't then we need
                # to keep polling.
                elsif ($fcadmin_poll_hash{$port}{'state'} ne $state) {
                    $continue_polling = 1;
                    last;
                }

            }
            # This is for efficiency, if there's a reason to keep polling for
            # any reason there's no need to keep looking, just break out and
            # poll again.
            if ($continue_polling == 1) {
                last;
            }
        }

        # Sleep and re-poll.
        if ($continue_polling == 1) {
            sleep(3);
        }

    }

    # If continue polling is set, we must have timed out so return $ERR_STATUS
    if ($continue_polling == 1) {
        return $ERR_STATUS;

    }

    # Otherwise return good status.
    return $GOOD_STATUS;


}



## @name ::storage_load_balance(;$)
##
## @author Mathew Cattel (Mathew.Cattel@netapp.com)
##
## @description
## This method kicks off the storage load balance command which does an
## ra_recast of path weight for load balancing in VSeries.  It will either
## verify the load balance completes or can be run with a no-verify flag.
##
## @arg no_verify Optional: Default is undef, if a scalar is passed in to the
##                          call the command will not verify if a load balance
##                          was done.
##
## @synopsis
## $status = $filer_obj->storage_load_balance();
##
sub storage_load_balance(;$) {
    my $self  = shift;  # The filer object to shift in.
    my ($no_verify) = @_;


    my $after = '';     # Holds raw output returned from Tharn::say().
    my $poll_time = "";
    my $date = "";

    # If we want to verify that the command actually kicked off a rebalance
    # then we'll get the time off the filer and store it for checking after we
    # issue the load balancing command.
    if(!defined($no_verify)){
        ($date) = (Tharn::say($self->{connection}, "date", -swallow, '-timeout', 120))[2];
        # we sleep for 2 seconds here because it's possible to issue the load
        # balance command in less than a second after and then the dates are
        # the same.  This sleep makes sure that we know for a fact if the
        # command did kick off a rebalance.
        sleep 2;
    }

    # Take the third array slice from Tharn::say() and place it within
    # $after, it contains the raw output from our sent command.

    eval {
        ($after) = (Tharn::say($self->{connection}, "storage load balance", -swallow, '-timeout', 5))[2];
    };

    if ($@ =~ m/could not establish connection/) {  

        # This most likely means that the command was issued during a giveback
        # and rshd on the filer isn't ready yet. In this case, we'll start a 
        # polling loop using a short say() timeout in order to get the command
        # through to the filer quickly.

        my $timeout = time() +30;
        my $timeout_flag = 1;
        while (time <= $timeout) {
            eval {
                ($after) = (Tharn::say($self->{connection}, "storage load balance", -swallow, '-timeout', 3))[2];
            };
            if ($@ =~ m/could not establish connection/) { 
                sleep 2;
            } elsif ($@) {
                $self->{error} = $ERR_STATUS;
                $self->{errmsg} = "Error: filer::storage_load_balance: $@";
            } else {
                $timeout_flag = 0;
                last;
            }
        }

        if ($self->{error} == $ERR_STATUS) {
            return $ERR_STATUS;
        } elsif ($timeout_flag == 1) {
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "Error: filer::storage_load_balance: Timed out trying to issue storage load balance command.";
            return $ERR_STATUS;
        }

    } elsif ($@) {
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "Error: filer::storage_load_balance: $@";
        return $ERR_STATUS;
    }

    # If we're to verify that the command actually kicked off a rebalance
    # we'll enter this block.
    if(!defined($no_verify)){

        my $last_balanced_date = "";

        # We're going to issue a storage load show here because at the end of
        # the output is the timestamp for the last time a load balance was done
        # The line looks like this:
        # I/O was last balanced at: Mon Feb  2 14:26:24 EST 2009
        my ($output) = (Tharn::say($self->{connection}, "storage load show", -swallow, '-timeout', 30))[2];

        # Split on newlines so each element in the array is a line of output.
        my @tmp_output = split(/\n/, $output);
        foreach my $line (@tmp_output){

            # Next until we find the line we're looking for.
            if($line !~ m/balanced at\:/){next;}

            # Split out the date to the right of the 'at:'
            $last_balanced_date = (split(/at\: /, $line))[1];
            last;
        }

        # This is where our error checking is performed, if $last_balanced_date
        # is before $date or if it's still the empty string (unexpected error)
        # the ret_val here for compare_dates will be something other than 2
        # and we'll error accordingly.  A return value of 2 means the second
        # date is later than the first date.
        my $ret_val = VSeries::utils::compare_dates($date, $last_balanced_date);

        if($ret_val != 2){

            # If we've entered this code path it's because we expected Ontap
            # to have issued a load balance and it did not.  The only tolerable
            # case where this can happen is if Ontap had asynchronously
            # issue a load balance that had completed within 10 seconds of
            # when we issued our command.  So we'll check for that case by
            # taking the timestamp we took right before issuing the command
            # and subtracting 10 seconds and seeing if that is before
            # the date Ontap is saying is when the last load balance
            # completed.  Any other cases of why a load balance
            # didn't complete when we expected one is still treated as a
            # failure case.


            # Split out the date string into tokens
            my @date_tokens = split(/ +/, $date);

            # These are the units that the date string will have.
            my ($seconds, $minutes, $hours, $days, $months, $years) = 0;

            # It's easier to assign numerical values to the months for
            # comparison.
            my %month_vals = ('JAN' => 1, 'FEB' => 2, 'MAR' => 3, 'APR' => 4, 'MAY' => 5, 'JUN' => 6,
                              'JUL' => 7, 'AUG' => 8, 'SEP' => 9, 'OCT' => 10, 'NOV' => 11, 'DEC' => 12);

            # Eventually at the end we'll want to re-map from month values to
            # strings for months again.
            my %reversed_months = reverse %month_vals;

            # For reference here's a date string and then the index into the
            # token array so that the code below is easier to follow.
            # Mon Feb  2  14:26:24 EST 2009
            # (0) (1) (2)  (3)     (4)  (5)
            #

            # Convert months into a numerical value for manipulation.
            $months = $month_vals{uc($date_tokens[1])};

            # Should be obvious based on sample above.
            $days = $date_tokens[2];
            $years = $date_tokens[5];

            # This teases out the values because they're all part of one token.
            $date_tokens[3] =~ m/(\d+):(\d+):(\d+)/;
            $hours = $1;
            $minutes = $2;
            $seconds = $3;


            # Ok so now we want to subtract 10 seconds from the seconds value,
            # this is easy unless seconds is under 10 seconds then we begin
            # to wrap and affect other units as we roll time backwards.

            # This is the easy case.
            if($seconds >= 10){$seconds = $seconds - 10;}

            # This is where seconds wraps but minutes doesn't since it's
            # greater than 1.
            elsif($minutes != 0){
                $minutes = $minutes - 1;
                $seconds = 60 - (10 - $seconds);
            }

            # This is where seconds and minutes wrap but hours does not
            # since it's greater than 0
            elsif($hours != 0){
                $hours = $hours - 1;
                # if minutes wraps it'll always be the 59th minute of the
                # previous hour.
                $minutes = 59;
                $seconds = 60 - (10 - $seconds);
            }
            # This is where seconds, minutes, and hours wrap but days does not
            # because it's greater than 1.
            elsif($days != 1){

                $days = $days - 1;
                # If days wrap the hours will always be the 23rd hour of the
                # previous day.
                $hours = 23;

                # if minutes wraps it'll always be the 59th minute of the
                # previous hour.
                $minutes = 59;
                $seconds = 60 - (10 - $seconds);
            }
            # This is where seconds, minutes, hours, days and months rollback
            # but not years since months was greater than 1 (Jan).  This does
            # get tricky because the last day of the preceding month isn't
            # always the same value, in some cases its 31 or 30 or 28, or
            # even 29 in a leap year.
            elsif($months != 1){
                $months = $months -1;

                # Ok so let's rollback the days looking at the month value
                # to determine what the days should be set as.
                # Feb, leap year
                if($months == 2 && $years % 4 == 0){$days = 29;}
                # Feb, non-leap year
                elsif($months == 2){$days = 28;}
                # Months that have 30 days.
                elsif($months == 4 || $months == 6 || $months == 9 || $months == 11){$days = 30;}
                # All other months have 31.
                else{$days = 31;}
                $hours = 23;
                $minutes = 59;
                $seconds = 60 - (10 - $seconds);
            }
            # The last case is where everything wraps including the year
            # so basically its Dec 31 23:59: and some seconds!
            else{
                $years = $years - 1;
                $months = 12;
                $days = 31;
                $hours = 23;
                $minutes = 59;
                $seconds = 60 - (10 - $seconds);
            }

            # For formatting these units are padded with a leading zero if
            # they are single digits.
            if($seconds < 10){$seconds = "0" . $seconds;}
            if($minutes < 10){$minutes = "0" . $minutes;}
            if($hours < 10){$hours = "0" . $hours;}

            # Build the new date string to look like the old and recompare
            my $new_date = $date_tokens[0] . " " . $reversed_months{$months} . " " . $days . " " . $hours . ":" . $minutes . ":" . $seconds . " " . $date_tokens[4] . " " . $years;

            # Recompare the dates anything but a ret_val of 2 is a failure.
            # If it's a 2 that means by subracting 10 seconds from date
            # new_date was now "earlier" than the last balanced date and thus
            # within our tolerance.
            $ret_val = VSeries::utils::compare_dates($new_date, $last_balanced_date);

            if($ret_val != 2){
         
               $self->{error} = $ERR_STATUS;
               $self->{errmsg} = "Error: filer::storage_load_balance: Unable to verify storage load balance completion.  Filer date: $date Last Load Balance Date: $last_balanced_date";
               return $ERR_STATUS;
            }
        }                
    }


    return($GOOD_STATUS);
}
#--------------------------------------------------------------------------------
## @name ::online_vols(;%)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Walks a filer obj internal vol list onlining them. (@{$self->{vol_list}})
## Will return an error after 3 tries to online a given volume.  Waits 10
## second between retries.
##
## @arg %ops Optional: A hash of options that can be passed into this method.
##                     Current supported options are:
##                     'f' which would force the volume online.
##
## @synopsis
## $status = $filer_obj->online_vols();
##
sub online_vols(;%) {
    my $self  = shift;      # The filer object to shift in.
    my $ops_ref = shift;  # This lives here for any future expansion.
    my $cmd = "";
    my $status = "";
    my @output_array = ();
    my %status_hash = ();
    my $counter = 0;

    my %ops =  %{$ops_ref} if (defined $ops_ref);    

    # ontap usage:
    # vol online {<vol-name> [-f] | <plex-name>}
    # - bring volume <vol-name> or plex <plex-name> online

    foreach my $vol (@{$self->{vol_list}}) {
        do{
            # In the event that the first time through here the online failed,
            # delay for 10 seconds and retry in case it was a transient
            # condition
            if($counter > 0){
                sleep(10);
            }
            $cmd = "vol online " . $vol;

            # Check if the user specified the force option
            if (defined $ops{'f'}) {$cmd .= " -f";}

            # Send the command.
            $status = Tharn::say($self->{connection}, $cmd, "-wait", $vol, "-ok");
            
            # Call vol_status, then parse the output.
            @output_array = @{$self->vol_status()};
            %status_hash = %{VSeries::utils::parse_vol_status("", @output_array)};
            $counter++;
            
        }while($status_hash{$vol}{state} ne "online" && $counter <= 3);
            
        if($counter > 3){
            $self->{error} = $ERR_STATUS;
            $self->{errmsg} = "ERROR: Unable to online volume $vol after multiple tries";
            return ($ERR_STATUS);
        }
    }
    return $GOOD_STATUS;
}    
#--------------------------------------------------------------------------------
## @name ::online_aggrs(;%)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Will attempt to online the aggr held in $self->{aggr}.
## Also handles options via the %ops hash which can be used
## to support future enhancements.
##
## @arg %ops Optional: A hash of options that can be passed into this method.
##                     Current supported options are:
##                     'f' which would force the aggr online.
##
## @synopsis
## $status = $filer_obj->online_aggrs();
##
sub online_aggrs(;%) {
    my $self  = shift;      # The filer object to shift in.
    my $ops_ref = shift;  # This lives here for any future expansion.
    my $cmd = "";
    my $status = "";
    my @output_array = ();
    my %status_hash = ();
    my $counter = 0;
    
    my %ops =  %{$ops_ref} if (defined $ops_ref);

    # ontap usage:
    # aggr online {<atv-name> [-f] | <plex-name>}
    #   - bring aggregate or traditional volume <atv-name> or
    #     plex <plex-name> online


    do{
        # In the event that the first time through here the online failed,
        # delay for 10 seconds and retry in case it was a transient
        # condition
        if($counter > 0){
            sleep(10);
        }
        $cmd = "aggr online " . $self->{aggr};
            
        # Check if the user specified the force option
        if (defined $ops{'f'}) {$cmd .= " -f";}
            
        # Send the command.
        $status = Tharn::say($self->{connection}, $cmd, "-wait", $self->{aggr}, "-ok");
            
        # Call aggr_status, then parse the output.
        @output_array = @{$self->aggr_status()};
        %status_hash = %{VSeries::utils::parse_aggr_status("", @output_array)};
        $counter++;
            
    }while($status_hash{$self->{aggr}}{state} ne "online" && $counter <= 3);
          
    if($counter > 3){
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} = "ERROR: Unable to online aggregate " . $self->{aggr} . " after multiple tries";
        return ($ERR_STATUS);
    }
    return $GOOD_STATUS;
}    
#--------------------------------------------------------------------------------
## @name ::offline_aggrs(;%)
##
## @author James Silva (silva@netapp.com)
##
## @description
## Will attempt to offline the aggr held in $self->{aggr}.
## Also handles options via the %ops hash which can be used
## to support future enhancements.
## I suggest using the -a and -f flags to offline aggrs.
##
## @arg %ops Optional: A hash of options that can be passed into this method.
##                     Current supported options are:
##                     'a' Usually used in combo with...
##                     'f' f, which will force the aggr offline.
##                     't' where you'd enter your cifs delay.
##
## @synopsis
## $status = $filer_obj->offline_aggrs(;%);
##
sub offline_aggrs(;%) {
    my $self  = shift;      # The filer object to shift in.
    my $ops_ref = shift;  # This lives here for any future expansion.
    my $cmd = "";
    my $status = "";
    my @output_array = ();
    my %status_hash = ();
    my $counter = 0;

    my %ops =  %{$ops_ref} if (defined $ops_ref);    
    
    # ontap usage:
    # aggr offline {<atv-name> | <plex-name>} [-t cifsdelay]
    #   - take aggregate or traditional volume <atv-name> or
    #     plex <plex-name> offline

    do{
        # In the event that the first time through here the offine failed,
        # delay for 10 seconds and retry in case it was a transient
        # condition
        if($counter > 0){
            sleep(10);
        }
        $cmd = "priv set -q diag;aggr offline " . $self->{aggr};
            
        # Check if the user specified any additional options.
        if (defined $ops{'f'}) {$cmd .= " -f";}
        if (defined $ops{'a'}) {$cmd .= " -a";}
        if (defined $ops{'t'}) {$cmd .= " -t " . $ops{'t'};}
            
        # Send the command.
        $status = Tharn::say($self->{connection}, $cmd, "-wait", $self->{aggr}, "-ok");
            
        # Call aggr_status, then parse the output.
        @output_array = @{$self->aggr_status()};
        %status_hash = %{VSeries::utils::parse_aggr_status("", @output_array)};
        $counter++;
            
    }while($status_hash{$self->{aggr}}{state} ne "offline" && $counter <= 3);
          
    if($counter > 3){
        $self->{error} = $ERR_STATUS;
        $self->{errmsg} .= "ERROR: Unable to offline aggregate $self->{aggr} after multiple tries";
        return ($ERR_STATUS);
    }
    return $GOOD_STATUS;    
}
#--------------------------------------------------------------------------------
## @name ::get_storage_show_disk($;$)
##
## @author James Silva (silva@netapp.com)
##
## @description
## For a given OWNED disk we will return a hash which contains dynamic keys.
## These keys are generated based the output of storage show disk cmd.
## Any keys which contain spaces will be transformed to 'spaceless' keys.
## It is up to the caller to verify that the data returned is good, otherwise
## fail at the .thpl layer.  Note if a line key exists and there is no value
## for it like "Pri Port" below, we'll place a "-" in that key's value.
## 
## @arg $disk Required: Owned disk you are requesting info for.
## @arg $options Optional: options you'd like to prepend to this cmd i.e "-a".
##
## @synopsis
## %ssd_hash = %{$filer_obj->get_storage_show_disk($disk, "-a")};
##
sub get_storage_show_disk() {
    my $self = shift; 
    my ($disk, $options) = @_;
    my $cmd = "storage show disk ";
    $cmd .= $options if (defined $options);
    $cmd .= " " . $disk;
    
    # Send cmd via Tharn::say() to the filer.
    my $after = (Tharn::say($self->{connection}, "$cmd", -swallow, '-timeout', 30))[2];
    
    # $after is now this (minus filer prompts):
    #vra-v3050-2b> storage show disk -a vrs-br200e-3:15.126L0
    #Disk:             vrs-br200e-3:15.126L0
    #Shelf:            -  
    #Bay:              - 
    #Serial:           600E000000A800000066700700000000
    #Vendor:           FUJITSU 
    #Model:            E3000           
    #Rev:              0000
    #RPM:              N/A
    #WWN:              2:000:00e000:a81377
    #UID:              600E0000:00A80000:00667007:00000000:00000000:00000000:00000000:00000000:00000000:00000000
    #Downrev:          no
    #Pri Port:          
    #Sec Name:         vrs-br200e-4:15.126L0   
    #Sec Port:          
    #Power-on Hours:   N/A
    #Blocks read:      0
    #Blocks written:   0
    #Time interval:    00:00:00
    #Glist count:      0
    #Scrub last done:  00:00:00
    #Scrub count:      0
    #LIP count:        0
    #Dynamically qualified:  No
    #vra-v3050-2b> 
    
    # Split the output on end lines:
    my @output = split(/\n/, $after);
    
    # Init the return hash:
    my %return_hash = ();
    
    # Build our dynamic return hash now.
    foreach my $line (@output) {
        # Strip leading whitespace (if any exists).
        $line =~ s/^\s+//;
        my ($line_key, $line_value) = undef;
        ($line_key, $line_value) = split(/:\s+/, $line);
        # Paranoia, Paranoia, Paranoia comin' to get me :)
        if (!defined $line_key) {next;}
        # Strip any spaces or colons in the key:
        $line_key =~ s/[\s+, :]//g;
        # Populate this $line_key with the $line_value
        if (defined $line_value && $line_value ne "") { $return_hash{$line_key} = $line_value; }
        else { $return_hash{$line_key} = "-"; }
    }
    return(\%return_hash);
}

#-------------------------------------------------------------------------------
## @name ::check_messages($$;$$)
##
## @author Mark LaBonte (labonte@netapp.com)
##
## @description
## Reads in a file (passed in as the first argument) and looks for instances
## of one or more messages (passed in as the second argument). A hash reference
## containing the messages and the total count for each is returned. This routine
## does not validate format of date/time passed to it.
##
## @arg $file          Required: Path to the file to be searched.
## @arg $messages_ref  Required: A pointer to an array of messages to look for.
## @arg $start_time    Optional: Start time in Unix 'date' format "Tue Jun  7 06:40:14 EDT 2011"
## @arg $end_time      Optional: End time in Unix 'date' format "Tue Jun  7 06:40:14 EDT 2011"
##
## @synopsis
## $messages_counts_ref = $filer_obj->check_messages($file, \@messages, $start_time, $end_time);

sub check_messages ($$;$$) {
    # Local variables declaration
    my ($self, $file, $messages_ref, $start_time, $end_time) = @_;
    my %messages_to_find = ();
    my @messages = @$messages_ref;
    my $time_zone_defined = 0;
    my $epoch_start_time = 0;
    my $epoch_end_time = 0;
    # Extract year as messages don't indicate year
    $start_time =~ /(\d\d\d\d)$/;
    my $year = $1;
    
    # Create a hash to keep a count of specific message in messages file
    foreach my $message (@messages) {
        $messages_to_find{$message} = 0;
    }

    if ( defined ($start_time) && defined ($end_time) ) {
        $time_zone_defined = 1;

        # Convert start_time to seconds since epoch
        $epoch_start_time = VSeries::utils::unix_time_to_epoch($start_time);
        # Convert end_time to seconds since epoch
        $epoch_end_time = VSeries::utils::unix_time_to_epoch($end_time);
    }

    my $output = $self->rdfile($file);
    # Split on end lines.
    my @output = split(/\n/, $output);

    foreach my $line (@output) {
        # Make sure that we parse only those line that begin with date
        if ( $time_zone_defined == 1 && $line =~ /^Sun |^Mon |^Tue |^Wed |^Thu |^Fri |^Sat /) {
            # Get time in message line of format 'Tue Jun  7 06:40:14 EDT 2011'
            $line =~ /(^[A-Z][a-z][a-z] [A-Z][a-z][a-z]  ?\d\d? \d\d:\d\d:\d\d [A-Z][A-Z][A-Z]) .*/;
            my $msg_time = $1 . " $year";
            
            # Convert time in messages to seconds since epoch
            my $msg_time_epoch = VSeries::utils::unix_time_to_epoch($msg_time);
            if ( $epoch_start_time <= $msg_time_epoch && $msg_time_epoch <= $epoch_end_time ) {
                foreach my $message (keys %messages_to_find) {
                    if ($line =~ /\Q$message\E/) {
                        $messages_to_find{$message}++;
                    }
                }
            }
        } else {
            # If start_time and end_time are not defined we scan through
            # all lines.
            foreach my $message (keys %messages_to_find) {
                if ($line =~ /\Q$message\E/) {
                    $messages_to_find{$message}++;
                }
            }
        }
    }
    return \%messages_to_find;
}

########################################################################
#
# @name ::get_disk_serial($)
#
# @author: Pranav Ojha (pranavo@netapp.com)
#
# @description:
# This routine returns disk serial number for a disk. If a disk serial is not
# found, the routine returns 'undef'.
#
# @args: $disk_name (Required) Disk name for which the serial no. is required.
#
# @usage: $hash_ref = $filer_obj->get_disk_serial($disk);
#
########################################################################
sub get_disk_serial($){
  my $self = shift;
  my $disk_name = shift;
  my $disk_hash_ref = $self->{disks};
  my %disk_hash = %$disk_hash_ref;

  foreach my $disk_serial( keys %disk_hash ){
      my $disk_ref = $disk_hash{$disk_serial};
      my %disk = %$disk_ref;
      return $disk_serial if( $disk{'primary'} eq $disk_name );
  }

  # If we are here, then we didn't find the required disk
  return undef;
}

########################################################################
#
# @name ::wait_for_zeroing_to_complete()
#
# @author: Pranav Ojha (pranavo@netapp.com)
#
# @description:
# This routine waits for disk zeroing to complete. By default the wait
# is indefinite.
#
# @args:      $time (Optional): Disk zero timeout in seconds.
#
# @usage: $filer_obj->wait_for_zeroing_to_complete();
#
########################################################################
sub wait_for_zeroing_to_complete(;$){
    my $self = shift;
    my $wait_time = shift;
    my $zeroing_complete = 0;
    if (defined $wait_time){
        my $end_time = localtime() + $wait_time;
        while( localtime() < $end_time ){
            my (undef, undef, $output) = $self->send_filer_cmd("aggr status -s");
            if( $output !~ /zeroing,/ && $output !~ /\(not zeroed\)/ ){
                $zeroing_complete = 1;
                last;
            }
            sleep 600;
        }
        if( $zeroing_complete == 1 ){
            return $GOOD_STATUS;
        } else {
            return $ERR_STATUS;
        }
    }

    while(1){
        my (undef, undef, $output) = $self->send_filer_cmd("aggr status -s");
        last if( $output !~ /zeroing,/ && $output !~ /\(not zeroed\)/ );
        sleep 600;
    }
    return $GOOD_STATUS;
}

########################################################################
#
# @name ::wait_aggr_grow()
#
# @author: Pranav Ojha (pranavo@netapp.com)
#
# @description:
# This routine waits indefinitely for aggr to grow completely.
#
# @arg      $aggr_name (Required): Name of the aggregate under test.
#
# @usage: $filer_obj->wait_aggr_grow($aggr_name);
#
########################################################################
sub wait_aggr_grow($){
    my $self = shift;
    my $aggr_name = shift;

    while(1){
        my (undef, undef, $output) = $self->send_filer_cmd("aggr status $aggr_name");
        last if( $output !~ /growing/ );
        sleep 600;
    }
    return $GOOD_STATUS;
}

########################################################################
#
# @name ::wait_aggr_initialize($)
#
# @author: Pranav Ojha (pranavo@netapp.com)
#
# @description:
# This routine waits indefinitely for aggr to initialize completely.
#
# @arg      $aggr_name (Required): Name of the aggregate under test.
#
# @usage: $filer_obj->wait_aggr_initialize($aggr_name);
#
########################################################################
sub wait_aggr_initialize($){
    my $self = shift;
    my $aggr_name = shift;

    while(1){
        my (undef, undef, $output) = $self->send_filer_cmd("aggr status $aggr_name");
        last if( $output !~ /initializing/i );
        sleep 600;
    }
    return $GOOD_STATUS;
}


#--------------------------------------------------------------------------------
1;  # Needed for *.pm's

