# -*- perl -*-
# $Id$
#
## @summary switch object creater - basically creates an OO model of a switch
##   and has several methods that do cool switch things like enable zones
##   and disable zones.
##
## @description
##   This *.pm is meant to be something that creates switch objects (OO style)
##   and also holds other useful methods that help create zones on switches
##   or gather the active zone configs on a switch etc.  It is meant to be
##   something that will grow over time and will hold switch specific functions
##   created to support automated tests which will need to execute built in
##   switch functionality.
##
## @usage
##   use VSeries::switch_object;
##   use VSeries::switch_object qw(...);
##
## @keywords  switch_object test library
##
## @dependencies
##   use strict;
##   use Tharn;
##   use Expect;
##   use config_base;
##   use VSeries::filer;
##
## @author James Silva (silva@netapp.com )
##
## @status testme
##
##

package VSeries::switch_object;

use strict;
use Tharn;
use Expect;
use config_base;
use VSeries::filer;                      # For filer objects & methods
use VSeries::switch;                     # For switch objects & methods
use VSeries::utils;                      # For compare_arrays(@@);
use Data::Dumper;						 # Used for debug.

use vars ('@ISA','@EXPORT');
require Exporter;
our @ISA = ('Exporter');
@EXPORT = ('new',                        # Create a switch object
           'set_internal_variables',     # Set an objects data members
           'zone_config_enable',         # Enable a zone config
           'zone_config_disable',        # Disable a zone config
           'active_config_get',          # Get the active zone config
           'active_config_and_zone_get', # Get the active zone config & zones
           'zone_config_create_enable',  # Create a zone config & enable
           'zone_config_delete',         # Delete a zone config & its zones
           'zone_with_port_find',        # Get zone which contains some port
           'zone_save_config',           # Save zone config changes on a switch
           'zone_exists',                # Check if a zone exists
           'zone_config_exists',         # Check if a zone config exists
           'zone_add',                   # Add a zone to the existing config
           'zone_remove',                # Remove a zone from the existing config
           'cisco_go_to_normal_prompt',  # Bring cisco to normal prompt
           'get_error_state',            # Find if there is an error present
           'get_error_msg',              # Get an array of cascading errors
           'get_name',                   # Returns the name of the switch
           'get_host_type',              # Returns the hosttype of the switch
           'clear_error',                # Clears $error and @errmsg
           'set_switch_parameter',       # Set a parameter on a switch
           'so_switch_eport_get',        # Get switch interconnect port info
           'so_switch_disable',          # Disable entire switch, all ports
           'so_switch_enable',           # Enable entire switch, all ports
           'get_port_states',            # Get current state of all switch ports
           'set_switch_name',            # Set the hostname of the switch
           'get_switch_hw_sw_info',      # get hardware and software info
           'switch_reboot',              # reboot a switch
           'switch_fastboot',            # use the fastboot feature for Brocade
           'switch_connect',             # create a new connection to the switch
           'switch_disconnect',          # disconnect from a given switch
           'switch_port_disable',        # disable a given port on a switch
           'switch_port_enable',         # enable a given port on a switch
           'mcdata_clear_pending',       # Clears the pending zoneset for McData
           'get_seen_wwns',              # returns all wwns connected to switch
           'set_port_speed',             # sets the speed of a port on a switch
           'get_port_wwn_seen_on',       # get the port a wwn is connected to
           'get_port_fabric_id',         # gets the fabric id of a given port.
           'set_autologout_delay',       # sets the autologout delay in switch.
           'send_switch_cmd',            # miscellaneous method to execute whatever
                                         # command is passed in.
           );

## GLOBALS
my $GOOD_STATUS    = 0;
my $ERR_STATUS     = 1;
my $CISCO_VSAN_NUM = 1;
my $WWN_REG_EX = '(\w+:){7}(\w){2}';
# Sleep interval between switch port status check retries
my $PORT_STATUS_RETRY_WAIT = 10;
# Max number of retries (5 min total wait time for port to change state)
my $MAX_PORT_STATUS_RETRIES = 30;


## @name ::new()
##
## @description
## Initializes a new switch data structure and its data members.  It takes in
## a hash of arguments and if defined looks for a hostname for the given switch
## and will use that hostname to call _determine_switch_type(<hostname>) and
## take the returned value which are class types of switch_object that inherit
## and override functionality where it's necessary for vendor specific behavior
##
## @synopsis
## $switch_obj=VSeries::switch_object->new($args_ref);
##
## @arg $args_ref Optional: reference to a hash with key value pairs of options
##                          for the constructor.  Currently supported is the
##                          hostname key/value pair to dynamically determine
##                          switch vendor type and class for the switch object.
##
##
sub new (;$)
{
    my ($class, $args_ref) = @_;
    my ($self);

    # will be used to handle args_ref if specified and necessary.
    my %args = ();

    # create a new object only if necessary
    if (defined $args{-self}) {
        $self=$args{-self};
    } else  {
        $self={};

        # So here we've blessed self into VSeries::switch_object
        bless $self,$class;
    }

    # Now lets see if we were given args
    if(defined $args_ref){

        # we were so lets cast that to a hash and pull out hostname.
        %args = %$args_ref;

        # Ok lets use that hostname to determine the vendor type.
        my $returned_type = $self->_determine_switch_type($args{'hostname'});

        # If the returned type is Brocade Director then we re-bless the
        # object into the child class thus overriding necessary functionality
        # for that type of switch.
        if($returned_type =~ m/VSeries::Brocade_Director/){
            bless($self, $returned_type);
        }

    }

    # from Tharn::host();
    $self->{default_ip} = '';          # ip address of switch
    $self->{hostname} = '';            # dns name of switch
    $self->{hosttype} = '';            # i.e. switch-brocade, or switch-*
    $self->{id} = '';                  # a unique identifier
    $self->{irptchar} = '';            # interrupt charachter i.e. ctrl-c
    $self->{mac_address} = '';         # mac (physical) address of switch
    @{$self->{prompt_norm}} = ();      # list of normal prompts

    # from Tharn::connect();
    $self->{conntype} = '';            # type of conn i.e. rsh, telnet, etc.
    $self->{connport} = '';            # port num for telnet & serial else undef
    $self->{username} = '';            # username used for login
    $self->{password} = '';            # password used for login
    $self->{timeout} = '';             # timeout found in objects tharnhost file
    $self->{max_reconnect} = '';       # retry say() commands x times
    $self->{conn} = '';                # the full connectrec spec

    # error tracking
    $self->{error} = 0;                # error flag used as a bool
    @{$self->{errmsg}} = ();           # an array that stores cascading errors

    # used for topology
    @{$self->{filer_list}} = ();

    return($self);
}

## @name ::set_internal_variables()
##
## @description
## Parameters are all optional.  Defined as such in order to introduce readability
## and flexibility moving forward (i.e. you should be able to call this from
## anywhere with parameters missing and/or in the wrong order.  Errors will occur
## if a parameter is not labeled correctly.
##
## @synopsis
## $switch_obj=VSeries::switch_object->set_internal_variables(\%args);
##
## @arg %args Required: The actual key value pairs that you wish to change
## in the $switch_obj.
##
sub set_internal_variables(\%)
{
    # pick up our object
    my $self = shift;

    # catch the optional function parameters here
    my %args = @_;

    # parse function parameters that are present, putting into data members
    # could just say "$self->{$arg} = $val", but no check
    while (my ($arg, $val) = each %args)
    {
        if ($arg eq "hostname")
        {
            $self->{hostname} = $val;
        }
        elsif ($arg eq "default_ip")
        {
            $self->{filer_index} = $val;
        }
        elsif ($arg eq "hosttype")
        {
            $self->{hosttype} = $val;
        }
        elsif ($arg eq "id")
        {
            $self->{id} = $val;
        }
        elsif ($arg eq "irptchar")
        {
            $self->{irptchar} = $val;
        }
        elsif ($arg eq "mac_address")
        {
            $self->{mac_address} = $val;
        }
        elsif ($arg eq "prompt_norm")
        {
            @{$self->{prompt_norm}} = split(/,/,$val);
        }
        elsif ($arg eq "conntype")
        {
            $self->{conntype} = $val;
        }
        elsif ($arg eq "conn")
        {
            $self->{conn} = $val;
        }
        elsif ($arg eq "connport")
        {
            $self->{connport} = $val;
        }
        elsif ($arg eq "username")
        {
            $self->{username} = $val;
        }
        elsif ($arg eq "password")
        {
            $self->{password} = $val;
        }
        elsif ($arg eq "timeout")
        {
            $self->{timeout} = $val;
        }
        elsif ($arg eq "max_reconnect")
        {
            $self->{max_reconnect} = $val;
        }
        elsif ($arg eq "filer_list")
        {
            @{$self->{filer_list}} = split(/,/, $val);
        } else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "Invalid parameter: $arg => $val");
        }
    }
}

## @name ::zone_config_enable()
## @description
## Enables the zone configuration $cfg_name
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## $status = $switch_obj->zone_config_enable($cfg_name)
##
## @arg $cfg_name Required: A zone config name
##
sub zone_config_enable($)
{
    my ($self, $cfg_name) = @_;
    my $confirm_prompt = '^.*\(yes, y, no, n\): \[no\]\s*\n*';
    my ($match, $after, $prompt, $cmd) = '';
    my $hostname = $self->get_name();

    if ($self->{hosttype} =~ /switch-brocade/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # Enable the zone config $cfg_name
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "cfgEnable \"$cfg_name\"", '-prompt', $confirm_prompt, -swallow, '-timeout', 60);

        # - - - - - - -  - - - - - - - - - - - - - -
        # If this is a b200e switch (v5 firmware), we need to confirm stuff with "y"
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($prompt =~ /$confirm_prompt/) {
            ($match, undef, undef) = Tharn::say($self->{conn}, "y", "-wait", '^.*zone config.*is in effect', -swallow, '-timeout', 60);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # If it wasn't a v5.2.1 firmware we make sure $after contained thee following:
        # - - - - - - -  - - - - - - - - - - - - - -

        elsif ($after !~ /cfgEnable successfully completed/i) {
            return($ERR_STATUS);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Save the config just incase.
        # - - - - - - -  - - - - - - - - - - - - - -

        $self->zone_save_config();

        if ($self->get_error_state()) {
            return($ERR_STATUS);
        }
        else {
            return($GOOD_STATUS);
        }
    } # end ($self->{hosttype} =~ /switch-brocade/i) {
    elsif ($self->{hosttype} =~ /switch-cisco/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # In the case of Cisco you are 'activating a zoneset'
        #
        # The following code will automate the sequence below:
        #
        #vns-c9124-2# config
        #Enter configuration commands, one per line.  End with CNTL/Z.
        #vns-c9124-2(config)# zoneset activate name auto_config vsan 1
        #Zoneset activation initiated. check zone status
        #vns-c9124-2(config)# end
        #vns-c9124-2#
        #
        # Note we will only enable zonesets on vsan 1.
        # First get the switch into 'config' mode.
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

        # - - - - - - -  - - - - - - - - - - - - - -
        # Activate the effective 'zoneset' similar to brocades zone config
        # - - - - - - -  - - - - - - - - - - - - - -

        $cmd = "zoneset activate name " . $cfg_name . " vsan " . $CISCO_VSAN_NUM . "";

        #($match, undef, undef, $prompt) = Tharn::say($self->{conn}, "$cmd", "-wait", "Zoneset activation initiated",
        #                                                            '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);
        my $zoneset_prompts = "(Zoneset activation initiated|You are trying to activate zoneset|Do you want to continue|$hostname.*\#)";
                                                                    
        ($match, undef, undef, $prompt) = Tharn::say($self->{conn}, "$cmd", -swallow, ,'-noprompt', '-prompt', $zoneset_prompts, '-timeout', 30);

        if($prompt =~ m/$zoneset_prompts/){
            Tharn::say($self->{conn}, "y", '-noprompt', '-prompt', "$hostname.*\#", '-timeout', 30);
        }                                                            

        # - - - - - - -  - - - - - - - - - - - - - -
        # Bring the switch back to its normal mode
        # - - - - - - -  - - - - - - - - - - - - - -

        $self->cisco_go_to_normal_prompt();

        # - - - - - - -  - - - - - - - - - - - - - -
        # Make sure we activated the zoneset $cfg_name.
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($self->active_config_get() ne $cfg_name) {

            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: Unable to activate $cfg_name on switch " . $self->get_name() . "\.");
            return ($ERR_STATUS);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # If we got this return good status.
        # - - - - - - -  - - - - - - - - - - - - - -

        return ($GOOD_STATUS);

    } # end ($self->{hosttype} =~ /switch-cisco/i) {
    elsif ($self->{hosttype} =~ /switch-mcdata/i) {
        # - - - - - - -  - - - - - - - - - - - - - -
        # On mcdata it is only possible to enable a zoneset if
        # it is currently in the 'working' or 'pending' area.
        #
        # In the positive case we are returned the normal prompt:
        #
        #Root> config zoning activateZoneSet
        #Root>
        #
        # In the negative case there isn't a zoneset in the working area.
        #
        #Root> config zoning activatezoneset
        #Error 213: Number of Zones is Zero
        #Root>
        #
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after) = Tharn::say($self->{conn}, "config zoning activateZoneSet", -swallow, '-timeout', 30);

        # - - - - - - -  - - - - - - - - - - - - - -
        # Glance at $after to make sure we didn't have an error.
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($after =~ /Error \d+:/i){
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: Unable to activate $cfg_name on switch " . $self->get_name() . "\.");
            return ($ERR_STATUS);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Make sure we activated the zoneset $cfg_name.
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($self->active_config_get() ne $cfg_name) {

            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: Unable to activate $cfg_name on switch " . $self->get_name() . "\.");
            return ($ERR_STATUS);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # If we got this return good status.
        # - - - - - - -  - - - - - - - - - - - - - -

        return ($GOOD_STATUS);

    } # end ($self->{hosttype} =~ /switch-mcdata/i
    else {
        $self->{error} = $ERR_STATUS;
        push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
        return ($ERR_STATUS);
    }
}

## @name ::zone_config_disable()
##
## @description
## Disables the active zone configuration.  Returns $GOOD_STATUS for success
## and returns $ERR_STATUS on failure (to disable the zone).
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## $status = $switch_obj->zone_config_disable()
##
sub zone_config_disable()
{
    my $self = shift;
    my ($match, $after, $prompt) = '';
    my $confirm_prompt = '^.*\(yes, y, no, n\): \[no\]\s*\n*';
    my $hostname = $self->get_name();

    if ($self->{hosttype} =~ /switch-brocade/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # Disable the active zone config
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "cfgDisable", -swallow, '-prompt', $confirm_prompt, '-timeout', 30);

        # - - - - - - -  - - - - - - - - - - - - - -
        # If we're dealing with a b200e (v5 firmware) we need to confirm with 'y'
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($prompt =~ /$confirm_prompt/) {
            Tharn::say($self->{conn}, "y", -swallow, '-timeout', 30);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # If we're not dealing with b200e's (v5 firmware) then do the following:
        # - - - - - - -  - - - - - - - - - - - - - -

        elsif ($after =~ /cfgDisable successfully completed/i ||
               $after =~ /No effective configuration found/i) {
               # do nothing b/c we're good.
        }
        else {
            # otherwise return ERROR STATUS
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: Unable to disable the active zoneset on $hostname.");
            return($ERR_STATUS);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Save the config just incase.
        # - - - - - - -  - - - - - - - - - - - - - -

        $self->zone_save_config();

        if ($self->get_error_state()) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: Unable to save the zone config on $hostname.");
            return($ERR_STATUS);
        }
        else {
            return($GOOD_STATUS);
        }
    } # end ($self->{hosttype} =~ /switch-brocade/i)
    elsif ($self->{hosttype} =~ /switch-cisco/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # The following code will automate the sequence below:
        #
        #vns-c9124-2# config t
        #Enter configuration commands, one per line.  End with CNTL/Z.
        #vns-c9124-2(config)# no zoneset activate vsan 1
        #Zoneset deactivation initiated. check zone status
        #vns-c9124-2(config)# end
        #vns-c9124-2#
        # - - - - - - -  - - - - - - - - - - - - - -

        # - - - - - - -  - - - - - - - - - - - - - -
        # Note we assume that this zone config resides on vsan 1.
        # First get the switch into 'config' mode.
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

        # - - - - - - -  - - - - - - - - - - - - - -
        # Deactivate the effective 'zoneset' similar to brocades zone config
        # - - - - - - -  - - - - - - - - - - - - - -

        ($match, undef, undef, $prompt) = Tharn::say($self->{conn}, 'no zoneset activate vsan 1', '-wait', 'Zoneset deactivation initiated',
                                                                    '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);
        # - - - - - - -  - - - - - - - - - - - - - -
        # Bring the switch back to its normal mode
        # - - - - - - -  - - - - - - - - - - - - - -

        $self->cisco_go_to_normal_prompt();

        # - - - - - - -  - - - - - - - - - - - - - -
        # Make sure we disabled the active zoneset.
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($self->active_config_get() ne "no active config"){
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: Unable to disable the active zoneset on $hostname.");
            return($ERR_STATUS);
        }

        return ($GOOD_STATUS);

    }
    elsif ($self->{hosttype} =~ /switch-mcdata/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # The following code will automate the sequence below:
        #
        #Root> config zoning deactivateZoneSet
        #Root>
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, 'config zoning deactivateZoneSet', -swallow, '-timeout', 30);

        # - - - - - - -  - - - - - - - - - - - - - -
        # Make sure we disabled the active zoneset.
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($self->active_config_get() ne "no active config"){
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: Unable to disable the active zoneset on $hostname.");
            return($ERR_STATUS);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # If we got this far we successfully disabled the zoneset
        # - - - - - - -  - - - - - - - - - - - - - -

        return ($GOOD_STATUS);

    }
    else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
            return ($ERR_STATUS);
    }
}

## @name ::active_config_get()
##
## @description
## Returns the active zone configuration name in scalar context.  If no
## active config exists we return "no active config".
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## $config_name = $switch_obj->active_config_get()
##
sub active_config_get()
{
    my $self = shift;
    my ($after, $index, $line, $i, $config_name) = '';
    my @temp1 = ();
    my $noisy_switch = 1;   # used to see if the switch is buggy
    my $config_exists = 0;
    my $hostname = $self->get_name();
    my $no_config = "no active config";


    # - - - - - - -  - - - - - - - - - - - - - -
    # Determine what hosttype we're dealing with
    # i.e. Brocade, Cisco, McData.
    # - - - - - - -  - - - - - - - - - - - - - -

    if ($self->{hosttype} =~ /switch-brocade/i) {

        (undef, undef, $after) = Tharn::say($self->{conn}, "cfgActvShow", -swallow, '-timeout', 30);

        # - - - - - - -  - - - - - - - - - - - - - -
        # $after now contains the output of the brocade cli
        # command cfgActvShow.  A sample of this can be seen below.
        #
        #vss_b3800_7:admin> cfgActvShow
        #Effective configuration:
        # cfg:   Automation1
        # zone:  Auto1   1,0
        #                1,2
        # zone:  Auto2   1,1
        #                1,3
        # - - - - - - -  - - - - - - - - - - - - - -


        @temp1 = split(/\n/, $after);

        for ($index = 0; $index < @temp1; $index++) {
            $line = $temp1[$index];
            if ($line =~ /^\s*Effective configuration/) {
                $noisy_switch=0;
                for ($i = $index; $i < @temp1; $i++) {
                    $line = $temp1[$i];
                    if ($line =~ /^\s*cfg:\s+(\S+)\s*/) {
                        $config_name = $1;
                        $config_exists = 1;
                        last;
                    }
                }
            }
        } # for ($index = 0;

        if ($noisy_switch || !$config_exists) {
            push(@{$self->{errmsg}}, "WARN: Switch " . $self->get_name() . " console may be noisy check or there is no zone config currently active\.");
            return($no_config);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # If we got this far there is an actual config in place so return it.
        # - - - - - - -  - - - - - - - - - - - - - -

        return($config_name);

    } # end ($self->{hosttype} =~ /switch-brocade/i) {

    elsif ($self->{hosttype} =~ /switch-cisco/i) {
        # Fix for BURT 490772.
        # The output of show zoneset active  might be prompted by more. Pipe to no-more to avoid more. 

        (undef, undef, $after) = Tharn::say($self->{conn}, "show zoneset active | no-more", -swallow, '-timeout', 30);

        # - - - - - - -  - - - - - - - - - - - - - -
        # $after now contains the output of the cisco cli
        # command 'show zoneset active'.  In the positive case
        # we get something like this:
        #
        #vns-c9124-2# show zoneset active
        #zoneset name current vsan 1
        #  zone name Automation3 vsan 1
        #  * fcid 0x0e0300 [pwwn 50:0a:09:82:00:01:0b:0a]
        #  * fcid 0x0e0400 [pwwn 50:06:04:82:cb:1b:ce:2d]
        #vns-c9124-2#
        #
        # Or in the negative case (no zoneset enabled) we'll get
        # something like this:
        #
        #vns-c9124-2# show zoneset active
        #Zoneset not present
        #vns-c9124-2#
        # - - - - - - -  - - - - - - - - - - - - - -

        @temp1 = split(/\n/, $after);

        foreach $line (@temp1) {
            if ($line =~ /^.*Zoneset not present.*\n?$/){
                push(@{$self->{errmsg}}, "WARN: Switch" . $self->get_name() . " does not have an effective zone config in place\.");
                return($no_config);
            }
            elsif ($line =~ /^.*zoneset name (\S+) vsan \d+.*\n?$/) {
                $config_name = $1;
                $config_exists = 1;
                last;
            }
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Catch the case where there may be an effective zone config
        # but we just weren't able to get it.
        # - - - - - - -  - - - - - - - - - - - - - -

        if (!$config_exists) {
            push(@{$self->{errmsg}}, "WARN: Not able to get the effective zone config for switch " . $self->get_name() ."\.");
            return($no_config);
        }

    # - - - - - - -  - - - - - - - - - - - - - -
    # If we got this far there is an active config in place so return it.
    # - - - - - - -  - - - - - - - - - - - - - -

    return($config_name);

    } # end ($self->{hosttype} =~ /switch-cisco/i) {

    elsif ($self->{hosttype} =~ /switch-mcdata/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # The following sequence is automated below:
        #
        #Root> show zoning
        #Default Zone Enabled:    True
        #Zone Set:  Zone2
        #  Zone:  V3070_2A_ESS800
        #    Zone Member:  50:0A:09:82:00:00:76:31
        #    Zone Member:  50:05:07:63:00:C8:A1:79
        #  Zone:  v3050c
        #    Zone Member:  Domain  1, Port  8
        #    Zone Member:  Domain  1, Port  9
        #    Zone Member:  Domain  1, Port 10
        #    Zone Member:  Domain  1, Port  6
        #
        #Root>
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after) = Tharn::say($self->{conn}, "show zoning", -swallow, '-timeout', 30);

        @temp1 = split(/\n/, $after);

        foreach $line (@temp1) {
            if ($line =~ /^.*Zone Set:\s+(\S+)/){
                $config_name = $1;
                $config_exists = 1;
                last;
            }
            else { next; }


        } #foreach $line (@temp1)

        if (!$config_exists) {
            push(@{$self->{errmsg}}, "WARN: Not able to get the effective zone config for switch " . $self->get_name() ."\.");
            return($no_config);
        }
        else { return ($config_name); }

    } # end ($self->{hosttype} =~ /switch-mcdata/i) {

    else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
            return ($ERR_STATUS);
    }
}

## @name ::zone_config_create_enable()
##
## @description
## Creates and enables the zone config passed in hash form.  If the zone
## config was enabled we return $GOOD_STATUS otherwise we return $ERR_STATUS on error.
##
## --- A SAMPLE BROCADE CONFIG --------------------------------------
## 'config_name' -> ($ SCALAR) The name of your zone_config (c-style)
## 'zone_name1'  -> (@ ARRAY)  An array of ports/wwns.
## 'zone_nameN'  -> (@ ARRAY)  Same as above...
## 'zone_nameN+1'-> (@ ARRAY)  Same as above...
##
## --- A SAMPLE CISCO CONFIG --------------------------------------
## 'config_name' -> ($ SCALAR) The name of your zone_config (c-style)
## 'zone_name1'  -> (@ ARRAY)  An array of ports/wwns.
## 'zone_nameN'  -> (@ ARRAY)  Same as above...
## 'zone_nameN+1'-> (@ ARRAY)  Same as above...
##
## --- A SAMPLE MCDATA CONFIG --------------------------------------
## 'config_name' -> ($ SCALAR) The name of your zone_config (c-style)
## 'zone_name1'  -> (@ ARRAY)  An array of ports/wwns.
## 'zone_nameN'  -> (@ ARRAY)  Same as above...
## 'zone_nameN+1'-> (@ ARRAY)  Same as above...
##
##
## 'config_name' above is used for all zone configs but in cisco context,
## they call it a 'zoneset'
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## $status = $switch_obj->zone_config_create_enable(\%config)
##
## @arg \%config Required: A zone config, different format for each vendor
##
sub zone_config_create_enable(%)
{
    my ($self, $config_ref) = @_;
    my %config = %$config_ref;
    my ($cmd, $zone, $after, $prompt, $enabled_config, $zone_member, $match, $before) = '';
    my $confirm_prompt = '^.*\(yes, y, no, n\): \[no\]\s*\n*';
    my $unsupported_member = 0;
    my $hostname = $self->get_name();

    # - - - - - - -  - - - - - - - - - - - - - -
    # Determine what hosttype we're dealing with
    # i.e. Brocade, Cisco, McData.
    # - - - - - - -  - - - - - - - - - - - - - -

    if ($self->{hosttype} =~ /switch-brocade/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # For brocade switches the hash passed in should
        # have the following format key/value pairs:
        #
        # 'config_name' -> ($ SCALAR) The name of your zone_config (c-style)
        # 'zone_name1'  -> (@ ARRAY)  An array of ports/wwns.
        # 'zone_nameN'  -> (@ ARRAY)  Same as above...
        # 'zone_nameN+1'-> (@ ARRAY)  Same as above...
        #
        # Note*** wwns must be in this format 50:0a:09:80:00:01:43:88
        # Note*** ports must be in this format 1,1 where the first one
        # is a domain or blade number and the second 1 is the physical
        # port number.
        # Note*** depending on your ARRAYS of ports above you can create
        # your own hard and soft zones.  A hard zone would only consist of
        # domain/port pairs and a soft zone would have wwn's.
        # - - - - - - -  - - - - - - - - - - - - - -

        # - - - - - - -  - - - - - - - - - - - - - -
        # Format for creating a zone on brocade:
        #
        # zoneCreate "zoneName", "zoneMemberList"
        # - - - - - - -  - - - - - - - - - - - - - -

        foreach $zone (keys %config) {

            if ($zone eq 'config_name') {next;}

            $cmd = "zoneCreate \"" . $zone . "\", \"";

            foreach $zone_member (@{$config{$zone}}) {

                $cmd .= $zone_member . ";";
            }

            # need to strip the last semi off of the string.
            chop($cmd);

            $cmd .= "\"";

            # vss_b3800_7:admin> zonecreate "Auto1", "1,0;1,2;1,6"
            # error: "Auto1" duplicate name
            #
            # and if same zone and zone members, no response string.

            (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);

            if ($after =~ (/duplicate name/i ||
                           /Usage:/i         ||
                           /syntax error/i   ||
                           /^\s*error:/i     ||
                           /^\s*unmatched quote/i) ) {

                # error: "Auto1" duplicate name
                return($ERR_STATUS);
            }
        } # foreach my $zone

        # - - - - - - -  - - - - - - - - - - - - - -
        # Now we need create a zone config with $config{config_name}
        # - - - - - - -  - - - - - - - - - - - - - -

        # how to:
        # cfgCreate "cfgName", "cfgMemberList"

        # Clear $cmd for reuse.
        $cmd = "";

        $cmd = "cfgCreate \"$config{config_name}\", \"";

        #       cfgCreate "switch_zone_nate_test", "

        foreach $zone (keys %config) {

            if ($zone eq 'config_name') {next;}

            $cmd .= $zone . ";";
        }

        # need to strip the last semi off of the string.
        chop($cmd);

        $cmd .= "\"";

        (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);

        # error case below:
        # vss_b3800_7:admin> cfgcreate "switch_zone_nate_test", "Auto1;Auto2;Auto3"
        # error: "switch_zone_nate_test" duplicate name

        if ($after =~ (/duplicate name/i ||
                       /Usage:/i         ||
                       /syntax error/i   ||
                       /^\s*error:/i     ||
                       /^\s*unmatched quote/i) ) {


            # error: "Auto1" duplicate name
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: While creating $config{config_name} on " . $self->get_name() . "\.");
            return($ERR_STATUS);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Now enable the created zone config $config{config_name}
        # - - - - - - -  - - - - - - - - - - - - - -

        # how to:
        # cfgEnable cfgName

        # Clear $cmd for reuse.
        $cmd = "";

        $cmd = "cfgEnable \"$config{config_name}\"";

        my $match_string = '';
        (undef, undef, $match_string, $prompt) = Tharn::say($self->{conn}, "$cmd", -swallow, '-prompt', $confirm_prompt, '-timeout', 60);

        # - - - - - - -  - - - - - - - - - - - - - -
        # If this is a b200e switch, we need to confirm stuff with "y"
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($prompt =~ /$confirm_prompt/) {
            ($match_string, undef, undef) = Tharn::say($self->{conn}, "y", "-wait", "zone config", -swallow, '-timeout', 60);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # If its not a b200e then...
        #
        # WAIT FOR THIS:
        # cfgEnable successfully completed
        # Even when the same config is already enabled you get back
        # the message "cfgEnable successfully completed"
        # - - - - - - -  - - - - - - - - - - - - - -

        elsif ($match_string !~ /cfgEnable successfully completed/) {
            # error
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: '$zone_member' is an unsupported member type we only support wwns and interface zoning on " . $self->get_name() . "\.");
            return($ERR_STATUS);
        }

        # Check to see if the config is now really enabled:

        $enabled_config = $self->active_config_get();
        if ($enabled_config ne $config{config_name}) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: $config{config_name} was not successfully enabled on " . $self->get_name() . "\.");
            return($ERR_STATUS);
        }


        # - - - - - - -  - - - - - - - - - - - - - -
        # Save the config just incase.
        # - - - - - - -  - - - - - - - - - - - - - -

        $self->zone_save_config();

        if ($self->get_error_state()) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: While saving $config{config_name} zone configuration on " . $self->get_name() . "\.");
            return($ERR_STATUS);
        }
        else {
            return($GOOD_STATUS);
        }
    } # end if (self->{hosttype} =~ /switch-brocade/
    elsif ($self->{hosttype} =~ /switch-cisco/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # The following is the sequence of events that takes place
        # in this entire elsif block for switch-cisco:
        # - - - - - - -  - - - - - - - - - - - - - -
        #vns-c9124-2# config
        #Enter configuration commands, one per line.  End with CNTL/Z.
        #vns-c9124-2(config)# zone name jimmy1 vsan 1
        #vns-c9124-2(config-zone)# member fwwn 20:01:00:0d:ec:47:bb:00
        #vns-c9124-2(config-zone)# end
        #vns-c9124-2# config
        #Enter configuration commands, one per line.  End with CNTL/Z.
        #vns-c9124-2(config)# zoneset name jimmy_zoneset vsan 1
        #vns-c9124-2(config-zoneset)# member jimmy1
        #vns-c9124-2(config-zoneset)# end
        #vns-c9124-2# config
        #Enter configuration commands, one per line.  End with CNTL/Z.
        #vns-c9124-2(config)# zoneset activate name jimmy_zoneset vsan 1
        #Zoneset activation initiated. check zone status
        #vns-c9124-2(config-zone)# end
        #vns-c9124-2#
        # - - - - - - -  - - - - - - - - - - - - - -

        # - - - - - - -  - - - - - - - - - - - - - -
        # First check if the config or any zones already exist,
        # if they do we need to error out.
        # - - - - - - -  - - - - - - - - - - - - - -

        foreach $zone (keys %config) {

            if ($zone eq 'config_name'){
                if ($self->zone_config_exists($config{config_name}) == $GOOD_STATUS){
                    $self->{error} = $ERR_STATUS;
                    push(@{$self->{errmsg}}, "ERROR: '$config{config_name}' already exists on " . $self->get_name() . "\.");
                    return($ERR_STATUS);
                }
            }
            elsif ($self->zone_exists($zone) == $GOOD_STATUS){
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: '$zone' already exists on " . $self->get_name() . "\.");
                return($ERR_STATUS);
            }
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Create zones and add members to those zones.
        # - - - - - - -  - - - - - - - - - - - - - -

        foreach $zone (keys %config) {

            # - - - - - - -  - - - - - - - - - - - - - -
            # Enter config mode.
            # - - - - - - -  - - - - - - - - - - - - - -

            (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

            if ($zone eq 'config_name') { next; }

            $cmd = "zone name " . $zone . " vsan " . $CISCO_VSAN_NUM . "";

            (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

            foreach $zone_member (@{$config{$zone}}) {

                # - - - - - - -  - - - - - - - - - - - - - -
                # Need to find out what type of member this is first...
                # - - - - - - -  - - - - - - - - - - - - - -

                # 10:00:00:23:45:67:89:ab = wwn = used for soft zoning -> or
                # fc2/1 or fc 3/2 = interface zoning = hard zoning     -> or
                # unsupported member type

                SWITCH: {
                    if ($zone_member =~ /$WWN_REG_EX/) {$cmd = "member pwwn " . $zone_member;      last SWITCH;}
                    if ($zone_member =~ /fc ?\d+\/\d+/)     {$cmd = "member interface " . $zone_member; last SWITCH;}
                    $unsupported_member = 1;
                }

                if ($unsupported_member) {
                        $self->{error} = $ERR_STATUS;
                        push(@{$self->{errmsg}}, "ERROR: '$zone_member' is an unsupported member type we only support wwns and interface zoning on " . $self->get_name() . "\.");
                        return ($ERR_STATUS);
                }

                (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

            } # foreach $zone_member (@{config{zone}}) {

            # - - - - - - -  - - - - - - - - - - - - - -
            # Get us back to the normal prompt.
            # - - - - - - -  - - - - - - - - - - - - - -

            $self->cisco_go_to_normal_prompt();

        } # foreach $zone (keys %config)

        # - - - - - - -  - - - - - - - - - - - - - -
        # Get us back to the config prompt.
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

        # - - - - - - -  - - - - - - - - - - - - - -
        # Create a cisco 'zoneset' which is just a name
        # for a zone config.  We will use $config{config_name}
        # - - - - - - -  - - - - - - - - - - - - - -

        $cmd = "zoneset name " . $config{config_name} . " vsan " . $CISCO_VSAN_NUM . "";

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

        # - - - - - - -  - - - - - - - - - - - - - -
        # Add the appropriate zones to this zoneset.
        # - - - - - - -  - - - - - - - - - - - - - -

        foreach $zone (keys %config) {
            if ($zone eq 'config_name') { next; }
            $cmd = "member " . $zone;
            (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", -swallow, '-noprompt', '-prompt', "$hostname.*\#", '-timeout', 30);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Get us back to the normal prompt.
        # - - - - - - -  - - - - - - - - - - - - - -

        $self->cisco_go_to_normal_prompt();

        # - - - - - - -  - - - - - - - - - - - - - -
        # Enable the config.
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", -swallow, '-noprompt', '-prompt', "$hostname.*\#", '-timeout', 30);

        $cmd = "zoneset activate name " . $config{config_name} . " vsan " . $CISCO_VSAN_NUM . "";
        my $zoneset_prompts = "(Zoneset activation initiated|You are trying to activate zoneset|Do you want to continue|$hostname.*\#)";

        ($match, $before, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", -swallow, ,'-noprompt', '-prompt', $zoneset_prompts, '-timeout', 30);
        
        # check if we normally encountered a confirmation prompt.
        if($prompt =~ m/$zoneset_prompts/){
            Tharn::say($self->{conn}, "y", '-noprompt', '-prompt', "$hostname.*\#", '-timeout', 30);
        }

        $self->cisco_go_to_normal_prompt();

        # - - - - - - -  - - - - - - - - - - - - - -
        # Make sure we enabled the config & also validate it, i.e.
        # are the active config name and zones what we planned?
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($self->active_config_get() ne $config{config_name}){
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: $config{config_name} was not successfully enabled on " . $self->get_name() . "\.");
            return($ERR_STATUS);
        }

        foreach $zone (keys %config) {
            if ($zone eq 'config_name') {
                next;
            }
            elsif ($self->zone_exists($zone) == $ERR_STATUS){
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: $zone already exists on switch " . $self->get_name() . "\.");
                return($ERR_STATUS);
            }
        } #foreach $zone

        # - - - - - - -  - - - - - - - - - - - - - -
        # if we got this far, return $GOOD_STATUS
        # - - - - - - -  - - - - - - - - - - - - - -

        return($GOOD_STATUS);

    } # end ($self->{hosttype} =~ /switch-cisco/i) {
    elsif ($self->{hosttype} =~ /switch-mcdata/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # In order to tolerate hard zoning we first need to set the interop
        # mode on the mcdata switch to "mcdata".  The other option "open"
        # does not support hard zoning (using domain/port as members)
        # - - - - - - -  - - - - - - - - - - - - - -
        my %switch_params_to_set = ();
        $switch_params_to_set{interopmode} = "mcdata";

        if ($self->set_switch_parameter(\%switch_params_to_set)) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: Not able to set interopmode to $switch_params_to_set{interopmode} on " . $self->{hostname} . " to MCDATA.");
            return ($ERR_STATUS);
        }
        
        # Now clear the pending config so that we only create what the
        # caller passed in.
        my $return_status = $self->mcdata_clear_pending();
        if($self->get_error_state() == $ERR_STATUS){
            push(@{$self->{errmsg}}, "ERROR: Unable to clear the pending config for McData switch, aborting cfg and zoning creation");
            return ($ERR_STATUS);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # The following sequence is automated below:
        #
        #Root> config zoning addzone jimmy_zone1
        #Root> config zoning addportmem jimmy_zone1 1 1
        #Root> config zoning addportmem jimmy_zone1 1 2
        #Root> config zoning addzone jimmy_zone2
        #Root> config zoning addportmem jimmy_zone2 1 3
        #Root> config zoning addportmem jimmy_zone2 1 4
        #Root> config zoning renamezoneset automation_zoneset
        #Root>
        #Root>
        #Root> show zoning
        #Default Zone Enabled:    False
        #There are no zones in the active zoneset.
        #
        #Root> config zoning activatezoneset
        #Root> show zoning
        #Default Zone Enabled:    False
        #Zone Set:  automation_zoneset
        #  Zone:  jimmy_zone1
        #    Zone Member:  Domain  1, Port  1
        #    Zone Member:  Domain  1, Port  2
        #  Zone:  jimmy_zone2
        #    Zone Member:  Domain  1, Port  3
        #    Zone Member:  Domain  1, Port  4
        #
        #Root>
        # - - - - - - -  - - - - - - - - - - - - - -

        # First start by adding the zones
        foreach $zone (keys %config) {
            if ($zone eq 'config_name') { next; }
            (undef, undef, $after) = Tharn::say($self->{conn}, "config zoning addzone $zone", -swallow, '-timeout', 30);

            # - - - - - - -  - - - - - - - - - - - - - -
            # Does this zone already exist?
            # Root> config zoning addzone jimmy_zone1
            # Error 211: Duplicate Zone
            # - - - - - - -  - - - - - - - - - - - - - -

            if ($after =~ /Error\s+\d+:/i) {
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: The zone $zone already exists on McData switch $hostname.");
                return ($ERR_STATUS);
            }

            # - - - - - - -  - - - - - - - - - - - - - -
            # Next add the domain/port members to their specified zones.
            # We can also add wwn's as members here.
            # - - - - - - -  - - - - - - - - - - - - - -

            foreach $zone_member (@{$config{$zone}}) {

                # - - - - - - -  - - - - - - - - - - - - - -
                # Determine if each member is a domain/port or wwn.
                # - - - - - - -  - - - - - - - - - - - - - -

                if ($zone_member =~ /\d+\s\d+/) {
                    (undef, undef, $after) = Tharn::say($self->{conn}, "config zoning addportmem $zone $zone_member", -swallow, '-timeout', 30);
                }
                elsif ($zone_member =~ /$WWN_REG_EX/) {
                    (undef, undef, $after) = Tharn::say($self->{conn}, "config zoning addwwnmem $zone $zone_member", -swallow, '-timeout', 30);
                } else {
                    # unsupported member type
                    next;
                }

                # - - - - - - -  - - - - - - - - - - - - - -
                # Do any of these members already exist in this zone?
                # Doubt it, but lets check, cause we're that thorough :)
                #
                #Root> config zoning addportmem jimmy_zone1 1 1
                #Error 212: Duplicate Zone Member
                # - - - - - - -  - - - - - - - - - - - - - -

                if ($after =~ /Error\s+\d+:/i) {
                    $self->{error} = $ERR_STATUS;
                    push(@{$self->{errmsg}}, "Error: The zone $zone already contains member $zone_member on McData switch $hostname.");
                    return ($ERR_STATUS);
                }

            } # foreach $zone_member (@{$config{$zone}}) {
        } # foreach $zone (keys %config) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # Now rename whatever the current zoneset is (there may not be one).
        # Even if there is a zoneset in place we can rename no matter what.
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after) = Tharn::say($self->{conn}, "config zoning renamezoneset $config{config_name}", -swallow, '-timeout', 30);

        # - - - - - - -  - - - - - - - - - - - - - -
        # At this point we activate the zone set.
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after) = Tharn::say($self->{conn}, "config zoning activatezoneset", -swallow, '-timeout', 30);

        if ($after =~ /Error\s+\d+:/i) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: An error occurred while activating the zoneset $config{config_name} on $hostname.");
            return ($ERR_STATUS);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Verify our previous actions.
        # - - - - - - -  - - - - - - - - - - - - - -

        $config_ref = $self->active_config_and_zone_get();

        if ($config_ref eq 'no active config'){
            # active_config_and_zone_get() didn't find an active config
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "Error: There currently is no active zoneset on $hostname.");
            return ($ERR_STATUS);
        }

        my %compare_config = %$config_ref;

        foreach $zone (keys %config) {
            if ($zone eq 'config_name') {
                if ($config{$zone} ne $compare_config{$zone}) {
                    # config names don't match
                    $self->{error} = $ERR_STATUS;
                    push(@{$self->{errmsg}}, "ERROR: The config $config{config_name} is not active on $hostname.  Instead the active config is called: $compare_config{config_name}.");
                    return ($ERR_STATUS);
                }
            } else {
                # - - - - - - -  - - - - - - - - - - - - - -
                # Call compare_arrays() to verify our zones and zone members
                # are equal.  First sort both of our arrays.
                # - - - - - - -  - - - - - - - - - - - - - -

                @{$config{$zone}}         = sort(@{$config{$zone}});
                @{$compare_config{$zone}} = sort(@{$compare_config{$zone}});
                
                my @uppercase_zone = ();
                foreach my $element (@{$config{$zone}}){
                  if($element =~ m/$WWN_REG_EX/){
                      $element =~ tr/a-z/A-Z/; 
                  }
                  push @uppercase_zone, $element;
                  
                }

                if (compare_arrays(\@{$config{$zone}}, \@{$compare_config{$zone}}) ) {
                    if (compare_arrays(\@uppercase_zone, \@{$compare_config{$zone}}) ) {
                    
                        $self->{error} = $ERR_STATUS;
                        push(@{$self->{errmsg}}, "ERROR: The zone config currently active is not equal to the zone config we tried to enable." . (Dumper (@{$config{$zone}})) . "\n" . (Dumper (@{$compare_config{$zone}})));
                        return ($ERR_STATUS);
                    }
                    sleep 60;
                }
            }
        } # foreach $zone (keys %config)


        # - - - - - - -  - - - - - - - - - - - - - -
        # if we got this far, return $GOOD_STATUS
        # - - - - - - -  - - - - - - - - - - - - - -

        return($GOOD_STATUS);

    } # end ($self->{hosttype} =~ /switch-mcdata/i)
    else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
            return ($ERR_STATUS);
    }
}

## @name ::zone_config_delete()
##
## @description
## Deletes the config name and all the zones contained within it.
## The zone config passed in is a hash.  If the zone config was deleted
## we return $GOOD_STATUS otherwise we return $ERR_STATUS on error.
##
## For McData use 'zone_config_disable()' which is the closest to this
## type of functionality.
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## $switch_obj->zone_config_delete(\%config)
##
## @arg \%config Required: A zone config, different format for each vendor
##      See below for config format.
##
sub zone_config_delete(%) {
    my ($self, $config_ref) = @_;
    my %config = %$config_ref;
    my ($after, $prompt, $zone, $cmd) = '';
    my $hostname = $self->get_name();

    # - - - - - - -  - - - - - - - - - - - - - -
    # Determine what hosttype we're dealing with
    # i.e. Brocade, Cisco, McData.
    # - - - - - - -  - - - - - - - - - - - - - -

    if ($self->{hosttype} =~ /switch-brocade/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # For brocade switches the hash passed in should
        # have the following format key/value pairs:
        #
        # 'config_name' -> ($ SCALAR) The name of your zone_config (c-style)
        # 'zone_name1'  -> (@ ARRAY)  An array of ports/wwns.
        # 'zone_nameN'  -> (@ ARRAY)  Same as above...
        # 'zone_nameN+1'-> (@ ARRAY)  Same as above...
        # - - - - - - -  - - - - - - - - - - - - - -

        $cmd = "cfgDelete \"$config{config_name}\"";

        (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);
        if ($after =~ /^\s*error:/) { return($ERR_STATUS); }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Just deleted the config so now we need to delete each zone
        # - - - - - - -  - - - - - - - - - - - - - -

        $cmd = ''; # cleanup our command string holder.

        # zoneDelete "jimmyzone"

        foreach $zone (keys %config) {
            if ($zone eq 'config_name') {next;}

            $cmd = "zoneDelete \"$zone\"";

            (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);

            if ($after =~ /^\s*error:/) { return($ERR_STATUS); }
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Save the config just incase.
        # - - - - - - -  - - - - - - - - - - - - - -

        $self->zone_save_config();

        if ($self->get_error_state()) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: Unable to save the zone config $config{config_name} on switch $hostname.");
            return($ERR_STATUS);
        }
        else {
            return($GOOD_STATUS);
        }

    } # end (self->{hosttype} =~ /switch-brocade/i)
    elsif ($self->{hosttype} =~ /switch-cisco/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # The following code automates this similar procedure:
        #
        #vns-c9124-2# config
        #Enter configuration commands, one per line.  End with CNTL/Z.
        #vns-c9124-2(config)# no zoneset name jimmy_zoneset vsan 1
        #vns-c9124-2(config)# no zone name jimmy1 vsan 1
        #vns-c9124-2(config)# end
        #vns-c9124-2#
        # - - - - - - -  - - - - - - - - - - - - - -

        # - - - - - - -  - - - - - - - - - - - - - -
        # Get us to the config prompt.
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

        $cmd = "no zoneset name " . $config{config_name} . " vsan " . $CISCO_VSAN_NUM . "";

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

        foreach $zone (keys %config) {
            if ($zone eq 'config_name') {next;}

            $cmd = "no zone name " . $zone . " vsan " . $CISCO_VSAN_NUM . "";

            (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Get back to the normal prompt.
        # - - - - - - -  - - - - - - - - - - - - - -

        $self->cisco_go_to_normal_prompt();

        # - - - - - - -  - - - - - - - - - - - - - -
        # Do some error checking to make sure you deleted those configs:
        # - - - - - - -  - - - - - - - - - - - - - -

        $cmd = "show zoneset name " .$config{config_name}. " vsan " . $CISCO_VSAN_NUM . "";

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 30);

        if ($after !~ /Zoneset not present/) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: Not able to delete " . $config{config_name} . " on " . $self->get_name() . "\.");
            return($ERR_STATUS);
        }

        foreach $zone (keys %config) {

            if ($zone eq 'config_name') { next; }

            $cmd = "show zone name " . $zone . " vsan " . $CISCO_VSAN_NUM . "";

            (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 30);

            if ($after !~ /Zone not present/) {
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: Not able to delete zone name " . $zone . " on " . $self->get_name() . "\.");
                return($ERR_STATUS);
            }
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # If we got this far return good status.
        # - - - - - - -  - - - - - - - - - - - - - -

        return ($GOOD_STATUS);
    }
    else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
            return ($ERR_STATUS);
    }
}

## @name ::active_config_and_zone_get()
##
## @description
## Returns a reference to a hash which contains the active zone
## config and its contained zones in the following key/value format:
##
## 'config_name' -> ($ SCALAR) The name of your zone_config (c-style)
## 'zone_name1'  -> (@ ARRAY)  An array of ports/wwns.
## 'zone_nameN'  -> (@ ARRAY)  Same as above...
## 'zone_nameN+1'-> (@ ARRAY)  Same as above...
##
## In the above 'config_name' is a staticly named key... all other keys
## are zone names but not necessarily named 'zone_name1'... etc.
##
## If there is no active config we return the string "no active config".
## If there are errors worth mentioning we push them onto the errmsg array.
## We also only completely support hard zoning at this point for all vendors.
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## $config_ref = $switch_obj->active_config_and_zone_get()
##
sub active_config_and_zone_get() {
    my ($self) = @_;
    my ($after, $line, $member, $cur_zone_name) = '';
    my @array = ();
    my %zone_config   = ();
    my @cur_zone_memb = ();
    my @temp = ();
    my $loop_counter = 1;
    my $count = 0;
    my $no_config = "no active config";


    if ($self->{hosttype} =~ /switch-brocade/i) {
        # - - - - - - -  - - - - - - - - - - - - - -
        # Do a moreDisable (kills CR prompts.)
        # - - - - - - -  - - - - - - - - - - - - - -

        Tharn::say($self->{conn}, 'moreDisable', -swallow);

        # - - - - - - -  - - - - - - - - - - - - - -
        # Populate $output with the results of the cfgshow brocade switch command.
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, $after) = Tharn::say($self->{conn}, "cfgshow", -swallow, '-timeout', 60);

        # - - - - - - -  - - - - - - - - - - - - - -
        # $after = "Defined configuration:
        # cfg:   Automation1
        #               Auto1; Auto2
        # cfg:   ams500  ams500_2a_2b
        # cfg:   config  ic_zone1; ic_zone2; others
        # cfg:   gf825c_gf960c
        #               FAStT600; gf980
        # cfg:   jimmyconfig
        #               jimmyzone
        # zone:  Auto1   1,0; 1,2
        # zone:  Auto2   1,1; 1,3
        # zone:  FAStT600
        #               131,1; 131,5; 131,14; 131,11; 131,10; 131,0; 131,2; 131,8;
        #               131,12; 131,15
        # zone:  ams500_2a_2b
        #               103,13; 103,14; 103,15
        # zone:  gf980   131,6; 131,9
        # zone:  ic_zone1
        #               2,8; 11,8
        # zone:  ic_zone2
        #               2,9; 11,9
        # zone:  jimmyzone
        #               1,2
        # zone:  others  2,1; 2,3; 2,5; 2,6; 2,11; 2,12; 2,13; 2,14; 2,15; 11,1; 11,2;
        #               11,3; 11,5; 11,6; 11,7; 11,10; 11,11; 11,12; 11,13; 11,14;
        #               11,15; 2,7; 2,10; 2,0; 2,2; 11,0; 11,4; 2,4
        #
        #Effective configuration:
        # cfg:   Automation1
        # zone:  Auto1   1,0
        #               1,2
        # zone:  Auto2   1,1
        #               1,3";
        # - - - - - - -  - - - - - - - - - - - - - -

        @array = split (/\n/, $after);

        foreach (@array) {
            unless ($_ =~ /^\s*Effective configuration:/i) { $count++; next; }
            last;
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Strip off "Effective configuration:"
        # - - - - - - -  - - - - - - - - - - - - - -

        @array = @array[($count + 1) .. $#array];

        # - - - - - - -  - - - - - - - - - - - - - -
        # Now @array looks like this:
        #Effective configuration:  <- minus this, we stripped this line above
        # cfg:   Automation1
        # zone:  Auto1   1,0
        #                1,2
        # zone:  Auto2   1,1
        #                1,3
        # - - - - - - -  - - - - - - - - - - - - - -

        foreach $line (@array){
            if ($line =~ /^\s*cfg:/){
                @temp = split(/\s+/, $line);
                $zone_config{config_name} = $temp[2];
                next;
            }

            if ($line =~ /^\s*zone:/){

                # - - - - - - -  - - - - - - - - - - - - - -
                # If this is the first iteration, don't set the key b/c its null
                # - - - - - - -  - - - - - - - - - - - - - -

                if ($loop_counter > 1){$zone_config{$cur_zone_name} = [@cur_zone_memb];}
                @temp = split(/\s+/, $line);
                $cur_zone_name = $temp[2];
                @cur_zone_memb = ();
                foreach $member (@temp){

                    # - - - - - - -  - - - - - - - - - - - - - -
                    # Members look like this:
                    # 1,2 or 50:0a:09:80:00:01:43:88
                    # - - - - - - -  - - - - - - - - - - - - - -

                    if (($member =~ /\d,\d/) || ($member =~ /\w{2}:\w{2}:\w{2}:\w{2}:\w{2}:\w{2}:\w{2}:\w{2}/)){
                        push @cur_zone_memb, $member;
                    }
                } # foreach $member
                $loop_counter++;
                next;
            } # line =~ /zone:/

            if (($line !~ /^\s*cfg:/) && ($line !~ /^\s*zone:/)) {
                @temp = split(/\s+/, $line);
                foreach $member (@temp){
                    if (($member =~ /\d,\d/) || ($member =~ /\w{2}:\w{2}:\w{2}:\w{2}:\w{2}:\w{2}:\w{2}:\w{2}/)){
                        push @cur_zone_memb, $member;
                    } # $member =~
                } # foreach $member (@_)
            } # $line !~
        } # foreach $line (@array)

        # - - - - - - -  - - - - - - - - - - - - - -
        # Add the last value to the hash.
        # - - - - - - -  - - - - - - - - - - - - - -

        $zone_config{$cur_zone_name} = [@cur_zone_memb];

        # - - - - - - -  - - - - - - - - - - - - - -
        # Do a moreEnable on the switch. (return CR to normal)
        # - - - - - - -  - - - - - - - - - - - - - -

        Tharn::say($self->{conn}, 'moreEnable', -swallow);

        # - - - - - - -  - - - - - - - - - - - - - -
        # We should do some additional checking of our config returned
        # incase we didn't discover what we'd hope to.
        # - - - - - - -  - - - - - - - - - - - - - -

        # did we find a config name?
        if (!defined $zone_config{config_name}){
            push(@{$self->{errmsg}}, "WARN: Not able to find a config name on " . $self->get_name());
            return($no_config);
        }

        # did we find any zones?
        # reset our counter
        $count = 0;
        foreach (keys %zone_config){
            if ($_ eq 'config_name') {next;}
            $count++;
        }
        if ($count < 1){
            # this means we didn't find at least one zone
            push(@{$self->{errmsg}}, "WARN: Were not able to find any active zones in " . $zone_config{config_name} . " on " . $self->get_name() . "\.");
        }

        # do our zones have at least 1 member per zone?
        # clear the counter incase it was used in the previous block.
        $count = 0;
        foreach $cur_zone_name (keys %zone_config){
            if ($cur_zone_name eq 'config_name'){
                next;
            }
            elsif (scalar @{$zone_config{$cur_zone_name}} < 1){
                push(@{$self->{errmsg}}, "WARN: Zone name $cur_zone_name residing on " . $self->get_name() . " contains zero members\.");
            }
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # return the config even if errors occurred.  It is up to the user to
        # check that we didn't return "no active config".  If things eventually
        # go sour due to a config issue, when the errmsg array is printed out
        # warnings will be seen.
        # - - - - - - -  - - - - - - - - - - - - - -

        return(\%zone_config);

    } # end ($self->{hosttype} =~ /switch-brocade/i) {
    elsif ($self->{hosttype} =~ /switch-cisco/i) {

        # Fix for BURT 490772.
        # The output of show zoneset active  might be prompted by more. Pipe to no-more to avoid more.

        (undef, undef, $after) = Tharn::say($self->{conn}, "show zoneset active | no-more", -swallow, '-timeout', 60);

        # - - - - - - -  - - - - - - - - - - - - - -
        # $after will now contain something like this for the positive case:
        #
        #vns-c9124-2# show zoneset active
        #zoneset name auto_config vsan 1
        #  zone name Automation3 vsan 1
        #  * fcid 0x0e0300 [interface fc1/1 swwn 20:00:00:0d:ec:47:bb:00]
        #  * fcid 0x0e0400 [interface fc1/2 swwn 20:00:00:0d:ec:47:bb:00]
        #vns-c9124-2#
        #
        # Soft zoning example, with port logged and and without
        #zoneset name multi-array-1 vsan 1\r\n
	#  zone name v3070_4a_hds17159_CL2B vsan 1\r\n
	#    pwwn 50:0a:09:80:00:01:ae:f5\r\n
	#  * fcid 0x740700 [pwwn 50:06:0e:80:05:43:07:11]\r\n
        #
        # In the negative case $after should look like this:
        #
        #vns-c9124-2# show zoneset active
        #Zoneset not present
        #vns-c9124-2#
        #
        # Also if the tester does not have interface zoning configured
        # we will fail and log an informative error message stating that
        # the user needs to alter their zoning.  note 'interface-zoning'
        # is the name cisco uses for 'hard zoning'
        # - - - - - - -  - - - - - - - - - - - - - -

        @array = split (/\n/, $after);

        foreach $line (@array){
            if ($line =~ /^.*Zoneset not present.*\n?$/){
                return($no_config);
            }
            elsif ($line =~ /^.*zoneset name (\S+) vsan 1.*\n?$/) {
                $zone_config{config_name} = $1; # the (\S+) above is $1
                next;
            }
            elsif ($line =~ /^.*zone name (\S+) vsan 1.*\n?$/) {
                $cur_zone_name = $1;
                @cur_zone_memb = (); # used to init $zone_config{$cur_zone_name} as an array
                $zone_config{$cur_zone_name} = [@cur_zone_memb];
            }
            elsif ($line =~ /^\s*\*\s+fcid\s+\w+\s+\[interface\s+(fc\d+\/\d+)\s+.*\n?$/) {
                # we found a member of some zone which uses 'interface zoning'
                push (@{$zone_config{$cur_zone_name}}, $1);
            }
            elsif ($line =~ /^\s*\*\s+fcid\s+\w+\s+\[pwwn\s+($WWN_REG_EX)\].*\n?$/) {
	        # we found a member of some zone which uses soft zoning and 
	        # is logged in.
	        push (@{$zone_config{$cur_zone_name}}, $1);
            }
            elsif ($line =~ /^\s*pwwn\s+($WWN_REG_EX).*\n?$/) {
	        # we found a member of some zone which uses soft zoning and 
	        # is not logged in.
	        push (@{$zone_config{$cur_zone_name}}, $1);
            }            
            else {
                # could be some garbage line or the above conditions need work.
                next;
            }

        } # foreach $line (@array){

            # - - - - - - -  - - - - - - - - - - - - - -
            # We should do some additional checking of our config returned
            # incase we didn't discover what we'd hope to.
            # - - - - - - -  - - - - - - - - - - - - - -

            # did we find a config name?
            if (!defined $zone_config{config_name}){
                push(@{$self->{errmsg}}, "WARN: Not able to find a config name on " . $self->get_name());
                return($no_config);
            }

            # did we find any zones?
            foreach (keys %zone_config){
                if ($_ eq 'config_name') {next;}
                $count++;
            }
            if ($count < 1){
                # this means we didn't find at least one zone
                push(@{$self->{errmsg}}, "WARN: Were not able to find any active zones on " . $self->get_name() . "\.");
            }

            # do our zones have at least 1 member per zone?
            # clear the counter incase it was used in the previous block.
            $count = 0;
            foreach $cur_zone_name (keys %zone_config){
                if ($cur_zone_name eq 'config_name'){
                    next;
                }
                elsif ((scalar @{$zone_config{$cur_zone_name}}) < 1){
                    push(@{$self->{errmsg}}, "ERROR: Zone name $cur_zone_name in " . $zone_config{config_name} . " residing on " . $self->get_name() . " contains zero members\.");
                }
            }

            # - - - - - - -  - - - - - - - - - - - - - -
            # Return the config even if errors occurred.  It is up to the
            # caller to see if we returned "no active config" or a
            # reference to a hash.
            # - - - - - - -  - - - - - - - - - - - - - -

            return(\%zone_config);

    } # end ($self->{hosttype} =~ /switch-cisco/i)

    elsif ($self->{hosttype} =~ /switch-mcdata/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        #Root> show zoning
        #Default Zone Enabled:    True
        #Zone Set:  Zone2
        #  Zone:  V3070_2A_ESS800
        #    Zone Member:  50:0A:09:82:00:00:76:31
        #    Zone Member:  50:05:07:63:00:C8:A1:79
        #  Zone:  V3070_2B_ESS800
        #    Zone Member:  50:0A:09:82:00:00:76:3F
        #    Zone Member:  50:05:07:63:00:C4:A1:79
        #  Zone:  v6030_1
        #    Zone Member:  Domain  1, Port  0
        #    Zone Member:  Domain  1, Port  1
        #    Zone Member:  Domain  1, Port  4
        #  Zone:  v6030_2
        #    Zone Member:  Domain  1, Port  2
        #    Zone Member:  Domain  1, Port  3
        #    Zone Member:  Domain  1, Port  5
        #  Zone:  v3050c
        #    Zone Member:  Domain  1, Port  8
        #    Zone Member:  Domain  1, Port  9
        #    Zone Member:  Domain  1, Port 10
        #    Zone Member:  Domain  1, Port  6
        #  Zone:  v3050c-eva8000
        #    Zone Member:  Domain  1, Port  8
        #    Zone Member:  Domain  1, Port  9
        #    Zone Member:  Domain  1, Port 10
        #    Zone Member:  Domain  1, Port  6
        #
        #Root>
        #
        # In the negative case, there may not be any zoning in effect:
        #
        #Root> show zoning
        #Default Zone Enabled:    False
        #There are no zones in the active zoneset.
        # - - - - - - -  - - - - - - - - - - - - - -


        (undef, undef, $after) = Tharn::say($self->{conn}, "show zoning", -swallow, '-timeout', 60);

        @array = split (/\n/, $after);

        foreach $line (@array) {
            if ($line =~ /^.*Zone Set:\s+(\S+)/) {
                $zone_config{config_name} = $1; # the (\S+) above is $1
            }
            elsif ($line =~ /^.*Zone:\s+(\S+).*\n?$/) {
                $cur_zone_name = $1;
                # used to init $zone_config{$cur_zone_name} as an array
                @cur_zone_memb = ();
                $zone_config{$cur_zone_name} = [@cur_zone_memb];
            }
            elsif ($line =~ /^.*Zone Member:\s+Domain\s+(\d+),\s+Port\s+(\d+)/) {
                # found a member of zone which uses domain/port (hard) zoning
                # $member will be a cat of domain<space>port
                $member = $1 . " " . $2;
                push (@{$zone_config{$cur_zone_name}}, $member);
            }
            elsif ($line =~ /^.*Zone Member:\s+($WWN_REG_EX)/) {
                # found wwn members (soft zones) push them
                # on and have the caller deal with it...
                push (@{$zone_config{$cur_zone_name}}, $1);
            }
            elsif ($line =~ /There\s+are\s+no\s+zones\s+in\s+the\s+active\s+zoneset/) {
                # There is no active config in place.
                return($no_config);
            }

            else {
                # could be some garbage line or the above conditions need work.
                next;
            }
        } # foreach $line (@array) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # We should do some additional checking of our config returned
        # incase we didn't discover what we'd hope to.
        # - - - - - - -  - - - - - - - - - - - - - -

        # did we find a config name?
        if (!defined $zone_config{config_name}){
            push(@{$self->{errmsg}}, "WARN: Not able to find a zone config name on " . $self->get_name());
            return($no_config);
        }

        # did we find any zones?
        foreach (keys %zone_config){
            if ($_ eq 'config_name') {next;}
            $count++;
        }
        if ($count < 1){
            # this means we didn't find at least one zone
            push(@{$self->{errmsg}}, "WARN: Were not able to find any active zones on " . $self->get_name() . "\.");
        }

        # do our zones have at least 1 member per zone?
        foreach $cur_zone_name (keys %zone_config){
            if ($cur_zone_name eq 'config_name'){
                next;
            }
            elsif ((scalar @{$zone_config{$cur_zone_name}}) < 1){
                push(@{$self->{errmsg}}, "ERROR: Zone name $cur_zone_name in " . $zone_config{config_name} . " residing on " . $self->get_name() . " contains zero members\.");
            }
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Return the config even if errors occurred.  It is up to the
        # caller to see if we returned "no active config" or a
        # reference to a hash.
        # - - - - - - -  - - - - - - - - - - - - - -

        return(\%zone_config);

    } # end ($self->{hosttype} =~ /switch-mcdata/i) {
    else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
            return ($ERR_STATUS);
    }
}

## @name ::zone_with_port_find(\%config, $port_to_find)
##
## @description
## Returns the name of the zone in \%config which includes the $port_to_find.
## If it cannot find the port it returns "not found".
##
## Note: All cisco zone configuration is done on vsan 1.
##       Also, understand that there was a bug fixed in IronCity which is
##       OnTap version 7.3 where the Cisco blade was incremented by 1 to
##       reflect proper functionality.  If a test is calling this method
##       and the test may be run on prior versions of OnTap, i.e 7.2.x
##       then it is on the caller to manipulate the $port_to_find before
##       calling this method.  i.e. if your filer is 7.2.3 you should
##       manipulate the blade first.  If your 7.3 you shouldn't have any
##       issues.
##
## @synopsis
## $zone = VSeries::switch_object->zone_with_port_find(\%config, $port_to_find);
##
## @arg %config Required: The zone config which you wish to search through.
## @arg $port_to_find Required: The port to find in one of the zones.
##
sub zone_with_port_find(%$){
    my ($self, $config_ref, $port_to_find) = @_;
    my %config = %$config_ref;
    my ($domain, $port) = '';
    my $hostname = $self->get_name();
    my $target_zone = undef;


    if ($self->{hosttype} =~ /switch-brocade/i) {
        foreach my $zone (keys %config) {
            if ($zone eq 'config_name') {next;}
            foreach my $member (@{$config{$zone}}) {
                if ($member =~ /,/) {
                    ($domain, $port) = split(/,/, $member);
                    if ($port eq $port_to_find){$target_zone = $zone; last;}
                }
            }
        }
        if (!defined $target_zone){return("not found");}
        return($target_zone);

    } # end ($self->{hosttype} =~ /switch-brocade/i) {
    elsif ($self->{hosttype} =~ /switch-cisco/i) {

        my ($cisco_port, $cisco_blade) = '';

        # - - - - - - -  - - - - - - - - - - - - - -
        # First change the port to find so that it is in a form
        # that we find when calling active_config_and_zone_get().
        #
        # i.e. $port_to_find may look like '1-1' but in reality
        # this is actually 'fc1/1' on Cisco.
        # - - - - - - -  - - - - - - - - - - - - - -

        ($cisco_blade, $cisco_port) = split(/-/, $port_to_find);

        $port_to_find = "fc$cisco_blade\/$cisco_port";

        foreach my $zone (keys %config) {
            if ($zone eq 'config_name') {next;}
            foreach my $member (@{$config{$zone}}) {
                if ($member eq $port_to_find) {
                    $target_zone = $zone;
                    last;
                }
            }
        }
        if (!defined $target_zone){return("not found");}
        return($target_zone);

    } # end ($self->{hosttype} =~ /switch-cisco/i) {
    elsif ($self->{hosttype} =~ /switch-mcdata/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # McData members for domain/ports look like this 1<space>2
        # - - - - - - -  - - - - - - - - - - - - - -

        foreach my $zone (keys %config) {
            if ($zone eq 'config_name') {next;}
            foreach my $member (@{$config{$zone}}) {
                if ($member =~ /\d+\s\d+/) {
                    ($domain, $port) = split(/\s/, $member);
                    if ($port eq $port_to_find){$target_zone = $zone; last;}
                }
            }
        }
        if (!defined $target_zone){return("not found");}
        return($target_zone);
    }

    else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
            return ($ERR_STATUS);
    }
}

## @name ::zone_save_config()
##
## @description
## Save zone configuration changes on a fibre channel switch.  If an
## error were to occur it will place an error in the switch object's
## error flag.
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## VSeries::switch_object->zone_save_config();
##
##
sub zone_save_config(){
    my $self = shift;
    my ($after, $prompt) = '';
    my $confirm_prompt = '^.*\(yes, y, no, n\): \[no\]\s*\n*';

    if ($self->{hosttype} =~ /switch-brocade/i) {
        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "cfgSave", -swallow, '-prompt', $confirm_prompt, '-timeout', 60);

        # - - - - - - -  - - - - - - - - - - - - - -
        # If this is a b200e switch (v5 firmware), we need to confirm stuff with "y"
        # - - - - - - -  - - - - - - - - - - - - - -

        if ($prompt =~ /$confirm_prompt/) {
            # Just wait for a normal prompt.
            (undef, undef, $after) = Tharn::say($self->{conn}, "y", -swallow, '-timeout', 60);
        }
        if ($after !~ /cfgSave successfully completed/              &&
            $after !~ /Nothing changed: nothing to save, returning/ &&
            $after !~ /Updating flash/) {

            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "The switch config changes were not successfully saved.");
            return ($ERR_STATUS);
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # If we got this far return good status
        # - - - - - - -  - - - - - - - - - - - - - -

        return($GOOD_STATUS);
    }
    else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
            return ($ERR_STATUS);
    }

}

## @name ::zone_exists()
##
## @description
## Returns $GOOD_STATUS if the zone exists and $ERR_STATUS otherwise.
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## VSeries::switch_object->zone_exists($zone);
##
## @arg $zone Required: Check to see if this $zone exists.
##
sub zone_exists($){
    my ($self, $zone) = @_;
    my ($after, $cmd, $prompt) = '';
    my $hostname = $self->get_name();

    if ($self->{hosttype} =~ /switch-cisco/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # make sure we're at the normal prompt first
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, undef, $prompt) = Tharn::say($self->{conn}, " ", '-noprompt', '-prompt', "$hostname(\(config.*\))?\#", -swallow, '-timeout', 30);

        if ($1 =~ /config/){
            $self->cisco_go_to_normal_prompt();
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Check to see if this zone exists.
        # - - - - - - -  - - - - - - - - - - - - - -

        $cmd = "show zone name " . $zone . " vsan " . $CISCO_VSAN_NUM . "";
        (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);

        if ($after !~ /Zone not present/) {
            return($GOOD_STATUS);
        }
        else{ return($ERR_STATUS); }
    }
    else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
            return ($ERR_STATUS);
    }
}

## @name ::zone_config_exists()
##
## @description
## Returns $GOOD_STATUS if the zone_config exists and $ERR_STATUS otherwise.
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## VSeries::switch_object->zone_config_exists($zone_config);
##
## @arg $zone_config Required: Check to see if this $zone_config exists.
##
sub zone_config_exists($){
    my ($self, $zone_config) = @_;
    my ($after, $cmd, $prompt) = '';
    my $hostname = $self->get_name();

    if ($self->{hosttype} =~ /switch-cisco/i) {

        # - - - - - - -  - - - - - - - - - - - - - -
        # make sure we're at the normal prompt first
        # - - - - - - -  - - - - - - - - - - - - - -

        (undef, undef, undef, $prompt) = Tharn::say($self->{conn}, " ", '-noprompt', '-prompt', "$hostname(\(config.*\))?\#", -swallow, '-timeout', 30);

        if ($1 =~ /config/){
            $self->cisco_go_to_normal_prompt();
        }

        # - - - - - - -  - - - - - - - - - - - - - -
        # Check to see if this zoneset exists.
        # - - - - - - -  - - - - - - - - - - - - - -
        $cmd = "show zoneset name " . $zone_config . " vsan " . $CISCO_VSAN_NUM . "";
        (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);

        if ($after !~ /Zoneset not present/) {
            return($GOOD_STATUS);
        }
        else{ return($ERR_STATUS); }
    }
    else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
            return ($ERR_STATUS);
    }
}

## @name ::cisco_go_to_normal_prompt()
##
## @description
## Brings a cisco switch to its normal prompt
##
## @synopsis
## VSeries::switch_object->cisco_go_to_normal_prompt();
##
##
sub cisco_go_to_normal_prompt(){
    my $self = shift;
    my $after = '';

    (undef, undef, $after) = Tharn::say($self->{conn}, "end", -swallow, '-timeout', 60);

    return $GOOD_STATUS;
}

## @name ::get_error_state()
##
## @description
## Returns the value of the error flag
##
## @synopsis
## VSeries::switch_object->get_error_state();
##
##
sub get_error_state(){
    my $self = shift;
    return $self->{error};
}

## @name ::get_error_msg()
##
## @description
## Returns the error message for a given switch object.  This is useful
## for reporting failures or unexpected behavior for a given test. Its
## an array so several errors could be contained within the array. Also
## this is returned as a reference so the caller must dereference to get
## any serious use out of it.
##
## @synopsis
## VSeries::switch_object->get_error_msg();
##
##
sub get_error_msg(){
    my $self = shift;
    return (\@{$self->{errmsg}});
}

## @name ::get_name()
##
## @description
## Returns the name of a given switch object.
##
## @synopsis
## $name = VSeries::switch_object->get_name();
##
##
sub get_name(){
    my $self = shift;
    return($self->{hostname});
}

## @name ::get_host_type()
##
## @description
## Returns the host type of the switch.
##
## @synopsis
## $host_type = VSeries::switch_object->get_host_type();
##
##
sub get_host_type(){
    my $self = shift;
    return($self->{hosttype});
}

## @name ::clear_error()
##
## @description
## Method to clear the error state of a switch.  In some negative tests, it's
## expected that an error is generated.  This allows for gracefull continuation
## after an error.
##
## @synopsis
## VSeries::switch_object->clear_error();
##
##
sub clear_error(){
    my $self = shift;
    $self->{error} = 0;
    @{$self->{errmsg}} = ();

}

## @name ::set_switch_parameter()
##
## @description
## Method to set a parameter on a switch.  This is usually an odd ball
## binary setting on a switch for instance changing the setting opensysms
## on a mcdata switch to enabled/disabled. Takes in a hash where the key
## and value pairs are the settings you'd like to set.  It returns
## $GOOD_STATUS on success and $ERR_STATUS on failure.
##
## C-U-R-R-E-N-T S-U-P-P-O-R-T-E-D P-A-R-A-M-E-T-E-R-S
## MCDATA:
## - key->opensysms   value->1/0 for on/off respectively
## - key->interopmode value->open/mcdata for 2 types of interop
##
## BROCADE:
## - no support yet.
##
## CISCO:
## - no support yet.
##
## @synopsis
## $status = VSeries::switch_object->set_switch_parameter(\%parameters_ref);
##
##
sub set_switch_parameter(%){
    my ($self, $parameters_ref) = @_;
    my %parameters = ();
    my $param_not_found = 0;
    my ($cmd, $after) = '';

    # Dereference the hash:
    %parameters = %$parameters_ref;

    if ($self->{hosttype} =~ /switch-mcdata/i) {

        foreach my $param (keys %parameters) {

            SWITCH1: {
                if ($param =~ /opensysms/i) {
                    $cmd = "config features opensysms $parameters{$param}";
                    (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);
                    last SWITCH1;
                }
                if ($param =~ /interopmode/i) {
                    # Need to offline switch, apply cmd, then online switch.
                    $cmd = "maint system setonlinestate false";
                    (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);
                    $cmd = "config switch interopmode $parameters{$param}";
                    (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);
                    $cmd = "maint system setonlinestate true";
                    (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);
                    last SWITCH1;
                }
                # no support for current parameter
                $param_not_found = 1;

            } # SWITCH1:

            if ($param_not_found) {
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: Not able to set ". $param ." to ". $parameters{$param} .".");
                return $ERR_STATUS;
            }
        } # foreach my $param

        # - - - - - - -  - - - - - - - - - - - - - -
        # Now we need to verify each option that was set
        # - - - - - - -  - - - - - - - - - - - - - -

        foreach my $param (keys %parameters) {

            SWITCH2: {
                if ($param =~ /opensysms/i) {
                    # - - - - - - -  - - - - - - - - - - - - - -
                    #Root> show opensysms config
                    #OpenSysMS State:       Enabled
                    #Host Control State:    Enabled
                    #Root>
                    #
                    # OpenSysMS State should be Enabled/Disabled dependent
                    # on how we set it in %parameters
                    # - - - - - - -  - - - - - - - - - - - - - -

                    $cmd = "show opensysms config";
                    (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);

                    if ($parameters{$param} =~ /^1$/) {
                        if ($after !~ /OpenSysMS State:\s+Enabled/i) {
                            $self->{error} = $ERR_STATUS;
                            push(@{$self->{errmsg}}, "ERROR: Not able to set the state of OpenSysMS to " . $parameters{$param} . ".");
                            return $ERR_STATUS;
                        }
                    }
                    elsif ($parameters{$param} =~ /^0$/) {
                        if ($after !~ /OpenSysMS State:\s+Disabled/i) {
                            $self->{error} = $ERR_STATUS;
                            push(@{$self->{errmsg}}, "ERROR: Not able to set the state of OpenSysMS to " . $parameters{$param} . ".");
                            return $ERR_STATUS;
                        }
                    } else {
                        $self->{error} = $ERR_STATUS;
                        push(@{$self->{errmsg}}, "ERROR: " . $parameters{$param} . " is not a valid value for " . $param . ".");
                        return $ERR_STATUS;
                    }

                    last SWITCH2;
                } # ($param =~ /opensysms/i) {

                if ($param =~ /interopmode/i) {
                    # - - - - - - -  - - - - - - - - - - - - - -
                    #Root> show switch
                    #State:                  Online
                    #R_A_TOV:                100
                    #Rerouting Delay:        Disabled
                    #Interop Mode:           McDATA Fabric 1.0
                    #.
                    #. *more stuff*
                    #.
                    #API Management IP:      10.97.104.168
                    #Password Exp. Time:     0
                    #Root>
                    # - - - - - - -  - - - - - - - - - - - - - -

                    $cmd = "show switch";
                    (undef, undef, $after) = Tharn::say($self->{conn}, "$cmd", -swallow, '-timeout', 60);

                    if ($parameters{$param} =~ /^mcdata$/i) {
                        if ($after !~ /Interop Mode:\s+McDATA Fabric\s+\d+/i) {
                            $self->{error} = $ERR_STATUS;
                            push(@{$self->{errmsg}}, "Error: Not able to set the state of the Interop Mode to " . $parameters{$param} . ".");
                            return $ERR_STATUS;
                        }
                    }
                    elsif ($parameters{$param} =~ /^open$/i) {
                        if ($after !~ /Interop Mode:\s+Open Fabric\s+\d+/i) {
                            $self->{error} = $ERR_STATUS;
                            push(@{$self->{errmsg}}, "ERROR: Not able to set the state of the Interop Mode to " . $parameters{$param} . ".");
                            return $ERR_STATUS;
                        }
                    } else {
                        $self->{error} = $ERR_STATUS;
                        push(@{$self->{errmsg}}, "ERROR: " . $parameters{$param} . " is not a valid value for " . $param . ".");
                        return $ERR_STATUS;
                    }

                    last SWITCH2;
                }
            } # SWITCH2:
        } # foreach my $param

        # - - - - - - -  - - - - - - - - - - - - - -
        # If we got this far, the appropriate settings were set.
        # - - - - - - -  - - - - - - - - - - - - - -

        return ($GOOD_STATUS);

    } # end ($self->{hosttype} =~ /switch-mcdata/
    else {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: There is no support for " . $self->{hosttype} . " switches in this method.");
            return ($ERR_STATUS);
    }
}

## @name ::so_switch_eport_get ()
##
## @description:  get switch interconnect info.  Returns an array of hashes
##  having port number, its WWN and peer's WWN
##
## @synopsis
## $array_of_hashes_ref = VSeries::switch_object->so_switch_eport_get()
##
## Note: Cisco switch section supports Vsan 1 only
##
sub so_switch_eport_get () {
    my $self = shift;
    my $hostname = $self->get_name();
    my $hosttype = $self->get_host_type();
    my ($after, $line, $length_port_name) = "";
    my @info = ();
    my @port = ();
    my (@port_WWN, @peer_port_WWN) = ();
    my $version = "";
    my @eportlist = ();

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Cisco switch
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if ($hosttype =~ /switch-cisco/i) {

        Tharn::say($self->{conn}, "terminal length 0", -swallow, '-timeout', 60);

        # Fix for BURT 490772.
        # The output of show interface  might be prompted by more. Pipe to no-more to avoid more.

        (undef, undef, $after) = Tharn::say($self->{conn}, "show interface brief | no-more", -swallow,
                                                          '-timeout', 60);
        $after =~ s/\r//;
        my @info = split(/\n/, $after);

        # show interface brief sample output

        #-------------------------------------------------------------------------------
        #Interface  Vsan   Admin  Admin   Status          SFP    Oper  Oper   Port
        #                  Mode   Trunk                          Mode  Speed  Channel
        #                         Mode                                 (Gbps)
        #-------------------------------------------------------------------------------
        #fc1/1      1      auto   on      notConnected     swl    --           --
        #fc1/2      1      auto   on      notConnected     swl    --           --
        #fc1/3      1      auto   on      up               swl    F       4    --
        #fc1/4      1      auto   on      up               swl    F       4    --
        #fc1/5      1      auto   on      up               swl    F       4    --
        #fc1/6      1      auto   on      up               swl    F       4    --
        #fc1/7      1      auto   on      up               swl    F       4    --
        #fc1/8      1      auto   on      trunking         c1550  TE      4    --
        #fc1/9      1      auto   on      up               swl    F       2    --
        #fc1/10     1      auto   on      up               swl    F       2    --
        #fc1/11     1      auto   on      up               swl    F       2    --
        #fc1/12     1      auto   on      up               swl    F       2    --
        #fc1/13     1      auto   on      up               swl    F       4    --
        #fc1/14     1      auto   on      up               swl    F       4    --
        #fc1/15     1      auto   on      notConnected     swl    --           --
        #fc1/16     1      auto   on      notConnected     swl    --           --
        #fc1/17     1      auto   on      notConnected     swl    --           --
        #fc1/18     1      auto   on      notConnected     swl    --           --
        #fc1/19     1      auto   on      notConnected     swl    --           --
        #fc1/20     1      auto   on      notConnected     swl    --           --
        #fc1/21     1      auto   on      notConnected     swl    --           --
        #fc1/22     1      auto   on      notConnected     swl    --           --
        #fc1/23     1      auto   on      sfpAbsent        --     --           --
        #fc1/24     1      auto   on      notConnected     swl    --           --
        #
        #-------------------------------------------------------------------------------
        #Interface          Status                            Speed
        #                                                    (Gbps)
        #-------------------------------------------------------------------------------
        #sup-fc0            up                                1
        #
        #-------------------------------------------------------------------------------
        #Interface               Status     IP Address         Speed     MTU
        #-------------------------------------------------------------------------------
        #mgmt0                    up         10.97.103.234/24   100 Mbps 1500


        my $i = 0;
        foreach $line (@info)  {
            my $port_type = (split (/\s+/, $line))[6];
            # look for lines for the switch interconnect
            next unless (($port_type eq "E") || ($port_type eq "TE"));
            $port[$i] = (split (/\s+/, $line))[0];
            $length_port_name = length($port[$i]);
            # extract switch interconnect port info
            $port[$i] = substr($port[$i], 2, $length_port_name);
            ++$i;
        }

        $i = 0;

        # show interface fc1/8 sample output:

        # fc1/8 is trunking
        #     Hardware is Fibre Channel, SFP is CWDM-1550
        #     Port WWN is 20:08:00:0d:ec:47:e8:c0
        #     Peer port WWN is 20:10:00:0d:ec:47:bb:00
        #     Admin port mode is auto, trunk mode is on

        # Parse for and extract Port and Peer Port WWN's for each
        # switch interconnect port
        foreach my $port (@port) {
            Tharn::say($self->{conn}, "terminal length 0", -swallow, '-timeout', 60);
            (undef, undef, $after) = Tharn::say($self->{conn}, "show interface fc $port",
                                                  -swallow, '-timeout', 60);
            $after =~ s/\r//;
            my @info = split(/\n/, $after);
            foreach $line (@info)  {
                next unless ($line =~ "Port WWN is");
                $line =~ s/^\s+//g;
                $port_WWN[$i] = (split (/\s+/, $line))[3];
            }

            foreach $line (@info)  {
                next unless ($line =~ "Peer port WWN is");
                $line =~ s/^\s+//g;
                $peer_port_WWN[$i] = (split (/\s+/, $line))[4];
            }

            my %eport_info = ();
            $eport_info{"port"} = $port[$i];
            $eport_info{"port_WWN"} = $port_WWN[$i];
            $eport_info{"peer_port_WWN"} = $peer_port_WWN[$i];
            ++$i;
            push (@eportlist, \%eport_info);
         }
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Brocade switch
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    } elsif ($hosttype =~ /switch-brocade/i) {

        # version command sample output
        #Kernel:     2.4.19
        #Fabric OS:  v5.2.2
        #Made on:    Thu Jun 7 22:55:46 2007
        #Flash:      Thu Sep 20 15:21:08 2007
        #BootProm:   4.5.3

        (undef, undef, $after) = Tharn::say($self->{conn}, "version", -swallow,
                                                       '-timeout', 60);

        # parse the firmware version
        $after =~ /Fabric OS:\s+v([3-9])/;
        $version = "v" . $1;

        # switchShow command sample (abridged) output
        #switchName:     vns_b200e_8
        #switchType:     34.0
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Subordinate
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:05:1e:04:14:73
        #zoning:         ON (Metrocluster_V6030)
        #switchBeacon:   OFF
        #
        #Area Port Media Speed State     Proto
        #====================================
        #  0   0   id    AN   No_Sync
        #  1   1   id    N4   No_Light
        #  2   2   id    N4   No_Light
        #  3   3   id    N4   Online           F-Port  50:0a:09:83:00:01:71:9d
        #  4   4   id    N4   Online           F-Port  50:0a:09:81:00:01:71:9d
        #  8   8   id    N2   Online           LD E-Port  10:00:00:05:1e:04:15:29 "vns_b200e_9" (upstream)
        #  9   9   id    N2   Online           F-Port  50:06:0e:80:03:4f:e7:19
        # 10  10   id    N2   Online           F-Port  50:06:0e:80:03:4f:e7:0a
        # 15  15   --    N4   No_Module


        (undef, undef, $after) = Tharn::say($self->{conn}, "switchshow", -swallow,
                                                        '-timeout', 60);

        # handle output formatting differently between firmware versions
        if ($version eq "v3") {
           @info = split(/\n\r/, $after);
        } else {
           @info = split(/\r\n/, $after);
        }
        my $i = 0;

        # Extract the switch interconnect port info. Need to parse
        # differently between fw versions
        foreach $line (@info) {
           next unless ($line =~ "E-Port");
           $line =~ s/^\s+//g;
           if ($version eq "v3") {
               $port[$i] = (split (/\s+/, $line))[1];
               chop($port[$i]);
           } else {
               $port[$i] = (split (/\s+/, $line))[0];
           }
           ++$i;
        }

        my $j = 0;

        #portshow sample output (truncated):
        #
        #vns_b200e_8:admin> portshow 8
        #portName:
        #portHealth: HEALTHY
        #
        #Authentication: None
        #portDisableReason: None
        #portCFlags: 0x1
        #portFlags: 0x907         PRESENT ACTIVE E_PORT G_PORT U_PORT LOGICAL_ONLINE LOGIN
        #portType:  11.0
        #POD Port: Port is licensed
        #portState: 1    Online
        #portPhys:  6    In_Sync
        #portScn:   16   E_Port
        #port generation number:    4482
        #portId:    010800
        #portIfId:    43020008
        #portWwn:   20:08:00:05:1e:04:14:73
        #portWwn of device(s) connected:
        #
        #Distance:  auto (desired = 30 Km, actual = 31 Km)
        #portSpeed: N2Gbps

        foreach my $port (@port) {
            (undef, undef, $after) = Tharn::say($self->{conn}, "portshow $port",
                                            -swallow, '-timeout', 60);
            # handle output formatting differently between firmware versions
            if ($version eq "v3") {
               @info = split(/\n\r/, $after);
            } else {
               @info = split(/\r\n/, $after);
            }

            # parse for line with port wwn info
            foreach $line (@info) {
                next unless ($line =~ /portWwn:/i);
                $line =~ s/^\s+//g;
                $port_WWN[$j] = (split (/\s+/, $line))[1];
            }

            # Parse for line with "portWwn of device(s) connected:"
            # We then parse its followup lines to extract the Peer Port wwn
            # Need to parse differently between firmware versions
            for (my $i=0; $i<@info; ++$i) {
                $line = undef;
                $line = $info[$i];
                next unless ($line =~ /connected:/i);
                if ($version eq "v3" ) {
                    $line =~ s/^\s+//g;
                    $peer_port_WWN[$j] = (split (/\s+/, $line))[4];
                } else {
                    $line = undef;
                    $line = $info[$i + 1];
                    $line =~ s/^\s+//g;
                    $peer_port_WWN[$j] = (split (/\s+/, $line))[0];
                }
            }
            my %eport_info = ();
            $eport_info{"port"} = $port[$j];
            $eport_info{"port_WWN"} = $port_WWN[$j];
            $eport_info{"peer_port_WWN"} = $peer_port_WWN[$j];
            ++$j;
            push (@eportlist, \%eport_info);
        }
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Mcdata switch
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    } elsif  ($hosttype =~ /switch-mcdata/i) {
        my $director;
        my $peer_switch_wwn;

        #show system sample output:
        #Name:                vns-m4400-2
        #Description:         Fibre Channel Switch
        #Contact:             End User Contact (please configure)
        #Location:            End User Premise (please configure)
        #Date/Time:           10/08/2007 22:01:58
        #Serial Number:       4G00232T
        #Type Number:         004400
        #Model Name:          Sphereon 4400
        #Model Number:        001
        #EC Level:            1060918
        #Firmware Version:    09.06.00 22
        #Beaconing:           Disabled

        (undef, undef, $after) = Tharn::say($self->{conn}, "show system", -swallow,
                                                      '-timeout', 60);
        if ($after =~ /director/i) {
            $director = 1;
        } else {
            $director = 0;
        }

        # show port status sample output (abridged)
        #
        #Port  State               Type    Attached WWN             Beaconing  Reason
        #----  ------------------  ------  -----------------------  ---------  ------
        #0     Online              fPort   50:06:04:82:CB:1B:CE:3C  Disabled
        #1     Online              ePort   50:0A:09:82:00:01:43:88  Disabled
        #2     No Light            gxPort  None                     Disabled
        #3     No Light            gxPort  None                     Disabled
        #4     Not Installed       gxPort  None                     Disabled   54
        #5     Not Installed       gxPort  None                     Disabled   54
        #13    Inactive            gxPort  None                     Disabled   46
        #14    Inactive            gxPort  None                     Disabled   46
        #15    Inactive            gxPort  None                     Disabled   46

        (undef, undef, $after) = Tharn::say($self->{conn}, "show port status", -swallow,
                                                      '-timeout', 60);
        $after =~ s/\r//;
        @info = split(/\n/, $after);
        my $i = 0;
        foreach $line (@info) {
            next unless (($line =~ /Online/) && ($line =~ /ePort/));
            $port[$i] = (split(/\s+/, $line))[0];
            ++$i;
        }

        $i = 0;

        # show port into sample output (abridged)
        #
        #Port  WWN                      OpSpeed           SpeedCap
        #----  -----------------------  ----------------  -----------------------------------
        #0     20:00:08:00:88:E3:61:74  2G                1 2 4 Neg
        #1     20:01:08:00:88:E3:61:74  2G                1 2 4 Neg
        #2     20:02:08:00:88:E3:61:74  Not Established   1 2 4 Neg
        #3     20:03:08:00:88:E3:61:74  Not Established   1 2 4 Neg
        #13    20:0D:08:00:88:E3:61:74  N/A               1 2 4 Neg
        #14    20:0E:08:00:88:E3:61:74  N/A               1 2 4 Neg
        #15    20:0F:08:00:88:E3:61:74  N/A               1 2 4 Neg


        # show switch sample output (abridged)
        #
        #State:                  Online
        #R_A_TOV:                100
        #E_D_TOV:                20
        #Preferred Domain ID:    1
        #Domain ID Offset:       96
        #Switch Priority:        Default
        #Rerouting Delay:        Disabled
        #Interop Mode:           McDATA Fabric 1.0
        #Active Domain ID:       1
        #World Wide Name:        10:00:08:00:88:E3:61:74
        #Insistent Domain ID:    Disabled
        #Domain RSCN:            Disabled

        foreach my $port (@port) {
            (undef, undef, $after) = Tharn::say($self->{conn}, "show port info", -swallow,
                                                     '-timeout', 60);
            @info = split(/\r\n/, $after);
            # extract the wwn info for each switch interconnect port
            $line = $info[$port + 2];
            $port_WWN[$i] = (split (/\s+/, $line))[1];
            (undef, undef, $after) = Tharn::say($self->{conn}, "show switch", -swallow,
                                                     '-timeout', 60);
            @info = split(/\r\n/, $after);
            # extract the peer port wwn info
            foreach $line (@info) {
                next unless ($line =~ /World Wide Name:/);
                $peer_port_WWN[$i] = (split(/\s+/, $line))[3]; #my switch WWN
            }
            my %eport_info = ();
            $eport_info{"port"} = $port[$i];
            $eport_info{"port_WWN"} = $port_WWN[$i];
            $eport_info{"peer_port_WWN"} = $peer_port_WWN[$i];
            ++$i;
            push (@eportlist, \%eport_info);
        }
    } else {
        $self->{error} = $ERR_STATUS;
        push(@{$self->{errmsg}}, "ERROR: unsupported switch type: ($self->{hosttype}");
        return $ERR_STATUS;
    }

    return (@eportlist);
}

## @name ::so_switch_disable (;$)
##
## @description: Disables all ports of a switch
## Errors will be reported in the switch object's error flag and message will
## contain port number that failed to be properly enabled.
##
## @synopsis
## $status = VSeries::switch_object->so_switch_disable($no_verify_flag)
##
## Note: Cisco switch section supports Vsan 1 only
##
## @arg $no_verify Optional: bypass port state change verification
##                           and exit quickly after issuing disable
##                           command

sub so_switch_disable (;$) {

    my $self = shift;
    my ($no_verify) = @_;
    my $hostname = $self->get_name();
    my $hosttype = $self->get_host_type();
    my $line = "";
    my @cisco_port = ();
    my $length_port_name = "";
    my ($after, $prompt) = "";
    my $i = 0;
    my @info = ();
    my $fc_port = "";
    my $port_state = "";
    my $ld_state = "";
    my $port_not_disabled = 0;
    my $retry_port_verify = 0;
    my $nochange = 0;

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Cisco switch
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if ($hosttype =~ /switch-cisco/i) {

        # Sample output to parse, showing different fields

        #Interface  Vsan   Admin  Admin   Status          SFP    Oper  Oper   Port
        #                  Mode   Trunk                          Mode  Speed  Channel
        #                         Mode                                 (Gbps)
        #-------------------------------------------------------------------------------
        #fc1/7      1      auto   on      down             swl    --           --
        #fc1/8      1      auto   on      down             c1550  --           --
        #fc1/9      1      auto   on      down             swl    --           --
        #fc1/10     1      auto   on      down             swl    --           --
        #fc1/19     1      auto   on      down             swl    --           --
        #fc1/20     1      auto   on      down             swl    --           --
        #fc1/21     1      auto   on      down             swl    --           --
        #fc1/22     1      auto   on      outOfServc       swl    --           --
        #fc1/23     1      auto   on      sfpAbsent        --     --           --
        #fc1/24     1      auto   on      down             swl    --           --


        Tharn::say($self->{conn}, "terminal length 0", -swallow, '-timeout', 60);

        # Fix for BURT 490772.
        # The output of show interface might be prompted by more. Pipe to no-more to avoid more.

        (undef, undef, $after) = Tharn::say($self->{conn}, "show interface brief | no-more", -swallow, '-timeout', 60);

        $after =~ s/\r//;
        @info = split(/\n/, $after);

        my %blade_ports_hash = (); # Stores key/value pair of blade and number_of_ports.
        my %blade_port_id = (); # Stores starting port-id's of each blade.

        foreach my $line (@info)  {
            # Determine total number of ports by parsing for lines
            # containing leading fc characters
            if ($line =~ /^fc(\d+)\/(\d+)/) {
               $blade_port_id{$1} = $2 if (not exists $blade_port_id{$1});
               $blade_ports_hash{$1}++;
            }
        }

        # Sequence of Cisco commands to generate a scope containing all ports and
        # then disabling all ports within that scope
        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-noprompt',  -swallow, '-timeout', 30);

        foreach my $blade_port(keys %blade_ports_hash) {
           (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "int fc$blade_port/$blade_port_id{$blade_port} - $blade_ports_hash{$blade_port}", '-noprompt', -swallow, '-timeout', 30);
           (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "shutdown", '-noprompt',  -swallow, '-timeout', 30);
        }

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "end", '-noprompt', -swallow, '-timeout', 30);

        # Check if param is set to bypass port state changes and exit quickly
        if ( defined $no_verify && $no_verify ) {
            return $GOOD_STATUS;
        }

        # Loop until we verify all ports have been disabled or until
        # we exceed retry limit
        while ( ! $nochange ) {

        # Fix for BURT 490772.
        # The output of show interface might be prompted by more. Pipe to no-more to avoid more.

            (undef, undef, $after) = Tharn::say($self->{conn}, "show interface brief | no-more", -swallow, '-timeout', 60);
            @info = split(/\n/, $after);

            foreach my $line (@info)  {
                # Parse only lines containing fc module port information
                if ($line !~ /^fc/ ) { next; }
                # Pull out the 1st and 5th column data fields for each port
                $fc_port = (split (/\s+/, $line))[0];
                $port_state = (split (/\s+/, $line))[4];

                # Check if port has been disabled or is in acceptable alternate state
                if ( $port_state !~ /(down|sfpAbsent|outOfServc|notConnected|diagFailure)/i ) {
                    $port_not_disabled = 1;
                    $retry_port_verify++;
                    last;
               }
            }

            # Check if ports have changed state or if we have exceeded max number of retries
            if ( ! $port_not_disabled ) {
                $nochange = 1;  # All usuable ports are in disabled states, we're done
            } elsif ( $port_not_disabled && ($retry_port_verify > $MAX_PORT_STATUS_RETRIES) ) {
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: Cisco switch port $fc_port still enabled");
                return ($ERR_STATUS);
            } else { # Sleep for a bit and recheck port status
                $port_not_disabled = 0;
                 sleep $PORT_STATUS_RETRY_WAIT;
            }

        } # end while

        return $GOOD_STATUS;

    } elsif ($hosttype =~ /switch-mcdata/i) {

        # show switch sample output (abridged)
        #
        #State:                  Offline
        #R_A_TOV:                100
        #E_D_TOV:                20
        #Preferred Domain ID:    1
        #Domain ID Offset:       96
        #Switch Priority:        Default
        #Rerouting Delay:        Disabled
        #Interop Mode:           McDATA Fabric 1.0
        #Active Domain ID:       1
        #World Wide Name:        10:00:08:00:88:E3:61:74
        #Insistent Domain ID:    Disabled
        #Domain RSCN:            Disabled

        #Root> show port status
        #Port  State               Type    Attached WWN             Beaconing  Reason
        #----  ------------------  ------  -----------------------  ---------  ------
        #0     Offline             gxPort  None                     Disabled
        #1     Offline             gxPort  None                     Disabled
        #2     Offline             gxPort  None                     Disabled
        #3     Offline             gxPort  None                     Disabled
        #4     Not Installed       gxPort  None                     Disabled   54
        #5     Not Installed       gxPort  None                     Disabled   54
        #6     Not Installed       gxPort  None                     Disabled   54
        #7     Not Installed       gxPort  None                     Disabled   54
        #8     Inactive            gxPort  None                     Disabled   46
        #9     Inactive            gxPort  None                     Disabled   46
        #10    Inactive            gxPort  None                     Disabled   46

        (undef, undef, $after) = Tharn::say($self->{conn}, "maint system setOnlineState 0", -swallow, '-timeout', 60);

        # Check if param is set to bypass port state changes and exit quickly
        if ( defined $no_verify && $no_verify ) {
            return $GOOD_STATUS;
        }

        # Loop until we verify all ports have been enabled or until
        # we exceed retry limit
        while ( ! $nochange ) {

            # Check port status of all ports
            (undef, undef, $after) = Tharn::say($self->{conn}, "show switch", -swallow, '-timeout', 60);

            $after =~ s/\r//;
            @info = split(/\n/, $after);

            foreach $line (@info) {
                next unless ($line =~ /^State:/);
                if ( $line !~ /State:\s+Offline/ ) {
                    $port_not_disabled = 1;
                    $retry_port_verify++;
                    last;
                }
            }

            # Check if ports have changed state of if we have exceeed max number of retries
            if ( ! $port_not_disabled ) {
                $nochange = 1;  # Switch is in Offline state, we're done
            } elsif ( $port_not_disabled && ($retry_port_verify > $MAX_PORT_STATUS_RETRIES) ) {
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: Mcdata switch port not disabled as expected: $line");
                return ($ERR_STATUS);
            } else { # Sleep for a bit and recheck port status
                $port_not_disabled = 0;
                sleep $PORT_STATUS_RETRY_WAIT;
            }

        } # end while

        return $GOOD_STATUS

    } elsif ($hosttype =~ /switch-brocade/i) {

        # sample switchShow output (V5)
        #
        #switchName:     vrs_br200e_1
        #switchType:     34.0
        #switchState:    Offline
        #switchMode:     Native
        #switchRole:     Disabled
        #switchDomain:   1 (unconfirmed)
        #switchId:       fffc01
        #switchWwn:      10:00:00:05:1e:03:aa:e8
        #zoning:         ON (vra_v3050_cfg1)
        #switchBeacon:   OFF
        #Area Port Media Speed State     Proto
        #=====================================
        #  1   1   id    N2   No_Sync          Disabled
        # 22  22   id    N2   No_Sync          Disabled
        # 23  23   id    N4   No_Sync          Disabled
        # 24  24   id    N2   No_Sync          LD Disabled
        # 25  25   id    N4   No_Light         Disabled
        # 26  26   id    N4   No_Light         Disabled

        # sample switchShow output (V3)
        #
        #switchName:     vss_b3800_7
        #switchType:     9.2
        #switchState:    Offline
        #switchMode:     Native
        #switchRole:     Disabled
        #switchDomain:   1 (unconfirmed)
        #switchId:       fffc01
        #switchWwn:      10:00:00:60:69:51:f1:e4
        #switchBeacon:   OFF
        #Zoning:         ON (Automation1)
        #port  0: id N2 In_Sync        Disabled
        #port  1: id N2 In_Sync        Disabled
        #port  2: id N2 In_Sync        Disabled
        #port  3: id N2 In_Sync        Disabled

        (undef, undef, $after) = Tharn::say($self->{conn}, "switchDisable", -swallow, '-timeout', 60);

        # Check if param is set to bypass port state changes and exit quickly
        if ( defined $no_verify && $no_verify ) {
            return $GOOD_STATUS;
        }

        # Loop until we verify switchState is Offline or until
        # we exceed retry limit
        while ( ! $nochange ) {

            (undef, undef, $after) = Tharn::say($self->{conn}, "switchShow", -swallow, '-timeout', 60);

            $after =~ s/\r//;
            @info = split(/\n/, $after);

            foreach $line (@info)  {

                # Verify the switchState is set to Offline, if not then we'll repoll
                next unless ($line =~ /^switchState:/);
                if ( $line !~ /^switchState:\s+Offline/ ) {
                    $port_not_disabled = 1;
                    $retry_port_verify++;
                    last;
                }
            }

            # Check if ports have changed state of if we have exceeed max number of retries
            if ( ! $port_not_disabled ) {
                $nochange = 1;  # Switch is in Offline state, we're done
            } elsif ( $port_not_disabled && ($retry_port_verify > $MAX_PORT_STATUS_RETRIES) ) {
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: Brocade switch port $fc_port still enabled");
                return ($ERR_STATUS);
            } else { # Sleep for a bit and recheck port status
                $port_not_disabled = 0;
                sleep $PORT_STATUS_RETRY_WAIT;
            }

        } # end while

        return $GOOD_STATUS;

    } else {
        $self->{error} = $ERR_STATUS;
        push(@{$self->{errmsg}}, "ERROR: unsupported switch type: $hosttype");
        return $ERR_STATUS;
    }
}

## @name ::so_switch_enable(;$)
##
## @description: Enables all ports of an entire switch. The method will verify
## that each port connected to a device has reached an expected enable/online
## state.
## Errors will be reported in the switch object's error flag and message will
## contain port number that failed to be properly enabled.
##
## @synopsis
## $status = VSeries::switch_object->so_switch_enable($no_verify_flag)
##
## Notes:
## 1) Cisco switch section supports VSAN 1 only
##
## Enabled states are determined logically by this criteria:
##
## Cisco: connected port enabled states: "up", "trunking"
##        other enabled states: "notConnected", "sfpAbsent"
##        disabled state: "down"
## McData: connected port enabled states: "Online"
##         other enabled states: "Not Installed", "No Light",
##                               "Inactive"
##         disabled state: "Offline"
## Brocade: connected port enabled states: "Online"
##          other enabled states: "No_Sync", "No_Light",
##                                "No_Module"
##          disabled state: "Disabled" appears in Proto column
##
## @arg $no_verify Optional: bypass port state change verification
##                           and exit quickly after issuing enable
##                           command

sub so_switch_enable (;$) {

    my $self = shift;
    my ($no_verify) = @_;
    my $hostname = $self->get_name();
    my $hosttype = $self->get_host_type();
    my $line = "";
    my @cisco_port = ();
    my $length_port_name = "";
    my ($after, $prompt) = "";
    my @info = ();
    my $fc_port = "";
    my $port_state = "";
    my $port_status = "";
    my $port_status_opt_second_field = ""; # McData specific (ex. No Light)
    my $port_ld_status = "";
    my $ld_state = "";
    my $i = 0;
    my $port_not_enabled = 0;
    my $retry_port_verify = 0;
    my $nochange = 0;
    my $end_of_header = "";
    my @portlist = ();

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Cisco switch
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if ($hosttype =~ /switch-cisco/i) {

        # show interface brief sample output

        #-------------------------------------------------------------------------------
        #Interface  Vsan   Admin  Admin   Status          SFP    Oper  Oper   Port
        #                   Mode  Trunk                          Mode  Speed  Channel
        #                         Mode                                (Gbps)
        #-------------------------------------------------------------------------------
        #fc1/1      1      auto   on      notConnected     swl    --           --
        #fc1/2      1      auto   on      notConnected     swl    --           --
        #fc1/3      1      auto   on      up               swl    F       4    --
        #fc1/4      1      auto   on      up               swl    F       4    --
        #fc1/5      1      auto   on      notConnected     swl    --           --
        #fc1/6      1      auto   on      up               swl    F       4    --
        #fc1/7      1      auto   on      up               swl    F       4    --
        #fc1/8      1      auto   on      trunking         c1550  TE      4    --
        #fc1/9      1      auto   on      up               swl    F       2    --
        #fc1/10     1      auto   on      up               swl    F       2    --
        #fc1/11     1      auto   on      up               swl    F       2    --
        #fc1/12     1      auto   on      up               swl    F       2    --
        #fc1/13     1      auto   on      up               swl    F       4    --
        #fc1/14     1      auto   on      up               swl    F       4    --
        #fc1/15     1      auto   on      up               swl    F       2    --
        #fc1/16     1      auto   on      notConnected     swl    --           --
        #fc1/17     1      auto   on      notConnected     swl    --           --
        #fc1/18     1      auto   on      notConnected     swl    --           --
        #fc1/19     1      auto   on      notConnected     swl    --           --
        #fc1/20     1      auto   on      notConnected     swl    --           --
        #fc1/21     1      auto   on      outOfServc       swl    --           --
        #fc1/22     1      auto   on      notConnected     swl    --           --
        #fc1/23     1      auto   on      sfpAbsent        --     --           --
        #fc1/24     1      auto   on      notConnected     swl    --           --

        Tharn::say($self->{conn}, "terminal length 0", -swallow, '-timeout', 60);

        # Fix for BURT 490772.
        # The output of show interface might be prompted by more. Pipe to no-more to avoid more.

        (undef, undef, $after) = Tharn::say($self->{conn}, "show interface brief | no-more", -swallow, '-timeout', 60);

        $after =~ s/\r//;
        @info = split(/\n/, $after);

        my %blade_ports_hash = (); # Stores key/value pair of blade and number_of_ports.
        my %blade_port_id = (); # Stores starting port-id's of each blade.

        foreach my $line (@info)  {
            # determine total number of ports by parsing for lines
            # containing leading fc characters
            if ($line =~ /^fc(\d+)\/(\d+)/) {
               $blade_port_id{$1} = $2 if (not exists $blade_port_id{$1});
               $blade_ports_hash{$1}++;
            }
        }

        # Sequence of Cisco commands to generate a scope containing all ports and
        # then disabling all ports within that scope
        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-noprompt', -swallow, '-timeout', 30);

        foreach my $blade_port(keys %blade_ports_hash) {
           (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "int fc$blade_port/$blade_port_id{$blade_port} - $blade_ports_hash{$blade_port}", '-noprompt', -swallow, '-timeout', 30);
           (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "no shutdown", '-noprompt', -swallow, '-timeout', 30);
        }

        (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "end", '-noprompt', -swallow, '-timeout', 30);

        # Check if param is set to bypass port state changes and exit quickly
        if ( defined $no_verify && $no_verify ) {
            return $GOOD_STATUS;
        }

        # Verify all ports are enabled: enabled ports can have many different
        # steady state enabled states(ie. up, notConnected, sfpAbsent, trunking)
        while ( ! $nochange ) {

        # Fix for BURT 490772.
        # The output of show interface might be prompted by more. Pipe to no-more to avoid more.

            (undef, undef, $after) = Tharn::say($self->{conn}, "show interface brief | no-more", -swallow, '-timeout', 60);
            $after =~ s/\r//;
            @info = split(/\n/, $after);

            foreach my $line (@info)  {

                # Parse only lines containing fc module port information
                if ($line =~ /^fc/ ) {
                    # pull out the 1st and 5th column data fields for each port
                    $fc_port = (split (/\s+/, $line))[0];
                    $port_status = (split (/\s+/, $line))[4];

                    # Check if Cisco port has reached enabled state or is in acceptable alternate state
                    if ( $port_status =~ /^(up|trunking|notConnected|sfpAbsent|outOfServc)/i ) {
                        next;
                    } else {
                        $port_not_enabled = 1;
                        $retry_port_verify++;
                        last;
                    }
                }
            }

            if ( ! $port_not_enabled ) {
                $nochange = 1;  # All ports are verified to be up or in
                                # valid enabled states, we're done
            } elsif ( $port_not_enabled && ($retry_port_verify > $MAX_PORT_STATUS_RETRIES) ) {
                # Check if we have exceeded max number of retries
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: Cisco switch port $fc_port is still disabled");
                return ($ERR_STATUS);
            } else {
                # Sleep for a bit and recheck port status
                $port_not_enabled = 0;
                sleep $PORT_STATUS_RETRY_WAIT;
            }

        } # end while

        return $GOOD_STATUS;

    } elsif ($hosttype =~ /switch-mcdata/i) {

        # show port status sample output
        # ------------------------------

        #Port  State               Type    Attached WWN             Beaconing  Reason
        #----  ------------------  ------  -----------------------  ---------  ------
        #0     Online              fPort   50:06:04:82:CB:1B:CE:3C  Disabled
        #1     Online              fPort   50:0A:09:82:00:01:43:88  Disabled
        #4     Not Installed       gxPort  None                     Disabled   54
        #5     Not Installed       gxPort  None                     Disabled   54
        #6     Not Installed       gxPort  None                     Disabled   54
        #7     Not Installed       gxPort  None                     Disabled   54
        #  -------here are a couple of sample ports that would be disabled: ---------
        #8     Offline             gxPort  None                     Disabled
        #9     Offline             gxPort  None                     Disabled
        #10    No Light            gxPort  None                     Disabled   46
        #11    No Light            gxPort  None                     Disabled   46
        #12    Inactive            gxPort  None                     Disabled   46
        #13    Inactive            gxPort  None                     Disabled   46
        #14    Inactive            gxPort  None                     Disabled   46
        #15    Inactive            gxPort  None                     Disabled   46
        #
        #46: No key enabled
        #54: Optics not installed


        # show switch sample output
        # -------------------------

        #State:                  Online
        #R_A_TOV:                100
        #E_D_TOV:                20
        #Preferred Domain ID:    1
        #Domain ID Offset:       96
        #Switch Priority:        Default
        #Rerouting Delay:        Disabled
        #Interop Mode:           McDATA Fabric 1.0
        #Active Domain ID:       1
        #World Wide Name:        10:00:08:00:88:E3:61:74
        #Insistent Domain ID:    Disabled
        #Domain RSCN:            Disabled

        (undef, undef, $after) = Tharn::say($self->{conn}, "maint system setOnlineState 1", -swallow, '-timeout', 60);

        # Check if param is set to bypass port state changes and exit quickly
        if ( defined $no_verify && $no_verify ) {
            return $GOOD_STATUS;
        }

        # Loop until we verify all ports have been enabled or until
        # we exceed retry limit
        while ( ! $nochange ) {

            (undef, undef, $after) = Tharn::say($self->{conn}, "show switch", -swallow, '-timeout', 60);
            $after =~ s/\r//;
            @info = split(/\n/, $after);

            # Verify the switchState is set to Online
            foreach $line (@info) {
                next unless ($line =~ /^State:/);
                if ( $line !~ /State:\s+Online/ ) {
                    # Set $fc_port to SWITCH indicating switch online
                    # is not yet seen
                    $fc_port = "SWITCH";
                    $port_not_enabled = 1;
                    $retry_port_verify++;
                    last;
                }
            }

            # Verify the each port state has reach an acceptable enabled steady state
            (undef, undef, $after) = Tharn::say($self->{conn}, "show port status", -swallow, '-timeout', 60);

            $after =~ s/\r//;
            @info = split(/\n/, $after);

            foreach $line (@info) {

                # Skip over non-port lines
                if ( $line =~ /^Port/ || $line =~ /^---/ ) { next; }

                # If we hit a blank line, there is no more port info, so end loop
                if ( $line =~ /^\s+$/ ) { last; }

                # Parse only for line with specific port info
                if ( $line =~ /^(\w+)\s+(\w+)\s+(\w+)\s+\w+\s+\w+/  ) {

                    $fc_port = $1;
                    $port_status = $2;
                    $port_status_opt_second_field = $3;

                    # Check if McData port has reached enabled state
                    if ( $port_status eq "Online" || $port_status eq "Inactive" ||
                          ($port_status eq "No" && $port_status_opt_second_field eq "Light" ) ||
                            ($port_status eq "Not" && $port_status_opt_second_field eq "Installed" ) ) {
                        next;
                    } else {
                        $port_not_enabled = 1;
                        $retry_port_verify++;
                        last;
                    }
                }
            }

            if ( ! $port_not_enabled ) {
                $nochange = 1;  # Switch is in Online state, we're done
            } elsif ( $port_not_enabled && ($retry_port_verify > $MAX_PORT_STATUS_RETRIES) ) {
                # Check if we have exceeed max number of retries
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: Mcdata switch port $fc_port still disabled");
                return ($ERR_STATUS);
            } else {
                # Sleep for a bit and recheck port status
                $port_not_enabled = 0;
                sleep $PORT_STATUS_RETRY_WAIT;
            }

        } # end while

        return $GOOD_STATUS

    } elsif ($hosttype =~ /switch-brocade/i) {

        # switchShow sample output (v4,5)
        #
        #switchName:     vns_b200e_8
        #switchType:     34.0
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:05:1e:04:14:73
        #zoning:         ON (Metrocluster_V6030)
        #switchBeacon:   OFF
        #
        #Area Port Media Speed State     Proto
        #=====================================
        #  0   0   id    AN   No_Sync
        #  1   1   id    N4   No_Light
        #  2   2   id    N4   No_Light
        #  3   3   id    N4   Online           F-Port  50:0a:09:83:00:01:71:9d
        #  4   4   id    N4   Online           F-Port  50:0a:09:81:00:01:71:9d
        #  5   5   id    N4   In_Sync          Disabled
        #  8   8   id    N2   Online           LD E-Port  10:00:00:05:1e:04:15:29 "vns_b200e_9" (upstream)
        #                 - or same interconnect port when still disabled...... -
        #  8   8   id    N2   No_Sync          LD Disabled
        #  9   9   id    N4   In_Sync          Disabled (Switch not ready for F or L ports)
        # 21  21   id    N2   Online           F-Port  50:0a:09:82:00:00:de:9d
        # 22  22   id    N2   Online           F-Port  50:0a:09:82:00:01:91:cf
        # 23  23   --    N4   No_Module

        # switchShow sample output (v3)
        #
        #switchName:     vss_b3800_7
        #switchType:     9.2
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:60:69:51:f1:e4
        #switchBeacon:   OFF
        #Zoning:         ON (Automation1)
        #port  0: id N2 Port_Flt  (transitional state!!!)
        #port  1: id N1 In_Sync        Disabled (transitional state !!)
        #port  2: id N2 No_Sync
        #port  3: id N2 Online         F-Port 50:06:04:82:cb:1b:ce:2c
        #port  4: id N2 No_Light
        #port  5: id N2 No_Light       Disabled
        #port  6: id N2 No_Light
        #port  7: id N2 No_Light
        #port  8: id N2 No_Light
        #port  9: id N2 No_Light
        #port 10: id N2 No_Light
        #port 11: id N2 No_Light
        #port 12: id N2 No_Light

        #BURT 455978: extra address column    
        #switchShow sample output (v6.4.0a|v6.2.1)
        #
        #Index Port Address Media Speed State     Proto
        #==============================================
        #0   0   041700   id    N8   Online      FC  E-Port  10:00:00:05:1e:d3:ca:00 "vgbrdcxs84"
        #1   1   041500   id    N4   Online      FC  F-Port  21:00:00:24:ff:30:42:1d
        #2   2   041300   id    N4   Online      FC  F-Port  50:0a:09:81:00:02:59:e7
        #3   3   041100   id    N4   Online      FC  F-Port  50:0a:09:83:00:02:59:e7
        #4   4   041600   id    N8   Online      FC  F-Port  20:49:00:0b:5d:6a:02:0e
        #5   5   041400   id    N8   Online      FC  F-Port  21:40:00:0b:5d:6a:02:0e
        #6   6   041200   id    N8   Online      FC  E-Port  10:00:00:05:1e:d3:ca:00 "vgbrdcxs84" (upstream)
        #7   7   041000   id    N8   No_Light    FC
        #8   8   040f00   id    N8   No_Light    FC
        #9   9   040d00   id    N8   No_Light    FC
        #10  10   040b00   id    N8   No_Light    FC
        #11  11   040900   id    N8   No_Light    FC
        #12  12   040e00   id    N8   No_Light    FC
        #13  13   040c00   id    N8   No_Light    FC
        #14  14   040a00   id    N8   No_Light    FC
        #15  15   040800   id    N8   No_Light    FC
        #16  16   040700   id    N8   No_Light    FC
        #17  17   040500   id    N8   No_Light    FC
        #18  18   040300   id    N8   No_Light    FC
        #19  19   040100   id    N8   No_Light    FC
        #20  20   040600   id    N8   No_Light    FC
        #21  21   040400   id    N8   No_Light    FC
        #22  22   040200   id    N8   No_Light    FC
        #23  23   040000   id    N8   No_Light    FC
        #vgbr300s78:admin>


        # For Brocade switches, output is formatted differently depending
        # on firmware rev, so we need to determine the switch fw rev first

        (undef, undef, $after) = Tharn::say($self->{conn}, "version", -swallow,
                                                       '-timeout', 60);
        $after =~ /Fabric OS:\s+(v[3-9].*)/;
        my $version = $1;

        (undef, undef, $after) = Tharn::say($self->{conn}, "switchEnable", -swallow, '-timeout', 60);

        # Check if param is set to bypass port state changes and exit quickly
        if ( defined $no_verify && $no_verify ) {
            return $GOOD_STATUS;
        }

        # Loop until we verify all ports have been enabled or until
        # we exceed retry limit
        while ( ! $nochange ) {

            (undef, undef, $after) = Tharn::say($self->{conn}, "switchShow", -swallow, '-timeout', 60);
            $after =~ s/\r//;
            @info = split(/\n/, $after);

            foreach $line (@info)  {

                # Verify the switchState is set to Online, if not then we'll repoll
                if ( $line =~ /^switchState:/ && $line !~ /^switchState:\s+Online/ ) {
                    # Set $fc_port to SWITCH indicating switch online
                    # is not yet seen and we need to retry
                    $fc_port = "SWITCH";
                    $port_not_enabled = 1;
                    $retry_port_verify++;
                    last;
                }
            }

            # Now we need to verify each port state. The output format varies between
            # fw revs so we need to check the rev to decide how to parse command
            # output lines with port state

            if ( $version eq "v3" ) {

                (undef, undef, $after) = Tharn::say($self->{conn}, "switchShow", -swallow, '-timeout', 60);
                $after =~ s/\r//;
                @info = split(/\n/, $after);

                # Look for lines leading with 'port' and grab the
                # port number and its state
                foreach $line (@info)  {

                    $line =~ s/^\s+//g;
                    if ( $line =~ /^port/ ) {

                        $fc_port = (split (/\s+/, $line))[1];
                        # Remove trailing ":" char from port number
                        chop $fc_port;
                        $port_state =  (split (/\s+/, $line))[4];
                        $port_status =  (split (/\s+/, $line))[5];
                        # If port is a switch interconnect, we need to parse for an
                        # additional field for the port status
                        if ( defined $port_status && $port_status =~ /LD/ ) {
                            $port_ld_status =  (split (/\s+/, $line))[6];
                        }

                        # Check if Brocade V3 port has reached enabled state
                        if ( (($port_state eq "Online" || $port_state eq "No_Sync" || $port_state eq "No_Light" ||
                               $port_state eq "No_Module") && $port_status ne "Disabled" ) ||
                             ( $port_state eq "Online" && $port_status =~ /LD/ &&  $port_ld_status ne "Disabled")  ) {
                            next;
                        } else {
                            $port_not_enabled = 1;
                            $retry_port_verify++;
                            last;
                        }
                    }
                }      
            } else { # fw rev v4,5 and beyond

                (undef, undef, $after) = Tharn::say($self->{conn}, "switchShow", -swallow, '-timeout', 60);
                $after =~ s/\r//;
                @info = split(/\n/, $after);

                # Clear @portlist array and header flags in case we're looping and retrying
                # switchShow
                @portlist = ();
                $end_of_header = 0;
                
                # This bool variable is set to true if an "address" field is found
                my $address_field = 0;
                
                # Create @portlist subarray containing only the port info
                # output lines. We skip over all lines preceeding port info
                foreach $line (@info)  {

                    if ( $end_of_header )  {
                        push @portlist, $line;
                        next;
                    }

                    if ($line =~ /^======/ ) {
                        $end_of_header = 1;
                        next;
                    }
                    
                    if($line =~ /Port\s*Address\s*Media/) {
                        $address_field = 1;
                        next;
                    }
                }

                # Now parse port info lines for port number and its state
                foreach $line (@portlist)  {

                    # Delete any leading whitespaces
                    $line =~ s/^\s+//g;

                    $fc_port = (split (/\s+/, $line))[1];

                    $port_state = (split(/\s+/, $line))[4 + $address_field];
                    $port_status =  (split (/\s+/, $line))[5 + $address_field];
                    
                    # If port is a switch interconnect, we need to parse for an
                    # additional field for the port status
                    if ( $port_status =~ /LD/ ) {
                        $port_ld_status = (split (/\s+/, $line))[6 + $address_field];
                    }

                    # Check if Brocade V4,V5 port has reached enabled state
                    if ( (($port_state eq "Online" || $port_state eq "No_Sync" || $port_state eq "No_Light" ||
                          $port_state eq "No_Module") && $port_status ne "Disabled" ) ||
                         ( $port_state eq "Online" && $port_status =~ /LD/ &&  $port_ld_status ne "Disabled")  ) {
                        next;
                    } else {
                        $port_not_enabled = 1;
                        $retry_port_verify++;
                        last;
                    }
                }
            }

            if ( ! $port_not_enabled ) {
                $nochange = 1;  # switch ports are all in steady state online, we're done
            } elsif ( $port_not_enabled && ($retry_port_verify > $MAX_PORT_STATUS_RETRIES) ) {
                # Check if we have exceeed max number of retries
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: Brocade switch port $fc_port still disabled");
                return ($ERR_STATUS);
            } else {
                # Sleep for a bit and recheck port status
                $port_not_enabled = 0;
                sleep $PORT_STATUS_RETRY_WAIT;
            }

        } # end while

        return $GOOD_STATUS;

    } else {
        $self->{error} = $ERR_STATUS;
        push(@{$self->{errmsg}}, "ERROR: unsupported switch type: $hosttype");
        return $ERR_STATUS;
    }
}

## @name ::get_port_states ()
##
## @description: Gets current state of all switch ports. This
## method is meant to be used as a snapshot to verify port state
## changes (via comparing them to an earlier snapshot) after
## running switch_enable or switch_disable.
##
## @synopsis
## $port_state_hash_ref = VSeries::switch_object->get_port_states()
##
## Hash key: port  Hash Value: the port state
##
sub get_port_states () {

    my $self = shift;
    my $hostname = $self->get_name();
    my $hosttype = $self->get_host_type();
    my $after = "";
    my @info = ();
    my %port_state = ();
    my $port = "";
    my $state = "";
    my $line = "";

    if ($hosttype =~ /switch-cisco/i) {

        # show interface brief sample output

        #-------------------------------------------------------------------------------
        #Interface  Vsan   Admin  Admin   Status          SFP    Oper  Oper   Port
        #                   Mode  Trunk                          Mode  Speed  Channel
        #                         Mode                                (Gbps)
        #-------------------------------------------------------------------------------
        #fc1/1      1      auto   on      notConnected     swl    --           --
        #fc1/2      1      auto   on      notConnected     swl    --           --
        #fc1/3      1      auto   on      up               swl    F       4    --
        #fc1/4      1      auto   on      up               swl    F       4    --
        #fc1/5      1      auto   on      notConnected     swl    --           --
        #fc1/6      1      auto   on      up               swl    F       4    --
        #fc1/7      1      auto   on      up               swl    F       4    --
        #fc1/8      1      auto   on      trunking         c1550  TE      4    --
        #fc1/9      1      auto   on      up               swl    F       2    --
        #fc1/10     1      auto   on      up               swl    F       2    --
        #fc1/11     1      auto   on      up               swl    F       2    --
        #fc1/12     1      auto   on      up               swl    F       2    --
        #fc1/13     1      auto   on      up               swl    F       4    --
        #fc1/14     1      auto   on      up               swl    F       4    --
        #fc1/15     1      auto   on      up               swl    F       2    --
        #fc1/16     1      auto   on      notConnected     swl    --           --
        #fc1/17     1      auto   on      notConnected     swl    --           --
        #fc1/18     1      auto   on      notConnected     swl    --           --
        #fc1/19     1      auto   on      notConnected     swl    --           --
        #fc1/20     1      auto   on      notConnected     swl    --           --

        Tharn::say($self->{conn}, "terminal length 0", -swallow, '-timeout', 60);

        # Fix for BURT 490772.
        # The output of show interface might be prompted by more. Pipe to no-more to avoid more.

        (undef, undef, $after) = Tharn::say($self->{conn}, "show interface brief | no-more", -swallow, '-timeout', 60);
        @info = split(/\n/, $after);

        foreach my $line (@info)  {
            # Parse for lines containing leading "fc" characters
            if ( $line =~ /^fc/ ) {
                $port = (split (/\s+/, $line))[0];
                $state =  (split (/\s+/, $line))[4];
                $port_state{$port} = $state;
            }
        }

        return \%port_state;

    } elsif ($hosttype =~ /switch-mcdata/i) {


        # show port status sample output
        # ------------------------------

        #Port  State               Type    Attached WWN             Beaconing  Reason
        #----  ------------------  ------  -----------------------  ---------  ------
        #0     Online              fPort   50:06:04:82:CB:1B:CE:3C  Disabled
        #1     Online              fPort   50:0A:09:82:00:01:43:88  Disabled
        #2     No Light            gxPort  None                     Disabled
        #3     No Light            gxPort  None                     Disabled
        #4     Not Installed       gxPort  None                     Disabled   54
        #5     Not Installed       gxPort  None                     Disabled   54
        #6     Not Installed       gxPort  None                     Disabled   54
        #7     Not Installed       gxPort  None                     Disabled   54
        #8     Inactive            gxPort  None                     Disabled   46
        #9     Inactive            gxPort  None                     Disabled   46
        #10    Inactive            gxPort  None                     Disabled   46
        #11    Inactive            gxPort  None                     Disabled   46
        #12    Inactive            gxPort  None                     Disabled   46
        #13    Inactive            gxPort  None                     Disabled   46
        #14    Inactive            gxPort  None                     Disabled   46
        #15    Inactive            gxPort  None                     Disabled   46
        #
        #46: No key enabled          (this line appears if you run command from menu subdir)
        #54: Optics not installed    (this line appears if you run command from menu subdir)


        (undef, undef, $after) = Tharn::say($self->{conn}, "show port status", -swallow, '-timeout', 60);
        $after =~ s/\r//;
        @info = split(/\n/, $after);

        foreach my $line (@info)  {

            # Parse for port information lines
            if ( $line =~ /^(\d+)\s+(\w+)\s+\w+/ && $line !~ /^\d+:\s+/ ) {
                $port_state{$1} = $2;
            }
        }

        return \%port_state;

    } elsif ($hosttype =~ /switch-brocade/i) {

        my @portlist = ();     # for Brocade switch parsing
        my $end_of_header = 0; # for Brocade switch parsing

        # switchShow sample output (v5)
        #
        #switchName:     vns_b200e_8
        #switchType:     34.0
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:05:1e:04:14:73
        #zoning:         ON (Metrocluster_V6030)
        #switchBeacon:   OFF
        #
        #Area Port Media Speed State     Proto
        #=====================================
        #  0   0   id    AN   No_Sync
        #  1   1   id    N4   No_Light
        #  2   2   id    N4   No_Light
        #  3   3   id    N4   Online           F-Port  50:0a:09:83:00:01:71:9d
        #  4   4   id    N4   Online           F-Port  50:0a:09:81:00:01:71:9d
        #  5   5   id    N4   No_Light
        #  8   8   id    N2   Online           LD E-Port  10:00:00:05:1e:04:15:29 "vns_b200e_9" (upstream)
        # 21  21   id    N2   Online           F-Port  50:0a:09:82:00:00:de:9d
        # 22  22   id    N2   Online           F-Port  50:0a:09:82:00:01:91:cf
        # 23  23   --    N4   No_Module

        # switchShow sample output (v3)
        #
        #switchName:     vss_b3800_7
        #switchType:     9.2
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:60:69:51:f1:e4
        #switchBeacon:   OFF
        #Zoning:         ON (Automation1)
        #port  0: id N2 Port_Flt
        #port  1: id N1 No_Sync
        #port  2: id N2 Port_Flt
        #port  3: id N2 Online         F-Port 50:06:04:82:cb:1b:ce:2c
        #port  4: id N2 No_Light
        #port  5: id N2 No_Light
        #port  6: id N2 No_Light
        #port  7: id N2 No_Light
        #port  8: id N2 No_Light
        #port  9: id N2 No_Light
        #port 10: id N2 No_Light
        #port 11: id N2 No_Light
        #port 12: id N2 No_Light

        # For Brocade switches, output is formatted differently depending
        # on firmware rev, so we need to determine the switch fw rev first

        (undef, undef, $after) = Tharn::say($self->{conn}, "version", -swallow,
                                                       '-timeout', 60);

        # Get the firmware rev
        $after =~ /Fabric OS:\s+(v[3-9])/;
        my $version = $1;

        (undef, undef, $after) = Tharn::say($self->{conn}, "switchShow", -swallow, '-timeout', 60);
        @info = split(/\r\n/, $after);

        # Parse output depending on fw rev
        if ( $version eq "v3" ) {

            # Look for lines leading with 'port' and grab the
            # port number and its state
            foreach my $line (@info)  {
                if ( $line =~ /^port/ ) {
                    $port = (split (/\s+/, $line))[1];
                    chop $port;
                    $state =  (split (/\s+/, $line))[4];
                    $port_state{$port} = $state;
                }
            }
        } else { # fw rev v4,5 and beyond

            # Create @portlist subarray containing only the port info
            # output lines. We skip over all lines preceeding port info
            foreach my $line (@info)  {

                if ( $end_of_header )  {
                    push @portlist, $line;
                    next;
                }

                if ($line =~ /^======/ ) {
                    $end_of_header = 1;
                    next;
                }
            }

            # Now parse port info lines for port number and its state
            foreach my $line (@portlist)  {

                # Delete any leading whitespaces
                $line =~ s/^\s+//g;
                $port = (split (/\s+/, $line))[1];
                $state =  (split (/\s+/, $line))[4];
                $port_state{$port} = $state;
            }
        }

        return \%port_state;

    } else {
        $self->{error} = $ERR_STATUS;
        push(@{$self->{errmsg}}, "ERROR: unsupported switch type: $hosttype");
        return $ERR_STATUS;
    }

}

## @name ::set_switch_name($)
##
## @description
## Set the name the Switch
##
## @synopsis
## $switch_obj->switch_name_set($newname);
##
## @arg $name Required: New name of the switch
##
sub set_switch_name($) {
    my $self = shift;
    my($name) = @_;
    my ($temp, $user_id) = "";
    my $host_type = $self->{hosttype};
    my @brocade_v5_prompts = ();

   # get host type and determine proper method

   if ($host_type =~ /switch-mcdata/i) {
       Tharn::say($self->{conn}, "config system name $name", '-timeout' , 30);
   }
   elsif ($host_type =~ /switch-cisco/i) {
       Tharn::say($self->{conn}, "config t", -noprompt, '-timeout', 30);
       Tharn::say($self->{conn}, "switchname $name", -noprompt, '-timeout', 30);
       Tharn::say($self->{conn}, "end", -noprompt, '-timeout', 30);
   }
   elsif ($host_type =~ /switch-brocade/i) {
       ($temp, $user_id) = split(/:/, $self->{conn}->{hostp}->{prompt_norm}[0]);

       # need to get some switch info (model, vendor, firmware)
       # incase we need to do different stuff for different levels
       # of firmware or brocade switch models yada yada yada...
       my %switch_info = $self->get_switch_hw_sw_info();

       # if its firmware v5 (like b200e's and 3850's) we can't be
       # sensitive to the prompt to wait for because name changes
       # don't get reflected instantly they require a logout first.
       # see burt 242549
       if (($switch_info{"firmware"} =~ /v5/i) || ($switch_info{"firmware"} =~ /v6/i) || ($switch_info{"firmware"} =~ /v7/i)) {
           @brocade_v5_prompts = $self->{conn}->prompts();
           push(@brocade_v5_prompts, "${name}:${user_id}");
           Tharn::say($self->{conn}, "switchname \"$name\"", '-prompt', @brocade_v5_prompts, '-timeout', 30);
       }
       else{
           Tharn::say($self->{conn}, "switchname \"$name\"", -noprompt, -prompt, "${name}:${user_id}", '-timeout', 30);
       }
   }

}


## @name ::get_switch_hw_sw_info
## @description Returns hash having following info
##      "vendor" => <name>
##      "model" => <value>
##      "firmware" => <value>
## @synopsis
## switch_hw_sw_info_get($conn}
##
## @arg $conn Required: The switch connection to send this operation over
##
sub get_switch_hw_sw_info() {
    my $self = shift;
    my ($after, @info, $line, $length_port_name) = "";
    my $vendor = undef;
    my $model = undef;
    my $firmware = undef;
    my %switch_hw_sw_info = ();

    my $hosttype = $self->{hosttype};
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Cisco switch
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if ($hosttype =~ /switch-cisco/i) {
        $vendor = 'Cisco';
        chomp($vendor);
        Tharn::say($self->{conn}, "terminal length 0", -swallow, '-timeout', 60);

        # Fix for BURT 490772.
        # The output of show hardware might be prompted by more. Pipe to no-more to avoid more.

        (undef, undef, $after) = Tharn::say($self->{conn}, "show hardware | no-more", -swallow, '-timeout', 60);
        (@info) = split(/\r\n/, $after);
        foreach $line (@info)  {
            if ($line =~ /System version:/i) {
                $line =~ s/^\s+//g;
                ($firmware) = (split (/\s+/, $line))[2];
            }
            if ($line =~ /Model number is/i) {
                $line =~ s/^\s+//g;
                ($model) = (split (/\s+/, $line))[3];
            }
            if ((defined($model)) && (defined($firmware))) {
                last; #get out of the foreach loop, got the needed info
            }
        }
    } #cisco

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Mcdata switch
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if ($hosttype =~ /switch-mcdata/i) {
        $vendor = 'Mcdata';
        chomp($vendor);
        my $director;
        (undef, undef, $after) = Tharn::say($self->{conn}, "show system", -swallow, '-timeout', 60);
        if ($after =~ /director/i) {
            $director = 1;
        } else {
            $director = 0;
        }

        (undef, undef, $after) = Tharn::say($self->{conn}, "show system", -swallow, '-timeout', 60);
        (@info) = split(/\r\n/, $after);
        foreach $line (@info)  {
            if ($line =~ /Firmware Version:/i) {
                $line =~ s/^\s+//g;
                ($firmware) = (split (/\s+/, $line))[2];
            }
            if ($line =~ /Model Name:/i) {
                $line =~ s/^\s+//g;
                ($model) = (split (/\s+/, $line))[3];
            }
        }
    } #mcdata

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Brocade switch
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if ($hosttype =~ /switch-brocade/i) {
        $vendor = 'Brocade';
        chomp($vendor);
        my $version = undef;
        my @brocade_prompts = ('^\s*newname\d+:.*>');

        (undef, undef, $after) = Tharn::say($self->{conn}, "version", -swallow, '-prompt', @brocade_prompts, '-timeout', 60);

        # check for the Fabric OS version of the switch
        if ($after =~ /Fabric OS:\s*v7/i) {
            $version = "v7";
        } elsif ($after =~ /Fabric OS:\s*v6/i) {
            $version = "v6";
        } elsif ($after =~ /Fabric OS:\s*v5/i) {
            $version = "v5";
        } elsif ($after =~ /Fabric OS:\s*v4/i) {
            $version = "v4";
        } elsif ($after =~ /Fabric OS:\s*v3/i) {
            $version = "v3";
        } elsif ($after =~ /Fabric OS:\s*v2/i) {
            $version = "v2";
        } else { $version = "unsupported version"; }

        # need to handle a V3 OS differently than other versions...
        if ($version eq "v3") {
           @info = split(/\n\r/, $after);
        } elsif ($version eq "v5"||"v4"||"v2") {
           @info = split(/\r\n/, $after);
        }

        foreach $line (@info)  {
            if ($version eq "unsupported version") { last; }
            if ($line =~ /Fabric OS:/i) {
                $line =~ s/^\s+//g;
                ($firmware) = (split (/\s+/, $line))[2];
            }
        }

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # need to get the model name only possible in version 5 (not 4,3,2)
        #
        #vss-b200e-1:admin> chassisshow
        #
        #POWER SUPPLY  Unit: 1
        #Time Awake:             44 days
        #
        #FAN  Unit: 1
        #Time Awake:             44 days
        #
        #FAN  Unit: 2
        #Time Awake:             44 days
        #
        #FAN  Unit: 3
        #Time Awake:             44 days
        #
        #CHASSIS/WWN  Unit: 1
        #Header Version:         2
        #Factory Part Num:       40-0500027-06
        #Factory Serial Num:     RD060033215
        #Manufacture:            Day: 23  Month:  4  Year: 2006
        #Update:                 Day: 12  Month:  4  Year: 2007
        #Time Alive:             298 days
        #Time Awake:             44 days
        #ID:                     BRD0000CA
        #Part Num:               SLKWRM0000200
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        if ($version eq "v5") {
            (undef, undef, $after) = Tharn::say($self->{conn}, "chassisshow", -swallow, '-prompt', @brocade_prompts, '-timeout', 60);
            @info = ();
            $line = '';
            @info = split(/\n/, $after);

            foreach $line (@info) {
                if ($line =~ /^\s*Part Num:/i) {
                    $line =~ s/^s+//g;
                    ($model) = (split(/\s+/, $line))[2];
                }
            }
        } # if ($version eq "v5")
    } #Brocade

    $switch_hw_sw_info{"vendor"}   = $vendor;
    $switch_hw_sw_info{"model"}    = $model;
    $switch_hw_sw_info{"firmware"} = $firmware;

    return (%switch_hw_sw_info);
}

## @name ::switch_logout
## @description Logout of the Switch
## @synopsis
## $state = switch_logout{$conn}
## @arg $conn Required: The switch connection to send this operation over
## @example
## $state = switch_logout($sc);
sub so_switch_logout() {
   my $self = shift;

   my ($match, $before, $after) = "";
   my $host_type = $self->{hosttype};

   if ($host_type =~ /switch-mcdata/i) {
       # logout of the CLI
       ($match, $before, $after) = Tharn::say($self->{conn}, "logout", '-blind', '-timeout', 30);
   } else {
       # logout of the CLI
       ($match, $before, $after) = Tharn::say($self->{conn}, "exit", '-blind', '-timeout', 30);
   }

}




## @name ::switch_reboot
##
## @description Reboot the switch, works only for cisco and brocade.  Returns
## $GOOD_STATUS, at the moment it doesn't verify the switch rebooted only that
## it can kick off a reboot.
##
## @synopsis
## my $status = $switch_obj->switch_reboot()
##
sub switch_reboot() {

   my $self = shift;
    
   # get host type and determine proper method
   my $host_type = $self->{hosttype};

   if ($host_type =~ /switch-mcdata/i) {
       #say($conn, "maint system setOnlineState 0");
       #say($conn, "maint system setOnlineState 1");
   } elsif ($host_type =~ /switch-cisco/i) {
       my $CONF_PROMPT = "This command will reboot the system.";
       Tharn::say($self->{conn}, "reload", '-prompt', $CONF_PROMPT, '-timeout', 30);
       Tharn::say($self->{conn}, "y", '-noprompt', '-prompt', "y", '-timeout', 30);
   } elsif ($host_type =~ /switch-brocade/i) {
       my $conf_prompt = "Are you sure you want to reboot the switch";
       Tharn::say($self->{conn}, "reboot", '-prompt', $conf_prompt, '-timeout', 30);
       # Send "y" blindly in case the switch FW prompts for confirmation.
       # If not, no harm done.
       Tharn::say($self->{conn}, "y", '-blind');
   }

   return $GOOD_STATUS;
}

## @name ::switch_fastboot
##
## @description fastboot the switch, works only for brocade.  Returns
## $GOOD_STATUS, at the moment it doesn't verify the switch fastboot was
## successful.
##
## @synopsis
## my $status = $switch_obj->switch_fastboot();
##
sub switch_fastboot($) {

   my $self = shift;

   # get host type and determine proper method
   my $host_type = $self->{hosttype};

   if ($host_type =~ /switch-mcdata/i) {
       #say($conn, "maint system setOnlineState 0");
       #say($conn, "maint system setOnlineState 1");
   } elsif ($host_type =~ /switch-cisco/i) {
       ######
   } elsif ($host_type =~ /switch-brocade/i) {
       Tharn::say($self->{conn}, "fastboot", '-noprompt', '-prompt', "fastboot", '-timeout', 30);
       # Send "y" blindly in case the switch FW prompts for confirmation.
       # If not, no harm done.
       Tharn::say($self->{conn}, "y", '-blind');
   }
}


## @name ::switch_connect($)
##
## @description
## Takes in a connection type (#telnet, #rsh, etc) and calls Tharn::host on the
## hostname of the switch so that it then can call Tharn::connect with the
## passed in value for connection type.  The result will be a connection to the
## switch that the switch object represents.  It then calls set internal
## variables and places the newly made connection in the proper data members
## of the object.  This way subsequent calls to methods of the switch object
## will go over the new connection when called.  Doesn't have a return value.
##
## @synopsis
## $switch_obj->switch_connect($conn_type);
##
## @arg $conn_type Required: String representing the type of connection to be
##                           used to communicate with the given switch.
##                           Usually this is #telnet but may be something else
##                           like #rsh or #ssh in special circumstances.
##
sub switch_connect($){
    my $self = shift;
    my ($conn_type) = @_;

    my $conn_obj = "";
    if($conn_type ne ""){
        Tharn::host($self->{hostname});
        $conn_obj = Tharn::connect($conn_type);
    }
    else{
        my $hostrec_object = Tharn::host($self->{hostname});

        my $hash_ref = $hostrec_object->{connection_fields};
        my %hash = %$hash_ref;

        foreach my $connect_type (keys %hash){
            $conn_obj = Tharn::connect("#" . $connect_type);
            last;
        }

    }


    $self->set_internal_variables(

        conntype      => $conn_obj->conntype(),
        connport      => $conn_obj->connport(),
        username      => $conn_obj->username(),
        password      => $conn_obj->password(),
        timeout       => $conn_obj->timeout(),
        conn          => $conn_obj

    );



}


## @name ::switch_disconnect()
##
## @description
## This is a method to disconnect from the current connection in a switch
## object.  This is done as a way to gracefully close a connection rather than
## leaving it around stale and perhaps hogging up the only available connection
## to a switch because some switches are configured to only allow one
## connection at a time.  It then calls set internal variable to clear out the
## connection specific data members of the switch object.  Doesn't have a
## return type.
##
## @synopsis
## $switch_obj->switch_disconnect();
##
sub switch_disconnect(){
    my $self = shift;

    disconnect($self->{conn});
    $self->set_internal_variables(

        conntype      => undef,
        connport      => undef,
        username      => undef,
        password      => undef,
        timeout       => undef,
        conn          => undef

    );
}

## @name ::switch_port_disable()
##
## @description Disable a port on a switch
##
## @synopsis
## my $status = $switch_obj->switch_port_disable($port);
##
## @arg $port Required: The switch port number to disable
##
sub switch_port_disable($) {
    my $self = shift;
    my($port) = @_;

    # get host type and determine proper method
    my $host_type = $self->{'hosttype'};

    if ($host_type =~ /switch-mcdata/i) {
        Tharn::say($self->{conn}, "config port blocked $port true");
    }
    elsif ($host_type eq "switch-dell") {
        Tharn::say($self->{conn}, "configure");
        Tharn::say($self->{conn}, "interface ethernet 1/$port");
        Tharn::say($self->{conn}, "shutdown");
        Tharn::say($self->{conn}, "exit");
        Tharn::say($self->{conn}, "exit");
    }
    elsif ($host_type =~ /switch-cisco/i) {
        # If the switch port value has a dash in it we need to translate it
        # to a slash
        if($port =~ m/\-/){
            $port =~ s/\-/\//;
        }

        Tharn::say($self->{conn}, "conf t", -noprompt,  "-wait", "#");
        my (undef,undef,$after) = Tharn::say($self->{conn}, "int fc $port", -noprompt, "-wait", "#");
        # Checking if the port number is a valid one
        if($after =~ /invalid interface id format detected/i) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, $after);
            return($ERR_STATUS);
        }
        
        Tharn::say($self->{conn}, "shutdown", -noprompt, "-wait", "#");
        Tharn::say($self->{conn}, "end", -noprompt, "-wait", "#");
    }
    elsif ($host_type =~ /switch-brocade/i) {
        # brocade support
        Tharn::say($self->{conn}, "portdisable $port");
    }


    return $GOOD_STATUS;

}


## @name ::switch_port_enable()
##
## @description Enable port on switch
##
## @synopsis
## my $status = $switch_obj->switch_port_enable($port)
##
## @arg $port Required: The switch port number to enable
##
sub switch_port_enable($) {
    my $self = shift;
    my($port) = @_;

    # get host type and determine proper method
    my $host_type = $self->{'hosttype'};



    if ($host_type =~ /switch-mcdata/i) {
        Tharn::say($self->{conn}, "config port blocked $port false");
    }
    elsif ($host_type eq "switch-dell") {
        Tharn::say($self->{conn}, "configure");
        Tharn::say($self->{conn}, "interface ethernet 1/$port");
        Tharn::say($self->{conn}, "no shutdown");
        Tharn::say($self->{conn}, "exit");
        Tharn::say($self->{conn}, "exit");
    }
    elsif ($host_type =~ /switch-cisco/i) {
        # If the switch port value has a dash in it we need to translate it
        # to a slash
        if($port =~ m/\-/){
            $port =~ s/\-/\//;
        }

        Tharn::say($self->{conn}, "conf t", -noprompt,  "-wait", "#");
        my (undef,undef,$after) = Tharn::say($self->{conn}, "int fc $port", -noprompt, "-wait", "#");
        # Checking if the port number is a valid one
        if($after =~ /invalid interface id format detected/i) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, $after);
            return($ERR_STATUS);
        }
        
        Tharn::say($self->{conn}, "no shutdown", -noprompt, "-wait", "#");
        Tharn::say($self->{conn}, "end", -noprompt, "-wait", "#");
    }
    elsif ($host_type =~ /switch-brocade/i) {
        # brocade support
        Tharn::say($self->{conn}, "portenable $port");
    }


   return $GOOD_STATUS;

}



## @name ::_determine_switch_type($)
##
## @description This is a special method called from within the switch_object
## constructor to determine the vendor of the switch the given object will be
## communicating with.  It takes in a specific host name and gets a hostrec
## object for the hostname and from there will figure out whatever the
## specified connection protocol is for the switch.  Once it can connect
## it will do so and run the "switchShow" command.  This is a Brocade
## specific command and based on the output, either in the form of an error
## string or valid output will use that output to determine the vendor type
## for the switch.  With that it will return to the caller the properly
## scoped class for the switch_object to be cast into for support for that
## vendor.  This is an internal method with a specific intended purpose that is
## used to scope the switch object correctly it is not intended to be called
## outside the constructor.
##
## @synopsis
## my $class = VSeries::switch_object::_determine_switch_type($hostname);
##
## @arg $port Required: The switch port number to enable
##
sub _determine_switch_type($){
    my $self = shift;
    my ($hostname) = @_;


    # Get the host rec for the specified file name.
    # If there isn't one available host will fail so
    # there's no need to worry about error checking.
    my $hostrec_object = Tharn::host($hostname);

    # The connection fields attribute of a hostrec object is a hash table with
    # keys being the protocol types that have been specified in the connectrec
    # object within the tharnhost file.
    my $hash_ref = $hostrec_object->{connection_fields};
    my %hash = %$hash_ref;

    # This will be the connectrec object used to talk to the switch.
    my $connrec_object = "";

    # Loop through and take the first available protocol.
    foreach my $conn_type (keys %hash){
        $connrec_object = Tharn::connect("#" . $conn_type);
        last;
    }

    # Returned attributes from say calls, lets initialize them.
    my ($match, $before, $after, $prompt) = "";

    # Now lets call say and run the switchShow command.  Once it runs let's
    # look at what after contains to see what vendor we have.
    eval{
        ($match, $before, $after, $prompt) = Tharn::say($connrec_object, "switchShow");
    };
    if($@){
        # This is the Cisco specific error string, so that's still part of our
        # base object.
        if($after =~ m/invalid command detected at/){
            return "VSeries::switch_object";
        }
        # This is a McData specific error string, so that's also still part of
        # our base object.
        elsif($after =~ m/Error 234: Invalid command/){
            return "VSeries::switch_object";
        }
    }

    # Ok now we need to deal with Brocade Direct functionality.  If the output
    # of switchShow contains index values it's a director.
    my @output = split(/\n/, $after);

    # Check if any line has the Index column at the start of the output.
    foreach my $line (@output){
        #
        # switchShow output has an additional column called 'Index', which has been
        # introduced in all latest firmware versions of Brocade switches.
        # Instead of matching the Index column, we are getting the switchType and
        # matching it with the Brocade_Director Switch models.
        #
        # switchShow sample output (v3)
        #
        # switchName:     vss_b3800_7
        # switchType:     9.2
        #
        # How to find the brocade_switch models?
        # I have provided few reference links in the BURT.
        # Please refer to the BURT #488373 for more information.
        #
        if ($line =~ m/^switchType\s*:\s*(\d+)\.\d+/i) {
           my $switch_type = $1;

           if ($switch_type =~ m/^(10|21|36|42|47|55|62|68)$/) {
              return "VSeries::Brocade_Director";
           }
        }
    }

    # If it didn't contain an Index column then it's a non-director class
    # Brocade and that's part of our base object as well.
    return "VSeries::switch_object";

}

## @name ::mcdata_clear_pending()
##
## @description
## Clears the pending zone configuration on McData switches. This prevents the
## enabling of stale zones.
##
## @synopsis
## my $status = $switch_obj->mcdata_clear_pending()
##
sub mcdata_clear_pending() {

    my $self = shift;
    my $after;

    # Issue the command to clear the pending zoneset.
    Tharn::say($self->{conn}, "config zoning clearzoneset", -swallow, '-timeout', 30);

    # Look for evidence that the command was successful and return status.
    (undef, undef, $after) = Tharn::say($self->{conn}, "config zoning showpending", -swallow, '-timeout', 30);

    if ($after =~ /There are no zones in pending zoneset/ ) {
        return($GOOD_STATUS);
    } else {
        $self->{error} = $ERR_STATUS;
        push(@{$self->{errmsg}}, "ERROR: failed to clear pending zoneset on switch " . $self->get_name() . ".");
        return($ERR_STATUS);
    }
}


## @name ::get_seen_wwns()
##
## @description returns an array ref of all WWNs connected to the switch.
##
## @synopsis
## my $array_ref = $switch_obj->get_seen_wwns()
##
##
sub get_seen_wwns(){
    my ($self) = @_;
    
    # This is the array we'll return.
    my @seen_wwns = ();
    
    # Captures say output.
    my $after = "";
    
    # Array of single lines of output.
    my @info = ();

    # get host type and determine proper method
    my $host_type = $self->{'hosttype'};


    # Check the host type for the switch and given the vendor run the
    # appropriate command.
    if ($host_type =~ /switch-mcdata/i) {

        # show port status sample output
        # ------------------------------

        #Port  State               Type    Attached WWN             Beaconing  Reason
        #----  ------------------  ------  -----------------------  ---------  ------
        #0     Online              fPort   50:06:04:82:CB:1B:CE:3C  Disabled
        #1     Online              fPort   50:0A:09:82:00:01:43:88  Disabled
        #2     No Light            gxPort  None                     Disabled
        #3     No Light            gxPort  None                     Disabled
        #4     Not Installed       gxPort  None                     Disabled   54
        #5     Not Installed       gxPort  None                     Disabled   54
        #6     Not Installed       gxPort  None                     Disabled   54
        #7     Not Installed       gxPort  None                     Disabled   54
        #8     Inactive            gxPort  None                     Disabled   46
        #9     Inactive            gxPort  None                     Disabled   46
        #10    Inactive            gxPort  None                     Disabled   46
        #11    Inactive            gxPort  None                     Disabled   46
        #12    Inactive            gxPort  None                     Disabled   46
        #13    Inactive            gxPort  None                     Disabled   46
        #14    Inactive            gxPort  None                     Disabled   46
        #15    Inactive            gxPort  None                     Disabled   46
        #
        #46: No key enabled          (this line appears if you run command from menu subdir)
        #54: Optics not installed    (this line appears if you run command from menu subdir)

        
        (undef, undef, $after) = Tharn::say($self->{conn}, "show port status", -swallow, '-timeout', 60);
        
        # Strip off carriage returns.
        $after =~ s/\r//;
        
        # Split on new line for single line parsing.
        @info = split(/\n/, $after);

        # For each line of this output match on the WWN regex
        # and put what matches in the seen_wwns array.
        foreach my $line (@info)  {

            # Parse for port information lines
            if ( $line =~ /($WWN_REG_EX)/ ) {
                push(@seen_wwns, $1);
            }
        }

        return \@seen_wwns;

    }

    elsif ($host_type =~ /switch-brocade/i) {



        # switchShow sample output (v5)
        #
        #switchName:     vns_b200e_8
        #switchType:     34.0
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:05:1e:04:14:73
        #zoning:         ON (Metrocluster_V6030)
        #switchBeacon:   OFF
        #
        #Area Port Media Speed State     Proto
        #=====================================
        #  0   0   id    AN   No_Sync
        #  1   1   id    N4   No_Light
        #  2   2   id    N4   No_Light
        #  3   3   id    N4   Online           F-Port  50:0a:09:83:00:01:71:9d
        #  4   4   id    N4   Online           F-Port  50:0a:09:81:00:01:71:9d
        #  5   5   id    N4   No_Light
        #  8   8   id    N2   Online           LD E-Port  10:00:00:05:1e:04:15:29 "vns_b200e_9" (upstream)
        # 21  21   id    N2   Online           F-Port  50:0a:09:82:00:00:de:9d
        # 22  22   id    N2   Online           F-Port  50:0a:09:82:00:01:91:cf
        # 23  23   --    N4   No_Module

        # switchShow sample output (v3)
        #
        #switchName:     vss_b3800_7
        #switchType:     9.2
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:60:69:51:f1:e4
        #switchBeacon:   OFF
        #Zoning:         ON (Automation1)
        #port  0: id N2 Port_Flt
        #port  1: id N1 No_Sync
        #port  2: id N2 Port_Flt
        #port  3: id N2 Online         F-Port 50:06:04:82:cb:1b:ce:2c
        #port  4: id N2 No_Light
        #port  5: id N2 No_Light
        #port  6: id N2 No_Light
        #port  7: id N2 No_Light
        #port  8: id N2 No_Light
        #port  9: id N2 No_Light
        #port 10: id N2 No_Light
        #port 11: id N2 No_Light
        #port 12: id N2 No_Light

        (undef, undef, $after) = Tharn::say($self->{conn}, "switchShow", -swallow, '-timeout', 60);
        
        # Split buffer into single lines for parsing.
        @info = split(/\r*\n/, $after);

        # Same technique here match on the WWN regex, however
        # with brocade you need to be sure the line isn't
        # the wwn of the switch itself so handle that
        # case by skipping over that line.
        foreach my $line (@info)  {
            $line =~ s/\r//g;

            # Parse for port information lines
            if ( $line =~ /($WWN_REG_EX)/ ) {
                if($line =~ m/^switchWwn:/){ next;}
                push(@seen_wwns, $1);
            }
        }

        return \@seen_wwns;


    }
    elsif ($host_type =~ /switch-cisco/i) {
        #vnci9124s52#
        #vnci9124s52# show flogi database
        #--------------------------------------------------------------------------------
        #INTERFACE        VSAN    FCID           PORT NAME               NODE NAME
        #--------------------------------------------------------------------------------
        #fc1/3            1     0xcf0500  20:04:00:a0:b8:21:a9:fa 20:04:00:a0:b8:21:a9:f8
        #fc1/5            1     0xcf0600  50:0a:09:83:00:01:90:74 50:0a:09:83:00:01:90:74
        #fc1/6            1     0xcf0700  50:0a:09:83:00:01:91:be 50:0a:09:83:00:01:91:be

        # Fix for BURT 490772.
        # The output of show flogi database  might be prompted by more. Pipe to no-more to avoid more.

        (undef, undef, $after) = Tharn::say($self->{conn}, "show flogi database | no-more ", -swallow, '-timeout', 60);
        
        # Split into single lines for parsing.
        @info = split(/\r\n/, $after);
        
        # Same technique here except each line has the port and node wwns
        # so match on both but only push the port wwn onto the seen_wwns array.
        foreach my $line (@info)  {
            # Parse for port information lines, we want the port WWN.
            if ( $line =~ m/($WWN_REG_EX)\s+($WWN_REG_EX)/ ) {
                push(@seen_wwns, $1);
            }
        }

        return \@seen_wwns;
    }
    
}


## @name ::set_port_speed($$)
##
## @description sets the speed of a port on a switch
##
## @synopsis
## my $status = $switch_obj->set_port_speed($port, $speed)
##
## @arg $port Required: The switch port number to set the speed of.
##
## @arg $speed Required: The value to set the port at.  It is expected to be 
##                       of a legal value accepted by the switch.
##
sub set_port_speed($$){
    my $self = shift;
    my($port, $speed) = @_;

    # get host type and determine proper method
    my $host_type = $self->{'hosttype'};


    # The commands are different for different vendors so check the hosttype 
    # and act accordingly.
    if ($host_type =~ /switch-mcdata/i) {
        $speed .= "GB";
        Tharn::say($self->{conn}, "speed $port $speed");
    }
    elsif ($host_type =~ /switch-cisco/i) {
        # If the switch port value has a dash in it we need to translate it
        # to a slash
        if($port =~ m/\-/){
            $port =~ s/\-/\//;
        }

        # Ciscos need a translation from GB to MB.
        $speed = $speed * 1000;

        # The Cisco interface is a sub menu system, need to put the cli
        # in the right sub menu to configure the port speed.
        Tharn::say($self->{conn}, "conf t", -noprompt,  "-wait", "#");
        Tharn::say($self->{conn}, "int fc $port", -noprompt, "-wait", "#");
        Tharn::say($self->{conn}, "switchport speed $speed", -noprompt, "-wait", "#");
        Tharn::say($self->{conn}, "end", -noprompt, "-wait", "#");
    }
    elsif ($host_type =~ /switch-brocade/i) {
        # brocade support
        Tharn::say($self->{conn}, "portCfgSpeed $port $speed");
    }

   return $GOOD_STATUS;

}

## @name ::get_port_wwn_seen_on($)
##
## @description takes in a wwn and finds the port that wwn is connected to.
##              This is very useful in going between soft and hard zoning.
##
## @synopsis
## my $port = $switch_obj->get_port_wwn_seen_on($wwn)
##
## @arg $wwn Required: a wwn the caller wants to know what port on the switch
##                     its seen on.
##
sub get_port_wwn_seen_on($){

    my $self = shift;
    my($wwn) = @_;

    # Captures say output.
    my $after = "";
    
    # Array of single lines of output.
    my @info = ();

    # get host type and determine proper method
    my $host_type = $self->{'hosttype'};

    # Value to be returned to the caller.
    my $port_seen_on = "";


    if ($host_type =~ /switch-mcdata/i) {

        # show port status sample output
        # ------------------------------

        #Port  State               Type    Attached WWN             Beaconing  Reason
        #----  ------------------  ------  -----------------------  ---------  ------
        #0     Online              fPort   50:06:04:82:CB:1B:CE:3C  Disabled
        #1     Online              fPort   50:0A:09:82:00:01:43:88  Disabled
        #2     No Light            gxPort  None                     Disabled
        #3     No Light            gxPort  None                     Disabled
        #4     Not Installed       gxPort  None                     Disabled   54
        #5     Not Installed       gxPort  None                     Disabled   54
        #6     Not Installed       gxPort  None                     Disabled   54
        #7     Not Installed       gxPort  None                     Disabled   54
        #8     Inactive            gxPort  None                     Disabled   46
        #9     Inactive            gxPort  None                     Disabled   46
        #10    Inactive            gxPort  None                     Disabled   46
        #11    Inactive            gxPort  None                     Disabled   46
        #12    Inactive            gxPort  None                     Disabled   46
        #13    Inactive            gxPort  None                     Disabled   46
        #14    Inactive            gxPort  None                     Disabled   46
        #15    Inactive            gxPort  None                     Disabled   46
        #
        #46: No key enabled          (this line appears if you run command from menu subdir)
        #54: Optics not installed    (this line appears if you run command from menu subdir)


        (undef, undef, $after) = Tharn::say($self->{conn}, "show port status", -swallow, '-timeout', 60);
        
        # Strip out any carriage returns.
        $after =~ s/\r//;
        
        # Split on new lines.
        @info = split(/\n/, $after);

        # For each line match on the WWN regex.  If matched, compare that
        # with the passed in regex, if that matches then strip out the
        # port value for that line and return.
        foreach my $line (@info)  {

            # Parse for port information lines
            if ( $line =~ /($WWN_REG_EX)/ ) {
                if($1 =~ m/$wwn/i){
                    $line =~ m/^(\d+)/;
                    $port_seen_on = $1;
                    last;
                }
            }
        }

        return $port_seen_on;
    }

    elsif ($host_type =~ /switch-brocade/i) {



        # switchShow sample output (v5)
        #
        #switchName:     vns_b200e_8
        #switchType:     34.0
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:05:1e:04:14:73
        #zoning:         ON (Metrocluster_V6030)
        #switchBeacon:   OFF
        #
        #Area Port Media Speed State     Proto
        #=====================================
        #  0   0   id    AN   No_Sync
        #  1   1   id    N4   No_Light
        #  2   2   id    N4   No_Light
        #  3   3   id    N4   Online           F-Port  50:0a:09:83:00:01:71:9d
        #  4   4   id    N4   Online           F-Port  50:0a:09:81:00:01:71:9d
        #  5   5   id    N4   No_Light
        #  8   8   id    N2   Online           LD E-Port  10:00:00:05:1e:04:15:29 "vns_b200e_9" (upstream)
        # 21  21   id    N2   Online           F-Port  50:0a:09:82:00:00:de:9d
        # 22  22   id    N2   Online           F-Port  50:0a:09:82:00:01:91:cf
        # 23  23   --    N4   No_Module

        # switchShow sample output (v3)
        #
        #switchName:     vss_b3800_7
        #switchType:     9.2
        #switchState:    Online
        #switchMode:     Native
        #switchRole:     Principal
        #switchDomain:   1
        #switchId:       fffc01
        #switchWwn:      10:00:00:60:69:51:f1:e4
        #switchBeacon:   OFF
        #Zoning:         ON (Automation1)
        #port  0: id N2 Port_Flt
        #port  1: id N1 No_Sync
        #port  2: id N2 Port_Flt
        #port  3: id N2 Online         F-Port 50:06:04:82:cb:1b:ce:2c
        #port  4: id N2 No_Light
        #port  5: id N2 No_Light
        #port  6: id N2 No_Light
        #port  7: id N2 No_Light
        #port  8: id N2 No_Light
        #port  9: id N2 No_Light
        #port 10: id N2 No_Light
        #port 11: id N2 No_Light
        #port 12: id N2 No_Light

        (undef, undef, $after) = Tharn::say($self->{conn}, "switchShow", -swallow, '-timeout', 60);
        
        # Split into single lines.
        @info = split(/\r*\n/, $after);

        # Same technique.  Match on the wwn regex, make sure it's not the
        # switch wwn, but then based on which version of firmware the switch
        # is running stripping out hte port value is done differently.  Using
        # regexes to handle those cases.
        foreach my $line (@info)  {

            $line =~ s/\r//g;
            
            # Parse for port information lines
            if ( $line =~ /($WWN_REG_EX)/ ) {           
                if($line =~ m/^switchWwn\:/){ next;}
                if($1 =~ m/$wwn/){    
                    if($line =~ m/^port\s+(\d+)/){
                        $port_seen_on = $1;
                        last;
                    }
                    elsif($line =~ m/^\s*\d+\s+(\d+)/){
                        $port_seen_on = $1;
                        last;                                            
                    }    
                }
            }
        }

        return $port_seen_on;


    }
    elsif ($host_type =~ /switch-cisco/i) {
        #vnci9124s52#
        #vnci9124s52# show flogi database
        #--------------------------------------------------------------------------------
        #INTERFACE        VSAN    FCID           PORT NAME               NODE NAME
        #--------------------------------------------------------------------------------
        #fc1/3            1     0xcf0500  20:04:00:a0:b8:21:a9:fa 20:04:00:a0:b8:21:a9:f8
        #fc1/5            1     0xcf0600  50:0a:09:83:00:01:90:74 50:0a:09:83:00:01:90:74
        #fc1/6            1     0xcf0700  50:0a:09:83:00:01:91:be 50:0a:09:83:00:01:91:be

        # Fix for BURT 490772.
        # The output of show flogi database might be prompted by more. Pipe to no-more to avoid more.

        (undef, undef, $after) = Tharn::say($self->{conn}, "show flogi database | no-more", -swallow, '-timeout', 60);
        
        # Split onto new lines.
        @info = split(/\r\n/, $after);

        # Same technique.  Match on the wwn regex and then check if the wwn 
        # matches the wwn passed in.  If so then strip out the port seen on
        # and convert the slash to a dash and return.
        foreach my $line (@info)  {
        
            # Parse for port information lines, we want the port WWN.
            if ( $line =~ /($WWN_REG_EX)\s+$WWN_REG_EX/ ) {
                 if($1 =~ m/$wwn/){
                     $line =~ m/^fc(\d+\/\d+)/;
                     $port_seen_on = $1;
                     $port_seen_on =~ s/\//-/g;
                     last;
                }
            }

        }
        return $port_seen_on;
    }
}


## @name ::get_port_fabric_id($)
##
## @description takes in a wwn and finds the fabric id of the switch port
## that the given wwn is seen on.  This is useful in generating FITP bootargs.
##
## @synopsis
## my $fabric_id = $switch_obj->get_port_fabric_id($wwn)
##
## @arg $wwn Required: a wwn the caller wants to know the fabric id of the port
## this wwn is seen on.
##
sub get_port_fabric_id($){
    my $self = shift;
    my($wwpn) = @_;

    # Captures say output.
    my $after = "";
    
    # Array of single lines of output.
    my @info = ();

    # get host type and determine proper method
    my $host_type = $self->{'hosttype'};

    # Value to be returned to the caller.
    my $port_fabric_id = "";

    if ($host_type =~ /switch-mcdata/i) {
    
        # Sample entry of the show nameserver command:
        # Root> show nameserver
	# DID  Type  PortId  Port Name                Node Name                COS  FC4
	# ---  ----  ------  -----------------------  -----------------------  ---  ---
	# 1    N     610513  50:0A:09:82:00:00:E1:60  50:0A:09:82:00:00:E1:60  3    18 
	# 1    N     610613  50:05:07:63:03:13:81:24  50:05:07:63:03:FF:C1:24  2-3  2  
	# 1    N     610313  21:01:00:1B:32:27:CD:6D  20:01:00:1B:32:27:CD:6D  3    N/A
	# 1    N     610813  50:05:07:63:03:03:C1:24  50:05:07:63:03:FF:C1:24  2-3  2  
	# 1    N     610B13  50:06:0E:80:03:4F:E7:17  50:06:0E:80:03:4F:E7:17  3    2  
	# 1    N     610413  50:0A:09:82:00:00:E1:59  50:0A:09:82:00:00:E1:59  3    18 
        # 1    N     610F13  21:01:00:1B:32:25:27:99  20:01:00:1B:32:25:27:99  3    18
        
        (undef, undef, $after) = Tharn::say($self->{conn}, "show nameserver", -swallow, '-timeout', 60);
        
        # Strip out any carriage returns.
        $after =~ s/\r//;
        
        # Split on new lines.
        @info = split(/\n/, $after);

        foreach my $line (@info)  {
            # Parse for port information lines, we want the port WWN.
            if ( $line =~ /($WWN_REG_EX)\s+$WWN_REG_EX/ ) {
                     if($wwpn !~ m/$1/i){next;}      
                     
                     # Strip out the FCID field and return it to the caller.
                     $line =~ m/^\S+\s+\S+\s+(\S+)/;
                        
                     $port_fabric_id = $1;
                           
                     last;
                
            }

        }
        return $port_fabric_id;        
        
    
    }
    elsif ($host_type =~ /switch-brocade/i) {
    
    # Sample entry of the nsshow command:
    #
    #Type Pid    COS     PortName                NodeName                 TTL(sec)\r\n
    # N    0a0000;      3;50:0a:09:82:00:01:4b:57;50:0a:09:82:00:01:4b:57; na\r\n
    #    PortSymb: [40] "NetApp FC Storage Adapter vnv3040f6b:0a."\r\n
    #    Fabric Port Name: 20:00:00:05:1e:03:fc:26 \r\n
    #    Permanent Port Name: 50:0a:09:82:00:01:4b:57\r\n
    #    Port Index: 0\r\n
    #    Share Area: No\r\n
    #    Device Shared in Other AD: No\r\n
    #    Redirect: No \r\n
    
    
        (undef, undef, $after) = Tharn::say($self->{conn}, "nsshow", -swallow, '-timeout', 60);
        
        # Split onto new lines.
	@info = split(/\r?\n/, $after);
	
	
        # Will contain the current line of output.
        my $current_line = "";
        
        # Now loop over the lines of output we know the WWN is on the same line
        # the fabric ID.  In the sample output above the Pid on the first line 
        # is the fabric ID and the PortName value is the wwn seen on that port.
        # We can match by the commas being on both sides of the wwn passed in.
        for(my $i = 0; $i < scalar(@info); $i++){
            $current_line = $info[$i];
            
            # Strip out leading whitespace.
            $current_line =~ s/^\s+//;

            # If the line doesn't match our regex move on to the next one.
            if($current_line !~ m/\;$wwpn\;/){next;}
            
            # Match and use the regex to pull out the fabric id value.
            $current_line =~ m/^\S+\s+(\S+)\;/;
            
            # Set return variable and return it to the caller.
            $port_fabric_id = $1;
            last;
            
        }
        
        return $port_fabric_id;
    
    
    }
    elsif ($host_type =~ /switch-cisco/i) {
        #vnci9124s52#
        #vnci9124s52# show flogi database
        #--------------------------------------------------------------------------------
        #INTERFACE        VSAN    FCID           PORT NAME               NODE NAME
        #--------------------------------------------------------------------------------
        #fc1/3            1     0xcf0500  20:04:00:a0:b8:21:a9:fa 20:04:00:a0:b8:21:a9:f8
        #fc1/5            1     0xcf0600  50:0a:09:83:00:01:90:74 50:0a:09:83:00:01:90:74
        #fc1/6            1     0xcf0700  50:0a:09:83:00:01:91:be 50:0a:09:83:00:01:91:be

        # Fix for BURT 490772.
        # The output of show flogi database  might be prompted by more. Pipe to no-more to avoid more.

        (undef, undef, $after) = Tharn::say($self->{conn}, "show flogi database | no-more", -swallow, '-timeout', 60);
        
        # Split onto new lines.
        @info = split(/\r\n/, $after);

        # Same technique.  Match on the wwn regex and then check if the wwn 
        # matches the wwn passed in.  If so then strip out the port seen on
        # and return.
        foreach my $line (@info)  {
            # Parse for port information lines, we want the port WWN.
            if ( $line =~ /($WWN_REG_EX)\s+$WWN_REG_EX/ ) {
                     if($wwpn !~ m/$1/){next;}      
                     
                     # Strip out the FCID field and return it to the caller.
                     $line =~ m/^\S+\s+\S+\s+(\S+)/;
                     
                     my $matched_id = $1;
                     
                     # Removing the "0x" at the front for consistent format 
                     # with other vendor types.
                     $matched_id =~ s/^0x//;
                        
                     $port_fabric_id = $matched_id;
                           
                     last;
                
            }

        }
        return $port_fabric_id;
    }
    

}

## @name ::set_autologout_delay($)
##
## @description sets the autologout delay on a switch
##
## @synopsis
## my $status = $switch_obj->set_autologout_delay($timeout);
##
## @arg $timeout Required: The autologout timeout to be used.
##
##
sub set_autologout_delay($){
    my $self = shift;
    my($timeout) = @_;
    
    # get host type and determine proper method, logout is needed for the new
    # timeout to reflect
    my $host_type = $self->{'hosttype'};

    if ($host_type =~ /switch-brocade/i) {
        Tharn::say($self->{conn}, "timeout $timeout", -swallow, '-timeout', 60);
        Tharn::say($self->{conn}, "logout", -swallow, '-blind');    
    }
    
    return $GOOD_STATUS;
}

## @name ::zone_add($$)
## @description
## Adds zones to the configuration $cfg_name
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## $status = $switch_obj->zone_add($cfg_name, $zones_array_ref)
##
## @arg $cfg_name Required: Config name
## @arg $zones_members_hash_ref Required: Hash reference with key/value pair
##                                        of zones and array reference of 
##                                        members that are to be added
##                                        to the configuration $cfg_name.
##
sub zone_add($$) {
   my ($self, $cfg_name, $zones_members_hash_ref) = @_;

   my ($cmd, $zone, $after, $prompt, $enabled_config, $zone_member, $match, $before) = '';

   my $hostname = $self->get_name();

   if ($self->{hosttype} =~ /switch-brocade/i) {
      foreach my $zone(keys %$zones_members_hash_ref) {
         foreach my $member(@{$zones_members_hash_ref->{$zone}}){
            (undef, undef, $after) = Tharn::say($self->{conn}, "zoneadd $zone,$member", -swallow, '-timeout', 60);

            if ($after =~ (/Usage:/i         ||
                           /syntax error/i   ||
                           /^\s*error:/i     ||
                           /^\s*unmatched quote/i) ) {
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: While adding member '$member' to the zone '$zone' on" . $self->get_name());
                return($ERR_STATUS);
            }
         }

         (undef, undef, $after) = Tharn::say($self->{conn}, "cfgadd $cfg_name,$zone", -swallow, '-timeout', 60);

         if ($after =~ (/Usage:/i         ||
                        /syntax error/i   ||
                        /^\s*error:/i     ||
                        /^\s*unmatched quote/i) ) {
             $self->{error} = $ERR_STATUS;
             push(@{$self->{errmsg}}, "ERROR: While adding zone '$zone' to the config '$cfg_name' on" . $self->get_name());
             return($ERR_STATUS);
         }
      }
   } elsif ($self->{hosttype} =~ /switch-cisco/i) {
      (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

      foreach my $zone(keys %$zones_members_hash_ref) {
         $cmd = "zone name " . $zone . " vsan " . $CISCO_VSAN_NUM . "";
         (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

         if ($after =~ m/Invalid command at/) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: While accessing zone '$zone' in the config '$cfg_name' on" . $self->get_name());
            return($ERR_STATUS);
         }

         foreach my $member(@{$zones_members_hash_ref->{$zone}}){
            $cmd = "member pwwn $member";
            (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

            if ($after =~ m/Invalid command at/) {
               $self->{error} = $ERR_STATUS;
               push(@{$self->{errmsg}}, "ERROR: While adding member '$member' to the zone '$zone' on" . $self->get_name());
               return($ERR_STATUS);
            }
         }

         $self->cisco_go_to_normal_prompt();
         (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);
      }

      # - - - - - - -  - - - - - - - - - - - - - -
      # Create a cisco 'zoneset' which is just a name
      # for a zone config.  We will use $config{config_name}
      # - - - - - - -  - - - - - - - - - - - - - -

      $cmd = "zoneset name " . $cfg_name . " vsan " . $CISCO_VSAN_NUM . "";

      (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

      if ($after =~ m/Invalid command at/) {
         $self->{error} = $ERR_STATUS;
         push(@{$self->{errmsg}}, "ERROR: While accessing zoneset '$cfg_name' on" . $self->get_name());
         return($ERR_STATUS);
      }
      foreach my $zone(keys %$zones_members_hash_ref) {
         # - - - - - - -  - - - - - - - - - - - - - -
         # Add the appropriate zones to this zoneset.
         # - - - - - - -  - - - - - - - - - - - - - -

         $cmd = "member " . $zone;
         (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", -swallow, '-noprompt', '-prompt', "$hostname.*\#", '-timeout', 30);

         if ($after =~ m/Invalid command at/) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: While adding zone '$zone' to the config '$cfg_name' on" . $self->get_name());
            return($ERR_STATUS);
         }
      }

       # - - - - - - -  - - - - - - - - - - - - - -
       # Get us back to the normal prompt.
       # - - - - - - -  - - - - - - - - - - - - - -

       $self->cisco_go_to_normal_prompt();
   }

   return $GOOD_STATUS;
}

## @name ::zone_remove($$)
## @description
## Removes zones from the configuration $cfg_name
##
## Note: All cisco zone configuration is done on vsan 1
##
## @synopsis
## $status = $switch_obj->zone_remove($cfg_name, $zones_array_ref)
##
## @arg $cfg_name Required: Config name
## @arg $zones_members_hash_ref Required: Hash reference with key/value pair
##                                        of zones and array reference of 
##                                        members that are to be added
##                                        to the configuration $cfg_name.
##
sub zone_remove($$) {
   my ($self, $cfg_name, $zones_members_hash_ref) = @_;

   my ($cmd, $zone, $after, $prompt, $enabled_config, $zone_member, $match, $before) = '';

   my $hostname = $self->get_name();

   if ($self->{hosttype} =~ /switch-brocade/i) {
      foreach my $zone(keys %$zones_members_hash_ref) {
         foreach my $member(@{$zones_members_hash_ref->{$zone}}){
            (undef, undef, $after) = Tharn::say($self->{conn}, "zoneremove $zone,$member", -swallow, '-timeout', 60);

            if ($after =~ (/Usage:/i         ||
                           /syntax error/i   ||
                           /^\s*error:/i     ||
                           /^\s*unmatched quote/i) ) {
                $self->{error} = $ERR_STATUS;
                push(@{$self->{errmsg}}, "ERROR: While removing member '$member' from the zone '$zone' on" . $self->get_name());
                return($ERR_STATUS);
            }
         }

         (undef, undef, $after) = Tharn::say($self->{conn}, "cfgremove $cfg_name,$zone", -swallow, '-timeout', 60);

         if ($after =~ (/Usage:/i         ||
                        /syntax error/i   ||
                        /^\s*error:/i     ||
                        /^\s*unmatched quote/i) ) {
             $self->{error} = $ERR_STATUS;
             push(@{$self->{errmsg}}, "ERROR: While removing zone '$zone' from the config '$cfg_name' on" . $self->get_name());
             return($ERR_STATUS);
         }
      }

   } elsif ($self->{hosttype} =~ /switch-cisco/i) {
      (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

      foreach my $zone(keys %$zones_members_hash_ref) {
         $cmd = "zone name " . $zone . " vsan " . $CISCO_VSAN_NUM . "";
         (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

         if ($after =~ m/Invalid command at/) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: While accessing zone '$zone' in the config '$cfg_name' on" . $self->get_name());
            return($ERR_STATUS);
         }

         foreach my $member(@{$zones_members_hash_ref->{$zone}}){
            $cmd = "no member pwwn $member";
            (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

            if ($after =~ m/Invalid command at/) {
               $self->{error} = $ERR_STATUS;
               push(@{$self->{errmsg}}, "ERROR: While removing member '$member' from zone '$zone' on" . $self->get_name());
               return($ERR_STATUS);
            }
         }

         $self->cisco_go_to_normal_prompt();
         (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "config t", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);
      }

      # - - - - - - -  - - - - - - - - - - - - - -
      # Create a cisco 'zoneset' which is just a name
      # for a zone config.  We will use $config{config_name}
      # - - - - - - -  - - - - - - - - - - - - - -

      $cmd = "zoneset name " . $cfg_name . " vsan " . $CISCO_VSAN_NUM . "";

      (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", '-noprompt', '-prompt', "$hostname.*\#", -swallow, '-timeout', 30);

      if ($after =~ m/Invalid command at/) {
         $self->{error} = $ERR_STATUS;
         push(@{$self->{errmsg}}, "ERROR: While accessing zoneset '$cfg_name' on" . $self->get_name());
         return($ERR_STATUS);
      }

      foreach my $zone(keys %$zones_members_hash_ref) {
         # - - - - - - -  - - - - - - - - - - - - - -
         # Add the appropriate zones to this zoneset.
         # - - - - - - -  - - - - - - - - - - - - - -

         $cmd = "no member " . $zone;
         (undef, undef, $after, $prompt) = Tharn::say($self->{conn}, "$cmd", -swallow, '-noprompt', '-prompt', "$hostname.*\#", '-timeout', 30);

         if ($after =~ m/Invalid command at/) {
            $self->{error} = $ERR_STATUS;
            push(@{$self->{errmsg}}, "ERROR: While removing zone '$zone' from the config '$cfg_name' on" . $self->get_name());
            return($ERR_STATUS);
         }
      }

       # - - - - - - -  - - - - - - - - - - - - - -
       # Get us back to the normal prompt.
       # - - - - - - -  - - - - - - - - - - - - - -

       $self->cisco_go_to_normal_prompt();
   }

   return $GOOD_STATUS;
}

## @name ::send_switch_cmd($;$)
##
## @description
## Executes whatever command is passed in. Returns command output i.e.
## array ref of output of say, whatever that may be.
##
## @synopsis
## $status = $switch_obj->send_switch_cmd($input_cmd)
##
## @arg $input_cmd Required: string containing command to be executed
## @arg $timeout Optional: Sets command timeout value. Defaults to 60 secs.
##
## ASSUMPTIONS: since input command can be anything, the parsing of the command
## output is not done here and is the responsibility of the calling program
##
sub send_switch_cmd($;$)
{
   my ($self, $input_cmd, $timeout) = @_;
   my ($prompt, $after) = (undef, undef);

   my @say_output=();

   my $cmd_timeout = 60;
   $cmd_timeout = $timeout if (defined $timeout);

   #
   # After performing 'FirmwareRestart' the switch does not return
   # any prompt and so we will be using '-noprompt' for that operation.
   # If we do not use '-noprompt' then Tharn::say will fail.
   #
   $prompt = "-noprompt" if ($input_cmd =~ m/FirmwareRestart/i);

   $self->clear_error();

   # Execute the input command.
   (undef, undef, $after) = Tharn::say($self->{conn}, "$input_cmd", $prompt, -swallow, '-timeout', $cmd_timeout);

   $after =~ s/^\s+//;

   if ($after =~ (/Usage:/i         ||
                  /syntax error/i   ||
                  /^\s*error:/i     ||
                  /Invalid command at/i ||
                  /^\s*unmatched quote/i) ) {
      $self->{error} = $ERR_STATUS;
      push(@{$self->{errmsg}}, "ERROR: Failed to execute command '$input_cmd' from the switch " . $self->get_name());
      return($ERR_STATUS);
   } elsif ($after =~ /.*\r\n$/) {
      @say_output = split ("\r\n", $after);
   } elsif ($after ne q{} && $after !~ m/(\r|\n)/) {
      #
      # In case if switch output is returned in a scalar format then the following regex is
      # used to convert it into an array.
      # Actually we came across an issue with the armadillo switches where we need to retrieve
      # values of only online ports and as the switch output has returned in a scalar context i.e.
      # without any new line characters, we have used this regex to convert it into an array, which
      # helped us in retrieving the values of online ports from an array easily.
      #
      # Sample Output :-
      # -------------
      # 20120611 102251 [01_test] . Comment: After : Info^M
      # 20120611 102251 [01_test] . Device Status         = Good
      # 20120611 102251 [01_test] . Device                = "FibreBridge 6500N"
      # 20120611 102251 [01_test] . Serial Number         = FB6500N101113
      # 20120611 102251 [01_test] . Device Version        = A64Y
      # 20120611 102251 [01_test] . Build Number          = A64Y
      # 20120611 102251 [01_test] . Build Date            = "Mar 15 2012" 17:53:03
      # 20120611 102251 [01_test] . Flash Revision        = 1
      # 20120611 102251 [01_test] . CLI Revision          = 1.75
      # 20120611 102251 [01_test] . Base version          = 51.01
      # 20120611 102251 [01_test] . Version Number        = 1.50
      # 20120611 102251 [01_test] . User-defined name     = "        "
      # 20120611 102251 [01_test] . World Wide Name       = 20 00 00 10 86 60 2C 76
      # 20120611 102251 [01_test] . FC1 Node Name         = 20 00 00 10 86 60 2C 76
      # 20120611 102251 [01_test] . FC1 Port Name         = 21 00 00 10 86 60 2C 76
      # 20120611 102251 [01_test] . FC1 Data Rate         = 8Gb
      # 20120611 102251 [01_test] . FC1 Connection Mode   = ptp
      # 20120611 102251 [01_test] . FC2 Node Name         = 20 00 00 10 86 60 2C 76
      # 20120611 102251 [01_test] . FC2 Port Name         = 22 00 00 10 86 60 2C 76
      # 20120611 102251 [01_test] . FC2 Data Rate         = N/A
      # 20120611 102251 [01_test] . FC2 Connection Mode   = N/A
      # 20120611 102251 [01_test] . MP1 MAC Address       = 00 10 86 60 2C 76
      # 20120611 102251 [01_test] . MP1 IP Address        = 10.227.184.103
      # 20120611 102251 [01_test] . MP1 IP Subnet Mask    = 255.255.0.0
      # 20120611 102251 [01_test] . MP1 IP Gateway        = 10.227.184.1
      # 20120611 102251 [01_test] . MP1 IP DHCP           = disabled
      # 20120611 102251 [01_test] . MP2 MAC Address       = 00 10 86 60 2C 77
      # 20120611 102251 [01_test] . MP2 IP Address        = 192.168.10.29
      # 20120611 102251 [01_test] . MP2 IP Subnet Mask    = 255.255.0.0
      # 20120611 102251 [01_test] . MP2 IP Gateway        = 0.0.0.0
      # 20120611 102251 [01_test] . MP2 IP DHCP           = disabled
      # 20120611 102251 [01_test] . Active Configuration  = ATTO
      #
      @say_output = $after =~ /\G(.*\n|.+)/g;
   } else {
      @say_output = split ("\n", $after);
   }

   # If we're running on an older version of Nate prior to late July '09
   # we need to check if the say() output echoes the command line as part
   # of the output.  If that's the case then strip out the echoed line
   if ( $say_output[0] =~ m/^$input_cmd$/ ) {
       shift @say_output;
   }

   #
   # FirmwareRestart is an Armadillo command. After performing this operation, the switch 
   # sessions should be reconnected. Otherwise, the scripts will fail.
   #
   if ($input_cmd =~ m/FirmwareRestart/i) {
      sleep 30;

      $self->switch_disconnect();
      $self->switch_connect();
   }

   return \@say_output;
}

1;  # Needed for *.pm's
