# -*-perl-*x
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##     Library functions for wafliron
##
## @description
##     The wafliron library consists of the following functions:
##
##     wafliron_start - run wafliron on aggregate or a traditional volume
##
##     wafliron_status - return wafliron status
##
##     wafliron_stop - stop wafliron
##
##     wafliron_wait - wait until walfiron complete
##
##     wafliron_commit - commit change log for wafliron optional commit
##
##     wafliron_reject - reject change log for wafliron optional commit
##
##     wafliron_runcommit - wrapper to start wafliron in Optional commit
##     mode and then commit the change log.
##
##     wafliron_runreject - wrapper to start wafliron in Optional commit
##     mode and then reject the change log.
##
##     wafliron_review - run the 'aggr wafliron review' command
##
##     wafliron_start_on_root - run wafliron on the root volume
##
##     wafliron_on_root - Run wafliron from the boot menu
##
##     wafliron_pick_new_root - Use bootmenu to pick new root volume
##
##     wafliron_create_temp_root - Create a volume suitable to be a root volume
##                                  and copy all necessary system files to it
##
##     wafliron_setup_container - Creates a new tradvol/aggregate using commonIronParams
##
##     wafliron_setup_vvols - Creates a new tradvol/flexvol using commonIronParams
##
##     wafliron_cleanup_vvols - Cleansup all aggrs & vols created by
##                              wafliron_setup_container and wafliron_setup_vvols
##
##     wafliron_populate_volumes - Created requested files and snapshots using commonIronParams
##
##     wafliron_run_ironops - Run iron along with specific operations
##
## @usage
##     use wafliron;
##
## @author dl-johny@netapp.com (FSS WAFL Quality)
##
## @status public

package Wafliron;

use strict;

use Exporter qw(import);

BEGIN {
    our @EXPORT = qw(&wafliron_start
        &wafliron_status
        &wafliron_stop
        &wafliron_wait
        &wafliron_volume_order
        &wafliron_oc_supported
        &wafliron_commit
        &wafliron_reject
        &wafliron_runcommit
        &wafliron_runreject
        &wafliron_review
        &get_container
        &wafliron_runreject
        &wafliron_logfiles_present
        &wafliron_populate_vol
        &wafliron_cleanup_vol
        &wafliron_start_on_root
        &wafliron_on_root
        &wafliron_pick_new_root
        &wafliron_create_temp_root
        &wafliron_setup_container
        &wafliron_setup_vvols
        &wafliron_cleanup_vvols
        &wafliron_populate_volumes
        &wafliron_run_ironops);
} ## end BEGIN

use Tharn;
use Carp qw(croak);
use SPARSE::Util::Logger qw(logWarn logTrc logDbg logXXX);
use Volume
    qw(vol_exists vol_root :df vol_size vol_status vol_container vol_create_flex
       vol_create vol_disk_free vol_online vol_offline vol_purge);
use Aggregate
    qw(aggr_exists aggr_online aggr_offline aggr_status
       aggr_create aggr_root aggr_purge);
use Filer::Version qw(get_wafl_version
    WV_BOILERMAKER_MIN
    WV_ROLLINGROCK_MIN
);
use filer qw(filer_cli_test_say filer_cli_advanced_say filer_options_get
             filer_cli_diag_say filer_cli_admin_say filer_sync filer_options_set);
use FS::Util::Utilities qw(uSystem write_file);
use FS::Util::nb_tail;
use FS::Util::Setup qw(copy_system_files cold_start_filer);
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR BOOLEAN OBJECT HASHREF ARRAYREF);
use SnapCmds qw(snapshot_create snapshot_delete snapshot_status
                snapshot_list snaprestore_volume);
use Storage::Misc qw(misc_printflag misc_setflag);
use wafl_utils qw(df);
use boot qw(transit_to reboot_to halt reboot boot_state);
use Utils qw(make_conn make_conn_and_complain);
use ATI::ati_ontapi
    qw(ATI_SUCCESS ATI_ONTAPI_TRANSPORT_HTTP ATI_ONTAPI_VERSION_DEFAULT
    ATI_ONTAPI_PASSWORD_DEFAULT ati_ontapi_create_context ati_ontapi_call_api);
use FS::Util::Storage_Ems qw(ems_log_get_mark ems_log_dump_by_id);
use Utils::Units qw(:constants);    # UNIT_B UNIT_KB UNIT_MB etc.
use Admin::License qw(license);
use Hammer
    qw(nfs_hammer_start nfs_hammer_stop cifs_hammer_start cifs_hammer_stop);
use CIFS::Setup qw(is_cifs_setup cifs_setup);
use waflobjCreator qw(create_directory);
use Utils::Units qw(convert_to_bytes);
use Volume::Setup qw(setup_volumes);
use QtreeCmds qw(qtreeCreate qtreeSecurity);
use Admin::License qw(license is_licensed);
use NACL::MTask::SystemLogDetector;
use NACL::STask::Client::Mount;
use NACL::STask::Node;
use NACL::C::Volume;
use NACL::C::Client;

use Dense qw(dense_start dense_stop dense_undo dense_toggle
             dense_wait_for_status dense_off_undo dense_config);
use Dense::Clone
    qw(dense_sis_clone_start dense_sis_clone_status dense_sis_clone_wait_for_finish);
use filer qw(filer_date_get filer_vol_option_set);
use Perf::Reallocation qw(DONE);
use Compression::Constants qw(:all);
use Compression::CompressionScanner;
use Compression::Decompress;
use NOTE::Filer;
use Data::Dumper;
use RaidSqa;
use POSIX;
use constant LOCAL_WAFLIRON_TIMEOUT => 2400;

# %id_table is used to save ($fc, $container_name, $mount_root, $tail)
# when walfiron_start(wait=>0) is called. So the caller does not need to pass
# those arguments again when call wafliron_wait().

my %id_table = ();

# Need to save the original root volume and aggr
# for IRON_ON_ROOT
my ( $orig_root_aggr, $orig_root_volume, $flags_hash, @Nfs_hamid, @Cifs_hamid );

# after wafliron start, how long should we wait when the console output pauses
my $START_MAXIDLE = 600;

## @name ::wafliron_start($fc, $container_name[, %opts])
##
## @summary
##     start wafliron on aggregate or traditional volume
##
## @description
##
##     On Anchorsteam or later releases, the library uses command
##     "aggr wafliron start"(it works on both aggregate and traditional volume)
##     On pre-Achorsteam release, the library uses command
##     "vol wafliron start";
##
##     If the $container_name is a flexible volume, the library will
##     look up the containing aggregate.
##
##     Depends on options, the function can return immediately after
##     start wafliron or wait until wafliron is complete
##
## @arg $fc Required: filer connection object.
##
## @arg $container_name Required: name of aggregate, traditional
##                                volume or flexible volume
##
## @arg %opts Optional: options can be set:
##
##      wait=>1 or 0, default is 1
##         if wait is set to 1, the wafliron_start() will wait for
##         wafliron completed; otherwise, return immediately after start.
##
##      force=>1  (command line option -f)
##      triage=>1 (command line option -t)
##
##      compression=>1 (command line option -c).
##      Available from IC.0, to enable compression-related checks.
##
##      volume_order=>volumes list (command line option -v,
##                                  available from 7.3, e.g vol3,vol1,vol2).
##      From IC.0, volume ordering is available.
##
##      oc => 1 or 0 or prev_cp.
##       if oc is 1, wafliron is run in Optional Commit mode.
##       if oc is prev_cp, wafliron is run in Optional Commit prev_cp mode.
##       if oc is 0, regular wafliron is run. Default is 0.
##      From BR.0 and IC.1 optional commit is available
##
##      response_list => a hash with key-value pair as
##                       vol/aggr name and reponse to be sent to the
##                       vol/aggr during prev_cp
##
##      new_resp_list => a hash reference
##                       new_resp_list =>
##                                      {interactive => 0,
##                                       aggr_resp   => 'r',
##                                       response_list => {"volume_name1"  => snap_id1,
##                                                         "volume_name21" => snap_id2}}
##                       this runs "wafliron start testaggr -o -prev_cp volume_name1:snap_id1,volume_name2:snap_id2"
##
##      new_resp_list => a hash reference
##                       new_resp_list =>
##                                      {interactive => 1,
##                                       aggr_resp   => 'r',
##                                       response_list => {"volume_name1"  => snap_id1,
##                                                         "volume_name21" => snap_id2}}
##                       this runs "wafliron start testaggr -o -prev_cp" and inputs for volume_name1 and volume_name2
##                       will be asked in prev_cp selection prompts
##
##      new_resp_list => a hash reference
##                       new_resp_list =>
##                                      {interactive => 1,
##                                       aggr_resp   => 'r',
##                                       response_list => {"incomplete_list = "vol1:s1,vol2,vol3",
##                                                         "vol2"  => snap_id1,
##                                                         "vol3" => snap_id2}}
##                       this runs "wafliron start testaggr -o -prev_cp vol1:s1,vol2,vol3" and inputs for vol2 and vol3
##                       will be asked in prev_cp selection prompts (This is to give some vvol inputs with the prev_cp
##                       command and for other volumes input is given in the selection prompts.
##
##      inject_file=>'some_file' (command line option -i <inject-file>)
##
##      timeout   wafliron timeout value,
##                $WAFLIRON_TIMEOUT from commonParam.pl is recommended
##
##      maxidle   after wafliron start, how long should we wait when the console output pauses
##                Default to $START_MAXIDLE
##
##      scan_speed speed of the scan, should be between -1 and 10000. Default is 10000.
##                 scan_speed=0 means to set the default dynamic scan speed
##                 scan_speed=-1 means to suspend scanning
##
##      DSC       1/0  Enable/Disable DSC (Defer Summary Check) flag
##
##      LSF       1/0  Enable/Disable LSF (Local Status Files) flag
##
##      panic_on_corruption => 1/0 Enable/Disable iron panic on corruption
##
##      setflags => a hash with a list of flags and values to be set before running
##                  wafliron
##                  The script has to explicitly restore the flags to their original value
##                  The key is the name of the flag and the value is the value to be set
##
##      root_vol => For IRON_ON_ROOT this parameter should contain the name of the volume on
##                  the aggr that should become the root volume when ironing. For the corruption
##                  suite, the object of the iron is an aggr, but for IRON_ON_ROOT we need to setup
##                  the corrupted volume as the root. This parameter contains the name of the corrupted
##                  volume
##                  If this parameter is set to "AUTOROOT", the AUTOROOT volume which is automatically
##                  created by the filer after setting new root aggr without calling vol_pick_root will
##                  be selected as the root volume when ironing.
##
##      tmp_root => For IRON_ON_ROOT this parameter indicates a temporary root volume has already been
##                  established and therefore should not be done inside this library. For the corruption
##                  suite, when the corruption is on the root aggr, a temporary root volume is created by
##                  Corruption.pm, so there is no need to create one here in this library
##
## @return value
##      if (wait == 0), return id if wafliron start successfully;
##                      the id is used when caller call wafliron_wait() later.
##
##                      id is a reference to a list
##                      (connection_obj, container_name)
##
##      if (wait == 1), return (corruption_found, corruption_messages, cp_hash) if
##                      wafliron completed;
##                      corruption_found == 1 means "found corruption"
##                      cp_hash is a reference to a hash with key as vol/aggr name
##                      and value as a reference to an array containing the cp values
##                      that were provided by the filer during prev_cp.
##                      Contains nothing if oc is not set to prev_cp
##
## @example my $id = wafliron_start($fc, $container_name, wait=>0);
##          my ($corruption_found, $message) =
##                   wafliron_start($fc, $container_name);

sub wafliron_start ($$;%) {
    my ( $fc, $name, %opts ) = @_;
    my ( $corruption_found, $return_message, %cp_hash );
    my ( $ref, $fconn, $host_obj, $passwd, $ver );

    my $container_name = get_container( $fc, $name );
    my $container_type = wafliron_type( $fc, $name );
    my $container_status
        = ( $container_type == 'aggr' )
        ? aggr_status( $fc, $container_name )
        : vol_status( $fc, $container_name );

    my $wait   = defined( $opts{wait} ) ? $opts{wait} : 1;
    my $force  = $opts{force}           ? " -f"       : "";
    my $triage = $opts{triage}          ? " -t"       : "";

    my $node   = NACL::STask::Node->new(hostrec => $fc->hostp());

    # Enable/Disable Defer Summary Check
    my $dsc = defined( $opts{DSC} ) ? $opts{DSC} : 1;

    # Enable/Disable local status files
    my $lsf = defined( $opts{LSF} ) ? $opts{LSF} : 0;

    # Enable/Disable iron panic on corruption
    my $panic_on_corruption
        = defined( $opts{panic_on_corruption} )
        ? $opts{panic_on_corruption}
        : undef;

    # compression is available after IC.0
    my $compression = $opts{compression} ? " -c" : "";
    my $scan_speed = defined( $opts{scan_speed} ) ? $opts{scan_speed} : 10000;
    my $inject_file
        = defined( $opts{inject_file} ) ? " -i $opts{inject_file}" : "";
    my $timeout
        = defined( $opts{timeout} ) ? $opts{timeout} : LOCAL_WAFLIRON_TIMEOUT;
    my $maxidle = defined( $opts{maxidle} ) ? $opts{maxidle} : $START_MAXIDLE;
    ## volume ordering is available only after IC.0
    my $volume_order
        = defined( $opts{volume_order} ) ? " -v $opts{volume_order}" : "";
    ## optional commit is available  only after IC.1
    my $opt_commit = defined( $opts{oc} ) ? $opts{oc} : 0;

    # New parameters for IRON_ON_ROOT
    # root is the name of the volume that will become the root volume
    #  when IRON_ON_ROOT=core
    my $root
        = defined( $opts{root_vol} ) ? "$opts{root_vol}" : "$::FILER_VOL";

    # tmp_root indicates a new root volume has already been created and it
    #  should not be done in this routine
    my $tmp_root_done = defined( $opts{tmp_root} ) ? "$opts{tmp_root}" : 0;

    # When the function wafliron_pick_new_root is called and even though after
    # setting the flags If there is any panic when onlining
    # the aggregate when filer is up, then powercycle can be set so that the
    # aggregate will be onlined from maintenance mode
    my $powercycle = defined( $opts{powercycle} ) ? "$opts{powercycle}" : 0;

    #my $prev_cp = defined($opts{prev_cp})? $opts{prev_cp} : 0;
    my $response_list
        = defined( $opts{response_list} ) ? $opts{response_list} : 0;

    #New response list to support the new prev_cp behavior
    my $new_resp_list
        = defined( $opts{new_resp_list} ) ? $opts{new_resp_list} : 0;

    #Get the list of flags that need to be set before running wafliron
    my $setflags = defined( $opts{setflags} ) ? $opts{setflags} : undef;

    # validate if response list is provided
    croak("Response list is required for prev_cp mode.")
        if ( $opt_commit eq "prev_cp"
             && $response_list == 0
             && $new_resp_list == 0 );

    # set panic on corruption
    if ( defined( $opts{'panic_on_corruption'} ) ) {
        filer_cli_diag_say( $fc,
                  "setflag wafl_iron_panic_on_corruption $panic_on_corruption" );
    }

    # set all the required flags
    while ( my ( $flag, $value ) = each(%$setflags) ) {
        say( $fc, "priv set -q test; setflag $flag $value", -timeout, 30 );
    }

    if ($::IRON_ON_ROOT) {

        logDbg(
            "Wafliron::wafliron_start - IRON_ON_ROOT=$::IRON_ON_ROOT, name=$name, oc=$opt_commit, root=$root, tmp_root=$tmp_root_done"
        );

        my $stat_obj = aggr_status( $fc, $container_name );
        croak "Wafliron::wafliron_start - aggr_status failed: $@" if $@;

        # If we have not already saved the root volume and aggr
        # do it now.
        if ( !$orig_root_volume ) {
            $orig_root_volume = Volume::vol_root($fc);
            $orig_root_aggr   = Aggregate::aggr_root($fc);
        }

        # must use the console for new root operations
        my $conn_type = $fc->conntype();
        if ( $conn_type !~ m/console|pm_cons/ ) {
            $fconn = connect( $fc->hostp()->hostname() . "#console", -login );
        } else {
            $fconn    = $fc;
        }
        $host_obj = $fconn->hostp();
        $passwd   = $host_obj->default_password();
        $ver      = get_wafl_version($host_obj);

        # Store flags, to restore after filer reboot
        $flags_hash = misc_printflag (
                             host => $host_obj,
                             flag => "wafl_metadata_visible "
                                    ."wafl_iron_panic_on_corruption "
                                    ."wafl_inconsistent_panic_testing "
                                    ."wafl_metadata_checks_in_raid "
                                    ."wafl_iron_lazy_dist_checks "
                                    ."wafl_snapiron_enabled "
                                    ."wafl_maxvvbn_disabled "
                                    ."wafl_lwr_trip "
                         );
        logDbg("Wafliron::wafliron_start - flags_hash = ". Dumper $flags_hash);

        # Set nfs.persistent.enable
        filer_cli_diag_say( $fc, "registry set nfs.persistent.enable on" );

        # IRON_ON_ROOT can contain three values, core, multi, and spare
        if ( $::IRON_ON_ROOT =~ /core/ ) {

            logDbg(
                "Wafliron::wafliron_start - Calling wafliron_on_root on account of IRON_ON_ROOT=$::IRON_ON_ROOT"
            );
            ## volume name is only used when IRON_ON_ROOT
            croak(
                "Wafliron::wafliron_start - root_vol is required for IRON_ON_ROOT=core"
            ) if ( $root eq "" );

# For 'core' we will pick the object of the wafliron as the root aggr and the $root variable
# for the root volume. root_vol => $root is set by Corruption.pm
            wafliron_pick_new_root(
                $fconn, $name, $root,
                reboot        => 0,
                scan_speed    => $scan_speed,
                tmp_root_done => $tmp_root_done,
                powercycle    => $powercycle,
                type          => $container_status->type(),
                setflags      => $flags_hash
            );

# wafliron_on_root returns the $ref variable used when the caller has specified wait => 0
            if ( !$wait ) {
                return (
                    wafliron_on_root(
                        $fconn, $name,
                        root_vol      => $root,
                        oc            => $opt_commit,
                        force         => $force,
                        timeout       => $timeout,
                        wait          => $wait,
                        response_list => $response_list,
                        new_resp_list => $new_resp_list,
                        type          => $stat_obj->type()
                    )
                );
            } else {

# wafliron_on_root returns corruption_found, and the console messages from the wafliron start command
                ( $corruption_found, $return_message ) = wafliron_on_root(
                    $fconn, $name,
                    root_vol      => $root,
                    oc            => $opt_commit,
                    force         => $force,
                    timeout       => $timeout,
                    wait          => $wait,
                    response_list => $response_list,
                    new_resp_list => $new_resp_list,
                    type          => $stat_obj->type()
                );

           # If we have committed or rejected the changes already
           # go back to the original root volume, otherwise just return status
           # the commit will be done later

                if ( $opt_commit =~ /commit|reject|0/ ) {

                    # Set the root volume back to what it was
                    wafliron_pick_new_root( $fconn, $orig_root_aggr,
                        $orig_root_volume, reboot => 1, setflags => $flags_hash );

                    # Lets get rid of the previous root volume
                    vol_purge( $fconn, $orig_root_volume . "_temp" );
                    aggr_purge(
                        $fconn,
                        $orig_root_aggr . "_temp",
                        destroy_corrupt_volumes => 1
                    );
                } ## end if ( $opt_commit =~ /commit|reject|0/)
                return ( $corruption_found, $return_message );
            } ## end else [ if ( !$wait )

        } elsif ( $::IRON_ON_ROOT =~ /multi|spare/ ) {

# Corruption.pm will create the temp root when corruption_type=aggr
# so don't do it here, otherwise we need to create temporary aggr and volume and switch to them
# before we start wafliron from the console
            if ( !$tmp_root_done ) {

                # Create a temporary root aggr and volume
                wafliron_create_temp_root(
                    $fc,
                    aggr => $orig_root_aggr . "_temp",
                    vol  => $orig_root_volume . "_temp",
                    setflags => $flags_hash
                );
            } ## end if ( !$tmp_root_done )

            # fall thru to regular iron start
        } else {
            die("Wafliron::wafliron_start - Unknown IRON_ON_ROOT value");
        }
    } ## end if ($::IRON_ON_ROOT)

    my ( $tail, $start_cmd, $oc_message );

    # Make sure passed in scan_speed is a valid value.
    if ( $scan_speed < -1 ) {
        $scan_speed = -1;
    } elsif ( $scan_speed > 10000 ) {
        $scan_speed = 10000;
    }

    filer_cli_advanced_say( $fc, "wafl scan speed $scan_speed" );

    filer_cli_diag_say( $fc, "setflag wafl_iron_defer_summary_check $dsc" );

    filer_cli_diag_say( $fc, "setflag wafl_iron_status_local $lsf" );

    # fail if wafliron is already running
    if ( _wi_status( $fc, $container_type, $container_name ) ) {
        croak(
            "Wafliron::wafliron_start - wafliron is running on $container_name"
        );
    }

    # Create an object to tail-read messages file from wafliron start.
    # Since the corruption messages can have many different patterns,
    # the function will not try to match any specific pattern. Instead
    # it will return any message from wafliron Starting to wafliron completed,
    # leave it to the caller to decide what to look for

    $tail = NACL::MTask::SystemLogDetector->new(
                command_interface => $node,
                set_log_file      => ['messages'] );
    $tail->start();

    my ( $before, $after, $prompt );

# figure out what 'type' value to expect in the EMS event log (used for detecting wafliron start)
    my $start_match_type;
    {
        # use volume or aggregate?
        my $status_obj = aggr_status( $fc, $container_name );
        croak "Wafliron::wafliron_start - aggr_status failed: $@" if $@;

        my $object_type = $status_obj->type();
        if ( ( $object_type eq 'flex' ) or ( $object_type eq 'trad' ) ) {
            $start_match_type = 'volume';
        } elsif ( $object_type eq 'aggr' ) {
            $start_match_type = 'aggregate';
        } else {
            croak(
                "Wafliron::wafliron_start - unknown object type spec: $object_type (1)"
            );
        }
    }

    # track wafliron state
    my $wafliron_started = 0;

    # capture current position in the ems log (for wafl iron start notification)
    logcomment("Wafliron::wafliron_start - marking ems log");
    my $mark = ems_log_get_mark( host => $fc->{hostp} );

    my @say_args_for_start = ( -timeout, $timeout, -maxidle, $maxidle );
    #if ( ( $fc->conntype eq 'pm_cons' ) || ( $fc->conntype eq 'telnet' ) ) {
        # some connection types should not wait looking for a prompt
        # (because the console is noisy and prompts are often lost)
        push( @say_args_for_start, -blind );
    #} ## end if ( ( $fc->conntype eq...

    # start wafliron and monitor the ems log
    if ( !$opt_commit ) {
        $start_cmd
            = $container_type
            . " wafliron start "
            . $container_name
            . $compression
            . $force
            . $triage
            . $inject_file
            . $volume_order;
        say( $fc, "priv set -q advanced; $start_cmd", @say_args_for_start );
    } else {
        # setflag will be removed later if necessary
        say( $fc,
            "priv set -q test; setflag wafl_iron_optional_commit_enabled 1",
            -timeout, $timeout );

        my $status_obj = aggr_status( $fc, $container_name );
        if ( $status_obj->state() eq "online" ) {
            # check to see if any snapsnap is getting created and wait
            my ( $endtime, $sstatus, $snap_status, @snap_status );
            $endtime = time() + $timeout;
            $sstatus = 'creating';
            while ( $sstatus eq 'creating' && $endtime > time() ) {
                @snap_status = snapshot_status( $fc, " -A $container_name" );
                $sstatus = '';
                foreach $snap_status ( @snap_status ) {
                    if ( $snap_status->{status} eq 'creating' ) {
                        $sstatus = $snap_status->{status};
                        last;
                    } else {
                        $sstatus = $snap_status->{status};
                    }
                }
                Tharn::sleep(15);
            } ## end while
        } ## end if

        # it's required to offline aggregate before start optional commit
        eval {
            aggr_offline( $fc, $container_name, force => 1, all => 1 );
        };
        if ($@) {
            my $croak = 1;
            if ($@ =~ /Unrecognized format for aggr status/i) {
                $status_obj = aggr_status( $fc, $container_name );
                $croak = 0 if $status_obj->state() eq "offline";
            }
            if ( $croak ) {
                croak("Wafliron::wafliron_start - Could not offline "
                          . "aggregate for wafliron optional commit: $@");
            }
        }

        if ( $opt_commit == 1 ) {
            # prepare to match expected output
            $start_cmd
                = $container_type
                . " wafliron start "
                . $container_name
                . $compression
                . $force
                . $triage
                . $inject_file . " -o "
                . $volume_order;
            say( $fc, "priv set -q advanced; $start_cmd",
                @say_args_for_start );
        } elsif ( $opt_commit eq "prev_cp" ) {
            # must use the console for prev_cp operations
            my ($conn, $conn_type);
            $conn_type = $fc->conntype();
            if ( $conn_type !~ m/console|pm_cons/ ) {
                $conn = connect( $fc->hostp()->hostname() . "#console", -login );
            } else {
                $conn = $fc;
            }

            $start_cmd
                = "priv set -q advanced;"
                . $container_type
                . " wafliron start "
                . $container_name
                . $compression
                . $force
                . $triage
                . $inject_file
                . " -o -prev_cp "
                . $volume_order;

            # Go handle the prev_cp prompt
            _do_prev_cp_prompt( $conn, $start_cmd, $response_list, $new_resp_list );

            my $text = "";

            # Go handle the password prompt
            _wafliron_login( $conn, $text );

        } else {
            croak(
                "Wafliron::wafliron_start - Invalid value passed for oc. Valid values are 0,1 and prev_cp"
            );
        }
    } ## end else [ if ( !$opt_commit )

    my $start_timeout = 240;
    my $endtime       = time() + $start_timeout;
    my @wafl_check_info_events;
    $Log->trace(
        "Wafliron::wafliron_start - Waiting for wafliron to start...");
outer:
    while ( time() < $endtime ) {
        $Log->trace("Wafliron::wafliron_start - ...in start wait loop...");

        # update the marker
        my $old_mark = $mark;
        $mark = ems_log_get_mark( host => $fc->{hostp} );

        # fetch the new ems events
        my $events;
        eval {
            $events = ems_log_dump_by_id(
                host      => $fc->{hostp},
                timestamp => $old_mark
            );
        };
        if ( $@ =~ /TIMEOUT/i ) {
            $mark = $old_mark;
        } elsif ( $@ ) {
            $Log->warn("eval error = $@");
        }

        $Log->trace( "events hash is \n" . Dumper($events) . "\n" );

        foreach my $id ( keys %{$events} ) {
            my $hash = $events->{$id};
            if ((   exists( $hash->{__event_def} )
                    && ( $hash->{__event_def} eq 'wafl.iron.start' )
                ))
            {
                $wafliron_started = 1;
                last outer;
            } elsif (
                (   exists( $hash->{__event_def} )
                    && ( $hash->{__event_def} eq 'wafl.check.info' )
                )
                && (exists( $hash->{msg} )
                    && ( $hash->{msg}
                        =~ /\s+$start_match_type\s+$container_name/ )
                )
                )
            {
                if ($opt_commit
                    && (   ( $hash->{msg} =~ /not enough free space/i )
                        || ( $hash->{msg} =~ /cannot start iron/i )
                        || ( $hash->{msg} =~ /no space left on device/i ) )
                    )
                {

                    # dump out wafl.check.info events seen so far, if any
                    if (@wafl_check_info_events) {
                        $Log->warn(
                            "Wafliron::wafliron_start - matched wafl.check.info event(s) for $start_match_type $container_name. these may (or may not) be errors. if wafliron fails to start, these events may (or may not) contain the reason. if wafliron starts successfully, then this warning can probably be ignored. see burt 369173 for more info. matched events:\n"
                                . Dumper( \@wafl_check_info_events )
                                . "\n" );
                    } ## end if (@wafl_check_info_events)

                    $tail = undef;
                    croak(
                        "Wafliron::wafliron_start - failed with error (check logs for related wafl.check.info events): '"
                            . $hash->{msg}
                            . "'" );
                } else {
                    push( @wafl_check_info_events, $hash );
                }
            } else {

                # must be some unrelated/uninteresting event
            }
        } ## end foreach my $id

        Tharn::sleep(10);
    } ## end while ( time() < $endtime)

    if (@wafl_check_info_events) {
        $Log->warn(
            "Wafliron::wafliron_start - matched wafl.check.info event(s) for $start_match_type $container_name. these may (or may not) be errors. if wafliron fails to start, these events may (or may not) contain the reason. if wafliron starts successfully, then this warning can probably be ignored. see burt 369173 for more info. matched events:\n"
                . Dumper( \@wafl_check_info_events )
                . "\n" );
    } ## end if (@wafl_check_info_events)

    # Let's check syslog messages to see if wafliron started.
    # EMS logs is known to be unreliable, so we re-confirm
    # using syslog messages.
    if ( !$wafliron_started ) {

        my $tail_messages = $tail->get_logs( skip_cached_logs => 1 );
        logDbg("Wafliron::wafliron_start - syslog messages:\n$tail_messages");
        if ( $tail_messages =~ /wafl.iron.start/ ) {
            $Log->trace(
                "Wafliron::wafliron_start - wafliron start message found in syslog but not in ems log"
            );
            $wafliron_started = 1;
        } ## end if ( $tail_messages =~...
    } ## end if ( !$wafliron_started)

    # did it start, or did we time out?
    if ( !$wafliron_started ) {
        die (
                "Wafliron::wafliron_start - wafliron did not"
                       . " start within the timeout period: $timeout"
        );
    } ## end if ( !$wafliron_started)

    $Log->trace(
        "Wafliron::wafliron_start - Now, wafliron has started on '$container_name'."
    );

    $ref = [ $fc, $container_name ];
    my $mount_root = '';
    $id_table{$ref} = [
        (   $fc,      $container_name, $tail, $mount_root,
            $timeout, $container_type, $volume_order
        )
    ];

    if ( !$wait ) {
        return ($ref);
    }

    ( $corruption_found, $return_message ) = wafliron_wait($ref);

# If we are running IRON_ON_ROOT=multi or spare, but we have not committed the changes
# then we will let wafliron_commit restore the old root volume, otherwise we
# need to restore the root volume back to the original values here.
# The tmp_root_done flag comes from Corruption.pm and is used when it has
# created the temporary root volumes so it will also cause us to not restore
# the root volume here.
    my $ior_string
        = ( defined($::IRON_ON_ROOT) ) ? $::IRON_ON_ROOT : "undefined";
    logDbg(
        "Wafliron::wafliron_start - iron on root is: $ior_string, opt_commit is: $opt_commit, and tmp_root_done is: $tmp_root_done"
    );
    if ( $::IRON_ON_ROOT =~ /multi|spare/ ) {
        if ( $opt_commit =~ /commit|reject|0/ ) {
            if ( !$tmp_root_done ) {
                logDbg(
                    "Wafliron::wafliron_start - Restoring root volume, oc= $opt_commit and tmp_root_done is: $tmp_root_done"
                );

               # If this routine created the new root volume
               # then we need to return to using original root volume and aggr
               # before we return to the caller
               # Get console connection
                my ($conn_type, $conn);
                $conn_type = $fc->conntype();
                if ( $conn_type !~ m/console|pm_cons/ ) {
                    $conn = connect( $fc->hostp()->hostname() . "#console", -login );
                } else {
                    $conn = $fc;
                }
                transit_to( $conn, 'UP' );
                wafliron_pick_new_root( $conn, $orig_root_aggr,
                    $orig_root_volume, reboot => 1, setflags => $flags_hash );

                # Lets get rid of the previous root volume
                if ( $::IRON_ON_ROOT != /spare/ ) {
                    vol_purge( $conn, $orig_root_volume . "_temp" );
                }
                aggr_purge(
                    $conn,
                    $orig_root_aggr . "_temp",
                    destroy_corrupt_volumes => 1
                );
            }    # End of if (!$tmp_root_done)
        }    # End of $opt_commit =~ /commit|reject|0/
    }    # End of if ($::IRON_ON_ROOT =~ /multi|spare/)

    return ( $corruption_found, $return_message, \%cp_hash );
} ## end sub wafliron_start ($$;%)

## @name ::wafliron_wait($id)
##
## @summary
##     This function works with walfiron_start(). It accept the $id returned
##     from wafliron_start(). Look up the saved information include
##     $filer_connection and $container_name, wait for wafliron to complete
##     on the specified container.
##     return (corruption_found, corruption_messages);
##
## @description
##     check wafliron status on the specified aggregate or volume,
##     if still running, wait until complete.
##     return ($corruption_found, $corruption_messages);
##     $corruption_found == 1 means "found corruption"
##
##     $corruption_messages:
##     The wafliron corruption messages can have many different patterns,
##     ths function will not try to match any specific pattern. Instead
##     it will return any message saved from wafliron Starting to wafliron
##     completed on specified container. The following is an example of the
##     corruption_message:
##
##   Mar 29 12:22:12 GMT [wafl.iron.start:notice]: Starting wafliron on \
##   aggregate wafl_check_iron_aggr.
##   Tue Mar 29 12:22:12 GMT [wafl.check.info:error]: WAFLIRON, aggregate \
##   wafl_check_iron_aggr: Directory 103, block 0, unallocated entry 2 \
##   with nonzero fields. Clearing.
##   Tue Mar 29 12:22:12 GMT [wafl.vv.vol.config:notice]: Unable to configure \
##   volume in aggregate 'wafl_check_iron_aggr': missing container inode
##   Tue Mar 29 12:22:12 GMT [wafl.vv.vol.config:notice]: Unable to configure \
##   volume in aggregate 'wafl_check_iron_aggr': missing container inode
##   Tue Mar 29 12:22:13 GMT [wafl.delete.restart:info]: Debug-build-only \
##   message - restarting 0 zombies, aggr wafl_check_iron_aggr 0x95f33098
##   Tue Mar 29 12:22:53 GMT [wafl.iron.start:notice]: Starting wafliron on \
##   volume lost_volume1.
##   Tue Mar 29 12:22:54 GMT [wafl.delete.restart:info]: Debug-build-only \
##   message - restarting 0 zombies, vol lost_volume1 0x95f62098
##   Tue Mar 29 12:22:54 GMT [wafl.scan.start:info]: Starting container block \
##   reclamation on volume lost_volume1.
##   Tue Mar 29 12:29:04 GMT [wafl.scan.iron.done:info]: Volume lost_volume1, \
##   wafliron completed.
##   Tue Mar 29 12:33:48 GMT [wafl_filesystem_sca:info]: DBG: iron pvol \
##   wafl_check_iron_aggr - completing
##   Tue Mar 29 12:33:53 GMT [wafl.check.info:error]: WAFLIRON, aggregate \
##   wafl_check_iron_aggr: 1 lost files or directories relinked into lost+found
##   Tue Mar 29 12:33:53 GMT [wafl.scan.iron.done:info]: Aggregate \
##   wafl_check_iron_aggr, wafliron

##
## @arg $id Required: the id returned by wafliron_start(...wait=>0).
##
## @return value (corruption_found, corruption_messages)
##
## @example my ($corruption_found, $mesg) =
##                  wafliron_start($id);
##

sub wafliron_wait ($) {
    my ( $id ) = @_;
    my ( $found, $lines, $match, $before, $after, $prompt, $fh, $fconn, $mount_obj );

    if ( !($id) or !defined( $id_table{$id} ) ) {
        croak("The id passed to wafliron_wait is invalid");
    }

    my ( $fc, $container_name, $tail, $mount_root, $timeout,
         $type, $volume_order )
        = (
        $id_table{$id}[0], $id_table{$id}[1], $id_table{$id}[2],
        $id_table{$id}[3], $id_table{$id}[4], $id_table{$id}[5],
        $id_table{$id}[6]
        );

    my $endtime = time() + $timeout;
    my $container_status
        = ( $type == 'aggr' )
        ? aggr_status( $fc, $container_name )
        : vol_status( $fc, $container_name );

    # if iron_on_root look for wafliron completed messages from console
    if ( $::IRON_ON_ROOT =~ /core/ ) {

        # Get console connection
        my $conn_type = $fc->conntype();
        if ( $conn_type !~ m/console|pm_cons/ ) {
            $fconn = connect( $fc->hostp()->hostname() . "#console", -login );
        } else {
            $fconn = $fc;
        }

        # issue wafliron command and wait for it to complete
        ( $match, $before, $after, $prompt ) = say( $fconn,
            "priv set -q test;wafl scan speed 10000",
            '-wait',
            "(Aggregate|Volume).*$container_name.*wafliron completed|Wafliron did not find any corruption",
            '-noprompt',
            '-timeout',
            LOCAL_WAFLIRON_TIMEOUT
        );
        $tail .= $match . $before . $after . $prompt;

# wafliron is done and system rebooted so no need to keep hacking the console for messages
# hopefully lines contains everything we need already.
# matching the complete message for walfiron
        $found = 0;
        if ( $tail
            =~ s/(.*$container_name.*wafliron completed|Wafliron did not find any corruption).*/$1/s
            )
        {
            my @list = grep (/wafl.check.info:(error|ALERT)|wafl.iron.*:CRITICAL/, split(/\n/, $tail));
            logcomment ("Wafliron::wafliron_wait - Corruptions found :". Dumper \@list);
            foreach my $line ( @list )
            {
                $found = 1 unless $line =~ /wafl.iron.oc.deletedChangeLog:CRITICAL|wafl.iron.(oc|oc.root).noCorruption:CRITICAL/;
            }

            return ( $found, $tail );
        } ## end if ( $tail =~ ...
        logWarn("Did not find wafliron completed message. $tail");
        return ( $found, $tail );
    } else {
        # delete $id from $id_table
        delete( $id_table{$id} );

        my $lines = "";

        # wait until wafliron complete
        while ( time() < $endtime ) {
            if ( _wi_status( $fc, $type, $container_name ) == 1 ) {
                Tharn::sleep(45);
            } else {
                Tharn::sleep(45);
                $lines .= $tail->get_logs( skip_cached_logs => 1 );

                # Because of excessive syslog messages (ex. in case of summarymap
                # corruption) the following line matching could hang.
                # To avoid, we can split lines and search for start message.
                my @lines = split /\n/, $lines;
                my $length = scalar @lines;
                for ( my $i = 0; $i <= $length; $i += 50 ) {
                    my $lines_few = join( "\n", @lines[ $i .. $i + 50 ] );

                    # matching the starting message for walfiron
                    if ($lines_few !~ /.*\s*\w+\s+\w+\s+\d+\s+\d+:\d+:\d+
                                  \s+\w+\s+.*\[.*\]:\s+
                              Starting\s+wafliron\s+on\s+\w+\s+
                              $container_name.*/ixs)
                    {
                        logWarn(
                            "Did not find wafliron start message:\n$lines_few"
                        );
                    } else {
                        $lines_few =~ s/(.*\s*\w+\s+\w+\s+\d+\s+\d+:\d+:\d+)
                                  (\s+\w+\s+.*\[.*\]:\s+
                                  Starting\s+wafliron\s+on\s+\w+\s+
                                  $container_name.*)/$1$2/ixs;
                        logXXX("Found message - Starting wafliron on...");
                        last;
                    } ## end else [ if ( $lines_few !~ ...
                } ## end for ( my $i = 0; $i <= ...

                my $found = 0;

                # This is a fix to burt 388600 excessive syslog messages makes
                # wafliron completed message to be dropped as specified by the
                # kern_syslog_drop. At this time, logwarn only is not enough.
                # wafliron_wait should be able to tell whether there is any
                # corruption, with the available syslogs and console outputs.
                if ( $lines
                    !~ s/(.*$container_name, wafliron completed).*/$1/s )
                {
                    logWarn(
                        "Did not find wafliron completed message. This is expected in rare cases involving a lot of wafliron output, e.g. massive corruptions. For those cases, this condition can be ignored within the test script code. Otherwise, it should be considered a product bug"
                    );
                } ## end if ( $lines !~ ...
                my @list = grep (/wafl.check.info:(error|ALERT)|wafl.iron.*:CRITICAL/, split(/\n/, $lines));
                logcomment ("Wafliron::wafliron_wait - Corruptions found :". Dumper \@list);
                foreach my $line ( @list )
                {
                   $found = 1 unless $line =~ /wafl.iron.oc.deletedChangeLog:CRITICAL|wafl.iron.(oc|oc.root).noCorruption:CRITICAL/;
                }
                return ( $found, $lines );
            } ## end else [ if ( _wi_status( $fc, ...
        } ## end while ( time() < $endtime)

        $tail = undef;
        croak("wafliron timeout.");
    } ## end else [ if ( $::IRON_ON_ROOT =~...

} ## end sub wafliron_wait ($)

## @name ::_wi_status($fc,$type $container_name)
##
## @summary
##     Internal subroutine which is actually optimized wafliron_status
##	Is wafliron running on specified aggr/vol or not?
##
## @description
##     When evaluated in scalar context, this function returns a number indicating
##     the wafliron status:
##         0 - no wafliron in progress.
##         1 - wafliron is running.
##         2 - wafliron is idle (optional commit, waiting for commit/reject).
##
##     In array context, this function returns a list containing the status value
##     described above and any console output captured during operations.
##
## @arg $fc Required: filer connection object.
##
## @arg $container_name Required: aggregate or traditional volume name.
##
## @example $status=wafiron_status($fc, $aggr);
##

sub _wi_status ($$$) {
    my ( $fc, $type, $name ) = @_;
    my ( $ver, $boot_state );

    my $fh = $fc->hostp();
    $boot_state = boot_state( $fc, timeout => 300 );
    if ( $boot_state eq 'NORC' ) {
        $ver = WV_BOILERMAKER_MIN;
    } else {
        $ver = eval { get_wafl_version( $fc->hostp() ) } || "";
    }

    my $retval;
    if ( $ver >= WV_BOILERMAKER_MIN ) {

        # use ZAPI

        # use the user name and password from the given connection
        my $username = $fc->username();

        # fall back to root
        $username = 'root' if ( !$username );

        # ati_ontapi_create_context will die if passwd is undef
        my $passwd = $fh->default_password();
        $passwd = ATI_ONTAPI_PASSWORD_DEFAULT if ( !$passwd );

        # get context
        my $context = ati_ontapi_create_context(
            filer     => $fh->default_ip(),
            user      => $username,
            password  => $passwd,
            transport => ATI_ONTAPI_TRANSPORT_HTTP,
            version   => ATI_ONTAPI_VERSION_DEFAULT,
        );

        my @out = ati_ontapi_call_api(
            context => $context,
            api     => 'aggr-list-info',
            args    => [ 'aggregate' => $name ],
        );
        ( $out[0] == ATI_SUCCESS )
            || croak __PACKAGE__ . "::_wi_status: Error $out[2]: $out[3]\n";

        # we only asked about a single aggregate, so there had better be
        # only one entry in the returned list of responses
        croak "too many responses to aggr-list-info ZAPI"
            if ( scalar( @{ $out[1]->{aggregates} } ) > 1 );

        # extract the single entry
        my $aggregate_data = $out[1]->{aggregates}->[0];

        # the wafliron entry is optional in the aggr-info data structure,
        # handle this special-case here...
        if ( !exists $aggregate_data->{wafliron} ) {

           # wafliron is not running (and has never been run during this boot)
            $Log->trace(
                "_wi_status: wafliron is NOT running (and has never been run), returning 0"
            );
            return ( wantarray ? ( 0, '' ) : 0 );
        } else {

        # will we ever see more than one aggr-wafliron-info entry in the list?
        my $count = scalar( @{ $aggregate_data->{wafliron} } );
        croak "too many aggr-list-info entries for aggr '$name': $count"
                if ( $count > 1 );

            # check the wafliron state
            my $state = $aggregate_data->{wafliron}->[0]->{state};
            if (   ( $state eq "starting" )
                || ( $state eq "scanning" )
                || ( $state eq "checking_lost_blocks" )
                || ( $state eq "checking_lost_inodes" )
                || ( $state eq "finishing" )
                || ( $state eq "aborting" ) )
            {

                # wafliron is running
                $Log->trace("_wi_status: wafliron is running, returning 1");
                return ( wantarray ? ( 1, '' ) : 1 );
            } elsif ( ( $state eq "not_running" )
                || ( $state eq "waiting_for_commit" ) ) {
                my $mount_state = $aggregate_data->{'mount-state'};
                if ( $mount_state eq 'iron_restricted' ) {

                    # looks like IOC is waiting for commit/reject
                    $Log->trace(
                        "_wi_status: IOC is waiting for commit/reject, returning 2 (1)"
                    );
                    return ( wantarray ? ( 2, '' ) : 2 );
                } else {

                    # wafliron is not running (but has been run sometime in the past)
                    $Log->trace(
                        "_wi_status: wafliron is NOT running (but did run sometime earlier), returning 0"
                    );
                    return ( wantarray ? ( 0, '' ) : 0 );
                } ## end else [ if ( $mount_state eq 'iron_restricted')
            } else {
                croak("unexpected zapi wafliron state result: '$state'");
            }
        } ## end else [ if ( !exists $aggregate_data...
    } else {

        # use CLI
        my $cmd       = $type . " wafliron status " . $name;
        my $try_again = 1;
    again:

        # this code sometimes fails if -swallow is not used
        my ( $match, $before, $after, $prompt )
            = say( $fc, "priv set -q advanced; $cmd", -swallow, -timeout,
            300 );
        my $output = $before . $match . $after . $prompt;

        if (   $after =~ /is\s+active/
            || $after =~ /is\s+in\s+the\s+process\s+of\s+starting/ )
        {
            $Log->trace("_wi_status: wafliron is running, returning 1");
            return ( wantarray ? ( 1, $output ) : 1 );
        } elsif ( $after =~ /is\s+not\s+currently\s+active/
            || $after =~ /no\s+waflirons\s+in\s+progress/ )
        {
            $Log->trace(
                "_wi_status: wafliron is NOT running now, returning 0");
            return ( wantarray ? ( 0, $output ) : 0 );
        } elsif ( $after =~ /either\s+commit\s+or\s+reject/ ) {
            $Log->trace(
                "_wi_status: IOC is waiting for commit/reject, returning 2 (2)"
            );
            return ( wantarray ? ( 2, $output ) : 2 );
        } else {

      # after some operations, e.g. wafliron commit, it can take a bit for the
      # new status to be reflected in the wafliron status output. so, if we
      # end up here, we try just one more time before failing. see burt338391.
            if ($try_again) {
                $try_again = 0;
                goto again;
            } else {
                croak("unexpected result: '$after'");
            }
        } ## end else [ if ( $after =~ /is\s+active/...
    } ## end else [ if ( $ver >= WV_BOILERMAKER_MIN)
} ## end sub _wi_status ($$$)

## @name ::wafliron_stop($fc, $container_name[, %opts])
##
## @summary
##     stop wafliron running on specified aggr/vol
##
## @description
##     if wafliron successfully stopped, return 1;
##     if wafliron was not running before the call, return 0;
##     croak if other error
##
## @arg $fc Required: filer connection object.
##
## @arg $container_name  Required: aggregate or traditional volume name.
##
## @arg %opts Optional: options can be set:
##
##      timeout timeout value to stop wafliron
##
##      id      the id returned by wafliron_start(...wait=>0),
##
## @example my $stopped=wafliron_stop($fc, $aggr);
##          wafliron_stop($fc, $aggr, timeout=>120, id=>$id);
##

sub wafliron_stop ($$;%) {
    my ( $fc, $name, %opts ) = @_;

    my $timeout
        = defined( $opts{timeout} ) ? $opts{timeout} : LOCAL_WAFLIRON_TIMEOUT;
    my $container_name = get_container( $fc, $name );
    my $type = wafliron_type( $fc, $name );
    my $cmd = $type . " wafliron stop " . $container_name;

    say( $fc, "priv set -q advanced; $cmd", -timeout, $timeout );

    my $endtime = time() + $timeout;
    while ( _wi_status( $fc, $type, $container_name ) ) {
        if ( time() < $endtime ) {
            Tharn::sleep(3);
        } else {
            croak("wafliron stop timeout");
        }
    }

    # delete $id from $id_table
    if ( defined( $opts{id} ) ) {
        my $id   = $opts{id};
        my $tail = $id_table{$id}[2];
        delete( $id_table{$id} );
        $tail = undef;
    }

    return 1;
} ## end sub wafliron_stop ($$;%)

## @name ::wafliron_status($fc, $container_name)
##
## @summary
##     Is wafliron running on specified aggr/vol or not?
##
## @description
##     If wafliron is running on aggr/vol, returns 1;  if wafliron is idle returns 2;
##     if no wafliron in progress returns 0.
##
## @arg $fc Required: filer connection object.
##
## @arg $container_name Required: aggregate or traditional volume name.
##
## @example $status=wafiron_status($fc, $aggr);
##
sub wafliron_status ($$) {
    my ( $fc, $name ) = @_;
    my $type = wafliron_type( $fc, $name );
    my $container = get_container( $fc, $name );

    # capture result in a var to force scalar context - see _wi_status()
    my $status = _wi_status( $fc, $type, $container );

    return $status;
} ## end sub wafliron_status ($$)

## @name ::wafliron_volume_order($fc, $output, @order)
##
## @summary
##     verify that volumes have been iron-mounted in the given order
##
## @description
##
##     On Ironcity or later releases, volume ordering is available.
##	   This function accepts the filer connection, complete wafliron
##	   output, and volume list. It checks volume mounting messages against
##	   expected volume order. NOTE: when sepcifying volume order don't
##	   forget that sparse volume mounted first even it is not in the list
##	   and parent volume mounted before its clone.
##     return (true|false);
##
## @arg $fc Required: filer connection object to specify volume
##	   existance.
##
## @arg $output Required: all wafliron messages. To get these messages run wafliron_start()
##	   with options wait=>1.
##
## @arg @order Required: expected volume order, which is the array of volume names.
##
## @return value
##      true  - if actual volume order match to the expected one.
##	    false - otherwise
##		die if mounting messages are missing
##
## @example my (undef, $output) = wafliron_start($fc,$aggr,volume_order=>$list);
##          my $result = check_order($output,@vols);
sub wafliron_volume_order($$@) {
    my ( $fc, $output, @order ) = @_;
    my @lines = split /\n/, $output;
    my ( $line, $volume );
    while ( @order && @lines ) {
        unless ( vol_exists( $fc, $order[0] ) ) {
            logWarn("Volume $order[0] in the order list doesn't exist");
            shift @order;
            next;
        }
        $line = shift @lines;
        if ( $line =~ s/(.*?)(Wafliron\s+mounted\s+volume\s+)(\w+)(\..*)/$3/ )
        {
            $volume = $line;
            if ( $volume ne $order[0] ) {
                logcomment(
                    "Wafliron::wafliron_volume_order - Volume $volume doesn't match expected volume $order[0]"
                );
                return 0;
            } else {
                shift @order;
            }
        } ## end if ( $line =~ ...
    } ## end while ( @order && @lines )
    if (@order) {
        croak("Did not find all mounting messages:\n @order\n$output");
    } else {
        return 1;
    }
} ## end sub wafliron_volume_order($$@)

# decide to use "vol wafliron" or "aggr wafliron" based on wafl version
sub wafliron_type {

    # Always aggr in newer versions
    return ("aggr");

} ## end sub wafliron_type

## return the containing aggregate or traditional volume name
sub get_container ($$) {
    my ( $fc, $name ) = @_;

    if ( aggr_exists( $fc, $name ) ) {
        logDbg("$name is aggr");
        return $name;
    } else {
        logDbg("$name is not aggr, self container");
    }

    if ( vol_exists( $fc, $name ) ) {
        my $container_name = vol_container( $fc, $name );
        if ($@) {

            # it is a traditional volume
            logDbg("$name is a traditional volume");
            $container_name = $name;
        } else {
            logDbg(
                "$name is a flex volume, container is " . $container_name );
        }
        return $container_name;
    } else {
        croak( "$name does not exist (neither volume nor aggr): " . $@ );
    }
} ## end sub get_container ($$)

## @name ::wafliron_oc_supported($fc)
##
## @summary
##     query to find out whether optional commit is supported
##
## @description
##     This function returns 1 if wafliron optional commit is supported, and
##     0 if not.
##
## @arg $fc Required: filer connection object.
##
sub wafliron_oc_supported($) {
    my $fc = shift;

    # In case if any one using, keeping this function around
    # OC is supported in all new releases
    return 1;

} ## end sub wafliron_oc_supported($)

## @name ::check_oc_status($id)
##
## @summary
##     This function works with walfiron_start(). It accept the $id returned
##     from wafliron_start(). Look up the saved information include
##     $filer_connection and $container_name, wait for wafliron to commit or reject changes
##     on the specified container.
##     return (commit_reject_messages);
##
## @description
##     check  aggr  and wafliron status before commit/reject
##	commit or reject changes
##    check  aggr  and wafliron status after commit/reject
##     return (commit_reject_messages);

sub check_oc_status ($) {
    my ($id) = @_;
    if ( !($id) or !defined( $id_table{$id} ) ) {
        die("The id passed to optional_commit is invalid");
    }
    my $oc_message;
    my ( $fc, $container_name, $type, $oc_commit, $found, ) = (
        $id_table{$id}[0], $id_table{$id}[1], $id_table{$id}[5],
        $id_table{$id}[7], $id_table{$id}[8]
    );
    my ( $state, $status ) = aggr_status( $fc, $container_name );
    if ($found) {

        #check  aggr  and wafliron status before commit/reject
        unless ( _wi_status( $fc, $type, $container_name ) == 2
            && $state =~ /iron_restricted/ )
        {
            die("Unexpected status when inconsistency found, and before commit/reject"
            );
        } ## end unless ( _wi_status( $fc, ...

        # commit or reject changes
        $oc_message = commit_reject_wait($id);
        ( $state, $status ) = aggr_status( $fc, $container_name );

        #check  aggr  and wafliron status after commit/reject
        unless ( _wi_status( $fc, $type, $container_name ) == 0
            && $state =~ /offline/ )
        {
            die("Unexpected status for wafliron or $type $container_name after commit/reject changes"
            );
        } ## end unless ( _wi_status( $fc, ...
    } else {
        unless ( _wi_status( $fc, $type, $container_name ) == 0
            && $state =~ /offline/ )
        {
            croak(
                "Unexpected status for wafliron or $type $container_name,if no inconsistency found"
            );
        } ## end unless ( _wi_status( $fc, ...
    } ## end else [ if ($found)
    return $oc_message;
} ## end sub check_oc_status ($)

## @name ::commit_reject_wait($id)
##
## @summary
##     This function works with walfiron_start(). It accept the $id returned
##     from wafliron_start(). Look up the saved information include
##     $filer_connection and $container_name, wait for wafliron to commit or reject changes
##     on the specified container.
##     return (commit_reject_messages);
##
## @description
##     wait until wafliron  complete commit or reject changes
##    matching commit or reject  changes
##     return (commit_reject_messages);

sub commit_reject_wait ($) {
    my ($id) = @_;
    my ( $fc, $container_name, $tail, $mount_root,
                     $timeout, $type, $opt_commit ) = (
        $id_table{$id}[0], $id_table{$id}[1], $id_table{$id}[2],
        $id_table{$id}[3], $id_table{$id}[4], $id_table{$id}[5],
        $id_table{$id}[7]
    );

    logcomment("Wafliron::commit_reject_wait: trying $opt_commit...");

    delete( $id_table{$id} );
    my $endtime  = time() + $timeout;

    my $lines     = "";
    my $start_cmd = $type . " wafliron " . $opt_commit . " " . $container_name;
    my ( undef, undef, $after )
        = say( $fc, "priv set -q advanced; $start_cmd", -timeout, $timeout );

    # wait until wafliron complete commit or reject changes
    while ( time() < $endtime ) {
        if ( _wi_status( $fc, $type, $container_name ) == 1 ) {
            Tharn::sleep(30);
        } else {
            Tharn::sleep(30);
            $lines = $tail->get_logs( skip_cached_logs => 1 );

            # matching the delete log file message
            if ($lines !~ /.*\s*\w+\s+\w+\s+\d+\s+\d+:\d+:\d+
	                           \s+\w+\s+\[.*\]:\s+
							   Deleted\s+change\s+log\s+files\s+for\s+\w+\s+
	                            $container_name.*/ixs
                )
            {
                logWarn("Did not find delete log files message:$lines");
            } ## end if ( $lines !~ /.*\s*\w+\s+\w+\s+\d+\s+\d+:\d+:\d+ )

            # matching commit or reject  changes
            if ( $opt_commit =~ /commit/i ) {
                if ($lines =~ /Committed\s+change\s+log\s+files\s+for\s+\w+\s+
	                            $container_name.*/ixs)
                {
                    logcomment("Wafliron committed changes");
                } else {
                    die("Wafliron didn't commit changes:\n$lines");
                }
            } elsif ( $opt_commit =~ /reject/i ) {
                if ($lines =~ /Rejecting\s+change\s+log\s+files\s+for\s+\w+\s+
	                            $container_name.*/ixs)
                {
                    logcomment("Wafliron rejected changes");
                } else {
                    die("Wafliron didn't reject changes:\n$lines");
                }
            } else {
                die("Invalid subcommand $opt_commit");
            }
        } ## end else [ if ( _wi_status( $fc, ...
        return $lines;
    } ## end while ( time() < $endtime)

    $tail = undef;
    croak("wafliron commit/reject timeout.");

} ## end sub commit_reject_wait ($)

use constant RE_COMMITTED_FINISH =>
    "Committed change log files for (aggregate|volume)";
use constant RE_COMMIT_CHANGES =>
    "Commit changes for (aggregate|volume) .* to disk\\?";

## @name ::wafliron_commit($fc, $container_name[, %opts])
##
## @summary
##     commit the optional-commit status files to the aggregate/tradvol.
##
## @description
##     This function mimics the 'aggr wafliron commit <aggr/trad>' CLI
##     command. It's the responsibility of the caller to make sure
##     wafliron (scanners) is done on the aggregate.
##
## @arg $fc Required: filer connection object.
##
## @arg $container_name Required: name of aggregate or traditional
##                                volume
##
## @arg %opts Optional: options can be set:
##
##      wait=>1 or 0, default is 1 (NOT Available)
##         if wait is set to 1, the wafliron_commit() will wait for commit
##         to be completed; otherwise, return immediately after commit.
##      force=>1 or 0 default is 1
##         if force is 1, waits for commit message and answers y.
##         if force is 0 and we receive a prompt to commit, it answers n.
## @return
##      return (oc_msg, before_match)
##              or
##      return (oc_msg)
##

sub wafliron_commit($$;%) {
    my ( $fc, $name, %opts ) = @_;

    my ( $oc_message, $force, $tail, $messages_file,
         $mount_root, $already_halted, $mount_obj );
    my ( $console, $match, $before, $after );
    my $node = NACL::STask::Node->new(hostrec => $fc->hostp());

    $force = defined( $opts{force} ) ? defined( $opts{force} ) : 1;

    # New parameters for IRON_ON_ROOT
    # root is the name of the volume that will become the root volume
    # when IRON_ON_ROOT=core

    my $root
        = defined( $opts{root_vol} ) ? "$opts{root_vol}" : "$::FILER_VOL";

    # tmp_root indicates a new root volume has already been created and it
    #  should not be done in this routine
    my $tmp_root_done = defined( $opts{tmp_root} ) ? "$opts{tmp_root}" : 0;

    # If the delete_all_snaps option is set then we must get into boot menu
    #  and delete all snaps before the system will come up
    my $delete_all_snaps
        = defined( $opts{delete_all_snaps} ) ? "$opts{delete_all_snaps}" : 0;

    my $container_name = get_container( $fc, $name );
    my $end_time       = time() + LOCAL_WAFLIRON_TIMEOUT;
    my $type           = wafliron_type( $fc, $container_name );
    my $container_status
        = ( $type == 'aggr' )
        ? aggr_status( $fc, $container_name )
        : vol_status( $fc, $container_name );

# If IRON_ON_ROOT with delete_all_snaps then we need to halt the filer
# and delete the existing snaps before we can allow the filer to come all the way up
    $already_halted = 0;

    ## Assume that wafliron scanners are done. It's the caller's
    ## responsibility to make sure of that.
    my $start_cmd = "aggr wafliron commit " . $container_name;

    #
    # We need to make a console connect
    #
    my $conn_type = $fc->conntype();
    if ( $conn_type !~ m/console|pm_cons/ ) {
        $console = connect( $fc->hostp()->hostname() . "#console", -login );
    } else {
        $console = $fc;
    }

    # See if we need to enter the password
    # Handle any password prompt
    my $text = "";
    _wafliron_login( $fc, $text );

 # If we are running IRON_ON_ROOT=core, then we are in the boot menu and need
 # to send the command to the console and capture all the output in oc_message
    if ( $::IRON_ON_ROOT =~ /core/ ) {
        if ($delete_all_snaps) {

# issue wafliron command to commit or reject and stop the filer before it can come up
# so we can delete all snaps
            ( $match, $before, $after, undef )
                = say( $console, "priv set -q advanced; $start_cmd",
                '-wait', ["The system will reboot now"],
                '-noprompt', '-timeout',
                LOCAL_WAFLIRON_TIMEOUT
                );

            # Stop the automatic booting
            transit_to( $console, "FIRMWARE" );
            $already_halted = 1;

        } else {

            # issue wafliron command to commit or reject and wait for it to complete
            ( $match, $before, $after, undef ) = say( $console,
                "priv set -q advanced; $start_cmd",
                '-wait',
                [   "System rebooted after an internal reboot",
                    "not in the expected state",
                    RE_COMMIT_CHANGES
                ],
                '-noprompt',
                '-timeout',
                LOCAL_WAFLIRON_TIMEOUT
            );
            ## @step Go to bootmenu to prevent filer from coming up with current root vol
            logDbg(
                'Wafliron::wafliron_commit -  Go to bootmenu after wafliron commit to stop coming up fully'
            );
            transit_to( $console, 'BOOT_MENU' );
        } ## end else [ if ($delete_all_snaps)

        $oc_message = $before . $after;

        #
        # If we need to confirm commit and force is set confirm it else
        # say no.

        if ( $match =~ RE_COMMIT_CHANGES ) {
            if ( $force == 1 ) {
                say( $console,
                    "Y",
                    '-wait',
                    "System rebooted after an internal reboot|not in the expected state",
                    '-noprompt',
                    '-timeout',
                    LOCAL_WAFLIRON_TIMEOUT
                );
            } else {
                say( $console,
                    "N",
                    '-wait',
                    "System rebooted after an internal reboot|not in the expected state",
                    '-noprompt',
                    '-timeout',
                    LOCAL_WAFLIRON_TIMEOUT
                );

            } ## end else [ if ( $force == 1 )
        } ## end if ( $match =~ RE_COMMIT_CHANGES)

    } else {

        # get a handle on the /etc/messages file
        $tail = NACL::MTask::SystemLogDetector->new(
                    command_interface => $node,
                    set_log_file      => ['messages'] );
        $tail->start();

        ( $match, $before, $after ) = say( $console,
            "priv set -q advanced; $start_cmd",
            -timeout,
            LOCAL_WAFLIRON_TIMEOUT,
            '-noprompt',
            '-wait',
            [ RE_COMMITTED_FINISH, RE_COMMIT_CHANGES ]
        );

        # If we need to confirm commit and force is set confirm it else
        # say no.

        if ( $match =~ RE_COMMIT_CHANGES ) {
            if ( $force == 1 ) {
                say( $console, "Y", '-blind' );
            } else {
                say( $console, "N", '-blind' );
            }
        } ## end if ( $match =~ RE_COMMIT_CHANGES)

        # wait until commit is done or timeout
        while (( time() < $end_time )
            && ( _wi_status( $fc, $type, $container_name ) != 0 ) )
        {
            Tharn::sleep(15);
        }

        $oc_message = $tail->get_logs( skip_cached_logs => 1 );

        ## BURT 358901 fix to wait for $req_end_msg get append with $oc_message
        my $req_end_msg
            = "Deleted change log files for aggregate $container_name";
        my $log_fetch_timer = time() + 30;

        # wait till you get the message or 30 seconds whichever is earlier
        until (( $oc_message =~ m/$req_end_msg/g )
                || ( time() > $log_fetch_timer ))
        {

            # append the messages to $oc_message
            $oc_message .= $tail->get_logs( skip_cached_logs => 1 );

            # sleep a bit to avoid frequent check
            Tharn::sleep(1);
        } ## end until ( ( $oc_message =~ ...

        # $oc_message to be NATE logged, refer to enhancement burt 336996
        logcomment(
            "Wafliron::wafliron_commit - wafliron commit messages : $oc_message"
        );

        # Verify if wafliron committed the changes.
        if ($oc_message =~ /Committed\s+change\s+log\s+files\s+for\s+\w+\s+
                     	 .*$container_name.*/ixs)
        {
            logcomment("Wafliron committed changes");
        } elsif ( ( $match =~ RE_COMMIT_CHANGES ) && ( $force == 0 ) ) {
            logcomment( "Wafliron did not commit changes "
                    . "when prompted: force was $force" );
        } else {
            die("Wafliron failed to commit changes");
        }

    } ## end else [ if ( $::IRON_ON_ROOT =~...

    # If any kind of IRON_ON_ROOT we need to restore the original root volume
    # after committing
    if ( $::IRON_ON_ROOT && !$tmp_root_done ) {

        # We have to get back to the original root aggr and volume
        # Do not reboot since we may have to delete all snaps before
        # the filer will come up clean
        wafliron_pick_new_root(
            $console, $orig_root_aggr, $orig_root_volume,
            delete_all_snaps => $delete_all_snaps,
            already_halted   => $already_halted,
            reboot           => 1,
            setflags         => $flags_hash
        );

        # Make sure the filer is 'UP'
        transit_to( $console, 'UP' );

        if ( $::IRON_ON_ROOT =~ /spare|multi/ ) {

            # Lets get rid of the previous root volume
            vol_purge( $fc, $orig_root_volume . "_temp" );
        }

        # Lets get rid of the previous root aggr
        aggr_purge(
            $fc,
            $orig_root_aggr . "_temp",
            destroy_corrupt_volumes => 1
        );
    } ## end if ( $::IRON_ON_ROOT &&...

    # We do not disconnect here on the console because if ntest is
    # called with -consolelog option and we disconnect
    # other connects to the console within the test fail.

    if (wantarray) {
        return ( $oc_message, $before );
    } else {
        return ($oc_message);
    }
} ## end sub wafliron_commit($$;%)

## @name ::wafliron_reject($fc, $container_name[, %opts])
##
## @summary
##     reject the optional-commit status files for the aggregate/tradvol.
##
## @description
##     This function mimics the 'aggr wafliron reject <aggr/trad>' CLI
##     command. It's the responsibility of the caller to make sure
##     wafliron (scanners) is done on the aggregate.
##
## @arg $fc Required: filer connection object.
##
## @arg $container_name Required: name of aggregate or traditional
##                                volume
##
## @arg %opts Optional: options can be set:
##
##      (Not avail) wait=>1 or 0, default is 1
##         if wait is set to 1, the wafliron_reject() will wait for commit
##         to be completed; otherwise, return immediately after commit.
##

sub wafliron_reject($$;%) {
    my ( $fc, $name, %opts ) = @_;

    my ( $oc_message, $tail, $messages_file, $mount_root, $mount_obj );
    my $node = NACL::STask::Node->new(hostrec => $fc->hostp());

    # tmp_root indicates a new root volume has already been created and it
    #  should not be done in this routine
    my $tmp_root_done = defined( $opts{tmp_root} ) ? "$opts{tmp_root}" : 0;

    my ( $match, $before, $after );
    my $container_name = get_container( $fc, $name );
    my $end_time       = time() + LOCAL_WAFLIRON_TIMEOUT;
    my $type           = wafliron_type( $fc, $container_name );
    my $container_status
        = ( $type == 'aggr' )
        ? aggr_status( $fc, $container_name )
        : vol_status( $fc, $container_name );

    ## Assume that wafliron scanners are done. It's the caller's
    ## responsibility to make sure of that.
    my $start_cmd = "aggr wafliron reject " . $container_name;

    # We need to make a console connect
    my ($conn_type, $console);
    $conn_type = $fc->conntype();
    if ( $conn_type !~ m/console|pm_cons/ ) {
        $console = connect( $fc->hostp()->hostname() . "#console", -login );
    } else {
        $console = $fc;
    }

    # See if we need to enter the password
    # Handle any password prompt
    my $text = "";
    _wafliron_login( $fc, $text );

    # If we are running IRON_ON_ROOT=core, then we are in the boot menu and need
    # to send the command to the console and capture all the output in oc_message
    if ( $::IRON_ON_ROOT =~ /core/ ) {

        # issue wafliron command to commit or reject and wait for it to complete
        ( $match, $before, $after, undef )
            = say( $console, "priv set -q advanced; $start_cmd",
            '-wait',     "Rejected changes for aggregate",
            '-noprompt', '-timeout',
            LOCAL_WAFLIRON_TIMEOUT
            );
        $oc_message = $match . $before . $after;

        ## @step Go to bootmenu to switch to original good root vol
        logDbg(
            'Wafliron::wafliron_reject -  Go to bootmenu after wafliron reject to stop coming up fully'
        );
        transit_to( $console, 'BOOT_MENU' );

        ## Mark tmp_root_done to be not done to reset to orig_root_vol
        $tmp_root_done = 0;

    } else {

        # get a handle on the /etc/messages file
        $tail = NACL::MTask::SystemLogDetector->new(
                    command_interface => $node,
                    set_log_file      => ['messages'] );
        $tail->start();

        ( $match, $before, $after )
            = say( $console, "priv set -q advanced; $start_cmd",
            -timeout, LOCAL_WAFLIRON_TIMEOUT
            );

        # wait until reject is done or timeout
        while (( time() < $end_time )
            && ( _wi_status( $fc, $type, $container_name ) != 0 ) )
        {
            Tharn::sleep(15);
        }

        # append the messages to $oc_message
        $oc_message .= $tail->get_logs( skip_cached_logs => 1 );

        # BURT 359407 fix: Deletion of the change log files can be confirmed
        # through $chglog_del_msg. So, we need to wait for $chglog_del_msg and
        # also it should get append with $oc_message
        my $chglog_del_msg
            = "Deleted change log files for aggregate $container_name";
        my $log_fetch_timer = time() + 30;

        # wait till you get the message or 30 seconds whichever is earlier
        until (( $oc_message =~ m/$chglog_del_msg/g )
                || ( time() > $log_fetch_timer ))
        {

            # append the messages to $oc_message
            $oc_message .= $tail->get_logs( skip_cached_logs => 1 );

            # sleep a bit to avoid frequent check
            Tharn::sleep(1);
        } ## end until ( ( $oc_message =~ ...

        # $oc_message to be NATE logged, refer to enhancement burt 336996
        logcomment(
            "Wafliron::wafliron_reject - wafliron reject messages : $oc_message"
        );

        # Verify if wafliron rejected the changes.
        if ($oc_message =~ /Rejecting\s+change\s+log\s+files\s+for\s+\w+\s+
                     	 $container_name.*|Rejected\schanges\sfor\saggregate/ixs)
        {
            logcomment("Wafliron rejected changes");
        } else {
            $mount_obj->purge();
            die("Wafliron failed to reject changes");
        }

        # Verify the deletion of change log files.
        if ($oc_message
            =~ /Deleted\s+change\s+log\s+files\s+for\s+aggregate\s+
                          $container_name.*/ixs)
        {
            logcomment("Change logs files have been deleted");
        } else {
            die("Change logs files have not been deleted");
        }
    } ## end else [ if ( $::IRON_ON_ROOT =~...

    # If any kind of IRON_ON_ROOT we need to restore the original root volume
    # after rejecting
    if ( $::IRON_ON_ROOT && !$tmp_root_done ) {

        # Set the root volume back to what it was
        wafliron_pick_new_root(
            $console, $orig_root_aggr, $orig_root_volume,
            powercycle => 1,
            reboot     => 1,
            setflags   => $flags_hash
        );

        # Lets get rid of the previous root volume
        if ( $::IRON_ON_ROOT != /spare/ ) {
            vol_purge( $fc, $orig_root_volume . "_temp" );
        }
        aggr_purge(
            $fc,
            $orig_root_aggr . "_temp",
            destroy_corrupt_volumes => 1
        );
    } ## end if ( $::IRON_ON_ROOT &&...

    return $oc_message;
} ## end sub wafliron_reject($$;%)

## @name ::wafliron_runcommit($fc, $container_name[, %opts])
##
## @summary
##     start wafliron in optional commit mode and commit the status files
##     for the aggregate/tradvol.
##
## @description
##     This function mimics the 'aggr wafliron start <aggr/trad>' CLI
##     command, following by 'aggr wafliron commit <aggr/trad>'.
##
## @arg $fc Required: filer connection object.
##
## @arg $container_name Required: name of aggregate or traditional
##                                volume
##
sub wafliron_runcommit($$;%) {
    my ( $fc, $name, %opts ) = @_;
    my $timeout
        = defined( $opts{timeout} ) ? $opts{timeout} : LOCAL_WAFLIRON_TIMEOUT;
    my $node = NACL::STask::Node->new(hostrec => $fc->hostp());

    # New parameters for IRON_ON_ROOT
    # root is the name of the volume that will become the root volume
    #  when IRON_ON_ROOT=core
    my $root
        = defined( $opts{root_vol} ) ? "$opts{root_vol}" : $::FILER_VOL_FLEX;

    # tmp_root indicates a new root volume has already been created and it
    #  should not be reset in this routine
    my $tmp_root_done = defined( $opts{tmp_root} ) ? "$opts{tmp_root}" : 0;
    my ( $found, $output, $oc_message );

    my $container_name = get_container( $fc, $name );
    my $type = wafliron_type( $fc, $container_name );
    my $container_status
        = ( $type == 'aggr' )
        ? aggr_status( $fc, $container_name )
        : vol_status( $fc, $container_name );

    # get a handle on the /etc/messages file
    logcomment ("Trying to re-mount root volume");
    my $mount_obj = _mount_root_vol($fc);

    # Get the mount point
    my $mount_root = $mount_obj->mount_point();
    logcomment("mount_point for root vol - $mount_root");

    my $tail = NACL::MTask::SystemLogDetector->new(
                    command_interface => $node,
                    set_log_file      => ['messages'] );
    $tail->start();

    # ensure that caller didn't specify wait or oc option
    if ( defined( $opts{wait} ) || defined( $opts{oc} ) ) {
        croak("wait or oc option passed to wafliron_runcommit");
    }

    # now force wait and oc options to be on
    $opts{wait} = 1;
    $opts{oc}   = 1;

    ( $found, $oc_message ) = wafliron_start( $fc, $container_name, %opts );
    if ( _wi_status( $fc, $type, $container_name ) != 2 ) {
        $mount_obj->purge();
        die("Wafliron Optional Commit did not complete on volume $container_name"
        );
    }

    $oc_message .= wafliron_commit( $fc, $container_name,
        tmp_root => $tmp_root_done );

# If running IRON_ON_ROOT of any flavor, the tail created at the begining of this routine
# does not reflect any wafliron activity since the root volume may have changed during the
# wafliron start and has changed back after wafliron commit.
    if ( !$::IRON_ON_ROOT ) {
        $oc_message = $tail->get_logs( skip_cached_logs => 1 );
    }

    $mount_obj->purge();
    return ( $found, $oc_message );
} ## end sub wafliron_runcommit($$;%)

## @name ::wafliron_runreject($fc, $container_name[, %opts])
##
## @summary
##     start wafliron in optional commit mode and reject the status files
##     for the aggregate/tradvol.
##
## @description
##     This function mimics the 'aggr wafliron start <aggr/trad>' CLI
##     command, followed by 'aggr wafliron reject <aggr/trad>'.
##
## @arg $fc Required: filer connection object.
##
## @arg $container_name Required: name of aggregate or traditional
##                                volume
##
sub wafliron_runreject($$;%) {
    my ( $fc, $name, %opts ) = @_;
    my ( $found, $output, $oc_message );

    my $container_name = get_container( $fc, $name );
    my $type = wafliron_type( $fc, $container_name );
    my $node = NACL::STask::Node->new(hostrec => $fc->hostp());

    # get a handle on the /etc/messages file
    logcomment ("Trying to re-mount root volume");
    my $mount_obj = _mount_root_vol($fc);

    # Get the mount point
    my $mount_root = $mount_obj->mount_point();
    logcomment("mount_point for root vol - $mount_root");

    my $tail = NACL::MTask::SystemLogDetector->new(
                    command_interface => $node,
                    set_log_file      => ['messages'] );
    $tail->start();

    # ensure that caller didn't specify wait or oc option
    if ( defined( $opts{wait} ) || defined( $opts{oc} ) ) {
        croak("wait or oc option passed to wafliron_runreject");
    }

    # now force wait and oc options to be on
    $opts{wait} = 1;
    $opts{oc}   = 1;

    ( $found, $oc_message ) = wafliron_start( $fc, $container_name, %opts );
    if ( _wi_status( $fc, $type, $container_name ) != 2 ) {
        die("Wafliron Optional Commit did not complete on volume $container_name"
        );
    }

    wafliron_reject( $fc, $container_name );

    $oc_message = $tail->get_logs( skip_cached_logs => 1 );
    return ( $found, $oc_message );
} ## end sub wafliron_runreject($$;%)

## @name ::wafliron_review($fc, $container_name[, %opts])
##
## @summary
##     Run the 'aggr wafliron review' command for the specified volume
##     (trad/aggr) and return the output.
##
## @description
##     This function mimics the 'aggr wafliron review <aggr/trad>' CLI
##     command.
##
## @arg $fc Required: filer connection object.
##
## @arg $container_name Required: name of aggregate or traditional
##                                volume
##

sub wafliron_review($$;%) {
    my ( $fc, $name, %opts ) = @_;
    my $review_msg = "";

    my $container_name = get_container( $fc, $name );
    my $review_cmd = "aggr wafliron review " . $container_name;

    my $client = NACL::C::Client->find();
    my $client_apiset = $client->apiset();

    # Iron_on_root in-core does not produce a review file
    if ( !$::IRON_ON_ROOT || $::IRON_ON_ROOT =~ /spare|multi/ ) {

        # get a handle on the /etc/log/wafliron/oc-aggr-current file
        logcomment ("Trying to re-mount root volume");
        my $mount_obj = _mount_root_vol($fc);

        # Get the mount point
        my $mount_root = $mount_obj->mount_point();
        logcomment("mount_point for root vol - $mount_root");

        my $messages_file
            = $mount_root . "/etc/log/wafliron/oc-$container_name-current";
        $review_msg = $client_apiset->execute_command(
                             command => "cat $messages_file" );

        # Close mount obj
        $mount_obj->purge();

        # Adding $review_cmd to make it easier for log reviews
        logcomment("Wafliron::wafliron_review: $review_cmd");
        logcomment(
            "Wafliron::wafliron_review: review_msg is: \n" . "$review_msg" );
    } else {
        $review_msg = SayWrap::sayEOF( $fc, "priv set -q test; $review_cmd" );
    }

    return $review_msg;
} ## end sub wafliron_review($$;%)

## @name ::wafliron_logfiles_present($fh, $fc, $volume)
##
## @summary
##     Checks to see if logs files exist for a
##     volume specified, if volume is not defined, current
##	   root volume will be used
##
## @description
##     This function checks to see if the log files
##     are present for the given Volume.
##
## @arg $filer_host Required: filer host
## @arg $filer_connection Required: filer connection object.
##
## @arg $volume Optional: name of the volume
##
## @arg $wafliron_type=> optional "nlif" is used.
##                       Default "oc" is used.

sub wafliron_logfiles_present($$;$%) {

    #
    # Enable metadata visible flag so we can
    # check the wafliron metadir
    #

    my ( $filer_host, $filer_connection, $vol_name, %opts ) = @_;
    my ($iron_log_dir);
    my ($ret_val);
    my ($old_wafl_metadata_flag);
    my (@files);

    #
    # Initial local variables
    my ($wafliron_type)
        = defined( $opts{wafliron_type} ) ? $opts{wafliron_type} : "oc";

    logDbg(
        "Wafliron::wafliron_logfiles_present - Enable wafl_metadata_visible");

    $old_wafl_metadata_flag = misc_printflag(
        host => $filer_host,
        flag => 'wafl_metadata_visible'
    );

    $ret_val = misc_setflag(
        host  => $filer_host,
        flag  => 'wafl_metadata_visible',
        value => 1
    );

    if ( !($ret_val) ) {
        $@ = "Unable to set wafl_metadata_visible";
        return 0;
    }

    if ( defined($vol_name) ) {
        $iron_log_dir = $vol_name;
    } else {

        #
        # For now, the log files reside in the root directory
        #
        logDbg("Wafliron::wafliron_logfiles_present - Get Root Directory");

        $iron_log_dir = vol_root($filer_connection);
    } ## end else [ if ( defined($vol_name...

    if ($@) {
        return 0;
    }

    $iron_log_dir = "/vol/$iron_log_dir/metadir/irondir/";
    logDbg("Wafliron::wafliron_logfiles_present - Check for log files");

    my (undef, undef, $status_files_list) = say($filer_connection, "priv set -q test; ls $iron_log_dir ");

    #
    # Would like to check the uuid of the volume but nothing
    # is available, so for now, check to see that nothing is there.
    #

    logDbg( "Wafliron::wafliron_logfiles_present - ls $iron_log_dir results\n"
            . "$status_files_list" );

    @files = split( /\n/, $status_files_list);

    if ( $wafliron_type eq "nlif" ) {
        # since iron_container_bitmap is not present after SN
        # we will check if the iron metadir directory is empty or not
        # and return 1 or 0 accordingly

        # checking if folder is empty (. and .. dir excluded)
        if ( scalar(@files) > 2) {
            return 1;
        } else {
            return 0;
        }
    } elsif ( $wafliron_type eq "oc" ) {
        if ( grep ( /OC-Change-Log/, @files ) ) {
            return 1;
        }
    } else {
        return 0;
    }

    #
    # Restoring old value
    #

    logDbg(
        "Wafliron::wafliron_logfiles_present - Restoring wafl_metadata_visible to $old_wafl_metadata_flag"
    );
    misc_setflag(
        host  => $filer_host,
        flag  => 'wafl_metadata_visible',
        value => $old_wafl_metadata_flag
    );
    return 0;
} ## end sub wafliron_logfiles_present($$;$%)

## @name ::wafliron_populate_vol($fc, $volume, %opt)
##
## @summary
##     Populates a volume to with some files and snapshots.
##
## @description
##     This routine populates a volume  and takes snapshots during
##     the population.
##
## @arg $filer_connection Required: filer connection object
## @arg $vol_name Required: volume name
## @arg prefix_filename=> optional prefix filename to be used.
##                        Defaults wafliron.
## @arg num_files=> optional number of files to create. Defaults to 100.
## @arg prefix_snapshot=> optional prefix snapshot name to be used. Defaults
##                        wafliron_snapshot.
## @arg num_snapshots=> optional number of snapshots to create.
##                      Defaults to 100.
## @arg percent_full=> optional percentage of the filesystem to fill.
##                      Defaults to 10.
##
## @return %populate_status
##          {status} equal 1 if succeed 0 if failed
##          {filelist} array of the name of files created
##          {snaplist} array of the names of snapshot created
##

sub wafliron_populate_vol ($$;%) {

    my ( $fc, $vol_name, %opts ) = validate_mixed(
        @_,
        [ { type => OBJECT }, { type => SCALAR }, ],
        {   prefix_filename => {
                type     => SCALAR,
                optional => 1,
                default  => "wafliron"
            },
            num_files => { type => SCALAR, optional => 1, default => 100 },
            num_snapshots => { type => SCALAR, optional => 1, default => 0 },
            prefix_snapshotname => {
                type     => SCALAR,
                optional => 1,
                default  => "wafliron_snapshot"
            },
            percent_full => { type => SCALAR, optional => 1, default => 10 }
        }
    );

    my ( $num_files, $num_snapshots, $num_files_snapshot );
    my ( $file_size, $percent );
    my ( %diskfree, $avail, $used, $total );
    my (@snapshot_list) = ();
    my (@file_list)     = ();
    my %populate_status = (
        status   => 0,
        filelist => @file_list,
        snaplist => @snapshot_list
    );
    my ($WAFLIRON_MAX_SNAPSHOTS) = 20;

    #
    # Initial local variables
    my ($prefix_filename)     = $opts{prefix_filename};
    my ($prefix_snapshotname) = $opts{prefix_snapshotname};
    $num_files     = $opts{num_files};
    $num_snapshots = $opts{num_snapshots};
    $percent       = $opts{percent_full};

    #
    # Check for reasonable values for num_files and snapshots
    #

    if ( $num_files == 0 ) {
        logDbg(
            "Wafliron::wafliron_populate_vol - Changing $num_files to 100\n");
        $num_files = 100;
    }

    if ( $num_snapshots <= 0 ) {
        $num_snapshots      = 0;
        $num_files_snapshot = $num_files;
    } else {
        if ( $num_snapshots > $WAFLIRON_MAX_SNAPSHOTS ) {
            logcomment(
                "Wafliron::wafliron_populate_vol - Setting snapshots form $num_snapshots "
                    . "to $WAFLIRON_MAX_SNAPSHOTS" );
            $num_snapshots = $WAFLIRON_MAX_SNAPSHOTS;
        } ## end if ( $num_snapshots > ...

        if ( $num_snapshots > $num_files ) {
            logcomment(
                "Wafliron::wafliron_populate_vol - Setting snapshots form $num_snapshots "
                    . "to num_files_snapshot" );
            $num_files_snapshot = 1;
        } ## end if ( $num_snapshots > ...

        $num_files_snapshot = int( $num_files / $num_snapshots );

        logcomment(
            "Wafliron::wafliron_populate_vol - Number of Files: $num_files\n"
                . "Number of Snapshot: $num_snapshots\n"
                . "Number of File/Snapshot: $num_files_snapshot\n"
                . "Percentage of Volume: $percent\n" );

    } ## end else [ if ( $num_snapshots <=...

    logDbg("Wafliron::wafliron_populate_vol - Populate volume\n");

    eval {
        my ( $snapid, $fileid );
        my ($total_percent);

        #
        # Get volume information
        #
        logDbg("Wafliron::wafliron_populate_vol - Get df information\n");

        %diskfree      = df( $fc, $vol_name, options => '-k' );
        $used          = $diskfree{volume}->{used};
        $total         = $diskfree{volume}->{total};
        $avail         = $diskfree{volume}->{avail};
        $total_percent = ( $total * $percent ) / 100;

        logDbg(   "Wafliron::wafliron_populate_vol -df for volume $vol_name\n"
                . "   total: $total\n"
                . "   used:  $used\n"
                . "   avail: $avail\n"
                . "   percent: $percent\n"
                . "   total_percent: $total_percent\n" );

        if ( $total_percent <= $used ) {
            logcomment(
                "Wafliron::wafliron_populate_vol - Volume already at $percent full\n"
            );
            $populate_status{status} = 1;
            return (%populate_status);
        } ## end if ( $total_percent <=...

        #
        # Calculate size of file for the given number of files
        #

        $file_size
            = int( ( ( $total_percent - $used ) * 1024 ) / $num_files );

        for (
            $snapid = 0, $fileid = 1;
            ( $snapid < $num_snapshots ) || ( $snapid == 0 );
            $snapid++
            )
        {
            my ($num_end_snapshot);

            $num_end_snapshot = ( $snapid + 1 ) * $num_files_snapshot;

            for (
                ;
                ( $fileid <= $num_files ) && ( $fileid <= $num_end_snapshot );
                $fileid++
                )
            {

                my ($filename) = "/vol/$vol_name/$prefix_filename." . $fileid;

                write_file(
                    filer => $fc,
                    path  => $filename,
                    size  => $file_size
                );

                push( @file_list, $filename );
            } ## end for ( ; ( $fileid <= $num_files...

            if ( $snapid < $num_snapshots ) {
                my ($snapname) = "$prefix_snapshotname." . $snapid;
                logcomment(
                    "Wafliron::wafliron_populate_vol - Creating snapshot $snapname\n"
                );
                snapshot_create( $fc, $vol_name, $snapname );
                push( @snapshot_list, $snapname );
            } ## end if ( $snapid < $num_snapshots)
        } ## end for ( $snapid = 0, $fileid...
    };

    if ($@) {
        logcomment( "Wafliron::wafliron_populate_vol - Setup error:" . "$@" );
        $populate_status{status} = 0;
    } else {
        $populate_status{status}   = 1;
        $populate_status{snaplist} = [@snapshot_list];
        $populate_status{filelist} = [@file_list];
    }

    return %populate_status;
} ## end sub wafliron_populate_vol ($$;%)
## @name ::wafliron_cleanup_vol($fc, $volume, %opt)
##
## @summary
##     Removes the specified set of files and snapshots.
##
## @description
##     This routine removes the give set of files and snapshots.
##
## @arg $filer_connection Required: filer connection object
## @arg $vol_name Required: volume name
## @arg \@filelist Optional: File list
## @arg \@snaplist Optional: Snapshot list
##
## @return 1 if removed all files/snapshot else 0

sub wafliron_cleanup_vol ($$;%) {
    logcomment("Wafliron::wafliron_cleanup_vol");
    my ( $fc, $vol_name, %opts ) = validate_mixed(
        @_,
        [ { type => OBJECT }, { type => SCALAR }, ],
        {   filelist => {
                type     => ARRAYREF,
                optional => 1,
                default  => ()
            },
            snaplist => { type => ARRAYREF, optional => 1, default => () },
        }
    );

    my (@filelist);
    my (@snaplist);

    @filelist = @{ $opts{filelist} };
    @snaplist = @{ $opts{snaplist} };

    eval {
        for (@filelist)
        {
            my ($file_name) = $_;
            filer_cli_test_say( $fc, "rm $file_name" );
        } ## end for (@filelist)

        for (@snaplist) {
            my ($snap_name) = $_;
            logcomment(
                "Wafliron::wafliron_cleanup_vol - deleting snapshot is $_");
            snapshot_delete( $fc, $vol_name, $snap_name ) or die $@;
        } ## end for (@snaplist)

    };

    if ($@) {
        return 0;
    }
    return 1;
} ## end sub wafliron_cleanup_vol ($$;%)

## @name ::wafliron_start_on_root($fc, wait=>0, root => aggr0)
##
## @summary
##     Starts wafliron on the root volume
##
## @description
##      This routine runs wafliron on the root volume.
##
## @arg $filer_connection Required: Filer Connectrec
## @arg %opts
##      status_interval: Defaults to 1 min
##      root: name of the root aggr/vol.Defaults to aggr0
##      wait: wait for wafliron to complete or return immediately with an id
##            Defaults to 1
##      panic_on_corruption: Value for the bootarg wafl-iron-panic-on-corruption?
##      ignore_missing_complete_mesg: Doesn't look for "Wafliron completed" message
##                                    in the /etc/messages file. In some scenarios, this
##                                    message doesn't show up, hence added this option
##      setflags: A hash of flags that need to be set at FIRMWARE prompt.
##
## @return
##      $ref if wait == 0.
##      (-1,$msgs) if wait == 0 and wafliron has completed before the filer comes up
##      ($found,$corruption_msgs) in other cases
##       Returns captured messages if ignore_missing_complete_mesg => 1

sub wafliron_start_on_root($;%) {

    my ( $fconn, %opts ) = validate_mixed(
        @_,
        [ { type => OBJECT }, ],
        {   status_interval     => { type => SCALAR, default  => 1 },
            root                => { type => SCALAR, default  => 'aggr0' },
            iron_timeout        => { type => SCALAR, default  => 1600 },
            cmd_timeout         => { type => SCALAR, default  => 1600 },
            wait                => { type => SCALAR, default  => 1 },
            panic_on_corruption => { type => SCALAR, optional => 1 },
            ignore_missing_complete_mesg => { type => SCALAR, default  => 0 },
            setflags                     => { type => SCALAR, optional => 1 }
        }
    );

    my $conn_type = $fconn->conntype();
    my $host_obj  = $fconn->hostp();
    my $node      = NACL::STask::Node->new(hostrec => $fconn->hostp());
    my $client    = NACL::C::Client->find();
    my $client_apiset = $client->apiset();

    if ( $conn_type !~ m/console|pm_cons/ ) {
        $fconn = make_conn( $host_obj, '#console' );
    }

    my ( $end_time, $error, $ran, $tail, $mount_root, $wait, $opts_val );
    my $type = wafliron_type( $fconn, $opts{'root'} );
    $fconn->timeout( $opts{'cmd_timeout'} );
    $wait = $opts{'wait'};
    my $root     = $opts{'root'};
    my $setflags = $opts{setflags};
    my %transit_opts;
    $transit_opts{'timeout'} = 3600;
    $transit_opts{'haltcmd'} = 'halt -f -t 0';

    logcomment ("Trying to re-mount root volume");
    my $mount_obj = _mount_root_vol($fconn);

    # Get the mount point
    $mount_root = $mount_obj->mount_point();
    logcomment("mount_point for root vol - $mount_root");

    my $cmd;
    if ( -e "$mount_root/etc/messages" ) {
        $cmd
            = "$main::SUDO_PATH mv $mount_root/etc/messages $mount_root/etc/messages.bkp "
            . "\&\& $main::SUDO_PATH touch $mount_root/etc/messages";
    } else {
        $cmd = "$main::SUDO_PATH touch $mount_root/etc/messages";
    }

    $client_apiset->execute_command( command => $cmd );
    $mount_obj->purge();

    if ( defined( $opts{'panic_on_corruption'} ) ) {
        transit_to( $fconn, 'FIRMWARE', %transit_opts );
        if ( $opts{'panic_on_corruption'} == 1 ) {
            $opts_val = 'true';
        } elsif ( $opts{'panic_on_corruption'} == 0 ) {
            $opts_val = 'false';
        } else {
            $opts_val = $opts{'panic_on_corruption'};
        }
        say( $fconn,
            "setenv wafl-iron-panic-on-corruption? $opts_val",
            '-noprompt',
            '-wait',
            "CFE|LOADER",
            '-timeout',
            LOCAL_WAFLIRON_TIMEOUT
        );
    } ## end if ( defined( $opts{'panic_on_corruption'...

    # Burt 504389 workaround
    #say( $fconn, "setenv walloc_active_check_enabled false",
    #         '-noprompt', '-wait', "CFE|LOADER", '-timeout',
    #         LOCAL_WAFLIRON_TIMEOUT );

    # Burt 499666 workaround for IOC
    #say( $fconn, "setenv wafl-ioc-root-rsh-disable? false",
    #         '-noprompt', '-wait', "CFE|LOADER", '-timeout',
    #         LOCAL_WAFLIRON_TIMEOUT );

    while ( my ( $var, $value ) = each( %{$setflags} ) ) {
        if ( $var =~ /wafl_inconsistent_panic_testing|wafl_iron_panic_on_corruption/ ) {
            $var = 'wafl-inconsistent-panic-testing?'
                if ( $var =~ /wafl_inconsistent_panic_testing/ );
            $var = 'wafl-iron-panic-on-corruption?'
                if ( $var =~ /wafl_iron_panic_on_corruption/ );
            $var = 'wafl-snapiron-enabled?'
                if ( $var =~ /wafl_snapiron_enabled/ );
            $var = 'wafl-iron-lazy-dist-checks?'
                if ( $var =~ /wafl_iron_lazy_dist_checks/ );

            if ( $value == 1 ) {
                $value = 'true';
            } elsif ( $value == 0 ) {
                $value = 'false';
            }
        }
        if ( $value =~ /true|false/ ) {
            say( $fconn, "setenv $var $value",
                '-noprompt', '-wait', "CFE|LOADER", '-timeout',
                    LOCAL_WAFLIRON_TIMEOUT );
        }
    }

    transit_to( $fconn, 'BOOT_MENU', %transit_opts );

    ## @step Start wafliron from boot menu
    logDbg(
        'Wafliron::wafliron_start_on_root - Start wafliron from boot menu');
    say( $fconn, "wafliron -t $opts{'status_interval'}", '-blind' );

    ## @step Wait for filer to come up
    logDbg(
        'Wafliron::wafliron_start_on_root - Wait for filer to come up after wafliron start on root'
    );
    transit_to( $fconn, 'UP' );

    logcomment ("Trying to re-mount root volume");
    $mount_obj = _mount_root_vol($fconn);

    # Get the mount point
    $mount_root = $mount_obj->mount_point();
    logcomment("mount_point for root vol - $mount_root");

    $tail = NACL::MTask::SystemLogDetector->new(
                    command_interface => $node,
                    set_log_file      => ['messages'] );
    $tail->start();

    #Create the id and add to id_table
    my $ref = \( $fconn, $root );
    $id_table{$ref} = [
        (   $fconn, $opts{'root'}, $tail, $mount_root, $opts{'iron_timeout'},
            $type, " "
        )
    ];

    # If wafliron has completed before the filer has come up
    # check for errors.Also handle wait == 0

    my $etc_msgs_file = "$mount_root/etc/messages";

    if ( wafliron_status( $fconn, $opts{'root'} ) == 0 ) {
        my $lines = _get_all_lines_from_file($etc_msgs_file);
        my $found = 0;
        my @list = grep (/wafl.check.info:(error|ALERT)|wafl.iron.*:CRITICAL/, split(/\n/, $lines));
        logcomment ("Wafliron::wafliron_start_on_root - Corruptions found :". Dumper \@list);
        foreach my $line ( @list )
        {
            $found = 1 unless $line =~ /wafl.iron.oc.deletedChangeLog:CRITICAL|wafl.iron.(oc|oc.root).noCorruption:CRITICAL/;
        }
        if ( !$wait ) {
            $tail = undef;
            $mount_obj->purge();
            return ( -1, $lines );
        }

        $tail = undef;
        $mount_obj->purge();
        return ( $found, $lines );
    } ## end if ( wafliron_status( ...

    ## @step Wait for wafliron to finish
    if ($wait) {
        if ( $opts{ignore_missing_complete_mesg} ) {
            while ( wafliron_status( $fconn, $opts{'root'} ) ) {
                Tharn::sleep 5;
            }
            my $big_str_ret_val = _get_all_lines_from_file($etc_msgs_file);
            $mount_obj->purge();
            return ($big_str_ret_val);
        } else {
            my ( $corruption_found, $return_message ) = wafliron_wait($ref);
            $mount_obj->purge();
            return ( $corruption_found, $return_message );
        }
    } else {
        $mount_obj->purge();
        return ($ref);
    }
} ## end sub wafliron_start_on_root($;%)

## @name ::wafliron_on_root ($fc, $name, wait=>0, root_vol => <root vol>)
##
## @summary
##     Starts wafliron on either a root volume or root aggr
##
## @description
##      This routine runs wafliron on the root vol or aggr.
##
## @arg $filer_connection Required: Filer Connectrec
## @arg %opts
##      status_interval: Defaults to 1 min
##      root_vol: name of the root vol.
##      wait: wait for wafliron to complete or return immediately with an id
##            Defaults to 1
##      panic_on_corruption: Value for the bootarg wafl-iron-panic-on-corruption?
##      ignore_missing_complete_mesg: Doesn't look for "Wafliron completed" message
##                                    in the /etc/messages file. In some scenarios, this
##                                    message doesn't show up, hence added this option
##      setflags: A hash of flags that need to be set at FIRMWARE prompt.
##      oc => 1 or 0 or prev_cp.
##       if oc is 1, wafliron is run in Optional Commit mode.
##       if oc is prev_cp, wafliron is run in Optional Commit prev_cp mode.
##       if oc is 0, regular wafliron is run. Default is 0.
##      From BR.0 and IC.1 optional commit is available
##
## @return
##      $ref if wait == 0.
##      (-1,$msgs) if wait == 0 and wafliron has completed before the filer comes up
##      ($found,$corruption_msgs) in other cases
##       Returns captured messages if ignore_missing_complete_mesg => 1

sub wafliron_on_root($$;%) {

    my ( $fconn, $object, %opts ) = validate_mixed(
        @_,
        [ { type => OBJECT }, { type => SCALAR } ],
        {   status_interval => { type => SCALAR, default => 1 },
            root_vol        => { type => SCALAR, default => 'vol0' },
            iron_timeout    => { type => SCALAR, default => 1600 },
            cmd_timeout     => { type => SCALAR, default => 1600 },
            wait            => { type => SCALAR, default => 1 },
            oc              => { type => SCALAR, default => 0 },
            force           => { type => SCALAR, default => "" },
            timeout  => { type => SCALAR, default => LOCAL_WAFLIRON_TIMEOUT },
            tmp_root => { type => SCALAR, default => 0 },
            response_list => { type => SCALAR | HASHREF, default => 0 },
            new_resp_list => { type => SCALAR | HASHREF, default => 0 },
            type          => { type => SCALAR,           default => 'aggr' }
        }
    );

    my $conn_type = $fconn->conntype();
    my $host_obj  = $fconn->hostp();
    my $passwd    = $host_obj->default_password();

    if ( $conn_type !~ m/console|pm_cons/ ) {
        $fconn = make_conn( $host_obj, '#console' );
    }

    my ( %cp_hash, $match, $before, $after, $prompt, $cmd, $ref, $force );
    my $response_list = $opts{'response_list'};
    my $new_resp_list = $opts{'new_resp_list'};

# tmp_root indicates a new root volume has already been created and we should not online the aggr before iron
    my $tmp_root_done = defined( $opts{tmp_root} ) ? "$opts{tmp_root}" : 0;

    my ( $found, $lines );
    my ( $end_time, $error, $ran, $tail, $wait, $waitstring );

    logcomment(
              "Wafliron::wafliron_on_root: object=$object, oc=$opts{'oc'}, \n"
            . "root=$opts{'root_vol'}, wait=$opts{'wait'}, \n"
            . "force=$opts{'force'}, type=$opts{'type'}" );

    $fconn->timeout( $opts{'cmd_timeout'} );
    $wait  = $opts{'wait'};
    $force = $opts{force};

    # If AUTOROOT was selected as the root vol, we need to transit
    # to boot menu
    my $boot_state = boot_state( $fconn, timeout => 300 );
    if ( $boot_state eq 'UP' ) {
        transit_to( $fconn, 'BOOT_MENU' ) or croak $@;
    }

    # # At this point we should be at the boot_menu prompt
    ## @step Start wafliron from boot menu
    logDbg('Wafliron::wafliron_on_root -Start wafliron from boot menu');
    $cmd   = "wafliron ";
    $lines = "";
    $found = 0;
    if ( $opts{'type'} eq 'trad' ) {
        $waitstring = "Volume $object, wafliron completed|system will reboot|System rebooting..";
    } else {
        $waitstring = "Aggregate $object, wafliron completed|system will reboot|System rebooting..";
    }

    if ( $opts{'oc'} ) {
        $cmd .= "-o $force -t $opts{'status_interval'}";
        if ( $opts{'oc'} =~ /reject/ ) {
            croak(
                "Wafliron::wafliron_on_root - Reject is not a valid option for IRON_ON_ROOT=$::IRON_ON_ROOT"
            );
        }
        if ( $opts{'oc'} =~ /commit/ ) {

            # issue wafliron command and wait for it to complete
            ( $match, $before, $after, undef )
                = say( $fconn, $cmd, '-wait', $waitstring, '-noprompt',
                '-timeout', LOCAL_WAFLIRON_TIMEOUT );
            $lines .= $match . $before . $after;

            # If no corruption was found, then no need to commit changes
            if ( $lines !~ /Wafliron did not find any corruption/i ) {

                # See if we need to enter the password
                # Handle any password prompt
                $lines .= _wafliron_login( $fconn, $lines );

                # Look for explicit commit options
                if ( $opts{'oc'} =~ /commit|reject/ ) {
                    eval {

      # issue wafliron command to commit or reject and wait for it to complete
                        ( $match, $before, $after, undef ) = say( $fconn,
                            "priv set -q advanced;aggr wafliron $opts{'oc'} $object",
                            '-wait',
                            "The system will reboot|System rebooted after an internal reboot",
                            '-noprompt',
                            '-timeout',
                            120
                        );
                        $lines .= $match . $before . $after;
                    };
                    if ( $@ =~ /timeout/ ) {
                        ( $match, $before, $after, undef ) = say( $fconn,
                            "priv set -q advanced;aggr wafliron $opts{'oc'} $object",
                            '-wait',
                            "The system will reboot|System rebooted after an internal reboot",
                            '-noprompt',
                            '-timeout',
                            120
                        );
                    } ## end if ( $@ =~ /timeout/ )
                    $lines .= $match . $before . $after;

                    # Handle any password prompt
                    $lines .= _wafliron_login( $fconn, $lines );

                } else {

                    # issue wafliron command wait for rsh service to begin before returning
                    ( $match, $before, $after, undef )
                        = say( $fconn, "", '-wait', 'Enabling rsh service',
                        '-noprompt', '-timeout', LOCAL_WAFLIRON_TIMEOUT );
                    $lines .= $match . $before . $after;

                } ## end else [ if ( $opts{'oc'} =~ /commit|reject/)
            } ## end if ( $lines !~ /Wafliron did not find any corruption/i)
        } elsif ( $opts{'oc'} =~ /prev_cp/ ) {

            # issue wafliron command to commit from a prev_cp
            $cmd .= " -prev_cp ";
            $lines .= _do_prev_cp_prompt( $fconn, $cmd, $response_list, $new_resp_list );

            # wait for wafliron to complete
            if ($wait) {

                # See if we have already hit the waitstring
                if ( $lines !~ /$waitstring/ ) {
                    eval {
                        ( $match, $before, $after, undef ) = say( $fconn,
                            "",
                            '-wait',
                            "$waitstring|Password:",
                            '-noprompt',
                            '-timeout',
                            LOCAL_WAFLIRON_TIMEOUT
                        );
                        $lines .= $match . $before . $after;
                    };
                    if ( $@ =~ /timeout/ ) {

                        # If no corruption was found, then no need to commit changes
                        if ( $lines
                            !~ /Wafliron did not find any corruption/i )
                        {
                            logWarn(
                                "Wafliron time out - did not see the waitstring: $waitstring"
                            );
                        } ## end if ( $lines !~ ...
                    } ## end if ( $@ =~ /timeout/ )
                } ## end if ( $lines !~ /$waitstring/)
            } else {
                Tharn::sleep 30;
            }

            # Handle any password prompt
            $lines .= _wafliron_login( $fconn, $lines );
        } else {
            if ($wait) {

                # issue wafliron command and wait for it to complete
                ( $match, $before, $after, undef )
                    = say( $fconn, $cmd, '-wait', $waitstring, '-noprompt',
                    '-timeout', LOCAL_WAFLIRON_TIMEOUT );
            } else {

               # issue wafliron command wait for rsh service to begin before returning
               eval {
                ( $match, $before, $after, undef )
                    = say( $fconn, $cmd, '-wait', 'Enabling rsh service',
                    '-noprompt', '-timeout', LOCAL_WAFLIRON_TIMEOUT );
               };
            } ## end else [ if ($wait)
            $lines .= $match . $before . $after;
            if ( $match =~ /system will reboot/ ) {

                # wait for system to become stable before trying to logon
                snooze(60);
            }

            # Handle any password prompt
            $lines .= _wafliron_login( $fconn, $lines );

        } ## end else [ if ( $opts{'oc'} =~ /commit/)
    } else {

        # Issue wafliron without any opt_commit parameter
        $cmd .= "$force -t $opts{'status_interval'}";
        if ($wait) {

            # issue wafliron command and wait for it to complete
            ( $match, $before, $after, undef )
                = say( $fconn, $cmd, '-wait', $waitstring, '-noprompt',
                '-timeout', LOCAL_WAFLIRON_TIMEOUT );
            $lines .= $match . $before . $after;
        } ## end if ($wait)
        if ( $match =~ /system will reboot/ ) {

            # wait for system to become stable before trying to logon
            snooze(60);
        }

        # Handle any password prompt
        $lines .= _wafliron_login( $fconn, $lines );

    } ## end else [ if ( $opts{'oc'} )

    my @list = grep (/wafl.check.info:(error|ALERT)|wafl.iron.*:CRITICAL/, split(/\n/, $lines));
    logcomment ("Wafliron::wafliron_on_root - Corruptions found :". Dumper \@list);
    foreach my $line ( @list )
    {
        $found = 1 unless $line =~ /wafl.iron.oc.deletedChangeLog:CRITICAL|wafl.iron.(oc|oc.root).noCorruption:CRITICAL/;
    }

    if ( $lines =~ /vol_pick_root: .* in aggregate .* is now the root volume/ )
    {
        logDbg(
            "Wafliron::wafliron_on_root - Successfully chose new root volume!"
        );
    } ## end if ( $lines =~ ...

    ## @step Wait for wafliron to finish?
    if ($wait) {
        return ( $found, $lines, \%cp_hash );
    } else {
        # Create the id and add to id_table
        $ref = \( $fconn, $object );
        $id_table{$ref} = [
            (   $fconn, $object, $lines, undef, $opts{'iron_timeout'}, undef,
                " "
            )
        ];
        return ($ref);
    } ## end else [ if ($wait)
} ## end sub wafliron_on_root($$;%)

## @name
## ::wafliron_pick_new_root($conn, $aggr, $vol)
##
## @summary Pick a new root volume
##
## @description
##         Halt the system and pick a new root volume
##         Bring system back to up state unless there
##         is a panic. In the case of a panic, control
##         will be returned to the caller at that point.
##
## @example
##          wafliron_pick_new_root($conn, "aggr1", "vol1", reboot => 1|0);
##
## @arg $conn Required. The connection to the filer host,
##            where the old root aggr to be switched.
##
## @arg $aggr Required. The name of the aggr to be switched
##               as root aggr.
##
## @arg $vol Required. The name of the vol to be switched
##               as root volume.
##
## @opt reboot => 1|0
##
## @opt corrupted_vol => 1|0
##
## @return
##      Console text from the last successful
##      say command.
##      The filer may either be in CLI mode or
##      the LOADER depending upon the result of the commands
##

sub wafliron_pick_new_root ($$;@) {
    my $conn = make_conn_and_complain(shift);

    my ( $aggr, $vol, %opts ) = @_;

    my ( $match, $before, $after, $prompt, $text, $cmd, $boot_state );

    ## reboot option
    my $reboot = defined( $opts{reboot} ) ? $opts{reboot} : 0;
    ## scan_speed
    my $scan_speed = defined( $opts{scan_speed} ) ? $opts{scan_speed} : 10000;

# tmp_root indicates a new root volume has already been created and we should not online the aggr before iron
    my $tmp_root_done = defined( $opts{tmp_root} ) ? "$opts{tmp_root}" : 0;
    my $powercycle = defined( $opts{powercycle} ) ? "$opts{powercycle}" : 0;
    my $already_halted
        = defined( $opts{already_halted} ) ? "$opts{already_halted}" : 0;
    my $type = defined( $opts{type} ) ? "$opts{type}" : 'aggr';

    #This option can be used when we pick corrupted vol as root in which case
    #the volume wil be restricted
    my $corrupted_vol
        = defined( $opts{corrupted_vol} ) ? "$opts{corrupted_vol}" : 0;

    # If the delete_all_snaps option is set then we must get into boot menu
    #  and delete all snaps before the system will come up
    my $delete_all_snaps
        = defined( $opts{delete_all_snaps} ) ? "$opts{delete_all_snaps}" : 0;

    # Get the list of flags that need to be set
    my $setflags = defined( $opts{setflags} ) ? $opts{setflags} : undef;

    my %transit_opts;
    $transit_opts{'timeout'} = 3600;

    my $conn_type = $conn->conntype();
    my $host_obj  = $conn->hostp();
    my $root_aggr;

    # For AUTOROOT as the new root vol, we will need to reboot the filer
    # to come up with AUTOROOT as the new root vol. This overwrites
    # $reboot setting if it was set to 0.
    if ( $vol eq "AUTOROOT" ) {
        $reboot = 1;
    }

    if ( $type eq 'trad' ) {
        logcomment(
            "Wafliron::wafliron_pick_new_root: AGGR=$aggr, reboot=$reboot,\n"
                . "powercycle=$powercycle, type=$type, delete_all_snaps=$delete_all_snaps, \n"
                . "tmp_root=$tmp_root_done, already_halted=$already_halted" );
    } else {
        logcomment(
            "Wafliron::wafliron_pick_new_root: AGGR=$aggr, VOL=$vol, reboot=$reboot,\n"
                . "powercycle=$powercycle, type=$type,  delete_all_snaps=$delete_all_snaps, \n"
                . "tmp_root=$tmp_root_done, already_halted=$already_halted" );
    } ## end else [ if ( $type eq 'trad' )

    if ( !$already_halted ) {
        $boot_state = boot_state( $conn, timeout => 300 );
        if ( $boot_state eq 'UP' ) {

            # See if we need to login
            eval { _wafliron_login( $conn, $text ); };

            # Need to know the current root aggr
            eval { $root_aggr = aggr_root($conn); };
            logDbg("Wafliron::wafliron_pick_new_root: aggr_root Error is: $@") if ($@);

     #Corrupted volume will be restricted/not visible(when aggr is restricted)
     #Hence this check not required(volume will not be visible).
     #This check will be avoided only in case corrupted_vol options is set
            logDbg(
                "Wafliron::wafliron_pick_new_root: Checking corrupted volume"
            );
            if ( !$corrupted_vol ) {
                eval {
                    ( $match, $before, $after, undef )
                        = say( $conn, "vol options $vol", -blind );
                };
                if ( $@ || $after . $before =~ /No volume named .* exists/ ) {
                    logDbg(
                        "Wafliron::wafliron_pick_new_root: Intended root volume $vol does not exist"
                    );
                    $vol = ( $aggr eq 'aggr0' ) ? 'vol0' : 'AUTOROOT';
                    logDbg(
                        "Wafliron::wafliron_pick_new_root: Using $vol as root volume"
                    );
                } ## end if ( $@ || $after . $before...
            } ## end if ( !$corrupted_vol )
        } ## end if ( $boot_state eq 'UP')
        logDbg("Wafliron::wafliron_pick_new_root: checking root aggr");
        if ( $root_aggr ne $aggr ) {

            # aggr is not already the root volume so we have to
            # get into maint mode and make it root
            transit_to( $conn, 'MAINT', powercycle => $powercycle, %transit_opts )
                or croak $@;
            ( undef, undef, $text ) = say( $conn,
                "aggr online $aggr",
                '-wait',
                "(Volume|Aggregate).*is.*online|now online|already online|PANIC|Panic|You must specify a root aggregate",
                '-noprompt',
                '-timeout',
                LOCAL_WAFLIRON_TIMEOUT
            );

            if ( $text =~ /panic/i ) {
                logDbg(
                    "Wafliron::wafliron_pick_new_root: Filer panicked - returning before setting aggr as new root"
                );
                return ($text);
            } ## end if ( $text =~ /panic/i)

            ( undef, undef, $text ) = say( $conn, "aggr options $aggr root",
                '-noprompt', '-timeout', LOCAL_WAFLIRON_TIMEOUT );

            if ( $text =~ /panic/i ) {
                logDbg(
                    "Wafliron::wafliron_pick_new_root: Filer panicked - returning before vol_pick_root"
                );
                return 0;
            } ## end if ( $text =~ /panic/i)
        } ## end if ( $root_aggr ne $aggr)

        # Set the required bootargs and reboot to the boot menu
        logDbg(
            "Wafliron::wafliron_pick_new_root: Calling transit_to BOOT_MENU -- intended root vol is: $vol"
        );
        halt($conn);
    } ## end if ( !$already_halted )

# set scanner speed, of course, all we have is slow or not slow. If the use wants to set the
# scan speed to less then 1000 then we will make it slow, otherwise we will leave
# it at not slow
    if ( $scan_speed == -1 ) {
        say( $conn, "setenv wafl-scanners-slow? true",
            '-noprompt', '-wait', "CFE|LOADER", '-timeout',
            LOCAL_WAFLIRON_TIMEOUT );

    } elsif ( $scan_speed < 2000 ) {
        say( $conn, "setenv wafl-scanners-slow? true",
            '-noprompt', '-wait', "CFE|LOADER", '-timeout',
            LOCAL_WAFLIRON_TIMEOUT );

    } else {
        say( $conn, "setenv wafl-scanners-slow? false",
            '-noprompt', '-wait', "CFE|LOADER", '-timeout',
            LOCAL_WAFLIRON_TIMEOUT );

    } ## end else [ if ( $scan_speed == -1)

    # Must set env variable(s) to successfully boot up
    logcomment("Wafliron::wafliron_pick_new_root - setflags - ".Dumper $setflags);
    while ( my ($flag, $value) = each( %{$setflags} ) ) {
        if ( $flag =~ /wafl_iron_panic_on_corruption/ ) {
            $flag = 'wafl-iron-panic-on-corruption?';
        } elsif ( $flag =~ /wafl_inconsistent_panic_testing/ ) {
            $flag = 'wafl-inconsistent-panic-testing?';
        } elsif ( $flag =~ /wafl_snapiron_enabled/ ) {
            $flag = 'wafl-snapiron-enabled?';
        } elsif ( $flag =~ /wafl_iron_lazy_dist_checks/ ) {
            $flag = 'wafl-iron-lazy-dist-checks?';
        } else {
            next;
        }

        if ( $value == 1 ) {
            $value = 'true';
        } elsif ( $value == 0 ) {
            $value = 'false';
        }

        # Setenv
        if ( $value =~ /true|false/ ) {
            say( $conn, "setenv $flag $value",
                 '-noprompt', '-wait', "CFE|LOADER", '-timeout',
                 LOCAL_WAFLIRON_TIMEOUT );
        }
    } ## end while

    # Burt 504389 workaround
    #say( $conn, "setenv walloc_active_check_enabled false",
    #         '-noprompt', '-wait', "CFE|LOADER", '-timeout',
    #         LOCAL_WAFLIRON_TIMEOUT );

    # Burt 499666 workaround for IOC
    #say( $conn, "setenv wafl-ioc-root-rsh-disable? false",
    #         '-noprompt', '-wait', "CFE|LOADER", '-timeout',
    #         LOCAL_WAFLIRON_TIMEOUT );

    transit_to( $conn, 'BOOT_MENU' ) or croak $@;

    my $dhcp_str = 'Ctrl-C to skip DHCP search';
    my $start_setup = 'new hostname';

    if ( $vol eq 'AUTOROOT' ) {

     # Send '1' for normal boot, without running vol_pick_root to use AUTOROOT
        $cmd = '1';
        ( $match, $before, $after ) = say( $conn, $cmd, '-wait',
            "$dhcp_str",            -noprompt,  '-timeout',
            LOCAL_WAFLIRON_TIMEOUT, '-maxidle', $::CMDTIMEOUT
        );

        if ( $match =~ /$dhcp_str/ ) {
            logcomment(
                "Wafliron::wafliron_pick_new_root - Starting Setup\n\n");

            # Do a warm start on the filer
            cold_start_filer( newfs => 0 );
        } ## end if ( $match =~ /$dhcp_str/)

    } ## end if ( $vol eq 'AUTOROOT')

    # No need to do a vol_pick_root if the root is a traditional volume
    if ( $type ne 'trad' && $vol ne "AUTOROOT" ) {
        $cmd = "vol_pick_root $vol $aggr";
        eval {
            ( $match, $before, $after, undef )
                = say( $conn, $cmd, '-noprompt', '-wait', "Selection",
                '-timeout', LOCAL_WAFLIRON_TIMEOUT );
        };

        # If we had a timeout error try the command again
        if ( $@ && $@ =~ /IDLE|TIMEOUT/ ) {
            ( $match, $before, $after, $prompt )
                = say( $conn, $cmd, '-noprompt', '-wait', "Selection",
                '-timeout', LOCAL_WAFLIRON_TIMEOUT );
        }
    } ## end if ( $type ne 'trad' )

    if ($delete_all_snaps) {
        logDbg('Wafliron::wafliron_pick_new_root - Must delete snapshots');
        _delete_all_snaps( $conn, $aggr, $vol );
    }

    if ($reboot) {
        if ( $match =~ /Selection/ ) {

            # Need to re-boot to activate new root volume
            ( $match, $before, $after, undef ) = say( $conn,
                "1",
                '-wait',
                "mgr.boot.reason_ok:notice|.*System rebooted.*|Password:|Login|$start_setup",
                -noprompt,
                -timeout,
                LOCAL_WAFLIRON_TIMEOUT
            );
            $text = $match . $before . $after;

            if ( $match =~ /$dhcp_str|$start_setup/ ) {
                logcomment("Wafliron::wafliron_pick_new_root - Starting Setup\n\n");

                # Do a warm start on the filer
                cold_start_filer( newfs => 0 );
            }

# Make sure the system is in UP state - At this point it should only require a login
            transit_to( $conn, 'UP' ) or croak $@;
        } else {
            logDbg(
                "Wafliron::wafliron_pick_new_root: Calling transit_to UP -- if($reboot)"
            );

# Make sure the system is in UP state - At this point it should only require a login
            transit_to( $conn, 'UP' ) or croak $@;
        } ## end else [ if ( $match =~ /Selection/)
    } ## end if ($reboot)

    $boot_state = boot_state( $conn, timeout => 300 );
    if ( $boot_state eq 'UP' ) {
        # set user requested flags
        while ( my ( $flag, $value ) = each( %{$setflags} ) ) {
            say( $conn, "priv set -q test; setflag $flag $value", -timeout, 30 );
        }
    }

    return 1;
} ## end sub wafliron_pick_new_root ($$;@)

## @name
## ::wafliron_create_temp_root($frsh, $aggr, $vol)
##
## @summary Create a temporary root volume
##
## @description
##         Create a temporary root aggr / volume from available
##         disks
##
## @example
##          wafliron_create_temp_root($conn, aggr=>, volume=>, disks=> vol_size=> );
##
## @arg $conn Required. The connection to the filer host
##
## @arg $aggr Optional. The name of the aggr to be created
##
## @arg $vol Optional. The name of the vol to be created
##
## @arg disks=> Optional. The number of disks to use when creating the aggr
##
## @arg vol_size=> Optional. The size of the temp root volume
##
## @return
##	Name of the volume created
##

sub wafliron_create_temp_root ($;@) {

    my ( $frsh, %opts ) = @_;

    my ($cmd,       $err_msg1,  $err_msg2,  $err_msg3,    $err_msg4,
        $err_msg5,  $err_msg6,  $err_msg7,  $err_msg8,    $err_msg9,
        $err_msg10, $err_msg11, $err_msg12, $err_msg13,   $pass_msg,
        $fail_msg,  $query1,    $query2,    $start_setup, $dhcp_str,
        $match,     $before,    $after,     $text,        $fconn,
        $start_setup1
    );

    $err_msg1  = 'Invalid Volume name for command create_temp_root';
    $err_msg2  = 'No (disk|disks) specified for command create_temp_root';
    $err_msg3  = 'Invalid syntax for command create_temp_root';
    $err_msg4  = 'Invalid option .* for command create_temp_root';
    $err_msg5  = 'Invalid raid type.*for command create_temp_root';
    $err_msg6  = 'No disk or invalid disk name specified';
    $err_msg7  = 'Disk \w+\.\w+ is listed twice';
    $err_msg8  = 'Disk \w+\.\w+ does not exist\.';
    $err_msg9  = 'Disk \w+ does not exist';
    $err_msg10 = 'Disk \w+\.\w+ is not a spare disk';
    $err_msg11 = 'Disk \w+\.\w+ is not owned by local host';
    $err_msg12 = 'Assimilation failed';
    $err_msg13 = 'No raid-type specified for command create_temp_root';
    $pass_msg  = 'wafl\.vol\.add:notice';
    $fail_msg  = 'raid\.label\.io\.writeError';
    $query1    = 'Install a new file system\?';
    $query2    = 'This will erase all the data on the disks, are you sure\?';
    #$start_setup = 'aggregate \w+ has completed successfully';
    $start_setup  = 'Running configuration dialog';
    #$start_setup1 = 'System Storage Configuration';
    $start_setup1 = 'Contacting DHCP server';
    $dhcp_str     = 'Ctrl-C to skip DHCP search';

    my $fh    = $frsh->hostp();
    my $conn_type = $frsh->conntype();
    if ( $conn_type !~ m/console|pm_cons/ ) {
        $fconn = connect( $fh->hostname() . "#console", -login );
    } else {
        $fconn = $frsh;
    }

    my $aggr      = defined( $opts{aggr} )  ? $opts{aggr}  : 'temp_aggr';
    my $vol       = defined( $opts{vol} )   ? $opts{vol}   : 'temp_root';
    my $aggr_size = defined( $opts{disks} ) ? $opts{disks} : $::AGGR_SIZE;
    my $vol_size
        = defined( $opts{vol_size} ) ? $opts{vol_size} : $::FLEX_SIZE;

    # Get the list of flags that need to be set
    my $setflags = defined( $opts{setflags} ) ? $opts{setflags} : undef;

    logDbg(
        "Wafliron::wafliron_create_temp_root: IRON_ON_ROOT=$::IRON_ON_ROOT, AGGR=$aggr, VOL=$vol, disks=$aggr_size, vol_size=$vol_size"
    );

    # Need to know the current root vol and aggr
    my $root_vol  = vol_root($frsh);
    my $root_aggr = aggr_root($frsh);

    if ( $::IRON_ON_ROOT =~ /multi|core/ ) {

        # If the aggr does not already exist - create it
        if ( !aggr_exists( $frsh, $aggr ) ) {

            # If the volume and aggr name are the same create a trad volume
            if ( $aggr eq $vol ) {
                eval { vol_create( $frsh, $aggr, $aggr_size ); };
                if ( $@ && $@ =~ /Couldn't create .* disks needed/ ) {
                    vol_rename( $frsh, $::FILER_VOL_TRAD, $aggr )
                        or die(
                        "Unable to create trad vol of size $aggr_size : $@");
                }
            } else {
                eval { aggr_create( $frsh, $aggr, $aggr_size ); };
                if (   $@
                    && $@ =~ /Couldn't create aggregate: .* disks needed/ )
                {
                    aggr_rename( $frsh, $::FILER_AGGR, $aggr )
                        or
                        die("Unable to create aggr of size $aggr_size : $@");
                } ## end if ( $@ && $@ =~ ...
            } ## end else [ if ( $aggr eq $vol )
        } ## end if ( !aggr_exists( $frsh...

        # If the volume does not already exist - create it
        # force guarantee=none since it will become a root volume
        # This should get all the system files copied over as well
        if ( !vol_exists( $frsh, $vol ) ) {
            vol_create_flex( $frsh, $vol, $aggr, $vol_size )
                or die("Unable to create temp root vol");
        } else {
            if ( $vol_size == 0 ) {
                ## @step Get space requirements for new root
                my $df = vol_disk_free(
                    $frsh,
                    path        => $root_vol,
                    space_type  => DF_SPACE,
                    report_type => UNIT_MB
                );
                $vol_size = int( $df->{used} * 2 );
                $_        = $::FLEX_SIZE;
                s/(\d+)(\w)//;
                my $size  = $1;
                my $units = $2;
                if ( $units =~ /g|t/ || $vol_size < $size ) {
                    $vol_size = $size;
                }
                logDbg(
                    "Wafliron:: FLEX_SIZE is: $::FLEX_SIZE, so vol_size will be $vol_size of unit $units for new root"
                );
                $vol_size .= $units;
            } ## end if ( $vol_size == 0 )
            vol_size( $frsh, $vol, $vol_size );
        } ## end else [ if ( !vol_exists( $frsh...

        # Now change to the new root
        wafliron_pick_new_root( $fconn, $aggr, $vol, reboot => 1 );

    } elsif ( $::IRON_ON_ROOT =~ /spare/ ) {

        # If the aggr already exist and it is the root volume already
        # just return, else destroy it before creating a new one
        if ( Aggregate::aggr_exists( $frsh, $aggr ) ) {
            my %aggropts = Aggregate::aggr_options( $frsh, $aggr );
            if ( $aggropts{root} ) {
                return $aggr;
            }
            Aggregate::aggr_purge( $frsh, $aggr,
                destroy_corrupt_volumes => 1 );
        } ## end if ( Aggregate::aggr_exists...

        # Create a new trad root volume
        # save filer license information, need to restore it afterwards
        my ( @junk, @licenses, @disklist, @short_list );
        my $licenseText = SayWrap::sayEOF( $frsh, "license" );
        my @licenseArray = split( '\n', $licenseText );
        foreach my $lic (@licenseArray) {
            next if ( $lic =~ /not licensed/ );
            @junk = split( " ", $lic );
            push( @licenses, $junk[0] );
        }

        my @groups = RaidSqa::RaidSqa_get_group_list(
            host      => $fh,
            filer_vol => 'vol0'
        );

        # Won't work with mixed checksum disks
        foreach my $checksum_type ('block', 'advanced_zoned', 'zoned') {
            @disklist = RaidSqa::RaidSqa_get_disk_list_by_checksum_type(
                                        host       => $fh,
                                        spare_only => "yes",
                                        checksum_type => $checksum_type,
                                        group_name => $groups[0] );
            @short_list = ( $disklist[0], $disklist[1], $disklist[2] );
            last if $#short_list > 0 && defined $short_list[0] && defined $short_list[1];
        }

        # Going to boot menu for creation of temp root volume
        transit_to( $fconn, 'FIRMWARE' );
        say( $fconn, "setenv bootarg.admin.passwd.prompt.setup false",
             '-noprompt', '-wait', "CFE|LOADER", '-timeout',
             LOCAL_WAFLIRON_TIMEOUT );
        transit_to( $fconn, 'MAINT' ) or croak $@;

        # Offline the previous root aggr so we don't end up with two roots
        ( $match, undef, undef, undef ) = say( $fconn,
            "aggr offline $root_aggr",
            '-noprompt',
            '-wait',
            "Aggregate .* is (now|already) offline|(now|already) offline|Are you sure want to take the root (aggregate|volume) .* offline\\?|Are you sure"
        );
        if ( $match =~ /Are you sure/ ) {
            say( $fconn, "yes", '-noprompt' );
        }

        # Going to boot menu for creation of temp root volume
        transit_to( $fconn, 'BOOT_MENU' ) or die("$@");

        $cmd = "create_temp_root $aggr -t raid0 @short_list";
        logcomment(
            "Wafliron::wafliron_create_temp_root - Creating temp root volume"
        );
        ( $match, $before, $after ) = say( $fconn,
            $cmd,
            '-wait',
            "$start_setup|$start_setup1|$dhcp_str|$fail_msg|$err_msg13|$err_msg1|$err_msg2|$err_msg3|$err_msg4|$err_msg5|$err_msg6|$err_msg7|$err_msg8|$err_msg9|$err_msg10|$err_msg11|$err_msg12|$query1|$pass_msg",
            -noprompt,
            '-timeout',
            $::CMDTIMEOUT,
            '-maxidle',
            $::TIMEOUT
        );

        if ( $match
            =~ /$err_msg1|$err_msg2|$err_msg3|$err_msg4|$err_msg5|$err_msg6|$err_msg7|$err_msg8|$err_msg9|$err_msg10|$err_msg11|$err_msg12|$err_msg13/
            )
        {
            die("Temporary trad vol creation failed with error : $match");
        } ## end if ( $match =~ ...

        if ( $match =~ /$query1/i ) {
            ( $match, $before, $after )
                = say( $fconn, 'yes', qr/$query2/, -noprompt, '-timeout',
                $::CMDTIMEOUT * 10,
                '-maxidle', $::TIMEOUT );
        } ## end if ( $match =~ /$query1/i)
        if ( $match =~ /$query2/i ) {
            ( $match, $before, $after )
                = say( $fconn, 'yes', qr/$dhcp_str|$fail_msg/, -noprompt,
                '-timeout', $::CMDTIMEOUT * 10,
                '-maxidle', $::TIMEOUT );
        } ## end if ( $match =~ /$query2/i)
        if ( $match =~ /$dhcp_str|$start_setup|$start_setup1|$pass_msg/ ) {
            logcomment(
                "Wafliron::wafliron_create_temp_root - Temporary trad vol creation successfull. Starting Setup\n\n"
            );

            # Do a warm start on the filer
            logcomment("Wafliron::wafliron_create_temp_root - Calling cold_start_filer with newfs 0");
            eval { cold_start_filer( newfs => 0 ); };
            if ($@) {
                cold_start_filer( newfs => 0 );
            }
            transit_to( $fconn, 'UP' );

            SayWrap::sayEOF( $frsh,
                "priv set -q diag; setflag wafl_inconsistent_panic_testing 1" );
            SayWrap::sayEOF( $frsh,
                "priv set -q diag; setflag wafl_inconsistent_panic 0" );
            SayWrap::sayEOF( $frsh,
                "priv set -q diag; setflag wafl_iron_panic_on_corruption 0" );

            # Re-License the filer
            foreach (@licenses) {
                license( $frsh, $_ )
                    or logDbg(
                    "Wafliron::wafliron_create_temp_root - Unable to set license for $_ : $@"
                    );
            } ## end foreach (@licenses)
            reboot($fconn, timeout=>'2400') if ( grep( /syncmirror/, @licenses ) );

            # Root vol is now a traditional volume
            $vol = $aggr;

        } else {
            die("Temporary trad vol creation failed with error : $match");
        }

    } else {
        die("invalid IRON_ON_ROOT value. $::IRON_ON_ROOT");
    }

    # set user requested flags
    while ( my ( $flag, $value ) = each( %{$setflags} ) ) {
        say( $fconn, "priv set -q test; setflag $flag $value", -timeout, 30 );
    }

    return ($vol);
} ## end sub wafliron_create_temp_root ($;@)

##=========================================================================

=head2 _delete_all_snaps

This function will go delete all snapshots before allowing the system to come up

RETURNS
        0 on failure. $@ will be set with error message
        1 on success

ARGUMENTS

=over

=item %args

        Required
        conn     : Connection object. Should be an console connection.
        Setup arguments : Reference to a hash containing the setup
                          information like ip, dns, netmask etc


=back

=cut    

sub _delete_all_snaps {

    my ( $fconn, $aggr, $vol ) = @_;

    my ( $match, $before, $after, $say_error );
    my $retries = 6;

    ### @step Get to boot menu
    my %transit_opts;
    $transit_opts{'haltcmd'} = 'halt';
    $transit_opts{'timeout'} = 3600;

    transit_to( $fconn, 'BOOT_MENU', %transit_opts ) or croak $@;
    eval {

        # delete all snapshots option
        ( $match, $before, $after, undef )
            = say( $fconn, "boot_snap_delete", '-wait',
            "Delete|stop the mount", -noprompt, -timeout,
            660,                     -maxidle,  120
            );
    };
    $say_error = ($@) ? $@ : "";
    Tharn::sleep(15);
    while (1) {
        if ( $say_error !~ /IDLE|TIMEOUT/ ) {
            Tharn::sleep(5);
            if ( $match =~ /stop the mount/ ) {
                $match = "";
                eval {
                    logDbg("saying no to stop the mount");
                    $retries++;
                    ( $match, $before, $after, undef ) = say( $fconn,
                        "no",
                        '-wait',
                        'Delete|mgr.boot.disk_done|mgr.boot.reason_ok|Password:|Login',
                        -swallow,
                        -noprompt,
                        -prompt,
                        "Do you want to stop the mount of this volume",
                        -timeout,
                        360,
                        -maxidle,
                        10
                    );
                };
                $say_error = ($@) ? $@ : "";
            } elsif ( $match =~ /Delete/ ) {
                $match = "";
                $retries++;
                eval {
                    logDbg("saying yes to delete the snapshot");
                    ( $match, $before, $after, undef ) = say( $fconn,
                        "yes",
                        '-wait',
                        'stop the mount|mgr.boot.disk_done|mgr.boot.reason_ok|Password:|Login',
                        -swallow,
                        -noprompt,
                        -prompt,
                        "Delete?",
                        -timeout,
                        360,
                        -maxidle,
                        10
                    );
                };
                $say_error = ($@) ? $@ : "";
            } elsif ( $match
                =~ /mgr.boot.disk_done|mgr.boot.reason_ok|Password:|Login/
                || $retries <= 0 )
            {
                logDbg("End of delete_all_snaps, match is: $match");
                last;
            } else {

                # poke the console looking for a prompt
                $match = "";
                eval {
                    ( $match, $before, $after, undef ) = say( $fconn,
                        "\n",
                        '-wait',
                        'Delete|stop the mount|mgr.boot.disk_done|mgr.boot.reason_ok|Password:|Login',
                        -swallow,
                        -noprompt,
                        -timeout,
                        360,
                        -maxidle,
                        10
                    );
                };
                $retries--;
                if ( $match
                    =~ /mgr.boot.disk_done|mgr.boot.reason_ok|Password:|Login/
                    || $retries <= 0 )
                {
                    logDbg("End of delete_all_snaps, match is: $match");
                    last;
                } ## end if ( $match =~ ...
            } ## end else [ if ( $match =~ /stop the mount/)
        } else {

            # poke the console looking for a prompt
            $match = "";
            eval {
                ( $match, $before, $after, undef ) = say( $fconn,
                    "\n",
                    '-wait',
                    'Delete|stop the mount|mgr.boot.disk_done|mgr.boot.reason_ok|Password|Login',
                    -noprompt,
                    -timeout,
                    360,
                    -maxidle,
                    10
                );
            };
            $say_error = ($@) ? $@ : "";
            if ( $match
                =~ /mgr.boot.disk_done|mgr.boot.reason_ok|Password:|Login/
                || $retries <= 0 )
            {
                $retries--;
                logDbg("End of delete_all_snaps, match is: $match");
                last;
            } ## end if ( $match =~ ...
        } ## end else [ if ( $say_error !~ /IDLE|TIMEOUT/)
    } ## end while (1)

} ## end sub _delete_all_snaps

##=========================================================================

=head2 _wafliron_login

This function will login to the filer if the password prompt is seen

RETURNS
        Lines from the console incase this is an IRON_ON_ROOT run where
		the output needs to be sent back to the caller

ARGUMENTS

=over

=item %args

        Required
        lines     : Lines from the console up to this point so we can
					look for the password prompt if any

=back

=cut

sub _wafliron_login($$) {

    my ( $fc, $lines ) = @_;
    my ( $conn_type, $fconn, $host_obj, $passwd);

    $conn_type = $fc->conntype();
    if ( $conn_type !~ m/console|pm_cons/ ) {
        $fconn = connect( $fc->hostp()->hostname() . "#console", -login );
    } else {
        $fconn = $fc;
    }
    $host_obj = $fconn->hostp();
    $passwd   = $host_obj->default_password();

    my ( $match, $before, $after );

    # See we have been prompted for the password
    if ( $lines =~ /Password:|Pass|word:|too many login attempts/ ) {
        Tharn::sleep(5);
        eval {
            ( $match, $before, $after, undef )
                = say( $fconn, "$passwd", '-wait',
                "console_login_mgr:info|root logged in from console",
                -noprompt, -timeout, 120 );
            $lines .= $match . $before . $after;
        };
        if ( $@ && $@ =~ /IDLE|TIMEOUT/ ) {
            logDbg(
                "Wafliron::_wafliron_login - Filer is not waiting for a password or not ready to accept one"
            );

            # See if we get a prompt after hitting enter
            Tharn::sleep(5);
            eval {
                ( $match, $before, $after, undef )
                    = say( $fconn, "\n", '-wait',
                    "Password:|Login|too many login attempts",
                    -timeout, 120 );
                $lines .= $match . $before . $after;
                if ( $match =~ /Password|Login|too many login attempts/ ) {
                    Tharn::sleep(5);
                    ( $match, $before, $after, undef ) = say( $fconn,
                        "$passwd",
                        '-wait',
                        "console_login_mgr:info|root logged in from console",
                        -noprompt,
                        -prompt,
                        "Password:|Pass|word:",
                        '-timeout',
                        120
                    );
                    $lines .= $match . $before . $after;
                } ## end if ( $match =~ ...
            };
        } ## end if ( $@ && $@ =~ /IDLE|TIMEOUT/)
    } else {

# See if console is hung up - sleep a bit and feed it a password to see if it will un-hang
        Tharn::sleep(5);
        eval {
            ( $match, $before, $after, undef ) = say( $fconn,
                "$passwd",
                '-wait',
                "Type .* for a list of commands|console_login_mgr:info|root logged in from console",
                -noprompt,
                -timeout,
                120
            );
            $lines .= $match . $before . $after;
        };
        if ( $match =~ /Type .* for a list of commands/ ) {
            logDbg("Wafliron::_wafliron_login - Already logged in.");
        } elsif (
            $match !~ /console_login_mgr:info|root logged in from console/ )
        {
            eval {
                Tharn::sleep(5);
                ( $match, $before, $after, undef )
                    = say( $fconn, "\n", '-wait',
                    "Password:|Login|too many login attempts",
                    -timeout, 120 );
                $lines .= $match . $before . $after;
                if ( $match =~ /Password|Login|too many login attempts/ ) {
                    Tharn::sleep(5);
                    ( $match, $before, $after, undef ) = say( $fconn,
                        "$passwd",
                        '-wait',
                        "console_login_mgr:info|root logged in from console",
                        -noprompt,
                        -timeout,
                        120
                    );
                    $lines .= $match . $before . $after;
                } ## end if ( $match =~ ...
            };

            # Filer may not need a password
        } else {
            if ( $match
                !~ /console_login_mgr:info|root logged in from console/ )
            {
                logDbg(
                    "Wafliron::_wafliron_login - Did not see password request, skipping entering a password"
                );
            } ## end if ( $match !~ ...
        } ## end else [ if ( $match =~ /Type .* for a list of commands/)
    } ## end else [ if ( $lines =~ /Password:|Pass|word:|too many login attempts/)

    return ($lines);
} ## end sub _wafliron_login($$)

sub _get_all_lines_from_file($) {
    my ($filename) = @_;
    open( MSGS_FH, $filename ) or die("Unable open messages file");
    my @msgs = <MSGS_FH>;
    my $ret_lines = join( "", @msgs );
    close(MSGS_FH);
    return ($ret_lines);
} ## end sub _get_all_lines_from_file($)

sub _do_prev_cp_prompt($$) {

    my ( $conn, $cmd, $response_list, $new_resp_list ) = @_;
    my ($lines,      $match,   $before,  $after,
        $after_poke, $prompt,  %cp_hash, $response_timeout,
        $ret_val
    );
    my $say_error;
    my $loop_timeout;
    my $endtime;
    my $response;

    my $inofile_selection_prompt = 'Selection\\?';
    my $prevcp_selection_prompt  = 'prev_cp\\?';
    my $ino_selection_prompt     = 'inode file\\?';

   # issue the start command, and wait for the first inofile selection prompt.
   # use -noprompt to avoid matching any other prompts.
   # issue wafliron command and wait for it to complete

    if ($response_list) {

        $response_timeout = time() + 360;
        eval {
            ( $match, $before, $after, undef ) = say( $conn, $cmd,
                '-wait',           $inofile_selection_prompt,
                -swallow,          -verbose,
                -noprompt,         -timeout,
                $response_timeout, -maxidle, 90
            );

            # capture output
            $lines .= $before;
            $lines .= $after;
        };

        # capture any error
        $say_error = ($@) ? $@ : "NOERROR";
        logDbg("Wafliron::_do_prev_cp_prompt - say_error = $say_error");

        # If no corruption was found, then no need to commit changes
        if ( $lines !~ /Wafliron did not find any corruption/i ) {
            $loop_timeout = 660;
            $endtime      = time() + $loop_timeout;

            # error checking
            while ( $say_error && time() < $endtime ) {
                if ( $say_error =~ /IDLE|TIMEOUT|NOERROR/ ) {

                    eval {
                        $response_timeout = time() + 180;
                        logDbg(
                            "Wafliron::_do_prev_cp_prompt - Poking the console to see if the wafliron command worked..."
                        );
                        ( $match, $before, $after, undef ) = say( $conn, "\n",
                            '-wait',           $inofile_selection_prompt,
                            -swallow,          -verbose,
                            -noprompt,         -timeout,
                            $response_timeout, -maxidle,
                            90
                        );
                    };
                    $say_error = ($@) ? $@ : "";

                    # append any new output
                    $lines .= $before;
                    $lines .= $after;

                    # Go thru the loop one more time if we match the selection prompt.
                    last if ( $match =~ /$inofile_selection_prompt/ );
                } else {
                    croak(
                        "Wafliron::_do_prev_cp_prompt - failed during inofile selection dialogue: $say_error"
                    );
                }
            } ## end while ( $say_error && time...
            if ( $match =~ /$inofile_selection_prompt/ ) {

# expect an inofile prompt for each volume. respond to these one-by-one until timeout or completion.
                $endtime = time() + $loop_timeout;
                my $select_an_inofile = 1;
                while ( $select_an_inofile && ( time() < $endtime ) ) {
                    if ( $before
                        =~ m/Which\s+inofile\s+should\s+be\s+used\s+for\s+\w+\s+(\w+)\?/i
                        )
                    {
                        my $current_vol = $1;
                        logDbg(
                            "Wafliron::_do_prev_cp_prompt - Working on $current_vol"
                        );

                        # extract the list of valid responses from the output
                        my @cp_lines = split( /\n/, $before );
                        my @cp = "r";
                        foreach (@cp_lines) {
                            if (m/\"(.*)\":/) {
                                push @cp, $1;
                            }
                        }
                        $cp_hash{$current_vol} = \@cp;
                        $response
                            = defined( $$response_list{$current_vol} )
                            ? $$response_list{$current_vol}
                            : "r";

                        # OK. we have response
                        logDbg("Sending response: $response");
                        $response_timeout = time() + 180;
                        eval {
                            ( $match, $before, $after, undef ) = say( $conn,
                                $response,
                                '-wait',
                                "Starting wafliron on .* $current_vol|$inofile_selection_prompt",
                                -swallow,
                                -verbose,
                                -noprompt,
                                -prompt,
                                "$inofile_selection_prompt",
                                -timeout,
                                $response_timeout,
                                -maxidle,
                                90
                            );

                            # append any new output
                            $lines .= $before;
                            $lines .= $after;
                        };

                       # Clear out before so we have to poke the console again
                        $before = '';
                        $say_error = ($@) ? $@ : "";

                        # Found the Password prompt? or System rebooting
                        # Don't go thru the loop again
                        if ( $match
                            =~ /Password:|word:|Login|too many login attempts/
                            )
                        {
                            $lines .= _wafliron_login( $conn, $lines );
                        } ## end if ( $match =~ ...
                    } else {

       # try poking the console to get it to reprint the current prompt (which
       # sometimes gets lost in a deluge of console messages)
                        $before = '';
                        $match  = '';
                        eval {
                            $response_timeout = time() + 60;
                            logDbg(
                                "Wafliron::_do_prev_cp_prompt - Poking the console looking for the Which inode file message..."
                            );
                            ( $match, $before, $after, $prompt ) = say( $conn,
                                "\n",
                                '-wait',
                                "$inofile_selection_prompt|Password:|word:|Login|too many login attempts",
                                -noprompt,
                                -swallow,
                                -timeout,
                                $response_timeout,
                                -maxidle,
                                10
                            );
                        };
                        $say_error = ($@) ? $@ : "";

                        # append any new output
                        $lines .= $before;
                        $lines .= $after;
                        if ( $match
                            =~ /Password:|word:|Login|too many login attempts/
                            )
                        {
                            $lines .= _wafliron_login( $conn, $lines );
                		} ## end if ( $match =~ ...

                        $select_an_inofile
                            = ( $match =~ /$inofile_selection_prompt/ )
                            ? 1
                            : 0;
                    } ## end else [ if ( $before =~ ...
                } ## end while ( $select_an_inofile...
            } else {
                croak(
                    "Wafliron::_do_prev_cp_prompt - failed during inofile selection dialogue: $say_error"
                );
            }
        } ## end if ( $lines !~ /Wafliron did not find any corruption/i)
    } elsif ($new_resp_list) {

        # New prev_cp UI has two behaviors interactive and non - interactive
        if ( $new_resp_list->{interactive} )
        {    # when prev_cp interactiveness is needed

# in this prev_cp method, volume and vvol inputs are given
# in the prev_cp selection prompts. No vvol  input is given while starting prev_cp
            my ($no_of_vols, @vols, %vol_flag);
            $Log->comment(
                "Wafliron::_do_prev_cp_prompt - Processing the new prev_cp interactive request"
            );
            $Log->comment("Wafliron::_do_prev_cp_prompt - new_resp_list: " . Dumper $new_resp_list);
            eval {
                $response_timeout = time() + 360;
                eval {
                    ( $match, $before, $after, undef ) = say( $conn, $cmd,
                        '-wait',           $inofile_selection_prompt,
                        -swallow,          -verbose,
                        -noprompt,         -timeout,
                        $response_timeout, -maxidle, 90
                    );
                };
                # capture output
                my $lines .= $before . $after;

                eval {
                    ( $match, $before, $after, undef ) = say( $conn, "\n",
                        '-wait', $inofile_selection_prompt, -swallow, -verbose,
                        -noprompt, -timeout, $response_timeout, -maxidle, 90 );
                };
                # capture output
                $lines .= $before . $after;

                $response_timeout = time() + 180;
                if ( $lines
                    =~ m/Which\s+inofile\s+should\s+be\s+used\s+for\s+\w+\s+(\w+)\?/i
                    or $lines
                    =~ m/Which\s+inofile\s+should\s+be\s+used\s+for\s+(\w+)\?/i
                    )
                {
                    logDbg(
                        "Wafliron::_do_prev_cp_prompt - Working on $1"
                    );
                    ( undef, $before, $after, undef )
                        = say( $conn, $new_resp_list->{'aggr_resp'},
                        '-wait', "$inofile_selection_prompt|$prevcp_selection_prompt|$ino_selection_prompt",
                        -swallow, -noprompt, -prompt,
                        "$inofile_selection_prompt|$prevcp_selection_prompt|$ino_selection_prompt",
                        -timeout, 120 );
                    $lines .= $before . $after;
                } ## end if ( $after . $before ...
            };

            # Find no of volume inputs
            $no_of_vols = keys %{ $new_resp_list->{response_list} };
            logcomment("Number of volumes in respose list = $no_of_vols");

            # Initialize vol flag, to track resp list
            foreach my $val ( keys %{ $new_resp_list->{response_list} } ) {
                $vol_flag{$val} = 0;
            }

# arrange the volumes in csv format which could be given as input in prev_cp UI
            my $list_of_vols
                    = join( ',', keys %{ $new_resp_list->{response_list} } );

            # Iterate till the volume list exhaust
            while ( $no_of_vols ) {
                eval {
                    ( $match, $before, $after, undef ) = say( $conn,
                                "\n",
                                -noprompt,
                                -prompt,
                                "$inofile_selection_prompt|$ino_selection_prompt",
                                -timeout,
                                $response_timeout,
                                -maxidle,
                                90
                    );
                };
                # capture output
                $lines .= $before . $after;

                # Find out the prompt where the entire output of volumes can be given
                if ( $after
                    =~ m/Which\s+(flexvols|volumes)\s+in\s+aggregate\s+(.*)\s+(should|would)/i
                )
                {
                    logDbg( "Wafliron::_do_prev_cp_prompt - Working on $1" );
                    ( undef, $before, $after, undef )
                        = say( $conn, $list_of_vols, -noprompt, -prompt,
                        $inofile_selection_prompt );

                    # Append the logs into $lines so that this can finally be returned.
                    $lines .= $before . $after;
                } elsif ( $after
                        =~ m/Which\s+inofile\s+should\s+be\s+used\s+for\s+\w+\s+(\w+)\?/i
                        or $after
                        =~ m/Which\s+inofile\s+should\s+be\s+used\s+for\s+(\w+)\?/i
                        )
                {
                    logDbg( "Wafliron::_do_prev_cp_prompt - Working on $1" );
                    if ( $new_resp_list->{response_list}->{$1} eq "#0" )
                    {
# when #0 is the vvol option, it should be invalidated and better option to use is 'r' instead
                        $Log->comment(
                            "Invalid prev_vol_vbn no; Using r instead"
                        );
                        ( undef, $before, $after, undef )
                               = say( $conn, 'r', -noprompt, -prompt,
                                      $inofile_selection_prompt );
                        $lines .= $before . $after;
                        $no_of_vols--;
                    } elsif ( $new_resp_list->{response_list}->{$1} eq "#" )
                    {
                        # when # is the vvol option search for prev_volinfo_vbn
                        ( undef, $before, $after, undef )
                                    = say( $conn,
                                        $new_resp_list->{response_list}->{$1},
                                        -noprompt,
                                        -prompt,
                                        $inofile_selection_prompt
                                    );
                        $lines .= $before . $after;
                        ( undef, $before, $after, undef )
                                    = say( $conn,
                                        " ",
                                        -noprompt,
                                        -prompt,
                                        $inofile_selection_prompt
                                    );
                        $lines .= $before . $after;
                        if ( $after =~ m/prev_volinfo_vbn\s*:\s*(\d+)/i )
                        {
                            ( undef, $before, $after, undef )
                                        = say( $conn, "#" . $1,
                                            -timeout,  3600,
                                            -noprompt, -blind
                                        );
                            $lines .= $before . $after;
                            $no_of_vols--;
                        } else {
                            die("Invalid vvol matching done");
                        }
                    } else {
                        ( undef, $before, $after, undef )
                            = say( $conn,
                                   $new_resp_list->{response_list}->{$1},
                                           -noprompt, -blind );
                        $lines = $before . $after;
                        $no_of_vols-- unless $vol_flag{$1} eq 1;
                        $vol_flag{$1} = 1;
                    }
                } else {
                    $Log->comment("Wafliron::_do_prev_cp_prompt -"
                       . " ERROR: Finding the initial prev_cp strings");
                }
            } ## end while

            return $lines;

    # When an incomplete volume response list is given in prev_cp command line
        } elsif ( exists $new_resp_list->{response_list}->{incomplete_list} )
        {

 # in this case some of the vvol inputs are given with prev_cp command
 # for the rest of the volumes, vvol inputs will be asked in selection prompts
            $Log->comment("Semi interactive prev_cp");
            my $incomplete_hash;
            my $no_of_vols      = 0;
            my @incomplete_list = split( /\s*,\s*/,
                $new_resp_list->{response_list}->{incomplete_list} );
            foreach my $key (@incomplete_list) {
                if ( $key =~ /(.*)\s*:\s*(.*)/ ) {
                    $incomplete_hash->{$1} = $2;
                } else {
                    if ( $key =~ m/\s*(.*)\s*/ ) {
                        $incomplete_hash->{$1} = undef;
                        $no_of_vols++;
                    }
                } ## end else [ if ( $key =~ /(.*)\s*:\s*(.*)/)
            } ## end foreach my $key (@incomplete_list)

            $Log->comment(
                "Wafliron::_do_prev_cp_prompt - Total no of vols to be ironed in semi interactive mode: $no_of_vols"
            );
            die("incomplete_list and the volume snap inputs in response_list are not proper"
                ),
                if ( !$no_of_vols );

            # Add the incomplete list to the prev_cp $cmd
            $cmd .= " " . $new_resp_list->{response_list}->{incomplete_list};
            $Log->comment(
                "Wafliron::_do_prev_cp_prompt - Command to be entered in prev_cp $cmd"
            );
            eval {
                ( undef, $before, $after, undef )
                    = say( $conn, $cmd, -noprompt, -prompt,
                    $inofile_selection_prompt );
            };
            $say_error = ($@) ? $@ : "";
            if ($say_error) {
                logcomment(
                    "Did not see inofile_selection_prompt after sending command: $cmd"
                );
            }
            if ( $after . $before
                =~ m/Which\s+inofile\s+should\s+be\s+used\s+for\s+(\w+)/i )
            {
                logcomment(" what is aggregate :: $1 ::");
                ( undef, $before, $after, undef )
                    = say( $conn, $new_resp_list->{'aggr_resp'},
                    -noprompt, -prompt, $inofile_selection_prompt, -timeout,
                    1200 );
                $lines .= $before . $after;
            } ## end if ( $after . $before ...

            # Iterate till the volume list exhaust
            while ( $no_of_vols-- ) {
                if ( $after
                    =~ m/Which\s+inofile\s+should\s+be\s+used\s+for\s+(\w+)\?/i
                    )
                {
                    if ( !$no_of_vols )
                    { # For last volume gather the console output till the filer prompt
                        eval {
                            ( undef, $before, $after, undef ) = say( $conn,
                                $new_resp_list->{response_list}->{$1},
                                -noprompt,
                                -prompt,
                                $inofile_selection_prompt
                            );
                        };
                        $say_error    = ($@) ? $@ : "";
                        $loop_timeout = 660;
                        $endtime      = time() + $loop_timeout;

                        # error checking
                        if ( $say_error && time() < $endtime ) {
                            eval {
                                $response_timeout = time() + 60;
                                logDbg(
                                    "Wafliron::_do_prev_cp_prompt - Poking the console looking selection prompt..."
                                );
                                ( $match, $before, $after, $prompt )
                                    = say( $conn,
                                    "\n",
                                    '-wait',
                                    "$inofile_selection_prompt|Password:|word:|Login|too many login attempts",
                                    -noprompt,
                                    -swallow,
                                    -timeout,
                                    $response_timeout,
                                    -maxidle,
                                    10
                                    );
                            };

                            # append any new output
                            $lines .= $before;
                            $lines .= $after;
                            if ( $match
                                =~ /Password:|word:|Login|too many login attempts/
                                )
                            {
                                $lines .= _wafliron_login( $conn, $lines );
                            } ## end if ( $match =~ ...
                        } ## end if ( $say_error && time...
                    } else {
                        ( undef, $before, $after, undef )
                            = say( $conn,
                            $new_resp_list->{response_list}->{$1},
                            -noprompt, -prompt, $inofile_selection_prompt );
                    } ## end else [ if ( !$no_of_vols )
                    $lines .= $before . $after;
                } ## end if ( $after =~ ...
            } ## end while ( $no_of_vols-- )
            return $lines;

        } else {

# all inputs for prev_cp (volume and their vvol options) will be given as
# input in prev_cp command line itself so there is no need to wait for selection prompt
            $Log->comment(
                "Wafliron::_do_prev_cp_prompt - Going to new prev_cp non - interactive mode"
            );

            my $vol_snap_list;

            # Bind the volume and the input as : seperated csv pairs
            foreach my $vol_key ( keys %{ $new_resp_list->{response_list} } )
            {
                $vol_snap_list .= $vol_key . ":"
                    . $new_resp_list->{response_list}->{$vol_key} . ",";
            }
            chop($vol_snap_list);  # Remove the last , if any
                                   #$vol_snap_list = '"'. $vol_snap_list. '"';
            $cmd .= " " . $vol_snap_list;
            eval {
                ( $match, $before, $after, undef )
                    = say( $conn, $cmd, '-wait', $inofile_selection_prompt,
                    -swallow, -verbose, -noprompt, -timeout, 300, -maxidle,
                    90 );
                $match = $before;
            };
            $say_error = ($@) ? $@ : "";
            if ($say_error) {
                logcomment(
                    "Did not see inofile_selection_prompt after sending command: $cmd"
                );
            }
            if ( $before
                =~ m/Which\s+inofile\s+should\s+be\s+used\s+for\s+(\w+)/i )
            {
                ( undef, $match, $after, undef )
                    = say( $conn, $new_resp_list->{'aggr_resp'},
                    -timeout, 1200 );
            } ## end if ( $before =~ ...
            $match .= "\n" . $match . $after;
            return $match;
        } ## end else [ if ( $new_resp_list->...
    } else {
        die("Wafliron::_do_prev_cp_prompt - No valid prev_cp response list is given as input. response_list is:"
                . Dumper($response_list) );
    }

    return ($lines);
} ## end sub _do_prev_cp_prompt($$)

=head2 wafliron_setup_container

Creates a new tradvol/aggregate using the given number of disks.

RETURNS

Return href on success, or dies with error string in $@.

ARGUMENTS

=over 4

=item $filer_host

This is the connection to filer host.

=item $iron_cparams

Hash ref to iron container params to create new aggr.

=back

=cut

sub wafliron_setup_container($$) {

    my ( $filer_host, $iron_cparams ) = (@_);
    my ( $new_vols, $new_aggrs, $return_href );
    my $purge = 0; # purge existing aggr?
    my $wafl_flush_cmd = "wafl flush cache";
    logTrc(
        "wafliron_setup_container: iron_cparams = \n" . Dumper $iron_cparams );

    $purge = 1 unless ( $main::SETUP_DONE eq '1'
                        || $main::SETUP_DONE eq 'reuse' );

    if ( $iron_cparams->{IRON_CONTAINER_LARGE_AGGR} =~ /64-bit/i ) {
        $iron_cparams->{IRON_CONTAINER_LARGE_AGGR} = 1;
    } elsif ( $iron_cparams->{IRON_CONTAINER_LARGE_AGGR} =~ /32-bit/i ) {
        $iron_cparams->{IRON_CONTAINER_LARGE_AGGR} = 0;
    }

    eval {
        if ( $iron_cparams->{IRON_CONTAINER_TYPE} =~ /^trad$/i ) {
            ( $new_vols, $new_aggrs ) = setup_volumes(
                $filer_host,
                {   vol       => $iron_cparams->{IRON_CONTAINER_NAME},
                    type      => 'trad',
                    disks     => $iron_cparams->{IRON_CONTAINER_SIZE},
                    raid_type => $iron_cparams->{IRON_CONTAINER_RAID_TYPE},
                    purge_existing => $purge,
                }
            );
        } else {
            ( $new_vols, $new_aggrs ) = setup_volumes(
                $filer_host,
                {   aggr       => $iron_cparams->{IRON_CONTAINER_NAME},
                    type       => 'aggr',
                    disks      => $iron_cparams->{IRON_CONTAINER_SIZE},
                    large_aggr => $iron_cparams->{IRON_CONTAINER_LARGE_AGGR},
                    raid_type  => $iron_cparams->{IRON_CONTAINER_RAID_TYPE},
                    purge_existing => $purge,
                }
            );
        } ## end else [ if ( $iron_cparams->IRON_CONTAINER_TYPE...
    };

    die("wafliron_setup_container: $@\n") if ($@);

    eval {
        # Connect to filer host
        my $filer_conn = make_conn($filer_host);

        # reset wafl scan speed
        filer_cli_test_say ( $filer_conn, "wafl scan speed 10000" );

        my $mdlc_flag = defined($iron_cparams->{IRON_MDLC_DISABLE})
                       ? $iron_cparams->{IRON_MDLC_DISABLE} : 0;
        filer_cli_test_say ( $filer_conn, "setflag wafl_metadata_checks_in_raid $mdlc_flag" );

        my $maxvvbn_flag = defined($iron_cparams->{IRON_MAXVVBN_DISABLE})
                       ? $iron_cparams->{IRON_MAXVVBN_DISABLE} : 0;
        filer_cli_test_say ( $filer_conn, "setflag wafl_maxvvbn_disabled $maxvvbn_flag" );

        my $ldc_flag = defined($iron_cparams->{IRON_LDC_ENABLE})
                       ? $iron_cparams->{IRON_LDC_ENABLE} : 1;
        filer_cli_test_say ( $filer_conn, "setflag wafl_iron_lazy_dist_checks $ldc_flag" );

        my $snapiron_flag = defined($iron_cparams->{IRON_SNAPIRON_ENABLE})
                       ? $iron_cparams->{IRON_SNAPIRON_ENABLE} : 0;
        filer_cli_test_say ( $filer_conn, "setflag wafl_snapiron_enabled $snapiron_flag" );

        # Get aggr details and return
        $return_href->{IRON_CONTAINER_NAMES} = "@$new_aggrs";
        foreach my $aggr (@$new_aggrs) {
            my $wafl_touch_cmd = "wafl touch -A $aggr 73 -private 0 0 99999 0";

            # Setup aggr options
            if ( defined($iron_cparams->{IRON_CONTAINER_OPTIONS}) ) {
                next if $iron_cparams->{IRON_CONTAINER_OPTIONS} =~ /undef/i;
                logcomment("wafliron_setup_container: Setting aggr options...");
                if ( $iron_cparams->{IRON_CONTAINER_OPTIONS} =~ /nosnap/i ) {
                    Aggregate::aggr_options($filer_conn, $aggr, nosnap => 1)
                      or die "Failed to set nosnap option on aggr: $aggr";
                }
            }

            # Create snapshot
            if ( defined $iron_cparams->{IRON_CONTAINER_NUMOF_SNAPS}
                && $iron_cparams->{IRON_CONTAINER_NUMOF_SNAPS} ) {

                my ( @snap_names, @snap_ids, $snap_name, $snap_id );
                foreach
                  my $num ( 1 .. $iron_cparams->{IRON_CONTAINER_NUMOF_SNAPS} ) {
                    say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                    say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

                    # Create snap
                    $snap_name
                        = $iron_cparams->{IRON_CONTAINER_SNAP_NAME_PREFIX}
                        . $aggr
                        . $num ;
                    snapshot_create( $filer_conn, "-A " . $aggr, $snap_name );
                    my @snap_status = snapshot_status( $filer_conn, "-A " . $aggr );
                    foreach my $snap_status ( @snap_status ) {
                        if ($snap_status->{name} eq $snap_name) {
                            $snap_id = $snap_status->{snapid};
                            last;
                        }
                    }
                    push( @snap_names, $snap_name );
                    push( @snap_ids,   $snap_id );
                } ## end foreach my $num ( 1 .. $iron_cparams...

                say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

                if ( $iron_cparams->{IRON_CONTAINER_NUMOF_SNAPS} == 1 ) {
                    $return_href->{$aggr}{IRON_CONTAINER_SNAPFILENAME}
                                                                   = $snap_name;
                    $return_href->{$aggr}{IRON_CONTAINER_SNAPFILEID} = $snap_id;
                } else {
                    $return_href->{$aggr}{IRON_CONTAINER_SNAPFILENAME}
                                                                  = \@snap_names;
                    $return_href->{$aggr}{IRON_CONTAINER_SNAPFILEID}= \@snap_ids;
                }
            } ## end if ( defined $iron_cparams...

            my $status_obj = Aggregate::aggr_status ($filer_conn, $aggr);
            $return_href->{$aggr}{IRON_CONTAINER_NAME}   = $aggr;
            $return_href->{$aggr}{IRON_CONTAINER_TYPE}   = $status_obj->type();
            $return_href->{$aggr}{IRON_CONTAINER_STATUS} = $status_obj->status();
            $return_href->{$aggr}{IRON_CONTAINER_STATE}  = $status_obj->state();
        } ## end foreach my $aggr (@$new_aggrs)

        # Create & Delete aggr SNAPSHOT if needed before iron
        if ( $iron_cparams->{IRON_CONTAINER_SNAP_CREATE_PREIRON} ||
             $iron_cparams->{IRON_CONTAINER_SNAP_DELETE_PREIRON} ) {
            foreach my $aggr ( @$new_aggrs ) {
                logcomment("wafliron_setup_container: Creating snapshot on aggr $aggr...");
                my $wafl_touch_cmd = "wafl touch -A $aggr 73 -private 0 0 99999 0";
                my $snap_name = $aggr."_snapp.$$";
                snapshot_create($filer_conn, "-A " . $aggr, $snap_name);

                say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

                if ( $iron_cparams->{IRON_CONTAINER_SNAP_DELETE_PREIRON} ) {
                    logcomment("wafliron_setup_container: Deleting snapshot on aggr $aggr...");
                    snapshot_delete($filer_conn, "-A " . $aggr, $snap_name);
                    say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                    say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush
                }
            }
        }
    };
    die("wafliron_setup_container: $@\n") if ($@);
    logTrc(
        "wafliron_setup_container: return_href = \n" . Dumper $return_href );

    return ($return_href);
} ## end sub wafliron_setup_container($$)

=head2 wafliron_setup_vvols

Creates a new tradvol/flexvol using the given number of disks.

RETURNS

Return aggr of newly created volumes on success, or dies with error string in $@.

ARGUMENTS

=over 4

=item $filer_host

This is the connection to filer host.

=item $iron_vparams

Hash ref to iron volume params to create new volumes.

=back

=cut

sub wafliron_setup_vvols($$) {

    my ( $filer_host, $iron_vparams ) = (@_);
    my ( @new_vols, @dense_vol_list, @compress_vol_list, %vvol_type_hash );
    my $filer_conn = make_conn( $filer_host ); # Connect filer
    my $filer_name = $filer_host->hostname();
    my $purge = 0; # purge existing volume?
    logTrc( "wafliron_setup_vvols: iron_vparams = \n" . Dumper $iron_vparams );
    logTrc( "wafliron_setup_vvols: filer = \n" . Dumper $filer_host->{hostname} );

    $purge = 1 unless ( $main::SETUP_DONE eq '1'
                        || $main::SETUP_DONE eq 'reuse' );

    # Enable compression
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};
        $vvol_ref->{IRON_VVOL_A_SIS} = 0
                   if $vvol_ref->{IRON_VVOL_A_SIS} =~ /undef/i;
        $vvol_ref->{IRON_VVOL_COMPRESSION} = 0
             if $vvol_ref->{IRON_VVOL_COMPRESSION} =~ /undef/i;

        if ( $vvol_ref->{IRON_VVOL_A_SIS}
          || $vvol_ref->{IRON_VVOL_COMPRESSION} ) {
            eval {
                filer_cli_admin_say (
                  $filer_conn,
                  "options licensed_feature.nearstore_option.enable on" );
            };
            last;
        }
    }

    # Create Aggr_vol_list
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my ( @Aggr_vol_list, @vol_list );
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};
        logTrc( "wafliron_setup_vvols: vvol_ref "
                   . "for VVOL_SET$i = \n" . Dumper $vvol_ref );

        foreach ( my $j = 0; $j < $vvol_ref->{IRON_VVOL_NUMOF_VVOLS}; $j++ ) {
            my $vol_name = $vvol_ref->{IRON_VVOL_NAME_PREFIX} . "_VOL$i$j";
            $vvol_type_hash{$vol_name} = $vvol_ref->{IRON_VVOL_TYPE};

            unless ( $vvol_ref->{IRON_VVOL_TYPE} =~ /trad/i ) {
                eval {
                    dense_stop( host      => $main::FILER,
                                path_name => "/vol/$vol_name" );
                };
            } else {
                $vvol_ref->{IRON_VVOL_A_SIS} = 0;
            }

            push ( @dense_vol_list, $vol_name )
                       if $vvol_ref->{IRON_VVOL_A_SIS};
            push ( @compress_vol_list, $vol_name )
                       if $vvol_ref->{IRON_VVOL_COMPRESSION};

            push ( @Aggr_vol_list,
                    {   vol         => $vol_name,
                        type        => $vvol_ref->{IRON_VVOL_TYPE},
                        aggr        => $vvol_ref->{IRON_VVOL_CONTAINER_NAME},
                        vol_size    => $vvol_ref->{IRON_VVOL_SIZE},
                        spcres      => $vvol_ref->{IRON_VVOL_SPACE_GUARANTEE},
                        license_special_options => 1,
                        purge_existing => $purge,
                    }
                );
        } ## end foreach ( my $j = 0; $j < $vvol_ref...

        # Setup volumes
        eval {
            my ( $new_vols, $new_aggrs )
                = setup_volumes( $filer_host, @Aggr_vol_list );
            @new_vols = ( @new_vols, @$new_vols );
            $iron_vparams->{"VVOL_SET$i"}{IRON_VVOL_LIST} = $new_vols;
        };
        die("wafliron_setup_vvols fail: $@\n") if ($@);
    } ## end for ( my $i = 0; $i < $iron_vparams...

    # Auto snap option set
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};
        next if $vvol_ref->{IRON_VVOL_OPTIONS} =~ /undef/i;

        foreach my $vol_name ( @{ $vvol_ref->{IRON_VVOL_LIST} } ) {
            # skip if volume dosen't exists
            next unless Volume::vol_exists( $filer_conn, $vol_name );
            logcomment("wafliron_setup_vvols: Setting vol options for volume $vol_name");
            if ( $vvol_ref->{IRON_VVOL_OPTIONS} =~ /nosnap/i ) {
                Volume::vol_options($filer_conn, $vol_name, nosnap => 1)
                      or die "Failed to set nosnap option on vol: $vol_name";
            }
            if ( $vvol_ref->{IRON_VVOL_OPTIONS} =~ /no_atime_update/i ) {
                Volume::vol_options($filer_conn, $vol_name, no_atime_update => 1)
                      or die "Failed to set nosnap option on vol: $vol_name";
            }
            if ( $vvol_ref->{IRON_VVOL_OPTIONS} =~ /nvfail/i ) {
                Volume::vol_options($filer_conn, $vol_name, nvfail => 1)
                      or die "Failed to set nosnap option on vol: $vol_name";
            }
        }
    }

    # Compression
    foreach my $vol (@compress_vol_list) {
        # Skip if it not flexvol
        next if $vvol_type_hash{$vol} !~ /flex/i;
        logcomment("wafliron_setup_vvols: Setting compression...");

        eval {
            dense_toggle( host      => $main::FILER,
                          path_name => "/vol/$vol",
                          state     => "on" );

        compression::enable_compression_option(
                                  filer   => $main::FILER,
                                  volname => $vol );
        #    or die("Unable to enable compression on '$vol' : $@");
        };
    }

    # Set dense schedule
    foreach my $vol (@dense_vol_list) {
        # Skip if it not flexvol
        next if $vvol_type_hash{$vol} !~ /flex/i;
        logcomment("wafliron_setup_vvols: Setting dense...");

        eval {
            dense_toggle( host      => $main::FILER,
                          path_name => "/vol/$vol",
                          state     => "on" );
        };

        # Update sis config schedule
        eval {
            dense_config (
                      host => $filer_host->{hostname},
                      path_name       => "/vol/$vol",
                      schedule_string => "sun-sat\@0-23",
                      #'inline-compression' => 'true',
                      #'background-compression' => 'true',
                    );
        };
        logWarn("wafliron_setup_vvols: Error: $@") if $@;
    }

    logTrc(
        "wafliron_setup_vvols: return new_vols = \n" . Dumper \@new_vols );

    return (@new_vols);
} ## end sub wafliron_setup_vvols($$)


=head2 wafliron_cleanup_vvols

Cleansup all aggrs & vols created by wafliron_setup_container and wafliron_setup_vvols

RETURNS

Return nothing on success, or log warning with error string in $@.

ARGUMENTS

=over 4

=item $filer_host

This is the connection to filer host.

=item $iron_vparams

Hash ref to iron volume params that was used to create volumes.

=item $iron_cparams

Hash ref to iron container params that was used to create aggr.

=back

=cut

sub wafliron_cleanup_vvols($$$) {

    my ( $filer_host, $iron_vparams, $iron_cparams ) = (@_);
    my $filer_conn = make_conn($filer_host);
    #my $filer_conn = connect( '#rsh', -timeout, $main::TIMEOUT );
    logTrc(
        "wafliron_cleanup_vvols: iron_vparams = \n" . Dumper $iron_vparams);
    logTrc(
        "wafliron_cleanup_vvols: iron_cparams = \n" . Dumper $iron_cparams);

    return if $main::CLEANUP eq 'none';    # No cleanup requried

    logcomment("purge clone volumes...");
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};

        foreach my $vol_name ( @{$vvol_ref->{IRON_VVOL_CLONE_CLONE_LIST}},
                               @{$vvol_ref->{IRON_VVOL_CLONE_LIST}} ) {

            # skip if volume doesn't exists
            next unless Volume::vol_exists( $filer_conn, $vol_name );

            # Check if vol is online
            my $vol_status = vol_status( $filer_conn, $vol_name );
            logcomment( "wafliron_cleanup_vvols: $vol_name status - "
                       . $vol_status->state() );

            # Try onlining
            eval {
                vol_online( $filer_conn, $vol_name );
            };

            # Cleanup clone vol
            unless ( Volume::vol_purge( $filer_conn, $vol_name ) ) {
                Volume::vol_purge( $filer_conn, $vol_name )
                  or logWarn("Unable to purge the clone volume $vol_name: $@\n");
            }
        } ## end foreach my $vol_name
    } ## end for ( my $i = 0; $i < $iron_vparams...

    logcomment("purge volumes...");
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};

        foreach my $vol_name ( @{ $vvol_ref->{IRON_VVOL_LIST} } ) {

            # skip if volume doesn't exists
            next unless Volume::vol_exists( $filer_conn, $vol_name );

            # Check if vol is online
            my $vol_status = vol_status( $filer_conn, $vol_name );
            logcomment( "wafliron_cleanup_vvols: $vol_name status - "
                       . $vol_status->state() );

            # Try onlining
            eval {
                vol_online( $filer_conn, $vol_name );
            };

            # Cleanup vol
            unless ( Volume::vol_purge( $filer_conn, $vol_name ) ) {
                Volume::vol_purge( $filer_conn, $vol_name )
                  or logWarn("Unable to purge the volume $vol_name: $@\n");
            }
        } ## end foreach my $vol_name
    } ## end for ( my $i = 0; $i < $iron_vparams...

    return if $main::CLEANUP eq 'volume';    # No aggr cleanup requried

    logcomment("purge aggrs...");
    if ( Aggregate::aggr_exists(
      $filer_conn, $iron_cparams->{IRON_CONTAINER_NAME}) ) {
        # Cleanup aggr
        Aggregate::aggr_purge (
            $filer_conn,
            $iron_cparams->{IRON_CONTAINER_NAME} )
            or logWarn(
                "Unable to purge the aggregate $iron_cparams->{IRON_CONTAINER_NAME}: $@\n"
            );
    } ## end if ( Aggregate::aggr_exists...
} ## end sub wafliron_cleanup_vvols($$$)

=head2 wafliron_run_ironops

Run iron along with specific operations as user specified.

RETURNS

Return status hashref with iron operation status & messages

ARGUMENTS

=over 4

=item $filer_conn

Connection object to filer.

=item $iron_params

Hash ref to iron params that contains all testing aggr and volume information.

=item $iron_flags_methods_ref

Hash ref to iron flags and methods.

=item $iron_operations_ref

Hash ref to iron operations to be performed.

=back

=cut

sub wafliron_run_ironops($$$@) {

    my ( $filer_conn, $iron_params, $iron_flags_methods_ref,
         $iron_operations_ref, $iron_vvol_params, $iron_cparams ) = (@_);
    logTrc( "wafliron_run_ironops: iron_params = \n" . Dumper $iron_params );
    logTrc( "wafliron_run_ironops: iron_flags_methods_ref = \n"
            . Dumper $iron_flags_methods_ref );
    logTrc( "wafliron_run_ironops: iron_operations_ref = \n"
            . Dumper $iron_operations_ref ) if defined $iron_operations_ref;
    logTrc( "wafliron_run_ironops: iron_vvol_params = \n"
            . Dumper $iron_vvol_params ) if defined $iron_vvol_params;
    logTrc( "wafliron_run_ironops: iron_cparams = \n"
            . Dumper $iron_cparams ) if defined $iron_cparams;

    my ( %status, $id, $wafliron_running, $hamid, $status_obj, $subtests_ref,
         @subtests, $aggr_name, $vol_type, $vol_name, $filer_cons );
    my ( $corruption_found, $corruption_message, $mark_fail, $vol_list );
    my $wafl_flush_cmd = "wafl flush cache";

    # initialize status var
    $status{iron_corruption_found}     = 0;
    $status{recovery_corruption_found} = 0;

    # Aggar name
    if ( defined $iron_params->{IRON_CONTAINER_NAME} ) {
        $aggr_name = $iron_params->{IRON_CONTAINER_NAME};
    } elsif ( defined $iron_params->{AGGR_NAME} ) {
        $aggr_name = $iron_params->{AGGR_NAME};
        $vol_name  = $iron_params->{VOL_NAME};
        $vol_type  = $iron_params->{VOL_TYPE};
    } else {
        die("wafliron_run_ironops: Could not findout aggr name...");
    }

    # Get list of volumes
    $status_obj = Aggregate::aggr_status( $filer_conn, $aggr_name );
    $vol_list   = $status_obj->volumes() if defined $status_obj;

    # Online aggr & vols on req
    if ( $iron_flags_methods_ref->{IRON_ONLINE_CORRUPTED_VOL}
          && !$iron_flags_methods_ref->{IRON_OPTIONAL_COMMIT} ) {
        # Store flags, to restore after filer reboot
        logcomment("wafliron_run_ironops: Online corrupted volume...");
        my $flags_hash = misc_printflag (
                             host => $filer_conn->hostp(),
                             flag => "wafl_metadata_visible "
                                    ."wafl_iron_panic_on_corruption "
                                    ."wafl_inconsistent_panic_testing "
                                    ."wafl_metadata_checks_in_raid "
                                    ."wafl_iron_lazy_dist_checks "
                                    ."wafl_snapiron_enabled "
                                    ."wafl_maxvvbn_disabled "
                                    ."wafl_lwr_trip "
                         );
        logDbg("wafliron_run_ironops: flags_hash = ". Dumper $flags_hash);

        # We could expect panic here with corrupted vol/aggr
        say($filer_conn, "priv set test;bootargs set AUTOBOOT TRUE",
                          '-swallow', '-timeout', 30);

        # Get console session
        my $conn_type = $filer_conn->conntype();
        if ( $conn_type !~ m/console|pm_cons/ ) {
            $filer_cons = connect(
                  $filer_conn->hostp()->hostname() . "#console", -login );
        } else {
            $filer_cons = $filer_conn;
        }

        my ( $match, $before, $after, $cmd, $lines );
        if ( defined $iron_params->{IRON_CONTAINER_NAME} ) {
            $cmd = "aggr online $aggr_name -f";
        } else {
            $cmd = "vol online $iron_params->{VOL_NAME} -f";
        }

        eval {
            ( $match, $before, $after, undef ) = say( $filer_cons,
               $cmd, -noprompt, -prompt, 'Boot Loader version|is now online',
               -timeout, 1200 );
        };
        $lines = $match . $before . $after;
        Tharn::sleep 300; # reboot sleep

        eval {
            if ( defined $iron_params->{IRON_CONTAINER_NAME} ) {
                my $aggr_status = aggr_status( $filer_cons, $aggr_name );
                logDbg("wafliron_run_ironops: aggr_status = ". Dumper $aggr_status);
            } else {
                my $vol_status  = vol_status( $filer_cons, $iron_params->{VOL_NAME} );
                logDbg("wafliron_run_ironops: vol_status = ". Dumper $vol_status);
            }
        };

        # Check for panic
        my @list = grep (/wafl_metadata_check_local/, split(/\n/, $lines));
        logcomment ("Message found :". Dumper \@list);
        if ( scalar @list ) {
            $mark_fail = 1;
        } else {
            $mark_fail = 0;
        }
        # Check for panic
        transit_to( $filer_cons, 'UP' );

        # Must set env variables back after filer reboot
        while ( my ($flag, $value) = each( %{$flags_hash} ) ) {
            # Setflag
            misc_setflag(
                host  => $filer_conn->hostp(),
                flag  => $flag, value => $value,
            );
        } ## end while
        logcomment("wafliron_run_ironops: Corrupted volume is online...");
    }

    # Run iron on aggr/trad volume
    my $iron_aggr_or_vol;
    if ( defined $vol_type && $vol_type =~ /trad/i ) {
        $iron_aggr_or_vol = $vol_name;
    } else {
        $iron_aggr_or_vol = $aggr_name;
    }

    my %nfs_hammer_flag;
    eval {
        # Start NFS HAMMER if needed
        if ( defined $iron_operations_ref
            && !$iron_operations_ref->{IRON_POSTMOUNT_HAMMER}
            && $iron_operations_ref->{IRON_NFS_HAMMER} ) {

            # Get aggr status
            $status_obj = Aggregate::aggr_status( $filer_conn, $aggr_name );

            if ( defined $status_obj ) {
                # Get volume list
                my $vollist_ref
                    = $status_obj->volumes();  # returns reference to an array

                foreach my $vol ( @$vollist_ref ) {
                    $nfs_hammer_flag{$vol} = 0;

                    # Check if vol is online before we start hammer
                    my $vol_status = vol_status( $filer_conn, $vol );

                    # skip offline/restricted vols
                    logcomment( "wafliron_run_ironops: $vol status - "
                                               . $vol_status->state() );
                    next if $vol_status->state() !~ /online/i;
                    $nfs_hammer_flag{$vol} = 1;

                    logcomment( "wafliron_run_ironops: "
                                      . "Start NFS hammer on $vol..." );
                    push @subtests, subtest(
                      sub{ _client_load_vol($filer_conn, $vol, $vol_status); },
                           -bg, -runid, "load_$vol",
                                   '-iffatal', '-ok', '-ifscript', '-ok' );
                } ## end foreach my $vol (@$vollist_ref)
            } ## end if ( defined $status_obj)
        } ## end if ( defined $iron_operations_ref...
    };
    logDbg("wafliron_run_ironops: nfs_hammer_flag: ". Dumper \%nfs_hammer_flag);

    eval {
        if ( $iron_flags_methods_ref->{IRON_CONTAINER_SEGMENT_CLEANING} ) {
            my $realloc = new Perf::Reallocation (
                                   fc          => $filer_conn,
                                   aggr        => $aggr_name,
                                   target_type => 'aggr' );

            logcomment("Starting blkrealloc scanner on $aggr_name...");
            $realloc->start( once => 1, scanner => "blkrealloc" );
            logcomment("Wait for Reallocation on $aggr_name to complete...");
            $realloc->wait_for_status( DONE, timeout => $main::TIMEOUT );
        }

        if ( !$mark_fail && $iron_flags_methods_ref->{IRON_ONLINE_CORRUPTED_VOL} ) {
            logcomment( "Run opts before run iron..." );
            if ( defined $iron_operations_ref && defined $id ) {
                if ( $iron_operations_ref->{IRON_BUF_CACHE_REVALID} ) {
                    push @subtests, subtest( sub{ _wafl_flush_cache($filer_conn,
                          $iron_operations_ref->{IRON_BUF_CACHE_REVALID_SLEEPTIME}); },
                                         -bg, -runid, 'flush_cache',
                                         '-iffatal','-ok','-ifscript','-ok' );
                }

                if ( $iron_operations_ref->{IRON_CLIENT_LOAD_ROOT_VOL} ) {
                    my $root_vol = vol_root( $filer_conn );
                    my $vol_status = vol_status( $filer_conn, $root_vol );
                    push @subtests, subtest(
                          sub{ _client_load_vol($filer_conn, $root_vol, $vol_status); },
                             -bg, -runid, 'load_root_vol',
                                         '-iffatal', '-ok', '-ifscript', '-ok' );
                }

                # Perform ops while iron is running
                $subtests_ref = wafliron_perform_ops( $filer_conn,
                                                         $iron_params,
                                                         $iron_operations_ref,
                                                         $iron_flags_methods_ref,
                                                         $iron_vvol_params );
            }
        }
    };

    # flags
    my ( $mdlc_flag, $maxvvbn_flag, $ldc_flag, $snapiron_flag );
    $mdlc_flag = defined($iron_flags_methods_ref->{IRON_MDLC_DISABLE})
                       ? $iron_flags_methods_ref->{IRON_MDLC_DISABLE} : 0;
    filer_cli_test_say ( $filer_conn, "setflag wafl_metadata_checks_in_raid $mdlc_flag" );

    $maxvvbn_flag = defined($iron_flags_methods_ref->{IRON_MAXVVBN_DISABLE})
                       ? $iron_flags_methods_ref->{IRON_MAXVVBN_DISABLE} : 0;
    filer_cli_test_say ( $filer_conn, "setflag wafl_maxvvbn_disabled $maxvvbn_flag" );

    $ldc_flag = defined($iron_flags_methods_ref->{IRON_LDC_ENABLE})
                       ? $iron_flags_methods_ref->{IRON_LDC_ENABLE} : 1;
    $ldc_flag = 0 unless $iron_flags_methods_ref->{IRON_DSC_ENABLED}; # no LDC/MPT with out DSC
    filer_cli_test_say ( $filer_conn, "setflag wafl_iron_lazy_dist_checks $ldc_flag" );

    $snapiron_flag = defined($iron_flags_methods_ref->{IRON_SNAPIRON_ENABLE})
                       ? $iron_flags_methods_ref->{IRON_SNAPIRON_ENABLE} : 0;
    filer_cli_test_say ( $filer_conn, "setflag wafl_snapiron_enabled $snapiron_flag" );

    if ( $iron_flags_methods_ref->{IRON_OPTIONAL_COMMIT} ) {
        logcomment( "wafliron_run_ironops: Offline aggr for IOC" );
        eval {aggr_offline( $filer_conn, $iron_aggr_or_vol, force => 1, all => 1);};
    }

    # Get aggr status
    $status_obj = Aggregate::aggr_status( $filer_conn, $aggr_name );

    if ( defined $status_obj ) {
        # Get volume list
        my $vollist_ref
                    = $status_obj->volumes();  # returns reference to an array

        foreach my $vol ( @$vollist_ref ) {
            # Check if vol is online before we start hammer
            my $vol_status = vol_status( $filer_conn, $vol );
            logcomment( "wafliron_run_ironops: $vol status - "
                                               . Dumper $vol_status );
        }
    }

    logcomment( "wafliron_run_ironops: Starting wafliron "
                                 . "on aggr/vol $iron_aggr_or_vol" );
    eval {
        ( $corruption_found, $corruption_message ) = wafliron_start (
               $filer_conn,
               $iron_aggr_or_vol,
               DSC        => $iron_flags_methods_ref->{IRON_DSC_ENABLED},
               LSF        => $iron_flags_methods_ref->{IRON_LOCAL_STATUS_FILES},
               force      => $iron_flags_methods_ref->{IRON_FORCEFUL},
               wait       => $iron_flags_methods_ref->{IRON_WAIT},
               oc         => $iron_flags_methods_ref->{IRON_OPTIONAL_COMMIT},
               timeout    => $iron_flags_methods_ref->{IRON_TIMEOUT},
               scan_speed => $iron_flags_methods_ref->{IRON_SCANNER_SPEED},
               container_status_before_ironing => $iron_flags_methods_ref
                   ->{IRON_CONTAINER_STATUS_BEFORE_IRONING},
               panic_on_corruption =>
                   $iron_flags_methods_ref->{IRON_PANIC_ON_CORRUPTION},
               setflags => {
                   wafl_iron_scan_suspend_duration =>
                     $iron_flags_methods_ref->{IRON_SCANNER_SUSPEND_DURATION},
                   wafl_metadata_checks_in_raid => $mdlc_flag,
                   wafl_maxvvbn_disabled        => $maxvvbn_flag,
                   wafl_iron_lazy_dist_checks   => $ldc_flag,
                   wafl_snapiron_enabled        => $snapiron_flag,
               },
        );
    };
    if ( $@ ) {
        die ( $@ ) unless ( $iron_flags_methods_ref->{IRON_IGNORE_START_MSG} );
        undef $id;
    } else {
        $id = $corruption_found unless $iron_flags_methods_ref->{IRON_WAIT};
    }

    # See wafl scan status
    say($filer_conn, "priv set -q advanced; wafl scan status ");

    # Start NFS HAMMER if needed
    if ( defined $iron_operations_ref
            && $iron_operations_ref->{IRON_NFS_HAMMER} ) {

        # Get aggr status
        $status_obj = Aggregate::aggr_status( $filer_conn, $iron_aggr_or_vol );

        if ( defined $status_obj ) {
            # Get volume list
            my $vollist_ref
                    = $status_obj->volumes();  # returns reference to an array

            foreach my $vol ( @$vollist_ref ) {
                if (defined $nfs_hammer_flag{$vol} && $nfs_hammer_flag{$vol}) {
                    logDbg("wafliron_run_ironops: Hammer is already running on $vol");
                    next;
                }

                # Check if vol is online before we start hammer
                my $vol_status = vol_status( $filer_conn, $vol );

                if ( defined $vol_status ) {
                    # skip offline/restricted vols
                    logcomment( "wafliron_run_ironops: $vol status - " . Dumper $vol_status);
                    logcomment( "wafliron_run_ironops: $vol status - "
                                               . $vol_status->state() );

                    logcomment( "wafliron_run_ironops: "
                                      . "Start NFS hammer on $vol..." );
                    push @subtests, subtest(
                        sub{ _client_load_vol($filer_conn, $vol, $vol_status); },
                               -bg, -runid, "load_$vol",
                                   '-iffatal', '-ok', '-ifscript', '-ok' );
                }

                # See wafl scan status
                eval { say($filer_conn, "priv set -q advanced; wafl scan status -V $vol"); };
            } ## end foreach my $vol (@$vollist_ref)
        } ## end if ( defined $status_obj)
    } ## end if ( defined $iron_operations_ref...

    # Run CIFS HAMMER if needed
    if ( defined $iron_operations_ref
            && $iron_operations_ref->{IRON_CIFS_HAMMER}
            && $main::CIFS_CLIENT ne 'undef'
            && !$iron_flags_methods_ref->{IRON_WAIT} ) {

        # Get aggr status
        $status_obj = Aggregate::aggr_status( $filer_conn, $aggr_name );

        if ( defined $status_obj ) {

            # Complete the cifs setup on filer
            logcomment(
                    "Setting up CIFS on filer $main::FILER..."
            );
            if (!is_cifs_setup( host => $main::FILER ))
            {
                die("cifs_setup failed")
                        unless cifs_setup(
                        host             => $main::FILER,
                        passwd_group     => 'yes',
                        cifs_pass        => $main::CIFS_PASSWD,
                        'security-style' => 'multiprotocol',
                        workgroup        => $main::CIFS_WORKGROUP
                        );
            } ## end if ( !is_cifs_setup( host...

            # Get volume list
            my $vollist_ref
                = $status_obj->volumes();  # returns reference to an array
            logcomment(
                "wafliron_run_ironops: Start CIFS hammer on each volume..."
                );
            foreach my $vol_name (@$vollist_ref) {
                # Change security settings on test volume
                logcomment("Change security settings on $vol_name...");
                qtreeSecurity( $filer_conn, "/vol/$vol_name", 'mixed' );
                if ($@) {
                    # Skip hammer run on failed vols
                    logWarn(
                        "Failed to change security settings on $vol_name: $@"
                    );
                    next;
                } ## end if ($@)

                eval {
                    # Start cifs hammer
                    $hamid = cifs_hammer_start(
                            fc             => $filer_conn,
                            cifs_client    => $main::CIFS_CLIENT,
                            cifs_passwd    => $main::CIFS_PASSWD,
                            cifs_user      => $main::CIFS_USER,
                            cifs_workgroup => $main::CIFS_WORKGROUP,
                            vol_name       => $vol_name,
                            verify_traffic => 'yes'
                    );
                };
                if ($@) {
                    logWarn("Error starting hammer on $vol_name: $@");
                } else {
                    logcomment(
                        "wafliron_run_ironops: CIFS hammer started on $vol_name..."
                    );
                    push( @Cifs_hamid, $hamid );
                } ## end else [ if ($@)

                # See wafl scan status
                say($filer_conn, "priv set -q advanced; wafl scan status -V $vol_name");
            } ## end foreach my $vol_name (@$vollist_ref)
        } ## end if ( defined $status_obj)
    } ## end if ( defined $iron_operations_ref...

    # Let hammer run for few mins
    Tharn::sleep 120 if ( !$mark_fail
             && $iron_flags_methods_ref->{IRON_ONLINE_CORRUPTED_VOL} );

    if ( defined $iron_operations_ref && defined $id ) {
        # Make sure that iron is running before we start ops
        logcomment("wafliron_run_ironops: Check wafliron status...");
        $wafliron_running
                = wafliron_status( $filer_conn, $iron_aggr_or_vol );

        # See wafl scan status
        say($filer_conn, "priv set -q advanced; wafl scan status ");

        if ( $iron_operations_ref->{IRON_BUF_CACHE_REVALID} ) {
            push @subtests, subtest(
                    sub{ _wafl_flush_cache($filer_conn,
                         $iron_operations_ref->{IRON_BUF_CACHE_REVALID_SLEEPTIME}); },
                         -bg, -runid, 'flush_cache',
                                   '-iffatal','-ok','-ifscript','-ok' );
        }

        if ( $iron_operations_ref->{IRON_CLIENT_LOAD_ROOT_VOL} ) {
            my $root_vol = vol_root( $filer_conn );
            my $vol_status = vol_status( $filer_conn, $root_vol );
            push @subtests, subtest(
                    sub{ _client_load_vol($filer_conn, $root_vol, $vol_status); },
                         -bg, -runid, 'load_root_vol',
                                   '-iffatal','-ok','-ifscript','-ok' );
        }

        if ( $iron_operations_ref->{IRON_POLL_WAFLIRON_STATUS} ) {
            push @subtests, subtest(
                    sub{ _wafl_poll_wafliron_status($filer_conn,
                      $iron_operations_ref->{IRON_POLL_WAFLIRON_SLEEPTIME}); },
                         -bg, -runid, 'poll_wafliron_status',
                                   '-iffatal','-ok','-ifscript','-ok' );
        }

        if ( $iron_operations_ref->{IRON_START_KIREETI_SCAN} ) {
            push @subtests, subtest(
                    sub{ _wafl_scan_kireeti($filer_conn, $iron_aggr_or_vol); },
                         -bg, -runid, 'wafl_scan_kireeti',
                                   '-iffatal','-ok','-ifscript','-ok' );
        }

        # Perform ops while iron is running
        eval {
            $subtests_ref = wafliron_perform_ops( $filer_conn,
                                                  $iron_params,
                                                  $iron_operations_ref,
                                                  $iron_flags_methods_ref,
                                                  $iron_vvol_params );
        };
        logWarn("wafliron_perform_ops error: $@") if $@;

        # See wafl scan status
        say($filer_conn, "priv set -q advanced; wafl scan status ");
    }

    # reset wafl_iron_scan_suspend_duration
    if ( $iron_flags_methods_ref->{IRON_SCANNER_SUSPEND_DURATION} ) {
        filer_cli_test_say( $filer_conn,
                   "setflag wafl_iron_scan_suspend_duration 0" );
    }

    logcomment("wafliron_run_ironops: Check wafliron status...");
    $wafliron_running
                = wafliron_status( $filer_conn, $iron_aggr_or_vol );

    # reset wafl scan speed
    filer_cli_test_say ( $filer_conn, "wafl scan speed 10000" );
    # See wafl scan status
    say($filer_conn, "priv set -q advanced; wafl scan status ");

    my %id_table_save = %id_table; # save id table

    # Wait for wafliron complete
    eval {
        ( $corruption_found, $corruption_message ) = wafliron_wait($id);
        $wafliron_running
                    = wafliron_status( $filer_conn, $iron_aggr_or_vol );
    };
    logWarn("wafliron_run_ironops: wafliron_wait error - $@") if $@;

    %id_table = %id_table_save; # Copy id table back

    # Check if iron is in iron_restricted state
    if ( $wafliron_running == 2
               || $iron_flags_methods_ref->{IRON_OPTIONAL_COMMIT} ) {

        my ($fc, $container_name, $timeout, $type) = ( $id_table{$id}[0],
                    $id_table{$id}[1], $id_table{$id}[4], $id_table{$id}[5] );
        my $end_time = time() + $timeout;

        # Wait for wafliron complete
        while ( time() < $end_time ) {
            my $cmd = $type . " wafliron status " . $container_name;
            my ( $match, $before, $after, $prompt ) = say( $fc,
                    "priv set -q advanced; $cmd", -swallow, -timeout, 300 );
            my $output = $before . $match . $after . $prompt;
            Tharn::sleep 30;
            last if $output =~ /either commit or reject the changes/i;
            last if $output =~ /previous attempt to start wafliron/i;

            # See wafl scan status
            say($fc, "priv set -q advanced; wafl scan status ");
        }
        Tharn::sleep 120;

        # Commit/Reject the changes after iron run
        if ( $iron_flags_methods_ref->{IRON_COMMIT_CHANGES} ) {
            $id_table{$id}[7] = 'commit'; # Set to commit changes
        } else {
            $id_table{$id}[7] = 'reject'; # Set to reject changes
        }

        eval {
            # Dump wafliron review log
            my $review = wafliron_review($fc, $container_name);
            logTrc( "wafliron_run_ironops: "
                      . "wafliron_review output:\n " . Dumper $review );

            # See wafl scan status
            say($fc, "priv set -q advanced; wafl scan status ");
        };

        eval {
            # Commit or reject the changes
            commit_reject_wait( $id );
            Tharn::sleep 60;
        };

        logcomment( "Wafliron commit/reject complete." );
        if ( $iron_flags_methods_ref->{IRON_COMMIT_CHANGES} ) {
            my $status_obj;
            if ( defined $vol_type && $vol_type =~ /trad/i ) {
                vol_online( $filer_conn, $iron_aggr_or_vol );
                $status_obj = vol_status( $filer_conn, $iron_aggr_or_vol );
            } else {
                aggr_online( $filer_conn, $iron_aggr_or_vol );
                $status_obj = aggr_status( $filer_conn, $iron_aggr_or_vol );
            }

            # Check aggr/vol status
            die ("Unable to bring $iron_aggr_or_vol online\n")
                   unless $status_obj->state() eq "online";
        }

        # See wafl scan status
        say($fc, "priv set -q advanced; wafl scan status ");
    } elsif ( defined $id && $wafliron_running ) {
        # Wait for iron to complete if it is running
        logcomment(
            "wafliron_run_ironops: Wait for wafliron to complete...");

        # Just ignore timeout error. Some corruptions take more time to fix
        eval {
            ( $corruption_found, $corruption_message ) = wafliron_wait($id);
        };
        unless (($@ == '') || ($@ =~ /wafliron timeout/)) {
            die ("wafliron_run_ironops: Failed in wafliron_wait with: $@");
        } else {
            logWarn("wafliron_run_ironops: error in wafliron_wait with: $@");
        }

        # See wafl scan status
        say($filer_conn, "priv set -q advanced; wafl scan status ");

        # Check wafliron status
        $wafliron_running
                    = wafliron_status( $filer_conn, $iron_aggr_or_vol );

        if ( $wafliron_running ) {
            logcomment(
                "wafliron_run_ironops: Check if wafliron is still running..."
            );
            Tharn::sleep 180;
            %id_table = %id_table_save; # Copy id table back
            eval { ($corruption_found, $corruption_message) = wafliron_wait($id); };
            $wafliron_running
                    = wafliron_status( $filer_conn, $iron_aggr_or_vol );
            if ( $wafliron_running ) {
                $status_obj = Aggregate::aggr_status( $filer_conn, $iron_aggr_or_vol );
                # See wafl scan status
                say($filer_conn, "priv set -q advanced; wafl scan status ");
                logTrc( "wafliron_run_ironops: aggr_status status_obj = "
                            . Dumper $status_obj );

                if ( defined $iron_flags_methods_ref->{IRON_RECOVERY_WAIT} ) {
                    logWarn("Iron is still running: $@");
                } else {
                    die("Iron is still running: $@");
                }
            } ## end if
        } else {
            logcomment( "wafliron_run_ironops: iron completed successfully" );
        }
    } else {
        logWarn( "wafliron_run_ironops: Iron completed "
                              . "before operations complete..." )
          unless ($iron_flags_methods_ref->{IRON_ONLINE_CORRUPTED_VOL});
    }

        #if ( ref($corruption_found) =~ /ARRAY/ ) {
        #    %id_table = %id_table_save; # Copy id table back
        #    ( $corruption_found, $corruption_message ) = wafliron_wait( $id );
        #}

    # Assign return val
    $status{iron_corruption_found}   = $corruption_found;
    $status{iron_corruption_message} = $corruption_message;

    # See wafl scan status
    say($filer_conn, "priv set -q advanced; wafl scan status ");

    # Stop all subtests
    foreach my $subtest ( @subtests ) {
        if ( $subtest->is_alive() ) {
            logcomment( "wafliron_run_ironops: kill subtest..." );
            $subtest->kill();
        }
    }

    # Stop NFS hammer if requried
    foreach my $hamid ( @Nfs_hamid ) {
        logcomment( "wafliron_run_ironops: Stop NFS hammer..." );
        nfs_hammer_stop( id => $hamid ) if defined($hamid);
    }

    # Stop CIFS hammer if requried
    foreach my $hamid (@Cifs_hamid) {
        logcomment( "wafliron_run_ironops: Stop CIFS hammer..." );
        cifs_hammer_stop( id => $hamid ) if defined($hamid);
    }

    _iron_wait( $filer_conn, $iron_aggr_or_vol, $iron_flags_methods_ref )
        if ( defined $iron_flags_methods_ref->{IRON_RECOVERY_WAIT} );

    # See wafl scan status
    eval { say($filer_conn, "priv set -q advanced; wafl scan status"); };

    # Wait for sis checkpoint complete
    # Get volume list
    $status_obj = aggr_status( $filer_conn, $aggr_name );
    my $vollist_ref = $status_obj->volumes();
    my @status_list = ("IDLE", "NOSTATUS");
    foreach my $vol_name ( @$vollist_ref ) {
        # Delete lost volume(s)
        if ( defined $iron_flags_methods_ref->{IRON_RECOVERY_LOSTVOL_DELETE}
                 && $iron_flags_methods_ref->{IRON_RECOVERY_LOSTVOL_DELETE}
                 && $vol_name =~ /lost_volume/ ) {
            logcomment("wafliron_run_ironops: Deleting lost volume...");
            eval {
                # Cleanup clone vol
                unless ( Volume::vol_purge( $filer_conn, $vol_name ) ) {
                    Volume::vol_purge( $filer_conn, $vol_name )
                      or logWarn("Unable to purge the clone volume $vol_name: $@\n");
                }
            };
            next;
        }

        eval {
            logcomment("wafliron_run_ironops: sis stop on volume...");
            dense_stop( host      => $main::FILER,
                        path_name => "/vol/$vol_name" );

            logcomment("wafliron_run_ironops: wait for sis scanner...");
            dense_wait_for_status( host => $main::FILER,
                               path_name => "/vol/$vol_name",
                               status => \@status_list,
                               time_gap => '30',
                               time_out => $main::TIMEOUT ) or logWarn($@);
        };
    }

    # Run recovery tool again
    if ( defined $iron_flags_methods_ref->{IRON_SECOND_RUN_RECOVERY_TOOL}
               && $iron_flags_methods_ref->{IRON_SECOND_RUN_RECOVERY_TOOL} ) {

        # Split volume clone
        if ( defined $iron_flags_methods_ref->{IRON_RECOVERY_CLONE_SPLIT_ALL}
                 && $iron_flags_methods_ref->{IRON_RECOVERY_CLONE_SPLIT_ALL} ) {
            logcomment("wafliron_run_ironops: Volume clone split on all volumes...");
            foreach my $vol_name ( @$vollist_ref ) {
                my $volstat_obj = vol_status( $filer_conn, $vol_name );
                if ( $volstat_obj->status() =~ /Volume has clones: (.*)Volume/i ) {
                    my @cvol_list = split(', ', $1);
                    if ( scalar @cvol_list ) {
                        foreach my $cvol ( @cvol_list ) {
                            logcomment("Performing $cvol split...");
                            Volume::Clone::vol_clone_split_start (
                                              conn             => $filer_conn,
                                              clone            => $cvol,
                                              wait_to_complete => 1,
                                              validate         => 1 )
                               or logWarn("Error: Split operation failed.");
                        }
                    }
                }
            }
        }
        elsif ( defined $iron_flags_methods_ref->{IRON_RECOVERY_CLONE_SPLIT}
                 && $iron_flags_methods_ref->{IRON_RECOVERY_CLONE_SPLIT} ) {
            logcomment("wafliron_run_ironops: Volume clone split...");
        }

        # Delete all snapshots
        if ( defined $iron_flags_methods_ref->{IRON_RECOVERY_SNAPSHOT_DELETE_ALL}
                 && $iron_flags_methods_ref->{IRON_RECOVERY_SNAPSHOT_DELETE_ALL} ) {
            logcomment("wafliron_run_ironops: Delete all snapshots...");
            eval { snapshot_delete( $filer_conn,
                                    "-A $iron_aggr_or_vol", 'all', '-a' ); };
            snapshot_list( $filer_conn, "-A $iron_aggr_or_vol" );
            foreach my $vol_name ( @$vollist_ref ) {
                eval { snapshot_delete( $filer_conn, $vol_name, 'all', '-a' ); };
                snapshot_list( $filer_conn, $vol_name );
            }
        }
        elsif ( defined $iron_flags_methods_ref->{IRON_RECOVERY_SNAPSHOT_DELETE}
                 && $iron_flags_methods_ref->{IRON_RECOVERY_SNAPSHOT_DELETE} ) {
            logcomment("wafliron_run_ironops: Delete snapshots...");

            if ( defined $iron_params->{IRON_CONTAINER_NAME} ) {
                # Get aggr snap shots list
                if ( ref($iron_params->{IRON_CONTAINER_SNAPFILENAME}) =~ /ARRAY/ ) {
                    my $sname_ref = $iron_params->{IRON_CONTAINER_SNAPFILENAME};
                    foreach my $snap_name ( @$sname_ref ) {
                        eval { snapshot_delete( $filer_conn,
                                       "-A $aggr_name", $snap_name ); };
                    }
                } else {
                    my $snap_name = $iron_params->{IRON_CONTAINER_SNAPFILENAME};
                    eval { snapshot_delete( $filer_conn,
                                       "-A $aggr_name", $snap_name ); };
                }
            } else {
                # Get snap shots list
                my $snap_list = $iron_params->{SNAP_FILES};
                foreach my $snap_name ( @$snap_list ) {
                    eval {
                        snapshot_delete( $filer_conn,
                                        $vol_name, $snap_name );
                    };
                }
            }
            Tharn::sleep 10;
        }

        # See wafl scan status
        say($filer_conn, "priv set -q advanced; wafl scan status ");

        if ( defined $iron_flags_methods_ref->{IRON_RECOVERY_RUN}
                 && $iron_flags_methods_ref->{IRON_RECOVERY_RUN} ) {
            logcomment("wafliron_run_ironops: Run iron recovery...");
            my ( $id, $found, $output );
            eval {
                $id = wafliron_start (
                            $filer_conn,
                            $iron_aggr_or_vol,
                            timeout => $main::TIMEOUT,
                            DSC     => $iron_flags_methods_ref->{IRON_DSC_ENABLED},
                            timeout => $iron_flags_methods_ref->{IRON_TIMEOUT},
                            wait    => 0,
                            panic_on_corruption => 0,
                            setflags => {
                                wafl_inconsistent_panic_testing => 1,
                            },
                );
            };
            if ( $@ ) {
                die ( $@ ) unless $@ =~ /TIMEOUT/i;
                _iron_wait( $filer_conn, $iron_aggr_or_vol, $iron_flags_methods_ref );
            }

            Tharn::sleep 60; # Wait
            #%id_table = %id_table_save; # Copy id table back
            eval {
                ( $corruption_found, $corruption_message ) = wafliron_wait( $id );
            };

            # Make sure that iron is running before we start ops
            logcomment("wafliron_run_ironops: Check wafliron status...");
            $wafliron_running
                    = wafliron_status( $filer_conn, $iron_aggr_or_vol );

            $status{recovery_corruption_found}   = $corruption_found;
            $status{recovery_corruption_message} = $corruption_message;

            # If found corruption
            if (( defined $corruption_found && $corruption_found )
              && $iron_flags_methods_ref->{IRON_RECOVERY_IGNORE_SNAPSHOT_ERROR} ) {
                my $ver = eval { get_wafl_version( $filer_conn->hostp() ) } || "";
                if ( $ver > WV_ROLLINGROCK_MIN ) {
                    $status{recovery_corruption_found} = 0;
                    my @list =
                       grep (/wafl.check.info:(error|ALERT)|wafl.iron.*:CRITICAL/,
                               split(/\n/, $output));
                    foreach my $line ( @list ) {
                        $status{recovery_corruption_found} = 1
                        unless $line
                          =~ /please delete the snapshot|Snap \d+, private inode|Protecting summary block/; 
                    }
                }
            }

            # See wafl scan status
            say($filer_conn, "priv set -q advanced; wafl scan status ");

            logWarn( "wafliron_run_ironops: Found corruption "
                       . "after running iron recovery tool...") if $found;
            logcomment(
              "wafliron_run_ironops: recovery iron run completed successfully..."
            );
        }
    } ## end if ( defined $iron_flags_methods_ref...

        # Create & Delete VVOL/AGGR SNAPSHOT if needed after iron run
    if ( defined $iron_operations_ref &&
       ($iron_operations_ref->{IRON_SNAP_CREATE_POSTIRON} ||
        $iron_operations_ref->{IRON_SNAP_DELETE_POSTIRON}) ) {
        if ( defined $vol_name )
        {
            logcomment("wafliron_run_ironops: Creating snapshot on vol $vol_name...");
            my $wafl_touch_cmd = "wafl touch -V $vol_name 73 -private 0 0 99999 0";
            my $snap_name = $vol_name."_snapp.$$";
            snapshot_create($filer_conn, $vol_name, $snap_name);

            say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
            say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

            if ($iron_operations_ref->{IRON_SNAP_DELETE_POSTIRON}) {
                logcomment("wafliron_run_ironops: Deleting snapshot on vol $vol_name...");
                snapshot_delete($filer_conn, $vol_name, $snap_name);
                say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush
            }
        } elsif (defined $iron_params->{IRON_CONTAINER_NAME}) {
            logcomment("wafliron_run_ironops: Creating snapshot on aggr $aggr_name...");
            my $snap_name = $aggr_name."_snapp.$$";
            my $wafl_touch_cmd = "wafl touch -A $aggr_name 73 -private 0 0 99999 0";
            snapshot_create($filer_conn, "-A " . $aggr_name, $snap_name);

            say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
            say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

            if ($iron_operations_ref->{IRON_SNAP_DELETE_POSTIRON}) {
                logcomment("wafliron_run_ironops: Deleting snapshot on aggr $aggr_name...");
                snapshot_delete( $filer_conn, "-A " . $aggr_name, $snap_name );
                say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush
            }
        }
    }

    logWarn( "wafliron_run_ironops: ERROR: $@" ) if ($@);
    # See wafl scan status
    say($filer_conn, "priv set -q advanced; wafl scan status ");

    # Stop NFS hammer if requried
    foreach my $hamid (@Nfs_hamid) {
        logcomment("wafliron_run_ironops: Stop NFS hammer...");
        nfs_hammer_stop( id => $hamid ) if defined($hamid);
    }

    # Stop CIFS hammer if requried
    foreach my $hamid (@Cifs_hamid) {
        logcomment("wafliron_run_ironops: Stop CIFS hammer...");
        cifs_hammer_stop( id => $hamid ) if defined($hamid);
    }

    return \%status;
} ## end sub wafliron_run_ironops($$$@)

=head2 wafliron_perform_ops

Perform operations on test aggr or volume.

RETURNS

Return nothing on success.

ARGUMENTS

=over 4

=item $filer_conn

Connection object to filer.

=item $iron_params

Hash ref to iron params.

=item $iron_operations_ref

Hash ref to iron operations params.

=back

=cut

sub wafliron_perform_ops($$$@) {

    my ( $filer_conn, $iron_params, $iron_operations_ref,
         $iron_flags_methods_ref, $iron_vparams ) = (@_);
    my ( $subtests_ref );
    my $wafl_flush_cmd = "wafl flush cache";

    logTrc( "wafliron_perform_ops: iron_params = " . Dumper $iron_params);
    logTrc( "wafliron_perform_ops: iron_operations_ref = "
            . Dumper $iron_operations_ref);

    # root volume
    my $root_vol = vol_root( $filer_conn );

    # sync
    if ( $iron_operations_ref->{IRON_FORCE_CP} ) {
        eval { filer_sync( $filer_conn ); };
    }

    if ( $iron_operations_ref->{IRON_ACCESS_STATUS_FILES}
        && $iron_flags_methods_ref->{IRON_LOCAL_STATUS_FILES} ) {
        logcomment("wafliron_perform_ops: Start accessing iron status files...");
        push @$subtests_ref, subtest (
                        sub { _read_iron_status_files( $filer_conn, $root_vol ); },
                        -bg, -runid, 'read_iron_status_files',
                        '-iffatal', '-ok', '-ifscript', '-ok' );
    }

    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};
        logTrc ("wafliron_perform_ops: iron_vparams = ". Dumper $vvol_ref);

        foreach my $vol_name ( @{$vvol_ref->{IRON_VVOL_LIST}},
                               @{$vvol_ref->{IRON_VVOL_CLONE_LIST}} ) {
            # Check if vol is online
            my $vol_status = vol_status( $filer_conn, $vol_name );
            unless (( $vol_status->state() =~ /restricted/i )
                          || ( $vol_status->state() =~ /offline/i )) {
                if ( $iron_operations_ref->{IRON_RUN_SIO_ALL_FILES} ) {
                    logcomment("wafliron_perform_ops: Run sio on files in volume $vol_name...");
                    push @$subtests_ref, subtest (
                        sub { _run_sio_all_files($filer_conn, $vol_name);},
                               -bg, -runid, 'run_sio' . $vol_name,
                               '-iffatal', '-ok', '-ifscript', '-ok' );
                }
            }
        }
    }

    # Create & Delete VVOL SNAPSHOT if needed
    if ( $iron_operations_ref->{IRON_VVOL_SNAP_CREATE}
        || $iron_operations_ref->{IRON_VVOL_SNAP_DELETE} ) {
        if ( defined $iron_params->{VOL_NAME} ) {
            my $vol_name = $iron_params->{VOL_NAME};

            # Check if vol is online
            my $vol_status = vol_status( $filer_conn, $vol_name );
            unless (( $vol_status->state() =~ /restricted/i )
                          || ( $vol_status->state() =~ /offline/i )) {
                my $wafl_touch_cmd = "wafl touch -V $vol_name 73 -private 0 0 99999 0";
                logcomment("wafliron_perform_ops: Creating snapshot on vol $vol_name...");
                my $snap_name = $vol_name . "_snap.$$";
                snapshot_create( $filer_conn, $vol_name, $snap_name );

                say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

                if ( $iron_operations_ref->{IRON_VVOL_SNAP_DELETE} ) {
                    logcomment(
                    "wafliron_perform_ops: Deleting snapshot on vol $vol_name..."
                );
                    snapshot_delete( $filer_conn, $vol_name, $snap_name );
                    say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                    say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush
                } ## end if ( $iron_operations_ref...
            } ## end if ( defined $iron_params...
        } ## end unless
    } ## end if ( $iron_operations_ref...

    # Create & Delete CONTAINER SNAPSHOT if needed
    if ( $iron_operations_ref->{IRON_CONTAINER_SNAP_CREATE}
        || $iron_operations_ref->{IRON_CONTAINER_SNAP_DELETE} ) {
        if ( defined $iron_params->{IRON_CONTAINER_NAME} ) {
            my $aggr_name = $iron_params->{IRON_CONTAINER_NAME};

            # Check if aggr is online
            my ( $state, $status ) = aggr_status( $filer_conn, $aggr_name );
            unless (( $state =~ /restricted/i ) || ( $state =~ /offline/i )) {
                logcomment( "wafliron_perform_ops:
                      Creating snapshot on aggr $aggr_name..." );
                my $snap_name = $aggr_name . "_snap.$$";
                my $wafl_touch_cmd = "wafl touch -A $aggr_name 73 -private 0 0 99999 0";
                snapshot_create( $filer_conn, "-A " . $aggr_name, $snap_name );

                say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

                if ( $iron_operations_ref->{IRON_CONTAINER_SNAP_DELETE} ) {
                    logcomment( "wafliron_perform_ops:
                        Deleting snapshot on aggr $aggr_name..." );
                    snapshot_delete( $filer_conn, "-A " . $aggr_name, $snap_name );
                    say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                    say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush
                } ## end if ( $iron_operations_ref...
            } ## end if ( defined $iron_params...
        } ## end unless
    } ## end if ( $iron_operations_ref...

    logTrc("returning from wafliron_perform_ops...");
    logcomment("returning from wafliron_perform_ops...");
    return $subtests_ref;
} ## end sub wafliron_perform_ops($$$)

=head2 wafliron_populate_volumes

Crete reuested test files and snapshots on test volume.

RETURNS

Return hash ref on success, or dies with error string in $@.

ARGUMENTS

=over 4

=item $filer_host

This is the connection to filer host.

=item $iron_vparams

Hash ref to iron volume params.

=back

=cut

sub wafliron_populate_volumes($$) {

    my ( $filer_host, $iron_vparams ) = (@_);
    my ( %return_ref, $filer, $rsh_conn );
    logTrc( "wafliron_populate_volumes: iron_vparams = \n"
            . Dumper $iron_vparams);

    my $blk_size_bytes = '64k';
    my $file_size      = 1;
    my $files_per_snap = 1;
    my $filer_conn     = make_conn( $filer_host ); # connect
    my $wafl_flush_cmd = "wafl flush cache";

    # Install license if needed
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};
        if ( $vvol_ref->{IRON_VVOL_SIS_CLONES}
             || $vvol_ref->{IRON_VVOL_CLONES} ) {
            $filer = host( $main::FILER );
            $rsh_conn = connect( '#rsh', -timeout, $main::TIMEOUT );
            license( $rsh_conn, "flex_clone" );
            last;
        }
    }

    my $purge = 0; # purge existing clone volume?
    $purge = 1 unless ( $main::SETUP_DONE eq '1'
                           || $main::SETUP_DONE eq 'reuse' );

    # Work on each vvol set
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};
        logTrc( "wafliron_populate_volumes: vvol_ref for VVOL_SET$i = \n"
                . Dumper $vvol_ref);

        next unless $vvol_ref->{IRON_VVOL_NUMOF_VVOLS}; # Skip in case of zero vvols
        my $vol_size_bytes = int convert_to_bytes( $vvol_ref->{IRON_VVOL_SIZE} );

        my $vol_usable_space =
            int( $vol_size_bytes * ($vvol_ref->{IRON_VVOL_PCT_USAGE}/100) );

        foreach my $vvol_name ( @{ $vvol_ref->{IRON_VVOL_LIST} } ) {
            my $wafl_touch_cmd = "wafl touch -V $vvol_name 73 -private 0 0 99999 0";
            my ($total, $used, $avail, $capacity)
                 = vol_disk_free($filer_conn, path => "/vol/$vvol_name",
                           space_type  => DF_SPACE, report_type => UNIT_MB);
            my $vol_usable_space = int ($avail * ($vvol_ref->{IRON_VVOL_PCT_USAGE}/100));
            my $files_per_snap = 1;
            if ( defined $vvol_ref->{IRON_VVOL_NUMOF_SNAPS}
                        && $vvol_ref->{IRON_VVOL_NUMOF_SNAPS} ) {
                if ( $vvol_ref->{IRON_VVOL_NUMOF_FILES}
                       > $vvol_ref->{IRON_VVOL_NUMOF_SNAPS} ) {
                    $files_per_snap = int ($vvol_ref->{IRON_VVOL_NUMOF_FILES}
                                            / $vvol_ref->{IRON_VVOL_NUMOF_SNAPS});
                }
                $file_size = int ( $vol_usable_space
                      / ($files_per_snap * $vvol_ref->{IRON_VVOL_NUMOF_SNAPS}) );
            } else {
                $file_size = int ( $vol_usable_space / $files_per_snap );
            }

            $return_ref{$vvol_name}{VOL_SIZE} = $total . 'M';
            $return_ref{$vvol_name}{VOL_NAME} = $vvol_name;
            $return_ref{$vvol_name}{VOL_TYPE} = $vvol_ref->{IRON_VVOL_TYPE};
            $return_ref{$vvol_name}{AGGR_NAME}
                                    = $vvol_ref->{IRON_VVOL_CONTAINER_NAME};

            eval {
                my $filer_path = "/vol/" . $vvol_name . "/";
                my $qtree_path = $filer_path . "qtree";
                $file_size = $file_size . 'm';

                # Create qtree
                qtreeCreate( $filer_conn, $qtree_path );
                my $ret_val = filer_cli_advanced_say( $filer_conn,
                           "ls $qtree_path" ) if ( $@ =~ /qtree exists/ );
                my @files = split( /\n/, $ret_val );

                my ( @subtests, @test_file_list, @snap_file_list, $loop_count );
                if ( defined $vvol_ref->{IRON_VVOL_NUMOF_SNAPS}
                        && $vvol_ref->{IRON_VVOL_NUMOF_SNAPS} ) {
                    $loop_count = $vvol_ref->{IRON_VVOL_NUMOF_SNAPS};
                } else {
                    $loop_count = 1;
                }

                for ( my $i = 0; $i < $loop_count; $i++ ) {
                    my @snap_test_files = ();

                    my ( $fbsize, $fnofiles );
                    for ( my $j = 0; $j < $files_per_snap; $j++ ) {

                        my ( $filer_io_cmd, $file_name, $file_path );
                        $file_name
                                = $vvol_ref->{IRON_VVOL_FILE_NAME_PREFIX}
                                    . "_$vvol_name" . "_file$i" . "$j";
                        $file_path = $qtree_path . "/" . $file_name;

                        if ( $vvol_ref->{IRON_FILE_CREATE_METHOD}
                                                           eq "filersio" ) {
                            # Write file using filersio
                            $filer_io_cmd
                                    = "priv set -q test; filersio asyncio_active "
                                      . "0 -r 0 $blk_size_bytes 0 $file_size "
                                      . "$main::IRON_FILERSIO_RUNTIME 20 "
                                      . "$file_path -create -synchronous -fillonce 65";
                        } elsif ( $vvol_ref->{IRON_FILE_CREATE_METHOD}
                                                                   eq "mkfile" ) {
                            # Write file using mkfile
                            if ( $vvol_ref->{IRON_USE_VBN_ZERO} ) {
                                $filer_io_cmd
                                    = "priv set -q test; mkfile -z $file_size $file_path";
                            } else {
                                $filer_io_cmd
                                    = "priv set -q test; mkfile $file_size $file_path";
                            }
                        } else {
                            croak("wafliron_populate_volumes: file create tool"
                                   . " should be either mkfile or filersio.");
                        }

                        push @subtests, subtest( sub{ _write_file($filer_conn,
                           $filer_io_cmd, $file_path, $vvol_ref); }, -bg, -runid,
                           "$file_name", '-iffatal', '-ok', '-ifscript', '-ok' );

                        push( @test_file_list, $file_name );
                        $return_ref{$vvol_name}{$file_name}{TEST_FILE_NAME} = $file_name;
                        $return_ref{$vvol_name}{$file_name}{TEST_FILE_PATH} = $file_path;
                        $return_ref{$vvol_name}{$file_name}{TEST_FILE_SIZE} = $file_size;
                        push( @snap_test_files, $file_name )
                                        if $vvol_ref->{IRON_VVOL_NUMOF_SNAPS};

                        if (($filer_io_cmd =~ /mkfile/ && ($j % 8) eq 7)
                              || ($filer_io_cmd =~ /filersio/ && ($j % 60) eq 59)) {
                            Subtest::wait_finish(subtests => [@subtests]);
                        }
                    } ## end for ( my $j = 0;

                    # Wait for subtests
                    logcomment("Waiting for subtest to get finish....");
                    Subtest::wait_finish(subtests => [@subtests]);
                    SayWrap::sayEOF( $filer_conn, "priv set -q diag; df -h" );

                    # Polulate with snapshots
                    if ( defined $vvol_ref->{IRON_VVOL_NUMOF_SNAPS}
                        && $vvol_ref->{IRON_VVOL_NUMOF_SNAPS} ) {
                        # Create snapshot
                        my $snap_name = $vvol_ref->{IRON_VVOL_SNAP_NAME_PREFIX}
                                        . "_$vvol_name"
                                        . "_snap$i";

                        snapshot_create( $filer_conn, $vvol_name, $snap_name );
                        push( @snap_file_list, $snap_name );

                        say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                        say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

                        # Update return ref
                        $return_ref{$vvol_name}{$snap_name}{SNAP_FILE_NAME} = $snap_name;
                        $return_ref{$vvol_name}{$snap_name}{SNAP_TEST_FILES} = \@snap_test_files;
                        my @snap_status = snapshot_status( $filer_conn, $vvol_name );
                        foreach my $snap_status ( @snap_status ) {
                            if ($snap_status->{name} eq $snap_name) {
                                $return_ref{$vvol_name}{$snap_name}{SNAP_FILE_ID}
                                             = $snap_status->{snapid};
                                last;
                            }
                        }
                    } ## end if ( defined $vvol_ref...
                } ## end for ( my $i = 0;

                $return_ref{$vvol_name}{TEST_FILES} = \@test_file_list;
                $return_ref{$vvol_name}{SNAP_FILES} = \@snap_file_list;

                # Create & Delete VVOL SNAPSHOT if needed before iron
                if ( $vvol_ref->{IRON_VVOL_SNAP_CREATE_PREIRON} ||
                     $vvol_ref->{IRON_VVOL_SNAP_DELETE_PREIRON} ) {
                    logcomment(
                      "wafliron_populate_volumes: Creating snapshot on vol $vvol_name..."
                    );
                    my $snap_name = $vvol_name."_snapp.$$";
                    snapshot_create( $filer_conn, $vvol_name, $snap_name );

                    say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                    say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

                    if ( $vvol_ref->{IRON_VVOL_SNAP_DELETE_PREIRON} ) {
                        logcomment("wafliron_populate_volumes: Deleting snapshot on vol $vvol_name...");
                        snapshot_delete( $filer_conn, $vvol_name, $snap_name );
                        say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                        say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush
                    }
                }
            };
            die("wafliron_populate_volumes: $@\n") if ($@);
            # Buftree info
            eval {
                SayWrap::sayEOF( $filer_conn,
                    "priv set -q test; wafl buftree check -s 0 -V $vvol_name priv:73;" );
                SayWrap::sayEOF( $filer_conn,
                    "priv set -q test; wafl buftree check -s 0 -V $vvol_name priv:76;" );
            };
        } # end of foreach loop
    } ## end for ( my $i = 0; $i < $iron_vparams...

    # Now create clone vols if needed
    logcomment("wafliron_populate_volumes: create clone & pfs vols...");
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my (@Aggr_vol_list, @CAggr_vol_list, @snaplist, %pfs_parent);
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};

        # create pfs volume(s) for each vvol
        foreach ( my $j = 0; $j < $vvol_ref->{IRON_VVOL_NUMOF_VVOLS}; $j++ ) {
            foreach ( my $k = 0; $k < $vvol_ref->{IRON_VVOL_PFSVOL}; $k++ ) {
                # License check
                if (!is_licensed($filer_conn, "flexcache_nfs")) {
                    license($filer_conn, "flexcache_nfs") or croak $@;
                    logcomment("Flexcache license installed");
                }

                # Enable flexcache license & options
                filer_options_set($filer_conn, 'licensed_feature.flexcache_nfs.enable', 'on');
                filer_options_set($filer_conn, 'flexcache.enable', 'on');
                filer_options_set($filer_conn, 'flexcache.access', 'all');
                filer_options_set($filer_conn, 'sparse.unlimited_vols', 'on');

                # Only for flex vols
                next unless $vvol_ref->{IRON_VVOL_TYPE} =~ /flex/i;
                my $parent_vol_name = ${ $vvol_ref->{IRON_VVOL_LIST} }[$j];
                my $vol_name        = "SET$i" . "_PVOL$i$j$k";
                $pfs_parent{$vol_name} = $parent_vol_name;

                if ( defined $purge && $purge ) {
                    eval {
                        # Cleanup pfs vol
                        unless ( Volume::vol_purge( $filer_conn, $vol_name ) ) {
                            Volume::vol_purge( $filer_conn, $vol_name )
                              or logWarn("Unable to purge pfs volume $vol_name: $@\n");
                        }
                    };
                }

                eval {
                    vol_create_flex($filer_conn, $vol_name,
                          $vvol_ref->{IRON_VVOL_CONTAINER_NAME},
                          $vvol_ref->{IRON_VVOL_SIZE},
                          source => "$main::FILER:$parent_vol_name");
                    push (@{$vvol_ref->{IRON_VVOL_PFSVOL_LIST}}, $vol_name);
                };
            } ## end foreach ( my $k = 0; $k < $vvol_ref...
        } ## end foreach ( my $j = 0; $j < $vvol_ref...

        # create clone volume(s) for each vvol
        foreach ( my $j = 0; $j < $vvol_ref->{IRON_VVOL_NUMOF_VVOLS}; $j++ ) {
            foreach ( my $k = 0; $k < $vvol_ref->{IRON_VVOL_CLONES}; $k++ ) {
                # clone for only flex vols
                next unless $vvol_ref->{IRON_VVOL_TYPE} =~ /flex/i;
                my $parent_vol_name = ${ $vvol_ref->{IRON_VVOL_LIST} }[$j];
                my $vol_name        = "SET$i" . "_CVOL$i$j$k";

                # Check if clone vol exists
                if ( vol_exists( $filer_conn, $vol_name ) ) {
                    # check if the volume is a clone or not
                    my ($parent, $snapname) =
                         Volume::Clone::get_clone_info(
                                         conn  => $filer_conn,
                                         clone => $vol_name );
                    if ( $@ ) {
                        logWarn ("volume $vol_name already"
                               . " exists but is not a clone. purging");
                        $purge = 1;
                    }
                }

                push(
                    @Aggr_vol_list,
                    {   vol      => $vol_name,
                        parent   => $parent_vol_name,
                        type     => "clone",
                        aggr     => $vvol_ref->{IRON_VVOL_CONTAINER_NAME},
                        vol_size => $vvol_ref->{IRON_VVOL_SIZE},
                        spcres   => $vvol_ref->{IRON_VVOL_SPACE_GUARANTEE},
                        purge_existing => $purge,
                    }
                );
            } ## end foreach ( my $k = 0; $k < $vvol_ref...
        } ## end foreach ( my $j = 0; $j < $vvol_ref...

        eval {
            logcomment("wafliron_populate_volumes: creating " . Dumper @Aggr_vol_list);
            my ( $new_vols, $new_aggrs )
                = setup_volumes( $filer_host, @Aggr_vol_list );
            $vvol_ref->{IRON_VVOL_CLONE_LIST} = $new_vols;
        };
        logWarn("wafliron_populate_volumes failed to create clones: $@\n") if ($@);

        foreach ( my $j = 0; $j < $vvol_ref->{IRON_VVOL_NUMOF_VVOLS}; $j++ ) {
            foreach ( my $k = 0; $k < $vvol_ref->{IRON_VVOL_CLONES}; $k++ ) {
                # clone for only flex vols
                next unless $vvol_ref->{IRON_VVOL_TYPE} =~ /flex/i;
                my $parent_vol_name = "SET$i" . "_CVOL$i$j$k";
                foreach ( my $l = 0; $l < $vvol_ref->{IRON_VVOL_CLONES_CLONE}; $l++ ) {
                    my $vol_name        = "SET$i" . "_CCVOL$i$j$k$l";
                    push(
                        @CAggr_vol_list,
                        {   vol      => $vol_name,
                            parent   => $parent_vol_name,
                            type     => "clone",
                            aggr     => $vvol_ref->{IRON_VVOL_CONTAINER_NAME},
                            vol_size => $vvol_ref->{IRON_VVOL_SIZE},
                            spcres   => $vvol_ref->{IRON_VVOL_SPACE_GUARANTEE},
                            purge_existing => $purge,
                        }
                    );
                }
            } ## end foreach ( my $k = 0; $k < $vvol_ref...
        } ## end foreach ( my $j = 0; $j < $vvol_ref...

        eval {
            my ( $new_vols, $new_aggrs )
                = setup_volumes( $filer_host, @CAggr_vol_list );
            $vvol_ref->{IRON_VVOL_CLONE_CLONE_LIST} = $new_vols;
            my @new_list = (@{$vvol_ref->{IRON_VVOL_CLONE_LIST}}, @{$new_vols});
            $vvol_ref->{IRON_VVOL_CLONE_LIST} = \@new_list;
        };
        logWarn("wafliron_populate_volumes failed to create clones: $@\n") if ($@);

        # Update wafliron_populate_volumes return ref with pfs vol info
        #foreach my $vol ( @{$vvol_ref->{IRON_VVOL_PFSVOL_LIST}} ) {
        #    my $status_obj = vol_status( $filer_conn, $vol );
        #    logTrc("wafliron_populate_volumes: status_obj = ". Dumper $status_obj);
        #    my ($total, $used, $avail, $capacity)
        #         = vol_disk_free($filer_conn, path => "/vol/$vol",
        #                   space_type  => DF_SPACE, report_type => UNIT_MB);

        #    $return_ref{$vol}{VOL_NAME}  = $vol;
        #    $return_ref{$vol}{VOL_SIZE}  = $total . 'M';
        #    $return_ref{$vol}{VOL_TYPE}  = 'flexcache';
        #    $return_ref{$vol}{AGGR_NAME} = $vvol_ref->{IRON_VVOL_CONTAINER_NAME};
        #    $return_ref{$vol}{VOL_PARENT_NAME} = $status_obj->parent();
        #    $return_ref{$vol}{VOL_SOURCE} = $pfs_parent{$vol}
        #}

        # Update wafliron_populate_volumes return ref with clone vol info
        foreach my $vol ( @{$vvol_ref->{IRON_VVOL_CLONE_LIST}} ) {
            my $status_obj = vol_status( $filer_conn, $vol );
            logTrc("wafliron_populate_volumes: status_obj = ". Dumper $status_obj);
            my ($total, $used, $avail, $capacity)
                 = vol_disk_free($filer_conn, path => "/vol/$vol",
                           space_type  => DF_SPACE, report_type => UNIT_MB);

            $return_ref{$vol}{VOL_NAME} = $vol;
            $return_ref{$vol}{VOL_SIZE} = $total . 'M';
            $return_ref{$vol}{VOL_TYPE} = 'clone';
            $return_ref{$vol}{AGGR_NAME} = $vvol_ref->{IRON_VVOL_CONTAINER_NAME};
            $return_ref{$vol}{VOL_PARENT_NAME} = $status_obj->parent();

            # Get snap shot name
            my @snap_status = snapshot_status( $filer_conn, $vol );
            my $snap_name = $snap_status[0]->{name};
            logcomment( "wafliron_populate_volumes: snap_name = $snap_name" );

            # Wait for snapshot to complete
            my ( $snap_id, $endtime, $sstatus, $snap_status );
            $endtime = time() + $main::TIMEOUT;
            $sstatus = 'creating';
            while ( $sstatus eq 'creating' ) {
                @snap_status = snapshot_status( $filer_conn, $vol );
                last unless scalar @snap_status;
                foreach my $snap_status ( @snap_status ) {
                    if ( $snap_status->{name} eq $snap_name ) {
                        $sstatus = $snap_status->{status};
                        $snap_id = $snap_status->{snapid};
                        last;
                    }
                }
                if ( $endtime < time() ) {
                    die( "wafliron_populate_volumes: Failed to "
                       . "create snapshot: $snap_name on volume: $vol" );
                }
                Tharn::sleep(5);
            } ## end while

            push( @snaplist, $snap_name ) if defined $snap_name;
            $return_ref{$vol}{SNAP_FILES} = \@snaplist if ( @snaplist );
            $return_ref{$vol}{$snap_name}{SNAP_FILE_ID}   = $snap_id if defined $snap_id;
            $return_ref{$vol}{$snap_name}{SNAP_FILE_NAME} = $snap_name if defined $snap_name;
        } ## end foreach
    } ## end for ( my $i = 0; $i < $iron_vparams...


    # Now Create snapshots under clone vols
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};

        foreach my $vvol_name ( @{$vvol_ref->{IRON_VVOL_CLONE_LIST}} ) {
            eval {
                my $wafl_touch_cmd
                      = "wafl touch -V $vvol_name 73 -private 0 0 99999 0";
                my $filer_path = "/vol/" . $vvol_name . "/";
                my $qtree_path = $filer_path . "qtree";

                # Create qtree
                qtreeCreate( $filer_conn, $qtree_path );
                my $ret_val = filer_cli_advanced_say( $filer_conn,
                           "ls $qtree_path" ) if ( $@ =~ /qtree exists/ );
                my @files = split( /\n/, $ret_val );

                my (@test_file_list, @snap_file_list, $loop_count);
                if ( defined $vvol_ref->{IRON_VVOL_NUMOF_SNAPS}
                         && $vvol_ref->{IRON_VVOL_NUMOF_SNAPS} ) {
                    $loop_count = $vvol_ref->{IRON_VVOL_NUMOF_SNAPS};
                } else {
                    $loop_count = 1;
                }

                for ( my $i = 0; $i < $loop_count; $i++ ) {
                    my @snap_test_files = ();

                    # Need to discuss this more. Is it right to comment & create snapshot?
                    if ( $vvol_ref->{IRON_NEW_CLONE_TESTFILES} ) {
                        my ( @subtests, $filer_io_cmd, $fbsize );
                        for ( my $j = 0; $j < $files_per_snap; $j++ ) {
                            my $file_name = $vvol_ref->{IRON_VVOL_FILE_NAME_PREFIX}
                                                 . "_$vvol_name" . "_file$i" . "$j";
                            my $file_path = $qtree_path . "/" . $file_name;

                            if ( $vvol_ref->{IRON_FILE_CREATE_METHOD}
                                                         eq "filersio" ) {
                                # Write file using filersio
                                $filer_io_cmd
                                    = "priv set -q test; filersio asyncio_active "
                                      . "0 -r 0 $blk_size_bytes 0 $file_size "
                                      . "$main::IRON_FILERSIO_RUNTIME 20 "
                                      . "$file_path -create -synchronous -fillonce 65";
                            }
                            elsif ( $vvol_ref->{IRON_FILE_CREATE_METHOD}
                                                            eq "mkfile" ) {
                                # Write file using mkfile
                                if ( $vvol_ref->{IRON_USE_VBN_ZERO} ) {
                                    $filer_io_cmd
                                          = "priv set -q test; mkfile -z $file_size $file_path";
                                } else {
                                    $filer_io_cmd
                                          = "priv set -q test; mkfile $file_size $file_path";
                                }
                            }
                            else {
                                croak("wafliron_populate_volumes: file create tool"
                                   . " should be either mkfile or filersio.");
                            }

                            push @subtests, subtest( sub{ _write_file($filer_conn,
                                 $filer_io_cmd, $file_path, $vvol_ref); }, -bg, -runid,
                                 "$file_name", '-iffatal', '-ok', '-ifscript', '-ok' );

                            push (@test_file_list, $file_name);
                            $return_ref{$vvol_name}{$file_name}{TEST_FILE_NAME} = $file_name;
                            $return_ref{$vvol_name}{$file_name}{TEST_FILE_PATH} = $file_path;
                            $return_ref{$vvol_name}{$file_name}{TEST_FILE_SIZE} = $file_size;
                            push (@snap_test_files, $file_name) if $vvol_ref->{IRON_VVOL_NUMOF_SNAPS};

                            if (($filer_io_cmd =~ /mkfile/ && ($j % 8) eq 7)
                                  || ($filer_io_cmd =~ /filersio/ && ($j % 60) eq 59)) {
                                Subtest::wait_finish(subtests => [@subtests]);
                            }
                        }

                        # Wait for subtests
                        logcomment("Waiting for subtest to get finish....");
                        Subtest::wait_finish(subtests => [@subtests]);
                    } ## end of if


                    if ( defined $vvol_ref->{IRON_VVOL_NUMOF_SNAPS}
                             && $vvol_ref->{IRON_VVOL_NUMOF_SNAPS} ) {
                        # Create snapshot
                        my $snap_name = $vvol_ref->{IRON_VVOL_SNAP_NAME_PREFIX}
                                        . "_$vvol_name" . "_snap$i";

                        say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                        say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

                        snapshot_create( $filer_conn, $vvol_name, $snap_name );
                        push (@snap_file_list, $snap_name);

                        say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                        say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

                        # Update return ref
                        $return_ref{$vvol_name}{$snap_name}{SNAP_FILE_NAME}
                                                                   = $snap_name;
                        $return_ref{$vvol_name}{$snap_name}{SNAP_TEST_FILES}
                                                            = \@snap_test_files;
                        my @snap_status = snapshot_status($filer_conn, $vvol_name);
                        foreach my $snap_status ( @snap_status ) {
                            if ($snap_status->{name} eq $snap_name) {
                                $return_ref{$vvol_name}{$snap_name}{SNAP_FILE_ID}
                                        = $snap_status->{snapid};
                                last;
                            }
                        }
                    }
                }

                $return_ref{$vvol_name}{TEST_FILES} = \@test_file_list;
                $return_ref{$vvol_name}{SNAP_FILES} = \@snap_file_list;
            };
            croak("wafliron_populate_volumes: $@\n") if ($@);
        } # end of foreach loop
    } # end of snap create for clone vols

    # vol size
    SayWrap::sayEOF( $filer_conn, "priv set -q diag; df -h" );

    # Create & Delete VVOL SNAPSHOT on clone vol
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};

        foreach my $vol_name ( @{$vvol_ref->{IRON_VVOL_CLONE_LIST}} ) {
            my $wafl_touch_cmd
                = "wafl touch -V $vol_name 73 -private 0 0 99999 0";
            if ($iron_vparams->{IRON_VVOL_SNAP_CREATE_PREIRON} ||
                  $iron_vparams->{IRON_VVOL_SNAP_DELETE_PREIRON} ) {
                logcomment(
                   "wafliron_populate_volumes: Creating snapshot on vol $vol_name..."
                );
                my $snap_name = $vol_name . "_snapp.$$";
                snapshot_create( $filer_conn, $vol_name, $snap_name );

                say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush

                if ( $iron_vparams->{IRON_VVOL_SNAP_DELETE_PREIRON} ) {
                    logcomment("wafliron_populate_volumes: Deleting snapshot on vol $vol_name...");
                    snapshot_delete( $filer_conn, $vol_name, $snap_name );
                    say( $filer_conn, $wafl_touch_cmd, -noprompt, -timeout, 200 ); # Do a touch
                    say( $filer_conn, $wafl_flush_cmd, -noprompt, -timeout, 200 ); # flush
                }
            }
        }
    } # end of Create & Delete snapshot

    # Install snaprestore license if neeed
    my $install_snaprestore_licese_flag = 1;

    # Do snap restore if requested on regular vols
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};

        if ( $vvol_ref->{IRON_VVOL_SNAP_RESTORE} ) {
            # Install licesne
            if ( $install_snaprestore_licese_flag ) {
                logcomment("Install snaprestore license...");
                license( $filer_conn, 'snaprestore') or
                  die("Unable to Install the 'Snap Restore' License : $@");
                $install_snaprestore_licese_flag = 0;
            }

            foreach my $vvol_name ( @{$vvol_ref->{IRON_VVOL_LIST}} ) {
                # Do volume clone split in case we have clone
                my @cvol_list = Volume::Clone::get_clones(
                                      conn => $filer_conn,
                                      vol  => $vvol_name );
                if ( scalar @cvol_list ) {
                    foreach my $cvol ( @cvol_list ) {
                        logcomment("Performing $cvol split...");
                        Volume::Clone::vol_clone_split_start (
                                              conn             => $filer_conn,
                                              clone            => $cvol,
                                              wait_to_complete => 1,
                                              validate         => 1 )
                               or die("Error: Split operation failed.");
                    }
                }
                my $vol_ref = $return_ref{$vvol_name}; # vol details from return ref
                my @snap_list = @{$vol_ref->{SNAP_FILES}};
                if ( defined $snap_list[0] )
                {
                    logcomment("snap restore $vvol_name to $snap_list[0]...");
                    snaprestore_volume( $filer_conn, $vvol_name, $snap_list[0] )
                       or die("Could not snap restore $vvol_name to $snap_list[0]");
                }

                # Update retrun ref after restore
                my @new_snap_list = snapshot_list( $filer_conn, $vvol_name );
                my @new_list = ();
                foreach my $snap ( @snap_list ) {
                    if ( grep(/$snap/, @new_snap_list) )
                    {
                        push (@new_list, $snap);
                    } else {
                        # Cleanup retrun ref - non existing snaps
                        delete $return_ref{$vvol_name}->{$snap};
                    }
                }
                $return_ref{$vvol_name}->{SNAP_FILES} = \@new_list;
            }
        }
    }

    # Do snap restore if requested on clone vols
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};

        if ( $vvol_ref->{IRON_VVOL_SNAP_RESTORE} ) {
            # Install licesne
            if ( $install_snaprestore_licese_flag ) {
                logcomment("Install snaprestore license...");
                license( $filer_conn, 'snaprestore') or
                  die("Unable to Install the 'Snap Restore' License : $@");
                $install_snaprestore_licese_flag = 0;
            }

            foreach my $vvol_name ( @{$vvol_ref->{IRON_VVOL_CLONE_LIST}} ) {
                my $vol_ref = $return_ref{$vvol_name}; # vol details from return ref
                my @snap_list = @{$vol_ref->{SNAP_FILES}};
                my @new_slist = snapshot_list($filer_conn, $vvol_name);
                if ( (scalar @new_slist) && (defined $snap_list[0]) &&
                      grep (/$snap_list[0]/, @new_slist) )
                {
                    logcomment("snap restore $vvol_name to $snap_list[0]...");
                    snaprestore_volume( $filer_conn, $vvol_name, $snap_list[0] )
                       or die("Could not snap restore $vvol_name to $snap_list[0]");
                }

                # Update retrun ref after restore
                my @new_snap_list = snapshot_list( $filer_conn, $vvol_name );
                my @new_list = ();
                foreach my $snap ( @snap_list ) {
                    if ( grep(/$snap/, @new_snap_list) ) {
                        push (@new_list, $snap);
                    } else {
                        # Cleanup retrun ref - non existing snaps
                        delete $return_ref{$vvol_name}->{$snap};
                    }
                }
                $return_ref{$vvol_name}->{SNAP_FILES} = \@new_list;
            }
        }
    }

    my $fconn = NOTE::Filer->connect( host => $main::FILER );
    my @status_list = ("IDLE", "NOSTATUS");
    for ( my $i = 0; $i < $iron_vparams->{IRON_NUMOF_VVOL_SETS}; $i++ ) {
        my $vvol_ref = $iron_vparams->{"VVOL_SET$i"};

        foreach my $vol_name ( @{$vvol_ref->{IRON_VVOL_LIST}},
                            @{$vvol_ref->{IRON_VVOL_CLONE_LIST}} ) {
            # Give pref to compression first
            if ( $vvol_ref->{IRON_START_COMPRESSION_SCANNER} &&
                        $vvol_ref->{IRON_VVOL_COMPRESSION} ) {
                logcomment("wafliron_populate_volumes: "
                      . "Start compression scanner on $vol_name...");
                eval {
                    # Wait for IDLE status before compression
                    dense_wait_for_status( host => $main::FILER,
                                       path_name => "/vol/$vol_name",
                                       status => \@status_list,
                                       time_gap => '30',
                                       time_out => $main::TIMEOUT ) or die $@;

                    my $comp_obj = Compression::CompressionScanner->new (
                                                    conn   => $fconn,
                                                    volume => $vol_name );
                    $comp_obj->start( volume   => $vol_name,
                                      expected => C_RC_SUCCESS );
                };
                logDbg("wafliron_populate_volumes CompressionScanner error: $@") if $@;
            }

            # de-compress on request
            if ( $vvol_ref->{IRON_START_DECOMPRESSION_SCANNER} &&
                        $vvol_ref->{IRON_VVOL_COMPRESSION} ) {
                logcomment("wafliron_populate_volumes: "
                       . "Start decompress scanner on $vol_name...");
                eval {
                    # Wait for IDLE status before decompress
                    dense_wait_for_status( host => $main::FILER,
                                       path_name => "/vol/$vol_name",
                                       status => \@status_list,
                                       time_gap => '30',
                                       time_out => $main::TIMEOUT ) or die $@;

                    my $comp_obj = Compression::Decompress->new (
                                                    conn   => $fconn,
                                                    volume => $vol_name );
                    $comp_obj->start( volume   => $vol_name,
                                      expected => C_RC_SUCCESS );
                };
                logWarn("Decompress error: $@") if $@;
            }

            # SIS scanner (TODO: Make it run parallel)
            if ( $vvol_ref->{IRON_START_SIS_SCANNER} ) {
                logcomment("wafliron_populate_volumes: "
                      . "Start volume sis scanner on $vol_name...");
                eval {
                    logcomment("wafliron_populate_volumes: sis on the volume");
                    dense_toggle( host      => $main::FILER,
                                  path_name => "/vol/$vol_name",
                                  state     => "on" );

                    logcomment("wafliron_populate_volumes: do sis start");
                    #my $block = 1 if ( defined $vvol_ref->{IRON_SIS_INCLUDE_TRAPPED_BLOCKS}
                    #                && $vvol_ref->{IRON_SIS_INCLUDE_TRAPPED_BLOCKS} );
                    dense_start( host      => $main::FILER,
                                 path_name => "/vol/$vol_name",
                                 #block     => $block,
                                 scan      => 1 );

                    logcomment("wafliron_populate_volumes: wait for sis complete");
                    dense_wait_for_status( host  => $main::FILER,
                                       path_name => "/vol/$vol_name",
                                       status    => \@status_list,
                                       #status    => "IDLE",
                                       time_gap => '30',
                                       time_out  => $main::TIMEOUT ) or die $@;
                };
                logWarn("wafliron_populate_volumes dense_start error: $@") if $@;
            }

            if ( $vvol_ref->{IRON_START_SISUNDO_SCANNER} ) {
                logcomment("wafliron_populate_volumes: "
                    . "Start volume sis undo scanner on $vol_name...");
                eval {
                    #logcomment("wafliron_populate_volumes: sis stop & undo on volume");
                    logcomment("wafliron_populate_volumes: sis stop on volume");
                    dense_stop( host      => $main::FILER,
                                path_name => "/vol/$vol_name" );
                    #dense_undo( host => $main::FILER,
                    #            path_name => "/vol/$vol_name" );

                    logcomment("wafliron_populate_volumes: wait for sis complete");
                    dense_wait_for_status( host => $main::FILER,
                                       path_name => "/vol/$vol_name",
                                       #status => "NOSTATUS",
                                       status    => \@status_list,
                                       time_gap => '30',
                                       time_out => $main::TIMEOUT ) or die $@;
                };
                logWarn("wafliron_populate_volumes dense_stop error: $@") if $@;
            }
        }
    }

    logTrc( "wafliron_populate_volumes: return iron_vparams = \n"
            . Dumper $iron_vparams);
    logTrc(
        "wafliron_populate_volumes: return_ref = \n" . Dumper \%return_ref );

    return \%return_ref;
} ## end sub wafliron_populate_volumes($$)


sub _run_sio_all_files($$) {
    my ($filer_conn, $vol) = (@_);
    my $filer = host( $main::FILER );
    $filer_conn = connect( '#rsh', -timeout, $main::TIMEOUT );
    my $qtree_name = "qtree";
    my $dir_path = "/vol/" . $vol . "/" . $qtree_name;
    logcomment("_run_sio_all_files: run sio...");

    while (1) {
        my (undef, undef, $files)
            = say( $filer_conn, "priv set test; ls $dir_path" );
        my @file_list = split /\n/,$files;
        my $file_size = "10k";
        foreach my $file (@file_list){
            logcomment("Writing to file $file...");
            my $file_path = $dir_path."/".$file;
            my $filer_sio = "priv set test; filersio asyncio_active "
                     . "0 -r 100 4k 0 $file_size "
                     . "$main::IRON_FILERSIO_RUNTIME 10 $file_path";
            filer_cli_test_say( $filer_conn, $filer_sio );
        }
    }
    disconnect($filer_conn, '-ok');
}

sub _read_iron_status_files($$) {
    my ($filer_conn, $vol_hosting_status_files) = (@_);
    my $filer = host( $main::FILER );
    $filer_conn = connect( '#rsh', -timeout, $main::TIMEOUT );
    my $iron_log_dir = "/vol/$vol_hosting_status_files/metadir/irondir/";
    my $filer_host = $filer_conn->hostp();
    logcomment("_read_iron_status_files: read iron status files...");

    my $ret_val = misc_setflag( host  => $filer_host,
                                flag  => 'wafl_metadata_visible',
                                value => 1 );

    if ( !($ret_val) ) {
        $@ = "Unable to set wafl_metadata_visible";
        disconnect($filer_conn, '-ok');
        return 0;
    }

    while( 1 ) {
        eval {
            $ret_val = wafliron_logfiles_present( $filer,
                          $filer_conn, $vol_hosting_status_files );
            if ( $ret_val ) {
                logcomment("wafliron log files are present.");
            } else {
                logcomment("wafliron log files are not present.");
            }
            eval { $ret_val = filer_cli_advanced_say( $filer_conn,
                                              "ls $iron_log_dir" ); };

            logcomment("list $iron_log_dir results\n" . Dumper $ret_val);
        };
    }
}

sub _wafl_flush_cache($$) {
    logcomment("_wafl_flush_cache: satrt flush cache in the back ground...");
    my ($filer_conn, $sleep) = (@_);
    my $filer = host( $main::FILER );
    $filer_conn = connect( '#rsh', -timeout, $main::TIMEOUT );
    while (1) {
        filer_cli_test_say($filer_conn, "priv set -q diag; wafl flush cache");
        logcomment("Sleeping for $sleep sec...");
        Tharn::sleep($sleep);
    }
    disconnect($filer_conn, '-ok');
}

sub _wafl_poll_wafliron_status($$) {
    logcomment("_wafl_poll_wafliron_status: poll wafl iron status in the back ground...");
    my ($filer_conn, $sleep) = (@_);
    my $filer = host( $main::FILER );
    $filer_conn = connect( '#rsh', -timeout, $main::TIMEOUT );
    while (1) {
        filer_cli_test_say($filer_conn, "priv set -q test; aggr wafliron status");
        logcomment("Sleeping for $sleep sec...");
        Tharn::sleep($sleep);
    }
    disconnect($filer_conn, '-ok');
}

sub _wafl_scan_kireeti($$) {
    logcomment("_wafl_scan_kireeti: try to start kireeti in the back ground...");
    my ($filer_conn, $container_name) = (@_);
    my $filer = host( $main::FILER );
    $filer_conn = connect( '#rsh', -timeout, $main::TIMEOUT );
    my $stat_obj = aggr_status( $filer_conn, $container_name );
    my $vols = $stat_obj->{volumes};
    while (1) {
        foreach my $vol ( @$vols ) {
            filer_cli_test_say($filer_conn, "priv set -q test; wafl scan kireeti $vol");
        }
        logcomment("Sleeping for 60 sec...");
        Tharn::sleep(60);
    }
    disconnect($filer_conn, '-ok');
}

sub _client_load_vol($$$) {
    my ($filer_conn, $vol, $vol_status) = (@_);
    my ($filer, $nfs_hamid);
    $filer = host( $main::FILER );
    $filer_conn = connect( '#rsh', -timeout, $main::TIMEOUT );
    logcomment("_client_load_vol: Running traffic on vol $vol");
    logcomment("_client_load_vol: Volume status: " . Dumper $vol_status);

    # Skip if vol is not online
    my $retry_start = 1;
    while ( $retry_start ) {
        if ( $vol_status->state() =~ /online/i ) {
            eval {
                $nfs_hamid = nfs_hammer_start (
                          fc             => $filer_conn,
                          client         => $main::CLIENT,
                          verify_traffic => 'yes',
                          mode           => 'mixed',
                          hammer_time    => -1,
                          vol_name       => $vol );
            };
            if ($@) {
                logWarn("_client_load_vol: Failed to start hammer - $@");
                Tharn::sleep 120;
            } else {
                $retry_start = 0;
            }
        }
    }
    disconnect($filer_conn, '-ok');
}

sub _write_file($) {
    my ($filer_conn, $filer_io_cmd, $file_path, $vvol_ref) = @_;
    my $create_file = 0;
    my $filer = host( $main::FILER );
    $filer_conn = connect( '#rsh', -timeout, $main::TIMEOUT );

    my $out = SayWrap::sayEOF( $filer_conn, "priv set -q test; slist $file_path" );
    unless ( $out =~ /Invalid/i ) {
        unless ( $vvol_ref->{IRON_VVOL_TESTFILE_REUSE} ) {
            SayWrap::sayEOF( $filer_conn, "priv set -q test; rm $file_path" );
            eval { filer_sync( $filer_conn ); };
            $create_file++;
        }
    } else {
        $create_file++;
    }

    SayWrap::sayEOF( $filer_conn, "priv set -q diag; df -h" );
    $out = SayWrap::sayEOF( $filer_conn, $filer_io_cmd,
                TIMEOUT => LOCAL_WAFLIRON_TIMEOUT ) if $create_file;

    unless ( $out =~ /file creation failed/ ) {
        if ( $vvol_ref->{IRON_VVOL_SIS_CLONES} ) {
            my $sis_clone = $file_path . "sisclone";
            my $clone_id = dense_sis_clone_start( conn   => $filer_conn,
                                                  source => $file_path,
                                                  clone  => $sis_clone,
                                                 );

            dense_sis_clone_wait_for_finish( conn   => $filer_conn,
                                             vol    => $sis_clone,
                                             id     => $clone_id,
            ) unless $clone_id
                =~ /file does not exist|Clone file exists|SIS clone not supported/i;
        }
    } else {
        logcomment("_write_file: Failed to write file with $out");
        disconnect($filer_conn, '-ok');
        return 1;
    }
    disconnect($filer_conn, '-ok');
}


sub _iron_wait($$$) {
    my ($filer_conn, $iron_aggr_or_vol, $iron_flags_methods_ref) = @_;
    logcomment( "_iron_wait: sleeping for max "
                . "$iron_flags_methods_ref->{IRON_RECOVERY_WAIT}..." );

    my $node   = NACL::STask::Node->new(hostrec => $filer_conn->hostp());

    my $wafliron_running = 0;
    my $endtime  = time() + $iron_flags_methods_ref->{IRON_RECOVERY_WAIT};
    while ( time() < $endtime ) {
        # See wafl scan status
        say($filer_conn, "priv set -q advanced; wafl scan status ");

        # If iron is running, just wait
        $wafliron_running
              = wafliron_status( $filer_conn, $iron_aggr_or_vol );
        last unless ( $wafliron_running == 1 );
        Tharn::sleep 120;
    }

    my $tailmsg = NACL::MTask::SystemLogDetector->new(
                    command_interface => $node,
                    set_log_file      => ['messages'] );
    $tailmsg->start();

    my $try_again = 0;
    $endtime  = time() + $iron_flags_methods_ref->{IRON_RECOVERY_WAIT};
    while ( time() < $endtime && $try_again < 5 ) {
        Tharn::sleep 120;
        my $lines .= $tailmsg->get_logs( skip_cached_logs => 1 );
        logDbg( "_iron_wait: log messages "
                             . "for last 120sec = \n". Dumper $lines );
        unless ( $lines =~ m/info.error|WAFLIRON/ ) {
            $try_again++;
            last if $try_again > 5;
        }
    }
}

sub _mount_root_vol($) {
    my ( $fc ) = @_;
    logDbg( "_mount_root_vol: Mounting root volume to client" );

    my $node   = NACL::STask::Node->new(hostrec => $fc->hostp());
    my $client = NACL::C::Client->find();

    # Create root volume object
    my $cur_root_vol     = vol_root($fc);
    my $cur_root_vol_obj = NACL::C::Volume->find(
                                command_interface => $node,
                                filter => {
                                  volume  => $cur_root_vol,
                                });

    # Create root vol mount object
    return NACL::STask::Client::Mount->create(
                    command_interface        => $client,
                    server_command_interface => $node,
                    volume                   => $cur_root_vol_obj,
                    chmod                    => '777' );
}

1;
