# Hey emacs, this is basically -*- perl -*-
#$Id$
#
#Copyright (c) 2002 Network Appliance, Inc.
#All rights reserved.
#
## @summary Library of functions for basic filer commands
## @description
## This library provides three types of functions for interfacing
## with NetAPp filers or Netcaches
## 1 ) Basic functions to provide a robust interface
## for executing commands under a specified privilege mode
## 2) A function, filer_cli_error_check, for invoking commands
## and testing for their expected return code and output text.
## This function is the one recommended to verify
## expected error message text from the CLI.
## This function is expected to PASS or FAIL
## 3) Functions to return status information
## These functions may issue FAIL errors if they detect filer problems
## 4) Functions to get a filer to a requested state.
## These functions will issue FATAL errors via ExitError if and only if
## the requested state is not present by the time the function ends.
## Thus,
## a) function calls to delete something that does not exist
## are not FATAL.
## b) function calls to create something that already exists
## are not FATAL.
##
## Specific support for Full Sail has been added.
## DO NOT copy this file back to earlier Perforce branches.
##
## All errors not associated with a failed state change
## are announced by a call to ExitSoftError.
##
## All calls to these functions that do not specify valid parameters
## will result in a call to ExitScriptError that also terminates the test script.
## @usage
##   use filer ;
## @keywords CLI, say, uptime, sysstat
## @dependencies Filer:
##   use config_base;
##   use Tharn;
## @author dl-central-libs
## @burt 68700 intermittent failure of rsh to return anything
## @burt 74433 DAFS and FCP headings needs alignment for sysstat -x output
## @burt 92338 partial rsh output, especially sysconfig
## @burt 93884 console response has: 'INTERRUPT PRINTF:serialhandler: COM 0: wei## rd interrupt: 0x5
## @burt 103863 'vfiler run' loses result code from unsuccessful commands
## @status public
## @change 2002-03-11 jhayden : created
## @change 2003-01-06 hedstrom: add filer_reboot filer_vol_snapdir_disable filer_vol_snapdir_enable
## @change 2003-01-13 hedstrom: add filer_file_copy
## @change 2003-01-14 hedstrom: fix filer_service_is_started for DAFS
## @change 2003-01-15 hedstrom: fix filer_service_license for DAFS
## @change 2003-01-18 hedstrom: implement new space reservation syntax
## @change 2003-02-28 hedstrom: allow options that are set 'off' to be '' or 'off'
## @change 2003-03-04 hedstrom: license codes in hash with gateway filer support
## @change 2003-03-06 hedstrom: filer_power functions with RTP and Sunnyvale support
## @change 2003-03-17 hedstrom: improve ability to find result in filer_cli_say stderr and stdout
## @change 2003-03-23 hedstrom: add filer_model_get, skip license code lookup is code was given
## @change 2003-03-24 jlatta: changed WORM references in license hash tables to snaplock and added licenses
## @change 2003-03-24 jlatta: added filer_vol_online to export list.
## @change 2003-03-24 hedstrom: add add verification to power off operations
## @change 2003-03-30 hedstrom: add comment to report license status
## @change 2003-03-31 jlatta: fixed regexp problems in filer_vol_create, filer_vol_is_offline, and filer_vol_is_online
## @change 2003-04-10 hedstrom: look for new quotas still initializing message
## @change 2003-04-11 hedstrom: recognize Mellanox cluster interconnect
## @change 2003-04-22 jlatta: added snaplock license for NearStore
## @change 2003-05-16 hedstrom: fix filer_cluster_enable function
## @change 2003-06-03 hedstrom: fix filer_vol_snap_exists and filer_vol_snap_is_busy to allow special characters in snapshot names
## @change 2003-07-08 hedstrom: change vfiler service name to multistore
## @change 2003-07-09 hedstrom: change filer_vol_create to no longer wait for 'normal' to see vol was created
## @change 2003-07-31 hedstrom: detect filer is not alive when commands have no response
## @change 2003-08-10 hedstrom: support rsh-socket that does not return \r on output lines
## @change 2003-08-20 hedstrom: use code from lib/nfs/rsh_lib.pm for filer_quotas_file_append
## @change 2003-09-03 hedstrom: add connection type to filer_cli_say error messages
## @change 2003-09-04 hedstrom: add filer_cli_default_answer_set, cope with some prompts within filer_cli_say
## @change 2003-09-05 hedstrom: make FILER_CLI_SAY_CHOP variables with param interface
## @change 2003-09-13 hedstrom: fix filer_cli_say bug that could drop first command line output
## if it included the original input command (some message prefix values did this)
## @change 2003-09-15 vadim: got rid of CHOP globals and made them arguments to
## filer_cli_say() since its the only function that needs them.
## @change 2003-09-23 hedstrom: vfilerize quota and df commands from here
## @change 2003-12-03 hedstrom: add applies_to_filer_with_debug_kernel applies_to_filer_with_dns_enabled
##                              filer_cli_clean filer_dns_is_enabled filer_nfs_initialize
##                              filer_vol_quota_is_enabled filer_vol_wait_until_online
## @change 2004-01-06 hedstrom: retry any filer commands that fail because filer was out of rsh threads
## @change 2004-01-29 hedstrom: burt 109997 - implement interface to allow filer_cli_say requests to retry only when a repeated command would not fail after the first invocation.
## @change 2004-06-11 hedstrom: verify giveback finishes in filer_cluster_giveback
## @change 2004-08-10 hedstrom: burt138619 return PASS/FAIL status, use logresult(PASS) within filer_cli_error_check
## @change 2004-08-27 hedstrom: update filer_sysstat_x_ops_get to return a hash of results when asked
## @change 2004-08-30  boppana: Added the functions for getting snapshot space and wafl scan information:
##                              filer_vol_totalspace_get,filer_vol_usedspace_get,filer_vol_snap_totalspace_get,filer_vol_snap_freespace_get,filer_vol_snap_usedspace_get,
##                              filer_vol_spaceinfo_get,filer_vol_snap_waflscan_complete,filer_vol_snap_reclaimspace_get. Modified filer_vol_freespace_get
## @change 2004-08-30 hedstrom: no longer issue FATAL error when requested state exists to begin with
## @change 2004-09-10 hedstrom: support expected timeout within filer_cli_say functions
## @change 2004-10-04 sli: modify filer_vol_exists to handle "vol status: Volume 'xx' is temporarily busy" situation
## @change 2004-12-13 hedstrom: extend filer_vol_snap_create wait time, revise giveback support for easier maintenance
## @change 2005-01-05 hedstrom: remove duplicate functions, alphabetize others
## @change 2005-01-05 hedstrom: add filer_vols_get
## @change 2005-01-05 hedstrom: alphabetize the functions
## @change 2005-01-21 hedstrom: support simulator use in filer_version_get
## @change 2005-01-25 hedstrom: report burt68700 recovered errors during POSTTEST or POSTJOB as warnings
## @change 2005-01-28 hedstrom: increase timeout for filer_vol_snap_create and allow an override
## @change 2005-01-29 hedstrom: filer_version_get was not supposed to try printenv from simulators
## @change 2005-01-29 hedstrom: report all burt68700 recovered errors as warnings
## @change 2005-02-01 hedstrom: burt151665 handle long volume names
## @change 2005-02-10 sli: add filer_ping_seq_time_get
## @change 2005-02-10 hedstrom: support volume search by attributes that are not present
## @change 2005-02-16 hedstrom: add filer_file_append, filer_ip_ping
## @change 2005-02-23 hedstrom: support netcache that does not have 'echo' command
## @change 2005-03-01 hedstrom: use NATE::Result
## @change 2005-03-08 hedstrom: add functions to support revert and its required preparations
## @change 2005-03-17 hedstrom: fix search string in filer_vol_snap_exists & filer_vol_snap_is_busy
## @change 2005-03-18 hedstrom: add filer_vol_option_set filer_vol_options_get
## @change 2005-03-23 corbin: add functions filer_netgroup_add cleanup on filer_if_netmask_get
## @change 2005-04-04 hedstrom: remove burt tracking of 68700 - too much for crusher
## @change 2005-04-05 corbin: add functions filer_access_cache_flush, filer_access_cache_get, filer_actual_storage_get, filer_client_cache.
## @change 2005-06-01 archana: Updated author tag
## @change 2005-06-22 hedstrom: filer_options_set change on verifying passwords
## @change 2005-06-29 corbin: add functions filer_lock_status_file_get and filer_volume_dump
## @change 2005-07-11 hedstrom: filer_options_set fix verifying passwords
## @change 2005-07-20 korry: add longer timeout for filer_cli_say call in filer_cli_error_check
## @change 2005-08-05 corbin: burt169522: added retry capability to filer_access_cache_get
## @change 2005-08-19 hedstrom: add filer_name function to extract filer host name from connection
## @change 2005-09-08 corbin: add function filer_volume_restore
## @change 2005-09-15 korry: add longer timeout for filer_cli_say call in filer_vol_snap_delete
## @change 2005-10-05 corbin: add functions filer_export_fence_disable filer_export_fence_enable
## @change 2005-10-17 corbin: export function filer_get_qtree and have it return qtree volume information and updated filer_lock_status_file_get
## @change 2005-11-11 corbin: add functions filer_cifs_access filer_cifs_shares_get and fix filer_netgroup_add
## @change 2006-01-05 corbin: add function filer_resource_agitation_set
## @change 2006-01-08 hedstrom: allow filer_options_set to skip verification (burt169732)
## @change 2006-01-08 hedstrom: replace FSR::FSRutils regx_escape with quotemeta
## @change 2006-02-03 corbin: add function filer_vol_snap_sched_get
## @change 2006-03-01 korry: remove DAFS in 7.1.0.1 and GB.
## @change 2006-03-01 korry: add status option to deprecated 'quota'.
## @change 2006-05-13 corbin: add filer_if_mtu_set
## @change 2006-08-29 corbin: add filer_cifs_homedir filer_cifs_load_homedir filer_cifs_lookup filer_cifs_sessions filer_cifs_share_change filer_domainuser_delete
## @change 2006-09-29 corbin: add filer_ipfw_disable filer_ipfw_enable filer_ipfw_rule_add filer_ipfw_rule_delete filer_flexcache_delegations filer_flexcache_eject
## @change 2006-10-30 corbin: add function filer_netstat_get
## @change 2006-11-15 korry: burt208611 add filer_cluster_is_in_unsync and filer_cluster_takeover to handle take/give unsynchonized log state.
## @change 2007-01-15 korry: add -v to filer_version_get() to determine installed kernel.
## @change 2007-05-16 hedstrom: make release agnostic back to Scrimshaw.
## @change 2007-09-18 hedstrom: add disk_sanitization, env_exception, flex_scale,
##                    operations_manager, protection_manager, provisioning_manager,
##                    smdomino, snaplock_enterprise,
##                    snapmanager_oracle, snapmanager_sap, snapmanager_sharepoint,
##                    snapmover, sv_nbu_sec
##                    services
## @change 2007-10-02 corbin: add function filer_stats_vol_show
## @change 2007-11-08 hedstrom: add compression service and promoted_simulator license codes
##                    fix regular expression in filer_machine_type_get
## @change 2007-11-20 korry: using 'result -l' in filer_cli_say() to handle the output result consistently.
## @change 2008-01-17 korry: add 'cf' license to support cluster in BR
## @change 2008-01-29 korry: change the logic in filer_cli_say to parse munched up result output
## @change 2008-03-25 korry: change invalid HASH ref in filer_service_license() and fix outscope hash licenses
## @change 2008-05-08 corbin: add function filer_ifstat_get
## @change 2008-11-03 karthik2: add 'flexcache_nfs' license to support SA appliance
## @change 2009-11-12 sajitha: burt364872: Replaced logcomment() with $Log->debug() to prevent the swelling of log file.
## @change 2012-02-17 giuntoli: burt570126 added parsing to license cleanup code to handle new output of licensing v2
## @change 2013-01-18 amal: burt670298 added cmode supported subroutine filer_cmd()

# When adding new functions, please:
# 1) Maintain the alphabetical order of the functions
# 2) Define new function names that start with filer_ and
# 3) Add additional descriptive nouns to the function name to define the
#  portion of the filer that the new function deals with and
# 4) End the function name with a verb. (Try to reuse the same verbs that appear on other functions)
#    The following verbs  are reserved for functions that can cope with unknown initial states:
#    initialize - completely sets up a service, returns 0 if the service is not available or cannot be started
#    make - create a new object if it does not exist, fail only if the object does not exist at the end
#    vaporize - destroys an object only reporting a failure if the object still exists at the end
# 5) Write functions to meet specific needs of a filer user, rather than to simply expose
# the filer interface to our test scripts. This should allow us to write test scripts from
# an operational perspective, rather than from the low level CLI. This will permit us to
# maintain those operational functions within this package while minimizing changes to
# individual test scripts.
# 6) Check input parameter values. Use ExitScriptError to report out of range values.
# 7) Use ExitError to report the inability to perform a request that would change the filer state.
#    or a serious problem that indicates an unexpected change in filer state
#    (For example, lost of the ability to communicate with the filer)
# 8) Use ExitSoftError to report other unexpected results.

use strict;

package filer;

use Burt;
use config_base;
use Tharn;
use Filer::Version qw(:releases are_we_ontapng get_wafl_version);
use boot qw(boot reboot wait_for_rsh);
use Utils qw(make_conn);
use Data::Dumper;
use Socket;
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate validate_pos SCALAR BOOLEAN ARRAYREF HASHREF OBJECT);

use vars qw(
    @ISA @EXPORT
    $FILER_COMMAND_COUNT
    $FILER_CLI_SAY_TIMEOUT
    %FILER_MODELS
    $RESPONSE $RESPONSE_PROMPT
    %filer_licenses %filer_1day_licenses %filer_90day_licenses
    %cache_licenses %cache_1day_licenses %cache_90day_licenses
    %gateway_licenses %gateway_1day_licenses %gateway_90day_licenses
    %nearstore_licenses %nearstore_1day_licenses %nearstore_90day_licenses
    %promoted_simulator_licenses %deprecated_licenses
    %sa_filer_license %vsa_90day_licenses %promoted_vsa_licenses
    );

require Exporter;
@ISA = ('Exporter');
@EXPORT = qw(
    cmdOkReturn cmdFailedReturn cmdBadUsageReturn cmdBadArgReturn cmdNotFoundReturn cmdDefaultReturn

    arg_vol_remove_slashvol
    applies_to_filer_with_cluster_installed
    applies_to_filer_with_debug_kernel
    applies_to_filer_with_dns_enabled
    filer_access_cache_flush
    filer_access_cache_get
    filer_actual_storage_get
    filer_aggr_snap_delete filer_aggr_snap_delete_all
    filer_aggr_spaceinfo_get
    filer_aggr_totalspace_get
    filer_aggrs_get
    filer_bootargs_get
    filer_bootargs_set
    filer_bootargs_unset
    filer_ciob_revision_get
    filer_cifs_access filer_cifs_homedir filer_cifs_is_started
    filer_cifs_load_homedir filer_cifs_lookup
    filer_cifs_sessions filer_cifs_share_change filer_cifs_shares_get
    filer_cifs_start filer_cifs_stop
    filer_cli_admin_say filer_cli_advanced_say
    filer_cli_clean
    filer_cli_default_answer_set filer_cli_diag_say
    filer_cli_say filer_cli_test_say
    filer_cli_priv
    filer_cli_error_check filer_cli_split
    filer_client_cache
    filer_cluster_disable filer_cluster_enable filer_cluster_is_enabled
    filer_cluster_giveback
    filer_cluster_initialize
    filer_cluster_is_in_takeover
    filer_cluster_is_in_unsync
    filer_cluster_is_installed filer_cluster_is_licensed
    filer_cluster_license
    filer_cluster_restore
    filer_cluster_unlicense
    filer_cluster_takeover
    filer_crosscheck_disable filer_crosscheck_enable filer_crosscheck_is_enabled
    filer_date_advance filer_date_get filer_date_set
    filer_dbg_level_set filer_dbg_to_set
    filer_disk_capacity_get
    filer_dns_is_enabled
    filer_domainuser_delete
    filer_download
    filer_export_fence_disable
    filer_export_fence_enable
    filer_file_append
    filer_file_copy
    filer_file_destroy
    filer_file_exists
    filer_file_make
    filer_file_read
    filer_file_space_reserve_disable filer_file_space_reserve_enable
    filer_flexcache_delegations filer_flexcache_eject
    filer_get_vol_name_from_path
    filer_if_alias filer_if_down
    filer_if_ip
    filer_if_ips_get
    filer_if_mtu_set
    filer_if_status
    filer_if_up
    filer_if_netmask_get
    filer_ifinfo_vi_check
    filer_ifstat_get
    filer_ifstat_vi_check
    filer_ifstat_zero
    filer_ip_ping
    filer_ipfw_disable filer_ipfw_enable
    filer_ipfw_rule_add filer_ipfw_rule_delete
    filer_kernel_is_debug
    filer_lock_status_file_get
    filer_mem_stats_cleanup
    filer_mem_stats_diff
    filer_mem_stats_dump
    filer_mem_stats_save
    filer_model_get
    filer_name
    filer_netgroup_add
    filer_netstat_get
    filer_netstat_if_get
    filer_nfs_initialize filer_nfs_is_started filer_nfs_start filer_nfs_stop
    filer_nvram_fw_revision2_get
    filer_nvram_fw_revision_get
    filer_nvram_revision_get
    filer_options_get filer_options_set
    filer_panic
    filer_ping_seq_time_get
    filer_power_cycle filer_power_off filer_power_on
    filer_processor_count_get
    filer_qtree filer_qtree_count
    filer_qtree_create filer_qtree_delete filer_qtree_exists
    filer_qtree_get
    filer_qtree_oplocks_disable
    filer_qtree_oplocks_enable filer_qtree_oplocks_is_enabled filer_qtree_oplocks_set
    filer_qtree_security_mixed_set
    filer_qtree_security_ntfs_set
    filer_qtree_security_set
    filer_qtree_security_unix_set
    filer_qtree_security_style_get
    filer_qtree_space_reserve_disable filer_qtree_space_reserve_enable
    filer_quota_scramble_conf
    filer_quota_tree_add filer_quota_tree_get filer_quota_tree_report
    filer_quotas_file_append filer_quotas_file_read
    filer_randomize
    filer_reboot
    filer_registry_get
    filer_release_revert
    filer_resource_agitation_set
    filer_serial_number_get
    filer_service_is_licensable filer_service_is_licensed filer_service_is_started
    filer_service_license filer_service_start filer_service_status_get filer_service_stop
    filer_service_unlicense
    filer_services_get
    filer_services_get_licensable
    filer_services_get_licensed
    filer_showfh
    filer_snapmirror_dest_break
    filer_snapmirror_is_started filer_snapmirror_start
    filer_snapmirror_status
    filer_snapmirror_status_all
    filer_snapmirror_stop
    filer_snaprestore_initialize
    filer_software_delete filer_software_exists filer_software_list
    filer_sparecore_disable filer_sparecore_enable filer_sparecore_is_ready filer_sparecore_status_get
    filer_spares_get
    filer_statit_get
    filer_statit_start
    filer_statit_stop
    filer_stats_vol_show
    filer_stats_aggr_show
    filer_sync
    filer_sysconfig_A_check filer_sysconfig_a_check
    filer_sysconfig_c_check filer_sysconfig_v_check
    filer_sysstat_check filer_sysstat_get filer_sysstat_ops_get filer_sysstat_x_ops_get
    filer_systemboard_fw_revision_get
    filer_uptime_check filer_uptime_ops_get
    filer_user_add filer_user_delete filer_user_exists
    filer_version_get
    filer_vfiler_command
    filer_vfiler_context_get filer_vfiler_context_off filer_vfiler_context_set
    filer_vfiler_context_simulate
    filer_vfiler_create filer_vfiler_destroy

    filer_vfilers_get
    filer_vfilers_ips_parse
    filer_vfiler_limit_get filer_vfiler_limit_set
    filer_vfiler_start filer_vfiler_status_get filer_vfiler_stop
    filer_vol_create filer_vol_destroy filer_vol_exists
    filer_vol_file_exists
    filer_vol_freespace_get
    filer_vol_extent_enable
    filer_vol_extent_disable
    filer_vol_extent_size_set
    filer_vol_is_offline
    filer_vol_is_online
    filer_vol_offline
    filer_vol_online
    filer_vol_option_set
    filer_vol_options_get
    filer_vol_quota_disable filer_vol_quota_enable filer_vol_quota_is_enabled
    filer_vol_restrict
    filer_vol_size_get
    filer_vol_snap_create
    filer_vol_snap_delete filer_vol_snap_delete_all
    filer_vol_snap_exists
    filer_vol_snap_freespace_get
    filer_vol_snap_is_busy
    filer_vol_snap_recreate
    filer_vol_snap_rename
    filer_vol_snap_sched_get
    filer_vol_snap_sched_set
    filer_vol_snap_reclaimspace_get
    filer_vol_snap_reserve_set
    filer_vol_snap_restore
    filer_vol_snap_totalspace_get
    filer_vol_snap_usedspace_get
    filer_vol_snap_waflscan_complete
    filer_vol_snapdir_disable filer_vol_snapdir_enable
    filer_vol_space_reserve_disable filer_vol_space_reserve_enable filer_vol_space_reserve_is_enabled
    filer_vol_space_reserve_fill_disable
    filer_vol_space_reserve_fill_enable
    filer_vol_space_reserve_fill_is_enabled
    filer_vol_space_reserve_overwrite_disable
    filer_vol_space_reserve_overwrite_enable
    filer_vol_space_reserve_overwrite_is_enabled
    filer_vol_spaceinfo_get
    filer_vol_totalspace_get
    filer_vol_usedspace_get
    filer_vol_vaporize
    filer_vol_wafliron_stop
    filer_vol_wait_until_online
    filer_vol_wait_until_ready
    filer_vols_get
    filer_volume_dump
    filer_volume_restore
    );

# Data ONTAP command line Return codes
use constant cmdOkReturn => 0;
use constant cmdFailedReturn => 1;
use constant cmdBadUsageReturn => 2;
use constant cmdBadArgReturn => 3;
use constant cmdNotFoundReturn => 127;
use constant cmdDefaultReturn => 255;

use constant NETGROUP_FILE => '/etc/netgroup';

$FILER_COMMAND_COUNT   = 0;

param('FILER_CLI_SAY_TIMEOUT','-default',60);
# Define filer license codes for all filer, cache and nearstore models
# Valid License keys are defined at:
# http://web.netapp.com/engineering/design-depot/appliance-mgmt/license-matrix.html
# Codes are generated for each of these license keys using decrypt command on UNIX
# or see values at: http://wikid.netapp.com/w/HHG/Software-4.0/Procedures-3.0/License_FAQ
%deprecated_licenses = (
    remotesyncmirror => 1,   # slot 30 - Diamond & 6.2 only
    syncmirror_remote => 1,  # slot 30 - 6.3 only
    snapmanager => 1,        # slot 15 - CV, Hummer, Turdle, 6.1.1
    syncmirror => 1,         # slot 17 - pre 6.3
    volcopy => 1,            # slot 9 - reserved
    rtsp => 1                # slot 14 - replaced by real
    );
%cache_licenses = (
    netcache => "GXVPYWA",            # slot 7
    nntp => "WJSRXFF",                # slot 12 - 6.0.x and up
    mms => "EKMAREM",                 # slot 13 - Microsoft Streaming 6.0.x and up
    real => "YMWRXFF",                # slot 14 - Real Streaming 6.0.x and up
    icap => "GFCJHSI",                # slot 19 - and up
    grm => "WRQLGBN",                 # slot 20 - and up
    quicktime => "MOWPYWA",           # slot 21 - Apple QuickTime Streaming
    ncagent => "GJOLDQC",             # slot 22 - GRM Agent
    mms_pro => "",             # slot 23 - Microsoft Media Streaming high speed
    https => "IQWRXFF",
    nfscache => "SZNNCZG",            # slot 29 - 6.4 and up
    real_pro => "",            # slot 33 - and up
    mms_ultra => "",           # slot 34 - and up
    real_ultra => "",          # slot 35 - and up
    'com-e' => "",               # slot 36 - and up
    https_proxy => "",         # slot 37 - and up
    stache => "",               # slot 38 - and up
    mpeg4_isma => "",               # slot 39 - and up
    mpeg4_isma_pro => "",              # slot 39 - and up
    mpeg4_isma_ultra => "",               # slot 39 - and up
    gateway_storage => "",               # slot 42 - 5.4 and up
    streaming_broadcast => "",               # slot 43 - and up
    streaming_archive => ""               # slot 44 - and up
    );
%cache_1day_licenses = (
    netcache => "",            # slot 7
    nntp => "",                # slot 12 - 6.0.x and up
    mms => "",                 # slot 13 - Microsoft Streaming 6.0.x and up
    real => "",                # slot 14 - Real Streaming 6.0.x and up
    icap => "",                # slot 19 - and up
    grm => "",                 # slot 20 - and up
    quicktime => "",           # slot 21 - Apple QuickTime Streaming
    ncagent => "",             # slot 22 - GRM Agent
    mms_pro => "",             # slot 23 - Microsoft Media Streaming high speed
    https => "",
    nfscache => "",            # slot 29 - 6.4 and up
    real_pro => "",            # slot 33 - and up
    mms_ultra => "",           # slot 34 - and up
    real_ultra => "",          # slot 35 - and up
    'com-e' => "",               # slot 36 - and up
    https_proxy => "",         # slot 37 - and up
    stache => "",              # slot 38 - and up
    mpeg4_isma => "",              # slot 39 - and up
    mpeg4_isma_pro => "",              # slot 39 - and up
    mpeg4_isma_ultra => "",              # slot 39 - and up
    gateway_storage => "",            # slot 42 - 5.4 and up
    streaming_broadcast => "",             # slot 43 - and up
    streaming_archive => ""            # slot 44 - and up
    );
%cache_90day_licenses = (
    netcache => "",            # slot 7
    nntp => "",                # slot 12 - 6.0.x and up
    mms => "",                 # slot 13 - Microsoft Streaming 6.0.x and up
    real => "",                # slot 14 - Real Streaming 6.0.x and up
    icap => "",                # slot 19 - and up
    grm => "",                 # slot 20 - and up
    quicktime => "",           # slot 21 - Apple QuickTime Streaming
    ncagent => "",             # slot 22 - GRM Agent
    mms_pro => "",             # slot 23 - Microsoft Media Streaming high speed
    https => "",
    nfscache => "",            # slot 29 - 6.4 and up
    real_pro => "",            # slot 33 - and up
    mms_ultra => "",           # slot 34 - and up
    real_ultra => "",          # slot 35 - and up
    'com-e' => "",               # slot 36 - and up
    https_proxy => "",         # slot 37 - and up
    stache => "",              # slot 38 - and up
    mpeg4_isma => "",              # slot 39 - and up
    mpeg4_isma_pro => "",              # slot 39 - and up
    mpeg4_isma_ultra => "",              # slot 39 - and up
    gateway_storage => "",            # slot 42 - 5.4 and up
    streaming_broadcast => "",             # slot 43 - and up
    streaming_archive => ""            # slot 44 - and up
   );
%filer_licenses = (      # Exclude any NetCache entries, not accepted, even though code can be generated
    nfs => "UYTCNCG",                 # slot 4
    cifs => "GTLYRVH",                # slot 5
    http => "AODUWOJ",                # slot 6
    netcache => "1",                  # slot 7
    cluster => "KDQLGBN",             # slot 8
    cf => "KDQLGBN",                  # clustered failover to replace cluster
    snapmirror => "UUNLDQC",          # slot 10
    snaprestore => "EOYPBIL",         # slot 11
    nntp => "1",                      # slot 12 - 6.0.x and up
    mms => "1",                       # slot 13 - Microsoft Streaming 6.0.x and up
    real => "1",                      # slot 14 - Real Streaming 6.0.x and up
    snapmanagerexchange => "GFMYRVH", # slot 15 - VB.2, 6.2 and up
    dafs => "OBOLDQC",                # slot 16
    syncmirror_local => "UELJHSI",    # slot 17 - 6.3 and up
    multistore => "SPQLGBN",              # slot 18
    vfiler => "SPQLGBN",              # slot 18
    icap => "1",                      # slot 19 - and up
    grm => "1",                       # slot 20 - and up
    quicktime => "1",                 # slot 21 - Apple QuickTime Streaming
    ncagent => "1",                   # slot 22 - GRM Agent
    mms_pro => "1",                   # slot 23 - Microsoft Media Streaming high speed
    vld => "QYOAREM",                 # slot 24 - Virtual Local Disk, VB.2, 6.3 and up
    smsql => "YCTPBIL",               # slot 25 - SnapManager for SQL
    sv_ontap_pri => "KXQPNOJ",        # slot 26 - 6.4 and up
    sv_ontap_sec => "YSPCZIE",        # slot 27 - 6.4 and up
    sv_dr_ontap_pri => "WHRTIVH",     # slot 28 - AnchorSteam (6.6?) and up
    nfscache => "1",                  # slot 29 - 6.4 and up
    fcp => "ICRNCZG",                 # slot 31 - 6.3 and up (or KFPEYRI)
    cluster_remote => "GTFJHSI",      # slot 30 - 6.3 and up  see burt 80994 code=1
    iscsi => "IKVAREM",               # slot 32 - 6.3 and up (or YMUITYG)
    real_pro => "1",                  # slot 33 - and up
    mms_ultra => "1",                 # slot 34 - and up
    real_ultra => "1",                # slot 35 - and up
    'com-e' => "1",                     # slot 36 - and up
    https_proxy => "1",               # slot 37 - and up
    stache => "1",                    # slot 38 - and up
    gateway_storage => "",     # slot 42 - 6.4 and up
    sv_unix_pri => "IUILGBN",         # slot 45 - 6.4 and up
    sv_linux_pri => "GROPYWA",        # slot 46 - 6.4 and up
    sv_windows_pri => "SLGLDQC",      # slot 47 - 6.4 and up
    snaplock => "OXRNCZG",            # slot 48 - Fullsail.1 and up (pre 7.0.0.1 IASLGBN ; after KJHHIJE )
    raindrd => "",              # slot 49 - Anchorsteam and up
    #XXX double check
    # License PFS under its various aliases
    access_on_demand => "WLHHIJE",    # slot 51
    flex_cache => "WLHHIJE",          # slot 51 (old free license)
    flexcache_nfs => "SKWPYWA",       # slot 82 (new pay license)

    snapmover => "KBNNOFF",           # slot 58
    snapmirror_sync => "IEOAOTB",     # slot 59
    sv_windows_ofm_pri => "",         # slot 62
    snapvalidator => "MCGWSMD",       # slot 63
    socks => "",                      # slot 64
    http_compression => "",           # slot 65
    flex_clone => "QNFYRVH",          # slot 66
    rapid_restore => "QBBSXFF",       # slot 67
    advanced_sis => "EJCFMLK",        # slot 68
    nearstore_asis1 => "EJCFMLK",     # slot 68
    nearstore_option => "ERKJHSI",    # slot 69
    nearstore_asis2 => "OSFYRVH",     # slot 70
    a_sis => "OSFYRVH",
    disk_sanitization => "CPUPBIL",
    env_exception => "KUVPBIL",
    flex_scale => "MZGLGBN",
    operations_manager => "QNVRXFF",
    persistent_archive => "WWHLSHL",
    protection_manager => "QVDWSMD",
    provisioning_manager => "IDIJHSI",
    smdomino => "CFZRXFF",
    snaplock_enterprise => "AOXRXFF",
    snapmanager_oracle => "EPTEMLK",
    snapmanager_sap => "KYLAOTB",
    snapmanager_sharepoint => "IFNNCZG",
    sv_nbu_sec => "MJWGIJE",
    wafl_compression => "KVTAOTB",    # See change 648989 (was wafl_compression)
    compression => "KVTAOTB",         # See change 648989 (was wafl_compression)
   );
%filer_1day_licenses = (
    nfs => "BYYHKWH",                 # slot 4
    cifs => "BGHMFDG",                # slot 5
    http => "JOPQAKE",                # slot 6
    netcache => "1",                  # slot 7
    cluster => "VEJZQXA",             # slot 8
    cf => "VEJZQXA",                  # slot 8
    snapmirror => "ZSLZTIL",          # slot 10
    snaprestore => "BCBVVQC",         # slot 11
    nntp => "1",                      # slot 12 - 6.0.x and up
    mms => "1",                       # slot 13 - Microsoft Streaming 6.0.x and up
    real => "1",                      # slot 14 - Real Streaming 6.0.x and up
    snapmanagerexchange => "JGXDPPJ", # slot 15 - VB.2, 6.2 and up
    dafs => "FQPZQXA",                # slot 16
    syncmirror_local => "BWJVYBN",    # slot 17 - 6.3 and up
    multistore => "JESZTIL",          # slot 18
    vfiler => "JESZTIL",          # slot 18
    icap => "1",                      # slot 19 - and up
    grm => "1",                       # slot 20 - and up
    quicktime => "1",                 # slot 21 - Apple QuickTime Streaming
    ncagent => "1",                   # slot 22 - GRM Agent
    mms_pro => "1",                   # slot 23 - Microsoft Media Streaming high speed
    vld => "VUFKGUB",                 # slot 24 - Virtual Local Disk, VB.2, 6.3 and up
    smsql => "VKMZTIL",               # slot 25 - SnapManager for SQL
    sv_ontap_pri  => "ZIIKJFM",       # slot 26 - 6.4 and up
    sv_ontap_sec => "BSXFLND",        # slot 27 - 6.4 and up
    nfscache => "1",                  # slot 29 - 6.4 and up
    cluster_remote => "1",            # slot 30 - 6.3 and up  see burt 80994
    fcp => "PTNSLZG",                 # slot 31 - 6.3 and up
    iscsi => "PBSFAFM",               # slot 32 - 6.3 and up
    real_pro => "1",                  # slot 33 - and up
    mms_ultra => "1",                 # slot 34 - and up
    real_ultra => "1",                # slot 35 - and up
    'com-e' => "1",                   # slot 36 - and up
    https_proxy => "1",               # slot 37 - and up
    stache => "1",                    # slot 38 - and up
    gateway_storage => "",            # slot 42 - 6.4 and up
    sv_unix_pri => "",                # slot 45 - 6.4 and up
    sv_linux_pri => "",               # slot 46 - 6.4 and up
    sv_windows_pri => "",             # slot 47 - 6.4 and up
    snaplock => "OXRNCZG",            # slot 48 - Fullsail.1 and up
    raindrd => ""                     # slot 49 - Anchorsteam and up
    );
%filer_90day_licenses = (
    nfs => "TZBGCKE",                 # slot 4
    cifs => "RGDTQPJ",                # slot 5
    http => "ZOLXLWH",                # slot 6
    netcache => "1",                  # slot 7
    cluster => "JEYOVIL",             # slot 8
    cf => "JEYOVIL",                  # slot 8
    snapmirror => "TVVOSXA",          # slot 10
    snaprestore => "PBQKACN",         # slot 11
    nntp => "1",                      # slot 12 - 6.0.x and up
    mms => "1",                       # slot 13 - Microsoft Streaming 6.0.x and up
    real => "1",                      # slot 14 - Real Streaming 6.0.x and up
    snapmanagerexchange => "BIACHDG", # slot 15 - VB.2, 6.2 and up
    dafs => "TPEPVIL",                # slot 16
    syncmirror_local => "VYTKXQC",    # slot 17 - 6.3 and up
    multistore => "DHCPSXA",          # slot 18
    vfiler => "DHCPSXA",          # slot 18
    icap => "1",                      # slot 19 - and up
    grm => "1",                       # slot 20 - and up
    quicktime => "1",                 # slot 21 - Apple QuickTime Streaming
    ncagent => "1",                   # slot 22 - GRM Agent
    mms_pro => "1",                   # slot 23 - Microsoft Media Streaming high speed
    vld => "JUUZKFM",                 # slot 24 - Virtual Local Disk, VB.2, 6.3 and up
    smsql => "RYYOVIL",               # slot 25 - SnapManager for SQL
    sv_ontap_pri => "TLSZHUB",        # slot 26 - 6.4 and up
    sv_ontap_sec => "TBZOVIL",        # slot 27 - 6.4 and up
    nfscache => "1",                  # slot 29 - 6.4 and up
    cluster_remote => "1",            # slot 30 - 6.3 and up  see burt 80994
    fcp => "FKGEGMK",                 # slot 31 - 6.3 and up
    iscsi => "JUYZHUB",               # slot 32 - 6.3 and up
    real_pro => "1",                  # slot 33 - and up
    mms_ultra => "1",                 # slot 34 - and up
    real_ultra => "1",                # slot 35 - and up
    'com-e' => "1",                   # slot 36 - and up
    https_proxy => "1",               # slot 37 - and up
    stache => "1",                    # slot 38 - and up
    gateway_storage => "",            # slot 42 - 6.4 and up
    sv_unix_pri => "",                # slot 45 - 6.4 and up
    sv_linux_pri => "",               # slot 46 - 6.4 and up
    sv_windows_pri => "",             # slot 47 - 6.4 and up
    snaplock => "OXRNCZG",            # slot 48 - Fullsail.1 and up
    raindrd => "",                    # slot 49 - Anchorsteam and up
    flex_cache => "DNLTQPJ",          # slot 51 (old free license)
    flexcache_nfs => "KGCPSXA",       # slot 82 (new pay license)
    snapvalidator => "FYLVMND",       # slot 63
    rapid_restore => "KCCGRBN",       # slot 67
    disk_sanitization => "FIBNUIL",
    );
%gateway_licenses = (
    nfs => "DXCSYJM",                 # slot 4
    cifs => "DXCQNUJ",                # slot 5
    http => "LFLUIBI",                # slot 6
    netcache => "",                   # slot 7 (or )
    cluster => "XDNJFLF",             # slot 8
    cf => "XDNJFLF",                  # slot 8
    snapmirror => "FMVLPCB",          # slot 10
    snaprestore => "BSPHXGN",         # slot 11
    nntp => "",                       # slot 12 - 6.0.x and up
    mms => "GJUCNCG",                 # slot 13 - Microsoft Streaming 6.0.x and up
    real => "ORCHIJE",                # slot 14 - Real Streaming 6.0.x and up
    snapmanagerexchange => "NYZYDIG", # slot 15 - VB.2, 6.2 and up
    dafs => "TNTHUVC",                # slot 16
    syncmirror_local => "HPTHUVC",    # slot 17 - 6.3 and up
    multistore => "PXBMPCB",          # slot 18
    vfiler => "PXBMPCB",              # slot 18
    icap => "1",                      # slot 19 - and up (or )
    grm => "1",                       # slot 20 - and up (or )
    quicktime => "1",                 # slot 21 - Apple QuickTime Streaming
    ncagent => "1",                   # slot 22 - GRM Agent
    mms_pro => "1",                   # slot 23 - Microsoft Media Streaming high speed
    vld => "VKUWHKM",                 # slot 24 - Virtual Local Disk, VB.2, 6.3 and up
    smsql => "XTJSJSD",               # slot 25 - SnapManager for SQL
    sv_ontap_pri => "FCSWEZB",        # slot 26 - 6.4 and up
    sv_ontap_sec => "FSYLSNL",        # slot 27 - 6.4 and up
    nfscache => "1",                  # slot 29 - 6.4 and up
    fcp => "NOWJTEH",                 # slot 31 - 6.3 and up (or KFPEYRI)
    cluster_remote => "NGOFYXI",      # slot 30 - 6.3 and up  see burt 80994 code=1
    iscsi => "",                      # slot 32 - 6.3 and up (or YMUITYG)
    real_pro => "1",                  # slot 33 - and up
    mms_ultra => "1",                 # slot 34 - and up
    real_ultra => "1",                # slot 35 - and up
    'com-e' => "1",                   # slot 36 - and up
    https_proxy => "1",               # slot 37 - and up
    stache => "1",                    # slot 38 - and up
    gateway_storage => "NSOHUVC",     # slot 42 - 6.4 and up
    sv_unix_pri => "VCCZZDB",         # slot 45 - 6.4 and up
    sv_linux_pri => "ZIWUHIN",        # slot 46 - 6.4 and up
    sv_windows_pri => "BSLQJQE",      # slot 47 - 6.4 and up
    raindrd => "",                    # slot 48 - Anchorsteam and up
    snaplock => "PVMDZOE",            # slot 48 - Fullsail.1 and up (pre 7.0.0.1 NMXHXGN ; after PVMDZOE )
    flex_cache => "PDLQNUJ",          # slot 51 (old free license)
    flexcache_nfs => "WWBMPCB",       # slot 82 (new pay license)
    a_sis => "WCHHJNL",
    nearstore_option => "OMASVYB",
    snapmirror_sync => "NQTWEZB",     # slot 59

    disk_sanitization => "HBAMSNL",
    env_exception => "OGBMSNL",
    flex_scale => "EDTHUVC",
    snapvalidator => "ROLSJSD",       # slot 63
    flex_clone => "WYMFYXI",          # slot 66
    operations_manager => "UZAOOLF",
    protection_manager => "UHJSJSD",
    provisioning_manager => "MPNFYXI",
    smdomino => "HREOOLF",
    snaplock_enterprise => "FADOOLF",
    snapmanager_oracle => "IBZADRK",
    snapmanager_sap => "OKRWEZB",
    snapmanager_sharepoint => "MRSJTEH",
    );
%gateway_1day_licenses = (
    );
%gateway_90day_licenses = (
    );
%nearstore_licenses = (
    nfs => "JCHZPUE",
    cifs => "HJIMEAK",
    http => "PRQQZGI",
    netcache => "TLIMEAK",
    cluster => "HVAIGIB",
    cf => "HVAIGIB",
    snapmirror => "JYAIGIB",
    snapmirror_sync => "RCZSVEC",
    snaprestore => "FEVDOMN",
    nntp => "PZWQZGI",
    mms => "PHFVUNG",
    real => "XPNZPUE",
    snapmanagerexchange => "VWOMEAK",
    dafs => "JSJIJTL",
    syncmirror_local => "NMBEOMN",
    multistore => "TJHIGIB",
    vfiler => "TJHIGIB",
    icap => "XDZDLBD",
    grm => "NENBPDJ",
    quicktime => "NMVFKKH",
    ncagent => "VUDKFRF",
    mms_pro => "TBFXTWK",
    vld => "ZWZSYPM",
    sms => "HBEIJTL",
    smsql => "YCTPBIL",               # slot 25 - SnapManager for SQL
    sv_ontap_pri => "JOXSVEC",
    sv_ontap_sec => "JEEIJTL",
    pnfs => "1",
    cluster_remote => "RSTBPDJ",
    fcp => "BKFBBKH",
    iscsi => "RIGTYPM",
    real_pro => "ZMKIJTL",
    mms_ultra => "DLGTYPM",
    real_ultra => "FUVOAYD",
    'com-e' => "VUJMEAK",
    https_proxy => "VCSQZGI",
    stache => "DLAVUNG",
    sv_unix_pri => "XXHKFRF",         # slot 45 - 6.4 and up (site license)
    sv_linux_pri => "FGQOAYD",        # slot 46 - 6.4 and up (site license)
    sv_windows_pri => "FOYSVEC",      # slot 47 - 6.4 and up (site license)
    snaplock => "XVCGKKH",            # slot 48 - Fullsail.1 and up (pre 7.0.0.1 RYCEOMN ; after THSZPUE )
    flex_cache => "HNKMEAK",          # slot 51 (old free license)
    flexcache_nfs => "AJHIGIB",       # slot 82 (new pay license)
    nearstore_asis1 => "MHNXTWK",
    nearstore_asis2 => "WQQQZGI",
    a_sis => "WQQQZGI",
    disk_sanitization => "LNFIJTL",
    env_exception => "SSGIJTL",
    snapvalidator => "VAROAYD",       # slot 63
    flex_clone => "ALSBPDJ",          # slot 66
    operations_manager => "YLGKFRF",
    protection_manager => "YTOOAYD",
    provisioning_manager => "QBTBPDJ",
    smdomino => "LDKKFRF",
    snaplock_enterprise => "JMIKFRF",
    snapmanager_oracle => "MNEXWK",
    snapmanager_sap => "SWWSVEC",
    snapmanager_sharepoint => "QDYFKKH",
    sv_nbu_sec => "UHHZPUE",
    wafl_compression => "XTBVGUE",    # See change 648989 (was wafl_compression)
    compression => "XTBVGUE",         # See change 648989 (was wafl_compression)
    );
%nearstore_1day_licenses = (
    nfs => "JWJASHI",                 # slot 4
    cifs => "JESENOG",                # slot 5
    http => "RMAJIVE",                # slot 6
    netcache => "PTBWWAK",            # slot 7
    cluster => "DDURYIB",             # slot 8
    cf => "DDURYIB",                  # slot 8
    snapmirror => "LJDIJTL",          # slot 10
    snaprestore => "NSSDLBD",         # slot 11
    nntp => "LHQASHI",                # slot 12 - 6.0.x and up
    mms => "LPYENOG",                 # slot 13 - Microsoft Streaming 6.0.x and up
    real => "TXGJIVE",                # slot 14 - Real Streaming 6.0.x and up
    snapmanagerexchange => "REIWWAK", # slot 15 - VB.2, 6.2 and up
    dafs => "NOASYIB",                # slot 16
    syncmirror_local => "NMBEOMN",    # slot 17 - 6.3 and up
    multistore => "RCDSBUL",          # slot 18
    vfiler => "RCDSBUL",          # slot 18
    icap => "TLSNDCD",                # slot 19 - and up
    grm => "JMGLHEJ",                 # slot 20 - and up
    quicktime => "JUOPCLH",           # slot 21 - Apple QuickTime Streaming
    ncagent => "RCXTXRF",             # slot 22 - GRM Agent
    mms_pro => "PJYGMXK",             # slot 23 - Microsoft Media Streaming high speed
    vld => "DTQCOFC",                 # slot 24 - Virtual Local Disk, VB.2, 6.3 and up
    sms => "DJXRBUL",                 # slot 25 - SnapManager for SQL
    sv_ontap_pri => "LZZSYPM",        # slot 26
    sv_ontap_pri => "NIPOAYD",        # slot 27
    pnfs => "1",                      # slot 29
    cluster_remote => "1",            # slot 30 - 6.3 and up  see burt 80994
    fcp => "RUEHMXK",                 # slot 31 - 6.3 and up
    iscsi => "VEXCOFC",               # slot 32 - 6.3 and up
    real_pro => "VUDSBUL",            # slot 33 - and up
    mms_ultra => "ZSZCRQM",           # slot 34 - and up
    real_ultra => "BCPYSYD",          # slot 35 - and up
    'com-e' => "RCDWWAK",             # slot 36 - and up
    https_proxy => "RKLASHI",         # slot 37 - and up
    stache => "ZSTENOG",              # slot 38 - and up
    nearstore => "XVCGKKH"
    );
%nearstore_90day_licenses = (
    );
# See: http://wikid.netapp.com/w/Simulator_licenses
%promoted_simulator_licenses = (   # Exclude
    nfs => "BQOEAZL",                 # slot 4
    cifs => "DZDACHD",                # slot 5
    http => "NAZOMKC",                # slot 6
    netcache => "1",                  # slot 7
    cluster => "JQIGWWF",             # slot 8
    cf => "JQIGWWF",                  # slot 8
    snapmirror => "DFVXFJJ",          # slot 10
    snaprestore => "DNDCBQH",         # slot 11
    nntp => "1",                      # slot 12 - 6.0.x and up
    mms => "1",                       # slot 13 - Microsoft Streaming 6.0.x and up
    real => "1",                      # slot 14 - Real Streaming 6.0.x and up
    snapmanagerexchange => "BCJEAZL", # slot 15 - VB.2, 6.2 and up
    dafs => "TBPGWWF",                # slot 16
    syncmirror_local => "RIQTKCL",    # slot 17 - 6.3 and up
    multistore => "NQBYFJJ",          # slot 18
    vfiler => "NQBYFJJ",              # slot 18
    icap => "1",                      # slot 19 - and up
    grm => "1",                       # slot 20 - and up
    quicktime => "1",                 # slot 21 - Apple QuickTime Streaming
    ncagent => "1",                   # slot 22 - GRM Agent
    mms_pro => "1",                   # slot 23 - Microsoft Media Streaming high speed
    vld => "JGFRLTG",                 # slot 24 - Virtual Local Disk, VB.2, 6.3 and up
    smsql => "HNGEAZL",               # slot 25 - SnapManager for SQL
    sv_ontap_pri => "PVOIVFK",        # slot 26 - 6.4 and up
    sv_ontap_sec => "PDXMQMI",        # slot 27 - 6.4 and up
    sv_dr_ontap_pri => "",     # slot 28 - AnchorSteam (6.6?) and up
    nfscache => "1",                  # slot 29 - 6.4 and up
    fcp => "BKHEXNB",                 # slot 31 - 6.3 and up (or KFPEYRI)
    cluster_remote => "BCZZBHD",      # slot 30 - 6.3 and up  see burt 80994 code=1
    iscsi => "BSLRLTG",               # slot 32 - 6.3 and up (or YMUITYG)
    real_pro => "1",                  # slot 33 - and up
    mms_ultra => "1",                 # slot 34 - and up
    real_ultra => "1",                # slot 35 - and up
    'com-e' => "1",                     # slot 36 - and up
    https_proxy => "1",               # slot 37 - and up
    stache => "1",                    # slot 38 - and up
    gateway_storage => "",     # slot 42 - 6.4 and up
    sv_unix_pri => "RQAYBFE",         # slot 45 - 6.4 and up
    sv_linux_pri => "ZYICXLC",        # slot 46 - 6.4 and up
    sv_windows_pri => "ZOPRKAM",      # slot 47 - 6.4 and up
    snaplock => "ZOJPPVM",            # slot 48 - Fullsail.1 and up (pre 7.0.0.1 BIICBQH )
    raindrd => "",              # slot 49 - Anchorsteam and up
    #XXX double check
    # License PFS under its various aliases
    access_on_demand => "HFAYFJJ",    # slot 51
    flex_cache => "HFAYFJJ",          # slot 51 (old free license)
    flexcache_nfs => "EVMEXNB",       # slot 82 (new pay license)

    snapmover => "",           # slot 58
    snapmirror_sync => "XJQIVFK",     # slot 59
    sv_windows_ofm_pri => "",         # slot 62
    snapvalidator => "JQAACHD",       # slot 63
    socks => "UXBNQMI",               # slot 64
    http_compression => "SEDAFSN",           # slot 65
    flex_clone => "ANLEAZL",          # slot 66
    rapid_restore => "CMRGWWF",       # slot 67
    advanced_sis => "",        # slot 68
    nearstore_asis1 => "",     # slot 68
    nearstore_option => "ELNRLTG",    # slot 69
    nearstore_asis2 => "",     # slot 70
    a_sis => "",
    disk_sanitization => "PZKEAZL",
    env_exception => "",
    flex_scale => "",
    operations_manager => "CYLGWWF",
    protection_manager => "CGUKRDE",
    provisioning_manager => "UYNXFJJ",
    smdomino => "RKBAFSN",
    snaplock_enterprise => "",
    snapmanager_oracle => "QZJTKCL",
    snapmanager_sap => "WICPMKC",
    snapmanager_sharepoint => "UPDCBQH",
    sv_nbu_sec => "",
   );

%sa_filer_license = (
    flexcache_nfs => "QLULMHM",
    );

%vsa_90day_licenses = (
    a_sis => "YJTNOSE",
    cifs => "PPBSJZC",
    disk_sanitization => "DDWEYEI",
    flex_clone => "MDJWHRL",
    flexcache_nfs => "MTFHXNM",
    http => "ZQWGUCC",
    iscsi => "NIJJTLG",
    multistore => "ZGZPNBJ",
    nfs => "NGMWHRL",
    operations_manager => "OOJYDPF",
    protection_manager => "OWRCZVD",
    provisioning_manager => "GEWPNBJ",
    smdomino => "DBZRMKN",
    smsql => "TDEWHRL",
    snapdrive_unix => "UPTRJZC",
    snapdrive_windows => "SWUEYEI",
    snapmanagerexchange => "NSGWHRL",
    snapmanager_hyperv => "OEGJTLG",
    snapmanager_oracle => "CQHLSUK",
    snapmanager_sap => "IZZGUCC",
    snapmanager_sharepoint => "GGBUIIH",
    snapmirror => "PVSPNBJ",
    snapmirror_sync => "JAOADYJ",
    snapmover => "JSFWHRL",
    snaprestore => "PDBUIIH",
    snapvalidator => "VGYRJZC",
    sv_ontap_pri => "BMMADYJ",
    vld => "VWCJTLG",
   );

%promoted_vsa_licenses = (
    a_sis => "GONOURE",
    cifs => "XTVSPYC",
    disk_sanitization => "LHQFEEI",
    flex_clone => "UHDXNQL",
    flexcache_nfs => "UXZHDNM",
    http => "HVQHACC",
    iscsi => "VMDKZKG",
    multistore => "HLTQTAJ",
    nfs => "VKGXNQL",
    operations_manager => "WSDZJOF",
    protection_manager => "WAMDFVD",
    provisioning_manager => "OIQQTAJ",
    smdomino => "LFTSSJN",
    smsql => "BIYWNQL",
    snapdrive_unix => "CUNSPYC",
    snapdrive_windows => "ABPFEEI",
    snapmanagerexchange => "VWAXNQL",
    snapmanager_hyperv => "WIAKZKG",
    snapmanager_oracle => "KUBMYTK",
    snapmanager_sap => "QDUHACC",
    snapmanager_sharepoint => "OKVUOHH",
    snapmirror => "XZMQTAJ",
    snapmirror_sync => "REIBJXJ",
    snapmover => "RWZWNQL",
    snaprestore => "XHVUOHH",
    snapvalidator => "DLSSPYC",
    sv_ontap_pri => "JQGBJXJ",
    vld => "DBXJZKG",
   );

# Some system properties are exposed by multiple interfaces
# If $filer_crosscheck is not zero, it requests all interfaces to be compared when a value is queried
my $filer_crosscheck = 0;
my $filer_cli_answer = 'no';  # The default answer if a yes|no prompt is detected in filer_cli_say
my $filer_cli_answer_prompt = 'Please answer yes or no';
my $cant = "Can\'t";

my $waiting_for_giveback_string = "Waiting for giveback\.\.\.";

# Local subroutines

# Check that file arguments are valid
sub arg_file ($) {
    my $arg = shift;

    if ($arg eq '') {
        ExitScriptError("File name must be specified");
    }
    if ($arg =~ m#/#) {
        ExitScriptError("A valid file name without slashes is required, not $arg");
    }
    return $arg;
}

# Check that the IP address list argument is valid
sub arg_ipaddrs ($) {
    my $arg = shift;

    if (!defined($arg) or $arg eq '') {
        ExitScriptError("An IP address list argument must be specified");
    }
    if ($arg =~ m/\s/ ) {
        ExitScriptError("A valid IP address list argument is required, not '$arg'");
    }
    return $arg;
}

# Check that the ipspace argument is valid
sub arg_ipspace ($) {
    my $arg = shift;

    # OK to be undefined. We should verify the format if it is defined
    return $arg;
}

# Check that the storage paths list argument is valid
sub arg_paths ($) {
    my $arg = shift;

    if (!defined($arg) or $arg eq '') {
        ExitScriptError("A storage path list argument must be specified");
    }
    # space separators are OK to get multiple paths in a create
    return $arg;
}

# Check that snapshot arguments are valid
sub arg_snapshot ($) {
    my $arg = shift;

    if ($arg eq '') {
        ExitScriptError("Snapshot name must be specified");
    }
    if ($arg =~ '/vol' || $arg eq '/vol/') {
#      ExitScriptError("A valid snapshot name (without '/vol') is required, not $arg");
    }
    return $arg;
}


# Check that vfiler argument is valid
sub arg_vfiler ($) {
    my $arg = shift;

    if (!defined($arg) or $arg eq '') {
        ExitScriptError("A vfiler name must be specified");
    }
    if ($arg =~ m/\s/ ) {
        ExitScriptError("A valid vfiler name is required, not '$arg'");
    }
    return $arg;
}

# Check that volume arguments are valid and remove /vol/ if it is there
sub arg_vol_remove_slashvol ($) {
    my $arg = arg_vol(shift);

    $arg =~ s/^\/vol\///;
    return $arg;
}

# Check that volume arguments are valid
sub arg_vol ($) {
    my $arg = shift;

    if ($arg eq '') {
        ExitScriptError("Volume name must be specified");
    }
    if ($arg eq '/vol' || $arg eq '/vol/') {
        ExitScriptError("A valid volume name is required, not $arg");
    }
    return $arg;
}

## @name ::applies_to_filer_with_cluster_installed
## @summary Skips the current test case if filer does not have cluster installed
## @arg $fc Required: The filer connection object for the command line interface
sub applies_to_filer_with_cluster_installed  ($) {
    my($fc) = shift;
    ExitNA ('Cluster service must be installed.') if (!filer_cluster_is_installed($fc));
}

## @name ::applies_to_filer_with_debug_kernel
## @summary Skips the current test case if filer is not running a debug kernel
## @arg $fc Required: The filer connection object for the command line interface
sub applies_to_filer_with_debug_kernel  ($) {
    my($fc) = shift;
    ExitNA ('ONTAP debug kernel is required for this test') if (!filer_kernel_is_debug($fc));
}

## @name ::applies_to_filer_with_dns_enabled
## @summary Skips the current test case if filer does not have DNS enabled
## @arg $fc Required: The filer connection object for the command line interface
sub applies_to_filer_with_dns_enabled  ($) {
    my($fc) = shift;
    ExitNA ('DNS must be enabled.') if (!filer_is_dns_enabled($fc));
}

## @name    ::filer_access_cache_flush
## @summary Flush NFS access cache of filer
## @arg     $fc Required: Filer's connection object.
## @arg     retry Optional: Number of times to retry (default: 4)
## @arg     timeout Optional: retry timeout(default: 300)
## @return  1 if succeeds
sub filer_access_cache_flush ($;%) {
    my ($fc, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT }
        ],
        {
            retry => { type => SCALAR, default => 4, regex => qr(^\d+$) }
        }
    );
    my $text;
    # The retry option is any number >= 0
    # at least try flushing the access cache once.
    my $retries = $opts{retry} + 1;
    my $retry_timeout = $opts{timeout};
    $retry_timeout = 300 if(!defined($retry_timeout));
    my $end_time=time()+$retry_timeout;
    while(time() < $end_time) {
        $text = filer_cli_say($fc, "exportfs -f");
        if ( $text =~ /Cache in use, please try again/ ) {
            # Wait for 10 seconds before trying again
            logcomment("Retrying NFS access cache flush after 10secs...");
            sleep 10;
        } elsif ( ! $text ) {
            # Flush was successful
            return 1;
        }
    }
    ExitError($text);
}

## @name    ::filer_access_cache_get
## @summary Get NFS access cache of filer
## @arg $fc Required: Filer's connection object.
## @arg retry Optional: Number of times to retry (default: 4)
## @return  Hash table of access cache entries:
##    hits   => number of hits,
##    misses => number of misses,
##    total  => lookup requests,
##    ...
sub filer_access_cache_get ($;) {
    my ($fc, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT }
        ],
        {
            retry => { type => SCALAR, default => 4, regex => qr(^\d+$) }
        }
    );
    my ($text, %access_cache, @name, @value, @namval);
    # The retry option is any number >= 0
    # at least try getting the access cache once.
    my $retries = $opts{retry} + 1;

    # Try the command a few times before failing
    foreach my $i (1..$retries) {
        # Get access cache from the filer
        $text = filer_cli_say($fc, "nfsstat -d");
        if ( $i != $retries and $text =~ /Cache in use, please try again/ ) {
            # Wait for 1 second before trying again
            logcomment("Retrying get NFS access cache...");
            sleep 1;
        } elsif ( $text !~ /access cache\s+\(hits,.* misses/ ) {
            # SEE BURT:251213
            # Output is not what it is expected
            ExitError($text);
        } else {
            # Everything is fine
            last;
        }
    }

    # Create a hash table of access cache entries;
    #  i.e., each line that contains 'access cache',
    #  (name1, name2, ...) = (value1, value2, ...)
    #  will be mapped to
    #      name1 => value1, name2 => value2, ...
    %access_cache = map {
                          @namval = /\((.*)\)\s+=\s+\((.*)\)/;
                          @name   = split /,\s/, $namval[0];
                          @value  = split /,\s/, $namval[1];
                          map { $name[$_], $value[$_] } 0..$#name;
                        } grep { /^access cache/ } split /\n/, $text;

    return %access_cache;
}

## @name      ::filer_actual_storage_get
## @summary   Get actual storage for given path
## @arg $fc   Required: Filer's connection object.
## @arg $path Required: Path to get the actual storage.
## @return    Actual storage of path if succeeds else an empty string.
sub filer_actual_storage_get($$;) {
    my ($fc, $path) = validate_pos(@_,
        { type => OBJECT },
        { type => SCALAR }
    );
    my ($cmd, $storage, $adv_path);

    $cmd = "exportfs -s $path";
    # Get actual storage of path from filer
    my $text = filer_cli_say($fc, $cmd);

    ($adv_path, $storage) = ($text =~ /storage of (\S+) is (\S+)/);
    $storage = "" if ($adv_path ne $path);

    return $storage;
}

## @name ::filer_aggr_snap_delete
## @summary Delete an aggregate snapshot on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $aggr Required: The aggregate name
## @arg $snapshot Required: name of the snapshot
## @arg $timeout Optional: seconds to wait for snapshot to be deleted, 0= no wait
## default = 240, 0 means do not wait
sub filer_aggr_snap_delete ($$$;$) {
    my $fc  = shift;
    my $aggr = arg_vol_remove_slashvol(shift);
    my $snapshot = arg_snapshot(shift);
    my $timeout = shift;

    if (!defined($timeout)) {
        $timeout = 240;
    }

    my $cmd = "snap delete -A $aggr $snapshot";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /No such snapshot/) {
            logcomment("Snapshot appears to already be deleted");
            last;
        } elsif ($line =~ /does not exist or is not online|Snapshot in use by / ) {
            ExitError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /usage:|snap:|snap delete:/ ) {
            # This form of error message has not been observed with this command
            ExitError("Unexpected error from '$cmd': $line");
        }
    }

    if ($timeout >0) {
        my $endtime = time() + $timeout;
        while (filer_vol_snap_exists($fc,$aggr,$snapshot)) {
            if ($endtime < time()) {
                ExitError("Snap delete did not complete within $timeout seconds for snapshot: $snapshot on aggregate: $aggr");
            }
            sleep(15);
        }
    }
}

## @name ::filer_aggr_snap_delete_all
## @summary Deletes all snapshots on the aggregate on this filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $aggr Required: The aggregate name
## @arg $timeout Optional: seconds to wait for snapshot to be deleted, 0= no wait
## default = 240, 0 means do not wait
sub filer_aggr_snap_delete_all ($$;$) {
    my $fc  = shift;
    my $aggr = arg_vol_remove_slashvol(shift);
    my $timeout = shift;

    if (!defined($timeout)) {
        $timeout = 240;
    }

    my $cmd = "snap delete -A -a -f $aggr";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /usage:|snap:|snap delete:/ ) {
            # This form of error message has not been observed with this command
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_aggr_spaceinfo_get
## @summary Returns a hash of the current space statistics on a aggregate
## @arg $fc Required: The filer connection object for the command line interface
## @arg $aggr Required: The aggr name
sub filer_aggr_spaceinfo_get($$){
    my $fc  = shift;
    my $aggr  = shift;

    ##my $vol = arg_vol_remove_slashvol(shift);;
    ##my $cmd = "df /vol/$vol";
    my $cmd = "df -A $aggr";
    ##$cmd = filer_vfiler_command($fc,$cmd);
    #my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    my (undef,undef,$result) = say($fc,$cmd);
    ##my @tmprslt = filer_cli_split($text);
    my @tmprslt = split(/\n/,$result);
    my @volout = split(/ +/,$tmprslt[1]);
    my @volsnapout = split(/ +/,$tmprslt[2]);
    my %dfoutput = (TOTAL => int($volout[1]), # total space on volume
            USED => int($volout[2]),  # Used space on the volume
            FREE => int($volout[3]),  # Free space on the volume
            SNAPTOTAL => int($volsnapout[1]), # total space in snapshots area on volume
            SNAPUSED => int($volsnapout[2]),  # used space in snapshots area on volume
            SNAPFREE => int($volsnapout[3])); # free space in snapshots area on volume
    return (%dfoutput);
}

## @name ::filer_aggr_totalspace_get
## @summary Returns the total space on a aggregate in bytes
## @arg $fc Required: The filer connection object for the command line interface
## @arg $aggr Required: The aggregate name
sub filer_aggr_totalspace_get($$){
    my $fc  = shift;
    my $aggr = shift;

    my %outval = filer_aggr_spaceinfo_get($fc,$aggr);
    return(int(int($outval{"TOTAL"}) * 1024));
}

## @name ::filer_aggrs_get
## @summary Returns a list of aggregates on the specified filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $aggr_types Optional: aggregate attributes to look for
sub filer_aggrs_get ($;$) {
    my $fc  = shift;
    my $type_ref = shift;

    my @aggr_types;
    if (defined($type_ref)) {
        $type_ref =~ s/,/ /g;
        $type_ref =~ s/^\s+//;
        $type_ref =~ s/\s+$//;
        $type_ref =~ s/\s+/ /g;
        @aggr_types = split(" ",$type_ref);
    }

    my @aggrs;
    my %found_aggr_type;
    my $cmd = "aggr status -v";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my $aggr = '';
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /No aggregate named/) {
            last;
        } elsif ($line =~ m/^\s+Volume\s+State/) {
            next;
        } elsif ($line =~ m/^\s+Containing\s+aggregate/) {
            next;
        } elsif ($line =~ m/^\s+Plex\s+/) {
            next;
        } elsif ($line =~ m/^\s+RAID\s+group\s+/) {
            next;
        } elsif ($line =~ m/^\s*$/) {
            next;
        } elsif ($line =~ m/aggr status:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
            next;
        } elsif ($line =~ m/^\s*(\S+)\s+online\s+(\S+),?\s+(\S+)/) {
            my $new_aggr = $1;
            if ($aggr ne '') {
                # Make sure all required 'types' are present
                my $all_types_found = 1;
                foreach my $required_type (@aggr_types) {
                    if (!defined($found_aggr_type{$required_type})) {
                        $all_types_found = 0;
                        last;
                    }
                }
                if ($all_types_found) {
                    push(@aggrs,$aggr);
                }
            }
            $aggr = $new_aggr;
            # Drop aggregate name from the current line that also has 'type' values
            $line =~ s/^\s*\S+\s+//;
            %found_aggr_type = (); # reset type data from previous aggregate
            # Now record status and options that appear on this line
        }
        # Assume this line shows additional aggregate 'types'
        $line =~ s/,/ /g;
        $line =~ s/^\s+//;
        $line =~ s/\s+$//;
        $line =~ s/\s+/ /g;
        my @new_types = split(' ',$line);
        foreach my $new_type (@new_types) {
            $found_aggr_type{$new_type} = 1;
        }
    }
    if ($aggr ne '') {
        # Make sure all required 'types' are present
        my $all_types_found = 1;
        foreach my $required_type (@aggr_types) {
            if ($required_type =~ m/^\!/) {
                # This type was requested to not exist
                # Throw this aggregate out if it does exist
                $required_type =~ s/^\!//;
                if (defined($found_aggr_type{$required_type})) {
                    $all_types_found = 0;
                    last;
                }
            } else {
                if (!defined($found_aggr_type{$required_type})) {
                    $all_types_found = 0;
                    last;
                }
            }
        }
        if ($all_types_found) {
            push(@aggrs,$aggr);
        }
    }
    return @aggrs;
}

## @name ::filer_bootargs_get
## @summary Returns a hashref of arg/value pair of the bootargs for the filer.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_bootargs_get {
    my $fc = shift;

    my $cmd = "bootargs dump";
    my($text,$result) = filer_cli_test_say($fc,$cmd,{timeout=>120,retry=>'ok'});

    my %bootargs = ();
    my @lines = split( /\n/, $text );
    for my $line (@lines) {
        next if ($line =~ /Printing boot options:/);
        if ($line =~ /(.*)(\s+)==(\s+)(.*)/) {
            $bootargs{$1} = $4;
            $bootargs{$1} =~ s/(\s+)$//;       # Remove trailing spaces.
        }
    }
    if (! %bootargs) {
        ExitError("Unexpected output from '$cmd': $text");
    }
    return \%bootargs;
}

## @name ::filer_bootargs_set
## @summary Sets the specified bootargs on the filer.
## WARNING: Setting bootargs parameters may require reboot and/or
## re-installation of ONTAP with zeroing of disks.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $args Required: A hashref of arg/value pairs to be set.
sub filer_bootargs_set {
    my ($fc, $args) = validate_pos (@_, {type => OBJECT}, {type => HASHREF});

    logcomment("WARNING: Setting bootargs parameters may require reboot and/or " .
               "re-installation of ONTAP with zeroing of disks.");

    foreach my $arg (keys %$args) {
        my $value = $args->{$arg};
        my $cmd = "bootargs set $arg $value";
        my($text,$result) = filer_cli_test_say($fc,$cmd,{timeout=>120,retry=>'ok'});
        $text =~ s/^\s*//;
        $text =~ s/\s*$//;
        $text =~ s/\r//g;
        if ($text !~ /(\S+)(\s+)==(\s+)(.*)/ || ($1 ne $arg) || ($4 ne $value)) {
            ExitError("Unexpected output from '$cmd': $text");
        }
    }
}

## @name ::filer_bootargs_unset
## @summary Unsets the specified bootargs on the filer, ie. sets them to 'undefined'.
## WARNING: Setting bootargs parameters may require reboot and/or
## re-installation of ONTAP with zeroing of disks.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $args Required: A arrayref of bootargs to be unset.
sub filer_bootargs_unset {
    my ($fc, $args) = validate_pos (@_, {type => OBJECT}, {type => ARRAYREF});

    logcomment("WARNING: Setting bootargs parameters may require reboot and/or " .
               "re-installation of ONTAP with zeroing of disks.");

    foreach my $arg (@$args) {
        my $cmd = "bootargs unset $arg";
        my($text,$result) = filer_cli_test_say($fc,$cmd,{timeout=>120,retry=>'ok'});
        if ($text ne q{}) {
            ExitError("Unexpected output from '$cmd': $text");
        }
    }
}

## @name ::filer_cifs_access
## @summary Add or delete an ACL for a share on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg share Required: The share on filer to set access on
## @arg user Optional: Name of user or group
## @arg rights Optional: The rights being granted to user for share
## @arg group Optional: If group is yes, indicates $user is name of UNIX group
## @arg delete_flag Optional: Specifies access rights being removed on share
## @arg win_machine Opional: Specifies grant access to windows machine accounts
sub filer_cifs_access ($%) {
    my ($fc, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT }
        ],
        {
            share       => { type => SCALAR },
            user        => { type => SCALAR, optional => 1 },
            rights      => { type => SCALAR, optional => 1,
                             regex => qr/^(No Access|Full Control|Read|Change|r|w|x|rw|rx|rwx)$/ },
            group       => { type => SCALAR, optional => 1,
                             regex => qr/^(yes|no)$/ },
            delete_flag => { type => SCALAR, optional => 1,
                             regex => qr/^(yes|no)$/ },
            win_machine => { type => SCALAR, optional => 1,
                             regex => qr/^(yes|no)$/ },
        }
    );

    my ($cmd, $text, $share, $user, $rights, $ret_val);

    $ret_val = 1;

    # Validate parameter combinations
    if (defined($opts{win_machine}) && (defined($opts{user}) ||
        defined($opts{rights}) || defined($opts{group}))) {
        ExitScriptError("Invalid combination of params, win_machine can " .
                       "only be set with share and delete_flag");
    } elsif (!defined($opts{win_machine}) && !defined($opts{user}) && !defined($opts{rights})) {
        ExitScriptError("User and rights params not defined");
    }

    $share  = $opts{share};
    $user   = $opts{user};
    $rights = $opts{rights};

    # Build command based on named parameter settings
    if (defined($opts{win_machine})) {
        if (defined($opts{delete_flag}) && ($opts{delete_flag} eq 'yes')) {
            $cmd = "cifs access -delete $share -m";
        } else {
            $cmd = "cifs access $share -m ";
        }
    } elsif (defined($opts{delete_flag}) && ($opts{delete_flag} eq 'yes')) {
        if (defined($opts{group}) && ($opts{group} eq 'yes')) {
            $cmd = "cifs access -delete $share -g $user $rights";
        } else {
            $cmd = "cifs access -delete $share $user $rights";
        }
    } else {
        if (defined($opts{group}) && ($opts{group} eq 'yes')) {
            $cmd = "cifs access $share -g $user $rights";
        } else {
            $cmd = "cifs access $share $user $rights";
        }
    }

    $text = filer_cli_say($fc, $cmd, {retry=>'ok'});

    if ($text !~ /share\(s\) have been successfully modified/) {
        $ret_val = 0;
    }

    return $ret_val;
}

## @name ::filer_cifs_homedir
## @summary Return home directory paths, if any exist on filer
## @arg fc Required: The filer connection object for the command line interface
sub filer_cifs_homedir($) {
    my $fc  = shift;
    my $cmd = "cifs homedir";

    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    my($line);
    my @info = filer_cli_split($text);

    return 1;
}

## @name ::filer_cifs_is_started
## @summary Determines if CIFS is started on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cifs_is_started ($) {
    my($fc) = shift;
    return filer_service_is_started($fc,'cifs');
}

## @name ::filer_cifs_load_homedir
## @summary Load the cifs_homedir.cfg file on filer
## @arg fc Required: The filer connection object for the command line interface
sub filer_cifs_load_homedir(%) {
    my %opts = validate(
        @_,
        {
            fc => { type => OBJECT },
        }
    );

    my($fc) = $opts{fc};
    my $ret = 0;
    my $cmd = "cifs homedir load -f";

    my($text) = filer_cli_say($fc,$cmd);
    if ($text eq '') {
        $ret = 1;
    }
    return $ret;
}

## @name ::filer_cifs_lookup
## @summary Look up the SID entry from lclgroups.cfg and return the name
##          that corresponds to the SID
## @arg fc Required: The filer connection object for the command line interface
## @arg sid Required: The SID entry from the lclgroups.cfg to lookup
sub filer_cifs_lookup($$) {
    my $fc = shift;
    my $sid = shift;
    my ($cmd, $text);

    $cmd = "cifs lookup $sid";
    $text = filer_cli_say($fc, $cmd, {retry=>'ok'});

    return $text;
}

## @name ::filer_cifs_sessions
## @summary Look up the active cifs sessions on the filer
## @arg fc Required: The filer connection object for the command line interface
sub filer_cifs_sessions($) {
    my $fc = shift;
    my ($cmd, $text, @info, @info1, %HashInfo, $line, $user, $client);

    $cmd = "cifs sessions";
    $text = filer_cli_say($fc, $cmd, {retry=>'ok'});

    @info = filer_cli_split($text);
    # Remove the headers from the array
    if ($info[0] =~ /Server Registers as/i &&
        $info[1] =~ /Root volume language/i &&
       ($info[2] =~ /Selected domain controller/i ||
        $info[2] =~ /Using Local/i) &&
        $info[3] =~ /=========/ &&
        $info[4] =~ /PC IP/) {
        shift @info;
        shift @info;
        shift @info;
        shift @info;
        shift @info
    } else {
        ExitError("Bad header lines: ".
                  "\n$info[0]\n$info[1]\n$info[2]\n$info[3]\ninfo[4]\n");
    }

    foreach $line (@info) {
        chomp $line;
        if ($line =~ /^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+/) {
            @info1 = split(/\s+/, $line);
            ($client, $user) = @info1;
            $user =~ s/\(//g;
            $HashInfo{$client}{user} = $user;
        }
    }

    return %HashInfo;
}

## @name ::filer_cifs_share_change
## @summary Change the caching attribute on a cifs share
## @arg fc Required: The filer connection object for the command line interface
## @arg share Required: The cifs share to change caching option on.
## @arg option Required: The caching option to change to.
sub filer_cifs_share_change($$$) {
    my $fc = shift;
    my $share = shift;
    my $option = shift;
    my $ret = 0;
    my $cmd = "cifs shares -change $share $option";

    my($text) = filer_cli_say($fc,$cmd);
    if ($text eq '1 share(s) have been successfully modified') {
        $ret = 1;
    }
    return $ret;
}

## @name ::filer_cifs_shares_get
## @summary Get cifs share information
## @arg $fc Required: The filer connection object for the command line interface
## @arg share Optional: Share to get info for, if not specified return all info
sub filer_cifs_shares_get ($%) {
    my ($fc, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT }
        ],
        {
            share => { type => SCALAR, optional => 1 }
        }
    );

    my ($cmd, $cur_share, $flg_cnt, @info, $line, $name, $rights);
    my ($req_share_name, %shares, $text);

    $req_share_name = quotemeta ($opts{share});

    $cmd = "cifs shares $opts{share}";
    $text = filer_cli_say($fc, $cmd, {retry=>'ok'});

    @info = filer_cli_split($text);
    # Remove the headers from the array
    # Remove the headers from the array
    if ($info[0] =~ /Name/i && $info[1] =~ /----/) {
        shift @info;
        shift @info;
    } else {
        ExitError("Bad header lines:\n$info[0]\n$info[1]");
    }

    # Format of the output
    # ShareName MountPoint Description
    #           ... flags
    #               Rights Line 1 (FORM: name / rights)
    #                 ...
    #               Rights Line N
    foreach $line (@info) {
        # Check if the line starts with the share name
        if ($line =~ /^(\S+)/) {
            # Start processing of a new share name so zero the counts
            $cur_share = $1;
            if (!$req_share_name || $cur_share =~ /^$req_share_name$/i) {
                $flg_cnt = 0;
                ($shares{$cur_share}{mountpoint},
                    $shares{$cur_share}{description}) =
                    ($line =~ /^\S+\s+(\S+)\s*(.*$)/);
            } else {
                $cur_share = undef;
            }
            next;
        }
        if ($cur_share) {
            # Process share flags
            if ($line =~ /\s+\.\.\.\s+/) {
                ($shares{$cur_share}{flags}[$flg_cnt]) =
                    ($line =~ /\s+\.\.\.\s+(.*$)/);
                $flg_cnt++;
                next;
            }
            # Process the rights
            ($name, $rights) = ($line =~ m#^\s+(\S+)\s+/\s+(.*$)#);
            if ($name && $rights) {
                $shares{$cur_share}{acl}{$name} = $rights;
            }
        }
    }
    return %shares;
}

## @name ::filer_cifs_start
## @summary Starts CIFS service on the filer if possible
## @description Returns 1 if service is now started, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cifs_start ($;$) {
    my($fc) = shift;
    return filer_service_start ($fc,'cifs');
}

## @name ::filer_cifs_stop
## @summary Stops CIFS service on the filer if possible
## @description Returns 1 if service is already stopped, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cifs_stop ($;$) {
    my($fc) = shift;
    return filer_service_stop ($fc,'cifs');
}

## @name ::filer_ciob_revision_get
## @summary Returns the CIOB revision number of the specified filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_ciob_revision_get ($) {
    my $fc = shift;

    my $slot = 0;

    my $cmd = "sysconfig -v $slot";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $field = '';
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /\s+CIOB Revision ID:/) {
            $field = $line;
            $field =~ s/^\s+CIOB Revision ID:\s+(\w+).*/$1/;
            last;
        }
    }
    return $field;
}

## @name ::filer_cli_admin_say
## @summary sends a self terminating command to the filer's command line interface
## @description
## Issues the specified command using the admin privilege level and returns the output.
## If array output was expected the result code is also returned.
## Terminates as a fatal error if command does not return.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $command Required: The command to issue
## @arg %opts Optional: Options hash:
## show => 'all' or 'debug' or 'syslog' to include debug or syslog output
## default excludes both debug and syslog messages from the returned output.
## retry => 'ok' if command can be retried and get the same expected results
## timeout => n to override the $FILER_CLI_SAY_TIMEOUT timeout (in seconds) value
## allowerror =>
##    0 (default) to not allow errors
##    1 if errors (other than timeout) are expected
##    2 if 'waiting for giveback' message may appear and is OK
##    3 if a TIMEOUT is OK (The returned result code will be -1 if a timeout is seen)
sub filer_cli_admin_say ($$;%) {
    my($fc) = shift;
    my($command) = shift;
    my($hashref) = shift;

    my $cmd = "priv set -q admin;$command";
    return (filer_cli_say($fc,$cmd,$hashref));
}

## @name ::filer_cli_advanced_say
## @summary sends a self terminating command to the filer's command line interface
## @description
## Issues the specified command using the advanced privilege level and returns the output.
## If array output was expected the result code is also returned.
## Terminates as a fatal error if command does not return.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $command Required: The command to issue
## @arg %opts Optional: Options hash:
## show => 'all' or 'debug' or 'syslog' to include debug or syslog output
## default excludes both debug and syslog messages from the returned output.
## retry => 'ok' if command can be retried and get the same expected results
## timeout => n to override the $FILER_CLI_SAY_TIMEOUT timeout (in seconds) value
## allowerror =>
##    0 (default) to not allow errors
##    1 if errors (other than timeout) are expected
##    2 if 'waiting for giveback' message may appear and is OK
##    3 if a TIMEOUT is OK (The returned result code will be -1 if a timeout is seen)
sub filer_cli_advanced_say ($$;%) {
    my($fc) = shift;
    my($command) = shift;
    my($hashref) =shift;

    return (filer_cli_say($fc,"priv set -q advanced;$command",$hashref));
}

## @name ::filer_cli_clean
## @summary removes any syslog messages from a text string
## @arg $line Required: The filer output text from the command line interface
sub filer_cli_clean($) {
    my $line = shift;

    # Remove syslog messages
    # Note that vfiler output has an extra 'vfiler_name@filer_name:\s' at the start of the square bracketed text, for example:
    # 'Tue Sep 23 07:23:41 EDT [vf2@rtpqa-f880-3: lun.destroy:warning]: LUN /vol/vol0/vf2/snapshot_lun_offline destroyed'
    # 'Tue Sep 23 07:45:19 EDT [rtpqa-f880-3: mlnx_trace_wedge_pr:info]: VI IC: Cluster standby IB link (port1) down for too long,please check cable/SFP on port1 in slot 10.'
    # 'Mon Sep 29 16:10:59 EDT [asup.smtp.sent:notice]: System Notification mail sent: System Alert from rtpqa-f880-3'
    # 'Mon Dec  1 13:22:12 PST [rtpqa-fas980-1: asup.general.reminder:info]: Autosupport is not configured to send to Network Appliance. Enabling Autosupport can significantly speed problem determination and resolution should a problem occur on your system. For further information on Autosupport, please see: http://now.netapp.com/autosupport/'
    # These may occur in the middle of output, so don't insist on a new line at the start
    $line =~ s/(Mon|Tue|Wed|Thu|Fri|Sat|Sun) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s+\d{1,2} \d\d\:\d\d\:\d\d [A-Z]{3} \[\S+?\:\s*\S+?\]\:\s+\S.*?(\r*\n|$)//mg;
    # 'Wed Oct  1 09:13:22 EDT last message repeated 2 times'
    $line =~ s/(Mon|Tue|Wed|Thu|Fri|Sat|Sun) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s+\d{1,2} \d\d\:\d\d\:\d\d [A-Z]{3} last message repeated \d+ times(\r*\n|$)//mg;
    return $line;
}

## @name ::filer_cli_default_answer_set
## @summary Establishes an answer for a command that causes a prompt on filer's command line interface
## @arg $fc Required: The filer connection object for the command line interface
## @arg $answer Required: The response to be issued if prompted
## @arg $prompt Optional: The prompt string to detect the need for the answer
sub filer_cli_default_answer_set ($$;$) {
    my($fc) = shift;
    my($answer) = shift;
    my($prompt) = shift;

    $filer_cli_answer = $answer;
    if (defined($prompt) and $prompt ne '') {
        $filer_cli_answer_prompt = $prompt;
    } else {
        $filer_cli_answer_prompt = 'Please answer yes or no';
    }
}

## @name ::filer_cli_diag_say
## @summary sends a self terminating command to the filer's command line interface
## @description
## Issues the specified command using the diag privilege level and returns the output.
## If array output was expected the result code is also returned.
## Terminates as a fatal error if command does not return.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $command Required: The command to issue
## @arg %opts Optional: Options hash:
## show => 'all' or 'debug' or 'syslog' to include debug or syslog output
## default excludes both debug and syslog messages from the returned output.
## retry => 'ok' if command can be retried and get the same expected results
## timeout => n to override the $FILER_CLI_SAY_TIMEOUT timeout (in seconds) value
## allowerror =>
##    0 (default) to not allow errors
##    1 if errors (other than timeout) are expected
##    2 if 'waiting for giveback' message may appear and is OK
##    3 if a TIMEOUT is OK (The returned result code will be -1 if a timeout is seen)
sub filer_cli_diag_say ($$;%) {
    my($fc) = shift;
    my($command) = shift;
    my($hashref) = shift;

    return (filer_cli_say($fc,"priv set -q diag;$command",$hashref));
}

## @name ::filer_cli_error_check
## @summary sends an invalid command to the filer's command line interface
## @description
## Issues the specified command.
## Verifies the expected error message(s).
## Returns the command output.
## If an array of output is requested,
## it returns the return code, the observed text output
## and a PASS/FAIL status (0/1)
## Terminates as a fatal error if command does not return.
## If there is expected output, this treats each call
## as a separate test case that will have either a PASS or FAIL
## test result logged with a message.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $command Required: The command to issue
## @arg $expectedrc Optional: The expected return code from this command.
## Set to '' if you do not wish to test the result.
## @arg @expected Optional: expected output text (regular expressions) for this command
sub filer_cli_error_check ($$$;@) {
    my($fc) = shift;
    my($command) = shift;
    my($expectedrc) = shift;
    my(@expected) = @_;

    my $got_expected_msg = 1; # Assume all messages were found
    my $fail_status = 0;      # Assume this call does not fail

    # Allow long timeout and accept errors for this command.
    # Return here to verify expected output.
    my($text,$result) = filer_cli_say($fc,$command,480,1);

    my @info = filer_cli_split($text);
    if (0 < @info) {
        foreach my $expected (@expected) {
            if (0 == grep { /$expected/ } @info) {
                ExitSoftError("Missing expected text output: '$expected'");
                $got_expected_msg = 0;
                $fail_status = 1; # Note that this test failed
            }
        }
    } else {
        if (0 < @expected) {
            ExitSoftError("No output text to verify '$expected[0]' from: '$command'");
            $got_expected_msg = 0;
            $fail_status = 1; # Note that this test failed
        }
    }
    if (defined($expectedrc) and $expectedrc ne '') {
        if ($result ne $expectedrc) {
            if ($command =~ m/vfiler\s+run\s+/ and $result eq '1') {
                if ($expectedrc ne '0') {
                    if ($got_expected_msg) {
                        # If we got all of the messages, issue message for the wrong result code
                        ExitSoftError('-burt',103863,"Expected return code: ".$expectedrc." got, $result from: $command");
                        $fail_status = 1; # Note that this test failed
                    } else {
                        # Just issue a comment that return code was also not expected
                        logcomment("Expected result code \'$expectedrc\'. Got \'$result\' from \'$command\'");
                    }
                } else {
                }
            } else {
                # Make sure there is a result code to verify
                if ($result !~ m/result not found/) {
                    # If we got all of the messages, verify the return code too
                    if ($got_expected_msg) {
                        ExitSoftError("Expected result code \'$expectedrc\'. Got \'$result\' from \'$command\'");
                        $fail_status = 1; # Note that this test failed
                    } else {
                        # Just issue a comment that return code was also not expected
                        logcomment("Expected result code \'$expectedrc\'. Got \'$result\' from \'$command\'");
                    }
                }
            }
        }
    }

    if (!$fail_status) {
        if (defined($expectedrc) and $expectedrc ne '') {
            if (0 < @expected) {
                logresult('PASS',"Verified return code $expectedrc and expected text from: $command");
            } else {
                logresult('PASS',"Verified return code $expectedrc from: $command");
            }
        } else {
            # There was no expected return code
            if (0 < @expected) {
                logresult('PASS',"Verified expected text from: $command");
            } else {
                # There was no output to verify, thus this was not a test
            }
        }
    }

    if (wantarray) {
        return ($text,$result,$fail_status);
    } else {
        return $text;
    }
}

## @name ::filer_cli_priv
## @summary Returns command line interface privilege level and optionally establishes a new privilege
## @arg $fc Required: The filer connection object for the command line interface
## @arg $priv Optional: The privilege level to be established: 'admin' 'advanced' 'diag' or 'test'
sub filer_cli_priv ($;$) {
    my($fc) = shift;
    my($priv) = shift;

    my $old_priv = filer_cli_say($fc,'priv',{retry=>'ok'});
    if (defined($priv) and $priv ne '') {
        my $cmd = "priv set -q $priv";
        my $text = filer_cli_say($fc,$cmd,{retry=>'ok'});
    }
    return $old_priv;
}

## @name ::filer_cli_say
## @summary sends a self terminating command to the filer's command line interface
## @description
## Issues the specified command and returns the output.
## If array output was expected the result code is also returned.
## Terminates as a fatal error if command does not return.
## Terminates as a fatal error if command returns cmdNotFoundReturn.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $command Required: The command to issue
## @arg %opts Optional: Options hash:
## show => 'all' or 'debug' or 'syslog' to include debug or syslog output
## default excludes both debug and syslog messages from the returned output.
## retry => 'ok' if command can be retried and get the same expected results
## timeout => n to override the $FILER_CLI_SAY_TIMEOUT timeout (in seconds) value
## ignore => 'panic' to prevent reporting of PANIC strings found in the output
## useresult => 0 to prevent use of the ontap result command, default is 1
## allowerror =>
##    0 (default) to not allow errors
##    1 if errors (other than timeout) are expected
##    2 if 'waiting for giveback' message may appear and is OK
##    3 if a TIMEOUT is OK (The returned result code will be -1 if a timeout is seen)
sub filer_cli_say ($$;$$$) {
    my($fc) = shift;
    my($command) = shift;

    my $result = '';

    my $options = '';

    # Determine default timeout
    my($timeout);
    if (!defined($FILER_CLI_SAY_TIMEOUT)) {
        $FILER_CLI_SAY_TIMEOUT = 60;
    }
    $timeout = $FILER_CLI_SAY_TIMEOUT;

    # Parse optional parameters
    # An older version of this function supported $timeout and $allowerror as two positional parameters
    # This support is continued, however these values are now also accepted as hash entries
    my $allowerror =0;
    my($opts) = shift;
    if (defined($opts)) {
        if (ref($opts) ne 'HASH') {
            $timeout = $opts;
            $options .= " timeout =$timeout";
            $opts = shift;
        }
    } else {
        $opts = shift;
        if (defined($opts)) {
            if (ref($opts) ne 'HASH') {
                $allowerror = $opts;
                $options .= " allowerror =$allowerror";
                $opts = shift;
            }
        } else {
            $opts = shift;
        }
    }
    no strict 'refs';
    if (defined($opts)) {
        if (defined($opts->{'timeout'})) {
            $timeout = $opts->{'timeout'};
            if ($timeout !~ m/^\d+$/) {
                ExitScriptError("Unrecognized timeout option: $opts->{'timeout'}, must be numeric");
            }
            $options .= " timeout=$timeout";
        }
        if (defined($opts->{'allowerror'})) {
            $allowerror = $opts->{allowerror};
            $options .= " allowerror=$allowerror";
        }
    }
    if ($allowerror !~ m/^[0123]$/ ) {
        ExitScriptError("Unrecognized allowerror option: $opts->{'allowerror'}, must be 0 , 1, 2 or 3");
    }
    my ($show_debug, $show_syslog) = (0,0);
    if (defined(my $show = $opts->{'show'})) {
        if ($show eq 'all') {
            $show_debug = $show_syslog = 1;
            $options .= ' show=all';
        } elsif ($show eq 'debug') {
            $show_debug = 1;
            $options .= ' show=debug';
        } elsif ($show eq 'syslog') {
            $show_syslog = 1;
            $options .= ' show=syslog';
        } else {
            ExitScriptError("Unrecognized show option: $opts->{'show'}, must be 'all' 'debug' or 'syslog'");
        }
    }

    my $use_result = 1;
    if (defined($opts->{'useresult'})) {
        $use_result = $opts->{'useresult'};
    }
    my $result_cmd;
    if ($use_result) {
        $result_cmd = 'result -l';
    } else {
        $result_cmd = '';
    }
    my $result_cmd_stack = ';'.$result_cmd;

    my $retry_ok = 0;
    if (defined(my $retry = $opts->{'retry'})) {
        if ($retry =~ m/^ok$/i) {
            $retry_ok = 1;
            $options .= ' retry=ok';
        } else {
            ExitScriptError("Unrecognized retry option: $opts->{'retry'}, must be 'ok'");
        }
    }
    use strict 'refs';

    if (length($command) > 1024 ) {
        ExitScriptError("The filer command line interface does not support input longer than 1024 characters");
    }

    # Determine severity of non-fatal errors.
    # This allows use of this function by filer_cli_error_check function
    my $errorlevel = '-fail';
    if ($allowerror) {
        $errorlevel = '-warn';
    }
    my @before_more = ();  # Burt 72264, needs to pass extra error message to $before

    # Provide command count in log file to help trace test progress
    $FILER_COMMAND_COUNT++;
    logcomment("Command input ($FILER_COMMAND_COUNT): \'$command\' $options");

    my $conn_type = $fc-> conntype();
    my($match,$before,$after);
    my $inputline;
    if (length($command) < 1005) {
        $inputline = "$command$result_cmd_stack;echo D'o'Ne";
        # We have room for both result and echo
        # Retry this command as long as the filer appears to have no more rsh threads available
        my $retry_pause = 1;
        do {
            # 'echo not found' appears on netcache machines
            ($match,$before,$after) = say($fc,
                $inputline,
                qr/DoNe|echo not found|syntax error: .* inside quotes\.|Connection timed out| Connection reset by peer|$waiting_for_giveback_string|$filer_cli_answer_prompt/,
                '-timeout', $timeout,
                "-comment","Attempted '$command'");
            if (defined($match) and $match =~ m/ Connection reset by peer/) {
                # This message is only returned by rsh connections (perhaps only rsh-socket connections)
                # for example: Error reading "IO::Socket::INET=GLOB(0x13a2558)": Connection reset by peer
                # It appears the filer is out of rsh threads.
                # This means the command was not really invoked
                # and it is still safe to attempt the command for the first time.
                sleep($retry_pause);
                if ($retry_pause == 1) {
                    logcomment("Command will be retried until a filer rsh thread becomes available");
                }
                # Extend the next retry pause
                $retry_pause++;
            } else {
                if ($retry_pause > 1) {
                    logcomment("Command filer rsh thread became available after $retry_pause attempts");
                }
                $retry_pause = 0;
            }
        } until ($retry_pause == 0);
    } else {
        # No room for result and echo DoNe, use two command line invocations to get 'result'
        ($match,$before,$after) = say($fc,
            "$command",
            qr/syntax error: end of command inside quotes\.|Connection timed out|$waiting_for_giveback_string|echo not found/,
            '-timeout', $timeout,
            "-comment","Attempted '$command'");
        $before .= $after;
        if ($before =~ /Protocol error, .* closed connection/) {
            my $cmdlength = length($command);
            logcomment("Attempted command: $command");
            if ($allowerror) {
                ExitSoftError("Protocol error over CLI interface. Command is probably too long. It was $cmdlength bytes.");
            } else {
                ExitError("Protocol error over CLI interface. Command is probably too long. It was $cmdlength bytes.");
            }
        }
        if ($use_result) {
            ($match, my $before2, my $after2) = say($fc,
                "$result_cmd;echo D'o'Ne",
                qr/DoNe|echo not found|Connection timed out|$waiting_for_giveback_string|$filer_cli_answer_prompt/,
                '-timeout', 5,
                "-comment","Attempted 'result after $command'");
            $before .= $before2;
        }
    }
    if (defined($match) and $match =~ m/$filer_cli_answer_prompt/) {
        # It appears the command caused an expected prompt
        # Use the answer that can be provided
        ($match,$before,$after) = say($fc,$filer_cli_answer,
            qr/DoNe|echo not found|syntax error: end of command inside quotes\.|Connection timed out|$waiting_for_giveback_string/,
            '-timeout', $timeout,
            "-comment","Attempted '$filer_cli_answer' answer to '$command'");
    }
    if (!defined($match) or $match eq '') {
        #print("before='$before'\n"); #DEBUG
        #print("response='$RESPONSE'\n"); #DEBUG
        #print("after='$after'\n"); #DEBUG

        # Caller may be expecting a timeout
        if (3 == $allowerror) {
            logcomment("Command timed out: '$command'");
            $result = -1;
        } else {
            if (!defined($after) or $after eq '') {
                # burt 68700 rsh intermittently fails to return anything
                # 2004-01-06 - it was announced that debug kernels support up to 192 rsh sessions
                # while non-debug kernels only support 24
                # Failures due to this limit may have messages like: Error reading "IO::Socket::INET=GLOB(0xac887c)": Connection reset by peer
                # Note: a PANIC can also cause no answer
                my $fh = $fc->{'hostp'};
                my $filer_ip = $fh->default_ip();
                if (!host_is_alive($filer_ip)) {
                    # Wait for host to respond to ping
                    logresult('WARN',"Host does not respond to ping after command '$command' via $conn_type.");
                    host_wait($filer_ip);
                }
                if (host_is_alive($filer_ip)) {
                    # Burt feedback requested 'rshstat -a' and 'netstat -rn' output
                    say($fc,'priv set -q diag;rshstat -a','-ok');
                    say($fc,'priv set -q diag;rshstat -t','-ok');
                    say($fc,'netstat -rn','-ok');
                    # Only retry if the command has been marked as retriable
                    if ($retry_ok) {
                        sleep(15);  # Take a break, and try again
                        ($match,$before,$after) = say($fc,
                            "$command$result_cmd_stack;echo D'o'Ne",
                            qr/DoNe|echo not found|syntax error: end of command inside quotes\.|Connection timed out| Connection reset by peer|$waiting_for_giveback_string|$filer_cli_answer_prompt/,
                            '-timeout', $timeout,
                            "-comment","Retry of attempted '$command'");
                        if (!defined($match) or $match eq '') {
                            if ($allowerror) {
                                ExitSoftError("Retry of command '$command' did not return with $conn_type.");
                            } else {
                                ExitError("Retry of command '$command' did not return with $conn_type.");
                            }
                        } else {
                            # Credit burt 68700 only if rsh intermittently fails to return anything
                            # This happens too often in the crusher runs - it overloads burt to check its status
                            # burt_found(68700);
                            logresult('WARN',"Command '$command' did not return until second try with $conn_type. See burt 68700");
                        }
                    } else {
                        if ($allowerror) {
                            ExitSoftError("Command '$command' did not return with $conn_type and is not marked to be retried.");
                        } else {
                            ExitError("Command '$command' did not return with $conn_type and is not marked to be retried.");
                        }
                    }
                } else {
                    # May want to wait for filer to wake up
                    # then look for new core files

                    if ($allowerror) {
                        ExitSoftError("Filer does not respond to ping after $conn_type command '$command'");
                    } else {
                        ExitError("Filer does not respond to ping after $conn_type command '$command'");
                    }
                }
            } else {
                # Filer can hang after a reboot
                if ($after =~ m/Probing devices$/) {
                    # If this was the last thing heard from the filer console, it hung during reboot
                    # burt 81571 - Filer hangs 'Probing devices' on boot after iscsi stop
                    if ($ENV{PRE_POST_TYPE} eq "PRETEST" || $ENV{PRE_POST_TYPE} eq 'PREJOB' ) {
                        burt_found(81571);
                        logresult('WARN',"Command '$command' did not complete, filer appears hung due to burt 81571");
                    } else {
                        ExitSoftError('-burt',81571,"Command '$command' did not complete, filer appears hung 'Probing devices'");
                    }
                    filer_power_cycle($fc);
                    ($match,$before,$after) = say($fc,
                        "$command$result_cmd_stack;echo D'o'Ne",
                        qr/DoNe|result: Command not found|echo not found|syntax error: end of command inside quotes\.|Connection timed out| Connection reset by peer|$waiting_for_giveback_string|$filer_cli_answer_prompt/,
                        '-timeout', $timeout,
                        "-comment","Retry of '$command' after filer hang");
                    if ($match eq '') {
                        if ($allowerror) {
                            ExitSoftError("Retry of command '$command' after filer hang still did not return");
                        } else {
                            ExitError("Retry of command '$command' after filer hang still did not return");
                        }
                    } else {
                        logcomment("Retry of command after filer hang was successful");
                    }
                } else {
                    # rsh commands with large output intermittently do not complete
                    my $burt = 92338;     # Was 76034
                    if ($ENV{PRE_POST_TYPE} eq "PRETEST" || $ENV{PRE_POST_TYPE} eq 'PREJOB' ) {
                        burt_found($burt);
                        logresult('WARN',"Command '$command' did not complete with $conn_type, but did output: $after");
                    } else {
                        ExitSoftError('-burt',$burt,"Command '$command' did not complete with $conn_type, but did output: $after");
                    }
                    ($match,$before,$after) = say($fc,
                        "$command$result_cmd_stack;echo D'o'Ne",
                        qr/DoNe|result: Command not found|echo not found|syntax error: end of command inside quotes\.|Connection timed out| Connection reset by peer|$waiting_for_giveback_string|$filer_cli_answer_prompt/,
                        '-timeout', $timeout,
                        "-comment","Retry of '$command' after partial output");
                    if ($match eq '') {
                        if ($allowerror) {
                            ExitSoftError("Retry of command '$command' did not return using $conn_type.");
                        } else {
                            ExitError("Retry of command '$command' did not return using $conn_type.");
                        }
                    } else {
                        logcomment("Retry of command with partial output was successful using $conn_type");
                    }
                }
            }
        }
    }
    if (defined($before) and $before =~ /Protocol error, .* closed connection/) {
        ExitScriptError("Script problem on filer CLI: $before");
    }
    if (defined($before) and $before =~ /INTERRUPT PRINTF:serialhandler: COM 0: weird interrupt:/) {
        ExitSoftError('-burt',93884,"Detected 'weird interrupt' in console output using $conn_type");
        # Remove the extraneous text
        $before =~ s/INTERRUPT PRINTF:serialhandler: COM 0: weird interrupt: 0x\d\r*\n//;
    }
    if (defined($match) and $match ne 'DoNe' and $match ne 'echo not found') {
        $before = $before.$match;
    }
    # stdout and stderr output have a race condition.
    # Sometimes stderr is slow and the last error message may appear
    # after the '$result and 'DoNe'
    if (defined($after) and $after ne '' && $after !~  /^\r*\n$/) {
        #logcomment("Command '$command' returned extra text: '$after'");
        # Include the extra error message with other text output
        # See Burt 72264 for a discussion
        # Remove end of line characters from after the 'DoNe' match string
        @before_more = split(/\r*\n/, $after);
    }

    if (defined($before)) {
        # Remove this know debug message that can interfer with seeing expected output
        if ($before =~ /tavorvia_cpeer_reinit: hung while flushing cpeer vi sendq/) {
            ExitSoftError('-burt',136552,"Encountered extraneous debug message about tavorvia_cpeer_reinit");
            $before =~ s/tavorvia_cpeer_reinit: hung while flushing cpeer vi sendq\r*\n*//g;
        }
        # Remove debug messages if requested by global setting
        if (!$show_debug) {
            $before =~ s/^DBG\:.*?(\n|$)//mg;
        }
        # Remove syslog messages if requested by global setting
        if (!$show_syslog) {
            $before = filer_cli_clean($before);
        }
    } else {
        $before = '';
    }

    my @info = filer_cli_split($before);
    if (@info >0) {
        if ($conn_type ne 'rsh-persay' and $conn_type ne 'rsh-socket') {
            # logcomment("Connection type: $conn_type\n");
            # The CLI command string is being returned with the $before data.
            # In almost all cases this is unwanted data so we shift off the
            # $command output before doing anything else.  We also test to be
            # sure we have not removed valid data.  The first line should
            # always contain the $command string.
            my $echoed_command = shift(@info);
            #print("Removed this echoed command: $echoed_command\n");#DEBUG
            my $meta_command = quotemeta($command);
            if ($echoed_command !~ /$meta_command/ or ($echoed_command =~ m/:/ and $command !~ m/:/)) {
                # Echoed command doesn't match input $command
                # rsh connections do not echo the command
                # Put it back, it was not the original command
                print("Please verify this line of output should not be ignored: $echoed_command\n");#DEBUG
                @info= ($echoed_command,@info);
            } else {
                #logcomment("Dropped input command line from output: '$echoed_command'\n");
            }
        }

        if ($use_result && defined($match) and $match eq 'DoNe' and $match ne 'echo not found') {
            $result = pop @info;
            chomp $result;
            if ($result =~ s/^result=(0|1|2|3|127|255)\r*\n*$/$1/) {
            } else {
                # Make sure that 'result' command was recognized before trying to cope with where its value is.
                if ($result =~ m/result: Command not found|result not found/) {
                    # We have removed this line from the @info. Should be good to go
                } else {
                    # Error output and stdout output have intermixed, separate them.
                    # Issue burt message and try to recover output without result and 'DoNe'
                    my $result_on_line_split = '';
                    my @result_stack;
                    while ($result !~ /^result=\d+$/) {
                        if ($result_on_line_split eq '') {
                            # Only the following return code are possible
                             # CMD_RESULT_OK             0    // success
                             # CMD_RESULT_FAILED         1    // unspecified generic failure
                             # CMD_RESULT_BAD_USAGE      2    // cmd usage error
                             # CMD_RESULT_BAD_ARG        3    // argument failed validation
                             # CMD_RESULT_CMD_NOT_FOUND  127    // cmd doesn't exist
                             # CMD_RESULT_NOT_SET        255    // cmd did not set return code
                             if ($result =~ s/=([^2]*)(2)([^5]*)(5)([^5]*)(5)/=$1$3$5/
                             || $result =~ s/=([^1]*)(1)([^2]*)(2)([^7]*)(7)/=$1$3$5/
                             || $result =~ s/=([^3]*)(3)()()()()/=$1/
                             || $result =~ s/=([^2]*)(2)()()()()/=$1/
                             || $result =~ s/=([^1]*)(1)()()()()/=$1/
                             || $result =~ s/=([^0]*)(0)()()()()/=$1/) {
                                # Remember closest preceeding line that ends with a numeric value
                                $result_on_line_split = $2.$4.$6;
                                # Attempt to remove "result=" also
                                $result =~ s/(.*)r(.*)e(.*)s(.*)u(.*)l(.*)t(.*)=/$1$2$3$4$5$6$7/;
                            }
                        }
                        push (@result_stack,$result);
                        if (@info >0) {
                            $result = pop @info;
                            chomp $result;
                        } else {
                            # result code is not present or has been separated from the end-of-line character
                            if ($result_on_line_split eq '') {
                                if (!$allowerror) {
                                    # This is a terminating error.
                                    # It needs to be skipped for some error testing scripts that try invalid subcommands
                                    ExitError("no result code was returned using $conn_type for \'$command\'");
                                } else {
                                    ExitSoftError("no result code was returned using $conn_type for \'$command\'");
                                }
                                $result = '';
                            } else {
                                # Assume preceeding line has return code on the end
                                $result = $result_on_line_split;
                            }
                            last;
                        }
                    }
                    # restore non return-code data to the rest of the text output
                    while (@result_stack > 0) {
                        push (@info,pop(@result_stack));
                    }
                }
                logcomment("Default result code returned from \'$command\'") if ($result eq cmdDefaultReturn);
                # The following return code indicates the command is not implemented
                if (!$allowerror) {
                    # This is a terminating error.
                    # It needs to be skipped for some error testing scripts that try invalid subcommands
                    ExitError("cmdNotFoundReturn result code returned from \'$command\'") if ($result eq cmdNotFoundReturn);
                }
            }
        } elsif (defined($match) and $match =~ m/$waiting_for_giveback_string/) {
            if (!$allowerror) {
                # This is a terminating error.
                ExitError("Commands cannot be entered while waiting for giveback");
            } elsif ($allowerror eq '2') {
                logcomment("Detected '$waiting_for_giveback_string'");
            } else {
                ExitSoftError("Commands cannot be entered while waiting for giveback");
            }
         }
        # Now append extra $after lines to the end
        @info = (@info,@before_more);

        # Look for generic filer errors
        foreach my $line (@info) {
            if ($line =~ /Not yet implemented/i) {
                if ($command =~ /lun\s+move/) {
                    burt_found(71372);
                }
                if ($allowerror) {
                    ExitSoftError("Command not implemented: $command");
                } else {
                    ExitError("Command not implemented: $command");
                }
            }
            if ($line =~ /is deprecated/i) {
                # download command as part of a revert may report deprecated options
                # do not trap them here
                if ($line !~ /option\s+\S+\s+is deprecated/i) {
                    if ($allowerror) {
                        ExitSoftError("Command is deprecated: $command");
                    } else {
                        ExitError("Command is deprecated: $command");
                    }
                }
            }

            no strict 'refs';
            if (!defined($opts->{'ignore'}) or $opts->{'ignore'} !~ m/panic/i) {
                if ($line =~ /PANIC: |Generating \/etc\/crash\/mini-core/) {
                    # Make sure this is not one of the PANIC exceptions that does not indicate a PANIC just happened
                    if ($line !~ /relog syslog PANIC: /) {
                        if ($allowerror) {
                            logresult('WARN',"PANIC detected: $line.");
                        } else {
                            ExitSoftError("PANIC detected: $line.");
                        }
                    }
                }
            }
            use strict 'refs';
        }
        $before = join("\r\n", @info);
    }
    # let's make sure result only returns value
    if ($result =~ /result=(\d+)/) { $result = $1;}

    if (wantarray) {
        return ($before,$result);
    } else {
        return $before;
    }
}

## @name ::filer_cli_split
## @summary splits the filer's command line output into an array of separate lines
## @arg $text Required: The command output string to be split
sub filer_cli_split ($) {
    my($text) = shift;

    my @output = split(/\r*\n/,$text);
    return @output;
}

## @name ::filer_cli_test_say
## @summary sends a self terminating command to the filer's command line interface
## @description
## Issues the specified command using the test privilege level and returns the output.
## If array output was expected the result code is also returned.
## Terminates as a fatal error if command does not return.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $command Required: The command to issue
## @arg $timeout Deprecated: use timeout entry in hash instead
## @arg $allowerror Deprecated: use allowerror entry in hash instead
## @arg %opts Optional: Options hash:
## show => 'all' or 'debug' or 'syslog' to include debug or syslog output
## default excludes both debug and syslog messages from the returned output.
## retry => 'ok' if command can be retried and get the same expected results
## timeout => n to override the $FILER_CLI_SAY_TIMEOUT timeout (in seconds) value
## allowerror => 0 (default) to not allow errors
##               1 if errors (other than timeout) are expected
##               2 if 'waiting for giveback' message may appear and is OK
sub filer_cli_test_say ($$;%) {
    my($fc) = shift;
    my($command) = shift;
    my($hashref) = shift;

    return (filer_cli_say($fc,"priv set -q test;$command",$hashref));
}

## @name    ::filer_client_cache
## @summary Caches host on filer and returns if the client has the
##          given access to path.
## @arg $fc        Required: Filer's connection object.
## @arg $client_ip Required: Client's ip address.
## @arg $path      Required: Path to cache access for client.
## @arg $access    Required: Cache client for given access.
##                 Must be ro, rw, root or mount.
## @return 1 if client has given access to path.
##         0 if client does not have given access to path.
sub filer_client_cache ($$$$;) {
    my ($fc, $client_ip, $path, $access) = validate_pos(@_,
        { type => OBJECT },
        { type => SCALAR },
        { type => SCALAR },
        { type => SCALAR, regex => qr/^(r[ow]|root|mount)$/ }
    );
    my $access_op;

    # To cache client for mount, command expects no access
    $access_op = ($access eq 'mount') ? "" : $access;

    # Cache client
    my $text = filer_cli_say($fc, "exportfs -c $client_ip $path $access_op");

    # Parse output of command
    if ( $text =~ /$client_ip has $access access to $path/ ) {
        # Client has given access to path
        return 1;
    } elsif ( $text =~ /$client_ip does not have $access access to $path/ ) {
        # Client does not have given access to path
        return 0;
    } else {
        ExitError($text);
    }
}

## @name ::filer_cluster_disable
## @summary Enable cluster on the filer.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_disable ($) {
    my($fc) = shift;

    my $cmd = "cf disable";
    my($text,$result) = filer_cli_say($fc,$cmd,$FILER_CLI_SAY_TIMEOUT);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ( $line =~ /cf: /) {
            ExitError("Unexpected error attempting to disable clustering");
        }
    }
}

## @name ::filer_cluster_enable
## @summary Enable cluster on the filer.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_enable ($) {
    my($fc) = shift;

    my $cmd = "cf enable";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /clustering already enabled|/) {
            last;
        } elsif ( $line =~ /cf: /) {
            ExitError("Unexpected error attempting to enable clustering");
        }
    }
}

## @name ::filer_cluster_giveback
## @summary Giveback to the filer's partner
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_giveback ($) {
    my($fc) = shift;

    # -f to force giveback even if this call is made before
    # the partner is ready for giveback
    my $cmd = "cf giveback -f";
    my($text,$result) = filer_cli_say($fc,$cmd,$FILER_CLI_SAY_TIMEOUT);

    # Wait for giveback to complete
    my $count = 0;
    while($count<10) {
        sleep(15);
        if (filer_cluster_is_in_takeover($fc)) {
            $count++;
            if ($count>=10) {
                ExitError("Giveback failed");
            }
        } else {
            last;
        }
    }
}

## @name ::filer_cluster_initialize
## @summary Licenses and starts cluster service on the filer if possible
## @description Returns 1 if service is now started, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_initialize ($) {
    my($fc) = shift;

    my($initialized) = 0;
    if (filer_cluster_is_installed($fc)) {
        filer_cluster_license($fc) if (!filer_cluster_is_licensed($fc));
        filer_cluster_enable($fc) if (!filer_cluster_is_enabled($fc));
        # Restore cluster if it is in takeover/giveback
        filer_cluster_restore($fc);
        $initialized = 1;
        print("This system is clustered\n");
    } else {
        print("This system is not clustered\n");
    }
    return $initialized;
}

## @name ::filer_cluster_is_enabled
## @summary Determines if cluster is enabled on the filer.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_is_enabled ($) {
    my($fc) = shift;

    my $cmd = "cf status";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /Cluster enabled|Controller Failover enabled/) {
            return 1;
        }
    }
    return 0;
}

## @name ::filer_cluster_is_in_takeover
## @summary Determines if filer has taken over for its partner
## @description
## Returns 1 if this filer is taking over or has taken over for its partner
## Returns 2 if this filer been taken over or is waiting for giveback,
## Returns 3 if this filer is waiting for partner to recover
## otherwise returns 0.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_is_in_takeover ($) {
    my($fc) = shift;

    my $cmd = "cf status";
    # Allow the 'Waiting for giveback' message (4th parameter is '2')
    my($text,$result) = filer_cli_say($fc,$cmd,{allowerror=>2,retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        $line = $line;
        if ($line =~ /Cluster enabled.* is up|Cluster must be licensed|takeover disabled because of reason |has disabled takeover by /) {
            return 0;
        } elsif ($line =~ /waiting for partner to recover|unsynchronized log/) {
            # We cannot attempt giveback yet
            return 3;
        } elsif ($line =~ /attempting takeover| has taken over/) {
            return 1;
        } elsif ($line =~ / has been taken over by | is ready for giveback|$waiting_for_giveback_string/) {
            return 2;
        }
    }
    return 0;
}

## @name ::filer_cluster_is_in_unsync
## @summary Determines if filer is in unsynchronized log state
## @description
## Returns 1 if this filer is in unsync log state
## otherwise returns 0.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_is_in_unsync ($) {
    my ($fc) = shift;

    my $cmd = "cf status";
    my ($text,$result) =
      filer_cli_say($fc, $cmd, { allowerror=>2, retry=>'ok' });

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        $line = $line;
        if ($line =~ /\(unsynchronized log\)/) {
            return 1;
        }
    }
    return 0;
}

## @name ::filer_cluster_is_installed
## @summary Determines if cluster is installed on the filer.
## @description
## Determines if cluster is installed on the filer by looking
## for the existence of a Cluster Adapter.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_is_installed ($) {
    my($fc) = shift;

    my $cmd = "sysconfig -v";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        $line = $line;
        # "VI Cluster Adapter" "Troika FCV1"  OR "VI Cluster Adapter(Rev B) (Troika FCVI hw rev 83, fw rev 45410001.2b)"
        # "Interconnect HBA (Mellanox IB: MT21108) FW version 3.2.3::3.2.3"
        #  Jivaros (FAS270) systems have no interconnect HBAs, just use  'partner ID:' for all systems
        if ($line =~ /partner ID:/) {
            return 1;
        } elsif ($line =~ /This system has been taken over by /) {
            logresult('WARN',"Filer has been taken over");
            return 1;  # cluster is installed if filer is taken over
        }
    }
    return 0;
}

## @name ::filer_cluster_is_licensed
## @summary Determines if cluster is licensed on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_is_licensed ($) {
    my($fc) = shift;
    return eval(filer_service_is_licensed ($fc,'cluster'));
}

## @name ::filer_cluster_license
## @summary Licenses cluster on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_license ($) {
    my($fc) = shift;
    filer_service_license ($fc,'cluster');
}

## @name ::filer_cluster_restore
## @summary Restore cluster status if either filer has taken over for its partner
## @description
## Returns 1 if this filer is taking over or has taken over for its partner,
## Returns 2 if this filer been taken over or is waiting for giveback,
## otherwise returns 0.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_restore ($) {
    my($fc) = shift;

    my $cf_status = filer_cluster_is_in_takeover($fc);

    # We must wait for the partner to be ready for giveback
    my $count = 0;
    while (1) {
        sleep(15);
        $cf_status = filer_cluster_is_in_takeover($fc);
        if ($cf_status eq '3') {
            $count++;
            if ($count>=10) {
                 ExitError("Giveback failed");
            }
        } else {
            last;
        }
    }

    if ($cf_status eq '0') {
    } elsif ($cf_status eq '1') {
        filer_cluster_giveback($fc);
    } elsif ($cf_status eq '2') {
        # Get the partner to issue a giveback
        my $partner = $fc->hostp()->part_host();
        if (defined($partner) and $partner ne '') {
            host($partner);
            my $fcb = connect('#rsh-socket');
            filer_cluster_giveback($fcb);
            disconnect($fcb);
        } else {
            my $filer = filer_name($fc);
            ExitConfigError("Clustered filers must define their partner in their tharnhosts file: $filer");
        }
    } else {
        ExitScriptError("Unknown cluster status");
    }
    return $cf_status;
}

## @name ::filer_cluster_takeover
## @summary Takeover the filer's partner
## @arg $fc Required: The filer connection object for the command line interface
sub filer_cluster_takeover ($) {
    my($fc) = shift;

    # -f to force takeover
    my $cmd = "cf takeover -f";
    my($text,$result) = filer_cli_say($fc,$cmd,$FILER_CLI_SAY_TIMEOUT);

    # Wait for giveback to complete
    my $count = 0;
    while($count<10) {
        sleep(15);
        if (! filer_cluster_is_in_takeover($fc)) {
            $count++;
            if ($count>=10) {
                ExitError("Takeover failed");
            }
        } else {
            last;
        }
    }
}

## @name ::filer_cluster_unlicense
## @summary Unlicenses cluster on the filer
## @arg $fc Required: The filer connection where this function is executed
sub filer_cluster_unlicense ($) {
    my($fc) = shift;
    filer_service_unlicense ($fc,'cluster');
}

## @name ::filer_crosscheck_disable
## @summary Disable crosschecking of filer attributes as they are retrieved
sub filer_crosscheck_disable () {
    $filer_crosscheck = 0;
}

## @name ::filer_crosscheck_enable
## @summary Enable crosschecking of filer attributes as they are retrieved
sub filer_crosscheck_enable () {
    $filer_crosscheck = 1;
}

## @name ::filer_crosscheck_is_enabled
## @summary Determine if crosschecking of filer attributes as they are retrieved is enabled
sub filer_crosscheck_is_enabled () {
    return $filer_crosscheck;
}

## @name ::filer_dafs_is_started
## @summary Determines if DAFS is started on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_dafs_is_started ($) {
    my($fc) = shift;
    return filer_service_is_started($fc,'dafs');
}

## @name ::filer_dafs_start
## @summary Licenses and starts DAFS service on the filer if possible
## @description Returns 1 if service is now started, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_dafs_start ($) {
    my($fc) = shift;
    return filer_service_start ($fc,'dafs');
}

## @name ::filer_dafs_stop
## @summary Licenses and starts DAFS service on the filer if possible
## @description Returns 1 if service is already stopped, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_dafs_stop ($) {
    my($fc) = shift;
    return filer_service_stop ($fc,'dafs');
}

## @name ::filer_date_advance
## @summary Sets the future date for the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $add_day optional: can be 0 to 28 days, default is 0
## @arg $timezone optional: defaults to local timezone
sub filer_date_advance ($;$$) {
    my($fc) = shift;
    my $add_day = shift;
    my $timezone = shift;

    if (!defined($add_day)) {
        $add_day =0;
    }

    my ($sec,$min,$hour,$mday,$mon,$year) = (localtime(time()))[0..5];

    $mday=$mday + $add_day;   #added day
    $mon = $mon +1;  #real month

    if ($mon==1 ||$mon ==3||$mon==5||$mon==7||$mon==8||$mon==10||$mon==12 ) {
        if ($mday>31) {
            $mday = $mday -31;
            $mon = $mon +1;
            if ($mon>12) {
                $mon =$mon-12;
                $year = $year +1;
            }
        }
    }
    elsif ($mon==4 ||$mon ==6||$mon==9||$mon==11 ) {
        if ($mday>30) {
            $mday = $mday -30;
            $mon = $mon +1;
            if ($mon>12) {
                $mon =$mon-12;
                $year = $year +1;
            }
        }
    }
    elsif ($mon==2) {
        if ($year %4 ==0) { # leap year
            if ($mday>29) {
                $mday = $mday -29;
                $mon = $mon +1;
                if ($mon>12) {
                    $mon =$mon-12;
                    $year = $year +1;
                }
            }
        }
        else { # non leap year
            if ($mday>28) {
                $mday = $mday -28;
                $mon = $mon +1;
                if ($mon>12) {
                    $mon =$mon-12;
                    $year = $year +1;
                }
            }
        }
    }

    # format to 4 digit
    my $yyyy = $year + 1900;

    # And format it for two digits
    my $mm = '0'.$mon;
    $mm =~ s/.*(\d\d)$/$1/;

    my $dd = '0'.$mday;
    $dd =~ s/.*(\d\d)$/$1/;

    my $hh = '0'.$hour;
    $hh =~ s/.*(\d\d)$/$1/;
    my $nn = '0'.$min;
    $nn =~ s/.*(\d\d)$/$1/;
    my $ss = '0'.$sec;
    $ss =~ s/.*(\d\d)$/$1/;
    my $date = $yyyy.$mm.$dd.$hh.$nn.".".$ss;

    if (!defined($timezone)) {

    }
    print("Setting filer time to: $date");
    filer_cli_say ($fc,"date $date");
}

## @name ::filer_date_get
## @summary Gets the date from the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_date_get ($) {
    my($fc) = shift;
    my ($text) = filer_cli_say ($fc,'date',{retry=>'ok'});
    return $text;
}

## @name ::filer_date_set
## @summary Sets the date on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $date optional: [[[[CC]yy]mm]dd]hhmm[.ss]] format of date command input,
## defaults to local time where the script is running
## @arg $timezone optional: defaults to local timezone
sub filer_date_set ($;$) {
    my($fc) = shift;
    my $date = shift;
    my $timezone = shift;

    if (!defined($date)) {
        my ($sec,$min,$hour,$mday,$mon,$year) = (localtime(time()))[0..5];
        $year = $year + 1900;
        my $yyyy = $year;
        my $mm = '0'.$mon;
        $mm =~ s/.*(\d\d)$/$1/;
        # Get the real month number, 1 origin
        $mm=$mm+1;
        # And format it for two digits
        $mm = '0'.$mm;
        $mm =~ s/.*(\d\d)$/$1/;

        my $dd = '0'.$mday;
        $dd =~ s/.*(\d\d)$/$1/;

        my $hh = '0'.$hour;
        $hh =~ s/.*(\d\d)$/$1/;
        my $nn = '0'.$min;
        $nn =~ s/.*(\d\d)$/$1/;
        my $ss = '0'.$sec;
        $ss =~ s/.*(\d\d)$/$1/;
        $date = $yyyy.$mm.$dd.$hh.$nn.".".$ss;
    }
    if (!defined($timezone)) {

    }
    print("Setting filer time to: $date");
    my($text,$result) = filer_cli_say ($fc,"date $date");

    if ($text =~ /date: Invalid date specified/) {
        ExitScriptError("Invalid date specified: '$date'");
    }
}

## @name ::filer_dbg_level_set
## @summary Sets the debug flag on a debug kernel.
## @description
##  Returns 0 if the command succeeded or not running a debug kernel
##  Returns 1 if the command failed
## @arg $fc Required: The filer connection object for the command line interface
## @arg $mod Required: The kernel module to set debug level on
## @arg $level Required: The level to set
sub filer_dbg_level_set ($$$) {
    my($fc) = shift;
    my($mod) = shift;
    my($level) = shift;
    my ($cmd, $ret, $text);

    $ret = 0;
    if (filer_kernel_is_debug($fc)) {
        $cmd = "dbg level $mod=$level";
        ($text) = filer_cli_diag_say ($fc, $cmd, {retry=>'ok'});
        if ($text !~ /$mod := $level/i) {
            $ret = 1;
        }
    }
    return $ret;
}

## @name ::filer_dbg_to_set
## @summary Sets the debug to 'console | log | off'
## @description
##  Returns 0 if the command succeeded or not running a debug kernel
##  Returns 1 if the command failed
## @arg $fc Required: The filer connection object for the command line interface
## @arg $target Required: String of where to send debug, can be 'console', 'log', 'off', 'console log'
sub filer_dbg_to_set ($$) {
    my($fc) = shift;
    my($where) = shift;
    my ($cmd, $ret, $text);

    $ret = 0;
    if (filer_kernel_is_debug($fc)) {
        # No output from command
        $cmd = "dbg to $where ";
        filer_cli_diag_say ($fc, $cmd, {retry=>'ok'});

        $cmd = "dbg to";
        ($text) = filer_cli_diag_say ($fc, $cmd, {retry=>'ok'});
        if ($text !~ /$where/i) {
            $ret = 1;
        }
    }
    return $ret;
}

## @name ::filer_disk_capacity_get
## @summary Returns the size of the smallest disk (undefined if no disks are found.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_disk_capacity_get ($) {
    my $fc  = shift;

    my $cmd = "vol status -r";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my $min_capacity;
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /\s+\d+\/\d+\s+\d+\/\d+/) {
            my $mb = $line;
            $mb =~ s/.*\s\d+\/\d+\s+(\d+)\/\d+.*/$1/ ;
            my $blks = $line;
            $blks =~ s/.*\s\d+\/\d+\s+\d+\/(\d+).*/$1/ ;
            if (defined($min_capacity)) {
                if ($min_capacity > 512*$blks) {
                    $min_capacity = 512*$blks;
                }
            } else {
               $min_capacity = 512 * $blks;
               print("capacity=$min_capacity ($mb/$blks)\n");
            }
        }
    }

    return $min_capacity;
}

## @name ::filer_dns_is_enabled
## @summary Determines if DNS is enabled on the filer.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_dns_is_enabled ($) {
    my($fc) = shift;

    my $cmd = "dns info";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /DNS is enabled/) {
            return 1;
        }
    }
    return 0;
}

## @name ::filer_domainuser_delete
## @summary Delete the specified user from the domain group
## @arg fc Required: The filer connection object for the command line interface
## @arg user Required: The domain user to delete
## @arg group Required: The domain group from which to delete user
sub filer_domainuser_delete($$$) {
    my $fc = shift;
    my $user = shift;
    my $group = shift;
    my ($cmd, $text);

    $cmd = "useradmin domainuser delete $user -g $group";
    $text = filer_cli_say($fc, $cmd, {retry=>'ok'});

    return $text;
}

## @name ::filer_download
## @summary Perform a download on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $timeout Optional: a timeout in seconds, default is 1200
sub filer_download ($;$) {
    my($fc) = shift;
    my($timeout) = shift;
    if (!defined($timeout)) {
        $timeout = 1200;
    }

    my ($text) = filer_cli_say ($fc,'download',{timeout=>$timeout});
    return $text;
}

## @name ::filer_export_fence_disable
## @summary Unfence the hosts from the paths on an exports rule
## @arg $fc Required: The filer connection object for the command line interface
## @arg hosts Optional: The set of hosts to unfence, default is all hosts
## @arg paths Optional: The set of paths to unfence, default is all paths
## @arg persistent Optional: Save to the exports file, default is not to save
sub filer_export_fence_disable ($;%) {
    my ($fc, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT }
        ],
        {
            hosts      => { type => ARRAYREF, optional => 1 },
            paths      => { type => ARRAYREF, optional => 1 },
            persistent => { type => SCALAR,    default => 0 },
        }
    );
    my $hosts = 'allhosts';
    my $paths = 'allpaths';
    my $save  = $opts{persistent} ? 'save' : 'nosave';

    # Create list of hosts to unfence
    if ( defined($opts{hosts}) ) {
        $hosts = join(':', @{$opts{hosts}});
    }
    # Create list of paths to unfence
    if ( defined($opts{paths}) ) {
        $paths = join(':', @{$opts{paths}});
    }

    # Disable fencing on the filer
    my $cmd  = "exportfs -b disable $save $hosts $paths";
    my $text = filer_cli_say ($fc, $cmd, {retry => 'ok'});
    if ( $text ) {
        ExitError($text);
    }
    return 1;
}

## @name ::filer_export_fence_enable
## @summary Fence the hosts from the paths on an exports rule
## @arg $fc Required: The filer connection object for the command line interface
## @arg hosts Optional: The set of hosts to fence, default is all hosts
## @arg paths Optional: The set of paths to fence, default is all paths
## @arg persistent Optional: Save to the exports file, default is not to save
sub filer_export_fence_enable ($;%) {
    my ($fc, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT }
        ],
        {
            hosts      => { type => ARRAYREF, optional => 1 },
            paths      => { type => ARRAYREF, optional => 1 },
            persistent => { type => SCALAR,    default => 0 },
        }
    );
    my $hosts = 'allhosts';
    my $paths = 'allpaths';
    my $save  = $opts{persistent} ? 'save' : 'nosave';

    # Create list of hosts to fence
    if ( defined($opts{hosts}) ) {
        $hosts = join(':', @{$opts{hosts}});
    }
    # Create list of paths to fence
    if ( defined($opts{paths}) ) {
        $paths = join(':', @{$opts{paths}});
    }

    # Enable fencing on the filer
    my $cmd  = "exportfs -b enable $save $hosts $paths";
    my $text = filer_cli_say ($fc, $cmd, {retry => 'ok'});
    if ( $text !~ /exportfs: Turning on NFS for fencing/ ) {
        ExitError($text);
    }
    return 1;
}

## @name ::filer_file_append
## @summary Appends line(s) to a file on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg @lines Required: The lines to be added
sub filer_file_append ($$@) {
    my($fc) = shift;
    my $file = shift;
    my(@lines) = @_;

    # Handle the vfiler context - find the vfiler's root
    my $vfiler_context;
    # See if we are simulating a vfiler context
    my $simulated_context = $fc->{hostp}->{'vfiler_simulated_context'};
    if ($simulated_context eq '') {
        $vfiler_context = filer_vfiler_context_get($fc);
    } else {
        $vfiler_context = $simulated_context;
    }
    if ($vfiler_context ne 'vfiler0' and $vfiler_context ne '') {
        # Determine the path to the file
        my %vfiler_data = filer_vfiler_status_get($fc,$vfiler_context);
        if (defined($vfiler_data{'Path'})) {
            $file = $vfiler_data{'Path'}.$file;
        }
    }

    my $fh = $fc->{'hostp'};
    my $filer_ip = $fh->default_ip();
    my $filer_pw = $fh->default_password();

    # Be sure to quote line to be added
    my $cmd = "wrfile $file";
    if (! open(RSH, "|rsh $filer_ip -l root:$filer_pw $cmd")) {
        ExitError("couldn't open rsh socket to $filer_ip for $cmd");
    }

    foreach my $line (@lines) {
        if (! print(RSH $line)) {
            ExitError("couldn't print through rsh socket to $filer_ip for $cmd");
            close(RSH);        # ignore any additional errors
        }
        $Log->debug("Wrote to $file: $line");
    }

    if (! close(RSH)) {
        logresult('WARN', "couldn't close rsh socket to $filer_ip for $cmd");
    }
}

## @name ::filer_file_copy
## @summary Copies an existing file to another file
## @arg $fc Required: The filer connection object for the command line interface
## @arg $in_file Required: The file name including /vol/<volume>/ if not in the root directory.
## @arg $out_file Required: The file name including /vol/<volume>/ if not in the root directory.
## @arg $block_count Optional: The number of blocks to copy. Default is a very large number, to force all blocks.
sub filer_file_copy ($$$;$) {
    my $fc  = shift;
    my($in_file) = shift;
    my($out_file) = shift;
    my $count = shift;

    if (defined($count) ) {
        if ($count eq '') {
            $count = 999999999999999;
        }
    } else {
        $count = 999999999999999;
    }

    if (!filer_file_exists($fc,$out_file) ) {
        filer_file_make($fc,$out_file)
    }

    my $cmdname = "dd";
    my $cmd = "$cmdname if=$in_file of=$out_file count=$count";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,3600);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /$cmdname:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_file_destroy
## @summary Destroys an existing file
## @arg $fc Required: The filer connection object for the command line interface
## @arg $file Required: The file name including /vol/<volume>/ if not in the root directory.
sub filer_file_destroy ($$) {
    my $fc  = shift;
    my($file) = shift;

    my $cmdname = "rm";
    my $cmd = "$cmdname $file";
    #my($text,$result) = filer_cli_diag_say($fc,$cmd);
    my $text = filer_cmd(host=>$fc->hostp(), cmd=>"priv set diag; $cmd");
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /$cmdname:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_file_exists
## @summary Returns 1 if the specified file exists
## @arg $fc Required: The filer connection object for the command line interface
## @arg $file Required: The file name including /vol/<volume>/ if not in the root directory.
sub filer_file_exists ($$) {
    my $fc  = shift;
    my($file) = shift;

    my $cmdname = "file reservation";
    my $cmd = "$cmdname $file";
    my $text;

    eval{
    $text = filer_cmd(host=>$fc->hostp(), cmd=>"$cmd");
      };
    if ($text =~ "$cant find path") {
        return 0;
    }
    return 1;
}

## @name ::filer_file_make
## @summary Make a file returns 1 if the mkfile succeeds
## @arg $fc Required: The filer connection object for the command line interface
## @arg $file Required: The file name including /vol/<volume>/ if not in the root directory.
## @arg $size Optional: The byte count size requested for the file, default is 1024 i.e. 1kb, else size can also be specified in K/k[B/b], M/m[B/b], or G/g[B/b]; Decimal values are also accepted.
sub filer_file_make ($$;$) {
    my $fc  = shift;
    my($file) = shift;
    my($size) = shift;

    if (!defined($size) || ($size eq '')){
    	$size = 1024;
    }elsif($size =~ /^\d+(\.\d+)?gb?$/i){
    	# Checking for files of sizes in Gb
   		$size*=1024*1024*1024;
    }elsif($size =~ /^\d+(\.\d+)?mb?$/i){
    	# Checking for files of sizes in Mb
    	$size*=1024*1024;
    }elsif($size =~ /^\d+(\.\d+)?kb?$/i){
    	# Checking for files of sizes in Kb
    	$size*=1024;
    }elsif($size=~ /^\d+(\.\d+)?$/){
    	# do nothing
    }else{
    	# Any other input variables are considered invalid
    	my $msg ="File size Invalid";
    	$Log->warn($msg);
    	return 0;
    }


    # Define a larger timeout for larger files
    my $timeout = 60 + $size/10000;

    my $cmdname = "mkfile";
    my $cmd = "$cmdname $size $file";



    #my($text,$result) = filer_cli_diag_say($fc,$cmd,$timeout);
    my $text = filer_cmd(host=>$fc->hostp(), cmd=>"priv set diag; $cmd", timeout=>$timeout); 


    if ($text =~ /(write|file creation) failed/) {
        return 0;
    }
    return 1;
}


## @name ::filer_file_read
## @summary Gets the file contents as an array or text string
## @arg $fc Required: The filer connection object for the command line interface
## @arg $file Required: The file name including /vol/<volume>/ if not in the root directory.
sub filer_file_read ($$) {
    my($fc) = shift;
    my($file) = shift;
    my $text;
    my $result;

    if (my $server = are_we_ontapng($fc)) {
        $text = $server->rd_file(file=>$file);
    }
    else {
        # As some may use this to read message logs, we need to ignore panic string detection
        ($text,$result) = filer_cli_diag_say($fc,"rdfile $file",{retry=>'ok',ignore=>'panic'});
    }
    if (wantarray) {
        my @info = filer_cli_split($text);
        return @info;
    } else {
        return $text;
    }
}

## @name ::filer_file_space_reserve_disable
## @summary Turn off space reservations for a file
## @arg $fc Required: The filer connection object for the command line interface
## @arg $path Required: The file name including its path
sub filer_file_space_reserve_disable ($$) {
    my $fc  = shift;
    my $path = shift;

    my $cmd = "file reservation $path disable";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        #space reservations for file /vol/blocks_fvt/cbh: off
        if ($line =~ /space reservations for file.*:\s+off/) {
            return;
        } elsif ($line =~ /file reservation:|file:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_file_space_reserve_enable
## @summary Turn on space reservations for a file
## @arg $fc Required: The filer connection object for the command line interface
## @arg $path Required: The file name including its path
sub filer_file_space_reserve_enable ($$) {
    my $fc  = shift;
    my $path = shift;

    my $cmd = "file reservation $path enable";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        #space reservations for file /vol/blocks_fvt/cbh: on
        if ($line =~ /space reservations for file.*:\s+on/) {
            return;
        } elsif ($line =~ /file reservation:|file:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_flexcache_delegations
## @summary Get flexcache delegations from filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $cache Required: The flexcache volume to check for delegations
## @return An array of file ids representing the delegations held. Array size
##         is the number of delegations held.
sub filer_flexcache_delegations ($$) {
    my ($delegations, @delegations);
    my $fc = shift;
    my $cache = shift;

    my $cmd = "priv set -q test; flexcache delegations $cache";
    my ($text, $result) = filer_cli_say($fc, $cmd, {retry=>'ok'});

    # Empty string means no delegations found, return empty array
    return @delegations if ($text eq '');

    if ($text =~ /not a mapped FlexCache volume/) {
        ExitError("$cache is not a mapped FlexCache volume");
    }

    my @lines = split("\n", $text);
    foreach my $line (@lines) {
        if ($line =~ /Locked delegations \((\d+) total\)/) {
            # Get the number of delegations as reported by command
            $delegations = $1;
        } else {
            # Add file ids to list of delegations
            foreach my $file (split(' ', $line)) {
                push @delegations, $file;
            }
        }
    }

    # Verify array contains correct number of entries
    if (scalar(@delegations) != $delegations) {
        ExitError("File id count does not match return string");
    }

    return @delegations;
}

## @name ::filer_flexcache_eject
## @summary Eject a file from a flexcache
## @arg $fc Required: The filer connection object for the command line interface
## @arg $file Required: Filer path of file to be ejected
## @return 1 if success, 0 otherwise
sub filer_flexcache_eject ($$) {
    my $fc = shift;
    my $file = shift;

    my $cmd = "priv set -q test; flexcache eject $file";
    my ($text, $result) = filer_cli_say($fc, $cmd);

    if ($text =~ /$file: ejected/) {
        return 1;
    }

    if ($text =~ /Object does not exist/ || $text =~ /not a FlexCache volume/) {
        ExitError($text);
    }

    return 0;
}

## @name ::filer_get_vol_name_from_path
## @summary Returns the volume name from the path
## @arg $vol Required: The volume name
sub filer_get_vol_name_from_path ($) {
    my $arg = shift;
    if($arg =~ /\//){
        my @out = split(/\//,$arg);
        return ($out[2]);
    }
    return $arg;
}

## @name ::filer_if_alias
## @summary Create an IP alias for the specified filer interface
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: The filer interface name
## @arg $ipaddr Required: an IP address in dot octet format
## @arg $netmask Optional: a netmaks value in dot octet format, 255.255.255.0 is the default
sub filer_if_alias ($$$;$) {
    my $fc  = shift;
    my $interface = shift;
    my $ipaddr =shift;
    my $netmask =shift;
    if (!defined($netmask)) {
        $netmask = '255.255.255.0';
    }

    my $cmd = "ifconfig $interface -alias $ipaddr netmask $netmask";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /ifconfig:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_if_down
## @summary Take a filer interface down
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: The filer interface name
sub filer_if_down ($$) {
    my $fc  = shift;
    my $interface = shift;

    my $cmd = "ifconfig $interface down";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /ifconfig:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_if_ip
## @summary Establish the IP address associated with a filer interface
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: The filer interface name
## @arg $ip Required: The filer IP address for this interface
## @arg $netmask Optional: The netmask, default is 0xffffff00
sub filer_if_ip ($$$;$) {
    my $fc  = shift;
    my $interface = shift;
    my $ip = shift;
    my $netmask = shift;

    if (!defined($netmask)) {
        $netmask = '0xffffff00';
    }

    my $cmd = "ifconfig $interface $ip netmask $netmask";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /ifconfig:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_if_ips_get
## @summary Returns a list of IP addresses for the specified filer interface
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: The filer interface name
sub filer_if_ips_get ($$) {
    my $fc  = shift;
    my $interface = shift;

    my @ips = ();

    my $cmd = "ifconfig -a";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $if = '';
    foreach my $line (@info) {
        if ($line =~ /ifconfig:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /^$interface:/) {
            $if = $interface;
        } elsif ($line =~ /^\S+:\s/) {
            # We have started a new interface description
            if ($if eq $interface) {
                # We have just finished the requested interface
                last;
            }
        } elsif ($if eq $interface) {
            # Only process output associated with the requested interface
            if($line =~ /^\s+inet\s+(\d+\.\d+\.\d+\.\d+)\s/) {
                #        inet 10.60.153.24 netmask 0xffffff00 broadcast 10.60.153.255
                @ips = (@ips,$1);
            }
        }
    }
    return @ips;
}

## @name ::filer_if_mtu_set
## @summary Set the MTU for the specified filer interface which may or may not
## succeed depending on the MTU supported by the specified interface.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: The filer interface name
## @arg $mtu Required: mtu size in bytes to set for filer interface
## @return 1 if setting the MTU succeeded
##         0 if setting the MTU failed
sub filer_if_mtu_set ($$$) {
    my $fc  = shift;
    my $interface = shift;
    my $mtu =shift;

    my $cmd = "ifconfig $interface mtusize $mtu";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /ifconfig:/) {
            return 0;
        }
    }
    return 1;
}

## @name ::filer_if_status
## @summary Determine filer interface up or down status
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: The filer interface name
sub filer_if_status ($$) {
    my $fc  = shift;
    my $interface = shift;

    my $cmd = "ifconfig $interface";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my $status = 'unknown';
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /ifconfig:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ m/\(.*\)/) {
            $status = $line;
            $status =~ s/.*\((.*)\).*/$1/;
            $status =~ s/.*-//g;
            last;
        }
    }
    return $status;
}

## @name ::filer_if_up
## @summary Bring a filer interface up
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: The filer interface name
sub filer_if_up ($$) {
    my $fc  = shift;
    my $interface = shift;

    my $cmd = "ifconfig $interface up";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /ifconfig:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_if_netmask_get
## @summary Return filer netmask if possible or default of 255.255.255.0/24
## @arg $fc Required: The filer connection object for the command line interface
## @arg $fc_ipaddr Required: The IP address of the interface to get the
##      netmask for.
## @return netmask in format xxx.xxx.xxx.xxx/yy
## @notes IP V4 only
sub filer_if_netmask_get ($$) {
    my $fc  = shift;
    my $fc_ipaddr = shift;
    my $netmask = "255.255.255.0/24";

    my $cmd = "ifconfig -a";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /ifconfig:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ / $fc_ipaddr /) {
            # Only process output associated with the filers IP address
            if($line =~ /\snetmask\s0x(........)\sbroadcast/) {
                # inet 10.60.153.24 netmask 0xffffff00 broadcast 10.60.153.255

                # convert netmask from hex string to binary rep to ascii form
                my $netmask_bin = pack('H8', $1);
                my $mask = inet_ntoa($netmask_bin);
                # Convert netmask to binary form "1111...10000"
                my $netmask_bits = unpack('B32', $netmask_bin);
                # Append subnet bits
                $netmask = "$mask/" . ($netmask_bits =~ tr/1/1/);
            }
        }
    }
    return $netmask;
}

## @name ::filer_ifinfo_vi_check
## @summary Check 'ifinfo $FILER_VI_DEV' output
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: VI Device ($FILER_VI_DEV)
## @arg $all Optional: If defined, run with the -a option instead of interface
sub filer_ifinfo_vi_check ($$;$) {
    my $fc = shift;
    my $interface = shift;
    my $all = shift;
    my $options;
    if (!defined($all)) {
        $options = "$interface";
    } else {
        $options = "-a";
    }

    my $cmd = "ifinfo $options";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    my $found_interface = 0;
    my $interface_checks = 0;
    foreach my $line (@info) {
        if ($line =~ /-- interface /) {
            if ($line =~ /interface  $interface/) {
               $found_interface = 1;
            } else {
               $found_interface = 0;
            }
        }
        if ($found_interface == 1) {
            # Let's make sure all the sections are here
            ($line =~ /MAC address/) and $interface_checks++;
            ($line =~ /Driver name: VI\/IP/) and $interface_checks++;
            ($line =~ /NIC Soft State Structure/) and $interface_checks++;
            # Should be 2 matches for status
            ($line =~ /Status.*RUNNING/) and $interface_checks++;
            ($line =~ /Command Queue/) and $interface_checks++;
            ($line =~ /Interrupt Queue/) and $interface_checks++;
            ($line =~ /NIC info from EEPROM/) and $interface_checks++;
            ($line =~ /VI Soft State Structure/) and $interface_checks++;
            ($line =~ /VI Event Queue/) and $interface_checks++;
            ($line =~ /TPT page-list/) and $interface_checks++;
            ($line =~ /Attributes/) and $interface_checks++;
            ($line =~ /Statistics/) and $interface_checks++;
            ($line =~ /VI\/IP Exception Report Begin/) and $interface_checks++;
        }
    }
    logcomment("interface_checks are $interface_checks");
    if ($interface_checks == 14) {
        return 1;
    } else {
        return 0;
    }
}

## @name ::filer_ifstat_get
## @summary Return a hash of the ifstat output
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: The interface to get stats for
## @arg verbose Optional: Run ifstat with the verbose flag.
sub filer_ifstat_get ($$;%) {
    my ($fc, $interface, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR }
        ],
        {
            verbose => { type => SCALAR, default => 0 },
        }
    );
    my ($cmd, @fields, $label, %ret, $section, $val, $vflag);

    if ($opts{verbose}) {
        $vflag = '-v';
    } else {
        $vflag = '';
    }
    $cmd = "ifstat $vflag $interface";
    #my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});
    my($text,$result) = filer_cli_advanced_say($fc,$cmd,{retry=>'ok'});
    my @info = filer_cli_split($text);
    my $found_interface = 0;
    %ret = ();
    foreach my $line (@info) {
        if ($line =~ /-- interface /) {
            if ($line =~ /interface  $interface/) {
                $found_interface = 1;
            } else {
                $found_interface = 0;
            }
        }
        if ($found_interface == 1) {
            # Set the section string
            if ($line =~ /(RECEIVE|TRANSMIT|DEVICE|LINK_INFO|RSS|Queue 0|Queue 1|Queue 2|Queue 3Queue 4)/) {
                $section = $1;
                next;
            }
            # Do not start saving data until first section encountered
            next if (!defined($section));
            @fields = split(/\|/, $line);
            foreach $line (@fields) {
                $line =~ s/\s*?$//;
                $line =~ /\s+(.*):\s+(.*)$/;
                $label = $1;
                $val = $2;
                $label =~ s/ /_/g;
                # Convert a numeric string to a number if the number
                # has a unit suffix
                my ($num, $unit) = $val =~ /^(\d+)\s*(t|g|m|k)*$/;
                if (defined($unit)) {
                    if ($unit eq 't') {
                        $num *= 1000000000000;
                    } elsif ($unit eq 'g') {
                        $num *= 1000000000;
                    } elsif ($unit eq 'm') {
                        $num *= 1000000;
                    } elsif ($unit eq 'k') {
                        $num *= 1000;
                    }
                }
                $val = $num if (defined($num));
                $ret{$section}{$label} = $val;
            }
        }
    }
    return (\%ret);
}

## @name ::filer_ifstat_vi_check
## @summary Check 'ifstat $FILER_VI_DEV' output
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: VI Device ($FILER_VI_DEV)
## @arg $all Optional: If defined, run with the -a option instead of interface
sub filer_ifstat_vi_check ($$;$) {
    my $fc = shift;
    my $interface = shift;
    my $all = shift;
    my $options;
    if (!defined($all)) {
        $options = "$interface";
    } else {
        $options = "-a";
    }

    my $cmd = "ifstat $options";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    my $found_interface = 0;
    my $interface_checks = 0;
    foreach my $line (@info) {
        if ($line =~ /-- interface /) {
            if ($line =~ /interface  $interface/) {
                $found_interface = 1;
            } else {
                $found_interface = 0;
            }
        }
        if ($found_interface == 1) {
            # Let's make sure all the sections are here
            ($line =~ /RECEIVE/) and $interface_checks++;
            ($line =~ /TRANSMIT/) and $interface_checks++;
            ($line =~ /DEVICE/) and $interface_checks++;
            ($line =~ /LINK_INFO/) and $interface_checks++;
        }
    }
    if ($interface_checks == 4) {
        return 1;
    } else {
        return 0;
    }
}

## @name ::filer_ifstat_zero
## @summary Check 'ifstat -z <interface>' output
## @arg $fc Required: The filer connection object for the command line interface
## @arg $interface Required: Interface to zero ifstat's on
sub filer_ifstat_zero ($$) {
    my $fc = shift;
    my $interface = shift;

    my $cmd = "ifstat -z $interface";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $return = 0;
    foreach my $line (@info) {
        if ($line =~ /-- interface  $interface/) {
            $return=1;
        }
    }
    return $return;
}

## @name ::filer_ip_ping
## @summary "ping" an IP address from a filer, returns 1 if alive
## @arg $fc Required: The filer connection object for the command line interface
## @arg $ip Required: IP address
sub filer_ip_ping ($$) {
    my $fc = shift;
    my $ip = shift;

    my $cmd = "ping $ip 2";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $return = 0;
    foreach my $line (@info) {
        if ($line =~ / is alive/) {
            $return=1;
            last;
        }
    }
    return $return;
}

## @name ::filer_ipfw_disable
## @summary Disables the use of ipfw
## @arg $fc Required: The filer connection object for the command line interface
## @return 1 if success, 0 otherwise
sub filer_ipfw_disable ($) {
    my $fc = shift;

    filer_options_set($fc, 'ipfw.enable', 'off');
    my $return = filer_options_get($fc, 'ipfw.enable');
    if ($return eq 'off') {
        return 1;
    } else {
        return 0;
    }
}

## @name ::filer_ipfw_enable
## @summary Enables the use of ipfw
## @arg $fc Required: The filer connection object for the command line interface
## @return 1 if success, 0 otherwise
sub filer_ipfw_enable ($) {
    my $fc = shift;

    filer_options_set($fc, 'ipfw.enable', 'on');
    my $return = filer_options_get($fc, 'ipfw.enable');
    if ($return eq 'on') {
        return 1;
    } else {
        return 0;
    }
}

## @name ::ipfw_rule_add
## @summary Add an ipfw rule (see ipfw help for more information)
## @arg $fc Required: The filer connection object for the command line interface
## @arg $rule_id Required: An ipfw rule id number (1..65534)
## @arg $action Required: Specifies to 'allow' or 'deny' access
## @arg $proto Required: Protocol to which the rule will be applied
## @arg $from Required: From client identifier, may be 'any' or an ip address
## or a hostname
## @arg $to Required: To client identifier, my be 'any' or an ip address
## or a hostname
## @arg $options Optional: Additional parameters passed to the ipfw add command.
## @return 1 if success, 0 otherwise
sub filer_ipfw_rule_add($$$$$$;$) {
    my $fc      = shift;
    my $rule_id = shift;
    my $action  = shift;
    my $proto   = shift;
    my $from    = shift;
    my $to      = shift;
    my $options = shift;

    my $cmd = "priv set -q test;" .
              "ipfw add $rule_id $action $proto from $from to $to $options";

    my $return = filer_cli_say($fc, $cmd,{retry=>'ok'});
    if ($return =~ /$rule_id $action $proto from $from to $to/) {
        return 1;
    } else {
        return 0;
    }
}

## @name ::ipfw_rule_delete
## @summary Delete an ipfw rule
## @arg $fc Required: The filer connection object for the command line interface
## @arg $rule_id Required: The numerical identifier of the rule to delete
## @return 1 if success, 0 otherwise
sub filer_ipfw_rule_delete($$) {
    my $fc = shift;
    my $rule_id = shift;

    my $cmd = "priv set -q test;" .
              "ipfw delete $rule_id";

    my $return = filer_cli_say($fc, $cmd,{retry=>'ok'});
    # The 'setsockopt(IP_FW_DEL)' return value happens when we try to delete
    # a rule number that does not exist or has already been deleted.
    # Consider this success.
    if ($return eq '' || $return =~ /rule $rule_id: setsockopt(IP_FW_DEL)/) {
        return 1;
    } else {
        return 0;
    }
}

## @name ::filer_kernel_is_debug
## @summary Returns 1 if the filer's ONTAP kernel is a "debug" build.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_kernel_is_debug ($) {
    my $fc  = shift;

    my $version = filer_version_get($fc);
    if ($version =~ m/<.*d.*>/) {
        return 1;
    }
    return 0;
}

## @name ::filer_lock_status_file_get
## @summary Get the lock status of the specified file
## @arg $fc Required: The filer connection object.
## @arg file Optional: The file to get the lock status.
## If this option is not given, the lock status for all files is returned.
## @arg partner Optional: Check lock status of partner in a cluster pair.
## @return  List of locks:
##    dev     => Device id of file,
##    inode   => Inode of file,
##    ipaddr  => IP address of client holding lock,
##    host    => Hostname of client holding lock,
##    proto   => Protocol of lock,
##    mode    => Mode of lock,
##    sbyte   => Starting byte of range lock,
##    blength => Length of range lock,
##    ltype   => Lock type (read: 0, write 1),
##    owner   => Owner of lock,
##    path    => Full path of file (oplock only),
##    oplock  => Type of oplock granted,
##    excl    => Exclusive lock (yes, no),
##    pid     => Process id of lock owner,
##    state   => State of lock.
sub filer_lock_status_file_get ($;%) {
    my ($fc, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT }
        ],
        {
            file    => { type => SCALAR,  optional => 1 },
            partner => { type => BOOLEAN,  default => 0 },
        }
    );
    my (@locks, $dev, $inode, $file, $text, $cmd, $op);
    # File to get lock status for
    $file = $opts{file} if defined($opts{file});
    # Get the lock status from the partner filer instead
    $cmd  = $opts{partner} ? 'partner ' : '';

    # Get the lock status of file from filer
    $cmd .= "lock status -f $file";
    $text = filer_cli_say($fc, $cmd);

    # Process each line from the output of command
    foreach ( split /\r?\n/, $text ) {
        if ( /^===/ ) {
            # Get device id and inode
            ($dev, $inode) = /^=+(.*):(.*)/;
            $dev   = hex($dev);
            $inode = hex($inode);
        } elsif ( $dev and $inode ) {
            my (%h, @out);
            # Get lock protocol
            ($h{proto}) = /^(\w*)/;
            if ( $h{proto} eq 'NLM' ) {
                # Get host and owner holding the NLM lock
                ($h{host}, $h{owner}) = /\[([^,]+),(.*)\]/;
            } else {
                # Get client's ip address and owner holding the delegation
                if ( @out = /\[IP=(\d+\.\d+\.\d+\.\d+),(.*)\]:/ ) {
                    # This is a delegation
                    ($h{ipaddr}, $h{owner}) = @out;
                    ($h{state}) = /\]:\s+(\S+)/;
                } else {
                    # This is an oplock
                    ($h{ipaddr}) = /\shost=(\d+\.\d+\.\d+\.\d+)/;
                    ($h{path})   = /\spath=[^(]*\(([^)]*)/;
                    ($h{owner})  = /\sowner=(\S*)/;
                    ($h{state})  = /\sstate=(\S*)/;
                    ($h{host})   = /\shost=\S+\((.*)\)/;
                    # Get optional info
                    ($op) = /\soplock=(\S*)/;
                    $h{oplock} = $op if $op;
                    if (@out = /pid=(\S+) offset=(\S+) len=(\S+) excl=(\S+)/) {
                        ($h{pid}, $h{sbyte}, $h{blength}, $h{excl}) = @out;
                    }
                }
                # Get lock mode
                ($op) = /\bmode=(.*)\s/;
                $h{mode} = $op if $op;
            }
            if ( @out = /^(?:NFSv4|NLM).*\]: (\d+):(\d+) (\d+) (\S+)/ ) {
                ($h{sbyte}, $h{blength}, $h{ltype}, $h{state}) = @out;
            }
            %h = (%h, dev=>$dev, inode=>$inode);
            push @locks, \%h;
        }
    }
    return @locks;
}

## @name ::filer_machine_type_get
## @summary Returns the (IBM) machine type of the specified filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_machine_type_get ($) {
    my $fc = shift;

    my $cmd = "sysconfig -v 0";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $field = '';
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ m/\s+Machine Type:\s+(\S+)/) {
            $field = $1;
            last;
        }
    }
    return $field;
}

## @name ::filer_mem_stats_cleanup
## @summary Cleanup the copy of the stats (release memory)
## @arg $fc Required: The filer connection object for the command line interface
sub filer_mem_stats_cleanup ($) {
    my $fc = shift;

    my $cmd = "mem_stats -c";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    return $result;
}

## @name ::filer_mem_stats_diff
## @summary Compare the current mem_stats statistics to the copy and report differences
## @arg $fc Required: The filer connection object for the command line interface
sub filer_mem_stats_diff ($) {
    my $fc = shift;

    my $cmd = "mem_stats -e";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

   return $text;
}

## @name ::filer_mem_stats_dump
## @summary Dump the mem_stats statistics in /etc/log/mem_stats.log
## @arg $fc Required: The filer connection object for the command line interface
sub filer_mem_stats_dump ($) {
    my $fc = shift;

    my $cmd = "mem_stats -f";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    return $result;
}

## @name ::filer_mem_stats_save
## @summary Make a copy of the mem_stats statistics and save it
## @arg $fc Required: The filer connection object for the command line interface
sub filer_mem_stats_save ($) {
    my $fc = shift;

    my $cmd = "mem_stats -b";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    return $result;
}

## @name ::filer_model_get
## @summary Returns the model of the specified filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_model_get ($) {
    my $fc = shift;

    my $cmd = "sysconfig -v 0";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $field = '';
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /\s+Model Name:\s+(\w+)/) {
            $field = $1;
            last;
        }
    }
    return $field;
}

## @name ::filer_name
## @summary Returns the filer's host name
## @arg $fc Required: The filer connection object for the command line interface
sub filer_name ($) {
    my($fc) = shift;

    my $filer = $fc->hostp()->id();
    # Remove any extra host parameters to expose the filer host name
    $filer =~ s/;.*//;
    # Remove any domain name on the filer host name
    $filer =~ s/\..*//;
    return $filer;
}

## @name ::filer_netgroup_add
## @summary Adds/replaces given netgroup on the filer's /etc/netgroup file.
## @arg $fc Required: Filer's connection object.
## @arg $netgroup Required: Netgroup to add the clients to.
## @arg @clients Required: List of client names to add to netgroup.
## @return  1 if succeeds
sub filer_netgroup_add($$\@;) {
    my ($fc, $netgroup, $clients) = validate_pos(@_,
        { type => OBJECT },
        { type => SCALAR },
        { type => ARRAYREF }
    );
    my (@cont, @tmp, $op);

    # Read current contents of the netgroup file
    @cont = filer_file_read($fc, NETGROUP_FILE);

    my $file_re = quotemeta(NETGROUP_FILE);

    foreach (@cont) {
        next if /^$file_re:/;
        s/\s*$//;	# Get rid off trailing spaces
        if ( /\\$/ ) {
            # Found '\' at the end of line, multi-line continues.
            chop;			# Get rid off '\' at the end.
            push @tmp, $_;
            next;
        } elsif (@tmp) {
            # This is the last line of a multi-line netgroup.
            push @tmp, $_;
            $_ = join("", @tmp);	# Merge all lines in multi-line
        }
        if ( /^$netgroup\b/ ) {
            # Netgroup found, remove it
            undef @tmp;		# Clear multi-line buffer
            next;
        }
        if ( @tmp ) {
            # Multi-line netgroup which must be kept
            $op .= "$_\n" foreach @tmp;
        } else {
            # Single-line netgroup which must be kept
            $op .= "$_\n";
        }
    }

    $op .= "$netgroup";
    foreach (@$clients) {
        $op .= " ($_,,)";
    }
    $op .= "\n";

    # Write the contents of the netgroup file
    filer_file_append($fc, NETGROUP_FILE, $op);

    return 1;
}

## @name ::filer_netstat_get
## @summary Calls 'netstat -a' on a filer
## @arg $fc Required: Filer's connection object.
## @arg $nopt Optional: If set, show network addresses as numbers
## @return A hashref containing the result of the netstat -a call. Keys are
##         based on protocol, values are arrays of hashes for each entry.
sub filer_netstat_get ($;$) {
    my $fc      = shift;
    my $nopt    = shift;
    my ($cmd, $return, @ret_arr, @columns, $group, $columns, $line, %netstat_h);

    # Generate the command string
    $cmd = "netstat -a";
    if (defined($nopt)) {
        $cmd .= 'n';
    }

    $return = filer_cli_say($fc, $cmd, {retry=>'ok'});

    @ret_arr = split(/\r*\n/, $return);

    # Drop all lines until a line beginning with "Active..." is found.
    # Sometimes after a reboot, there might be additional lines before the
    # actual command result. We need to ignore these.
    while ($line = shift @ret_arr) {
        if ($line =~ /^Active/) {
            $group = $line;
            last;
        }
    }

    foreach $line (@ret_arr) {
        # A blank line indicates start of a new group, reset key variables
        if ($line eq '') {
            # Reset variables
            undef $group;
            undef $columns;
            @columns = ();
            next;
        }

        # Make sure we know the group we are working with
        if (!defined $group) {
            if ($line =~ /^Active/) { # Group line begins with "Active"
                $group = $line;
            } else {
                ExitError("filer_netstat_get: output does not match " .
                          "group header pattern: <$line>");
            }
        # Once we have the group, the next line should be the column headers
        } elsif (!defined $columns) {
            # Column header line contains "Local Address", "Remote Address",
            # and a variable number of other headers
            if ($line =~ /^Local Address/) {
                push @columns, 'Local Address';
                $line =~ s/^Local Address\s+//;
                $columns++;
                if ($line =~ s/^Remote Address\s+//) {
                    push @columns, 'Remote Address';
                    $columns++;
                }
                while ($line =~ s/(\S+)\s*//) {
                    push @columns, $1;
                    $columns++;
                }
            } else {
                ExitError("filer_netstat_get: could not determine " .
                          "column headings: <$line>");
            }
        # Then we parse the lines accordingly
        } else {
            # Each line should contain the same number of columns as the
            # column header line for this group
            my %data_h;
            for (0..($columns-1)) {
                $line =~ s/(\S+)\s*//;
                $data_h{$columns[$_]} = $1;
            }
            if ($line ne '') {
                ExitError("filer_netstat_get: data line has too many columns");
            }
            # Put the hash of data onto the netstat hash
            push @{$netstat_h{$group}}, \%data_h;
        }
    }

    # Return the reference to the hash, or undef if hash is empty
    return (%netstat_h) ? \%netstat_h : undef;
}

## @name ::filer_netstat_if_get
## @summary Calls 'netstat -i' or 'netstat -I <if>' on a filer
## @arg $fc Required: Filer's connection object.
## @arg $if Optional: Interface to call netstat against.
## @return A hashref containing the result of the 'netstat -i' call. Keys are
##         based on protocol, values are arrays of hashes for each entry.
sub filer_netstat_if_get ($;$) {
    my $fc = shift;
    my $if = shift;
    my ($cmd, $return, @ret_arr, $line, @hdr_elems, @line_elems,
        $if_name, %netstat_h);

    # Generate the command string
    if (defined $if) {
        # Use capital I and send the interface name
        $cmd = "netstat -I $if";
    } else {
        # Use lowercase i and parse all the interfaces
        $cmd = "netstat -i";
    }

    $return = filer_cli_say($fc, $cmd, {retry=>'ok'});

    if ($if && $return =~ /No such interface/) {
        ExitError("filer_netstat_if_get: Invalid interface specified: $if");
    }

    @ret_arr = split(/\r*\n/, $return);

    # Make sure we find the heading line
    while ($line = shift @ret_arr) {
        if ($line =~ /^Name/) {
            @hdr_elems = split(/\s+/, $line);
            # We don't need to keep the 'Name' header
            shift @hdr_elems;
            last;
        }
    }
    if (!@hdr_elems) {
        ExitError("filer_netstat_if_get: could not find header line in " .
                  "reply: <$return>");
    }

    foreach $line (@ret_arr) {
        # Split the line and create a hash entry from it
        @line_elems = split(/\s+/, $line);

        # The first element on the line is the if name
        $if_name = shift @line_elems;

        # Add each remaining element to the return hash
        for my $idx (0..$#line_elems) {
            $netstat_h{$if_name}{$hdr_elems[$idx]} = $line_elems[$idx];
        }
    }

    # Return the reference to the hash, or undef if hash is empty
    return (%netstat_h) ? \%netstat_h : undef;
}

## @name ::filer_nfs_initialize
## @summary Licenses and starts NFS service on the filer if possible
## @description Returns 1 if service is now started, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_nfs_initialize ($) {
    my($fc) = shift;

    my($initialized)=0;
    filer_service_license($fc,'nfs') if (!filer_service_is_licensed($fc,'nfs'));
    filer_nfs_start($fc) if (!filer_nfs_is_started($fc,));
    $initialized=1;
    return $initialized;
}

## @name ::filer_nfs_is_started
## @summary Determines if NFS is started on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_nfs_is_started ($) {
    my($fc) = shift;
    return filer_service_is_started($fc,'nfs');
}

## @name ::filer_nfs_start
## @summary Starts NFS service on the filer if possible
## @description Returns 1 if service is now started, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_nfs_start ($) {
    my($fc) = shift;
    return filer_service_start ($fc,'nfs');
}

## @name ::filer_nfs_stop
## @summary Stops NFS service on the filer if possible
## @description Returns 1 if service is already stopped, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_nfs_stop ($) {
    my($fc) = shift;
    return filer_service_stop ($fc,'nfs');
}

## @name ::filer_nvram_fw_revision2_get
## @summary Returns the NVRAM secondary microcode revision number of the specified filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_nvram_fw_revision2_get ($) {
    my $fc = shift;

    my $cmd = "sysconfig -v";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $field = '';
    my $found_slot = 0;
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /\s+slot\s.* NVRAM III/) {
            # NVRAM III does not support this field
            last;
        } elsif ($line =~ /\s+slot\s.* NVRAM /) {
            $found_slot = 1;
        }
        if ($found_slot eq 1) {
            if ($line =~ /^\s+Secondary Microcode Revision:/) {
                $found_slot = 1;
                $field = $line;
                $field =~ s/^\s+Secondary Microcode Revision:\s+(\w+).*/$1/;
                last;
            }
        }
    }
    return $field;
}
#        slot 7: NVRAM (NetApp NVRAM IV)\r\n
#                Revision:           B0\r\n
#                Serial Number:      46209\r\n
#                Memory Size:        256 MB\r\n
#                Battery Status:      OK\r\n
#                Primary Microcode Revision: 0114\r\n
#                 Microcode Revision: 0114\r\n

## @name ::filer_nvram_fw_revision_get
## @summary Returns the NVRAM revision number of the specified filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_nvram_fw_revision_get ($) {
    my $fc = shift;

    my $cmd = "sysconfig -v";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $field = '';
    my $found_slot = 0;
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /\s+slot\s.* NVRAM III/) {
            # NVRAM III does not support this field
            last;
        } elsif ($line =~ /\s+slot\s.* NVRAM /) {
            $found_slot = 1;
        }
        if ($found_slot eq 1) {
            if ($line =~ /^\s+Primary Microcode Revision:/) {
                $found_slot = 1;
                $field = $line;
                $field =~ s/^\s+Primary Microcode Revision:\s+(\w+).*/$1/;
                last;
            }
        }
    }
    return $field;
}

## @name ::filer_nvram_revision_get
## @summary Returns the NVRAM revision number of the specified filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_nvram_revision_get ($) {
    my $fc = shift;

    my $cmd = "sysconfig -v";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $field = '';
    my $found_slot = 0;
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /\s+slot\s.*NVRAM Revision ID:/) {
            $found_slot = 1;
        }
        if ($found_slot eq 1) {
            if ($line =~ /\s+Revision:/) {
                $found_slot = 1;
                $field = $line;
                $field =~ s/^\s+Revision:\s+(\w+).*/$1/;
                last;
            }
        }
    }
    return $field;
}

#                Revision:           A0\r\n
#                Serial Number:      50706\r\n
#                Memory Size:        128 MB\r\n
#                Battery Status:      OK\r\n

## @name ::filer_options_get
## @summary Get the value for option given
## @arg $fc Required: The filer connection name where this function is executed
## @arg $option Required: The name of the option
sub filer_options_get ($$) {
    my($fc) = shift;
    my ($option) = shift;

    my $cmd = "options $option";
    #my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    my $text = filer_cmd(host=>$fc->hostp(), cmd=>$cmd);

    my $value = "";
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /$option /) {
            (undef, $value) = split(/\s+/, $line);
        }
    }
    return $value;
}

## @name ::filer_options_set
## @summary Set the option to the given value
## @arg $fc Required: The filer connection name where this function is executed
## @arg $option Required: The name of the option
## @arg $value Required: The value to set the option to
## @arg $verify Optional: set to 0 to skip verification that value was set
## @return 0 if verification determined the set operation failed
sub filer_options_set ($$$;$) {
    my($fc) = shift;
    my ($option) = shift;
    my ($value) = shift;
    my ($verify) = shift;

    if (!defined($verify)) {
        $verify = 1;
    }

    my $status = 1;

    # Define list of options that require diag privilege
    my @diag_options = qw(ipfw);
    # Determine the name for the group of options
    my $option_group = $option;
    $option_group =~ s/\..*//g;
    # See if the input option is in a diag option group
    my $diag_option = grep(/^$option_group$/,@diag_options);

    my $cmd = "options $option $value";
    #my ($text,$result);
    my $hostrec = $fc->hostp();
    if ($diag_option) {
      #    ($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});
           filer_cmd(host=>$hostrec, cmd=>"priv set diag; $cmd");
    } else {
      #    ($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
           filer_cmd(host=>$hostrec, cmd=>"$cmd");
    }

    if ($verify) {
        #Verify values have been changed
        my $checkvalue = filer_options_get($fc, $option);
        # passwords are not displayed except for ldap.base.passwd
        if ($option =~ m/\.passwd$/ and $option !~ m/^ldap\.base\.passwd$/) {
            if ($checkvalue !~ m/^\*+$/) {
                ExitSoftError("Option $option displays as $checkvalue, but should have been displayed as all asterisks");
                $status = 0;
            }
        } else {
            if ($value =~ /^off$/i) {
                if ($checkvalue !~ m/^off$|^$/i ) {
                    # some variables disappear when set to 'off'
                    ExitSoftError("$option is $checkvalue, but should have been removed since it was set to $value");
                    $status = 0;
                }
            } elsif ($value ne $checkvalue) {
                ExitSoftError("$option is $checkvalue, but should have been changed to $value");
                $status = 0;
            } else {
                # Note that some diag options do not get saved in the registry
                if (!$diag_option) {
                    if (filer_crosscheck_is_enabled()) {
                        # Verify values have also been changed in registry
                        my $registrycheckvalue = filer_registry_get($fc, $option);
                        if ($value ne $registrycheckvalue) {
                            ExitSoftError("Option $option is now $registrycheckvalue, but should have been changed to $value");
                            $status = 0;
                        }
                    }
                }
            }
        }
    }
    return $status;
}

## @name ::filer_panic
## @summary Panics the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_panic ($) {
    my($fc) = shift;
    say ($fc, "priv set advanced; panic",-blind);
    sleep (320);
}

## @name ::filer_ping_seq_time_get
## @summary Get array of icmp_seq or transaction time for ping -s host(or ipaddrss)
## @example filer_ping_seq_time_get($fc, "192.168.2.5", "SEQ", 60)
## @arg $fc Required: The filer connection object for the command line interface
##    should be telnet or pm_cons for ping to work
## @arg $target: target host name or ip address
## @arg $op Required: SEQ (for icmp_seq) or TIME (for transaction time) you requested
## @arg $test_time Optional: the time to run your ping. default is 60
sub filer_ping_seq_time_get ($$$;$) {
    my $fc  = shift;
    my $target =shift;
    my $seq_or_time = shift;
    my $test_time = shift;
    $test_time = 60 unless (defined($test_time));

    my @resultarray =();
    my $cmd = "ping -s $target";
    my $end =time() + $test_time;

    # just send ping, don't wait for return
    say($fc,$cmd, -blind);

    my $line;
    while (time()< $end) {
        $line = say ($fc, "",".*\n", -noprompt,-nocr, -timeout => ($test_time +60 ), -ok, $cmd);

        if(not defined($line) or ($line eq "")){
            tstlog(-fail, "Did not receive output of ping for $cmd");
        }

        if ($seq_or_time =~/SEQ/i) {
            if ($line =~ /icmp_seq=(\d+)/) {
  	        push (@resultarray, $1);
            }
        } elsif ($seq_or_time =~/TIME/i) {
            if ($line =~ /time=(\S+)/){
  	        push (@resultarray, $1);
            }
        }
    }
    # now send ctrl-C (ctrl-X for simulators) to stop the "ping".
    say($fc,"\003", -blind);  # send control-C to stop program

    return @resultarray;
}


## @name ::filer_power_cycle
## @summary Turns off power to a filer and turns it back on
## @arg $fc Required: The filer connection object for the command line interface
sub filer_power_cycle ($) {
    my($fc) = shift;
    host_power($fc);
}

## @name ::filer_power_off
## @summary Turns off power to a filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_power_off ($) {
    my($fc) = shift;
    host_power($fc,'off');
}

## @name ::filer_power_on
## @summary Turns on power to a filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_power_on ($) {
    my($fc) = shift;
    host_power($fc,'on');
}


## @name ::filer_processor_count_get
## @summary Returns the number of processors on the specified filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_processor_count_get ($) {
    my $fc = shift;

    my $slot = 0;

    my $cmd = "sysconfig -v $slot";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $field = '';
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /\s+Processors:/) {
            $field = $line;
            $field =~ s/^\s+Processors:\s+(\d+).*/$1/;
            last;
        }
    }
    return $field;
}

## @name ::filer_qtree
## @summary Gets the output from the qtree command
## @arg $fc Required: The filer connection object for the command line interface
sub filer_qtree ($) {
    my($fc) = shift;

    my $cmd = "qtree status";
    $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    my ($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);

    if (wantarray) {
        return @info;
    } else {
        return $text;
    }
}

## @name ::filer_qtree_count
## @summary Count the number of qtrees found
## @arg $fc Required: The filer connection object for the command line interface
## @arg $volume Optional: If this is supplied the function will return
## the number of qtrees for that volume only
sub filer_qtree_count ($;$) {
    my($fc) = shift;
    my($volume) = shift;

    my $count = 0;
    my %HashOfInfo = filer_qtree_get($fc);
    foreach my $vol (keys %HashOfInfo) {
        foreach my $tree (keys %{$HashOfInfo{$vol}}) {
            if(defined($volume)) {
                $volume=~ s#.*/(\w+)#$1#;
                if($vol eq "$volume") {
                    $count += $HashOfInfo{$vol}{$tree}{'Count'};
                }
            } else {
                $count += $HashOfInfo{$vol}{$tree}{'Count'};
            }
        }
    }

    return $count;
}

## @name ::filer_qtree_create
## @summary Creates a qtree
## @description Creates a qtree
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtree Required: The name of the qtree to create
## @arg $volume Optional: The name of the volume to create the qtree on.
## Defaults to the root volume.
sub filer_qtree_create ($$;$) {
    my($fc) = shift;
    my ($qtree) = shift;
    my ($volume) = shift;

    my $qtreepath;
    if (defined($volume)) {
        $volume=~ s#.*/(\w+)#$1#;
        if($qtree =~ /\//) {
            ExitSoftError("Qtree given ($qtree) contains a '/' and a volume arg was also given. If volume arg is supplied, qtree should not be a path");
            $qtreepath = "$qtree"
        } else {
            $qtreepath = "/vol/${volume}/$qtree"
        }
    } else {
        $qtreepath = "$qtree"
    }
    if (!filer_qtree_exists($fc,$qtree,$volume)) {
        # Only attempt to create it if it does not already exist
        # This avoids getting an error message:
        #qtree create: Can't create qtree (File exists).
        my $cmd = "qtree create $qtreepath";
        $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
        my($text,$result) = filer_cli_say($fc,$cmd);

        my($line);
        my @info = filer_cli_split($text);
        # assume it was successful since it doesn't give any output.
        foreach $line (@info) {
            if ($line =~ /qtree create:|invalid qtree name|$cant create qtree/) {
                ExitError("Qtree creation failed: $line");
            }
        }
    }
}

## @name ::filer_qtree_delete
## @summary Deletes a qtree
## @description Deletes a qtree using 'qtree delete'.
## Formerly used 'rm'. If rm functionality is needed, see
## filer_file_destroy.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtree Required: The name or full path of the qtree to delete
## @arg $volume Optional: The name of the volume that has the qtree.
## Cannot be specified if $qtree is a path. Defaults to vol0 if $qtree is
## not a path.
## @arg $force Optional: If set to anything that evaluates to true, will
## perform a recursive qtree delete (force all contents to be removed).
sub filer_qtree_delete ($$;$$) {
    my $fc = shift;
    my $qtree = shift;
    my $volume = shift;
    my $force = shift;

    my $qtreepath;
    if (defined($volume)) {
        $volume=~ s#.*/(\w+)#$1#;
        if($qtree =~ /\//) {
            ExitScriptError("Qtree given ($qtree) contains a '/' and a volume arg was also given. If volume arg is supplied, qtree must not be a path");
        } else {
            $qtreepath = "/vol/${volume}/$qtree";
        }
    } else {
        # If no volume is specified, $qtree can be either a full path
        # or a qtree name. If only a qtree name, then the qtree will
        # be removed from the root volume.
        $qtreepath = "$qtree";
        # A leading slash is needed.
        $qtreepath =~ s/^\/*/\//;
    }
    if ($force) { $force = '-f' };
    my $cmd = "qtree delete $force $qtreepath";
    my($text,$result) = filer_cli_diag_say($fc,$cmd);

    if ($text) {
        ExitError("Qtree deletion failed: $text");
        }

    if (filer_qtree_exists($fc, $qtree, $volume)) {
        ExitError("Qtree delete did not return an error, but qtree still exists");
    }
}

## @name ::filer_qtree_exists
## @summary Determines if qtree exists
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtree Required: The name of the qtree to check
## @arg $volume Optional: The name of the volume to check for the qtree
sub filer_qtree_exists ($$;$) {
    my($fc) = shift;
    my($qtree) = shift;
    my($volume) = shift;
    if (!defined($qtree) || $qtree eq '') {
        ExitScriptError("A qtree parameter is required");
    }
    if (!defined($volume)) {
        if ($qtree =~ "/") {
            $volume = $qtree;
            $volume =~ s#/vol/## ;
            $volume =~ s#/.*##g ;
            $qtree =~ s#^/vol/$volume/##;
            if ($qtree =~ m#/#) {
                print("Parsed Volume: '$volume', Qtree: '$qtree'\n");#DEBUG
                ExitScriptError("A qtree name cannot contain a '/'");
            }
        } else {
            # volume must be filer's root volume
            $volume = '';
        }
    } else {
        # remove /vol/ if present
        $volume =~ s#^/vol/## ;
    }

    my $return = 0;
    my %HashOfInfo = filer_qtree_get($fc);
    foreach my $vol (keys %HashOfInfo) {
        foreach my $tree (keys %{$HashOfInfo{$vol}}) {
            if ($HashOfInfo{$vol}{$tree}{'Name'} eq $qtree) {
                if ($vol eq $volume || $vol eq '') {
                    $return = 1;
                }
            }
        }
    }

    return $return;
}

## @name ::filer_qtree_get
## @summary Gets output of the qtree command
## @description Gets output of qtree command and returns hash of info
## @arg $fc Required: The filer connection object for the command line interface
## @arg $volume Optional: The name of the volume to check for the qtree
## @arg vols Optional: If true the qtree information of volume(s) is
##      returned as well (default is 0)
sub filer_qtree_get ($;%) {
    my ($fc, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT }
        ],
        {
            volume => { type => SCALAR, default => '' },
            vols   => { type => SCALAR, default => 0 },
        }
    );
    my $cmd = "qtree status $opts{volume}";
    $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my($line);
    my @info = filer_cli_split($text);
    my %HashOfInfo;
    my($vol, $style, $oplocks, $status);
    my $totalcount=0;
    my $oldvol="";
    foreach $line (@info) {
        my $qtree = "";
        if ($line =~ /Volume.*Tree.*Style|-------/) {next;}
        $line =~ s/\s+$//; # remove extra space at the end
        my @info = split(/\s+/, $line);
        if ( @info == 5 ) {
            # This is the qtree information
            ($vol, $qtree, $style, $oplocks, $status) = @info;
        } elsif ( @info == 4 and $opts{vols} ) {
            # This is the volume information
            ($vol, $style, $oplocks, $status) = @info;
        } else {
            # This is not a qtree and no vol
            # information must be returned
            next;
        }
        if($oldvol ne $vol) {$totalcount=0;}
        $oldvol=$vol;
        $HashOfInfo{$vol}{$qtree}{'Name'}=$qtree;
        $HashOfInfo{$vol}{$qtree}{'Style'}=$style;
        $HashOfInfo{$vol}{$qtree}{'Oplocks'}=$oplocks;
        $HashOfInfo{$vol}{$qtree}{'Status'}=$status;
        $HashOfInfo{$vol}{$qtree}{'Count'}=++$totalcount;
    }

    return %HashOfInfo;
}

## @name ::filer_qtree_oplocks_disable
## @summary Disable oblocks
## @description Disables Oplocks on qtree and returns 1 if successful
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtree Required: The name of the qtree to create
## @arg $volume Optional: The name of the volume to create the qtree on.
## Defaults to the root volume.
sub filer_qtree_oplocks_disable ($$;$) {
    my($fc) = shift;
    my ($qtree) = shift;
    my ($volume) = shift;

    my $qtreepath;
    if (defined($volume)) {
        $volume=~ s#.*/(\w+)#$1#;
        if($qtree =~ /\//) {
            ExitSoftError("Qtree given ($qtree) contains a '/' and a volume arg was also given. If volume arg is supplied, qtree should not be a path");
            $qtreepath = "$qtree"
        } else {
            $qtreepath = "/vol/${volume}/$qtree"
        }
    } else {
        $qtreepath = "$qtree"
    }
    my $return = filer_qtree_oplocks_set($fc, $qtreepath, "disable");
}

## @name ::filer_qtree_oplocks_enable
## @summary Enable oplocks
## @description Enable Oplocks on qtree and returns 1 if successful
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtree Required: The name of the qtree to enable oplocks
## @arg $volume Optional: The name of the volume to enable oplocks qtree on.
## Defaults to the root volume.
sub filer_qtree_oplocks_enable ($$;$) {
    my($fc) = shift;
    my ($qtree) = shift;
    my ($volume) = shift;

    my $qtreepath;
    if (defined($volume)) {
        $volume=~ s#.*/(\w+)#$1#;
        if ($qtree =~ /\//) {
            ExitSoftError("Qtree given ($qtree) contains a '/' and a volume arg was also given. If volume arg is supplied, qtree should not be a path");
            $qtreepath = "$qtree"
        } else {
            $qtreepath = "/vol/${volume}/$qtree"
        }
    } else {
       $qtreepath = "$qtree"
    }
    my $return = filer_qtree_oplocks_set($fc, $qtreepath, "enable");
}

## @name ::filer_qtree_oplocks_is_enabled
## @summary Determines if oplocks on qtree are enabled
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtree Required: The name of the qtree to check oplocks
## @arg $volume Optional: The name of the volume to check for the oplocks qtree
sub filer_qtree_oplocks_is_enabled ($$;$) {
    my($fc) = shift;
    my($qtree) = shift;
    my($volume) = shift;

    my $return = 0;
    my %HashOfInfo = filer_qtree_get($fc);
    foreach my $vol (keys %HashOfInfo) {
        foreach my $tree (keys %{$HashOfInfo{$vol}}) {
            if (($HashOfInfo{$vol}{$tree}{'Oplocks'} eq "enabled")  &&
               ($HashOfInfo{$vol}{$tree}{'Name'} eq $qtree)) {
                if(defined($volume)) {
                    $volume=~ s#.*/(\w+)#$1#;
                    if($vol eq "$volume") {
                        $return = 1;
                    }
                } else {
                    $return = 1;
                }
            }
        }
    }

    return $return;
}

## @name ::filer_qtree_oplocks_set
## @summary Disable or Enable oplocks
## @description Disables Oplocks on qtree and returns 1 if successful
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtreepath Required: The path including the qtree
## @arg $state Required: Enable or disable oplocks
sub filer_qtree_oplocks_set($$$) {
    my($fc) = shift;
    my ($qtreepath) = shift;
    my ($state) = shift;

    my $cmd = "qtree oplocks $qtreepath $state";
    $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    my($text,$result) = filer_cli_say($fc,$cmd);

    my($line);
    my @info = filer_cli_split($text);
    # assume it was successful since it doesn't give any output.
    my $return = 1;
    foreach $line (@info) {
        if ($line =~ /qtree:/) {
            $return=0;
        }
    }

    return $return;
}

## @name ::filer_qtree_security_mixed_set
## @summary Change security to mixed on qtree
## @description Change security to mixed on qtree and return 1 if succesfull
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtree Required: The name of the qtree to change security
## @arg $volume Optional: The name of the volume to change security
## Defaults to the root volume.
sub filer_qtree_security_mixed_set ($$;$) {
    my($fc) = shift;
    my ($qtree) = shift;
    my ($volume) = shift;

    my $qtreepath;
    if (defined($volume)) {
        $volume=~ s#.*/(\w+)#$1#;
        if($qtree =~ /\//) {
            ExitSoftError("Qtree given ($qtree) contains a '/' and a volume arg was also given. If volume arg is supplied, qtree should not be a path");
            $qtreepath = "$qtree"
        } else {
            $qtreepath = "/vol/${volume}/$qtree"
        }
    } else {
        $qtreepath = "$qtree"
    }
    my $return = filer_qtree_security_set($fc, $qtreepath, "mixed");
}

## @name ::filer_qtree_security_ntfs_set
## @summary Change security to ntfs on qtree
## @description Change security to ntfs on qtree and return 1 if succesfull
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtree Required: The name of the qtree to change security
## @arg $volume Optional: The name of the volume to change security
## Defaults to the root volume.
sub filer_qtree_security_ntfs_set ($$;$) {
    my($fc) = shift;
    my ($qtree) = shift;
    my ($volume) = shift;

    my $qtreepath;
    if (defined($volume)) {
        $volume=~ s#.*/(\w+)#$1#;
        if($qtree =~ /\//) {
            ExitSoftError("Qtree given ($qtree) contains a '/' and a volume arg was also given. If volume arg is supplied, qtree should not be a path");
            $qtreepath = "$qtree"
        } else {
            $qtreepath = "/vol/${volume}/$qtree"
        }
    } else {
        $qtreepath = "$qtree"
    }
    my $return = filer_qtree_security_set($fc, $qtreepath, "ntfs");
}

## @name ::filer_qtree_security_set
## @summary Change security for qtree
## @description Changes security for qtree returns 1 if successful
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtreepath Required: The path including the qtree
## @arg $mode Required: unix, mixed, ntfs
sub filer_qtree_security_set($$$) {
    my($fc) = shift;
    my ($qtreepath) = shift;
    my ($mode) = shift;

    my $cmd = "qtree security $qtreepath $mode";
    $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    my($text,$result) = filer_cli_say($fc,$cmd);

    my($line);
    my @info = filer_cli_split($text);
    # assume it was successful since it doesn't give any output.
    my $return = 1;
    foreach $line (@info) {
        if ($line =~ /$cant find path/) {
            $return=0;
        }
    }

    return $return;
}

## @name ::filer_qtree_security_style_get
## @summary Determines if oplocks on qtree are enabled
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtree Required: The name of the qtree to check oplocks
## @arg $volume Optional: The name of the volume to check for the oplocks qtree
sub filer_qtree_security_style_get ($$;$) {
    my($fc) = shift;
    my($qtree) = shift;
    my($volume) = shift;

    my $style = "";
    my %HashOfInfo = filer_qtree_get($fc);
    foreach my $vol (keys %HashOfInfo) {
        foreach my $tree (keys %{$HashOfInfo{$vol}}) {
            if ($HashOfInfo{$vol}{$tree}{'Name'} eq $qtree) {
                if(defined($volume)) {
                    $volume=~ s#.*/(\w+)#$1#;
                    if($vol eq "$volume") {
                        $style = "$HashOfInfo{$vol}{$tree}{'Style'}";
                    }
                } else {
                    $style = "$HashOfInfo{$vol}{$tree}{'Style'}";
                }
            }
        }
    }

    return $style;
}

## @name ::filer_qtree_security_unix_set
## @summary Change security to unix on qtree
## @description Change security to unix on qtree and return 1 if succesfull
## @arg $fc Required: The filer connection object for the command line interface
## @arg $qtree Required: The name of the qtree to change security
## @arg $volume Optional: The name of the volume to change security
## Defaults to the root volume.
sub filer_qtree_security_unix_set ($$;$) {
    my($fc) = shift;
    my ($qtree) = shift;
    my ($volume) = shift;

    my $qtreepath;
    if (defined($volume)) {
        $volume=~ s#.*/(\w+)#$1#;
        if($qtree =~ /\//) {
            ExitSoftError("Qtree given ($qtree) contains a '/' and a volume arg was also given. If volume arg is supplied, qtree should not be a path");
            $qtreepath = "$qtree"
        } else {
            $qtreepath = "/vol/${volume}/$qtree"
        }
    } else {
        $qtreepath = "$qtree"
    }
    my $return = filer_qtree_security_set($fc, $qtreepath, "unix");
}

## @name ::filer_qtree_space_reserve_disable
## @summary Turn off space reservations for a qtree
## @arg $fc Required: The filer connection object for the command line interface
## @arg $path Required: The qtree name including its path
sub filer_qtree_space_reserve_disable ($$) {
    my $fc  = shift;
    my $path = shift;

    my $cmd = "qtree reservation $path disable";
    $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
       if ($line =~ /The reservation option is no longer a supported option/) {
           logresult('WARN',"This test may have been deprecated, 'qtree reservation' is no longer supported");
           return;
       } elsif ($line =~ /(qtree|file) reservation:/) {
           # The 'file" message prefix was lased used in Scrimshaw
           ExitError("Unexpected error from '$cmd': $line");
       }
    }
}

## @name ::filer_qtree_space_reserve_enable
## @summary Turn on space reservations for a qtree
## @arg $fc Required: The filer connection object for the command line interface
## @arg $path Required: The qtree name including its path
sub filer_qtree_space_reserve_enable ($$) {
    my $fc  = shift;
    my $path = shift;

    my $cmd = "qtree reservation $path enable";
    $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /The reservation option is no longer a supported option/) {
            logresult('WARN',"This test may have been deprecated, 'qtree reservation' is no longer supported");
            return;
        } elsif ($line =~ /(qtree|file) reservation:/) {
            # The 'file" message prefix was lased used in Scrimshaw
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_quota_scramble_conf
## @summary Executes the 'quota scramble_conf' command on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_quota_scramble_conf ($;) {
    my ($fc) = shift;
    my $cmd = 'quota scramble_conf';

    my $result = filer_cli_diag_say($fc, $cmd, {retry => 'ok'});
    return $result;
}

## @name ::filer_quota_tree_add
## @summary Adds quota values for a tree into the filer's /etc/quotas file
## @arg $fc Required: The filer connection object for the command line interface
## @arg $tree Required: The qtree name (* is OK as the last qualifier).
## @arg $disk Required: The disk use in bytes, or '-' for no limit
## @arg $files Optional: The file count, or '-' for no limit
## @arg $thold Optional: The disk use warning message threshold in bytes, or '-' for no limit
## @arg $sdisk Optional: The soft disk size limit in bytes, or '-' for no limit
## @arg $sfile Optional: The soft file count limit
sub filer_quota_tree_add ($$;$$$$$) {
    my($fc) = shift;
    my($tree) = shift;
    my($disk) = shift;
    if (defined($disk)) {
        if ($disk eq '') {
            $disk = '-';
        }
    } else {
        $disk = '-';
    }
    my($files) = shift;
    if (defined($files)) {
        if ($files eq '') {
            $files = '-';
        }
    } else {
        $files = '-';
    }
    my($thold) = shift;
    if (defined($thold)) {
        if ($thold eq '') {
            $thold = '-';
        }
    } else {
        $thold = '-';
    }
    my($sdisk) = shift;
    if (defined($sdisk)) {
        if ($sdisk eq '') {
            $sdisk = '-';
        }
    } else {
        $sdisk = '-';
    }
    my($sfile) = shift;
    if (defined($sfile)) {
        if ($sfile eq '') {
            $sfile = '-';
        }
    } else {
        $sfile = '-';
    }

    # Parse tree to see if @ location suffix is needed
    my $quota_id = $tree;
    my $type = 'tree';
    if ($tree =~ /\*$/) {
        $quota_id = '*';
        $tree =~ s/\*$//;
        $type = 'tree@'.$tree;
    }

    # Add new entry to existing file data
    filer_quotas_file_append($fc,"$quota_id $type $disk $files $thold $sdisk $sfile");
}

## @name ::filer_quota_tree_get
## @summary Gets the quota definition values for a tree from the filer's /etc/quotas file as an array
## @arg $fc Required: The filer connection object for the command line interface
## @arg $tree Required: The qtree name. For example, /vol/sledgehammer/my_qtree
sub filer_quota_tree_get ($$) {
    my($fc) = shift;
    my($tree) = shift;

    # Parse tree to see if @ location suffix is needed
    my $quota_id = $tree;
    my $type = 'tree';
    if ($tree =~ /\*$/) {
        $quota_id = '*';
        $tree =~ s/\*$//;
        $type = 'tree@'.$tree;
        $type = quotemeta($type);
    }
    $quota_id = quotemeta($quota_id);

    my @info = filer_quotas_file_read($fc);
    my @answer = ();
    my $column;

    #print("Looking thru /etc/quotas for '$quota_id' '$type'\n");   #DEBUG
    foreach my $line (@info) {
        if ($line =~ /^#/) {
            next;
        } elsif ($line =~ m#^$quota_id\s+$type\s+#) {
            # Drop the first two columns
            $line =~ s#^$quota_id\s+$type\s+##;
            # Columns that may follow include:
            # disk  files thold sdisk sfile
            @answer = split(/\s+/,$line);
        } else {
            #print("No match for $line\n");   #DEBUG
        }
    }

    return @answer;
}

## @name ::filer_quota_tree_report
## @summary Gets the quota report values for a tree.
## (KbytesUsed,Limit,FilesUsed,Limit, Quota Specifier)
## @arg $fc Required: The filer connection object for the command line interface
## @arg $tree Required: The qtree name. For example, /vol/sledgehammer/my_qtree
sub filer_quota_tree_report ($$) {
    my($fc) = shift;
    my($tree) = shift;

    my ($vol,$tree_name);
    $vol = $tree;
    $vol =~ s/\/vol\/(\w+)\/.*/$1/;
    $vol = quotemeta($vol);
    $tree_name = $tree;
    $tree_name =~ s/\/vol\/\w+\/(.*)/$1/;
    $tree_name = quotemeta($tree_name);

    my @answer = ();
    my $cmd = "quota report";
    $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context

    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    # Second message text appeared with FullsailN_030409_2330
    while ($text =~ /quotas are initializing|some quotas are still initializing/) {
        sleep(15);
         ($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    }

    my($line);
    my @info = filer_cli_split($text);
    # assume it was successful since it doesn't give any output.
    my $return = 1;
    foreach $line (@info) {
        if ($line =~ /^quota:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /^\s|^Type|^--/) {
            next; # A heading line was seen
        } elsif ($line =~ /tree\s+\d+\s+$vol\s+$tree_name\s+/) {
            # This line is reporting about the tree we want
            my $report_line = $line;
            # Type       ID    Volume    Tree  Used     Limit    Used    Limit   Quota Specifier
            # tree        11     vol0 pcmds_lun_clone         0 419430400       1       - /vol/vol0/pcmds_lun_clone
            $report_line =~ s/tree\s+\d+\s+$vol\s+$tree_name\s+//;
            # Return (KbytesUsed,Limit,FilesUsed,Limit, Quota Specifier)
            @answer = split(/\s+/,$report_line);
            print("KbytesUsed=$answer[0], Limit=$answer[1], FilesUsed=$answer[2], Limit=$answer[3], QuotaSpecifier=$answer[4]\n");
            last;
        } elsif ($line =~ /tree\s+\d+/) {
            next;  # a valid entry for another tree was seen
        } elsif ($line =~ /^\s*$/ or $line =~ /^===== /) {
            # vfiler header lines
        } else {
            ExitSoftError("Unrecognized quota report line: $line");
        }
    }
    if (@answer == 0) {
        ExitSoftError("No quota report data for volume: $vol, tree: $tree_name");
    }

    return @answer;
}

## @name ::filer_quotas_file_append
## @summary Appends line(s) to the filer's /etc/quotas file
## @arg $fc Required: The filer connection object for the command line interface
## @arg @lines Required: The lines to be added
sub filer_quotas_file_append ($@) {
    filer_file_append(shift,'/etc/quotas',@_);
}

## @name ::filer_quotas_file_read
## @summary Gets the filer's /etc/quotas file as an array or text string
## @arg $fc Required: The filer connection object for the command line interface
sub filer_quotas_file_read ($) {
    my($fc) = shift;

    my $quotas_file = '/etc/quotas';

    # Handle the vfiler context - find the vfiler's root
    my $vfiler_context;
    # See if we are simulating a vfiler context
    my $simulated_context = $fc->{hostp}->{'vfiler_simulated_context'};
    if ($simulated_context eq '') {
        $vfiler_context = filer_vfiler_context_get($fc);
    } else {
        $vfiler_context = $simulated_context;
    }
    if ($vfiler_context ne 'vfiler0' and $vfiler_context ne '') {
        # Determine the path to the quotas file
        my %vfiler_data = filer_vfiler_status_get($fc,$vfiler_context);
        if (defined($vfiler_data{'Path'})) {
            $quotas_file = $vfiler_data{'Path'}.$quotas_file;
        }
    }

    my $text = filer_file_read($fc,$quotas_file);

    if (wantarray) {
        my @info = filer_cli_split($text);
        return @info;
    } else {
        return $text;
    }
}

## @name ::filer_randomize
## @summary Return the output from randomize command
## @arg $fc Required: The filer connection object for the command line interface
## @arg path Optional: Filer path to be randomize (in /vol/volume[/path] format)
## (mutually exclusive with status and abort)
## @arg status Optional: Set to 1 to obtain randomize status message
## (mutually exclusive with path and abort)
## @arg abort Optional: Set to 1 to abort current randomize operation
## (mutually exclusive with path and status)
sub filer_randomize($%;) {
    my ($fc, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
        ],
        {
            path => { type => SCALAR, optional => 1 },
            status => { type => SCALAR, optional => 1 },
            abort => { type => SCALAR, optional => 1 },
        }
    );

    # User must specify one option, and only one option
    if (scalar keys %opts != 1) {
        ExitScriptError("One and only one option must be specified: " .
                        "path, status, or abort.");
    }

    # Build the command
    my $cmd = 'randomize';
    if ($opts{path}) {
        $cmd .= " $opts{path}";
    } elsif ($opts{status}) {
        $cmd .= ' status';
    } elsif ($opts{abort}) {
        $cmd .= ' abort';
    } else {
        ExitScriptError("Option must be specified to evaluate to TRUE");
        # e.g. specifying {status => 0}, {abort => 0}, or {path => ''} is invalid
    }

    my $result = filer_cli_test_say($fc, $cmd, {retry => 'ok'});
    return $result;
}

## @name ::filer_reboot
## @summary Reboots the filer, returns a current connection object
## @arg $fc Required: The filer connection object for the command line interface
sub filer_reboot ($) {
    my($fc) = shift;

    my $need_reconnect = 0;
    my $filer = filer_name($fc);
    my $fh = $fc->{'hostp'};
    my $cmd;
    my $comment;
    if (filer_cluster_is_enabled($fc)) {
        # disallow takeover by the partner
        $comment = "Rebooting without takeover";
        $cmd = "halt -f -t 0";
    } else {
        # check for unsync log state
        if (filer_cluster_is_in_unsync($fc)) {
           $comment = "Cluster is in unsynchronized log state, rebooting without takeover";
           $cmd = "halt -f -t 0";
    	} else {
           $comment = "Rebooting";
           $cmd = "halt -t 0";
	}
    }

    my $conn_type = $fc-> conntype();
    if ($conn_type =~ m/pm_cons|rsh-persay|rsh-socket/) {
        if ($conn_type =~ m/pm_cons/) {
            logcomment($comment,"Rebooting from the filer $conn_type connection");
            reboot($fc,haltcmd=>"$cmd")
                or ExitError("Could not reboot filer. $@");
        } else {
            logcomment($comment,"Rebooting from the filer host object");
            reboot($fh,haltcmd=>"$cmd")
                or ExitError("Could not reboot filer. $@");
        }
    } else {
        logcomment("Performing disconnect of $conn_type connection before reboot");
        disconnect($fc,'-logout');
        $need_reconnect = 1;
        logcomment($comment,"Rebooting from the filer host object");
        reboot($fh,haltcmd=>"$cmd")
            or ExitError("Could not reboot filer. $@");
    }

    wait_for_rsh($filer);  # Give the filer time for rsh to wake up

    # Verify that the filer did not get stuck during reboot
    # See burt 81571 for an example
    # Issue a command that is valid on any filer or cache appliance
    my $count = 0;
    while ($count <5) {
        $count++;
        sleep(15);
        if ($need_reconnect) {
            host($filer);
            $fc = connect("#$conn_type", -timeout, 60);
            $need_reconnect = 0;
        }
        my($match,$before,$after) = say($fc, "?", $waiting_for_giveback_string,
            -timeout, 30 ,  # Must be long enough to see the "Waiting for giveback"
            -commline, "No prompt after reboot");
        if ((!defined($before) or $before eq '') and (!defined($after) or $after eq '')) {
            if ($count == 5 ) {
                burt_found(81571);
                filer_power_cycle($fc);
                # Re-establish connection if necessary
                if ($conn_type !~ m/pm_cons|rsh-persay|rsh-socket/) {
                    host($filer);
                    $fc = connect("#$conn_type", -timeout, 60);
                }
            }
        } elsif (defined($match) and $match =~ m/$waiting_for_giveback_string/) {
            ExitError("Filer was taken over during reboot");
        } else {
           logcomment("Verified filer connection after reboot");
           last;
        }
    }
    return $fc;
}

## @name ::filer_registry_get
## @summary Get the value for option given
## @arg $fc Required: The filer connection name where this function is executed
## @arg $option Required: The name of the option
sub filer_registry_get ($$) {
    my($fc) = shift;
    my ($option) = shift;

    my $cmd = "registry get options.${option}";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my $value = "";
    my @val;
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /$option/) {
            #(undef, $value) = split(/=/, $line);
            (undef, @val) = split(/=/, $line);
            #need this since some values have '=' in them
            #and we want to get the whole thing (for example host=all)
            $value = join('=', @val);
        }
    }
    return $value;
}

## @name ::filer_release_revert
## @summary Set the option to the given value
## @arg $fc Required: The filer connection name where this function is executed
## @arg $release Required: The name of the option
## @arg $value Required: The value to set the option to
sub filer_release_revert ($$;$) {
    my($fc) = shift;
    my ($release) = shift;
    my ($timeout) = shift;
    if (!defined($timeout)) {
	$timeout = 1200;
    }

    my $cmd = "revert_to -i -f $release";
    say ($fc, $cmd,
	'-noprompt',
	'-prompt',"System shut down because",
	'-timeout',$timeout);
    boot($fc) or die $@;
}

## @name ::filer_resource_agitation_set
## @summary Set resource agitation levels.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $res Required: resource (whose agitation level is to be set)
## @arg consume_pct Optional: Consume this percent of resource.
## @arg random_N Optional: Random deny|drop|collect|sleep 1 out of N.
## @arg value Optional: Only used for kma_sleep, number of msecs to sleep.
sub filer_resource_agitation_set ($$;%) {
    my ($fc, $res, %opts) = validate_mixed (
        @_,
        [
            { type => OBJECT },
            { type => SCALAR },
        ],
        {
            consume_pct => { type => SCALAR, optional => 1,
                                                regex => qr/^\d+$/ },
            random_N    => { type => SCALAR, optional => 1,
                                                regex => qr/^\d+$/ },
            value       => { type => SCALAR, optional => 1,
                                                regex => qr/^\d+$/ },
        }
    );
    # Table to map specified resource to proper
    # sub-command and required arguments needed
    my %val_h = (
        cpu       => { cmd => 'cpu',       args => [qw(consume_pct)] },
        malloc    => { cmd => 'malloc',    args => [qw(consume_pct random_N)] },
        kma_sleep => { cmd => 'kma_sleep', args => [qw(random_N value)] },
        mbuf      => { cmd => 'mbuf',      args => [qw(consume_pct random_N)] },
        waflbuf   => { cmd => 'waflbuf',   args => [qw(consume_pct random_N)] },
        fd        => { cmd => 'fd',        args => [qw(consume_pct random_N)] },
        pBlk      => { cmd => 'pBlk',      args => [qw(consume_pct random_N)] },
        raidrm    => { cmd => 'raidrm',    args => [qw(consume_pct random_N)] },
        java      => { cmd => 'java',      args => [qw(random_N)] },
        gc        => { cmd => 'gc',        args => [qw(random_N)] },
        page      => { cmd => 'page',      args => [qw(random_N)] },
        hseg      => { cmd => 'hseg',      args => [qw(random_N)] },
        socket    => { cmd => 'socket',    args => [qw(random_N)] },
        nwdrop    => { cmd => 'nwdrop',    args => [qw(random_N)] },
        nscorrupt => { cmd => 'nscorrupt', args => [qw(random_N)] },
        nwcorrupt => { cmd => 'nscorrupt', args => [qw(random_N)] },
        reset     => { cmd => 'reset',     args => [] },
    );

    # Validate resource
    if ( !exists($val_h{$res}) ) {
        ExitScriptError("Resource '$res' is not supported");
    }

    # Command to set agitation
    my $cmd = "setral $val_h{$res}{cmd}";

    # Validate arguments
    foreach my $arg ( @{$val_h{$res}{args}} ) {
        ExitScriptError("Required argument '$arg' is not defined")
            if ( !defined($opts{$arg}) );
        $cmd .= " $opts{$arg}";
    }

    # Set resource agitation on the filer
    my $text = filer_cli_test_say($fc, $cmd, {retry=>'ok'});

    if ( $text !~ /agitation (started|stopped)/ ) {
        # If agitation is already started, the new settings
        # are displayed instead
        $text =~ s/setral settings: [^(]+\((?:in)?active\)//;
        ExitError($text) if $text;
    }

    return 1;
}

## @name ::filer_serial_number_get
## @summary Returns the serial number of the specified filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_serial_number_get ($) {
    my $fc = shift;

    my $slot = 0;

    my $cmd = "sysconfig -v $slot";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $field = '';
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /\s+Serial Number:/) {
            $field = $line;
            $field =~ s/^\s+Serial Number:\s+(\w+).*/$1/;
            last;
        }
    }
    return $field;
}

## @name ::filer_service_is_licensable
## @summary Determines if service can be licensed on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $service Required: The name of the service
sub filer_service_is_licensable ($$) {
    my($fc) = shift;
    my ($service) = shift;

    # the 'priv set diag' avoids the issue of command not rsh'able, see Burt 66069
    my $cmd = "license";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /$service /) {
            return 1;
        }
    }
    return 0;
}

## @name ::filer_service_is_licensed
## @summary Determines if service is licensed on the filer
## @description Returns 0 if service is not licensed. Returns 2 is it is 'site' licensed, else returns 1.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $service Required: The name of the service
sub filer_service_is_licensed ($$) {
    my($fc) = shift;
    my ($service) = shift;
    my $cmd = "license";
    # the 'priv set diag' avoids the issue of command not rsh'able, see Burt 66069
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok', useresult => 0});

    my($line);
    my @info = filer_cli_split($text);
    foreach $line (@info) {
        # make sure to match either cf | cluster license
        $service = 'cf'
           if ($line =~ /\s+cf\s+.*/ and $service eq 'cluster');
        $service = 'cluster'
           if ($line =~ /\s+cluster\s+.*/ and $service eq 'cf');
        if ($line =~ /^$service\s|\s+$service\s/i) {
            if ($line =~ /not licensed|expired/) {
                logcomment("Service $service is not licensed.");
                return 0;
            } elsif ($line =~ /\ssite\s/) {
                logcomment("Service $service is site licensed.");
                return 2;
            } else {
                logcomment("Service $service is  licensed.");
                return 1;
            }
        }
    }
    logcomment("No mention of $service service in license command output.");
    return 0;
}

## @name ::filer_service_is_started
## @summary Returns the status of the requested service on the filer
## @description Returns a 1 if the service is started
## or 0 if it is not.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $service Required: The name of the service (cifs, fcp, iscsi or nfs)
sub filer_service_is_started ($$) {
    my($fc) = shift;
    my ($service) = shift;

    my $status = 'status';  # subcommand syntax for the service
    if ($service =~ m/cifs/i) {
        $status = 'sessions';
    }

    my $cmd = "$service $status";
    # Add vfiler simulated context support if this service is vfilerized
    if ($service =~ m/^cifs$|^iscsi$|^nfs$/i) {
        $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    }
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
       if ($line =~ /$service\s*:\s+/) {
          ExitSoftError("Error message presented to the console: $line.");
       }
       if ($service =~ m/fcp|iscsi|dafs/i) {
          if ($line =~ /$service service is not running|$service not found|$service is not licensed/i) {
             return 0;
          }
          if ($line =~ /$service service is stopped/i) {
             return 0;
          }
          if ($line =~ /$service service is stopped[\r\n]/i) {
             ExitSoftError('-burt',71469,"Output should end in a '.'" );
             return 0;
          }
          if ($line =~ /$service service is not running/i) {
             #burt_begin(66192);
             #ExitSoftError("Service name should be in lowercase letters." );
             #burt_end(66192);
             return 0;
          }
          if ($line =~ /$service service is running/) {
             return 1;
          }
          if ($line =~ /$service service is started/i) {
             return 1;
          }
          if ($line =~ /$service service is started[\r\n]/i) {
             ExitSoftError('-burt',71469,"Output should end in a '.'" );
             return 1;
          }
          if ($line =~ /$service service is running/i) {
             #ExitSoftError('-burt',66192,"Service name should be in lowercase letters." );
             return 1;
          }
       } elsif ($service =~ m/cifs/i) {
          if ($line =~ m/CIFS\s.*enabled|Server Registers as /) {
             return 1;
          } elsif ($line =~ m/CIFS not running/) {
             return 0;
          }
       } elsif ($service =~ m/nfs/i) {
          if ($line =~ m/server is running/) {
             return 1;
          } elsif ($line =~ m/server is NOT running/) {
             return 0;
          }
       } elsif ($service =~ m/snapmirror/i) {
          if ($line =~ m/Snapmirror is on/i) {
             return 1;
          } elsif ($line =~ m/Snapmirror is off/i) {
             return 0;
          }
       } else {
           # Other services may not use these messages,
           # but report on them if they are seen:
           if ($line =~ / service is not running| service is stopped/) {
              return 0;
           }
           if ($line =~ / service is running| service is started/) {
              return 1;
           }
       }
    }
    logresult('WARN',"Could not verify status of $service service");
    # This could be a back level ONTAP version without this service
    # Report that the service is not started
    return 0;
}

## @name ::filer_service_license
## @summary Licenses a service on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $service Required: The name of the service to license, 'cifs' for example
## @arg $code Optional: The code to enable the service
## @arg $days Optional: The number of days (1 or 90) for a temporary license to the service.
## @arg $allow_error Optional: 1 if it is OK to see an error, 0 is default - no errors allowed
## Default is for a permanent license.
sub filer_service_license ($$;$$$) {
    my($fc) = shift;
    my ($service) = shift;
    my ($code) = shift;
    my ($days) = shift;
    my ($allow_error) = shift;
    if (! defined($code) ) {
       $code = '';
    }
    if (! defined($days) ) {
        $days = '';
    }
    if (! defined($allow_error) ) {
        $allow_error = 0;
    }

    my $verify_license = 1; # default to verify the license was added
    my $filer_model = '';
    my %day_licenses = ();
    my $license_name = '';

#print Data::Dumper->Dump([%filer_1day_licenses], [qw(filer_1day_licenses)]);
    if ($code eq '') {
        # Convert to lowercase letters to support mixed case input
        $service = lc($service);

        if (defined($deprecated_licenses{$service})) {
            ExitScriptError("Service '$service' is not supported on current releases of ONTAP.");
        }

        # If license was not deprecated, we need to look up a valid value
        # based on model and OEM vendor
        # Different codes for filer, netcache, NearStore, Gateway Filer, IBM or promoted simulator use
        $filer_model = $fc->{hostp}->{'MODEL'};
        if (!defined($filer_model) ) {
            $filer_model = $fc->{hostp}->{'model'};
            if (!defined($filer_model) ) {
                $filer_model = filer_model_get($fc);
            }
        }

        if ($fc->hostp->hosttype() =~ /(filer|cache)-simulator/) {
            if ($1 eq 'cache') {
                $filer_model = 'C';  # Use prefix of the netcache models
            } else {
                # check if the filer simulator is promoted or not
                # one way is to see if test privilege is available or not
                my($text,$result) = filer_cli_say($fc,"priv set -q test",{retry=>'ok', useresult => 0});
                if ($text =~ /priv: unrecognized privilege: test/i) {
                    $filer_model = 'promoted_simulator';
                } else {
                    $filer_model = 'F';  # Use prefix of the standard "filer" models
                }
            }
        }
        if ($filer_model =~ m/^R/i ) { # For NearStore products:
            if ($days eq '') {
                if (defined($nearstore_licenses{$service})) {
                    $code = $nearstore_licenses{$service};
                    if ($code eq '1') {
                        ExitScriptError("Service '$service' is not supported for a license on current releases of NearStore products.");
                    }
                } else {
                    ExitScriptError("Service '$service' is not recognized by the test script for NearStore products.");
                }
            } else { # For NearStore products:
                $license_name = '%nearstore_'.$days.'day_licenses';
                %day_licenses = eval($license_name);
                if (defined($day_licenses{$service})) {
                    $code = $day_licenses{$service};
                    if ($code eq '1') {
                        ExitScriptError("Service '$service' is not supported for a temporary license on current releases of NearStore products.");
                    }
                } else {
                    ExitScriptError("Service '$service' is not recognized by the test script for $days day license of NearStore products.");
                }
            }
        } elsif ($filer_model =~ /^(G|V)/i and $filer_model !~ m/VSIM|VSA/i) { # For Gateway Filers
            if ($days eq '') {
                if (defined($gateway_licenses{$service})) {
                    $code = $gateway_licenses{$service};
                    if ($code eq '1') {
                        ExitScriptError("Service '$service' is not supported for a license on current releases of Gateway Filer products.");
                    }
                } else {
                    ExitScriptError("Service '$service' is not recognized by the test script for Gateway Filer products.");
                }
            } else {
                $license_name = '%gateway_'.$days.'day_licenses';
                %day_licenses = eval($license_name);
                if (defined($day_licenses{$service})) {
                    $code = $day_licenses{$service};
                    if ($code eq '1') {
                        ExitScriptError("Service '$service' is not supported for a temporary license on current releases of Gateway Filer products.");
                    }
                } else {
                    ExitScriptError("Service '$service' is not recognized by the test script for $days day license of Gateway Filer products.");
                }
            }
        } elsif ($filer_model =~ /^C/i) { # For NetCache
            if ($days eq '') {
               if (defined($cache_licenses{$service})) {
                   $code = $cache_licenses{$service};
                   if ($code eq '1') {
                        ExitScriptError("Service '$service' is not supported for a license on current releases of NetCache products.");
                   }
               } else {
                    ExitScriptError("Service '$service' is not recognized by the test script for NetCache products.");
               }
            } else {
                $license_name = '%cache_'.$days.'day_licenses';
                %day_licenses = eval($license_name);
                if (defined($day_licenses{$service})) {
                    $code = $day_licenses{$service};
                    if ($code eq '1') {
                        ExitScriptError("Service '$service' is not supported for a temporary license on current releases of NetCache products.");
                    }
                } else {
                    ExitScriptError("Service '$service' is not recognized by the test script for $days day license of NetCache products.");
                }
            }
        } elsif ($filer_model =~ /^N/i) { # For IBM rebranded filer models
            my $machine_type = filer_machine_type_get($fc);
            # IBM-2862-A20
            if ($machine_type =~ /IBM-\d\d\d\d-(A|G|C)\d\d/) {
                if ($1 eq 'A'
                    || $1 eq 'C') {
                    # non-Gateway
                   if ($days eq '') {
                       if (defined($filer_licenses{$service})) {
                           $code = $filer_licenses{$service};
                           if ($code eq '1') {
                               ExitScriptError("Service '$service' is not supported for a license on current releases of IBM rebranded filer machine types.");
                           }
                       } else {
                           ExitScriptError("Service '$service' is not recognized by the test script for IBM rebranded filer machine types.");
                       }
                   } else {
                       $license_name = '%filer_'.$days.'day_licenses';
                       %day_licenses = eval($license_name);
                       if (defined($day_licenses{$service})) {
                           $code = $day_licenses{$service};
                           if ($code eq '1') {
                               ExitScriptError("Service '$service' is not supported for a temporary license on current releases of IBM rebranded filer machine types.");
                           }
                       } else {
                           ExitScriptError("Service '$service' is not recognized by the test script for $days day license of IBM rebranded machine types.");
                       }
                   }
                } else {
                    # Gateway
                   if ($days eq '') {
                       if (defined($gateway_licenses{$service})) {
                           $code = $gateway_licenses{$service};
                           if ($code eq '1') {
                               ExitScriptError("Service '$service' is not supported for a license on current releases of IBM rebranded gateway machine types.");
                           }
                       } else {
                           ExitScriptError("Service '$service' is not recognized by the test script for IBM rebranded gateway machine types.");
                       }
                   } else {
                       $license_name = '%gateway_'.$days.'day_licenses';
                       %day_licenses = eval($license_name);
                       if (defined($day_licenses{$service})) {
                           $code = $day_licenses{$service};
                           if ($code eq '1') {
                               ExitScriptError("Service '$service' is not supported for a temporary license on current releases of IBM rebranded gateway machine types.");
                           }
                       } else {
                           ExitScriptError("Service '$service' is not recognized by the test script for $days day license of IBM rebranded gateway machine types.");
                       }
                   }
                }
            } else {
                ExitScriptError("IBM machines must define a Machine Type");
            }
        } elsif ($filer_model eq 'promoted_simulator') {
            if ($days eq '') {
                if (defined($promoted_simulator_licenses{$service})) {
                    $code = $promoted_simulator_licenses{$service};
                    if ($code eq '1') {
                        ExitScriptError("Service '$service' is not supported for a license on current releases of promoted simulators.");
                    }
                } else {
                    ExitScriptError("Service '$service' is not recognized by the test script for promoted simulators.");
                }
            } else {
                $license_name = '%promoted_simulator_'.$days.'day_licenses';
                %day_licenses = eval($license_name);
                if (defined($day_licenses{$service})) {
                    $code = $day_licenses{$service};
                    if ($code eq '1') {
                        ExitScriptError("Service '$service' is not supported for a temporary license on current releases of promoted simulators.");
                    }
                } else {
                    ExitScriptError("Service '$service' is not recognized by the test script for $days day license ofpromoted simulators.");
                }
            }
        } elsif ($filer_model =~ /^SA/i) {
            # For SA Appliance
            if (defined($sa_filer_license{$service})) {
               $code = $sa_filer_license{$service};
               if ($code eq '1') {
                    ExitScriptError("Service '$service' is not supported for a license on current releases of SA Appliance Filer products.");
               }
             } else {
               ExitScriptError("Service '$service' is not not recognized by the test script for a license on current releases of SA Appliance Filer products.");
             }
        } elsif ($filer_model =~ /VSA/i) { #for promoted, non-promoted VSA's including Flexcache VSA
            # check if the VSA model is promoted or not
            # one way is to see if test privilege is available or not
            my($text,$result) = filer_cli_say($fc,"priv set -q test",{retry=>'ok', useresult => 0});
            if ($text =~ /priv: unrecognized privilege: test/i) {
                $filer_model = 'promoted_vsa';
            }
            if ($days eq '' ) {
                if ($filer_model =~ /promoted_vsa/) {
                    #for promoted VSA
                    if (defined($promoted_vsa_licenses{$service})) {
                        $code = $promoted_vsa_licenses{$service};
                        if ($code eq '1') {
                            ExitScriptError("Service '$service' is not supported for a license on current releases of promoted VSAs.");
                        }
                    } else {
                        ExitScriptError("Service '$service' is not recognized by the test script for a license on current releases of promoted VSAs.");
                    }
                } else {
                    #for non-promoted VSA
                    if (defined($filer_licenses{$service})) {
                        $code = $filer_licenses{$service};
                        if ($code eq '1') {
                            ExitScriptError("Service '$service' is not supported for a license on current releases of non-promoted VSAs.");
                        }
                    } else {
                        ExitScriptError("Service '$service' is not recognized by the test script for a license on current releases of non-promoted VSAs.");
                    }
                }
            } else {
                if ($filer_model =~ /promoted_vsa/) {
                    #expiry licenses for promoted VSA
                    $license_name = '%vsa_'.$days.'day_licenses';
                    %day_licenses = eval($license_name);
                    if (defined($day_licenses{$service})) {
                        $code = $day_licenses{$service};
                        if ($code eq '1') {
                            ExitScriptError("Service '$service' is not supported for a temporary license on current releases of VSAs.");
                        }
                    } else {
                        ExitScriptError("Service '$service' is not recognized by the test script for $days day license of VSA products.");
                    }
                } else {
                    #expiry license for non promoted VSA
                    $license_name = '%filer_'.$days.'day_licenses';
                    %day_licenses = eval($license_name);
                    if (defined($day_licenses{$service})) {
                        $code = $day_licenses{$service};
                        if ($code eq '1') {
                            ExitScriptError("Service '$service' is not supported for a temporary license on current releases of Filer products.");
                        }
                    } else {
                        ExitScriptError("Service '$service' is not recognized by the test script for $days day license of Filer products.");
                    }
                }
            }
        } else {
            # Assume it is a filer model such as:  F880, FAS3070, S500
            if ($days eq '') {
                if (defined($filer_licenses{$service})) {
                    $code = $filer_licenses{$service};
                    if ($code eq '1') {
                        ExitScriptError("Service '$service' is not supported for a license on current releases of Filer products.");
                    }
                } else {
                    ExitScriptError("Service '$service' is not recognized by the test script for Filer products.");
                }
            } else {
                $license_name = '%filer_'.$days.'day_licenses';
                %day_licenses = eval($license_name);
                if (defined($day_licenses{$service})) {
                    $code = $day_licenses{$service};
                    if ($code eq '1') {
                        ExitScriptError("Service '$service' is not supported for a temporary license on current releases of Filer products.");
                    }
                } else {
                    ExitScriptError("Service '$service' is not recognized by the test script for $days day license of Filer products.");
                }
            }
        }
    }

    # The 'priv set diag' avoids the issue of command not rsh'able, see Burt 66069
    my $cmd;
    if ($code) {
        # Beginning with fullsail, 'license $service=$code' is replaced with 'license add $code'
        $cmd = "license add $code";

        my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok', useresult => 0});

        my($line);
        my @info = filer_cli_split($text);
        # In clustered environment, verify license change includes prompt to change partner
        my $foundprompt = 0;
        my $foundpartnerprompt = 0;
        foreach $line (@info) {
            if ($line =~ /license not found/) {
                ExitError('-burt',66069,'license command disappeared.');
            }
            if ($line =~ / not licensed/) {
                if ($allow_error eq '0') {
                    ExitError("Command failed ($cmd): $line");
                } elsif ($allow_error eq '1') {
                    logcomment("License for $service service could not be added");
                }
                $verify_license = 0; # Skip license verification, it will fail
            }
            if ($line =~ /Promotional license installation limit exceeded /) {
                if ($result eq cmdOkReturn) {
                    ExitSoftError('-burt',105976,"Non-zero return code should accompany a license add command that fails");
                }
                if ($allow_error eq '0') {
                    ExitError("Command failed ($cmd): $line");
                } elsif ($allow_error eq '1') {
                    logcomment("License for $service service could not be added, promotional installation limit exceeded");
                }
                $verify_license = 0; # Skip license verification, it will fail
            }

            # Exact text not known
            if ($line =~ / partner /) {
                $foundpartnerprompt = 1;
            }
            # Look for extra prompt for specific services
            if ($service =~ /fcp|iscsi/) {
                if ($line =~ /Run \'$service start\' to start .* service./) {
                    $foundprompt = 1;
                }
            } elsif ($service =~ /dafs/) {
                if ($line =~ /Use \'dafs setup\' to configure DAFS/) {
                    $foundprompt = 1;
                }
            }
        }
        if ($verify_license) {
            if ($service =~ /fcp|iscsi|dafs/) {
                if (!$foundprompt) {
                    ExitSoftError("Missing prompt to start $service.");
                } else {
                    logcomment("Verified prompt to start $service.");
                }
            }

            if (!filer_service_is_licensed($fc,$service)) {
                ExitError("Could not license $service service with $code.");
            }

            if (wantarray) {
                return ($text,$result);
            } else {
                return $text;
            }
        }
    } else {
        ExitScriptError("No license code defined for $service service on model: $filer_model");
    }
}

## @name ::filer_service_start
## @summary Starts a service on the filer
## @description Starts the service and returns 0 if it was started
## or 1 if it was already started.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $service Required: The name of the service to start (cifs, fcp, iscsi or nfs)
sub filer_service_start ($$;$) {
    my($fc) = shift;
    my ($service) = shift;

    my $start = 'start';  # subcommand syntax for the service
    if ($service =~ m/nfs|snapmirror/i) {
        $start = 'on';
    } elsif ($service =~ m/cifs/i) {
        $start = 'restart';
    }

    my $cmd = "$service $start";
    # Add vfiler simulated context support if this service is vfilerized
    if ($service =~ m/^cifs$|^iscsi$|^nfs$/i) {
        $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    }

    my($text,$result,$opts);
    if ($opts = shift) {
        ($text,$result) = filer_cli_say($fc,$cmd,$opts);
    } else {
        ($text,$result) = filer_cli_say($fc,$cmd);
    }

    my($line);
    my @info = filer_cli_split($text);
    foreach $line (@info) {
        if ($line =~ /$service\s*:\s+/) {
            ExitSoftError("Error message presented to the console: $line.");
        }
        if ($service =~ m/nfs/i) {
            if ($line =~ /server is running/) {
                return 0;
            }
        } elsif ($service =~ m/cifs/i) {
            if ($line =~ /CIFS is already running/) {
                return 1;
            } elsif ($line =~ /CIFS local server is running|Server Registers as /) {
                return 0;
            }
        } else {
            if ($line =~ /$service service started\./) {
                return 0;
            }
            if ($line =~ / service started\./) {
                #ExitSoftError('-burt',66192,"Service name should be in lowercase letters." );
                return 0;
            }
            # 'FCP service is already running.' has been observed
            if ($line =~ /$service service is already running\./) {
                return 1;
            }
            if ($line =~ / service is already running\./) {
                # ExitSoftError('-burt',66192,"Service name should be in lowercase letters.");
                return 1;
            }
        }
    }
}

## @name ::filer_service_status_get
## @summary Determines the status of a service on the filer
## @description Determines the service status and returns 1 if is was started
## or 0 if it is not running.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $service Required: The name of the service to start (cifs, fcp, iscsi or nfs)
sub filer_service_status_get ($$) {
    my($fc) = shift;
    my ($service) = shift;

    return filer_service_is_started($fc,$service);
}

## @name ::filer_service_stop
## @summary Stops a service on the filer
## @description Stops the service and returns 0 if it was stopped
## or 1 if it was already stopped.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $service Required: The name of the service to stop (cifs, fcp, iscsi or nfs)
sub filer_service_stop ($$;$) {
    my($fc) = shift;
    my ($service) = shift;

    my $stop = 'stop';  # subcommand syntax for the service
    if ($service =~ m/nfs|snapmirror/i) {
        $stop = 'off';
    } elsif ($service =~ m/cifs/i) {
        $stop = 'terminate -t 0';
    }

    my $cmd = "$service $stop";
    # Add vfiler simulated context support if this service is vfilerized
    if ($service =~ m/^cifs$|^iscsi$|^nfs$/i) {
            $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    }

    my($text,$result,$opts);
    if ($opts = shift) {
        ($text,$result) = filer_cli_say($fc,$cmd,$opts);
    } else {
        ($text,$result) = filer_cli_say($fc,$cmd);
    }

    my($line);
    my @info = filer_cli_split($text);
    foreach $line (@info) {
        if ($line =~ /DBG:\./) {
            ExitSoftError("Debug output presented to the console: $line.");
        }
        if ($service =~ m/fcp|iscsi|dafs/i) {
            if ($line =~ /$service\s*:\s+/) {
                ExitSoftError("Error message presented to the console: $line.");
            }
            if ($line =~ /$service service stopped\./) {
                return 0;
            } elsif ($line =~ /$service not found/) {
                return 1;    # It is probably not licenced!
            }
            if ($line =~ / service stopped\./) {
                #ExitSoftError('-burt',66192,"Service name should be in lowercase letters." );
                return 0;
            }
            if ($line =~ /$service service is not running\./) {
                return 1;
            }
            if ($line =~ / service is not running\./) {
                #ExitSoftError('-burt',66192,"Service name should be in lowercase letters." );
                return 1;
            }
        } elsif ($service =~ m/cifs/i) {
            if ($line =~ m/CIFS local server has shut down/) {
                return 0;
            } elsif ($line =~ m/Server Registers as /) {
                return 1;
            }
        } elsif ($service =~ m/nfs/i) {
            if ($line =~ m/server is NOT running/) {
                return 0;
            }
        } elsif ($service =~ m/snapmirror/i) {
            if ($text =~ m/Snapmirror is off/i) {
                return 0;
            }
        }
    }
}

## @name ::filer_service_unlicense
## @summary Unlicenses a service on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $service Required: The name of the service to unlicense
sub filer_service_unlicense ($$) {
    my($fc) = shift;
    my ($service) = shift;

    # the 'priv set diag' avoids the issue of command not rsh'able, see Burt 66069
    my $cmd;
    $cmd = "license delete $service";
    my($text,$result) = filer_cli_diag_say($fc,$cmd);

    # If license has already been deleted, command returns 0 and no message
    if ($result ne 0) {
        my $got_message = 0;

        if ($text =~ /license not found/) {
            ExitError('-burt',66069,'license command disappeared.');
        } elsif ($text =~ /PANIC: assertion "\(newState == MBX_TAKEOVER_DISABLED\) \|\| \(newState == MBX_TAKEOVER_ENABLED\)" failed: file "\.\.\/common\/ha\/fmmbx\.c", line 1079/) {
            ExitError('-burt',68352,"Panic during: &cmd.");
        } elsif ($text =~ /unlicensed $service\./i) {
            # Got the expected verification, no error
            $got_message = 1;
        } else {
            if ($service =~/^cifs$/i) {
               if ($text =~ /cifs will be disabled upon reboot/) {
                   $got_message = 1;
                   # Should we do the reboot here?
                   # Otherwise license test below will fail

               }
           }
        }
        if ($got_message == 0) {
            # Verification string comes from license man page
            ExitError( "No confirmation message seen for '$cmd'");
        }
        if ($service !~/^cifs$/i) {
            # CIFS needs a reboot to finish, so don't verify it
            if (filer_service_is_licensed($fc,$service)) {
                ExitError("Could not disable license to $service service.");
            }
        }
    }
}

## @name ::filer_services_get
## @summary Returns a hash of services.
## @description The keys are the service names, the values are the text from the license output.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_services_get ($) {
    my($fc) = shift;

    my %services;

    my $cmd = "license";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my($line);
    my @info = filer_cli_split($text);
    foreach $line (@info) {
        if ($line =~ /license:\s+/) {
           ExitSoftError("Error message presented to the console: $line.");
        }
        $line =~ s/^\s*//g;
        $line =~ s/\s*$//g;
        $line =~ s/\s\s/ /g;
        if ($line !~ m/^\s*$/ && $line !~ /---------/ && $line !~ /Package/) {
            my ($key,$value) = split(' ',$line,2);
            if (defined($key) and $key !~ m/^\s*$/) {
                if (defined($value)and $value !~ m/^\s*$/) {
                    $services{$key} = $value;
                }
            }
        }
    }
    return %services;
}

## @name ::filer_services_get_licensable
## @summary Returns a list of licensed services.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_services_get_licensable ($) {
    my($fc) = shift;

    my %services = filer_services_get($fc);
    return keys %services;
}

## @name ::filer_services_get_licensed
## @summary Returns a list of licensed services.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_services_get_licensed ($) {
    my %services = filer_services_get(shift);

    my @licensed;
    foreach my $service (keys %services) {
        if ($services{$service} !~ /not licensed|expired/) {
            push(@licensed,$service);
        }
    }
    return @licensed;
}

## @name ::filer_showfh
## @summary show file handle information (ADVANCED MODE)
##          returns a hash of (flags snapid fileid gen fsid)
## @arg $fc Required: The filer connection object for the command line interface
## @arg $file Required: The file name including /vol/<volume>/ if not in the root directory.
sub filer_showfh($$) {
    my $fc  = shift;
    my($file) = shift;

    my $cmd = "showfh $file";
    my($text) = filer_cli_advanced_say($fc,$cmd);
    if ($text =~ /flags=(\S+)\s+snapid=(\S+)\s+fileid=(\S+)\s+gen=(\S+)\s+fsid=(\S+)/) {
       my %fh;

       $fh{flags}=$1;
       $fh{snapid}=$2;
       $fh{fileid}=$3;
       $fh{gen}=$4;
       $fh{fsid}=$5;

       return %fh;
    }

    ExitError("Unexpected result for showfh: $@");
}

## @name ::filer_snapmirror_dest_break
## @summary Breaks the snapmirror to the specified destination
## @description This also includes quiesing the destination
## @arg $fc Required: The filer connection object for the command line interface
## @arg $dest Required: The destination volume or qtree
sub filer_snapmirror_dest_break ($$) {
    my $fc  = shift;
    my $dest  = shift;
    if ($dest eq '') {
        ExitScriptError("A valid volume or qtree must be specified");
    }

    if ($dest =~ m#/#) {
        my $cmd = "snapmirror quiesce $dest";
        my($text,$result) = filer_cli_say($fc,$cmd);
    }

    my $cmd = "snapmirror break $dest";
    my($text,$result) = filer_cli_say($fc,$cmd);
}

## @name ::filer_snapmirror_dest_quiesce
## @summary Allow in-progress transfers to snapmirror destination to complete
## after whcih new transfers are not started.
## @description Control is returned after the quiesce has completed
## @arg $fc Required: The filer connection object for the command line interface
sub filer_snapmirror_dest_quiesce ($$) {
    my $fc  = shift;
    my $dest  = shift;
    if ($dest eq '') {
        ExitScriptError("A valid volume or qtree must be specified");
    }

    my $cmd = "snapmirror quiesce $dest";
    my($text,$result) = filer_cli_say($fc,$cmd);

    while () {
        my $cmd = "snapmirror status $dest";
        my($text,$result) = filer_cli_say($fc,$cmd);
        if ($text =~ m/is quiescing/) {
            # Allow for timeout to be specified

            sleep(5);
        } elsif ($text =~ m/has quiesced|Snapmirror is off/) {
            last;
        }
    }
}

## @name ::filer_snapmirror_dest_status
## @summary Determines status of a snapmirror destination
## @arg $fc Required: The filer connection object for the command line interface
sub filer_snapmirror_dest_status ($$) {
    my $fc  = shift;
    my $dest  = shift;
    if ($dest eq '') {
        ExitScriptError("A valid volume or qtree must be specified");
    }

    my $cmd = "snapmirror status $dest";
    my($text,$result) = filer_cli_say($fc,$cmd);
}

## @name ::filer_snapmirror_initialize
## @summary starts snapmirror data transfer to the specified destination
## @arg $fc Required: The filer connection object for the command line interface
## @arg $dest Required: The destination volume or qtree
## @arg $dest Optional: [ -S source ] [ -k kilobytes ] [ -s src_snap ] [ -c create_dest_snap ] [ -w ] [ -L ]
sub filer_snapmirror_initialize ($$;$) {
    my $fc  = shift;
    my $dest  = shift;
    my $opt  = shift;
    if (!defined($opt)) {
        $opt = '';
    }

    my $cmd = "snapmirror initialize $opt $dest";
    my($text,$result) = filer_cli_say($fc,$cmd);
}

## @name ::filer_snapmirror_off
## @summary stops snapmirror service on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_snapmirror_off ($) {
    my $fc  = shift;

    my $cmd = "snapmirror off";
    my($text,$result) = filer_cli_say($fc,$cmd);

}

## @name ::filer_snapmirror_on
## @summary starts snapmirror on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_snapmirror_on ($) {
    my $fc  = shift;

    my $cmd = "snapmirror on";
    my($text,$result) = filer_cli_say($fc,$cmd);
}

## @name ::filer_snapmirror_is_started
## @summary Determines if snapmirror is started on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_snapmirror_is_started ($) {
    my($fc) = shift;
    return filer_service_is_started($fc,'snapmirror');
}

## @name ::filer_snapmirror_start
## @summary Starts snapmirror service on the filer if possible
## @description Returns 1 if service is now started, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_snapmirror_start ($) {
    my($fc) = shift;
    return filer_service_start ($fc,'snapmirror');
}

## @name ::filer_snapmirror_status
## @summary Determines snapmirror status
## @description Returns 1 if service is now started, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_snapmirror_status ($;$) {
    my $fc  = shift;

    my $cmd = "snapmirror status";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    if ($text =~ m/Snapmirror is on/) {
        return 1;
    } else {
        return 0;
    }
}

## @name ::filer_snapmirror_status_all
## @summary Reports the status of all snapmirrored relationships with a source
## and or destination  on this filer
## @description Returns a hash of all status attributes of all snapmirrored
## relationship on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_snapmirror_status_all ( $ ) {
    my $fc = shift;

    my ($text,$result) = filer_cli_say($fc,"snapmirror status", {retry=>'ok'});
    my (@lines) = filer_cli_split($text);

    # verify first line
    my $line = shift(@lines);

    ExitError("No snapmirror transfers ")
        if (!scalar(@lines));

    # verify second line
    $line = shift(@lines);
    ExitError("Invalid header line : $line")
        if ($line !~ /^Source\s+Destination\s+State\s+Lag\s+Status\s*$/);

    my %HoH;
    foreach $line (@lines) {
        ExitError("Invalid format : $line")
            if ($line !~ qr{^([-:/\w]+)\s+
                            ([-:/\w]+)\s+
                            ([a-zA-Z-]+)\s+
                            (-|\d\d:\d\d:\d\d)\s+
                            ([-a-zA-Z]+)\s*$}x);

        my $dest = $2;
        $HoH{$dest}{source} = $1;
        $HoH{$dest}{state} = $3;
        $HoH{$dest}{lag} = $4;
        $HoH{$dest}{status} = $5;
    }

    return(%HoH);
}

## @name ::filer_snapmirror_stop
## @summary Stops snapmirror service on the filer if possible
## @description Returns 1 if service is already stopped, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_snapmirror_stop ($) {
    my($fc) = shift;
    return filer_service_stop ($fc,'snapmirror');
}

## @name ::filer_snaprestore_initialize
## @summary Licenses and starts snaprestore service on the filer if possible
## @description Returns 1 if service is now started, otherwise 0
## @arg $fc Required: The filer connection object for the command line interface
sub filer_snaprestore_initialize ($) {
    my($fc) = shift;

    my($initialized) = 0;
    filer_service_license($fc,'snaprestore') if (!filer_service_is_licensed($fc,'snaprestore'));
    $initialized = 1;
    return $initialized;
}

## @name ::filer_software_delete
## @summary Returns the list of software files in /etc/software.
## @arg $fc Required: The filer connection object for the command line interface
## @arg @file Required: The file(s) to be deleted from the /etc/softare directory
sub filer_software_delete ($@) {
    my $fc  = shift;
    my @files  = @_;

    my $file = join(' ',@files);

    my $cmd = "software delete $file";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /software delete:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_software_exists
## @summary Returns 1 if the specified file is in the /etc/software directory.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $file Required: The file to be found in the /etc/softare directory
sub filer_software_exists ($$) {
    my $fc  = shift;
    my $file  = shift;

    my $cmd = "software list";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my(@files);
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /software list:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ( $line eq $file ) {
            return 1;
        }
    }
    return 0;
}

## @name ::filer_software_list
## @summary Returns the list of software files in /etc/software.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_software_list ($) {
    my $fc  = shift;

    my $cmd = "software list";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my(@files);
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /software list:/) {
           ExitSoftError("Unexpected error from '$cmd': $line");
        } else {
            push(@files, $line);
        }
    }

    return @files;
}

## @name ::filer_sparecore_disable
## @summary Disables the use of sparecore
## @arg $fc Required: The filer connection object for the command line interface
sub filer_sparecore_disable ($) {
    my $fc  = shift;

    my($text,$result) = filer_options_set($fc,"cf.takeover.on_panic","off");
}

## @name ::filer_sparecore_enable
## @summary Enables the use of sparecore
## @arg $fc Required: The filer connection object for the command line interface
sub filer_sparecore_enable ($) {
    my $fc  = shift;

    my($text,$result) = filer_options_set($fc,"cf.takeover.on_panic","on");
}

## @name ::filer_sparecore_is_ready
## @summary Returns 1 if next savecore will use sparecore approach
## @arg $fc Required: The filer connection object for the command line interface
sub filer_sparecore_is_ready ($) {
    my $fc  = shift;

    my $cmd = "savecore -i";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    if ($text =~ /sparecore dumpcore feature will be used/) {
        return 1;
    } else {
        return 0;
    }
}

## @name ::filer_sparecore_status_get
## @summary Returns percent complete value for the current sparecore
## @arg $fc Required: The filer connection object for the command line interface
sub filer_sparecore_status_get ($) {
    my $fc  = shift;

    my $complete = 0;

    my $cmd = "savecore -s";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    # What is the message text that appears???
    if ($text =~ /No savecore in progress/) {
        $complete = 100;
    } elsif ($text =~ /(\d+)%/) {
        $complete = $1;
    }
    return $complete;
}

## @name ::filer_spares_get
## @summary Returns the list of spare disks.
## @arg $fc Required: The filer connection object for the command line interface
sub filer_spares_get ($) {
    my $fc  = shift;

    my $cmd = "vol status -r";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my(@disks);
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /Spare disks (empty)/) {
           last;
        } elsif ($line =~ /^zero|^spare/) {
            my @words = split(/\s+/, $line);
            push(@disks, $words[1]);
        }
    }
    return @disks;
}

## @name ::filer_statit_get
## @summary Stops statit and gets values returned
## @example filer_statit_get($fc, "DAFS operations")
## @arg $fc Required: The filer connection object for the command line interface
## @arg $stat Optional: Statistic you want
## if $stat is missing, this function returns a hash of values in the form of
## $Hash{$heading}{Statistic}. For example, if you want to get the DAFS ops in
## the Miscellaneous Statistics section, you would get it like this:
## %hash=filer_statit_get($fc); $ops=$hash{'Miscellaneus'}{'DAFS operations'}
sub filer_statit_get ($;$) {
    my $fc  = shift;
    my $stat = shift;

    my $cmd = "statit -e";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});
    my %statitValues;
    my $heading;
    # I only have 6 that are returned in the hash. More can easily be added.
    # For example, if you want to get idle time in the hash, just add
    # '"idle time",' to the list. None of this is necessary if you are going
    # to use the optional arg $stat.
    my @statistics = (
        "DAFS operations",
        "NFS operations",
        "CIFS operations",
        "HTTP operations",
        "FCP operations",
        "iSCSI operations",
        );

    sub getvalue ($$) {
        my $op = shift;
        my $line = shift;
        my $value = $line;
        $value =~ s/.*\s+(\d+\.?\d+) $op.*/$1/;
        return $value;
    }

    my($line);
    my @info = filer_cli_split($text);
    foreach $line (@info) {
        if (defined($stat)) {
            if ($line =~ /$stat/) {
                my $value = getvalue($stat,$line);
                return $value;
            }
        }
        if ($line =~ /usage: statit/) {
            ExitSoftError("Command failed: $line");
        }
        if ($line =~ /Statistics/) {
            $heading = $line;
            $heading =~ s/.*\s+(\w+) Statistics.*/$1/;
        }

        foreach my $statistic (@statistics) {
            if ($line =~ /$statistic/) {
                my $value = getvalue($statistic,$line);
                $statitValues{$heading}{$statistic} = $value;
            }
        }
    }

    # shouldn't get this far if $stat was found
    if (defined($stat)) {
        ExitSoftError("Didn't find the value you were looking for");
    }
    return %statitValues;
}

## @name ::filer_statit_start
## @summary Start statistics gathering
## @description
## Runs 'statit -b' to gather statistics
## @arg $fc Required: The filer connection object for the command line interface
sub filer_statit_start {
    my($fc) = shift;

    my $cmd = "statit -b";
    my($text,$result) = filer_cli_diag_say($fc,$cmd);

    my @info = filer_cli_split($text);
    my $return = 1;
    foreach my $line (@info) {
        if ($line =~ /usage/) {
            $return=0;
        }
    }
    return $return;
}

## @name ::filer_statit_stop
## @summary Stop statit
## @description
## Runs 'statit -e' to gather statistics
## @arg $fc Required: The filer connection object for the command line interface
sub filer_statit_stop {
    my($fc) = shift;
    my $cmd = "statit -e";
    my($text,$result) = filer_cli_diag_say($fc,$cmd);

    my @info = filer_cli_split($text);
    my $return = 1;
    foreach my $line (@info) {
        if ($line =~ /usage/) {
            $return=0;
        }
    }
    return $return;
}

## @name ::filer_stats_vol_show
## @summary Show volume statistics
## @description
## Runs 'stats show volume:<vol>' to gather volume statistics
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Optional: Show stats of volume name. Default is all volumes
sub filer_stats_vol_show($;$) {
    my $fc = shift;
    my $vol = shift;

    my ($line, %stats);

    $vol = '' if (!defined($vol));
    my $cmd = "stats show volume:$vol";
    my ($text, undef) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});
    if ($text =~ /No instance|invalid object name|Unknown/i) {
        ExitError("Could not get volume stats: $text");
    }
    my @vol_stats = filer_cli_split($text);
    foreach $line (@vol_stats) {
        my (undef, $name, $key, $value) = split(/:/, $line);
        if (defined($name) && defined($key) && defined($value)) {
            $stats{$name}{$key} = $value;
        }
    }
    if (!%stats) {
        ExitError('Could not get volume stats');
    }
    return %stats;
}

## @name ::filer_stats_aggr_show
## @summary Show aggr statistics
## @description
## Runs 'stats show aggregate:<aggr>' to gather aggr statistics
## @arg $fc Required: The filer connection object for the command line interface
## @arg $aggr Optional: Show stats of aggr name. Default is all aggrs
sub filer_stats_aggr_show($;$) {
    my $fc = shift;
    my $aggr = shift;

    my ($line, %stats);

    $aggr = '' if (!defined($aggr));
    my $cmd = "stats show aggregate:$aggr";
    my ($text, undef) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});
    if ($text =~ /No instance|invalid object name|Unknown/i) {
        ExitError("Could not get aggr stats: $text");
    }
    my @aggr_stats = filer_cli_split($text);
    foreach $line (@aggr_stats) {
        my (undef, $name, $key, $value) = split(/:/, $line);
        if (defined($name) && defined($key) && defined($value)) {
            $stats{$name}{$key} = $value;
        }
    }
    if (!%stats) {
        ExitError('Could not get aggr stats');
    }
    return %stats;
}

## @name ::filer_sysconfig_A_check
## @summary Check 'sysconfig -A' output and return text output.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $slot Optional: You can provide a slot number for this command
sub filer_sysconfig_A_check ($) {
    my $fc = shift;
    my $slot = shift;
    $slot = "" unless(defined($slot));

    my $cmd = "sysconfig -A $slot";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ( ($line =~ /sysconfig:\s/) && ($line !~ /sysconfig: There are no configuration errors/) ) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        }
    }
    return $text;
}

## @name ::filer_sysconfig_a_check
## @summary Check 'sysconfig -a' output and return text output.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $slot Optional: You can provide a slot number for this command
sub filer_sysconfig_a_check ($;$) {
    my $fc = shift;
    my $slot = shift;
    $slot = "" unless(defined($slot));

    my $cmd = "sysconfig -a $slot";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        }
    }
    return $text;
}

## @name ::filer_sysconfig_c_check
## @summary Check 'sysconfig -c' output and return text output.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $expected Optional: expected output. Default is no errors expected.
sub filer_sysconfig_c_check ($;$) {
    my $fc = shift;
    my $expected = shift;
    if (!defined($expected)) {
        $expected = qr/sysconfig: There are no configuration errors\./;
    }

    my $cmd = "sysconfig -c";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my $found = 0;
    if ($text !~ /$expected/) {
        ExitSoftError("Output from '$cmd' did not match '$expected'");
    }
    return $text;
}

## @name ::filer_sysconfig_v_check
## @summary Check 'sysconfig -v' output
## @arg $fc Required: The filer connection object for the command line interface
## @arg $slot Optional: adapter name
sub filer_sysconfig_v_check ($;$) {
    my $fc = shift;
    my $slot = shift;
    if (!defined($slot)) {
       $slot = ""
    }

    my $cmd = "sysconfig -v $slot";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $slotlinecount = 0;
    my $got_slot = "";
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /\s+slot .*:/) {
            # A new slot is being described
            # See if we need to verify the previous results
            if ($slotlinecount > 0) {
            }
            # Determine new slot id
            $got_slot = $line;
            $got_slot =~ s/\s+slot (.*):/$1/;
            # If only a specific slot was requested, verify we got it
            if ($slot ne "") {
                if ($got_slot !~ /$slot/) {
                    ExitSoftError("Unexpected slot: '$got_slot' displayed, expected: '$slot'");
                }
            }
        }
    }
    return $text;
}

## @name ::filer_sysstat_check
## @summary Check sysstat output for requested protocol
## @example filer_sysstat_check($fc, "CIFS")
## @arg $fc Required: The filer connection object for the command line interface
## @arg $protocol Required: Protocol you want to verify with systat -x and -?
sub filer_sysstat_check ($$) {
    my $fc = shift;
    my $protocol = shift;

    my $option = "";
    ($protocol =~ /DAFS/i)  and $option = "-d";
    ($protocol =~ /FCP/i)   and $option = "-f";
    ($protocol =~ /iSCSI/i) and $option = "-i";
    my $cmd = "sysstat -c 1 $option 1";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    if ($text !~ /$protocol/msi) {
        ExitSoftError("Output from ($cmd) is missing $protocol protocol heading");
    }

    $cmd = "sysstat -c 1 -x 1";
    ($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    if ($text !~ /$protocol/msi) {
        ExitSoftError("Output from ($cmd) is missing $protocol protocol heading");
    }
}

## @name ::filer_sysstat_get
## @summary Get protocol information from sysstat -x
## @arg $fc Required: The filer connection object for the command line interface
## @arg $options Optional: Options to the sysstat command.
## Currently supports FILER_SYSSTAT_COUNT and FILER_SYSSTAT_INTERVAL
## Example:
##    my %sysstat_output = filer_sysstat_get($fc,
##                    {
##                    'FILER_SYSSTAT_COUNT'=> $FILER_SYSSTAT_COUNT,
##                    'FILER_SYSSTAT_INTERVAL'=> $FILER_SYSSTAT_INTERVAL,
##                    });
##
## And then info can be obtained like:
## my $average_nfs_ops = $sysstat_output{'NFS'}{'Avg'};
##
## Supported protocols are NFS,CIFS,DAFS,FCP,HTTP,iSCSI and also CPU and Disk Utilization.
## Supported values foreach protocol are Total,Min,Max, and Avg.
##
sub filer_sysstat_get ($;$) {
    my ($fc, $options) = @_;
    my $count = $options->{'FILER_SYSSTAT_COUNT'};
    $count = 1 unless ( (defined($count)) && ($count ne ""));
    my $interval = $options->{'FILER_SYSSTAT_INTERVAL'};
    $interval = 15 unless ( (defined($interval)) && ($interval ne ""));

    my $after;
    my @protocols = qw(CPU NFS CIFS HTTP FCP iSCSI DAFS DISK_UTIL);
    my %sysstatHash;

    # intialize all the values
    foreach my $protocol (@protocols) {
        $sysstatHash{$protocol}{'Total'} = 0;
        $sysstatHash{$protocol}{'Avg'} = 0;
        $sysstatHash{$protocol}{'Min'} = 0;
        $sysstatHash{$protocol}{'Max'} = 0;
    }

    my $cmd = "sysstat -c $count -s -x $interval";
    my($text,$result) = filer_cli_say($fc,$cmd,$count*$interval+60);
    my $value_type = "Total";
    my $got_DAFS = 0;
    foreach my $line (split(/\n/, $text)) {
        if ($line =~ m/\sDAFS|OTHER\s/) {
            $got_DAFS = 1;
        }
        if ($line =~ /%/) {
            $line =~ s/^\s+//g;
            my @opsline = split(/\s+/, $line);
            $opsline[0] =~ s/%//;
            $sysstatHash{'CPU'}{$value_type} = $sysstatHash{'CPU'}{$value_type}
            + $opsline[0];
            $sysstatHash{'NFS'}{$value_type} = $sysstatHash{'NFS'}{$value_type}
            + $opsline[1];
            $sysstatHash{'CIFS'}{$value_type}=$sysstatHash{'CIFS'}{$value_type}
            + $opsline[2];
            $sysstatHash{'HTTP'}{$value_type}=$sysstatHash{'HTTP'}{$value_type}
            + $opsline[3];
            $opsline[15] =~ s/%//;
            $sysstatHash{'DISK_UTIL'}{$value_type} = $sysstatHash{'DISK_UTIL'}{$value_type}
            + $opsline[15];
            if ($got_DAFS) {
                # DAFS is absolete in GB release
                $sysstatHash{'DAFS'}{$value_type}=$sysstatHash{'DAFS'}{$value_type} + $opsline[16];
            }
            $sysstatHash{'FCP'}{$value_type} = $sysstatHash{'FCP'}{$value_type}
                + $opsline[16 + $got_DAFS];
            $sysstatHash{'iSCSI'}{$value_type}=$sysstatHash{'iSCSI'}{$value_type}
                + $opsline[17 + $got_DAFS];
        } elsif ($line =~ /Min/) {
            $value_type = "Min";
        } elsif ($line =~ /Max/) {
            $value_type = "Max";
        } elsif ($line =~ /Avg/) {
            $value_type = "Avg";
        } elsif ($line =~ /Summary Statistics/) {
            $sysstatHash{'Summary Statistics'} = $line;
        }
    }
    return %sysstatHash;
}

## @name ::filer_sysstat_ops_get
## @summary Get operation count for a protocol using sysstat (without -x option)
## @example filer_sysstat_ops_get($fc, "NFS", 10)
## @arg $fc Required: The filer connection object for the command line interface
## @arg $protocol Required: Protocol you want to get average ops
## @arg $count Optional: The number of lines of systat output to get average from
sub filer_sysstat_ops_get ($$;$) {
    my $fc  = shift;
    my $protocol = shift;
    my $count = shift;
    my $after;
    my $ops = 0;
    my $totalops = 0;
    my $counter = 0;
    my $option = "";
    ($protocol =~ /DAFS/i)  and $option = "-d";
    ($protocol =~ /FCP/i)   and $option = "-f";
    ($protocol =~ /iSCSI/i) and $option = "-i";
    $count = 1 unless (defined($count));

    my $cmd = "sysstat -c $count $option 1";
    my($text,$result) = filer_cli_say($fc,$cmd);
    foreach my $line (split(/\n/, $text)) {
        if ($line =~ /%/) {
            $counter++;
            my @opsline = split(/\s+/, $line);
            ($protocol =~ /NFS/i)   and $ops = $opsline[2];
            ($protocol =~ /CIFS/i)  and $ops = $opsline[3];
            ($protocol =~ /HTTP/i)  and $ops = $opsline[4];
            ($protocol =~ /DAFS/i)  and $ops = $opsline[4];
            ($protocol =~ /FCP/i)   and $ops = $opsline[4];
            ($protocol =~ /iSCSI/i) and $ops = $opsline[4];
            $totalops += $ops;
        }
    }
    return $totalops/$counter;
}

## @name ::filer_sysstat_x_ops_get
## @summary Get average operation count for a protocol using sysstat -x
## @description If an array is requested for a specific protocol,
## the following values are returned:
## Total, Min, Max, Average
## @arg $fc Required: The filer connection object for the command line interface
## @arg $protocol Required: the protocol of interest
## @arg $count Optional: The number of lines of systat output to get average from
sub filer_sysstat_x_ops_get ($$;$) {
    my $fc  = shift;
    my $protocol = shift;
    my $count = shift;
    my $after;
    $count = 1 unless ( (defined($count)) && ($count ne ""));

    my @protocols = qw(NFS CIFS HTTP FCP iSCSI DAFS);
    my %sysstatHash;

    # intialize all the values
    foreach my $proto (@protocols) {
        $sysstatHash{$proto}{'Total'} = 0;
        $sysstatHash{$proto}{'Avg'} = 0;
        $sysstatHash{$proto}{'Min'} = 0;
        $sysstatHash{$proto}{'Max'} = 0;
    }

    my $cmd = "sysstat -c $count -s -x 1";
    my($text,$result) = filer_cli_say($fc,$cmd,$count+60);
    my $value_type = "Total";
    my $got_DAFS = 0;
    foreach my $line (split(/\n/, $text)) {
        if ($line =~ m/\sDAFS\s/) {
            $got_DAFS = 1;
        }
        if ($line =~ /%/) {
            $line =~ s/^\s+//g;
            my @opsline = split(/\s+/, $line);
            $sysstatHash{'NFS'}{$value_type} = $sysstatHash{'NFS'}{$value_type} + $opsline[1];
            $sysstatHash{'CIFS'}{$value_type} = $sysstatHash{'CIFS'}{$value_type} + $opsline[2];
            $sysstatHash{'HTTP'}{$value_type} = $sysstatHash{'HTTP'}{$value_type} + $opsline[3];
            if ($got_DAFS) {
                # DAFS removed from 7.1.0.1 release
                $sysstatHash{'DAFS'}{$value_type} = $sysstatHash{'DAFS'}{$value_type} + $opsline[16];
            }
            $sysstatHash{'FCP'}{$value_type} = $sysstatHash{'FCP'}{$value_type} + $opsline[16 + $got_DAFS];
            $sysstatHash{'iSCSI'}{$value_type} = $sysstatHash{'iSCSI'}{$value_type} + $opsline[17 + $got_DAFS];
        } elsif ($line =~ /Min/) {
            $value_type = "Min";
        } elsif ($line =~ /Max/) {
            $value_type = "Max";
        } elsif ($line =~ /Avg/) {
            $value_type = "Avg";
        } elsif ($line =~ /Summary Statistics/) {
            $sysstatHash{'Summary Statistics'} = $line;
        }
    }

    if ($protocol =~ /NFS/i) {
        if (wantarray) {
            return (%{$sysstatHash{'NFS'}});
        } else {
            return $sysstatHash{'NFS'}{'Avg'};
        }
    } elsif ($protocol =~ /CIFS/i) {
        if (wantarray) {
            return (%{$sysstatHash{'CIFS'}});
        } else {
            return $sysstatHash{'CIFS'}{'Avg'};
        }
    } elsif ($protocol =~ /HTTP/i) {
        if (wantarray) {
            return (%{$sysstatHash{'HTTP'}});
        } else {
            return $sysstatHash{'HTTP'}{'Avg'};
        }
    } elsif ($protocol =~ /DAFS/i) {
        if (wantarray) {
            return (%{$sysstatHash{'DAFS'}});
        } else {
            return $sysstatHash{'DAFS'}{'Avg'};
        }
    } elsif ($protocol =~ /FCP/i) {
        if (wantarray) {
            return (%{$sysstatHash{'FCP'}});
        } else {
            return $sysstatHash{'FCP'}{'Avg'};
        }
    } elsif ($protocol =~ /iSCSI/i) {
        if (wantarray) {
            return (%{$sysstatHash{'iSCSI'}});
        } else {
            return $sysstatHash{'iSCSI'}{'Avg'};
        }
    } elsif ($protocol =~ /all/i) {
        return %sysstatHash;
    } else {
        logcomment("Bad Protocol");
        return 0;
    }
}

## @name ::filer_systemboard_fw_revision_get
## @summary Returns the System Board firmware revision number of the specified filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_systemboard_fw_revision_get ($) {
    my $fc = shift;

    my $cmd = "sysconfig -v 0";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $field = '';
    my $found_slot = 0;
    foreach my $line (@info) {
        if ($line =~ /sysconfig:\s/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /\s+slot\s0: System Board /) {
            $found_slot = 1;
        }
        if ($found_slot eq 1) {
            if ($line =~ /^\s+ Firmware release:/) {
                $field = $line;
                $field =~ s/^\s+ Firmware release:\s+(\S+).*/$1/;
                last;
            }
        }
    }
    return $field;
}

## @name ::filer_uptime_check
## @summary Verify uptime output has all the correct fields
## @example filer_uptime_check($fc);
## @arg $fc Required: The filer connection object for the command line interface
sub filer_uptime_check ($) {
    my $fc = shift;
    my $cmd = "uptime";
    # Possible output formats
    #  2:58pm up  7 days, 19:16 795217260 NFS ops, 1017333 CIFS ops, 1639 HTTP ops, 0 DAFS ops, 0 FCP ops, 0 iSCSI ops
    #  3:49pm up 20 mins, 28 NFS ops, 0 CIFS ops, 0 HTTP ops, 0 DAFS ops, 2 FCP ops, 0 iSCSI ops
    # 10:43am up  1:02 49567 NFS ops, 0 CIFS ops, 0 HTTP ops, 0 DAFS ops, 10 FCP ops, 0 iSCSI ops
    my ($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    if ($text !~ /^\s+\d+:\d\d[a|p]m\sup\s+\d+/) {
        ExitSoftError("Missing the current time or uptime value at the beginning of uptime output");
    }
    # remove DAFS from TT.0.1 and GordonBiersch release
    foreach my $protocol ("NFS","CIFS","HTTP","FCP","iSCSI") {
        if ($text !~ /\s+\d+\s+$protocol ops/) {
            ExitSoftError("Missing $protocol ops in the uptime output");
        }
    }
}

## @name ::filer_uptime_ops_get
## @summary Get total ops for requested protocol
## @example filer_uptime_ops_get($fc, "CIFS");
## @arg $fc Required: The filer connection object for the command line interface
## @arg $protocol Required: Protocol you want op count for (Ex. FCP, iSCSI, NFS)
sub filer_uptime_ops_get ($$) {
    my $fc = shift;
    my $protocol = shift;
    my $cmd = "uptime";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    if ($text =~ / (\d+) $protocol ops/) {
        $text =~ s/.* (\d+) $protocol ops.*/$1/
    } else {
        ExitScript("Invalid protocol $protocol, not found in uptime output");
    }
    return $text;
}

## @name ::filer_user_add
## @summary Add a user
## @arg $fc Required: The filer connection object for the command line interface
## @arg $user Required: Name of the user to add
## @arg $password Required: Password for the user
## @arg $group Optional: Group for the user, default is "Administrators"
sub filer_user_add ($$$;$) {
    my $fc  = shift;
    my $user = shift;
    my $password = shift;
    my $group = shift;
    $group = "Administrators" unless ($group);
    my $useradded = 0;

    my $ver = get_wafl_version($fc->hostp());
    if ($fc->{conntype} =~ /rsh/) {
        my $cmd;
        if ($ver < WV_ANCHORSTEAM_MIN) {
            $cmd = "useradmin useradd $user $password";
        } else{
            $cmd = "useradmin user add $user -p $password -g $group";
        }
        my($text,$result) = filer_cli_say($fc,$cmd,$FILER_CLI_SAY_TIMEOUT);

        my @info = filer_cli_split($text);
        my $userdeleted = 0;
        foreach my $line (@info) {
            if ($line =~ /User <$user> added./) {
                $useradded = 1;
            }
        }
    } else {
        if ($ver < WV_ANCHORSTEAM_MIN) {
            say($fc, "useradmin useradd $user", -blind);
        } else{
            say($fc, "useradmin user add $user -g $group", -blind);
        }
        while (1) {
            my $prompt;
            (undef, undef, undef, $prompt) = say($fc, "", -prompt,
                ['New password:','Retype new password:']);
            if ($prompt =~ 'New password') {
                say($fc, "$password", -blind)
            } elsif ($prompt =~ 'Retype new password:') {
                say($fc, "$password", "User <$user> added.");
                $useradded = 1;
                last;
            } else {
                say($fc, "", "user add failed");
            }
        }
    }
    return $useradded;
}

## @name ::filer_user_delete
## @summary Delete a user
## @arg $fc Required: The filer connection object for the command line interface
## @arg $user Required: Name of the user to delete
sub filer_user_delete ($$) {
    my $fc  = shift;
    my $user = shift;

    my $ver = get_wafl_version($fc->hostp());
    my $cmd = "";
    if ($ver < WV_ANCHORSTEAM_MIN) {
        $cmd = "useradmin userdel $user";
    } else{
        $cmd = "useradmin user delete $user";
    }
    my($text,$result) = filer_cli_say($fc,$cmd,$FILER_CLI_SAY_TIMEOUT);

    my @info = filer_cli_split($text);
    my $userdeleted = 0;
    foreach my $line (@info) {
        if ($line =~ /User <$user> deleted./) {
            $userdeleted = 1;
        }
    }
    return $userdeleted;
}

## @name ::filer_user_exists
## @summary Check to see if a user exists
## @arg $fc Required: The filer connection object for the command line interface
## @arg $user Required: Name of the user to check
sub filer_user_exists ($$) {
    my $fc  = shift;
    my $user = shift;

    my $cmd = "useradmin userlist $user";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    my $userexists = 0;
    foreach my $line (@info) {
        if ($line =~ /Name: $user/) {
            $userexists = 1;
        }
    }

    return $userexists;
}

## @name ::filer_version_get
## @summary Obtain version of ONTAP on a filer, returns '' if version command fails
## @arg $fc Required: The filer connection object for the command line interface
sub filer_version_get ($) {
    my $fc  = shift;
    my $version = '';

    my $cmd;
    if ($fc->hostp()->hosttype() =~ m/filer-ng/) {
        $cmd = "modvers -v";
        my $filer_version = filer_cli_say( $fc, $cmd, {retry=>'ok',useresult=>0});
        # maytag.ko, Build 664514, BuildTime 050707220600PDT, BuildPath /rlse/ontapng/RtrickyN_070507_2000/bundledefs/.., BuildOptions
        # maytag.ko, Build p4down, BuildTime 081307201351PDT, BuildPath /rlse/ontapng/RtrickyN_070813_2000, BuildOptions
        # maytag.ko, Build , BuildTime , BuildPath /rlse/ontapng/boilerms1N_071006_0030, BuildOptions x86
        # maytag.ko, Build , BuildTime , BuildPath /x/eng/rlse11/ontapng/nightly/mainngN_081022_1330/, BuildOptions x86_64.debug
        if ($filer_version =~ m/maytag\.ko, Build (\S*), BuildTime (\S*), BuildPath (\S*), BuildOptions\s*(.*?)\s*$/m) {
            my $change_number = $1;  # change number or 'p4down'
            my $build_timestamp = $2;
            my $release = $3;
            my $build_options = $4;
            # "/abc/def/xyz/" -> "/abc/def/xyz"
            $release =~ s#/$##g;
            # "/abc/def/xyz/bundledefs/.." -> "/abc/def/xyz"
            $release =~ s#/[^/]+/\.\.$##g;
            # "/abc/def/xyz" -> "xyz"
            $release =~ s#.*/##g;
            $version = $release;
            if ($build_options) {
                $version .= "<$build_options>";
            }
        }
    } else {
        $cmd = "version -v";
        my($text,$result) = filer_cli_say($fc,$cmd,{allowerror=>1,retry=>'ok'});

        if ($text !~ m/NetApp Release|Data ONTAP/) {
            # burt 142788 - "version.thpl needs to account for base states of 'ok' prompt and maintenance mode"
            # (BUT ONLY if NOT using a simulator
            if ($fc->hostp()->hosttype() !~ m/simulator/ and $fc->conntype() =~ m/serial|pm_cons/) {
                # Try the OK/CFE prompt command to return the version
                #ok printenv last-OS-booted-ver
                #last-OS-booted-ver    NetApp Release RanchorsteamN_040922_2215
                $cmd = "printenv last-OS-booted-ver";
                ($text,$result) = filer_cli_say($fc,$cmd,{allowerror=>1,retry=>'ok'});
            }
       }
       my @info = filer_cli_split($text);
       foreach my $line (@info) {
           if ($line =~ /(?:NetApp|Data ONTAP) Release\s+(.*)/) {
               #Data ONTAP Release R7.1.0.1xN_060201_0100: Wed Feb  1 05:21:02 PST 2006 (IBM)  <8>
               $version = $1;
               # note: date field has colons, but they are not followed by a space
               $version =~ s/[\s:].*//;
           }
           if ($line =~ /\s*\((.*)\)/) {
               my $brand = $1;
               # When a kernel has been re-branded
               # the re-branded company id is enclosed in parentheses
               $version = "$version($brand)";
           }
           if ($line =~ /\s*<(.*)>/) {
               my $kernel = $1;
               # note When special kernels have been installed
               # the kernel id follows on a second line enclosed in angle brackets
               # This has been seen on the same line with Scrimshaw and with simulator private builds:
               # NetApp Release /x/eng/build5/scratch/doucette/doucette_gb_145605_390901: Fri Jan 21 10:02:20 PST 2005  <td>
               return "$version<$kernel>";
           }
       }
    }
   return $version;
}



## @name ::filer_vfiler_command
## @summary converts the input command into its current simulated vfiler context form
## @arg $fc Required: The filer connection object for the command line interface
sub filer_vfiler_command ($$) {
    my $fc  = shift;
    my $cmd = shift;

    my $vfiler = filer_vfiler_context_simulate($fc);
    if ($vfiler ne '') {
        $cmd = "vfiler run $vfiler $cmd";
    }
    return $cmd
}

## @name ::filer_vfiler_context_get
## @summary Determine the vfiler context on the filer using the specified connection
## @arg $fc Required: The filer connection object for the command line interface
sub filer_vfiler_context_get ($) {
    my $fc  = shift;

    # Determine the vfiler context from the current prompt
    my $current_vfiler = 'vfiler0';
    my $conn_type = $fc-> conntype();
    if ($conn_type !~ /rsh/) {
        say($fc, "\r", -nocr, -swallow);
        if ($RESPONSE_PROMPT) {
            if ($RESPONSE_PROMPT =~ m/@/) {
                ($current_vfiler) = split('@',$RESPONSE_PROMPT,1);
            }
        } else {
            print("No RESPONSE_PROMPT seen while testing for vfiler context\n");
        }
    }
    return $current_vfiler;
}

## @name ::filer_vfiler_context_off
## @summary Drop any vfiler context from the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_vfiler_context_off ($) {
    my $fc  = shift;

    my $conn_type = $fc-> conntype();
    if ($conn_type !~ /rsh/) {
        my $cmd = "vfiler context vfiler0";
        my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

        my @info = filer_cli_split($text);
        foreach my $line (@info) {
            if ($line =~ /vfiler context:/) {
                ExitError("Unexpected error from '$cmd': $line");
            }
        }
    }
}

## @name ::filer_vfiler_context_set
## @summary Establish a new vfiler context on the filer, return the old one
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vfiler Required: The vfiler name
sub filer_vfiler_context_set ($$) {
    my $fc  = shift;
    my $vfiler = arg_vfiler(shift);

    # Determine the vfiler context
    my $current_vfiler = filer_vfiler_context_get($fc);

    if ($current_vfiler ne $vfiler) {
        my $cmd = "vfiler context $vfiler";
        my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

        my @info = filer_cli_split($text);
        foreach my $line (@info) {
            if ($line =~ /vfiler context:/) {
                ExitError("Unexpected error from '$cmd': $line");
            }
        }
        print("Vfiler context changed from '$current_vfiler' to '$vfiler'\n");
    }
    $fc->{hostp}->{'vfiler_simulated_context'} = $vfiler;
    return $current_vfiler;
}

## @name ::filer_vfiler_context_simulate
## @summary Defines the vfiler context for future vfilerized commands the filer and returns the previous value.
## @description
## This function defines the vfiler context for a filer whenever future vfilerized commands
## are to be issued to that filer. The filer_vfiler_command function uses this setting
## to vfilerize the command to be executed.
## Setting the $vfiler to '' turns off the use of 'vfiler run $vfiler $cmd'
## within the filer_vfiler_command function.
## Otherwise,any vfiler name including 'vfiler0' may be specified.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vfiler Optional: The vfiler name to be used with all vfilerized commands.
## If none is given this function just returns the current value.
sub filer_vfiler_context_simulate ($;$) {
    my $fc  = shift;
    my $vfiler = shift;

    my $previous_context = $fc->{hostp}->{'vfiler_simulated_context'};

    if (!defined($previous_context)) {
        $previous_context = '';
    }

    if (defined($vfiler)) {
        if ($vfiler ne '') {
            $vfiler = arg_vfiler($vfiler);
        }
        # The value passed needs to be stored within the $fh object so that it can be
        # managed from one place for each filer
        if ($vfiler eq 'vfiler0') {
            $vfiler = '';
        }
        $fc->{hostp}->{'vfiler_simulated_context'} = $vfiler;
    }
    # Make sure we have the filer OUT of vfiler context
    filer_vfiler_context_off ($fc);
    return $previous_context;
}

## @name ::filer_vfiler_create
## @summary Create a vfiler on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vfiler Required: The vfiler name
## @arg $ipaddrs Required: a string of blank delimited IP addresses
## @arg $paths Required: a string of blank delimited paths (volumes or qtrees)
## @arg $ipspace Optional: an ipspace name to define appropriate interfaces
sub filer_vfiler_create ($$$$;$) {
    my $fc  = shift;
    my $vfiler = arg_vfiler(shift);
    my $ipaddrs = arg_ipaddrs(shift);
    my $paths = arg_paths(shift);
    my $ipspace = arg_ipspace(shift);

    # vfiler create vfilername [-n] [-s ipspace ] -i ipaddr [-i ipaddr ]... path [ path ...]
    my $cmd = "vfiler create $vfiler";
    if ($ipspace) {
        $cmd .= " -s ".$ipspace;
    }
    $ipaddrs =~ s/\s+/ /g;
    my @ips = split(" ",$ipaddrs);
    foreach my $ip (@ips) {
        # Add this IP address to the vfiler command
        $cmd .= " -i $ip";
    }
    $cmd .= " $paths";
    my($text,$result) = filer_cli_say($fc,$cmd);
    # This was shown to be an interactive command under AS.1:
    # exportfs [vf1]: Could not open for reading /etc/exports
    #  The etc configuration directory for vfiler "vf1" is /vol/vf1_vol/etc.
    #
    # Setting up vfiler vf1
    # Wed Feb 23 16:47:17 GMT [vf1@rc:notice]: Vfiler vf1 initialized.
    # Configure vfiler IP address 10.60.153.125? [y]:
    # y
    # Interface to assign this address to {e0a, e0b, e0c, e0d, tvr3}: e0a
    # Netmask to use: [255.255.255.0]:
    #         The administration host is given root access to the filer's
    #         /etc files for system administration.  To allow /etc root access
    #         to all NFS clients enter RETURN below.
    # Please enter the name or IP address of the administration host:
    # Do you want to run DNS resolver? [n]:
    # Do you want to run NIS client? [n]:
    # ifconfig (warning): route already exists
    #    (probably due to another interface on the same subnet.)
    # Wed Feb 23 17:11:48 GMT [10/100/1000-VI/e0a:info]: Ethernet e0a: Link up
    # Wed Feb 23 17:11:48 GMT [10/100/1000-VI/e0a:warning]: Autonegotiation failed, link operating at 100 Mbps half-duplex
    # Default password for root on vfiler vf1 is "".
    # New password:Wed Feb 23 17:11:48 GMT [vf1@useradmin.added.deleted:info]: The user 'root' has been added.

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ m/does not exist/) {
            # ^Error: ipspace e5 does not exist
            ExitError("Vfiler not created by '$cmd': $line");
        } elsif ($line =~ m/The IP addresses of vfiler ".*" are still unconfigured./) {
            # This seems to be a warning message with result = 0
        # Don't warn since this is normal operation.  Should handle the
        # ifconfig down/up in this function.
        # logresult('WARN',"Vfiler IP address not configured by '$cmd': $line");
        } elsif ($line =~ m/ not created|would be exceeded| and try again/) {
            # 'Maximum allowed vfilers would be exceeded.'
            # 'Operation would result in movement of mapped LUNs in /vol/vol0/vf2. Unmap any LUNs affected by this operation and try again.'
            # 'Vfiler vf2 not created'    <- message appears after all known messages
            ExitError("Vfiler not created by '$cmd': $line");
        } elsif ($line =~ /vfiler create:/) { # This message format is not supported at Scrimshaw release
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vfiler_destroy
## @summary Destroy a vfiler on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vfiler Required: The vfiler name
sub filer_vfiler_destroy ($$) {
    my $fc  = shift;
    my $vfiler = arg_vfiler(shift);

    # vfiler destroy [-f] vfilername
    my $cmd = "vfiler destroy -f $vfiler";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vfiler destroy:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vfiler_limit_get
## @summary Determine the limit on the number of vfilers
## @description
## Returns the scalar current limit, or the array of (current limit, platform limit, current count)
## @arg $fc Required: The filer connection object for the command line interface
sub filer_vfiler_limit_get ($) {
    my $fc  = shift;

    my $limit = 0;
    my $platform_limit = 0;
    my $current = 0;

    # vfiler limit [<limit>]
    my $cmd = "vfiler limit";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vfiler limit:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /Current limit:\s*(\d+)/) {
            $limit = $1;
        } elsif ($line =~ /Platform hard limit:\s*(\d+)/) {
            $platform_limit = $1;
        } elsif ($line =~ /Current in use:\s*(\d+)/) {
            $current = $1;
        } else {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
    if (wantarray) {
        return ($limit,$platform_limit,$current);
    } else {
        return $limit;
    }
}

## @name ::filer_vfiler_limit_set
## @summary Determine the limit on the number of vfilers
## @arg $fc Required: The filer connection object for the command line interface
sub filer_vfiler_limit_set ($$) {
    my $fc  = shift;
    my $limit  = shift;

    # vfiler limit [<limit>]
    my $cmd = "vfiler limit $limit";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vfiler limit:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vfiler_start
## @summary Start a vfiler on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vfiler Required: The vfilertemplate (vfiler name or *)
sub filer_vfiler_start ($$) {
    my $fc  = shift;
    my $vfiler = arg_vfiler(shift);

    # vfiler start vfilertemplate
    my $cmd = "vfiler start $vfiler";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /does not exist/) {
            ExitError("Vfiler, $vfiler, has not been defined");
        } elsif ($line =~ /vfiler start:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vfiler_status_get
## @summary Returns a hash of all vfiler attributes for the specified vfiler
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vfiler Required: The vfiler name connection object for the command line interface
sub filer_vfiler_status_get ($$) {
    my $fc  = shift;
    my $vfiler  = shift;

    my %vfiler_data;

    my $cmd = "vfiler status -a";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    if ($text =~ m/(Use of vfilers not licensed|Use of vfilers requires a multistore license|vfiler command not allowed|VFilers are not yet fully initialized|vfiler cannot be run)/) {
	return();
    }
    if ($text !~ m/Use of vfilers not licensed/) {
        my @info = filer_cli_split($text);
        my $current_vfiler = '';
        foreach my $line (@info) {
            if ($text =~ m/vfiler status:/s) {
                ExitSoftError("Unexpected error from '$cmd': $text");
                last;
            }
            if ($line =~ /^\S+\s+running/) {
                # We are starting the description of a new vfiler
                $current_vfiler =$line;
                $current_vfiler =~ s/\s.*//;
            } else {
                if ($current_vfiler eq $vfiler) {
                    if ($line =~ m/:/) {
                        $line =~ s/^\s*//;
                        $line =~ s/:\s+/:/;
                        my ($key,$value) = split(":",$line,2);
                        $key =~ s/^\s*//;
                        # copy over pertinent data to the has
                        if ($key =~ m/ipspace/i) {
                            $vfiler_data{$key} = $value;
                        } elsif ($key =~ m/IP address/i) {
                            # IP address: 10.60.157.146 [e4]
                            my ($ip,$ifc) = split(" ",$value,2);
                            $ifc =~ s/[\[\]]//g;
                            $vfiler_data{'IP'} = $ip;
                            $vfiler_data{'IFC'} = $ifc;
			    push(@{ $vfiler_data{'Interfaces'} }, $ifc);
			    push(@{ $vfiler_data{'IP Adresses'}{$ifc} }, $ip);
                        } elsif ($key =~ m/Path/i) {
                            # Path: /vol/vol0/kc3050-1-vf1 [/etc]\n
			    if ($value =~ /(\S+)\s+\[\/etc\]/) {
                                $vfiler_data{$key} = $1;
                                #print("Vfiler{$key}='$1'\n");
			    }
                            $value =~ s/\s.*//;  # drop off [/etc] text
			    push(@{ $vfiler_data{'Paths'} }, $value);
                            #print("\@Vfiler{'Paths'}<'$value'\n");
                        } elsif ($key =~ m/UUID/) {
                            $vfiler_data{$key} = $value;
                            #print("Vfiler{$key}='$value'\n");
                        } elsif ($key =~ m/Protocols allowed/) {
                            $vfiler_data{$key} = $value;
                            #print("Vfiler{$key}='$value'\n");
                        } elsif ($key =~ m/Protocols disallowed/) {
                            $vfiler_data{$key} = $value;
                            #print("Vfiler{$key}='$value'\n");
                        }
                    }
                }
            }
        }
    }
    if (0 == keys %vfiler_data) {
        ExitSoftError("Vfiler $vfiler does not appear to be defined");
    }
    return %vfiler_data;
}

## @name ::filer_vfiler_stop
## @summary Start a vfiler on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vfiler Required: The vfilertemplate (vfiler name or *)
sub filer_vfiler_stop ($$) {
    my $fc  = shift;
    my $vfiler = arg_vfiler(shift);

    # vfiler stop vfilertemplate
    my $cmd = "vfiler stop $vfiler";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /does not exist/) {
            ExitError("Vfiler, $vfiler, has not been defined");
        } elsif ($line =~ /vfiler stop:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_sync
## @summary Issues a sync command on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_sync ($) {
    my $fc = shift;
    my $cmd = "sync";
    return filer_cli_diag_say($fc,$cmd);
}

## @name ::filer_vfilers_get
## @summary Returns a list of all vfilers on the filer
## @arg $fc Required: The filer connection object for the command line interface
sub filer_vfilers_get ($) {
    my $fc  = shift;

    my @existing;

    my $cmd = "vfiler status";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    if ($text =~ m/(Use of vfilers not licensed|Use of vfilers requires a multistore license|vfiler command not allowed|VFilers are not yet fully initialized|vfiler cannot be run)/) {
        return ();
    }
    if ($text =~ m/vfiler status:/s) {
        ExitSoftError("Unexpected error from '$cmd': $text");
        return ();
    }

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
    	my @tmp = split(/\s+/, $line);
    	push(@existing,$tmp[0]);
    }
    return @existing;
}

## @name ::filer_vfilers_ips_parse
## @summary Returns a hash of all vfiler definitions from the input FILER_VFILER_IPS value
## @arg $fc Required: The filer connection object for the command line interface
## @arg $param Required: The FILER_VFILER_IPS param value
## @arg $test_volume_vfiler0 Optional: A volume in the default vfiler for new qtrees for new vfilers
sub filer_vfilers_ips_parse ($$;$) {
    my $fc  = shift;   # Later, we should verify the values exist here, or create them
    my $param  = shift;
    my $test_volume_vfiler0  = shift;
    if (!defined($test_volume_vfiler0) or $test_volume_vfiler0 eq '') {
        if (defined($main::FILER_VOL) and $main::FILER_VOL ne '') {
            my $vol = $main::FILER_VOL;
            $vol =~ s#^/vol/##;
            $test_volume_vfiler0 = "/vol/$vol";
        } else {
            $test_volume_vfiler0 = '/vol/vol0';
        }
    }

    my %Vfiler;
    no strict 'refs';
    if ($param ne '') {
        # Verify vfiler configuration uses these IP addresses
        # or establish them

        # Define a hash of vfilers containing their [interface,IP address,vol or qtree]
        my $unique = 0;
        foreach my $entry (split(/\s+/,$param)) {
            use strict 'refs';
            $unique++;
            my $vfiler = "vf" . $unique;
            my ($ifc,$ip,$path) = split(/:/,$entry);
            if (!defined($ip)) {
                ExitConfigError("Each FILER_VFILER_IPS entry must name a vfiler:ipaddress at a minimun, $ifc has no IP address");
            } else {
                # Default a qtree on the current FILER_VOL for this vfiler's path
                if (!defined($path)) {
                    $path = "$test_volume_vfiler0/$vfiler";
                }
                $Vfiler{$vfiler} = [$ifc, $ip, $path];
                print("$vfiler: $ifc, $ip, $path\n");
            }
        }
    } else {
        print("No vfilers configured\n");
    }
    return %Vfiler;
}

## @name ::filer_vol_create
## @summary Create a volume on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
## @arg $options Optional: 'vol create' command options, defaults to "2" to request a minimum size volume
sub filer_vol_create ($$;$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);
    my $options = shift;

    # /vol/ is dropped even if it was specified
    $vol =~ s/\/vol\///;

    my $cmd = "vol create $vol";
    if ($options) {
        $cmd .= " ".$options;
    } else {
        $cmd .= " 2";  # Assume a minimum disk count volume is to be added
    }
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /volume '\S+' already exists/) {
            logcomment("Volume $vol already exists");
            last;
        } elsif ($line =~ /vol create:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
    # Now wait for volume to be initialized and brought online
    my $done = filer_vol_wait_until_online($fc,$vol);
}

## @name ::filer_vol_destroy
## @summary Destroys an existing (offline) volume
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex to be destroyed
sub filer_vol_destroy ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol destroy $vol -f";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /No volume named '\S+' exists/) {
            logcomment("Volume $vol does not exist");
            last;
        } elsif ($line =~ /vol destroy:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_exists
## @summary Returns 1 if the specified volume exists
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex
sub filer_vol_exists ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol status $vol";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /No volume named/) {
            return 0;
        } elsif ($line =~ /vol status: Volume '\S+' is temporarily busy/) {
            return 1;
        } elsif ($line =~ /vol status:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        }
    }
    return 1;
}

## @name ::filer_vol_extent_disable
## @summary Turn extents off for a volume.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_extent_disable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol options $vol extent off";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vol options:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_extent_enable
## @summary Turn extents on for a volume.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_extent_enable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol options $vol extent on";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vol options:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_extent_size_set
## @summary Set the extent size for a volume.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
## @arg $extent_size Required: The extent size to set
sub filer_vol_extent_size_set ($$$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);
    my $extent_size = shift;

    my $cmd = "vol options $vol extent_size $extent_size";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vol options:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_file_exists
## @summary Returns 1 if the specified volume has the specfied file
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex
## @arg $file Required: The file name
sub filer_vol_file_exists ($$$) {
    my $fc  = shift;
    my $vol = arg_vol_remove_slashvol(shift);
    my $file = arg_file(shift);

    my $cmd = "ls /vol/$vol/$file";
    my($text,$result) = filer_cli_diag_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ "$cant open directory") {
            return 0;
        } elsif ($line =~ /ls:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /^$file$/) {
            # This could be a directory instead of a file?
            return 1;
        }
    }
    return 0;
}

## @name ::filer_vol_freespace_get
## @summary Returns the amount of free space on a volume in bytes
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name
sub filer_vol_freespace_get ($$) {
    my $fc  = shift;
    my $vol = shift;

    my %outval = filer_vol_spaceinfo_get($fc,$vol);
    return($outval{"FREE"}* 1024);
}

#Previous function before modification
#sub filer_vol_freespace_get ($$) {
#   my $fc  = shift;
#   my $vol = arg_vol_remove_slashvol(shift);
#
#   my $cmd = "df /vol/$vol";
#   $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
#   my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

#   my @info = filer_cli_split($text);
#   foreach my $line (@info) {
      # Filesystem          kbytes   used     avail   capacity Mounted on
      # |/vol/sledgehammer/    50119928       1924   50118004       0%  /vol/sledgehammer/
#      if ($line =~ /^Filesystem/) {
#         next;
#      } elsif ($line =~ m/^\/vol\/$vol\/\s/) {
#         $line =~ s/\s+/ /g;
#         my (undef,undef,undef,$avail,undef) = split(" ",$line);
#         if ($avail =~ m/[0-9]+/) {
#            # available space is in blocks of 1024 bytes
#            return 1024*$avail;
#         } else {
#            ExitSoftError("Found the invalid value '$avail' for the number of blocks available freespace on $vol");
            # Command ouput format may have changed?
#            last;
#         }
#         last;
#      } elsif ($line =~ /vol options:/) {
#         ExitError("Unexpected error from '$cmd': $line");
#      } else {
#         logcomment("Unrecognized '$cmd' output: $line");
#      }
#   }
#   ExitSoftError("Could not determine available freespace of $vol");
#   return -1;
#}

## @name ::filer_vol_is_offline
## @summary Returns 1 if a volume or plex is offline.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex
sub filer_vol_is_offline ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol status $vol";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ m/^\s*$vol\s+offline/) {
            return 1;
        }
    }
    return 0;
}

## @name ::filer_vol_is_online
## @summary Returns 1 if a volume or plex is online.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex
sub filer_vol_is_online ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol status $vol";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        #rtpqa-fas940-4> vol status sledgehammer
        #         Volume State      Status            Options
        #   sledgehammer restricted raid_dp           raidsize=8
        #                           initializing
        #
        #                Plex /sledgehammer/plex0: offline, empty, active
        if ($line =~ /vol status:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /^\s*$vol\s+online/ ) {
            return 1;
        }
    }
    return 0;
}

## @name ::filer_vol_offline
## @summary Brings a volume or plex offline
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex
## @arg $timedelay Optional: The number of seconds to wait after sending message to volume users.
## Default is 0
sub filer_vol_offline ($$;$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);
    my $timedelay  = shift;
    if (!defined($timedelay)) {
        $timedelay = 0;
    }

    my $cmd = "vol offline $vol -t $timedelay";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vol offline: Volume '\S+' is already offline/) {
            logcomment("Volume $vol is already offline");
            last;
        } elsif ($line =~ /vol offline:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_online
## @summary Brings a volume or plex online.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex
sub filer_vol_online ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    # Operation is forced to avoid prompt for volume inconsistencies
    my $cmd = "vol online $vol -f";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
       if ($line =~ /vol online: Volume '\S+' is already online/) {
           logcomment("Volume $vol is already online");
           last;
       } elsif ($line =~ /vol online:/) {
            ExitError("Unexpected error from '$cmd': $line");
       }
    }
}

## @name ::filer_vol_option_set
## @summary Brings a volume or plex online.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex
## @arg $option Required: The option name
## @arg $value Optional: the value for hte named option. Note option root has no value
sub filer_vol_option_set ($$$;$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);
    my $option = shift;
    my $value = shift;

    my $cmd = "vol options $vol $option";
    if (defined($value) and $value !~ m/^\s*$/) {
        $cmd = "$cmd $value";
    }
    my($text,$result) = filer_cli_say($fc,$cmd,{timeout=>120,retry=>'ok'});
}

## @name ::filer_vol_options_get
## @summary Returns a list of options for a volume.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex
sub filer_vol_options_get ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    # Use an array rather than a hash since the 'root' option has no value
    my @options = ();

    my $cmd = "vol options $vol";
    my($text,$result) = filer_cli_say($fc,$cmd,{timeout=>120,retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ m/!vol options:\s*(.*)/) {
            #vol options: No volume named 'xxx' exists.
            ExitSoftError("Unexpected error from vol options: $1");
            return undef;
        } else {
            # Assume this is a valid output line
            $line =~ s/\s//g; # drop all spaces
            my @new_options = split(',',$line);
            push(@options,@new_options);
        }
    }
    return @options;
}

## @name ::filer_vol_quota_disable
## @summary Turns on quotas for the specified volume on the filer.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_quota_disable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);
    # Remove /vol/ from volume name to match syntax on quota command
    $vol =~ s/\/vol\///;

    my $cmd = "quota off $vol";
    $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        #
        if ($line =~ /quotas already off/) {
            logcomment("Quotas were already off");
            last;
        } elsif ($line =~ /usage:|quota:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_quota_enable
## @summary Turns on quotas for the specified volume on the filer.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_quota_enable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);
    # Remove /vol/ from volume name to match syntax on quota command
    $vol =~ s/\/vol\///;

    my $cmd = "quota on $vol";
    $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my $wait_for_enable = 1;
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /quota: Starting quota initialization/) {
            # Normal success message
        } elsif ($line =~ /quota:\s+.*quotas are already on/) {
            # Message is already enabled
            $wait_for_enable = 0; # no wait needed
        } elsif ($line =~ /usage:|quota:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
    if ($wait_for_enable) {
        if (!filer_vol_quota_is_enabled($fc,$vol)) {
            ExitError("Failed to enable quotas for volume: $vol");
        }
    }
}

## @name ::filer_vol_quota_is_enabled
## @summary Returns 1 if quotas for the specified volume on the filer are enabled.
## @description Waits for quota initialization to end, returns 0 if quotas are not enabled.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_quota_is_enabled ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);
    # Remove /vol/ from volume name to match syntax on quota command
    $vol =~ s/\/vol\///;

    #my $cmd = "quota $vol";  # Does not work, see burt 102733 The syntax to get quota status for a single volume fails
    my $cmd;
    # "status" sub command is needed beginning with Gordonbiersh,
    # not there for Anchorsteam or Tsingtao
    my $ver = get_wafl_version($fc->hostp());
    if ($ver < WV_GORDONBIERSCH_MIN) {
            $cmd = "quota";
    } else{
            $cmd = "quota status";
    }
    $cmd = filer_vfiler_command($fc,$cmd);  # Convert command if simulating a vfiler context

    try: while (1) {
        my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

        my @info = filer_cli_split($text);
        foreach my $line (@info) {
            if ($line =~ /^$vol: quotas are off/) {
                return 0; # This volume is listed as 'off' in the quota command output
                # Normal success message
            } elsif ($line =~ /^$vol: quotas are on/) {
                return 1; # This volume is listed as 'on' in the quota command output
            } elsif ($line =~ /^$vol: quotas are initializing/) {
                sleep(15);
                next try;
            } elsif ($line =~ /\[quota: warning\]:/) {
                # Normal warning message
            } elsif ($line =~ /usage:|quota:/) {
                ExitSoftError("Unexpected error from '$cmd': $line");
            }
        }
        return 0; # This volume is not included in the quota command output
    }
}

## @name ::filer_vol_restrict
## @summary Marks a volume or plex as restricted.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex
## @arg $timedelay Optional: The number of seconds to wait after sending message to volume users.
## Default is 0
sub filer_vol_restrict ($$;$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);
    my $timedelay  = shift;
    if (!defined($timedelay)) {
        $timedelay = 0;
    }

    my $cmd = "vol restrict $vol -t $timedelay";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vol restrict: Volume '\S+' is already restricted/) {
            logcomment("Volume $vol is already restricted");
            last;
        } elsif ($line =~ /vol restrict:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_size_get
## @summary Returns size of volume in bytes
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex
sub filer_vol_size_get ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $size = 0;

    my $cmd = "vol status -b $vol";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /No volume named/) {
            ExitSoftError("No volume named '$vol'");
            return 0;
        } elsif ($line =~ /---/) {
            next;
        } elsif ($line =~ /^\w+\s+\d+\s+\d+\s+\d+/) {
            my $blocksize = $line;
            $blocksize =~ s/^\w+\s+(\d+)\s+\d+\s+\d+.*/$1/;
            my $blockcount = $line;
            $blockcount =~ s/^\w+\s+\d+\s+(\d+)\s+\d+.*/$1/;
            $size = $blocksize * $blockcount;
            last;
        }
    }
    return $size;
}

## @name ::filer_vol_snap_create
## @summary Create a volume snapshot on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
## @arg $snapshot Required: name of the snapshot
## @arg $timeout Optional: seconds to wait for snapshot to be created,
## default = 1200, 0 means do not wait
sub filer_vol_snap_create ($$$;$) {
    my $fc  = shift;
    my $vol = arg_vol_remove_slashvol(shift);
    my $snapshot = arg_snapshot(shift);
    my $timeout = shift;

    if (!defined($timeout)) {
        $timeout = 1200;
    }

    my $cmd = "snap create $vol $snapshot";
    my($text,$result) = filer_cli_say($fc,$cmd,{timeout=>120});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /Snapshot already exists/) {
            # The test can continue
            logcomment("Snapshot already exists");
        } elsif ($line =~ /does not exist or is not online/ ) {
            ExitError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /usage:|snap:|snap create:/ ) {
            # This form of error message has not been observed with this command
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
    if ($timeout >0) {
        my $endtime = time() + $timeout;
        while (!filer_vol_snap_exists($fc,$vol,$snapshot)) {
            if ($endtime < time()) {
                ExitError("Failed to create snapshot: $snapshot on volume: $vol within $timeout seconds");
            }
            sleep(15);
        }
    }
}

## @name ::filer_vol_snap_delete
## @summary Delete a volume snapshot on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name
## @arg $snapshot Required: name of the snapshot
## @arg $timeout Optional: seconds to wait for snapshot to be deleted, 0= no wait
## default = 240, 0 means do not wait
sub filer_vol_snap_delete ($$$;$) {
    my $fc  = shift;
    my $vol = arg_vol_remove_slashvol(shift);
    my $snapshot = arg_snapshot(shift);
    my $timeout = shift;

    if (!defined($timeout)) {
        $timeout = 240;
    }

    my $cmd = "snap delete $vol $snapshot";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /No such snapshot/) {
            logcomment("Snapshot appears to already be deleted");
            last;
        } elsif ($line =~ /does not exist or is not online|Snapshot in use by / ) {
            ExitError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /usage:|snap:|snap delete:/ ) {
            # This form of error message has not been observed with this command
            ExitError("Unexpected error from '$cmd': $line");
        }
    }

    if ($timeout >0) {
        my $endtime = time() + $timeout;
        while (filer_vol_snap_exists($fc,$vol,$snapshot)) {
            if ($endtime < time()) {
                ExitError("Snap delete did not complete within $timeout seconds for snapshot: $snapshot on volume: $vol");
            }
            sleep(15);
        }
    }
}

## @name ::filer_vol_snap_delete_all
## @summary Deletes all snapshots on the volume on this filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name
## @arg $timeout Optional: seconds to wait for snapshot to be deleted, 0= no wait
## default = 240, 0 means do not wait
sub filer_vol_snap_delete_all ($$;$) {
    my $fc  = shift;
    my $vol = arg_vol_remove_slashvol(shift);
    my $timeout = shift;

    if (!defined($timeout)) {
        $timeout = 240;
    }

    my $cmd = "snap delete -a -f $vol";
    my($text,$result) = filer_cli_say($fc,$cmd,{timeout=>$timeout,retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /usage:|snap:|snap delete:/ ) {
            # This form of error message has not been observed with this command
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_snap_exists
## @summary Determines if the specified volume has the specified snapshot file
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
## @arg $snapshotRequired: name of the snapshot
sub filer_vol_snap_exists ($$$) {
    my $fc  = shift;
    my $vol = arg_vol_remove_slashvol(shift);
    my $snapshot = arg_snapshot(shift);

    # Distinguish volume from qtree
    my $option = '';
    if ($vol =~ /\//) {
        $option = '-o';
        # Put /vol/ prefx back on for -o option
        $vol = "/vol/$vol";
    }

    my $cmd = "snap list $option $vol";
    # 60 seconds was not enough time
    my($text,$result) = filer_cli_say($fc,$cmd,{timeout=>120,retry=>'ok'});

    #     1% ( 1%)    0% ( 0%)  Jun 03 10:59  shelley(0016791277)_smdst.1 (snapmirror)
    # As you can see above, Perl regular expression characters may appear in the snapshot name
    my $snapshot_search = quotemeta($snapshot);
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /Volume .* does not exist or is not online/) {
            return 0;
        } elsif ($line =~ /usage:|snap:|snap list:/ ) {
            ExitError("Unexpected error from '$cmd': $line");
        } else {
            if ($option eq '-o') {
                if ($line =~ /\s+$snapshot_search\s*$/ ) {
                    return 1;
                }
            } else {
                if ($line =~ /^\s+\d+\%\s+\(.*\)\s+\d+\%\s+\(.*\)\s+\w+\s+\d+\s\d\d:\d\d\s+$snapshot_search\s+/ ) {
                    # This line may end after the snapshot name and several spaces or include the text '(busy,snapmirror)'
                    return 1;
                }
            }
        }
    }
    return 0;
}

## @name ::filer_vol_snap_freespace_get
## @summary Returns the amount of free space available for snapshots on a volume in bytes
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name
sub filer_vol_snap_freespace_get($$){
    my $fc  = shift;
    my $vol = filer_get_vol_name_from_path(shift);

    my %outval = filer_vol_spaceinfo_get($fc,$vol);
    return($outval{"SNAPFREE"}* 1024);
}

## @name ::filer_vol_snap_is_busy
## @summary Determines if the specified snapshot is busy
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name with or without its path
## @arg $snapshotRequired: name of the snapshot
sub filer_vol_snap_is_busy ($$$) {
    my $fc  = shift;
    my $vol = arg_vol_remove_slashvol(shift);
    my $snapshot = arg_snapshot(shift);

    my $cmd = "snap list $vol";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    # As you can see above, Perl regular expression characters may appear in the snapshot name
    my $snapshot_search = quotemeta($snapshot);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /Volume .* does not exist or is not online/) {
            ExitError("Volume $vol does not exist");
        } elsif ($line =~ /usage:|snap:|snap list:/ ) {
            ExitError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /^\s+\d+\%\s+\(.*\)\s+\d+\%\s+\(.*\)\s+\w+\s+\d+\s\d\d:\d\d\s+$snapshot_search\s*/ ) {
            # This line may end after the snapshot name and several spaces or include the text '(busy)'
            if ($line =~ /\(busy\)/ ) {
                return 1;
            } elsif ($line =~ /\(busy,LUNs\)/) {
                burt_fixed(85195);
                return 1;
            } else {
                return 0;
            }
        }
    }
    return 0;
}

## @name ::filer_vol_snap_reclaimspace_get
## @summary Returns the amount of space reclaimed if a snapshot is deleted
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name with or without its path
## @arg $vol Required: The name of the sanpshot to be deleted
sub filer_vol_snap_reclaimspace_get($$$){
    my $fc  = shift;
    my $vol  = filer_get_vol_name_from_path(shift);
    my $snapname = shift;

    my $cmd = "snap reclaimable $vol $snapname";
    #my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    my ($result,undef,undef) = say($fc,$cmd,'-wait', "Approximately [0-9]* Kbytes would be freed");
    #my ($start,$end,$result) = say($fc,$cmd,'-prompt');

    $result =~ /(\d+)/;
    return int($1* 1024);
}

## @name ::filer_vol_snap_recreate
## @summary Creates or recreates a volume snapshot on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
## @arg $snapshotRequired: name of the snapshot
sub filer_vol_snap_recreate ($$$) {
    my $fc  = shift;
    my $vol = arg_vol_remove_slashvol(shift);
    my $snapshot = arg_snapshot(shift);

    if (filer_vol_snap_exists($fc,$vol,$snapshot) ) {
        if (filer_vol_snap_is_busy($fc,$vol,$snapshot)) {
            # Rename old snapshot so new one will have the lun we just created.
            # One reason to be busy is that it is used by a lun such as $snaplun
            # that should have been deleted above.
            # So rename it instead
            my $id = time_id();
            filer_vol_snap_rename($fc,$vol,$snapshot,"$id"."_$snapshot");
        } else {
            filer_vol_snap_delete($fc,$vol,$snapshot);
        }
    }
    filer_vol_snap_create($fc,$vol,$snapshot);

    my $waitloop = 20;
    while (!filer_vol_snap_exists($fc,$vol,$snapshot)) {
        $waitloop--;
        if ($waitloop <0) {
            ExitError("Failed to create snapshot $snapshot on volume $vol");
        }
        sleep(10);
    }
}

## @name ::filer_vol_snap_rename
## @summary Delete a volume snapshot on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
## @arg $snapshot Required: Name of the snapshot
## @arg $to_snapshot Required: New name of the snapshot
sub filer_vol_snap_rename ($$$$) {
    my $fc  = shift;
    my $vol = arg_vol_remove_slashvol(shift);
    my $snapshot = arg_snapshot(shift);
    my $tosnapshot = arg_snapshot(shift);

    my $cmd = "snap rename $vol $snapshot $tosnapshot";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /usage:|snap:|snap rename:/) {
            # This form of error message has not been observed with this command
            ExitError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /does not exist or is not online|No such snapshot/ ) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_snap_reserve_set
## @summary Change a volume's snapshot reserve on the filer
## @arg $fc          Required: The filer connection object for the command line interface
## @arg $vol         Required: The volume name
## @arg $reserve_pct Required: The reserve percent
sub filer_vol_snap_reserve_set ($$$) {
    my $fc          = shift;
    my $vol         = arg_vol_remove_slashvol(shift);
    my $reserve_pct = shift;

    # /vol/ is dropped if it was specified
    $vol =~ s/\/vol\///;

    my $cmd = "snap reserve $vol $reserve_pct";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /usage:|snap:|snap reserve:/) {
            # This form of error message has not been observed with this command
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_snap_restore
## @summary Restore a volume from a snapshot on the filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
## @arg $snapshot Required: Name of the snapshot
sub filer_vol_snap_restore ($$$) {
    my $fc  = shift;
    my $vol = arg_vol_remove_slashvol(shift);
    my $snapshot = arg_snapshot(shift);

    my $cmd = "snap restore -f -t vol -s $snapshot $vol";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /usage:|snap:|snap restore:/) {
            # This form of error message has not been observed with this command
            ExitError("Unexpected error from '$cmd': $line");
        } elsif ($line =~ /SNAPRESTORE is not licensed/ ) {
            ExitConfigError("SNAPRESTORE must be licensed prior to running '$cmd': $line");
        } elsif ($line =~ /does not exist or is not online|No such snapshot/ ) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_snap_sched_get
## @summary Returns a volume's snapshot schedule on the filer
## @arg $fc    Required: The filer connection object for the command line interface
## @arg $vol   Required: The volume name
sub filer_vol_snap_sched_get ($$) {
    my $fc    = shift;
    my $vol   = arg_vol_remove_slashvol(shift);

    my $cmd = "snap sched $vol";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /usage:|snap:|snap sched:/) {
            # This form of error message has not been observed with this command
            ExitSoftError("Unexpected error from '$cmd': $line");
        }
        if ($line =~ /$vol/) {
            $line =~ s/^Volume $vol: //;
            return $line;
        }
    }
    ExitSoftError("Could not get snap schedule for $vol");
}

## @name ::filer_vol_snap_sched_set
## @summary Change a volume's snapshot schedule on the filer
## @arg $fc    Required: The filer connection object for the command line interface
## @arg $vol   Required: The volume name
## @arg $sched Required: The snapshot schedule
sub filer_vol_snap_sched_set ($$$) {
    my $fc    = shift;
    my $vol   = arg_vol_remove_slashvol(shift);
    my $sched = shift;

    # /vol/ is dropped even if it was specified
    $vol =~ s/\/vol\///;

    my $cmd = "snap sched $vol $sched";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /usage:|snap:|snap sched:/) {
            # This form of error message has not been observed with this command
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_snap_totalspace_get
## @summary Returns the total amount space on a volume in bytes
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name
sub filer_vol_snap_totalspace_get($$){
    my $fc  = shift;
    my $vol = shift;

    my %outval = filer_vol_spaceinfo_get($fc,$vol);
    return($outval{"SNAPTOTAL"}* 1024);
}

## @name ::filer_vol_snap_usedspace_get
## @summary Returns the amount of snapshot area used space on a volume in bytes
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or path
sub filer_vol_snap_usedspace_get($$){
    my $fc  = shift;
    my $vol = filer_get_vol_name_from_path(shift);

    my %outval = filer_vol_spaceinfo_get($fc,$vol);
    return($outval{"SNAPUSED"}* 1024);
}

## @name ::filer_vol_snap_waflscan_complete
## @summary Returns status of WAFL scans relevant to snapshot creation
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name with or without its path
sub filer_vol_snap_waflscan_complete($$){
    my $fc  = shift;
    my $vol  = filer_get_vol_name_from_path(shift);

    my $cmd = "priv set diag;wafl scan status $vol";
    my (undef,undef,$result) = say($fc,$cmd);
    #$cmd = filer_vfiler_command($fc,$cmd);
    if ($result =~ /blocks used summary update/ || $result =~ /snap create summary update/ ){
        return 0; # Scan still running
    }
    return 1; # Scan complete
}

## @name ::filer_vol_snapdir_disable
## @summary Turn off visiblity to .snapshot directory on a volume.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_snapdir_disable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol options $vol nosnapdir on";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vol options:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_snapdir_enable
## @summary Turn on visiblity to .snapshot directory on a volume.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_snapdir_enable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol options $vol nosnapdir off";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vol options:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_space_reserve_disable
## @summary Turn off space reservations for a volume
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_space_reserve_disable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol options $vol create_reserved off";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vol options:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_space_reserve_enable
## @summary Turn on space reservations for a volume
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_space_reserve_enable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol options $vol create_reserved on";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vol options:/) {
            ExitError("Unexpected error from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_space_reserve_fill_disable
## @summary Turn off space reservations of file holes for newly created files on a volume.
## (new for Full Sail)
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_space_reserve_fill_disable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    logresult('WARN',"The filer_vol_space_reserve_fill_disable function is deprecated");

    #my $cmd = "vol options $vol reserve_fill off";
    # Replace the previous call with syntax that provides similar functionality
    filer_vol_space_reserve_disable($fc,$vol);
}

## @name ::filer_vol_space_reserve_fill_enable
## @summary Turn on space reservations of file holes for newly created files on a volume
## (new for Full Sail)
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_space_reserve_fill_enable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    logresult('WARN',"The filer_vol_space_reserve_fill_enable function is deprecated");

    #my $cmd = "vol options $vol reserve_fill on";
    # Replace the previous call with syntax that provides similar functionality
    filer_vol_space_reservate_enable($fc,$vol);
}

## @name ::filer_vol_space_reserve_fill_is_enabled
## @summary Returns 1 if space reserve fill is enabled on the specified volume
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_space_reserve_fill_is_enabled ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    logresult('WARN',"The filer_vol_space_reserve_fill_is_enabled function is deprecated");

    #my $cmd = "vol options $vol";
    # Replace the previous call with syntax that provides similar functionality
    return filer_vol_space_reserve_is_enabled($fc,$vol);
}

## @name ::filer_vol_space_reserve_is_enabled
## @summary Returns 1 if create_reserve is enabled on the specified volume
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_space_reserve_is_enabled ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol status $vol";
    my($text,$result) = filer_cli_say($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /create_reserved=on/) {
            return 1;
        } elsif ($line =~ /create_reserved=off/) {
            return 0;
        } elsif ($line =~ /vol status:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
        }
    }
    return undef;
}

## @name ::filer_vol_space_reserve_overwrite_disable
## @summary Turn on space reservations of all written areas of a file for newly created files on a volume
## (new for Full Sail)
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_space_reserve_overwrite_disable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    logresult('WARN',"The filer_vol_space_reserve_overwrite_disable function is deprecated");

    #my $cmd = "vol options $vol reserve_overwrite off";
    # Replace the previous call with syntax that provides similar functionality
    filer_vol_space_reserve_disable($fc,$vol);
}

## @name ::filer_vol_space_reserve_overwrite_enable
## @summary Turn on space reservations of all written areas of a file for newly created files on a volume
## (new for Full Sail)
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_space_reserve_overwrite_enable ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    logresult('WARN',"The filer_vol_space_reserve_overwrite_enable function is deprecated");

    #my $cmd = "vol options $vol reserve_overwrite on";
    # Replace the previous call with syntax that provides similar functionality
    filer_vol_space_reserve_enable($fc,$vol);
}

## @name ::filer_vol_space_reserve_overwrite_is_enabled
## @summary Returns 1 if space reserve overwrite is enabled on the specified volume
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
sub filer_vol_space_reserve_overwrite_is_enabled ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    logresult('WARN',"The filer_vol_space_reserve_overwrite_is_enabled function is deprecated");

    #my $cmd = "vol options $vol";
    # Replace the previous call with syntax that provides similar functionality
    return filer_vol_space_reserve_is_enabled($fc,$vol);
}

## @name ::filer_vol_spaceinfo_get
## @summary Returns a hash of the current space statistics on a volume
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or path
sub filer_vol_spaceinfo_get($$){
    my $fc  = shift;
    my $vol  = filer_get_vol_name_from_path(shift);

    ##my $vol = arg_vol_remove_slashvol(shift);;
    ##my $cmd = "df /vol/$vol";
    my $cmd = "df $vol";
    while (filer_vol_snap_waflscan_complete($fc,$vol) != 1){
        sleep(60);
    }
    ##$cmd = filer_vfiler_command($fc,$cmd);
    #my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});
    my (undef,undef,$result) = say($fc,$cmd);
    ##my @tmprslt = filer_cli_split($text);
    my @tmprslt = split(/\n/,$result);
    my @volout = split(/ +/,$tmprslt[1]);
    my @volsnapout = split(/ +/,$tmprslt[2]);
    my %dfoutput = (TOTAL => int($volout[1]), # total space on volume
            USED => int($volout[2]),  # Used space on the volume
            FREE => int($volout[3]),  # Free space on the volume
            SNAPTOTAL => int($volsnapout[1]), # total space in snapshots area on volume
            SNAPUSED => int($volsnapout[2]),  # used space in snapshots area on volume
            SNAPFREE => int($volsnapout[3])); # free space in snapshots area on volume
    return (%dfoutput);
}

## @name ::filer_vol_totalspace_get
## @summary Returns the total space on a volume in bytes
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name
sub filer_vol_totalspace_get($$){
    my $fc  = shift;
    my $vol = filer_get_vol_name_from_path(shift);

    my %outval = filer_vol_spaceinfo_get($fc,$vol);
    return(int(int($outval{"TOTAL"}) * 1024));
}

## @name ::filer_vol_usedspace_get
## @summary Returns the amount of used space on a volume in bytes
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name
sub filer_vol_usedspace_get($$){
    my $fc  = shift;
    my $vol = filer_get_vol_name_from_path(shift);

    my %outval = filer_vol_spaceinfo_get($fc,$vol);
    return($outval{"USED"}* 1024);
}

## @name ::filer_vol_vaporize
## @summary Forces destruction of a volume if it exists
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name or plex to be destroyed
sub filer_vol_vaporize ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    if (filer_vol_exists($fc,$vol)) {
        # Make sure volume is offline
        if (!filer_vol_is_offline($fc,$vol)) {
            # Files on this volume open for use by ONTAP can abort this request
            # Can we detect this and wait for it to clear?

            filer_vol_offline($fc,$vol);
        }
        filer_vol_destroy($fc,$vol);
    }
}

## @name ::filer_vol_wafliron_stop
## @summary Stops a wafliron operation on a volume.
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name
sub filer_vol_wafliron_stop ($$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);

    my $cmd = "vol wafliron stop $vol";
    my($text,$result) = filer_cli_advanced($fc,$cmd);

    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        if ($line =~ /vol wafliron:/) {
            ExitSoftError("Unexpected message from '$cmd': $line");
        }
    }
}

## @name ::filer_vol_wait_until_online
## @summary Wait until a volume on the filer is online
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
## @arg $timeout Optional: The number of seconds to wait, defaults to a VERY LONG time
sub filer_vol_wait_until_online ($$;$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);
    my $timeout = shift;

    # /vol/ is dropped even if it was specified
    $vol =~ s/\/vol\///;

    if (!defined($timeout) or $timeout =~ m/^\s*$/) {
        $timeout = 60*60*3; # 3 hours
    }
    my $endtime = time() + $timeout;

    # Now wait for volume to be initialized and brought online
    my $done =0;
    while (!$done) {
        $done = filer_vol_is_online ($fc,$vol);
        if (time() > $endtime) {
            last;
        } else {
            sleep 30;
        }
    }
    return $done;
}

## @name ::filer_vol_wait_until_ready
## @summary Wait until a volume on the filer is online and not growing
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol Required: The volume name including its path
## @arg $timeout Optional: The number of seconds to wait, defaults to a VERY LONG time
sub filer_vol_wait_until_ready ($$;$) {
    my $fc  = shift;
    my $vol = arg_vol(shift);
    my $timeout = shift;

    # /vol/ is dropped even if it was specified
    $vol =~ s/\/vol\///;

    if (!defined($timeout) or $timeout =~ m/^\s*$/) {
       $timeout = 60*60*3; # 3 hours
    }
    my $endtime = time() + $timeout;

    # Now wait for volume to be initialized and brought online
    my $online =0;
    my $done =0;
    while (!$done) {
        my $cmd = "vol status $vol";
        my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

        my @info = filer_cli_split($text);
        foreach my $line (@info) {
            #rtpqa-fas940-4> vol status sledgehammer
            #         Volume State      Status            Options
            #   sledgehammer restricted raid_dp           raidsize=8
            #                           initializing
            #
            #                Plex /sledgehammer/plex0: offline, empty, active
            if ($line =~ /vol status:/) {
                ExitError("Unexpected error from '$cmd': $line");
            } elsif ($line =~ /^\s*$vol\s+online\s+/ ) {
                $online = 1;
                $done = 1;
            } elsif ($line =~ /^\s+growing/ ) {
                $done =0;
                last;
            }
        }
        if (!$done) {
            if (time() > $endtime) {
                last;
            } else {
                sleep 30;
            }
        }
    }
    return $done;
}

## @name ::filer_vols_get
## @summary Returns a list of volumes on the specified filer
## @arg $fc Required: The filer connection object for the command line interface
## @arg $vol_types Optional: volume attributes to look for
sub filer_vols_get ($;$) {
    my $fc  = shift;
    my $type_ref = shift;

    my @vol_types;
    if (defined($type_ref)) {
        $type_ref =~ s/,/ /g;
        $type_ref =~ s/^\s+//;
        $type_ref =~ s/\s+$//;
        $type_ref =~ s/\s+/ /g;
        @vol_types = split(" ",$type_ref);
    }

    my @vols;
    my %found_vol_type;
    my $cmd = "vol status -v";
    my($text,$result) = filer_cli_say($fc,$cmd,{retry=>'ok'});

    my $vol = '';
    my @info = filer_cli_split($text);
    foreach my $line (@info) {
        #  rtpqa-fas940-4*> vol status -v
        #          Volume State      Status            Options
        #            vol0 online     raid4, trad       root, diskroot, nosnap=off,
        #                                              nosnapdir=off, minra=off,
        #                                              no_atime_update=off,
        #                                              raidtype=raid4, raidsize=8,
        #                                              nvfail=off,
        #                                              snapmirrored=off,
        #                                              resyncsnaptime=60,
        #                                              create_ucode=on,
        #                                              convert_ucode=off,
        #                                              maxdirsize=31457,
        #                                              fs_size_fixed=off,
        #                                              svo_enable=off,
        #                                              svo_checksum=off,
        #                                              svo_allow_rman=off,
        #                                              svo_reject_errors=off,
        #                                              fractional_reserve=100
        #                 Containing aggregate: <N/A>
        #
        #                 Plex /vol0/plex0: online, normal, active
        #                     RAID group /vol0/plex0/rg0: normal
        #
        if ($line =~ /No volume named/) {
            last;
        } elsif ($line =~ m/^\s+Volume\s+State/) {
            next;
        } elsif ($line =~ m/^\s+Containing\s+aggregate/) {
            next;
        } elsif ($line =~ m/^\s+Plex\s+/) {
            next;
        } elsif ($line =~ m/^\s+RAID\s+group\s+/) {
            next;
        } elsif ($line =~ m/^\s*$/) {
            next;
        } elsif ($line =~ m/vol status:/) {
            ExitSoftError("Unexpected error from '$cmd': $line");
            next;
        } elsif ($line =~ m/^\s*(\S+)\s+online\s+(\S+),?\s+(\S+)/) {
            my $new_vol = $1;
            if ($vol ne '') {
                # Make sure all required 'types' are present
                my $all_types_found = 1;
                foreach my $required_type (@vol_types) {
                    if (!defined($found_vol_type{$required_type})) {
                        $all_types_found = 0;
                        last;
                    }
                }
                if ($all_types_found) {
                    push(@vols,$vol);
                }
            }
            $vol = $new_vol;
            # Drop volume name from the current line that also has 'type' values
            $line =~ s/^\s*\S+\s+//;
            %found_vol_type = (); # reset type data from previous volume
            # Now record status and options that appear on this line
        }
        # Assume this line shows additional volume 'types'
        $line =~ s/,/ /g;
        $line =~ s/^\s+//;
        $line =~ s/\s+$//;
        $line =~ s/\s+/ /g;
        my @new_types = split(' ',$line);
        foreach my $new_type (@new_types) {
            $found_vol_type{$new_type} = 1;
        }
    }
    if ($vol ne '') {
        # Make sure all required 'types' are present
        my $all_types_found = 1;
        foreach my $required_type (@vol_types) {
            if ($required_type =~ m/^\!/) {
                # This type was requested to not exist
                # Throw this volume out if it does exist
                $required_type =~ s/^\!//;
                if (defined($found_vol_type{$required_type})) {
                    $all_types_found = 0;
                    last;
                }
            } else {
                if (!defined($found_vol_type{$required_type})) {
                    $all_types_found = 0;
                    last;
                }
            }
        }
        if ($all_types_found) {
            push(@vols,$vol);
        }
    }
    return @vols;
}

## @name ::filer_volume_dump
## @summary Dump the contents of the volume.
## @arg $fc Required: The filer connection object.
## @arg $volume Required: The volume to dump.
## @arg $device Required: The device to dump contents to.
## @arg $level  Optional: The dump level to perform (default: 0).
sub filer_volume_dump ($$$;$) {
    my ($fc, $volume, $device, $level) = validate_pos(@_,
        { type => OBJECT },
        { type => SCALAR },
        { type => SCALAR },
        { type => SCALAR, default => 0 }
    );
    my ($cmd);

    # Options for dump
    my $opts = 'f';

    $cmd = "dump $level$opts $device $volume";
    # Dump the contents of volume to device
    my $text = filer_cli_say($fc, $cmd, {retry=>'ok'});

    # Make sure the dump is done
    if ( $text !~ /DUMP IS DONE/ ) {
        ExitError("Volume dump failed");
    }
    return 1;
}

## @name ::filer_volume_restore
## @summary Restore the contents of a dump.
## @arg $fc Required: The filer connection object.
## @arg $device Required: The device to get dump contents from.
## @arg $dir  Optional: The directory to restore in. The default is to restore from where the dump was done.
sub filer_volume_restore ($$$;) {
    my ($fc, $device, $dir) = validate_pos(@_,
        { type => OBJECT },
        { type => SCALAR },
        { type => SCALAR, default => '' },
    );
    my ($cmd);

    # Options for dump
    my $opts = 'rf';

    # If the directory to restore in is specified then add the option
    if ($dir ne '') {
        $opts .= 'D';
    }
    $cmd = "restore $opts $device $dir";
    # Restore the contents of volume from device
    my $text = filer_cli_say($fc, $cmd, {retry=>'ok'});

    # Make sure the restore is done
    if ( $text !~ /RESTORE IS DONE/ ) {
        ExitError("Volume restore failed");
    }
    return 1;
}

1;

#print "\n ".__FILE__." LOADED \n";

END {
    my $save_return_code = $?;
    #print("Starting END routine for ".__FILE__." with exit code $save_return_code\n");

    # This is a place to put generic Filer cleanup operations
    # Be sure to control these actions via parameters so that
    # users can opt out of the cleanup

    # This is also the place to process statistics that were gathered
    # during the use of this module.
    if (defined($FILER_COMMAND_COUNT) and $FILER_COMMAND_COUNT ne '0') {
        # Avoid problems if there is nothing to report
        my @params;
        push(@params,"FILER_CLI_COUNT=$FILER_COMMAND_COUNT");
        my $result_obj = NATE::Result::Info -> new([@params]);
        $result_obj -> store();
    }

    # Restore the initial return code
    $? = $save_return_code;
}


sub filer_cmd {
    my (%opts) = validate(@_,{
            host => {type=>OBJECT, isa=>'Hostrec'},
            conntype => {type=>SCALAR,default=>"#rsh",optional=>1},
            timeout => {type=>SCALAR, default=>300,optional=>1},
            filermode => {type=>SCALAR,default=>"normal",optional=>1 },
            privmode => {type=>SCALAR,default=>"diag",optional=>1 },
            noprompt => {type=>SCALAR,default=>0,optional=>1 },
            cmd => {type=>SCALAR},
            sp_check => {type=>SCALAR,default=>0,optional=>1 },
            prompt_check => {type=>SCALAR,default=>0,optional=>1 }
            }
            );
    my $output;
    my $conn;
    if ($opts{filermode} eq "maint") { 
        $opts{conntype} = "#console";
        $conn = make_conn($opts{host},$opts{conntype});
        my $cmd = "$opts{cmd}";
        if($opts{noprompt}){
            (undef, undef, $output) = say($conn, "$cmd", -noprompt , "-timeout", $opts{timeout});
        }
        else{
            my $MAINT_PROMPT = "\\*\>";
            (undef, undef, $output) = say($conn, "$cmd", -prompt, "$MAINT_PROMPT", "-timeout", $opts{timeout});
        }

    }
    elsif ($opts{filermode} eq "normal"){ 
        if($opts{host}->{hosttype} =~ /filer-ng/){
            my $server= new Server;
            $server->set_execution_server(-Server => $opts{host}->{hostname});
            my $cmd = "run local  \"priv set -q $opts{privmode};$opts{cmd}\"";
            if($opts{noprompt}){
                $output = $server->cli(cmd=>"$cmd",-Timeout=>$opts{timeout},noprompt=>1);
            }
            else{
                $output = $server->cli(cmd=>"$cmd",-Timeout=>$opts{timeout});
            }

        }
        else{
            $conn = make_conn($opts{host},$opts{conntype});
            my $cmd = "priv set -q $opts{privmode};$opts{cmd}";
            if($opts{noprompt}){
                (undef, undef, $output) = say($conn, "$cmd",-noprompt,"-timeout", $opts{timeout});
            }elsif($opts{sp_check}){
                my $cmd = "$opts{cmd}";
                if($opts{prompt_check}){
                    (undef, undef, undef, $output) = say($conn, "$cmd", -prompt, "$opts{sp_check}", "-timeout", $opts{timeout});
                } else{
                    (undef, undef, $output) = say($conn, "$cmd", -prompt, "$opts{sp_check}", "-timeout", $opts{timeout});
                }
            }
            else{
                (undef, undef, $output) = say($conn, "$cmd","-timeout", $opts{timeout});
            }

        }
    }
    elsif($opts{filermode} eq "ngsh"){
        logcomment("NOT YET IMPLEMENTED");
    }
    else{
        die "Unsupported filermode";
    }
    return $output;
}
#end                  
