# $Id$

#
# Copyright (c) 2012 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Volume Component Module
## @author dl-nacl-dev@netapp.com, pkj@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Volume

=head1 DESCRIPTION

C<NACL::C::Volume> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Volume.  A related class is
L<NACL::CS::Volume|lib-NACL-CS-Volume-pm>, which represents
the past state of an ONTAP volume.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 volume

The name of the volume, used to identify it uniquely.  (Volume
names are assumed to be unique within a node or within a cluster in 7mode
and in a vserver in Cmode)

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the volume is present.

=cut

package NACL::C::Vol;

use strict;
use warnings;
use Params::Validate qw(:all);
use NACL::CS::Volume;
use NACL::CS::VolumeExplore;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument ();
use base 'NACL::C::Component::ONTAP';
use NACL::C::Job;
use NACL::STask::VolumeMove;
use NACL::STask::VolumeClone;
use NACL::STask::Mkfile;
use NACL::C::Vserver;
use Tharn qw($Log);
use NACL::Exceptions::InvalidFilterField qw(:try);
use NACL::APISet::Exceptions::CommandFailedException ();
use NACL::C::UnitNormalization;
use Data::Dumper ();

# Imports from mix-ins
use NACL::C::_Mixins::AggregateVolume qw(_translate_options_values);
use NACL::C::_Mixins::Volume qw(:all);
use NACL::C::_Mixins::VolumeFileDirectory qw(:find_state);
use NACL::C::_Mixins::Common qw(_determine_is_repository);

use NACL::STask::VolumeFile;
use NACL::C::VolumeDirectory;
use NACL::CS::VolumeListFilesInDirectory;

# Component exceptions
use NACL::C::Exceptions::Volume::AlreadyExists                ();
use NACL::C::Exceptions::Volume::DoesNotExist                 ();
use NACL::C::Exceptions::Volume::UnmountRoot                  ();
use NACL::C::Exceptions::Volume::AlreadyModified              ();
use NACL::C::Exceptions::Volume::UnsupportedUnmount           ();
use NACL::C::Exceptions::Volume::StateUnknown                 ();
use NACL::C::Exceptions::Volume::FileOperationsInProgress     ();
use NACL::C::Exceptions::Volume::CurrentlyBusy                ();
use NACL::C::Exceptions::Volume::DeleteSnapMirrorRelationship ();
use NACL::C::Exceptions::Volume::AlreadyMounted               ();
use NACL::C::Exceptions::Volume::VolumeOffline                ();
use NACL::C::Exceptions::Volume::VserverLocked                ();
use NACL::C::Exceptions::Volume::NotExported                  ();
use NACL::C::Exceptions::Volume::HasClones                    ();
use NACL::C::Exceptions::Volume::CannotDeleteRootVolume       ();
use NACL::C::Exceptions::Volume::AlreadyRestricted            ();
use NACL::C::Exceptions::Volume::MustBeUnmounted              ();

=head1 Linked components

(See
http://wikid.netapp.com/w/QA/projects/Libraries_Initiative/Project/Common_Infrastructure/Users_Guide/Component_Layer_Users_Guide#Linked_components
for a more detailed explanation on linked components)

This component is linked to the following components:

=begin html

<table border="1">
<tr> <th>Prefix</th> <th>Linked Component</th> </tr>
<tr> <td>efficiency</td> <td>VolumeEfficiency</td> </tr>
<tr> <td>snapshot</td> <td>VolumeSnapshot</td> </tr>
</table>

=end html

Examples:

    # Invoking the "start" method of the VolumeEfficiency component
    $vol_obj->efficiency_start(%opts);

    # Invoking the "create" method of the VolumeSnapshot component
    my $snapshot_obj = $vol_obj->snapshot_create(snapshot => $snap_name, %opts);

=head2 Linkage to file and directory components

Functionality of the L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm>
component can be accessed by using the "file_" prefix and provided
the filename through the C<filename> argument. (Note that this
should be the filename, not the full path)

Examples:

    my $file_obj = $vol_obj->file_create(filename => 'file', %other_opts);
    
    $vol_obj->file_delete(filename => 'file');
    
Functionality of the L<NACL::C::VolumeDirectory|lib-NACL-C-VolumeDirectory-pm>
component can be accessed by using the "directory_" prefix and provided
the filename through the C<dirname> argument. (Note that this
should be the directory name, not the full path)

Examples:

    my $dir_obj = $vol_obj->directory_create(dirname => 'dir', %other_opts);
    
    $vol_obj->directory_delete(dirname => 'dir');

=cut

sub can {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my @opts = $pkg_or_obj->_can_validation(@args);

    my $code;
    my ($method, $error_ptr) = @opts;
    if ($code = $pkg_or_obj->NACL::C::Component::can(@opts)) {
        $Log->exit();
        return $code;
    }

    # Link to implementations in NACL::C::VolumeFile and VolumeDirectory
    my $file_directory_handle = sub {
        my %args = validate_with(
            params => \@_,
            spec   => {
                linked_component => {type => SCALAR},
                linked_method    => {type => SCALAR},
                name_argument    => {type => SCALAR},
            }
        );
        my $linked_component = $args{linked_component};
        my $linked_method    = $args{linked_method};
        my $name_argument    = $args{name_argument};

        if ($linked_component->can($linked_method, $error_ptr)) {
            $code = sub {
                $Log->enter();
                my ($self, @args) = @_;
                my %opts = $self->_common_validate_with(
                    params          => \@args,
                    additional_spec => {$name_argument => {type => SCALAR}},

                    # Allow other options, which are those accepted by the
                    # NACL::C::Volume(File|Directory) method
                    allow_extra => 1
                );

                # The form of the file path is /vol/$vol_name/$file_name
                $opts{path} = "/vol/$opts{volume}/$opts{$name_argument}";
                delete $opts{volume};
                delete $opts{$name_argument};

                my @ret = $linked_component->$linked_method(%opts);
                $Log->exit();
                return wantarray ? @ret : $ret[0];
            };
        } else {
            $error_ptr .= "Could not find the method '$linked_method' in "
                . "$linked_component\n";
        }
    };

    if ($method =~ /file_(.*)/) {
        my $method = $1;
        my $linked_component;
        if ($pkg_or_obj->isa('NACL::STask::STask')) {
            $linked_component = 'NACL::STask::VolumeFile';
        } else {
            $linked_component = 'NACL::C::VolumeFile';
        }
        $file_directory_handle->(
            linked_component => $linked_component,
            linked_method    => $method,
            name_argument    => 'filename'
        );
    } elsif ($method =~ /directory_(.*)/) {
        my $method = $1;
        $file_directory_handle->(
            linked_component => 'NACL::C::VolumeDirectory',
            linked_method    => $method,
            name_argument    => 'dirname'
        );
    }

    if ($code) {
        $Log->exit();
        return $code;
    }

    $code = $pkg_or_obj->_handle_component_aggregation(
        method    => $method,
        error_ptr => $error_ptr,
        table     => [
            {   component => 'NACL::C::VolumeEfficiency',
                prefix    => 'efficiency'
            },
            {   component => 'NACL::C::VolumeSnapshot',
                prefix    => 'snapshot'
            },
        ],
    );

    if ($code) {
        $Log->exit();
        return $code;
    } elsif ($code = $pkg_or_obj->SUPER::can(@opts)) {
        $Log->exit();
        return $code;
    } else {
        $Log->exit();
        return undef;
    }
} ## end sub can

=head2 clone_create
    
    my $clone_c_obj     = $vol_c_obj->clone_create(flexclone => $clone_name);
    
    my $clone_stask_obj = $vol_stask_obj->clone_create();
    
Creates a clone of the volume on which the method was invoked. The
C<command_interface>, C<vserver> and C<parent-volume> arguments do not
need to be passed since these are extracted from the volume object.

If invoked with a component object, then the arguments that can be passed
to this method are those accepted by
L<the component|lib-NACL-C-VolumeClone-pm/create> and the object returned is
of type C<NACL::C::VolumeClone>.

If invoked with a task object, then the arguments that can be passed
to this method are those accepted by
L<the task|lib-NACL-STask-VolumeClone-pm/create> and the object returned is
of type C<NACL::STask::VolumeClone>.

=cut

sub clone_create {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params      => \@args,
        allow_extra => 1
    );
    $opts{'parent-volume'} = delete $opts{volume};
    my $clone_pkg
        = $pkg_or_obj->_get_package_at_same_layer(file => 'VolumeClone');
    my $clone_obj = $clone_pkg->create(%opts);
    $Log->exit();
    return $clone_obj;
}

=head2 clone_find

    # Scalar context, called with component object
    my $clone_c_obj = $vol_c_obj->clone_find();
    
    # List context, called with component object
    my @clone_c_objs = $vol_c_obj->clone_find();
    
    # Scalar context, called with task object
    my $clone_stask_obj = $vol_stask_obj->clone_find();
    
    # List context context, called with task object
    my @clone_stask_objs = $vol_stask_obj->clone_find();
    
    # Accepts all the options of find(). (Example here shows using
    # allow_empty)
    $vol_obj->clone_find(allow_empty => 1);
    
    # Also callable on a VolumeClone object, provides the same features
    # (scalar/list context; returns C/STask object based on type of object
    # with which method was called; accepts all options of find)
    my @clones = $clone_obj->clone_find();
    
Finds and returns objects for any flexclones of this volume.

=cut

sub clone_find {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {filter => {type => HASHREF, default => {}}},
        allow_extra     => 1
    );
    my %filter;
    $pkg_or_obj->_move_primary_keys(
        source => \%opts,
        target => \%filter
    );
    $filter{'parent-volume'} = delete $filter{volume};
    $opts{filter} = {%{$opts{filter}}, %filter};
    my $clone_pkg
        = $pkg_or_obj->_get_package_at_same_layer(file => 'VolumeClone');
    my @clones = $clone_pkg->find(%opts);
    $Log->exit();
    return wantarray ? @clones : $clones[0];
}

=head1 METHODS

=head2 new

  my $volume = NACL::C::Volume->new(command_interface=>$ci,volume => $vol_name,
                                  vserver => $vserver_name);

Return a new Volume component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'volume',
    scalar => 'vserver',

    # Not a primary key, but used to determine which ZAPI to invoke.
    # Hence, is a private attribute and is undocumented.
    scalar => '_is_repository',

    scalar => [
        {   '-read_cb' => sub {
                my $self     = $_[0];
                my $aggr_obj = $self->get_aggregate();
                return $aggr_obj->aggregate();
            },
            '-store_cb' => sub {
                my ($self, $value) = @_;
                $Log->warn('Trying to store \'aggregate\' information in '
                        . 'volume object persistently may cause this information '
                        . 'to go stale at any point. Eg: volume move operation '
                        . "\nPlease update your call not to store this "
                        . 'information in the object. When you invoke '
                        . '$obj->aggregate, Component will automatically '
                        . 'fetch the current value by querying the filer.');
                return $value;
            },
        },
        'aggregate'
    ],
];

# Backwards compatibility for the "is_repository" method
no strict 'refs';
*is_repository = *_is_repository;
use strict 'refs';

sub init {
    $Log->enter();
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit();
} ## end sub init

=head2 create

  my $volume  = NACL::C::Volume->create(command_interface=>$ci,volume =>$vol_name,
                                      vserver => $vserver_name);

(Class method) Request creation of a new volume element, and  return a new
volume component object that refers to it.  Does not wait for volume
creation to complete.

Uses a 7Mode CLI/ZAPI, CMode CLI/ZAPI APISet.

Invokes volume-create-async/volume-create API's for CMode ZAPI, 7Mode ZAPI.

Invokes "vol create" command for 7Mode ZAPI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: volume-name

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< disklist => [$disk1, $disk2, $disk3, ... ] >>

Applicable for CMode 7Mode CLI/ZAPI.

Maps to:

7M CLI : -d <value> (Programmatic name: "disks")

7M ZAPI: disks

=item C<< "mirror-disks" => [$disk1, $disk2, $disk3, ... ] >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: mirror-disks

7M CLI : mirror disks (Programmatic name: "mirror-disks")

=item C<< "force" => $boolean >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -f (Programmatic name: "force")

7M ZAPI: force

=item C<< size => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: size

7M CLI : volume size (Programmatic name: "size")

7M ZAPI: size

=item C<< state => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-state

=item C<< type => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-type

=item C<< policy => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: export-policy

=item C<< user => $string >>

If the user specified 'user' as string(user name, instead of user-id)
then CMode ZAPI executes NACL::CS::VserverServicesUnixUser->fetch()
method to get the user-id.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: user-id


=item C<< group => $string >>

If the user specified 'group' as string(group name, instead of group id)
then CMode ZAPI executes NACL::CS::VserverServicesUnixGroup->fetch()
method to get the group-id.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: group-id

=item C<< "security-style" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-security-style

=item C<< "unix-permissions" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: unix-permissions

=item C<< "junction-path" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: junction-path

=item C<< "junction-active" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-junction-active

=item C<< vsroot => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-vserver-root

=item C<< comment => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-comment

=item C<< "maxdir-size" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-dir-size

=item C<< "space-guarantee" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: space-reserve

7M CLI : space reserved (Programmatic name: "space-reserve")

=item C<< "snapshot-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot-policy

=item C<< "stripe-count" => $string >>

Applicable for CMode CLI

=item C<< "stripe-width" => $string >>

Applicable for CMode CLI

=item C<< concurrency => $string >>

Applicable for CMode CLI

=item C<< optimize => $string >>

Applicable for CMode CLI

=item C<< foreground => $string >>

Applicable for CMode CLI

=item C<< "antivirus-on-access-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: antivirus-on-access-policy

=item C<< "stripe-format" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: antivirus-on-access-policy

=item C<< "index-dir-enabled" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: index-dir-enabled

=item C<< "max-write-alloc-blocks" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-write-alloc-blocks

=item C<< aggregate => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: containing-aggr-name

7M CLI : aggregate name (Programmatic name: "containing-aggr-name")

7M ZAPI: containing-aggr-name

=item C<< language => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -l <value> (Programmatic name: "language-code-for-create")

7M ZAPI: language-code

=item C<< "is-mirrored" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -m (Programmatic name: "is-mirrored")

7M ZAPI: is-mirrored

=item C<< "space-reserve" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: space-reserve

7M CLI : -s <value>(Programmatic name: "space-reserve")

7M ZAPI: space-reserve

=item C<< "remote-location" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -S (Programmatic name: "remote-location")

7M ZAPI: remote-location

=item C<< "snaplock-type" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -S <value> (Programmatic name: "remote-location")

7M ZAPI: remote-location

=item C<< "raid-size" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -r <value> (Programmatic name: "raid-size")

7M ZAPI: raid-size

=item C<< "is-display-command" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -n (Programmatic name: "is-display-command")

7M ZAPI: is-display-command

=item C<< diskcount => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : number of disks (Programmatic name: "disk-count")

7M ZAPI: disk-count

=item C<< "disk-size" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : disk size (Programmatic name: "disk-size")

7M ZAPI: disk-size

=item C<< raidtype => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -t <value> (Programmatic name: "volume-raid-type")

7M ZAPI: volume-raid-type

=item C<<"disk-size" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : disk size (Programmatic name: "disk-size")

7M ZAPI: disk-size

=item C<< "is-snaplock" => $string >>

Applicable for 7Mode ZAPI.

=item C<< "disk-size-with-unit" => $string >>

Applicable for 7Mode ZAPI.

=item C<< "flexcache-origin-volume" => $string >>

Flexcache Origin Volume Name

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: flexcache-origin-volume-name

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'autosize' => $string >>

Applicable for CMode CLI

=item C<< 'max-autosize' => $string >>

Applicable for CMode CLI

=item C<< 'nvfail' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-nvfail-enabled

=item C<< 'flexcache-fill-policy' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: flexcache-fill-policy

=item C<< 'flexcache-cache-policy' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: flexcache-cache-policy

=item C<< 'autosize-increment' => $string >>

Applicable for CMode CLI

=item C<< 'directory-bitwidth' => $string >>

Applicable for CMode CLI

=item C<< 'flexcache-min-reserve' => $string >>

Applicable for CMode CLI

=item C<< 'constituent-role' => $string >>

Applicable for CMode CLI

=item C<< 'vm-align-suffix' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -u <value> (Programmatic name: "vm-align-suffix")

=item C<< 'vm-align-sector' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -e <value> (Programmatic name: "vm-align-sector")

=item C<< 'percent-snapshot-space' => $string >>

Applicable for CMode CLI

=item C<< 'is-repository' => $string >>

This boolean option (i.e. takes values of 0/1) describes
whether the volumes is on a repository vserver or not.

For CMode ZAPI, async ZAPIs must be run for repository vservers,
while the normal ZAPI must be run for normal vservers.
This option is optional, so if not provided then the
filer is queried to determine if the vserver is a repository vserver

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-repository

=item C<< 'autosize-mode' => $string >>

Applicable for CMode CLI

=item C<< 'min-autosize' => $string >>

Applicable for CMode CLI

=item C<< 'autosize-shrink-threshold-percent' => $string >>

Applicable for CMode CLI

=item C<< 'autosize-grow-threshold-percent' => $string >>

Applicable for CMode CLI

=item C<< 'ols-aggr-list' => [ $ols-aggr-1, $ols-aggr-2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'object-write-sync-period' => $string >>

Applicable for CMode CLI

=item C<< 'is-reserve-constituent' => $string >>

Applicable for CMode CLI

=item C<< 'enable-object-store' => $string >>

Applicable for CMode CLI

=item C<< 'ols-constituent-count' => $string >>

Applicable for CMode CLI

=item C<< 'ols-constituent-size' => $string >>

Applicable for CMode CLI

=item C<< 'qos-policy-group' => $string >>

Applicable for CMode CLI

=item C<< 'is-managed-by-service' => $string >>

Applicable for CMode CLI

=item C<< 'storage-service' => $string >>

Applicable for CMode CLI

=item C<< 'enable-snapdiff' => $string >>

Applicable for CMode CLI

=item C<< 'external-cache' => $string >>

External Cache Policy Name

Applicable for CMode CLI

=item C<< 'data-aggr-list' => [ $data-aggr-1, $data-aggr-2, ... ] >>

*List of Aggregates for Data Constituents
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'ns-mirror-aggr-list' => [ $ns-mirror-aggr-1, $ns-mirror-aggr-2, ... ] >>

*List of Aggregates for Namespace Mirrors
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'max-namespace-constituent-size' => $string >>

*Maximum Size of Namespace Constituent, possible value(s) are
<integer>[KB,MB,GB,TB,PB]

Applicable for CMode CLI

=item C<< 'namespace-aggregate' => $string >>

*Namespace Aggregate

Applicable for CMode CLI

=item C<< 'use-new-algorithm' => $string >>

*Use New Algorithm, possible value(s) are true,false

Applicable for CMode CLI

=item C<< 'percent-reserve-constituent' => $string >>

*Percent Space For Reserve Data Constituent

Applicable for CMode CLI

=item C<< 'max-data-constituent-size' => $string >>

*Maximum Size of each Data Constituent, possible value(s) are
<integer>[KB,MB,GB,TB,PB]

Applicable for CMode CLI

=item C<< 'autosize-increment-percent' => $string >>

Autosize Increment Percent (for flexvols only)

Applicable for CMode CLI

=item C<< 'unreachable-attr-action' => $string >>


Action When Attributes Are Not Reachable possible value(s) are, return-generated,wait
Applicable for CMode CLI

=item C<< 'caching-policy' => $string >>

Caching Policy Name
Applicable for CMode CLI

=item C<< 'junctionFileHandle' => $string >>

Junction File Handle
Applicable for CMode CLI

=item C<< 'snapshot-policy-uuid' => $string >>

Snapshot policy uuid
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume create" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Volume::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
volume that already exists.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter();
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "data-aggr-list"      => {type => ARRAYREF, optional => 1},
            "ns-mirror-aggr-list" => {type => ARRAYREF, optional => 1},
            "ols-aggr-list" => {type => ARRAYREF | SCALAR, optional => 1},
            disklist        => {type => ARRAYREF,          optional => 1},
            "mirror-disks"  => {type => ARRAYREF,          optional => 1},
            force           => {type => BOOLEAN,           optional => 1},
            "is-snaplock"   => {type => BOOLEAN,           optional => 1},
            job_component   => {type => SCALARREF,         optional => 1},
            _optional_scalars
                qw(size state type policy user group security-style
                unix-permissions junction-path junction-active vsroot comment
                maxdir-size space-guarantee snapshot-policy stripe-count stripe-width
                concurrency optimize foreground antivirus-on-access-policy stripe-format
                index-dir-enabled max-write-alloc-blocks aggregate language
                is-mirrored space-reserve remote-location snaplock-type raid-size
                is-display-command diskcount disk-size raidtype disk-size
                is-snaplock disk-size-with-unit flexcache-origin-volume autosize
                max-autosize nvfail flexcache-fill-policy flexcache-cache-policy
                autosize-increment directory-bitwidth flexcache-min-reserve
                constituent-role vm-align-suffix vm-align-sector
                percent-snapshot-space is-repository
                autosize-mode min-autosize autosize-shrink-threshold-percent
                autosize-grow-threshold-percent
                object-write-sync-period is-reserve-constituent enable-object-store
                ols-constituent-count ols-constituent-size qos-policy-group
                is-managed-by-service storage-service enable-snapdiff
                external-cache
                max-namespace-constituent-size namespace-aggregate use-new-algorithm
                percent-reserve-constituent max-data-constituent-size
                autosize-increment-percent unreachable-attr-action

                caching-policy junctionFileHandle snapshot-policy-uuid
                ),
        },
        api => 'volume_create',
    );

    my $ret;
    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_create_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },

            ],
            frontend => 'NACL::C::Volume::create',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Volume::AlreadyExists->detect_convert_and_throw(
            exception => $exception);
        $Log->exit();
        $exception->throw();
    };

    $Log->exit();
    return $ret;

} ## end sub create

sub _create_7mode_cli {
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};

    my $apiset = delete $opts{apiset};

    my $vol_args = $pkg->_hash_copy(
        source => \%opts,
        copy   => [
            qw(volume force is-mirrored space-reserve size
                remote-location snaplock-type raid-size mirror-disks
                is-display-command disk-size vm-align-sector vm-align-suffix)
        ],
        map => {
            "disklist"        => "disks",
            "aggregate"       => "containing-aggr-name",
            "raidtype"        => "volume-raid-type",
            "diskcount"       => "disk-count",
            'space-guarantee' => 'space-reserve',
            "language"        => "language-code-for-create",
        }
    );

    %opts = %$vol_args;
    if (defined($opts{disks})) {
        $opts{disks} = join(" ", @{$opts{disks}});
    }

    if (defined($opts{"mirror-disks"})) {
        $opts{"mirror-disks"} = join(" ", @{$opts{"mirror-disks"}});
    }

    $pkg->_translate_size_suffixes(
        options_to_translate => [qw(size)],
        api_opts             => \%opts,
    );

    $apiset->vol_create(%opts);

    return $pkg->new(
        volume            => $opts{volume},
        vserver           => undef,
        command_interface => $command_interface
    );
} ## end sub _create_7mode_cli

sub _create_cmode_cli {
    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $job_component     = delete $opts{job_component};

    my %cmode_args = $pkg->get_api_options(
        apiset  => $apiset,
        api     => 'volume_create',
        options => \%opts,
    );

    my $vserver = $pkg->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'create',
    );

    my $response = $apiset->volume_create(%cmode_args);
    my $jobid    = $response->get_job_ids();
    if ($jobid) {
        $$job_component = NACL::C::Job->new(
            id                => $jobid,
            command_interface => $command_interface
        );
    } ## end if ($jobid)

    $Log->exit();
    return $pkg->new(
        volume            => $opts{volume},
        vserver           => $vserver,
        command_interface => $command_interface
    );
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts              = $pkg->_validate_backend_opts(@args);
    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $allowed_cmode_options = [
        qw(antivirus-on-access-policy constituent-role
            flexcache-cache-policy flexcache-fill-policy index-dir-enabled
            junction-path max-write-alloc-blocks size snapshot-policy
            storage-service unix-permissions vm-align-sector vm-align-suffix
            volume)
    ];
    my $cmode_args = $pkg->_hash_copy(
        source => \%opts,
        copy   => $allowed_cmode_options,
        map    => {
            'type'                    => 'volume-type',
            'state'                   => 'volume-state',
            'security-style'          => 'volume-security-style',
            'comment'                 => 'volume-comment',
            'user'                    => 'user-id',
            'space-guarantee'         => 'space-reserve',
            'maxdir-size'             => 'max-dir-size',
            'vsroot'                  => 'is-vserver-root',
            'nvfail'                  => 'is-nvfail-enabled',
            'junction-active'         => 'is-junction-active',
            'group'                   => 'group-id',
            'flexcache-origin-volume' => 'flexcache-origin-volume-name',
            'policy'                  => 'export-policy',
            'aggregate'               => 'containing-aggr-name',
            'percent-snapshot-space'  => 'percentage-snapshot-reserve',
            'qos-policy-group'        => 'qos-policy-group-name',
        },
    );

    $pkg->_convert_multiple_to_bytes(
        api_opts     => $cmode_args,
        byte_options => [qw(max-dir-size)]
    );

    $pkg->_strip_percentage_from_end_for_multiple(
        api_opts        => $cmode_args,
        percent_options => [qw(percentage-snapshot-reserve)]
    );

    my $vserver = $pkg->_handle_zapi_vserver_context(
        api_opts          => $cmode_args,
        vserver           => $opts{vserver},
        command_interface => $command_interface
    );

    #This code gets the user-id, and group-id,
    #if the user specifed these options as strings(names, instead of id's)
    if (defined $opts{'user-id'} && $opts{'user-id'} !~ /\d+/) {
        require NACL::CS::VserverServicesUnixUser;
        my $cs_obj = NACL::CS::VserverServicesUnixUser->fetch(
            command_interface => $command_interface,
            filter            => {'user' => $opts{'user-id'}},
            requested_fields  => [qw ( id )],
            apiset_should     => {interface => 'ZAPI'},
        );
        $cmode_args->{'user-id'} = $cs_obj->id();
    } ## end if ( defined $opts{'user-id'...})

    if (defined $opts{'group-id'} && $opts{'group-id'} !~ /\d+/) {
        require NACL::CS::VserverServicesUnixGroup;
        my $cs_obj = NACL::CS::VserverServicesUnixGroup->fetch(
            command_interface => $command_interface,
            filter            => {'name' => $opts{'group-id'}},
            requested_fields  => [qw ( id )],
            apiset_should     => {interface => 'ZAPI'},
        );
        $cmode_args->{'group-id'} = $cs_obj->id();
    } ## end if ( defined $opts{'group-id'...})

    my $response;
    my $value = $pkg->_determine_is_repository(%opts);
    if ($value) {
        $cmode_args = $pkg->_hash_move(
            source => $cmode_args,
            map    => {
                'volume'        => 'volume-name',
                'space-reserve' => 'space-guarantee'
            },
            target => $cmode_args,
        );
        $response = $apiset->volume_create_async(%$cmode_args);
    } else {
        $response = $apiset->volume_create(%$cmode_args);
    }

    $pkg->_populate_job_components(
        command_interface => $command_interface,
        response          => $response,
        job_component     => $opts{job_component}
    );

    $Log->exit();

    my $obj = $pkg->new(
        volume            => $opts{volume},
        vserver           => $vserver,
        command_interface => $command_interface,
    );

    $obj->is_repository($value);
    return $obj;
} ## end sub _create_cmode_zapi

sub _create_7mode_zapi {
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};

    my $apiset = delete $opts{apiset};

    my $vol_args = [
        qw(volume force is-mirrored space-reserve size
            remote-location snaplock-type raid-size mirror-disks
            is-display-command disk-size disk-size-with-unit is-snaplock)
    ];

    $vol_args = $pkg->_hash_copy(
        source => \%opts,
        copy   => $vol_args,
        map    => {
            "language"  => "language-code",
            "disklist"  => "disks",
            "aggregate" => "containing-aggr-name",
            "raidtype"  => "volume-raid-type",
            "diskcount" => "disk-count",
        }
    );
    %opts = %$vol_args;

    $pkg->_translate_size_suffixes(
        options_to_translate => [qw(size)],
        api_opts             => \%opts,
    );

    $apiset->volume_create(%opts);

    return $pkg->new(
        volume            => $opts{volume},
        vserver           => undef,
        command_interface => $command_interface
    );
} ## end sub _create_7mode_zapi

=head2 delete

  NACL::C::Volume->delete(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name);

                            or

  $volume_object->delete(...);

(Class or instance method)This method deletes a volume element.

Uses a 7Mode CLI/ZAPI, CMode CLI/ZAPI APISet.

Invokes "volume-destroy,volume-destroy-async" API for CMode ZAPI,7Mode ZAPI.

Invokes "vol destroy" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: name

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: name

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< foreground =>$string >>

Applicable for CMode CLI

=item C<< "disable-offline-check " =>$string >>

Applicable for CMode CLI

=item C<< force => $boolean >>

(Optional) A boolean when true will do a forceful deleting of the volume .

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : -f (Programmatic name: "force")

7M ZAPI: force

=item C<< destroy_corrupt_volume => $boolean >>

(Optional defaults to 0)

If 7-Mode CLI command is used to delete the inconsistent volume,filer throws the
prompt 'This volume is marked corrupt and could be useful for debugging
 purposes, are you sure you want to proceed? ' and asks for user confirmation.

Set destroy_corrupt_volume to 1 to delete inconsistent volume.

This option is not applicable and ignored for 7-Mode ZAPI and C-Mode CLI/ZAPI.

=item C<< "job_component"=> \$scalar >>

=item C<< 'is-repository' => 0 | 1 >>

(Applicable only for CMode ZAPI)
This boolean option (i.e. takes values of 0/1) describes
whether the volume is on a repository vserver or not.

For CMode ZAPI, async ZAPIs must be run for repository vservers,
while the normal ZAPI must be run for normal vservers.
This option is optional, so if not provided then the
filer is queried to determine if the vserver is a repository vserver.

Applicable for CMode ZAPI.

Maps to:

CM ZAPI: is-repository

=item C<< '*' >>

This method takes all other options supported by
"volume delete" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Volume::DoesNotExist>

This type of exception is thrown when an attempt is made to delete
a non-existent volume.

=item C<NACL::C::Exceptions::Volume::DeleteSnapMirrorRelationship>

This type of exception gets thrown if an attempt is made to delete a
volume that has a snapmirror relationship.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            volume                 => {type => SCALAR},
            force                  => {type => BOOLEAN, optional => 1},
            job_component          => {type => SCALARREF, optional => 1},
            destroy_corrupt_volume => {type => BOOLEAN, default => 0},
            _optional_scalars
                qw( foreground disable-offline-check is-repository
                ),
        },
        api => 'volume_delete',
    );
    my $ret;

    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_delete_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },

            ],
            frontend => 'NACL::C::Volume::delete',
        );

    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        $Log->enter();
        my $exception_object = $_[0];
        foreach my $type (
            qw(DoesNotExist DeleteSnapMirrorRelationship
            HasClones CannotDeleteRootVolume)
            )
        {
            my $full_type = "NACL::C::Exceptions::Volume::$type";
            $full_type->detect_convert_and_throw(
                exception => $exception_object);
        }
        $Log->exit();
        $exception_object->throw();
    };

    $Log->exit();
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $job_component     = delete $opts{job_component};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_delete',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'delete',
    );

    my $response = $apiset->volume_delete(%cmode_args);
    my $jobid    = $response->get_job_ids();
    if ($jobid) {
        $$job_component = NACL::C::Job->new(
            id                => $jobid,
            command_interface => $command_interface
        );
    } ## end if ($jobid)
    $Log->exit();
} ## end sub _delete_cmode_cli

sub _delete_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface   = delete $opts{command_interface};
    my $apiset              = delete $opts{apiset};
    my $destroy_corrupt_vol = delete $opts{destroy_corrupt_volume};

    my $prompt_answer_yes = ['.*volume is marked corrupt .*' => 'y'];
    my $prompt_answer_no  = ['.*volume is marked corrupt .*' => 'n'];

    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume force)],
    );
    %opts = %$vol_args;
    $opts{force} = 1;

    if ($destroy_corrupt_vol) {
        $opts{'connectrec-match_table'} = $prompt_answer_yes;
    } else {
        $opts{'connectrec-match_table'} = $prompt_answer_no;
    }

    $apiset->vol_destroy(%opts);
} ## end sub _delete_7mode_cli

sub _delete_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface   = delete $opts{command_interface};
    my $apiset              = delete $opts{apiset};
    my $destroy_corrupt_vol = delete $opts{destroy_corrupt_volume};

    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            volume => "name",
            force  => "force",
        }
    );
    %opts = %$vol_args;
    $opts{force} = 1;

    $apiset->volume_destroy(%opts);
} ## end sub _delete_7mode_zapi

sub _delete_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_repo_builder(
        @args,
        api => 'volume-destroy',
        map => {volume => 'name'}
    );

    $Log->exit();
} ## end sub _delete_cmode_zapi

=head2 unmount

  NACL::C::Volume->unmount(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name);

                            or

  $volume_object->unmount(...);

(Class or instance method) This method unmounts a volume element.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-unmount" API for CMode ZAPI.

Invokes "exportfs" command for 7Mode CLI.

For agnostic purposes, this function is a no-op for 7Mode

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: volume-name

7M CLI : volume name (Programmatic name: "volume")

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< force => $boolean >>

(Optional) A boolean when true will do a forceful unmounting of the volume .

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: force

7M CLI : -f (Programmatic name: "force")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume unmount" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Volume::UnmountRoot>

This type of exception gets thrown if an attempt is made to unmount a
vserver's root volume.

=item C<NACL::C::Exceptions::Volume::DoesNotExist>

This type of exception gets thrown if an attempt is made to unmount a
non-existing volume.

=item C<NACL::C::Exceptions::Volume::UnsupportedUnmount>

This type of exception gets thrown if an attempt is made to unmount a
volume of type DP or DC.

=item C<NACL::C::Exceptions::Volume::NotExported>

This type of exception gets thrown if an attempt is made to unmount a
volume which is not exported.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub unmount {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {force => {type => BOOLEAN, optional => 1},},
        api             => 'volume_unmount',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_unmount_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_unmount_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_unmount_7mode_cli',
                    set       => '7Mode',
                    interface => 'CLI',
                },
            ],
            frontend => 'NACL::C::Volume::unmount',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        foreach my $type (
            qw(UnmountRoot DoesNotExist UnsupportedUnmount NotExported))
        {
            my $full_type = "NACL::C::Exceptions::Volume::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }

        $Log->exit();
        $exception->throw();
    };

    $Log->exit();
} ## end sub unmount

sub _noop { }    # null action for 7Mode use of CMode-only actions.

sub _unmount_cmode_zapi {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vserver           = delete $opts{vserver};

    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(force)],
        map    => {volume => "volume-name",}
    );
    %opts = %$vol_args;

    $pkg_or_obj->_handle_zapi_vserver_context(
        api_opts          => \%opts,
        vserver           => $vserver,
        command_interface => $command_interface
    );

    $apiset->volume_unmount(%opts);
    $Log->exit();
} ## end sub _unmount_cmode_zapi

sub _unmount_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_unmount',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'unmount',
    );

    $apiset->volume_unmount(%cmode_args);
    $Log->exit();
} ## end sub _unmount_cmode_cli

sub _unmount_7mode_cli {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %cmd_args;
    $cmd_args{path} = "/vol/" . $opts{volume};
    if (exists($opts{force}) and $opts{force} =~ /true/) {
        $cmd_args{zap} = "true";
    } else {
        $cmd_args{unexport} = "true";
    }

    my $response = $apiset->exportfs(%cmd_args);
    $Log->exit();
} ## end sub _unmount_7mode_cli

=head2 mount

  NACL::C::Volume->mount(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name);

                            or

  $volume_object->mount(...);

(Class or instance method)

For CMode, this method mounts a volume element on a particular vserver's
junction point. (not to be confused with I<client>-side mounting, such as
over NFS).

In 7Mode, exports the volume using "exportfs"

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-mount" API for CMode ZAPI.

Invokes "exportfs,rdfile" commands for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-name

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< junction-path=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: junction-path

=item C<< policy-override=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: export-policy-override

=item C<< active=>$string >>

Applicable for CMode CLI

=item C<< activate-junction=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: activate-junction

=item C<< '*' >>

This method takes all other options supported by
"volume mount" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item L<NACL::C::Exceptions::Volume::AlreadyMounted|lib-NACL-C-Exceptions-Volume-AlreadyMounted-pm>

This type of exception is thrown when an attempt is made to mount a volume
that has already been mounted at a different junction-path.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub mount {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw ( junction-path active policy-override
                activate-junction ),
        },
        api => 'volume_mount',
    );
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_mount_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_mount_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_mount_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::Volume::mount'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Volume::AlreadyMounted->detect_convert_and_throw(
            exception => $exception);

        $Log->exit();
        $exception->throw();
    };

    $Log->exit();
} ## end sub mount

sub _mount_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_mount',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'mount',
    );

    $apiset->volume_mount(%cmode_args);
    $Log->exit();
} ## end sub _mount_cmode_cli

sub _mount_cmode_zapi {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vserver           = delete $opts{vserver};

    my $cmode_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(junction-path activate-junction)],
        map    => {
            "volume"          => "volume-name",
            "policy-override" => "export-policy-override",
        }
    );

    %opts = %$cmode_args;

    $pkg_or_obj->_handle_zapi_vserver_context(
        api_opts          => \%opts,
        vserver           => $vserver,
        command_interface => $command_interface
    );

    $apiset->volume_mount(%opts);
    $Log->exit();
} ## end sub _mount_cmode_zapi

sub _mount_7mode_cli {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %cmd_args;

    my $response = $apiset->rdfile(file => "/etc/exports");
    my $output = $response->get_parsed_output();

    if (grep (/\/vol\/$opts{volume}/, split(/\n/, $output->[0]->{contents})))
    {
        $cmd_args{path} = "/vol/$opts{volume}";
        $apiset->exportfs(%cmd_args);
    } else {
        $cmd_args{persistent} = "/vol/" . $opts{volume};
        $apiset->exportfs(%cmd_args);
    } ## end else [ if ( grep ( /\/vol\/$opts{volume}/...))]

    $Log->exit();
} ## end sub _mount_7mode_cli

=head2 modify

  NACL::C::Volume->modify(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name,...);

                            or

  $volume_object->modify(...);

(Class or instance method) Modify a volume element.

For CMode, the capabilities of the "volume modify" command are available.
For 7Mode, "modify" only knows so far how to modify
the "state" attribute of the  volume element, setting it to be "online",
"offline", or  "restricted".

Uses a 7Mode CLI/ZAPI, CMode CLI/ZAPI APISet.

Invokes "volume-set-total-files" API for CMode ZAPI.

Invokes "volume-online,volume-offline,volume-restrict" API's for 7Mode ZAPI.

Invokes "vol online,vol offline,vol restrict,vol autosize"
commands for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: name

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: name

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: owning-vserver-name

=item C<< size => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: size

7M CLI : volume size (Programmatic name: "size")

7M ZAPI: size

=item C<< state => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: state

7M ZAPI: state

=item C<< policy => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy

=item C<< user => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: user-id

=item C<< group => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: group-id

=item C<< "security-style" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: style

=item C<< "unix-permissions" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: permissions

=item C<< "junction-active" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-junction-active

=item C<< comment => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: comment

=item C<< autosize => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: is-enabled

7M ZAPI: is-enabled

=item C<< "max-autosize" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: maximum-size

7M CLI : -m <value> (Programmatic name: "maximum-size")

7M ZAPI: maximum-size

=item C<< "autosize-increment" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: increment-size

7M CLI : -i <value> (Programmatic name: "increment-size")

7M ZAPI: increment-size

=item C<< files => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: files-total

=item C<< "maxdir-size" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-dir-size

=item C<< "space-guarantee" => $string >>

Applicable for CMode CLI

=item C<< "min-readahead" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: minimal-read-ahead

=item C<< "atime-update" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-atime-update-enabled

=item C<< "snapdir-access" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapdir-access-enabled

=item C<< "percent-snapshot-space" => $string >>

Applicable for CMode CLI

=item C<< "snapshot-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot-policy

=item C<< "stripe-width" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: stripe-width

=item C<< concurrency => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: concurrency

=item C<< optimize => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: optimize

=item C<< "index-dir-enabled" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: index-dir-enabled

=item C<< "max-write-alloc-blocks" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-write-alloc-blocks

=item C<< "antivirus-on-access-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-antivirus-attributes

=item C<< "cifs-delay" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -t <value> (Programmatic name: "cifs-delay")

7M ZAPI: cifs-delay

=item C<< language => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: language

=item C<< "stripe-format" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: stripe-format

=item C<< "upgrade-64bit-available-space" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: available-space

=item C<< "upgrade-64bit-capacity" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: capacity

=item C<< "upgrade-64bit-grow-space" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: grow-space

=item C<< "upgrade-64bit-percent-used-space" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: used-space

=item C<< "root-dir-gen" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: root-dir-gen

=item C<< "block-type" => $string  >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: block-type

=item C<< "files-private-used" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: files-private-used

=item C<< "files" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: files-total

=item C<< "files-used" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: files-used

item C<< inodefile-private-capacity => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: inodefile-private-capacity

=item C<< inodefile-public-capacity => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: inodefile-public-capacity

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'read-realloc' => $string >>

Valid values are "on/off/space-optimized".

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM CLI : "read-realloc"

CM ZAPI: "read-realloc"

<volume-attributes>
<volume-performance-attributes>
<read-realloc>space-optimized</read-realloc>
</volume-performance-attributes>
</volume-attributes>

Invokes "volume-modify-iter" API for CMode ZAPI.

7M CLI: "read_realloc"

Invokes "vol options $vol_name read_realloc $val" command for 7Mode CLI.

7M ZAPI: "read_realloc"

<volume-set-option>
<option-name>read_realloc</option-name>
<option-value>space_optimized</option-value>
<volume>$vol_name</volume>
</volume-set-option>

Invokes "volume-set-option" API for 7Mode ZAPI.

=item C<< 'sched-snap-name' => $string >>

Naming Scheme for Automatic Snapshots. Valid values are "create_time" and "ordinal".

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7Mode CLI: "schedsnapname" of "vol options"

7Mode ZAPI: "schedsnapname" option-name of "volume-set-options"

=item C<< 'extent-enabled' => $string >>

Extent Option. Valid values are "on", "off" and "space-optimized".

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CMode ZAPI:

<volume-performance-attributes>
<extent-enabled>
</extent-enabled>
</volume-performance-attributes>

7Mode CLI: "extent" option of "vol options"

7Mode ZAPI: "extent" option-name of "volume-set-options"

Note that in 7Mode, "space_optimized" is a valid value, not
"space-optimized". So, if a call is made like:
NACL::C::Volume->modify(%opts, 'extent-enabled' => 'space-optimized');
the value automatically gets translated to 'space_optimized' in 7Mode.

=item C<< 'snapshot-clone-dependency' => $string >>

Snapshot Cloning Dependency. Valid values are "on" and "off".

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CMode ZAPI:

<volume-snapshot-attributes>
<snapshot-clone-dependency>
</snapshot-clone-dependency>
</volume-snapshot-attributes>
of "volume-modify-iter"

7Mode CLI: "snapshot_clone_dependency" option of "vol options"

7Mode ZAPI: "snapshot_clone_dependency" option-name of "volume-set-option"

=item C<< 'auto-snapshots-enabled' => $string >>

Should Automatic Snapshots be enabled? Valid values are "true" and "false".

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CMode ZAPI:

<volume-snapshot-attributes>
<auto-snapshots-enabled>
</auto-snapshots-enabled>
</volume-snapshot-attributes>
of "volume-modify-iter"

7Mode CLI: "nosnap" option of "vol options"

7Mode ZAPI: "nosnap" option-name of "volume-set-option"

Note that since "nosnap" means the opposite of "auto-snapshots-enabled"
("nosnap" means "should automatic snapshots be disabled?")
in 7Mode the value are flipped, i.e.
'auto-snapshots-enabled' => 'true' internally gets translated to
'nosnap' => 'false'.

=item C<< 'create-ucode' => $true_or_false >>

Use Unicode Directory Format. Valid values are "true" and "false".

Applicable for CMode ZAPI and 7Mode CLI.

Maps to:

CMode ZAPI:

<volume-language-attributes>
<create-ucode> $value </create-ucode>
</volume-language-attributes>
of "volume-modify-iter"

7Mode CLI: "create_ucode" option of "vol options"

=item C<< 'instance-uuid' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: instance-uuid

=item C<< 'provenance-uuid' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: provenance-uuid

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'flexcache-min-reserve' => $string >>

=item C<< 'is-repository' => 0 | 1 >>

Applicable for CMode CLI

This boolean option (i.e. takes values of 0/1) describes
whether the volumes is on a repository vserver or not.

For CMode ZAPI, async ZAPIs must be run for repository vservers,
while the normal ZAPI must be run for normal vservers.
This option is optional, so if not provided then the
filer is queried to determine if the vserver is a repository vserver

=item C<< 'autosize-mode' => $string >>

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : autosize mode (Programmatic name: "autosize-mode")

7M ZAPI: mode

=item C<< 'force-nvfail-on-dr' => $string >>

Applicable for CMode CLI

=item C<< 'min-autosize' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: minimum-size

7M CLI : -minimum-size <value> (Programmatic name: "minimum-size")

7M ZAPI: minimum-size

=item C<< 'autosize-grow-threshold-percent' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: grow-threshold-percent

=item C<< 'autosize-shrink-threshold-percent' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: shrink-threshold-percent

=item C<< 'ols-aggr-list' => [ $ols-aggr-1, $ols-aggr-2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'object-write-sync-period' => $string >>

Applicable for CMode CLI

=item C<< 'is-reserve-constituent' => $string >>

Applicable for CMode CLI

=item C<< 'enable-object-store' => $string >>

Applicable for CMode CLI

=item C<< 'ols-constituent-count' => $string >>

Applicable for CMode CLI

=item C<< 'ols-constituent-size' => $string >>

Applicable for CMode CLI

=item C<< 'minimum-size' => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -minimum-size <value> (Programmatic name: "minimum-size")

7M ZAPI: minimum-size

=item C<< 'grow-threshold-percent' => $string >>

Applicable for CMode CLI

=item C<< 'shrink-threshold-percent' => $string >>

Applicable for CMode CLI

=item C<< 'shrink' => $string >>

Applicable for CMode CLI

=item C<< 'on' => $string >>

Applicable for CMode CLI

=item C<< 'off'=> $string >>

Applicable for CMode CLI

=item C<< 'grow-shrink'=> $string >>

Applicable for CMode CLI

=item C<< 'reset'=> $string >>

Applicable for CMode CLI

=item C<< 'qos-policy-group' => $string >>

Applicable for CMode CLI

=item C<< 'enable-snapdiff' => $string >>

Applicable for CMode CLI

=item C<< 'is-managed-by-service' => $string >>

Applicable for CMode CLI

=item C<< 'storage-service' => $string >>

Applicable for CMode CLI

=item C<< 'fc-delegs-enabled' => $string >>

*FlexCache Delegations, possible value(s) are true,false

Applicable for CMode CLI

=item C<< 'autosize-reset' => $string >>

Autosize Reset

Applicable for CMode CLI

=item C<< 'external-cache' => $string >>

External Cache Policy Name

Applicable for CMode CLI

=item C<< 'ns-mirror-aggr-list' => [ $ns-mirror-aggr-1, $ns-mirror-aggr-2, ... ] >>

*List of Aggregates for Namespace Mirrors
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'max-namespace-constituent-size' => $string >>

*Maximum Size of Namespace Constituent, possible value(s) are
<integer>[KB,MB,GB,TB,PB]

Applicable for CMode CLI

=item C<< 'use-new-algorithm' => $string >>

*Use New Algorithm, possible value(s) are true,false

Applicable for CMode CLI

=item C<< 'percent-reserve-constituent' => $string >>

*Percent Space For Reserve Data Constituent

Applicable for CMode CLI

=item C<< 'max-data-constituent-size' => $string >>

*Maximum Size of each Data Constituent, possible value(s) are
<integer>[KB,MB,GB,TB,PB]

Applicable for CMode CLI

=item C<< 'space-nearly-full-threshold-percent' => $string >>

Volume Nearly Full Threshold Percent

Applicable for CMode CLI

=item C<< 'space-full-threshold-percent' => $string >>

Volume Full Threshold Percent

Applicable for CMode CLI

=item C<< 'autosize-increment-percent' => $string >>

Autosize Increment Percent (for flexvols only)

Applicable for CMode CLI

=item C<< 'unreachable-attr-action' => $string >>


Action When Attributes Are Not Reachable possible value(s) are, return-generated,wait
Applicable for CMode CLI

=item C<< 'caching-policy' => $string >>

Caching Policy Name
Applicable for CMode CLI

=item C<< 'snapshot-policy-uuid' => $string >>

Snapshot policy uuid
Applicable for CMode CLI

=item C<< 'junctionFileHandle' => $string >>

Junction File Handle
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume modify" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Volume::DoesNotExist>

This type of exception is thrown when an attempt is made to modify the
attributes of a non-existent volume.

=item C<NACL::C::Exceptions::Volume::AlreadyModified>

This type of exception is thrown when an attempt is made to modify the
attributes of a volume into a state it's already in. For example, attempting
to offline an already offline volume, or force-online an already online
volume will both result in this sort of exception.

=item C<NACL::C::Exceptions::Volume::StateUnknown>

This type of exception is thrown when an attempt is made to modify the
attributes of a volume whose state is unknown ("-")

=item C<NACL::C::Exceptions::Volume::CurrentlyBusy>

This type of exception is thrown when an attempt is made to modify the
attributes of a volume when the volume is "dynamically growing state" or busy.

=item C<NACL::C::Exceptions::Volume::VolumeOffline>

This type of exception is thrown when an attempt is made to modify the
attributes of a volume which is in offline state.

=item C<NACL::C::Exceptions::Volume::VserverLocked>

This type of exception is thrown when an attempt is made to modify the
attributes of a volume when its vserver is locked (either config-locked
or baseline is being generated. See burt 637356)

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "ns-mirror-aggr-list" => {type => ARRAYREF, optional => 1},
            "ols-aggr-list" => {type => ARRAYREF | SCALAR, optional => 1},
            job_component   => {type => SCALARREF,         optional => 1},
            _optional_scalars(
                qw (size state policy user group security-style unix-permissions
                    junction-active comment autosize max-autosize autosize-increment
                    files maxdir-size space-guarantee min-readahead atime-update
                    snapdir-access percent-snapshot-space snapshot-policy
                    stripe-width concurrency optimize antivirus-on-access-policy
                    flexcache-cache-policy flexcache-fill-policy nvfail
                    filesys-size-fixed fractional-reserve index-dir-enabled
                    max-write-alloc-blocks convert-ucode space-mgmt-try-first
                    in-nvfailed-state snap-autodelete-enabled
                    snap-autodelete-commitment snap-autodelete-defer-delete
                    snap-autodelete-delete-order snap-autodelete-prefix
                    snap-autodelete-target-free-space snap-autodelete-trigger
                    snap-autodelete-destroy-list cifs-delay auto-snapshots-enabled
                    read-realloc extent-enabled
                    sched-snap-name snapshot-clone-dependency instance-uuid provenance-uuid
                    foreground
                    flexcache-min-reserve is-repository
                    autosize-mode force-nvfail-on-dr min-autosize
                    autosize-grow-threshold-percent autosize-shrink-threshold-percent i2p-enabled
                    object-write-sync-period is-reserve-constituent enable-object-store
                    ols-constituent-count ols-constituent-size
                    qos-policy-group
                    enable-snapdiff is-managed-by-service storage-service
                    fc-delegs-enabled autosize-reset
                    external-cache
                    max-namespace-constituent-size use-new-algorithm
                    percent-reserve-constituent max-data-constituent-size
                    space-nearly-full-threshold-percent space-full-threshold-percent
                    autosize-increment-percent create-ucode

                    unreachable-attr-action

                    caching-policy snapshot-policy-uuid junctionFileHandle
                    )
            ),
        },
        api => 'volume_modify',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_modify_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_modify_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_modify_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
                {   method    => '_modify_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::Volume::modify',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        if (defined $opts{state} && $opts{state} eq "offline") {
            if ($exception->text() =~ /volume\s+(.*?\s)*is offline/i) {
                NACL::C::Exceptions::Volume::AlreadyModified
                    ->convert_and_throw(exception => $exception);
            }
        }

        foreach my $type (
            qw(DoesNotExist AlreadyModified StateUnknown VolumeOffline
            VserverLocked MustBeUnmounted)
            )
        {
            my $full_type = "NACL::C::Exceptions::Volume::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }

        my $fileops = 'NACL::C::Exceptions::Volume::FileOperationsInProgress';
        if ($fileops->detect(exception => $exception)) {

            # This error implies that "force-offline" is supported for
            # "volume modify -state <state>"
            $opts{command_interface}->version_manager()
                ->features_set('force_offline_for_volume' => 1);

            $Log->exit();
            $fileops->convert_and_throw(exception => $exception);
        }

        $Log->exit();
        $exception->throw();
    };

    $Log->exit();
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $job_component     = delete $opts{job_component};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_modify',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'modify',
    );

    my $response = $apiset->volume_modify(%cmode_args);
    my $jobid    = $response->get_job_ids();
    if ($jobid) {
        $$job_component = NACL::C::Job->new(
            id                => $jobid,
            command_interface => $command_interface
        );
    } ## end if ($jobid)
    $Log->exit();
} ## end sub _modify_cmode_cli

sub _modify_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $volume            = $opts{volume};

    # This code must be run before running any of the other APIs
    # This is because, in CMode, if we attempt modifying the state along
    # with other fields, it attempts to change the state first before
    if (exists $opts{state}) {
        my $state = delete $opts{state};
        try {
            if ($state eq 'online') {
                $apiset->vol_online(
                    volume => $volume,
                    force  => $opts{force}
                );
            } else {
                my %args_offline_restrict;
                $pkg_or_obj->_hash_copy(
                    source => \%opts,
                    copy   => [qw(volume cifs-delay)],
                    target => \%args_offline_restrict
                );
                if ($state eq 'offline') {
                    $apiset->vol_offline(%args_offline_restrict);
                } elsif ($state eq 'restricted') {
                    $apiset->vol_restrict(%args_offline_restrict);
                } else {
                    NATE::Exceptions::Argument->throw("'state' can be any of "
                            . "'online', 'offline' or 'restricted' "
                            . "but the value supplied was '$state'");
                } ## end else [ if ( $state eq 'offline')]
            } ## end else [ if ( $state eq 'online')]
        } ## end try
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $ex = shift;
            $Log->debug($ex->text());

# For pre-SN, CMode does not throw an exception, so 7Mode should behave the same way
            _check_if_pre_SN($ex, $command_interface);
        };
    } ## end if ( exists $opts{state...})

    # Modifying the snapshot policy here. CMode allows us to set a snapshot
    # policy for each volume. 7Mode doesn't allow this, but it does allow
    # us to control whether a volume should have automatic snapshot enabled
    # or not. A CMode snapshot-policy of 'none' would in 7Mode mean
    # "switch off auto-snapshots", and any other snapshot-policy would mean
    # "switch on snapshots".

    if (exists $opts{'read-realloc'}) {
        $opts{'read-realloc'} =~ s/-/_/;
    }

    if (exists $opts{'snapshot-policy'}) {
        if ($opts{'snapshot-policy'} eq 'none') {
            $apiset->vol_options(
                volume         => $opts{volume},
                'option-name'  => 'nosnap',
                'option-value' => 'on'
            );
        } else {
            $apiset->vol_options(
                volume         => $opts{volume},
                'option-name'  => 'nosnap',
                'option-value' => 'off'
            );
        } ## end else [ if ( $opts{'snapshot-policy'...})]
        delete $opts{'snapshot-policy'};
    } ## end if ( exists $opts{'snapshot-policy'...})

    # Translate size suffixes to the 7Mode equivalents
    $pkg_or_obj->_translate_size_suffixes(
        api_opts => \%opts,
        options_to_translate =>
            [qw(size max-autosize autosize-increment min-autosize)],
    );

    if (exists $opts{size}) {
        $apiset->vol_size(size => $opts{size}, volume => $volume);
    }

    # Copy and translate the option names
    my $volopts_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => _vol_options_copy(),
        map    => _vol_options_map(),
    );

    my @options_to_flip_then_translate
        = @{_options_to_flip(mode_of_options => 'SEVEN')};

    _flip_option_values(
        hashref        => $volopts_args,
        mode_of_values => 'CLUSTER',
        values_to_flip => \@options_to_flip_then_translate
    );

    _translate_options_values(
        hashref             => $volopts_args,
        direction           => 'CLUSTERTOSEVEN',
        values_to_translate => \@options_to_flip_then_translate
    );

    # Call 'vol options' to set each of the options provided
    while (my ($key, $value) = each %$volopts_args) {
        $apiset->vol_options(
            volume         => $volume,
            'option-name'  => $key,
            'option-value' => $value
        );
    } ## end while ( my ( $key, $value...))

    # Update the hashref to have on/off/reset to be set to 1 depending on the
    # value of the key specified through "variable_name".
    # $hashref->{$variable_name} = 'true' becomes $hashref->{on} = 1
    # $hashref->{$variable_name} = 'false' becomes $hashref->{off} = 1
    # $hashref->{$variable_name} = 'reset' becomes $hashref->{reset} = 1
    my $populate_hash_true_false = sub {
        my %opts = validate(
            @_,
            {   hashref       => {type => HASHREF},
                variable_name => {type => SCALAR},
            }
        );
        my $hash          = $opts{hashref};
        my $variable_name = $opts{variable_name};
        my $variable      = delete $hash->{$variable_name};

        if ($variable eq 'true') {
            $hash->{on} = 1;
        } elsif ($variable eq 'false') {
            $hash->{off} = 1;
        } else {
            NATE::Exceptions::Argument->throw(
                      "Value of '$variable_name' must be 'true' or 'false' "
                    . "but value provided was '$variable'");
        }
    };

    my %autosize_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%autosize_args,
        copy   => [qw(autosize shrink autosize-mode)],
        map    => {
            'max-autosize'                      => 'maximum-size',
            'autosize-increment'                => 'increment-size',
            'min-autosize'                      => 'minimum-size',
            'autosize-grow-threshold-percent'   => 'grow-threshold-percent',
            'autosize-shrink-threshold-percent' => 'shrink-threshold-percent',
        },
    );

    if (exists $autosize_args{autosize} && !$autosize_args{'autosize-mode'}) {
        $populate_hash_true_false->(
            hashref       => \%autosize_args,
            variable_name => 'autosize'
        );
    } ## end if ( exists $autosize_args...)

    if (keys %autosize_args) {
        $apiset->vol_autosize(volume => $volume, %autosize_args);
    }

    if (exists $opts{'snap-autodelete-enabled'}) {

        # Use _translate_options_values to convert true/false to on/off
        _translate_options_values(
            hashref             => \%opts,
            direction           => 'CLUSTERTOSEVEN',
            values_to_translate => [qw(snap-autodelete-enabled)],
        );
        $apiset->snap_autodelete(
            'vol-name' => $volume,
            action     => delete $opts{'snap-autodelete-enabled'}
        );
    } ## end if ( exists $opts{'snap-autodelete-enabled'...})

    my %snap_autodel_map;
    foreach my $snap_autodel_opt (@{_snap_autodelete_fields()}) {
        my $sevenmode_equivalent = $snap_autodel_opt;
        $sevenmode_equivalent =~ s/^snap\-autodelete\-//;
        $sevenmode_equivalent =~ s/-/_/g;
        $snap_autodel_map{$snap_autodel_opt} = $sevenmode_equivalent;
    } ## end foreach my $snap_autodel_opt...
    my %snap_autodelete_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%snap_autodelete_args,
        map    => \%snap_autodel_map,
    );

    # Call 'snap autodelete' to set each of the options provided
    # while ( my ( $key, $value ) = each %snap_autodelete_args ) {
    # $apiset->snap_autodelete(
    # volume         => $volume,
    # 'option-name'  => $key,
    # 'option-value' => $value
    # );
    # }
} ## end sub _modify_7mode_cli

sub _modify_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $volume            = $opts{volume};

    # This code must be run before running any of the other APIs
    # This is because, in CMode, if we attempt modifying the state along
    # with other fields, it attempts to change the state first before
    if (exists $opts{state}) {
        try {
            my $state = delete $opts{state};
            if ($state eq 'online') {
                $apiset->volume_online(name => $volume);
            } else {
                my %args_offline_restrict;
                $pkg_or_obj->_hash_copy(
                    source => \%opts,
                    map    => {'volume' => 'name'},
                    copy   => [qw(cifs-delay)],
                    target => \%args_offline_restrict
                );
                if ($state eq 'offline') {
                    $apiset->volume_offline(%args_offline_restrict);
                } elsif ($state eq 'restricted') {
                    $apiset->volume_restrict(%args_offline_restrict);
                } else {
                    NATE::Exceptions::Arguments->throw(
                              "'state' can be any of "
                            . "'online', 'offline' or 'restricted' "
                            . "but the value supplied was '$state'");
                } ## end else [ if ( $state eq 'offline')]
            } ## end else [ if ( $state eq 'online')]
        } ## end try
        catch NACL::APISet::Exceptions::ResponseException with {
            my $ex = shift;
            $Log->debug($ex->text());

# For pre-SN, CMode does not throw an exception, so 7Mode should behave the same way.
            _check_if_pre_SN($ex, $command_interface);

        };
    } ## end if ( exists $opts{state...})

    if (exists $opts{'read-realloc'}) {
        $opts{'read-realloc'} =~ s/-/_/;
    }

    # Translate size suffixes to the 7Mode equivalents
    $pkg_or_obj->_translate_size_suffixes(
        api_opts => \%opts,
        options_to_translate =>
            [qw(size max-autosize autosize-increment min-autosize)],
    );

    if (exists $opts{size}) {
        $apiset->volume_size(volume => $volume, 'new-size' => $opts{size});
    }

    # Copy and translate the option names
    my $volopts_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => _vol_options_copy(),
        map    => _vol_options_map(),
    );

    _translate_options_values(
        hashref   => $volopts_args,
        direction => 'CLUSTERTOSEVEN',
        values_to_translate =>
            _option_values_to_translate(mode_of_options => 'SEVEN')
    );

    my @options_to_flip_then_translate
        = @{_options_to_flip(mode_of_options => 'SEVEN')};

    _flip_option_values(
        hashref        => $volopts_args,
        mode_of_values => 'CLUSTER',
        values_to_flip => \@options_to_flip_then_translate
    );

    _translate_options_values(
        hashref             => $volopts_args,
        direction           => 'CLUSTERTOSEVEN',
        values_to_translate => \@options_to_flip_then_translate
    );

    # Call 'volume-set-option' to set each of the options provided
    while (my ($key, $value) = each %$volopts_args) {
        $apiset->volume_set_option(
            volume         => $volume,
            'option-name'  => $key,
            'option-value' => $value
        );
    } ## end while ( my ( $key, $value...))

    my %autosize_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%autosize_args,
        map    => {
            'max-autosize'                      => 'maximum-size',
            'autosize-increment'                => 'increment-size',
            'autosize'                          => 'is-enabled',
            'min-autosize'                      => 'minimum-size',
            'autosize-grow-threshold-percent'   => 'grow-threshold-percent',
            'autosize-shrink-threshold-percent' => 'shrink-threshold-percent',
            'autosize-mode'                     => 'mode'
        },
    );

    if (keys %autosize_args) {
        $apiset->volume_autosize_set(volume => $volume, %autosize_args);
    }

    if (exists $opts{files}) {
        $apiset->volume_set_total_files(
            volume                  => $volume,
            'requested-total-files' => $opts{files}
        );
    } ## end if ( exists $opts{files...})

    my %snap_autodel_map;
    foreach my $snap_autodel_opt (@{_snap_autodelete_fields()}) {
        my $sevenmode_equivalent = $snap_autodel_opt;
        $sevenmode_equivalent =~ s/^snap\-autodelete\-//;
        $sevenmode_equivalent =~ s/-/_/g;
        $snap_autodel_map{$snap_autodel_opt} = $sevenmode_equivalent;
    } ## end foreach my $snap_autodel_opt...
    my %snap_autodelete_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%snap_autodelete_args,
        map    => \%snap_autodel_map,
    );

    # Call 'snap autodelete' to set each of the options provided
    while (my ($key, $value) = each %snap_autodelete_args) {
        $apiset->snapshot_autodelete_set_option(
            volume         => $volume,
            'option-name'  => $key,
            'option-value' => $value
        );
    } ## end while ( my ( $key, $value...))
} ## end sub _modify_7mode_zapi

sub _modify_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    my $apiset = $opts{apiset};
    my $volume = $opts{volume} || '';

    $pkg_or_obj->_convert_multiple_to_bytes(
        api_opts     => \%opts,
        byte_options => [
            qw(upgrade-64bit-grow-space
                upgrade-64bit-percent-used-space autosize-increment min-autosize
                max-autosize maxdir-size compression-space-saveddedupe-space-saved
                sis-space-saved filesys-size-fixed overwrite-reserve
                overwrite-reserve-required overwrite-reserve-used
                overwrite-reserve-used-actual size available
                size-available-for-snapshots total used snapshot-reserve-size)
        ],
    );

    my $value;

    # Guaranteed to be a single value.
    # XXX Currently not handling the case of "vserver" not being provided:
    # which would be possible if extended_query is provided
    if (exists $opts{vserver}) {

        # We should invoke the async form of the ZAPI for repository vservers
        $value = $pkg_or_obj->_determine_is_repository(%opts);
    }

    my $api;
    if ($value) {
        $api = 'volume-modify-iter-async';
    } else {
        $api = 'volume-modify-iter';
    }

    my $common_map = $pkg_or_obj->_common_zapi_map();

    $pkg_or_obj->_iter_helper(
        %opts,
        api                  => $api,
        map                  => $common_map,
        does_not_exist_error => {
            sub => sub {
                my ($vol_name) = validate_pos(@_, {type => SCALAR});
                return "Volume '$vol_name' not found. Reason: entry "
                    . "doesn't exist";
            },
            args => [$volume]
        },
        does_not_exist_exception =>
            'NACL::C::Exceptions::Volume::DoesNotExist'
    );

    $Log->exit();
} ## end sub _modify_cmode_zapi

sub _check_if_pre_SN {
    my ($ex, $command_interface) = validate_pos(
        @_,
        {isa => 'NACL::APISet::Exceptions::ResponseException'},
        {isa => 'NACL::C::CommandInterface::ONTAP'}
    );

    if ($ex->text() !~ /is already/) {
        $Log->exit();
        $ex->throw();
    } else {
        my $codename = $command_interface->get_codename();
        if (   $codename =~ /Rollingrock_0/i
            || $codename =~ /Boilermaker/i)
        {

            # Suppress the exception
        } else {
            $Log->exit();
            $ex->throw();
        }
    } ## end else [ if ( $ex->text() !~ /is already/)]

} ## end sub _check_if_pre_SN

=head2 start_check

  NACL::C::Volume->start_check(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name,...);

                            or

  $volume_object->start_check(...);

(Class or instance method) This method starts checking a volume element.

Uses a CMode CLI, 7Mode CLI/ZAPI APIset.

Invokes "volume-verify-start" API for 7Mode ZAPI.

Invokes "vol verify start" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI

=item C<< foreground => $string >>

Applicable for CMode CLI

=item C<< repair => $string >>

Applicable for CMode CLI

=item C<< resume => $string >>

Applicable for CMode CLI

=item C<< inode => $string >>

Applicable for CMode CLI

=item C<< "plex-number" => $string >>

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : -f <value> (Programmatic name: "plex-number")

7M ZAPI: fix-plex

=item C<< "log-only" => $string >>

Applicable for 7Mode ZAPI.

=item C<< '*' >>

This method takes all other options supported by
"volume start-check" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub start_check {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            volume => {type => SCALAR},
            _optional_scalars qw( repair resume foreground inode plex-number
                log-only)
        },
        api => 'volume_start_check',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_start_check_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_start_check_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_start_check_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::Volume::start_check',
    );
    $Log->exit();
    return $ret;
} ## end sub start_check

sub _start_check_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume log-only)],
        map    => {"plex-number" => "fix-plex",}
    );
    %opts = %$vol_args;

    $apiset->volume_verify_start(%opts);

} ## end sub _start_check_7mode_zapi

sub _start_check_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->vol_verify_start(%opts);

} ## end sub _start_check_7mode_cli

sub _start_check_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_start_check',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'start_check',
    );
    $apiset->volume_start_check(%cmode_args);
    $Log->exit();
} ## end sub _start_check_cmode_cli

=head2 stop_check

  NACL::C::Volume->stop_check(command_interface=>$ci,volume =>$aggr_name,
                              vserver => $vserver_name);

                            or

  $volume_object->stop_check(...);

(Class or instance method) This method stops checking a volume element.

Uses a CMode CLI APISet, 7Mode CLI/ZAPI APISet.

Invokes "volume-verify-stop" API for 7Mode ZAPI.

Invokes "vol verify stop" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume stop-check" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub stop_check {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        api    => 'volume_stop_check',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_stop_check_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_stop_check_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_stop_check_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::Volume::stop_check',
    );
    $Log->exit();
    return $ret;
} ## end sub stop_check

sub _stop_check_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->vol_verify_stop(%opts);

} ## end sub _stop_check_7mode_cli

sub _stop_check_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->volume_verify_stop(%opts);

} ## end sub _stop_check_7mode_zapi

sub _stop_check_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_stop_check',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'stop_check',
    );

    $apiset->volume_stop_check(%cmode_args);
    $Log->exit();
} ## end sub _stop_check_cmode_cli

=head2 suspend_check

  NACL::C::Volume->suspend_check(command_interface=>$ci,volume=>$vol_name);

                            or

  $volume_object->suspend_check();

(Class or instance method) This method suspends checking a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-verify-suspend" API for 7Mode ZAPI.

Invokes "vol verify suspend" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub suspend_check {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {volume => {type => SCALAR},},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_suspend_check_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_suspend_check_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit();
    return $ret;
} ## end sub suspend_check

sub _suspend_check_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->vol_verify_suspend(%opts);

} ## end sub _suspend_check_7mode_cli

sub _suspend_check_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->volume_verify_suspend(%opts);

} ## end sub _suspend_check_7mode_zapi

=head2 resume_check

  NACL::C::Volume->resume_check(command_interface=>$ci,volume=>$vol_name);

                            or

  $volume_object->resume_check();

(Class or instance method) This method resumes checking a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-verify-resume" API for 7Mode ZAPI.

Invokes "vol verify resume" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub resume_check {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {volume => {type => SCALAR},},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_resume_check_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_resume_check_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit();
    return $ret;
} ## end sub resume_check

sub _resume_check_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->vol_verify_resume(%opts);

} ## end sub _resume_check_7mode_cli

sub _resume_check_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->volume_verify_resume(%opts);

} ## end sub _resume_check_7mode_zapi

=head2 offline

  NACL::C::Volume->offline(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name,...);

                            or

  $volume_object->offline(...);

(Class or instance method) This method offlines a volume element.

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APIset.

Invokes "volume-offline-async,volume-offline" API for CMode ZAPI.

Invokes "volume-offline" API for 7Mode ZAPI.

Invokes "volume offline" command for 7Mode CLI.

Same operation can be performed using modify method also.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: <option name>

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: name

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI

=item C<< "cifs-delay" => $string >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'force' => $string >>

Applicable for CMode CLI

=item C<< 'is-repository' => 0 | 1 >>

(Applicable only for CMode ZAPI)
This boolean option (i.e. takes values of 0/1) describes
whether the volume is on a repository vserver or not.

For CMode ZAPI, async ZAPIs must be run for repository vservers,
while the normal ZAPI must be run for normal vservers.
This option is optional, so if not provided then the
filer is queried to determine if the vserver is a repository vserver.

=item C<< '*' >>

This method takes all other options supported by
"volume offline" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub offline {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(qw(cifs-delay foreground force is-repository )),
        },
        api => 'volume_offline',
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_offline_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_offline_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_offline_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_offline_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::offline',
    );
    $Log->exit();
    return $ret;
} ## end sub offline

sub _offline_cmode_cli {
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $job_component     = delete $opts{job_component};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_offline',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'offline',
    );

    my $response = $apiset->volume_offline(%cmode_args);
    my $jobid    = $response->get_job_ids();
    if ($jobid) {
        $$job_component = NACL::C::Job->new(
            id                => $jobid,
            command_interface => $command_interface
        );
    } ## end if ($jobid)
    $Log->exit();
} ## end sub _offline_cmode_cli

sub _offline_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset   = delete $opts{apiset};
    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume cifs-delay)],
    );
    %opts = %$vol_args;

    $apiset->vol_offline(%opts);

} ## end sub _offline_7mode_cli

sub _offline_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset   = delete $opts{apiset};
    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(cifs-delay)],
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;

    $apiset->volume_offline(%opts);

} ## end sub _offline_7mode_zapi

sub _offline_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_repo_builder(
        @args,
        api => 'volume-offline',
        map => {volume => 'name'}
    );
    $Log->exit();
}

=head2 online

  NACL::C::Volume->online(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name,...);

                            or

  $volume_object->online(...);

(Class or instance method) This method onlines a volume element.

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APIset.

Invokes "volume-online, volume-online-async" API for CMode ZAPI.

Invokes "volume-online" API for 7Mode ZAPI.

Invokes "vol online" command for 7Mode CLI.

Same function can be performed using modify method also.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: name

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: name

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< force => $boolean >>

(Optional) A boolean when true forcefully onlines the volume element.

Applicable for 7Mode CLI.

Maps to:

7M CLI: -f (Programmatic name: "force")

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'is-repository' => 0 | 1 >>

(Applicable only for CMode ZAPI)
This boolean option (i.e. takes values of 0/1) describes
whether the volumes is on a repository vserver or not.

For CMode ZAPI, async ZAPIs must be run for repository vservers,
while the normal ZAPI must be run for normal vservers.
This option is optional, so if not provided then the
filer is queried to determine if the vserver is a repository vserver

=item C<< '*' >>

This method takes all other options supported by
"volume online" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub online {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( foreground is-repository
                    )
            ),
            volume => {type => SCALAR},
            force  => {type => BOOLEAN, optional => 1},
        },
        api => 'volume_online',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_online_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_online_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_online_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_online_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::online',
    );
    $Log->exit();
    return $ret;
} ## end sub online

sub _online_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_online',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'online',
    );

    $apiset->volume_online(%cmode_args);
    $Log->exit();

} ## end sub _online_cmode_cli

sub _online_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset   = delete $opts{apiset};
    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume force)],
    );
    %opts = %$vol_args;
    $apiset->vol_online(%opts);
} ## end sub _online_7mode_cli

sub _online_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset   = delete $opts{apiset};
    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;
    $apiset->volume_online(%opts);

} ## end sub _online_7mode_zapi

sub _online_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_repo_builder(
        @args,
        api => 'volume-online',
        map => {volume => 'name'}
    );
    $Log->exit();
}

=head2 rename

  NACL::C::Volume->rename(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name,...);

                            or

  $volume_object->rename(...);

(Class or instance method) This method renames a volume element.

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APIset.

Invokes "volume-rename,volume-rename-async" API for CMode ZAPI.

Invokes "volume-rename" API for CMode ZAPI.

Invokes "vol rename" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: volume

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< newname => $new_name >>

(Required) A scalar to which the volume element is to be renamed.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: new-volume-name

7M CLI : new volume name (Programmatic name: "new-volume-name")

7M ZAPI: new-volume-name

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'is-repository' => 0 | 1 >>

=item C<< 'is-repository' => $string >>

(Applicable only for CMode ZAPI)
This boolean option (i.e. takes values of 0/1) describes
whether the volumes is on a repository vserver or not.

For CMode ZAPI, async ZAPIs must be run for repository vservers,
while the normal ZAPI must be run for normal vservers.
This option is optional, so if not provided then the
filer is queried to determine if the vserver is a repository vserver

=item C<< '*' >>

This method takes all other options supported by
"volume rename" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( foreground is-repository
                    )
            ),
            newname       => {type => SCALAR},
            job_component => {type => SCALARREF, optional => 1},
        },
        api => 'volume_rename',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_rename_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_rename_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_rename_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_rename_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::rename',
    );

    $pkg_or_obj->volume($opts{newname}) if (ref $pkg_or_obj);
    $Log->exit();
    return $ret;
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $job_component     = delete $opts{job_component};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_rename',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'rename',
    );

    my $response = $apiset->volume_rename(%cmode_args);
    my $jobid    = $response->get_job_ids();
    if ($jobid) {
        $$job_component = NACL::C::Job->new(
            id                => $jobid,
            command_interface => $command_interface
        );
    } ## end if ($jobid)

    $Log->exit;
} ## end sub _rename_cmode_cli

sub _rename_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"  => "volume",
            "newname" => "new-volume-name",
        }
    );
    %opts = %$vol_args;

    $apiset->vol_rename(%opts);
    $pkg_or_obj->volume($opts{"new-volume-name"}) if ref($pkg_or_obj);

} ## end sub _rename_7mode_cli

sub _rename_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"  => "volume",
            "newname" => "new-volume-name",
        }
    );
    %opts = %$vol_args;

    $apiset->volume_rename(%opts);

} ## end sub _rename_7mode_zapi

sub _rename_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_repo_builder(
        @args,
        api => 'volume-rename',
        map => {
            volume  => 'volume',
            newname => 'new-volume-name',
        }
    );
    $Log->exit();
}

=head2 autosize

   NACL::C::Volume->autosize(
                                command_interface => $ci,
                                     volume => $volume,
                                     vserver => $vserver,
                                  ...);

                                                                 or

    $Volume_object->autosize(...);

(Class or instance method)
This method is used to set the autosize settings of the flexible volume..

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-autosize-set" API for CMode ZAPI.

Invokes "vol autosize" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "volume" => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: volume

7M CLI : volume name (Programmatic name: "volume")

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "increment-size" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: increment-size

7M CLI : -i <value> (Programmatic name: "increment-size")

=item C<< "maximum-size" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: maximum-size

7M CLI : -m <value> (Programmatic name: "maximum-size")

=item C<< "is-enabled" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-enabled

=item C<< 'shrink-threshold-percent' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: shrink-threshold-percent

7M CLI : -shrink-threshold-percent <value>
(Programmatic name: "shrink-threshold-percent")

=item C<< 'mode' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: mode

7M CLI : -mode <value> (Programmatic name: "mode")

=item C<< 'grow-threshold-percent' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: grow-threshold-percent

7M CLI : -grow-threshold-percent <value>
(Programmatic name: "grow-threshold-percent")

=item C<< 'minimum-size' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: minimum-size

7M CLI : -minimum-size <value> (Programmatic name: "minimum-size")

=item C<< 'reset' => $string >>

Autosize Reset

Applicable for CMode CLI, 7Mode CLI

Maps to:

7M CLI: reset (Programmatic name: reset)

=item C<< 'increment-percent' => $string >>

Increment Percentage

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume autosize" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub autosize {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( increment-size maximum-size is-enabled

                    shrink-threshold-percent mode grow-threshold-percent minimum-size

                    reset

                    increment-percent
                    )
            ),
        },
        api => 'volume_autosize',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_autosize_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_autosize_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_autosize_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::Volume::autosize',
    );

    $Log->exit();
    return $ret;
} ## end sub autosize

sub _autosize_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_autosize',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'autosize',
    );

    $apiset->volume_autosize(%cmode_args);
    $Log->exit();
} ## end sub _autosize_cmode_cli

sub _autosize_cmode_zapi {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vserver           = delete $opts{vserver};

    if (defined $opts{'is-enabled'}) {
        $opts{'is-enabled'} =~ s/on/True/i;
        $opts{'is-enabled'} =~ s/off/False/i;
    }

    my $allowed_cmode_options = [
        qw( volume increment-size maximum-size is-enabled
            shrink-threshold-percent mode grow-threshold-percent minimum-size
            )
    ];
    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_cmode_options,
        target => \%cmode_args,
    );
    %opts = %cmode_args;

    $pkg_or_obj->_handle_zapi_vserver_context(
        api_opts          => \%opts,
        vserver           => $vserver,
        command_interface => $command_interface
    );

    $apiset->volume_autosize_set(%opts);
    $Log->exit();
} ## end sub _autosize_cmode_zapi

sub _autosize_7mode_cli {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [
        qw( volume increment-size maximum-size
            minimum-size shrink-threshold-percent grow-threshold-percent
            )
    ];

    $pkg_or_obj->_translate_size_suffixes(
        options_to_translate =>
            [qw(maximum-size increment-size minimum-size)],
        api_opts => \%opts,
    );

    my %mode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        map    => {'mode' => 'autosize-mode'},
        target => \%mode_args,
    );

    if ($opts{'is-enabled'} && !$opts{'mode'}) {
        $mode_args{'on'}    = 1 if ($opts{'is-enabled'} eq "on");
        $mode_args{'off'}   = 1 if ($opts{'is-enabled'} eq "off");
        $mode_args{'reset'} = 1 if ($opts{'is-enabled'} eq "reset");
    }

    if ($opts{reset} && $opts{reset} eq "true") {
        $mode_args{'reset'} = 1;
    }
    $apiset->vol_autosize(%mode_args);
    $Log->exit();
} ## end sub _autosize_7mode_cli

=head2 scrub_start

  NACL::C::Volume->scrub_start(command_interface=>$ci,
                             volume=>$vol_name);

                            or

  $volume_object->scrub_start();

(Class or instance method) This method starts scrubbing a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-scrub-start" API for 7Mode ZAPI.

Invokes "vol scrub start" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : volume name <value> (Programmatic name: "volume")

7M ZAPI: name

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scrub_start {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_scrub_start_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_scrub_start_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit();
    return $ret;
} ## end sub scrub_start

sub _scrub_start_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],

    );
    %opts = %$vol_args;

    $apiset->vol_scrub_start(%opts);

} ## end sub _scrub_start_7mode_cli

sub _scrub_start_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;

    $apiset->volume_scrub_start(%opts);

} ## end sub _scrub_start_7mode_zapi

=head2 scrub_stop

  NACL::C::Volume->scrub_stop(command_interface=>$ci,
                             volume=>$vol_name);

                            or

  $volume_object->scrub_stop();

(Class or instance method) This method stops scrubbing a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-scrub-stop" API for 7Mode ZAPI.

Invokes "vol scrub stop" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : volume name <value> (Programmatic name: "volume")

7M ZAPI: name

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scrub_stop {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {volume => {type => SCALAR},},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_scrub_stop_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_scrub_stop_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit();
    return $ret;
} ## end sub scrub_stop

sub _scrub_stop_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],

    );
    %opts = %$vol_args;

    $apiset->vol_scrub_stop(%opts);

} ## end sub _scrub_stop_7mode_cli

sub _scrub_stop_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;

    $apiset->volume_scrub_stop(%opts);

} ## end sub _scrub_stop_7mode_zapi

=head2 scrub_resume

  NACL::C::Volume->scrub_resume(command_interface=>$ci,
                             volume=>$vol_name);

                            or

  $volume_object->scrub_resume();

(Class or instance method) This method resumes scrubbing a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-scrub-resume" API for 7Mode ZAPI.

Invokes "vol scrub resume" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : volume name <value> (Programmatic name: "volume")

7M ZAPI: name

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scrub_resume {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {volume => {type => SCALAR},},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_scrub_resume_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_scrub_resume_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit();
    return $ret;
} ## end sub scrub_resume

sub _scrub_resume_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],

    );
    %opts = %$vol_args;

    $apiset->vol_scrub_resume(%opts);

} ## end sub _scrub_resume_7mode_cli

sub _scrub_resume_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;

    $apiset->volume_scrub_resume(%opts);

} ## end sub _scrub_resume_7mode_zapi

=head2 scrub_suspend

  NACL::C::Volume->scrub_suspend(command_interface=>$ci,
                             volume=>$vol_name);

                            or

  $volume_object->scrub_suspend();

(Class or instance method) This method suspends scrubbing a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-scrub-suspend" API for 7Mode ZAPI.

Invokes "vol scrub suspend" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : volume name <value> (Programmatic name: "volume")

7M ZAPI: name

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scrub_suspend {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {volume => {type => SCALAR},},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_scrub_suspend_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_scrub_suspend_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit();
    return $ret;
} ## end sub scrub_suspend

sub _scrub_suspend_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],

    );
    %opts = %$vol_args;

    $apiset->vol_scrub_suspend(%opts);

} ## end sub _scrub_suspend_7mode_cli

sub _scrub_suspend_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;

    $apiset->volume_scrub_suspend(%opts);

} ## end sub _scrub_suspend_7mode_zapi

=head2 restrict

  NACL::C::Volume->restrict(command_interface=>$ci,
                             volume=>$vol_name);

                            or

  $volume_object->restrict();

(Class or instance method) This method restricts a volume element.

Uses a 7Mode CLI/ZAPI, CMode CLI/ZAPI APISet.

Invokes "volume-restrict, volume-restrict-async" API for CMode ZAPI.

Invokes "volume-restrict" API for 7Mode ZAPI.

Invokes "vol restrict" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: name

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: name

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'is-repository' => 0 | 1 >>

(Applicable only for CMode ZAPi)
This boolean option (i.e. takes values of 0/1) describes
whether the volumes is on a repository vserver or not.

For CMode ZAPI, async ZAPIs must be run for repository vservers,
while the normal ZAPI must be run for normal vservers.
This option is optional, so if not provided then the
filer is queried to determine if the vserver is a repository vserver

=item C<< '*' >>

This method takes all other options supported by
"volume restrict" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Volume::AlreadyRestricted>

This type of exception gets thrown if an attempt is made to restrict a
volume which is already restricted.

=back

=back

=cut

sub restrict {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(cifs-delay
                foreground is-repository
                ),
        },
        api => 'volume_restrict',
    );
    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_restrict_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_restrict_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
                {   method    => '_restrict_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_restrict_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::Volume::restrict',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Volume::AlreadyRestricted
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit();
        $exception->throw();
    };
    $Log->exit();
    return $ret;
} ## end sub restrict

sub _restrict_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume cifs-delay)],

    );
    %opts = %$vol_args;

    $apiset->vol_restrict(%opts);

} ## end sub _restrict_7mode_cli

sub _restrict_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"     => "name",
            "cifs-delay" => "cifs-delay",
        }
    );
    %opts = %$vol_args;

    $apiset->volume_restrict(%opts);

} ## end sub _restrict_7mode_zapi

sub _restrict_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_repo_builder(
        @args,
        api => 'volume-restrict',
        map => {volume => 'name'}
    );
    $Log->exit();
} ## end sub _delete_cmode_zapi

sub _restrict_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_restrict',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'restrict',
    );

    $apiset->volume_restrict(%cmode_args);
    $Log->exit();
} ## end sub _restrict_cmode_cli

=head2 move

  NACL::C::Volume->move(command_interface=>$ci,
                             volume=>$vol_name ...);

                            or

  $volume_object->move();

(Class or instance method) This method moves a volume element to another
aggregate.

Uses a CMode CLI APIset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI

=item C<< "destination-aggregate" => $string >>

Applicable for CMode CLI

=item C<< "destination-volume" => $string >>

Applicable for CMode CLI

=item C<< "destination-stripe-format" => $string >>

Applicable for CMode CLI

=item C<< "cutover-window" => $string >>

Applicable for CMode CLI

=item C<< "cutover-attempts" => $string >>

Applicable for CMode CLI

=item C<< "foreground" => $string >>

Applicable for CMode CLI

=item C<< "cutover-action" => $string >>

Applicable for CMode CLI

=item C<< "perform-validation-only" => $string >>

Applicable for CMode CLI

=item C<< "extended_query" => $string  >>

(Optional)
L<< NACL::C::VolumeMove->start|lib-NACL-C-VolumeMove-pm/start >>
If invoked with an STask package or object, 
then see the implementation and documentation of provided functionality is present in 
L<< NACL::STask::VolumeMove->start|lib-NACL-STask-VolumeMove-pm/start >>

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub move {
    $Log->enter();
    $Log->warn(
        "volume move should not be used ,volume move start should be used");
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params      => \@args,
        allow_extra => 1
    );
    my $is_c_or_t = $pkg_or_obj->is_c_or_t();
    my $move_pkg  = "NACL::${is_c_or_t}::VolumeMove";

    my $ret = $move_pkg->start(%opts);

    $Log->exit();
    return $ret;
}

=head2 make_vsroot

    NACL::C::Volume->make_vsroot(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->make_vsroot(...);

(Class or instance method)
This method is used to designate a non-root volume as a root volume of the Vserver.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume make-vsroot" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub make_vsroot {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        api    => 'volume_make_vsroot',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_make_vsroot_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::make_vsroot',
    );

    $Log->exit();
} ## end sub make_vsroot

sub _make_vsroot_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_make_vsroot',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'make_vsroot',
    );

    $apiset->volume_make_vsroot(%cmode_args);
    $Log->exit();
} ## end sub _make_vsroot_cmode_cli

=head2 add_members

    NACL::C::Volume->add_members(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->add_members(...);

(Class or instance method)
This method is used to add members to an existing striped volume.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'count' => $string >>

Applicable for CMode CLI

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume add-members" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_members {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( count foreground
                    )
            ),
        },
        api => 'volume_add_members',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_members_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::add_members',
    );

    $Log->exit();
} ## end sub add_members

sub _add_members_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $job_component     = delete $opts{job_component};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_add_members',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $command_interface,
        frontend_method   => 'add_members',
    );

    my $response = $apiset->volume_add_members(%cmode_args);
    my $jobid    = $response->get_job_ids();
    if ($jobid) {
        $$job_component = NACL::C::Job->new(
            id                => $jobid,
            command_interface => $command_interface
        );
    } ## end if ($jobid)
    $Log->exit();
} ## end sub _add_members_cmode_cli

=head2 add_other_volumes

    NACL::C::Volume->add_other_volumes(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to add 7-mode volumes to the cluster database.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume add-other-volumes" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_other_volumes {
    $Log->enter();
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( node
                    )
            ),
        },
        ignore_primary_keys => 1,
        api                 => 'volume_add_other_volumes',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_other_volumes_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::add_other_volumes',
    );

    $Log->exit();
} ## end sub add_other_volumes

sub _add_other_volumes_cmode_cli {
    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg->get_api_options(
        apiset  => $apiset,
        api     => 'volume_add_other_volumes',
        options => \%opts,
    );
    $pkg->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'add_other_volumes',
    );

    $apiset->volume_add_other_volumes(%cmode_args);
    $Log->exit();
} ## end sub _add_other_volumes_cmode_cli

=head2 transition

    NACL::C::Volume->transition(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->transition(...);

(Class or instance method)
This method is used to transition a volume from 7-Mode to Cluster-Mode.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'affinity-node' => $string >>

Applicable for CMode CLI

=item C<< 'check-only' => $string >>

Applicable for CMode CLI

=item C<< 'junction-path' => $string >>

Applicable for CMode CLI

=item C<< 'source-node' => $string >>

Applicable for CMode CLI

=item C<< 'nondisruptive' => $string >>

Applicable for CMode CLI

=item C<< 'suppress-warnings' => $string >>

Applicable for CMode CLI

=item C<< 'operation-type' => $string >>

Applicable for CMode CLI

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'allow-7mode' => $string >>

Applicable for CMode CLI

=item C<< 'destination-vserver' => $string >>

*Destination vserver for a 7-mode transition

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume transition" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub transition {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( affinity-node check-only junction-path source-node nondisruptive
                    suppress-warnings operation-type foreground

                    allow-7mode

                    destination-vserver
                    )
            ),
        },
        api => 'volume_transition',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_transition_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::transition',
    );

    $Log->exit();
} ## end sub transition

sub _transition_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_transition',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'transition',
    );

    $apiset->volume_transition(%cmode_args);
    $Log->exit();
} ## end sub _transition_cmode_cli

=head2 size

    NACL::C::Volume->size(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->size(...);

(Class or instance method)
This method is used to set/Display the size of flexible volume..

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-size-async, volume-size" API for CMode ZAPI.

Invokes "vol size" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: volume-name

7M CLI : volume name (Programmatic name: "volume-name")

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< 'vol-size' => $string >>

Applicable for CMode CLI

=item C<< 'is-replica' => $string >>

Applicable for CMode CLI

=item C<< 'is-rdonly' => $string >>

Applicable for CMode CLI

=item C<< 'new-size' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: new-size

7M CLI : volume size (Programmatic name: "size")

=item C<< 'is-fixed' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'is-repository' => 0 | 1 >>

(Applicable only for CMode ZAPI)
This boolean option (i.e. takes values of 0/1) describes
whether the volumes is on a repository vserver or not.

For CMode ZAPI, async ZAPIs must be run for repository vservers,
while the normal ZAPI must be run for normal vservers.
This option is optional, so if not provided then the
filer is queried to determine if the vserver is a repository vserver

=item C<< '*' >>

This method takes all other options supported by
"volume size" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub size {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vol-size is-replica is-rdonly new-size is-fixed is-repository
                    )
            ),
        },
        api => 'volume_size',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_size_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_size_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_size_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::size',
    );

    $Log->exit();
} ## end sub size

sub _size_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_size',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'size',
    );

    $apiset->volume_size(%cmode_args);
    $Log->exit();
} ## end sub _size_cmode_cli

sub _size_7mode_cli {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_mode_options = [qw(volume)];
    my %mode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_mode_options,
        target => \%mode_args,
        map    => {'new-size' => 'size',},
    );

    $apiset->vol_size(%mode_args);
    $Log->exit();
} ## end sub _size_7mode_cli

sub _size_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_repo_builder(
        @args,
        api  => 'volume-size',
        map  => {volume => 'volume'},
        copy => [qw(new-size)],
    );
    $Log->exit();
}

=head2 remove_other_volume

    NACL::C::Volume->remove_other_volume(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->remove_other_volume(...);

(Class or instance method)
This method is used to remove 7-mode volume from the cluster database.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume remove-other-volume" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_other_volume {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        api    => 'volume_remove_other_volume',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_other_volume_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::remove_other_volume',
    );

    $Log->exit();
} ## end sub remove_other_volume

sub _remove_other_volume_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_remove_other_volume',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'remove_other_volume',
    );

    $apiset->volume_remove_other_volume(%cmode_args);
    $Log->exit();
} ## end sub _remove_other_volume_cmode_cli

=head2 remove_junction

    NACL::C::Volume->remove_junction(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

(Class method)
This method is used to remove a dead junction by junction path.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required)

Applicable for CMode CLI

=item C<< 'junction-path' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume remove-junction" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_junction {
    $Log->enter();
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vserver junction-path
                    )
            ),
        },
        ignore_primary_keys => 1,
        api                 => 'volume_remove_junction',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_junction_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::remove_junction',
    );

    $Log->exit();
} ## end sub remove_junction

sub _remove_junction_cmode_cli {
    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg->get_api_options(
        apiset  => $apiset,
        api     => 'volume_remove_junction',
        options => \%opts,
    );
    $pkg->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'remove_junction',
    );

    $apiset->volume_remove_junction(%cmode_args);
    $Log->exit();
} ## end sub _remove_junction_cmode_cli

=head2 destroy

    NACL::C::Volume->destroy(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->destroy(...);

(Class or instance method)
This method is used to destroy an existing volume.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'force' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'disable-offline-check' => $string >>

=item C<< '*' >>

This method takes all other options supported by
"volume destroy" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub destroy {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( force

                    disable-offline-check
                    )
            ),
        },
        api => 'volume_destroy',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_destroy_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::destroy',
    );

    $Log->exit();
} ## end sub destroy

sub _destroy_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_destroy',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'destroy',
    );

    $apiset->volume_destroy(%cmode_args);
    $Log->exit();
} ## end sub _destroy_cmode_cli

=head2 transition_recover

    NACL::C::Volume->transition_recover(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->transition_recover(...);

(Class or instance method)
This method is used to initiates recovery for a volume which is in an incompletely transitioned state.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'affinity-node' => $string >>

Applicable for CMode CLI

=item C<< 'action' => $string >>

Applicable for CMode CLI

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'source-node' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume transition-recover" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub transition_recover {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( affinity-node action foreground source-node
                    )
            ),
        },
        api => 'volume_transition_recover',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_transition_recover_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::transition_recover',
    );

    $Log->exit();
} ## end sub transition_recover

sub _transition_recover_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_transition_recover',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'transition_recover',
    );

    $apiset->volume_transition_recover(%cmode_args);
    $Log->exit();
} ## end sub _transition_recover_cmode_cli

=head2 explore

    my $vol_explore_obj = NACL::C::Volume->explore(
                                command_interface => $ci,
                                format => 'volume',
                                ...);

                or

    my @vol_explore_objs = NACL::C::Volume->explore(
                                command_interface => $ci,
                                format => 'volume',
                                ...);

(Class method)
This method is used to view meta-data within volumes.

Uses a CMode CLI APISet.

In an array context it creates a Component State object of the type
L<NACL::CS::VolumeExplore|lib-NACL-CS-VolumeExplore-pm>. for each
present element and returns them.
In scalar context it creates a Component State object of the type
L<NACL::CS::VolumeExplore|lib-NACL-CS-VolumeExplore-pm>. for each
present present but returns only the first element.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'verbose' => $string >>

Applicable for CMode CLI

=item C<< 'dump' => $string >>

Applicable for CMode CLI

=item C<< 'find' => $string >>

Applicable for CMode CLI

=item C<< 'noinomap' => $string >>

Applicable for CMode CLI

=item C<< 'export-file' => $string >>

Applicable for CMode CLI

=item C<< 'print' => $string >>

Applicable for CMode CLI

=item C<< 'rawtimes' => $string >>

Applicable for CMode CLI

=item C<< 'recursive' => $string >>

Applicable for CMode CLI

=item C<< 'fill' => $string >>

Applicable for CMode CLI

=item C<< 'format' => $string >>

Applicable for CMode CLI

=item C<< 'import-file' => $string >>

Applicable for CMode CLI

=item C<< 'set' => $string >>

Applicable for CMode CLI

=item C<< 'scope' => $string >>

Applicable for CMode CLI

=item C<< 'list' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume explore" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub explore {
    $Log->enter();
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( verbose dump find noinomap export-file print rawtimes recursive fill format
                    import-file set scope list
                    )
            ),
        },
        ignore_primary_keys => 1,
        api                 => 'volume_explore',
    );
    my @objs = $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_explore_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::explore',
    );

    $Log->exit();
    return wantarray ? @objs : $objs[0];
} ## end sub explore

sub _explore_cmode_cli {
    $Log->enter();
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params      => \@args,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );

    my $ci     = delete $opts{command_interface};
    my $apiset = $opts{apiset};

    my %cmode_args = $pkg->get_api_options(
        apiset  => $apiset,
        api     => 'volume_explore',
        options => \%opts,
    );

    my $response = $apiset->volume_explore(%cmode_args);
    my $output   = $response->get_parsed_output();

    my @state_objs;
    foreach my $row (@$output) {
        my $obj = NACL::CS::VolumeExplore->new(command_interface => $ci);
        $obj->_set_fields(row => $row);
        push(@state_objs, $obj);
    }

    $Log->exit();
    return @state_objs;
} ## end sub _explore_cmode_cli

=head2 transition_protect

    NACL::C::Volume->transition_protect(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->transition_protect(...);

(Class or instance method)
This method is used to fence access to volume during Transition..

Uses a CMode CLI/ZAPI APISet.

Invokes 'volume-transition-protect' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: vserver

=item C<< 'is-enabled' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-enabled

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume transition-protect" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub transition_protect {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( is-enabled
                    )
            ),
        },
        api => 'volume_transition_protect',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_transition_protect_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_transition_protect_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::transition_protect',
    );

    $Log->exit();
} ## end sub transition_protect

sub _transition_protect_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_transition_protect',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'transition_protect',
    );

    $apiset->volume_transition_protect(%cmode_args);
    $Log->exit();
} ## end sub _transition_protect_cmode_cli

sub _transition_protect_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    my $apiset = $opts{apiset};

    my $allowed_cmode_options = [qw( volume vserver is-enabled )];

    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_cmode_options,
        target => \%cmode_args,
    );

    my $convert_opts = {'on' => 'true', 'off' => 'false'};

    if (defined $convert_opts->{$cmode_args{'is-enabled'}}) {
        $cmode_args{'is-enabled'}
            = $convert_opts->{$cmode_args{'is-enabled'}};
    }

    $apiset->volume_transition_protect(%cmode_args);
    $Log->exit();
} ## end sub _transition_protect_cmode_zapi

=head2 set_msid

    NACL::C::Volume->set_msid(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->set_msid(...);

(Class or instance method)
This method is used to set MSID of a volume.

Uses a CMode CLI/ZAPI APISet.

Invokes "volume-set-msid" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

Applicable for CMode CLI/ZAPI

Maps to :

CM ZAPI : name

(Required for class method, Not Applicable for instance method)

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< 'new-msid' => $string >>

Applicable for CMode CLI/ZAPI

Maps to :

CM ZAPI : new-msid

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume set-msid" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub set_msid {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( new-msid
                    )
            ),
        },
        api => 'volume_set_msid',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_set_msid_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_set_msid_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::set_msid',
    );

    $Log->exit();
} ## end sub set_msid

sub _set_msid_cmode_cli {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    my $apiset = delete $opts{apiset};

    my %cmode_args = $pkg_or_obj->get_api_options(
        apiset  => $apiset,
        api     => 'volume_set_msid',
        options => \%opts,
    );
    $pkg_or_obj->_handle_cli_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $cmode_args{vserver},
        command_interface => $opts{command_interface},
        frontend_method   => 'set_msid',
    );

    $apiset->volume_set_msid(%cmode_args);
    $Log->exit();
} ## end sub _set_msid_cmode_cli

sub _set_msid_cmode_zapi {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    my $apiset  = $opts{apiset};
    my $vserver = $opts{vserver};

    my $allowed_cmode_options = [qw( new-msid )];
    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_cmode_options,
        map    => {'volume' => 'name',},
        target => \%cmode_args,
    );
    $pkg_or_obj->_handle_zapi_vserver_context(
        api_opts          => \%cmode_args,
        vserver           => $vserver,
        command_interface => $opts{command_interface},
    );

    $apiset->volume_set_msid(%cmode_args,);
    $Log->exit();
} ## end sub _set_msid_cmode_zapi

=head2 show_space

    my $show_space_obj = NACL::C::Volume->show_space(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_space_objs = NACL::C::Volume->show_space(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_space_obj = $Volume_obj->show_space();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeShowSpace-pm/ATTRIBUTES>
of NACL::CS::VolumeShowSpace as input.
This method returns objects of the type L<NACL::CS::VolumeShowSpace|lib-NACL-CS-VolumeShowSpace-pm>.

=cut

sub show_space {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::VolumeShowSpace;
    my @state_objs = NACL::CS::VolumeShowSpace->fetch(%opts);
    $Log->exit();
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_space

=head2 show_footprint

    my $show_footprint_obj = NACL::C::Volume->show_footprint(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_footprint_objs = NACL::C::Volume->show_footprint(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_footprint_obj = $Volume_obj->show_footprint();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeShowFootprint-pm/ATTRIBUTES>
of NACL::CS::VolumeShowFootprint as input.
This method returns objects of the type L<NACL::CS::VolumeShowFootprint|lib-NACL-CS-VolumeShowFootprint-pm>.

=cut

sub show_footprint {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::VolumeShowFootprint;
    my @state_objs = NACL::CS::VolumeShowFootprint->fetch(%opts);
    $Log->exit();
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_footprint

=head2 show_space_old

    my $show_space_old_obj = NACL::C::Volume->show_space_old(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_space_old_objs = NACL::C::Volume->show_space_old(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_space_old_obj = $Volume_obj->show_space_old();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeShowSpaceOld-pm/ATTRIBUTES>
of NACL::CS::VolumeShowSpaceOld as input.
This method returns objects of the type L<NACL::CS::VolumeShowSpaceOld|lib-NACL-CS-VolumeShowSpaceOld-pm>.

=cut

sub show_space_old {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::VolumeShowSpaceOld;
    my @state_objs = NACL::CS::VolumeShowSpaceOld->fetch(%opts);
    $Log->exit();
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_space_old

=head2 show_footprint_old

    my $show_footprint_old_obj = NACL::C::Volume->show_footprint_old(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_footprint_old_objs = NACL::C::Volume->show_footprint_old(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_footprint_old_obj = $Volume_obj->show_footprint_old();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeShowFootprintOld-pm/ATTRIBUTES>
of NACL::CS::VolumeShowFootprintOld as input.
This method returns objects of the type L<NACL::CS::VolumeShowFootprintOld|lib-NACL-CS-VolumeShowFootprintOld-pm>.

=cut

sub show_footprint_old {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::VolumeShowFootprintOld;
    my @state_objs = NACL::CS::VolumeShowFootprintOld->fetch(%opts);
    $Log->exit();
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_footprint_old

=head2 cluster_max_volume

    NACL::C::Volume->cluster_max_volume(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to get or set the maximum number of volumes allowed in the cluster.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'cluster-max-volume' => $string >>

Maximum number of volumes allowed in the cluster possible value(s) are, integer

Applicable for CMode CLI

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume cluster-max-volume" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub cluster_max_volume {
    $Log->enter();
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'cmd_execution_time' => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( cluster-max-volume
                    )
            ),
        },
        ignore_primary_keys => 1,
        api                 => 'volume_cluster_max_volume',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cluster_max_volume_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::cluster_max_volume',
    );

    $Log->exit();
} ## end sub cluster_max_volume

sub _cluster_max_volume_cmode_cli {
    $Log->enter();
    my ($pkg, @opts) = @_;
    my %opts = $pkg->_validate_backend_opts(@opts);

    my $apiset         = $opts{apiset};
    my $execution_time = $opts{cmd_execution_time};
    my %cmode_args     = $pkg->get_api_options(
        apiset  => $apiset,
        api     => 'volume_cluster_max_volume',
        options => \%opts,
    );

    my $response = $apiset->volume_cluster_max_volume(%cmode_args);
    if (defined $execution_time) {
        $$execution_time = $response->get_execution_time();
    }

    $Log->exit();
} ## end sub _cluster_max_volume_cmode_cli

=head2 language

    my $language_obj = NACL::C::Volume->language(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @language_objs = NACL::C::Volume->language(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $language_obj = $Volume_obj->language();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeLanguage-pm/ATTRIBUTES>
of NACL::CS::VolumeLanguage as input.
This method returns objects of the type L<NACL::CS::VolumeLanguage|lib-NACL-CS-VolumeLanguage-pm>.

=cut

sub language {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeLanguage;
    my @state_objs = NACL::CS::VolumeLanguage->fetch(%opts);
    $Log->exit();
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 show_df

    my $show_df_obj = NACL::C::Volume->show_df(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_df_objs = NACL::C::Volume->show_df(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_df_obj = $Volume_obj->show_df();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeShowDf-pm/ATTRIBUTES>
of NACL::CS::VolumeShowDf as input.
This method returns objects of the type L<NACL::CS::VolumeShowDf|lib-NACL-CS-VolumeShowDf-pm>.

=cut

sub show_df {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeShowDf;
    my @state_objs = NACL::CS::VolumeShowDf->fetch(%opts);
    $Log->exit();
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 status

    my $status_obj = NACL::C::Volume->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @status_objs = NACL::C::Volume->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $status_obj = $Volume_obj->status();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeStatus-pm/ATTRIBUTES>
of NACL::CS::VolumeStatus as input.
This method returns objects of the type L<NACL::CS::VolumeStatus|lib-NACL-CS-VolumeStatus-pm>.

=cut

sub status {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeStatus;
    my @state_objs = NACL::CS::VolumeStatus->fetch(%opts);
    $Log->exit();
    return wantarray ? @state_objs : $state_objs[0];
}

sub _primary_keys_validate_spec {
    return (
        volume  => {type => SCALAR},
        vserver => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

=head2 get_aggregate

  my $aggr_object = $volume_object->get_aggregate();

(Instance method only) This method returns a STask/Component object
corresponding to the parent aggregate for the volume modeled by this object.
If the volume object is a Component object, then object of type
NACL::C::StorageAggregate will be returned. If the volume object is a STask
object, then object of type NACL::STask::Aggregate will be returned.

=cut

sub get_aggregate {
    my $self = shift;
    $Log->enter();

    $self->_verify_invocation(style => 'instance_only');

    # Find whom the caller is - STask/C??.
    my $c_or_t = $self->is_c_or_t();

    # Get aggregate object
    my $aggr_obj = $self->build_associated_object(
        type => "NACL::${c_or_t}::Aggregate");
    $Log->exit();
    return $aggr_obj;
} ## end sub get_aggregate

=head2 get_vserver

  my $vserver_object = $volume_object->get_vserver();

(Instance method only) This method returns a Component object
corresponding to the vserver to which the volume modeled by this object belongs
to. It returns an object of type NACL::C::Vserver irrespective of whether
this volume object is a STask or Component object. When a STask for Vserver
is added, this method will return a Vserver STask object for the callers
from STask layer.

=cut

sub get_vserver {
    my $self = shift;
    $Log->enter();

    $self->_verify_invocation(style => 'instance_only');

    if ($self->command_interface()->mode() eq 'CMode') {

        my $c_or_t = $self->is_c_or_t();
        my $type   = "NACL::${c_or_t}::Vserver";

        # Get vserver object
        my $vserver_obj = $self->emerge_contained_object(type => $type);
        $Log->exit();
        return $vserver_obj;
    } else {

        # Do nothing for 7Mode
        $Log->exit();
        return;
    } ## end else [ if ( $self->command_interface...)]
} ## end sub get_vserver

=head2 is_hybrid

    $ret = $volume->is_hybrid();

(Instance method only) This method checks if the volume lives on a
hybrid aggregate or not.  Will return 1 if its hybrid and 0 otherwise.

=cut

sub is_hybrid {
    $Log->enter();
    my $self = shift;

    # Get aggregate containing the volume
    my $aggr = $self->get_aggregate();

    # Check if container aggregate is hybrid
    my $ret = $aggr->is_hybrid();
    $Log->exit();
    return $ret;
} ## end sub is_hybrid

# Implementation for the following find/state methods are from
# NACL::C::_Mixins::VolumeFileDirectory

=head3 find_files

    # List context
    my @file_objs = $vol_obj->find_files();

    # Scalar context
    my $file_obj = $vol_obj->find_files();

    # Accepts the same options as normal "find" calls
    my $file_obj = $vol_obj->find_files(allow_empty => 1);

    my $file_obj = $vol_obj->find_files(filter => { 'file-size' => '>1MB' });

Finds files present in the volume and returns objects of type
L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm>. Accepts arguments accepted
by "find".

=head3 find_directories

    # List context
    my @dir_objs = $vol_obj->find_directories();

    # Scalar context
    my $dir_obj = $vol_obj->find_directories();

    # Accepts the same options as normal "find" calls
    my $dir_obj = $vol_obj->find_directories(allow_empty => 1);

    my $dir_obj = $vol_obj->find_directories(filter => { 'owner-id' => $id });

Finds directories present in the volume and returns objects of type
L<NACL::C::VolumeDirectory|lib-NACL-C-VolumeDirectory-pm>. Accepts arguments accepted
by "find".

=head3 find_files_and_directories

    # List context
    my @objs = $vol_obj->find_files_and_directories();

    # Scalar context
    my $obj = $vol_obj->find_files_and_directories();

    # Accepts the same options as normal "find" calls
    my $obj = $vol_obj->find_files_and_directories(allow_empty => 1);

    my $obj = $vol_obj->find_files_and_directories(filter => { 'acl-type' => $acl });

Finds files and directories present in the volume and returns objects of type
L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> for files found and objects
of type 
L<NACL::C::VolumeDirectory|lib-NACL-C-VolumeDirectory-pm> for directories found.
Accepts arguments accepted by "find".

=head3 state_files

    # List context
    my @files_cs = $vol_obj->state_files();

    # Scalar context
    my $file_cs = $vol_obj->state_files();

    # Accepts arguments accepted by "fetch"
    my @files_cs = $vol_obj->state_files(
        requested_fields => [ qw(acl-type) ],
        filter           => { 'file-size' => '>1MB' },
        allow_empty      => 1
    );

Fetches the state of contained files and returns objects of type
L<NACL::CS::VolumeListFilesInDirectory|lib-NACL-CS-VolumeListFilesInDirectory-pm>.
Accepts the arguments accepted by the "fetch" method.

=head3 state_directories

    # List context
    my @dirs_cs = $vol_obj->state_directories();

    # Scalar context
    my $dir_cs = $vol_obj->state_directories();

    # Accepts arguments accepted by "fetch"
    my @dirs_cs = $vol_obj->state_directories(
        requested_fields => [ qw(acl-type) ],
        filter           => { 'inode-number' => num },
        allow_empty      => 1
    );

Fetches the state of contained directories and returns objects of type
L<NACL::CS::VolumeListFilesInDirectory|lib-NACL-CS-VolumeListFilesInDirectory-pm>.
Accepts the arguments accepted by the "fetch" method.

=head3 state_files_and_directories

    # List context
    my @cs = $vol_obj->state_files_and_directories();

    # Scalar context
    my $cs = $vol_obj->state_files_and_directories();

    # Accepts arguments accepted by "fetch"
    my @cs = $vol_obj->state_files_and_directories(
        requested_fields => [ qw(acl-type) ],
        filter           => { 'inode-number' => num },
        allow_empty      => 1
    );

Fetches the state of contained files and directories and returns objects of type
L<NACL::CS::VolumeListFilesInDirectory|lib-NACL-CS-VolumeListFilesInDirectory-pm>.
Accepts the arguments accepted by the "fetch" method.

=cut

sub _file_dir_find_state_wrapper {
    $Log->enter();
    my ($self, @args) = @_;
    my %opts = $self->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'file-type' => {type => SCALAR,  optional => 1},
            filter      => {type => HASHREF, default  => {}},
        },

        # Allow other options for find
        allow_extra => 1
    );
    my $volume = delete $opts{volume};
    my %filter = %{$opts{filter}};
    $filter{'directory-path'} = "/vol/$volume";
    $self->_hash_move(
        source => \%opts,
        target => \%filter,
        move   => [qw(vserver file-type)]
    );
    $opts{filter} = \%filter;
    my @cs = NACL::CS::VolumeListFilesInDirectory->fetch(%opts);
    $Log->exit();
    return @cs;
}

=head2 name_attribute

Returns 'volume'.

=cut

sub name_attribute {
    return 'volume';
}

=head2 get_local_node_name

    # Class method call
    
    my $node_name = NACL::C::Volume->get_local_node_name(
        command_interface => $ci,
        volume            => $vol_name,
        vserver           => $vs_name   # Necessary only for CMode
    );
    
    # Instance call
    my $node_name = $vol_obj->get_local_node_name();

(Applicable only for CMode, for 7Mode it simply returns the name of the
node modeled by the command_interface)

Determines the node on which the volume is present and returns the node
name. This is necessary for Nodescope operations, since they need to
be performed in the context of the node where the volume is present.

=cut

sub get_local_node_name {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(params => \@args);
    if ($opts{command_interface}->is_7mode()) {
        $Log->exit();
        return $opts{command_interface}->hostrec->hostname;
    } elsif (!defined $opts{vserver}) {
        $Log->warn('For a CMode system, the vserver value '
                . 'is required to uniquely identify the volume but was not '
                . 'passed. Please update the call to get_local_node_name to pass '
                . 'the vserver value. Note that the value returned for this '
                . 'method is not guaranteed to be correct since the vserver '
                . 'value was not provided');
    }

    # Move the primary keys into the filter, all other options are
    # sent through directly to fetch (these should just be the common options)
    my %filter;
    $pkg_or_obj->_move_primary_keys(
        source => \%opts,
        target => \%filter,
    );

    # The field "node" is returned only in CLI, so force to run in CLI
    my $cs_object = NACL::CS::Volume->fetch(
        %opts,
        filter           => \%filter,
        requested_fields => [qw(node)],
    );

    $Log->exit();
    return $cs_object->node();
} ## end sub get_local_node_name

=head2 mkfile

B<DEPRECATED>. Please use the method C<file_create> instead.

    $vol_obj->mkfile(filename => $name, %other_opts);

(Instance method only)
This method is used to make files on a volume. It uses the volume name
and the file-name provided to construct the file path (of the form
/vol/$vol_name/$file_name).

This method is a wrapper around
L<< NACL::C::Mkfile->mkfile|lib-NACL-C-Mkfile-pm/mkfile >>. Therefore,
this method accepts all of the options accepted by
the C<NACL::C::Mkfile> method.

=over

=item Options

=over

=item C<< filename => $filename >>

(Required)
This is the name of the file to be created. Note that this is <not> the
path to the file. The path to the file gets constructed internally based
on the volume name and file name.

=item Other options

All of the options accepted by
L<< NACL::C::Mkfile->mkfile|lib-NACL-C-Mkfile-pm/mkfile >> are accepted
by this method as well.

=back

=back

=cut

sub mkfile {
    $Log->enter();
    my $self = shift;
    $Log->warn('This method is deprecated. Please use the method '
            . "'file_create' instead");
    $self->_verify_invocation(style => 'instance_only');
    my %opts = $self->_common_validate_with(
        params          => \@_,
        additional_spec => {filename => {type => SCALAR}},

        # Allow other options, which are those accepted by mkfile
        allow_extra => 1
    );

    # The form of the file path is /vol/$vol_name/$file_name
    # The "filename" argument accepted by the mkfile APISet is more
    # correctly the file-path. Here we're constructing the file path
    # given the file name.
    $opts{filename} = "/vol/$opts{volume}/$opts{filename}";
    delete $opts{volume};

    # Figure out whether to invoke NACL::C::Mkfile or NACL::STask::Mkfile
    # based on whether this was called by a C or STask Volume object.
    my $is_c_or_t  = $self->is_c_or_t();
    my $mkfile_pkg = "NACL::${is_c_or_t}::Mkfile";
    my $ret        = $mkfile_pkg->mkfile(%opts);

    $Log->exit();
    return $ret;
} ## end sub mkfile

# CMode ZAPI back-end builder for those methods that have an equivalent
# repository ZAPI. Is a wrapper around "_base_backend_cmode_zapi".
# Accepts the following arguments:
# 1. "map". Expects "map" to be provided with "volume" specified.
# 2. "api" : The normal API
sub _backend_cmode_zapi_repo_builder {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            map => {type => HASHREF},
            api => {type => SCALAR},
        },
        allow_extra => 1
    );

    my $is_repository = $pkg_or_obj->_determine_is_repository(%opts);
    if ($is_repository) {
        $opts{map} = {%{$opts{map}}, volume => 'volume-name'};
        $opts{api} .= '-async';
    }

    my $vserver;
    $opts{vserver_operated_on} ||= \$vserver;
    $pkg_or_obj->_base_backend_cmode_zapi(%opts);

    $Log->exit();
} ## end sub _delete_cmode_zapi

=head2 construct_dblade_path

    # Class method call
    my $dblade_path = NACL::C::Volume->construct_dblade_path(
        command_interface => $ci,
        volume            => $vol_name,
        vserver           => $vs_name
    );
    
                  or
                  
    # Instance call
    my $dblade_path = $vol_obj->construct_dblade_path();
    
(Class or instance method)
Returns the dblade path of a volume while accounting for the "name-ordinal"
value of the volume. For 7Mode, it simply returns "/vol/$vol_name" (there's
no concept of a dblade or a "name-ordinal" for 7Mode)

While most scripts assume the path to be "/vol/$vol_name", this assumption
is not always correct. It is possible to create multiple volumes of the same
name on the same node. In this case, the "name-ordinal" value has to be
accounted for. The first volume created will have a "name-ordinal" value of
"base", while successive volumes of that name will be assigned consecutive
integer "name-ordinal" values. For example, let's say we create 3 volumes on a
node, all named "myvol". In this case, the respective "name-ordinal" and
dblade paths are:

First creation:
    name-ordinal: base
    dblade path:  /vol/myvol
    
Second creation:
    name-ordinal: 1
    dblade path:  /vol/myvol(1)
    
Third creation:
    name-ordinal: 2
    dblade path:  /vol/myvol(2)
    
=over
    
=item Options

=over

=item Class method call

The primary keys (i.e volume (and vserver if CMode)) along with the
command_interface are mandatory. The "common options" (such as
method-timeout, apiset_must, apiset_should) are also accepted.

=item Instance method call

The "common options" (such as method-timeout, apiset_must, apiset_should) are
accepted.

=back

=back

=cut

# This and "construct_dblade_file_path" are unit tested in
# VolumeFileDirectory.thpl since they're closely tied to files.
sub construct_dblade_path {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(params => \@args);
    my $path = "/vol/$opts{volume}";

    # No concept of "name-ordinal" in 7Mode, so do this only for CMode
    if ($opts{command_interface}->is_cmode()) {
        my %primary_keys;
        $pkg_or_obj->_move_primary_keys(
            source => \%opts,
            target => \%primary_keys
        );

        my $vol_cs = NACL::CS::Volume->fetch(
            %opts,
            filter           => \%primary_keys,
            requested_fields => ['name-ordinal']
        );
        my $name_ordinal = $vol_cs->name_ordinal();
        if ($name_ordinal ne 'base') {

            # If it's not "base", then it's a number. The dblade path will
            # contain this name-ordinal within brackets
            $path .= "($name_ordinal)";
        }
    }

    $Log->exit();
    return $path;
}

=head2 construct_dblade_file_path

    # Class method call
    my $dblade_file_path = NACL::C::Volume->construct_dblade_file_path(
        command_interface => $ci,
        volume            => $vol_name,
        vserver           => $vs_name,
        filename          => $filename
    );
    
                  or
                  
    # Instance call
    my $dblade_file_path = $vol_obj->construct_dblade_file_path(filename => $filename);
    
(Class or instance method)
Returns the dblade file path for a given filename. This internally uses
L<construct_dblade_path> to account for the "name-ordinal".

In addition to the options accepted by L<construct_dblade_path>, this method
accepts C<filename> as argument.
    
=cut

sub construct_dblade_file_path {
    $Log->enter();
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {filename => {type => SCALAR}},

        # All other options get passed through to "construct_dblade_path"
        allow_extra => 1
    );
    my $filename = delete $opts{filename};

    my $dblade_path = $pkg_or_obj->construct_dblade_path(%opts);
    $dblade_path .= "/$filename";

    $Log->exit();
    return $dblade_path;
}

1;
